#! /usr/bin/env python 

import difflib
import os
import sys
import argparse 

def getPackageDirAndMake(path, base, outDir):
  packageDir = os.path.dirname(path).replace(base, "")
  outPackageDir = os.path.join(outDir, packageDir)
  if not os.path.isdir(outPackageDir):
    os.makedirs(outPackageDir)
  return outPackageDir



parser = argparse.ArgumentParser("""
Take two release directories and analyze the contents to find differences.  
Input: later release, earlier release, output directory
Output: A directory structure in --out which contains directories for packages 
which have differences.  Sub-directory structure is maintained.  Three types of 
files will exist in the structure.  The later type is _diff.txt which shows 
the differences in unified diff format.  The other two types end with _added and
_removed for files which were added/removed between the earlier and later 
releases.   
""")
parser.add_argument('--later', '-L', type=str, required=True, help="""
    "Later" of the two releases, files in this release but not the earlier
    will show up as _added""")
parser.add_argument('--earlier', '-E', type=str, required=True, help="""
    "Earlier" of the two releases, files in this release but not the earlier
    will show up as _removed.""")
parser.add_argument('--out', '-o', type=str, required=True, help="""
    Output directory for all differences.  Output will mirror release directory 
    structure""")


ignoreStrings = ["CMakeLists.txt", "_dict.cc"]

args = parser.parse_args()


later = os.path.join(args.later, "")
earlier = os.path.join(args.earlier, "")

outDir = args.out

laterBase = os.path.basename(os.path.dirname(later))
earlierBase = os.path.basename(os.path.dirname(earlier))


for directory in [later,earlier]:
  if not os.path.isdir(directory):
    raise Exception("Error -- directory not found: " + directory)

for root, dirs, files in os.walk(later, topdown=True):
  if ".svn" in root:
    continue
  if "Linux2.6" in root:
    continue
  
  for name in files:
    ignore = False
    for string in ignoreStrings:
      if string in name: ignore = True
    if ignore: continue


    laterPath = os.path.join(root, name)
    earlierPath = laterPath.replace(later, earlier)
    if not os.path.isfile(earlierPath):
      outPackageDir = getPackageDirAndMake(laterPath, later, outDir)
      outFilePath = os.path.join(outPackageDir, name + "_added")
      open(outFilePath,'w').write(open(laterPath, 'r').read())
      continue

    laterLines = open(laterPath, 'r').readlines()
    earlierLines = open(earlierPath, 'r').readlines()

    diff = list(difflib.unified_diff(earlierLines, laterLines,
                                fromfile=earlierBase, tofile=laterBase))
    if len(diff) > 0:
      outPackageDir = getPackageDirAndMake(laterPath, later, outDir)
      outFilePath = os.path.join(outPackageDir, name + "_diff")
      open(outFilePath, 'w').writelines(diff)

        



for root, dirs, files in os.walk(earlier, topdown=True):
  if ".svn" in root:
    continue
  if "Linux2.6" in root:
    continue

  
  for name in files:
    ignore = False
    for string in ignoreStrings:
      if string in name: ignore = True
    if ignore: continue


    earlierPath = os.path.join(root, name)
    laterPath = earlierPath.replace(earlier, later)
    if not os.path.isfile(laterPath):
      outPackageDir = getPackageDirAndMake(earlierPath, earlier, outDir)
      outFilePath = os.path.join(outPackageDir, name + "_removed")
      open(outFilePath,'w').write(open(earlierPath, 'r').read())



