#!/bin/bash

# make and show plots of DSO output, to allow shifter to check before
# installing results in database

# usage: shifterDSOCheck.sh [-f file_timestamp] [-d db_timestamp]
# where file_timestamp is of the form 2011-04-04_15_03_42
#    if omitted, the latest directory will be used
# if omitted, DB will not be accessed

usage () {
     echo "shifterDSOCheck.sh [-c <current_source>] [-r <reference_source>]     "
     echo "                    [-t] [-p <level>] [-d <detector>] [-P ,<partition>]"
	 echo "    where each source is of the format file:<timestamp> or db:<timestamp>"
	 echo "    The timestamp is in the form used by the DSO output:"
	 echo "        yyyy-mm-dd_HH_MM_SS, or \"latest\" for the most recent"
	 echo "  ****  NOTE!  db times are UTC, file times are Local. ****"
	 echo "    Default for -c <current_source> is file:latest "
	 echo "    Default for -r <reference_source> is null (no reference)"
	 echo "    Default for -d <detector> is NOVADAQ_ENVIRONMENT, then NDOS"
	 echo "                -P P<partition>"
	 echo "                -t : test mode - executables replaced with echo"
	 echo "                -p : print level [0]"
	 echo ""
     echo "examples:"
     echo "    shifterDSOCheck.sh               |    plots the latest file"
     echo "    shifterDSOCheck.sh -c file:latest   |    plots the latest file"
     echo "    shifterDSOCheck.sh -c file:latest  -r db:latest | plots the latest file, comparing to current database as reference"
	 
}


get_vars() {
# Parse command line options.
	while getopts hc:r:tp:d:P: OPT; do
    	case "$OPT" in
        	h)
            	usage
            	exit 0
            	;;
        	c)
            	current_source=$OPTARG
            	;;
        	r)
            	reference_source=$OPTARG
            	;;
        	d)
            	detID=$OPTARG
            	;;
        	d)
            	detID=$OPTARG
            	;;
	        P)
            	partition=$OPTARG
            	;;
        	p)
            	print_level=$OPTARG
            	;;
	        t)
		echo=echo
		;;
        	\?)
            	# getopts issues an error message
            	usage
            	exit 1
            	;;
    	esac
	done
	
	# Remove the switches we parsed above.
	shift `expr $OPTIND - 1`


}

set_environment() {
    source ~/.bashrc
    setup_novasoft_onmon_dso
    ##Directory to tmpfiles
    srt_setup -a
    SRT_PRIVATE_RELEASE=${PWD}  
    cd ../../tmpfiles/
    TMPFILE=${PWD}

    NOVADAQ_ENVIRONMENT=$(sh $SRT_PRIVATE_RELEASE/DSOResultPlotter/scripts/returnOnlineBins.sh 4)
	#echo $NOVADAQ_ENVIRONMENT
    if [ "x$detID" == "x" ]; then
	if [ "$NOVADAQ_ENVIRONMENT" != "x" ]; then
	    detID=$NOVADAQ_ENVIRONMENT
			#echo $NOVADAQ_ENVIRONMENT
	else
	    detID=NDOS
	fi
    fi
    

#Setup the NOVA DB environment
setup_file=`$SRT_PRIVATE_RELEASE/DSOResultPlotter/scripts/returnOnlineBins.sh 1`
print_db_env=`$SRT_PRIVATE_RELEASE/DSOResultPlotter/scripts/returnOnlineBins.sh 2`
runonline=$SRT_PRIVATE_RELEASE/DSOResultPlotter/scripts/doInOnline.sh
#source "$setup_file" -d $detID <<END > /dev/null
#echo setup_file=$setup_file
source $setup_file -d $detID  <<END 
y
END

	if [ $print_level -ge $debug_level ]; then
		$print_db_env
	fi

	detmode=DCS
	if [ "$detID" == "NDSBTest" ]; then
		dsoPath=/data/PedestalData/PedScans
		gconfigname=NDSBTestGlobalConfig
		dcms="dcm-4-01-01 dcm-4-01-02"
 		detNum=4
        elif [ "$detID" == "NearDet" ]; then
                dsoPath=/data1/PedestalData/PedScans
                gconfigname=InitialGlobalConfig
                dcms=""
                for (( db=1; db<=3; db++))
                do
                        for (( ldcm=1; ldcm<=4; ldcm++))
                        do
                                dcms="$dcms dcm-1-`printf %2.2d $db`-`printf %2.2d $ldcm`"
                        done
                done
                dcms="$dcms dcm-1-04-01 dcm-1-04-02"
                detNum=1
	elif [ "$detID" == "FarDet" ]; then
		dsoPath=/data1-b/PedestalData/PedScans
		echo "$dsoPath"
                gconfigname=FarDetGlobalConfig
		gconfigname="$gconfigname$partition"
		dcms=""
		for (( db=1; db<=14; db++))
		do
			for (( ldcm=1; ldcm<=12; ldcm++))
			do
				dcms="$dcms dcm-2-`printf %2.2d $db`-`printf %2.2d $ldcm`"
			done
		done
 		detNum=2
	else
		dsoPath=/data2/PedestalData/PedScans
		gconfigname=InitialGlobalConfig
		detID=NDOS
		dcms="dcm-3-01-01 dcm-3-01-02 dcm-3-01-03 dcm-3-02-01 dcm-3-02-02 dcm-3-02-03 dcm-3-03-01 dcm-3-03-02 dcm-3-03-03 dcm-3-04-01 dcm-3-04-02"
		detNum=3
	fi
	if [ $print_level -ge $debug_level ]; then
		echo "dcms list=\"$dcms\""
	fi

}

set_defaults() {

	# this needs a special test release, since it uses ROOT, which is not
	# installed on all nodes.  We don't want to cause compile error messages
	# in the normal test release.

	#workdir=/home/novadaq/plotTestOffline/

	dbPath=/tmp/db_$$
	current_source="file:latest"
	print_level=1
	debug_level=2 
	partition=""
}

make_plots_from_files() {

	local file_timestamp=$1

	local olddir=$PWD
# get latest files, or use file_timestamp argument provided by user
	if [ "$file_timestamp" != "latest" ]; then
		local timestamp=$file_timestamp
	else
       	cd $dsoPath
		local dirs=`ls -t |head -100`
        	for dir in $dirs
		do
			local res=`find $dir -name threshold\*.csv` 
			if [ "$res" != "" ]; then
				local timestamp=$dir
				cd $olddir
				break
			fi
		done
		if [ "$timestamp" == "" ]; then
			echo "shifterDSOCheck.sh: Error - can't find recent .csv file!"
			$echo exit 1
		fi
	fi
	
	# get rid of any trailing /
	timestamp=`basename $timestamp`

	# make canvases for values in the csv files
	which plotDSOResult
	$echo plotDSOResult "$dsoPath/$timestamp" "$detID"

	# exit if failed to produce
	if [ $? -ne 0 ]; then
		$echo exit 1
	fi
	
	# name of histo file 
	histos=$timestamp\_hists.root

}

make_plots_from_db() {
        dumper=`$SRT_PRIVATE_RELEASE/DSOResultPlotter/scripts/returnOnlineBins.sh 3`
	db_timestamp_in=$1
	dbtprefix=""
	dbt=""
	#if [ "$db_timestamp_in" != "latest" ]; then
		#fix_db_timestamp $db_timestamp_in
		#db_timestamp=$timestamp_tmp
		#dbtprefix="-t"
		#dbt='''$db_timestamp'''
	#fi
	local tag=dbdso_$db_timestamp_in
	local dbDir=$dbPath/$tag
	local olddir=$PWD
	histos=$tag\_hists.root
#
	if [ $print_level -ge $debug_level ]; then
		echo "make_plots_from_db: db time argument=$dbt"
	fi
	

	mkdir -p $dbDir
    cd $dbDir
     	echo "WARNING!  THIS DIRECTORY IS SUBJECT TO AUTOMATIC DELETION!" > README
       	#$echo $dumper "$dbt" $gconfigname $detmode $detID $dcms << END
		echo " "
		echo "No need to answer the following question!"
		if [ "$db_timestamp_in" != "latest" ]; then
			if [ $print_level -ge $debug_level ]; then
			    echo "$dumper -t "$db_timestamp" $gconfigname $detmode $detID $dcms"
			fi
			fix_db_timestamp $db_timestamp_in
			db_timestamp=$timestamp_tmp
	       	$runonline "$dumper -t \"$db_timestamp\" $gconfigname $detmode $detID $dcms" << END
yes
END
			status=$?
			if [ $print_level -ge $debug_level ]; then
				echo "$dumper status=$status"
			fi

		else
			if [ $print_level -ge $debug_level ]; then
				echo "$dumper $gconfigname $detmode $detID $dcms"
			fi
	       	$runonline "$dumper $gconfigname $detmode $detID $dcms" << END
yes
END
		
			status=$?
			if [ $print_level -ge $debug_level ]; then
				echo "$dumper status=$status"
			fi
		fi

# rename the output files to have the names in the same format as the DSO output

	$echo mv febmask_dump.csv febmask_$tag.csv
	$echo mv pixelmask_dump.csv pixelmask_$tag.csv
	$echo mv thresholds_dump.csv thresholds_$tag.csv

	cd $olddir
	cmd="$echo plotDSOResult "$dbDir" "$detID""
	if [ $print_level -ge $debug_level ]; then
		echo $cmd
		fi
	$cmd
	if [ ! -e $histos ]; then
		echo "Database results requested, but file $histos not found."
		$echo histos=""
	fi
}

plot_results () {
    dso_csv="dso-histos.csv"
    if [ "$NOVADAQ_ENVIRONMENT" == "FarDet" ]; then
	OMdetID="FD"
    elif [ "$NOVADAQ_ENVIRONMENT" == "NearDet" ]; then
	OMdetID="ND"
    else
	OMdetID="NDOS"
    fi
    
   echo "Running onmon_viewer..." 
    if [ -f $current_histos ]; then
    	if [ "x$reference_histos" == "x" ]; then
	    echo ${current_histos} 
	    echo ${OMdetID}
	    echo ${dso_csv}
	    cd $SRT_PRIVATE_RELEASE
	    #plot the latest DSO on disk
	    onmon_viewer -f $TMPFILE/${current_histos} -d ${OMdetID} -v DSOResultPlotter/scripts/${dso_csv}

	else
	    #plot the latest DSO on disk and save DB histo, the reference plot will have to be opened in Onmon until this reference can be passed here
	    cd $SRT_PRIVATE_RELEASE
	    onmon_viewer -f $TMPFILE/${current_histos} -d ${OMdetID} -v DSOResultPlotter/scripts/${dso_csv} -r $TMPFILE/${reference_histos}
	fi
    else
	echo "Could not find histogram file $current_histos"
	$echo exit 1
    fi
}

fix_db_timestamp () {
	#convert from yyyy-mm-dd_HH_MM_SS to yyyymmdd HH:MM:SS
	local ts=$1
	if [ $print_level -ge $debug_level ]; then
		echo "fix_db_timestamp in: $ts"
	fi

	local y=${ts:0:4}
	local m=${ts:5:2}
	local d=${ts:8:2}
	local H=${ts:11:2}
	local M=${ts:14:2}
	local S=${ts:17:2}

	timestamp_tmp="$y$m$d $H:$M:$S"
	if [ $print_level -ge $debug_level ]; then
		echo "fix_db_timestamp out: $timestamp_tmp"
	fi

}

parse_source () {

	local source=$1
	type=${source:0:4}
	if [ $type":" == "file:" ]; then
		timestamp=${source:5}
	else 
		type=${source:0:2}
		if [ $type":" == "db:" ]; then
			timestamp=${source:3}
		else
			echo "Could not parse type ($type) in source=$source"	
			exit 1
		fi
		#if [ $timestamp != "latest" ]; then
			#fix_db_timestamp $timestamp
			#timestamp=$timestamp_tmp
		#fi
	fi

}


makeplots () {
	
	for input in current reference
	do
		sourcevar=$input\_source
		if [ $print_level -ge $debug_level ]; then 
			echo "input=$input, ${!sourcevar}"
		fi
		if [ ! -z ${!sourcevar} ]; then
			parse_source ${!sourcevar}
			if [ $print_level -ge $debug_level ]; then 
				echo "parse result: type=$type, timestamp=$timestamp"
			fi
			if [ "x$type" == "xfile" ]; then
				cmd="make_plots_from_files $timestamp"
				if [ $print_level -ge $debug_level ]; then
					echo $cmd
				fi
				$cmd
			elif [ "x$type" == "xdb" ]; then
				cmd="make_plots_from_db $timestamp"
				if [ $print_level -ge $debug_level ]; then
					echo $cmd
				fi
				$cmd
			else
				echo "Type $type not understood. "
				usage
				exit 1
			fi
			if [ $print_level -ge $debug_level ]; then 
				echo "input=$input, histos=$histos"
			fi
			histvar=$input\_histos
			read -r -a $histvar <<< $histos
		fi
	done

}


main () {

	set_defaults
	get_vars $*
	set_environment
	makeplots
	plot_results

}

main $*

exit 0
