////////////////////////////////////////////////////////////////////////
/// \brief  Accumulation space for HoughVertex
/// \author Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#ifndef HV_LITETH2_H
#define HV_LITETH2_H

#include "TH2.h"

#include <cassert>

namespace hv
{

/// \brief  Accumulation space for HoughVertex
///
/// TH2F is a bad temporary storage for a fast algorithm. There's a lot of
/// overhead. Using this simple version is at least 2x faster.
class LiteTH2
{
public:
  LiteTH2(int nx, float x0, float x1,
          int ny, float y0, float y1)
      : fNx(nx), fX0(x0), fX1(x1),
        fNy(ny), fY0(y0), fY1(y1)
  {
    fBins = new float[(nx+2)*(ny+2)];
    memset(fBins, 0, (nx+2)*(ny+2)*sizeof(float));
  }

  LiteTH2(const LiteTH2& h)
    : fNx(h.fNx), fX0(h.fX0), fX1(h.fX1),
      fNy(h.fNy), fY0(h.fY0), fY1(h.fY1)
  {
    fBins = new float[(fNx+2)*(fNy+2)];
    memcpy(fBins, h.fBins, (fNx+2)*(fNy+2)*sizeof(float));
  }

  ~LiteTH2()
  {
    delete[] fBins;
  }

  int GetNbinsX() const {return fNx;}
  int GetNbinsY() const {return fNy;}

  float GetBinCenterX(int ix) const
  {
    assert(ix >= 0 && ix < fNx+2);
    return fX0+(ix-.5)/fNx*(fX1-fX0);
  }

  float GetBinCenterY(int iy) const
  {
    assert(iy >= 0 && iy < fNy+2);
    return fY0+(iy-.5)/fNy*(fY1-fY0);
  }

  float GetBinContent(int ix, int iy) const
  {
    const int bin = iy*(fNx+2)+ix;
    if(bin < 0 || bin > (fNx+2)*(fNy+2)) return 0;
    return fBins[bin];
  }

  int FindBinX(float x) const
  {
    int ret = fNx*(x-fX0)/(fX1-fX0)+1;
    if(ret < 0) return 0;
    if(ret > fNx+1) return fNx+1;
    return ret;
  }

  int FindBinY(float y) const
  {
    int ret = fNy*(y-fY0)/(fY1-fY0)+1;
    if(ret < 0) return 0;
    if(ret > fNy+1) return fNy+1;
    return ret;
  }

  void SetBinContent(int ix, int iy, float val)
  {
    const int bin = iy*(fNx+2)+ix;
    assert(bin >= 0 && bin < (fNx+2)*(fNy+2));
    fBins[bin] = val;
  }

  void AddBinContent(int ix, int iy, float val)
  {
    const int bin = iy*(fNx+2)+ix;
    assert(bin >= 0 && bin < (fNx+2)*(fNy+2));
    fBins[bin] += val;
  }

  TH2F* ToTH2F() const
  {
    TH2F* ret = new TH2F("", "", fNx, fX0, fX1, fNy, fY0, fY1);

    for(int ix = 0; ix < fNx+2; ++ix)
      for(int iy = 0; iy < fNy+2; ++iy)
        ret->SetBinContent(ix, iy, GetBinContent(ix, iy));

    return ret;
  }

protected:
  float* fBins;
  int fNx;
  float fX0, fX1;
  int fNy;
  float fY0, fY1;
};

} // namespace

#endif
