////////////////////////////////////////////////////////////////////////
/// \file NumuEAlg.h
//
/// \version $Id: NumuEAlg.h,v 1.10 2012-11-15 20:46:24 lein Exp $
/// \author  lein@physics.umn.edu
////////////////////////////////////////////////////////////////////////
#ifndef NumuEAlg_h
#define NumuEAlg_h

#include <string>

// Framework includes
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindOneP.h"

// NOvA includes 
#include "fhiclcpp/ParameterSet.h"
#include "RecoBase/Track.h"
#include "RecoBase/Energy.h"
#include "NumuEnergy/NumuE.h"
#include "MuonRemove/art/TrackCleanUpAlg.h"
#include "CosRej/CosRejObj.h"
#include "TMVA/Reader.h"

namespace numue {

  /// An algorithm to determine the energy of a slice
  class NumuEAlg {
  public:
    explicit NumuEAlg(fhicl::ParameterSet const &pset);
    virtual ~NumuEAlg();                        

    /// \brief Returns various energy estimations for FD. See comments at top of function in .cxx for full explanation
    NumuE FDEnergy(std::vector< art::Ptr<rb::Track> > const sliceTracks,
                   art::PtrVector< rb::CellHit >            sliceHits,
                   art::Ptr< rb::Cluster>                   sliceCluster,
                   art::Event                        const& e,
                   murem::TrackCleanUpAlg*                  trkCleanUpAlg);
    
    /// \brief Returns QE energy estimation using formula. See comments at top of function in .cxx for full explanation
    rb::Energy QEFormulaEnergy(std::vector< art::Ptr<rb::Track> > const sliceTracks,
                               art::PtrVector< rb::CellHit >            sliceHits,
			       double                                  &error,
                               art::Event                        const& e);

    /// \brief Returns various energy estimations for ND. See comments at top of function in .cxx for full explanation
    NumuE NDEnergy(std::vector< art::Ptr<rb::Track> > const sliceTracks,
                   art::PtrVector< rb::CellHit >            sliceHits,
                   art::Ptr< rb::Cluster>                   sliceCluster,
		   art::Event                        const& e,
                   murem::TrackCleanUpAlg*                  trkCleanUpAlg);

    /// \brief Function that, given muon track length in cm, returns muon energy in GeV. This is done with a spline fit to muon track length -> energy.
    double      MuonEFromTrackLength(double trkLen) const;

    /// \brief Function that, given muon energy in GeV, returns an effective track length in active material in cm. This is done by reversing a spline fit of muon track length -> energy.
    double      ActiveTrackLengthFromMuonE(double muonE) const;

    /// \brief Function that, given muon track length in cm, returns muon energy in GeV. This distinguishes between the three types of muon tracks and makes the approriate function calls.
    double      NDMuonEFromTrackLength(double actTrkLen, double catTrkLen, double trkCalAct, double trkCalTran, double trkCalCat,
				       MuonType*                        muonType = NULL) const;
    double      NDMuonEFromTrackLength(art::Ptr<rb::Track>       const& track,
				       MuonType*                        muonType = NULL) const;

    /// \brief Function that, given muon track length in cm, returns muon energy in GeV. This is only for ND muons that are entirely contained in the active region and the transition plane.
    double      NDMuonEInActiveOnly(double actTrkLen) const;

    /// \brief Function that, given muon track length in cm, returns muon energy in GeV. This is only for ND muons that are entirely contained in the muon catcher region.
    double      NDMuonEInCatcherOnly(double catTrkLen) const;

    /// \brief Function that, given muon track length in cm, returns muon energy in GeV. This is only for ND muons that have activity in the both the active region and the muon catcher region.
    double      NDMuonEInActiveAndCatcher(double actTrkLen, double catTrkLen) const;

    /// \brief Function that returns energy for each track given the assumption that it is a muon. It uses the functions above, like MuonEFromTrackLength and NDMuonEFromTrackLength.
    std::vector<rb::Energy> MuonEnergies(std::vector< art::Ptr<rb::Track> > const sliceTracks) const;

    /// \brief Function that returns the plane number of the transition plane - the last active plane before the steel of the muon catcher.
    int         NDTransitionPlane() const;

    /// \brief Function that returns the z position of the middle of the transition plane - the last active plane before the steel of the muon catcher.
    double      NDTransitionZPos() const;

    /// \brief Get values from cosrej object and use them to get TMVA trained value of uncontained energy for FD
    double      GetUCE(art::FindManyP<cosrej::CosRejObj> CRO, size_t slicenum, float totalgev, int nhitslice); 

    /// \brief Fernanda's scattering energy.  Scattering variables themselves stored in CosRej
    double      GetFSE(art::FindManyP<cosrej::CosRejObj> CRO, size_t slicenum);

  private:

    std::string     fPIDModuleLabel;         ///< Label for module writing ReMId objects to file
    bool            fPIDMethod;              ///< If true, use highest PID to identify muon track instead of longest track (Only FD currently)
    std::string     fTMVApath;               ///< path to TMVA weight file
    TMVA::Reader*   fReaderE;                ///< Reader for TMVA to get uncontained energy
    float           Evars[8];               ///< Inputs for uncontained energy BDT (TMVA)

    double          FDMstitch1 = 334;        ///< cm; First stitch location on 4 spline fit for muon trk len -> energy in FD
    double          FDMstitch2 = 539;        ///< cm; Second stitch location on 4 spline fit for muon trk len -> energy in FD
    double          FDMstitch3 = 1064;       ///< cm; Third stitch location on 4 spline fit for muon trk len -> energy in FD
    double          FDMoffset  = 0.1503;     ///< GeV; Offset for 4 spline fit for muon trk len -> energy in FD
    double          FDMslope1  = 0.001910;   ///< GeV/cm; First slope for 4 spline fit for muon trk len -> energy in FD
    double          FDMslope2  = 0.001987;   ///< GeV/cm; Second slope for 4 spline fit for muon trk len -> energy in FD
    double          FDMslope3  = 0.002039;   ///< GeV/cm; Third slope for 4 spline fit for muon trk len -> energy in FD
    double          FDMslope4  = 0.002159;   ///< GeV/cm; Fourth slope for 4 spline fit for muon trk len -> energy in FD

    bool            fFD;                     ///< Is detector FD?
    bool            fND;                     ///< Is detector ND?
    double          fMaxTrkLen;              ///< Maximum track length allowed in detector (diagonal length)
    TVector3        fNumiDir;                ///< Direction of NuMI Beam
    int             fTranPlane;              ///< Last plane of scintillator before steel begins in ND
    double          fTranZPos;               ///< Z position of middle of transition plane

    
    /// \brief Function that makes cluster from all hits in slice that do not belong to track
    rb::Cluster MakeVertexCluster(art::PtrVector< rb::CellHit > const& sliceHits,
                                  art::PtrVector< rb::CellHit > const& trackHits) const;

    /// \brief Function that finds the longest track that is still smaller than maxTrkLen
    int         LongestTrack(std::vector< art::Ptr<rb::Track> > const& sliceTracks,
                             double&                                   longestTrackLength)  const;

    /// \brief Function for ND that divides track length into length in active material and length in muon catcher. Also records x,y position of track if crosses transition plane. This can be used for containment purposes.
    void        NDTrackLength(art::Ptr<rb::Track>       const& track,
                              double*                          actLength = NULL,
                              double*                          catLength = NULL,
                              double*                          transitionX = NULL,
                              double*                          transitionY = NULL) const;
    /// \brief Function for ND that divides track visible energy into three regions: active, transiton, and catcher
    void        NDTrackEnergySplitting(art::Ptr<rb::Track>       const& track,
                                       double*                          trkCalAct  = NULL,
                                       double*                          trkCalTran = NULL,
			               double*                          trkCalCat  = NULL) const;
    /// \brief Function that, given total hadron cal energy on and off muon track in GeV, returns weighted hadron energy in GeV. This is done with a spline fit to the hadron energy vs. (True Neutrino Energy - Reco Muon Energy). The fit was done for a QE reco population.
    double      HadEQE(double calE)   const; 
    /// \brief Function that, given total hadron cal energy on and off muon track in GeV, returns weighted hadron energy in GeV. This is done with a spline fit to the hadron energy vs. (True Neutrino Energy - Reco Muon Energy). The fit was done for a NonQE reco population.
    double      HadENonQE(double calE)   const; 
    /// \brief Function that, given total hadron cal energy on and off muon track in GeV, returns weighted hadron energy in GeV. This is done with a spline fit to the hadron energy vs. (True Neutrino Energy - Reco Muon Energy). The fit was done for a reco population that had both QE and nonQE.
    double      HadECC(double calE)   const; 
    /// \brief Function that, given total hadron cal energy on and off muon track in GeV, returns weighted hadron energy in GeV. This is done with a spline fit to the hadron energy vs. (True Neutrino Energy - Reco Muon Energy). The fit was done for a QE reco population. The fit only used slices with ALMOST NO hadronic activity in the transition plane or muon catcher!!
    double      NDHadEQE(double calE)   const; 
    /// \brief Function that, given total hadron cal energy on and off muon track in GeV, returns weighted hadron energy in GeV. This is done with a spline fit to the hadron energy vs. (True Neutrino Energy - Reco Muon Energy). The fit was done for a NonQE reco population. The fit only used slices with ALMOST NO hadronic activity in the transition plane or muon catcher!!
    double      NDHadENonQE(double calE)   const; 
    /// \brief Function that, given total hadron cal energy on and off muon track in GeV, returns weighted hadron energy in GeV. This is done with a spline fit to the hadron energy vs. (True Neutrino Energy - Reco Muon Energy). The fit was done for a reco population that had both QE and nonQE. The fit does not try to account for large amounts of hadronic activity in the muon catcher.
    double      NDHadECC(double calE)   const; 
  };//End of class NumuEAlg

}//End of namespace

#endif /* NumuEAlg_h */
////////////////////////////////////////////////////////////////////////
