//////////////////////////////////////////////////////////////////////////////
// \file    KalmanGeoHelper.h
// \brief   A class of useful track geometry functions for the KalmanTrack algorithm
// \version $Id: KalmanGeoHelper.h,v 1.1 2012-11-16 16:57:10 raddatz Exp $
// \author  Nick Raddatz
////////////////////////////////////////////////////////////////////////////
#ifndef TRK_KALMANGEOHELPER_H
#define TRK_KALMANGEOHELPER_H

#include "Geometry/Geometry.h"
#include "GeometryObjects/OfflineChan.h"
#include "ChannelInfo/BadChanList.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"

#include "TF1.h"

namespace trk
{
  /// \brief A collection of geometry functions used by the KalmanTrack tracking algorithm
  class KalmanGeoHelper
  {
  public:
    KalmanGeoHelper();
    KalmanGeoHelper(bool badChan);
    virtual ~KalmanGeoHelper();
    
    /// \brief Estimate the position on a line within a box
    /// 
    /// This method takes a line, defined by its start point z0,v0 and direction dz and dv, and 2d box, defined by its center z,v and half lengths halfz, halfv and fills the bestPos with the best estimation of the closest point in that box t  the line. 
    /// For lines that do not intersect the box, this is defined as the corner of the box closest to the line.
    /// For lines that do intersect the box, the best position is the the point on the line bisecting the intersection points of the box.
    ///
    /// \param bestPos A vector that is filled with the best position of the line in the box with the first entry corresponding to z and the second to v
    /// \return  Total intersection length of the line with the box
    double BestTrackPoint(double z, double v, double halfz, double halfv, std::vector<double> & bestPos, double z0,double v0, double dz, double dv);
    
    /// \brief Variation of geo::CountCellsOnlineFast function specific to KalmanTrack algorithm
    int CountMissedCellsOnLine(geo::View_t view, double vi, double zi, double vf, double zf, int maxCells);
    
    /// \brief Variation of geo::CountCellsOnlineFast function specific to KalmanTrack algorithm
    int CountMissedCellsOnLine(geo::View_t view, double vi, double zi, double vf, double zf, std::vector<geo::OfflineChan> &hitsonline, std::vector<double> &distance); 
    
    /// \brief Estimation of the mip missed hit fraction as a function of the dist through a cell and cell depth w
    double CalcMissFrac(double dist, double w, geo::View_t view);
    
    /// \brief Finds the plane best matched to a trajectory point
    /// This method takes a trajectory point and finds the plane that closest matches this position in either view between minplane and maxplane
    /// \param traj Trajectory point to be matched to a plane
    /// \param minplane Minimum plane to search for a match to the trajectory point
    /// \param maxplane Maximum plane to search for a match to the trajectory point
    /// \return Plane that matches the trajectory point between minplane and maxplane
   int  MatchTrajectoryToPlane(TVector3 traj, int minplane, int maxplane);

   /// \Brief Variation of MatchTrajectoryToPlane that limits the matching plane to be in the input view
   int  MatchTrajectoryToPlaneInView(TVector3 traj, geo::View_t view, int minplane, int maxplane);


  private:
    bool  fBadChannels;
    art::ServiceHandle<geo::Geometry> fgeom;
    art::ServiceHandle<chaninfo::BadChanList> fbadc;
    bool ndos;
    bool nd;
    bool fd;
    int  fnmaxplane;
    double fclipperLength;
    double fwlow;
    double fwhigh;
    double flengthmax;
    TF1  effpar0;  // parameter of the eff function that depends on cell depth
    TF1  effpar1;  // parameter of the eff function that depends on cell depth
    TF1  effpar2;  // parameter of the eff function that depends on cell depth
    TF1  effpar3;  // parameter of the eff function that depends on cell depth
    void   SetupHelper(); // function to setup efficiency calculations and other variables based on detector
    double Efficiency(double w, double length); // returns efficiency as a function of w and length in cell
  };
}
#endif
