#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Persistency/Common/Assns.h"
#include "art/Framework/Principal/Handle.h"	
#include "art/Framework/Principal/Event.h"
#include "DDTUtilities/AssociationUtil.h"

#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/GroupedHitList.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"

#include "RawData/RawTrigger.h"

#include <boost/format.hpp>

#include <string>
#include <vector>
#include <algorithm>

namespace highe {
  class HighEnergyFilt;
}

class highe::HighEnergyFilt : public art::EDFilter {
public:
  explicit HighEnergyFilt(fhicl::ParameterSet const & p);
  virtual ~HighEnergyFilt();

  virtual bool filter(art::Event & e) override;

private:
  bool send_trigger(unsigned const sum_adc, 
		    unsigned const n_hits_total) const;

  std::string slice_label_, slice_instance_;
  unsigned prescale_, trigger_counts_;
  unsigned sum_adc_trigger_threshold_;
  unsigned n_hits_trigger_threshold_;
};

highe::HighEnergyFilt::HighEnergyFilt(fhicl::ParameterSet const & p) :
  slice_label_(p.get<std::string>("slice_label")),
  slice_instance_(p.get<std::string>("slice_instance")),
  prescale_(p.get<unsigned>("prescale")),
  trigger_counts_(0),
  sum_adc_trigger_threshold_
  (p.get<unsigned>("sum_adc_trigger_threshold")),
  n_hits_trigger_threshold_
  (p.get<unsigned>("n_hits_trigger_threshold"))
{
  produces<std::vector<novaddt::TriggerDecision>>();
}

highe::HighEnergyFilt::~HighEnergyFilt()
{
  std::cout << "\n\tNumber of High Energy Filt Triggers = " << trigger_counts_
  	    << "\n" << std::endl;
}

bool highe::HighEnergyFilt::filter(art::Event & e)	
{
  bool result = false;

  std::unique_ptr<std::vector<novaddt::TriggerDecision>> 
    trigger_decisions(new std::vector<novaddt::TriggerDecision>);

  art::Handle<std::vector<novaddt::HitList> > slices;
  e.getByLabel(slice_label_, slice_instance_, slices);

  // We need to subtract off the trigger time, so that we know what 
  // time stamp in the EVD this corresponds to:
  // art::Handle<std::vector<rawdata::RawTrigger>> RawTriggers;
  // e.getByLabel("daq", RawTriggers);
  // assert(RawTriggers->size() == 1);
  // TDC event_time(RawTriggers->at(0).fTriggerTimingMarker_TimeStart);
  // boost::format hitform("%4u %4u %5u %20u");

  unsigned n_slice = 0;
  unsigned n_triggered_slices = 0;
  for (auto const& slice : *slices)
  {
    unsigned sum_adc = 0;

    for (auto const& hit : slice)
      sum_adc += hit.ADC().val;

    double n_hits_total = slice.size();

    boost::format form("%5u %3u %8u %8.2f");
    std::cout << "MF: " <<
      form % e.id().event() % ++n_slice % n_hits_total % sum_adc
	      << std::endl;
    
    if (send_trigger(sum_adc, n_hits_total))
    {
      ++n_triggered_slices;

      ++trigger_counts_;
      if (trigger_counts_ % prescale_ == prescale_ - 1) 
      {
	auto time_interval = 
	  std::minmax_element(slice.begin(), slice.end(), 
			      novaddt::CompareDAQHit<novaddt::TDC>());
	auto t0 = time_interval.first->TDC();
	auto t1 = time_interval.second->TDC();
	auto dt = t1 - t0;
	
	// double tdc_to_us = 1.0 / 64e6 * 1e6;
	// std::cout << "\nRun    " << e.run()
	// 	  << "\nSubrun " << e.subRun()
	// 	  << "\nEvent  " << e.id().event()
	// 	  << "\nt0 =   " << (t0 - event_time) * tdc_to_us
	// 	  << "\nt1 =   " << (t1 - event_time) * tdc_to_us
	// 	  << "\ndt =   " << dt
	// 	  << std::endl;

	trigger_decisions->emplace_back
	  (t0, dt, daqdataformats::trigID::TRIG_ID_DATA_ENERGY, prescale_);

	result = true;
      }
    }
  }

  std::cout << "MF: Event " << e.event() << " has " << n_triggered_slices
	    << " triggered slices (Space Slice)."
	    << std::endl;

  e.put(std::move(trigger_decisions));
  return result;
}

bool highe::HighEnergyFilt::send_trigger
(unsigned const sum_adc,
 unsigned const n_hits_total) const
{
  if (sum_adc < sum_adc_trigger_threshold_)
    return false;

  if (n_hits_total < n_hits_trigger_threshold_)
    return false;

  return true;
}

DEFINE_ART_MODULE(highe::HighEnergyFilt)
