////////////////////////////////////////////////////////////////////////
/// \brief   Simple analyzer module to print to text file information 
///          required for c++ based block alignment study.
///          It does very rudimentary stuf, loop over tracks and garnering
///          track information - all passed to a .txt.
/// \author  gsdavies@iastate.edu
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <ostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cstdlib>
#include <vector>

// NOvASoft includes
#include "CMap/CMap.h"
#include "Geometry/Geometry.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Track.h"
#include "NovaDAQConventions/DAQConventions.h"

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"


namespace align {
  class AlignInspect : public art::EDAnalyzer {
    public:
      explicit AlignInspect(fhicl::ParameterSet const& pset);
      virtual ~AlignInspect();                        
      
      void analyze(art::Event const& evt);
      void reconfigure(const fhicl::ParameterSet& pset);
      
      void beginJob();
      void endJob();
      
    private:
      
      bool IsGoodTrack(art::Ptr<rb::Track> t, double x, double y, double z); ///< yes/no is "track" entering within a buffer window
      std::string   fCellHitsListLabel; ///< Where to find CellHit
      std::string   fTrackLabel;        ///< Where to find Tracks

      unsigned int  fMuonNhit;          ///< Required # hits on muon track
      unsigned int  fMuonNhitX;         ///< # required in x view
      unsigned int  fMuonNhitY;         ///< # required in y view
      double        fDetEdgeBuffer;     ///< Detector edge buffer zone for entering cosmic ray track decision [cm]
      unsigned int  fNevents;
      
  };
};

////////////////////////////////////////////////////////////////////////
namespace align{

  AlignInspect::AlignInspect(fhicl::ParameterSet const& pset): 
    EDAnalyzer(pset),
    fCellHitsListLabel(pset.get< std::string  >("CellHitsListLabel")             ),
    fTrackLabel       (pset.get< std::string  >("TrackLabel")                    ),
    fMuonNhit         (pset.get< unsigned int >("MuonNhit")                      ),
    fMuonNhitX        (pset.get< unsigned int >("MuonNhitX")                     ),
    fMuonNhitY        (pset.get< unsigned int >("MuonNhitY")                     ),
    fDetEdgeBuffer    (pset.get< double       >("DetEdgeBuffer")                 ), fNevents(0)
  {
    reconfigure(pset); 
  }

  //......................................................................
  AlignInspect::~AlignInspect()
  {
  }
      
  //......................................................................
  void AlignInspect::beginJob()
  {
  }
  //......................................................................
  
  void AlignInspect::analyze(art::Event const& evt)
  {
    ++fNevents;
    art::ServiceHandle<geo::Geometry>         geom;
    art::ServiceHandle<cmap::CMap>            cmap;
    
    // Get the tracks that were made by CosmicTrack
    // and test that there are tracks in the event
    art::Handle< std::vector<rb::Track> > trks;
    evt.getByLabel(fTrackLabel,trks);
    if (trks->size()==0) {
      mf::LogWarning("AlignWarn") << "There are " << trks->size() 
 	                          << " Tracks in this event, skip";
      return;
    }
    LOG_DEBUG("AlignInspect") << "There are " << trks->size() << " Tracks in this event";
    
    
    std::vector<art::Ptr<rb::CellHit> >  chits; 
    const unsigned int trkMax = trks->size();
    //const unsigned int trkMax = 1;
    if(trkMax < 2) return;
    for(unsigned int trkIdx = 0; trkIdx < trkMax; ++trkIdx){
      art::Ptr<rb::Track> track(trks, trkIdx);
      
      TVector3 start(track->Start());
      TVector3 end(track->Stop());
      if(end.Y() > start.Y()) std::swap(start, end);
      
      // Is this a good track?? i.e. entering and extends across at least 8 planes
      if (this->IsGoodTrack(track,start.X(),start.Y(),start.Z()) == false) return;
      assert(track->ExtentPlane() >= 8); 

      const unsigned int hitMax = track->NCell();  // maximum number of CellHits on track
      const unsigned int hitX   = track->NXCell(); // maximum number of CellHits on track in XZ
      const unsigned int hitY   = track->NYCell(); // maximum number of CellHits on track in YZ

      // Do NOT select tracks if max number of hits or XZ- YZ-hits less than configurable values
      if ( (hitMax < fMuonNhit) || (hitX < fMuonNhitX) || (hitY < fMuonNhitY) ) return;
      assert(hitMax >= fMuonNhit);
      const int detid = novadaq::cnv::kNDOS;
      // Grab the block ID here from DAQChannelMap
      daqchannelmap::lchan logchan = cmap->Map()->encodeLChan(detid, track->MinPlane(),1);
      const int bID = cmap->Map()->computeBlock(logchan);
      //      std::cout << bID << ", plane: " << track->MinPlane() << ", det: " << detid << std::endl;
      // write out track start/end positions and track index + event number, run number
      // Note: hard-coded output directory an output names here!
      std::ostringstream oss;
      oss << "/nova/data/users/gsdavies/alignment/mc/inspect/blkshift10_" << evt.subRun() << ".txt";

      std::string fname(oss.str());
      std::ofstream fout(fname.c_str(),std::ios::app);

      if(fout){
	fout << evt.subRun() << " " << evt.id().event() << " " << trkIdx << " " 
	     << bID << " " << start.X() << " " << start.Y() << " " << start.Z() << " " 
	     << end.X() << " " << end.Y() << " " << end.Z() << " " 
	     << track->MinTNS() << " " << track->MaxTNS() << std::endl;
      }
      fout.close();
      
    } // end for trkIdx
    
    return;    
  }
  
  //......................................................................
    
  void AlignInspect::reconfigure(const fhicl::ParameterSet& pset)
  {
  }  
  //......................................................................
  
  void AlignInspect::endJob() 
  {
  }
    
  //.....................................................................
  
  bool AlignInspect::IsGoodTrack(art::Ptr<rb::Track> t, double x, double y, double z)
  {        
    art::ServiceHandle<geo::Geometry> geom;
    if (t->ExtentPlane() >= 8)                         return true;
    else return false;         
    if (y > geom->DetHalfHeight()-fDetEdgeBuffer)      return true;
    if (fabs(x) > geom->DetHalfWidth()-fDetEdgeBuffer) return true;
    if (z < fDetEdgeBuffer)                            return true;        
    if (z > geom->DetLength()-fDetEdgeBuffer)          return true;
    return false;        	
  } 
} //namespace align
////////////////////////////////////////////////////////////////////////

namespace align
{
  DEFINE_ART_MODULE(AlignInspect);
}
///////////////////////////////////////////////////////////////////////

