////////////////////////////////////////////////////////////////////////
// \file    BPFEnergyAna_module.cc
// \brief   Module to create the TTree and histos used to examine and
//          validate the performance of the BPFEnergyEstimator module.
// \version 
// \author  $Author: mbaird42 $
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <vector>

// ROOT includes
#include "TH1F.h"
#include "TH2F.h"
#include "TTree.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "RecoBase/Cluster.h"
#include "BreakPointFitter/BPFEnergy.h"

#include "Utilities/AssociationUtil.h"

#include "MCCheater/BackTracker.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"



// BPFEnergyAna header
namespace bpfit {
  class BPFEnergyAna : public art::EDAnalyzer {
  public:
    explicit BPFEnergyAna(fhicl::ParameterSet const& pset);    
    ~BPFEnergyAna();                        
    
    void analyze(const art::Event& evt); 
    void reconfigure(const fhicl::ParameterSet& pset);
    void beginJob();
    void resetVars();
    void setTruthVars(const std::vector<cheat::NeutrinoEffPur>& nuEP);

  private:
    std::string fSlicerLabel; ///< label for module that made the slices
    std::string fEnergyLabel; ///< label for module that made the energy objects
    bool        fIsMC;        ///< are you running over MC?

    TTree *fBPFEnergyTree;    ///< TTree with slice level info (for energy estimator training)


    
    //
    // Variables to go into the Energy TTree.
    //

    // variables from the BPFEnergy object
    float fEreco1;    ///< first reconstructed event energy
    float fEres1;     ///< first estimated event energy resolution
    float fEreco2;    ///< second reconstructed event energy
    float fEres2;     ///< second estimated event energy resolution
    float fEreco3;    ///< third reconstructed event energy
    float fEres3;     ///< third estimated event energy resolution
    float fPmu;       ///< reconstructed muon momentum from the track selected to be the muon
    float fDirZmu;    ///< track.dirZ() from the track selected to be the muon
    float fN3Dprongs; ///< number of fuzzyk 3D prongs
    float fE3Dprongs; ///< sum of hit energies in 3D prongs (excluding the "muon" prong) - this does NOT account for dead material
    float fEremain;   ///< sum of remaining slice energy not in 3D prongs - this does NOT account for dead material
    float fSumPE;     ///< sum of PE from all hits not on the muon track
    float fEventID;   ///< best BPF muon PID value for the event

    // MCTruth and slice variables used for cutting etc.
    float fE;     ///< true energy
    float fMinX;  ///< minimum hit X value in the slice
    float fMinY;  ///< minimum hit Y value in the slice
    float fMinZ;  ///< minimum hit Z value in the slice
    float fMaxX;  ///< maximum hit X value in the slice
    float fMaxY;  ///< maximum hit Y value in the slice
    float fMaxZ;  ///< maximum hit Z value in the slice
    int fNHitX;   ///< number of X-view hits in the slice
    int fNHitY;   ///< number of Y-view hits in the slice
    int fCCNC;    ///< is this CC or NC (from truth)
    int fIntType; ///< neutrino interaction type (from truth)
    int fnuPDG;   ///< PDG code for the best matched nu (from truth)

  };
}



namespace bpfit
{
  //.......................................................................
  BPFEnergyAna::BPFEnergyAna(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset)
  {
    this->reconfigure(pset);
  }

  //......................................................................
  BPFEnergyAna::~BPFEnergyAna()
  {
    //======================================================================
    // Clean up any memory allocated by your module... ...if you dare...
    //======================================================================
  }

  //......................................................................
  void BPFEnergyAna::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSlicerLabel = pset.get<std::string>("SlicerLabel");
    fEnergyLabel = pset.get<std::string>("EnergyLabel");
    fIsMC        = pset.get<bool>       ("IsMC"); 
  }
      
  //......................................................................
  void BPFEnergyAna::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    //
    // Book the Energy TTree
    //
    fBPFEnergyTree = tfs->make<TTree>("BPFEnergyTree","Tree containing BPFEnergy object info");

    // attach the variables:
    fBPFEnergyTree->Branch("Ereco1",&fEreco1,"Ereco1/F");
    fBPFEnergyTree->Branch("Eres1",&fEres1,"Eres1/F");
    fBPFEnergyTree->Branch("Ereco2",&fEreco2,"Ereco2/F");
    fBPFEnergyTree->Branch("Eres2",&fEres2,"Eres2/F");
    fBPFEnergyTree->Branch("Ereco3",&fEreco3,"Ereco3/F");
    fBPFEnergyTree->Branch("Eres3",&fEres3,"Eres3/F");
    fBPFEnergyTree->Branch("Pmu",&fPmu,"Pmu/F");
    fBPFEnergyTree->Branch("DirZmu",&fDirZmu,"DirZmu/F");
    fBPFEnergyTree->Branch("N3Dprongs",&fN3Dprongs,"N3Dprongs/F");
    fBPFEnergyTree->Branch("E3Dprongs",&fE3Dprongs,"E3Dprongs/F");
    fBPFEnergyTree->Branch("Eremain",&fEremain,"Eremain/F");
    fBPFEnergyTree->Branch("SumPE",&fSumPE,"SumPE/F");
    fBPFEnergyTree->Branch("EventID",&fEventID,"EventID/F");

    fBPFEnergyTree->Branch("E",&fE,"E/F");
    fBPFEnergyTree->Branch("MinX",&fMinX,"MinX/F");
    fBPFEnergyTree->Branch("MinY",&fMinY,"MinY/F");
    fBPFEnergyTree->Branch("MinZ",&fMinZ,"MinZ/F");
    fBPFEnergyTree->Branch("MaxX",&fMaxX,"MaxX/F");
    fBPFEnergyTree->Branch("MaxY",&fMaxY,"MaxY/F");
    fBPFEnergyTree->Branch("MaxZ",&fMaxZ,"MaxZ/F");
    fBPFEnergyTree->Branch("NHitX",&fNHitX,"NHitX/I");
    fBPFEnergyTree->Branch("NHitY",&fNHitY,"NHitY/I");
    fBPFEnergyTree->Branch("CCNC",&fCCNC,"CCNC/I");
    fBPFEnergyTree->Branch("IntType",&fIntType,"IntType/I");
    fBPFEnergyTree->Branch("nuPDG",&fnuPDG,"nuPDG/I");

  }

  //......................................................................
  void BPFEnergyAna::resetVars()
  {
    fEreco1    = -5.0;
    fEres1     = -5.0;
    fEreco2    = -5.0;
    fEres2     = -5.0;
    fEreco3    = -5.0;
    fEres3     = -5.0;
    fPmu       = -5.0;
    fDirZmu    = -5.0;
    fN3Dprongs = -5.0;
    fE3Dprongs = -5.0;
    fEremain   = -5.0;
    fSumPE     = -5.0;
    fEventID   = -5.0;

    fE         = -5.0;
    fMinX      = 1.0e9;
    fMinY      = 1.0e9;
    fMinZ      = 1.0e9;
    fMaxX      = -1.0e9;
    fMaxY      = -1.0e9;
    fMaxZ      = -1.0e9;
    fNHitX     = -5;
    fNHitY     = -5;
    fCCNC      = -5;
    fIntType   = -5;
    fnuPDG     = 0;
  }

  //......................................................................
  void BPFEnergyAna::setTruthVars(const std::vector<cheat::NeutrinoEffPur>& nuEP)
  {
    //
    // This function sets any variables for the TTree that come from truth.
    // Currently, the most pure neutrino is assumed to be the best
    // match.
    //

    float mostPure = 0.0;
    unsigned int match = 0;
    for(unsigned int i = 0; i < nuEP.size(); ++i) {
      if(nuEP[i].purity > mostPure && nuEP[i].neutrinoInt->NeutrinoSet()) {
	mostPure = nuEP[i].purity;
	match = i;
      }
    }
    
    if(mostPure > 0 && nuEP[match].neutrinoInt->NeutrinoSet()) {
      simb::MCNeutrino nu = nuEP[match].neutrinoInt->GetNeutrino();
      simb::MCParticle Nu = nu.Nu();
      fE       = Nu.E();
      fCCNC    = nu.CCNC();
      fIntType = nu.InteractionType();
      fnuPDG   = Nu.PdgCode(); 
    }
    
  }

  //......................................................................
  void BPFEnergyAna::analyze(const art::Event& evt)
  {

    // get the slices
    art::Handle< std::vector< rb::Cluster > > slices;
    evt.getByLabel(fSlicerLabel, slices);

    // create the FindMany object for getting the energy objects
    art::FindManyP<bpfit::BPFEnergy> energyFMP(slices, evt, fEnergyLabel);

    // Make the required service handles...
    art::ServiceHandle<cheat::BackTracker> BT;

    // Put the slices into a PtrVector and get the list neutrinos matched
    // to each slice from backtracker.
    art::PtrVector<rb::Cluster> ARTslices;
    for(unsigned int i = 0; i < slices->size(); ++i) {
      ARTslices.push_back(art::Ptr<rb::Cluster>(slices,i));
    }
    std::vector< std::vector< cheat::NeutrinoEffPur > > ALLnuEffPur;
    if(fIsMC) {
      ALLnuEffPur = BT->SlicesToMCTruthsTable(ARTslices);
    }

    
    // loop over all slices and get the fuzzyk 3D prongs
    for(unsigned int islice = 0; islice < slices->size(); ++islice) {

      // As usual, skip the noise slice...
      if((*slices)[islice].IsNoise()) continue;
      
      // NOTE: Currently, the fundamental object of analysis is the slice.
      // If in the future we change things so that the fundamental object
      // of analysis becomes the vertex, then we will have to get prongs
      // associated with the vertex instead.

      // Reset all TTree variables.
      this->resetVars();

      // Set slice level and truth variables for the TTrees.
      fMinX  = (*slices)[islice].MinX();
      fMinY  = (*slices)[islice].MinY();
      fMinZ  = (*slices)[islice].MinZ();
      fMaxX  = (*slices)[islice].MaxX();
      fMaxY  = (*slices)[islice].MaxY();
      fMaxZ  = (*slices)[islice].MaxZ();
      fNHitX = (*slices)[islice].NXCell();
      fNHitY = (*slices)[islice].NYCell();
      if(fIsMC) {
	this->setTruthVars(ALLnuEffPur[islice]);
      }

      // get the energy object associated with this slice
      std::vector< art::Ptr< bpfit::BPFEnergy > > energies;
      if(energyFMP.isValid()) energies = energyFMP.at(islice);

      // There can be only one!!!
      // Something went wrong if there is more than one energy object
      // per slice.
      if(energies.size() != 1) continue;

      // Set remaining branch values and fill the TTree.
      fEreco1    = energies[0]->E1();
      fEres1     = energies[0]->Eres1();
      fEreco2    = energies[0]->E2();
      fEres2     = energies[0]->Eres2();
      fEreco3    = energies[0]->E3();
      fEres3     = energies[0]->Eres3();
      fPmu       = energies[0]->PMuon();
      fDirZmu    = energies[0]->DirZMuon();
      fN3Dprongs = energies[0]->N3DProngs();
      fE3Dprongs = energies[0]->EFuzzyK3D();
      fEremain   = energies[0]->ERemain();
      fSumPE     = energies[0]->SumPE();
      fEventID   = energies[0]->EventID();

      fBPFEnergyTree->Fill();

    } // end loop over slices (islice)
    
    return;

  }

} // end namespace bpfit
////////////////////////////////////////////////////////////////////////



#include "art/Framework/Core/ModuleMacros.h"

namespace bpfit
{
  DEFINE_ART_MODULE(BPFEnergyAna);
}
