////////////////////////////////////////////////////////////////////////
// \file    BPFdEdxHistoMaker_module.cc
// \brief   Module to create dE/dx vs. beta*gamma plots from BPF tracks
// \version 
// \author  $Author: mbaird42 $
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <vector>

// ROOT includes
#include "TH2F.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "RecoBase/Cluster.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Track.h"
#include "RecoBase/FitSum.h"

#include "Utilities/AssociationUtil.h"
#include "Geometry/Geometry.h"
#include "Calibrator/Calibrator.h"

#include "BreakPointFitter/dEdxCalculator.h"



// BPFdEdxHistoMaker header
namespace bpfit {
  class BPFdEdxHistoMaker : public art::EDAnalyzer {
  public:
    explicit BPFdEdxHistoMaker(fhicl::ParameterSet const& pset);    
    ~BPFdEdxHistoMaker();                        
    
    void analyze(const art::Event& evt); 
    void reconfigure(const fhicl::ParameterSet& pset);
    void beginJob();
    void endJob();

  private:
    std::string   fSlicerLabel;
    std::string   fProngLabel;
    std::string   fProngInstance;
    std::string   fTrackLabel;

    double       fdxTOL;      ///< cut off value for dx, if dx is less than this, don't compute dE/dx
    bool         fSmooth;     ///< Smooth the likelihood histos at the end?
    unsigned int fNsmooth;    ///< number of bins on either side of the current bin to smooth over 
    bool         fNormalize;  ///< Normalize the likelihood histos at the end?

    TH2F  *fdEdxVSbgMu[2][2]; ///< Computed dE/dx for cell hits (vs. beta*gamma) for the muon assupmtion.
    TH2F  *fdEdxVSbgPi[2][2]; ///< Computed dE/dx for cell hits (vs. beta*gamma) for the pion assupmtion.
    TH2F  *fdEdxVSbgPr[2][2]; ///< Computed dE/dx for cell hits (vs. beta*gamma) for the proton assupmtion.

    bpfit::dEdxCalculator fdEdxCalc; ///< helper class for computing dEdx values

  };
}



namespace bpfit
{
  //.......................................................................
  BPFdEdxHistoMaker::BPFdEdxHistoMaker(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset),
      fdEdxCalc()
  {
    this->reconfigure(pset);
  }

  //......................................................................
  BPFdEdxHistoMaker::~BPFdEdxHistoMaker()
  {
    //======================================================================
    // Clean up any memory allocated by your module... ...if you dare...
    //======================================================================
  }

  //......................................................................
  void BPFdEdxHistoMaker::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSlicerLabel   = pset.get<std::string> ("SlicerLabel");
    fProngLabel    = pset.get<std::string> ("ProngLabel");
    fProngInstance = pset.get<std::string> ("ProngInstance");
    fTrackLabel    = pset.get<std::string> ("TrackLabel");
    fdxTOL         = pset.get<double>      ("dxTOL");
    fSmooth        = pset.get<bool>        ("Smooth");
    fNsmooth       = pset.get<unsigned int>("Nsmooth");
    fNormalize     = pset.get<bool>        ("Normalize");
  }
      
  //......................................................................
  void BPFdEdxHistoMaker::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    //
    // Book the NTuples and histograms...
    //

    // We are dividing the dE/dx info from cell hits into multiple histograms
    // based on whether or not the track is steep (large dx) and whether the
    // cell hit is near to (W <= 0.0) or far from (W > 0.0) the readout. This
    // is to reflect the fact that the dE/dx calculations might be slightly
    // different depending on these conditions.

    // A note on naming of the histos below:
    //
    // * {Mu,Pi,Pr} obviously refers to the particle assumption
    // * W{0,1} refers to whether the track W values was <= 0 (W0) or > 0 (W1)
    // * X{0,1} refers to whether the dx value was <= 8 cm (X0) or > 8 cm (X1)

    for(unsigned int w = 0; w < 2; ++w) {
      for(unsigned int x = 0; x < 2; ++x) {

	char histoname[30];

	sprintf(histoname, "fdEdxVSbgMu_W%.1u_X%.1u", w, x);
	fdEdxVSbgMu[w][x] = tfs->make<TH2F>(histoname,
					    "dE/dx vs. #beta#gamma (muon track assumption);ln(#beta#gamma);dE/dx [GeV/cm]",
					    160,-2.0,6.0,400,0.0,0.04);

	sprintf(histoname, "fdEdxVSbgPi_W%.1u_X%.1u", w, x);
	fdEdxVSbgPi[w][x] = tfs->make<TH2F>(histoname,
					    "dE/dx vs. #beta#gamma (pion track assumption);ln(#beta#gamma);dE/dx [GeV/cm]",
					    160,-2.0,6.0,400,0.0,0.04);

	sprintf(histoname, "fdEdxVSbgPr_W%.1u_X%.1u", w, x);
	fdEdxVSbgPr[w][x] = tfs->make<TH2F>(histoname,
					    "dE/dx vs. #beta#gamma (proton track assumption);ln(#beta#gamma);dE/dx [GeV/cm]",
					    160,-2.0,6.0,400,0.0,0.04);
	

      } // end loop over x
    } // end loop over w

  }

  //......................................................................
  void BPFdEdxHistoMaker::endJob()
  {

    //
    // Apply simple smoothing if requested.
    //
    // The smoothing is simple: set each bin to the average of that bin
    // with the bin immedately above and below. Do NOT include the
    // overflow bins in the smoothing.
    //

    if(fSmooth) {
      
      unsigned int NXbins = 0.0;
      unsigned int NYbins = 0.0;
      double ave = 0.0;
      
      for(unsigned int w = 0; w < 2; ++w) {
	for(unsigned int x = 0; x < 2; ++x) {

	  // muon histos
	  NXbins = fdEdxVSbgMu[w][x]->GetNbinsX();
	  NYbins = fdEdxVSbgMu[w][x]->GetNbinsY();
	  for(unsigned int i = 1; i <= NXbins; ++i) {
	    for(unsigned int j = 1+fNsmooth; j+fNsmooth <= NYbins; ++j) {
	      ave = 0.0;
	      for(unsigned int s = j-fNsmooth; s <= j+fNsmooth; ++s)
		ave += fdEdxVSbgMu[w][x]->GetBinContent(i,s);
	      ave = ave/(2*fNsmooth+1);
	      fdEdxVSbgMu[w][x]->SetBinContent(i,j,ave);
	    }
	  }

	  // pion histos
	  NXbins = fdEdxVSbgPi[w][x]->GetNbinsX();
	  NYbins = fdEdxVSbgPi[w][x]->GetNbinsY();
	  for(unsigned int i = 1; i <= NXbins; ++i) {
	    for(unsigned int j = 1+fNsmooth; j+fNsmooth <= NYbins; ++j) {
	      ave = 0.0;
	      for(unsigned int s = j-fNsmooth; s <= j+fNsmooth; ++s)
		ave += fdEdxVSbgPi[w][x]->GetBinContent(i,s);
	      ave = ave/(2*fNsmooth+1);
	      fdEdxVSbgPi[w][x]->SetBinContent(i,j,ave);
	    }
	  }

	  // proton histos
	  NXbins = fdEdxVSbgPr[w][x]->GetNbinsX();
	  NYbins = fdEdxVSbgPr[w][x]->GetNbinsY();
	  for(unsigned int i = 1; i <= NXbins; ++i) {
	    for(unsigned int j = 1+fNsmooth; j+fNsmooth <= NYbins; ++j) {
	      ave = 0.0;
	      for(unsigned int s = j-fNsmooth; s <= j+fNsmooth; ++s)
		ave += fdEdxVSbgPr[w][x]->GetBinContent(i,s);
	      ave = ave/(2*fNsmooth+1);
	      fdEdxVSbgPr[w][x]->SetBinContent(i,j,ave);
	    }
	  }

	} // end loop over x
      } // end loop over w
      
    } // end smoothing



    //
    // Normalize the likelihood histos if requested.
    //
    // The normalization is so that the area of each vertical stip
    // (one bin wide) is 1.0. Since each bin represents a probability,
    // we will include the overflow bins in the normalization.
    //

    if(fNormalize) {
      
      unsigned int NXbins = 0.0;
      unsigned int NYbins = 0.0;
      
      for(unsigned int w = 0; w < 2; ++w) {
	for(unsigned int x = 0; x < 2; ++x) {

	  // muon histos
	  NXbins = fdEdxVSbgMu[w][x]->GetNbinsX();
	  NYbins = fdEdxVSbgMu[w][x]->GetNbinsY();
	  for(unsigned int i = 0; i <= NXbins+1; ++i) {
	    float integral = 0.0;
	    for(unsigned int j = 0; j <= NYbins+1; ++j)
	      integral += fdEdxVSbgMu[w][x]->GetBinContent(i,j);
	    if(integral == 0.0) continue;
	    for(unsigned int j = 0; j <= NYbins+1; ++j)
	      fdEdxVSbgMu[w][x]->SetBinContent(i,j,fdEdxVSbgMu[w][x]->GetBinContent(i,j)/integral);
	  }
	  
	  // pion histos
	  NXbins = fdEdxVSbgPi[w][x]->GetNbinsX();
	  NYbins = fdEdxVSbgPi[w][x]->GetNbinsY();
	  for(unsigned int i = 0; i <= NXbins+1; ++i) {
	    float integral = 0.0;
	    for(unsigned int j = 0; j <= NYbins+1; ++j)
	      integral += fdEdxVSbgPi[w][x]->GetBinContent(i,j);
	    if(integral == 0.0) continue;
	    for(unsigned int j = 0; j <= NYbins+1; ++j)
	      fdEdxVSbgPi[w][x]->SetBinContent(i,j,fdEdxVSbgPi[w][x]->GetBinContent(i,j)/integral);
	  }
	  
	  // proton histos
	  NXbins = fdEdxVSbgPr[w][x]->GetNbinsX();
	  NYbins = fdEdxVSbgPr[w][x]->GetNbinsY();
	  for(unsigned int i = 0; i <= NXbins+1; ++i) {
	    float integral = 0.0;
	    for(unsigned int j = 0; j <= NYbins+1; ++j)
	      integral += fdEdxVSbgPr[w][x]->GetBinContent(i,j);
	    if(integral == 0.0) continue;
	    for(unsigned int j = 0; j <= NYbins+1; ++j)
	      fdEdxVSbgPr[w][x]->SetBinContent(i,j,fdEdxVSbgPr[w][x]->GetBinContent(i,j)/integral);
	  }
      
	} // end loop over x
      } // end loop over w

    } // end normalization
    
  }

  //......................................................................
  void BPFdEdxHistoMaker::analyze(const art::Event& evt)
  {

    // get the slices
    art::Handle< std::vector< rb::Cluster > > slices;
    evt.getByLabel(fSlicerLabel, slices);

    // create the FindMany object for getting prongs
    art::FindManyP<rb::Prong> prongFMP(slices, evt, art::InputTag(fProngLabel,fProngInstance));

    // make a geometry service handle
    art::ServiceHandle<geo::Geometry> geom;

    // make a calibrator service handle
    art::ServiceHandle<calib::Calibrator> cal;



    // loop over all slices and get the fuzzyk 3D prongs
    for(unsigned int islice = 0; islice < slices->size(); ++islice) {

      // As usual, skip the noise slice...
      if((*slices)[islice].IsNoise()) continue;
      
      // NOTE: Currently, the fundamental object of analysis is the slice.
      // If in the future we change things so that the fundamental object
      // of analysis becomes the vertex, then we will have to get prongs
      // associated with the vertex instead.

      // get the 3D prongs associated with this slice
      std::vector< art::Ptr< rb::Prong > > prongs;
      if(prongFMP.isValid()) prongs = prongFMP.at(islice);

      // create the FindMany object for getting tracks
      art::FindManyP<rb::Track> trackFMP(prongs, evt, fTrackLabel);

      // loop over all prongs and get the tracks
      for(unsigned int iprong = 0; iprong < prongs.size(); ++iprong) {

	// get the tracks associated with this prong
	std::vector< art::Ptr< rb::Track > > tracks;
	if(trackFMP.isValid()) tracks = trackFMP.at(iprong);

	// create the FindMany object for getting FitSum objects
	art::FindManyP<rb::FitSum> fitsumFMP(tracks, evt, fTrackLabel);

	// Loop over all tracks...
	for(unsigned int itrack = 0; itrack < tracks.size(); ++itrack) {

	  // get the FitSum object associated with this track
	  std::vector< art::Ptr< rb::FitSum > > fitsums;
	  if(fitsumFMP.isValid()) fitsums = fitsumFMP.at(itrack);

	  // There can be only one!!!
	  // (something went really wrong with BPF if there's not one and
	  // only one of these per track.)
	  if(fitsums.size() != 1) abort();

	  // Note to the user:
	  // It is intended that this program only runs over a sample of
	  // muons that have been tracked only under the muon assumption. I
	  // haven't forced that asumption here.

	  //
	  // Perform the dE/dx calculations with the helper class.
	  //
	  fdEdxCalc.computeDEDX(tracks[itrack],fitsums[0]->PDG());
	  
	  std::vector<double> dE;
	  std::vector<double> dx;
	  std::vector<double> W;
	  std::vector<double> BG;

	  fdEdxCalc.getResults(dE,dx,W,BG,fdxTOL);

	  for(unsigned int a = 0; a < dE.size(); ++a) {

	    unsigned int w = 0;
	    unsigned int x = 0;

	    if(W[a]  > 0.0) w = 1;
	    if(dx[a] > 8.0) x = 1;

	    if(fitsums[0]->PDG() == 13)
	      fdEdxVSbgMu[w][x]->Fill(log(BG[a]),dE[a]/dx[a]);
	    else if(fitsums[0]->PDG() == 211)
	      fdEdxVSbgPi[w][x]->Fill(log(BG[a]),dE[a]/dx[a]);
	    else if(fitsums[0]->PDG() == 2212)
	      fdEdxVSbgPr[w][x]->Fill(log(BG[a]),dE[a]/dx[a]);
	    else {
	      std::cout << "\n\nERROR: PDG code " << fitsums[0]->PDG()
			<< " is not supported! Aborting...\n\n";
	      abort();
	    }
	  }

	} // end loop over tracks (itrack)
	
      } // end loop over prongs (iprong)
      
    } // end loop over slices (islice)
    
    return;

  }

} // end namespace bpfit
////////////////////////////////////////////////////////////////////////



#include "art/Framework/Core/ModuleMacros.h"

namespace bpfit
{
  DEFINE_ART_MODULE(BPFdEdxHistoMaker);
}
