///
/// \file    TrackBasis.cxx
/// \brief   Construct a track-based (x,y,z) coordinate system
/// \author  messier@indiana.edu
/// \version $Id:$
///
#include "BreakPointFitter/func/TrackBasis.h"
#include "BreakPointFitter/func/BPException.h"
#include "TVector3.h"
#include "Utilities/func/MathUtil.h"
using namespace bpfit;

#define SQR(x) ((x)*(x))

void TrackBasis::FindAxis(const std::vector<double>& x,
			  const std::vector<double>& y,
			  const std::vector<double>& z,
			  const std::vector<double>& dx,
			  const std::vector<double>& dy,
			  const std::vector<double>& dz,
			  TVector3& z1)
{
  unsigned int i;
  std::vector<double> w(z.size()); // Hit weights in x/z then y/z views
  
  //
  // Space point vectors must all have same number of entries
  //
  if(x.size() != z.size() || y.size() != z.size() ||
     dx.size() != x.size() || dy.size() != y.size() ||
     dz.size() != z.size()) {
    throw BPException(__FILE__, __LINE__, kINDEX);
  }
  
  //
  // Make line fit in z-x projection
  //
  double mzx; // Slope in x-z view
  double bzx; // Intercept in x-z view
  for (i=0; i<w.size(); ++i) {
    w[i] = dx[i]*dx[i] + dz[i]+dz[i];
  }
  util::LinFitTS(z, x, w, mzx, bzx, 2);

  //
  // Make line fit in z-y projection
  //
  double mzy; // Slope in y-z view
  double bzy; // Intercept in y-z view
  for (i=0; i<w.size(); ++i) {
    w[i] = dy[i]*dy[i] + dz[i]+dz[i];
  }
  util::LinFitTS(z, y, w, mzy, bzy, 2);
  
  z1.SetXYZ( mzx,  mzy,  1.0);
  z1.SetMag(1.0);
}

//......................................................................

void TrackBasis::MakeBasis(const std::vector<double>& x,
			   const std::vector<double>& y,
			   const std::vector<double>& z,
			   const std::vector<double>& dx,
			   const std::vector<double>& dy,
			   const std::vector<double>& dz,
			   double vx, double vy, double vz)
{
  this->SetOrigin(vx,vy,vz);

  //
  // The detector based basis vectors
  //
  static const TVector3 x0(1,0,0);
  static const TVector3 y0(0,1,0);
  static const TVector3 z0(0,0,1);

  //
  // Set the z-axis to follow the track
  //  
  TrackBasis::FindAxis(x, y, z, dx, dy, dz, fZ1);
  
  //
  // Compute some cross products to find the detector axis that is
  // most complementary
  //
  TVector3 ZxX(x0.Cross(fZ1));
  TVector3 YxZ(y0.Cross(fZ1));
  TVector3 ZxZ(z0.Cross(fZ1));

  double ZxXmag = ZxX.Mag2();
  double YxZmag = YxZ.Mag2();
  double ZxZmag = ZxZ.Mag2();
  
  //
  // Choose the other two axes to maximize complementarity with the
  // detector coordinates
  //
  if ((YxZmag>=ZxXmag) && (YxZmag>=ZxZmag)) {
    fX1 = YxZ.Unit();
    fY1 = fZ1.Cross(fX1);
  }
  else if ((ZxXmag>=YxZmag) && (ZxXmag>=ZxZmag)) {
    fY1 = ZxX.Unit();
    fX1 = fY1.Cross(fZ1);
  }
  else {
    fX1 = ZxZ.Unit();
    fY1 = fZ1.Cross(fX1);
  }

  //
  // Use the unit vectors to find the rotation matrix
  //
  fU[0][0] = fUT[0][0] = x0.Dot(fX1);
  fU[0][1] = fUT[1][0] = y0.Dot(fX1);
  fU[0][2] = fUT[2][0] = z0.Dot(fX1);

  fU[1][0] = fUT[0][1] = x0.Dot(fY1);
  fU[1][1] = fUT[1][1] = y0.Dot(fY1);
  fU[1][2] = fUT[2][1] = z0.Dot(fY1);

  fU[2][0] = fUT[0][2] = x0.Dot(fZ1);
  fU[2][1] = fUT[1][2] = y0.Dot(fZ1);
  fU[2][2] = fUT[2][2] = z0.Dot(fZ1);

  //
  // Check that that the basis vectors point in the intended direction
  // of the track (based on the vertex location) and adjust the rotation
  // matrix if needed.
  //
  this->AdjustBasisDir(x,y,z,dx,dy,dz);
}

//......................................................................

void TrackBasis::AdjustBasisDir(const std::vector<double>& x,
				const std::vector<double>& y,
				const std::vector<double>& z,
				const std::vector<double>& dx,
				const std::vector<double>& dy,
				const std::vector<double>& dz)
{
  //
  // Compute the average Z coordinate in the new basis (weighted by dz.)
  //
  unsigned int i;
  double z1ave = 0.0;
  double sumdz = 0.0;
  double tol   = 0.1;
  double w;
  double x1,  y1,  z1;
  double dx1, dy1, dz1;
  
  for(i=0; i<x.size(); ++i) {    
    this->DetToTrk(x[i], y[i], z[i], dx[i], dy[i], dz[i],
		   &x1,  &y1,  &z1,  &dx1,  &dy1,  &dz1);
    if (dz1<tol) dz1 = tol;
    w = 1.0/(dz1*dz1);
    z1ave += w*z1;
    sumdz += w;
  }
  z1ave /= sumdz;

  //
  // If z1ave is behind the vertex (<0) then the new Z axis points away
  // from the intended track direction.  So reverse the Z axis in the
  // rotation matrix, and reverse the Y axis as well to maintain a
  // right handed coordinate system.
  //
  if(z1ave < 0.0) {
    fY1 = -fY1;
    fZ1 = -fZ1;

    fU[1][0] = fUT[0][1] = -fU[1][0];
    fU[1][1] = fUT[1][1] = -fU[1][1];
    fU[1][2] = fUT[2][1] = -fU[1][2];
    
    fU[2][0] = fUT[0][2] = -fU[2][0];
    fU[2][1] = fUT[1][2] = -fU[2][1];
    fU[2][2] = fUT[2][2] = -fU[2][2];
  }
}

//......................................................................

void TrackBasis::SetOrigin(double x, double y, double z) 
{
  fO.SetXYZ(x, y, z);
}

//......................................................................

void TrackBasis::DetToTrk(double  x,   double  y,   double  z,
			  double  dx,  double  dy,  double  dz,
			  double* x1,  double* y1,  double* z1,
			  double* dx1, double* dy1, double* dz1)
{
  x -= fO.X();
  y -= fO.Y();
  z -= fO.Z();

  *x1 = fU[0][0]*x + fU[0][1]*y + fU[0][2]*z;
  *y1 = fU[1][0]*x + fU[1][1]*y + fU[1][2]*z;
  *z1 = fU[2][0]*x + fU[2][1]*y + fU[2][2]*z;
  
  *dx1 = SQR(fU[0][0]*dx) + SQR(fU[0][1]*dy) + SQR(fU[0][2]*dz);
  *dy1 = SQR(fU[1][0]*dx) + SQR(fU[1][1]*dy) + SQR(fU[1][2]*dz);
  *dz1 = SQR(fU[2][0]*dx) + SQR(fU[2][1]*dy) + SQR(fU[2][2]*dz);
  
  *dx1 = sqrt(*dx1);
  *dy1 = sqrt(*dy1);
  *dz1 = sqrt(*dz1);
}

//......................................................................

void TrackBasis::TrkToDet(double  x,   double  y,   double  z,
			  double  dx,  double  dy,  double  dz,
			  double* x1,  double* y1,  double* z1,
			  double* dx1, double* dy1, double* dz1)
{
  *x1 = fUT[0][0]*x + fUT[0][1]*y + fUT[0][2]*z;
  *y1 = fUT[1][0]*x + fUT[1][1]*y + fUT[1][2]*z;
  *z1 = fUT[2][0]*x + fUT[2][1]*y + fUT[2][2]*z;
  
  *x1 += fO.X();
  *y1 += fO.Y();
  *z1 += fO.Z();
  
  *dx1 = SQR(fUT[0][0]*dx) + SQR(fUT[0][1]*dy) + SQR(fUT[0][2]*dz);
  *dy1 = SQR(fUT[1][0]*dx) + SQR(fUT[1][1]*dy) + SQR(fUT[1][2]*dz);
  *dz1 = SQR(fUT[2][0]*dx) + SQR(fUT[2][1]*dy) + SQR(fUT[2][2]*dz);
  
  *dx1 = sqrt(*dx1);
  *dy1 = sqrt(*dy1);
  *dz1 = sqrt(*dz1);
}

////////////////////////////////////////////////////////////////////////
				
