///
/// \file   dEdxTables.cxx
/// \brief  Implement energy loss tables using GEANT4
/// \author messier@indiana.edu
///
/// These tables were calculated using novasoft on June 16, 2015 using GEANT4:
///
/// Geant4 version Name: geant4-09-06-patch-04a    (04-February-2015)
/// Copyright : Geant4 Collaboration
/// Reference : NIM A 506 (2003), 250-303
/// WWW : http://cern.ch/geant4
///
/// In particular, they come from G4EmCalculator::ComputeTotalDEDX()
/// for muons in various NOvA materials. This is implemented into the
/// g4nova/PrintTables functions.
///
#include "BreakPointFitter/func/dEdxTable.h"
#include <string>
#include <cmath>
#include <iostream>
using namespace bpfit;
//
// Muon mass used by Groom, Mokhov, Striganov to compute their tables
//
static const double gsMmu = 105.6583568E-3;

//......................................................................

dEdxTable::dEdxTable(const char* mat) : 
  fSz(0),
  fT(0),
  fdEdx(0),
  fRho(0)
{
  static const std::string scint ("Scintillator");
  static const std::string pvc   ("PVC");
  static const std::string glue  ("Glue");
  static const std::string steel ("Steel");
  static const std::string vac   ("Vacuum");

  if      (scint==mat) this->InitScintillator();
  else if (pvc  ==mat) this->InitPVC();
  else if (steel==mat) this->InitSteel();
  else if (glue ==mat) this->InitGlue();
  else if (vac  ==mat) this->InitVac();
  else {
    static unsigned int nwarn = 10;
    if (nwarn>0) {
      --nwarn;
      std::cerr << __FILE__ << ":" << __LINE__ 
		<< " Have no table for material " << mat
		<< " Defaulting to Vacuum.";
      if (nwarn==0) {
	std::cerr << " Last warning.";
      }
      std::cerr << std::endl;
      this->InitVac();
    }
  }
}

//......................................................................

double dEdxTable::TtoBetaGamma(double t, double m) const
{
  double x = t/m;
  return sqrt(x*(x+2));
}

//......................................................................

double dEdxTable::BetaGammaToT(double bg, double m) const
{
  return m*(sqrt(bg*bg+1)-1);
}

//......................................................................

double dEdxTable::Interp(double t,
			 double m,
			 const double T[],
			 const double dedx[],
			 unsigned int imax) const
{
  //
  // Compute equivalent muon kinetic energy. Ie. a muon with
  // the same beta*gamma
  //
  double tmu = this->BetaGammaToT( this->TtoBetaGamma(t, m), gsMmu );
  
  unsigned int i0=0, i1=1;
  if      (tmu<=T[0])      { i0=0;      i1=1;}
  else if (tmu>=T[imax-1]) { i0=imax-2; i1=imax-1;}
  else {
    for (; i1<imax; ++i0, ++i1) {
      if (tmu>=T[i0] && tmu<T[i1]) break;
    }
  }

  double x  = log(tmu);
  double y;
  double x0 = log(T[i0]);
  double x1 = log(T[i1]);
  double y0 = log(dedx[i0]);
  double y1 = log(dedx[i1]);
  
  y = y0 + (x-x0)*(y1-y0)/(x1-x0);

  return exp(y)*1E-3;
}

//......................................................................
//
// Initialize tables for PVC
//
void dEdxTable::InitPVC()
{
  fSz = 90;
  static const double T[] = {
    1e-05,1.26219e-05,1.59312e-05,2.01081e-05,2.53802e-05, 3.20345e-05,4.04335e-05,5.10347e-05,6.44153e-05,8.13041e-05, 
    0.000102621,0.000129527,0.000163487,0.000206351,0.000260454, 0.000328741,0.000414933,0.000523722,0.000661035,0.00083435, 
    0.00105311,0.00132922,0.00167772,0.00211759,0.0026728, 0.00337357,0.00425808,0.00537449,0.00678361,0.00856218, 
    0.0108071,0.0136405,0.0172169,0.0217309,0.0274285, 0.0346199,0.0436968,0.0551535,0.069614,0.0878659, 
    0.110903,0.13998,0.176681,0.223005,0.281474, 0.355273,0.44842,0.56599,0.714385,0.901688, 
    1.1381,1.43649,1.81312,2.2885,2.88851, 3.64584,4.60173,5.80825,7.33109,9.2532, 
    11.6793,14.7414,18.6064,23.4848,29.6422, 37.414,47.2234,59.6048,75.2323,94.9572, 
    119.854,151.278,190.941,241.003,304.191, 383.946,484.611,611.67,772.041,974.46, 
    1229.95,1552.43,1959.45,2473.2,3121.63, 3940.09,4973.12,6277.01,7922.76,10000
  };
  static const double dEdx[] = {
    424.556,426.33,418.121,402.057,380.783, 354.619,324.262,292.096,260.04,229.628, 
    201.168,175.142,151.723,130.909,112.39, 95.9017,81.3085,68.6823,57.7386,48.3363, 
    40.3657,33.6323,27.9845,23.2492,19.3059, 16.022,13.3128,11.0783,9.23704,7.72416, 
    6.48557,5.47401,4.65107,3.98441,3.44721, 3.01727,2.6763,2.40924,2.20372,2.04953, 
    1.93088,1.84586,1.79012,1.75821,1.74543, 1.74773,1.76164,1.78427,1.8132,1.8465, 
    1.88409,1.92359,1.964,2.00473,2.04536, 2.08564,2.12548,2.16482,2.20376,2.24249, 
    2.28131,2.32067,2.36118,2.40366,2.4492, 2.49918,2.5554,2.62016,2.69636,2.78768, 
    2.89878,3.03552,3.20529,3.41742,3.68375, 4.01927,4.44294,4.97872,5.65696,6.51601, 
    7.60428,8.98301,10.7296,12.9413,15.7416, 19.2858,23.7704,29.4432,36.6169,45.6867
  };
  fT    = T;
  fdEdx = dEdx;
  fRho  = 1.49;
}

//......................................................................
///
/// Initialize tables for steel
///
void dEdxTable::InitSteel()
{
  fSz = 90;
  static const double T[] = {
    1e-05,1.26219e-05,1.59312e-05,2.01081e-05,2.53802e-05, 3.20345e-05,4.04335e-05,5.10347e-05,6.44153e-05,8.13041e-05, 
    0.000102621,0.000129527,0.000163487,0.000206351,0.000260454, 0.000328741,0.000414933,0.000523722,0.000661035,0.00083435, 
    0.00105311,0.00132922,0.00167772,0.00211759,0.0026728, 0.00337357,0.00425808,0.00537449,0.00678361,0.00856218, 
    0.0108071,0.0136405,0.0172169,0.0217309,0.0274285, 0.0346199,0.0436968,0.0551535,0.069614,0.0878659, 
    0.110903,0.13998,0.176681,0.223005,0.281474, 0.355273,0.44842,0.56599,0.714385,0.901688, 
    1.1381,1.43649,1.81312,2.2885,2.88851, 3.64584,4.60173,5.80825,7.33109,9.2532, 
    11.6793,14.7414,18.6064,23.4848,29.6422, 37.414,47.2234,59.6048,75.2323,94.9572, 
    119.854,151.278,190.941,241.003,304.191, 383.946,484.611,611.67,772.041,974.46, 
    1229.95,1552.43,1959.45,2473.2,3121.63, 3940.09,4973.12,6277.01,7922.76,10000
  };
  static const double dEdx[] = {
    166.621,173.054,176.681,177.944,176.322, 171.568,163.869,153.938,142.771,130.502, 
    118.099,105.801,93.8621,82.1472,72.984, 63.7723,55.2818,47.4775,40.6667,34.6532, 
    29.2846,24.4571,20.7757,17.5254,14.7265, 12.3562,10.3518,8.68083,7.28429,6.11978, 
    5.15553,4.3687,3.73155,3.20907,2.7856, 2.44403,2.17498,1.96034,1.7941,1.66847, 
    1.57685,1.51361,1.47388,1.45346,1.44863, 1.45618,1.47331,1.49765,1.5272,1.56037, 
    1.59776,1.63637,1.67576,1.71547,1.7552, 1.79481,1.83432,1.87383,1.91362,1.95409, 
    1.99586,2.03976,2.08691,2.13877,2.19727, 2.26488,2.34478,2.44105,2.55885,2.70477, 
    2.88715,3.11656,3.40638,3.77349,4.23926, 4.83072,5.58208,6.53659,7.74894,9.28825, 
    11.2418,13.7199,16.8619,20.8432,25.8857, 32.2697,40.3486,50.5686,63.4931,79.8329
  };
  fT    = T;
  fdEdx = dEdx;
  fRho  = 7.87;
}

//......................................................................
//
// Vaccuum is a valid GEANT material. Set things very close to
// zero. Keep the tables small for sake of efficiency
//
void dEdxTable::InitVac()
{
  fSz = 2;
  static const double T[]    = { 10.0E-3,  800.0 };
  static const double dEdx[] = { 1.0E-6,   1.0E-6};
  fT    = T;
  fdEdx = dEdx;
  fRho  = 1.0E-6;
}

//......................................................................
//
// Tables for our scintillator
//
void dEdxTable::InitScintillator()
{
  fSz = 90;
  static const double T[] = {
    1e-05,1.26219e-05,1.59312e-05,2.01081e-05,2.53802e-05, 3.20345e-05,4.04335e-05,5.10347e-05,6.44153e-05,8.13041e-05, 
    0.000102621,0.000129527,0.000163487,0.000206351,0.000260454, 0.000328741,0.000414933,0.000523722,0.000661035,0.00083435, 
    0.00105311,0.00132922,0.00167772,0.00211759,0.0026728, 0.00337357,0.00425808,0.00537449,0.00678361,0.00856218, 
    0.0108071,0.0136405,0.0172169,0.0217309,0.0274285, 0.0346199,0.0436968,0.0551535,0.069614,0.0878659, 
    0.110903,0.13998,0.176681,0.223005,0.281474, 0.355273,0.44842,0.56599,0.714385,0.901688, 
    1.1381,1.43649,1.81312,2.2885,2.88851, 3.64584,4.60173,5.80825,7.33109,9.2532, 
    11.6793,14.7414,18.6064,23.4848,29.6422, 37.414,47.2234,59.6048,75.2323,94.9572, 
    119.854,151.278,190.941,241.003,304.191, 383.946,484.611,611.67,772.041,974.46, 
    1229.95,1552.43,1959.45,2473.2,3121.63, 3940.09,4973.12,6277.01,7922.76,10000
  };
  static const double dEdx[] = {
    625.37,625.141,610.22,582.585,548.627, 507.575,460.727,411.514,362.908,317.502, 
    275.446,237.44,203.651,173.939,147.6, 124.668,104.636,87.5435,72.9925,60.6649, 
    50.3317,41.699,34.5093,28.5405,23.5999, 19.5213,16.1625,13.4027,11.1403,9.28975, 
    7.77981,6.55106,5.55427,4.74871,4.10085, 3.58309,3.17283,2.85153,2.60401,2.41783, 
    2.27711,2.17609,2.10876,2.06872,2.05046, 2.0492,2.06086,2.08205,2.10996,2.1424, 
    2.17885,2.2173,2.25641,2.29556,2.33438, 2.37263,2.41026,2.44727,2.48375,2.51991, 
    2.556,2.59229,2.62906,2.66668,2.70569, 2.74683,2.79103,2.83952,2.89386,2.95606, 
    3.02861,3.11472,3.21842,3.34483,3.50041, 3.69335,3.93403,4.23556,4.61457,5.09208, 
    5.69461,6.45575,7.41802,8.63478,10.1737, 12.1202,14.582,17.6953,21.6317,26.6083
  };
  fT    = T;
  fdEdx = dEdx;
  fRho  = 0.859;
}

//......................................................................
//
// Tables for glue
///
void dEdxTable::InitGlue()
{
  fSz = 90;
  static const double T[] = {
    1e-05,1.26219e-05,1.59312e-05,2.01081e-05,2.53802e-05, 3.20345e-05,4.04335e-05,5.10347e-05,6.44153e-05,8.13041e-05, 
    0.000102621,0.000129527,0.000163487,0.000206351,0.000260454, 0.000328741,0.000414933,0.000523722,0.000661035,0.00083435, 
    0.00105311,0.00132922,0.00167772,0.00211759,0.0026728, 0.00337357,0.00425808,0.00537449,0.00678361,0.00856218, 
    0.0108071,0.0136405,0.0172169,0.0217309,0.0274285, 0.0346199,0.0436968,0.0551535,0.069614,0.0878659, 
    0.110903,0.13998,0.176681,0.223005,0.281474, 0.355273,0.44842,0.56599,0.714385,0.901688, 
    1.1381,1.43649,1.81312,2.2885,2.88851, 3.64584,4.60173,5.80825,7.33109,9.2532, 
    11.6793,14.7414,18.6064,23.4848,29.6422, 37.414,47.2234,59.6048,75.2323,94.9572, 
    119.854,151.278,190.941,241.003,304.191, 383.946,484.611,611.67,772.041,974.46, 
    1229.95,1552.43,1959.45,2473.2,3121.63, 3940.09,4973.12,6277.01,7922.76,10000
  };
  static const double dEdx[] = {
    436.22,442.344,438.318,424.359,403.738, 379.328,349.232,316.072,281.948,249.32, 
    218.403,189.912,164.166,141.342,121.098, 103.311,87.184,73.3553,61.4308,51.1857, 
    42.5767,35.3589,29.3237,24.2963,20.1227, 16.6686,13.818,11.4713,9.54423,7.96558, 
    6.67559,5.62437,4.77042,4.07933,3.52265, 3.07688,2.72119,2.44033,2.22241,2.05669, 
    1.93419,1.84736,1.78978,1.75596,1.74115, 1.74126,1.75282,1.77291,1.79914,1.82958, 
    1.86386,1.90012,1.93715,1.97437,2.01139, 2.04793,2.0839,2.1192,2.15387,2.18801, 
    2.2218,2.25552,2.28953,2.32434,2.36061, 2.39917,2.44105,2.48751,2.54017,2.60109, 
    2.67287,2.75883,2.86316,2.99115,3.14952, 3.34678,3.59369,3.90388,4.29458,4.78763, 
    5.41056,6.19822,7.19473,8.45547,10.0507, 12.0689,14.6221,17.8513,21.9348,27.0978
  };
  fT    = T;
  fdEdx = dEdx;
  fRho  = 1.34;
}

//......................................................................

double dEdxTable::operator()(double T, double m) const
{
  if (fT==0) return 0.0;
  return fRho*this->Interp(T, m, fT, fdEdx, fSz);
}

////////////////////////////////////////////////////////////////////////
