///
/// \file    testTrackBasis.cc
/// \brief   Unit tests for TrackBasis class
/// \author  messier@indiana.edu
/// \version $Id:$
///
#include <iostream>
#include <fstream>
#include <cmath>
#include "TRandom.h"
#include "TVector3.h"
#include "BreakPointFitter/func/TrackBasis.h"

#define FAILED \
  { std::cerr << "Failed at line " << __LINE__ << std::endl; return false; }

//......................................................................

static TRandom* gsRandom = 0;

//......................................................................

void makeTrack(double*              x0,
	       double*              y0,
	       double*              z0,
	       TVector3&            axis,
	       std::vector<double>& x,
	       std::vector<double>& y,
	       std::vector<double>& z,
	       std::vector<double>& dx,
	       std::vector<double>& dy,
	       std::vector<double>& dz,
	       bool useinputs=false)
{
  unsigned int i;
  double ds = 6.0;
  double costh, sinth, phi;
  
  if (!useinputs) {
    *x0 = gsRandom->Uniform(-100,100);
    *y0 = gsRandom->Uniform(-100,100);
    *z0 = gsRandom->Uniform(-100,100);
    costh = gsRandom->Uniform(-1,1);
    phi   = gsRandom->Uniform(-M_PI, M_PI);
    sinth = sqrt(1.0-costh*costh);
    
    axis = TVector3(sinth*cos(phi),sinth*sin(phi),costh);
  }
  
  for (i=0; i<z.size(); ++i) {
    if (i%2==0) {
      dx[i] =  4.0/sqrt(12.0);
      dy[i] = 20.0/sqrt(12.0);
    }
    else {
      dx[i] = 20.0/sqrt(12.0);
      dy[i] =  4.0/sqrt(12.0);
    }
    dz[i] = 6.0/sqrt(12.0);

    x[i] = *x0 + ds*i*axis.X() + gRandom->Gaus(0.0, dx[i]);
    y[i] = *y0 + ds*i*axis.Y() + gRandom->Gaus(0.0, dy[i]);
    z[i] = *z0 + ds*i*axis.Z() + gRandom->Gaus(0.0, dz[i]);
  }
}

//......................................................................

bool checkBasis(TVector3& x,
		TVector3& y,
		TVector3& z)
{
  double eps = 1.0E-6;
  double tol = 1.0-eps;
  //
  // Check all are unit vectors
  //
  if (x.Mag2()<tol) FAILED;
  if (y.Mag2()<tol) FAILED;
  if (z.Mag2()<tol) FAILED;
  
  //
  // Check orthogonality
  //
  if (x.Dot(y)>eps) FAILED;
  if (x.Dot(z)>eps) FAILED;
  if (y.Dot(z)>eps) FAILED;

  // Check right handed
  if (x.Cross(y).Dot(z)<eps) FAILED;
  if (y.Cross(z).Dot(x)<eps) FAILED;
  if (z.Cross(x).Dot(y)<eps) FAILED;
  
  return true;
}

//......................................................................

bool testFindAxis() 
{
  static const double tol = 0.90;
  static const unsigned int npts = 100;
  double x0, y0, z0;
  std::vector<double> x(npts);
  std::vector<double> y(npts);
  std::vector<double> z(npts);
  std::vector<double> dx(npts);
  std::vector<double> dy(npts);
  std::vector<double> dz(npts);

  unsigned int itrial;
  for (itrial=0; itrial<1000; ++itrial) {
    TVector3 axisin;
    makeTrack(&x0, &y0, &z0, axisin, x, y, z, dx, dy, dz);

    TVector3 axisout;
    bpfit::TrackBasis::FindAxis(x, y, z, dx, dy, dz, axisout);
    
    double dot = fabs(axisin.Dot(axisout));
    if (dot<tol) {
      std::cerr << "axisin doesn't match axisout! " << dot << std::endl;
      std::cerr << "in:  " << 
	axisin.X() << "\t" <<
	axisin.Y() << "\t" <<
	axisin.Z() << std::endl;
      std::cerr << "out: " << 
	axisout.X() << "\t" <<
	axisout.Y() << "\t" <<
	axisout.Z() << std::endl;
      FAILED;
    }
  }
  return true;
}

//......................................................................

bool testMakeBasis()
{
  static double tol = 0.90;

  bpfit::TrackBasis sys;
  
  static const int npts = 50;
  std::vector<double> x(npts);
  std::vector<double> y(npts);
  std::vector<double> z(npts);
  std::vector<double> dx(npts);
  std::vector<double> dy(npts);
  std::vector<double> dz(npts);

  double x0 = 0;
  double y0 = 0;
  double z0 = 0;
  double dot;

  //
  // Put tracks along the detector x/y/z axes. In those cases I know
  // what the transformed frames should look like.
  //
  TVector3 zaxis(0,0,1);
  makeTrack(&x0, &y0, &z0, zaxis, x, y, z, dx, dy, dz, true);
  sys.MakeBasis(x, y, z, dx, dy, dz, x0, y0, z0);
  if (!checkBasis(sys.fX1, sys.fY1, sys.fZ1)) FAILED;

  dot = sys.fX1.Dot(TVector3(1,0,0)); if (dot<tol) FAILED;
  dot = sys.fY1.Dot(TVector3(0,1,0)); if (dot<tol) FAILED;
  dot = sys.fZ1.Dot(TVector3(0,0,1)); if (dot<tol) FAILED;

  if (sys.fU [0][0]<tol) FAILED;
  if (sys.fU [1][1]<tol) FAILED;
  if (sys.fU [1][1]<tol) FAILED;
  if (sys.fUT[0][0]<tol) FAILED;
  if (sys.fUT[1][1]<tol) FAILED;
  if (sys.fUT[1][1]<tol) FAILED;

  TVector3 xaxis(1,0,0);
  makeTrack(&x0, &y0, &z0, xaxis, x, y, z, dx, dy, dz, true);
  sys.MakeBasis(x, y, z, dx, dy, dz, x0, y0, z0);
  if (!checkBasis(sys.fX1, sys.fY1, sys.fZ1)) FAILED;

  dot = sys.fX1.Dot(TVector3(0,0,1));  if (fabs(dot)<tol) FAILED;  
  dot = sys.fY1.Dot(TVector3(0,1,0));  if (fabs(dot)<tol) FAILED;
  dot = sys.fZ1.Dot(TVector3(-1,0,0)); if (fabs(dot)<tol) FAILED;
  
  TVector3 yaxis(0,1,0);
  makeTrack(&x0, &y0, &z0, yaxis, x, y, z, dx, dy, dz, true);
  sys.MakeBasis(x, y, z, dx, dy, dz, x0, y0, z0);
  if (!checkBasis(sys.fX1, sys.fY1, sys.fZ1)) FAILED;
  
  dot = sys.fX1.Dot(TVector3(-1,0,0)); if (fabs(dot)<tol) FAILED;
  dot = sys.fY1.Dot(TVector3(0,0,1));  if (fabs(dot)<tol) FAILED;
  dot = sys.fZ1.Dot(TVector3(0,1,0));  if (fabs(dot)<tol) FAILED;
  
  return true;
}

//......................................................................

bool testDetToTrk() 
{
  double x0 = 10.0;
  double y0 = 20.0;
  double z0 = 30.0;
  double th  = 60.0*3.14/180.0;
  double phi = 30.0*3.14/180.0;
  TVector3 dir(sin(th)*cos(phi), sin(th)*sin(phi), cos(th));

  static const int npts = 10;
  std::vector<double> x(npts), dx(npts);
  std::vector<double> y(npts), dy(npts);
  std::vector<double> z(npts), dz(npts);
  makeTrack(&x0, &y0, &z0, dir, x, y, z, dx, dy, dz, true);

  bpfit::TrackBasis sys;
  sys.SetOrigin(x0, y0, z0);
  sys.MakeBasis(x, y, z, dx, dy, dz, x0, y0, z0);
  
  //
  // Check that transforming into and out of the track system gives us
  // back the same point
  //
  static const double tol = 1.0E-6;
  unsigned int i;
  double x1,   y1,   z1;
  double dx1,  dy1,  dz1;
  double x11,  y11,  z11;
  double dx11, dy11, dz11;
  for (i=0; i<npts; ++i) {
    sys.DetToTrk(x[i],  y[i],  z[i],
		 dx[i], dy[i], dz[i],
		 &x1,   &y1,   &z1,
		 &dx1,  &dy1,  &dz1);

    sys.TrkToDet(x1,    y1,    z1,
		 dx1,   dy1,   dz1,
		 &x11,  &y11,  &z11,
		 &dx11, &dy11, &dz11);

    if (fabs(x[i]-x11)>tol) FAILED;
    if (fabs(y[i]-y11)>tol) FAILED;
    if (fabs(z[i]-z11)>tol) FAILED;
  }
  return true;
}

//......................................................................

void makeVisualization() 
{
  static const int npts = 50;
  unsigned int i;
  double x0, y0, z0;
  TVector3 dir;
  std::vector<double>  x(npts),  y(npts),  z(npts);
  std::vector<double> dx(npts), dy(npts), dz(npts);
  makeTrack(&x0, &y0, &z0, dir, x, y, z, dx, dy, dz);

  bpfit::TrackBasis sys;
  sys.SetOrigin(x0, y0, z0);
  sys.MakeBasis(x, y, z, dx, dy, dz, x0, y0, z0);
  
  std::vector<double>  x1(npts),  y1(npts),  z1(npts);
  std::vector<double> dx1(npts), dy1(npts), dz1(npts);
  for (i=0; i<npts; ++i) {
    sys.DetToTrk(x[i],   y[i],   z[i],   dx[i],   dy[i],   dz[i],
		 &x1[i], &y1[i], &z1[i], &dx1[i], &dy1[i], &dz1[i]);
  }

  std::ofstream f;
  f.open("track.C");
  f << "{\n"
    << "TPolyMarker3D* p1 = new TPolyMarker3D(" << npts << ",20);\n"
    << "TPolyMarker3D* p2 = new TPolyMarker3D(" << npts << ",24);\n"
    << "p1->SetMarkerColor(kBlue);\n" 
    << "p2->SetMarkerColor(kRed);\n" 
    << "p1->SetMarkerSize(3.0);\n" 
    << "p2->SetMarkerSize(3.0);\n"
    << "TMarker3DBox* m1;\n"
    << "TMarker3DBox* m2;\n";
  
  for (i=0; i<npts; ++i) {
    f << "p1->SetPoint(" 
      << i << "," << x[i] << "," << y[i] << "," << z[i] 
      << ");\n";
    
    f << "p2->SetPoint(" 
      << i << "," << x1[i] << "," << y1[i] << "," << z1[i] 
      << ");\n";

    f << "m1 = new TMarker3DBox(" 
      << x[i]  << "," << y[i]  << "," << z[i]  << ","
      << dx[i] << "," << dy[i] << "," << 0     << ","
      << "0,0); m1->Draw();\n";
    
    f << "m1 = new TMarker3DBox(" 
      << x[i]  << "," << y[i]  << "," << z[i]  << ","
      << dx[i] << "," << 0     << "," << dz[i] << ","
      << "0,0); m1->Draw();\n";

    f << "m1 = new TMarker3DBox(" 
      << x[i]  << "," << y[i]  << "," << z[i]  << ","
      << 0     << "," << dy[i] << "," << dz[i] << ","
      << "0,0); m1->Draw();\n";

    f << "m2 = new TMarker3DBox(" 
      << x1[i]  << "," << y1[i]  << "," << z1[i]  << ","
      << dx1[i] << "," << dy1[i] << "," << 0     << ","
      << "0,0); m2->Draw();\n";
    
    f << "m2 = new TMarker3DBox(" 
      << x1[i]  << "," << y1[i] << "," << z1[i]  << ","
      << dx1[i] << "," << 0     << "," << dz1[i] << ","
      << "0,0); m2->Draw();\n";

    f << "m2 = new TMarker3DBox(" 
      << x1[i] << "," << y1[i]  << "," << z1[i]  << ","
      << 0     << "," << dy1[i] << "," << dz1[i] << ","
      << "0,0); m2->Draw();\n";
    
  }
  f << "p1->Draw();\n"
    << "p2->Draw();\n"
    << "}\n";
  f.close();
}

//......................................................................

int main(int argc, char** argv) 
{
  int seed = 11121970;
  if (argc>1) seed = atoi(argv[1]);
  gsRandom = new TRandom(seed);

  bool ok = true;

  std::cout << "** Starting TrackBasis tests ** " << std::endl;

  ok |= testFindAxis();
  std::cout << "testFindAxis: " << (ok? "OK":"FAILED") << std::endl;

  ok |= testMakeBasis();
  std::cout << "testMakeBasis: " << (ok? "OK":"FAILED") << std::endl;
  
  ok |= testDetToTrk();
  std::cout << "testDetToTrk: " << (ok? "OK":"FAILED") << std::endl;
  
  makeVisualization();
  
  if (!ok) return 1;
  return 0;
}

////////////////////////////////////////////////////////////////////////
