////////////////////////////////////////////////////////////////////////
/// \brief   This module creates Common Analysis Files.
/// \author  rocco@physics.umn.edu
/// \date    July 2012
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <algorithm>
#include <cmath>
#include <fenv.h>
#include <iomanip>
#include <iostream>
#include <iostream>
#include <map>
#include <string>
#include <time.h>
#include <vector>
#include <string>
#include <sstream>

// ROOT includes
#include "TFile.h"
#include "TH1D.h"
#include "TSystem.h"
#include "TTree.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/FileBlock.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindMany.h"
#include "art/Framework/Core/FindOneP.h"
#include "art/Framework/Core/FindOne.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "cetlib/exception.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <IFDH_service.h>

// NOvASoft includes
// Geo/Cal/Util
#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "GeometryObjects/Geo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "Metadata/MetadataManager.h"
#include "Utilities/func/MathUtil.h"
#include "RawData/RawTrigger.h"
#include "RunHistory/service/RunHistoryService.h"
#include "ChannelInfo/BadChanList.h"

// Sim/BackTracker
#include "SimulationBase/GTruth.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/GENIEReweightTable.h"
#include "MCCheater/BackTracker.h"

// RecoBase
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Energy.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/Prong.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Shower.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"

// Other Reco
#include "CosRej/CosRejObj.h"
#include "CosRej/NueCosRej.h"
#include "RecoJMShower/EID.h"
#include "RecoJMShower/JMShower.h"
#include "RecVarPID/RVP.h"
#include "ReMId/ReMId.h"
#include "LEM/PIDDetails.h"
#include "ShowerLID/EventLID.h"
#include "ShowerLID/ShowerLID.h"
#include "ShowerLID/ShowerPID.h"
#include "ShowerLID/DedxDistribution.h"
#include "MichelEFilter/MichelECluster.h"
#include "MichelEFilter/SliceMEFCluster.h"
#include "MuonRemove/art/MRCCParent.h"
#include "NueSandbox/NueSandObj.h"
#include "NumuSandbox/NumuSandObj.h"
#include "NumuEnergy/NumuE.h"
#include "QEEventFinder/QePId.h"
#include "Preselection/NuePreselObj.h"
#include "Preselection/Veto.h"
#include "BreakPointFitter/BPFEnergy.h"

// SummData
#include "SummaryData/CosmicExposure.h"
#include "SummaryData/EventQuality.h"
#include "SummaryData/POTSum.h"
#include "SummaryData/SpillData.h"

// OscLib - for woscdumb
#include "OscLib/func/OscCalculatorDumb.h"

// StandardRecord
#include "StandardRecord/StandardRecord.h"
// CAFMaker support
#include "CAFMaker/Blinding.h"



// note it is convention to NOT declare ANY "using namespace xxx" here
// especially "using namespace std"
namespace caf
{
  /// Module to create Common Analysis Files from ART files
  class CAFMaker : public art::EDProducer {
  public:
    explicit CAFMaker(fhicl::ParameterSet const& pset); // Required! explicit tag tells the compiler this is not a copy constructor
    virtual ~CAFMaker();
    
    // Optional, read/write access to event
    void produce(art::Event& evt) noexcept;
    
    void respondToOpenInputFile(const art::FileBlock& fb);
    
    // Optional if you want to be able to configure from event display, for example
    void reconfigure(const fhicl::ParameterSet& pset);
    
    // Optional use if you have histograms, ntuples, etc you want around for every event
    void beginJob();
    void endJob();
    virtual void beginRun(art::Run &r);
    virtual void endSubRun(art::SubRun& sr);
    
    
  private:
    std::string fCafFilename;
    std::string fDataTier;
    std::string fGeneratorLabel;
    std::string fReweightLabel;
    std::string fExposureLabel;
    std::string fSpillQualLabel;
    std::string fClusterLabel;
    std::string fHitsLabel;
    std::string fKalmanTrackLabel;
    std::string fDiscreteTrackLabel;
    std::string fDiscreteTrack2dLabel;
    std::string fCosmicTrackLabel;
    std::string fFuzzyKLabel;
    std::string fFuzzyK3dAssnLabel;
    std::string fFuzzyK2dAssnLabel;
    std::string fSliceMEFLabel;
    std::string fElasticArmsLabel;
    std::string fHoughVertexLabel;
    std::string fJMShowerLabel;
    std::string fNumuEnergyLabel;
    std::string fNuePreselLabel;
    std::string fVetoLabel;
    std::string fRemidLabel;
    std::string fQepidLabel;
    std::string fRawDataLabel;
    std::string fNueSelLabel;
    std::string fLidLabel;
    std::string fSPidLabel;
    std::string fLemLabel;
    std::string fRvpLabel;
    std::string fCosRejLabel;
    std::string fNueCosRejLabel;
    std::string fKalmanMichelELabel;
    std::string fCosmicMichelELabel;
    std::string fDiscreteMichelELabel;
    std::string fNueSandboxLabel;
    std::string fNumuSandboxLabel;
    std::string fNumiBeamLabel;
    std::string fMrccParentLabel;
    std::string fFileExtension;
    std::string fBPFEnergyLabel;
    
    bool fGeneratorInfo;
    bool fFullTruthInfo;
    bool fEnableBlindness;
    bool fIsRealData;
    double fTotalPOT;
    double fTotalEvents;
    double fTotalLivetime;
    int fCycle;
    
    TFile*          fFile;
    TTree*          fRecTree;
    TH1D*           hPOT;
    TH1D*           hEvents;
    TH1D*           hLivetime;
    art::ServiceHandle<cheat::BackTracker>           bt;
    art::ServiceHandle<calib::Calibrator>            cal;
    art::ServiceHandle<geo::Geometry>                geom;
    art::ServiceHandle<geo::LiveGeometry>            livegeom;
    art::ServiceHandle<nova::dbi::RunHistoryService> rh;
    art::ServiceHandle<chaninfo::BadChanList>        bc;
    
    novadaq::cnv::DetId fDetID;  ///< Detector ID in nova daq convention namespace typedef
    Det_t               fDet;    ///< Detector ID in caf namespace typedef
    
    //Creating separate friend tree and branches for spill info.
    //This tree will store one entry per event, even empty ones
    //so that POT counting can be done right.
    TTree*              fSpillTree;

    void InitializeOutfile();
    
    void AddMetadataToFile(std::map<std::string, std::string> metadata);
    
    void AddCosmicTruthToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                             std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRCosmic>* vec, const art::Event& evt);
    void AddGenInfoToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                         std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRGenerator>* vec, const art::Event& evt);
    
    void CalculateVisEFraction(std::vector<SRGenerator>* gen);
    
    void AddMCTruthToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                         std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRNeutrino>* vec, const art::Event& evt);
    
    void AddParticleToVec(const sim::Particle &part, std::vector<cheat::TrackIDE> &allTracks,
                          std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRTrueParticle>* vec);
    
    void AddTrueMichelEToVec(const sim::Particle &michel, const sim::Particle &motherMuon,
                             std::vector<cheat::TrackIDE> &allTracks, std::vector<SRTrueMichelE>* vec);
    
    void FindAndAddMichels(std::vector<const sim::Particle*> particles,std::vector<cheat::TrackIDE> &allTracks,
                           std::vector<SRTrueMichelE>* michelVec);

  std::vector<int>
  getFaveIds(const std::vector<cheat::NeutrinoWithIndex>& allNus,
             const std::vector<std::vector<cheat::NeutrinoEffPur>>& sliceTruthTable,
             std::function<double(const cheat::NeutrinoEffPur&)> sliceMetric,
             std::function<double(const cheat::NeutrinoEffPur&)> nuMetric);

    void drawSliceTruthTable(const std::vector< std::vector<cheat::NeutrinoEffPur> > &sliceTruthTable,
                             const std::vector<int> &faveIds);
    void drawSliceTruthTable(const std::vector< std::vector<cheat::NeutrinoEffPur> > &sliceTruthTable,
                             const std::vector<int> &faveIds, const std::vector<int> &otherFaveIds);

    /// \brief Retrieve an object from an association, with error handling
    ///
    /// This can go wrong in two ways: either the FindManyP itself is invalid,
    /// or the result for the requested index is empty. In most cases these
    /// have the same response, so conflating them here saves redundancy
    /// elsewhere.
    ///
    /// \param      fm  The FindManyP object describing the association
    /// \param      idx Which element of the FindManyP to look it
    /// \param[out] ret The product retrieved
    /// \return          Whether \a ret was filled
    template<class T> bool GetAssociatedProduct(const art::FindManyP<T>& fm,
                                                int idx, T& ret);

    void FillSliceVars(const rb::Cluster& slice, caf::SRSlice& srslice, bool allowEmpty = false) const;

    void AddTrkMEClusterToVec(const me::MichelECluster& michel, std::vector<SRTrkME>* vec, size_t trackId);
    void AddSlcMEClusterToVec(const me::SliceMEFCluster& michel, std::vector<SRSlcME>* vec);
    
    void FillTrackVars(const rb::Track& trk, SRTrack& srTrk);
    
    /// Acts on any object that has a a superset of the same fields as the old SRProng
    template<class T> void FillProngVars(const rb::Prong &trk, T& srPrng);
    
    
    static bool EssentiallyEqual(double a, double b, double precision = 0.0001)
    { return a <= (b+precision) && a >= (b-precision); }
    
    static bool sortTrackLength(const SRTrack& a, const SRTrack& b) {return a.len>b.len; }
    
    static bool sortRBTrackLength(const art::Ptr<rb::Track>& a, const art::Ptr<rb::Track>& b) {return a->TotalLength()>b->TotalLength(); }
    
    static bool sortRBProngLength(const art::Ptr<rb::Prong>& a, const art::Ptr<rb::Prong>& b) {return a->TotalLength()>b->TotalLength(); }
    
    static bool sortShowerEnergy(const SRJMShower& a, const SRJMShower& b) {return a.shwE>b.shwE; }
    
    static bool sortElasticProng(const SRElastic& a, const SRElastic& b) {return a.png3d.size()>b.png3d.size(); }
    
    static bool sortMuIdLength(const SRMuId& a, const SRMuId& b) {return a.len>b.len; }
    
    static bool sortNuWithIdx(const cheat::NeutrinoWithIndex& a, const cheat::NeutrinoWithIndex& b)
    {return a.truthColIndex<b.truthColIndex; }
    
    
    double SimpleOscProb(const simb::MCFlux& flux, const simb::MCNeutrino& nu) const;
    
    std::vector<caf::SRGenieWeights> GenieReweightTable(const rwgt::GENIEReweightTable& table) const;

    void FillSpillVars(art::Handle<sumdata::SpillData> spillPot,
                       art::Handle<sumdata::EventQuality> spillQual,
                       art::Handle<std::vector<rawdata::RawTrigger>> trigs,
                       SRSpill& spill,
                       art::Event& evt) const;

    void FillMuIdVars(const remid::ReMId& remid,
                      const rb::Track& track,
                      caf::SRMuId& muid) const;

    void FillJMShowerVars(const jmshower::JMShower& shw,
                          caf::SRJMShower& srshw) const;

    void FillJMEIDVars(const jmshower::EID& eid, caf::SRJMEid& sreid) const;

    void FillNueSandEIDVars(const jmshower::EID& eid,
                            caf::SRNueSandbox& sand) const;

    void FillShowerVars(const rb::Shower& shw,
                        const slid::ShowerLID& slid,
                        caf::SRShower& srshw) const;

    void FillSlidVars(const slid::ShowerLID& slid, caf::SRSLid& srslid) const;

    void FillSpidVars(const slid::ShowerPID& spid, caf::SRSPid& srspid) const;

    void FillLIDEnergyVars(const slid::ShowerLID& slid,
                           caf::SRSLidEnergy& en) const;

    void FillLEMVars(const lem::PIDDetails& lem, caf::SRLem& srlem) const;

    void FillRVPVars(const rvp::RVP& rvp, caf::SRRvp& srrvp) const;

    void FillCosRejVars(const cosrej::CosRejObj& cosrej,
                        caf::SRCosRej& srcosrej,
                        caf::SRContain& srcontain) const;

    void FillNueCosRejVars(const cosrej::NueCosRej& nuecosrej,
                           caf::SRNueCosRej& srnuecosrej) const;

    void FillQepidVars(const qeef::QePId& qepid, caf::SRQepid& rsqepid) const;

    void FillNuePreselVars(const presel::NuePreselObj& nuepre, 
			   caf::SRNuePresel& srnuepre) const;

    void FillVetoVars(const presel::Veto& veto, caf::SRVeto& srveto) const;

    void FillNumuSimpVars(const numue::NumuE& e, caf::SRNumuEnergy& sre) const;

    void FillBPFEnergyVars(const bpfit::BPFEnergy& e, caf::SRBPFEnergy& sre) const;

    void FillNueSandVars(const nuesand::NueSandObj& sand,
                         caf::SRNueSandbox& srsand) const;

    void FillNumuSandVars(const numusand::NumuSandObj& sand,
                          caf::SRNumuSandbox& srsand,
                          caf::SRContain& srcontain) const;

    void CopyRemidVars(const caf::SRMuId& muid, caf::SRRemid& remid);
  };
  
  //.......................................................................
  
  //.......................................................................
  CAFMaker::CAFMaker(fhicl::ParameterSet const& pset):
  fIsRealData(false),
  fFile(0),
  bt(),
  cal(),
  geom()
  {
    reconfigure(pset);

    // Normally CAFMaker is run without an output ART stream, so these go
    // nowhere, but can be occasionally useful for filtering as part of an ART
    // job.
    produces<std::vector<caf::StandardRecord>>();
    produces<caf::SRSpill>();
  }
  
  //......................................................................
  CAFMaker::~CAFMaker()
  {
  }
  
  //......................................................................
  void CAFMaker::reconfigure(const fhicl::ParameterSet& pset)
  {
    // Keep everything lined up for clarity (stylistic)
    fCafFilename          =   pset.get< std::string    >("CAFFilename");
    fDataTier             =   pset.get< std::string    >("DataTier");
    fGeneratorLabel       =   pset.get< std::string    >("GeneratorInput");
    fReweightLabel        =   pset.get< std::string    >("ReweightInput");
    fExposureLabel        =   pset.get< std::string    >("ExposureInput");
    fSpillQualLabel       =   pset.get< std::string    >("SpillQualInput");
    fClusterLabel         =   pset.get< std::string    >("ClusterInput");
    fHitsLabel            =   pset.get< std::string    >("HitInput");
    fKalmanTrackLabel     =   pset.get< std::string    >("KalmanInput");
    fDiscreteTrackLabel   =   pset.get< std::string    >("DiscreteInput");
    fDiscreteTrack2dLabel =   pset.get< std::string    >("Discrete2dInput");
    fCosmicTrackLabel     =   pset.get< std::string    >("CosmicInput");
    fFuzzyKLabel          =   pset.get< std::string    >("FuzzyInput");
    fFuzzyK3dAssnLabel    =   pset.get< std::string    >("Fuzzy3DAssnInput");
    fFuzzyK2dAssnLabel    =   pset.get< std::string    >("Fuzzy2DAssnInput");
    fSliceMEFLabel        =   pset.get< std::string    >("SliceMEFInput");
    fElasticArmsLabel     =   pset.get< std::string    >("ElasticInput");
    fHoughVertexLabel     =   pset.get< std::string    >("HoughVertexInput");
    fJMShowerLabel        =   pset.get< std::string    >("JmShowerInput");
    fNumuEnergyLabel      =   pset.get< std::string    >("NumuEnergyInput");
    fNuePreselLabel       =   pset.get< std::string    >("NuePreselInput");
    fVetoLabel            =   pset.get< std::string    >("VetoInput");
    fRemidLabel           =   pset.get< std::string    >("RemidInput");
    fQepidLabel           =   pset.get< std::string    >("QepidInput");
    fRawDataLabel         =   pset.get< std::string    >("RawDataInput");
    fNueSelLabel          =   pset.get< std::string    >("NueSelInput");
    fLidLabel             =   pset.get< std::string    >("LIDInput");
    fSPidLabel            =   pset.get< std::string    >("SPIDInput");
    fLemLabel             =   pset.get< std::string    >("LemInput");
    fRvpLabel             =   pset.get< std::string    >("RvpInput");
    fCosRejLabel          =   pset.get< std::string    >("CosRejInput");
    fNueCosRejLabel       =   pset.get< std::string    >("NueCosRejInput");
    fKalmanMichelELabel   =   pset.get< std::string    >("KalmanMichelEInput");
    fCosmicMichelELabel   =   pset.get< std::string    >("CosmicMichelEInput");
    fDiscreteMichelELabel =   pset.get< std::string    >("DiscreteMichelEInput");
    fNueSandboxLabel      =   pset.get< std::string    >("NueSandboxInput");
    fNumuSandboxLabel     =   pset.get< std::string    >("NumuSandboxInput");
    fNumiBeamLabel        =   pset.get< std::string    >("NuMIBeamInput");
    fGeneratorInfo        =   pset.get< bool           >("GeneratorInfo");
    fFullTruthInfo        =   pset.get< bool           >("FullTruthInfo");
    fEnableBlindness      =   pset.get< bool           >("EnableBlindness");
    fMrccParentLabel      =   pset.get<std::string     >("MRCCParentInput");
    fFileExtension        =   pset.get<std::string     >("FileExtension");
    fBPFEnergyLabel       =   pset.get<std::string     >("BPFEnergyInput");
  }
  
  //......................................................................
  void CAFMaker::respondToOpenInputFile(const art::FileBlock& fb)
  {
    if(!fFile)
    {
      // Filename wasn't set in the FCL, and this is the first file we've seen
      
      fCafFilename = basename(fb.fileName().c_str());
      const size_t dotpos = fCafFilename.find('.');
      assert(dotpos != std::string::npos); // Must have a dot, surely?
      fCafFilename.resize(dotpos);
      fCafFilename += fFileExtension;
      
      InitializeOutfile();
    }
  }

  //......................................................................
  void CAFMaker::beginJob()
  {
    if(!fCafFilename.empty()) InitializeOutfile();
  }
  
  //......................................................................
  void CAFMaker::beginRun(art::Run &r)
  {
    fDetID = geom->DetId();
    fDet   = (Det_t) fDetID;
  }
  
  //......................................................................
  void CAFMaker::InitializeOutfile()
  {
    assert(!fFile);
    assert(!fCafFilename.empty());
    
    mf::LogInfo("CAFMaker") << "Output filename is " << fCafFilename;
    
    fFile = new TFile(fCafFilename.c_str(), "RECREATE");
    
    hPOT = new TH1D("TotalPOT", "TotalPOT;; POT", 1, 0, 1);
    hEvents = new TH1D("TotalEvents", "TotalEvents;; Events", 1, 0, 1);
    hLivetime = new TH1D("TotalLivetime", "TotalLiveTime;;Seconds", 1, 0, 1);
    
    fRecTree = new TTree("recTree", "records");
    
    // Tell the tree it's expecting StandardRecord objects
    StandardRecord* rec = 0;
    fRecTree->Branch("rec", "caf::StandardRecord", &rec);
    
    // Create tree to store spill info for POT counting
    fSpillTree = new TTree("spillTree", "spills");

    caf::SRSpill* dummy = 0;
    fSpillTree->Branch("spill", "caf::SRSpill", &dummy);

    // Make the spillTree a friend of the fRecTree
    fRecTree->AddFriend("spillTree");
    
    
    fTotalPOT      = 0;
    fTotalEvents   = 0;
    fTotalLivetime = 0;
    fCycle         = -5;
    
    // Global information about the processing details:
    
    std::map<TString, TString> envmap;
    std::string envstr;
    // Environ comes from unistd.h
    for(char** penv = environ; *penv; ++penv)
    {
      const std::string pair = *penv;
      envstr += pair;
      envstr += "\n";
      const size_t split = pair.find("=");
      if(split == std::string::npos) continue; // Huh?
      const std::string key = pair.substr(0, split);
      const std::string value = pair.substr(split+1);
      envmap[key] = value;
    }
    
    // Get the command-line we were invoked with. What I'd really like is just
    // the fcl script and list of input filenames in a less hacky fashion. I'm
    // not sure that's actually possible in ART. TODO: ask the artists.
    FILE* cmdline = fopen("/proc/self/cmdline", "rb");
    char* arg = 0;
    size_t size = 0;
    std::string cmd;
    while(getdelim(&arg, &size, 0, cmdline) != -1)
    {
      cmd += arg;
      cmd += " ";
    }
    free(arg);
    fclose(cmdline);
    
    fFile->mkdir("env")->cd();
    
    TObjString(envmap["USER"]).Write("user");
    TObjString(envmap["HOSTNAME"]).Write("hostname");
    TObjString(envmap["PWD"]).Write("pwd");
    TObjString(envmap["SRT_PUBLIC_CONTEXT"]).Write("publiccontext");
    TObjString(envmap["SRT_PRIVATE_CONTEXT"]).Write("privatecontext");
    // default constructor is "now"
    TObjString(TTimeStamp().AsString()).Write("date");
    TObjString(cmd.c_str()).Write("cmd");
    TObjString(fCafFilename.c_str()).Write("output");
    TObjString(envstr.c_str()).Write("env");
  }
  
  void CAFMaker::AddMetadataToFile(std::map<std::string, std::string> metadata)
  {
    assert(fFile && "CAFMaker: Trying to add metadata to an uninitialized file");
    
    fFile->mkdir("metadata")->cd();

    for(auto const& pair:metadata){
      TObjString(pair.second.c_str()).Write(pair.first.c_str());
    }
  }
  
  
  
  //......................    //......................
  void CAFMaker::AddParticleToVec(const sim::Particle &part, std::vector<cheat::TrackIDE> &allTracks,
                                  std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRTrueParticle>* vec)
  {
    int trackId = part.TrackId();
    //find visible energy
    double visE = 0;
    for (std::vector<cheat::TrackIDE>::iterator  it = allTracks.begin();
         it != allTracks.end(); it++)
    {
      if(trackId == it->trackID)
      {
        visE += it->energy;
        break;
      }
      
    }
    //find visible energy in slice
    double inSliceVisE = 0;
    for (std::vector<cheat::TrackIDE>::iterator  it = sliceTracks.begin();
         it != sliceTracks.end(); it++)
    {
      if(trackId == it->trackID)
      {
        inSliceVisE += it->energy;
        break;
      }
    }
    //make a new SRTrueParticle and get an accessor
    vec->push_back(SRTrueParticle());
    SRTrueParticle& srPart = vec->back();
    //Set fields
    srPart.pdg = part.PdgCode();
    srPart.visE = visE;
    srPart.visEinslc = inSliceVisE;
    srPart.p = part.Momentum();
    
  }
  
  //......................    //......................
  void CAFMaker::AddTrueMichelEToVec(const sim::Particle &michel, const sim::Particle &motherMuon,
                                     std::vector<cheat::TrackIDE> &allTracks, std::vector<SRTrueMichelE>* vec)
  {
    
    vec->push_back(SRTrueMichelE());
    SRTrueMichelE& srMichel = vec->back();
    
    // find the visible energy from the FLS hits
    float visE = 0;
    for (std::vector<cheat::TrackIDE>::iterator  it = allTracks.begin();
         it != allTracks.end(); it++)
    {
      if(michel.TrackId() == it->trackID)
      {
        visE = it->energy;
        break;
      }
    }
    
    srMichel.E       = michel.E();
    srMichel.visE    = visE;
    srMichel.time    = michel.T();
    srMichel.mustop  = motherMuon.EndPosition();
    srMichel.p       = michel.Momentum();
    
  }
  
  //......................................................................
  
  void CAFMaker::AddMCTruthToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                                 std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRNeutrino>* vec, const art::Event& evt)
  {
    //get information about truNu
    
    vec->push_back(SRNeutrino());
    SRNeutrino& srTruth = vec->back();
    
    const art::Ptr<simb::MCTruth> &tru = effPur->neutrinoInt;
    
    srTruth.pdg          = tru->GetNeutrino().Nu().PdgCode();
    srTruth.hitnuc       = tru->GetNeutrino().HitNuc();
    
    srTruth.mode         = tru->GetNeutrino().Mode();  // to avoid depending on simb
    srTruth.iscc         = (tru->GetNeutrino().CCNC() == simb::kCC);
    srTruth.inttype      = tru->GetNeutrino().InteractionType();
    
    srTruth.q2           = tru->GetNeutrino().QSqr();
    srTruth.x            = tru->GetNeutrino().X();
    srTruth.y            = tru->GetNeutrino().Y();
    srTruth.W2           = tru->GetNeutrino().W() * tru->GetNeutrino().W();
    
    srTruth.E            = tru->GetNeutrino().Nu().Momentum().E();
    srTruth.visE         = effPur->energyTotal;
    srTruth.visEinslc    = effPur->energySlice;
    srTruth.eff          = effPur->efficiency;
    srTruth.pur          = effPur->purity;
    srTruth.nhitslc      = effPur->nSliceHits;
    srTruth.nhittot      = effPur->nTotalHits;
    srTruth.time         = tru->GetNeutrino().Nu().Position().T();
    srTruth.vtx          = tru->GetNeutrino().Nu().Position().Vect();
    srTruth.p            = tru->GetNeutrino().Nu().Momentum();
    
    TVector3 nuVtx = tru->GetNeutrino().Nu().Position().Vect();    
    bool inDet = fabs(nuVtx.X()) <= geom->DetHalfWidth()
      && fabs(nuVtx.Y()) <= geom->DetHalfHeight()
      && nuVtx.Z() >= 0 && nuVtx.Z() <= geom->DetLength();
    srTruth.isvtxcont    = inDet;

    art::PtrVector<simb::MCTruth> pv;
    pv.push_back(tru);
    art::FindManyP<simb::MCFlux> fmFlux(pv, evt, fGeneratorLabel);
    if(fmFlux.isValid())
    {
      std::vector< art::Ptr<simb::MCFlux> > fluxes = fmFlux.at(0);
      
      if(!fluxes.empty())
      {
        const simb::MCFlux& flux = *fluxes[0];
        
        srTruth.pdgorig  = flux.fntype;
        srTruth.woscdumb = SimpleOscProb(flux, tru->GetNeutrino());
        
        SRBeam& beam = srTruth.beam;
        
        beam.tv       = TVector3(flux.ftvx, flux.ftvy, flux.ftvz);
        beam.tp       = TVector3(flux.ftpx, flux.ftpy, flux.ftpz);
        beam.runjob   = flux.frun;
        beam.potnum   = flux.fevtno;
        beam.tptype   = flux.ftptype;
        beam.nimpwt   = flux.fnimpwt;
        beam.ndecay   = flux.fndecay;
        beam.v        = TVector3(flux.fvx, flux.fvy, flux.fvz);
        beam.pdp      = TVector3(flux.fpdpx, flux.fpdpy, flux.fpdpz);
        beam.ppdxdz   = flux.fppdxdz;
        beam.ppdydz   = flux.fppdydz;
        beam.pppz     = flux.fpppz;
        beam.ppenergy = flux.fppenergy;
        beam.ppmedium = flux.fppmedium;
        beam.ptype    = flux.fptype;
        beam.ppv      = TVector3(flux.fppvx, flux.fppvy, flux.fppvz);
        beam.muparp   = TVector3(flux.fmuparpx, flux.fmuparpy, flux.fmuparpz);
        beam.mupare   = flux.fmupare;
        beam.necm     = flux.fnecm; 
        beam.tgen     = flux.ftgen; 
        beam.tgptype  = flux.ftgptype; 
	beam.dk2gen   = flux.fdk2gen;
	beam.gen2vtx  = flux.fgen2vtx;
	beam.dk2vtx   = flux.fdk2gen + flux.fgen2vtx;

      }
      else{
        throw cet::exception("Missing Association Error", "Failed to find MCFlux object assoicated with MCTruth.  This probably means something went wrong with overlays.");
      }
    }
    
    
    //get the neutrino's daughters (and other progeny)
    std::vector<const sim::Particle*> particles = bt->MCTruthToParticles(tru);
    
    // We're going to loop over daughters to add other primaries to truNu.
    // First we're gonna get the lepton and add it as the zeroth entry,
    // so let's get its information.
    int lepPdg = tru->GetNeutrino().Lepton().PdgCode();
    double lepEnergy = tru->GetNeutrino().Lepton().E();
    int lepCount = 0;  //count the found leptons, for debugging.
    int lepTrackId = 0; // We'll set this later
    // now we'll loop over the lepton and add it
    for(const sim::Particle* part: particles)
    {
      //Add particle if it's a primary and it has the correct PDG code and the correct energy.
      if(part->Mother() == 0 && part->PdgCode() == lepPdg &&
         EssentiallyEqual(part->E(), lepEnergy))
      {
        lepTrackId = part->TrackId();
        ++lepCount;
        AddParticleToVec(*part, allTracks, sliceTracks, &srTruth.prim);
      }
    }
    
    
    if(lepCount != 1)
    {
      switch (lepCount) {
        case 0:
          if (!srTruth.iscc) break;
          
          mf::LogError("CAFMaker")
          << "*****************************************" << "\n"
          << "Uh, oh, CAFMaker did not find primary lepton. "<< "\n"
          << "We're going to need to find a way to deal with this."  << "\n"
          << "Info about the particle: "   << "\n"
          << "  pdg: " << srTruth.pdg   << "\n"
          << "  pdgorig: " << srTruth.pdgorig   << "\n"
          << "  E: "<< srTruth.E        << "\n"
          << "  visE: "<< srTruth.visE        << "\n"
          << "  nhittot: "<< srTruth.nhittot        << "\n"
          << "  X: "<< srTruth.vtx.X()  << "\n"
          << "  Y: "<< srTruth.vtx.Y()  << "\n"
          << "  Z: "<< srTruth.vtx.Z()  << "\n"
          << "  iscc: "<< srTruth.iscc  << "\n"
          << "  infid(nd): "<< (srTruth.vtx.X() < 200 && srTruth.vtx.X() > -200 &&
                                srTruth.vtx.Y() < 200 && srTruth.vtx.Y() > -200 &&
                                srTruth.vtx.Z() > 0 && srTruth.vtx.X() < 1600  ) << "\n"
          
          << "*****************************************" << "\n" ;
          break;
        default:
          mf::LogError("CAFMaker")
          << "*****************************************" << "\n"
          << "Uh, oh, CAFMaker found too many primary leptons in event. "<< "\n"
          << "We're going to need to find a way to deal with this."  << "\n"
          << "*****************************************";
          break;
      }
      //      assert(lepCount == 1);
    }
    
    for(const sim::Particle* part: particles){
      //Add particle if it's a primary and it's not the lepton
      if(part->TrackId() != lepTrackId &&
         part->Mother() == 0)
        AddParticleToVec(*part, allTracks, sliceTracks, &srTruth.prim);
    }
    
    FindAndAddMichels(particles, allTracks, &srTruth.michel);
    
  }
  
  void CAFMaker::AddGenInfoToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                                 std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRGenerator>* vec, const art::Event& evt) {
    
    vec->push_back(SRGenerator());
    SRGenerator& gentruth = vec->back();
    
    const art::Ptr<simb::MCTruth> &tru = effPur->neutrinoInt;
    
    //Build the particle list first
    simb::MCNeutrino neutrino = tru->GetNeutrino();
    simb::MCParticle nu       = neutrino.Nu();
    
    for(int i = 0; i < tru->NParticles(); i++) {
      simb::MCParticle mcpart = tru->GetParticle(i);
      caf::SRGenParticle gpart;
      
      //int gmid = mcpart.PdgCode();
      int ndaughters = mcpart.NumberDaughters();
      int gmmo = mcpart.Mother();
      
      int fdtrkid = 0;
      int ldtrkid = 0;
      
      if(ndaughters !=0) {
        fdtrkid = mcpart.Daughter(0);
        if(ndaughters == 1) {
          ldtrkid = 1;
        }
        else if(ndaughters >1) {
          ldtrkid = mcpart.Daughter(ndaughters-1);
        }
      }
      int gmfd = -1;
      int gmld = -1;
      //Genie uses the index in the particle array to reference the daughter particles.
      //MCTruth keeps the particles in the same order so use the track ID to find the proper index.
      for(int j = 0; j < tru->NParticles(); j++) {
        simb::MCParticle temp = tru->GetParticle(i);
        if(temp.TrackId() == fdtrkid) {
          gmfd = j;
        }
        if(temp.TrackId() == ldtrkid) {
          gmld = j;
        }
      }
      gpart.fP4            = mcpart.Momentum();
      gpart.fVtx           = mcpart.Position();
      gpart.fPolz          = mcpart.Polarization();
      gpart.fNuclVtx       = mcpart.GetGvtx();
      
      gpart.fPdgCode       = mcpart.PdgCode();
      gpart.fStatus        = mcpart.StatusCode();
      gpart.fRescatterCode = mcpart.Rescatter();
      gpart.fFirstDaughter = gmfd;
      gpart.fLastDaughter  = gmld;
      gpart.fMother        = gmmo;
      
      gentruth.fParticleList.push_back(gpart);
      if(i==0) {
        gentruth.fProbeIndx = 0; //&gentruth->fParticleList[0];
      }
      if( (mcpart.Mother()==tru->GetNeutrino().Nu().TrackId())
         && ((mcpart.PdgCode()  == tru->GetNeutrino().Nu().PdgCode())
             || ( mcpart.PdgCode()  == (tru->GetNeutrino().Nu().PdgCode())-1) ) ){
        gentruth.fFSLeptonIndx = i; //&gentruth->fParticleList[i];
      }
    }
    
    //Now the rest of the MCTruth stuff
    gentruth.fHitNucleon = tru->GetNeutrino().HitNuc();
    gentruth.fHitQuark   = tru->GetNeutrino().HitQuark();
    
    
    //GTruth stuff here
    art::PtrVector<simb::MCTruth> pv;
    pv.push_back(tru);
    art::FindManyP<simb::GTruth> fmGtruth(pv, evt, fGeneratorLabel);
    if(fmGtruth.isValid()) {
      std::vector< art::Ptr<simb::GTruth> > gtruths = fmGtruth.at(0);
      
      if(!gtruths.empty()) {
        const simb::GTruth& gtruth = *gtruths[0];
        
        gentruth.fVisEfraction = 1.0;
        gentruth.visEinslc    = effPur->energySlice;
        gentruth.fProbePDG    = gtruth.fProbePDG;
        gentruth.fProbeP4     = gtruth.fProbeP4;
        gentruth.fVertex      = gtruth.fVertex;
        gentruth.fHitNucP4    = gtruth.fHitNucP4;
        gentruth.ftgtPDG      = gtruth.ftgtPDG;
        gentruth.ftgtA        = gtruth.ftgtA;
        gentruth.ftgtZ        = gtruth.ftgtZ;
        gentruth.fIsSeaQuark  = gtruth.fIsSeaQuark;
        gentruth.fFShadSystP4 = gtruth.fFShadSystP4;
        gentruth.fgY          = gtruth.fgY;
        gentruth.fgX          = gtruth.fgX;
        gentruth.fgT          = gtruth.fgT;
        gentruth.fgW          = gtruth.fgW;
        gentruth.fgq2         = gtruth.fgq2;
        gentruth.fgQ2         = gtruth.fgQ2;
        gentruth.fResNum      = gtruth.fResNum;
        gentruth.fIsCharm     = gtruth.fIsCharm;
        gentruth.fNumNeutron  = gtruth.fNumNeutron;
        gentruth.fNumProton   = gtruth.fNumProton;
        gentruth.fNumPi0      = gtruth.fNumPi0;
        gentruth.fNumPiMinus  = gtruth.fNumPiMinus;
        gentruth.fNumPiPlus   = gtruth.fNumPiPlus;
        gentruth.fweight      = gtruth.fweight;
        gentruth.fprobability = gtruth.fprobability;
        gentruth.fXsec        = gtruth.fXsec;
        gentruth.fDiffXsec    = gtruth.fDiffXsec;
        gentruth.fGint        = gtruth.fGint;
        gentruth.fGscatter    = gtruth.fGscatter;
        
      }
    }
  }
  
  void CAFMaker::CalculateVisEFraction(std::vector<SRGenerator>* gens) {
    double totalVisEinSlice = 0;
    for(const SRGenerator& gen: *gens){
      totalVisEinSlice += gen.visEinslc;
    }
    for(SRGenerator& gen: *gens){
      gen.fVisEfraction = gen.visEinslc/totalVisEinSlice;
    }
  }
  
  //.....................................................................
  void CAFMaker::AddCosmicTruthToVec(const cheat::NeutrinoEffPur* effPur, std::vector<cheat::TrackIDE> &allTracks,
                                     std::vector<cheat::TrackIDE> &sliceTracks, std::vector<SRCosmic>* vec, const art::Event& evt)
  {
    //get information about true cosmic
    
    vec->push_back(SRCosmic());
    SRCosmic& srTruth = vec->back();
    
    const art::Ptr<simb::MCTruth> &tru = effPur->neutrinoInt;
    
    // BUT GetParticle(int i) requires looping over list of particles in
    // single MCTruth object. So we have made an MCTruth object for each
    // "cosmic" and GetParticle(0) corresponds to the only particle then
    srTruth.pdg          = tru->GetParticle(0).PdgCode();
    
    srTruth.E            = tru->GetParticle(0).Momentum().E();
    srTruth.visE         = effPur->energyTotal;
    srTruth.visEinslc    = effPur->energySlice;
    srTruth.eff          = effPur->efficiency;
    srTruth.pur          = effPur->purity;
    srTruth.nhitslc      = effPur->nSliceHits;
    srTruth.nhittot      = effPur->nTotalHits;
    
    srTruth.time         = tru->GetParticle(0).Position().T();
    srTruth.vtx          = tru->GetParticle(0).Position().Vect();
    srTruth.p            = tru->GetParticle(0).Momentum();
    
    
    srTruth.azimuth      = atan2(tru->GetParticle(0).Momentum().Pz(),
                                  tru->GetParticle(0).Momentum().Px());
    srTruth.zenith       = -(tru->GetParticle(0).Momentum().Py())/tru->GetParticle(0).Momentum().P();
    
    std::vector<const sim::Particle*> particles = bt->MCTruthToParticles(tru);
    
    
    const simb::MCTrajectory& traj = particles[0]->Trajectory();
    
    
    
    bool entered = 0;
    bool exited = 0;
    
    for(size_t pt = 0; pt < traj.size(); ++pt)
    {
      TVector3 point = traj.Position(pt).Vect();
      
      // Check to see if trajectory point is in the detector
      bool inDet = fabs(point.X()) <= geom->DetHalfWidth()   // Check X inside detector
      && fabs(point.Y()) <= geom->DetHalfHeight()  // Then check Y
      && point.Z() >= 0 && point.Z() <= geom->DetLength(); // then check Z
      
      if(inDet && !entered)
      {
        // then we just entered the detector
        entered = true;
        
        if(pt > 0)
        {
          // Put xyz from last point into an array for geo function call
          TVector3 lastPoint = traj.Position(pt - 1).Vect();
          double pointArray[3] = {point.X(), point.Y(), point.Z()};
          
          // Put xyz from direction vector into an array.
          TVector3 dir = (lastPoint - point).Unit();
          double dirArray[3] = {dir.X(), dir.Y(), dir.Z()};
          
          double enterArray[3] = {0,0,0};
          
          geo::ProjectToBoxEdge(pointArray, dirArray,
                                - geom->DetHalfWidth()  , geom->DetHalfWidth(),
                                - geom->DetHalfHeight() , geom->DetHalfHeight(),
                                0  , geom->DetLength(),
                                enterArray);
          
          srTruth.enter.SetXYZ(enterArray[0], enterArray[1], enterArray[2]);
          // No good way to interpolate between momentum at pt-1 and pt.
          srTruth.penter = traj.Momentum(pt-1);
        }
        else
        {// if this is the case, we started inside the detector.
          srTruth.enter = point;
          srTruth.penter = traj.Momentum(pt);
        }
      }
      
      if(!inDet && entered && !exited){
        // then we just exited the detector
        exited= true;
        
        // Put xyz from last point into an array for geo function call
        TVector3 lastPoint = traj.Position(pt - 1).Vect();
        double lastPointArray[3] = {lastPoint.X(), lastPoint.Y(), lastPoint.Z()};
        
        // Put xyz from direction vector into an array.
        TVector3 dir = (point - lastPoint).Unit();
        double dirArray[3] = {dir.X(), dir.Y(), dir.Z()};
        
        double exitArray[3] = {0,0,0};
        
        geo::ProjectToBoxEdge(lastPointArray, dirArray,
                              - geom->DetHalfWidth()  , geom->DetHalfWidth(),
                              - geom->DetHalfHeight() , geom->DetHalfHeight(),
                              0  , geom->DetLength(),
                              exitArray);
        
        srTruth.exit.SetXYZ(exitArray[0], exitArray[1], exitArray[2]);
      }
    }
    srTruth.stop = traj.Position(traj.size() - 1).Vect();
    
    // make sure the particle actually entered the detector
    if(!entered){
      // if the primary never entered the detector, set the coordinates to -5*10^9
      
      srTruth.enter.SetXYZ(-5e9, -5e9, -5e9);
      srTruth.exit.SetXYZ(-5e9, -5e9, -5e9);
    }
    
    // If the particle entered and did not exit, set the exit position to stop
    if(entered && !exited){
      srTruth.exit = srTruth.stop;
    }
    // Find the Michel electrons
    FindAndAddMichels(particles, allTracks, &srTruth.michel);
  }
  //......................................................................
  
  void CAFMaker::FindAndAddMichels(std::vector<const sim::Particle*> particles,std::vector<cheat::TrackIDE> &allTracks,
                                   std::vector<SRTrueMichelE>* michelVec)
  {
    for(const sim::Particle* part: particles){
      // Check to see if we're dealing with an electron, continue if not.
      int pdg = part->PdgCode();
      if (abs(pdg) != 11) continue;

      // Check to see if mother is a muon, continue if not
      int motherId = part->Mother();
      if(motherId == 0) continue;
      const sim::Particle* mother = bt->TrackIDToParticle( motherId );
      int pdgMother = 0;
      if(mother) pdgMother = mother->PdgCode();
      if(abs(pdgMother) != 13) continue;

      // Check to see if particle was produced in decay, continue if not.
      if(part->Process().compare("Decay")) continue;

      // Add particle if it passed all criteria
      AddTrueMichelEToVec(*part, *bt->TrackIDToParticle( motherId ),
                          allTracks, michelVec);
    }
  }

  //......................................................................
  void CAFMaker::AddTrkMEClusterToVec(const me::MichelECluster& michel, std::vector<SRTrkME>* vec, size_t trackId)
  {
    vec->push_back(SRTrkME());
    SRTrkME& srME = vec->back();

    srME.nhit         =                     michel.NCell();
    srME.calE         =                     michel.CalorimetricEnergy();
    srME.deltat       =                     michel.DeltaT();
    srME.disttotrack  =                     michel.DistanceToTrackEnd();
    srME.atend        =                     michel.AtTrackEnd();
    srME.view         = (caf::_plane_proj)  michel.View();
    srME.meanpos      =                     michel.MeanXYZ();
    srME.trkidx       =      (short)        trackId;
  }

  //......................................................................
  void CAFMaker::AddSlcMEClusterToVec(const me::SliceMEFCluster& michel, std::vector<SRSlcME>* vec)
  {
    vec->push_back(SRSlcME());
    SRSlcME& srME = vec->back();

    srME.mid          =                     michel.MID();
    srME.nhit         =                     michel.NCell();
    srME.nretriggers  =                     michel.NRetriggeredCells();
    srME.nmatches     =    (unsigned short) michel.NSlcMatches();
    srME.calE         =                     michel.CalorimetricEnergy();
    srME.pe           =                     michel.TotalPE();
    srME.adc          =                     michel.TotalADC();
    srME.adcretriggers=                     michel.ADCRetriggeredCells();
    srME.deltat       =                     michel.DeltaT();
    srME.disttoslc    =                     michel.DistToSlc();
    srME.assocwithslc =                     michel.AssocWithSlc();
    srME.meanpos      =                     michel.MeanXYZ();
  }

  //......................................................................
  
  template<class T> void CAFMaker::FillProngVars(const rb::Prong &prng, T& srPrng)
  {
    srPrng.nhit         = prng.NCell();
    srPrng.nhitx        = prng.NXCell();
    srPrng.nhity        = prng.NYCell();
    srPrng.nplane       = prng.ExtentPlane();
    srPrng.maxplanecont = prng.MostContiguousPlanes(geo::kXorY);
    srPrng.maxplanegap  = prng.MostMissingPlanes(geo::kXorY);
    srPrng.start        = prng.Start();
    srPrng.dir          = prng.Dir();
    srPrng.calE         = prng.CalorimetricEnergy();
    
    srPrng.len          = prng.TotalLength();
    
    if      (prng.NXCell() == 0)  srPrng.view = kY;
    else if (prng.NYCell() == 0)  srPrng.view = kX;
    else                           srPrng.view = kXorY;
  }
  
  //......................................................................
  
  void CAFMaker::FillTrackVars(const rb::Track &trk, SRTrack& srTrk)
  {
    FillProngVars(trk, srTrk);
    srTrk.stop         = trk.Stop();
    srTrk.lenE         = srTrk.len * 0.00157; // From N. Raddatz and K. Sachdev, see DocDB 7658-v2, slide 11
                                              // trkLength * MIP dE/dx
    srTrk.nplanegap    = trk.NMissingPlanes(geo::kXorY);

  }
  
  double EfficiencyMetric(const cheat::NeutrinoEffPur& ep)
  {
    return ep.efficiency;
  }

  double PurityMetric(const cheat::NeutrinoEffPur& ep)
  {
    return ep.purity;
  }

  double EfficiencyTimesPurityMetric(const cheat::NeutrinoEffPur& ep)
  {
    return ep.efficiency*ep.purity;
  }

  double EnergyMetric(const cheat::NeutrinoEffPur& ep)
  {
    return ep.energySlice;
  }

  //......................................................................
  std::vector<int> CAFMaker::
  getFaveIds(const std::vector<cheat::NeutrinoWithIndex>& allNus,
             const std::vector<std::vector<cheat::NeutrinoEffPur>>& sliceTruthTable,
             std::function<double(const cheat::NeutrinoEffPur&)> sliceMetric,
             std::function<double(const cheat::NeutrinoEffPur&)> nuMetric)
  {
    // double check that slice truth table is safe
    for(auto slc: sliceTruthTable)
    {
      assert(slc.size() == allNus.size() &&
             "Slice/truth table must be properly shaped");
    }

    // We'll make an array with an entry for every subevt to store the unique
    // identifier of the favourite neutrino within the subevt, with the correct
    // size and all entries initialized to -1
    std::vector<int> faveIds(sliceTruthTable.size(), -1);

    // if no neutrinos, there are no favorites
    if(allNus.empty()) return faveIds;

    // List of best slice for each nu
    std::vector<int> bestSlices(allNus.size());
    //First we find the best slice for each neutrino
    for(size_t iNu = 0; iNu < allNus.size(); ++iNu){
      float maxValue = 0;
      int bestSlice = -1;
      for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc){
        const cheat::NeutrinoEffPur& effpur = sliceTruthTable[iSlc][iNu];
        if(!effpur.neutrinoInt.isNull()){
          if(sliceMetric(effpur) > maxValue){
            maxValue = sliceMetric(effpur);
            bestSlice = iSlc;
          }
        }
      } // end for iSlc
      bestSlices[iNu] = bestSlice;
    } // end for iNu

    // List of best nu for each slice

    // now we find the best neutrino for each slice.
    for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc){
      float maxValue = 0;
      int bestNu = -1;
      for(size_t iNu = 0; iNu < sliceTruthTable[iSlc].size(); ++iNu){
        const cheat::NeutrinoEffPur& effpur = sliceTruthTable[iSlc][iNu];
        if(!effpur.neutrinoInt.isNull()){
          if(nuMetric(effpur) > maxValue){
            maxValue = nuMetric(effpur);
            bestNu = iNu;
          }
        }
      } // end for iNu
      if(bestNu >= 0 && bestSlices[bestNu] == (int)iSlc){
        faveIds[iSlc] = bestNu;
      }
    }

    return faveIds;
  }

  //......................................................................
  void CAFMaker::FillSliceVars(const rb::Cluster& slice,
                               caf::SRSlice& srslice,
			       bool allowEmpty) const
  {
    const double wx = (slice.NXCell() > 0) ? slice.W(slice.XCell(0).get()) : 0;
    const double wy = (slice.NYCell() > 0) ? slice.W(slice.YCell(0).get()) : 0;

    //Variables needed:
    //Holds contiguous planes info:
    unsigned int  cellsFromEdge = -1;

    //Hit counters:
    int nCalHit = 0;
    int nMipHit = 0;

    unsigned int minCellsFromEdge = 99999999;

    //Loop over slice in question here:
    for(unsigned int hitIdx = 0; hitIdx < slice.NCell(); ++hitIdx)
    {
      const art::Ptr<rb::CellHit>& chit = slice.Cell(hitIdx);
      //Get calbrated reco hit:
      const rb::RecoHit rhit(cal->MakeRecoHit(*chit,
                                              chit->View() == geo::kX ? wx : wy));

      const int planeNum = chit->Plane();
      cellsFromEdge = std::min((unsigned int)chit->Cell(), geom->Plane(planeNum)->Ncells() - 1 - chit->Cell());
      
      if(cellsFromEdge < minCellsFromEdge)
      {
        minCellsFromEdge = cellsFromEdge;
      }

      //Make sure things have been calibrated:
      if(!rhit.IsCalibrated())
      {
        LOG_DEBUG("CAFMaker") << "Not calibrated?! "
        << chit->Plane() << " " << chit->Cell();
        continue;
      }
      
      //Count up both the total calibrated hits
      //and the MIP hits from 100-245 PECorr
      //Count up total hits:
      nCalHit++;
      //Count up MIP hits:
      if ( (rhit.PECorr()>100.0) && (rhit.PECorr()<245.0) ) nMipHit++;
    } //END OF SLICE LOOP HERE

    srslice.nhit           = slice.NCell();
    srslice.ncalhit        = nCalHit;
    srslice.nmiphit        = nMipHit;
    srslice.ncontplanes    = slice.MostContiguousPlanes(geo::kXorY);
    if (!(allowEmpty && (slice.NCell()==0))){

      srslice.firstplane     = slice.MinPlane();
      srslice.lastplane      = slice.MaxPlane();
      srslice.firstcell      = slice.MinCell(geo::kXorY);
      srslice.lastcell       = slice.MaxCell(geo::kXorY);
      srslice.ncellsfromedge = minCellsFromEdge;

      srslice.calE           = slice.CalorimetricEnergy();
      srslice.starttime      = slice.MinTNS();
      srslice.endtime        = slice.MaxTNS();
      srslice.meantime       = slice.MeanTNS();

      slice.MinMaxMeanXYZ(srslice.boxmin, srslice.boxmax, srslice.meanpos);

      //TBrowser can't draw DBL_MAX
      if (allowEmpty && (slice.NXCell()==0)){
	srslice.boxmin.SetX(-5555);
	srslice.boxmax.SetX(5555);
      }
      if (allowEmpty && (slice.NYCell()==0)){
	srslice.boxmin.SetY(-5555);
	srslice.boxmax.SetY(5555);
      }
      
    }//End of loop over variables to always fill for slices / only fill for hadclust if it has hits

    else{

      srslice.firstplane     = 0;
      srslice.lastplane      = 0;
      srslice.firstcell      = 0;
      srslice.lastcell       = 0;
      srslice.ncellsfromedge = 0;

      srslice.calE           = -5;
      srslice.starttime      = -5;
      srslice.endtime        = -5;
      srslice.meantime       = -5;

      srslice.boxmin         = TVector3(-5555,-5555,-5555);
      srslice.boxmax         = TVector3(-5555,-5555,-5555);
      srslice.meanpos        = TVector3(-5555,-5555,-5555);

    }

 
  }

  //......................................................................
  /// This function is here for debugging purposes.
  /// It will draw a table of with slices as rows and neutrinos as columns.
  /// The favorites will be written with a star next to them.
  /// If two faveIds vectors are provided, the second will have an @ drawn next to them
  void CAFMaker::drawSliceTruthTable(const std::vector< std::vector<cheat::NeutrinoEffPur> > &sliceTruthTable,
                                     const std::vector<int> &faveIds)
  {
    std::vector<int> otherFaveIds(faveIds.size(), -1);
    drawSliceTruthTable(sliceTruthTable, faveIds, otherFaveIds);
  }
  
  //......................................................................

  /// This function is here for debugging purposes.
  /// It will draw a table of with slices as rows and neutrinos as columns.
  /// The favorites will be written with a star next to them.
  /// If two faveIds vectors are provided, the second will have an @ drawn next to them
  void CAFMaker::drawSliceTruthTable(const std::vector< std::vector<cheat::NeutrinoEffPur> > &sliceTruthTable,
                                     const std::vector<int> &faveIds, const std::vector<int> &otherFaveIds)
  {
    std::cout.precision(2);
    std::cout.width(2);
    std::cout.setf(std::ios::fixed,std::ios::floatfield);
    int iSlc = 0;
    for(auto slc: sliceTruthTable){
      int iNu  = 0;
      for(auto nu = slc.begin(); nu != slc.end(); ++nu, ++iNu)
      {
        std::cout << nu->efficiency;
        if(faveIds[iSlc] == iNu)
          std::cout << "*";
        else
          std::cout << " ";
        if(otherFaveIds[iSlc] == iNu)
          std::cout << "@";
        else
          std::cout << " ";
        std::cout << "         " ;
      }
      std::cout << std::endl;
    }
  }

  //......................................................................
  template<class T> bool CAFMaker::
  GetAssociatedProduct(const art::FindManyP<T>& fm, int idx, T& ret)
  {
    if(!fm.isValid()) return false;

    const std::vector<art::Ptr<T>> prods = fm.at(idx);

    if(prods.empty()) return false;

    ret = *prods[0];

    return true;
  }

  //......................................................................
  
  void CAFMaker::produce(art::Event& evt) noexcept
  {
    // Normally CAFMaker is run without an output ART stream, so these go
    // nowhere, but can be occasionally useful for filtering as part of an ART
    // job.
    std::unique_ptr<std::vector<caf::StandardRecord>> srcol(new std::vector<caf::StandardRecord>);
    std::unique_ptr<caf::SRSpill> spillcol(new SRSpill);

    fTotalEvents += 1;
    fIsRealData = evt.isRealData();
    
    //get a handle on the slices
    art::Handle<std::vector<rb::Cluster> > hSlices;
    evt.getByLabel(fClusterLabel, hSlices);
    if(hSlices.failedToGet())
    {
      mf::LogError("CAFMaker") << "Slices not found in CAFMaker";
    }
    
    // define a vector holding slices
    art::PtrVector<rb::Cluster> subevts;
    for(unsigned int i = 0; i < hSlices->size(); ++i)
      subevts.push_back(art::Ptr<rb::Cluster>(hSlices, i));
      
      
    //Get all of the hits from calhits
    art::Handle< std::vector<rb::CellHit> > hitsHandle;
    evt.getByLabel(fHitsLabel, hitsHandle);
    art::PtrVector<rb::CellHit> allHits;
    for(size_t i = 0; i < hitsHandle->size(); ++i)
    {
      art::Ptr<rb::CellHit> hit(hitsHandle, i);
      allHits.push_back(hit);
    }
    
    // We're gonna get a vector of NeutrinoEffPurs for allHits
    std::vector<cheat::NeutrinoWithIndex> allNus = bt->allMCTruth();
    
    std::vector< std::vector<cheat::NeutrinoEffPur> > sliceTruthTable = bt->SlicesToMCTruthsTable(subevts);
    
    
    
    // Find favorite neutrino IDs using several metrics.  Total energy is the
    // one we will use, for the favorite neutrino, but the others will be
    // stored.
    std::vector<int> faveIdsEff        = getFaveIds(allNus, sliceTruthTable, EfficiencyMetric, EfficiencyMetric);
    std::vector<int> faveIdsEnergy     = getFaveIds(allNus, sliceTruthTable, EnergyMetric, EnergyMetric);
    std::vector<int> faveIdsPur        = getFaveIds(allNus, sliceTruthTable, PurityMetric, PurityMetric);
    std::vector<int> faveIdsEffPur     = getFaveIds(allNus, sliceTruthTable, EfficiencyTimesPurityMetric, EfficiencyTimesPurityMetric);
    std::vector<int> faveIdsEffThenPur = getFaveIds(allNus, sliceTruthTable, EfficiencyMetric, PurityMetric);

    //Get all truth track hits in the event
    std::vector<cheat::TrackIDE> allTracks;
    if(bt->HaveTruthInfo()) allTracks = bt->HitsToTrackIDE(allHits);
      
      
      
      
    //try to get pot spill info if this is real data
    art::Handle<sumdata::SpillData> spillPot;

    if(bt->HaveTruthInfo()){
      evt.getByLabel(fGeneratorLabel, spillPot);
    }
    else{
      evt.getByLabel(fNumiBeamLabel, spillPot);
    }

    // Get the EventQuality information
    art::Handle<sumdata::EventQuality> spillQual;
    evt.getByLabel(fSpillQualLabel, spillQual);

    art::Handle<std::vector<rawdata::RawTrigger>> trigs;
    evt.getByLabel(fRawDataLabel, trigs);

    caf::SRSpill spill;
    FillSpillVars(spillPot, spillQual, trigs, spill, evt);

    //Fill the spill tree once per event so that POT counting can be done properly.
    //This is tree will have an entry even for empty events with no physics slices.
    caf::SRSpill* pspill = &spill;
    fSpillTree->SetBranchAddress("spill", &pspill);
    fSpillTree->Fill();
    
    
    // Get all association finder objects
    art::FindManyP<rb::Track>               fmDiscrete2d(hSlices, evt, fDiscreteTrack2dLabel);
    art::FindManyP<rb::Track>               fmDiscrete(hSlices, evt, fDiscreteTrackLabel);
    art::FindManyP<rb::Track>               fmKalman(hSlices, evt, fKalmanTrackLabel);
    art::FindManyP<rb::Track>               fmCosmic(hSlices, evt, fCosmicTrackLabel);
    art::FindManyP<rb::Vertex>              fmElastic(hSlices, evt, fElasticArmsLabel);
    art::FindManyP<rb::Vertex>              fmHoughVertex(hSlices, evt, fHoughVertexLabel); // TODO
    art::FindManyP<me::SliceMEFCluster>     fmSliceMEF(hSlices, evt, fSliceMEFLabel);
    art::FindManyP<jmshower::JMShower>      fmJmShower(hSlices, evt, fJMShowerLabel);
    art::FindManyP<jmshower::EID>           fmNueSel(hSlices, evt, fNueSelLabel);
    art::FindOneP<slid::EventLID>           foLID(hSlices, evt, fLidLabel); //one lid object per slice
    art::FindMany<slid::ShowerLID>          fmShwLID(hSlices, evt, fLidLabel);
    art::FindMany<slid::ShowerPID>          fmShwPID(hSlices, evt, fSPidLabel);
    art::FindManyP<numue::NumuE>            fmNumuEnergy(hSlices, evt, fNumuEnergyLabel);
    art::FindManyP<presel::NuePreselObj>    fmNuePresel(hSlices,evt, fNuePreselLabel);
    art::FindManyP<presel::Veto>            fmVeto(hSlices,evt, fVetoLabel);
    art::FindManyP<qeef::QePId>             fmQepid(hSlices, evt, fQepidLabel);
    art::FindManyP<nuesand::NueSandObj>     fmNueSand(hSlices, evt, fNueSandboxLabel);
    art::FindManyP<numusand::NumuSandObj>   fmNumuSand(hSlices, evt, fNumuSandboxLabel);
    art::FindManyP<lem::PIDDetails>         fmLem(hSlices, evt, fLemLabel);
    art::FindManyP<rvp::RVP>                fmRvp(hSlices, evt, fRvpLabel);
    art::FindManyP<cosrej::CosRejObj>       fmCosRej(hSlices,evt,fCosRejLabel);
    art::FindManyP<cosrej::NueCosRej>       fmNueCosRej(hSlices,evt,fNueCosRejLabel);
    art::FindManyP<murem::MRCCParent>       fmMrccPar(hSlices, evt, fMrccParentLabel);
    art::FindManyP<bpfit::BPFEnergy>        fmBPFEnergy(hSlices, evt, fBPFEnergyLabel);
    
    int nNonNoise = 0;
    for(art::Ptr<rb::Cluster> subevt: subevts){
      if(subevt->IsNoise() || subevt->NCell() == 0)
        continue;
      ++nNonNoise;
    }
    for(unsigned int subevtId = 0; subevtId < subevts.size(); ++subevtId )
    {
      if(subevts[subevtId]->IsNoise() || subevts[subevtId]->NCell() == 0)  continue;
      // because we don't care about the noise slice and slices with no hits.
      
      StandardRecord rec;
      StandardRecord* prec = &rec; // TTree wants a pointer-to-pointer
      fRecTree->SetBranchAddress("rec", &prec);
      
      //#######################################################
      // Fill header.
      //#######################################################
      // Get metadata information for header
      unsigned int run = evt.run();
      unsigned int subrun = evt.subRun();
      unsigned int spillNum = evt.id().event();
      bool isMC = bt->HaveTruthInfo();
      
      art::Timestamp ts = evt.time();
      const time_t timeSec = ts.timeHigh();
      
      struct tm* timeStruct = localtime(&timeSec);
      
      unsigned int maskstatus = 0;
      unsigned int dibmask = (rh->GoodDiBlockMask()&rh->GetConfiguration());

      if(rh->GoodDiBlockMask() == ((1<<16)-1)) maskstatus = 0;
      else if(rh->GoodDiBlockMask() == ((1<<17)-1)) maskstatus = 2;
      else maskstatus = 1;

      unsigned int dibfirst = 0;
      unsigned int diblast = 0;
      if (dibmask) {
        int iD;
        iD=0; while (!((dibmask>>iD)&1)) iD++; dibfirst=iD+1;
        iD=0; while (dibmask>>iD)        iD++; diblast=iD;
      }
      
      unsigned int nbadchan = bc->NBadInSubRun(subrun);
      unsigned int ntotchan = rh->NAnalysisChannels();

      // Pull the MC cycle number from the metadata
      // Assumes metadata module is being run...
      std::map<std::string, std::string> metadata = meta::MetadataManager::getInstance().GetMetadata();
      // Check it exists in the metadata, otherwise it defaults to -5
      // in the InitializeOutfile() function.
      // Convert the std::string to int
      if(metadata.count("simulated.cycle"))
	fCycle = std::stoi(metadata["simulated.cycle"]);      


      rec.hdr = SRHeader();
      
      rec.hdr.run             = run;
      rec.hdr.subrun          = subrun;
      rec.hdr.cycle           = fCycle;
      rec.hdr.evt             = spillNum;
      rec.hdr.subevt          = subevtId;
      rec.hdr.ismc            = isMC;
      rec.hdr.det             = fDet;
      rec.hdr.blind           = 0;
      rec.hdr.filt            = rb::IsFiltered(evt, subevts[subevtId]);
      
      rec.hdr.subevtstarttime = subevts[subevtId]->MinTNS();
      rec.hdr.subevtendtime   = subevts[subevtId]->MaxTNS();
      rec.hdr.subevtmeantime  = subevts[subevtId]->MeanTNS();
      rec.hdr.unixtime        = timeSec;
      
      rec.hdr.dibfirst        = dibfirst;
      rec.hdr.diblast         = diblast;
      rec.hdr.dibmask         = dibmask;
      rec.hdr.maskstatus      = maskstatus;
      
      rec.hdr.year            = timeStruct->tm_year + 1900;
      rec.hdr.month           = timeStruct->tm_mon + 1;
      rec.hdr.day             = timeStruct->tm_mday;
      rec.hdr.doy             = timeStruct->tm_yday + 1;
      rec.hdr.hour            = timeStruct->tm_hour + 1;
      rec.hdr.minute          = timeStruct->tm_min;
      rec.hdr.second          = timeStruct->tm_sec;
      
      rec.hdr.nbadchan        = nbadchan;
      rec.hdr.ntotchan        = ntotchan;

      //#######################################################
      // Fill Spill info.
      //#######################################################
      //Store info on NuMI beam quality and if the pot as good or bad.
      //This is meant to be used to refine a good spill, not for
      //pot counting since events have multiple slices and empty
      //spill are missing from this tree.  Use the friend tree for
      //pot counting.
      rec.spill = spill;
      
      //#######################################################
      // Add slice info.
      //#######################################################
      FillSliceVars(*subevts[subevtId], rec.slc);
      rec.slc.nnonnoise = nNonNoise;

      if(fDet == kFARDET){
        rec.sel.contain.nplanestofront = rec.slc.firstplane - (dibfirst-1)*64;
        rec.sel.contain.nplanestoback = ((diblast)*64)-1 - rec.slc.lastplane;
      }

      if(fDet == kNEARDET){
        rec.sel.contain.nplanestofront = rec.slc.firstplane;
        rec.sel.contain.nplanestoback = 214 - rec.slc.lastplane;
      }

      //#######################################################
      // Adding reconstructed objects.
      //#######################################################
      
      // Get all the tracks associated with this slice from discrete tracker.
      //This depends on the findMany object created above.
      if(fmDiscrete.isValid())
      {
        std::vector< art::Ptr<rb::Track> > discTracks = fmDiscrete.at(subevtId);
        // Sort the tracks in descending length order
        std::sort(discTracks.begin(), discTracks.end(), sortRBTrackLength);
        art::FindManyP<me::MichelECluster> fmDiscreteMichelE(discTracks, evt, fDiscreteMichelELabel);
        rec.trk.discrete.resize(discTracks.size());
        for(size_t trackId = 0; trackId < discTracks.size(); ++trackId)
        {
          FillTrackVars((*discTracks[trackId]), rec.trk.discrete[trackId]);
          if(fmDiscreteMichelE.isValid())
          {
            //Get MichelE from filter
            std::vector< art::Ptr<me::MichelECluster> > michels = fmDiscreteMichelE.at(trackId);
            
            for(art::Ptr<me::MichelECluster> michel: michels){
              AddTrkMEClusterToVec(*michel, &rec.me.trkdiscrete, trackId);
            }
          }
        }
      }
      if(fmDiscrete2d.isValid())
      {
        std::vector< art::Ptr<rb::Track> > discTracks = fmDiscrete2d.at(subevtId);
        // Sort the tracks in descending length order
        std::sort(discTracks.begin(), discTracks.end(), sortRBTrackLength);
        rec.trk.discrete2d.resize(discTracks.size());
        for(size_t trackId = 0; trackId < discTracks.size(); ++trackId){
          FillTrackVars(*discTracks[trackId], rec.trk.discrete2d[trackId]);
        }
      }
      
      if(fmCosmic.isValid())
      {
        std::vector<art::Ptr<rb::Track>> cosTracks = fmCosmic.at(subevtId);
        // Sort the tracks in descending length order
        std::sort(cosTracks.begin(), cosTracks.end(), sortRBTrackLength);
        
        art::FindManyP<me::MichelECluster> fmCosmicMichelE(cosTracks, evt, fCosmicMichelELabel);
        rec.trk.cosmic.reserve(cosTracks.size());
        rec.trk.cosmic.resize(cosTracks.size());
        for(size_t trackId = 0; trackId < cosTracks.size(); ++trackId)
        {
          FillTrackVars((*cosTracks[trackId]), rec.trk.cosmic[trackId]);
          if(fmCosmicMichelE.isValid())
          {
            //Get MichelE from filter
            std::vector< art::Ptr<me::MichelECluster> > michels = fmCosmicMichelE.at(trackId);
            
            for(art::Ptr<me::MichelECluster> michel: michels){
              AddTrkMEClusterToVec(*michel, &rec.me.trkcosmic, trackId);
            }
          }
        } // end for trackId
      }
      
      // now do the same for kalmanTrack
      if(fmKalman.isValid())
      {
        std::vector< art::Ptr<rb::Track> > kalTracks = fmKalman.at(subevtId);
        // Sort the tracks in descending length order
        std::sort(kalTracks.begin(), kalTracks.end(), sortRBTrackLength);
        art::FindManyP<remid::ReMId> fmRemid(kalTracks, evt, fRemidLabel);
        art::FindManyP<rb::Energy> fmMuTrkE(kalTracks, evt, fNumuEnergyLabel);
        art::FindManyP<me::MichelECluster> fmKalmanMichelE(kalTracks, evt, fKalmanMichelELabel);
        
        size_t nTracks2d = 0;
        size_t nTracks3d = 0;
        
        //Give SRRemid the pid values of track with largest PID
        unsigned int bestPidIdx = remid::HighestPIDTrack(kalTracks, 
                                  fRemidLabel, evt);

        unsigned int bestPidIdx3d = 999;

        for(size_t trackId = 0; trackId < kalTracks.size(); ++trackId)
        {
          if(kalTracks[trackId]->View() != geo::kXorY)
          {
            rec.trk.kalman2d.push_back(SRTrack());
            FillTrackVars((*kalTracks[trackId]), rec.trk.kalman2d[nTracks2d]);
            nTracks2d++;
          }
          else
          { 
            rec.trk.kalman.push_back(SRTrack());
            FillTrackVars((*kalTracks[trackId]), rec.trk.kalman[nTracks3d]);            
            if(trackId == bestPidIdx) bestPidIdx3d = nTracks3d; 

            if(fmRemid.isValid())
            {
              //Get MuID from ReMId
              std::vector< art::Ptr<remid::ReMId> > remids = fmRemid.at(trackId);
              
              if(!remids.empty())
              {
                rec.sel.remid.muid.push_back(SRMuId());
                SRMuId& srMuId = rec.sel.remid.muid.back();

                FillMuIdVars(*remids[0], *kalTracks[trackId], srMuId);

              }
              else
              {
                throw cet::exception("Missing Association Error", "Failed to find a ReMId object with this track.");
              }
            }
            
            
            if(fmMuTrkE.isValid())
            {
              //Get energy for track assuming it is a muon
              std::vector< art::Ptr<rb::Energy> > mutrkEs = fmMuTrkE.at(trackId);
              
              if(!mutrkEs.empty())
              {
                rec.energy.mutrkE.push_back(SREnergy());
                rec.energy.mutrkE.back().E = mutrkEs[0]->E();
              }
              else{
                throw cet::exception("Missing Association Error", "Failed to find a muon track energy object with this track.");
              }
            }
            
            if(fmKalmanMichelE.isValid()){
              //Get MichelE from filter
              std::vector< art::Ptr<me::MichelECluster> > michels = fmKalmanMichelE.at(trackId);
              
              for(art::Ptr<me::MichelECluster> michel: michels){
                AddTrkMEClusterToVec(*michel, &rec.me.trkkalman, trackId);
              }
            }
            nTracks3d++;
          }
        } // end for trackId
        

        if(!rec.sel.remid.muid.empty() && bestPidIdx3d != 999){
          

          CopyRemidVars(rec.sel.remid.muid[bestPidIdx3d], rec.sel.remid);
          rec.sel.remid.bestidx      = bestPidIdx3d;
        }
        else{
          rec.sel.remid.setDefault();
        }
      }
      else{
        rec.sel.remid.setDefault();
      }
      rec.trk.fillSizes();
      rec.sel.remid.fillSizes();
      
      
      //  Get all vertices associated with slice from FuzzyKVertex and ElasticArms
      if(fmElastic.isValid()){
        std::vector< art::Ptr<rb::Vertex> > elastics = fmElastic.at(subevtId);
        art::FindManyP<rb::Prong> fmFuzzyProng3D(elastics, evt, art::InputTag(fFuzzyKLabel, fFuzzyK3dAssnLabel));
        art::FindManyP<rb::Prong> fmFuzzyProng2D(elastics, evt, art::InputTag(fFuzzyKLabel, fFuzzyK2dAssnLabel));
        
        for(unsigned int vtxId = 0; vtxId < elastics.size(); ++vtxId){
          rec.vtx.elastic.push_back(SRElastic());
          SRElastic& srVtx = rec.vtx.elastic.back();
          
          srVtx.time = elastics[vtxId]->GetT();
          srVtx.vtx = elastics[vtxId]->GetXYZ();
          if(fmFuzzyProng3D.isValid()){
            std::vector< art::Ptr<rb::Prong> > prongs = fmFuzzyProng3D.at(vtxId);
            //short prongs in descending length order
            std::sort(prongs.begin(), prongs.end(),sortRBProngLength);
            for(art::Ptr<rb::Prong> prong: prongs){
              if(prong->NXCell() == 0 || prong->NYCell() == 0){
                srVtx.png2d.push_back(SRProng());
                FillProngVars(*prong, srVtx.png2d.back());
              }//end 2d prong case
              else{
                srVtx.png3d.push_back(SRProng());
                FillProngVars(*prong, srVtx.png3d.back());
              }//end 3d prong case
            }//end loop over prongs
          }//end if prongs are valid for this vertex
          //check if there is a valid instance label for 2D prongs
          if(fmFuzzyProng2D.isValid() && fFuzzyK2dAssnLabel != ""){
            std::vector< art::Ptr<rb::Prong> > prongs = fmFuzzyProng2D.at(vtxId);
            //short prongs in descending length order
            std::sort(prongs.begin(), prongs.end(),sortRBProngLength);
            for(art::Ptr<rb::Prong> prong: prongs){
              if(prong->NXCell() == 0 || prong->NYCell() == 0){
                srVtx.png2d.push_back(SRProng());
                FillProngVars(*prong, srVtx.png2d.back());
              }//end 2d prong case
              else{
                srVtx.png3d.push_back(SRProng());
                FillProngVars(*prong, srVtx.png3d.back());
              }//end 3d prong case
            }//end loop over prongs
          }//end if prongs are valid for this vertex
        } // end for vtxId
        //sort the vector
        std::sort(rec.vtx.elastic.begin(), rec.vtx.elastic.end(), sortElasticProng);
      }

      if(fmHoughVertex.isValid()){
        std::vector<art::Ptr<rb::Vertex>> vtxs = fmHoughVertex.at(subevtId);

        for(unsigned int vtxId = 0; vtxId < vtxs.size(); ++vtxId){
          rec.vtx.hough.push_back(SRHoughVertex());
          SRHoughVertex& srVtx = rec.vtx.hough.back();

          srVtx.time = vtxs[vtxId]->GetT();
          srVtx.vtx = vtxs[vtxId]->GetXYZ();
        }
      }
      rec.vtx.fillSizes();
      
      if(fmSliceMEF.isValid()){
        std::vector< art::Ptr<me::SliceMEFCluster> > michels = fmSliceMEF.at(subevtId);
        for(art::Ptr<me::SliceMEFCluster> michel: michels){
          AddSlcMEClusterToVec(*michel, &rec.me.slc);
        }
      }
      rec.me.fillSizes();
            
      //  Get all showers associated with slice from JMShower
      if(fmJmShower.isValid()){
	std::vector< art::Ptr<jmshower::JMShower> > showers = fmJmShower.at(subevtId);
	for(art::Ptr<jmshower::JMShower> shower: showers){
	  rec.sand.nue.jm.push_back(SRJMShower());
	  FillJMShowerVars(*shower, rec.sand.nue.jm.back());
	}
	
	//sort the vector
	std::sort(rec.sand.nue.jm.begin(), rec.sand.nue.jm.end(), sortShowerEnergy);
      }
      
      rec.sand.nue.njm = rec.sand.nue.jm.size();
     
      // Get the JM EID objects and add them.
      jmshower::EID eid;
      if(GetAssociatedProduct(fmNueSel, subevtId, eid)){
        FillJMEIDVars(eid, rec.sand.nue.jmid);
	
        rec.sand.nue.time     = subevts[subevtId]->MeanTNS();
	
        FillNueSandEIDVars(eid, rec.sand.nue);
      }
      else{
        rec.sand.nue.jmid    .setDefault();
      }
      
      //  Get slice level PID from LIDBuilder
      if(foLID.isValid()){
	cet::maybe_ref<art::Ptr<slid::EventLID> const> rlid(foLID.at(subevtId));
	art::Ptr<slid::EventLID> elid = rlid.ref();
	//if the pointer is valid
        if (elid){
          rec.sel.elecid.ann      = elid->Value();
          rec.sel.elecid.anne     = elid->ValueWithE();
          if(fmShwLID.isValid()){
	    std::vector<const  slid::ShowerLID* > shwlids = fmShwLID.at(subevtId);
            //sort here since the vector may not be
            sort(shwlids.begin(),shwlids.end(), slid::CompareByEnergy);
            art::FindOneP<rb::Shower> fos(shwlids, evt, fLidLabel);
            for(unsigned int shwId = 0; shwId < shwlids.size(); ++shwId){
              cet::maybe_ref<art::Ptr<rb::Shower> const> rshw(fos.at(shwId));
              art::Ptr<rb::Shower> shw = rshw.ref();
              if (shw){
                rec.shw.shwlid.push_back(SRShower());
                FillShowerVars(*shw, *shwlids[shwId], rec.shw.shwlid.back());
              }
	      
              rec.sel.elecid.shwlid.push_back(SRSLid());
              FillSlidVars(*shwlids[shwId], rec.sel.elecid.shwlid.back());
	      
              //fill energy information for leading energy shower which also
              //servers a slice pid
              if (shwId == 0){
                FillLIDEnergyVars(*shwlids[shwId], rec.energy.elecid);
              }
	      
              //fill energy vector with the other shower energies
              rec.energy.shwlid.push_back(SRSLidEnergy());
              FillLIDEnergyVars(*shwlids[shwId], rec.energy.shwlid.back());
            } //end for loop on showerlids
          } //end if showerlids are valid

          if(fmShwPID.isValid()){
            std::vector<const  slid::ShowerPID* > shwpids = fmShwPID.at(subevtId);
            //sort here since the vector may not be
            sort(shwpids.begin(),shwpids.end(), slid::CompareByEnergy);
            art::FindOneP<rb::Shower> fos(shwpids, evt, fSPidLabel);
            for(unsigned int shwId = 0; shwId < shwpids.size(); ++shwId){
              rec.sel.elecid.shwpid.push_back(SRSPid());
              FillSpidVars(*shwpids[shwId], rec.sel.elecid.shwpid.back());
            } //end for loop on showerpids
          } //end if showerpids are valid

        } //end if event id is valid
        else{
          rec.sel.elecid    .setDefault();
          rec.energy.elecid .setDefault();
        }
      } //end if slice lids are valid
      else{
        rec.sel.elecid    .setDefault();
        rec.energy.elecid .setDefault();
      }
      
      rec.shw.fillSizes();
      rec.energy.fillSizes();
      rec.sel.elecid.fillSizes();
      
      lem::PIDDetails lem;
      if(GetAssociatedProduct(fmLem, subevtId, lem))
        FillLEMVars(lem, rec.sel.lem);
      else
        rec.sel.lem.setDefault();
      
      rvp::RVP rvp;
      if(GetAssociatedProduct(fmRvp, subevtId, rvp))
        FillRVPVars(rvp, rec.sel.rvp);
      else
        rec.sel.rvp.setDefault();

      numue::NumuE numuE;
      if(GetAssociatedProduct(fmNumuEnergy, subevtId, numuE)){
        FillNumuSimpVars(numuE, rec.energy.numusimp);
        if(rec.energy.numusimp.trkccE > 0) {
          if(bt->HaveTruthInfo()) rec.energy.numusimp.shiftedtrkccE = rec.energy.numusimp.trkccE;
          else rec.energy.numusimp.shiftedtrkccE = rec.energy.numusimp.trkccE - (0.19*(rec.energy.numusimp.trkccE-rec.energy.mutrkE[rec.sel.remid.bestidx].E));
        }
      }
      else{
        rec.energy.numusimp.setDefault();
      }

      bpfit::BPFEnergy bpfE;
      if(GetAssociatedProduct(fmBPFEnergy, subevtId, bpfE)){
        FillBPFEnergyVars(bpfE, rec.energy.bpfenergy);
      }
      else{
        rec.energy.bpfenergy.setDefault();
      }

      cosrej::CosRejObj cosrej;
      if(GetAssociatedProduct(fmCosRej, subevtId, cosrej)){
        FillCosRejVars(cosrej, rec.sel.cosrej, rec.sel.contain);

        if(fDet == kFARDET) {
          if(cosrej.CosFwdCell() >  0 && cosrej.CosBakCell() > 0 &&
             cosrej.KalFwdCell() > 10 && cosrej.KalBakCell() > 10 &&
             rec.slc.ncellsfromedge > 1 &&
             rec.sel.contain.nplanestofront > 1 &&
             rec.sel.contain.nplanestoback > 1){
            rec.sel.contain.numucontain = true;
          }
        }
        else if(fDet == kNEARDET) {
          if(cosrej.KalFwdCellND() > 4 && cosrej.KalBakCellND() > 8 &&
             rec.slc.ncellsfromedge > 1 && rec.slc.firstplane > 1 &&
             rec.slc.lastplane < 212 && rec.trk.kalman[rec.sel.remid.bestidx].start.Z() < 1150 &&
             (rec.trk.kalman[rec.sel.remid.bestidx].stop.Z()<1275||rec.sel.contain.kalyposattrans<55) &&
             (rec.energy.numusimp.ndhadcaltranE+rec.energy.numusimp.ndhadcalcatE)<0.03) {
            rec.sel.contain.numucontain = true;
          }
        }
        else rec.sel.contain.numucontain = false;
      }
      else{
        rec.sel.cosrej.setDefault();
      }

      cosrej::NueCosRej nuecosrej;
      if(GetAssociatedProduct(fmNueCosRej, subevtId, nuecosrej)){
        FillNueCosRejVars(nuecosrej, rec.sel.nuecosrej);

        // nue containment boolean - nue group please feel free to update.  If
        // variables move from sandbox this needs to be changed too
        if(fDet == kFARDET){
          if(nuecosrej.ProngMaxX() < 750 && nuecosrej.ProngMinX() > -750 &&
             nuecosrej.ProngMaxY() < 750 && nuecosrej.ProngMinY() > -750 &&
             std::min(nuecosrej.StartDistToFront(),
                      nuecosrej.StopDistToFront()) > 50 &&
             std::min(nuecosrej.StartDistToBack(),
                      nuecosrej.StopDistToBack()) > 50){
            rec.sel.contain.nuecontain = true;
          }
        }

        if(fDet == kNEARDET){
          if(nuecosrej.ProngMaxX() < 185 && nuecosrej.ProngMinX() > -185 &&
             nuecosrej.ProngMaxY() < 185 && nuecosrej.ProngMinY() > -185 &&
             std::min(nuecosrej.StartDistToFront(),
                      nuecosrej.StopDistToFront()) > 50 &&
             std::min(nuecosrej.StartDistToBack(),
                      nuecosrej.StopDistToBack()) > 50){
            rec.sel.contain.nuecontain = true;
          }
        }
      }
      else{
        rec.sel.nuecosrej.setDefault();
      }


      qeef::QePId qepid;
      if(GetAssociatedProduct(fmQepid, subevtId, qepid)){
        FillQepidVars(qepid, rec.sel.qepid);
      }
      else{
        rec.sel.qepid.setDefault();
      }

      presel::NuePreselObj nuepre;
      if(GetAssociatedProduct(fmNuePresel, subevtId, nuepre)){
        FillNuePreselVars(nuepre, rec.sel.nuepre);
      }
      else{
        rec.sel.nuepre.setDefault();
      }      

      presel::Veto veto;
      if(GetAssociatedProduct(fmVeto, subevtId, veto)){
        FillVetoVars(veto, rec.sel.veto);
      }
      else{
        rec.sel.veto.setDefault();
      }

      murem::MRCCParent parent;
      if(GetAssociatedProduct(fmMrccPar, subevtId, parent)){
        rec.parent.mrccpar.eff    = parent.ParentEff();
        rec.parent.mrccpar.pur    = parent.ParentPur();
        rec.parent.mrccpar.slcidx = parent.ParentIndex();
      }
      else{
        rec.parent.mrccpar.setDefault();
      }

      nuesand::NueSandObj nueSand;
      if(GetAssociatedProduct(fmNueSand, subevtId, nueSand)){
        FillNueSandVars(nueSand, rec.sand.nue);
      }

      numusand::NumuSandObj numuSand;
      if(GetAssociatedProduct(fmNumuSand, subevtId, numuSand)){
        FillNumuSandVars(numuSand, rec.sand.numu, rec.sel.contain);
      }
      else{
        rec.sand.numu.setDefault();
      }
      
      //#######################################################
      //Fill truth information
      //#######################################################
      
      // Set mc branch values to default
      rec.mc.setDefault();
      if(bt->HaveTruthInfo())
      {
        //Get all truth track hits associated with slice
        std::vector<cheat::TrackIDE> sliceTracks = bt->HitsToTrackIDE(subevts[subevtId]->AllCells());
        
        int           currentNuIdx        = 0;
        unsigned int  nNuGlobal           = 0; // count of neutrino interactions in this slice
        
        for(unsigned int iEffPur = 0; iEffPur < sliceTruthTable[subevtId].size(); ++iEffPur)
        {
          // continue if truth left no hits in slice
          if(! (sliceTruthTable[subevtId][iEffPur].nSliceHits > 0) ) continue;
          
          if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEnergy[subevtId])
          {
            cheat::NeutrinoEffPur* faveNu = &sliceTruthTable[subevtId][iEffPur];
            
            //If NOT Neutrinos
            if(! faveNu->neutrinoInt->NeutrinoSet())
            {
              // Add faveNu to allcosmics
              if(fFullTruthInfo) AddCosmicTruthToVec(faveNu, allTracks, sliceTracks, &rec.mc.allcosmics, evt);

              // Add faveNu to cosmic
              AddCosmicTruthToVec(faveNu, allTracks, sliceTracks, &rec.mc.cosmic, evt);
            }
            // Otherwise we are a neutrino
            else
            {
              // Add faveNu to allnus
              if(fFullTruthInfo) AddMCTruthToVec(faveNu, allTracks, sliceTracks, &rec.mc.allnus, evt);

              // Add faveNu to nu
              AddMCTruthToVec(faveNu, allTracks, sliceTracks, &rec.mc.nu, evt);
              rec.mc.faveidxenergy = currentNuIdx;

              if(fGeneratorInfo) {
                //Add faveNu to allgen generator info branch
                if(fFullTruthInfo) AddGenInfoToVec(faveNu, allTracks, sliceTracks, &rec.mc.allgen, evt);

                //Add faveNu to generator info branch
                AddGenInfoToVec(faveNu, allTracks, sliceTracks, &rec.mc.gen, evt);
              }

              art::PtrVector<simb::MCTruth> pv;
              pv.push_back(faveNu->neutrinoInt);
              art::FindManyP<rwgt::GENIEReweightTable> fmRW(pv, evt, fReweightLabel);
              rwgt::GENIEReweightTable rw;
              if(GetAssociatedProduct(fmRW, 0, rw)) rec.mc.rwgt = GenieReweightTable(rw);

              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEff[subevtId])
                rec.mc.faveidxeff        = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsPur[subevtId])
                rec.mc.faveidxpur        = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEffPur[subevtId])
                rec.mc.faveidxeffpur     = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEffThenPur[subevtId])
                rec.mc.faveidxeffthenpur = currentNuIdx;

              currentNuIdx++;
            }
          }
        }
        
        for(unsigned int iEffPur = 0; iEffPur < sliceTruthTable[subevtId].size(); ++iEffPur)
        {
          if(sliceTruthTable[subevtId][iEffPur].neutrinoInt->NeutrinoSet())
            nNuGlobal++;
          // continue if truth left no hits in slice
          if(! (sliceTruthTable[subevtId][iEffPur].nSliceHits > 0) ) continue;
          
          if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex != faveIdsEnergy[subevtId])
          {
            // Cosmics
            if(! sliceTruthTable[subevtId][iEffPur].neutrinoInt->NeutrinoSet())
            {
              if(fFullTruthInfo) AddCosmicTruthToVec(&sliceTruthTable[subevtId][iEffPur], allTracks, sliceTracks, &rec.mc.allcosmics,  evt);
            }
            // Neutrino
            else
            {
              if(fFullTruthInfo) AddMCTruthToVec(&sliceTruthTable[subevtId][iEffPur], allTracks, sliceTracks, &rec.mc.allnus, evt);

              if(fGeneratorInfo && fFullTruthInfo) {
                AddGenInfoToVec(&sliceTruthTable[subevtId][iEffPur], allTracks, sliceTracks, &rec.mc.allgen, evt);
              }

              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEff[subevtId])
                rec.mc.faveidxeff     = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsPur[subevtId])
                rec.mc.faveidxpur     = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEffPur[subevtId])
                rec.mc.faveidxeffpur     = currentNuIdx;
              if((int)sliceTruthTable[subevtId][iEffPur].truthColIndex == faveIdsEffThenPur[subevtId])
                rec.mc.faveidxeffthenpur = currentNuIdx;

              currentNuIdx++;
            }
          }
        }// end loop to fill all* vectors
        
        // fill size fields for mc branch
        rec.mc.nnu           =  rec.mc.nu         .size();
        rec.mc.ncosmic       =  rec.mc.cosmic     .size();
        rec.mc.ngen          =  rec.mc.gen        .size();
        if(fFullTruthInfo){
          rec.mc.nallnus     =  rec.mc.allnus     .size();
          rec.mc.nallcosmics =  rec.mc.allcosmics .size();
          rec.mc.nallgen     =  rec.mc.allgen     .size();
        }
        rec.mc.global.nnu    =  nNuGlobal;
      }//end MC block
      
      if (fEnableBlindness) BlindThisRecord(&rec);
        
      fRecTree->Fill();
      srcol->push_back(rec);
    }// end loop over slices

    *spillcol = spill;

    evt.put(std::move(srcol));
    evt.put(std::move(spillcol));
  }
  
  
  void CAFMaker::endSubRun(art::SubRun& sr)
  {
    art::Handle<sumdata::POTSum> pot;
    //if this is mc get the spill info from the generator
    if (bt->HaveTruthInfo())
    {
      sr.getByLabel(fGeneratorLabel, pot);
    }
    //if we have real data try to get the spill info from the NuMI beam
    else {
      sr.getByLabel(fNumiBeamLabel, pot);
    }
    
    if(!pot.failedToGet())
    {
      //NOTE: The pot that comes out of IFDBSpillInfo has an implied factor of 1E12 that is being left off
      //For now the data POT is being multiplied by this factor so that the number is consistent with what we get in MC
      //Is this what we want to do in the future?
      float subrunpot = pot->totgoodpot;
      if (!bt->HaveTruthInfo()) subrunpot *= 1e12;
      fTotalPOT += subrunpot;
      
      mf::LogInfo("CAFMaker") << "POT in subrun " << sr.subRun()
      << " = " << pot->totgoodpot;
    }
    
    art::Handle<sumdata::CosmicExposure> expo;
    sr.getByLabel(fExposureLabel, expo);
    
    if(!expo.failedToGet())
    {
      fTotalLivetime += expo->totlivetime;
      mf::LogInfo("CAFMaker") << "Exposure in subrun " << sr.subRun()
      << " = " << expo->totlivetime;
    }
  }
  
  
  
  //......................................................................
  void CAFMaker::endJob()
  {
    if(fTotalEvents == 0){
      // Among other things this makes us unable to determine the file_type
      // metdata field.
      std::cerr << "No events processed in this file. Aborting rather than produce an empty CAF." << std::endl;
      abort();
    }

    fFile->cd();
    hPOT->Fill(.5, fTotalPOT);
    hEvents->Fill(.5, fTotalEvents);
    hLivetime->Fill(.5, fTotalLivetime);
    
    hPOT->Write();
    hEvents->Write();
    hLivetime->Write();
    fFile->Write();
    
    std::map<std::string, std::string> metadata = meta::MetadataManager::getInstance().GetMetadata();
    metadata["data_tier"] = fDataTier;
    if(fIsRealData) metadata["file_type"] = "importedDetector";
    else            metadata["file_type"] = "importedSimulated";
    
    // Gymnastics to mark put "restrictedbox" in nova.special for FD numi data
    if(metadata.count("online.stream"))
    {
      // Then this is a real data file and the stream exists in the metadata
      if( ! metadata["online.stream"].compare("0") &&  
           fDetID == novadaq::cnv::kFARDET)
      {
        // Then this is a numi trigger file from the far detector
        if(!fEnableBlindness)
        {
          // Then we are not blinding it, need to mark it as such in the metadata
          std::string tag("restrictedbox");


          if(metadata.count("nova.special") && 
            std::string("").compare(metadata["nova.special"]) &&
            std::string("none").compare(metadata["nova.special"]))
          { // Then the metadata is not the same as none, we need to append to it
            // Use a stringstream to jam the strings together and append the tag
            
            std::stringstream ss;
            ss.str();
            ss << metadata["nova.special"] << "-" << tag;
            metadata["nova.special"] = ss.str();
          }
          else
            metadata["nova.special"] = tag;
        }
      }
    }
    
    AddMetadataToFile(metadata);
  }
  
  //......................................................................
  double CAFMaker::SimpleOscProb(const simb::MCFlux& flux,
                                 const simb::MCNeutrino& nu) const
  {
    if(nu.CCNC() == 1 && nu.Nu().PdgCode() == flux.fntype)
    {
      // NC
      return 1;
    }
    if(nu.CCNC() == 1 && nu.Nu().PdgCode() != flux.fntype)
    {
      // Swapped NC. Doesn't exist. If people want to play clever reweighting
      // games to reclaim these, they can figure that out, but this allows the
      // naive implementation to work.
      return 0;
    }

    osc::OscCalculatorDumb calc;
    return calc.P(flux.fntype, nu.Nu().PdgCode(), nu.Nu().E());
  }
  
  //......................................................................
  std::vector<caf::SRGenieWeights> CAFMaker::
  GenieReweightTable(const rwgt::GENIEReweightTable& table) const
  {
    const unsigned int N = table.NShifts();
    
    std::vector<caf::SRGenieWeights> ret(N);
    
    for(unsigned int i = 0; i < N; ++i){
      ret[i].minus2sigma = table.Minus2Sigma(i);
      ret[i].minus1sigma = table.Minus1Sigma(i);
      ret[i].plus1sigma = table.Plus1Sigma(i);
      ret[i].plus2sigma = table.Plus2Sigma(i);
    }
    
    return ret;
  }

  //......................................................................
  void CAFMaker::FillSpillVars(art::Handle<sumdata::SpillData> spillPot,
                               art::Handle<sumdata::EventQuality> spillQual,
                               art::Handle<std::vector<rawdata::RawTrigger>> trigs,
                               SRSpill& spill,
                               art::Event& evt) const
  {
    //Add in run, subrun, evt information so that the spill tree can be
    //mapped to the record tree.
    spill.run    = evt.run();
    spill.subrun = evt.subRun();
    spill.evt    = evt.id().event();

    //if we fail to identify a spill, we want it set to bad

    // sumdata::SpillData variables
    spill.det = fDet;
    spill.ismc = bt->HaveTruthInfo();
    spill.isgoodspill = false;
    spill.spilltimesec = 0;
    spill.spilltimensec = 0;
    spill.gpsspilltimesec = 0;
    spill.gpsspilltimensec = 0;
    spill.deltaspilltimensec = 0;
    spill.spillpot = 0.0;
    spill.livetime = 0.0;
    spill.hornI = 0.0;
    spill.posx = 0.0;
    spill.posy = 0.0;
    spill.widthx = 0.0;
    spill.widthy = 0.0;

    //Calculate diblock masking.  Also in header, but needed here for mass 
    //accounting.
    unsigned int maskstatus = 0;
    unsigned int dibmask = (rh->GoodDiBlockMask()&rh->GetConfiguration());

    if(rh->GoodDiBlockMask() == ((1<<16)-1)) maskstatus = 0;
    else if(rh->GoodDiBlockMask() == ((1<<17)-1)) maskstatus = 2;
    else maskstatus = 1;

    unsigned int dibfirst = 0;
    unsigned int diblast = 0;
    if (dibmask) {
      int iD;
      iD=0; while (!((dibmask>>iD)&1)) iD++; dibfirst=iD+1;
      iD=0; while (dibmask>>iD)        iD++; diblast=iD;
    }

    spill.dibfirst        = dibfirst;
    spill.diblast         = diblast;
    spill.dibmask         = dibmask;
    spill.maskstatus      = maskstatus;

    // sumdata::EventQuality vars, filled by DQSpillFlags
    spill.nmissingdcms = 0;
    spill.fracdcm3hits = 0.0;
    spill.nouttimehits = 0;
    spill.nnoisyapds   = 0;

    spill.nmicroslices = 0;
    spill.ndcms        = 0;

    spill.nmissingdcmslg   = 0; // LiveGeometry check
    spill.nbaddcmslg       = 0; // LiveGeometry check
    spill.dcmedgematchfrac = 0;

    //need to flag each spill if the pot was good or bad
    //each slice will get the same tag

    //if there was no spill info for this slice mark it bad. log message this should not happen
    if (spillPot.failedToGet()){
      mf::LogError("CAFMaker") << "Spill Data not found for real data event";
      spill.isgoodspill = false;
    }
    //if we have a SpillData object, get all the variables
    else {
      spill.isgoodspill        = spillPot->goodbeam;
      spill.spilltimesec       = spillPot->spilltimesec;
      spill.spilltimensec      = spillPot->spilltimensec;
      spill.gpsspilltimesec    = spillPot->gpsspilltimensec;
      spill.gpsspilltimensec   = spillPot->gpsspilltimensec;
      spill.deltaspilltimensec = spillPot->deltaspilltimensec;
      spill.spillpot           = spillPot->spillpot;
      spill.hornI              = spillPot->hornI;
      spill.posx               = spillPot->posx;
      spill.posy               = spillPot->posy;
      spill.widthx             = spillPot->widthx;
      spill.widthy             = spillPot->widthy;

      if(bt->HaveTruthInfo()){
        //mc will flag each spill as good since mc does not make bad spills
        spill.isgoodspill = true;
      }
      else{
        //NOTE: there is an implied factor of 1e12 when getting IFDBSpillInfo
        //POT.  Adding this back in to avoid confusion between data and mc
        spill.spillpot *= 1e12;
      }
    }


    //if there was no spill info for this slice mark it bad. log message this should not happen
    if(spillQual.failedToGet()){
      mf::LogError("CAFMaker") << "Spill EventQuality not found for real data event";
    }
    else{
      spill.nmissingdcms = spillQual->nmissingdcms;
      spill.fracdcm3hits = spillQual->fracdcm3hits;
      spill.nouttimehits = spillQual->nouttimehits;
      spill.nnoisyapds   = spillQual->nnoisyapds;
      spill.dcmedgematchfrac = spillQual->dcmedgematchfrac;
      spill.nmicroslices = spillQual->nmicroslices;
    }

    if(trigs.failedToGet()){
      mf::LogError("CAFMaker") << "No trigger info found for this event";
    }
    else{
      bool once = true;
      for(const rawdata::RawTrigger& trig: *trigs){
        // Units of trigger field are 500ns
        spill.livetime += trig.fTriggerRange_TriggerLength*500e-9;
        if(once){
          once = false;
          spill.trigger = trig.fTriggerMask_TriggerType;
        }
        else{
          if(spill.trigger != trig.fTriggerMask_TriggerType){
            mf::LogError("CAFMaker") << "Multiple trigger types found in one Event. Storing the first value seen.";
          }
        }
      }
    }

    spill.nmissingdcmslg = livegeom->NDropouts();
    spill.nbaddcmslg     = livegeom->NBadDCMBC();
    spill.ndcms          = rh->NDCMs();
  }

  //......................................................................
  void CAFMaker::FillMuIdVars(const remid::ReMId& remid,
                              const rb::Track& track,
                              caf::SRMuId& muid) const
  {
    muid.pdg      = remid.Pdg();
    muid.pid      = remid.Value();
    muid.scatllh  = remid.ScatSeparation();
    muid.dedxllh  = remid.DedxSeparation();
    muid.len      = track.TotalLength();
    muid.measfrac = remid.MeasurementFraction();
    muid.cont     = remid.Contained();
  }

  //......................................................................
  void CAFMaker::FillJMShowerVars(const jmshower::JMShower& shw,
                                  caf::SRJMShower& srshw) const
  {
    srshw.nhit  = shw.NCell();
    srshw.len   = shw.TotalLength();
    srshw.width = shw.Radius();
    srshw.calE  = shw.CalorimetricEnergy();
    srshw.vtxE  = shw.VtxGeV();
    srshw.shwE  = shw.Energy();
    srshw.start = shw.Start();
    srshw.dir   = shw.Dir();
    srshw.stop  = shw.Stop();

    if      (shw.NXCell() == 0)  srshw.view = kY;
    else if (shw.NYCell() == 0)  srshw.view = kX;
    else                         srshw.view = kXorY;
  }

  //......................................................................
  void CAFMaker::FillJMEIDVars(const jmshower::EID& eid,
                               caf::SRJMEid& sreid) const
  {
    sreid.ann       = eid.ANN();
    sreid.annE      = eid.ANNE();
    sreid.annNoCos  = eid.ANNNoCos();
    sreid.annENoCos = eid.ANNENoCos();
    sreid.ismuon   = eid.IsMuon();
    sreid.eglll    = eid.EGLLL();
    sreid.egllt    = eid.EGLLT();
    sreid.emulll   = eid.EMuLLL();
    sreid.emullt   = eid.EMuLLT();
    sreid.epi0lll  = eid.EPi0LLL();
    sreid.epi0llt  = eid.EPi0LLT();
    sreid.eplll    = eid.EPLLL();
    sreid.epllt    = eid.EPLLT();
    sreid.enlll    = eid.ENLLL();
    sreid.enllt    = eid.ENLLT();
    sreid.epilll   = eid.EPiLLL();
    sreid.epillt   = eid.EPiLLT();
    sreid.vtxgev   = eid.VtxGeV();
    sreid.pi0mass  = eid.Pi0Mass();
    sreid.shwEFrac = eid.ShE();
    sreid.gap      = eid.Gap();
    sreid.elll     = eid.ELLL();
    sreid.ellt     = eid.ELLT();
    sreid.mulll    = eid.MuLLL();
    sreid.mullt    = eid.MuLLT();
  }

  //......................................................................
  void CAFMaker::FillNueSandEIDVars(const jmshower::EID& eid,
                                    caf::SRNueSandbox& sand) const
  {
    sand.vtx   = eid.Vtx();

    sand.E     = eid.NuEnergy();
    sand.depE  = eid.DetEnergy();
    sand.shwE  = eid.Sh1Energy();

    sand.sh1totalL   = eid.Sh1TotalLength();
    sand.sh1start    = eid.Sh1Start();
    sand.sh1stop     = eid.Sh1Stop();
    sand.sh1dir      = eid.Sh1Dir();

    sand.evtncell = eid.EvtNCell();
    sand.evtsumcostheta = eid.EvtSumCosTheta();
    sand.evtsump = eid.EvtSumP();
    sand.evtsumpt = eid.EvtSumPt();
    sand.evtsump0 = eid.EvtSumP0();
    sand.evtsumpt0 = eid.EvtSumPt0();
    sand.evtetot = eid.EvtEtot();
    sand.evtminx = eid.EvtMinX();
    sand.evtminy = eid.EvtMinY();
    sand.evtminz = eid.EvtMinZ();
    sand.evtmaxx = eid.EvtMaxX();
    sand.evtmaxy = eid.EvtMaxY();
    sand.evtmaxz = eid.EvtMaxZ();

    sand.evtgaptns = eid.EvtGapTNS();
    sand.evtncelltoedge = eid.EvtNCellToEdge();

    sand.sh1energy = eid.Sh1Energy();
    sand.sh1exclenergy = eid.Sh1ExclEnergy();
    sand.sh1totalL = eid.Sh1TotalLength();
    sand.sh1start = eid.Sh1Start();
    sand.sh1stop = eid.Sh1Stop();
    sand.sh1dir = eid.Sh1Dir();
    sand.sh1vtxdoca = eid.Sh1VtxDoca();
    sand.sh1gap = eid.Sh1Gap();
    sand.sh1nplane = eid.Sh1NPlane();
    sand.sh1xnplane = eid.Sh1XNPlane();
    sand.sh1ynplane = eid.Sh1YNPlane();
    sand.sh1ncell = eid.Sh1NCell();
    sand.sh1xncell = eid.Sh1XNCell();
    sand.sh1yncell = eid.Sh1YNCell();
    sand.sh1pid = eid.Sh1PID();
    sand.sh1ncelltoedge = eid.Sh1NCellToEdge();
    sand.sh1sh2dang = eid.Sh1Sh2Dang();

    sand.sh2energy = eid.Sh2Energy();
    sand.sh2exclenergy = eid.Sh2ExclEnergy();
    sand.sh2totalL = eid.Sh2TotalLength();
    sand.sh2start = eid.Sh2Start();
    sand.sh2stop = eid.Sh2Stop();
    sand.sh2dir = eid.Sh2Dir();
    sand.sh2vtxdoca = eid.Sh2VtxDoca();
    sand.sh2gap = eid.Sh2Gap();
    sand.sh2nplane = eid.Sh2NPlane();
    sand.sh2xnplane = eid.Sh2XNPlane();
    sand.sh2ynplane = eid.Sh2YNPlane();
    sand.sh2ncell = eid.Sh2NCell();
    sand.sh2xncell = eid.Sh2XNCell();
    sand.sh2yncell = eid.Sh2YNCell();
    sand.sh2pid = eid.Sh2PID();
    sand.sh2ncelltoedge = eid.Sh2NCellToEdge();

    sand.elll        = eid.ELLL();
    sand.sh1d2edge   = eid.Sh1DistToEdge();
    sand.eedge2cell  = eid.EEdge2Cell();
    sand.eedge5cell  = eid.EEdge5Cell();
    sand.eedge10cell = eid.EEdge10Cell();
  }

  //......................................................................
  void CAFMaker::FillShowerVars(const rb::Shower& shw,
                                const slid::ShowerLID& slid,
                                caf::SRShower& srshw) const
  {
    srshw.nhit         = shw.NCell();
    srshw.nhitx        = shw.NXCell();
    srshw.nhity        = shw.NYCell();
    srshw.nplane       = shw.ExtentPlane();
    srshw.nplanex      = shw.ExtentPlane(geo::kX);
    srshw.nplaney      = shw.ExtentPlane(geo::kY);
    srshw.gap          = slid.Gap();
    srshw.maxplanecont = shw.MostContiguousPlanes(geo::kXorY);
    srshw.maxplanegap  = shw.MostMissingPlanes(geo::kXorY);
    srshw.len          = shw.TotalLength();
    srshw.width        = slid.Radius();
    srshw.calE         = shw.CalorimetricEnergy();
    srshw.start        = shw.Start();
    srshw.dir          = shw.Dir();
    srshw.stop         = shw.Stop();
    if      (shw.NXCell() == 0)  srshw.view = kY;
    else if (shw.NYCell() == 0)  srshw.view = kX;
    else                         srshw.view = kXorY;
  }

  //......................................................................
  void CAFMaker::FillSlidVars(const slid::ShowerLID& slid,
                              caf::SRSLid& srslid) const
  {
    srslid.ann      = slid.Value();
    srslid.anne     = slid.ValueWithE();
    srslid.ismuon   = slid.IsMuon();
    srslid.eglll    = slid.EGLLL();
    srslid.egllt    = slid.EGLLT();
    srslid.emulll   = slid.EMuLLL();
    srslid.emullt   = slid.EMuLLT();
    srslid.epi0lll  = slid.EPi0LLL();
    srslid.epi0llt  = slid.EPi0LLT();
    srslid.eplll    = slid.EPLLL();
    srslid.epllt    = slid.EPLLT();
    srslid.enlll    = slid.ENLLL();
    srslid.enllt    = slid.ENLLT();
    srslid.epilll   = slid.EPiLLL();
    srslid.epillt   = slid.EPiLLT();
    srslid.vtxgev   = slid.VertexEnergy();
    srslid.pi0mass  = slid.Pi0mass();
    srslid.shwEFrac = slid.ShowerEFrac();
    srslid.gap      = slid.Gap();
    srslid.costheta = slid.CosTheta();

    std::map<int, float> tPartLongLL = slid.fPartLongLL;
    std::map<int, float> tPartTransLL = slid.fPartTransLL;

    std::map<int, float>::iterator itr;

    itr = tPartLongLL.find(slid::DedxParticleType::kELECTRON);
    if (itr != tPartLongLL.end()) srslid.elll = itr->second;
    else srslid.elll = -5;

    itr = tPartTransLL.find(slid::DedxParticleType::kELECTRON);
    if (itr != tPartTransLL.end()) srslid.ellt = itr->second;
    else srslid.ellt = -5;

    itr = tPartLongLL.find(slid::DedxParticleType::kMUON);
    if (itr != tPartLongLL.end()) srslid.mulll = itr->second;
    else srslid.mulll = -5;

    itr = tPartTransLL.find(slid::DedxParticleType::kMUON);
    if (itr != tPartTransLL.end()) srslid.mullt = itr->second;
    else srslid.mullt = -5;

    srslid.invglll  = slid.InverseGLLL();
  }

  void CAFMaker::FillSpidVars(const slid::ShowerPID& spid,
                              caf::SRSPid& srspid) const
  {
    srspid.annepi0  = spid.ValueEPi0();
    srspid.annepi0el= spid.ValueEPi0EL();

    srspid.elll = spid.ELLL();
    srspid.ellt = spid.ELLT();
    srspid.glll = spid.GLLL();
    srspid.gllt = spid.GLLT();
    srspid.mulll = spid.MuLLL();
    srspid.mullt = spid.MuLLT();
    srspid.pi0lll =spid.Pi0LLL();
    srspid.pi0llt = spid.Pi0LLT();
    srspid.plll = spid.PLLL();
    srspid.pllt = spid.PLLT();
    srspid.nlll = spid.NLLL();
    srspid.nllt = spid.NLLT();
    srspid.pilll = spid.PiLLL();
    srspid.pillt = spid.PiLLT();
  }



  //......................................................................
  void CAFMaker::FillLIDEnergyVars(const slid::ShowerLID& slid,
                                   caf::SRSLidEnergy& en) const
  {
    en.E    = slid.NueEnergy();
    en.depE = slid.ShowerDepEnergy();
    en.shwE = slid.ShowerEnergy();
    en.hadE = slid.HadronicEnergy();
  }

  //......................................................................
  void CAFMaker::FillLEMVars(const lem::PIDDetails& lem,
                             caf::SRLem& srlem) const
  {
    srlem.pid               = lem.Value();
    srlem.pidfit            = lem.vars.fPIDFit;
    srlem.pidexp            = lem.vars.fPIDExp;
    srlem.pidexpfit         = lem.vars.fPIDExpFit;
    srlem.meany             = lem.vars.fMeanY;
    srlem.meanyfit          = lem.vars.fMeanYFit;
    srlem.meanyexp          = lem.vars.fMeanYExp;
    srlem.meanysig          = lem.vars.fMeanYSig;
    srlem.meanysigexp       = lem.vars.fMeanYSigExp;
    srlem.meanqfrac         = lem.vars.fMeanQFrac;
    srlem.meanqfracfit      = lem.vars.fMeanQFracFit;
    srlem.meanqfracexp      = lem.vars.fMeanQFracExp;
    srlem.meanqfracsig      = lem.vars.fMeanQFracSig;
    srlem.meanqfracsigexp   = lem.vars.fMeanQFracSigExp;
    srlem.meanqfracbkg      = lem.vars.fMeanQFracBkg;
    srlem.meanqfracbkgexp   = lem.vars.fMeanQFracBkgExp;
    srlem.energydiff        = lem.vars.fEnergyDiff;
    srlem.energydiffexp     = lem.vars.fEnergyDiffExp;
    srlem.enrichfrac        = lem.vars.fEnrichFrac;
    srlem.enrichfracexp     = lem.vars.fEnrichFracExp;
    srlem.fannid            = lem.vars.fFannId;
    srlem.fannidenrich      = lem.vars.fFannIdEnrich;
    srlem.chisig            = lem.vars.fChiSig;
    srlem.chibkg            = lem.vars.fChiBkg;
    srlem.chisigexp         = lem.vars.fChiSigExp;
    srlem.chibkgexp         = lem.vars.fChiBkgExp;
    srlem.avginvE           = lem.vars.fAvgInvE;
    srlem.avgexpE           = lem.vars.fAvgExpE;
    srlem.avgsigE           = lem.vars.fAvgSigE;
    srlem.tmvabdtg          = lem.vars.fTMVABDTG;
    srlem.tmvabdt           = lem.vars.fTMVABDT;
    srlem.tmvabdtd          = lem.vars.fTMVABDTD;
    srlem.tmvabdtmf         = lem.vars.fTMVABDTMF;
    srlem.tmvamlp           = lem.vars.fTMVAMLP;
    srlem.dectree           = lem.vars.fDecTreePID;
  }

  //......................................................................
  void CAFMaker::FillRVPVars(const rvp::RVP& rvp, caf::SRRvp& srrvp) const
  {
    srrvp.pid         = rvp.Value();
    srrvp.ncell       = rvp.NCells();
    srrvp.recoE       = rvp.RecoE();
    srrvp.longtr      = rvp.LongTr();
    srrvp.longtrfrac  = rvp.LongTrFrac();
    srrvp.mipfrac     = rvp.MipFrac();
    srrvp.miphits     = rvp.MipHits();
    srrvp.epl20frac   = rvp.EPl20Frac();
    srrvp.efrac2plwin = rvp.EFrac2PlWin();
    srrvp.efrac6plwin = rvp.EFrac6PlWin();
    srrvp.efrac2sigrd = rvp.EFrac2SigRd();
    srrvp.prongs3D    = rvp.Prongs3D();
    srrvp.prongEbal3D = rvp.ProngEBal3D();
    srrvp.prongs2D    = rvp.Prongs2D();
    srrvp.prongEbal2D = rvp.ProngEBal2D();
    srrvp.eiso3sig    = rvp.EIso3Sig();
    srrvp.rvp12       = rvp.RVP12();
  }

  //......................................................................
  void CAFMaker::FillCosRejVars(const cosrej::CosRejObj& cosrej,
                                caf::SRCosRej& srcosrej,
                                caf::SRContain& srcontain) const
  {
    srcosrej.anglekal      = cosrej.AngleKal();
    srcosrej.anglecos      = cosrej.AngleCos();
    srcosrej.numucontpid   = cosrej.ConCosPID();
    srcosrej.numuoldpid    = cosrej.OldCosPID();
    srcosrej.numuprepid    = cosrej.PreCosPID();
    srcosrej.numupostpid   = cosrej.PostCosPID();
    srcosrej.numuuncontpid = cosrej.UnconCosPID();
    srcosrej.nkal3d        = cosrej.NTracks3D();
    srcosrej.mindist       = cosrej.MinDist();
    srcosrej.mincell       = cosrej.MinCell();
    srcosrej.eratio        = cosrej.ERatio();
    srcosrej.scatt         = cosrej.Scatt();
    srcosrej.hadE          = cosrej.StartAct();
    srcosrej.costhetatrue  = cosrej.CosThetaTrue();
    srcosrej.kalthetatrue  = cosrej.KalThetaTrue();
    srcosrej.pdgbest       = cosrej.PDGBest();
    srcosrej.fscattmax     = cosrej.FScattMax();
    srcosrej.fscattsum     = cosrej.FScattSum();
    srcosrej.fscattext     = cosrej.FScattExt();
    srcosrej.fscattsig     = cosrej.FScattSig();
    srcosrej.kfitspeed    = cosrej.KFitSpeed();
    srcosrej.cfitspeed    = cosrej.CFitSpeed();
    srcosrej.kdirscore    = cosrej.KDirScore();
    srcosrej.cdirscore    = cosrej.CDirScore();
    srcosrej.kscorediff   = cosrej.KScoreDiff();
    srcosrej.cscorediff   = cosrej.CScoreDiff();

    srcontain.vtxdist       = cosrej.VtxDist();
    srcontain.enddist       = cosrej.EndDist();
    srcontain.cosfwddist    = cosrej.CosFwdDist();
    srcontain.cosbakdist    = cosrej.CosBakDist();
    srcontain.cosfwdcell    = cosrej.CosFwdCell();
    srcontain.cosfwdcellnd  = cosrej.CosFwdCellND();
    srcontain.cosbakcell    = cosrej.CosBakCell();
    srcontain.cosbakcellnd  = cosrej.CosBakCellND();
    srcontain.cosfwdair     = cosrej.CosFwdAir();
    srcontain.cosfwdsteel   = cosrej.CosFwdSteel();
    srcontain.cosbakair     = cosrej.CosBakAir();
    srcontain.cosbaksteel   = cosrej.CosBakSteel();
    srcontain.kalfwdair     = cosrej.KalFwdAir();
    srcontain.kalfwdsteel   = cosrej.KalFwdSteel();
    srcontain.kalbakair     = cosrej.KalBakAir();
    srcontain.kalbaksteel   = cosrej.KalBakSteel();
    srcontain.kalfwddist    = cosrej.KalFwdDist();
    srcontain.kalbakdist    = cosrej.KalBakDist();
    srcontain.kalfwdcell    = cosrej.KalFwdCell();
    srcontain.kalfwdcellnd  = cosrej.KalFwdCellND();
    srcontain.kalbakcell    = cosrej.KalBakCell();
    srcontain.kalbakcellnd  = cosrej.KalBakCellND();
    srcontain.cosyposattrans= cosrej.CosYPosAtTrans();
    srcontain.kalyposattrans= cosrej.KalYPosAtTrans();
  }

  //......................................................................
  void CAFMaker::FillNueCosRejVars(const cosrej::NueCosRej& nuecosrej,
                                   caf::SRNueCosRej& srnuecosrej) const
  {
    srnuecosrej.hitsperplane = nuecosrej.HitsPerPlane();
    srnuecosrej.pngptp       = nuecosrej.ProngTransMom();
    srnuecosrej.partptp      = nuecosrej.ParticleShowerTransMom();
    srnuecosrej.photptp      = nuecosrej.PhotonShowerTransMom();
    srnuecosrej.starttop     = nuecosrej.StartDistToTop();
    srnuecosrej.startbottom  = nuecosrej.StartDistToBottom();
    srnuecosrej.startfront   = nuecosrej.StartDistToFront();
    srnuecosrej.startback    = nuecosrej.StartDistToBack();
    srnuecosrej.startwest    = nuecosrej.StartDistToWest();
    srnuecosrej.starteast    = nuecosrej.StartDistToEast();
    srnuecosrej.stoptop      = nuecosrej.StopDistToTop();
    srnuecosrej.stopbottom   = nuecosrej.StopDistToBottom();
    srnuecosrej.stopfront    = nuecosrej.StopDistToFront();
    srnuecosrej.stopback     = nuecosrej.StopDistToBack();
    srnuecosrej.stopwest     = nuecosrej.StopDistToWest();
    srnuecosrej.stopeast     = nuecosrej.StopDistToEast();
    srnuecosrej.cosdang      = nuecosrej.CosAngleToNextProng();
    srnuecosrej.vtxdoca	     = nuecosrej.ProngDistToVtx();
    srnuecosrej.prongmaxx    = nuecosrej.ProngMaxX();
    srnuecosrej.prongmaxy    = nuecosrej.ProngMaxY();
    srnuecosrej.prongmaxz    = nuecosrej.ProngMaxZ();
    srnuecosrej.prongminx    = nuecosrej.ProngMinX();
    srnuecosrej.prongminy    = nuecosrej.ProngMinY();
    srnuecosrej.prongminz    = nuecosrej.ProngMinZ();
  }

  //......................................................................
  void CAFMaker::FillQepidVars(const qeef::QePId& qepid,
                               caf::SRQepid& srqepid) const
  {
    srqepid.pdg    = qepid.Pdg();
    srqepid.mode   = qepid.Mode();
    srqepid.ntrk   = qepid.Ntrk();
    srqepid.pid    = qepid.Value();
    srqepid.offE   = qepid.OffTrkE();
    srqepid.ediff  = qepid.Ediff();
    srqepid.ediffz = qepid.EdiffZ();
    srqepid.dedx   = qepid.Dedx();
  }

  //......................................................................
  void CAFMaker::FillNuePreselVars(const presel::NuePreselObj& nuepre,
                              caf::SRNuePresel& srnuepre) const
  {
    srnuepre.passpre = nuepre.PassNuePresel();
  }

  //......................................................................
  void CAFMaker::FillVetoVars(const presel::Veto& veto,
                              caf::SRVeto& srveto) const
  {
    srveto.ncell      = veto.NCell();
    srveto.fwddist    = veto.FwdDist();
    srveto.bakdist    = veto.BakDist();
    srveto.ratio      = veto.Ratio();
    srveto.diry       = veto.DirY();
    srveto.angle      = veto.Angle();
    srveto.anglevar   = veto.AngleVar();
    srveto.anglevarold= veto.AngleVarOld();
    srveto.passthru   = veto.PassThru();
    srveto.passthruold= veto.PassThruOld();
    srveto.passangle  = veto.PassAngle();
    srveto.passanglefirstanal = veto.PassAngleFirstAnal();
    srveto.keep       = veto.Keep();
  }

  //......................................................................
  void CAFMaker::FillNumuSimpVars(const numue::NumuE& e,
                                  caf::SRNumuEnergy& sre) const
  {
    sre.E             =  e.E();
    sre.calccE        =  e.CalCCE();
    sre.trkqeE        =  e.TrkQEE();
    sre.trknonqeE     =  e.TrkNonQEE();
    sre.trkccE        =  e.TrkCCE();
    sre.angleE        =  e.AngleQEE();
    sre.angleerror    =  e.AngleQEError();
    sre.hadcalE       =  e.HadCalE();
    sre.hadtrkE       =  e.HadTrkE();
    sre.ucccE         =  e.UCCCE();
    sre.fsE           =  e.FScatterE();
    sre.ndtrklenact   =  e.NDTrkLenAct();
    sre.ndtrklencat   =  e.NDTrkLenCat();
    sre.ndtrkcalactE  =  e.NDTrkCalAct();
    sre.ndtrkcaltranE =  e.NDTrkCalTran();
    sre.ndtrkcalcatE  =  e.NDTrkCalCat();
    sre.ndhadcalactE  =  e.NDHadCalAct();
    sre.ndhadcaltranE =  e.NDHadCalTran();
    sre.ndhadcalcatE  =  e.NDHadCalCat();
    sre.ndhadtrkactE  =  e.NDHadTrkAct();
    sre.ndhadtrktranE =  e.NDHadTrkTran();
    sre.ndhadtrkcatE  =  e.NDHadTrkCat();
    sre.ndtrktranx    =  e.NDTrkTranX();
    sre.ndtrktrany    =  e.NDTrkTranY();
    FillSliceVars(e.HadCluster(),sre.hadclust,true);
  }

  //......................................................................
  void CAFMaker::FillBPFEnergyVars(const bpfit::BPFEnergy& e,
				   caf::SRBPFEnergy& sre) const
  {
    sre.E1        = e.E1();
    sre.Eres1     = e.Eres1();
    sre.E2        = e.E2();
    sre.Eres2     = e.Eres2();
    sre.E3        = e.E3();
    sre.Eres3     = e.Eres3();
    sre.EventID   = e.EventID();
    sre.PMuon     = e.PMuon();
    sre.DirZMuon  = e.DirZMuon();
    sre.N3DProngs = e.N3DProngs();
    sre.EFuzzyK3D = e.EFuzzyK3D();
    sre.ERemain   = e.ERemain();
    sre.SumPE     = e.SumPE();
  }

  //......................................................................
  void CAFMaker::FillNueSandVars(const nuesand::NueSandObj& sand,
                                 caf::SRNueSandbox& srsand) const
  {
    srsand.infid      = sand.InFiducialVolume();
    srsand.slicemeanx = sand.SliceMeanPosition().X();
    srsand.slicemeany = sand.SliceMeanPosition().Y();
    srsand.slicemeanz = sand.SliceMeanPosition().Z();

    srsand.npi0 = sand.fNPi0;
    srsand.npip = sand.fNPiPlus;
    srsand.npim = sand.fNPiMinus;

    srsand.pi0E = sand.fPi0Energy;
    srsand.pipE = sand.fPiPlusEnergy;
    srsand.pimE = sand.fPiMinusEnergy;

    srsand.phot0E  = sand.fPhot0E;
    srsand.phot1E  = sand.fPhot1E;
    srsand.opencos = sand.fOpenCos;
    srsand.conv0   = sand.fConvL0;
    srsand.conv1   = sand.fConvL1;

    srsand.mincellcosmic     = sand.fMincellCosmic;
    srsand.mincellwallcosmic = sand.fMincellWallCosmic;
    srsand.anglecosmic       = sand.fAngleCosmic;
    srsand.nplanesfilled     = sand.fFilledPlanes;
    srsand.nxplanes          = sand.fExtentXPlanes;
    srsand.nyplanes          = sand.fExtentYPlanes;
    srsand.ncellrow          = sand.fMaxRowOfCells;
    srsand.ncellsum          = sand.fMaxCellsInPlane;
    srsand.mincellavg        = sand.fMincellAvg;
    srsand.mincellkalman     = sand.fMincellKalman;
    srsand.nmiphits          = sand.fNMIPHits;
    srsand.mipmin            = sand.fMIPMin;
    srsand.mipmax            = sand.fMIPMax;

    srsand.fracangchanges      = sand.fFracAngChanges;
    srsand.fracangchangesloose = sand.fFracAngChangesLoose;
    srsand.fracmodalhits       = sand.fFracModalHits;
    srsand.fracnoncontig       = sand.fFracNonContig;
    srsand.fracnoncontigloose  = sand.fFracNonContigLoose;
    srsand.maxgap              = sand.fMaxGap;

    srsand.ecfNu = sand.ECFNu;
    
    srsand.dedxpng1 = sand.fdEdxProng1;
    srsand.dedxpng2 = sand.fdEdxProng2;
  }

  //......................................................................
  void CAFMaker::FillNumuSandVars(const numusand::NumuSandObj& sand,
                                  caf::SRNumuSandbox& srsand,
                                  caf::SRContain& srcontain) const
  {
    srsand.pimudecay = sand.PiMuDecay();
    srsand.nprotons  = sand.NProtons();
    srsand.nmutrks   = sand.NMuTrks();
    srsand.pdgbest   = sand.PDGBest();
    srsand.pdgsecondbest   = sand.PDGSecondBest();
    srsand.ndropouts = sand.NDropped();
    srsand.nbaddcms  = sand.NBad();

    srcontain.missE   = sand.MissingE();

    srcontain.xmin2   = sand.XMin2();
    srcontain.ymin2   = sand.YMin2();
    srcontain.zmin2   = sand.ZMin2();
    srcontain.xmax2   = sand.XMax2();
    srcontain.ymax2   = sand.YMax2();
    srcontain.zmax2   = sand.ZMax2();

    srcontain.xmint   = sand.XMint();
    srcontain.ymint   = sand.YMint();
    srcontain.zmint   = sand.ZMint();
    srcontain.xmaxt   = sand.XMaxt();
    srcontain.ymaxt   = sand.YMaxt();
    srcontain.zmaxt   = sand.ZMaxt();
  }

  //......................................................................
  void CAFMaker::CopyRemidVars(const caf::SRMuId& muid, caf::SRRemid& remid)
  {
    remid.pdg      = muid.pdg;
    remid.pid      = muid.pid;
    remid.scatllh  = muid.scatllh;
    remid.dedxllh  = muid.dedxllh;
    remid.len      = muid.len;

    remid.measfrac = muid.measfrac;
    remid.cont     = muid.cont;
  }

  DEFINE_ART_MODULE(CAFMaker);
  
} // end namespace caf
////////////////////////////////////////////////////////////////////////
