///////////////////////////////////////////////////////////////////////
/// \brief   A simple module to monitor the performance of Calhit. 
/// \author  Kanika Sachdev (ksachdev@physics.umn.edu)
/// \date    Dec. 2012
////////////////////////////////////////////////////////////////////////

#include "RecoBase/CellHit.h"
#include "MCCheater/BackTracker.h"
#include "Geometry/Geometry.h"
#include "CalibrationUtils/CalibUtil.h"

#include "art/Framework/Core/EDAnalyzer.h"
#include <art/Framework/Principal/Handle.h>
#include <art/Framework/Principal/Event.h>
#include <art/Framework/Services/Registry/ServiceHandle.h>
#include <art/Framework/Services/Optional/TFileService.h>
#include <art/Framework/Services/Optional/TFileDirectory.h>
#include <fhiclcpp/ParameterSet.h>
#include <messagefacility/MessageLogger/MessageLogger.h>
#include "art/Framework/Core/ModuleMacros.h"

#include <vector>
#include <cmath>

#include "TTree.h"

namespace calhit{
  
  class CalHitAna : public art::EDAnalyzer {

  public:
    explicit CalHitAna(fhicl::ParameterSet const& pset);
    virtual ~CalHitAna();
    
    void analyze(art::Event const& evt);
    void beginJob();
    void reconfigure(const fhicl::ParameterSet& pset);
    
  private:
    
    TTree* fHitTree;        ///<   Tree that gets filled for every hit
    TTree* fEvtTree;        ///<   Tree that gets filled for every event
    
    float  fPe;             ///<   PE of the CellHit
    int    fAdc;            ///<   ADC0 of the CellHit
    int    fTdc;            ///<   TDC0 of the CellHit
    float  fTns;            ///<   TNS of the CellHit   
    float  fFLSTime;        ///<   Time of energy deposit in cell
    float  fReadDist;       ///<   Distance to readout APD from energy deposit (with pigtail)
    int    fNoise;          ///<   Number of noise hits in event   
    int    fPhysics;        ///<   Number of physics hits in event   
    int    fAllHits;        ///<   Number of all hits in event   
    float  fPePerTrueE;     ///<   PE per True Energy in event 
    float  fAdcPerTrueE;    ///<   ADC per True Energy in event 
    unsigned int fPlane;    ///<   Plane of the CellHit
    unsigned int fCell;     ///<   Cell of the CellHit
    
    std::string fCellHitLabel; ///< Module label name from where to grab CellHits
  };
}

/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

namespace calhit {


  CalHitAna::CalHitAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    this->reconfigure(pset);
  }
  

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  
  CalHitAna::~CalHitAna() { }
  

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////


  void CalHitAna::reconfigure(const fhicl::ParameterSet& pset){

    fCellHitLabel  = pset.get<std::string>("CellHitLabel");
  }
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  
  void CalHitAna::beginJob(){
    
    art::ServiceHandle<art::TFileService> tfs;

    fHitTree = tfs->make<TTree>("fHitTree","fHitTree");

    fHitTree->Branch("Pe",    &fPe);
    fHitTree->Branch("Adc",   &fAdc);
    fHitTree->Branch("Tdc",   &fTdc);
    fHitTree->Branch("Tns" ,  &fTns);
    fHitTree->Branch("FLSTime", &fFLSTime);
    fHitTree->Branch("ReadDist", &fReadDist);
    fHitTree->Branch("Plane", &fPlane);
    fHitTree->Branch("Cell",  &fCell);

    fEvtTree = tfs->make<TTree>("fEvtTree","fEvtTree");

    fEvtTree->Branch("Noise",      &fNoise);
    fEvtTree->Branch("PePerTrueE", &fPePerTrueE);
    fEvtTree->Branch("AdcPerTrueE",&fAdcPerTrueE);
    fEvtTree->Branch("Physics",    &fPhysics);
    fEvtTree->Branch("AllHits",    &fAllHits);
  }
  
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////



  void CalHitAna::analyze(art::Event const& evt){

    art::ServiceHandle<cheat::BackTracker> bt;

    art::Handle< std::vector< rb::CellHit> > cHitCol;
    evt.getByLabel(fCellHitLabel, cHitCol);

    // Set default values

    fPlane      = -5;
    fCell       = -5;
    fPe         = -5;
    fTdc        = -5;
    fTns        = -5;
    fAdc        = -5;
    fNoise      = -5;
    fPePerTrueE = -5;

    art::ServiceHandle<geo::Geometry> geom;

    // Fill the hit tree
    for(int i = 0; i < (int)cHitCol->size(); ++i){
      art::Ptr< rb::CellHit> hit(cHitCol, i);
      fPlane = hit->Plane();
      fCell  = hit->Cell();
      fPe    = hit->PE();
      fAdc   = hit->ADC();
      fTdc   = hit->TDC();
      fTns   = hit->TNS();
      const std::vector< sim::FLSHit> flsHits = bt->HitToFLSHit(hit);
      fReadDist = -5;
      fFLSTime = -5;
      if (flsHits.size() == 1){
        double w = 0.0;
        if (hit->View() == geo::kX){
          w = (flsHits[0].GetEntryY() + flsHits[0].GetExitY())/2.0;
        }
        if (hit->View() == geo::kY){
          w = (flsHits[0].GetEntryX() + flsHits[0].GetExitX())/2.0;
        }
        double xyzd[4];
        calib::GetXYZD(hit->OfflineChan(),w,xyzd);
        fReadDist = xyzd[3];
        fFLSTime = (flsHits[0].GetEntryT() +flsHits[0].GetExitT())/2.0;
      }
      fHitTree->Fill();
    }
    
    // If truth information is available, fill event tree
    if( bt->HaveTruthInfo()){
      fNoise      = 0;  
      fPePerTrueE = 0;
      float totPe = 0;
      float totAdc= 0;
      float totE  = 0;
      fAllHits    = cHitCol->size();

      for(int i = 0; i < fAllHits; ++i){
	art::Ptr< rb::CellHit> hit(cHitCol, i);
	
	if( bt->IsNoise(hit) ){
	  fNoise += 1;
	  continue;
	}
	else
	  fPhysics += 1;
	
	const std::vector< sim::FLSHit > flsHits = bt->HitToFLSHit(hit);
	for(int j = 0; j < (int)flsHits.size(); j++){
	  totE += flsHits[j].GetEdep();
	}//End loop over flshits in hit
	
	totPe += hit->PE();
	totAdc+= hit->ADC();
      }// End loop over hits
      
      if( totE != 0){
	fPePerTrueE  = totPe/totE;
	fAdcPerTrueE = totAdc/totE;
      }
      fEvtTree->Fill();
    }// EndIf HaveTruthInfo
    
  }//  End of analyser
  

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  
  DEFINE_ART_MODULE(CalHitAna);    
  
} //End of namespace calhit

////////////////////////////////////////////////////////////////////////
