////////////////////////////////////////////////////////////////////////
/// \brief   This module creates CellHit object
/// \author  ??
/// \date    
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Core/ModuleMacros.h"

#include "Calibrator/Calibrator.h"
#include "ChannelInfo/BadChanList.h"
#include "RawData/RawDigit.h"
#include "RecoBase/CellHit.h"
#include "RunHistory/service/RunHistoryService.h"

#include <string>

#include "art/Framework/Core/EDProducer.h"


/// Calibrating RawData to Produce CellHits                                                                                                
namespace calhit {
  
  class CalHit : public art::EDProducer {
    
  public:
    
    explicit CalHit(fhicl::ParameterSet const& pset);
    virtual ~CalHit();

    virtual void beginRun(art::Run& run);
    void produce(art::Event& evt);

  private:
 
    int         fRemoveBadChans; ///< whether to remove bad channels                                                                      
    std::string fRawDataLabel;   ///< label of where to find RawData                                                                       

  }; // class CalHit                                                                                                                       
}

namespace calhit{

  CalHit::CalHit(fhicl::ParameterSet const& pset)
    : fRemoveBadChans(pset.get< bool        >("RemoveBadChans")        )
    , fRawDataLabel  (pset.get< std::string >("RawDataLabel"  )        )
  {
    produces< std::vector<rb::CellHit> >();
    produces<int, art::InRun>();
  }

  //......................................................................
  CalHit::~CalHit() { }

  //......................................................................

  void CalHit::beginRun(art::Run& run)
  {
    art::ServiceHandle<nova::dbi::RunHistoryService> rh;
    std::unique_ptr<int> mask(new int(rh->GoodDiBlockMask(true)));
    run.put(std::move(mask));
  }

  void CalHit::produce(art::Event& evt)
  {
    // get the RawDigit list
    art::Handle< std::vector<rawdata::RawDigit> > digitcol;
    evt.getByLabel(fRawDataLabel, digitcol);

    // Make a CellHit for every (non-bad) RawDigit and write it out
    const unsigned int Ndigits = digitcol->size();

    // define the unique_ptr holding the collection of rb::CellHits to be made
    std::unique_ptr< std::vector<rb::CellHit> > cellhitcol(new std::vector<rb::CellHit>);
    // Expect most of the cells to be good. Don't want to waste time resizing the array
    cellhitcol->reserve(Ndigits);

    // Get a Calibrator instance for this event
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<cmap::CMap> cmap;

    // Make the hits and write them out
    if(fRemoveBadChans){
      art::ServiceHandle<chaninfo::BadChanList> bcl;
      for(unsigned int iD = 0; iD < Ndigits; ++iD){
	const rawdata::RawDigit* digit = &(*digitcol)[iD];

	const int plane = cmap->GetPlane(digit);
	const int cell =  cmap->GetCell(digit);
	if(!bcl->IsBad(plane, cell)){
          cellhitcol->push_back(cal->MakeCellHit(digit));
	}
      }
    }
    else{
      for(unsigned int iD = 0; iD < Ndigits; ++iD){
        cellhitcol->push_back(cal->MakeCellHit(&(*digitcol)[iD]));
      }
    }

    evt.put(std::move(cellhitcol));
  }

  DEFINE_ART_MODULE(CalHit);

} // end namespace calhit
//////////////////////////////////////////////////////////////////////////
