////////////////////////////////////////////////////////////////////////
// Class:       FiberCalibration
// Module Type: producer
// File:        FiberCalibration_module.cc
//
// Generated at Mon Feb 11 09:01:32 2013 by Evan Niner using artmod
// from art v1_02_06.
////////////////////////////////////////////////////////////////////////

//NOvASoft includes
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "Utilities/func/MathUtil.h"
#include "CalibrationDataProducts/PCHit.h"
#include "CalibrationDataProducts/TCTrack.h"
#include "NovaDAQConventions/DAQConventions.h"

//framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Core/FindManyP.h"
#include "Utilities/AssociationUtil.h"

//ROOT includes
#include "TTree.h"
#include "TH2F.h"
#include "TH1F.h"

namespace calib {

  class FiberCalibration : public art::EDProducer {
  public:
    explicit FiberCalibration(fhicl::ParameterSet const & p);
    virtual ~FiberCalibration();

    void produce(art::Event & e) override;

    void reconfigure(const fhicl::ParameterSet& pset);
    void beginJob();
    void beginRun(art::Run& run);
    void clearNTuple();


  private:

    std::string   fPCHitLabel;        ///< Where to find calibration hits
    std::string   fQualXYLabel;        ///< Instance label for hits with xz quality
    std::string   fQualZLabel;         ///< Instance label for hits with z quality
    std::string   fQualAvgLabel;       ///< Instance label for hits with avg quality
    std::string   fTrackLabel;        ///< Where to find Tracks
            
    //variables to select track quality
    double fFitFrac;     ///<fraction of cells on track kept in final fiber speed fit
    int    fFitCells;    ///<number of cells used in final fiber speed fit for a track
    double  fAsy;         ///<maximum asymmetry between the x and y cells used in the final fit
    int    fTrackPlanes; ///<number of planes track passes through
    double fFiducialX;   ///<maximum distance in cm muon endpoints can be from detector edge in X
    double fFiducialY;   ///<maximum distance in cm muon endpoints can be from detector edge in Y
    double fFiducialZ;   ///<maximum distance in cm muon endpoints can be from detector edge in Z
    int    fStartPlaneDiff; ///<maximum difference between minimum plane in each view associated with the track
    int    fEndPlaneDiff;   ///<maximum difference between maximum plane in each view associated with the track
    double fMinZ;        /// hack for minimum z detector boundary for situation were front diblocks are missing
    double fMaxZ;        /// hack for maximum z in detector, which could be incorrectly calculated depending on missing diblocks

    double fTimeRes; ///<number of sigma timing resolution used to keep/reject hits in speed fit
    bool   fUseLiveGeo; ///<use the live geometry to find beginning and ending diblock

    //ntuple variables
    //event statistics
    int run;
    int subrun;
    int event;
    int cosmictr;
    //track start coordinate
    float sX;
    float sY;
    float sZ;
    //track end coordinate
    float eX;
    float eY;
    float eZ;
    //pathlength
    float plength;
    //min, max plane
    int minPl;
    int maxPl;
    int minPlX;
    int minPlY;
    int maxPlX;
    int maxPlY;
    //min, max cell
    int minCX;
    int minCY;
    int maxCX;
    int maxCY;
    //number of cells
    int ncellsT;
    int ncellsX;
    int ncellsY;
    //number of cells with good timing and calibrated
    int ncellscalT;
    int ncellscalX;
    int ncellscalY;
    //number of cells used in final fiber speed fit
    int ncellsfitT;
    int ncellsfitX;
    int ncellsfitY;
    //fiber speed calculated from fit
    float fiberspeed;
    // 1/beta calculated from fit of muon path using the fiber speed
    float beta;

    //tree to store ntuple statistics
    TTree* fOutTree;

    fhicl::ParameterSet fPSetND, fPSetFD;




   
    

  };
}

namespace calib {

  //define speed of light in cm/ns
  static const double sCmPerNsec = 29.9792458;

  FiberCalibration::FiberCalibration(fhicl::ParameterSet const & pset) 
  // Initialize member data here.
  {
    this->reconfigure(pset); 

    produces< std::vector<caldp::TCTrack> >();
    produces< art::Assns<caldp::TCTrack, rb::Track> >();
  }

  //---------------------------------------------------------------------------
  void FiberCalibration::reconfigure(const fhicl::ParameterSet& pset)
  {
    fPCHitLabel     = pset.get< std::string >("PCHitLabel");
    fTrackLabel     = pset.get< std::string >("TrackLabel");
    fQualXYLabel    = pset.get< std::string >("QualXYLabel");
    fQualZLabel     = pset.get< std::string >("QualZLabel");
    fQualAvgLabel   = pset.get< std::string >("QualAvgLabel");
    fFitFrac        = 0;
    fFitCells       = 0;
    fAsy            = 0;
    fTrackPlanes    = 0;
    fFiducialX      = 0;
    fFiducialY      = 0;
    fFiducialZ      = 0;
    fStartPlaneDiff = 0;
    fEndPlaneDiff   = 0;
    fTimeRes        = 0;
    fMinZ           = 0;
    fMaxZ           = 0;
    fUseLiveGeo     = false;
    fPSetND         = pset.get<fhicl::ParameterSet>("nd");
    fPSetFD         = pset.get<fhicl::ParameterSet>("fd");
  }  

  //---------------------------------------------------------------------------
  FiberCalibration::~FiberCalibration()
  {
    // Clean up dynamic memory and other resources here.
  }

  //---------------------------------------------------------------------------
  void FiberCalibration::beginJob()
  {
    TH1::SetDefaultSumw2();
    TH2::SetDefaultSumw2();

    art::ServiceHandle<art::TFileService> tfs;

    // define the tree
    fOutTree = tfs->make<TTree>("FiberCalib","Fiber Calibaration");

    //define ntuple branches
    fOutTree->Branch("run",    &run,    "run/I");
    fOutTree->Branch("subrun", &subrun, "subrun/I");
    fOutTree->Branch("event",  &event,  "event/I");
    fOutTree->Branch("cosmictr",  &cosmictr,  "cosmictr/I");
    fOutTree->Branch("sX", &sX, "sX/F");
    fOutTree->Branch("sY", &sY, "sY/F");
    fOutTree->Branch("sZ", &sZ, "sZ/F");
    fOutTree->Branch("eX", &eX, "eX/F");
    fOutTree->Branch("eY", &eY, "eY/F");
    fOutTree->Branch("eZ", &eZ, "eZ/F");
    fOutTree->Branch("plength", &plength, "plength/F");
    fOutTree->Branch("minPl", &minPl, "minPl/I");
    fOutTree->Branch("maxPl", &maxPl, "maxPl/I");
    fOutTree->Branch("minPlX", &minPlX, "minPlX/I");
    fOutTree->Branch("maxPlX", &maxPlX, "maxPlX/I");
    fOutTree->Branch("minPlY", &minPlY, "minPlY/I");
    fOutTree->Branch("maxPlY", &maxPlY, "maxPlY/I");
    //fOutTree->Branch("minCX", &minCX, "minCX/I");
    //fOutTree->Branch("minCY", &minCY, "minCY/I");
    //fOutTree->Branch("maxCX", &maxCX, "maxCX/I");
    //fOutTree->Branch("maxCY", &maxCY, "maxCY/I");
    //fOutTree->Branch("ncellsT", &ncellsT, "ncellsT/I");
    //fOutTree->Branch("ncellsX", &ncellsX, "ncellsX/I");
    //fOutTree->Branch("ncellsY", &ncellsY, "ncellsY/I");
    fOutTree->Branch("ncellscalT", &ncellscalT, "ncellscalT/I");
    fOutTree->Branch("ncellscalX", &ncellscalX, "ncellscalX/I");
    fOutTree->Branch("ncellscalY", &ncellscalY, "ncellscalY/I");
    fOutTree->Branch("ncellsfitT", &ncellsfitT, "ncellsfitT/I");
    fOutTree->Branch("ncellsfitX", &ncellsfitX, "ncellsfitX/I");
    fOutTree->Branch("ncellsfitY", &ncellsfitY, "ncellsfitY/I");
    fOutTree->Branch("fiberspeed", &fiberspeed, "fiberspeed/F");
    fOutTree->Branch("beta", &beta, "beta/F");

  }

  //----------------------------------------------------------------------
  void FiberCalibration::beginRun(art::Run&)
  {
    art::ServiceHandle<geo::Geometry> geom;

    fhicl::ParameterSet pset;
    switch(geom->DetId()){
    case novadaq::cnv::kNEARDET:
      pset = fPSetND;
      break;
    case novadaq::cnv::kFARDET:
      pset = fPSetFD;
      break;
    default:
      assert(0 && "Unknown detector");
    }

    fFitFrac        = pset.get< double      >("FitFrac");
    fFitCells       = pset.get< int         >("FitCells");
    fAsy            = pset.get< double      >("Asy");
    fTrackPlanes    = pset.get< int         >("TrackPlanes");
    fFiducialX      = pset.get< double      >("FiducialX");
    fFiducialY      = pset.get< double      >("FiducialY");
    fFiducialZ      = pset.get< double      >("FiducialZ");
    fStartPlaneDiff = pset.get< int         >("StartPlaneDiff");
    fEndPlaneDiff   = pset.get< int         >("EndPlaneDiff");
    fTimeRes        = pset.get< double      >("TimeRes");
    fMinZ           = pset.get< double      >("MinZ");
    fMaxZ           = pset.get< double      >("MaxZ");
    fUseLiveGeo     = pset.get< bool        >("UseLiveGeo");
  }

  //---------------------------------------------------------------------------
  void FiberCalibration::produce(art::Event & evt)
  {

    std::unique_ptr< std::vector<caldp::TCTrack>  > tctracks(new std::vector<caldp::TCTrack>);
      
    std::unique_ptr<art::Assns<caldp::TCTrack, rb::Track> > tcAssns(new art::Assns<caldp::TCTrack, rb::Track>);

    art::ServiceHandle<geo::Geometry>         geom;
    art::ServiceHandle<geo::LiveGeometry>     livegeom;
    novadaq::cnv::DetId detId = geom->DetId();

    //get the cosmic tracks associated with each event
    art::Handle<std::vector<rb::Track> > trackcol;
    evt.getByLabel(fTrackLabel,trackcol);

    //get associationgs between cosmic tracks and pchits
    art::FindManyP<caldp::PCHit> fmpcxy(trackcol, evt, art::InputTag(fPCHitLabel, fQualXYLabel));
    art::FindManyP<caldp::PCHit> fmpcz(trackcol, evt, art::InputTag(fPCHitLabel, fQualZLabel));
    art::FindManyP<caldp::PCHit> fmpcavg(trackcol, evt, art::InputTag(fPCHitLabel, fQualAvgLabel));

    std::vector< art::Ptr<rb::Track> > tracks;
    art::fill_ptr_vector(tracks, trackcol);

    //store event info in ntuple
    run = evt.run();
    subrun = evt.subRun();
    event = evt.id().event();

    //if we are in data use the live geometry to find front and back of the volume.
    if((geom->DetId() == novadaq::cnv::kFARDET) && (fUseLiveGeo)){
      fMinZ = livegeom->InstrumentedDetFront();
      fMaxZ = livegeom->InstrumentedDetBack();
    }

    //now loop over the tracks
    for (unsigned int i=0; i<tracks.size(); ++i){

      this->clearNTuple();

      const art::Ptr<rb::Track> tr = tracks.at(i);
      //if ((tr->NXCell()==0) || (tr->NYCell()==0)) continue;
      //get the pc hits associated with the cosmic track
      std::vector<art::Ptr<caldp::PCHit> > pcxy = fmpcxy.at(i);
      std::vector<art::Ptr<caldp::PCHit> > pcz = fmpcz.at(i);
      std::vector<art::Ptr<caldp::PCHit> > pcavg = fmpcavg.at(i);
      //empty pchit container to combine hits from the lists in
      std::vector<art::Ptr<caldp::PCHit> >  pch;
      //if we failed to get pc hits skip slice
      if (fmpcxy.isValid()){
        for (unsigned int j=0; j<pcxy.size(); ++j) pch.push_back(pcxy[j]);
      }
      if (fmpcz.isValid()){
        for (unsigned int j=0; j<pcz.size(); ++j) pch.push_back(pcz[j]);
      }
      if (fmpcavg.isValid()){
        for (unsigned int j=0; j<pcavg.size(); ++j) pch.push_back(pcavg[j]);
      }
      //must be at least two pchits with this track
      if (pch.size() < 2) continue;


      cosmictr = i; //store track number
      //passed initial round of cuts, now get track information
      TVector3 start(tr->Start());
      TVector3 end(tr->Stop());
      sX = start.X();
      sY = start.Y();
      sZ = start.Z();
      eX = end.X();
      eY = end.Y();
      eZ = end.Z();
      plength = (end - start).Mag();
      /*minPl = tr->MinPlane();
      maxPl = tr->MaxPlane();
      minCX = tr->MinCell(geo::kX);
      minCY = tr->MinCell(geo::kY);
      maxCX = tr->MaxCell(geo::kX);
      maxCY = tr->MaxCell(geo::kY);
      ncellsT = tr->NCell();
      ncellsX = tr->NXCell();
      ncellsY = tr->NYCell();*/

      //track quality cuts 
      //muon starpoint must be near either the ceiling or one of the walls of the detector
      //cut assumes muons cannot enter from bottom of detector
      //this selects muons that are not stoppers and pass all the way through the detector
      if ((sX < geom->DetHalfWidth() - fFiducialX) &&
          (sX > -(geom->DetHalfWidth() - fFiducialX)) &&
          (sY < geom->DetHalfHeight() - fFiducialY) &&
          (sZ < fMaxZ - fFiducialZ) &&
          (sZ > std::max(0.0, fMinZ) + fFiducialZ) ) continue;
      if ((eX < geom->DetHalfWidth() - fFiducialX) &&
          (eX > -(geom->DetHalfWidth() - fFiducialX)) &&
          (eY > -(geom->DetHalfHeight() - fFiducialY)) &&
          (eZ < fMaxZ - fFiducialZ) &&
          (eZ > std::max(0.0, fMinZ) + fFiducialZ) ) continue;

      caldp::TCTrack tc;

      tc.SetStart(start);
      tc.SetStop(end);
      tc.SetTotalLength(plength);

      //now loop over cell hits to get the necessary ones for fiber calibration
      for(unsigned int j=0; j<pch.size(); ++j){
        //if the hit had a bad timing fit, throw it out
        //if (!(pch[j]->GoodTime())) continue;

        //if the distance to readout is less then 0, don't use
        if (pch[j]->ReadDist() < 0.0) continue;

        tc.Add(pch[j]);

      }//end loop over cells

      ncellscalT = tc.NPCHit();
      ncellscalX = tc.NXPCHit();
      ncellscalY = tc.NYPCHit();

      minPlX = tc.MinPlane(geo::kX);
      minPlY = tc.MinPlane(geo::kY);
      maxPlX = tc.MaxPlane(geo::kX);
      maxPlY = tc.MaxPlane(geo::kY);
      minPl  = tc.MinPlane(geo::kXorY);
      maxPl  = tc.MaxPlane(geo::kXorY);

      //track must have traveled through a minimum number of planes
      if ((maxPl-minPl) < fTrackPlanes) continue;

      //Require the minimum and maximum planes in each view to agree
      //Cut in place to reduce case of poor reconstruction or view matching
      //In regions with lots of dead channels this cut would need to be reconsidered
      if (abs(minPlX-minPlY) > fStartPlaneDiff ) continue;
      if (abs(maxPlX-maxPlY) > fEndPlaneDiff ) continue;

      //perform the fit for fiber speed for the track, do NOT drop hits outside specified residual window
      //NOTE: When DCMs are in sync this can be configured to drop outlying hits and calculate the fiber speed. This is not desired when calculating dcm offsets
      std::vector<double> residuals;
      double ficept = 0.0;
      double chisqr = 0.0;
      fiberspeed = tc.CalculateFiberVelocity(fTimeRes,&residuals, &ficept, &chisqr, detId, false,true);

      ncellsfitT = residuals.size();

      //more quality cuts on the fit
      //make sure cells used in fit are above threshold
      if (ncellscalT < fFitCells) continue;
      //make sure cells used in the fit represent a certain fraction of total hits
      if (((double)ncellsfitT/(double)ncellscalT) < fFitFrac) continue;

      ncellsfitX = tc.NXPCHit();
      ncellsfitY = tc.NYPCHit();

      //cut to check difference between number of x and y cells used in the fit
      double asy = ((double)(ncellsfitY-ncellsfitX))/((double)(ncellsfitY+ncellsfitX));
      if (fabs(asy) > fAsy) continue;

      double dummy=0.0;
      residuals.clear();
      beta = tc.CalculateMuonVelocity(&residuals, &dummy, &chisqr, detId, tc.FiberVelocity());

      tctracks->push_back(tc);

      util::CreateAssn(*this, evt, *tctracks, tr, *tcAssns);

      fOutTree->Fill();
      
    
    }//end loop over slices

    evt.put(std::move(tctracks));
    evt.put(std::move(tcAssns));
  } 

  //---------------------------------------------------------------------------
  void FiberCalibration::clearNTuple(){

    //track start coordinate
    sX = -999.0;
    sY = -999.0;
    sZ = -999.0;
    //track end coordinate
    eX = -999.0;
    eY = -999.0;
    eZ = -999.0;
    //pathlength variables
    plength = -5.0;
    //min, max plane
    minPl = -5;
    maxPl = -5;
    //min, max cell
    minCX = -5;
    minCY = -5;
    maxCX = -5;
    maxCY = -5;
    //number of cells
    ncellsT = 0;
    ncellsX = 0;
    ncellsY = 0;
    //number of cells with good timing and calibrated
    ncellscalT = 0;
    ncellscalX = 0;
    ncellscalY = 0;
    //number of cells used in final fiber speed fit
    ncellsfitT = 0;
    ncellsfitX = 0;
    ncellsfitY = 0;
    //fit parameters
    fiberspeed = -99.0;
    beta = -5.0;


  }

} //end namespace

//-----------------------------------------------------------------------------

DEFINE_ART_MODULE(calib::FiberCalibration)
