////////////////////////////////////////////////////////////////////////
// \file    SumRunsCalib_module.cc
// \brief   Sum up AttenProfilesMap objects
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"

// NOvAsoft includes
#include "CalibrationDataProducts/AttenProfiles.h"

namespace calib
{
  /// Sum up AttenProfilesMap objects
  class SumRunsCalib: public art::EDFilter
  {
  public:
    explicit SumRunsCalib(const fhicl::ParameterSet& pset);
    ~SumRunsCalib();

    void reconfigure(const fhicl::ParameterSet& pset);
    virtual bool filter(art::Event& evt) {return false;}
    virtual bool endSubRun(art::SubRun& sr);
    virtual bool endRun(art::Run& run);

  protected:
    std::string fAttenHistsMapLabel;

    int fMinBlock, fMaxBlock;

    unsigned int fLastRun;

    std::map<int, std::unique_ptr<caldp::AttenProfilesMap>> fChannelMapProf;
  };

  //......................................................................
  SumRunsCalib::SumRunsCalib(const fhicl::ParameterSet& pset)
  {
    reconfigure(pset);

    for(int block = fMinBlock; block <= fMaxBlock; ++block){
      produces<caldp::AttenProfilesMap, art::InRun>(TString::Format("block%d", block).Data());
    }
  }

  //......................................................................
  SumRunsCalib::~SumRunsCalib()
  {
  }

  //......................................................................
  void SumRunsCalib::reconfigure(const fhicl::ParameterSet& pset)
  {
    fAttenHistsMapLabel = pset.get<std::string>("AttenHistsMapLabel");
    fMinBlock = pset.get<int>("MinBlock");
    fMaxBlock = pset.get<int>("MaxBlock");
    fLastRun = pset.get<int>("LastRun");

    assert(fMinBlock >= 0);
    assert(fMaxBlock < 0 || fMaxBlock >= fMinBlock);
    // I don't think we can determine the detector in time
    if(fMaxBlock < 0) fMaxBlock = 27;
  }

  //......................................................................
  bool SumRunsCalib::endSubRun(art::SubRun& sr)
  {
    for(int block = fMinBlock; block <= fMaxBlock; ++block){
      art::Handle<caldp::AttenProfilesMap> profsmap;
      sr.getByLabel(fAttenHistsMapLabel, TString::Format("block%d", block).Data(), profsmap);
      if(profsmap.failedToGet()) continue;

      if(!fChannelMapProf[block]) fChannelMapProf[block] = std::unique_ptr<caldp::AttenProfilesMap>(new caldp::AttenProfilesMap(profsmap->MinW(), profsmap->MaxW()));

      *fChannelMapProf[block] += *profsmap;

      // No one else will ever want this. Save the memory
      sr.removeCachedProduct(profsmap);
    }

    return false;
  }

  //......................................................................
  bool SumRunsCalib::endRun(art::Run& run)
  {
    for(int block = fMinBlock; block <= fMaxBlock; ++block){
      art::Handle<caldp::AttenProfilesMap> profsmap;
      run.getByLabel(fAttenHistsMapLabel, TString::Format("block%d", block).Data(), profsmap);
      if(profsmap.failedToGet()) continue;

      if(!fChannelMapProf[block]) fChannelMapProf[block] = std::unique_ptr<caldp::AttenProfilesMap>(new caldp::AttenProfilesMap(profsmap->MinW(), profsmap->MaxW()));

      *fChannelMapProf[block] += *profsmap;

      // No one else will ever want this. Save the memory
      run.removeCachedProduct(profsmap);
    }

    if(run.run() == fLastRun){
      for(auto& it: fChannelMapProf){
	const int block = it.first;
	run.put(std::move(it.second), TString::Format("block%d", block).Data());
      }

      fChannelMapProf.clear();

      return true;
    }

    return false;
  }

  DEFINE_ART_MODULE(SumRunsCalib);

} // namespace
