////////////////////////////////////////////////////////////////////////
// \file    ThresholdAna_module.cc
// \brief   Make plots of threshold effect from PCHits files
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

//C++ and C:
#include <cmath>
#include <vector>

//Framework:
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"

//NovaSoft:
#include "CalibrationUtils/CalibUtil.h"
#include "CalibrationDataProducts/PCHit.h"
#include "Geometry/Geometry.h"

// ROOT
#include "TH2.h"

namespace calib
{
  /// Make plots of threshold and shielding effects from PCHits files
  class ThresholdAna: public art::EDAnalyzer
  {
  public:
    explicit ThresholdAna(const fhicl::ParameterSet& pset);
    virtual ~ThresholdAna();

    virtual void beginRun(const art::Run& run) override;

    virtual void analyze(const art::Event& evt) override;

    virtual void reconfigure(const fhicl::ParameterSet& pset) override;

  protected:
    int GetView(int plane, art::ServiceHandle<geo::Geometry> geom);
    TH1* fNCalibVsW[4];
    TH1* fNZerosVsW[4];

    TH1* fNCalibVsCell[4];
    TH1* fNZerosVsCell[4];

    TH2* fRecoVsWTot[4];
    TH2* fTrueVsWTot[4];
    TH2* fThreshVsWTot[4];
    TH2* fShadowVsWTot[4];
    TH2* fRatioVsWTot[4];

    TH2* fRecoVsW[4][32*12];
    TH2* fTrueVsW[4][32*12];
    TH2* fThreshVsW[4][32*12];
    TH2* fShadowVsW[4][32*12];
    TH2* fRatioVsW[4][32*12];
  };

  //---------------------------------------------------------------------
  ThresholdAna::ThresholdAna(const fhicl::ParameterSet& pset)
    : EDAnalyzer(pset)
  {
    reconfigure(pset);

    // Flag that beginRun hasn't run yet
    fNZerosVsW[0] = 0;
  }

  //---------------------------------------------------------------------
  ThresholdAna::~ThresholdAna()
  {
  }

  //---------------------------------------------------------------------
  void ThresholdAna::reconfigure(const fhicl::ParameterSet& pset)
  {
  }

  //---------------------------------------------------------------------
  void ThresholdAna::beginRun(const art::Run& run)
  {
    // We've already been called for a previous run, no need to make histograms
    // again.
    if(fNZerosVsW[0]) return;

    art::ServiceHandle<geo::Geometry> geom;

    double maxW = -1;
    int nCell = -1;
    if(geom->DetId() == novadaq::cnv::kFARDET){
      maxW = 900;
      nCell = 32*12;
    }
    else if(geom->DetId() == novadaq::cnv::kNEARDET ||
	    geom->DetId() == novadaq::cnv::kNDOS){
      maxW = 250;
      nCell = 32*3;
    }
    else{
      assert(0 && "Unknown detector");
    }

    art::ServiceHandle<art::TFileService> tfs;

    const int nW = 100;

    const int nY = 1000;

    const TString viewStrArr[4]={"_xview","_yview","_muxview","_muyview"};
    bool isND=geom->DetId() == novadaq::cnv::kNEARDET;
    for(int view = geo::kX; view <= (geo::kY+2*isND); ++view){
      const TString viewStr = viewStrArr[view];

      fNZerosVsW[view] = tfs->make<TH1F>("nZerosVsW"+viewStr, ";W", nW, -maxW, +maxW);
      fNCalibVsW[view] = tfs->make<TH1F>("nCalibVsW"+viewStr, ";W;Number of PCHits", nW, -maxW, +maxW);

      fNZerosVsCell[view] = tfs->make<TH1F>("nZerosVsCell"+viewStr, ";Cell", nCell, 0, nCell);
      fNCalibVsCell[view] = tfs->make<TH1F>("nCalibVsCell"+viewStr, ";Cell;Number of PCHits", nCell, 0, nCell);

      fRatioVsWTot[view] = tfs->make<TH2F>(TString::Format("ratioVsW%s_tot", viewStr.Data()), ";W (cm);Reco / Ideal", nW, -maxW, +maxW, nY, 0, 10);

      fRecoVsWTot[view] = tfs->make<TH2F>(TString::Format("recoVsW%s_tot", viewStr.Data()), ";W (cm);PE / MIP expectation for path", nW, -maxW, +maxW, nY, 0, 100);
      fTrueVsWTot[view] = tfs->make<TH2F>(TString::Format("trueVsW%s_tot", viewStr.Data()), ";W (cm);PE poisson mean / true deposit", nW, -maxW, +maxW, nY, 0, 50);

      fThreshVsWTot[view] = tfs->make<TH2F>(TString::Format("threshVsW%s_tot", viewStr.Data()), ";W (cm);PE / poisson lambda", nW, -maxW, +maxW, nY, 0, 10);
      fShadowVsWTot[view] = tfs->make<TH2F>(TString::Format("shadowVsW%s_tot", viewStr.Data()), ";W (cm);True E / MIP expectation", nW, -maxW, +maxW, nY, 0, 10);

      for(int cell = 0; cell < nCell; ++cell){
        fRatioVsW[view][cell] = tfs->make<TH2F>(TString::Format("ratioVsW%s_cell%d", viewStr.Data(), cell), ";W (cm);Reco / Ideal", nW, -maxW, +maxW, nY, 0, 10);

        fRecoVsW[view][cell] = tfs->make<TH2F>(TString::Format("recoVsW%s_cell%d", viewStr.Data(), cell), ";W (cm);PE / MIP expectation for path", nW, -maxW, +maxW, nY, 0, 100);
        fTrueVsW[view][cell] = tfs->make<TH2F>(TString::Format("trueVsW%s_cell%d", viewStr.Data(), cell), ";W (cm);PE poisson mean / true deposit", nW, -maxW, +maxW, nY, 0, 50);

	fThreshVsW[view][cell] = tfs->make<TH2F>(TString::Format("threshVsW%s_cell%d", viewStr.Data(), cell), ";W (cm);PE / poisson lambda", nW, -maxW, +maxW, nY, 0, 10);
	fShadowVsW[view][cell] = tfs->make<TH2F>(TString::Format("shadowVsW%s_cell%d", viewStr.Data(), cell), ";W (cm);True E / MIP expectation", nW, -maxW, +maxW, nY, 0, 10);
      } // end for cell
    } // end for view
  }
  //......................................................................
  int ThresholdAna::GetView(int plane, art::ServiceHandle<geo::Geometry> geom){//get the view for the current combination of view and plane (x,y,mux, or muy)
    int view=0;
    if(geom->DetId() == novadaq::cnv::kNEARDET){
      int firstPlane=geom->FirstPlaneInMuonCatcher();
      int isMuonCatcher=firstPlane <= plane;
      view = geom->Plane(plane)->View()+(2*isMuonCatcher);
    }
    else{
      view = geom->Plane(plane)->View();
    } 
    return view;
  }
  //---------------------------------------------------------------------
  void ThresholdAna::analyze(const art::Event& evt)
  {
    art::ServiceHandle<geo::Geometry> geom;
    art::Handle<std::vector<caldp::PCHit>> hitsXY;
    evt.getByLabel("pclist", "pathQualXY", hitsXY);

    for(unsigned int n = 0; n < hitsXY->size(); ++n){
      const caldp::PCHit& hit = (*hitsXY)[n];
      int plane = hit.Plane();
      int view = GetView(plane,geom);
      // Saturated, skip
      if(hit.PE() > 2000) continue;

      fNCalibVsW[view]->Fill(hit.W());
      fNCalibVsCell[view]->Fill(hit.Cell());

      // This is the expected number of photons produced by 1 MIP. No need for
      // high number of decimal places, the absolute calibration actually sorts
      // out the MIP scale.
      const double mipExp = 1.78*hit.Path();

      // This is essentially what the actual attenuation fit uses (except the
      // factor to convert PEs to MIPs).
      const double recoPart = hit.PE()/mipExp;

      // This is what we actually want to measure, the true transmission
      // function.
      const double truePart = hit.PoissonLambda()/hit.TrueMeV();

      // Correction due to threshold effects: that we get a biased-high number
      // of PEs because of fluctuations sometimes needed to get over threshold.
      const double threshPart = hit.PE()/hit.PoissonLambda();

      // Correction due to "shadowing" (that the true energy deposit isn't
      // equally MIPish everywhere).
      const double shadowPart = hit.TrueMeV()/mipExp;

      // Written this way, it's clear that applying this ratio to the
      // observations will correct them to be the attenuation function we
      // actually want to measure. Could also have written as
      // threshPart*shadowPart, which makes the physical origins of the
      // corrections clearer.
      const double ratio = recoPart/truePart;

      // We fill all the different combinations, because when the distributions
      // are broad, different ways of taking the mean, eg <reco/true> vs
      // <reco>/<true> make a difference.

      fRecoVsWTot  [view]->Fill(hit.W(), recoPart);
      fTrueVsWTot  [view]->Fill(hit.W(), truePart);
      fThreshVsWTot[view]->Fill(hit.W(), threshPart);
      fShadowVsWTot[view]->Fill(hit.W(), shadowPart);
      fRatioVsWTot [view]->Fill(hit.W(), ratio);

      fRecoVsW  [view][hit.Cell()]->Fill(hit.W(), recoPart);
      fTrueVsW  [view][hit.Cell()]->Fill(hit.W(), truePart);
      fThreshVsW[view][hit.Cell()]->Fill(hit.W(), threshPart);
      fShadowVsW[view][hit.Cell()]->Fill(hit.W(), shadowPart);
      fRatioVsW [view][hit.Cell()]->Fill(hit.W(), ratio);
    }

    art::Handle<std::vector<caldp::PCHit>> hitsXYThresh;
    evt.getByLabel("pclist", "pathQualXYThresh", hitsXYThresh);

    for(unsigned int n = 0; n < hitsXYThresh->size(); ++n){
      const caldp::PCHit& hit = (*hitsXYThresh)[n];
      int plane = hit.Plane();
      int view = GetView(plane,geom);
      fNZerosVsW[view]->Fill(hit.W());
      fNZerosVsCell[view]->Fill(hit.Cell());
    }
  }

  DEFINE_ART_MODULE(calib::ThresholdAna);

} //end namespace
