////////////////////////////////////////////////////////////////////////
// Class:       TimingCalFilter
// Module Type: producer
// File:        TimingCalFilter_module.cc
//
// Generated at Mon Feb 11 09:01:32 2013 by Evan Niner using artmod
// from art v1_02_06.
////////////////////////////////////////////////////////////////////////

//NOvASoft includes
#include "Geometry/Geometry.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "Utilities/func/MathUtil.h"
#include "CalibrationDataProducts/PCHit.h"
#include "CalibrationDataProducts/TCTrack.h"
#include "CalibrationDataProducts/DCMSummary.h"
#include "CalibrationDataProducts/DCMStat.h"
#include "NovaDAQConventions/DAQConventions.h"

//framework includes
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Core/FindManyP.h"
#include "Utilities/AssociationUtil.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"

#include <limits>


//ROOT includes
#include "TTree.h"
#include "TH2F.h"
#include "TH1F.h"

namespace calib {

  class TimingCalFilter : public art::EDFilter {
  public:
    explicit TimingCalFilter(fhicl::ParameterSet const & p);
    virtual ~TimingCalFilter();

    virtual bool endSubRun(art::SubRun& sr);
    virtual bool filter(art::Event & e);

    void reconfigure(const fhicl::ParameterSet& pset);
    virtual void beginJob();


  private:

    std::string   fDCMLabel;        ///< Where to find DCM stats
            
    //map relating unique dcm to collection of pchits
    std::map<int, std::vector<art::Ptr<caldp::PCHit> > > fDCMHitMap;

    uint32_t minTimeRun;
    uint32_t maxTimeRun; 

    std::vector<std::vector<int> > fDCMOffsetCount;
    ///vector running sum of relative dcm offsets
    std::vector<std::vector<float> > fDCMOffsetSum;
    ///vector running sum squared of dcm offsets
    std::vector<std::vector<float> > fDCMOffsetSumSquare;
    ///vector of times a dcm is used
    std::vector<int > fDCMUse;

  };
}

namespace calib {

  //define speed of light in cm/ns
  static const double sCmPerNsec = 29.9792458;

  TimingCalFilter::TimingCalFilter(fhicl::ParameterSet const & pset)
  :
   minTimeRun(std::numeric_limits<uint32_t>::max()-1),
   maxTimeRun(std::numeric_limits<uint32_t>::min()),
   fDCMOffsetCount(caldp::kDCMNumMax, std::vector<int>(caldp::kDCMNumMax, 0)),
   fDCMOffsetSum(caldp::kDCMNumMax, std::vector<float>(caldp::kDCMNumMax, 0)),
   fDCMOffsetSumSquare(caldp::kDCMNumMax, std::vector<float>(caldp::kDCMNumMax, 0)),
   fDCMUse(caldp::kDCMNumMax, 0)
  {
    this->reconfigure(pset); 

    produces<caldp::DCMSummary, art::InSubRun>();
  }

//---------------------------------------------------------------------------
  void TimingCalFilter::reconfigure(const fhicl::ParameterSet& pset)
  {
    fDCMLabel  = pset.get< std::string >("DCMLabel");
  }  

  //---------------------------------------------------------------------------
  TimingCalFilter::~TimingCalFilter()
  {
  }

  //---------------------------------------------------------------------------
  void TimingCalFilter::beginJob()
  {
  }

  //---------------------------------------------------------------------------
  bool TimingCalFilter::endSubRun(art::SubRun& sr)
  {
    //object suming stats for absolute timing calibration
    caldp::DCMSummary fDCMSummary;

    fDCMSummary.SetDCMUse(fDCMUse);
    fDCMSummary.SetDCMOffsetCount(fDCMOffsetCount);
    fDCMSummary.SetDCMOffsetSum(fDCMOffsetSum);
    fDCMSummary.SetDCMOffsetSumSquare(fDCMOffsetSumSquare);
    fDCMSummary.SetStartTime(minTimeRun);
    fDCMSummary.SetEndTime(maxTimeRun);

    std::unique_ptr<caldp::DCMSummary> dsum(new caldp::DCMSummary(fDCMSummary));
    sr.put(std::move(dsum));

    for (unsigned int i=0; i<caldp::kDCMNumMax; ++i){
      fDCMUse[i] = 0;
      for (unsigned int j=0; j<caldp::kDCMNumMax; ++j){
        fDCMOffsetCount[i][j] = 0;
        fDCMOffsetSum[i][j] = 0;
        fDCMOffsetSumSquare[i][j] = 0;
      }
    }
    minTimeRun = std::numeric_limits<uint32_t>::max()-1;
    maxTimeRun = std::numeric_limits<uint32_t>::min();

    return true;
  }

  //---------------------------------------------------------------------------
  bool TimingCalFilter::filter(art::Event & evt)
  {
    art::ServiceHandle<geo::Geometry>         geom;

    //get the cosmic tracks associated with each event
    art::Handle<std::vector<caldp::DCMStat> > dcmcol;
    evt.getByLabel(fDCMLabel,dcmcol);

    std::vector<art::Ptr<caldp::DCMStat> > dcms;
    art::fill_ptr_vector(dcms, dcmcol);

    uint32_t evtTime  = evt.time().timeHigh();
    if (evtTime < minTimeRun) minTimeRun = evtTime;
    if (evtTime > maxTimeRun) maxTimeRun = evtTime;


    //now loop over the dcms
    for (unsigned int i=0; i<dcms.size(); ++i){
      const art::Ptr<caldp::DCMStat> dcm = dcms.at(i);
      for (unsigned int dcm1=0; dcm1<dcm->dcmList.size(); ++dcm1){
        //shift index to fill arrya starting at 0
        int d1Indx = dcm->dcmList[dcm1].dcmNum-1;
        fDCMUse[d1Indx]++;
        for (unsigned int dcm2=dcm1+1; dcm2<dcm->dcmList.size(); ++dcm2){
          //shift index for filling array
          int d2Indx = dcm->dcmList[dcm2].dcmNum-1;
          float offset = dcm->dcmList[dcm1].dcmTime - dcm->dcmList[dcm2].dcmTime;
          fDCMOffsetCount[d1Indx][d2Indx]++;
          fDCMOffsetSum[d1Indx][d2Indx] += offset;
          fDCMOffsetSumSquare[d1Indx][d2Indx] += offset*offset;
        }
      }
    } //end loop over dcm stat object

    // For automatic output file naming to work, need to allow at least one
    // event per subrun for now. ART issue 5427
    static unsigned int prevR = -1, prevSR = -1;
    if(evt.run() != prevR || evt.subRun() != prevSR){
      prevR = evt.run();
      prevSR = evt.subRun();
      return true;
    }
    return false;

  } 

} //end namespace

//-----------------------------------------------------------------------------

DEFINE_ART_MODULE(calib::TimingCalFilter)
