///////////////////////////////////////////////////////////////////////////
/// \brief   Histograms used by attenuation calibration
/// \author  Christopher Backhouse - bckhouse@caltech.edu
/// \date    Apr 2013
///////////////////////////////////////////////////////////////////////////
#include "CalibrationDataProducts/AttenProfiles.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

namespace caldp
{
  LiteProfile::LiteProfile(float w0, float w1)
      : fMinW(w0)
      , fMaxW(w1)
    {
    }

  void LiteProfile::Fill(float x, float y)
  {
    const unsigned int bin = ((x-fMinW)*kNumBins)/(fMaxW-fMinW);
    if(x < fMinW || bin > kNumBins) return; // Drop overflows
    
    assert(y >= 0);
    
    // only reserve the memory necessary for the 
    // the vectors when needed, not at construction
    // assume if the fCount vector is not at the size
    // of kNumBins, then none of them are
    if(fCount.size() < kNumBins){
      fCount.resize(kNumBins, 0 );
      fSum  .resize(kNumBins, 0.);
      fSqSum.resize(kNumBins, 0.);
    }
    
    ++fCount[bin];
    fSum[bin]   += y;
    fSqSum[bin] += y*y;
  }

  LiteProfile& LiteProfile::operator+=(const LiteProfile& rhs)
  {
    assert(rhs.fMinW == fMinW);
    assert(rhs.fMaxW == fMaxW);
    
    // only reserve the memory necessary for the 
    // the vectors when needed, not at construction
    // assume if the fCount vector is not at the size
    // of kNumBins, then none of them are
    if(fCount.size() < kNumBins){
      fCount.resize(kNumBins, 0 );
      fSum  .resize(kNumBins, 0.);
      fSqSum.resize(kNumBins, 0.);
    }

    if(fCount.size() != rhs.fCount.size() || 
       fSum.size()   != rhs.fSum.size()   || 
       fSqSum.size() != rhs.fSqSum.size() ){
      LOG_WARNING("AttenProfiles") << "vector sizes in LiteProfile + operator are not equal\n"
				   << fCount.size() << "/" << rhs.fCount.size() << "\n"
				   << fSum.size()   << "/" << rhs.fSum.size()   << "\n"
				   << fSqSum.size() << "/" << rhs.fSqSum.size() << "\n"
				   << "don't perform the sum - something is very wrong";
      return *this;
    }

    for(size_t i = 0; i < fCount.size(); ++i){
      //LOG_VERBATIM("AttenProfiles") << i << " " << rhs.fCount[i] << " " << rhs.fSum[i] << " " << rhs.fSqSum[i];
      fCount[i] += rhs.fCount[i];
      fSum[i]   += rhs.fSum[i];
      fSqSum[i] += rhs.fSqSum[i];
    }
    
    return *this;
  }

  TH1F* LiteProfile::ToTH1F() const
  {
    TH1F* h = new TH1F("", "", kNumBins, fMinW, fMaxW);
    
    for(size_t i = 0; i < kNumBins; ++i){
      const int N = fCount[i];
      if(N < 1) continue;
      h->SetEntries(h->GetEntries()+N);
      
      h->SetBinContent(i+1, fSum[i]/N);
      
      const float arg = fSqSum[i]-fSum[i]*fSum[i]/N;
      if(arg >= 0){
	h->SetBinError(i+1, std::sqrt(arg)/N);
      }
      else{
	h->SetBinError(i+1, fSum[i]/N);
      }
    }
    
    return h;
  }
  
  int LiteProfile::TotalHits() const
  {
    int ret = 0;
    for(int n: fSum) ret += n;
    return ret;
  }
  
  int LiteProfile::TotalEntries()
  {
    int ret =0;
    for(size_t i = 0; i < kNumBins; ++i){
      ret+=fCount[i];
    }
    
    return ret;
  }

  AttenProfiles::AttenProfiles(float minW, float maxW)
    : WPE(minW, maxW),
      WPE_corr(minW, maxW),
      WPE_corr_xy(minW, maxW),
      WPE_corr_z(minW, maxW),
      WPE_corr_xy_truth(minW, maxW)
  {
  }

  AttenProfiles& AttenProfiles::operator+=(const AttenProfiles& rhs)
  {
    WPE += rhs.WPE;
    WPE_corr += rhs.WPE_corr;
    WPE_corr_xy += rhs.WPE_corr_xy;
    WPE_corr_z += rhs.WPE_corr_z;
    WPE_corr_xy_truth += rhs.WPE_corr_xy_truth;
    
    return *this;
  }

  AttenProfilesMap::AttenProfilesMap(float w0, float w1) : fMinW(w0), fMaxW(w1) {}

  bool AttenProfilesMap::HasProfiles(const geo::OfflineChan& c)
  {
    return fMap.find(c) != fMap.end();
  }
  
  bool AttenProfilesMap::HasProfiles(int plane, int cell)
  {
    return HasProfiles(geo::OfflineChan(plane, cell));
  }

  AttenProfiles& AttenProfilesMap::GetProfiles(const geo::OfflineChan& c)
  {
    auto it = fMap.find(c);
    if(it != fMap.end()) return it->second;
    
    fMap.emplace(std::make_pair(c, AttenProfiles(fMinW, fMaxW)));
    return fMap.find(c)->second;
  }
  
  AttenProfiles& AttenProfilesMap::GetProfiles(int plane, int cell)
  {
    return GetProfiles(geo::OfflineChan(plane, cell));
  }
  
  /// List of channels with profiles on
  std::vector<geo::OfflineChan> AttenProfilesMap::Channels() const
  {
    std::vector<geo::OfflineChan> ret;
    ret.reserve(fMap.size());
    for(auto& it: fMap) ret.push_back(it.first);
    return ret;
  }
  
  /// Intended for writing out/drawing all the histograms
  std::map<int, std::map<int, std::vector<TH1F*>>> AttenProfilesMap::GetAllProfilesByPlaneAndCell()
  {
    std::map<int, std::map<int, std::vector<TH1F*>>> ret;
    for(auto& it: fMap){
      const int plane = it.first.Plane();
      const int cell = it.first.Cell();
      const AttenProfiles& hs = it.second;
      std::vector<TH1F*>& v = ret[plane][cell];
      /*        if(hs.WPE.TotalHits()){
		v.push_back(hs.WPE.ToTH1F());
		v.back()->SetName(TString::Format("h_WPE_%03d_%03d", plane, cell));
		v.back()->SetTitle(TString::Format("Plane %d, Cell %d;W;PE", plane, cell));
		}
		
		if(hs.WPE_corr.TotalHits()){
		v.push_back(hs.WPE_corr.ToTH1F());
		v.back()->SetName(TString::Format("h_WPE_corr_avg_%03d_%03d", plane, cell));
		v.back()->SetTitle(TString::Format("Avg Corr, Plane %d, Cell %d;W;PE/cm", plane, cell));
		}*/
      
      if(hs.WPE_corr_xy.TotalHits()){
	v.push_back(hs.WPE_corr_xy.ToTH1F());
	v.back()->SetName(TString::Format("h_WPE_corr_xy_%03d_%03d", plane, cell));
	v.back()->SetTitle(TString::Format("XY, Plane %d, Cell %d;W;PE/cm", plane, cell));
      }
      
      if(hs.WPE_corr_z.TotalHits()){
	v.push_back(hs.WPE_corr_z.ToTH1F());
	v.back()->SetName(TString::Format("h_WPE_corr_z_%03d_%03d", plane, cell));
	v.back()->SetTitle(TString::Format("Z, Plane %d, Cell %d;W;PE/cm", plane, cell));
      }
      /*
        if(hs.WPE_corr_xy_truth.TotalHits()){
	v.push_back(hs.WPE_corr_xy_truth.ToTH1F());
	v.back()->SetName(TString::Format("h_WPE_corr_xy_truth_%03d_%03d", plane, cell));
	v.back()->SetTitle(TString::Format("XY, Plane %d, Cell %d;W;MeV/cm", plane, cell));
	}*/
    }
    return ret;
  }

  AttenProfilesMap& AttenProfilesMap::operator+=(const AttenProfilesMap& rhs)
  {
    for(auto& it: rhs.fMap){
      GetProfiles(it.first) += it.second;
    }
    return *this;
  }
  
  void AttenProfilesMap::Clear()
  {
    fMap.clear();
  }
  
}
