// \file    PCHit.h
// \brief   A DriftResponse contains the basic information from cosmic track data needed for calibration.
// \version $Id: DriftResponse.cxx,v 1.0 2013/01/13 22:06:10 rtoner Exp $  
// \author  $Author: rtoner $
// \date    $Date: 2013/01/13 22:06:10 $
 
#include "CalibrationDataProducts/DriftResponse.h"
#include <iostream>
#include <iomanip>

namespace caldp 
{

  //------------------------------------------------------------
  void MeanStruct::SetResponse( double fMean, double fMeanErr, double fRMS, int fNHits ){
    mean = fMean;
    meanerr = fMeanErr;
    rms = fRMS;
    nhits = fNHits;
  }


  //------------------------------------------------------------
  DriftResponse::DriftResponse():
    run      (-1),
    startTime(0),
    endTime  (0)
  {  }


  //------------------------------------------------------------
  DriftResponse::DriftResponse(int fRun, uint32_t fStartTime, uint32_t fEndTime):
    run      (fRun),
    startTime(fStartTime),
    endTime  (fEndTime)
  {  }
   
   
  //------------------------------------------------------------
  DriftResponse::~DriftResponse(){  }
  

  //------------------------------------------------------------
  void DriftResponse::AddChannelResponse  ( int fOffChan, double fMean, double fMeanErr, double fRMS, int fNHits ){

    if ( meanMap.find(fOffChan) == meanMap.end() ) {

      MeanStruct meanResp;

      meanResp.SetResponse(fMean,fMeanErr,fRMS,fNHits);

      //Insert a new entry into the map.
      meanMap.insert(std::make_pair(fOffChan,meanResp));
    } else {
      std::cout << "Print an error message here about the entry already existing." << std::endl;
    }
  } //End AddChannelResponse

   
  //------------------------------------------------------------
  bool DriftResponse::GetChannelResponse  ( int fOffChan, double &fMean, double &fMeanErr, double &fRMS, int &fNHits ) const {

    fMean     = -1;
    fMeanErr  = -1;
    fRMS      = -1;
    fNHits    = -1;

    if ( meanMap.find(fOffChan) == meanMap.end() ) {
      return false;
    } 

    const MeanStruct meanResp = meanMap.at(fOffChan);

    fMean     = meanResp.Mean();
    fMeanErr  = meanResp.MeanErr();
    fRMS      = meanResp.RMS();
    fNHits    = meanResp.NHits();

    return true;


  } //End AddChannelResp

  //------------------------------------------------------------
  double DriftResponse::Mean (int fOffChan) const {

    double fMean = -1;
     
    if ( meanMap.find(fOffChan) != meanMap.end() ) {
      const MeanStruct meanAtChan = meanMap.at(fOffChan);
      fMean = meanAtChan.Mean();
    }

    return fMean;       

  };

  //------------------------------------------------------------
  double DriftResponse::MeanErr (int fOffChan) const {

    double fMeanErr = -1;
     
    if ( meanMap.find(fOffChan) != meanMap.end() ) {
      const MeanStruct meanAtChan = meanMap.at(fOffChan);
      fMeanErr = meanAtChan.MeanErr();
    }

    return fMeanErr;       

  };

  //------------------------------------------------------------
  double DriftResponse::RMS (int fOffChan) const {

    double fRMS = -1;
     
    if ( meanMap.find(fOffChan) != meanMap.end() ) {
      const MeanStruct meanAtChan = meanMap.at(fOffChan);
      fRMS = meanAtChan.RMS();
    }

    return fRMS;       

  };

  //------------------------------------------------------------
  double DriftResponse::NHits (int fOffChan) const {

    double fNHits = -1;
     
    if ( meanMap.find(fOffChan) != meanMap.end() ) {
      const MeanStruct meanAtChan = meanMap.at(fOffChan);
      fNHits = meanAtChan.NHits();
    }

    return fNHits;       

  };


  //------------------------------------------------------------
  bool  DriftResponse::CheckForChannel ( int fOffChan ) const {

    if ( meanMap.find(fOffChan) == meanMap.end() ) return false;

    return true;       

  };


  //------------------------------------------------------------
  std::ostream& operator<< (std::ostream& o, const DriftResponse& dr)
  {
    o << std::setiosflags(std::ios::fixed) << std::setprecision(2);
    o << "Run = "        << std::setw(1) << std::right << dr.Run()
      << " StartTime = " << std::setw(1) << std::right << dr.StartTime()
      << " EndTime = "   << std::setw(3) << std::right << dr.EndTime();

    MeanStruct firstChan = dr.ReturnMap().begin()->second;

    o << "\nResponse for channel " << dr.ReturnMap().begin()->first << " is" 
      << "\n\tMean: "    << firstChan.Mean()
      << "\n\tMeanErr: " << firstChan.MeanErr()
      << "\n\tRMS: "     << firstChan.RMS()
      << "\n\tNHits: "   << firstChan.NHits()
      << std::endl;

    return o;
  }


} // end namespace caldp
