////////////////////////////////////////////////////////////////////////
// \file    CalibUtil.h
// \brief   Functions of general use in calibration
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#ifndef CALIB_CALIBUTIL_H
#define CALIB_CALIBUTIL_H

#include <vector>
#include <map>

#include "GeometryObjects/OfflineChan.h"
#include "TVector3.h"

namespace calib
{
  /// \brief Methods used to estimate path length of a track through a cell
  ///
  /// The recommended method to perform a calibration is to accumulate separate
  /// histograms or averages for each cell in each quality level. If there is
  /// insufficient statistics in "xy" mode (there will be zero for cells along
  /// the sides of the detector), fall back to the "z" mode calculations. If
  /// there is insufficient stats here too (should only happen in the detector
  /// corners) fall back (reluctantly) to the kAverage calculations.
  enum EPathQuality{
    /// There are adjacent hits on the same plane. This means the track must
    /// have entered in one side wall of the cell and gone out the other. The
    /// path length estimate is likely very precise.
    kXYAdjacents,
    /// There are no adjacent cells in the same plane, but the cells of the
    /// same index in the upstream and downstream planes are hit. There is high
    /// confidence that the track went in the front wall of the cell and out
    /// the back. The path length estimate should be good. The kXYAdjacents
    /// case is preferred only because it is more common.
    kZAdjacents,
    /// The cell has no special properties, and the path length has been
    /// estimated using the \ref geo::AverageCellPathLength method. The result
    /// is likely pretty poor. Only use cells with this estimate if you have no
    /// choice.
    kAverage,
    /// Something was wrong with the track object, the pathlength will be
    /// zero. Do not use this cell.
    kBadTrack
  };

  /// \brief Calculate the best path length estimate for each hit in a track,
  /// and indicate how they were obtained.
  ///
  /// \param trk The track whose cells should be used
  /// \param[out] quals How the path length of each hit was determined. Please
  ///                   read the documentation for \ref EPathQuality
  /// \param extra Other cells that should be considered to be on the track
  ///              (optional). Results for these will be at the end of returned
  ///              vectors.
  /// \return Path lengths (in cm). Indices correspond to those passed to
  ///         \ref rb::Track::Cell
  std::vector<double> BestPathEstimates(const std::vector<geo::OfflineChan>& chans,
                                        const std::map<unsigned int, TVector3>& directions,
                                        std::vector<EPathQuality>& quals,
                                        const std::vector<geo::OfflineChan>& extras = {});

  /// \brief Find empty cells that would have contributed to the calibration if
  /// they'd been hit. We assume they fell below threshold.
  ///
  /// TODO: Can only currently find "xy" quality candidates
  ///
  /// \param[out] quals How the path length of each cell was determined. Please
  ///                   read the documentation for \ref EPathQuality
  /// \param[out] paths Estimated path length for each selected cell
  /// \param[out] chans Which cells were selected
  void FindBelowThresholdCalibCandidates(const std::vector<geo::OfflineChan>& trkchans,
                                         const std::map<unsigned int, TVector3>& directions,
                                         std::vector<EPathQuality>& quals,
                                         std::vector<double>& paths,
                                         std::vector<geo::OfflineChan>& chans);

  /// \brief Return the direction at a plane using the map provided by rb::Track
  TVector3 const DirectionAtPlane(const std::map<unsigned int, TVector3>& directions, 
				  unsigned int plane);

  /// \brief Return position in world coordninates and distance to the readout
  ///
  /// \param cellhit A hit on a track
  /// \param w The unknown coordinate from the opposite view
  /// \param xyzd[4] Returns x,y,z,d where d is the distance to readout
  void GetXYZD(geo::OfflineChan chan,
               double w,
               double *xyzd);
}


#endif // CALIB_CALIBUTIL_H
////////////////////////////////////////////////////////////////////////

