////////////////////////////////////////////////////////////////////////
/// \file    Calibrator.h
/// \brief   Service that handles calibrations for a particular
///          validity context (eg: detector+timestamp).
/// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////
#ifndef CALIBRATOR_H
#define CALIBRATOR_H

#include "TTimeStamp.h"

// ART includes
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Registry/ServiceMacros.h"
#include "fhiclcpp/ParameterSet.h"

#include "CMap/CMap.h"
#include "Geometry/Geometry.h"
#include "RawData/RawDigit.h"

#include "TF1.h"

namespace rb {
  class CellHit;
  class RecoHit;
}

///Calibration service
namespace calib {

  class AttenCurve;
  class AbsCache;
  class AttenCache;
  class ADCShapeFitTable;
  class DriftCache;
  class TimingCache;

  class Calibrator {

  public:
    Calibrator(fhicl::ParameterSet const& pset, art::ActivityRegistry& reg);
    ~Calibrator();

    void preEvent(art::Event const& evt);
    void postBeginRun(art::Run const& run);
    void postBeginSubRun(const art::SubRun& subrun);

    void reconfigure(const fhicl::ParameterSet& pset);

    // Get the timestamp currently assumed by the Calibrator (unused for now)
    TTimeStamp GetTimeStamp() const { return fCurrTimeStamp; }

    //---------------------------------------------
    // RawDigit to CellHit methods
    //---------------------------------------------

    // Main thing you'll want:
    rb::CellHit MakeCellHit(const rawdata::RawDigit *rawdigit); // Do everything

    float GetPE (rb::CellHit const&);
    float GetPE (const rawdata::RawDigit *rawdigit);
    /// \param goodTime Did the timing fit go well?
    /// \param maxadc   Internal hook for use of \ref GetPE
    float GetTNS(const rawdata::RawDigit *rawdigit, bool& goodTime,
		 double* maxadc = 0);

    //---------------------------------------------
    // CellHit to RecoHit methods
    //---------------------------------------------
    // (NOTE: w != dist_to_readout.  Those are
    // different arguments.)
    //---------------------------------------------

    // Main thing you'll want:

    // Do everything
    rb::RecoHit MakeRecoHit(rb::CellHit const& cellhit, double w);
    rb::RecoHit MakeRecoHit(rb::CellHit const& cellhit, double *xyz);
    
    double GetT       (rb::CellHit const& cellhit, 
		      double dist_to_readout);
    double GetPECorr  (rb::CellHit const& cellhit, double w);

    double GetTimeRes(rb::CellHit const& cellhit);

    double GetAttenScale(rb::CellHit const& cellhit, double w); ///< for PE->PECorr conversion
    double GetAttenScale(const calib::AttenCurve* curve, const double minW, const double maxW,double w);

    /// Part of the PE->PECorr conversion
    double GetDriftScale(rb::CellHit const& cellhit);

    // For PECorr->GeV conversion
    double GetPECorrToGeVScale(rb::CellHit const& cellhit);

    /// For GeV->MIP conversion
    double GetGeVToMIPScale(rb::CellHit const& cellhit);

    /// \brief For use by RecoBase classes
    ///
    /// Conversion from TotalGeV to CalorimetricEnergy. Takes account of dead
    /// material and threshold effects in an averaged fashion. Makes no sense
    /// to apply to individual hits. You probably don't want to use this, you
    /// probably want to look at functions of your RecoBase object.
    double GetGeVToCalorimetricScale() const;

    /// \brief Estimate of true shower energy, adapted from RecoJMShower
    ///
    /// Accounts for dead material and threshold effects that can't be done at
    /// the individual cell level.
    /// Pass in the TVectors and energy explicitly rather than the rb::Track
    /// object to avoid a circular dependency with RecoBase
    double GetShowerEnergy(TVector3 const& start,
			   TVector3 const& stop,
			   double   const& totalGeV);

    /// Given plane, cell and whether the data is real,
    /// get attenuation curve and Max w for the plane
    bool  GetAttenCurve(const int plane, const int cell, const bool is_realdata
			,const calib::AttenCurve*&, double& minW,double& maxW) const;

    /// Get the timing offset for a given plane, cell.  Useful downstream to check calibration
    double GetTimingOffset(unsigned int plane, unsigned int cell, bool isData);


  private:
    TTimeStamp fCurrTimeStamp; ///< current time stamp

    /// Root file \ref ADCShapeFitTable will load from
    std::string fShapeTableFilenameFDMC, fShapeTableFilenameFDData, fShapeTableFilenameNDMC, fShapeTableFilenameNDData;
    // Not initialized by default, must check and create if required
    ADCShapeFitTable* fShapeTable;

    AttenCache* fAttenCache;
    DriftCache* fDriftCache;
    AbsCache* fAbsCache;
    TimingCache* fTimingCache;
    bool fTimingCacheVldTimeSet; ///< Is fTimingCache set to the right time?
    int fCurrentRun; //get the current run to make sure that we know if we have perfect MC

    /// Helper function for the two public MakeRecoHit methods
    rb::RecoHit MakeRecoHit(rb::CellHit const& cellhit,
			    double w,
			    double *xyz);

    void EnsureTimingCacheVldTime();

    bool fUseFlatCalibOffset;
    double fFlatCalibOffset;

    bool fUseXSlopeOffset;
    double fXSlopeOffset;

    bool fUseYSlopeOffset;
    double fYSlopeOffset;

    bool fUseXSlopeFit;
    std::string fXSlopeFitForm;
    std::vector<float> fXSlopeFitParams;
    TF1* fXSlopeFit;

    bool fUseYSlopeFit;
    std::string fYSlopeFitForm;
    std::vector<float> fYSlopeFitParams;
    TF1* fYSlopeFit;

    bool fUseCellByCellCalibRandomOffset;
    bool fCellByCellCalibRandomOffsetHighResolution;
    double fCellByCellCalibRandomOffset;
    int fCellByCellCalibRandomOffsetSeed;


    //forces mc channels to be uncalibrated if there data equivalent is
    bool fMaskUncalibratedChannelsInMC;

    /// When used by CalHit to transform RawDigits into CellHits, produce
    /// CellHits that look like v0 data, whatever the input
    bool fMakeSinglePointCellHits;

    /// Expert toggle switch to turn off use of timing offsets for purpose of running the 
    /// timing calibration
    bool fUseTimingOffsets;

    // Cache service handles for speed (we're in the inner loop of lots of
    // algorithms)
    art::ServiceHandle<cmap::CMap> fCMap;
    art::ServiceHandle<geo::Geometry> fGeom;

    /// Number of calibrated hits returned
    unsigned int fStatsNumCalibrated;
    /// Number of uncalibrated hits returned
    unsigned int fStatsNumUncalibrated;
    /// Number of hits from uncalibrated diblocks using averaged calibrations
    unsigned int fStatsNumAveraged;

    bool fUseDrift;
    std::string fDriftCSV;

  };
} // end namespace calib
DECLARE_ART_SERVICE(calib::Calibrator, LEGACY)
#endif
///////////////////////////////////////////////////////////////////////////  
