////////////////////////////////////////////////////////////////////////
/// \brief   Function to perform, and table to cache, timing fits to ADC values
/// \author  bckhouse@caltech.edu
/// \date
////////////////////////////////////////////////////////////////////////

// Please, do not introduce any art dependencies into this file. It needs to be
// able to be compiled standalone (as part of genADCShapeFitLookupTable) to
// produce the lookup table file in the first place. It should be possible to
// do any required artyness in Calibrator.cxx

#include "Calibrator/func/ADCShapeFit.h"

#include <cassert>
#include <cmath>
#include <iostream>

#include "TFile.h"
#include "TTree.h"

template<class T> inline T sqr(T x){return x*x;}

namespace calib
{
  // Number of points we expect to receive, to find the best peak match with
  const unsigned int kNumFineTimingADCPoints = 4; // Including the DCS origin

  // Empirically, the best fit is never found outside this range. +3 is an
  // upper limit on the positive side because it means the pulse starts beyond
  // the last ADC observation.
  const double kZeroOffsetSamples = -1;
  // Have the capacity to consider 65,536 offsets, which is way too many to
  // give good performance.
  const int kOffsetStep = 16;
  const double kSamplesPerOffset = 1./(256*64);
  // With 4096 steps that means the maximum offset is +3 and step size is
  // 0.49ns (in the FD).

  //......................................................................
  /// \brief Helper function for \a ADCShapeFit inner loop.
  ///
  /// Also used independently to calculate the best fit adc peak height (the
  /// return value).
  inline double GetExpectations(double t0,
				double riseTime, double fallTime,
                                const int16_t* obs,
                                double* exps, double& base)
  {
    for(unsigned int t = 0; t < kNumFineTimingADCPoints; ++t){
      // This is the ideal ASIC formula from SimpleReadout
      exps[t] = (t > t0) ? exp(-(t-t0)/fallTime)*(1-exp(-(t-t0)/riseTime)) : 0;
    }

    // We can calculate the best fit baseline and normalization at this offset
    // analytically (just write the chisq expression, and take derivatives with
    // norm or base), like so:
    double U, V, W, X, Z;
    U = V = W = X = Z = 0;
    for(unsigned int i = 0; i < kNumFineTimingADCPoints; ++i){
      U += exps[i];
      V += 1;
      W += obs[i];
      X += exps[i]*exps[i];
      Z += obs[i]*exps[i];
    }
    const double norm = (W*U-V*Z)/(U*U-X*V);
    base = (W*X-Z*U)/(X*V-U*U);

    // Correct all the expectations for the best guess baseline and norm
    for(unsigned int n = 0; n < kNumFineTimingADCPoints; ++n){
      exps[n] = norm*exps[n]+base;
    }

    return norm;
  }

  //......................................................................
  uint16_t ADCShapeFit(int16_t adc1, int16_t adc2, int16_t adc3,
                       double riseTime, double fallTime,
                       bool& goodTime)
  {
    // The observations, in chronological order
    const int16_t obs[kNumFineTimingADCPoints] = {0, adc1, adc2, adc3};

    double bestchisq = 1e10; // infinity
    // Sane default
    uint16_t bestoffset = -kZeroOffsetSamples/kSamplesPerOffset;
    goodTime = false;

    // Scan through possible offsets of the peak, looking for the best match
    uint16_t offset = 0;
    do{
      // In TDC
      const double t0 = (double(offset)*kSamplesPerOffset+kZeroOffsetSamples);

      // Expectations
      double exps[kNumFineTimingADCPoints];
      double junk;
      const double norm = GetExpectations(t0, riseTime, fallTime, obs, exps, junk);

      if(norm < 0) continue;

      // Assuming equal, uncorrelated, errors on all points. (Including
      // correlations was not found to improve resolution).
      double chisq = 0;
      for(unsigned int i = 0; i < kNumFineTimingADCPoints; ++i)
	chisq += sqr(obs[i]-exps[i]);
      assert(chisq >= 0);

      if(chisq < bestchisq){
	bestchisq = chisq;
	bestoffset = offset;
	goodTime = true;
      }
    } while((offset += kOffsetStep) != 0);

    if(bestoffset == 0 || bestoffset == 256*256-kOffsetStep){
      /*
      std::cerr << "Best time offset is at edge of range. "
                << "If this happens repeatedly there's a "
                << "timing problem. Details: "
                << adc0 << " " << adc1 << " " << adc2 << std::endl;
      */
      bestoffset = -kZeroOffsetSamples/kSamplesPerOffset;
      goodTime = false;
    }

    return bestoffset;
  }

  //......................................................................
  ADCShapeFitTable::ADCShapeFitTable(const std::string& fname, bool isND, bool isMC)
    : fIsND(isND), fIsMC(isMC), fNHit(0), fNMiss(0)
  {
    TFile f(fname.c_str());
    assert(!f.IsZombie());

    TTree* tr = (TTree*)f.Get("shape_lookup");
    assert(tr);

    const unsigned int N = tr->GetEntries();

    assert(N == kNumTableEntries);

    // Offsets are all two bytes to save space, which is far more resolution
    // ADCShapeFit needs.
    uint16_t offset;
    bool good;
    tr->SetBranchAddress("offset", &offset);
    tr->SetBranchAddress("good", &good);

    for(unsigned int n = 0; n < N; ++n){
      tr->GetEntry(n);
      fTable[n] = offset;
      fGood[n] = good;
    }
  }

  //......................................................................
  ADCShapeFitTable::~ADCShapeFitTable()
  {
  }

  //......................................................................
  double ADCShapeFitTable::TNS(double tns0, int16_t adc1, int16_t adc2, int16_t adc3,
                               double& adcpeak, bool& goodTime, double& base) const
  {
    uint16_t offset;

    const double sampleTime = fIsND ? 125 : 500; // ns
    /***/ double riseTime = (fIsND ?  140 :  380)/sampleTime; // samples
    const double fallTime = (fIsND ? 4500 : 7000)/sampleTime; // samples

    if(!fIsND && !fIsMC) riseTime = 460/sampleTime;

    // These are the conditions for the best offset to be in the table
    if(adc3 >= 0 && adc3 < kADC3Max &&
       adc1 >= kADC1Min && adc1 < adc3 &&
       adc2 >= adc3+kADC2RelMin && adc2 < adc3+kADC2RelMax){
      ++fNHit;

      // And this is the offset it'll be at, based on aforementioned geometry
      const int tableIdx = (kADC2RelMax-kADC2RelMin)* // Wedge thickness
        (-kADC1Min*(adc3+1)+ // Rectangular part
         (adc3*(adc3-1))/2+adc1)+ // Triangular part
        (adc2-adc3-kADC2RelMin); // Distance through wedge

      assert(tableIdx >= 0 && tableIdx < int(kNumTableEntries));

      offset = fTable[tableIdx];
      goodTime = fGood[tableIdx];

      // One time in 400, do the full calculation just to sanity check that the
      // table (and our indexing into it) agrees. We have about a 5% miss rate
      // anyway, so this isn't too much additional computation.
      if(adc3%20 == 0 && adc1%20 == 0){
	assert(ADCShapeFit(adc1, adc2, adc3, riseTime, fallTime, goodTime) == offset);
      }
    }
    else{
      // It's not in our table, just do the fit ourselves.
      ++fNMiss;
      offset = ADCShapeFit(adc1, adc2, adc3, riseTime, fallTime, goodTime);
    }

    // Empirically, if the best-fit t0 is before the first sample, then this is
    // a bad fit. See docdb 12601.
    if(offset < -kZeroOffsetSamples/kSamplesPerOffset) goodTime = false;

    // We need to extract the best normalization, so repeat the relevant part
    // of the inner loop at that offset.
    if(goodTime){
      // The observations, in chronological order
      const int16_t obs[kNumFineTimingADCPoints] = {0, adc1, adc2, adc3};
      const double t0 = (double(offset)*kSamplesPerOffset+kZeroOffsetSamples);
      double junk[kNumFineTimingADCPoints]; // Expectations
      const double norm = GetExpectations(t0, riseTime, fallTime, obs, junk, base);

      // The value the peak of the curve would have before scaling by norm
      // Analytic calculation from the functional form in ReadoutSim
      static const double peak = pow(riseTime/(riseTime+fallTime), riseTime/fallTime)-
	pow(riseTime/(riseTime+fallTime), 1+riseTime/fallTime);

      adcpeak = norm*peak;
    }
    else{
      // Unless the fit failed, in which case we're probably safest just doing
      // this.
      adcpeak = adc3;
      return tns0;
    }

    return tns0+(offset*kSamplesPerOffset+kZeroOffsetSamples)*sampleTime;
  }

  //......................................................................
  double ADCShapeFitTable::TableHitFraction() const
  {
    return fNHit ? fNHit/double(fNHit+fNMiss) : 0;
  }
} // namespace
