////////////////////////////////////////////////////////////////////////
/// \brief   Function to perform, and table to cache, timing fits to ADC values
/// \author  bckhouse@caltech.edu
/// \date
////////////////////////////////////////////////////////////////////////

// Please, do not introduce any art dependencies into this file. It needs to be
// able to be compiled standalone (as part of genADCShapeFitLookupTable) to
// produce the lookup table file in the first place. It should be possible to
// do any required artyness in Calibrator.cxx

#ifndef CALIB_ADCSHAPEFIT_H
#define CALIB_ADCSHAPEFIT_H

#include "stdint.h"

#include <array>
#include <bitset>
#include <string>

namespace calib
{
  /// Scans a pulseheight shape across the given ADC values and returns the
  /// best offset. Offset is in internal units. Use \ref OffsetToTNS to convert
  /// to an absolute time in nanoseconds.
  ///
  /// Rise and fall time should be in units of the spacing between samples
  uint16_t ADCShapeFit(int16_t adc1, int16_t adc2, int16_t adc3,
                       double riseTime, double fallTime,
                       bool& goodTime);

  class ADCShapeFitTable
  {
  public:
    ADCShapeFitTable(const std::string& fname, bool isND, bool isMC);
    ~ADCShapeFitTable();
    /// Returns equivalent of OffsetToTNS(tdc, ADCShapeFit(adc1...)) but in
    /// most cases can do so more rapidly via a lookup table. \a adcpeak is the
    /// estimated peak curve value at the best fit time.
    double TNS(double tns0, int16_t adc1, int16_t adc2, int16_t adc3,
               double& adcpeak, bool& goodTime, double& base) const;
    /// What fraction of calls to \a TNS went via the table?
    double TableHitFraction() const;

    // Sadly this stuff needs to be in the header to allow the size of the
    // members to be known.

    // Bounds within which adc values must lie in order for them to be in our
    // table. adc1 must be less than adc3. adc2 is measured relative to
    // adc3. See genADCShapeFitLookupTable for more detail about the looping
    // order.
    const static int kADC3Max = 800;
    const static int kADC1Min = -10;
    const static int kADC2RelMin = -140;
    const static int kADC2RelMax = +10;
    // To understand this you need to understand the constraints the looping in
    // genADCShapeFitLookupTable and be ready to think about the 3D shape this
    // produces in ADC space.
    const static unsigned int kNumTableEntries =
      (-kADC1Min*kADC3Max + // Rectangular part, where adc1 < 0
       (kADC3Max*(kADC3Max-1))/2)* // Triangular part, where 0 <= adc1 < adc0
      (kADC2RelMax-kADC2RelMin); // Thickness of the wedge due to adc2

  protected:
    // No copy constructors
    ADCShapeFitTable(const ADCShapeFitTable&) = delete;
    ADCShapeFitTable& operator=(const ADCShapeFitTable&) = delete;

    bool fIsND;
    bool fIsMC;
    std::array<uint16_t, kNumTableEntries> fTable;
    std::bitset<kNumTableEntries> fGood;
    mutable int fNHit, fNMiss;
  };
}

#endif
