////////////////////////////////////////////////////////////////////////
/// \brief   Helper for AttenCurve
/// \author  Chris Backhouse - bckhouse@caltech.edu
///
////////////////////////////////////////////////////////////////////////
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "Calibrator/func/AttenCache.h"
#include "Calibrator/func/AttenCurve.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/GeometryBase.h"
#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif

namespace calib
{
  //......................................................................
  AttenCache::AttenCache(int maxdiblock)
    : fMaxDiblock(maxdiblock), fInitialized(false), fInitializedUncalibratedMask(false)
  {
  }

  //......................................................................
  AttenCache::~AttenCache()
  {
    for(std::map<Key_t, AttenCurve*>::iterator it = fMap.begin();
	it != fMap.end(); ++it){
      delete it->second;
    }
  }

  //......................................................................
  void AttenCache::Add(AttenCurve* res)
  {
    fMap[Key_t(res->det, res->chan)] = res;
  }

  //......................................................................
  const AttenCurve* AttenCache::ForData(const geo::GeometryBase* geom,
					geo::OfflineChan chan)
  {
    // TODO - this won't behave right if we ask for more than one detector
    // TODO - or data and MC in the same job.

    if(fInitialized){
      if(!fIsRealData){
	std::cerr << "Attempting to mix data and MC hits in Calibrator. Aborting" << std::endl;
	abort();
      }
    }
    else{
      fIsRealData = true;
      fInitialized = true;
      LoadFromDatabase(geom, true);
    }

    return Get(TranslateChannelData(chan, geom));
  }

  //......................................................................
  const AttenCurve* AttenCache::ForMC(const geo::GeometryBase* geom,
				      geo::OfflineChan chan, bool maskUncalibratedChannels)
  {
    // TODO - this won't behave right if we ask for more than one detector
    // TODO - or data and MC in the same job.
    if(fInitialized){
      if(fIsRealData){
	std::cerr << "Attempting to mix data and MC hits in Calibrator. Aborting" << std::endl;
	abort();
      }
    }
    else{
      fIsRealData = false;
      fInitialized = true;
      LoadFromDatabase(geom, false);
    }
     
    if(!fInitializedUncalibratedMask && maskUncalibratedChannels){
      fInitializedUncalibratedMask = true;
      LoadFromDatabaseUncalibratedMask(geom);
    }

    if(maskUncalibratedChannels &&
       fMapUncalibratedMask.count(TranslateChannelData(chan, geom))){
      return 0;
    } 

    return Get(TranslateChannelMC(chan, geom));
  }

  //......................................................................
  const AttenCurve* AttenCache::Get(Key_t key) const
  {
    auto it = fMap.find(key);
    if(it != fMap.end()) return it->second;
    return 0;
  }

  //......................................................................
  std::string AttenCache::GetCSVPath(novadaq::cnv::DetId det,
                                     bool isData, bool points) const
  {
    if(!fUseCSVsFromUPS){
      if(points) return fPointsCSVSource; else return fConstsCSVSource;
    }

    const char* ups_path = getenv("CALIBCSVS_CSV_PATH");
    assert(ups_path);
    const char* pointsStr = points ? "points" : "consts";
    const char* detStr = (det == novadaq::cnv::kFARDET) ? "fd" : "nd";
    const char* mcStr = isData ? "data" : "mc";
    return TString::Format("%s/%s/calib_atten_%s.%s.%s.%s.csv", ups_path, fTag.c_str(), pointsStr, detStr, mcStr, fTag.c_str()).Data();
  }

  //......................................................................
  void AttenCache::LoadFromDatabase(const geo::GeometryBase* geom, bool isData)
  {
    // For the FD this will make a cache ~80MB. That's OK, but denser LOWESS
    // points would increase this rapidly.

    mf::LogInfo("AttenCache")
      << "Initializing attenuation cache for " << novadaq::cnv::DetInfo::GetName(geom->DetId()) << " from database...\n";


    nova::dbi::Table tblConsts("Calibration/tables/CalibAttenConstsVld.xml");
    tblConsts.SetDetector(geom->DetId());
    tblConsts.SetDataTypeMask(isData ? nova::dbi::kDataOnly : nova::dbi::kMCOnly);

    if(!fTag.empty()) tblConsts.SetTag(fTag);
    tblConsts.SetVerbosity(100);

    tblConsts.SetMinTSVld(1);
    tblConsts.SetMaxTSVld(100);

    bool ok = true;

    if(fUseCSVsFromUPS || !fConstsCSVSource.empty()){
      const std::string constsCSVSource = GetCSVPath(geom->DetId(),
                                                     isData, false);
      mf::LogInfo("AttenCache") << "Loading attenuation cache from "
				<< constsCSVSource << std::endl;
      ok = ok && tblConsts.LoadFromCSV(constsCSVSource);
      mf::LogInfo("AttenCache") << "Got " << tblConsts.NRow() << " rows from CSV" << std::endl;
    }
    else{
      ok = ok && tblConsts.Load();
    }

    if(!ok){
      // No safe way to handle this
      mf::LogError("AttenCache") << "Unable to load attenuation tables. Aborting"
                                 << std::endl;
      // I've definitely seen mf lose races with shutdown
      std::cerr << "Unable to load attenuation tables. Aborting"
                << std::endl;
      abort();
    }

    const int coeffCol   = tblConsts.GetColIndex("coeff_exp");
    const int attenCol   = tblConsts.GetColIndex("atten_length");
    const int bkgCol     = tblConsts.GetColIndex("background");
    const int loEdgeCol  = tblConsts.GetColIndex("edge_low");
    const int hiEdgeCol  = tblConsts.GetColIndex("edge_high");
    const int loCoeffCol = tblConsts.GetColIndex("coeff_low");
    const int hiCoeffCol = tblConsts.GetColIndex("coeff_high");
    const int chisqCol   = tblConsts.GetColIndex("chisq");

    const int Nconsts = tblConsts.NRow();
    for(int nconsts = 0; nconsts < Nconsts; ++nconsts){
      nova::dbi::Row* row = tblConsts.GetRow(nconsts);

      AttenCurve* res = AttenCurve::Uninitialized(geom->DetId(),
						  geo::OfflineChan(-1, -1));

      res->chan = geo::OfflineChan::FromDBValidityChan(row->Channel());

      ok = ok && row->Col(coeffCol  ).Get(res->coeff_exp);
      ok = ok && row->Col(attenCol  ).Get(res->atten_length);
      ok = ok && row->Col(bkgCol    ).Get(res->background);

      ok = ok && row->Col(loEdgeCol ).Get(res->edge_low);
      ok = ok && row->Col(hiEdgeCol ).Get(res->edge_high);
      ok = ok && row->Col(loCoeffCol).Get(res->coeff_low);
      ok = ok && row->Col(hiCoeffCol).Get(res->coeff_high);
      ok = ok && row->Col(chisqCol  ).Get(res->chisq);
      assert(ok);

      bool inMuonCatcher = false;
      const geo::PlaneGeo* geoplane = 0;

      const int plane = res->chan.Plane();
      if(isData && plane < 1000){
        // Normal data just has regular channel numbers
        geoplane = geom->Plane(plane);
        if(geom->DetId() == novadaq::cnv::kNEARDET &&
           plane >= int(geom->FirstPlaneInMuonCatcher())){
          inMuonCatcher = true;
        }
      }
      else{
        // Otherwise is merged views or MC
        geoplane = RepresentativePlane(geom, plane, inMuonCatcher);
      }

      if(!geoplane) continue;
      const geo::CellGeo* geocell = geoplane->Cell(res->chan.Cell());
      if(!geocell) continue;

      res->cell_length = 2*geocell->HalfL();

      if(inMuonCatcher){
        // Find a plane from the same view that's not in the muon catcher, to
        // find out how much the muon catcher cell must be displaced.
        int i = 0;
        const geo::PlaneGeo* regularPlane = geom->Plane(i);
        while(regularPlane->View() != geoplane->View())
          regularPlane = geom->Plane(++i);
        const geo::CellGeo* regularCell = regularPlane->Cell(res->chan.Cell());

        // Assume that the cells are aligned on the readout side
        res->center_offset = regularCell->HalfL()-geocell->HalfL();
      }

      res->initialized = true;

      Add(res);

      // Figure out which diblock, mark it calibrated
      const int db = plane/64+1;
      fDiblocks.insert(db);
    } // end for nconsts

    // Now fetch the LOWESS points and add them to the curves we've just
    // created.

    int totPoints = 0;

    // Only do chunked loading for FD data from DB
    if(!isData || geom->DetId() != novadaq::cnv::kFARDET || fUseCSVsFromUPS || !fConstsCSVSource.empty())
      fMaxDiblock = -1;

    for(int diblock = 1; diblock <= (fMaxDiblock >= 0 ? fMaxDiblock : 1); ++diblock){
      nova::dbi::Table tblPoints("Calibration/tables/CalibAttenPointsVld.xml");
      tblPoints.SetDetector(geom->DetId());
      tblPoints.SetDataTypeMask(isData ? nova::dbi::kDataOnly : nova::dbi::kMCOnly);

      if(!fTag.empty()) tblPoints.SetTag(fTag);
      tblPoints.SetVerbosity(100);

      tblPoints.SetMinTSVld(1);
      tblPoints.SetMaxTSVld(100);

      if(fMaxDiblock > 0){
	// Difference in vldChan encoding block-to-block
	const int stride = 1000*64*1000;
	tblPoints.SetChannelRange((diblock-1)*stride, diblock*stride-1);
	if(diblock == fMaxDiblock){
	  // If this is the last iteration, make sure to also pick up the
	  // special averaged constants out at plane 1000+
	  tblPoints.SetMaxChannel(1000*1000*1000+stride);
	}
      }

      if(fUseCSVsFromUPS || !fPointsCSVSource.empty()){
        const std::string pointsCSVSource = GetCSVPath(geom->DetId(),
                                                       isData, true);
        mf::LogInfo("AttenCache") << "Loading attenuation cache from "
                                  << pointsCSVSource << std::endl;
        ok = ok && tblPoints.LoadFromCSV(pointsCSVSource);
	mf::LogInfo("AttenCache") << "Got " << tblPoints.NRow() << " rows from CSV" << std::endl;
      }
      else{
        ok = ok && tblPoints.Load();
      }

      if(!ok){
	// No safe way to handle this
	mf::LogError("AttenCache") << "Unable to load attenuation tables. Aborting"
				   << std::endl;
	// I've definitely seen mf lose races with shutdown
	std::cerr << "Unable to load attenuation tables. Aborting"
		  << std::endl;
	abort();
      }

      const int Npoints = tblPoints.NRow();
      if(Npoints == 0) continue;
      totPoints += Npoints;

      const int wColPts      = tblPoints.GetColIndex("w");
      const int factorColPts = tblPoints.GetColIndex("factor");

      for(int npoints = 0; npoints < Npoints; ++npoints){
	nova::dbi::Row* row = tblPoints.GetRow(npoints);

	const geo::OfflineChan chan = geo::OfflineChan::FromDBValidityChan(row->Channel()/1000);

	float w, factor;
	ok = ok && row->Col(wColPts     ).Get(w);
	ok = ok && row->Col(factorColPts).Get(factor);
	assert(ok);

	auto it = fMap.find(Key_t(geom->DetId(), chan));
	assert(it != fMap.end());
	AttenCurve* curve = it->second;

	const unsigned int interpIdx = row->Channel()%1000;
	if(curve->interp_pts.size() <= interpIdx)
	  curve->interp_pts.resize(interpIdx+1);
	curve->interp_pts[interpIdx] = {w, factor};
      } // end for npoints
    } // end for diblock

    // Check that the number of points we got from the DB is the same as the
    // total in all the AttenCurves. If these numbers don't match, it means
    // there were duplicates or gaps.
    int pointsCount = 0;
    for(auto it: fMap) pointsCount += it.second->interp_pts.size();
    assert(pointsCount == totPoints);
  }

  //......................................................................
  void AttenCache::LoadFromDatabaseUncalibratedMask(const geo::GeometryBase* geom)
  {
    mf::LogInfo("AttenCache")
      << "Initializing data attentuation fit chisq values for " << novadaq::cnv::DetInfo::GetName(geom->DetId()) << " from database to mask uncalibrated channels...\n";
    
    nova::dbi::Table tblConsts("Calibration/tables/CalibAttenConstsVld.xml");
    tblConsts.SetDetector(geom->DetId());
    tblConsts.SetDataTypeMask(nova::dbi::kDataOnly);

    if(!fTag.empty()) tblConsts.SetTag(fTag);
    tblConsts.SetVerbosity(100);

    tblConsts.SetMinTSVld(1);
    tblConsts.SetMaxTSVld(100);

    bool ok = true;

    if(!fConstsCSVSource.empty()){
      mf::LogInfo("AttenCache") << "Trying to get data constants from the MC CSV file, no good way to handle this if you want to use masking. Aborting." << std::endl;
      abort();
    }
    else{
      if(fUseCSVsFromUPS){
        // Always ask for data, because we're masking the MC to match real data
        // conditions.
        const std::string constsCSVSource = GetCSVPath(geom->DetId(),
                                                       true, false);
        mf::LogInfo("AttenCache") << "Loading attenuation cache from "
                                  << constsCSVSource << std::endl;
        ok = ok && tblConsts.LoadFromCSV(constsCSVSource);
        mf::LogInfo("AttenCache") << "Got " << tblConsts.NRow() << " rows from CSV" << std::endl;
      }
      else{
        ok = ok && tblConsts.Load();
      }
    }

    if(!ok){
      // No safe way to handle this
      mf::LogError("AttenCache") << "Unable to load attenuation tables. Aborting"
                                 << std::endl;
      // I've definitely seen mf lose races with shutdown
      std::cerr << "Unable to load attenuation tables. Aborting"
                << std::endl;
      abort();
    }

    const int chisqCol = tblConsts.GetColIndex("chisq");

    const int Nconsts = tblConsts.NRow();
    for(int nconsts = 0; nconsts < Nconsts; ++nconsts){
      nova::dbi::Row* row = tblConsts.GetRow(nconsts);

      double chisq;
      ok = ok && row->Col(chisqCol).Get(chisq);
      assert(ok);

      const geo::OfflineChan chan = geo::OfflineChan::FromDBValidityChan(row->Channel());

      if(chisq <= 0 || chisq > 0.2){
	fMapUncalibratedMask.insert(Key_t(geom->DetId(), chan));
      } 

      // Figure out which diblock, mark it calibrated
      const int db = chan.Plane()/64+1;
      fDiblocks.insert(db);
    } // end for nconsts

  }

  //......................................................................
  int AttenCache::GetView(int plane, const geo::GeometryBase* geom) const
  {
    if(geom->DetId() == novadaq::cnv::kNEARDET){
      const int firstPlane = geom->FirstPlaneInMuonCatcher();
      const int isMuonCatcher = (firstPlane <= plane);
      return geom->Plane(plane)->View()+(2*isMuonCatcher);
    }

    return geom->Plane(plane)->View();
  }

  //......................................................................
  const geo::PlaneGeo* AttenCache::
  RepresentativePlane(const geo::GeometryBase* geom,
                      int view,
                      bool& inMuonCatcher) const
  {
    // Unfortunately a bunch of different conventions competing here

    // View-merged MC, or average constants data
    if(view == 0 || view == 1 || view == 1000 || view == 1001){
      inMuonCatcher = false;
      return geom->Plane(*geom->GetPlanesByView(geo::View_t(view)).begin());
    }

    // View-merged MC, or some scheme I'm not sure is still in use, or
    // averaged-constants MC
    if(view == 2 || view == 3 || view == 10 || view == 11 ||
       view == 1002 || view == 1003){
      assert(geom->DetId() == novadaq::cnv::kNEARDET);

      inMuonCatcher = true;
      view = view%2;

      const int mcp = geom->FirstPlaneInMuonCatcher();
      const geo::PlaneGeo* geoplane = geom->Plane(mcp);
      if(geoplane->View() != geo::View_t(view))
        geoplane = geom->Plane(geom->NextPlaneOtherView(mcp, +1));

      return geoplane;
    }

    assert(0 && "Bad encoded view");
  }

  //......................................................................
  AttenCache::Key_t AttenCache::
  TranslateChannelData(geo::OfflineChan chan,
		       const geo::GeometryBase* geom) const
  {
    const int db = chan.Plane()/64+1;
    if(!fDiblocks.count(db)){
      // No calibration for this diblock, use the averages instead.
      // Stored at channel 1,000,000+1000*view+cell
      const int view = GetView(chan.Plane(), geom);
      return Key_t(geom->DetId(),
		   geo::OfflineChan(1000+int(view), chan.Cell()));
    }

    return Key_t(geom->DetId(), chan);
  }

  //......................................................................
  AttenCache::Key_t AttenCache::
  TranslateChannelMC(geo::OfflineChan chan,
		     const geo::GeometryBase* geom) const
  {
    // If it's a muon catcher plane, check to see if there is a special entry
    // for it
    if(geom->DetId() == novadaq::cnv::kNEARDET &&
       chan.Plane() >= int(geom->FirstPlaneInMuonCatcher())){
      const int viewInt = GetView(chan.Plane(), geom);
      const Key_t ret(geom->DetId(), geo::OfflineChan(viewInt, chan.Cell()));
      if(Get(ret)) return ret;
    }
    // If not, fall back to the non-muon-catcher constants
      
    // This is the way we encode the MC parameters. Plane represents the view.
    // TODO, we should move to plane+1000 like the data?
    return Key_t(geom->DetId(),
		 geo::OfflineChan(geom->Plane(chan.Plane())->View(),
				  chan.Cell()));
  }

  //......................................................................
  bool AttenCache::Key_t::operator<(const Key_t& rhs) const
  {
    if(fDet < rhs.fDet) return true;
    if(fDet > rhs.fDet) return false;
    return fChan < rhs.fChan;
  }

} // end namespace calib
////////////////////////////////////////////////////////////////////////
