#include <iostream>
//#include <fstream>
#include <iomanip>
#include <vector>
#include <string>
#include <unordered_map>
#include <cstdlib>
#include <ctime>
#include <getopt.h>
#include <boost/filesystem.hpp>

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif
#include "RunHistory/RunHistory.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "DAQChannelMap/HardwareDisplay.h"

#include "TFile.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TTree.h"
#include "TMath.h"

std::string det="";
std::string csvFileName="";
std::string relFile="";

bool write2DB=false;
bool saveHistos=true;
bool write2CSV=false;
bool prodMode=false;
bool check4MissingFiles=false;

std::vector<int> runList;
std::vector<std::string> rel;
std::vector<std::string> prefix;
std::vector<std::string> suffix1;
std::vector<std::string> suffix2;
std::vector<std::string> rawSuffix;

//------------------------------------------------------------

void PrintUsage()
{
  std::cout << "Usage: getBadChannels [options] [releases] [detector] [run1] [run2] ... [runN]" << std::endl;
  std::cout << "options:\n";
  std::cout << "\t -d (--write2DB): enable database commits" << std::endl;
  std::cout << "\t -h (--nohistos): disable histogram filling/writing" << std::endl;
  std::cout << "\t -p (--prod): run in production mode, abort if any histo file is missing" << std::endl;
  std::cout << "\t -v (--check): only search for missing subrun OnMon files" << std::endl;
  std::cout << "\t -c (--tocsv) [filename]: write results to csv file" << std::endl;
}

//------------------------------------------------------------

bool ParseCLArgs(int argc, char* argv[])
{
  if (argc == 1) {
    PrintUsage();
    exit(10);
  }

  struct option long_options[] = {
    {"write2DB", no_argument, 0, 'd'},
    {"nohistos", no_argument, 0, 'h'},
    {"tocsv", 0, 0, 'c'},
    {"prod", 0, 0, 'p'},
    {"check", 0, 0, 'v'},
    {0,0,0,0}
  };
  
  while (1) {
    int optindx;
    int c = getopt_long(argc,argv,"dhc:pv",long_options,&optindx);
        
    if (c==-1) break;
    
    switch(c) {
    case 'h':
      saveHistos=false;
      break;
    case 'd':
      write2DB=true;
      break;
    case 'c':
      write2CSV=true;
      csvFileName=optarg;
      break;
    case 'p':
      prodMode=true;
      break;
    case 'v':
      check4MissingFiles=true;
      break;
    default:
      break;
    }
  }
  
  if (write2CSV) {
    std::cout << csvFileName << std::endl;
  }

  if ((argc-optind) < 3) {
    std::cerr << "Missing arguments." << std::endl;
    PrintUsage();
    exit(10);
  }
  
  if ((argc-optind) < 3) {
    std::cerr << "Unrecognized arguments." << std::endl;
    PrintUsage();
    exit(10);
  }

  relFile = argv[optind++];
  det = argv[optind++];
  for (; optind<argc; ++optind)
    runList.push_back(atoi(argv[optind]));

  if (runList.empty()) {
    std::cerr << "ERROR: run list is empty!" << std::endl;
    PrintUsage();
    exit(11);
  }

  return true;
}

//------------------------------------------------------------

std::string getOnMonFile(int run, int isr, std::string det)
{
  int irel=(int)rel.size()-1;
  bool foundFile = false;
  bool foundSMALL = false;
  bool foundMISSING = false;

  const char* dir1 = getenv("NOVA_DATA");
  if (!dir1) {
    std::cerr << "$NOVA_DATA undefined!" << std::endl;
    exit(9);

  }

  char fname[256];
  std::string detStr = "";

  for ( ; irel>=0; --irel) {
    sprintf(fname,"%s/nearline-OnMon/%s/%s/%06d/%08d/%s%08d%s%03d%s",dir1,det.c_str(),rel[irel].c_str(),int(run/100),run,prefix[irel].c_str(),run,suffix1[irel].c_str(),isr,suffix2[irel].c_str());
    
    if (boost::filesystem::exists(fname)) {
      foundFile = true;
      break;
    }

    // now look for SMALL files
    size_t pos = prefix[irel].find('_');
    
    if (pos != std::string::npos) {
      detStr = prefix[irel].substr(0,prefix[irel].find('_'));
      
      sprintf(fname,"%s/nearline-OnMon/%s/%s/%06d/%08d/%s_r%08d_s%02d_%s.raw.SMALL",dir1,det.c_str(),rel[irel].c_str(),int(run/100),run,detStr.c_str(),run,isr,rawSuffix[irel].c_str());
      if (boost::filesystem::exists(fname)) {
	foundSMALL = true;
      }

      sprintf(fname,"%s/nearline-OnMon/%s/%s/%06d/%08d/%s_r%08d_s%02d_%s.raw.MISSING",dir1,det.c_str(),rel[irel].c_str(),int(run/100),run,detStr.c_str(),run,isr,rawSuffix[irel].c_str());
      if (boost::filesystem::exists(fname)) {
	foundMISSING = true;
      }
    }
  }

  std::string retStr="";
  if (foundFile) 
    retStr = fname;
  else {
    if (foundSMALL)
      retStr = "SMALL";
    else if (foundMISSING)
      retStr = "MISSING";
  }
  
  return retStr;
  
}

//------------------------------------------------------------

int main(int argc, char *argv[])
{
  ParseCLArgs(argc, argv);

  std::ifstream relList;
  relList.open(relFile.c_str());

  std::string str1;
  std::string str2;
  std::string str3;
  std::string str4;
  std::string str5;

  while(relList >> str1 >> str2 >> str3 >> str4 >> str5) {
    rel.push_back(str1);
    prefix.push_back(str2);
    suffix1.push_back(str3);
    suffix2.push_back(str4);
    rawSuffix.push_back(str5);
  }
  relList.close();
    
  if (check4MissingFiles) {
    //    int nMissing=0;
    std::string OnMonFile;
    for (size_t i=0; i<runList.size(); ++i) {
      nova::dbi::RunHistory rh(runList[i],det);
      int nsr = rh.NSubruns();    
      for (int isr=0; isr<nsr; ++isr) {
	OnMonFile = getOnMonFile(runList[i],isr,det); 
	if (OnMonFile.empty()) {
	  std::cout 
	    // << "Missing OnMon file for " 
	    << runList[i] << " " << isr << std::endl;
	  //	  ++nMissing;
	}
	else if (OnMonFile == "MISSING") {
	  std::cout << "*MISSING* " << runList[i] << " " << isr << std::endl;
	}
      }
    }
    exit(0);
  }

  nova::dbi::Table* t = new nova::dbi::Table("ChannelInfo/ChanRateState.xml");
  t->SetDetector(det);
  t->SetDataTypeMask(nova::dbi::kMCOnly|nova::dbi::kDataOnly);
  int stateIdx = t->GetColIndex("state");
  nova::dbi::Row* r = t->NewRow(); 

  int detId = novadaq::cnv::DetInfo::GetId(det);
  daqchannelmap::DAQChannelMap* daqcmap = daqchannelmap::DAQChannelMap::getInstance(detId);
  
  daqchannelmap::HardwareDisplay hwd;
  hwd.SetupDet(detId);

  std::unordered_map<int,int> state;
  
  bool isFirstSR = true;
  for (size_t ir=0; ir<runList.size(); ++ir) {
    int run = runList[ir];
    nova::dbi::RunHistory rh(run,det);
    if (!rh.LoadDAQRunHistory()) {
      std::cerr << "RunHistory::LoadDAQRunHistory failed, aborting." << std::endl;
      exit(2);
    }
    int nsr = rh.NSubruns();
    
    rh.LoadHardwareInfo();

    char fname[256];

    sprintf(fname,"%s_BadChan_%d.root",det.c_str(),run);
    TFile* fout = 0;
    TH1F* hRateDist=0;
    TH1F* hRateDistAPD=0;
    TH1F* hRateDistAPDEnabled=0;
    
    if (saveHistos) {
      fout = new TFile(fname,"RECREATE");    
      
      hRateDist = new TH1F("hRateDist","Hit Rate Distribution;log_{10}(Hit Rate)",240,-6.,6.);
      hRateDistAPD = new TH1F("hRateDistAPD","Hit Rates for Channels w/ APDs; log_{10}(Hit Rate)",240,-6.,6.);
      hRateDistAPDEnabled = new TH1F("hRateDistAPDEnabled","Hit Rates for Enabled Channels w/ APDs; log_{10}(Hit Rate)",240,-6.,6.);
    }
    
    TH1F* hRateDistDCM[200];
    TH2F* hBadChanMask[200];
    for (int i=0; i<200; ++i) { 
      hRateDistDCM[i]=0;
      hBadChanMask[i]=0;
    }
    
    TH2F* h2;
    char hname[256];
    char htitle[256];
    float rate;

    int ndb = rh.NDiBlocks();

    std::unordered_map<int,int> badMap;
    
    for (int isr=0; isr<nsr; ++isr) {

      std::string OnMonFile = getOnMonFile(runList[ir],isr,det);
      if (OnMonFile.empty()) {
	std::cerr << "File not found for run.subrun " << run << "." << isr 
		  << ".  ";
	if (prodMode) {
	  std::cerr << "Aborting." << std::endl;
	  exit(3);
	}
	else {
	  std::cerr << "Skipping." << std::endl;
	  continue;
	}
      }
      else if (OnMonFile == "SMALL") {
	std::cerr << "Only SMALL files found for this run, skipping."
		  << std::endl;
	continue;
      }
      else if (OnMonFile == "MISSING") {
	std::cerr << "Only MISSING files found for this subrun, skipping."
		  << std::endl;
	continue;
      }

      std::cout << "Using file: " << OnMonFile << std::endl;

      TFile* fin = new TFile(OnMonFile.c_str());
      if (!fin) {
	std::cerr << "Cannot open file " << OnMonFile 
		  << ".  Skipping." << std::endl;
	continue;
      }
	  
      time_t tstart = run*10000+isr*100;
      r->SetVldTime(tstart);
      
      int dcmIdx=0;
      int nbad=0;
      
      for (int i=0; i<ndb; ++i) {
	
	const nova::dbi::RunHistory::DiBlock& db = rh.GetDiBlock(i,false);
	int idb = db.num;
	
	int unsigned ndcm =  db.dcm.size() ;
	if(detId == novadaq::cnv::kNEARDET && i == 3)
	  ndcm = 2;
	
	for (unsigned int j=0; j< ndcm; ++j, ++dcmIdx) {
	  int idcm = db.dcm[j].position;
	  
	  sprintf(hname,"PixelsRateDCM_%02d_%02d",idb,idcm);
	  h2 = (TH2F*)fin->Get(hname);
	  if (!h2) continue;
	  
	  if (saveHistos) {
	    if (!hRateDistDCM[dcmIdx]) {
	      fout->cd();
	      sprintf(hname,"hRateDist_%s_%02d_%02d",det.c_str(),idb,idcm);
	      
	      sprintf(htitle,"Hit Rate Distribution for %s;log_{10}(Hit Rate)",
		      db.dcm[j].appname.c_str());
	      hRateDistDCM[dcmIdx] = new TH1F(hname,htitle,240,-6.,6.);
	    }
	    sprintf(hname,"hBadChanMask_%s_%02d_%02d_%02d",det.c_str(),idb,idcm,
		    isr);
	    hBadChanMask[dcmIdx] = (TH2F*)h2->Clone(hname);	  
	    hBadChanMask[dcmIdx]->SetName(hname);
	  }
	  
	  unsigned int nfebs = db.dcm[j].feb.size();
	  
	  for (int ix=1; ix<=h2->GetXaxis()->GetNbins(); ++ix) {
	    for (int iy=1; iy<=h2->GetYaxis()->GetNbins(); ++iy) {
	      unsigned int ifeb;
	      unsigned int ipix;
	      hwd.XYFEB(&ifeb,ix-1,iy-1);
	      hwd.XYPix(&ipix,ix-1,iy-1);
	      
	      bool hasAPD = ( ifeb < nfebs && db.dcm[j].feb[ifeb].hasAPD );
	      
	      rate = h2->GetBinContent(ix,iy);
	      
	      if (rate>0.) 
		rate = TMath::Log10(rate);
	      else
		rate = -5.;
	      
	      if (saveHistos) {
		hRateDist->Fill(rate);
		if (hasAPD) { 
		  hRateDistAPD->Fill(rate);
		  hRateDistDCM[dcmIdx]->Fill(rate);
		}
	      }
	      
	      if (hasAPD) {
		
		daqchannelmap::dchan daqChan = daqcmap->encodeDChan(detId,idb,
								    idcm,ifeb,
								    ipix);
		daqchannelmap::lchan logChan = daqcmap->encodeLChan(daqChan);
		
		int plane = daqcmap->getPlane(logChan);
		int cell = daqcmap->getCell(logChan);
		geo::OfflineChan chan(plane, cell);
		int tchan = chan.ToDBValidityChan();
		
		bool didChange=true;
		int tstate=0;
		if (rate > 3.5) {
		  tstate=1;
		  ++nbad;
		}
		else if (rate < 0.5) {
		  tstate=2;
		  ++nbad;
		}

		// check if this state has been seen yet, if not, set the state
		if (isFirstSR ||
		    (state.find(tchan) == state.end())) {
		  state[tchan] = tstate;
		  badMap[tchan] = tstate;
		}
		else {
		  if (tstate != 0) badMap[tchan] = tstate;
		  
		  if (state[tchan] != tstate) {
		    state[tchan] = tstate;
		  }
		  else {
		    didChange=false;
		  }
		}
		
		if (didChange) {    
		  r->SetChannel(tchan);
		  r->Set(stateIdx,tstate);
		  t->AddRow(r);
		}
		
		if (saveHistos) {
		  hRateDistAPDEnabled->Fill(rate);
		  hBadChanMask[dcmIdx]->SetBinContent(ix,iy,tstate);
		}
	      }
	    }
	  }
	}
      }
      
      if (isFirstSR)
	isFirstSR = false;
      
      if (saveHistos) {
	fout->cd();
	for (int i=0; i<200; ++i) 
	  if (hBadChanMask[i]) hBadChanMask[i]->Write();
      }
      
      fin->Close();
      std::cout << "Run " << run << ", Subrun " << isr << " has " << t->NRow() << " rows, nbad = " << nbad << "." << std::endl;
    }
    
    std::unordered_map<int,int>::iterator badMapItr = badMap.begin();
    int nBadUnique=0;
    for (; badMapItr != badMap.end(); ++badMapItr) {
      if (badMapItr->second != 0) ++nBadUnique;
    }
    std::cout << nBadUnique << " unique bad channels" << std::endl;

    if (saveHistos) {
      fout->cd();
      for (int i=0; i<200; ++i) 
	if (hRateDistDCM[i]) hRateDistDCM[i]->Write();
      if (hRateDist) hRateDist->Write();
      if (hRateDistAPD) hRateDistAPD->Write();
      if (hRateDistAPDEnabled) hRateDistAPDEnabled->Write();
      
      fout->Close();
      delete fout;
    }
        
  }

  std::cerr << "Table has " << t->NRow() << " rows." << std::endl;

  if (write2DB) {
    if (t->NRow() > 0) {
      if (! t->Write()) {
	std::cerr << "Failed to write to database." << std::endl;
      }
    }
  }
  
  if (write2CSV) {
    if (t->NRow() > 0) {
      if (! t->WriteToCSV(csvFileName)) {
	std::cerr << "Failed to write to CSV file.  Aborting." << std::endl;
      }
    }
  }
  
  return 0;
}
