#include <iostream>
//#include <fstream>
#include <iomanip>
#include <vector>
#include <string>
#include <map>
#include <cstdlib>
#include <ctime>
#include <getopt.h>
#include <boost/filesystem.hpp>

#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif
#include "RunHistory/RunHistory.h"
#include "NovaDAQConventions/DAQConventions.h"

#include "TFile.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TTree.h"
#include "TMath.h"

int run=0;
std::string det="";
std::string csvFileName="";
std::string relFile="";

int nextRun=0;

bool write2DB=false;
bool saveHistos=true;
bool write2CSV=false;

//------------------------------------------------------------

void PrintUsage()
{
  std::cout << "Usage: getFEBDropouts [options] [releases] [detector] [run] " << std::endl;
  std::cout << "options:\n";
  std::cout << "\t -d (--write2DB): enable database commits" << std::endl;
  std::cout << "\t -h (--nohistos): disable histogram filling/writing" << std::endl;
  std::cout << "\t -c (--tocsv) [filename]: write results to csv file" << std::endl;
  std::cout << "\t -n (--nextrun) [run]: next run number" << std::endl;
}

//------------------------------------------------------------

bool ParseCLArgs(int argc, char* argv[])
{
  if (argc == 1) {
    PrintUsage();
    exit(10);
  }

  struct option long_options[] = {
    {"write2DB", no_argument, 0, 'd'},
    {"nohistos", no_argument, 0, 'h'},
    {"tocsv", 0, 0, 'c'},
    {"nextrun", 0, 0, 'n'},
    {0,0,0,0}
  };
  
  while (1) {
    int optindx;
    int c = getopt_long(argc,argv,"dhc:n:",long_options,&optindx);
        
    if (c==-1) break;
    
    switch(c) {
    case 'h':
      saveHistos=false;
      break;
    case 'd':
      write2DB=true;
      break;
    case 'c':
      write2CSV=true;
      csvFileName=optarg;
      break;
    case 'n':
      nextRun=atoi(optarg);
      break;
    default:
      break;
    }
  }
  
  if (write2CSV) {
    std::cout << csvFileName << std::endl;
  }

  if ((argc-optind) < 3) {
    std::cerr << "Missing arguments." << std::endl;
    PrintUsage();
    exit(10);
  }
  
  if ((argc-optind) > 3) {
    std::cerr << "Unrecognized arguments." << std::endl;
    PrintUsage();
    exit(10);
  }

  relFile = argv[optind++];
  det = argv[optind++];
  run = atoi(argv[optind]);

  return true;
}

//------------------------------------------------------------

bool GetBinRange(TH2F* h2, double fStartHour, int& ix0, int& ix1)
{
  ix0=0;
  ix1=0;

  if (!h2) return false;
  int nxbins = h2->GetNbinsX();

  // get lowest/first bin that looks reasonable

  int ix = h2->GetXaxis()->FindBin(fStartHour);
  int ixstart = ix-5;
  if (ixstart<0) ixstart+=nxbins;
  int ic=0;
  TH1F* h1=0;
  TH1F* h1n=0;
  for (; ic<10; ++ic) {
    int tx = ixstart+ic;
    if (tx>nxbins) tx %= nxbins;
    h1 = (TH1F*)h2->ProjectionY("dummy",tx,tx);
    int nent = h1->GetEntries();
    if (h1->GetEntries() == 0) continue;
    ix0 = tx;
    TH1F* h1n = (TH1F*)h2->ProjectionY("dummy2",(tx+1)%nxbins,(tx+1)%nxbins);
    float ratio = (float)nent/(float)h1n->GetEntries();
    if (ratio < 0.2) ix0 = (tx+1)%nxbins;
    break;
  }
  if (ic==10)
    return false;

  // get highest/last bin that looks reasonable

  ic=0;
  for (; ic<50; ++ic) {
    int tx = ix0+ic;
    if (tx>nxbins) tx %= nxbins;
    h1 = (TH1F*)h2->ProjectionY("dummy",tx,tx);
    int nent = h1->GetEntries();
    if (h1->GetEntries() > 0) continue;
    ix1 = tx;
    int tx2 = tx-1;
    if (tx2 == 0) tx2 = nxbins;
    TH1F* h1n = (TH1F*)h2->ProjectionY("dummy2",tx2,tx2);
    float ratio = (float)h1n->GetEntries()/(float)nent;
    if (ratio < 0.2) ix1 = tx2;
    break;
  }
  if (ic==10)
    return false;

  return true;
}

//------------------------------------------------------------

int main(int argc, char *argv[])
{
  ParseCLArgs(argc, argv);

  std::ifstream relList;
  relList.open(relFile.c_str());

  std::vector<std::string> rel;
  std::vector<std::string> prefix;
  std::vector<std::string> suffix1;
  std::vector<std::string> suffix2;

  std::string str1;
  std::string str2;
  std::string str3;
  std::string str4;
  std::string dummy;

  while(relList >> str1 >> str2 >> str3 >> str4 >> dummy) {
    rel.push_back(str1);
    prefix.push_back(str2);
    suffix1.push_back(str3);
    suffix2.push_back(str4);
  }
  relList.close();
  
  uint32_t fStartYear;  ///< Header: actual years since 0 A.D.
  uint16_t fStartMonth; ///< Header: month of first processed event (1-12)
  uint16_t fStartDay;   ///< Header: day of first processed event (01-31)
  double   fStartHour;  ///< Header: hour of first processed event (0.0 - 23.99999)
  uint32_t fEndYear;  ///< Header: actual years since 0 A.D.
  uint16_t fEndMonth; ///< Header: month of last processed event (1-12)
  uint16_t fEndDay;   ///< Header: day of last processed event (01-31)
  double   fEndHour;  ///< Header: hour of last processed event (0.0 - 23.99999)
  
  char fname[256];
  const char* dir1 = getenv("NOVA_DATA");
  if (!dir1) {
    std::cerr << "$NOVA_DATA undefined!" << std::endl;
    exit(9);
  }

  nova::dbi::RunHistory rh(run,det);
  if (!rh.LoadDAQRunHistory()) {
    std::cerr << "RunHistory::LoadDAQRunHistory failed, aborting." << std::endl;
    exit(2);
  }
  int nsr = rh.NSubruns();

  if (nsr == 0) {
    std::cerr << "getFEBDropouts: 0 subruns found for this run, skipping this run." << std::endl;
    exit(2);
  }
  if ((rh.TStop() - rh.TStart()) < 30*60) {
    std::cerr << "getFEBDropouts: this run lasted less than 30 minutes.  Skipping this run." << std::endl;
    exit(2);
  }

  int irel=(int)rel.size()-1;
  bool foundFile = false;

  for ( ; irel>=0; --irel) {
    for (int isr=0; isr < nsr; ++isr) {
      sprintf(fname,"%s/nearline-OnMon/%s/%s/%06d/%08d/%s%08d%s%03d%s",dir1,det.c_str(),rel[irel].c_str(),int(run/100),run,prefix[irel].c_str(),run,suffix1[irel].c_str(),isr,suffix2[irel].c_str());
      if (boost::filesystem::exists(fname)) {
	foundFile = true;
	break;
      }
    }
    if (foundFile) break;
  }

  if (!foundFile) {
    std::cerr << "File not found!" << std::endl;
    exit(0);
  }

  TFile* fin = new TFile(fname);
  if (!fin) {
    std::cerr << "File " << fname << " could not be opened! Aborting!" << std::endl;
    exit(3);
  }
  fin->Close();

  nova::dbi::Table* t = new nova::dbi::Table("ChannelInfo/FEBDropout.xml");
  t->SetDetector(det);
  t->SetDataTypeMask(nova::dbi::kMCOnly|nova::dbi::kDataOnly);
  int stateIdx = t->GetColIndex("state");

  rh.LoadHardwareInfo();

  nova::dbi::Row* r = t->NewRow();

  time_t t0 = rh.TStart();
  time_t tstart = t0;
  time_t tstart_save = t0;

  time_t tMax = rh.TStop();

  if (nextRun > 0) {
    nova::dbi::RunHistory rhNext(nextRun,det);
    if (rhNext.LoadDAQRunHistory()) 
      tMax = rhNext.TStart();
  }

  sprintf(fname,"%s_FEBDropout_%d.root",det.c_str(),run);
  TFile* fout = 0;
  TH1F* hRateDist=0;
  TH1F* hRateDistAPD=0;
  TH1F* hRateDistAPDEnabled=0;

  if (saveHistos) {
    fout = new TFile(fname,"RECREATE");    
    
    hRateDist = new TH1F("hFEBRateDist","FEB Hit Rate Distribution;log_{10}(Hit Rate)",240,-6.,6.);
    hRateDistAPD = new TH1F("hFEBRateDistAPD","FEB Hit Rates for Channels w/ APDs; log_{10}(Hit Rate)",240,-6.,6.);
    hRateDistAPDEnabled = new TH1F("hFEBRateDistAPDEnabled","FEB Hit Rates for Enabled Channels w/ APDs; log_{10}(Hit Rate)",240,-6.,6.);
  }
  /*  
  TH1F* hRateDistDCM[200];
  for (int i=0; i<200; ++i) { 
    hRateDistDCM[i]=0;
  }
  */

  TH2F* h2;
  TH1F* h1;
  char hname[256];

  int delta_t = 120;
  if (nsr > 1) {
    delta_t = (rh.TStop() - rh.TStart())/nsr;
  }

  int ndb = rh.NDiBlocks();

  std::map<int,bool> state;
  std::map<int,bool> badMap;

  bool isFirst = true;
  
  for (int isr=0; isr<nsr; ++isr) {
    sprintf(fname,"%s/nearline-OnMon/%s/%s/%06d/%08d/%s%08d%s%03d%s",dir1,det.c_str(),rel[irel].c_str(),int(run/100),run,prefix[irel].c_str(),run,suffix1[irel].c_str(),isr,suffix2[irel].c_str());

    if (!boost::filesystem::exists(fname)) {
      std::cerr << "File not found for run " << run << "." << isr 
		<< ".  Skipping." << std::endl;
      continue;
    }

    //    continue;

    fin = new TFile(fname);
    if (!fin) {
      std::cerr << "Cannot open file " << fname << ".  Skipping." << std::endl;
      continue;
    }

    TTree* hdr = (TTree*)fin->Get("Header");
    hdr->SetBranchAddress("StartYear",&fStartYear);
    hdr->SetBranchAddress("StartMonth",&fStartMonth);
    hdr->SetBranchAddress("StartDay",&fStartDay);
    hdr->SetBranchAddress("StartHour",&fStartHour);
    hdr->SetBranchAddress("EndYear",&fEndYear);
    hdr->SetBranchAddress("EndMonth",&fEndMonth);
    hdr->SetBranchAddress("EndDay",&fEndDay);
    hdr->SetBranchAddress("EndHour",&fEndHour);
    hdr->GetEvent(0);
    if (fStartYear < 110) {
      std::cerr << "WARNING: subrun " << isr << " appears to have an invalid "
		<< "start time recorded in OnMon.  Setting time to start of "
		<< "NOvA epoch" << std::endl;
      fStartYear = 110;
      fStartMonth = 0;
      fStartDay = 1;
      fStartHour = 0.;
    }

    if (fStartYear > 1900) {
      fStartYear -= 1900;
      --fStartMonth;
      fEndYear -= 1900;
      --fEndMonth;
    }

    int hour = int(fStartHour);
    int min = int((fStartHour-hour)*60.);
    int sec = int((fStartHour-hour-(float)min/60.)*3600.);
 
    struct tm t1;
    t1.tm_sec = sec;
    t1.tm_min = min;
    t1.tm_hour = hour;
    t1.tm_mday = fStartDay;
    t1.tm_mon = fStartMonth;
    t1.tm_year = fStartYear;

    tstart = timegm(&t1);

    hour = int(fEndHour);
    min = int((fEndHour-hour)*60.);
    sec = int((fEndHour-hour-(float)min/60.)*3600.);
 
    t1.tm_sec = sec;
    t1.tm_min = min;
    t1.tm_hour = hour;
    t1.tm_mday = fEndDay;
    t1.tm_mon = fEndMonth;
    t1.tm_year = fEndYear;

    time_t tend = timegm(&t1);

    int tdiff = tstart-tstart_save;

    // catch case where we have a strange start time in the first subrun
    if (isr == 0 && fStartYear < 113) {
      tstart = t0;
    }
    
    if (isr > 0 && (tdiff < 0 || tdiff > isr*delta_t*1.2)) {
      if (isr == 1) {
	std::cout << "***WARNING! Validity time is off by " << tdiff 
		  << " seconds! Run " << run << ", Subrun " << isr 
		  << ".  Assuming a delta_t of " << tdiff << " seconds."
		  << std::endl;
	tstart = tstart_save + tdiff;
      }
      else {
	std::cout << "***WARNING! Validity time is off by " << tdiff 
		  << " seconds! Run " << run << ", Subrun " << isr 
		  << ".  Assuming a delta_t of " << delta_t << " seconds."
		  << std::endl;
	tstart = tstart_save+isr*delta_t;
      }
    }
    else
      tstart_save = tstart;

    if (tstart > tMax) tstart = tMax-1;

    int nbad=0;
    int ngood=0;

    int dt=0;

    for (int i=0; i<ndb; ++i) {

      const nova::dbi::RunHistory::DiBlock& db = rh.GetDiBlock(i,false);
      int idb = db.num;

      sprintf(hname,"FEBHitsVsHourDB_%02d",idb);
      h2 = (TH2F*)fin->Get(hname);
      if (!h2) continue;

      int nybins = h2->GetNbinsY();

      // loop over x-axis (time) bins 
      bool isFirstx = true;
      int nxbins = h2->GetNbinsX();

      int ix0;
      int ix1;
      if (i==0 && isFirstx) {
	if (!GetBinRange(h2,fStartHour,ix0,ix1)) {
	  continue;
	}
	dt = (int)(h2->GetXaxis()->GetBinWidth(1)*3600);
	isFirst = false;
      }      

      float tx1 = h2->GetXaxis()->GetBinLowEdge(ix0);
      int hour = int(tx1);
      int min = int(float(tx1-hour)*60.);
      int sec = int(3600.*(float(tx1-hour)-float(min)/60.));
      
      struct tm t2;
      t2.tm_sec = sec;
      t2.tm_min = min;
      t2.tm_hour = hour;
      t2.tm_mday = fStartDay;
      t2.tm_mon = fStartMonth;
      t2.tm_year = fStartYear;
      
      time_t tv = timegm(&t2);

      int nx = ix1-ix0;
      if (nx < 0) {
	nx += nxbins;
      }

      for (int ix=0; ix<=nx; ++ix, tv+=dt) {
	int xbin = ix0 + ix;
	if (xbin > nxbins)
	  xbin %= nxbins;
	h1 = (TH1F*)h2->ProjectionY("dummy",xbin,xbin);
	
	if (tv > tend) break;

	//	std::cout << "xbin = " << xbin << std::endl;
	//	std::cout << tend << ", " << tv << std::endl;

	r->SetVldTime(tv);

	// now loop over all channels and determine if the FEB has dropped
	for (int iy=1; iy<= nybins; ++iy) {
	  int idcm = (iy-1)/64;
	  int ifeb = (iy-1)%64;
	  if (!db.dcm[idcm].isActive) continue;
	  bool hasAPD = db.dcm[idcm].feb[ifeb].hasAPD; 
	  if (!hasAPD) {
	    continue;
	  }
	  float rate = h1->GetBinContent(iy);
	  bool isLive = (rate > 0.);
	  
	  if (saveHistos) {	    
	    hRateDist->Fill(rate);
	    if (hasAPD) { 
	      hRateDistAPD->Fill(rate);
	      /*
		if (!hRateDistDCM[idcm]) {
		sprintf(hname,"RateDistDCM_%02d_%03d",idb,idcm);
		hRateDistDCM[idbidcm*4] = new TH1F(hname,hname,240,-6.,6.);
		}		
		hRateDistDCM[idcm]->Fill(rate);
	      */
	    }	   
	  }
	  
	  int tchan = idb*1000+iy-1;

	  if (isFirst && isFirstx) {	    
	    state[tchan] = isLive;
	    badMap[tchan] = isLive;
	    r->SetChannel(tchan);
	    r->Set(stateIdx,isLive);
	    t->AddRow(r);
	    if (!isLive) ++nbad;
	  }
	  else {
	    if (!isLive) {
	      badMap[tchan] = isLive;
	    }

	    if (state[tchan] != isLive) {
	      state[tchan] = isLive;
	      r->SetChannel(tchan);
	      r->Set(stateIdx,isLive);
	      t->AddRow(r);
	      if (!isLive) 
		++nbad;
	      else
		++ngood;
	    }
	  }

	  if (saveHistos) {
	    hRateDistAPDEnabled->Fill(rate);
	  }
	}
	isFirstx = false;
      }
    }
    isFirst = false;

    fin->Close();
    std::cout << "Run " << run << ", Subrun " << isr << " has " << t->NRow() << " rows, nDropOuts = " << nbad << ", nReturns = " << ngood << std::endl;

  }
  
  std::map<int,bool>::iterator badMapItr = badMap.begin();
  int nBadUnique=0;
  for (; badMapItr != badMap.end(); ++badMapItr) {
    if (!badMapItr->second) ++nBadUnique;
  }
  std::cout << nBadUnique << " unique FEB dropouts" << std::endl;
  
  if (write2DB) {
    if (t->NRow() > 0) {
      if (! t->Write()) {
	std::cerr << "Failed to write to database." << std::endl;
      }
    }
  }
  
  if (write2CSV) {
    if (t->NRow() > 0) {
      if (! t->WriteToCSV(csvFileName)) {
	std::cerr << "Failed to write to CSV file.  Aborting." << std::endl;
      }
    }
  }
  
  if (saveHistos) {
    fout->cd();
    /*
    for (int i=0; i<200; ++i) 
      if (hRateDistDCM[i]) hRateDistDCM[i]->Write();
    */

    if (hRateDist) hRateDist->Write();
    if (hRateDistAPD) hRateDistAPD->Write();
    if (hRateDistAPDEnabled) hRateDistAPDEnabled->Write();
    
    fout->Close();
  }
  
  return 0;
}
