#! /bin/bash

###############################################################################
# prints help text for the script
###############################################################################

help_text () {
    echo "" >&2
    echo "This script checks database bad channel masks for a specified " >&2
    echo "run range, or files in a dataset, or runs within a dataset.   " >&2
    echo "Note that you need to provide a detector and the path to the  " >&2
    echo "badchanvalidatejob.fcl file. " >&2
    echo "" >&2
    echo "usage: badChanValidate.sh [options] " >&2
    echo "options: " >&2
    echo "       -h|--help     : prints this help message " >&2
    echo "       -r|--runrange : comma separated run range limits." >&2
    echo "                       must provide only two run numbers" >&2
    echo "                       it can be used with -s|--samdef  " >&2
    echo "       -s|--samdef   : sam dataset. it can  be used with" >&2
    echo "                       -r|--runrange                    " >&2
    echo "       -d|--det      : detector name. available options:" >&2
    echo "                       near or far. required argument   " >&2
    echo "       -p|--path     : path to the badchanvalidate jobfcl">&2
    echo "                       default is current directory      ">&2
    echo "       -t|--threshold: Define a threshold in percentage " >&2
    echo "                       of mismatch. Subruns with mismatch">&2
    echo "                       % below threshold won't be reported">&2
    echo "" >&2
    echo "Example:                                                 " >&2
    echo "" >&2
    echo "source /path/to/badChanValidate.sh -s my-sam-dataset     " >&2
    echo "       -r 10548,10549 -d near -p /path/to/jobfcl/ -t 0.2 " >&2
    exit
}

###############################################################################
# process script arguments using getopt
###############################################################################

process_args () {

    ARGS=$(getopt -o s:r:d:p:t:h -l "samdef:runrange:det:path:threshold:help" -n "badChanValidate.sh" -- "$@");

    eval set -- "$ARGS";

    # Check for strange arguments
    if [ $? -ne 0 ] ; then
	echo -e "\e[01;31mNo arguments provided. Terminating...\e[0m" >&2 ;
	exit 1 ;
    fi

    # Check for non-zero number of arguments
    if [ $# -le 1 ] ; then
	echo -e "\e[01;31mNo arguments provided. Please see the help menu\e[0m" >&2 ;
	help_text
	exit 1 ;
    fi

    while true ; do
	case "$1" in
	    -s|--samdef)
		defname=$2 ; echo "def $2" ; shift 2;;
	    -r|--runrange)
		runs=$2 ; echo "run $2" ; shift 2;;
	    -d|--det)
		det=$2 ; shift 2;;
	    -p|--path)
		path=$2 ; shift 2;;
	    -t|--threshold)
		threshold=$2 ; shift 2;;
	    -h|--help)
		help_text ; shift;;
	    --)	shift ; break ;;
            *) echo "Internal error!" ; exit 1 ;;
	esac
    done

    # parse comma separated run numbers into an array
    rarray=(${runs//,/ })

    # Check that only two run numbers are provided, if not using a sam dataset
    if [ -z ${defname} ]; then
	if [ ${#rarray[@]} -ne "2" ]; then
     	    echo -e "\e[01;31mError: Only two run numbers acceptable!\e[0m" >&2 ;
     	    exit 1 ;
	fi
    elif [ ${#rarray[@]} -ne "2" ] && [ ${#rarray[@]} -ne "0" ]; then
	echo -e "\e[01;31mError: Only two or no run numbers acceptable!\e[0m" >&2 ;	
     	exit 1 ;
    fi
    
    
    # Check that detector name is set
    if [ -z ${det} ] || ([ "${det}" != "far" ] && [ "${det}" != "near" ]) ; then
	echo -e "\e[01;31mError: Detector name not provided \e[0m" >&2 ;
	echo -e "\e[01;31m       or is invalid. Use far or near\e[0m" >&2 ;
	exit 1 ;
    fi

    # Check if the job fcl exists in dir
    if [ -z "${path}" ]; then path="${PWD}"; fi
    fcl=${path}/"badchanvalidatejob.fcl"
    if [ ! -f "${fcl}" ]; then
	echo -e "\e[01;31mError: badchanvalidate.fcl does not exist\e[0m" >&2 ;
	echo -e "\e[01;31m       in ${path}                        \e[0m" >&2 ;
    fi

    if [ ${rarray[0]} -le ${rarray[1]} ]; then
	runlow=${rarray[0]}
	runhigh=${rarray[1]}
    else
	runlow=${rarray[1]}
	runhigh=${rarray[0]}
    fi

    # if no threshold for percentage mismatch is specified, set it to 0.0
    if [ -z "${threshold}" ]; then
	threshold=0.00
    fi
}

###############################################################################
# Find files to validate badchans on and dump them in a text file
###############################################################################

list_files () {
    if [ -n "${defname}" ] && [ -z "${runs}" ] ; then
	`samweb list-definition-files $defname > list_of_files.txt`;
    elif [ -n "${defname}" ] && [ -n "${runs}" ] ; then
	`samweb list-files "defname: $defname and online.runnumber >= $runlow and online.runnumber <= $runhigh" > list_of_files.txt`;
    else
	# we will now have to construct file names
	tier="artdaq"
	# use cosmic trigger for far and numi for near
	if [ "${det}" == "far" ] ; then
	    trig="02"
	    release="S14-07-18"
	    prefix="fardet"
	    runinfo_cmd="NovaRunInfo -d FarDet "
	else
	    trig="00"
	    release="FA14-10-03"
	    prefix="neardet"
	    runinfo_cmd="NovaRunInfo -d NearDet "
	fi

	for run in `seq $runlow $runhigh`; do

	    # Check how many subruns are in this run with NovaRunInfo
	    runinfo=`${runinfo_cmd} -r ${run} `
	    parse_nsubrun=$( echo $runinfo | sed -e 's#.*Number of subruns: \([0-9]*\).*#\1# ' )
	    nsubruns=`expr $parse_nsubrun - 1`

	    for subrun in `seq 0 $nsubruns`; do

		filename=$(printf "%s_r%.8d_s%.2d_t%s_%s_v1_data.%s.root\n" ${prefix} ${run} ${subrun} ${trig} ${release} ${tier});
		# Check that the files exists first
		locate_status=`ifdh locateFile $filename`
		if [ $? == 0 ]; then
		    echo "$filename" >> list_of_files.txt
		fi
	    done
	done
    fi
}

###############################################################################
# run badchanvalidate job on the list of files
###############################################################################

run_validate() {

    outputfile="bad_chan_stats.txt"
    echo -e "Run Subrun Mismatch% " | column -t >> $outputfile

    while read filename
    do
	echo $filename
	# extract run and subrun info, so we may find the onmon file
	run_plus=${filename#*_r000}
	run=${run_plus%_s*}
	subrun_plus=${filename#*_s}
	subrun=${subrun_plus%_t*}

	temp_fcl="badchanvalidate_r${run}_s${subrun}.fcl"
	cp $fcl $temp_fcl
        hist_file="badchanval_r${run}_s${subrun}_hist.root"

	omname=$(printf "*_nearline_r%.8d_s%.3d*onmon.root\n" ${run#0} ${subrun#0});
	if [ "$det" == "far" ] ; then
	    omdir="/nova/data/nearline-OnMon/FarDet"
	else
	    omdir="/nova/data/nearline-OnMon/NearDet/FA14-09-23"
	    omdir2="/nova/data/nearline-OnMon/NearDet/S15-02-05"
	fi

	# the "-print -quit" part makes it quit after finding the 
        # first instance
	echo "find ${omdir} -iname ${omname} -print -quit"
	omfile=`find ${omdir} -iname ${omname} -print -quit`

	if [ -z "$omfile" ] ; then
	    omfile=`find ${omdir2} -iname ${omname} -print -quit`
	    if [ -z "$omfile" ] ; then
		echo -e "\e[01;31mOnmon file for run ${run}, subrun ${subrun} not found.\e[0m"
		echo -e "\e[01;31mFailed search: find ${omdir} -iname ${omname} -print -quit .\e[0m"
	    fi
	fi

	echo 'physics.analyzers.badchanval.OnmonFilename: "'${omfile}'"' >> $temp_fcl
	echo 'services.TFileService.fileName: "'${hist_file}'"' >> $temp_fcl

	if [ -n "$temp_fcl" ] ; then
	    output=$((nova -c ${temp_fcl} `samweb2xrootd ${filename}`) 2>&1 )
	    maskdiff=$(echo $output | sed -e 's#.*Difference between DB and OnMon masks : \([0-9]*\.[0-9]*\).*#\1# ')
	    echo "threshold and maskdiff ${threshold} ${maskdiff} "
	    if [ $(bc <<< "$maskdiff>$threshold") == 1 ]; then
		echo -e "${run} ${subrun} ${maskdiff}% " | column -t >> $outputfile
	    fi
	fi

    done < "list_of_files.txt"
}

###############################################################################
# main function
###############################################################################

main () {
    process_args $*

    if [ ! -d badchan_temp ]; then
	mkdir badchan_temp
    fi
    cd badchan_temp

    list_files
    run_validate

}

main $*
