////////////////////////////////////////////////////////////////////////
/// \brief   Module to monitor APD behaviour in realtime
/// \author  Kirk Bays
/// \date    May 2012
////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <math.h>
#include <sstream>
#include <string>

// ROOT includes
#include "TTree.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "CMap/CMap.h"
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "NovaDatabase/Table.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "NovaDAQConventions/DAQConventions.h"

#define NCHAN 500000

static double grate, grate_low, grate_high;  ///< keep track of hit rate
static int grun, gsubrun, numev;  ///< run, subrun and number of events
static int chan[NCHAN];  ///< channel number
static int hlo[NCHAN], hhi[NCHAN], nn[NCHAN], tfhits[NCHAN];  ///< channel specific: hit rates, nearest neighbor rate, num hits in fit tracks
static float pe[NCHAN], pecorr[NCHAN], minpe[NCHAN];  ///< channel specific pe, pecorr, minpe
static float pesq[NCHAN], pecorrsq[NCHAN];  ///< channel specific pe, pecorr squared
static time_t gtfirst, gtlast;  ///<  first and last even times
static unsigned int Idx;  ///< keep track of the total number of channels seen


namespace apdw
{
  class APDWatch : public art::EDAnalyzer
  {
  public:
    explicit APDWatch(fhicl::ParameterSet const &pset);
    virtual ~APDWatch();

    void beginJob();
    void analyze(art::Event const& evt);
    void endJob();

  protected:

    TTree* fNtTree;

  private:

    float convertAPDTemp(unsigned short adcCounts);
    void FillDAQDB();

  };

}

/////////////////////////////////////////////////////////////////////////
namespace apdw{

  //----------------------------------------------------------------------
  APDWatch::APDWatch(fhicl::ParameterSet const &pset)
    : EDAnalyzer(pset)
  {
  }

  //----------------------------------------------------------------------
  APDWatch::~APDWatch()
  {
  }

  //----------------------------------------------------------------------
  void APDWatch::beginJob()
  {
    grate = grate_low = grate_high = numev = Idx = 0;  // initialize

    for(unsigned int n = 0; n < NCHAN; ++n){
      chan[n] = hlo[n] = hhi[n] = pe[n] = pecorr[n] = nn[n] = tfhits[n] = 0;
      minpe[n] = 9999.;
    }
  }

  //----------------------------------------------------------------------
   void APDWatch::analyze(art::Event const& evt)
  {

    art::Timestamp ts = evt.time();
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<cmap::CMap> cmap;

    numev++;  // tally number of events

    time_t uevt_sec = ts.timeHigh(); //utval >> 32 & 0xFFFFFFFFF;

    int nncheck[10000][3] = {{0}};

    if(grate == 0) {  // get some info only once, for first event only
      gtfirst = uevt_sec;
      grun = evt.run();
      gsubrun = evt.subRun();
    }

    if(grate > 0) gtlast = uevt_sec;

    art::Handle< std::vector<rb::CellHit> > chits;
    evt.getByLabel("calhit", chits);

    //Get Track
        art::Handle< std::vector<rb::Track> >track;
        try{
          evt.getByLabel("cosmictrack", track);}
        catch(...) {
          mf::LogWarning("APDWatch") <<"Could not get a track ";
        }

    for(unsigned int n = 0; n < chits->size(); ++n){  // loop over all hits for basic info
      art::Ptr<rb::CellHit> chit(chits,n);

      int32_t tchan = chit->DaqChannel();
      bool exist = false;  // if channel already exists in list

      nncheck[n][0] = tchan;
      nncheck[n][1] = chit->TNS();

      for(unsigned int x = 0; x < Idx; ++x){  // check if channel already in list
        if(chan[x]==tchan) {  // it is, so just add info for this hit
          exist = true;
          pe[x] = pe[x] + chit->PE();
          pesq[x] = pesq[x] + (chit->PE() * chit->PE());
          nncheck[n][2] = x;
          if(chit->PE() < minpe[x]) minpe[x] = chit->PE();
          if(chit->PE() > 100) hhi[x]++;
          else hlo[x]++;
          break;
        }
      }

      if(!exist) {  // channel wasn't already in list, so add it
        chan[Idx] = tchan;
        pe[Idx] = chit->PE();
        pesq[Idx] = chit->PE() * chit->PE();
        nncheck[n][2] = Idx;
        if(chit->PE() < minpe[Idx]) minpe[Idx] = chit->PE();
        if(chit->PE() > 100) hhi[Idx] = 1;
        else hlo[Idx] = 1;
        if(Idx<NCHAN) Idx++;
        else {
          std::cout << "too many channels!" << std::endl;
          break;
        }
      }

      grate = grate + 1;  // total hit rate for event

      if(chit->PE() > 100) {
        grate_high = grate_high + 1;
      }
      else {
        grate_low = grate_low + 1;
      }
    }

    for(unsigned int i = 0; i < track->size(); ++i) {

      art::Ptr<rb::Track> trk(track,i);
      TVector3 vtx = trk->Start();
      TVector3 tkstop = trk->Stop();

      TVector3 p0 = (tkstop-vtx).Unit();

      unsigned int tracksize = 0;

      if(trk->Is3D()) tracksize = trk->NCell();  // need a 3-D track to calibrate with
      else if(trk->Is2D()) tracksize = 0;
      else std::cout << "Track isn't 2-D or 3-D???" << std::endl;

      for(unsigned int n = 0; n < tracksize; ++n){
        art::Ptr<rb::CellHit> chit = trk->Cell(n);

        int32_t tchan = chit->DaqChannel();

        bool exist = false;  // if channel already exists in list

        double xyz[3];
        geom->Plane(chit->Plane())->Cell(chit->Cell())->GetCenter(xyz);

        const double w = (vtx + p0*((xyz[2]-vtx.Z())/p0.Z()))[1-chit->View()];  // w from cosmic track fit

        for(unsigned int x = 0; x < Idx; ++x){  // check if channel already in list
          if(chan[x]==tchan) {  // it is, so just add info for this hit
            exist = true;
            rb::CellHit chit2 = *chit;
            chit2.SetMC();
	    
            rb::RecoHit rhit(cal->MakeRecoHit(*chit, w));

            tfhits[x]++;
            //            std::cout << "PECORR???" << rhit.PECorr() << std::endl;
            if(rhit.IsCalibrated()) {
              pecorr[x] = pecorr[x] + rhit.PECorr();
              pecorrsq[x] = pecorrsq[x] + (rhit.PECorr() * rhit.PECorr());
            }
            break;
          }
        }

        if(!exist) {  // channel wasn't already in list??  Should be for second loop
          std::cout << " FOUND AN EVENT IN TRACK CLUSTER NOT IN ALL HITS CLUSTER!?!?" << std::endl;
        }
      }
    }

    for(unsigned int n = 0; n < chits->size(); ++n){  // loop over hits again to look for NN hits
      double countnn = 0;  // keep track of how many times a neighboring cell is hit in time
      for(unsigned int x = 0; x < chits->size(); ++x){  // loop over hits to compare to
        if(abs(nncheck[n][1]-nncheck[x][1]) < 600 && nncheck[n][0]!=nncheck[x][0]) {  // check time (600 ns)
          if(cmap->Map()->getDiBlock(nncheck[n][0])==cmap->Map()->getDiBlock(nncheck[x][0]) &&
             cmap->Map()->getDCM(nncheck[n][0])==cmap->Map()->getDCM(nncheck[x][0]) &&
             cmap->Map()->getFEB(nncheck[n][0])==cmap->Map()->getFEB(nncheck[x][0]) ) { // check same APD
            int pix1 = cmap->Map()->getPixel(nncheck[n][0]);
            int pix2 = cmap->Map()->getPixel(nncheck[x][0]);

            /*
            if( (pix1==0 &&pix2==8)  || (pix1==1 &&pix2==9)  || (pix1==2 &&pix2==10) ||  (pix1==3 &&pix2==11) ||
                (pix1==4 &&pix2==12) || (pix1==5 &&pix2==13) || (pix1==6 &&pix2==15) ||  (pix1==7 &&pix2==14) ||
                (pix1==8 &&pix2==0)  || (pix1==8 &&pix2==23) || (pix1==9 &&pix2==1)  ||  (pix1==9 &&pix2==22) ||
                (pix1==10&&pix2==2)  || (pix1==10&&pix2==21) || (pix1==11&&pix2==3)  ||  (pix1==11&&pix2==20) ||
                (pix1==12&&pix2==4)  || (pix1==12&&pix2==19) || (pix1==13&&pix2==5)  ||  (pix1==13&&pix2==18) ||
                (pix1==14&&pix2==7)  || (pix1==14&&pix2==17) || (pix1==15&&pix2==6)  ||  (pix1==15&&pix2==16) ||
                (pix1==16&&pix2==15) || (pix1==16&&pix2==25) || (pix1==17&&pix2==14) ||  (pix1==17&&pix2==24) ||
                (pix1==18&&pix2==13) || (pix1==18&&pix2==26) || (pix1==19&&pix2==12) ||  (pix1==19&&pix2==27) ||
                (pix1==20&&pix2==11) || (pix1==20&&pix2==28) || (pix1==21&&pix2==10) ||  (pix1==21&&pix2==29) ||
                (pix1==22&&pix2==9)  || (pix1==22&&pix2==30) || (pix1==23&&pix2==8)  ||  (pix1==23&&pix2==31) ||
                (pix1==24&&pix2==17) || (pix1==25&&pix2==16) || (pix1==26&&pix2==18) ||  (pix1==27&&pix2==19) ||
                (pix1==28&&pix2==20) || (pix1==29&&pix2==21) || (pix1==30&&pix2==22) ||  (pix1==31&&pix2==23) ) {  // look for hits that are neighbors
              countnn++;
            }
            */

            if(pix1 == 0) {
              if(pix2 == 1 || pix2 == 3 || pix2 == 8 || pix2 == 9 || pix2 == 11) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 1) {
              if(pix2 == 0 || pix2 == 9 || pix2 == 8) countnn = countnn + 1.0/3.0;
            }

            if(pix1 == 2) {
              if(pix2 == 5 || pix2 == 3 || pix2 == 10 || pix2 == 13 || pix2 == 11) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 3) {
              if(pix2 == 0 || pix2 == 2 || pix2 == 8 || pix2 == 10 || pix2 == 11) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 4) {
              if(pix2 == 5 || pix2 == 7 || pix2 == 12 || pix2 == 13 || pix2 == 14) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 5) {
              if(pix2 == 4 || pix2 == 2 || pix2 == 12 || pix2 == 13 || pix2 == 10) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 6) {
              if(pix2 == 14 || pix2 == 7 || pix2 == 15) countnn = countnn + 1.0/3.0;
            }

            if(pix1 == 7) {
              if(pix2 == 6 || pix2 == 4 || pix2 == 12 || pix2 == 15 || pix2 == 14) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 8) {
              if(pix2 == 0 || pix2 == 3 || pix2 == 1 || pix2 == 9 || pix2 == 11) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 9) {
              if(pix2 == 1 || pix2 == 0 || pix2 == 8) countnn = countnn + 1.0/3.0;
            }

            if(pix1 == 10) {
              if(pix2 == 2 || pix2 == 3 || pix2 == 5 || pix2 == 13 || pix2 == 11) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 11) {
              if(pix2 == 0 || pix2 == 3 || pix2 == 2 || pix2 == 8 || pix2 == 10) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 12) {
              if(pix2 == 4 || pix2 == 5 || pix2 == 7 || pix2 == 13 || pix2 == 14) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 13) {
              if(pix2 == 4 || pix2 == 5 || pix2 == 2 || pix2 == 12 || pix2 == 10) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 14) {
              if(pix2 == 4 || pix2 == 6 || pix2 == 7 || pix2 == 12 || pix2 == 15) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 15) {
              if(pix2 == 6 || pix2 == 7 || pix2 == 14) countnn = countnn + 1.0/3.0;
            }

            if(pix1 == 16) {
              if(pix2 == 17 || pix2 == 24 || pix2 == 25) countnn = countnn + 1.0/3.0;
            }

            if(pix1 == 17) {
              if(pix2 == 16 || pix2 == 19 || pix2 == 24 || pix2 == 25 || pix2 == 27) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 18) {
              if(pix2 == 21 || pix2 == 19 || pix2 == 29 || pix2 == 26 || pix2 == 27) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 19) {
              if(pix2 == 17 || pix2 == 18 || pix2 == 24 || pix2 == 26 || pix2 == 27) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 20) {
              if(pix2 == 21 || pix2 == 23 || pix2 == 28 || pix2 == 29 || pix2 == 31) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 21) {
              if(pix2 == 18 || pix2 == 20 || pix2 == 26 || pix2 == 28 || pix2 == 29) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 22) {
              if(pix2 == 23 || pix2 == 30 || pix2 == 31) countnn = countnn + 1.0/3.0;
            }

            if(pix1 == 23) {
              if(pix2 == 20 || pix2 == 22 || pix2 == 28 || pix2 == 30 || pix2 == 31) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 24) {
              if(pix2 == 16 || pix2 == 17 || pix2 == 19 || pix2 == 25 || pix2 == 27) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 25) {
              if(pix2 == 16 || pix2 == 17 || pix2 == 24) countnn = countnn + 1.0/3.0;
            }

            if(pix1 == 26) {
              if(pix2 == 18 || pix2 == 19 || pix2 == 21 || pix2 == 27 || pix2 == 29) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 27) {
              if(pix2 == 17 || pix2 == 18 || pix2 == 19 || pix2 == 24 || pix2 == 26) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 28) {
              if(pix2 == 20 || pix2 == 21 || pix2 == 23 || pix2 == 29 || pix2 == 31) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 29) {
              if(pix2 == 18 || pix2 == 20 || pix2 == 21 || pix2 == 26 || pix2 == 28) countnn = countnn + 1.0/5.0;
            }

            if(pix1 == 30) {
              if(pix2 == 22 || pix2 == 23 || pix2 == 31) countnn = countnn + 1.0/3.0;
            }

            if(pix1 == 31) {
              if(pix2 == 20 || pix2 == 22 || pix2 == 23 || pix2 == 28 || pix2 == 30) countnn = countnn + 1.0/5.0;
            }

          }
        }
      }
      nn[nncheck[n][2]] += countnn;
    }
  }
  //  return;

  //----------------------------------------------------------------------
  void APDWatch::endJob()
  {
    //    std::cout << " APDWatch ended and saw : " << Idx << " unique channels " << std::endl;
    //    std::cout << " Run: " << grun << " Subrun: " << gsubrun << " Number of Events: " << numev << std::endl;
    //    std::cout << " rates: all: " << grate << " low: " << grate_low << " high: " << grate_high << std::endl;
    //    std::cout << "Data: i, chan num, pe, pecorr, minpe, tfhits, hlo, hhi" << std::endl;
    //    for (int i = 1; i<50; i++) {
    //      std::cout << " " << i << " " << chan[i] << " " << pe[i] << " " << pecorr[i] << " " << minpe[i] << " " << tfhits[i] << " " << hlo[i] << " " << hhi[i] << std::endl;
    //    }

    char fFirstEventTimeStampStr[256];
    char fLastEventTimeStampStr[256];

    //    FillDAQDB();  // in case we need to remake; adds 5-10 minutes

    struct tm * fFirstEventTimeStampinfo;
    struct tm * fLastEventTimeStampinfo;

    fFirstEventTimeStampinfo = localtime(&gtfirst);
    strftime(fFirstEventTimeStampStr,256,"%Y-%m-%d %H:%M:%S", fFirstEventTimeStampinfo);

    std::cout<<"Time of First Event=           "    <<fFirstEventTimeStampStr<<std::endl;

    fLastEventTimeStampinfo = localtime(&gtlast);
    strftime(fLastEventTimeStampStr,256,"%Y-%m-%d %H:%M:%S", fLastEventTimeStampinfo);

    std::cout<<"Time of Last Event=           "    <<fLastEventTimeStampStr<<std::endl;

    int telapsed = gtlast-gtfirst;

    std::cout << "Time elapsed: " << telapsed << " seconds" << std::endl;

    if(telapsed==0) {
      telapsed=1;
      std::cout << "Time elapsed is 0 seconds; cheat and call it 1 second to prevent INFs" << std::endl;
    }

    nova::database::Table* t = new nova::database::Table(std::string("/Commissioning/tables/Header.xml"));

    //    t->SetVerbosity(100);
    t->SetIgnoreEnvVar(true);
    t->SetDetector(novadaq::cnv::kFARDET);
    t->SetDBInfo("nova_dev","ifdbdev.fnal.gov","5433","bays");
    //    t->SetPasswordFile("/home/bays/dev/APDWatch/dbpwd.txt");

    nova::database::Row* r = t->NewRow();
    r->Set("subrun",                    gsubrun);
    r->Set("run",                       grun);
    r->Set("time_elapsed",              telapsed);
    r->Set("time_first",                fFirstEventTimeStampStr);
    r->Set("time_last",                 fLastEventTimeStampStr);
    r->Set("nevents",                   numev);
    r->Set("rate_low",                  grate_low/telapsed);
    r->Set("rate_hi",                   grate_high/telapsed);
    r->Set("num_hit_low",               grate_low);
    r->Set("num_hit_hi",                grate_high);

    t->AddRow(r);
    delete r;
    t->WriteToDB();

    std::cout << "Wrote header to DB..." << std::endl;

    nova::database::Table* ch = new nova::database::Table(std::string("/Commissioning/tables/Channel.xml"));

    //    ch->SetVerbosity(100);
    ch->SetIgnoreEnvVar(true);
    ch->SetDetector(novadaq::cnv::kFARDET);
    ch->SetDBInfo("nova_dev","ifdbdev.fnal.gov","5433","bays");
    //    ch->SetPasswordFile("/home/bays/dev/APDWatch/dbpwd.txt");

    for(unsigned int n = 0; n < Idx; ++n){

      double thi = hhi[n];
      double tlo = hlo[n];
      double num_nn = nn[n];

      nova::database::Row* r = ch->NewRow();
      r->Set("subrun",                    gsubrun);
      r->Set("channel",                   chan[n]);
      r->Set("run",                       grun);
      r->Set("pe",                        pe[n]/(hhi[n]+hlo[n]));
      r->Set("pecorr",                    pecorr[n]/(hhi[n]+hlo[n]));
      r->Set("rate_low",                  tlo/telapsed);
      r->Set("rate_hi",                   thi/telapsed);
      r->Set("nhits",                     hhi[n]+hlo[n]);
      r->Set("nn_rate",                   num_nn/telapsed);
      r->Set("pesq",                      pesq[n]/(hhi[n]+hlo[n]));
      r->Set("pecorrsq",                  pecorrsq[n]/(hhi[n]+hlo[n]));
      r->Set("ntrackhits",                tfhits[n]);
      r->Set("minpe",                     minpe[n]);

      ch->AddRow(r);
      //      delete r;
    }
    std::cout << "Write channel info to DB... this might take a bit..." << std::endl;
    ch->WriteToDB();
  }


  float APDWatch::convertAPDTemp(unsigned short adcCounts) {
    /* The following formula will be used to convert ADC counts to degrees
     celsius:
       temp = a + bx + cx^2 + dx^3 + ex^4
    */
    const float a = 80.55652;
    const float b = -9.769789e-2;
    const float c = 4.462396e-5;
    const float d = -1.054843e-8;
    const float e = 8.3267575e-13;

    return a + b * adcCounts + c * powf(adcCounts, 2.0) + d *
      powf(adcCounts, 3.0) + e * powf(adcCounts, 4.0);
  }

  void APDWatch::FillDAQDB()  // write out DAQ Channel Map info to database; only needs to be done once ever
  {
    nova::database::Table* t = new nova::database::Table(std::string("/Commissioning/tables/DAQ.xml"));
    art::ServiceHandle<cmap::CMap>    cmap;

    //  t->SetVerbosity(100);
    t->SetIgnoreEnvVar(true);
    t->SetDBInfo("nova_dev","ifdbdev.fnal.gov","5433","bays");
    //  t->SetPasswordFile("/home/bays/dev/APDWatch/dbpwd.txt");

    int32_t chan;

    for(unsigned int m = 0; m < 4; ++m){
      for(unsigned int n = 0; n < 3; ++n){
        for(unsigned int o = 0; o < 64; ++o){
          for(unsigned int p = 0; p < 32; ++p){

	    chan = cmap->Map()->encodeDChan(3,m+1,n+1,o,p);

            nova::database::Row* r = t->NewRow();
            r->Set("channel",                    chan);
            r->Set("diblock",                    m+1);
            r->Set("dcm",                        n+1);
            r->Set("port",                       o);
            r->Set("pixel",                      p);

            t->AddRow(r);
            delete r;
            t->WriteToDB();
          }
        }
      }
    }
  }

}

//------------------------------------------------

///////////////////////////////////////////////////////////
namespace apdw
{
  DEFINE_ART_MODULE(APDWatch);
}
