/////////////////////////////////////////////////////////////////////////
/// \file    Cana_module.cc
/// \brief   Commisionning analysis module - looks for interesting events
/// \author  messier@indiana.edu
////////////////////////////////////////////////////////////////////////
#include <string>
#include <vector>

#include "TH1F.h"
#include "TH2F.h"
#include "TTree.h"

// ART includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"

// NOvA includes
#include "ChannelInfo/BadChanList.h"
#include "Calibrator/art/Calibrator.h"
#include "CMap/CMap.h"
#include "Commissioning/CanaNt.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RawData/RawDigit.h"
#include "RawData/RawTrigger.h"
#include "RecoBase/Prong.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCTruth.h"
#include "SummaryData/POTSum.h"
#include "SummaryData/SpillData.h"
#include "TrackFit/RLFit.h"
#include "Utilities/func/MathUtil.h"
#include "Utilities/RawUtil.h"


namespace comi {
  
  /// Analysis module to hunt for neutrino candidates
  class Cana : public art::EDProducer {
  public:
    explicit Cana(fhicl::ParameterSet const& pset);
    virtual ~Cana();
    
    void beginJob();
    void endSubRun(art::SubRun &sr);

    void produce(art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
    
  private:
    void WipeNt();
    void LoadMCInfo(art::Event& evt);

  private:
    TTree*       fNtTree;
    TTree*       fPotTree;
    CanaNt*      fNt;
    sumdata::POTSum* fPOTSum;
    TH1F*        fNhitTotal;
    TH1F*        fNSlices;
    TH1F*        fNhitInTime;
    TH1F*        fNhitOutTime;
    TH1F*        fNtimeWin;
    TH1F*        fNhitTimeWin;
    TH1F*        fHitTimesAll;
    TH1F*        fHitTimesGood;
    TH2F*        fNhitTotalVsEvent;
    TH2F*        fNtimeWinVsEvent;

  private:
    int          fApplyBadChan;   ///< Apply bad channel masks?
    std::string  fRawDataLabel;   ///< Module that produced the raw data
    std::string  fMCLabel;        ///< Module that produced the mc truth
    std::string  fSpillLabel;     ///< Module that produced the spill info
    std::string  fPotLabel;       ///< Module that produced the POTSum object
    float        fTimeWindow;     ///< Size of sliding time window	       
    unsigned int fNhit;       	  ///< Total number of intime hits req'd on track
    unsigned int fNhitX;      	  ///< Required number of hits in window on track
    unsigned int fNhitY;      	  ///< Required number of hits in window on track
  };

  // NOvA uses a 64 Mhz clock
  static const double kUSEC_PER_TDC = (1.0E6/64.0E6);
  static const int    kTDC_PER_USEC = 64;
  
  //......................................................................
  
  Cana::Cana(fhicl::ParameterSet const& p)
  {
    this->reconfigure(p);
    this->produces< std::vector<CanaNt> >();
    this->produces< std::vector<rb::Prong> >();
  }
  
  //......................................................................
  
  void Cana::beginJob() 
  {
    art::ServiceHandle<art::TFileService> f;
    
    // Some event-level histograms
    fNhitTotal = f->make<TH1F>("fNhitTotal",  
			       ";Nhit;Events",
			       250, 0.0, 2500.0);
    
    fNhitInTime = f->make<TH1F>("fNhitInTime", 
				";Nhit;Events",
				200, 0.0, 200.0);
    
    fNhitOutTime = f->make<TH1F>("fNhitOutTime",
				 ";Nhit;Events",
			       500,0.0,5000.0);
    
    fNtimeWin = f->make<TH1F>("fNtimeWin",
			      ";N time slices;Events",
			      10, 0.0, 10.0);
    
    fNhitTimeWin = f->make<TH1F>("fNhitTimeWin",
				 ";Nhit;Time slices",
				 200, 0.0, 200.0);
    
    fHitTimesAll = f->make<TH1F>("fHitTimesAll",
				 ";time (used);hits / 10 usec",
				 70, -100.0, 600.0);
    
    fHitTimesGood = f->make<TH1F>("fHitTimesGood",
				  ";time (used);hits / 10 usec",
				  70, -100.0, 600.0);
    
    fNhitTotalVsEvent = f->make<TH2F>("fNhitTotalVsEvent",
				      ";Event Number;Nhit",
				      300,0.0,30000.0,
				      250,0.0,2500.0);
    
    fNtimeWinVsEvent = f->make<TH2F>("fNtimeWinVsEvent",
				     ";Event Number; Nwin",
				     300,0.0,30000.0,
				     50,0.0,50.0);
    fNt = new CanaNt;
    fNtTree = f->make<TTree>("cana","Cana Analysis Tree");
    fNtTree->Branch("fCanaNt", "comi::CanaNt", &fNt, 1600, 99);
    
    fPOTSum = new sumdata::POTSum();
    fPotTree = f->make<TTree>("pottree","Total POT Tree");
    fPotTree->Branch("fPOTSum","sumdata::POTSum", &fPOTSum,1600,99);
  }
  
  //......................................................................
  
  Cana::~Cana() { }
  
  //......................................................................
  
  void Cana::endSubRun(art::SubRun &sr)
  {
    art::Handle< sumdata::POTSum >p;
    try {
      sr.getByLabel(fPotLabel,p);
      fPOTSum->totpot     = p->totpot;
      fPOTSum->totgoodpot = p->totgoodpot;
      fPOTSum->totspills  = p->totspills;
      fPOTSum->goodspills = p->goodspills;
      fPotTree->Fill();
    }
    catch (...) {
    }
  }
  
  //......................................................................
  
  void Cana::produce(art::Event& evt)
  {
    unsigned register int i;
    std::unique_ptr< std::vector<CanaNt> >    ntvec(new std::vector<CanaNt>);
    std::unique_ptr< std::vector<rb::Prong> > prong(new std::vector<rb::Prong>);

    art::ServiceHandle<calib::Calibrator> cal;

    this->WipeNt();
    
    //
    // Load event ID information into ntuple
    //
    fNt->run    = evt.run();
    fNt->subrun = evt.subRun();
    fNt->evt    = evt.id().event();
    
    //
    // ART time stamps are 32-bits of UNIX-seconds (high part) and
    // 32-bits of nanoseconds (low part)
    //
    fNt->utc.SetSec(   (evt.time().value()>>32)&0xFFFFFFFFUL );
    fNt->utc.SetNanoSec(evt.time().value()     &0xFFFFFFFFUL );
    
    //
    // If there is information from the simulation, load it
    //
    this->LoadMCInfo(evt);
    
    //
    // Get trigger information to extract trigger type
    //
    art::Handle< std::vector<rawdata::RawTrigger> > trigv;
    evt.getByLabel(fRawDataLabel, trigv);
    const rawdata::RawTrigger& trig = (*trigv)[0];
    
    fNt->trgsrc = trig.fTriggerMask_TriggerType;
    fNt->tstart = trig.fTriggerTimingMarker_TimeStart;
    
    //
    // Get spill information
    //
    art::Handle< sumdata::SpillData > spill;
    try { 
      evt.getByLabel(fSpillLabel, spill);
      fNt->pot = spill->spillpot;
      fNt->dt  = spill->deltaspilltimensec/1000.;
      fNt->gbeam=spill->goodbeam;
    }
    catch (...) { }
    
    //
    // Pull out the raw digits from the event
    //
    art::Handle< std::vector<rawdata::RawDigit> > digidummy;
    evt.getByLabel(fRawDataLabel, digidummy);
    
    //
    // Shuffle the data into a vector we can manipulate
    //
    std::vector< art::Ptr<rawdata::RawDigit> > digi;
    for (i=0; i<digidummy->size(); ++i) {
      digi.push_back(art::Ptr<rawdata::RawDigit>(digidummy, i));
    }
    if (digi.size()==0) {
      //
      // Make sure each event gets at least one entry in the ntuple even
      // if we failed to find anything interesting at all
      //
      if (ntvec->size()==0) ntvec->push_back(*fNt);
      evt.put(std::move(ntvec));
      evt.put(std::move(prong));
      fNtTree->Fill();
      return;
    }
    
    //
    // Fill the histogram showing the times of all hits
    //
    for (i=0; i<digi.size(); ++i) {
      bool goodtime;
      fHitTimesAll->Fill(cal->GetTNS(digi[i].get(), goodtime)/1000);
    }
    
    unsigned int tdcwin = (unsigned int)rint(fTimeWindow*kTDC_PER_USEC);
    double       t;
    
    //
    // Filter out the bad channels for this run
    //
    if (fApplyBadChan!=0) {
      art::ServiceHandle<chaninfo::BadChanList> badc;
      badc->Apply(digi);
    }
    
    if (digi.size()==0) {
      //
      // Make sure each event gets at least one entry in the ntuple even
      // if we failed to find anything interesting at all
      //
      if (ntvec->size()==0) ntvec->push_back(*fNt);
      evt.put(std::move(ntvec));
      evt.put(std::move(prong));
      fNtTree->Fill();
      return;
    }
    for (i=0; i<digi.size(); ++i) {
      bool goodtime;
      fHitTimesGood->Fill(cal->GetTNS(digi[i].get(), goodtime)/1000);
    }
    
    //
    // Time sort the hits and look for in-time activity
    //
    std::vector<util::RawSlice> win;
    util::TimeSort(digi);
    util::TimeSlice(digi,tdcwin,fNhit,fNhitX,fNhitY,win);
    
    // Log some information about the numbers of hits and the number of
    // windows
    fNtimeWin->Fill(win.size());
    
    unsigned int nhittot   = digi.size();
    unsigned int nhitwin   = 0;
    unsigned int nhitinwin = 0;
    fNhitTotal->Fill(nhittot);
    fNhitTotalVsEvent->Fill(fNt->evt, nhittot);
    fNtimeWinVsEvent->Fill (fNt->evt, win.size());
    for (i=0; i<win.size(); ++i) {
      nhitwin = (win[i].second+1)-win[i].first;
      fNhitTimeWin->Fill(nhitwin);
    nhitinwin += nhitwin;
    }
    fNhitInTime->Fill(nhitinwin);
    fNhitOutTime->Fill(nhittot-nhitinwin);
    
    //
    // Evaluate the activity inside the time window
    //
    for (i=0; i<win.size(); ++i) {
      double sumt  = 0.0;
      double sumt2 = 0.0;
      double sumw  = 0.0;
      double wcut  = 0.01;
      fNt->iwin    = i;
      fNt->nhit    = 0;
      fNt->nhitx   = 0;
      fNt->nhity   = 0;
      fNt->nunused = 0;
      fNt->t1      = 999999.;
      fNt->t2      =-999999.;
      fNt->cell1x  = 999999;
      fNt->cell1y  = 999999;
      fNt->plane1x = 999999;
      fNt->plane1y = 999999;
      fNt->cell2x  =-999999;
      fNt->cell2y  =-999999;
      fNt->plane2x =-999999;
      fNt->plane2y =-999999;
      
      //
      // Skip the line fitting if we don't have enough hits in each view
      //
      unsigned int nx, ny;
      util::CountXY(digi, win[i].first, win[i].second, &nx, &ny);
      if (nx<2 || ny<2) continue;
      
      //
      // Make a straight line fit to the hits in the window
      //
      trk::RLFit line(digi, win[i].first, win[i].second);
      line.Fit();
      
      //
      // Compute some information about the track fit
      //
      art::ServiceHandle<cmap::CMap> cmap;
      for (unsigned int j=0; j<line.fXhit.size(); ++j) {
	//
	// Check weight to see if hit contributed to the track
	//
	if (line.fWx[j]>wcut) {
	  ++fNt->nhit;
	  ++fNt->nhitx;
	  
	  int p = cmap->GetPlane(line.fXhit[j]);
	  int c = cmap->GetCell (line.fXhit[j]);
	  if (p<fNt->plane1x) fNt->plane1x = p;
	  if (p>fNt->plane2x) fNt->plane2x = p;
	  if (c<fNt->cell1x)  fNt->cell1x  = c;
	  if (c>fNt->cell2x)  fNt->cell2x  = c;
	  
	  //
	  // Record the hit time in units of usec
	  //
	  bool goodtime;
	  t = cal->GetTNS(line.fXhit[j], goodtime)/1000;

	  if (t<fNt->t1) fNt->t1 = t;
	  if (t>fNt->t2) fNt->t2 = t;
	  
	  sumt  += line.fWx[j]*t;
	  sumt2 += line.fWx[j]*t*t;
	  sumw  += line.fWx[j];
	}
	else {
	  ++fNt->nunused;
	}
      }
      for (unsigned int j=0; j<line.fYhit.size(); ++j) {
	//
	// Check weight to see if hit contributed to the track
	//
	if (line.fWy[j]>wcut) {
	  ++fNt->nhit;
	  ++fNt->nhity;
	  
	  int p = cmap->GetPlane(line.fYhit[j]);
	  int c = cmap->GetCell (line.fYhit[j]);
	  if (p<fNt->plane1y) fNt->plane1y = p;
	  if (p>fNt->plane2y) fNt->plane2y = p;
	  if (c<fNt->cell1y)  fNt->cell1y  = c;
	  if (c>fNt->cell2y)  fNt->cell2y  = c;
	  
	  //
	  // Record the hit time in units of usec
	  //
	  bool goodtime;
	  t = cal->GetTNS(line.fYhit[j], goodtime)/1000;

	  if (t<fNt->t1) fNt->t1 = t;
	  if (t>fNt->t2) fNt->t2 = t;
	  
	  sumt  += line.fWy[j]*t;
	  sumt2 += line.fWy[j]*t*t;
	  sumw  += line.fWy[j];
	}
	else {
	  ++fNt->nunused;
	}
      }
      // Leave at -999999 if there is somehow no weight
      if(sumw){
	fNt->tave = sumt / sumw;
	fNt->trms = sqrt(sumt2-sumt*sumt/sumw)/(sumw-1.0);
      }
      
      fNt->cosdet = line.CosThetaBeam();
      fNt->phidet = (180.0/M_PI)*line.PhiBeam();
      
      fNt->coscosmic = line.CosThetaCosmic();
      fNt->phicosmic = (180.0/M_PI)*line.PhiCosmic();
      
      art::ServiceHandle<geo::Geometry> geom;
      //
      // Approximate directions of the BNB beam in the NDOS
      // frame. NuMI direction from GeometryBase::NuMIBeamDirection(). BNB direction from
      // Zelimir Djurcic <zdjurcic@hep.anl.gov> 2/26/2011, private
      // communication.
      //
      static const double bnbdir[]  = { -0.367, 0.012, 0.930};
      static const double numidir[] = {geom->NuMIBeamDirection().X(), geom->NuMIBeamDirection().Y(), geom->NuMIBeamDirection().Z()};
      
      //
      // Both the Booster and NuMI beamlines send neutrinos generally
      // north so analyze the track with a north-going assumption
      //
      line.AssumeNorthGoing();
      
      //
      // Compute angles wrt the NuMI and Booster beams
      //
      static const double magbnb = util::pythag(bnbdir[0],bnbdir[1],bnbdir[2]);
      static const double magnumi= util::pythag(numidir[0],numidir[1],numidir[2]);

      double magtrk = util::pythag((line.fX2-line.fX1), 
			          (line.fY2-line.fY1),
			          (line.fZ2-line.fZ1));
      
      fNt->cosbnb = 
	((line.fX2-line.fX1)*bnbdir[0] +
	 (line.fY2-line.fY1)*bnbdir[1] +
	 (line.fZ2-line.fZ1)*bnbdir[2]) / magbnb / magtrk;
      
      fNt->cosnumi = 
	((line.fX2-line.fX1)*numidir[0] +
	 (line.fY2-line.fY1)*numidir[1] +
	 (line.fZ2-line.fZ1)*numidir[2]) / magnumi / magtrk;
      
      //
      // Record projected distances to detector edges
      //
      bool indetector = true;
      if (line.fX1 < -geom->DetHalfWidth())  indetector = false;
      if (line.fX1 > +geom->DetHalfWidth())  indetector = false;
      if (line.fY1 < -geom->DetHalfHeight()) indetector = false;
      if (line.fY1 > +geom->DetHalfHeight()) indetector = false;
      if (line.fZ1 < 0)                      indetector = false;
      if (line.fZ1 > geom->DetLength())      indetector = false;
      if (indetector) {
	double x[3] = {line.fX1, line.fY1, line.fZ1};
	double dxf[3] = {line.fX2-line.fX1,
			 line.fY2-line.fY1,
			 line.fZ2-line.fZ1};
	double dxb[3] = {-dxf[0], -dxf[1], -dxf[2]};
	fNt->dedgef = geom->DEdge(x, dxf);
	fNt->dedgeb = geom->DEdge(x, dxb);
      }
      
      //
      // Now analyze the track assuming its a down-going cosmic ray
      //
      line.AssumeDownGoing();
      fNt->x1 = line.fX1;
      fNt->y1 = line.fY1;
      fNt->z1 = line.fZ1;
      fNt->x2 = line.fX2;
      fNt->y2 = line.fY2;
      fNt->z2 = line.fZ2;
      
      fNtTree->Fill();
      ntvec->push_back(*fNt);
      
      //
      // Store a prong to show the cana fit results
      // TODO CJB - prong doesn't store the endpoint which Cana knows
      // but then Cana doesn't store the hits anyway, so we're equally
      // at fault.
      //
      TVector3 xyz1(fNt->x1, fNt->y1, fNt->z1);
      TVector3 xyz2(fNt->x2, fNt->y2, fNt->z2);
      rb::Prong p(art::PtrVector<rb::CellHit>(), xyz1, xyz2-xyz1);
      prong->push_back(p);
    } // end loop i=0...win.size()
    
    //
    // Make sure that every event gets at least one entry in the ntuple
    //
    if (win.size()==0) fNtTree->Fill();
    
    //
    // Store the ntuple and the prongs into the event
    //
    evt.put(std::move(ntvec));
    evt.put(std::move(prong));
  }
  
  //......................................................................
  
  void Cana::reconfigure(const fhicl::ParameterSet& p)
  {
    fApplyBadChan   = p.get<int>("ApplyBadChan");
    fTimeWindow     = p.get<float>("TimeWindow"); 		     
    fNhit           = p.get<int>("Nhit");	    		     
    fNhitX          = p.get<int>("NhitX");	    		     
    fNhitY          = p.get<int>("NhitY");        		     
    fRawDataLabel   = p.get<std::string>("RawDataLabel", "daq"); 
    fPotLabel       = p.get<std::string>("PotLabel");	     
    fMCLabel        = p.get<std::string>("MCLabel", "generator");
    fSpillLabel     = p.get<std::string>("SpillLabel");
  }
  
  //......................................................................
  
  ///
  /// Fill the ntuple with clearly invalid data
  ///
  void Cana::WipeNt() 
  {
    fNt->run       = -99;
    fNt->subrun    = -99;
    fNt->evt       = -99;
    fNt->iwin      = -99;
    fNt->trgsrc    = -99;
    fNt->nhit      = -99;
    fNt->nhitx     = -99;
    fNt->nhity     = -99;
    fNt->plane1x   =  999999;
    fNt->plane2x   = -999999;
    fNt->plane1y   =  999999;
    fNt->plane2y   = -999999;
    fNt->cell1x    =  999999;
    fNt->cell2x    = -999999;
    fNt->cell1y    =  999999;
    fNt->cell2y    = -999999;
    fNt->t1        =  999999;
    fNt->t2        = -999999;
    fNt->tave      = -999999;
    fNt->trms      = -999999;
    fNt->cosnumi   = -999999;
    fNt->cosbnb    = -999999;
    fNt->cosdet    = -999999;
    fNt->phidet    = -999999;
    fNt->coscosmic = -999999;
    fNt->phicosmic = -999999;
    fNt->pot       = -999999;
    fNt->dt        = -999999;
    fNt->gbeam     = -999999;
    fNt->x1        = -999999;
    fNt->y1        = -999999;
    fNt->z1        = -999999;
    fNt->x2        = -999999;
    fNt->y2        = -999999;
    fNt->z2        = -999999;
    fNt->dedgef    = -999999;
    fNt->dedgeb    = -999999;
    fNt->nunused   = -99;
    fNt->nux       = -999999;
    fNt->nuy       = -999999;
    fNt->nuz       = -999999;
    fNt->nupx      = -999999;
    fNt->nupy      = -999999;
    fNt->nupz      = -999999;
    fNt->leppx     = -999999;
    fNt->leppy     = -999999;
    fNt->leppz     = -999999;
    fNt->nupdg     = -99;
    fNt->iscc      = -99;
    fNt->imode     = -99;
    fNt->tstart    = 0;
    fNt->utc.Set(0,0,0,0,0,0,0,1,0);
  }
  
  //......................................................................
  
  void Cana::LoadMCInfo(art::Event& evt)
  {
    //
    // Pull the objects holding MC truth out of the event
    //
    art::Handle< std::vector<simb::MCTruth> > mc;
    try {
      evt.getByLabel(fMCLabel, mc);
      //
      // The code only knows how to deal with 1 and exactly 1 MC truth
      // particle list
      //
      if (mc->size()!=1) return;
    }
    catch (...) { 
      //
      // Read failed. Skip setting MC information.
      //
      return; 
    }
    
    //
    // Check if the neutrino kinematics are set. If yes, use them
    //
    if ((*mc)[0].NeutrinoSet()) {
      const simb::MCNeutrino& gen = (*mc)[0].GetNeutrino();
      const simb::MCParticle& nu  = gen.Nu();
      fNt->nux   = nu.Vx();
      fNt->nuy   = nu.Vy();
      fNt->nuz   = nu.Vz();
      fNt->nupx  = nu.Px();
      fNt->nupy  = nu.Py();
      fNt->nupz  = nu.Pz();
      fNt->nupdg = nu.PdgCode();
      fNt->iscc  = (gen.CCNC()==simb::kCC);
      fNt->imode = gen.Mode();
      fNt->leppx = gen.Lepton().Px();
      fNt->leppy = gen.Lepton().Py();
      fNt->leppz = gen.Lepton().Pz();
      return;
    }
    
    //
    // Other cases fall through to here. Assume the first particle in
    // the list is the one we're interested in...
    //
    const simb::MCParticle& p  = (*mc)[0].GetParticle(0);
    fNt->nux   = p.Vx();
    fNt->nuy   = p.Vy();
    fNt->nuz   = p.Vz();
    fNt->nupx  = -999999;
    fNt->nupy  = -999999;
    fNt->nupz  = -999999;
    fNt->nupdg = p.PdgCode();
    fNt->iscc  = -1;
    fNt->imode = -1;
    fNt->leppx = p.Px();
    fNt->leppy = p.Py();
    fNt->leppz = p.Pz();
  }
} // end namespace comi  
////////////////////////////////////////////////////////////////////////
DEFINE_ART_MODULE(comi::Cana);
////////////////////////////////////////////////////////////////////////
