/////////////////////////////////////////////////////////////////////
/// \brief   Module to analyze low energy events and noise
/// \author  messier@indiana.edu
/// \date    May 2011
////////////////////////////////////////////////////////////////////
#include <vector>

// ROOT includes
#include "TH1D.h"
#include "TH2D.h"
#include "TTree.h"

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"

// NOvA includes
#include "ChannelInfo/BadChanList.h"
#include "CMap/CMap.h"
#include "Commissioning/LeanaNt.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"
#include "RawData/RawDigit.h"
#include "RawData/RawTrigger.h"
#include "TrackFit/RLFit.h"
#include "Utilities/func/MathUtil.h"
#include "Utilities/RawUtil.h"


namespace comi {
  class LeanaNt;
  
  /// Summary information about previous muons
  struct MuonCand {
    unsigned char fId;    ///< Stop? Through-going?
    float         fX1;    ///< Start x position (cm)
    float         fY1;    ///< Start y position (cm)
    float         fZ1;    ///< Start z position (cm)
    float         fX2;    ///< End x position (cm)
    float         fY2;    ///< End y position (cm)
    float         fZ2;    ///< End z position (cm)
    double        fT;     ///< Time in spill (usec)
    double        fNOvAT; ///< "Global" NOvA time (usec)
  };
  
  /// Analysis module for low energy events and noise
  class Leana : public art::EDAnalyzer {
  public:
    explicit Leana(fhicl::ParameterSet const& pset);
    virtual ~Leana();
    
    void beginJob();
    void analyze(art::Event const& evt);
    void reconfigure(const fhicl::ParameterSet& p);

  private:
    void Ana1Hit(unsigned int iwin);
    void AnaNHit(unsigned int iwin);
    void MuonCandidate(unsigned int iwin,
		       unsigned int sz,
		       unsigned int nx,
		       unsigned int ny);
    void FindClosestMuon(unsigned int i,
			 double x,
			 double y,
			 double z,
			 double t,
			 double* dstop,
			 double* tstop,
			 unsigned int* istop,
			 double* dperp,
			 double* tperp,
			 unsigned int* iperp,
			 bool assn=true);
    void DumpMuon(unsigned int i);
    void DumpMuonTable();
    void DumpMichel(unsigned int i);
    bool IsPointOnEdge(double x, double y, double z);

  private:
    TH1D* fEventCount;     ///< Event counter
    TH1D* fSliceSz;        ///< Histogram of Nhit per slice
    TH2D* fSliceSzVsEvent; ///< Histogram of slice size vs. event

    TH1D* f1HitADC;      ///< ADC spectrum for 1 hit slices
    TH1D* f1HitTDC;      ///< TDC spectrum for 1 hit slices
    TH2D* f1HitADCvsTDC; ///< ADC vs TDC for 1 hit slices

    TH1D* f2HitDplane; ///< Distance from first to last plane
    TH1D* f3HitDplane; ///< Distance from first to last plane

    TH1D* f3HitSamePlaneDcell; ///< Max-min cell in same plane
    TH1D* f3HitDiffPlaneDcell; ///< Max-min cell in different planes
    
    LeanaNt* fNt;     ///< Ntuple for hit clusters
    TTree*   fNtTree; ///< Ntuple for hit clusters
    
  private:
    //
    // Buffer size should be the calibration trigger rate * the size
    // of fDeltaTnamu. For 20 seconds at 10 Hz, this suggests a
    // minimum 2000
    //
    static const unsigned int MUON_BUFF_SZ = 4096;
    static const char kNotMu  = 0; ///< Not a good muon candidate
    static const char kStopMu = 1; ///< A stopping muon candidate
    static const char kThruMu = 2; ///< A through-going muon candidate

  private:
    std::vector< art::Ptr<rawdata::RawDigit> > fDigi;

    std::vector<util::RawSlice> fWin; ///< The raw slices for this event
    double                      fT0;  ///< Event NOvA trigger time (usec)

    bool fDoFill; ///< Should we be filling ntuple / histogram?
    unsigned int fMuIndx1;   ///< Current muon list
    unsigned int fMuIndx2;   ///< Old non-associated muons
    std::vector<MuonCand> fMuon[MUON_BUFF_SZ]; ///< Table of muon data

  private:    
    int fDumpMichel; ///< Dump table of Michel electrons?

  private:
    double       fTimeWindow; ///< [usec] Time window size
    unsigned int fNhit;       ///< Number of hits req'd in window
    unsigned int fNhitX;      ///< Number of hits req'd in x view
    unsigned int fNhitY;      ///< Number of hits req'd in y view
    
    int fDcellCut;            ///< Allowed gap between hit cells
    int fDplaneCut;           ///< Allowed gap between hit planes

    unsigned int fMuonNhit;   ///< Required # hits on muon track
    unsigned int fMuonNhitX;  ///< # required in x view
    unsigned int fMuonNhitY;  ///< # required in y view
    double       fEdgeBuffer; ///< [cm] Edge distance for stopping/entering decision
    double       fDeltaTnamu; ///< Time for non-association (s)
  };

  // NOvA uses a 64 Mhz clock
  static const double kUSEC_PER_TDC = (1.0E6/64.0E6);
  static const int    kTDC_PER_USEC = 64;
  
  //......................................................................
  
  Leana::Leana(fhicl::ParameterSet const& p) :
    EDAnalyzer(p),
    fDoFill(true),
    fMuIndx1(0),
    fMuIndx2(0)
  {
    this->reconfigure(p);
  }
  
  //......................................................................
  
  void Leana::reconfigure(const fhicl::ParameterSet& p) 
  {
    fTimeWindow = p.get<double>("TimeWindow");
    fNhit       = p.get<unsigned int>("Nhit");
    fNhitX      = p.get<unsigned int>("NhitX");
    fNhitY      = p.get<unsigned int>("NhitY");
    fDcellCut   = p.get<unsigned int>("DcellCut");
    fDplaneCut  = p.get<unsigned int>("DplaneCut");
    fMuonNhit   = p.get<unsigned int>("MuonNhit");
    fMuonNhitX  = p.get<unsigned int>("MuonNhitX");
    fMuonNhitY  = p.get<unsigned int>("MuonNhitY");
    fEdgeBuffer = p.get<double>("EdgeBuffer");
    fDeltaTnamu = p.get<double>("DeltaTnamu");
    fDumpMichel = p.get<int>("DumpMichel");
  }
  
  //......................................................................
  
  void Leana::beginJob()
  {
    //
    // Some macros to make booking histograms a little easier
    //
#define H1(NAME,LABEL,N,X1,X2)			\
    NAME = f->make<TH1D>(#NAME,LABEL,N,X1,X2)
    
#define H2(NAME,LABEL,NX,X1,X2,NY,Y1,Y2)		\
    NAME = f->make<TH2D>(#NAME,LABEL,NX,X1,X2,NY,Y1,Y2)
    
    art::ServiceHandle<art::TFileService> f;
    //
    // General event-level histograms
    //
    H1(fEventCount,    ";;Events Seen",       1,   -0.5,0.5);
    H1(fSliceSz,       ";N_{hit};N_{slices}", 101, -0.5,100.5);
    H2(fSliceSzVsEvent,";Event;N_{hit}",      100, 0.0,1000000,101,-0.5,100.5);
    
    //
    // 1-hit slice histograms
    //
    H1(f1HitADC,";ADC;N_{hit}",4096,-0.5,4095.5);
    H1(f1HitTDC,";t [#musec]; N_{hit}",700,-100,600);
    H2(f1HitADCvsTDC,";t [#musec]; ADC",
       700,-100,600,4096,-0.5,4095.5);
    
    H1(f2HitDplane,";#Delta_{plane}",51,-0.5,50.5);
    H1(f3HitDplane,";#Delta_{plane}",51,-0.5,50.5);
    
    H1(f3HitSamePlaneDcell,";#Delta_{cell}",51,-0.5,50.5);
    H1(f3HitDiffPlaneDcell,";#Delta_{cell}",51,-0.5,50.5);
    
    fNt = new LeanaNt;
    fNtTree = f->make<TTree>("le","Leana Analysis Tree");
    fNtTree->Branch("fLeanaNt","comi::LeanaNt",&fNt,1600,99);
  }
  
  //......................................................................
  
  Leana::~Leana() { }
  
  //......................................................................
  
  void Leana::analyze(art::Event const& evt)
  {
    unsigned register int i;
    unsigned int fMaxSz = 10;
    
    fNt->run    = evt.run();
    fNt->subrun = evt.subRun();
    fNt->evt    = evt.id().event();
    
    //
    // Get trigger and t0 data from the raw trigger information
    //
    art::Handle< std::vector<rawdata::RawTrigger> > trigv;
    evt.getByLabel("daq", trigv);
    const rawdata::RawTrigger& trig = (*trigv)[0];
    fT0 = trig.TDCT0()*kUSEC_PER_TDC;
    
    //
    // Pull out the raw digits from the event
    //
    art::Handle< std::vector<rawdata::RawDigit> > digidummy;
    evt.getByLabel("daq", digidummy);
    
    //
    // Shuffle the data into a vector we can manipulate
    //
    fDigi.clear();
    for (i=0; i<digidummy->size(); ++i) {
      fDigi.push_back(art::Ptr<rawdata::RawDigit>(digidummy, i));
    }
    art::ServiceHandle<chaninfo::BadChanList> badc;
    badc->Apply(fDigi);
    if (fDigi.size()==0) return;
    
    if (fDoFill) fEventCount->Fill(0.0);
    
    //
    // Time sort and slice hits into windows
    //
    fWin.clear();
    util::TimeSort(fDigi);
    unsigned int tdcwin = rint(fTimeWindow*kTDC_PER_USEC);
    util::TimeSlice(fDigi,tdcwin,fNhit,fNhitX,fNhitY,fWin);
    if (fWin.size()<1) return;
    
    //
    // Clear the muon table
    //
    fMuon[fMuIndx1].resize(fWin.size());
    for (i=0; i<fWin.size(); ++i) {
      fMuon[fMuIndx1][i].fId    = kNotMu;
      fMuon[fMuIndx1][i].fNOvAT = fT0;
      fMuon[fMuIndx1][i].fT     = 0.0;
    }
    
    //
    // Analyze the activity in each window in turn
    //
    for (i=0; i<fWin.size(); ++i) {
      unsigned int sz = fWin[i].second-fWin[i].first+1;
      unsigned int nx = 0;
      unsigned int ny = 0;
      util::CountXY(fDigi, fWin[i].first, fWin[i].second, &nx, &ny);
      
      fNt->iwin = i;
      fNt->nx   = nx;
      fNt->ny   = ny;
      
      if (fDoFill) fSliceSz->Fill(sz);
      if (fDoFill) fSliceSzVsEvent->Fill(evt.id().event(), sz);
      
      // Analyze the small events
      if (sz==1)              this->Ana1Hit(i);
      if (sz>1 && sz<=fMaxSz) this->AnaNHit(i);
      
      // Check if this activity is a muon
      this->MuonCandidate(i, sz, nx, ny);
    }
    
    // this->DumpMuonTable();
    
    //
    // Compute the time between the muons at fMuIndx1 and fMuIndx2 in
    // units of seconds
    //
    double tnassn = fabs(fMuon[fMuIndx2][0].fNOvAT - 
			 fMuon[fMuIndx1][0].fNOvAT)/1.E6;
    //
    // Advance the muon indicies. The "current" entry (fMuIndx1) always
    // advances. The secondary index (fMuIndx2) only advances if enough
    // time have elapsed since the muon at the current index
    //
    fMuIndx1 = (fMuIndx1+1)%MUON_BUFF_SZ;
    if (tnassn>fDeltaTnamu) {
      fMuIndx2 = (fMuIndx2+1)%MUON_BUFF_SZ;
      //
      // Now that we have the data we need to compute backgrounds, we
      // can start filling histograms.
      //
      fDoFill = true;
    }
  }
  
  //......................................................................
  
  void Leana::DumpMuon(unsigned int i) 
  {
    std::cout << "[" << i << "] ";
    
    if (i>= fMuon[fMuIndx1].size()) {
      std::cout << " not valid muon" << std::endl;
      return;
    }
    
    if (fMuon[fMuIndx1][i].fId == kNotMu)  std::cout << "  not-mu ";
    if (fMuon[fMuIndx1][i].fId == kStopMu) std::cout << " stop-mu ";
    if (fMuon[fMuIndx1][i].fId == kThruMu) std::cout << " thru-mu ";
    
    std::cout << " (x1,y1,z1)=(" 
	      << fMuon[fMuIndx1][i].fX1 << ","
	      << fMuon[fMuIndx1][i].fY1 << ","
	      << fMuon[fMuIndx1][i].fZ1 << ") ";
    
    std::cout << " (x2,y2,z2)=("
	      << fMuon[fMuIndx1][i].fX2 << ","
	      << fMuon[fMuIndx1][i].fY2 << ","
	      << fMuon[fMuIndx1][i].fZ2 << ") ";
    
    std::cout << " L=" <<
      util::pythag((fMuon[fMuIndx1][i].fX1-fMuon[fMuIndx1][i].fX2),
                   (fMuon[fMuIndx1][i].fY1-fMuon[fMuIndx1][i].fY2),
		   (fMuon[fMuIndx1][i].fZ1-fMuon[fMuIndx1][i].fZ2));
    
    std::cout << " t=" << fMuon[fMuIndx1][i].fT << std::endl;
  }
  
  //......................................................................
  
  void Leana::DumpMuonTable() 
  {
    std::cout << "*= indx1=" << fMuIndx1
	      << " has " << fMuon[fMuIndx1].size() << " entries\n";
    
    for (unsigned int i=0; i<fMuon[fMuIndx1].size(); ++i) this->DumpMuon(i);
    
    std::cout << "   indx2=" << fMuIndx2
	      << " has " << fMuon[fMuIndx2].size() << " entries\n";
    std::cout << "   " << fMuon[fMuIndx2][0].fNOvAT << "\t"
	      << fMuon[fMuIndx2][0].fT << std::endl;
    
  }
  
  //......................................................................
  
  void Leana::DumpMichel(unsigned int imu) 
  {
    std::cout << "||"
	      << "==================================="
	      << "==================================="
	      << std::endl;
    std::cout << fNt->run << "/" 
	      << fNt->subrun << "/"
	      << fNt->evt 
	      << " Michel Candidate" 
	      << std::endl;
    
    std::cout << "(nx,ny)=(" 
	      << fNt->nx << "," 
	      << fNt->ny << ") "
	      << "(x,y,z,t)=" 
	      << fNt->x << ","
	      << fNt->y << ","
	      << fNt->z << ","
	      << fNt->t << ") "
	      << " (dstop,tstop)=("
	      << fNt->dstop << ","
	      << fNt->tstop << ")"
	      << std::endl;
    
    this->DumpMuon(imu);
  }
  
  //......................................................................
  
  void Leana::Ana1Hit(unsigned int iwin)
  {
    unsigned int i = fWin[iwin].first;
    
    double t = fDigi[i]->TDC()*kUSEC_PER_TDC;
    double q = fDigi[i]->ADC();
    
    if (fDoFill) f1HitADC->Fill(q);
    if (fDoFill) f1HitTDC->Fill(t);
    if (fDoFill) f1HitADCvsTDC->Fill(t,q);
  }
  
  //......................................................................
  
  void Leana::AnaNHit(unsigned int iwin)
  {
    unsigned register int i;
    
    art::ServiceHandle<cmap::CMap>    cmap;
    art::ServiceHandle<geo::Geometry> geom;
    
    unsigned int i1 = fWin[iwin].first;
    unsigned int i2 = fWin[iwin].second;
    unsigned int j1, j2;
    //
    // Look for the smallest gap between each hit and its neighbors in
    // terms of cell and plane numbers. Flag isolated hits as bad.
    //
    static const int kUNINIT = 999999;
    std::vector<bool> goodhit(i2+1);
    for (j1=i1; j1<=i2; ++j1) {
      int dcell  = kUNINIT; // distance to closest cell
      int dplane = kUNINIT; // distance to closest plane
      
      int p1 = cmap->GetPlane(fDigi[j1].get());
      int c1 = cmap->GetCell (fDigi[j1].get());
      int v1 = geom->Plane(p1)->View();
      
      for (j2=i1; j2<=i2; ++j2) {
	if (j1==j2) continue;
	
	int p2 = cmap->GetPlane(fDigi[j2].get());
	int c2 = cmap->GetCell (fDigi[j2].get());
	int v2 = geom->Plane(p2)->View();      
	
	if (v1==v2) {
	  if (abs(c2-c1)<dcell) dcell = abs(c2-c1);
	}
	if (abs(p2-p1)<dplane)  dplane = abs(p2-p1);
      }
      //
      // If dcell and dplane are still uninitialized, its because there
      // is only one hit in this view. Reset the values to 0.
      //
      if (dcell  == kUNINIT) dcell = 0;
      if (dplane == kUNINIT) dplane = 0;
      
      if (dcell<=fDcellCut && dplane<=fDplaneCut) goodhit[j1] = true;
      else                                        goodhit[j1] = false;
    }
    
    int nx = 0;
    int ny = 0;
    int c1x = 999999, c2x = -999999;
    int c1y = 999999, c2y = -999999;
    int p1x = 999999, p2x = -999999;
    int p1y = 999999, p2y = -999999;
    double x=0, y=0, z=0, t=0;
    double qx=0, qy=0;
    double xyz[3];
    for (i=i1; i<=i2; ++i) {
      if (goodhit[i]==false) continue;
      
      int p = cmap->GetPlane(fDigi[i].get());
      int c = cmap->GetCell (fDigi[i].get());
      
      geo::View_t v;
      geom->CellInfo(p, c, &v, xyz);
      
      if (v==geo::kX) {
	++nx; 
	x  += xyz[0];
	qx += fDigi[i]->ADC();
	if (c<c1x) c1x = c;
	if (p<p1x) p1x = p;
	if (c>c2x) c2x = c;
	if (p>p2x) p2x = p;
      }
      if (v==geo::kY) {
	++ny; 
	y  += xyz[1]; 
	qy += fDigi[i]->ADC();
	if (c<c1y) c1y = c;
	if (p<p1y) p1y = p;
	if (c>c2y) c2y = c;
	if (p>p2y) p2y = p;
      }
      z += xyz[2];
      t += fDigi[i]->TDC()*kUSEC_PER_TDC;
    }
    //
    // Require at least one good hit in each view
    //
    if (nx<1 || ny<1) return;
    
    fNt->iwin = iwin;
    fNt->nx   = nx;
    fNt->ny   = ny;
    fNt->x    = x = x/(float)nx;
    fNt->y    = y = y/(float)ny;
    fNt->z    = z = z/(float)(nx+ny);
    fNt->t    = t = t/(float)(nx+ny);
    fNt->qx   = qx;
    fNt->qy   = qy;
    
    fNt->dcellx  = c2x-c1x;
    fNt->dcelly  = c2y-c1y;
    fNt->dplanex = p2x-p1x;
    fNt->dplaney = p2y-p1y;
    
    fNt->compx = (float)nx/(float)((fNt->dcellx+1)*(fNt->dplanex+1));
    fNt->compy = (float)ny/(float)((fNt->dcelly+1)*(fNt->dplaney+1));
    
    //
    // Find the closest muon to this cluster
    //
    double dstop, tstop;       // Distance and time to stop muon end point
    double dperp, tperp;       // Distance and time to muon track
    unsigned int istop, iperp; // Indicies of matched muons
    this->FindClosestMuon(iwin, 
			  x, y, z, t, 
			  &dstop, &tstop, &istop,
			  &dperp, &tperp, &iperp);
    fNt->dstop = dstop;
    fNt->tstop = tstop;
    fNt->dperp = dperp;
    fNt->tperp = tperp;
    
    //
    // Find the closest muon to this cluster in an unassociated event
    //
    unsigned int istopna, iperpna;
    this->FindClosestMuon(iwin, 
			  x, y, z, t, 
			  &dstop, &tstop, &istopna,
			  &dperp, &tperp, &iperpna,
			  false);
    fNt->dstopna = dstop;
    fNt->tstopna = tstop;
    fNt->dperpna = dperp;
    fNt->tperpna = tperp;
    
    if (fDoFill) fNtTree->Fill();
    
    //
    // Check if this is a Michel candidate, if yes, print it to a log
    // file
    //
    if (fDumpMichel==1) {
      bool ismichel = true;
      if ( fNt->dplanex>=(fNt->nx+fNt->ny) ) ismichel = false;
      if ( fNt->dplaney>=(fNt->nx+fNt->ny) ) ismichel = false;
      if ( fNt->dcellx >=(fNt->nx+fNt->ny) ) ismichel = false;
      if ( fNt->dcelly >=(fNt->nx+fNt->ny) ) ismichel = false;
      if ( 4*fNt->nx <= (fNt->nx+fNt->ny) )  ismichel = false;
      if ( 4*fNt->ny <= (fNt->nx+fNt->ny) )  ismichel = false;
      if ( fNt->dstop > 30.0 )               ismichel = false;
      if ( fNt->tstop > 10.0 )               ismichel = false;
      if (ismichel) this->DumpMichel(istop);
    }
  }
  
  //......................................................................
  
  void Leana::MuonCandidate(unsigned int iwin, 
			    unsigned int sz,
			    unsigned int nx,
			    unsigned int ny) 
  {
    if (sz<fMuonNhit || nx<fMuonNhitX || ny<fMuonNhitY) {
      fMuon[fMuIndx1][iwin].fId = kNotMu;
      return;
    }
    
    art::ServiceHandle<geo::Geometry> geom;
    
    trk::RLFit trkfit(fDigi, fWin[iwin].first, fWin[iwin].second);
    trkfit.Fit();
    trkfit.AssumeDownGoing();
    
    //
    // Does this look like an entering track?
    //
    bool isentering = false;
    if (this->IsPointOnEdge(trkfit.fX1,trkfit.fY1,trkfit.fZ1)) {
      isentering = true;
    }
    
    if (isentering==false) {
      fMuon[fMuIndx1][iwin].fId = kNotMu;
      return;
    }
    
    fMuon[fMuIndx1][iwin].fX1 = trkfit.fX1;
    fMuon[fMuIndx1][iwin].fY1 = trkfit.fY1;
    fMuon[fMuIndx1][iwin].fZ1 = trkfit.fZ1;
    
    fMuon[fMuIndx1][iwin].fX2 = trkfit.fX2;
    fMuon[fMuIndx1][iwin].fY2 = trkfit.fY2;
    fMuon[fMuIndx1][iwin].fZ2 = trkfit.fZ2;
    
    bool isstopping = false;
    if (this->IsPointOnEdge(trkfit.fX2,trkfit.fY2,trkfit.fZ2)==false) {
      isstopping = true;
    }
    
    if (isstopping) fMuon[fMuIndx1][iwin].fId = kStopMu;
    else            fMuon[fMuIndx1][iwin].fId = kThruMu;
    
    unsigned register int i;
    double sumt = 0.0;
    double sumw = 0.0;
    for (i=0; i<trkfit.fXhit.size(); ++i) {
      sumw += trkfit.fWx[i];
      sumt += trkfit.fWx[i]*trkfit.fXhit[i]->TDC()*kUSEC_PER_TDC;
    }
    for (i=0; i<trkfit.fYhit.size(); ++i) {
      sumw += trkfit.fWy[i];
      sumt += trkfit.fWy[i]*trkfit.fYhit[i]->TDC()*kUSEC_PER_TDC;
    }
    
    fMuon[fMuIndx1][iwin].fNOvAT = sumt/sumw;
    fMuon[fMuIndx1][iwin].fT     = fMuon[fMuIndx1][iwin].fNOvAT;
  }
  
  //......................................................................
  
  void Leana::FindClosestMuon(unsigned int iwin,
			      double x, double y, double z, double t,
			      double* dend, double* tend, unsigned int* istop,
			      double* dtrk, double* ttrk, unsigned int* iperp,
			      bool assn)
  {
    //
    // Decide if we're looking for real associations (assn=true) or are
    // looking for random coinicidences
    //
    unsigned int imu = 0;
    if (assn==true) imu = fMuIndx1;
    else            imu = fMuIndx2;
    
    unsigned register int i;
    double dstop;
    double dperp;
    *dend = 9E99;
    *tend = 9E99;
    *dtrk = 9E99;
    *ttrk = 9E99;
    *istop = 999999;
    *iperp = 999999;
    for (i=0; i<fMuon[imu].size(); ++i) {
      if (fMuon[imu][i].fId==kNotMu) continue;
      if (fMuon[imu][i].fT>=t)       continue;
      
      if (fMuon[imu][i].fId==kStopMu) {
	//
	// Find distance to stop mu end point
	//
	dstop =
	  util::pythag((x-fMuon[imu][i].fX2), 
		       (y-fMuon[imu][i].fY2),
		       (z-fMuon[imu][i].fZ2));
	if (dstop<*dend) {
	  *dend = dstop;
	  *tend = t-fMuon[imu][i].fT;
	  *istop = i;
	}
      }
      //
      // Find perpendicular distance to muon track
      //
      double point[3] = {x,y,z};
      double icept[3] = { fMuon[imu][i].fX1,
			  fMuon[imu][i].fY1,
			  fMuon[imu][i].fZ1
      };
      double slope[3] = {
	fMuon[imu][i].fX2 - fMuon[imu][i].fX1,
	fMuon[imu][i].fY2 - fMuon[imu][i].fY1,
	fMuon[imu][i].fZ2 - fMuon[imu][i].fZ1,
      };
      double closest[3] = {-999999.,-999999.,-999999.};
      dperp = geo::ClosestApproach(point, icept, slope, closest);
      //
      // If the point happens outside the limits of the track, compute
      // the distance to the track end points
      //
      if (closest[1]<fMuon[imu][i].fY2) {
	dperp = util::pythag((x-fMuon[imu][i].fX2),
			     (y-fMuon[imu][i].fY2),
			     (z-fMuon[imu][i].fZ2));
      }
      if (closest[1]>fMuon[imu][i].fY1) {
	dperp = util::pythag((x-fMuon[imu][i].fX1),
			     (y-fMuon[imu][i].fY1),
			     (z-fMuon[imu][i].fZ1));
      }
      if (dperp<*dtrk) {
	*dtrk = dperp;
	*ttrk = t-fMuon[imu][i].fT;
	*iperp = i;
      }
    }
  }
  
  //......................................................................
  
  bool Leana::IsPointOnEdge(double x, double y, double z)
  {
    art::ServiceHandle<geo::Geometry> geom;
    
    if (y>geom->DetHalfHeight()-fEdgeBuffer)      return true;
    if (fabs(x)>geom->DetHalfWidth()-fEdgeBuffer) return true;
    if (z<fEdgeBuffer)                            return true;
    //
    //! \todo Hardwired to the NDOS case...
    //
    if (z>810.0-fEdgeBuffer)                      return true;
    // if (z>geom->DetLength()-fEdgeBuffer)       return true;
    return false;
  }
} // end namespace comi
////////////////////////////////////////////////////////////////////////
DEFINE_ART_MODULE(comi::Leana);
////////////////////////////////////////////////////////////////////////
