import sys
from datetime import date, time, datetime, timedelta
from nova_uptime import connect_to_database
from nova_uptime import runs_for_period
from nova_uptime import nactive_chans
from nova_uptime import spills_between
from nova_uptime import pot_between

HTML_FILE = None
#
# Table format is
# row[0] = date/time
# row[1] = run number
# row[2] = active detector mass rounded to nearest kt
# row[3] = 10 minute running average NuMI power in kW
#
DELTA_T_MIN  = 20.
DELTA_T      = timedelta(minutes=DELTA_T_MIN)
HALF_DELTA_T = timedelta(minutes=DELTA_T_MIN/2)

def print_header():
    HTML_FILE.write(\
        "<html>"
        "<head>"
        "<script type=\"text/javascript\" src=\"https://www.google.com/jsapi\"></script>"
        "<script type=\"text/javascript\">"
        "google.load(\"visualization\", \"1\", {packages:[\"corechart\"]});"
        "google.setOnLoadCallback(drawChart);"
        "function drawChart() {"
        "var data = google.visualization.arrayToDataTable(["
        )

def print_footer():
    HTML_FILE.write(\
        "]);"
        "var options = {"
        "explorer: { axis: 'horizontal', maxZoomIn: 0.02857, maxZoomOut: 1.1}"
        "};"
        "var chart = new google.visualization.LineChart(document.getElementById('chart_div'));"
        "chart.draw(data, options);"
        "}"
        "</script>"
        "</head>"
        "<body>"
        "  <div id=\"chart_div\" style=\"width: 900px; height: 500px;\"></div>"
        "</body>"
        "</html>"
        )

def print_table(table):
    for row in table:
        line = "[ new Date("+\
               str(row[0].year)+","+\
               str(row[0].month-1)+","+\
               str(row[0].day)+","+\
               str(row[0].hour)+","+\
               str(row[0].minute)+","+\
               str(row[0].second)+"),"+\
               str(row[1])+","+\
               str(row[2])+","+\
               str(row[3])+"],"
        HTML_FILE.write(line)
        
def fill_table(table,run,t1,t2):
    if ((t2-t1)>DELTA_T):
        tr = t1;
        while (1):
            tr += DELTA_T
            table.append([tr,run,0,0])
            if (tr>t2):
                break
            
def do_fd_runs(t0,runs):
    tgap    = timedelta(minutes=1)
    table   = []
    tlast   = t0
    runlast = 0
    for r in runs:
        t1  = r['tstart']
        t2  = r['tstop']
        run = r['run']
        #
        # Check if we have a gap between this run and the previous one
        #
        if ((t1-tlast)>tgap):
            #
            # Close out the last run
            #
            table.append([tlast, runlast, 0, 0]) # Complete last run
            table.append([tlast, 0,       0, 0]) # Start downtime
            fill_table(table, 0, tlast, t1)      # Fill report intervals
            table.append([t1, 0, 0, 0])          # End of downtime
        #
        # Make entry for start of new run
        #
        table.append([t1,run,0,0])
        fill_table(table, run, t1, t2)
        #
        # Prepare for next iteration
        #
        runlast = run
        tlast   = t2
    return table

def do_pot(table):
    spills     = None
    last_spill = datetime(1970,1,1)
    for row in table:
        t1 = row[0]-HALF_DELTA_T
        t2 = row[0]+HALF_DELTA_T
        #
        # To optimize fetches from the beams database, grab spills 1
        # day at a time
        #
        if ( (last_spill==None)|(t2>last_spill) ):
            last_spill = t1 + timedelta(days=1)
            if (last_spill<t2):
                last_spill=t2
            spills = spills_between(t1, last_spill)
        pot = pot_between(t1, t2, spills)
        #
        # Normalize to kW. 700 kW is 1.3E17 POT/60 minutes. Our
        # 'pot' number is in units of 1e18 pot. Then scale x100 so numbers
        # fit on the plots with run numbers which are in 10000's range
        #
        row[3] = pot*(700.0/0.13)*(60./DELTA_T_MIN)*100.

def do_kiloton(table):
    """ Add active kilotons to the table. Round to the nearest kt
    """
    nchanmx = 14*12*64*32
    for row in table:
        if (row[1]==0):
            row[2] = 0
        else:
            n = nactive_chans(row[1])
            m = 1000*round(14*n/nchanmx+0.5,0)
            row[2] = m
        
def main():
    global HTML_FILE
    
    connect_to_database()

    #
    # By default, process the current month-1 week through "now"
    #
    t1 = datetime.utcnow()
    t0 = t1 - timedelta(days=33)
    if (len(sys.argv)>1):
        #
        # Set the default behavior - take the date given (like
        # 2014/08/01) and report for 1 days prior and for the rest of
        # the month extending beyong by one day
        #
        t00 = datetime.strptime(sys.argv[1], "%Y/%m/%d")
        t0  = t00 - timedelta(days=1)
        t1  = t00 + timedelta(days=33)
        
    outfile_name = "uptime_graphic_%02d.%02d.html" % (t1.year, t1.month)
    HTML_FILE = open(outfile_name,'w')
    print_header()

    HTML_FILE.write("['Date', 'FD Run Number', "\
                    "'FD Mass (tons)', 'NuMI (kW x 100)'],")

    runs = runs_for_period(t0, t1)
    table = do_fd_runs(t0, runs);
    do_pot(table)
    do_kiloton(table)
    print_table(table)

    print_footer()
    HTML_FILE.close()
    
if __name__ == "__main__":
    main()
