/////////////////////////////////////////////////////////
/// \brief  Cosmic Rejection PIDs
/// \author Kirk Bays
////////////////////////////////////////////////////////
#include <string>

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "CosRej/CosRejFxs.h"
#include "CosRej/CosRejObj.h"
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "MCCheater/BackTracker.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "ReMId/ReMId.h"
#include "TimingFit/TimingFitAlg.h"
#include "Utilities/AssociationUtil.h"
#include "Utilities/func/EnvExpand.h"

#include "TFile.h"
#include "TMVA/Reader.h"

namespace cosrej {

  class CosRej : public art::EDProducer {
  public:
    explicit CosRej(fhicl::ParameterSet const & pset);
    virtual ~CosRej();

    void produce (art::Event & evt);
    void beginRun(art::Run& run);

  protected:

    cosrej::CosRejFxs* fCosRejFxs;
    std::string        fSlicerLabel;
    std::string        fTrackLabel;
    std::string        fPIDLabel;
    std::string        fTMVApath;
    bool               fUseLongestTrack;
    TMVA::Reader*      fReaderOld;
    TMVA::Reader*      fReaderPre;
    TMVA::Reader*      fReaderPost;
    TMVA::Reader*      fReaderUC;
    float              TMVAvars[23];
    float              TMVAvarsNew[16];
  };
}


namespace cosrej
{

  //----------------------------------------------------------------------
  CosRej::CosRej(fhicl::ParameterSet const & pset):
    fCosRejFxs        (new CosRejFxs()),
    fSlicerLabel      (pset.get< std::string >("SlicerLabel")),
    fTrackLabel       (pset.get< std::string >("TrackLabel")),
    fPIDLabel         (pset.get< std::string >("PIDLabel")),
    fTMVApath         (pset.get< std::string >("tmvapath")),
    fUseLongestTrack  (pset.get< bool        >("UseLongestTrack")),
    fReaderOld        (nullptr),
    fReaderPre        (nullptr),
    fReaderPost       (nullptr),
    fReaderUC         (nullptr)
  {
    produces< std::vector<cosrej::CosRejObj> >();
    produces< art::Assns<cosrej::CosRejObj, rb::Cluster> >();

  }


  //----------------------------------------------------------------------
  CosRej::~CosRej()
  {
      if(fCosRejFxs)  delete fCosRejFxs;
      if(fReaderOld)  delete fReaderOld;
      if(fReaderPre)  delete fReaderPre;
      if(fReaderPost) delete fReaderPost;
      if(fReaderUC)   delete fReaderUC;
  }

  //-------------------------------------------------------------------
  void CosRej::beginRun(art::Run& run)
  {

    art::ServiceHandle<geo::Geometry>  geom;

    if(geom->DetId() == novadaq::cnv::kFARDET) {
      fReaderOld = new TMVA::Reader;
      fReaderOld->AddVariable("anglekal",&TMVAvars[0]);
      fReaderOld->AddVariable("anglecos",&TMVAvars[1]);
      fReaderOld->AddVariable("cosdiry",&TMVAvars[2]);
      fReaderOld->AddVariable("kaldiry",&TMVAvars[3]);
      fReaderOld->AddVariable("maxy",&TMVAvars[4]);
      fReaderOld->AddVariable("cosratio",&TMVAvars[5]);
      fReaderOld->AddVariable("cosfwdcell",&TMVAvars[6]);
      fReaderOld->AddVariable("cosbakcell",&TMVAvars[7]);
      fReaderOld->AddVariable("kalfwdcell",&TMVAvars[8]);
      fReaderOld->AddVariable("kalbakcell",&TMVAvars[9]);
      fReaderOld->AddVariable("scatt/tracklen",&TMVAvars[10]);
      fReaderOld->AddVariable("hade",&TMVAvars[11]);
      fReaderOld->AddVariable("visE",&TMVAvars[12]);
      fReaderOld->AddSpectator("remid",&TMVAvars[14]);
      fReaderOld->AddSpectator("osc",&TMVAvars[16]);
      fReaderOld->AddSpectator("trueE",&TMVAvars[19]);
      fReaderOld->AddSpectator("type",&TMVAvars[21]);
      std::string pidpath = util::EnvExpansion(fTMVApath);
      fReaderOld->BookMVA("CC_BDTA",pidpath+"OLD_BDTA.weights.xml");

      fReaderPre = new TMVA::Reader;

      fReaderPre->AddVariable("anglekal",&TMVAvarsNew[0]);
      fReaderPre->AddVariable("kaldiry",&TMVAvarsNew[1]);
      fReaderPre->AddVariable("maxy",&TMVAvarsNew[2]);
      fReaderPre->AddVariable("kalhits",&TMVAvarsNew[4]);
      fReaderPre->AddVariable("kallen",&TMVAvarsNew[5]);
      fReaderPre->AddVariable("kalbakcell",&TMVAvarsNew[6]);
      fReaderPre->AddVariable("kalfwdcell",&TMVAvarsNew[7]);
      fReaderPre->AddVariable("scatt/kallen",&TMVAvarsNew[8]);
      fReaderPre->AddVariable("nhit",&TMVAvarsNew[9]);
      fReaderPre->AddVariable("calE/nhit",&TMVAvarsNew[10]);

      fReaderPre->AddSpectator("remid",&TMVAvarsNew[12]);
      fReaderPre->AddSpectator("cospid",&TMVAvarsNew[13]);
      fReaderPre->AddSpectator("osc",&TMVAvarsNew[14]);
      fReaderPre->AddSpectator("trueE",&TMVAvarsNew[15]);
      fReaderPre->BookMVA("PRE_BDTA",pidpath+"PRE_BDTA.weights.xml");

      fReaderPost = new TMVA::Reader;
      fReaderPost->AddVariable("anglekal",&TMVAvarsNew[0]);
      fReaderPost->AddVariable("kaldiry",&TMVAvarsNew[1]);
      fReaderPost->AddVariable("maxy",&TMVAvarsNew[2]);
      fReaderPost->AddVariable("miny",&TMVAvarsNew[3]);
      fReaderPost->AddVariable("kalhits",&TMVAvarsNew[4]);
      fReaderPost->AddVariable("kallen",&TMVAvarsNew[5]);
      fReaderPost->AddVariable("kalbakcell",&TMVAvarsNew[6]);
      fReaderPost->AddVariable("kalfwdcell",&TMVAvarsNew[7]);
      fReaderPost->AddVariable("scatt/kallen",&TMVAvarsNew[8]);
      fReaderPost->AddVariable("nhit",&TMVAvarsNew[9]);
      fReaderPost->AddVariable("calE/nhit",&TMVAvarsNew[10]);
      fReaderPost->AddVariable("nkal",&TMVAvarsNew[11]);

      fReaderPost->AddSpectator("remid",&TMVAvarsNew[12]);
      fReaderPost->AddSpectator("cospid",&TMVAvarsNew[13]);
      fReaderPost->AddSpectator("osc",&TMVAvarsNew[14]);
      fReaderPost->AddSpectator("trueE",&TMVAvarsNew[15]);
      fReaderPost->BookMVA("POST_BDTA",pidpath+"POST_BDTA.weights.xml");

      fReaderUC = new TMVA::Reader;
      fReaderUC->AddVariable("anglekal",&TMVAvars[0]);
      fReaderUC->AddVariable("anglecos",&TMVAvars[1]);
      fReaderUC->AddVariable("cosdiry",&TMVAvars[2]);
      fReaderUC->AddVariable("kaldiry",&TMVAvars[3]);
      fReaderUC->AddVariable("maxy",&TMVAvars[4]);
      fReaderUC->AddVariable("cosratio",&TMVAvars[5]);
      fReaderUC->AddVariable("cosfwdcell",&TMVAvars[6]);
      fReaderUC->AddVariable("cosbakcell",&TMVAvars[7]);
      fReaderUC->AddVariable("kalfwdcell",&TMVAvars[8]);
      fReaderUC->AddVariable("kalbakcell",&TMVAvars[9]);
      fReaderUC->AddVariable("scatt/tracklen",&TMVAvars[10]);
      fReaderUC->AddVariable("hade",&TMVAvars[11]);
      fReaderUC->AddVariable("visE",&TMVAvars[12]);
      fReaderUC->AddSpectator("slicetime",&TMVAvars[13]);
      fReaderUC->AddSpectator("remid",&TMVAvars[14]);
      fReaderUC->AddSpectator("cbdtg",&TMVAvars[15]);
      fReaderUC->AddSpectator("osc",&TMVAvars[16]);
      fReaderUC->AddSpectator("misse",&TMVAvars[17]);
      fReaderUC->AddSpectator("nhit",&TMVAvars[18]);
      fReaderUC->AddSpectator("trueE",&TMVAvars[19]);
      fReaderUC->AddSpectator("ntracks3d",&TMVAvars[20]);
      fReaderUC->AddSpectator("type",&TMVAvars[21]);
      fReaderUC->AddSpectator("vtx",&TMVAvars[22]);
      fReaderUC->BookMVA("UC_BDTA",pidpath+"UC_BDTA.weights.xml");
    }

  } //end beginRun

  //-------------------------------------------------------------------
  void CosRej::produce(art::Event& evt)
  {

    art::ServiceHandle<geo::Geometry>  geom;
    art::ServiceHandle<geo::LiveGeometry>  livegeom;
    art::ServiceHandle<cheat::BackTracker> bt;
    art::Handle<std::vector<rb::Cluster> > slicevec;
    evt.getByLabel(fSlicerLabel,slicevec);

    if(slicevec->empty()) {
      mf::LogWarning ("No Slices")<<"No Slices in the input file";
      return;
    }

    art::FindManyP<rb::Track> trackAssnList(slicevec, evt, fTrackLabel);
    art::FindManyP<rb::Track> costrackAssnList(slicevec, evt, "cosmictrack");
    std::unique_ptr< std::vector<cosrej::CosRejObj> > outputObjects(new std::vector<cosrej::CosRejObj>);
    std::unique_ptr< art::Assns<cosrej::CosRejObj, rb::Cluster> > assoc(new art::Assns<cosrej::CosRejObj, rb::Cluster>);

    tf::TimingFitAlg tfalg;

    for(size_t i = 0; i < slicevec->size(); ++i){

      art::Ptr<rb::Cluster> slice(slicevec,i);

      if(slice->IsNoise()) {
        continue;
      }

      cosrej::CosRejObj cosrejobj; // initialization in constructor

      unsigned int bestIdx = 999;  // for index in list of the highest ReMId PID track
      TVector3 minp, maxp; // minimum and maximum extent vectors for slice
      int nhadtracks = 0;
      double fwddist = 0, bakdist = 0,  scatt = 0;
      int fwdcell = 0, bakcell = 0;
      bool badtrack = false;    // check for messed up tracks; a few kalman tracks have nans for directions

      const std::vector< art::Ptr<rb::Track> > cosTracks = costrackAssnList.at(i);

      minp = slice->MinXYZ();
      maxp = slice->MaxXYZ();

      if(!costrackAssnList.isValid()) {
        mf::LogError("CosRej") << "CosRej: No Cosmic Tracks! Do not go any further!";
        outputObjects->push_back(cosrejobj);
        util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));
        return;
      }

      if(cosTracks.size()==1) { // for straight line fit version of Cosmic Track, should only be one track
        const art::Ptr<rb::Track> track = cosTracks[0];
        double angleOfBestTrack = fCosRejFxs->getAngle(track->Dir());
        if(angleOfBestTrack>-6&&angleOfBestTrack<2) cosrejobj.SetAngleCos(angleOfBestTrack);

        TVector3 start = track->Start();
        TVector3 end = track->Stop();
        TVector3 dirbak = -track->Dir().Unit();
        TVector3 dirfor = (track->TrajectoryPoint(track->NTrajectoryPoints()-1) - track->TrajectoryPoint(track->NTrajectoryPoints()-2)).Unit();

        cosrejobj.SetCosFwdDist(livegeom->ProjectedLiveDistanceToEdge(end,dirfor));
        cosrejobj.SetCosBakDist(livegeom->ProjectedLiveDistanceToEdge(start,dirbak));
        cosrejobj.SetCosFwdCell(livegeom->ProjectedLiveCellsToEdge(end,dirfor));
        cosrejobj.SetCosBakCell(livegeom->ProjectedLiveCellsToEdge(start,dirbak));

        if(geom->DetId() == novadaq::cnv::kNEARDET) {
          TVector3 dirfor = (track->TrajectoryPoint(track->NTrajectoryPoints()-1) - track->TrajectoryPoint(track->NTrajectoryPoints()-2)).Unit();

          cosrejobj.SetCosFwdSteel(livegeom->ProjectedSteelDist(track->Stop(), dirfor));
          cosrejobj.SetCosBakSteel(livegeom->ProjectedSteelDist(track->Start(), -(track->Dir())));
          cosrejobj.SetCosFwdAir(livegeom->ProjectedAirDist(track->Stop(), dirfor));
          cosrejobj.SetCosBakAir(livegeom->ProjectedAirDist(track->Start(), -(track->Dir())));
          cosrejobj.SetCosYPosAtTrans(livegeom->YPositionAtTransition(track->Start(), track->Dir()));
          cosrejobj.SetCosFwdCellND(livegeom->FullNDProjectedCells(track->Stop(), dirfor));
          cosrejobj.SetCosBakCellND(livegeom->FullNDProjectedCells(track->Start(), -(track->Dir())));
        }

        TMVAvars[2] = track->Dir().Y();
        TMVAvars[5] = (1.0*track->NCell())/(1.0*slice->NCell());
        tf::TimingFitResult tfres = tfalg.HoughFit(track.get());
        cosrejobj.SetCFitSpeed(tfres.bestInvSpeed);
        cosrejobj.SetCDirScore(tfres.fwdScore-tfres.revScore);
        cosrejobj.SetCScoreDiff(tfres.bestScore-std::max(tfres.fwdScore,tfres.revScore));  // Chris's timing fit (Hough)
        cosrejobj.SetCosHitRatio(1.0*track->NCell()/(1.0*slice->NCell()));

        if(bt->HaveTruthInfo()) {
          art::PtrVector<rb::CellHit> trkhits = track->AllCells();
          const std::vector<const sim::Particle*> parts = bt->HitsToParticle(trkhits);

          if(!parts.empty()){
            const TVector3 trueDir = parts[0]->Momentum().Vect().Unit();
            cosrejobj.SetCosThetaTrue(trueDir.Dot(track->Dir()));
          }
        } // end cosmic truth
      } // end cosmics

      if(!trackAssnList.isValid()) {
        mf::LogWarning("CosRej") << "CosRej: No Kalman Tracks!";
        outputObjects->push_back(cosrejobj);
        util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));
        continue;
      }

      const std::vector< art::Ptr<rb::Track> > sliceTracks = trackAssnList.at(i);
      art::FindOneP<remid::ReMId> remidVec(sliceTracks, evt, fPIDLabel);

      if(fUseLongestTrack) {
        bestIdx = 0;
        cosrejobj.SetNTracks3D(sliceTracks.size());
      }

      else if(!remidVec.isValid()) {
        mf::LogWarning("CosRej") << "CosRej: No ReMId information!";
        outputObjects->push_back(cosrejobj);
        util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));
        continue;
      }

      else {
        bestIdx = remid::HighestPIDTrack(sliceTracks, fPIDLabel, evt);
        for(size_t j = 0; j < sliceTracks.size(); ++j){ // loop over Kalman tracks
          const art::Ptr<rb::Track> track = sliceTracks[j];
          art::Ptr<remid::ReMId> trackRemid = remidVec.at(j);
          if(!track->Is3D() || trackRemid->Value() < 0) continue;
          if(cosrejobj.NTracks3D()==-5) cosrejobj.SetNTracks3D(1);
          else cosrejobj.SetNTracks3D(cosrejobj.NTracks3D()+1);
          if(trackRemid->Value() < 0.5) nhadtracks++;
        }
      }

      if(cosrejobj.NTracks3D()>0 && bestIdx!=999) {

        const art::Ptr<rb::Track> track = sliceTracks[bestIdx];

        art::PtrVector<rb::CellHit> trkhits = track->AllCells();

        if(bt->HaveTruthInfo()) {
          const std::vector<const sim::Particle*> parts = bt->HitsToParticle(trkhits);
          const std::vector< cheat::TrackIDE > trids = bt->HitsToTrackIDE(trkhits);

          if(!parts.empty()){
            const TVector3 trueDir = parts[0]->Momentum().Vect().Unit();
            cosrejobj.SetKalThetaTrue(trueDir.Dot(track->Dir()));
            cosrejobj.SetPDGBest(parts[0]->PdgCode());
          }
        }

        double tdx = track->Dir().X();
        double tdy = track->Dir().Y();
        double tdz = track->Dir().Z();

        if(!(tdx>=-1&&tdx<=1&&tdy>=-1&&tdy<=1&&tdz>=-1&&tdz<=1)) badtrack = true;

        if(badtrack) continue;

        TVector3 start = track->Start();
        TVector3 end = track->Stop();
        TVector3 dirbak = -track->Dir().Unit();
        TVector3 dirfor = (track->TrajectoryPoint(track->NTrajectoryPoints()-1) - track->TrajectoryPoint(track->NTrajectoryPoints()-2)).Unit();

        cosrejobj.SetKalFwdDist(livegeom->ProjectedLiveDistanceToEdge(end,dirfor));
        cosrejobj.SetKalBakDist(livegeom->ProjectedLiveDistanceToEdge(start,dirbak));
        cosrejobj.SetKalFwdCell(livegeom->ProjectedLiveCellsToEdge(end,dirfor));
        cosrejobj.SetKalBakCell(livegeom->ProjectedLiveCellsToEdge(start,dirbak));

        cosrejobj.SetVtxDist(std::min(livegeom->DistToEdgeXY(start),livegeom->DistToEdgeZ(start)));
        cosrejobj.SetEndDist(std::min(livegeom->DistToEdgeXY(end),livegeom->DistToEdgeZ(end)));

        if(geom->DetId() == novadaq::cnv::kNEARDET) {
          cosrejobj.SetKalFwdSteel(livegeom->ProjectedSteelDist(track->Stop(), dirfor));
          cosrejobj.SetKalBakSteel(livegeom->ProjectedSteelDist(track->Start(), -(track->Dir())));
          cosrejobj.SetKalFwdAir(livegeom->ProjectedAirDist(track->Stop(), dirfor));
          cosrejobj.SetKalBakAir(livegeom->ProjectedAirDist(track->Start(), -(track->Dir())));
          cosrejobj.SetKalYPosAtTrans(livegeom->YPositionAtTransition(track->Start(), track->Dir()));
          cosrejobj.SetKalFwdCellND(livegeom->FullNDProjectedCells(track->Stop(), dirfor));
          cosrejobj.SetKalBakCellND(livegeom->FullNDProjectedCells(track->Start(), -(track->Dir())));
        }

        if(std::min(fwddist,bakdist) < cosrejobj.MinDist()) cosrejobj.SetMinDist(std::min(fwddist,bakdist));
        if(std::min(fwdcell,bakcell) < cosrejobj.MinCell()) cosrejobj.SetMinCell(std::min(fwdcell,bakcell));

        cosrejobj.SetERatio(track->TotalGeV()/slice->TotalGeV());
        scatt = fCosRejFxs->getScatt(track);
        if(track->TotalLength()>0) cosrejobj.SetScatt(scatt/track->TotalLength());

        double ferpars[4];
        bool fdef = 0;
        bool fFlag = fCosRejFxs->FScatterEstim(track,fdef,ferpars[0],ferpars[1],ferpars[2],ferpars[3]); // Fernandas scattering energy estimate
        if(fFlag==1) {
          cosrejobj.SetFScattExt(ferpars[0]);
          cosrejobj.SetFScattSig(ferpars[1]);
          cosrejobj.SetFScattMax(ferpars[2]);
          cosrejobj.SetFScattSum(ferpars[3]);
        }
        tf::TimingFitResult tfres = tfalg.HoughFit(track.get());
        cosrejobj.SetKFitSpeed(tfres.bestInvSpeed);
        cosrejobj.SetKDirScore(tfres.fwdScore-tfres.revScore);
        cosrejobj.SetKScoreDiff(tfres.bestScore-std::max(tfres.fwdScore,tfres.revScore));
        cosrejobj.SetStartAct(fCosRejFxs->getActivity(track->AllCells(),slice->AllCells()));
        cosrejobj.SetNHadTracks(nhadtracks);

        double angleOfBestTrack = fCosRejFxs->getAngle(sliceTracks[bestIdx]->Dir());
        if(angleOfBestTrack>-6&&angleOfBestTrack<2) cosrejobj.SetAngleKal(angleOfBestTrack);

        if(geom->DetId() == novadaq::cnv::kFARDET) {
          const art::Ptr<rb::Track> track = sliceTracks[bestIdx];
          TMVAvars[0] = cosrejobj.AngleKal();
          TMVAvars[1] = cosrejobj.AngleCos();
          TMVAvars[3] = track->Dir().Y();
          TMVAvars[4] = maxp.Y();
          TMVAvars[6] = cosrejobj.CosFwdCell();
          TMVAvars[7] = cosrejobj.CosBakCell();
          TMVAvars[8] = cosrejobj.KalFwdCell();
          TMVAvars[9] = cosrejobj.KalBakCell();
          TMVAvars[10] = cosrejobj.Scatt() / track->TotalLength();
          TMVAvars[11] = cosrejobj.StartAct();
          TMVAvars[12] = slice->CalorimetricEnergy();
          TMVAvars[13] = 0;  // after this is just spectator variables, don't need to be set even
          TMVAvars[14] = 0;
          TMVAvars[15] = 0;
          TMVAvars[16] = 0;
          TMVAvars[17] = 0;
          TMVAvars[18] = 0;
          TMVAvars[19] = 0;
          TMVAvars[20] = 0;
          TMVAvars[21] = 0;
          TMVAvars[22] = 0;

          cosrejobj.SetOldCosPID(fReaderOld->EvaluateMVA("CC_BDTA"));
          cosrejobj.SetUnconCosPID(fReaderUC->EvaluateMVA("UC_BDTA"));

          TMVAvarsNew[0] = cosrejobj.AngleKal();
          TMVAvarsNew[1] = track->Dir().Y();;
          TMVAvarsNew[2] = maxp.Y();
          TMVAvarsNew[3] = minp.Y();
          TMVAvarsNew[4] = track->NCell();
          TMVAvarsNew[5] = track->TotalLength();
          TMVAvarsNew[6] = cosrejobj.KalBakCell();
          TMVAvarsNew[7] = cosrejobj.KalFwdCell();
          TMVAvarsNew[8] = cosrejobj.Scatt() / track->TotalLength();
          TMVAvarsNew[9] = slice->NCell();
          TMVAvarsNew[10] = slice->CalorimetricEnergy()/slice->NCell();
          TMVAvarsNew[11] = cosrejobj.NTracks3D();

          TMVAvarsNew[12] = 0; // spectators
          TMVAvarsNew[13] = 0;
          TMVAvarsNew[14] = 0;
          TMVAvarsNew[15] = 0;

          cosrejobj.SetPreCosPID(fReaderPre->EvaluateMVA("PRE_BDTA"));
          cosrejobj.SetPostCosPID(fReaderPost->EvaluateMVA("POST_BDTA"));
          int frun = evt.run();
          if(frun > 10000 && frun < 17300) cosrejobj.SetConCosPID(cosrejobj.PreCosPID());
          else cosrejobj.SetConCosPID(cosrejobj.PostCosPID());
        }

      } // end if good track

      outputObjects->push_back(cosrejobj);  // write out object with all information
      util::CreateAssn(*this,evt,*(outputObjects.get()),slice,*(assoc.get()));

    } //end loop over slices

    evt.put(std::move(outputObjects));
    evt.put(std::move(assoc));

  } //end produce

} //end namespace


DEFINE_ART_MODULE(cosrej::CosRej)
