////////////////////////////////////////////////////////////////////////
/// \file  NOvARawInputSource.cxx
/// \brief Source to convert raw binary files to root files
///
/// \version $Id: NOvARawInputDriver.cxx,v 1.19 2012-12-01 07:05:47 bckhouse Exp $
/// \author  brebel@fnal.gov, soderber@fnal.gov, greenc@fnal.gov, paterno@fnal.gov
////////////////////////////////////////////////////////////////////////

//NOvA includes
#include "DAQ2RawDigit/NOvARawInputDriver.h"
#include "DAQ2RawDigit/OnlineUnpack.h"
#include "GeometryObjects/GeometryBase.h"
#include "NovaTimingUtilities/TimingUtilities.h"
#include "RawData/DAQHeader.h"
#include "RawData/RawDigit.h"
#include "RawData/RawTrigger.h"
#include "RawData/FlatDAQData.h"
#include "SummaryData/RunData.h"
#include "NovaDAQConventions/DAQConventions.h"

// Framework includes
#include "art/Framework/IO/Sources/put_product_in_principal.h"
#include "art/Utilities/Exception.h"
#include "art/Framework/Core/ProductRegistryHelper.h"
#include "art/Framework/Principal/EventPrincipal.h"
#include "art/Framework/Principal/RunPrincipal.h"
#include "art/Framework/Principal/SubRunPrincipal.h"
#include "art/Framework/Core/FileBlock.h"
#include "fhiclcpp/ParameterSet.h"


namespace daq2raw{

  //------------------------------------------------------------------------------
  // Required constructor.
  NOvARawInputDriver::NOvARawInputDriver(fhicl::ParameterSet const &pset,
					 art::ProductRegistryHelper &helper,
					 art::SourceHelper const &pm) 
    // look to see if the user provided a base file name for the gdml file, if not
    // pass along an emptry string to default to the behavior previously in NOvAInputDriverBase
    : NOvAInputDriverBase(pset.get<std::string>("GDMLFileBase",""))
    , fSourceHelper(pm)
    , fCurrentFilename()
    , fCurrentSubRunID(-1, -1) ///< Has to be distinguished from the ones that are acually used
    , fParser()
    , fMaxEventSize(pset.get<size_t>("maxEventSizeK", 65536) * 1024)
    , fFilterCorruptedNanoslices(pset.get<bool>("FilterCorruptedNanoslices"))
    , fDoFillFlatDAQData(pset.get<bool>("DoFillFlatDAQData"))
  {
    // Declare products that *may* be in the event (or run, or subRun)
    // here. Note the use of "reconstitutes" vs "produces": this is what
    // allows us to specify module label, process name and (optionally)
    // instance label for each product.
    helper.reconstitutes<rawdata::DAQHeader,                art::InEvent>("daq");
    helper.reconstitutes<std::vector<rawdata::RawDigit>,    art::InEvent>("daq");
    helper.reconstitutes<std::vector<rawdata::RawTrigger>,  art::InEvent>("daq");
    if(fDoFillFlatDAQData) helper.reconstitutes<std::vector<rawdata::FlatDAQData>, art::InEvent>("daq");
    helper.reconstitutes<sumdata::RunData,                  art::InRun  >("daq");
    // Turn off printf messages in the parser.
    fParser.setDebug(false);
  }

  //------------------------------------------------------------------------------
  // Required function: close currently-open file.
  void NOvARawInputDriver::closeCurrentFile() {
    fParser.close(); // Ignore error return
  }

  // Required function: open the file, "name" and construct and return a
  // new FileBlock object. MUST be successful or throw:
  // art::Exception(art::errors::FileOpenError) or
  // art::Exception(art::errors::FileReadError) are good candidates.
  void NOvARawInputDriver::readFile(std::string const &name,
				    art::FileBlock* &fb) {
    fCurrentFilename = name;

    int fd = fParser.open(fCurrentFilename.c_str());
    if (fd == -1) { // Failure
      throw art::Exception(art::errors::FileOpenError)
	<< "Unable to open file "
	<< name
	<< " due to system error, \""
	<< strerror(errno)
	<< "\"\n";
    }

    fb = new art::FileBlock(art::FileFormatVersion(1, "NOvARawInput 2011a"),
			    fCurrentFilename);
  }

  // Required function: read and fill Run, SubRun and/or Event as might be
  // appropriate.
  bool NOvARawInputDriver::readNext(art::RunPrincipal* const &inR,
				    art::SubRunPrincipal* const &inSR,
				    art::RunPrincipal* &outR,
				    art::SubRunPrincipal* &outSR,
				    art::EventPrincipal* &outE)
  {
    ////////////////////////////////////
    // Unpack the data.

    // NOTE: As of OnlineUnpack.cxx r1.9, the histogram pointers required
    // in the constructor are *not* filled. In the event that one *does*
    // want this kind of information, the data to be histogrammed should
    // be put into event products (not necessarily written out to file)
    // and turned into histograms by a downstream analyzer. This will
    // enable easy management of different histograms for diferent runs,
    // subruns, etc, etc and trivial control of whether histograns are
    // actually wanted or not.
    daq2raw::OnlineUnpack up(nullptr, nullptr, nullptr, &fCurrentFilename);
    up.fFilterCorruptedNanoslices = fFilterCorruptedNanoslices;

    /// Define data structures, which will be filled by OnlineUnpack::ProcessRawEvent
    std::unique_ptr<std::vector<rawdata::RawDigit> > digits(new std::vector<rawdata::RawDigit>);
    rawdata::RawTrigger                            raw_trigger;
    rawdata::FlatDAQData                           flat_daq_data;
    rawdata::DAQHeader                             daq_header;

    /// Fill the data structures for this event.
    if (!up.ProcessRawEvent(fParser, digits.get(), &raw_trigger, &flat_daq_data, &daq_header)) return false;

    // Get the actual timestamp here - use the current time as the
    // default, maybe the default should really be a ridiculous value
    struct timespec ts;

    novadaq::timeutils::
      convertNovaTimeToUnixTime(raw_trigger.fTriggerTimingMarker_TimeStart, ts);

    // art::Timestamp has the seconds since the beginning of the epoch
    // in the upper 32 bits and the fractional seconds in the lower 32
    // bits.
    art::Timestamp tstamp = (ts.tv_sec << 32) | ts.tv_nsec;

    ////////////////////////////////////
    // Assemble new principals as required.
    art::SubRunID newID(up.getRunNumber(), up.getSubrunNumber());
    if (fCurrentSubRunID.runID() != newID.runID()) { // New Run
      outR = fSourceHelper.makeRunPrincipal(up.getRunNumber(), tstamp);

      // get the base form of the geometry file for the current detector
      //\todo at some point we should query a database to get this, especially
      //\todo when we have instances of only part of a detector working
      std::string detFileBase = getDetectorGDML(up.getDetId());
      std::unique_ptr<sumdata::RunData> rundata(new sumdata::RunData(novadaq::cnv::DetId(up.getDetId()), detFileBase, geo::GeometryBase::ExtractGDML(detFileBase+".gdml", false)));
      art::put_product_in_principal(std::move(rundata),
				    *outR,
				    "daq"); // Module label
    }
    if (fCurrentSubRunID != newID) { // New SubRun
      outSR = fSourceHelper.makeSubRunPrincipal(up.getRunNumber(),
						  up.getSubrunNumber(),
						  tstamp);
      fCurrentSubRunID = newID;
    }
    outE = fSourceHelper.makeEventPrincipal(up.getRunNumber(),
					      up.getSubrunNumber(),
					      up.getEventNumber(),
					      tstamp);

    ////////////////////////////////////
    // Add the event products to the event:

    // Digitized data:
    art::put_product_in_principal(std::move(digits), *outE, "daq");

    // Trigger data:
    std::unique_ptr<std::vector<rawdata::RawTrigger> > rtcol(new std::vector<rawdata::RawTrigger>);
    rtcol->push_back(raw_trigger); // Read one, product is a vector.
    art::put_product_in_principal(std::move(rtcol),
				  *outE,
				  "daq"); // Module label

    // DAQ header:
    std::unique_ptr<rawdata::DAQHeader > daqcol(new rawdata::DAQHeader(daq_header));
    art::put_product_in_principal(std::move(daqcol),
				  *outE,
				  "daq"); // Module label

    /// Put FlatDAQData object
    if(fDoFillFlatDAQData) {
      std::unique_ptr<std::vector<rawdata::FlatDAQData> > flatcol(new std::vector<rawdata::FlatDAQData>);
      flatcol->push_back(flat_daq_data); // Read one, product is a vector.
      art::put_product_in_principal(std::move(flatcol),
				    *outE,
				    "daq"); // Module label
    }

    return true;
  }

}// end of namespace
