////////////////////////////////////////////////////////////////////////
// Class:       ClusterConvert
// Module Type: filter
// File:        ClusterConvert_module.cc
//
// Takes an offline rb::Cluster and turns it into a ddt HitList
//
// Generated at Wed Sep 24 06:29:23 2014 by Alec Habig using artmod
// from cetpkgsupport v1_07_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
// #include "art/Framework/Principal/Run.h"
// #include "art/Framework/Principal/SubRun.h"
// #include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "DDTBaseDataProducts/HitList.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"

#include <memory>

namespace novaddt {
  class ClusterConvert;
}

class novaddt::ClusterConvert : public art::EDFilter {
public:
  explicit ClusterConvert(fhicl::ParameterSet const & p);
  // The destructor generated by the compiler is fine for classes
  // without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  ClusterConvert(ClusterConvert const &) = delete;
  ClusterConvert(ClusterConvert &&) = delete;
  ClusterConvert & operator = (ClusterConvert const &) = delete;
  ClusterConvert & operator = (ClusterConvert &&) = delete;

  // Required functions.
  bool filter(art::Event & e) override;


private:

  // Declare member data here.
  std::string     fInputLabel;
  std::string     fInputInstance;
  int             fDetector;
  bool            fVerbose;
};


novaddt::ClusterConvert::ClusterConvert(fhicl::ParameterSet const & p):
  fInputLabel     (p.get< std::string       >("InputLabel"   )),
  fInputInstance  (p.get< std::string       >("InputInstance")),
  fDetector       (p.get< int               >("Detector")),
  fVerbose        (p.get< bool              >("Verbose"))
{
  std::cout << "--- novaddt::ClusterConvert begin" << std::endl;
  std::cout << "\t Input hitlists label:    " << fInputLabel    << std::endl;
  std::cout << "\t Input hitlists instance  " << fInputInstance << std::endl;
  std::cout << "\t Detector ID              " << fDetector      << std::endl;
  std::cout << "\t Verbose                  " << fVerbose       << std::endl;

  // Call appropriate produces<>() functions here.
  produces<std::vector<HitList>>("ConvertedClusters");
}

bool novaddt::ClusterConvert::filter(art::Event & event)
{
  art::Handle<std::vector<rb::Cluster> > clusters;
  event.getByLabel(fInputLabel, fInputInstance, clusters);

  //make a grouped hitlist out of hits after slicing                          
  std::unique_ptr< std::vector<HitList> >product(new std::vector<HitList>);

  // Loop over clusters
  if (fVerbose) std::cout << "ClusterConvert " << clusters->size()<<" clusters" << std::endl;
  for(unsigned int i = 0; i < clusters->size(); ++i) {
    // Treat each cluster as a slice.
    // Make a slice
    novaddt::HitList slice;
    // loop over hits, copy into slice
    if (fVerbose) std::cout << "ClusterConvert " << clusters->at(i).NCell()<<" cells" << std::endl;
    for(unsigned int j = 0; j < clusters->at(i).NCell(); ++j) {
      // manually transfer elements
      art::Ptr<rb::CellHit> hit(clusters->at(i).Cell(j));
      novaddt::DAQHit Dhit(fDetector,hit->DaqChannel(),hit->ADC(),hit->TDC());
      slice.push_back(Dhit);      
    }
    // sort the slice by TDC
    if (slice.size() > 0)
      std::sort(slice.begin(), slice.end(), CompareDAQHit<TDC>());

    if (fVerbose) for (auto const& Dhit : slice) {
      std::cout << "Channel:" << Dhit.Chan() ;
      std::cout << " ADC:"<< Dhit.ADC().val << " TDC:" << Dhit.TDC().val << " Plane:" << Dhit.Plane().val << " Cell:" << Dhit.Cell().val << std::endl;
    }
    // store slice in product
    product->emplace_back(slice);
  }

  event.put(std::move(product),"ConvertedClusters");

  return true;
}

DEFINE_ART_MODULE(novaddt::ClusterConvert)
