///
/// \file     DAQDataSummary_module.cc
/// \brief    Analysis module to walk the DAQ data and extract useful
///           information about how it was unpacked
/// \version: $Id:$
/// \author:  messier@indiana.edu
///

// ART Includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"

// ROOT Includes
#include "TTree.h"
#include "TH1F.h"

// NOvA Includes
#include "DAQDataFormats/RawEvent.h"
#include "DAQDataFormats/RawTrigger.h"
#include "DAQDataFormats/RawTriggerMask.h"
#include "DAQDataFormats/RawDataBlock.h"
#include "DAQDataFormats/RawMicroBlock.h"
#include "DAQDataFormats/RawMicroSlice.h"
#include "DAQDataFormats/RawNanoSlice.h"
#include "RawData/FlatDAQData.h"
#include "DQSpillFlags/DAQEventUnpack.h"
#include "DQSpillFlags/DAQEventSummary.h"

namespace dqsf {
  ///
  /// A module to help analysis of the raw DAQ data
  ///
  class DAQDataSummary : public art::EDAnalyzer
  {
  private:
    TTree* fDataSummaryTree;

    bool         eventIncomplete;           ///< Is the event incomplete?
    unsigned int emptyDataBlock;            ///< How many empty data blocks?
    unsigned int emptyMicroSlice;           ///< How many empty micro slices?
    unsigned int dataBlockMissingData;      ///< # of occurances of isMissingData
    unsigned int microSliceDataNotPresent;  ///< # of microslices with !DataPresen
    unsigned int totalNanoSlices;           ///< # of nano slices in the event
    unsigned int nanoSliceDataNotPresent;   ///< # of nanoslices reporting !DataPresent
    unsigned int nanoSliceNoLinkStatus;     ///< # of nanoslices reporting !LinkPresent
    unsigned int nanoSliceBufferEmpty;      ///< # of nanoslices reporting BufferEmpty
    unsigned int nanoSliceBufferFull;       ///< # of nanoslices reporting BufferFull
    unsigned int nanoSliceCommError;        ///< # of nanoslices reporting CommError
    unsigned int nanoSlicePackerError;      ///< # of nanoslices reporting PacketError
    unsigned int nanoSliceOverflowError;    ///< # of nanoslices reporting OverflowError
    unsigned int nanoSliceADCError;         ///< # of nanoslices reporting ADCError

  public:

    explicit DAQDataSummary(fhicl::ParameterSet const &p) : EDAnalyzer(p)
    {
      this->reconfigure(p);
    }

    //..................................................................

    virtual ~DAQDataSummary() { }

    //..................................................................

    void reconfigure(fhicl::ParameterSet const &p) { }

    //..................................................................

    void beginJob(){
      art::ServiceHandle<art::TFileService> tfs;

      fDataSummaryTree = tfs->make<TTree>("datasummary", "Data Summary Tree");

      fDataSummaryTree->Branch("fEventIncomplete",           &eventIncomplete,          "fNanoSlices/O");
      fDataSummaryTree->Branch("fNemptyDataBlock",           &emptyDataBlock,           "fNemptyDataBlock/i");
      fDataSummaryTree->Branch("fNemptyMicroSlice",          &emptyMicroSlice,          "fNemptyMicroSlice/i");
      fDataSummaryTree->Branch("fNdataBlockMissingData",     &dataBlockMissingData,     "fNdataBlockMissingData/i");
      fDataSummaryTree->Branch("fNmicroSliceDataNotPresent", &microSliceDataNotPresent, "fNmicroSliceDataNotPresent/i");
      fDataSummaryTree->Branch("fNtotalNanoSlices",          &totalNanoSlices,          "fNtotalNanoSlices/i");
      fDataSummaryTree->Branch("fNanoSliceDataNotPresent",   &nanoSliceDataNotPresent,  "fNanoSliceDataNotPresent/i");
      fDataSummaryTree->Branch("fNanoSliceNoLinkStatus",     &nanoSliceNoLinkStatus,    "fNanoSliceNoLinkStatus/i");
      fDataSummaryTree->Branch("fNanoSliceBufferEmpty",      &nanoSliceBufferEmpty,     "fNanoSliceBufferEmpty/i");
      fDataSummaryTree->Branch("fNanoSliceBufferFull",       &nanoSliceBufferFull,      "fNanoSliceBufferFull/i");
      fDataSummaryTree->Branch("fNanoSliceCommError",        &nanoSliceCommError,       "fNanoSliceCommError/i");
      fDataSummaryTree->Branch("fNanoSlicePackerError",      &nanoSlicePackerError,     "fNanoSlicePackerError/i");
      fDataSummaryTree->Branch("fNanoSliceOverflowError",    &nanoSliceOverflowError,   "fNanoSliceOverflowError/i");
      fDataSummaryTree->Branch("fNanoSliceADCError",         &nanoSliceADCError,        "fNanoSliceADCError/i");
    }

    bool UnpackRawDaqData(art::Event const& evt,
			  daqdataformats::RawEvent& rawevt)
    {
      try {
      	art::Handle<std::vector<rawdata::FlatDAQData> > flat_hdl;
      	evt.getByLabel("daq", flat_hdl);
      	if (flat_hdl->size()!=1) {
      	  std::cerr << __FILE__ << ":" << __LINE__
      		    << " Must have one and only one flat daq object."
      		    << " Found " << flat_hdl->size() << std::endl;
      	  return false;
      	}
      	rawevt.readData( (*flat_hdl)[0].getRawBufferPointer() );
      	return true;
      }
      catch (...) {
      	std::cerr << __FILE__ << ":" << __LINE__
      		  << " Bad unpack of flat DAQ data!"
      		  << std::endl;
      	return false;
      }
      //
      // Only errors fall through to here.
      //
      return false;
    }

    //..................................................................

    void analyze(art::Event const& evt)
    {
      static daqdataformats::RawEvent rawevt;

      bool aok = this->UnpackRawDaqData(evt, rawevt);
      if (aok==false) {
      	abort();
      }

      //
      // Unpack, analyze, and summarize the DAQ data
      //
      DAQEventUnpack  daqUnpack;
      DAQEventSummary daqSummary;
      daqUnpack.UnpackRawEvent(rawevt, daqSummary);
      daqSummary.Finalize();

      eventIncomplete          = daqSummary.fEventIncomplete;
      emptyDataBlock           = daqSummary.fNemptyDataBlock;
      emptyMicroSlice          = daqSummary.fNemptyMicroSlice;
      dataBlockMissingData     = daqSummary.fNdataBlockMissingData;
      microSliceDataNotPresent = daqSummary.fNmicroSliceDataNotPresent;
      totalNanoSlices          = daqSummary.fNtotalNanoSlices;
      nanoSliceDataNotPresent  = daqSummary.fNanoSliceDataNotPresent;
      nanoSliceNoLinkStatus    = daqSummary.fNanoSliceNoLinkStatus;
      nanoSliceBufferEmpty     = daqSummary.fNanoSliceBufferEmpty;
      nanoSliceBufferFull      = daqSummary.fNanoSliceBufferFull;
      nanoSliceCommError       = daqSummary.fNanoSliceCommError;
      nanoSlicePackerError     = daqSummary.fNanoSlicePackerError;
      nanoSliceOverflowError   = daqSummary.fNanoSliceOverflowError;
      nanoSliceADCError        = daqSummary.fNanoSliceADCError;

      fDataSummaryTree->Fill();

      //std::cout << "No. nano slices: " << totalNanoSlices << std::endl;
    }

  }; // class DAQDataSummary
  DEFINE_ART_MODULE(DAQDataSummary);
}

////////////////////////////////////////////////////////////////////////
