////////////////////////////////////////////////////////////////////////
// \file    ElasticArmsValidate.cxx
// \brief   Analysis module for performance of FuzzyK
// \version $Id: ElasticArmsValidate.cxx,v 1.2 2012-12-12 15:55:23 edniner Exp $
// \author  $Author: edniner $
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <map>
#include <vector>
#include <string>

// ROOT includes
#include "TTree.h"
#include "TH1F.h"
#include "TH2F.h"

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Core/FindManyP.h"

// NOvASoft includes
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "Utilities/func/MathUtil.h"
#include "MCCheater/BackTracker.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "GeometryObjects/Geo.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Vertex.h"

namespace earms {
  class ElasticArmsValidate : public art::EDAnalyzer {
    public:
      explicit ElasticArmsValidate(fhicl::ParameterSet const& pset);    
      ~ElasticArmsValidate();                        
      
      void analyze(const art::Event& evt); 
      void reconfigure(const fhicl::ParameterSet& pset);
      void beginJob();
      void ClearVectors();
      std::vector<int> getFaveIdsByEfficiency(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                              std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable);
      
    private:

      /// MC information
      std::string      fSliceLabel;
      std::string      fCellHitLabel;
      std::string      fVertexLabel;
      int              fCompMCNeutrino;
      
      TTree* fOutTree;
      //event summary info
      int run;
      int subrun;
      int event;

      //vertex position, reconstructed
      float rVx;
      float rVy;
      float rVz;

      //vertex histograms
      TH1F* fHistoRVX;//reco x position
      TH1F* fHistoRVY;//reco y position
      TH1F* fHistoRVZ;//reco z position
      TH1F* fHistoDVX;//delta x
      TH1F* fHistoDVY;//delta y
      TH1F* fHistoDVZ;//delta z
      TH1F* fHistoDVT;//total diff between reco and true vertex

      //for cosmics make histograms for endpoint as well
      TH1F* fHistoDEX;//delta x
      TH1F* fHistoDEY;//delta y
      TH1F* fHistoDEZ;//delta z
      TH1F* fHistoDET;//total diff between reco and true vertex

      //neutrino information
      float Vx;//start x
      float Vy;//start y
      float Vz;//start z
      float NuEff;//efficiency of neutrino for slice
      float NuPur;//purity of neutrino for slice
      float NuEVSl;//visible energy of neutrino in slice
      float NuEVTot;//visible energy of neutrino in all slices
      int NuPdg;//neutrino pdg
      int ccnc;//is cc or nc
      int mode;//interaction mode
      float NuE;//total energy of neutrino

      //for cosmics also store endpoint since elastic arms can get confused
      float Ex;//end x
      float Ey;//end y
      float Ez;//end z
    };
}

namespace earms
{
  //.......................................................................
  ElasticArmsValidate::ElasticArmsValidate(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset)
  {
    this->reconfigure(pset);
  }

  //......................................................................
  ElasticArmsValidate::~ElasticArmsValidate()
  {
    //======================================================================
    // Clean up any memory allocated by your module
    //======================================================================
  }

  //......................................................................
  void ElasticArmsValidate::reconfigure(const fhicl::ParameterSet& pset)
  {

    fSliceLabel        = pset.get< std::string >("SliceLabel");
    fCellHitLabel      = pset.get< std::string >("CellHitLabel");
    fVertexLabel       = pset.get< std::string >("VertexLabel");
    fCompMCNeutrino    = pset.get< int         >("CompMCNeutrino");
    
  }
      
  //......................................................................
  void ElasticArmsValidate::beginJob()
  {

    art::ServiceHandle<art::TFileService> tfs;

    // define the tree
    fOutTree = tfs->make<TTree>("ElasticArmsValidate","Elastic Arms Validation");

    //store event summary information
    fOutTree->Branch("run", &run, "run/I");
    fOutTree->Branch("subrun", &subrun, "subrun/I");
    fOutTree->Branch("event", &event, "event/I");

    //store vertex info
    fOutTree->Branch("rVx", &rVx);
    fOutTree->Branch("rVy", &rVy);
    fOutTree->Branch("rVz", &rVz);

    //vertex histograms
    fHistoRVX = tfs->make<TH1F>("fHistoRVX","Reconstructed X vertex coordinate;x(cm)",160,-800.0,800.0);
    fHistoRVY = tfs->make<TH1F>("fHistoRVY","Reconstructed Y vertex coordinate;y(cm)",160,-800.0,800.0);
    fHistoRVZ = tfs->make<TH1F>("fHistoRVZ","Reconstructed Z vertex coordinate;z(cm)",320,0.0,6400.0);
    fHistoDVX = tfs->make<TH1F>("fHistoDVX","Vertex True X - Reco X;(cm)",201,-100.5,100.5);
    fHistoDVY = tfs->make<TH1F>("fHistoDVY","Vertex True Y - Reco Y;(cm)",201,-100.5,100.5);
    fHistoDVZ = tfs->make<TH1F>("fHistoDVZ","Vertex True Z - Reco Z;(cm)",201,-100.5,100.5);
    fHistoDVT = tfs->make<TH1F>("fHistoDVT","Distance from Reco to True vertex;(cm)",150,0.0,300.0);
    fHistoDEX = tfs->make<TH1F>("fHistoDEX","Endpoint True X - Reco X;(cm)",201,-100.5,100.5);
    fHistoDEY = tfs->make<TH1F>("fHistoDEY","Endpoint True Y - Reco Y;(cm)",201,-100.5,100.5);
    fHistoDEZ = tfs->make<TH1F>("fHistoDEZ","Endpoint True Z - Reco Z;(cm)",201,-100.5,100.5);
    fHistoDET = tfs->make<TH1F>("fHistoDET","Distance from Reco to Endpoint;(cm)",150,0.0,300.0);

    //store neutrino info
    fOutTree->Branch("Vx", &Vx);//neutrino position info
    fOutTree->Branch("Vy", &Vy);
    fOutTree->Branch("Vz", &Vz);
    fOutTree->Branch("Ex", &Ex);
    fOutTree->Branch("Ey", &Ey);
    fOutTree->Branch("Ez", &Ez);
    fOutTree->Branch("NuEff", &NuEff);//info on quality of neutrino in slice
    fOutTree->Branch("NuPur", &NuPur);
    fOutTree->Branch("NuEVSl", &NuEVSl);
    fOutTree->Branch("NuEVTot", &NuEVTot);
    fOutTree->Branch("NuPdg", &NuPdg);
    fOutTree->Branch("ccnc", &ccnc);
    fOutTree->Branch("mode", &mode);
    fOutTree->Branch("NuE", &NuE);
    
  }

  //......................................................................
  void ElasticArmsValidate::analyze(const art::Event& evt)
  {
    art::ServiceHandle<cheat::BackTracker> bt;
    art::ServiceHandle<geo::Geometry> geom;

    art::Handle<std::vector<rb::Cluster> > sHandle;
    evt.getByLabel(fSliceLabel, sHandle);
    if(sHandle.failedToGet()){
      std::cout<<"No slices found, skipping event" << std::endl;
    }
    art::PtrVector<rb::Cluster> slices;
    for(unsigned int i=0; i<sHandle->size(); ++i){
      slices.push_back(art::Ptr<rb::Cluster>(sHandle, i));
    }

    //get all the hits to calculate neutrino efficiency/purity for slices
    art::Handle<std::vector<rb::CellHit>> htHandle;
    evt.getByLabel(fCellHitLabel, htHandle);
    art::PtrVector<rb::CellHit> allHits;
    for(unsigned int i=0; i<htHandle->size(); ++i){
      art::Ptr<rb::CellHit> hit(htHandle,i);
      allHits.push_back(hit);
    }

    //get associations between slices and vertices
    art::FindManyP<rb::Vertex> fmVert(sHandle, evt, fVertexLabel);

    //if these are events with neutrino information to compare too, get it
    std::vector<int> bestNuId;
    std::vector<cheat::NeutrinoWithIndex> nus;
    std::vector<std::vector<cheat::NeutrinoEffPur>> sEffPur;
    if(fCompMCNeutrino == 1){
      nus = bt->allMCTruth();
      //std::cout<<allNus.size()<<std::endl;
      sEffPur = bt->SlicesToMCTruthsTable(slices);
      //std::cout<<"Done matching"<<std::endl;
    
      bestNuId = getFaveIdsByEfficiency(nus, sEffPur); 
    }
    

    //loop over slices
    for( unsigned int i=0; i<slices.size(); ++i){
      this->ClearVectors();
      
      //skip noise slice
      if (slices[i]->IsNoise()) continue;
      //if we are comparing to truth, make sure there is a neutrino with this slice
      if(fCompMCNeutrino == 1){
        if (bestNuId[i] == -1) continue;
      }
      //make sure there is a vertex associated with this slice
      if(!(fmVert.isValid())) continue;

      std::vector<art::Ptr<rb::Vertex>> v = fmVert.at(i);
      //continue if there is no vertex or more then 1 vertex
      if (v.size() != 1) continue;


      //get reco vertex position
      rVx = v[0]->GetX();
      rVy = v[0]->GetY();
      rVz = v[0]->GetZ();
      //fill position histograms
      fHistoRVX->Fill(rVx);
      fHistoRVY->Fill(rVy);
      fHistoRVZ->Fill(rVz);

      //if we do not want to compare to truth, dump into to tree and move to next slice
      if(fCompMCNeutrino != 1){
        fOutTree->Fill();
        continue;
      }

      //now access true neutrino info for further validation
      art::Ptr<simb::MCTruth> nuTru = sEffPur[i][bestNuId[i]].neutrinoInt;

      NuEff = sEffPur[i][0].efficiency;
      NuPur = sEffPur[i][0].purity;
      NuEVSl = sEffPur[i][0].energySlice;
      NuEVTot = sEffPur[i][0].energyTotal;

      //if we have a neutrino in this mc truth, proceed
      if (nuTru->NeutrinoSet()){
        NuPdg = nuTru->GetNeutrino().Nu().PdgCode();
        ccnc = nuTru->GetNeutrino().CCNC();
        mode = nuTru->GetNeutrino().Mode();
        NuE = nuTru->GetNeutrino().Nu().Momentum().E();

        //now get vertex info and make comparisons
        Vx = nuTru->GetNeutrino().Nu().Vx();
        Vy = nuTru->GetNeutrino().Nu().Vy();
        Vz = nuTru->GetNeutrino().Nu().Vz();
        fHistoDVX->Fill(Vx-rVx);
        fHistoDVY->Fill(Vy-rVy);
        fHistoDVZ->Fill(Vz-rVz);
        float distdiff = util::pythag(Vx-rVx,Vy-rVy,Vz-rVz);
        fHistoDVT->Fill(distdiff);
      }
      else {
        TVector3 enter;
        TVector3 exit;
        TVector3 stop;
        //if this is a cosmic ray then we need to find the start and end points inside the detector
        std::vector<const sim::Particle*> particles = bt->MCTruthToParticles(nuTru);
        const simb::MCTrajectory& traj = particles[0]->Trajectory();

        
        bool entered = 0;    
        bool exited = 0;    

        for(size_t pt = 0; pt < traj.size(); ++pt)
        {
          TVector3 point = traj.Position(pt).Vect();
          
          // Check to see if trajectory point is in the detector 
          bool inDet = fabs(point.X()) <= geom->DetHalfWidth()   // Check X inside detector 
                    && fabs(point.Y()) <= geom->DetHalfHeight()  /// Then check Y
                    && point.Z() >= 0 && point.Z() <= geom->DetLength(); // then check Z

          if(inDet && !entered)
          {
            // then we just entered the detector 
            entered = true;
            
            if(pt > 0)
            {
              // Put xyz from last point into an array for geo function call
              TVector3 lastPoint = traj.Position(pt - 1).Vect();
              double pointArray[3] = {point.X(), point.Y(), point.Z()};
              
              // Put xyz from direction vector into an array.  
              TVector3 dir = (lastPoint - point).Unit();
              double dirArray[3] = {dir.X(), dir.Y(), dir.Z()};
              
              double enterArray[3] = {0,0,0};
              
              geo::ProjectToBoxEdge(pointArray, dirArray, 
                                               - geom->DetHalfWidth()  , geom->DetHalfWidth(), 
                                               - geom->DetHalfHeight() , geom->DetHalfHeight(), 
                                                                    0  , geom->DetLength(), 
                                               enterArray);
              
              enter.SetXYZ(enterArray[0], enterArray[1], enterArray[2]);
              
            }
            else
            {// if this is the case, we started inside the detector.
              enter = point;
            }
          }
          
          if(!inDet && entered && !exited){
            // then we just exited the detector 
            exited= true;
            
            // Put xyz from last point into an array for geo function call
            TVector3 lastPoint = traj.Position(pt - 1).Vect();
            double lastPointArray[3] = {lastPoint.X(), lastPoint.Y(), lastPoint.Z()};
            
            // Put xyz from direction vector into an array.  
            TVector3 dir = (point - lastPoint).Unit();
            double dirArray[3] = {dir.X(), dir.Y(), dir.Z()};
            
            double exitArray[3] = {0,0,0};
            
            geo::ProjectToBoxEdge(lastPointArray, dirArray, 
                                  - geom->DetHalfWidth()  , geom->DetHalfWidth(), 
                                  - geom->DetHalfHeight() , geom->DetHalfHeight(), 
                                  0  , geom->DetLength(), 
                                  exitArray);
            
            exit.SetXYZ(exitArray[0], exitArray[1], exitArray[2]);
            
          }
          
          
        }
        stop = traj.Position(traj.size() - 1).Vect();
        
        // make sure the particle actually entered the detector  
        if(!entered){
          /// if the primary never entered the detector, set the coordinates to -5*10^9
          
          enter.SetXYZ(-5e9, -5e9, -5e9);
          exit.SetXYZ(-5e9, -5e9, -5e9);
          
        }

        /// If the particle entered and did not exit, set the exit position to stop
        if(entered && !exited){
          exit = stop;
        }

        Vx = enter.X();
        Vy = enter.Y();
        Vz = enter.Z();
        Ex = exit.X();
        Ey = exit.Y();
        Ez = exit.Z();

        fHistoDVX->Fill(Vx-rVx);
        fHistoDVY->Fill(Vy-rVy);
        fHistoDVZ->Fill(Vz-rVz);
        float distdiff = util::pythag(Vx-rVx,Vy-rVy,Vz-rVz);
        fHistoDVT->Fill(distdiff);

        fHistoDEX->Fill(Ex-rVx);
        fHistoDEY->Fill(Ey-rVy);
        fHistoDEZ->Fill(Ez-rVz);
        float distdiff2 = util::pythag(Ex-rVx,Ey-rVy,Ez-rVz);
        fHistoDET->Fill(distdiff2);


      }
      
      //fill tree
      fOutTree->Fill();
    }//end loop over slices


    
  }

  //.........................................................................
  void ElasticArmsValidate::ClearVectors()
  {
    ccnc = -1;
    NuE = -1.0;
    NuEff = -1.0;
    NuPur = -1.0;
    NuEVSl = -1.0;
    NuEVTot = -1.0;
    Vx = -999.0;
    Vy = -999.0;
    Vz = -999.0;
    NuPdg = 0;
    mode = -1;
  }
  
  //.........................................................................
  std::vector<int> ElasticArmsValidate::getFaveIdsByEfficiency(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                                               std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable)
  {
    // double check that slice truth table is safe
    for(auto slc = sliceTruthTable.begin(); slc != sliceTruthTable.end(); ++slc)
    {
      assert(slc->size() == allNus.size() && "Slice/truth table must be properly shaped");
    }
    
    // We'll make an array with an entry for every subevt to store the unique identifier of 
    // the favourite neutrino within the subevt, with the correct size and all entries initialized to -1
    std::vector<int> faveIds(sliceTruthTable.size(), -1);
    
    // if no neutrinos, there are no favorites 
    if(allNus.size()  == 0) 
      return faveIds;
    
    
    // List of best slice for each nu by slicer efficiency 
    std::vector<int> bestSlices(allNus.size());
    //First we find the best slice for each neutrino
    for(size_t iNu = 0; iNu < allNus.size(); ++iNu)
    {
      
      float maxValue =  0;  // we will use total slicer efficiency  
      int bestSlice   = -1;
      for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].efficiency > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].efficiency;
            bestSlice = iSlc;
          }
        }
      }
      
      bestSlices[iNu] = bestSlice;
    }
    
    // List of best nu for each slice
    
    // now we find the best neutrino for each slice.  
    for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
    {
      
      float maxValue =  0;  // we will use slicer efficiency 
      int bestNu   = -1;
      for(size_t iNu = 0; iNu < sliceTruthTable[iSlc].size(); ++iNu)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].efficiency > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].efficiency;
            bestNu = iNu;
          }
        }
      }
      if(bestSlices[bestNu] == (int)iSlc)
      {
        faveIds[iSlc] = bestNu;
      }
    }
    
    
    
    return faveIds;
    
  }   

} // end namespace 
////////////////////////////////////////////////////////////////////////
namespace earms
{
  DEFINE_ART_MODULE(ElasticArmsValidate);
}
////////////////////////////////////////////////////////////////////////

