//////////////////////////////////////////////////////////////////////////////
/// \file    SimulationDrawer.cxx
/// \brief   Render the objects from the Simulation package
/// \author  messier@indiana.edu
/// \version $Id: SimulationDrawer.cxx,v 1.33 2012-12-05 22:14:05 bckhouse Exp $
/////////////////////////////////////////////////////////////////////////////

#include "EventDisplay/SimulationDrawer.h"

#include "TDatabasePDG.h"
#include "TLatex.h"
#include "TLine.h"
#include "TParticle.h"
#include "TPolyMarker.h"
#include "TPolyMarker3D.h"
#include "TPolyLine.h"
#include "TPolyLine3D.h"

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "EventDisplay/SimulationDrawingOptions.h"
#include "EventDisplay/Style.h"
#include "EventDisplayBase/Colors.h"
#include "EventDisplayBase/View2D.h"
#include "EventDisplayBase/View3D.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/OfflineChan.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "MCCheater/BackTracker.h"

namespace {
  // Utility function to make uniform error messages.
  void writeErrMsg(const char* fcn,
		   cet::exception const& e)
  {
    mf::LogVerbatim("SimulationDrawer")<< "RawDataDrawer::" << fcn
	      << " failed with message:\n"
	      << e;
  }
}

namespace evd
{
  //..................................................................
  SimulationDrawer::SimulationDrawer() 
  {
  }

  //......................................................................

  SimulationDrawer::~SimulationDrawer()
  {
  }

  //......................................................................

  void SimulationDrawer::MCTruthShortText(const art::Event& evt,
					  evdb::View2D*     view) 
  {
    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;
    if ( (drawopt->fText & 
	  evd::SimulationDrawingOptions::kTEXT_SHORT) == 0) {
      return;
    }

    art::ServiceHandle<evdb::Colors> colors;

    std::vector<simb::MCTruth> mctruth;
    this->GetMCTruth(evt, mctruth);

    // No way the user is going to be able to read these anyway, and it takes
    // forever to draw thousands of TLatexs
    if(mctruth.size() > 10){
      TLatex& latex = view->AddLatex(0.03, 0.2, "Too many MCTruths...");
      latex.SetTextColor(colors->Foreground(0));
      latex.SetTextSize(0.6);
      return;
    }

    for (unsigned int i=0; i<mctruth.size(); ++i) {
      std::string mctext;
      bool firstin  = true;
      bool firstout = true;
      std::string origin;
      std::string incoming;
      std::string outgoing;
      // Label cosmic rays -- others are pretty obvious
      if (mctruth[i].Origin()==simb::kCosmicRay) { origin = "c-ray: "; }
      for (int j=0; j<mctruth[i].NParticles(); ++j) {
	const simb::MCParticle& p = mctruth[i].GetParticle(j);
	char buff[1024];
	if (p.P()>0.05) {
	  sprintf(buff,"#color[%d]{%s #scale[0.75]{[%.1f GeV/c]}}", 
		  Style::ColorFromPDG(p.PdgCode()),
		  Style::LatexName(p.PdgCode()), 
		  p.P());
	}
	else {
	  sprintf(buff,"#color[%d]{%s}", 
		  Style::ColorFromPDG(p.PdgCode()),
		  Style::LatexName(p.PdgCode()));
	}
	if (p.StatusCode()==0) {
	  if (firstin==false) incoming += " + ";
	  incoming += buff;
	  firstin = false;
	}
	if (p.StatusCode()==1) {
	  if (firstout==false) outgoing += " + ";
	  outgoing += buff;
	  firstout = false;
	}
      } // loop on j particles
      if (origin=="" && incoming=="") {
	mctext = outgoing;
      }
      else {
	const int intType = mctruth[i].GetNeutrino().InteractionType();
	const std::string suffix = ShortInteractionSuffix(intType);
	mctext = origin+incoming+" #rightarrow "+outgoing+suffix;
      }
      TLatex& latex = view->AddLatex(0.03, 0.2, mctext.c_str());
      latex.SetTextColor(colors->Foreground(0));
      latex.SetTextSize(0.6);
    } // loop on i mctruth objects
  }

  //......................................................................

  // 'prefix' is what we need to add to our children in addition to what we draw
  void SimulationDrawer::PrintParticleAndOffspring(std::stringstream& ss,
                                                   const sim::Particle* part,
						   std::string prefix,
						   int depth) const
  {
    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;
    const int depthLimit = drawopt->fTextDepthLimit;
    const bool showVtxs  = drawopt->fTextIncludeVertex;

    art::ServiceHandle<cheat::BackTracker> bt;
    const sim::ParticleNavigator& pnav = bt->ParticleNavigator();

    // Find out what the particle is called and print it. Fallback is PDG code
    const TDatabasePDG* db = TDatabasePDG::Instance();
    const TParticlePDG* defn = db->GetParticle(part->PdgCode());
    if(defn){
      ss << defn->GetName();
    }
    else{
      switch(part->PdgCode()){
      case 1000010020: ss << "H2"; break;
      case 1000010030: ss << "H3"; break;
      case 1000020030: ss << "He3"; break;
      case 1000020040: ss << "He4"; break;
      case 1000040070: ss << "Be7"; break;
      case 1000040080: ss << "Be8"; break;
      case 1000040100: ss << "Be10"; break;
      case 1000050090: ss << "B9"; break;
      case 1000050100: ss << "B10"; break;
      case 1000050110: ss << "B11"; break;
      case 1000060100: ss << "C10"; break;
      case 1000060110: ss << "C11"; break;
      case 1000060120: ss << "C12"; break;
      case 1000060130: ss << "C13"; break;
      case 1000060140: ss << "C14"; break;
      case 1000070150: ss << "N15"; break;
      case 1000100220: ss << "Ne22"; break;
      case 1000140300: ss << "Si30"; break;
      case 1000150330: ss << "P33"; break;
      case 1000160340: ss << "S34"; break;
      case 1000160350: ss << "S35"; break;
      case 1000160360: ss << "S36"; break;
      case 1000170350: ss << "Cl35"; break;
      case 1000170360: ss << "Cl36"; break;
      case 1000170370: ss << "Cl37"; break;
      case 1000180380: ss << "Ar38"; break;
      case 1000210480: ss << "Sc48"; break;
      case 1000260560: ss << "Fe56"; break;
      case 1000220480: ss << "Ti48"; break;
      case 1000080160: ss << "O16"; break;
      case 1000070140: ss << "N14"; break;
      case 1000110230: ss << "Na23"; break;
      case 1000130270: ss << "Al27"; break;
      case 1000140280: ss << "Si28"; break;
      case 1000200400: ss << "Ca40"; break;
      case 1000561370: ss << "Ba137"; break;
      default:
        ss << part->PdgCode();
      }
    }

    // Print the kinetic energy, and the process that created this particle
    const double ke = part->E()-part->Mass();
    char Estr[1024];
    if(ke > 1)
      sprintf(Estr, "%.1lf GeV", ke);
    else if(ke > 1e-3)
      sprintf(Estr, "%.1lf MeV", 1e3*ke);
    else
      sprintf(Estr, "%.1lf keV", 1e6*ke);

    ss << "\t(" << Estr << ")\t[" << part->Process() << "]";
    if(showVtxs){
      ss << " at (" << part->Vx() << ", " << part->Vy() << ", " << part->Vz() << ")";
    }
    ss << std::endl;

    // Reached depth limit, shouldn't print daughters
    if(depth >= depthLimit && depthLimit > 0) return;

    // Unicode "light" box drawing characters
    const char* kHoriz  = "\u2500";
    const char* kVert   = "\u2502";
    const char* kCorner = "\u2514";
    const char* kTee    = "\u251c";

    // Print our daughters too, indented
    const int N = part->NumberDaughters();
    for(int n = 0; n < N; ++n){
      sim::ParticleNavigator::const_iterator childIt =  pnav.find(part->Daughter(n));
    
      // Should always be there because we created the daughter links ourselves
      // but better safe than sorry...
      if(childIt == pnav.end()) continue;

      if(n == N-1){
	ss << prefix << kCorner << kHoriz << kHoriz;
	PrintParticleAndOffspring(ss, childIt->second,
				  prefix+"     ", depth+1);
      }
      else{
	ss << prefix << kTee << kHoriz << kHoriz;
	PrintParticleAndOffspring(ss, childIt->second,
				  prefix+kVert+"  ", depth+1);
      }
    }

  }

  //......................................................................

  void SimulationDrawer::MCTruthLongText(const art::Event& evt,
					 evdb::View2D* /*view*/) 
  {
    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;
    // Skip drawing if option is turned off
    if ((drawopt->fText & 
	evd::SimulationDrawingOptions::kTEXT_LONG)==0) {
      return;
    }

    art::ServiceHandle<cheat::BackTracker> bt;
    const sim::ParticleNavigator& pnav = bt->ParticleNavigator();

    std::stringstream ss;
    // loop over all the primary particles
    for(int n = 0; n < pnav.NumberOfPrimaries(); ++n){
      ss << "\u2500\u2500";
      PrintParticleAndOffspring(ss, pnav.Primary(n), "  ", 1);
    }
    mf::LogVerbatim("SimulationDrawer") << ss.str();
  }

  //......................................................................

  void SimulationDrawer::MCTruthVectors2D(const art::Event& evt,
					  evdb::View2D*     xzview,
					  evdb::View2D*     yzview) 
  {
    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;
    if ( (drawopt->fDraw & 
	  evd::SimulationDrawingOptions::kDRAW_VECTORS)==0) {
      return;
    }

    // Unpack and draw the MC vectors
    std::vector<simb::MCTruth> mctruth;
    this->GetMCTruth(evt, mctruth);

    for (unsigned int i=0; i<mctruth.size(); ++i) {
      for (int j=0; j<mctruth[i].NParticles(); ++j) {
	const simb::MCParticle& p = mctruth[i].GetParticle(j);
      
	// Skip all but incoming and out-going particles
	if (!(p.StatusCode()==0 || p.StatusCode()==1)) continue;
      
	double r  = p.P()*100.0;          // Scale length so 1 m = 1 GeV/c
	if (r<0.1) continue;              // Skip very short particles
	if (p.StatusCode()==0) r = -r;    // Flip for incoming particles
      
	double x0 = p.Vx();
	double y0 = p.Vy();
	double z0 = p.Vz();
	double x1 = x0 + r * p.Px()/p.P();
	double y1 = y0 + r * p.Py()/p.P();
	double z1 = z0 + r * p.Pz()/p.P();
	if (xzview!=0) {
	  TLine& l = xzview->AddLine(z0, x0, z1, x1);
	  evd::Style::FromPDG(&l,p.PdgCode());
	}
	if (yzview!=0) {
	  TLine& l = yzview->AddLine(z0, y0, z1, y1);
	  evd::Style::FromPDG(&l,p.PdgCode());
	}
      } // loop on j particles
    } // loop on i mctruth objects
  }

  //......................................................................

  void SimulationDrawer::MCTruthVectors3D(const art::Event& evt,
					  evdb::View3D*     view)
  {
    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;
    if ( (drawopt->fDraw &
	  evd::SimulationDrawingOptions::kDRAW_VECTORS)==0) {
      return;
    }
    
    // Unpack and draw the MC vectors
    std::vector<simb::MCTruth> mctruth;
    this->GetMCTruth(evt, mctruth);
    
    for (unsigned int i=0; i<mctruth.size(); ++i) {
      for (int j=0; j<mctruth[i].NParticles(); ++j) {
	const simb::MCParticle& p = mctruth[i].GetParticle(j);
	
	// Skip all but incoming and out-going particles
	if (!(p.StatusCode()==0 || p.StatusCode()==1)) continue;
      
	double r  = p.P()*100.0;          // Scale length so 1 m = 1 GeV/c
	if (r<0.1) continue;              // Skip very short particles
	if (p.StatusCode()==0) r = -r; // Flip for incoming particles
      
	double x0 = p.Vx();
	double y0 = p.Vy();
	double z0 = p.Vz();
	double x1 = x0 + r * p.Px()/p.P();
	double y1 = y0 + r * p.Py()/p.P();
	double z1 = z0 + r * p.Pz()/p.P();
      
	int c = Style::ColorFromPDG(p.PdgCode());
	int s = Style::LineStyleFromPDG(p.PdgCode());
	int w = Style::LineWidthFromPDG(p.PdgCode());
	TPolyLine3D& l = view->AddPolyLine3D(2,c,w,s);
	l.SetPoint(0, x0, y0, z0);
	l.SetPoint(1, x1, y1, z1);
      } // loop on j particles
    } // loop on i mctruth objects
  }

  //......................................................................

  void SimulationDrawer::MCTruthTrajectoriesAnyD(const art::Event& evt,
						 evdb::View2D*     xzview,
						 evdb::View2D*     yzview,
						 evdb::View3D*     view3D)
  {
    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;
    if ( (drawopt->fDraw &
	  evd::SimulationDrawingOptions::kDRAW_TRAJECT)==0) {
      return;
    }
    
    // TODO: Seems this call is required for GetParticleList to work properly?
    std::vector<simb::MCTruth> mctruth;
    this->GetMCTruth(evt, mctruth);

    art::ServiceHandle<cheat::BackTracker> bt;
    const sim::ParticleNavigator& pnav = bt->ParticleNavigator();

    for(sim::ParticleNavigator::const_iterator j=pnav.begin(); 
	j != pnav.end();
	++j) {
      const sim::Particle part = *((*j).second);
      
      const int pdg = part.PdgCode();
      
      // Decay neutrinos are just messy
      if (abs(pdg) == 12 || abs(pdg) == 14 || abs(pdg) == 16) continue;
      
      const int N = part.NumberTrajectoryPoints();
      const int c = evd::Style::ColorFromPDG(pdg);
      const int w = evd::Style::LineWidthFromPDG(pdg);
      const int s = evd::Style::LineStyleFromPDG(pdg);
      
      // This drawing mode disables neutral particles
      bool draw_neutrals = (drawopt->fDraw &
			    evd::SimulationDrawingOptions::kDRAW_NEUTRALS);
      if (!draw_neutrals && s!=kSolid) continue;
      
      TPolyLine* xz = xzview ? &xzview->AddPolyLine(N, c, w, s) : 0;
      TPolyLine* yz = yzview ? &yzview->AddPolyLine(N, c, w, s) : 0;
      TPolyLine3D* xyz = view3D ? &view3D->AddPolyLine3D(N, c, w, s) : 0;
      
      for(int n = 0; n < N; ++n){
	const TLorentzVector& vec = part.Position(n);
	if(xz) xz->SetPoint(n, vec.Z(), vec.X());
	if(yz) yz->SetPoint(n, vec.Z(), vec.Y());
	if(xyz) xyz->SetPoint(n, vec.X(), vec.Y(), vec.Z());
      }
      
      if(xz) xz->SetBit(kCannotPick);
      if(yz) yz->SetBit(kCannotPick);
      if(xyz) xyz->SetBit(kCannotPick);
    } // end for j

  }

  void SimulationDrawer::MCTruthTrajectories2D(const art::Event& evt,
					       evdb::View2D*     xzview,
					       evdb::View2D*     yzview)
  {
    MCTruthTrajectoriesAnyD(evt, xzview, yzview, 0);
  }

  void SimulationDrawer::MCTruthTrajectories3D(const art::Event& evt,
					       evdb::View3D*     view)
  {
    MCTruthTrajectoriesAnyD(evt, 0, 0, view);
  }

  //......................................................................

  void SimulationDrawer::FLSHit2D(const art::Event& evt,
				  evdb::View2D*     xzview,
				  evdb::View2D*     yzview)
  {
    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;
    if ( (drawopt->fDraw & 
	  evd::SimulationDrawingOptions::kDRAW_HITS)==0) {
      return;
    }

    art::ServiceHandle<geo::Geometry> geo;

    // Otherwise, unpack and show the FLS hits
    for (unsigned int imod=0; imod<drawopt->fFLSHitListModules.size(); ++imod) {
      const char* which = drawopt->fFLSHitListModules[imod].c_str();
    
    std::vector<sim::FLSHitList> flshits;
    this->GetFLSHits(evt, which, flshits);

    // Indexed by color
    std::map<int, TPolyMarker*> pms_x, pms_y;

    for (unsigned int i=0; i<flshits.size(); ++i) {
      for (unsigned int j=0; j<flshits[i].fHits.size(); ++j) {
	const sim::FLSHit& hit = flshits[i].fHits[j];
	if (hit.GetEdep()<0.001*drawopt->fFLSHitThresh) continue;

	switch (drawopt->fFLSHitStyle) {
	case evd::SimulationDrawingOptions::kFLSHIT_AS_LINE:
          for(int view = 0; view < 2; ++view){
            evdb::View2D* v2d = (view == 0) ? xzview : yzview;
            if(!v2d) continue;

            // Dummy styling, overridden in next line
            TPolyLine& l = v2d->AddPolyLine(hit.GetNPoints(), kBlack, 0, kSolid);
            evd::Style::FromPDG(&l, hit.GetPDG());
            for(int i = 0; i < hit.GetNPoints(); ++i){
              double xyzloc[3] = {hit.GetX(i), hit.GetY(i), hit.GetZ(i)};
              double xyzwor[3];
              geo->Plane(hit.GetPlaneID())->Cell(hit.GetCellID())->LocalToWorld(xyzloc, xyzwor);
              l.SetPoint(i, xyzwor[2], xyzwor[view]);
            }
          }
	  break;
	default:
          double xyzloc[3] = {hit.GetXAverage(), hit.GetYAverage(), hit.GetZAverage()};
          double xyzwor[3];
          geo->Plane(hit.GetPlaneID())->Cell(hit.GetCellID())->LocalToWorld(xyzloc, xyzwor);

	  const int c = Style::ColorFromPDG(hit.GetPDG());
	  const double sz = 0.5;
	  
	  if(xzview){
	    TPolyMarker* pm = pms_x[c];
	    if(!pm){
	      pm = &xzview->AddPolyMarker(0, c, kFullDotMedium, sz);
	      pms_x[c] = pm;
	    }
	    pm->SetNextPoint(xyzwor[2], xyzwor[0]);
	  }
	  if(yzview){
	    TPolyMarker* pm = pms_y[c];
	    if(!pm){
	      pm = &yzview->AddPolyMarker(0, c, kFullDotMedium, sz);
	      pms_y[c] = pm;
	    }
	    pm->SetNextPoint(xyzwor[2], xyzwor[1]);
	  }
	  
	  if(fHighlite[hit.GetTrackID()]){
	    if(xzview) xzview->AddMarker(xyzwor[2],xyzwor[0],c,kOpenCircle,sz);
	    if(yzview) yzview->AddMarker(xyzwor[2],xyzwor[1],c,kOpenCircle,sz);
	  }

	} // FLSHit style
      } // loop on j hits in list
    } // loop on i hit lists
  } // loop o imod folders
}

  //......................................................................

  void SimulationDrawer::FLSHit3D(const art::Event& evt,
				  evdb::View3D*     view)
  {
    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;
    if ( (drawopt->fDraw &
	  evd::SimulationDrawingOptions::kDRAW_HITS)==0) {
      return;
    }

    art::ServiceHandle<geo::Geometry> geo;

    // Otherwise, unpack and show the FLS hits
    for (unsigned int imod=0; imod<drawopt->fFLSHitListModules.size(); ++imod) {
      const char* which = drawopt->fFLSHitListModules[imod].c_str();
    
      std::vector<sim::FLSHitList> flshits;
      this->GetFLSHits(evt, which, flshits);
      for (unsigned int i=0; i<flshits.size(); ++i) {
	for (unsigned int j=0; j<flshits[i].fHits.size(); ++j) {
	  const sim::FLSHit& hit = flshits[i].fHits[j];

	  if (hit.GetEdep()<0.001*drawopt->fFLSHitThresh) continue;

	  double xyzloc[3]; // Local cell xyz coordinates
	  double xyzwor[3]; // World coordinates
	  xyzloc[0] = 0.5*(hit.GetEntryX() + hit.GetExitX());
	  xyzloc[1] = 0.5*(hit.GetEntryY() + hit.GetExitY());
	  xyzloc[2] = 0.5*(hit.GetEntryZ() + hit.GetExitZ());
	
	  geo->Plane(hit.GetPlaneID())->Cell(hit.GetCellID())->LocalToWorld(xyzloc,xyzwor);
	
	  int   st = 7;
	  int   c  = Style::ColorFromPDG(hit.GetPDG());
	  float sz = 0.5;

	  TPolyMarker3D& p = view->AddPolyMarker3D(1, c, st, sz);
	  p.SetPoint(0, xyzwor[0], xyzwor[1], xyzwor[2]);

	  if (fHighlite[hit.GetTrackID()]) {
	    TPolyMarker3D& p = view->AddPolyMarker3D(1, c, 24, 2*sz);
	    p.SetPoint(0, xyzwor[0], xyzwor[1], xyzwor[2]);	  
	  }

	} // loop on j hits in list
      } // loop on i hit lists
    } // loop o imod folders
  }


  //......................................................................

  int SimulationDrawer::GetMCTruth(const art::Event& evt,
				   std::vector<simb::MCTruth>& mctruth) 
  {
    mctruth.clear();

    // going to use getManyByType so that we don't have to 
    // worry about using a module label
    std::vector< art::Handle< std::vector<simb::MCTruth> > > tcol;

    try{    
      evt.getManyByType(tcol);
      for(size_t mctc = 0; mctc < tcol.size(); ++mctc){
	for(size_t mct = 0; mct < tcol[mctc]->size(); ++mct){
	  mctruth.push_back(tcol[mctc]->at(mct));
	}
      }
    }// end try
    catch(cet::exception& e){
      writeErrMsg("GetMCTruth", e);
    }

    return mctruth.size();
  }

  //......................................................................

  int SimulationDrawer::GetFLSHits(const art::Event& evt,
				   const char* which,
				   std::vector<sim::FLSHitList>& flshits) 
  {

    flshits.clear();
    art::Handle< std::vector<sim::FLSHitList> > fcol;

    try{
      evt.getByLabel(which, fcol);
      std::vector<sim::FLSHitList> temp(fcol->size());
      for(size_t f = 0; f < fcol->size(); ++f){
	flshits.push_back(fcol->at(f));
      }
    }
    catch(cet::exception& e){
      writeErrMsg("GetFLSHits", e);
    }

    return flshits.size();
  }

  //......................................................................

  std::string SimulationDrawer::ShortInteractionSuffix(int iType) const
  {
    switch(iType){
    case simb::kQE:
    case simb::kCCQE:
      return " (QE)";

    case simb::kRes: 
    case simb::kResCCNuProtonPiPlus:
    case simb::kResCCNuNeutronPi0:
    case simb::kResCCNuNeutronPiPlus:
    case simb::kResCCNuBarNeutronPiMinus:
    case simb::kResCCNuBarProtonPi0:
    case simb::kResCCNuBarProtonPiMinus:
      return " (RES)";

    case simb::kDIS: 
    case simb::kCCDIS:
      return " (DIS)";

    case simb::kCoh: 
    case simb::kCCCOH:
      return " (COH)";

    case simb::kNCQE:
    case simb::kNCDIS:  
    case simb::kResNCNuProtonPi0:
    case simb::kResNCNuProtonPiPlus:
    case simb::kResNCNuNeutronPi0:
    case simb::kResNCNuNeutronPiMinus:
    case simb::kResNCNuBarProtonPi0:
    case simb::kResNCNuBarProtonPiPlus:
    case simb::kResNCNuBarNeutronPi0:
    case simb::kResNCNuBarNeutronPiMinus:
      return " (NC)";

    default: // simb::kNuElectronElastic and simb::kInverseMuDecay
      return "";
    }
  }

  //......................................................................

  void SimulationDrawer::HiLite(int trkId, bool dohilite)
  {
    fHighlite[trkId] = dohilite;
  }

  //......................................................................

  void SimulationDrawer::GetLimits(const art::Event* evt,
                                   double& xmin, double& xmax,
                                   double& ymin, double& ymax,
                                   double& zmin, double& zmax,
                                   const std::set<geo::OfflineChan>& hmap)
  {
    assert(evt);

    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;

    art::ServiceHandle<geo::Geometry> geom;

    xmin = +1e10; ymin = +1e10; zmin = +1e10;
    xmax = -1e10; ymax = -1e10; zmax = -1e10;

    for(unsigned int imod = 0; imod < drawopt->fFLSHitListModules.size(); ++imod){
      std::vector<sim::FLSHitList> lists;
      GetFLSHits(*evt, drawopt->fFLSHitListModules[imod].c_str(), lists);

      for(unsigned int i = 0; i < lists.size(); ++i){
        const std::vector<sim::FLSHit>& hits = lists[i].fHits;

        for(unsigned int j = 0; j < hits.size(); ++j){
          const sim::FLSHit& hit = hits[j];

          if(!hmap.empty() &&
             hmap.find(geo::OfflineChan(hit.GetPlaneID(),
                                        hit.GetCellID())) == hmap.end()) continue;

          geo::View_t view;
          double xyz[3];
          geom->CellInfo(hit.GetPlaneID(), hit.GetCellID(), &view, xyz);

          if(view == geo::kX && xyz[0] < xmin) xmin = xyz[0];
          if(view == geo::kX && xyz[0] > xmax) xmax = xyz[0];
          if(view == geo::kY && xyz[1] < ymin) ymin = xyz[1];
          if(view == geo::kY && xyz[1] > ymax) ymax = xyz[1];
          if(xyz[2] < zmin) zmin = xyz[2];
          if(xyz[2] > zmax) zmax = xyz[2];
        }
      }
    } // end for imod
  }

  //......................................................................

  void SimulationDrawer::GetTimeLimits(const art::Event* evt,
                                       double& tmin, double& tmax)
  {
    tmin = FLT_MAX;
    tmax = -FLT_MAX;

    assert(evt);

    art::ServiceHandle<evd::SimulationDrawingOptions> drawopt;

    for(unsigned int imod = 0; imod < drawopt->fFLSHitListModules.size(); ++imod){
      std::vector<sim::FLSHitList> lists;
      GetFLSHits(*evt, drawopt->fFLSHitListModules[imod].c_str(), lists);

      for(unsigned int i = 0; i < lists.size(); ++i){
        const std::vector<sim::FLSHit>& hits = lists[i].fHits;

        for(unsigned int j = 0; j < hits.size(); ++j){
          const sim::FLSHit& hit = hits[j];

          if(hit.GetEntryT() < tmin) tmin = hit.GetEntryT();
          if(hit.GetExitT()  > tmax) tmax = hit.GetExitT();
        }
      }
    }
  }

} //end namespace evd
////////////////////////////////////////////////////////////////////////
