////////////////////////////////////////////////////////////////////////
// \file   SliceNavigator_service.cc
// \brief  Look up products dependent on the current slice
// \author Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "EventDisplay/SliceNavigator.h"

#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"

#include "EventDisplayBase/NavState.h"
#include "EventDisplayBase/ServiceTable.h"

#include "Utilities/AssociationUtil.h"

#include "GeometryObjects/OfflineChan.h"
#include "Geometry/Geometry.h"
#include "MichelEFilter/MichelECluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/HoughResult.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "RecoBase/FilterList.h"

namespace evd
{
  //---------------------------------------------------------------
  SliceNavigator::SliceNavigator(const fhicl::ParameterSet& pset,
                                 art::ActivityRegistry& reg)
    : fFullEvent(true), fCurrentSlice(0), fMaxSliceId(0), fInitialSlice(-1)
  {
    reconfigure(pset);

    evdb::ServiceTable::OverrideCategory("SliceNavigator",
                                         evdb::kDRAWING_SERVICE);

    reg.sPostProcessEvent.watch(this, &SliceNavigator::postProcessEvent);
  }

  //---------------------------------------------------------------
  SliceNavigator::~SliceNavigator()
  {
    fSlicesToNavigate.clear();
  }

  //---------------------------------------------------------------

  void SliceNavigator::reconfigure(fhicl::ParameterSet const& pset)
  {
    fSliceLabel    = pset.get<std::string>("SliceLabel");
    fFilterResults = (pset.get<int>("SliceOpt.val") & 1);
    fDimHits       = (pset.get<int>("SliceOpt.val") & 2);
    fDimOthers     = (pset.get<int>("SliceOpt.val") & 4);

    fFullEvent = !pset.get<bool>("InitiallyEnabled");
    fInitialSlice = pset.get<int>("InitialSlice");
  }

  //---------------------------------------------------------------

  void SliceNavigator::postProcessEvent(const art::Event& evt)
  {
    // Clear the slices to navigate vector
    fSlicesToNavigate.clear();

    art::Handle<std::vector<rb::Cluster>> slices;
    evt.getByLabel( fSliceLabel, slices);

    if(slices.failedToGet()){
      std::cerr << "Couldn't find any slices under label "
                << fSliceLabel << std::endl;
      fMaxSliceId = 0;
      return;
    }

    int nSli = slices->size();
    for( int iSli = 0; iSli < nSli; ++iSli){
      // if we only want to look at slices that
      // survived filtering, throw the rest away
      if(!fFullEvent && fFilterResults && rb::IsFiltered(evt,slices,iSli) )
	continue;
      if(!fFullEvent && (*slices)[iSli].IsNoise())
	continue;

      fSlicesToNavigate.push_back( iSli );
      if(iSli == fInitialSlice){
        fCurrentSlice = fSlicesToNavigate.size()-1;
        fInitialSlice = -1; // Don't do this again
      }
    }
    if (fSlicesToNavigate.size() > 0) fMaxSliceId = fSlicesToNavigate.back();

    // Protecting against segfaults:  If we've advanced to the next event
    // with fCurrentSlice set to the max number of slices, and the next event
    // has fewer slices than the previous, we need to make sure we don't 
    // suffer from an index out of range problem.
    if( (fCurrentSlice + 1 > int(fSlicesToNavigate.size())) && !fSlicesToNavigate.empty() ) 
      fCurrentSlice = fSlicesToNavigate.size() - 1;

    if(fCurrentSlice == -1){
      if(!fSlicesToNavigate.empty())
    	// We pressed the prev slice button
    	fCurrentSlice = fSlicesToNavigate.size()-1;
      else
    	evdb::NavState::Set(evdb::kPREV_EVENT);
    }

    if(fCurrentSlice == -2 ){
      if(!fSlicesToNavigate.empty())
    	// We pressed the next button, but there
    	// no more slices left in the previous event
    	fCurrentSlice = 0;
      else
    	evdb::NavState::Set(evdb::kNEXT_EVENT);
    }

  }

  //---------------------------------------------------------------
  template<class T> size_t SliceNavigator::
  GetProducts(const art::Event&      evt,
              const std::string&     which,
              const std::string&     instance,
              std::vector<const T*>& prods)
  {
    return GetProductsImpl(evt, which, instance, prods);
  }

  //---------------------------------------------------------------
  template<class T> size_t SliceNavigator::
  GetProductsImpl(const art::Event&      evt,
                  const std::string&     which,
                  const std::string&     instance,
                  std::vector<const T*>& prods)
  {
    if(fFullEvent)
      return GetProductsImplFull(evt, which, instance, prods);
    else
      return GetProductsImplPart(evt, which, instance, prods);
  }


  //---------------------------------------------------------------
  // Need a special function for clusts again
  // because cluster-cluster assns are not allowed.
  // See generic GetProductsImplFull for the story

  template<> size_t SliceNavigator::
  GetProductsImplFull(const art::Event&      evt,
                      const std::string&     which,
                      const std::string&     instance,
                      std::vector<const rb::Cluster*>& clusts)
  {
    std::vector<const rb::Cluster*> temp(clusts);
    try{
      evt.getView(which, instance, temp);
      temp.swap(clusts);
    }
    catch(cet::exception& e){
      std::cerr << "Failed to get products of type " << typeid(rb::Cluster).name()
                << " under label \"" << which << "\": "
                << e << std::endl;
    }
    return clusts.size();
  }

  //---------------------------------------------------------------

  template<class T> size_t SliceNavigator::
  GetProductsImplFull(const art::Event&      evt,
                      const std::string&     which,
                      const std::string&     instance,
                      std::vector<const T*>& prods)
  {

    art::Handle<std::vector<T> > tempHandle;

    std::vector<art::Ptr<T> > temp;

    try{
      evt.getByLabel(art::InputTag(which, instance), tempHandle);
      art::fill_ptr_vector( temp,tempHandle);
    }
    catch(cet::exception& e){
      std::cerr << "Failed to get products of type " << typeid(T).name()
                << " under label \"" << which << "\": "
                << e << std::endl;
    }

    std::vector<art::Ptr<T> > tempcopy;
    // If fFilterResults is turned on, we don't want to
    // draw anything from a slice that was filtered out
    // So each product has to be mapped to its associated
    // slice cluster and check if the slice was filtered out.
    if( fFilterResults && typeid(T) != typeid(rb::CellHit)){

      art::Handle<std::vector<rb::Cluster> > slicesHandle;
      evt.getByLabel( fSliceLabel, slicesHandle);
      std::vector<art::Ptr<rb::Cluster> >    slices;
      art::fill_ptr_vector( slices, slicesHandle);
      art::FindOneP<rb::Cluster> fot(temp, evt, which);

      if(!fot.isValid()){
	std::cerr << "Nothing of type " << typeid(T).name()
		  << " with label \"" << which
		  << "\" associated with slices" << std::endl;
	return 0;
      }

      int nT = temp.size();
      for( int i = 0; i < nT; i++){

	art::Ptr<rb::Cluster> slice = fot.at( i );
	std::vector< art::Ptr< rb::Cluster> >::iterator slicePos =
	  std::find(slices.begin(), slices.end(), slice) ;
	int iSlice = std::distance( slices.begin(), slicePos );

       	if( fFilterResults && rb::IsFiltered(evt,slicesHandle, iSlice ) )
	  continue;
	tempcopy.push_back( temp[i] );

      }// end loop over temp
    }
    else
      tempcopy.swap( temp );

    for( art::Ptr<T> p: tempcopy)
      prods.push_back( p.get());

    return prods.size();
  }


  //---------------------------------------------------------------
  template<class T> size_t SliceNavigator::
  GetProductsImplPart(const art::Event&      evt,
                      const std::string&     which,
                      const std::string&     instance,
                      std::vector<const T*>& prods)
  {
    art::Handle<std::vector<rb::Cluster>> slices;
    evt.getByLabel(fSliceLabel, slices);

    if(slices.failedToGet()){
      std::cerr << "Couldn't find any slices under label "
                << fSliceLabel << std::endl;
      return 0;
    }

    art::FindManyP<T> fmt(slices, evt, art::InputTag(which, instance));
    if(!fmt.isValid()){
      std::cerr << "Nothing of type " << typeid(T).name()
                << " with label \"" << which
                << "\" associated with slices" << std::endl;
      return 0;
    }

    std::vector<art::Ptr<T>> prodPtrs = fmt.at( fSlicesToNavigate[fCurrentSlice] );
    prods.clear();
    for(art::Ptr<T> p: prodPtrs) prods.push_back(p.get());
    return prods.size();
  }

  //---------------------------------------------------------------
  // Special case for Tracks because there's this MichelECluster trick
  template<> size_t SliceNavigator::
  GetProducts(const art::Event&              evt,
              const std::string&             whichTrack,
	      const std::string&             whichMichel,
              const std::string&             instance,
              std::vector<const rb::Track*>& tracks)
  {
    std::vector<const me::MichelECluster*> clusts;
    GetProductsImpl(evt, whichTrack, instance, tracks);
    GetProductsImpl(evt, whichMichel, instance, clusts);

    for(const me::MichelECluster* c: clusts)
      tracks.push_back(c->ParentTrack().get());

    return tracks.size();
  }

  //---------------------------------------------------------------
  // Special case clusters because they *are* the slice, not just associated
  template<> size_t SliceNavigator::
  GetProducts(const art::Event&                evt,
              const std::string&               which,
              const std::string&               instance,
              std::vector<const rb::Cluster*>& clusts)
  {

    std::vector< const rb::Cluster*> temp;

    // If it's the actual slices, look them up by index.  Otherwise we'll get
    // nothing (not possible to associate Cluster with Cluster)
    if(which == fSliceLabel && instance == ""){
      art::Handle<std::vector<rb::Cluster>> slices;
      evt.getByLabel(fSliceLabel, slices);
      if(slices.failedToGet()){
        std::cerr << "Couldn't find any slices under label "
                  << fSliceLabel << std::endl;
        return 0;
      }
      if( fFullEvent ){
	int nSli = slices->size();
	for( int iSli=0; iSli < nSli; ++iSli){
	  if ( fFilterResults && rb::IsFiltered(evt,slices, iSli ) )
	    continue;
	  clusts.push_back((art::Ptr<rb::Cluster>(slices, iSli).get())) ;
	}
      }// if full event
      else{
	int idx = fSlicesToNavigate[ fCurrentSlice];
	if( fCurrentSlice >=0 )
	  clusts.push_back( (art::Ptr<rb::Cluster>(slices, idx)).get() );
      }

    }
    else
      GetProductsImplFull(evt, which, instance, clusts);

    return clusts.size();
  }

  //---------------------------------------------------------------
  bool SliceNavigator::FullEventMode() const
  {
    return fFullEvent;
  }

  //---------------------------------------------------------------
  void SliceNavigator::SetFullEventMode(bool mode)
  {
    if(mode == fFullEvent) return;

    fFullEvent = mode;
    evdb::NavState::Set(evdb::kRELOAD_EVENT);
  }

  //---------------------------------------------------------------
  void SliceNavigator::Previous()
  {
    --fCurrentSlice;

    if(fCurrentSlice < 0){
      fCurrentSlice = -1;
      evdb::NavState::Set(evdb::kPREV_EVENT);
      fSlicesToNavigate.clear();
    }
    else{
      // Go to the slice requested
      evdb::NavState::Set(evdb::kRELOAD_EVENT);
    }
  }

  //---------------------------------------------------------------
  void SliceNavigator::Next()
  {
    ++fCurrentSlice;

    if(fCurrentSlice >= int(fSlicesToNavigate.size())){
      fCurrentSlice = -2;
      fSlicesToNavigate.clear();
      evdb::NavState::Set(evdb::kNEXT_EVENT);
    }
    else{
      // Go to the slice requested
      evdb::NavState::Set(evdb::kRELOAD_EVENT);
    }
  }

  //---------------------------------------------------------------
  int SliceNavigator::CurrentSlice() const
  {
    if(fSlicesToNavigate.empty() || fCurrentSlice < 0) return 0;

    return fSlicesToNavigate [fCurrentSlice ];
  }

  //---------------------------------------------------------------

  int SliceNavigator::CurrentSliceIndex() const
  {
    return fCurrentSlice;
  }

  //---------------------------------------------------------------
  void SliceNavigator::SetCurrentSlice(int slice)
  {
    auto it = std::find(fSlicesToNavigate.begin(), fSlicesToNavigate.end(), slice);
    if(it != fSlicesToNavigate.end()){
      fCurrentSlice = it-fSlicesToNavigate.begin();
      evdb::NavState::Set(evdb::kRELOAD_EVENT);
    }
    else{
      // What else can we do?
      fCurrentSlice = 0;
      evdb::NavState::Set(evdb::kRELOAD_EVENT);
    }
  }

  //---------------------------------------------------------------
  int SliceNavigator::MaxSlice() const
  {
    return fMaxSliceId;
  }

  //---------------------------------------------------------------
  void SliceNavigator::GetBounds(const art::Event& evt,
                                 double& xmin, double& xmax,
                                 double& ymin, double& ymax,
                                 double& zmin, double& zmax)
  {
    // Fallback to the whole detector in case we don't get values below
    art::ServiceHandle<geo::Geometry> geom;
    xmin = -geom->DetHalfWidth();
    xmax = +geom->DetHalfWidth();
    ymin = -geom->DetHalfHeight();
    ymax = +geom->DetHalfHeight();
    zmin = 0;
    zmax = geom->DetLength();

    art::Handle<std::vector<rb::Cluster>> slices;
    evt.getByLabel( fSliceLabel, slices);

    if(slices.failedToGet()){
      std::cerr << "Couldn't find any slices under label "
                << fSliceLabel << std::endl;
      return;
    }

    if(fCurrentSlice >= 0 && fCurrentSlice < int(slices->size())){
      const rb::Cluster& slice = (*slices)[fCurrentSlice];

      const TVector3 r0 = slice.MinXYZ();
      const TVector3 r1 = slice.MaxXYZ();
      xmin = r0.X(); xmax = r1.X();
      ymin = r0.Y(); ymax = r1.Y();
      zmin = r0.Z(); zmax = r1.Z();
    }
  }

  //---------------------------------------------------------------

  std::vector<const rb::Cluster*> SliceNavigator::SlicesToNavigate( const art::Event& evt)
  {
    std::vector<const rb::Cluster*> clusts;
    art::Handle<std::vector<rb::Cluster>> slices;
    evt.getByLabel(fSliceLabel, slices);
    if(slices.failedToGet()) return {};

    for(unsigned int i = 0; i < fSlicesToNavigate.size(); ++i)
      clusts.push_back(&(*slices)[fSlicesToNavigate[i]]);

    return clusts;
  }// End of SlicesToNavigate

  //---------------------------------------------------------------


  // Need to explicitly instantiate every product that can be got. It's a lot
  // of typing, so use a macro
#define INSTANTIATE(T) template size_t SliceNavigator::GetProducts<T>(const art::Event&, const std::string&, const std::string&, std::vector<const T*>&)

  INSTANTIATE(geo::OfflineChan);
  INSTANTIATE(rb::CellHit);
  INSTANTIATE(rb::HoughResult);
  INSTANTIATE(rb::Prong);
  INSTANTIATE(rb::Vertex);
  INSTANTIATE(rb::Track);

  DEFINE_ART_SERVICE(SliceNavigator);

} // end namespace evd
////////////////////////////////////////////////////////////////////////
