////////////////////////////////////////////////////////////////////////
/// \class genie::flux::GFlavorOsc
/// \brief GENIE interface for flavor modification
///
///        Concrete instance of GFlavorMixerI that maps from
///        one flavor to another.
///        Supported config string formats:
///        1)  "[GFlavorOsc] blah blah blah "
///
/// physics.producers.generator.MixerConfig: "evgen::GFlavorOsc blah ..."
///
/// \author  Robert Hatcher <rhatcher \at fnal.gov>
///          Fermi National Accelerator Laboratory
///
/// \date    2012-08-03
////////////////////////////////////////////////////////////////////////

#ifndef GENIE_FLUX_GFLAVOROSC_H
#define GENIE_FLUX_GFLAVOROSC_H

#include <string>
#include "FluxDrivers/GFlavorMixerI.h"

namespace evgen {

  class GFlavorOsc : public genie::flux::GFlavorMixerI {
    
  public:
  
    GFlavorOsc();
    ~GFlavorOsc();

    //
    // implement the GFlavorMixerI interface:
    //

    /// each schema must take a string that configures it
    /// it is up to the individual model to parse said string
    /// and extract parameters (e.g. sin2th23, deltam12, etc)
    void      Config(std::string config);

    /// for any pair of PDG codes the model must calculate
    /// the transition probability.  This can also depend on
    /// neutrino energy (in GeV) and distance (in meters) from 
    /// the neutrino origin.
    double    Probability(int pdg_initial, int pdg_final, 
                          double energy, double dist);

    /// provide a means of printing the configuration
    void     PrintConfig(bool verbose=true);

  private:

    void         ParseConfig(std::string config);
    void         UnityProbMatrix();

    int          PDG2Indx(int pdg);
    int          Indx2PDG(int indx);
    const char*  IndxName(int indx);
    const char*  NuName(int pdg) { return IndxName(PDG2Indx(pdg)); }

    static const int kNFlavors = 7;

    double fLastEnergy;
    double fLastDist;
    double fProbMatrix[kNFlavors][kNFlavors];

  };

} // namespace evgen

//
//    Name        PDG   Indx
//    sterile       0   0
//    nu_e         12   1
//    nu_mu        14   2
//    nu_tau       16   3
//    nu_e_bar    -12   4
//    nu_mu_bar   -14   5
//    nu_tau_bar  -16   6
//
inline int evgen::GFlavorOsc::PDG2Indx(int pdg)
{
  switch ( pdg ) {
  case  12: return 1; break;
  case  14: return 2; break;
  case  16: return 3; break;
  case -12: return 4; break;
  case -14: return 5; break;
  case -16: return 6; break;
  default:  return 0; break;
  }
  return 0;
}
inline int evgen::GFlavorOsc::Indx2PDG(int indx)
{
  switch ( indx ) {
  case  1: return  12; break;
  case  2: return  14; break;
  case  3: return  16; break;
  case  4: return -12; break;
  case  5: return -14; break;
  case  6: return -16; break;
  default: return   0; break;
  }
  return 0;
}

#endif //GENIE_FLUX_GFLAVORSWAP_H
