////////////////////////////////////////////////////////////////////////
/// \brief  Use NuReweight to compute +/-1,2sigma shifts for all systematics
/// \author Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// GENIE: Macros in here conflict with art's MessageFacility. So include it
// first and then unset them again
#include "Messenger/Messenger.h"
#undef LOG_INFO
#undef LOG_DEBUG
#undef LOG_ERROR

#include <cassert>
#include <string>

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "Geometry/Geometry.h"
#include "NuReweight/art/NuReweight.h"
#include "SimulationBase/GTruth.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/GENIEReweightTable.h"
#include "Utilities/AssociationUtil.h"

namespace rwgt
{
  // It's a shame such a thing isn't officially defined in any of the headers
  const int kMaxReweightIdx = rwgt::fReweightTheta_Delta2Npi;

  /// \brief  Use NuReweight to compute +/-1,2sigma shifts for all systematics
  class MakeGENIEReweightTable : public art::EDProducer
  {
  public:
    explicit MakeGENIEReweightTable(const fhicl::ParameterSet& pset);
    virtual ~MakeGENIEReweightTable();

    void produce(art::Event& evt);
    void beginJob();

  protected:
    std::string fGeneratorLabel;

    bool fConsiderNDRock; ///< Include rock interaction in the ND?
    bool fConsiderFDRock; ///< Include rock interaction in the FD?

    std::vector<rwgt::NuReweight*> fRWsMinus2;
    std::vector<rwgt::NuReweight*> fRWsMinus1;
    std::vector<rwgt::NuReweight*> fRWsPlus1;
    std::vector<rwgt::NuReweight*> fRWsPlus2;
  };

  //___________________________________________________________________________
  MakeGENIEReweightTable::MakeGENIEReweightTable(const fhicl::ParameterSet& pset)
    : fGeneratorLabel(pset.get<std::string>("GeneratorLabel")),
      fConsiderNDRock(pset.get<bool>("ConsiderNDRock")),
      fConsiderFDRock(pset.get<bool>("ConsiderFDRock"))
  {
    produces<std::vector<rwgt::GENIEReweightTable>>();

    produces<art::Assns<rwgt::GENIEReweightTable, simb::MCTruth>>();
  }

  //___________________________________________________________________________
  MakeGENIEReweightTable::~MakeGENIEReweightTable()
  {
    for(rwgt::NuReweight* rw: fRWsMinus2) delete rw;
    for(rwgt::NuReweight* rw: fRWsMinus1) delete rw;
    for(rwgt::NuReweight* rw: fRWsPlus1) delete rw;
    for(rwgt::NuReweight* rw: fRWsPlus2) delete rw;
  }

  //___________________________________________________________________________
  void MakeGENIEReweightTable::beginJob()
  {
    mf::LogInfo("MakeGENIEReweightTable") << "Initializing GENIE reweight calculators...";

    // Don't need to see every tiny thing these modules are thinking...
    genie::Messenger::Instance()->SetPriorityLevel("PDG",
                                                   log4cpp::Priority::WARN);
    genie::Messenger::Instance()->SetPriorityLevel("ReW",
                                                   log4cpp::Priority::WARN);
    genie::Messenger::Instance()->SetPriorityLevel("Spline",
                                                   log4cpp::Priority::WARN);

    // The way NuReweight constructs the EventRecord it inevitably trips this
    genie::Messenger::Instance()->SetPriorityLevel("Target",
                                                   log4cpp::Priority::ERROR);

    genie::Messenger::Instance()->SetPriorityLevel("GENIEReweight",
                                                   log4cpp::Priority::WARN);

    fRWsMinus2.resize(kMaxReweightIdx+1, 0);
    fRWsMinus1.resize(kMaxReweightIdx+1, 0);
    fRWsPlus1. resize(kMaxReweightIdx+1, 0);
    fRWsPlus2. resize(kMaxReweightIdx+1, 0);

    // Start at 1. Zero is kReweightNull. Not worth dealing with shifting all
    // the indices by one everywhere, just leave it empty.
    for(int i = 1; i <= kMaxReweightIdx; ++i){
      for(int sigma: {-2, -1, +1, +2}){
        rwgt::NuReweight* rw = new rwgt::NuReweight;
        switch(sigma){
        case -2: fRWsMinus2[i] = rw; break;
        case -1: fRWsMinus1[i] = rw; break;
        case +1: fRWsPlus1 [i] = rw; break;
        case +2: fRWsPlus2 [i] = rw; break;
        }
	double shift = sigma;
	//Catch the cases where shape parameters are called instead of rate parameters.
	switch(i) {
	case rwgt::fReweightNormCCQE:
	case rwgt::fReweightNormCCQEenu:
	case rwgt::fReweightMaCCQEshape:
          rw->MaQEshape();
          break;

	case rwgt::fReweightNormCCRES:
	case rwgt::fReweightMaCCRESshape:
	case rwgt::fReweightMvCCRESshape:
          rw->CCRESshape();
          break;

	case rwgt::fReweightNormNCRES:
	case rwgt::fReweightMaNCRESshape:
	case rwgt::fReweightMvNCRESshape:
          rw->NCRESshape();
          break;

	case rwgt::fReweightAhtBYshape:
	case rwgt::fReweightBhtBYshape:
	case rwgt::fReweightCV1uBYshape:
	case rwgt::fReweightCV2uBYshape:
          rw->DIS_BYshape();
          break;

	case rwgt::fReweightCCQEMomDistroFGtoSF:
	case rwgt::fReweightTheta_Delta2Npi:
          // These are continuous knobs.
          // Encode -2sigma = .25, -1sigma = .5, +1sigma = .75, +2sigma = 1
	  shift = (shift < 0) ? (shift+3.0)/4.0 : (shift + 2)/4.0;
	  break;

        case rwgt::fReweightRnubarnuCC:
        case rwgt::fReweightDISNuclMod:
          // These are binary switches. Make the negative sigmas be "off" and
          // the positive be "on".
          shift = (shift < 0) ? 0 : 1;
          break;
	}

        rw->AddReweightValue(rwgt::ReweightLabel_t(i), shift);
        rw->Configure();
      }
    }

    mf::LogInfo("MakeGENIEReweightTable") << "Done initializing GENIE reweight calculators...";
  }

  //___________________________________________________________________________
  void MakeGENIEReweightTable::produce(art::Event& evt)
  {
    art::ServiceHandle<geo::Geometry> geom;

    std::unique_ptr<std::vector<rwgt::GENIEReweightTable>> tablecol(new std::vector<rwgt::GENIEReweightTable>);

    std::unique_ptr<art::Assns<rwgt::GENIEReweightTable, simb::MCTruth>> assns(new art::Assns<rwgt::GENIEReweightTable, simb::MCTruth>);


    art::Handle<std::vector<simb::MCTruth>> mctruths;
    evt.getByLabel(fGeneratorLabel, mctruths);
    assert(!mctruths.failedToGet());

    art::FindManyP<simb::GTruth> fmp(mctruths, evt, fGeneratorLabel);
    assert(fmp.isValid());

    for(unsigned int truthIdx = 0; truthIdx < mctruths->size(); ++truthIdx){
      const std::vector<art::Ptr<simb::GTruth>> gtruths = fmp.at(truthIdx);
      assert(gtruths.size() == 1);

      const simb::MCTruth& mctruth = (*mctruths)[truthIdx];
      const simb::GTruth& gtruth = *gtruths[0];

      // If we want to skip rock events in this detector
      if((geom->DetId() == novadaq::cnv::kNEARDET && !fConsiderNDRock) ||
         (geom->DetId() == novadaq::cnv::kFARDET  && !fConsiderFDRock)){
        // Where's the true vertex?
        const TVector3 vtx = mctruth.GetNeutrino().Nu().Position().Vect();

        // If the vertex is outside the detector, skip this event
        if(!geom->isInsideDetectorBigBox(vtx.X(), vtx.Y(), vtx.Z())) continue;
      }

      // Signal of a nu/e interaction, which can cause trouble
      if(gtruth.fgQ2 < 0) continue;

      GENIEReweightTable table;

      for(int systIdx = 1; systIdx <= kMaxReweightIdx; ++systIdx){
        // This one is known to be broken. Skipping it fills it with 1s
        if(systIdx == rwgt::fReweightFormZone) continue;

        const double wm2 = fRWsMinus2[systIdx]->CalcWeight(mctruth, gtruth);
        const double wm1 = fRWsMinus1[systIdx]->CalcWeight(mctruth, gtruth);
        const double wp1 = fRWsPlus1 [systIdx]->CalcWeight(mctruth, gtruth);
        const double wp2 = fRWsPlus2 [systIdx]->CalcWeight(mctruth, gtruth);

        table.SetWeights(systIdx, wm2, wm1, wp1, wp2);
      } // end for systIdx

      tablecol->push_back(table);

      util::CreateAssn(*this, evt, *tablecol,
                       art::Ptr<simb::MCTruth>(mctruths, truthIdx),
                       *assns);
    } // end for truthIdx


    evt.put(std::move(tablecol));
    evt.put(std::move(assns));
  }

  DEFINE_ART_MODULE(MakeGENIEReweightTable);

} // namespace
