///
/// \brief This package does not directly generate isotropic flux 
///        you have to add on an weight:|costheta|/pathLength in 
///        your analysis to make it isotropic
/// \author  Zukai Wang - ricken@fnal.gov
/// \date    Sept. 201
///

// C++ includes.
#include <iostream>
#include <string>
#include <cmath>
#include <memory>
#include <vector>

// ROOT includes.
#include "TDatabasePDG.h"
#include "TMath.h"
#include "TVector3.h"

#include "DAQChannelMap/DAQChannelMap.h"
#include "Geometry/Geometry.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "EventGeneratorBase/evgenbase.h"
#include "SummaryData/RunData.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Services/Optional/TFileService.h"

#include "CLHEP/Random/RandFlat.h"
#include "CLHEP/Random/RandGaussQ.h"

#include <fstream> 

#include "TTree.h"

namespace evgen {
  class SingleParticle;

  /// A module to check the results from the Monte Carlo generator
  class IsoGen : public art::EDProducer {
  public:
    explicit IsoGen(fhicl::ParameterSet const& pset);
    virtual ~IsoGen();                        

    void produce(art::Event& evt);  
    void reconfigure(const fhicl::ParameterSet& pset);
    void beginRun(art::Run& run);

  private:

    void Sample(simb::MCTruth &truth);

    std::vector<int>    fPDG;        ///< PDG code of particles to generate

    std::vector<double> fP0;         ///< Central momentum (GeV/c) to generate
    std::vector<double> fSigmaP;     ///< Variation in momenta (GeV/c)
    std::vector<int>    fPDist;      ///< How to distribute momenta (0=uniform, 1=gaussian)

    std::vector<double> fXhigh;      ///< Up limit of x position (cm)
    std::vector<double> fYhigh;      ///< Up limit of y position (cm)
    std::vector<double> fZhigh;      ///< Up limit of z position (cm)
    std::vector<double> fT0;         ///< Central t position (ns)
    std::vector<double> fXlow;       ///< Low limit of x position (cm)
    std::vector<double> fYlow;       ///< Low Limit of y position (cm)
    std::vector<double> fZlow;       ///< Low Limit of z position (cm)
    std::vector<double> fSigmaT;     ///< Variation in t position (ns)
    std::vector<int>    fPosDist;    ///< How to distribute xyz (0=uniform, 1=gaussian)

    TTree * isoGen_Tree;
    int surfID;
    double cosx, cosy, cosz;

  };
};


////////////////////////////////////////////////////////////////////////
namespace evgen{

  //____________________________________________________________________________
  IsoGen::IsoGen(fhicl::ParameterSet const& pset)
  {
    this->reconfigure(pset);

    // get the random number seed, use a random default if not specified    
    // in the configuration file.  
    int seed = pset.get< unsigned int >("Seed", evgb::GetRandomNumberSeed());
  
    createEngine( seed );
    
    produces< std::vector<simb::MCTruth> >();
    produces< sumdata::RunData, art::InRun >();
  }

  //____________________________________________________________________________
  IsoGen::~IsoGen() { }

  //____________________________________________________________________________
  void IsoGen::reconfigure(const fhicl::ParameterSet& pset)
  {
    fXlow       = pset.get< std::vector<double> >("Xlow");
    fYlow       = pset.get< std::vector<double> >("Ylow");
    fZlow       = pset.get< std::vector<double> >("Zlow");
    fXhigh      = pset.get< std::vector<double> >("Xhigh");
    fYhigh      = pset.get< std::vector<double> >("Yhigh");
    fZhigh      = pset.get< std::vector<double> >("Zhigh");
    fPDG        = pset.get< std::vector<int>    >("PDG");
    fP0         = pset.get< std::vector<double> >("P0");
    fSigmaP     = pset.get< std::vector<double> >("SigmaP");
    fPDist      = pset.get< std::vector<int>    >("PDist");
    fT0         = pset.get< std::vector<double> >("T0");
    fSigmaT     = pset.get< std::vector<double> >("SigmaT");
  }

  //____________________________________________________________________________
  void IsoGen::beginRun(art::Run& run)
  {
  
    // grab the geometry object to see what geometry we are using
    art::ServiceHandle<geo::Geometry> geo;
  
    std::unique_ptr<sumdata::RunData> runcol(new sumdata::RunData(geo->DetId(),
								  geo->FileBaseName(),
								  geo->ExtractGDML()));

    art::ServiceHandle<art::TFileService> tfs;
    isoGen_Tree = tfs->make<TTree>("isoGen_Tree","Information from Generator");
    isoGen_Tree -> Branch("surfID", &surfID,  "surfID/I");
    isoGen_Tree -> Branch("cosx",   &cosx,    "cosx/D");
    isoGen_Tree -> Branch("cosy",   &cosy,    "cosy/D");
    isoGen_Tree -> Branch("cosz",   &cosz,    "cosz/D");
  
    run.put(std::move(runcol));
  
    return;
  }

  //____________________________________________________________________________
  void IsoGen::produce(art::Event& evt)
  {
  
    ///unique_ptr allows ownership to be transferred to the art::Event after the ut statement
    std::unique_ptr< std::vector<simb::MCTruth> > truthcol(new std::vector<simb::MCTruth>);
  
    simb::MCTruth truth;
    truth.SetOrigin(simb::kSingleParticle);
    Sample(truth);
  
    //     std::cout << "put mctruth into the vector" << std::endl;
    truthcol->push_back(truth);
  
    //     std::cout << "add vector to the event " << truthcol->size() << std::endl;
    evt.put(std::move(truthcol));
  
    return;
  }

  //____________________________________________________________________________
  void IsoGen::Sample(simb::MCTruth& mct) 
  {
    // get the random number generator service and make some CLHEP generators
    art::ServiceHandle<art::RandomNumberGenerator> rng;

    CLHEP::HepRandomEngine &engine = rng->getEngine();
    CLHEP::RandFlat   flat(engine);
    CLHEP::RandGaussQ gauss(engine);

    std::ofstream ofs;
    ofs.open ("iso.txt", std::ofstream::out | std::ofstream::app);

    //
    // every event will have one of each particle species in the fPDG array
    //
    for (unsigned int i=0; i<fPDG.size(); ++i) {    
      // Choose momentum
      double p = 0.0;
      double m = 0.0;

      if (fPDist[i]) {
        p = gauss.fire(fP0[i], fSigmaP[i]);
      }
      else {
        p = flat.fire(fP0[i]-fSigmaP[i], fP0[i]+fSigmaP[i]);
      }     
    
      // ofs << "get the mass" << std::endl;
    
      static TDatabasePDG*  pdgt = TDatabasePDG::Instance();
      TParticlePDG* pdgp = pdgt->GetParticle(fPDG[i]);
      if (pdgp) m = pdgp->Mass();
    
      //Choose an incident surface:
      double lx,ly,lz;
      lx = fXhigh[i]-fXlow[i];
      ly = fYhigh[i]-fYlow[i];
      lz = fZhigh[i]-fZlow[i];

      double Surf[6];
      Surf[0] = lx*ly;
      Surf[1] = lx*lz;
      Surf[2] = ly*lz;
      Surf[3] = lx*lz;
      Surf[4] = ly*lz;
      Surf[5] = lx*ly;

      double sumA =0;
      for (int j=0; j!=6;++j) sumA += Surf[j];

      double dice = flat.fire(0,sumA);

      // Choose position and time
      double x[4];
      x[3] = flat.fire(fT0[i]-fSigmaT[i], fT0[i]+fSigmaT[i]);
      // Choose angles
      double phi;

      phi  = flat.fire(0,2.*TMath::Pi());
      //      double pass = flat.fire(0,1);

      //incident from front(0):
      if (dice<Surf[0]) {
	x[0] = flat.fire(fXlow[i],fXhigh[i]);
	x[1] = flat.fire(fYlow[i],fYhigh[i]);
        x[2] = fZhigh[i];

	cosz = flat.fire(-1,0);
	//	cosz = -sqrt(flat.fire(0,1));
	cosx = sqrt(1-cosz*cosz)*TMath::Cos(phi);
	cosy = sqrt(1-cosz*cosz)*TMath::Sin(phi);
	surfID=0;
      }

      //incident from end(5):
      else if (dice<Surf[0]+Surf[5]) {
	x[0] = flat.fire(fXlow[i],fXhigh[i]);
	x[1] = flat.fire(fYlow[i],fYhigh[i]);
        x[2] = fZlow[i];
	cosz = flat.fire(0,1);
	cosx = sqrt(1-cosz*cosz)*TMath::Cos(phi);
	cosy = sqrt(1-cosz*cosz)*TMath::Sin(phi);

	surfID=5;
      }

      //incident from top(1):
      else if (dice<Surf[0]+Surf[5]+Surf[1]) {
	x[0] = flat.fire(fXlow[i],fXhigh[i]);
	x[1] = fYhigh[i];
        x[2] = flat.fire(fZlow[i],fZhigh[i]);
	cosy = -flat.fire(0,1);
	cosz = sqrt(1-cosy*cosy)*TMath::Cos(phi);
	cosx = sqrt(1-cosy*cosy)*TMath::Sin(phi);

	surfID=1;
      }

      //incident from bottom(3):
      else if (dice<sumA-Surf[4]-Surf[2]) {
	x[0] = flat.fire(fXlow[i],fXhigh[i]);
	x[1] = fYlow[i];
        x[2] = flat.fire(fZlow[i],fZhigh[i]);
	cosy = flat.fire(0,1);
	cosz = sqrt(1-cosy*cosy)*TMath::Cos(phi);
	cosx = sqrt(1-cosy*cosy)*TMath::Sin(phi);

	surfID=3;
      }
    
      //incident from left(4):
      else if (dice<sumA-Surf[2]) {
	x[0] = fXlow[i];
	x[1] = flat.fire(fYlow[i],fYhigh[i]);
        x[2] = flat.fire(fZlow[i],fZhigh[i]);
	cosx = flat.fire(0,1);
	cosy = sqrt(1-cosx*cosx)*TMath::Cos(phi);
	cosz = sqrt(1-cosx*cosx)*TMath::Sin(phi);

	surfID=4;
      }
    
      //incident from right(2):
      else {
	x[0] = fXhigh[i];
	x[1] = flat.fire(fYlow[i],fYhigh[i]);
        x[2] = flat.fire(fZlow[i],fZhigh[i]);
	cosx = -flat.fire(0,1);
	cosy = sqrt(1-cosx*cosx)*TMath::Cos(phi);
	cosz = sqrt(1-cosx*cosx)*TMath::Sin(phi);

	surfID=2;
      }
     
      isoGen_Tree -> Fill();

      TLorentzVector pvec(cosx*p,
			  cosy*p,
			  cosz*p,
			  sqrt(p*p+m*m));


      //  ofs<<cosx<<" "<<cosy<<"  "<<cosz<<std::endl;

      TLorentzVector pos(x[0],
			 x[1],
			 x[2],
			 x[3]);

      //
      // set track id to -i as these are all primary particles and have
      // id <= 0
      //
      int trackid = -1*(i+1);

      std::string primary("primary");
      simb::MCParticle part(trackid, fPDG[i], primary);
      part.AddTrajectoryPoint(pos, pvec);
    
      mct.Add(part);
    } //end loop over particles

    // ofs.close();
  }


} //end namespace

////////////////////////////////////////////////////////////////////////
namespace evgen
{
  DEFINE_ART_MODULE(IsoGen);
}
////////////////////////////////////////////////////////////////////////

