///
/// \brief   Generator for supernovae
/// \author  m.tamsett@sussex.ac.uk
/// \date
///

// C++ includes.
#include <iostream>
#include <string>
#include <cmath>
#include <memory>
#include <vector>

// ROOT includes
#include "TDatabasePDG.h"
#include "TRandom3.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "Geometry/Geometry.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "EventGeneratorBase/evgenbase.h"
#include "SummaryData/RunData.h"
#include "NovaDAQConventions/DAQConventions.h"

#include "CLHEP/Random/RandFlat.h"
#include "CLHEP/Random/RandGaussQ.h"
#include "TH2F.h"
#include "TFile.h"

namespace evgen {
  class Supernova;

  /// A module to check the results from the Monte Carlo generator
  class SupernovaGen : public art::EDProducer {
  public:
    explicit SupernovaGen(fhicl::ParameterSet const& pset);
    virtual ~SupernovaGen();                        

    void produce(art::Event& evt);  
    void reconfigure(const fhicl::ParameterSet& pset);
    void beginRun(art::Run& run);

  private:

    void Sample(std::unique_ptr< std::vector<simb::MCTruth> > &truthcol);

    /// Depending on configuration of fPMeaning
    /// calculate the momentum of the particle
    double getMomentum(const double& pkf, const double& m) const;
    
    static const int    kUNIF = 0;
    static const int    kGAUS = 1;
    int                 fPDG;        ///< PDG code of particles to generate
    int                 fPMeaning;   ///< Meaning of P0, fSigmaP and fPDist. 
                                     ///< By default (fP0Meaning=0), P0 and sigmaP is momentum
                                     ///< If fPMeaning=1, then P0 and sigmaP is kinetic energy in GeV
                                     ///< If fPMeaning=2, then P0and sigmaP is total energy in GeV
    double              fXlow;       ///< Lower x position (cm)
    double              fXhigh;      ///< Upper x position (cm)
    double              fYlow;       ///< Lower y position (cm)
    double              fYhigh;      ///< Upper y position (cm)
    double              fZlow;       ///< Lower z position (cm)
    double              fZhigh;      ///< Upper z position (cm)
    double              fTlow;       ///< Lower t position (ns)
    double              fThigh;      ///< Upper t position (ns)
    double              fCosZ0;      ///< Cosine of central angle wrt z-axis
    double              fPhiXY0;     ///< Central angle in the x-y plane (degrees)
    double              fSigmaCosZ;  ///< Size of variation of cosz
    double              fSigmaPhiXY; ///< Size of variation in phixy (degrees)
    int                 fAngleDist;  ///< How to distribute angles (0=uniform, 1=gaussian)
    std::string         fPositronPDFFile; ///< ROOT file name where the positron PDFs are stored
    bool                fOnlyNoise;  ///< Generate no signal
    // Multiplicative scaling factor to change the distance of the SN 
    // the given rates are proportional to the distance^2 (10^2)
    double              fDistanceScaling; ///< Scaling factor to PDF values
    
    TH2F* fPositronPDF;
    int fNtimePoints, fNenergyBins;
    double fMinPositronEnergy = 0.0007488;   // GeV
    double fMaxPositronEnergy = 0.09975;     // GeV
    // Energy variation of the positrons will be defined by half the bin width
    double fSigmaP;
    int fTimeIndex = 0;
  };
};

////////////////////////////////////////////////////////////////////////

namespace evgen{

  //----------------------------------------------------------------------------
  SupernovaGen::SupernovaGen(fhicl::ParameterSet const& pset):
    fPMeaning(0)
  {
    this->reconfigure(pset);

    // get the random number seed, use a random default if not specified    
    // in the configuration file.  
    int seed = pset.get< unsigned int >("Seed", evgb::GetRandomNumberSeed());
  
    createEngine( seed );
    
    produces< std::vector<simb::MCTruth> >();
    produces< sumdata::RunData, art::InRun >();
  }

  //----------------------------------------------------------------------------
  SupernovaGen::~SupernovaGen() { }

  //----------------------------------------------------------------------------
  void SupernovaGen::reconfigure(const fhicl::ParameterSet& pset)
  {
    fPDG        = pset.get< int    >("PDG");
    try{
        fPMeaning = pset.get< int >("PMeaning");
        if     (fPMeaning == 1) {
            std::cout<<"SupernovaGen: Using Kinetic energy for the meaning of P0, SigmaP and PDist\n";
        }
        else if(fPMeaning == 2) {
            std::cout<<"SupernovaGen: Using Total energy for the meaning of P0, SigmaP and PDist\n";
        }
    }
    catch(...){
        fPMeaning = 0;
    }
    fXlow               = pset.get< double >("Xlow");
    fYlow               = pset.get< double >("Ylow");
    fZlow               = pset.get< double >("Zlow");
    fXhigh              = pset.get< double >("Xhigh");
    fYhigh              = pset.get< double >("Yhigh");
    fZhigh              = pset.get< double >("Zhigh");
    fTlow               = pset.get< double >("Tlow");
    fThigh              = pset.get< double >("Thigh");
    fCosZ0              = pset.get< double >("CosZ0");
    fPhiXY0             = pset.get< double >("PhiXY0");
    fSigmaPhiXY         = pset.get< double >("SigmaPhiXY");
    fSigmaCosZ          = pset.get< double >("SigmaCosZ");
    fAngleDist          = pset.get< int    >("AngleDist");
    fPositronPDFFile    = pset.get< std::string >("PositronPDFFile");
    fOnlyNoise          = pset.get< bool   >("OnlyNoise");
    fDistanceScaling    = pset.get< double >("DistanceScaling");
    
    TFile *root_file = new TFile( fPositronPDFFile.c_str() );
    root_file->GetObject("ibd",fPositronPDF);
    
    fNtimePoints = fPositronPDF->GetNbinsX();
    fNenergyBins = fPositronPDF->GetNbinsY();
    fSigmaP = (fMaxPositronEnergy - fMinPositronEnergy) / (2.*double(fNenergyBins));
  }

  //----------------------------------------------------------------------------
  void SupernovaGen::beginRun(art::Run& run)
  {
    std::cout << "--- SupernovaGen::begin run\n";
    // grab the geometry object to see what geometry we are using
    art::ServiceHandle<geo::Geometry> geo;
    std::cout << "\tDetector ID: " << geo->DetId() << std::endl;
  
    std::unique_ptr<sumdata::RunData> runcol(new sumdata::RunData(geo->DetId(),
                                             geo->FileBaseName(),
                                             geo->ExtractGDML()));
    run.put(std::move(runcol));
    
    return;
  }

  //----------------------------------------------------------------------------
  void SupernovaGen::produce(art::Event& evt)
  {
    std::cout << "--- SupernovaGen::produce, event: " << evt.id().event() << std::endl;
    
    // increment time index
    fTimeIndex++;
    
    // check that time index is within the available bounds
    // TODO exit gracefully rather than crashing
    if (fTimeIndex > fNtimePoints) return;
    
    //unique_ptr allows ownership to be transferred to the art::Event after the out statement
    std::unique_ptr< std::vector<simb::MCTruth> > truthcol(new std::vector<simb::MCTruth>);
  
    Sample(truthcol);
  
    //     std::cout << "add vector to the event " << truthcol->size() << std::endl;
    evt.put(std::move(truthcol));
    std::cout << "\tEnd of produce" << std::endl;
    return;
  }

  //----------------------------------------------------------------------------
  void SupernovaGen::Sample(std::unique_ptr< std::vector<simb::MCTruth> > &truthcol) 
  {
    std::cout << "--- SupernovaGen::Sample, time index: " << fTimeIndex << std::endl;
    // get the random number generator service and make some CLHEP generators
    art::ServiceHandle<art::RandomNumberGenerator> rng;
    CLHEP::HepRandomEngine &engine = rng->getEngine();
    CLHEP::RandFlat   flat(engine);
    CLHEP::RandGaussQ gauss(engine);
    
    //
    // every event will have the positron distribution from the fPositronPDF array
    //
    int generatedPositrons = 0;
    //// debug - make empty events
    //bool generate_noise = true;
    //if (generate_noise) return;
    //// end of debug
    for (int i_energy=0; i_energy<fNenergyBins; ++i_energy) {
      // the central momentum is given by the bin centre
      double P0 = fMinPositronEnergy + (double(i_energy+1)*2.*fSigmaP);
      double positronPDF = 0.;
      if (!fOnlyNoise) 
        positronPDF = fPositronPDF->GetBinContent(fTimeIndex,i_energy+1);
          
      std::cout << "\tEnergy bin[" << i_energy 
                << "] ("           << P0
                << " GeV), positron PDF: "  << positronPDF
                << std::endl;
      // Generate a poisson fluctated number of positrons based on this
      TRandom3 rnd(0); // zero seeds randomly
      int nPositrons = rnd.Poisson(positronPDF*fDistanceScaling);
      std::cout << "\tGenerating "  << nPositrons
                << " positrons "    << std::endl; 
      for (int i_positron=0; i_positron<nPositrons; ++i_positron) {
          generatedPositrons++;
	  simb::MCTruth mct;
	  mct.SetOrigin(simb::kSingleParticle);
          // std::cout << "picking a " << fPDG << std::endl
          //    << fP0 << " " << fSigmaP << " (" << fX0
          //    << "/" << fSigmaX << ", " << fY0 
          //    << "/" << fSigmaY << ", " << fZ0 
          //    << "/" << fSigmaZ << ") " << fTheta0XZ
          //    << "/" << fSigmaThetaXZ << " " << fTheta0YZ
          //    << "/" << fSigmaThetaYZ << std::endl;
          // std::cout << fPDist << " " << fPosDist << " " << fAngleDist << std::endl;
        
          // Momentum, kinetic energy or full energy
          double pkf = 0.0;
          pkf = flat.fire(P0-fSigmaP, P0+fSigmaP);
          
          // std::cout << "get the mass" << std::endl;
          const TDatabasePDG* pdgt = TDatabasePDG::Instance();
          const TParticlePDG* pdgp = pdgt->GetParticle(fPDG);
          // Mass in GeV
          double    m = 0.0;
          if (pdgp) m = pdgp->Mass();
    
          // Momentum in GeV/c
          const double p = getMomentum(pkf, m);
    
          //std::cout<<"Momentum = "<<p<<"\n";
    
          // Choose position
          double x[4];
          x[0] = flat.fire(fXlow, fXhigh);
          x[1] = flat.fire(fYlow, fYhigh);
          x[2] = flat.fire(fZlow, fZhigh);
          x[3] = flat.fire(fTlow, fThigh);
          // std::cout << "set the position" << std::endl;
          std::cout << "\t\tpositron[ " << generatedPositrons-1
                    << "]: p: "         << p
                    << " GeV, x: "      << x[0]
                    << " cm, y: "       << x[1]
                    << " cm, z: "       << x[2]
                    << " cm, t: "       << x[3]
                    << " ns"            << std::endl;
          const TLorentzVector pos(x[0], x[1], x[2], x[3]);
        
          // Choose angles
    
          double       cosz, phi;
          unsigned int itry;
          for (itry=0; itry<1000000; ++itry) {
            if (fAngleDist == kGAUS) {
              cosz = gauss.fire(fCosZ0,  fSigmaCosZ);
              phi  = gauss.fire(fPhiXY0, fSigmaPhiXY);
            } else {
              cosz = flat.fire(fCosZ0-fSigmaCosZ,
                       fCosZ0+fSigmaCosZ);
        
              phi  = flat.fire(fPhiXY0-fSigmaPhiXY,
                       fPhiXY0+fSigmaPhiXY);
            }
            if (cosz>=-1.0 && cosz<=1.0) break;
          }
          if (cosz<-1.0 || cosz>1.0) {
            mf::LogError("SupernovaGen") << __FILE__ << ":" << __LINE__
                          << " Failed to find an acceptable cos(theta_z)"
                          << " after many tries.\n"
                          << " Please adjust CosZ0 and SigmaCosZ0 in your"
                          << " SupernovaGen.fcl file and rerun";
            abort();
          }
        
          const double sinz    = sqrt(1.0-cosz*cosz);
          const double sinphi  = sin(phi*M_PI/180.0);
          const double cosphi  = cos(phi*M_PI/180.0);
          //
          // set track id to -i as these are all primary particles and have
          // id <= 0
          //
          const int    trackid = -1*(generatedPositrons+1);
    
          const std::string primary("primary");
          simb::MCParticle part(trackid, fPDG, primary);
    
          const TLorentzVector pvec(cosphi*sinz*p,
                                    sinphi*sinz*p,
                                    cosz*p,
                                    sqrt(p*p+m*m));
    
          part.AddTrajectoryPoint(pos, pvec);
        
          mct.Add(part);
	  truthcol->push_back(mct);
      } //end of loop over positrons
    } //end loop over energy bins particles
    std::cout << "Generated a total of " << generatedPositrons << " positrons" << std::endl;
  }

//------------------------------------------------------------------------------
double SupernovaGen::getMomentum(const double& pkf, const double& m) const{

    if     (fPMeaning == 0) return pkf;

    double total_energy = 0.0;
    if     (fPMeaning == 1){
        total_energy = pkf + m;
    }
    else if(fPMeaning == 2){
        total_energy = pkf;
    }
    else{
        total_energy = sqrt(pkf*pkf + m*m);
    }

    return sqrt(total_energy*total_energy - m*m);
}

//------------------------------------------------------------------------------
DEFINE_ART_MODULE(SupernovaGen);

} //end namespace


