////////////////////////////////////////////////////////////////////////
/// \file  LiveGeometry.h
/// \author  bays@caltech.edu
////////////////////////////////////////////////////////////////////////
#ifndef GEO_LIVEGEOMETRY_H
#define GEO_LIVEGEOMETRY_H

#include "GeometryObjects/GeometryBase.h"

#include "art/Framework/Principal/Run.h"
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Services/Registry/ServiceMacros.h"
#include "TVector3.h"

/// Detector geometry information
namespace geo {
  /// The instrumented geometry of one entire detector (near, far, ipnd)
  class LiveGeometry
  {
  public:
    LiveGeometry(fhicl::ParameterSet const& pset, art::ActivityRegistry& reg);
    ~LiveGeometry();

    void postBeginRun(art::Run const& run);
    void preBeginEvent(art::Event const& evt);

    /// \name Projected Distances
    //@{
    double ProjectedDistanceToEdge(TVector3 vertex, TVector3 dir);
    double ProjectedLiveDistanceToEdge(TVector3 vertex, TVector3 dir);
    double ProjectedDistanceToNextBadRegion(TVector3 vertex, TVector3 dir);
    bool   ProjectedDistance(TVector3 vertex, TVector3 dir, double &distToEdge, double &distToNextBad, double &distDead);
    //@}

    /// \name Projected Number of Cells
    //@{
    int ProjectedCellsToEdge(TVector3 vertex, TVector3 dir);
    int ProjectedLiveCellsToEdge(TVector3 vertex, TVector3 dir); ///< recommended projection to use
    int ProjectedCellsToNextBadRegion(TVector3 vertex, TVector3 dir);
    //@}

    /// \name Checking if Live
    //@{
    double GetBadBoxCorner(bool view, int coord, int i); ///< Get the corner of a bad box (hook for evend display)
    int  NumBadBoxesX(); ///< Size of the vector of bad boxes in the XZ view
    int  NumBadBoxesY(); ///< Size of the vector of bad boxes in the YZ view
    bool IsPointLiveInfo(TVector3 vertex, int &info);
    bool IsPointLive(TVector3 vertex);
    int  DCMStatus(unsigned int dib, unsigned int dcm); ///< returns: 0: RH says is bad; 1: good; 2: bad, either BC or DCM drop
    double DistToClosestBadRegion(TVector3 vertex);
    //@}

    /// \name Define a fully good volume; used to be in Geometry Service
    //@{
    double InstrumentedDetLength(); ///< get instrumented detector length of downstream part
    double InstrumentedDetFront(); ///< get instrumented detector front of downstream part
    double InstrumentedDetBack(); ///< get instrumented detector back of downstream part
    int    InstrumentedDetEnds(double &frontDwnstrm, double &backDwnstrm, double &frontUpstrm, double &backUpstrm); ///< give all detector ends information; most general
    void   SetInstrumentation(); ///< calculate the detector end information
    //@}

    /// \name Distances to Edge
    //@{
    double  DistToEdgeXPlus(TVector3 vertex);
    double  DistToEdgeXMinus(TVector3 vertex);
    double  DistToTop(TVector3 vertex);
    double  DistToBottom(TVector3 vertex);
    double  DistToFront(TVector3 vertex);
    double  DistToBack(TVector3 vertex);
    double  DistToEdgeX(TVector3 vertex);
    double  DistToEdgeY(TVector3 vertex);
    double  DistToEdgeXY(TVector3 vertex);
    double  DistToEdgeZ(TVector3 vertex);
    //@}

    /// \name Number of Cells/Planes to Edge
    //@{
    int  CellsToEdgeXPlus(TVector3 vertex);
    int  CellsToEdgeXMinus(TVector3 vertex);
    int  CellsToTop(TVector3 vertex);
    int  CellsToBottom(TVector3 vertex);
    int  PlanesToFront(TVector3 vertex);
    int  PlanesToBack(TVector3 vertex);
    int  CellsToEdgeX(TVector3 vertex);
    int  CellsToEdgeY(TVector3 vertex);
    int  CellsToEdgeXY(TVector3 vertex);
    int  PlanesToEdge(TVector3 vertex);
    //@}

    /// \name ND Muon Catcher Functions
    //@{
    bool   InMuonCatcher(TVector3 vertex);
    int    DoWeEnterMC(TVector3 vertex, TVector3 dir);
    double ProjectedDistanceInMC(TVector3 vertex, TVector3 dir);
    int    ProjectedCellsInMC(TVector3 vertex, TVector3 dir);
    int    FullNDProjectedCells(TVector3 vertex, TVector3 dir);
    double XPositionAtTransition(TVector3 vertex, TVector3 dir);
    double YPositionAtTransition(TVector3 vertex, TVector3 dir);
    double DistanceToEdgeInMC(TVector3 vertex, int &wall);
    int    CellsToEdgeInMC(TVector3 vertex, int &wall);
    double ProjectedAirDist(TVector3 vertex, TVector3 dir);
    double ProjectedSteelDist(TVector3 vertex, TVector3 dir);
    //@}

    /// \name Other
    //@{
    double LiveDetectorVolume(); ///< in ktons
    int ProjectedWall(TVector3 vertex, TVector3 dir); ///< if we assume the downstream InstrumentedDetEnds, what wall would this projected track exit from?
    int NBadDCMBC();
    int NDropouts();
    int TotalDropouts();
    //@}

  protected:
    double fEdgeFrontXZ=0, fEdgeBackXZ=0, fEdgeXMinusXZ=0, fEdgeXPlusXZ=0; ///< proper 2D detector boundaries in XZ
    double fEdgeFrontYZ=0, fEdgeBackYZ=0, fEdgeTopYZ=0, fEdgeBottomYZ=0;   ///< proper 2D detector boundaries in YZ
    double fMCFront=0, fMCBack=0, fMCEdge=0, fMCTop=0; ///< muon catcher boundaries
    bool fCheckBadChan = true;  ///< do we check bad channels and call DCMs bad if they have too many
    bool fVerbose = false;       ///< print out extra information
    bool fSetup = false;         ///< have we done setup yet this run
    bool fEvFlag = false;        ///< have we read in information for this event yet? (don't use preBeginEvent since race w/ BadChan)
    std::string fCalHitLabel;
    int fDCMs[12][14];
    int fFrontUpstream = -5;
    int fBackUpstream = -5;
    int fFrontDownstream = -5;
    int fBackDownstream = -5;
    int fNumDiblocks = 0;
    int fNumDCMs = 0;
    int fNumCells = 0;
    int fNumChannels = 0;
    int fWall = 0;
    int fNDropouts = 0;
    int fTotalDropouts = 0;
    int fNEvents = 0;
    int fNBadChan = 0;
    float fMassKtons;
    struct BadBox
    {

      float Z1;  ///< point 1 for minima, point 2 for maxima
      float Z2;
      float XY1;
      float XY2;
      unsigned int type; ///< 0 means uninstrumented, 1 means bad
      bool edge; ///< are we connected to the detector edge?
    };

    std::vector<BadBox> fBadBoxXZ, fBadBoxYZ;

    /// \name Private Functions for Determining Bad Boxes
    //@{
    void SetupLiveGeo();      ///< Necessary set up before anything; get boudnaries, find and fill bad boxes
    void FillBadBoxes();      ///< Fill bad and uninstrumented regions of detector
    void GetDetectorEdges();  ///< Get proper detector boundaries for each view
    void GetRHInfo();         ///< Get uninstrumented regions from RunHistory
    void GetBCInfo();         ///< Get bad regions from BadChannels
    void GetEvInfo();         ///< Get event by event information
    //@}

  };
}
DECLARE_ART_SERVICE(geo::LiveGeometry, LEGACY)
#endif // GEO_GEOMETRY_H
////////////////////////////////////////////////////////////////////////
