//////////////////////////////////////////////////////////////////////////
// \file    CellGeo.cxx
// \brief   Cell geometry construction and call routines
// \version $Id: CellGeo.cxx,v 1.7 2012-11-14 21:09:52 jcoelho Exp $
// \author  $Author: jcoelho $
//////////////////////////////////////////////////////////////////////////
#include <iostream>

#include "TGeoManager.h"
#include "TGeoBBox.h"
#include "TGeoMatrix.h"
#include "TGeoNode.h"
#include "TVector3.h"

#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/CellUniqueId.h"

namespace geo
{
  /// Construct a cell geometry
  CellGeo::CellGeo(std::vector<const TGeoNode*>& n, int depth)
  {
    fCellNode = n[depth];

    // Build the matrix that takes us to the top world frame
    fGeoMatrix = new TGeoHMatrix(*n[0]->GetMatrix());
    for (int i=1; i<=depth; ++i) {
      fGeoMatrix->Multiply(n[i]->GetMatrix());
    }

    // Copy fGeoMatrix to Original
    fGeoMatrixOriginal = new TGeoHMatrix(*fGeoMatrix);

    // fGeoMatrix is still unmodified
    fIsOriginal = true;

    // Construct the unique id for this cell
    fId = NodesToUniqueId(n, depth);
  }

  //......................................................................
  CellGeo::~CellGeo()
  {
    // trying to check if the fGeoMatrix exists and then deleting it causes a segfault, for some reason
    if(fGeoMatrix) delete fGeoMatrix;
    if(fGeoMatrixOriginal) delete fGeoMatrixOriginal;
  }

  //......................................................................

  /// Transform a position from local frame to world frame
  /// \param local : 3D array. Position in the local frame  Input.
  /// \param world : 3D array. Position in the world frame. Returned.
  void CellGeo::LocalToWorld(const double* local, double* world) const
  {
    fGeoMatrix->LocalToMaster(local,world);
  }

  //......................................................................

  /// Transform a 3-vector from local frame to world frame
  /// \param local : 3D array. Position in the local frame  Input.
  /// \param world : 3D array. Position in the world frame. Returned.
  void CellGeo::LocalToWorldVect(const double* local, double* world) const
  {
    fGeoMatrix->LocalToMasterVect(local,world);
  }

  //......................................................................

  /// Transform a position from world frame to local frame
  /// \param world : 3D array. Position in the world frame. Input.
  /// \param local : 3D array. Position in the local frame  Returned.
  void CellGeo::WorldToLocal(const double* local, double* world) const
  {
    fGeoMatrix->MasterToLocal(local,world);
  }

  //......................................................................

  /// Transform a 3-vector from world frame to local frame
  /// \param world : 3D array. Position in the world frame. Input.
  /// \param local : 3D array. Position in the local frame  Returned.
  void CellGeo::WorldToLocalVect(const double* local, double* world) const
  {
    fGeoMatrix->MasterToLocalVect(local,world);
  }

  //......................................................................

  void CellGeo::TranslateCell(double dx, double dy, double dz)
  {

    /// Set rotation angles to 0 for now
    double phi = 0;
    double theta = 0;
    double psi = 0;

    /// Build the translation and rotation matrices
    /// phi, theta and psi are euler angles
    TGeoRotation *rot = new TGeoRotation("alignmentRot",phi,theta,psi);
    /// dx, dy and dz are the translations we want
    TGeoTranslation *trans = new TGeoTranslation("",dx,dy,dz);
    TGeoCombiTrans *comb = new TGeoCombiTrans(*trans,*rot);

    this->TranslateCell(comb);

  }

  //......................................................................

  void CellGeo::TranslateCell(TGeoCombiTrans* comb)
  {

    /// Restart from original matrix
    fGeoMatrix = new TGeoHMatrix(*fGeoMatrixOriginal);

    /// Apply shift
    fGeoMatrix->MultiplyLeft(comb->MakeClone());

    // fGeoMatrix is now modified
    fIsOriginal = false;
  
  }
  
  //...................................................................... 
   
  void CellGeo::RestoreOriginal()
  {
     
    /// Restore original matrix
    fGeoMatrix = new TGeoHMatrix(*fGeoMatrixOriginal);
  
    // fGeoMatrix is back to original
    fIsOriginal = true;
    
  }

  //......................................................................

  /// Return the center position of a cell.
  /// \param xyz : 3-D array. The returned location.
  /// \param localz : Distance along the length of the cell
  /// (cm). Default is center of cell
  void CellGeo::GetCenter(double* xyz, double localz) const
  {
    static double xyzLocal[3] = {0,0,0};
    xyzLocal[2] = localz;
    this->LocalToWorld(xyzLocal, xyz);
  }

  /// Return the center position of a cell.
  /// \param xyz :a Tvector3 object. The returned location.
  /// \param localz : Distance along the length of the cell
  /// (cm). Default is center of cell
  void CellGeo::GetCenter(TVector3& xyz, double localz) const
  {
    double xyzArray[3];
    this->GetCenter(xyzArray, localz);
    xyz.SetXYZ(xyzArray[0], xyzArray[1], xyzArray[2]);
  }


  //......................................................................

  /// Return the distance from a point in the cell to the read out side
  /// of the cell
  /// \param localz : Z position in cm in local cell frame
  /// \returns Distance in cm
  double CellGeo::DistToReadOut(double localz) const
  {
    return this->HalfL()-localz;
  }

  //......................................................................

  double CellGeo::HalfW() const
  {
    return ((TGeoBBox*)fCellNode->GetVolume()->GetShape())->GetDX();
  }

  //......................................................................

  double CellGeo::HalfL() const
  {
    return ((TGeoBBox*)fCellNode->GetVolume()->GetShape())->GetDZ();
  }

  //......................................................................

  double CellGeo::HalfD() const
  {
    return ((TGeoBBox*)fCellNode->GetVolume()->GetShape())->GetDY();
  }

  //......................................................................

} // end namespace geo
////////////////////////////////////////////////////////////////////////
