/***************************************************
\file    HMatrixUser.cxx
\brief   HMatrix analyzer. 
This code is designed to run over FD beam files.
calls the hme::HMPid objects to get the chi2 values and fills in a tree.
then make a bunch of histograms and apply some analysis cuts and calculate and display a cut flow. 
It doesn't yet have appropriate POT weighting or oscillation weights.
\version $Id: HMatrixUser.cxx,v 1.10 2012/11/08 00:01:58 hsahoo Exp $
\author  $Author: hsahoo $

changes to HMatrix code
Jan 30 :
analyzer code : Get the 3 chi2 values from hme::HMPid object and fills in a tree..
These two values are now same.. corrected..
Jan 28 :
namespace is changed : hmatrixe --> hme
wrote a producer code, HMatrix calculation is done inside the producer 
and fills the 3 chi2 values in hme::HMPid object. 
histogram or tree filling is not done here.
Aug 28 : all matrix bins are initialized to zero for each event
         also total energy and total transverse energy
Aug 20 : New binning for HMatrix and matrix means
*************************************/


#include "HMPid.h"

// C/C++ includes
#include <vector>
#include <iostream>
#include <cmath>
#include <fstream>

// NOvASoft includes
#include "Calibrator/Calibrator.h"
#include "CMap/CMap.h"


#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "SimulationBase/MCParticle.h"
#include "SummaryData/POTSum.h"
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "NovaDAQConventions/DAQConventions.h"


#include "MCCheater/BackTracker.h"
#include "Utilities/func/MathUtil.h"
#include "RecoJMShower/JMShower.h"
#include "RecoJMShower/EID.h"
#include "RecVarPID/RVP.h"

// ROOT includes
#include "TH1.h"
#include "TH2.h"
#include "TTree.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Services/Registry/ServiceHandle.h" 
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "cetlib/search_path.h"
#include "fhiclcpp/ParameterSet.h"  
#include "messagefacility/MessageLogger/MessageLogger.h" 
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"

namespace hme{
  class HMatrixUser : public art::EDAnalyzer {

  public:
    explicit HMatrixUser(const fhicl::ParameterSet& pset); // Required! explicit tag tells the compiler this is not a copy constructor      
    virtual ~HMatrixUser();                        
    
    // Optional, read/write access to event
    void reconfigure(const fhicl::ParameterSet& pset);    
    void beginJob();
    void analyze(art::Event const& evt); 
    void endSubRun(const art::SubRun& sr);
    void endJob();

    double CalculateMIPPeak(art::Ptr<rb::Track> track);
    //void SetUpEnergyArrays(std::vector<int>, std::vector<int>, std::vector<double>, int);
    std::vector<double> GetHMChisq();
    bool isInVertexFiducial(art::Ptr<rb::Track> track) const;

    std::vector<double> SetUpEnergyArrays(std::vector<int>, std::vector<int>, std::vector<double>, int);

    double GetOscWeight(const simb::MCNeutrino& nu, art::Ptr<simb::MCFlux>& flux);
    int GetEventType(const simb::MCNeutrino& nu, art::Ptr<simb::MCFlux>& flux);

    double fTotalPOT;


  private:
    int GetMCTruthM(const art::Event& evt, art::PtrVector<simb::MCTruth>& mctruth); 
    ///< Sample configurable module parameters
    int              fCounter;
    //cut counts
    double pass_cuts[11][11];//0 is numucc, 1 is nc, 2 is beamnue, 3 is nuecc
    
    static const int long_bins;
    static const int trans_bins;
    static const int comb_bins;

    // wait : HMatricies : please check that these have right dimensions
    double HM[21][21];
    double HTM[42][42];
    double HC[55][55];
    
    // dimensions of the HMatrix mean file    
    double  lmeans[21];
    double trmeans[42];
    double ltmeans[55];
    
    // dimensions of the HMatrix bins
    //double lbin[21];
    //double transm[42];
    //double ltbin[55];


    //double etot;
    //double ttot; // for tran-only bins
    //double esumlay[1000];
    //double eplcell[1000][400];
    //double Tesum; // total energy sum of hits used in test


    //int nhits_tot;
    int nlay;
    int maxHcell[1000];
    std::vector<double> fhMeans;
    std::vector<double> fvMeans;
    std::vector<int> fmeans;
    float            fFloat;  
    double           fDouble; 
    double max_layers;
    double max_cells;

    std::string fSliceLabel;
    std::string fTrackLabel;         ///< Read Tracks
    std::string fLongHM;             ///< Longitudinal HMatrix
    std::string fTransHM;            ///< Transvers HMatrix
    std::string fCombHM;             ///< Combined HMatrix
    std::string fLongMeans;
    std::string fTransMeans;
    std::string fCombMeans;

    
    //double_t fmeans[6];
    
    TH2F*            flchisqLT;            ///> a histogram 
    TH2F*            flchisqLT_nc;            ///> a histogram 
    TH2F*            flchisqLT_numu;            ///> a histogram 
    TH2F*            flchisqLT_nue;            ///> a histogr
    
    TH1F*            flchisqL;            ///> a histogram 
    TH1F*            flchisqL_nc;            ///> a histogram 
    TH1F*            flchisqL_numu;            ///> a histogram 
    TH1F*            flchisqL_nue;            ///> a histogram 
    TH1F*            flchisqT;            ///> a histogram 
    TH1F*            flchisqT_nc;            ///> a histogram 
    TH1F*            flchisqT_numu;            ///> a histogram 
    TH1F*            flchisqT_nue;            ///> a histogram 
    TH1F*            flchisqC;            ///> a histogram 
    TH1F*            flchisqC_nc;            ///> a histogram 
    TH1F*            flchisqC_numu;            ///> a histogram 
    TH1F*            flchisqC_nue;            ///> a histogram 
    
    TH1F*            flchisqL_ecut;            ///> a histogram 
    TH1F*            flchisqL_nc_ecut;            ///> a histogram 
    TH1F*            flchisqL_numu_ecut;            ///> a histogram 
    TH1F*            flchisqL_nue_ecut;            ///> a histogram 
    TH1F*            flchisqT_ecut;            ///> a histogram 
    TH1F*            flchisqT_nc_ecut;            ///> a histogram 
    TH1F*            flchisqT_numu_ecut;            ///> a histogram 
    TH1F*            flchisqT_nue_ecut;            ///> a histogram 
    TH1F*            flchisqC_ecut;            ///> a histogram 
    TH1F*            flchisqC_nc_ecut;            ///> a histogram 
    TH1F*            flchisqC_numu_ecut;            ///> a histogram 
    TH1F*            flchisqC_nue_ecut;            ///> a histogram 
    
    TH1F*            fenergy_trans_bins;            ///> a histogram 
    TH1F*            fenergy_long_bins;            ///> a histogram 
    
    TH1F*            ftot_energy_event;
    TH1F*            ftot_energy_event_nc;
    TH1F*            ftot_energy_event_nue;
    TH1F*            ftot_energy_event_numu;
    
    TH1F*            flength_longest_track;
    TH1F*            flength_longest_track_nc;
    TH1F*            flength_longest_track_nue;
    TH1F*            flength_longest_track_numu;
    
    TH1F*            fmip_peak;
    TH1F*            fmip_peak_nc;
    TH1F*            fmip_peak_nue;
    TH1F*            fmip_peak_numu;
    
    TH1F*            fEnergy_hit;
    TH1F*            fEnergy_hit_nc;
    TH1F*            fEnergy_hit_nue;
    TH1F*            fEnergy_hit_numu;
    
    TH1F*            fntracks;
    TH1F*            fntracks_nc;
    TH1F*            fntracks_nue;
    TH1F*            fntracks_numu;
    TH1F*            fnhits_per_track;
    TH1F*            fnhits_per_track_nc;
    TH1F*            fnhits_per_track_nue;
    TH1F*            fnhits_per_track_numu;

    TTree*           tree_hmatrix;
    TTree*           tree_1;
    TTree*           partvar;
    //TTree*           tree_hmatrix22;

    

  };
}


////////////////////////////////////////////////////////////////////////


namespace hme{
  
  // define the tree global variables
  int gpdg, gmopdg;

  double gnslice = 0;
  double gntrks, gntrksl, gnhits, grhit;
  double gdepE, gtotE, gtotE2, gefrac_event, gmippeak, gtrkl;

  double glchi2, gtchi2, gcchi2;

  double grvp, geid;

  double gflux_type, gin_nu_e, gevent_type, gevent_no, gosc_weight;
  int gt_slice_el, gt_slice_mu, gt_slice_pion, gt_slice_pi0, gt_slice_proton, gt_slice_neutron, gt_slice_other; // seven particle type
  double gslc2, gstc2, gscc2, gslc2_el, gstc2_el, gscc2_el, gslc2_p0, gstc2_p0, gscc2_p0;

  double galc2, gatc2, gacc2, galc2_el, gatc2_el, gacc2_el, galc2_p0, gatc2_p0, gacc2_p0;
  int gt_alg_el, gt_alg_mu, gt_alg_pion, gt_alg_pi0, gt_alg_proton, gt_alg_neutron, gt_alg_other;

  double gx_pos_1, gy_pos_1, gz_pos_1, gdist_to_APD;
  double gcellhit_1, gview_1, gplane_1, gcell_1, gPE_1, gPECorr_1, gMIP_1, gGeV_1, gedep_1, gefrac_1;
  double gPECorr_PE_1, gMIP_PECorr_1, gGeV_PECorr_1, gGeV_MIP_1;
  double gPE_PECorr_1;
  
  HMatrixUser::HMatrixUser(const fhicl::ParameterSet& pset):
    EDAnalyzer(pset),
    fSliceLabel      (pset.get<std::string>("SliceLabel")    ),
    fTrackLabel      (pset.get<std::string>("TrackLabel")    ),
    fLongHM          (pset.get<std::string>("LongHM")        ),
    fTransHM         (pset.get<std::string>("TransHM")       ),
    fCombHM          (pset.get<std::string>("CombHM")        ),
    fLongMeans       (pset.get<std::string>("LongMeans")     ),
    fTransMeans      (pset.get<std::string>("TransMeans")    ),
    fCombMeans       (pset.get<std::string>("CombMeans")     )
  {   
    reconfigure(pset);
  }
  
  void HMatrixUser::reconfigure(const fhicl::ParameterSet& pset)
  {
    fLongHM      = pset.get<std::string>("LongHM");
    fTransHM     = pset.get<std::string>("TransHM");
    fCombHM      = pset.get<std::string>("CombHM");
    fLongMeans   = pset.get<std::string>("LongMeans");
    fTransMeans  = pset.get<std::string>("TransMeans");
    fCombMeans   = pset.get<std::string>("CombMeans");
  }
  //......................................................................
  HMatrixUser::~HMatrixUser()
  {
    //======================================================================
    // Clean up any memory allocated by your module
    //======================================================================
    
    std::cout << "#############################################" << std::endl;
    std::cout << "at the very end of the job:" << std::endl;
    std::cout << "cut_level " << " NC " << " numuCC " << " nueCC " << " other " << std::endl;

    for(int i=0; i<11; i++){
      std::cout<<i<< " ";
      for(int j=1; j<5; j++){
	//double cut_ratio = pass_cuts[j][i]/pass_cuts[j][2]*100;
	double cut_ratio = 0.0;
	std::cout << cut_ratio << " ";
      }
      //std::cout<<i<<" "<<pass_cuts[1][i]/pass_cuts[1][0]<<" "<<pass_cuts[2][i]/pass_cuts[2][0]<<" "<<pass_cuts[3][i]/pass_cuts[3][0]/<<" "<<pass_cuts[4][i]/pass_cuts[4][0]<<std::endl;
      std::cout<<" "<<std::endl;
    }

    std::cout << "cut_level " << " NC " << " numuCC " << " nueCC " << " other " << std::endl; 
    
    for(int i=0; i<11; i++){
      std::cout << i << " " << pass_cuts[1][i] << " " << pass_cuts[2][i] 
		<< " " <<pass_cuts[3][i] << " " <<pass_cuts[4][i] << std::endl;
    }

    std::cout << "#############################################" << std::endl;

  }

  //const int HMatrixUser::long_bins  = 16;
  //const int HMatrixUser::trans_bins = 28;
  //const int HMatrixUser::comb_bins  = 24;

  const int HMatrixUser::long_bins  = 21;
  const int HMatrixUser::trans_bins = 42;
  const int HMatrixUser::comb_bins  = 55;

  //......................................................................
  void HMatrixUser::beginJob()
  {
    fTotalPOT=0.0;

   //
    // Book histograms, ntuples, initialize counts etc., etc., ...
    //
    art::ServiceHandle<art::TFileService> tfs;

    max_layers = 1000;
    max_cells  = 400;
    //nlay       = 4;
    nlay       = 3; // nlay is changed to 3 on Aug 24

    flchisqLT = tfs->make<TH2F>("flchisqLT", ";Log Chi-squared Distribution (longitudinal)", 400, 0., 15,400,0., 15.);    
    flchisqL = tfs->make<TH1F>("flchisqL", ";Log Chi-squared Distribution (longitudinal)", 400, 0., 15.);    
    flchisqT = tfs->make<TH1F>("flchisqT", ";Log Long Chi-squared Distribution (transverse)", 400, 0., 15.);
        
    fenergy_trans_bins = tfs->make<TH1F>("energy_trans_bins", "Mean Transverse Energy Fraction;Transvers HMatrix Bin", 25, -0.5, 24.5);
    fenergy_long_bins = tfs->make<TH1F>("energy_long_bins", "Mean Longitudinal Energy Fraction;Longitudinal HMatrix Bin", 15, -0.5, 14.5);
    
    /***********************
    ftot_energy_event = tfs->make<TH1F>("tot_energy_event", "Total Energy (GEV) of the Event",100,0,20);
    ftot_energy_event_nc = tfs->make<TH1F>("tot_energy_event_nc", "Total Energy (GEV) of the Event (NC)",100,0,20);
    ftot_energy_event_numu = tfs->make<TH1F>("tot_energy_event_numu", "Total Energy (GEV) of the Event (NuMu)",100,0,20);
    ftot_energy_event_nue = tfs->make<TH1F>("tot_energy_event_nue", "Total Energy (GEV) of the Event (Nue)",100,0,20);

    flength_longest_track = tfs->make<TH1F>("length_longest_track", "Length of the longest track",100,0,2000);
    flength_longest_track_nc = tfs->make<TH1F>("length_longest_track_nc", "Length of the longest track (NC)",100,0,2000);
    flength_longest_track_numu = tfs->make<TH1F>("length_longest_track_numu", "Length of the longest track (NuMu)",100,0,2000);
    flength_longest_track_nue = tfs->make<TH1F>("length_longest_track_nue", "Length of the longest track (Nue)",100,0,2000);

    fmip_peak = tfs->make<TH1F>("mip_peak", "MIP Peak",100,0,1);
    fmip_peak_nc = tfs->make<TH1F>("mip_peak_nc", "MIP Peak (NC)",100,0,1);
    fmip_peak_numu = tfs->make<TH1F>("mip_peak_numu", "MIP Peak (NuMu)",100,0,1);
    fmip_peak_nue = tfs->make<TH1F>("mip_peak_nue", "MIP Peak (Nue)",100,0,1);
    
    fntracks = tfs->make<TH1F>("ntracks", "Number of Tracks",101,-0.5,100.5);
    fntracks_nc = tfs->make<TH1F>("ntracks_nc", "Number of Tracks (NC)",101,-0.5,100.5);
    fntracks_numu = tfs->make<TH1F>("ntracks_numu", "Number of Tracks (NuMu)",101,-0.5,100.5);
    fntracks_nue = tfs->make<TH1F>("ntracks_nue", "Number of Tracks (Nue)",101,-0.5,100.5);
    
    fEnergy_hit = tfs->make<TH1F>("Energy_hit", "Energy per cell hit (MeV)",100,0,50);
    fEnergy_hit_nc = tfs->make<TH1F>("Energy_hit_nc", "Energy per cell hit (MeV) (NC)",100,0,50);
    fEnergy_hit_numu = tfs->make<TH1F>("Energy_hit_numu", "Energy per cell hit (MeV) (NuMu)",100,0,50);
    fEnergy_hit_nue = tfs->make<TH1F>("Energy_hit_nue", "Energy per cell hit (MeV) (Nue)",100,0,50);

    fnhits_per_track= tfs->make<TH1F>("nhits_per_track", "Number of Hits per Track",100,0,500);
    fnhits_per_track_nc= tfs->make<TH1F>("nhits_per_track_nc", "Number of Hits per Track (NC)",100,0,500);
    fnhits_per_track_numu= tfs->make<TH1F>("nhits_per_track_numu", "Number of Hits per Track (NuMu)",100,0,500);
    fnhits_per_track_nue= tfs->make<TH1F>("nhits_per_track_nue", "Number of Hits per Track(Nue)",100,0,500);

    *********************/


    // wait : define the hmatrix tree ...
    // the variables in tree_hmatrix is calculated once for each event...
    tree_hmatrix = tfs->make<TTree>("hmatrix", "variables to analyze in HMatrix");

    tree_hmatrix->Branch("evtno",     &gevent_no);
    tree_hmatrix->Branch("fluxtype",  &gflux_type);
    tree_hmatrix->Branch("evttype",   &gevent_type);
    tree_hmatrix->Branch("oscw",      &gosc_weight);
    tree_hmatrix->Branch("nuE",       &gin_nu_e);
    tree_hmatrix->Branch("rvp",       &grvp);
    tree_hmatrix->Branch("eid",       &geid);
    tree_hmatrix->Branch("sel",       &gt_slice_el);
    tree_hmatrix->Branch("smu",       &gt_slice_mu);
    tree_hmatrix->Branch("spion",     &gt_slice_pion);
    tree_hmatrix->Branch("spi0",      &gt_slice_pi0);
    tree_hmatrix->Branch("sproton",   &gt_slice_proton);
    tree_hmatrix->Branch("sneutron",  &gt_slice_neutron);
    tree_hmatrix->Branch("sother",    &gt_slice_other);
    tree_hmatrix->Branch("ael",       &gt_alg_el);
    tree_hmatrix->Branch("amu",       &gt_alg_mu);
    tree_hmatrix->Branch("apion",     &gt_alg_pion);
    tree_hmatrix->Branch("api0",      &gt_alg_pi0);
    tree_hmatrix->Branch("aproton",   &gt_alg_proton);
    tree_hmatrix->Branch("aneutron",  &gt_alg_neutron);
    tree_hmatrix->Branch("aother",    &gt_alg_other);
    tree_hmatrix->Branch("depE",      &gdepE);
    tree_hmatrix->Branch("totE",      &gtotE);
    tree_hmatrix->Branch("totE2",     &gtotE2);
    tree_hmatrix->Branch("trkl",      &gtrkl);
    tree_hmatrix->Branch("mippeak",   &gmippeak);
    tree_hmatrix->Branch("ntrks",     &gntrks);
    tree_hmatrix->Branch("ntrksl",    &gntrksl);
    tree_hmatrix->Branch("nhits",     &gnhits); // total hits from tracker
    tree_hmatrix->Branch("nslice",    &gnslice); // total hits in slicer
    tree_hmatrix->Branch("slc2",      &gslc2);
    tree_hmatrix->Branch("stc2",      &gstc2);
    tree_hmatrix->Branch("scc2",      &gscc2);
    tree_hmatrix->Branch("slc2_el",   &gslc2_el);
    tree_hmatrix->Branch("stc2_el",   &gstc2_el);
    tree_hmatrix->Branch("scc2_el",   &gscc2_el);
    tree_hmatrix->Branch("slc2_p0",   &gslc2_p0);
    tree_hmatrix->Branch("stc2_p0",   &gstc2_p0);
    tree_hmatrix->Branch("scc2_p0",   &gscc2_p0);
    tree_hmatrix->Branch("alc2",      &galc2);
    tree_hmatrix->Branch("atc2",      &gatc2);
    tree_hmatrix->Branch("acc2",      &gacc2);
    tree_hmatrix->Branch("alc2_el",   &galc2_el);
    tree_hmatrix->Branch("atc2_el",   &gatc2_el);
    tree_hmatrix->Branch("acc2_el",   &gacc2_el);
    tree_hmatrix->Branch("alc2_p0",   &galc2_p0);
    tree_hmatrix->Branch("atc2_p0",   &gatc2_p0);
    tree_hmatrix->Branch("acc2_p0",   &gacc2_p0);

    //tree_hmatrix->Branch("efrac",     &gefrac_event);
    //tree_hmatrix->Branch("rhit",      &grhit);


    // This tree is for each hit..............
    tree_1 = tfs->make<TTree>("tree1", "tree1 variables");

    tree_1->Branch("cellhit",     &gcellhit_1);
    tree_1->Branch("view",        &gview_1);
    tree_1->Branch("x",           &gx_pos_1);
    tree_1->Branch("y",           &gy_pos_1);
    tree_1->Branch("z",           &gz_pos_1);
    tree_1->Branch("apd",         &gdist_to_APD);
    tree_1->Branch("plane",       &gplane_1);
    tree_1->Branch("cell",        &gcell_1);
    tree_1->Branch("PE",          &gPE_1);
    tree_1->Branch("PECorr",      &gPECorr_1);
    tree_1->Branch("MIP",         &gMIP_1);
    tree_1->Branch("GeV",         &gGeV_1);
    tree_1->Branch("edep",        &gedep_1);
    tree_1->Branch("efrac",       &gefrac_1);
    tree_1->Branch("PE2PECorr",   &gPECorr_PE_1);
    tree_1->Branch("PECorr2MIP",  &gMIP_PECorr_1);
    tree_1->Branch("MIP2GeV",     &gGeV_MIP_1);
    tree_1->Branch("PECorr2GeV",  &gGeV_PECorr_1);
    tree_1->Branch("PECorr2PE",   &gPE_PECorr_1);

    /***************
    // This tree is for each particles .............
    partvar = tfs->make<TTree>("partvar", "tree of particles");
    partvar->Branch("evtno",     &gevent_no);
    partvar->Branch("evttype",   &gevent_type);
    partvar->Branch("pdg",       &gpdg);
    partvar->Branch("mopdg",     &gmopdg);
    partvar->Branch("rvp",       &grvp);
    partvar->Branch("eid",       &geid);
    ***************/

    //initilize the histograms that will hold the horizontal and vertical matrix means to zero.
    for(int jj=0; jj<55; jj++){
      fhMeans.push_back(0);
      fvMeans.push_back(0);
    }
    
    //std::cout << "long_bins " << long_bins << "\t trans_bins " << trans_bins << "\t comb_bins " << comb_bins << std::endl;


    // All matrix bins are initialized separately to zero.........
    for (int i=0; i<long_bins; ++i){
      for (int j=0; j<long_bins; ++j){
	HM[i][j] = 0.;
      }
    }

    for (int i=0; i<trans_bins; ++i){
      for (int j=0; j<trans_bins; ++j){
	HTM[i][j] = 0.;
      }
    }

    for (int i=0; i<comb_bins; ++i){
      for (int j=0; j<comb_bins; ++j){
	HC[i][j] = 0.;
      }
    }


    /******************************************
// wait : old, only 16 bins were initialized..........
    for (int i=0; i<16; ++i)
      {
	for (int j=0; j<16; ++j)
	  {
	    HM[i][j] = 0.;
	    HTM[i][j] = 0.;
	    HC[i][j] = 0.;
	  }
      }
    ********************************/    

    std::string LongHM, TransHM, CombHM;
    std::string LongMeans, TransMeans, CombMeans;

    cet::search_path sp("FW_SEARCH_PATH");
    sp.find_file(fLongHM,LongHM);
    sp.find_file(fTransHM,TransHM);
    sp.find_file(fCombHM,CombHM);

    sp.find_file(fLongMeans,  LongMeans);
    sp.find_file(fTransMeans, TransMeans);
    sp.find_file(fCombMeans,  CombMeans);

    ifstream longhmi;  
    ifstream transhmi;
    ifstream combhmi;
    
    ifstream longmeans;
    ifstream transmeans;
    ifstream combmeans;

    // ############################################################
    //std::cout << "START : ########## Reading the HMatrixI files" << std::endl;
    // The matrixes are read from input directory.
    // make sure SRT_PRIVATE_CONTEXT flag is set properely.

    longhmi.open(LongHM.c_str());
    
    Double_t ldat;
    if (!longhmi.is_open()){
      throw cet::exception("UnresolvedMatrix")
        << "Can not open Long HMatrix file " << LongHM << "!";      
    }
    for (int i=0; i<long_bins; ++i){
      for (int j=0; j<long_bins; ++j){
	longhmi >> ldat;
	HM[i][j] = ldat;
	//std::cout << HM[i][j] << " ";
      }
      //std::cout << "\n";
    }
    longhmi.close();    
    //std::cout << "\n----------------------------" << std::endl;

    /********************************************
    for (int i=0; i<long_bins; ++i){
      for (int j=0; j<long_bins; ++j){
	std::cout << HM[i][j] << " ";
      }
      std::cout << "\n";
    }
    ****************************/

    transhmi.open(TransHM.c_str());
    Double_t tdat;
    if (!transhmi.is_open()){
      throw cet::exception("UnresolvedMatrix")
	<< "Can not open Trans HMatrix file " << TransHM << "!";
    }
    for (int i=0; i<trans_bins; ++i){
      for (int j=0; j<trans_bins; ++j){
	transhmi >> tdat;
	HTM[i][j] = tdat;
	//std::cout << HTM[i][j] << " ";
      }
      //std::cout << "\n";
    }
    transhmi.close();
    //std::cout << "\n----------------------------" << std::endl;


    /********************************************
    for (int i=0; i<trans_bins; ++i){
      for (int j=0; j<trans_bins; ++j){
	std::cout << HTM[i][j] << " ";
      }
      std::cout << "\n";
    }
    *******************************************/


    combhmi.open(CombHM.c_str());
    Double_t cdat;
    if (!combhmi.is_open()){
      throw cet::exception("UnresolvedMatrix")
	<< "Can not open Combined HMatrix file " << CombHM << "!";
    }
    for (int i=0; i<comb_bins; ++i){
      for (int j=0; j<comb_bins; ++j){
	combhmi >> cdat;
	HC[i][j] = cdat;
	//std::cout << HC[i][j] << " ";
      }
      //std::cout << "\n";
    }
    combhmi.close();

    //std::cout << "DONE : ########## Reading the HMatrixI files" << std::endl;
    // ############################################################


    /****************************
    std::cout << "\n----------------------------" << std::endl;
    for (int i=0; i<comb_bins; ++i){
      for (int j=0; j<comb_bins; ++j){
	std::cout << HC[i][j] << " ";
      }
      std::cout << "\n";
    }
    std::cout << "\n----------------------------" << std::endl;
    ***************************/



    // setup the matrix means-- these are the mean of the matrix bins that were used to make the matrix 
    // eventually, these should also be read by the matrix file

    // Longitudinal-only means
    //Double_t lmeans2[long_bins] = {.032, .093, .139, .151, .144, .119, .090, .068, .050, .036, .026, .018, .013, .009, .006, .004}; 
    //Double_t lmeans2[15] = {.108, .214, .208, .149, .101, .068, .047, .034, .023, .015, .010, .007, .004, .003, .002};

    /*************
    for(int i=0; i<long_bins; i++){
      lmeans[i]=lmeans2[i];
      lbin[i]=0;
    }
    ************/

    // transverse-only means
    //Double_t trmeans2[trans_bins] = {.0003, .0008, .0090, .1534, .0093, .0008, .0003, .0008, .0027, .0218, .2059, .0219, .0026, .0008, .0018, .0054, .0304, .2072, .0306, .0051, .0018, .0027, .0070, .0315, .1878, .0316, .0072, .0027};

    /*****************
    for(int i=0; i<trans_bins; i++){
      trmeans[i]=trmeans2[i];
      transm[i]=0;
    }
    ************/
      
    // Combined means (Long plus Trans)
    //Double_t ltmeans2[comb_bins] = {.032, .093, .0015, .0123, .1116, .0124, .0014, .0031, .0173, .1107, .0172, .0029, .144, .119, .090, .068, .050, .036, .026, .018, .013, .009, .006, .004};

    /*********************
    for(int i=0; i<comb_bins; i++){
      ltmeans[i]=ltmeans2[i];
      ltbin[i]=0;
    }
    *******************/


    // ############################################################
    //std::cout << "START : ########## Reading the HMatrix Mean files" << std::endl;
    
    longmeans.open(LongMeans.c_str());    
    Double_t in_lmeans;
    if (!longmeans.is_open()){
      throw cet::exception("UnresolvedMatrix")
        << "Can not open Long HMatrix Means file " << LongMeans << "!";      
    }
    for (int i=0; i<long_bins; ++i){
      longmeans >> in_lmeans;
      lmeans[i] = in_lmeans;
      //lbin[i] = 0;
      //std::cout << lmeans[i] << " ";
    }
    //std::cout << "\n";
    longmeans.close();
            
    transmeans.open(TransMeans.c_str());    
    Double_t in_trmeans;
    if (!transmeans.is_open()){
      throw cet::exception("UnresolvedMatrix")
        << "Can not open Trans HMatrix Means file " << TransMeans << "!";      
    }
    for (int i=0; i<trans_bins; ++i){
      transmeans >> in_trmeans;
      trmeans[i] = in_trmeans;
      //transm[i] = 0;
      //std::cout << trmeans[i] << " ";
    }
    //std::cout << "\n";
    transmeans.close();
    
    
    combmeans.open(CombMeans.c_str());    
    Double_t in_combmeans;
    if (!combmeans.is_open()){
      throw cet::exception("UnresolvedMatrix")
        << "Can not open Comb HMatrix Means file " << CombMeans << "!";      
    }
    for (int i=0; i<comb_bins; ++i){
      combmeans >> in_combmeans;
      ltmeans[i] = in_combmeans;
      //ltbin[i] = 0;
      //std::cout << ltmeans[i] << " ";
    }
    //std::cout << "\n";
    combmeans.close();

    /***************
    for ( int i=0; i<long_bins; ++i){
      std::cout << lmeans[i] << " "; 
    }
    std::cout << "\n";
    for ( int i=0; i<trans_bins; ++i){
      std::cout << trmeans[i] << " "; 
    }
    std::cout << "\n";
    for ( int i=0; i<comb_bins; ++i){
      std::cout << ltmeans[i] << " "; 
    }
    std::cout << "\n";
    *****************/
    //std::cout << "DONE : ########## Reading the HMatrix Mean files" << std::endl;
    // ############################################################



    /*************************
    for(int i=0; i<max_layers; i++){
      esumlay[i]=0;
      maxHcell[i]=0;
      
      for(int j=0; j<max_cells; j++){
	eplcell[i][j]=0;
      }
       
    }

    // initialize everything else to zero
    etot = 0.;  
    ttot = 0.; // for tran-only bins
    Tesum = 0.; // total energy sum of hits used in test
    ******************************/



    //nhits_tot=0; 
    fCounter=0;

    //initialize the cutflow bins to zero
    for(int i=0; i<11; i++){
      for(int j=0; j<10; j++){
	pass_cuts[i][j]=0;
      }
    }

  } // end of beginJob loop



  void HMatrixUser::endSubRun(const art::SubRun& sr){

    /****************************************************
    art::Handle<sumdata::POTSum> pot;
    sr.getByLabel("generator", pot);    
    fTotalPOT += pot->totgoodpot;    
    std::cout << "POT in subrun " << sr.subRun() 
	      << " = " << pot->totgoodpot << std::endl;
    ******************************/
  }

  int event_number = 0;
  //double lbin[21];
  //double transm[42];
  //double ltbin[55];

  //-------------------------------------------------------------------------
  void HMatrixUser::analyze(const art::Event& evt)
  {

    event_number++;
    gevent_no = event_number;
    //std::cout << "***** event_number " << event_number << " ************" << std::endl;

    /*********************
    // Reset here all matrix bins to zero for each event ...
    // all total energy, total transverse energy
    for ( int i=0; i<long_bins; ++i){
      lbin[i] = 0.0;
    }
    for ( int i=0; i<trans_bins; ++i){
      transm[i] = 0.0;
    }
    for ( int i=0; i<comb_bins; ++i){
      ltbin[i] = 0.0;
    }

    //etot = 0.0;
    //ttot = 0.0;

    //std::cout << "etot " << etot << "\t ttot " << ttot << std::endl;
    std::cout << "----------------------------" << std::endl;
    for ( int i=0; i<long_bins; ++i){
      std::cout << lbin[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<trans_bins; ++i){
      std::cout << transm[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<comb_bins; ++i){
      std::cout << ltbin[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    ********************/

    // These are the cuts, they can be changed around to play with the cutflow if you like

    double EnergyMin  = 0.5; //GeV
    double EnergyMax  = 3.5; //GeV
    double max_length = 400; //cm
    double MIP_CUT = 0.55;
    int ntracks  = 0;
    int ntracksl = 0;
    int event_type = 0;
    int flux_type = 0;
    double oscw = 1;

    int slice_el      = 0;
    int slice_mu      = 0;
    int slice_pi0     = 0;
    int slice_pion    = 0;
    int slice_neutron = 0;
    int slice_proton  = 0;
    int slice_others  = 0;

    int alg_el      = 0;
    int alg_mu      = 0;
    int alg_pi0     = 0;
    int alg_pion    = 0;
    int alg_neutron = 0;
    int alg_proton  = 0;
    int alg_others  = 0;


    art::ServiceHandle<geo::Geometry> geo;
    art::ServiceHandle<calib::Calibrator> cal;
    
    int ifirst = 1;  // first plane of shower - use this to start binning, it is changed later

    // these vectors will hold the energy, plane and cell, for each hit in the event
    std::vector<double> slicer_energy_vec;
    std::vector<int> slicer_plane_vec;
    std::vector<int> slicer_cell_vec; 
    //this vector holds the first plane in the event 
    //std::vector<double> min_plane_vec;

    std::vector<double> slicer_energy_vec_el;
    std::vector<int> slicer_plane_vec_el;
    std::vector<int> slicer_cell_vec_el; 

    std::vector<double> slicer_energy_vec_pi0;
    std::vector<int> slicer_plane_vec_pi0;
    std::vector<int> slicer_cell_vec_pi0; 


    std::vector<double> alg_energy_vec;
    std::vector<int>    alg_plane_vec;
    std::vector<int>    alg_cell_vec; 
    std::vector<double> alg_energy_vec_el;
    std::vector<int>    alg_plane_vec_el;
    std::vector<int>    alg_cell_vec_el; 
    std::vector<double> alg_energy_vec_pi0;
    std::vector<int>    alg_plane_vec_pi0;
    std::vector<int>    alg_cell_vec_pi0; 

 
    glchi2=0;
    gtchi2=0;
    gcchi2=0;

    /******************
    // Get the HMatrix chi2 from hme object
    art::Handle< std::vector<hme::HMPid> > hme_chi2_list;
    evt.getByLabel("hmatrix", hme_chi2_list);

    //std::cout << "hmatrix size : " << hme_chi2_list->size() << std::endl;
    if (hme_chi2_list->size() != 1) return;
    
    art::Ptr<hme::HMPid> hme_chi2(hme_chi2_list, 0);
    
    glchi2 = hme_chi2->LongChi2();
    gtchi2 = hme_chi2->TransChi2();
    gcchi2 = hme_chi2->CombChi2();
    
    **********************/

    //std::cout << "lchisq_new " << glchi2 << "\t tchisq " << gtchi2 << "\t cchisq " << gcchi2 << std::endl;
        
    //flchisqL->Fill(glchi2);
    //flchisqT->Fill(gtchi2);
    //flchisqC->Fill(gcchi2);
    //flchisqLT->Fill(glchi2,gtchi2);


    /**********************
       MCcheater.. Get Backtracker information
    *********************/
    art::ServiceHandle<cheat::BackTracker> mc_backtracker;

    /**********************
    MCTruth : Get event information.
              1:nue CC, 2:nuMu CC, 3:NC, 4:beam nue CC
    *********************/
    
    art::Handle< std::vector<simb::MCTruth> > mc_truth_list;
    evt.getByLabel("generator", mc_truth_list);
    
    //std::cout << "size of mc_truth_list : " << mc_truth_list->size() << std::endl;
    // the list size is one... you can use the 0 the entry w/o loop over the list
    if (mc_truth_list->size() != 1) return;
    
    //for (unsigned int i=0; i<mc_truth_list->size(); ++i){
    
    art::Ptr<simb::MCTruth> mc_truth(mc_truth_list, 0);
    
    //std::cout << "# of particles in the event [MCTruth] " << mc_truth->NParticles() << std::endl;    
    // this also adds the nucleus pdg code....
    //for (int i = 0; i<mc_truth->NParticles(); ++i){
      //std::cout << "particle  " << i << "  pdg code : " << mc_truth->GetParticle(i).PdgCode() << std::endl;
      //std::cout << "pdg code " << mc_truth->GetParticle(i).PdgCode() << "\t track id " << mc_truth->GetParticle(i).TrackId() << std::endl;
    //}
    
    
    /*******************************************
      if (mc_truth->Origin()==simb::kBeamNeutrino){
	std::cout << " Origin is beam neutrino" << std::endl;
      }else if (mc_truth->Origin()==simb::kCosmicRay){
	std::cout << "Origin is cosmic ray" << std::endl;
      }else if (mc_truth->Origin()==simb::kSuperNovaNeutrino){
	std::cout << "Origin is Supernova Neutrino" << std::endl;
      }else if (mc_truth->Origin()==simb::kSingleParticle){
	std::cout << " Origin is single particle" << std::endl;
	// single particles thrown at the detector....
      }else{
	std::cout << " unknown origin " << std::endl;
      }
    *****************************************/

    // there are no neutrinos in single particles.    
    if (mc_truth->Origin()==simb::kSingleParticle){
      // for single particles, there is only one in the particle list....
      // e_minus = 11, mu_minus = 13, photon = 22, pi0 = 111
      //std::cout << "particles " << mc_truth->GetParticle(0).PdgCode() << " \t energy " << mc_truth->GetParticle(0).E() << std::endl;
      gin_nu_e = mc_truth->GetParticle(0).E();
      
    }
    
    
    //const simb::MCNeutrino &mc_neutrino = mc_truth->GetNeutrino(); // incoming neutrino
    // .Nu() --> reference to MCParticle
    // CCNC : Neutrino interaction category : CC or NC
    // CCNC = 0 or charged current, = 1 or neutral current
    
    if(mc_truth->Origin()==simb::kBeamNeutrino){
      
      
      /**********************
           MCFlux : Get flux information.
      *********************/
      
      // found zero products matching this criteria
      // MCFlux part is moved here, because for single particles, there is no MCFlux
      art::Handle< std::vector<simb::MCFlux> > mc_flux_list;
      evt.getByLabel("generator", mc_flux_list);
      //std::cout << "size of mc_flux_list : " << mc_flux_list->size() << std::endl;
      // the size of the mc flux list is always one......
      
      //if (mc_flux_list->size() != 1) return;
      art::Ptr<simb::MCFlux> mc_flux(mc_flux_list, 0);
      //std::cout << "mc_flux type : " << mc_flux->fntype << std::endl;
      flux_type = mc_flux->fntype;
      gflux_type = flux_type;
      
      const simb::MCNeutrino &mc_neutrino = mc_truth->GetNeutrino(); // incoming neutrino
      gin_nu_e = mc_neutrino.Nu().E();
      
      //std::cout << " mode : " << mc_neutrino.Mode() << " interaction : " << mc_neutrino.InteractionType() << std::endl;
      //std::cout << "neutrino pdgcode : " << mc_neutrino.Nu().PdgCode() << " energy : " << mc_neutrino.Nu().E() << " CCNC : " << mc_neutrino.CCNC() << std::endl; 
      
      
      
      /******************************
       // this part of the code can be replaced by GetEventType() function ...
       // 1:nue CC, 2:NuMu CC, 3:NC, 4:others
       // nue CC includes beam nue CC	
	if (mc_neutrino.CCNC() == simb::kNC){ // or 1
	  event_type_old = 3; // NC : neutral current
	}else{
	  
	  if (fabs(mc_neutrino.Nu().PdgCode())==14){
	    event_type_old = 2; // 14 is for nu_mu : NuMu CC
	  }else if (fabs(mc_neutrino.Nu().PdgCode())==12){
	  ////event_type = 1; // 12 is for nu_e : Nue CC including beam nue CC
	    if(fabs(mc_flux->fntype)==12){ // if flux is nu_e
	      event_type_old = 4; // beam nue CC
	    }else{
	      event_type_old = 1; // Nue CC
	    }
	  }else{
	    event_type_old = 5; // others
	  }
	  	
	}
      ***********************/
      
	//const simb::MCParticle &mc_lepton = mc_neutrino.Lepton(); // outgoing lepton    
	//std::cout << "lepton pdgcode : " << mc_lepton.PdgCode() << " mass : " << mc_lepton.Mass() << " energy : " << mc_lepton.E() << std::endl;
      
	oscw = GetOscWeight(mc_neutrino, mc_flux);
	gosc_weight = oscw;

	event_type = GetEventType(mc_neutrino, mc_flux);
	//std::cout << "event_type : " << event_type << "\t osc : " << gosc_weight << std::endl;
	gevent_type = event_type;
      
      
    } // end of beamneutrino
    
    
    
      /////} // loop over mc_truth_list
    
    //gevent_type = event_type;
    //std::cout << "event_type : " << event_type << std::endl;

    pass_cuts[event_type][0]++;



    // PartilceNavigator gives same particles as from geantgen .........
    //const sim::ParticleNavigator &pnav = mc_backtracker->ParticleNavigator();
    //std::cout << "mc_backtracker : " << pnav.size() << std::endl;
    // particles generated during event simulation....
    
    // It gives all particles in an event, including secondaries....
    //for (unsigned int i=0; i<pnav.size(); ++i){
      //if (pnav.Particle(i)->PdgCode()>10000) continue;
      //std::cout << "pdg code " << pnav.Particle(i)->PdgCode() << "\t trackid " << pnav.Particle(i)->TrackId() << std::endl;
      //gpdg = pnav.Particle(i)->PdgCode();
    //}
    
    //gpdg=0;
    //std::cout << "mc_backtracker2 : " << pnav.NumberOfPrimaries() << std::endl;
    //for (int i=0; i<pnav.NumberOfPrimaries(); ++i){
      
    //const sim::Particle* gen_particle = pnav.Primary(i);
      //std::cout << "pdg code2 " << gen_particle->PdgCode() << "\t trackid " << gen_particle->TrackId() << std::endl;
      
      //int gt_el, gt_mu, gt_elnu, gt_munu, gt_pion, gt_pi0, gt_gamma;
      
      //gpdg = gen_particle->PdgCode();
      
      //std::cout << "pdg code2 " << gpdg << std::endl;

      //std::cout << "event type  " << gevent_type << " \t pdg code2 " << gpdg << std::endl;

      //partvar->Fill();

      /***********************************
      if(fabs(gpdg)==11){++gt_el;} // el
      else if(fabs(gpdg)==12){++gt_el_nu;} // el_nu
      else if(fabs(gpdg)==13){++gt_mu;} // mu
      else if(fabs(gpdg)==14){++gt_mu_nu;} // mu_nu
      else if(fabs(gpdg)==22){++gt_photon;} // photon
      else if(fabs(gpdg)==111){++gt_pi0;} // pi_zero
      else if(fabs(gpdg)==211){++gt_pion;} // pi_plus
      else if(fabs(gpdg)==2112){++gt_neutron;} // neutron
      else if(fabs(gpdg)==2212){++gt_proton;} // proton
      else{
	std::cout << "pdg code2 " << gen_particle->PdgCode() << std::endl;
      }
      **********************************/

    //}

    /**********************
        Get Vector of all particles in an event...........
      This is same as the list from particleNavigator ......
    *********************/
    /************************
    art::Handle< std::vector<sim::Particle> > mc_particle_handle;
    evt.getByLabel("geantgen", mc_particle_handle); // Geant produced MC

    std::cout << "size of mc_particle_list : " << mc_particle_handle->size() << std::endl;

    art::PtrVector<sim::Particle> mc_particle_list;
    for (unsigned int i=0; i<mc_particle_handle->size(); ++i){
      art::Ptr<sim::Particle> mc_particle(mc_particle_handle, i);
      mc_particle_list.push_back(mc_particle);

      //particle_list.push_back(art::Ptr<sim::Particle>(mc_particle_handle, i));
      //std::cout << "pdg code " << mc_particle->PdgCode() << "\t track id " << mc_particle->TrackId() << std::endl;
    }
    *****************************/
    // The track ID number assigned by the Monte Carlo.  This will be
    // unique for each Particle in an event. - 0 for primary particles

    //std::cout << "size of mc_particle_list222 : " << mc_particle_list->size() << std::endl;



    /**********************
    Get hits from slicer.
    *********************/

    art::Handle<std::vector<rb::CellHit> > all_calhits; // collect all cell_hits
    evt.getByLabel("calhit", all_calhits);
    art::PtrVector<rb::CellHit> all_hits;

    // cal_hits = sum of hits in slice + noise slice
    //std::cout << "all_calhits size " << all_calhits->size() << std::endl;
    for(unsigned int i=0; i<all_calhits->size(); ++i){
      all_hits.push_back(art::Ptr<rb::CellHit>(all_calhits, i));
    }


    art::Handle< std::vector<rb::Cluster> > cluster_list;
    //evt.getByLabel("slicer", cluster_list);
    evt.getByLabel(fSliceLabel, cluster_list);
    //int cluster_size = cluster_list->size();
    
    std::cout << "Analyzer : slicer is called : " << fSliceLabel.c_str() << std::endl;
    //std::cout << "no of clusters " << cluster_list->size() << std::endl;
    
    
    //if (!cluster_list->size()) return;
    if (cluster_list->size()<2) return;
    //added this to avoid problem when there is only noise slice
    
    // Find the cluster having largest number of hits ........
    art::Ptr<rb::Cluster> largest_cluster;
    unsigned int ncell_largest_cluster = 0;
    int best_slice_index = -1;

    for (unsigned int sliceIdx=0; sliceIdx<cluster_list->size(); ++sliceIdx){
      art::Ptr<rb::Cluster> cluster(cluster_list, sliceIdx);
      //std::cout << "cluster " << sliceIdx << "\t" << cluster->NCell() << std::endl; 
      
      if (cluster->IsNoise()==true) continue; // we don't want the noise slice...
      
      if (cluster->NCell()>ncell_largest_cluster){
	ncell_largest_cluster = cluster->NCell();
	largest_cluster       =  cluster;
	best_slice_index      = sliceIdx;
      }

      //std::cout << "cluster_ " << sliceIdx << "\t" << cluster->NCell() << std::endl; 
    }

    //std::cout << "cell_hits in the largest_cluster " << best_slice_index << "\t" << largest_cluster->NCell() << std::endl; 

    if (best_slice_index<0) return;

    gnslice = largest_cluster->NCell();

    //std::cout << "ADC "       << largest_cluster->TotalADC() << "\t PE "     << largest_cluster->TotalPE() << "\t PECorr " << largest_cluster->TotalPECorr() << "\t GeV "    << largest_cluster->TotalGeV() << std::endl; 

    // get truth information from the slice using backtracker ...
    //art::PtrVector< rb::CellHit > sliceHits = largest_cluster->AllCells();
    //std::vector<cheat::NeutrinoEffPur> neu_eff_pur = mc_backtracker->SliceToMCTruth(largest_cluster, all_hits);
    //std::cout << "eff : " << neu_eff_pur[0].efficiency
    //      << "\t purity : " << neu_eff_pur[0].purity << std::endl;


    // wait
    // call different PIDs.......
    double rvp = -1;
    double eid = -1;

    if(mc_truth->Origin()==simb::kBeamNeutrino){
      art::FindManyP<rvp::RVP> fmRvp(cluster_list, evt, "rvp");
      std::vector<art::Ptr<rvp::RVP>> rvps = fmRvp.at(best_slice_index);
      if(rvps.size()>0){
	rvp = rvps[0]->Value();
      }
      
      art::FindManyP<jmshower::JMShower> fmShower(cluster_list, evt, "recojmshower");
      //grab list of showers associated with best slice
      std::vector<art::Ptr<jmshower::JMShower> > jmshower_list = fmShower.at(best_slice_index);
      //std::cout << "no of jmshowers " << jmshower_list.size() << std::endl;

      double ANN_max = -1; 
      //int best_jmshower_index = -1;
      for (unsigned int jmsIdx=0; jmsIdx<jmshower_list.size(); ++jmsIdx){
	//std::cout << "jmshower " << jmsIdx << "\t" << jmshower_list[jmsIdx]->ANN(0) << std::endl;
	
	if (jmshower_list[jmsIdx]->ANN(0)>ANN_max){
	  ANN_max = jmshower_list[jmsIdx]->ANN(0);
	  //best_jmshower_index  = jmsIdx;
	}
	
      }
      eid = ANN_max;
      
      //std::cout << "best_jmshower " << best_jmshower_index << " eid   " << eid << std::endl;      
    }
    
    grvp = rvp;
    geid = eid;
    //std::cout << "rvp " << rvp << "\t eid " << eid << std::endl;


    /*************************
    for (unsigned int i=0; i<pnav.size(); ++i){
      sim::Particle* pn_part = pnav.Particle(i);
      if (pn_part->PdgCode()>10000) continue;
      //std::cout << "pdg code " << pnav.Particle(i)->PdgCode() << "\t trackid " << pnav.Particle(i)->TrackId() << std::endl;
      gpdg = pn_part->PdgCode();
      int mo_track_id = pn_part->Mother();
      gmopdg = -1;
      if(mo_track_id>0){
	gmopdg = pnav[pn_part->Mother()]->PdgCode();
      }
      //std::cout << "evt " << gevent_type << "  eid  " << geid << "  grvp  " << grvp << "  pdg code " << gpdg << "\t " << pn_part->TrackId() << "\t" <<mo_track_id << "\t" << gmopdg << std::endl;

    }
    ***************/


    const double wx = (largest_cluster->NXCell() > 0) ? largest_cluster->W(largest_cluster->XCell(0).get()) : 0;
    const double wy = (largest_cluster->NYCell() > 0) ? largest_cluster->W(largest_cluster->YCell(0).get()) : 0;

    for(unsigned int j=0; j<largest_cluster->NCell(); ++j){

      const art::Ptr<rb::CellHit>& cellhit = largest_cluster->Cell(j);


      // Get Recohit : makes ure the hit is calibrated :
      //rb::RecoHit rhit = largest_cluster->RecoHit(j);
      const rb::RecoHit recohit = cal->MakeRecoHit(*cellhit, cellhit->View() == geo::kX ? wx : wy);

      if(!recohit.IsCalibrated()){
	std::cerr << "Not calibrated?!" << std::endl;
	continue;
      }



      /*********************
      //double drift_scale = cal->GetDriftScale(cellhit.get());
      //double MIP_scale = cal->GetMIPScale(cellhit.get()); // PECorr->MIP conversion
      //double GeV_scale = cal->GetGeVScale(cellhit.get()); // MIP->GeV conversion
      
      //std::cout << "drift " << MIP_scale << "\t" << GeV_scale << std::endl;
      

      double energy = cellhit->PE();
      int plane     = cellhit->Plane();
      int cell      = cellhit->Cell();
      
      int view = 0;
      if (cellhit->View() == geo::kX){
	view = 1; // X view is odd planes
      }
      **********************/
      
      int plane     = cellhit->Plane();
      int cell      = cellhit->Cell();
      double recoPECorr = recohit.PECorr();

      //double recoMIP    = recohit.MIP();
      //double recoE      = recohit.GeV();
      
      //std::cout << plane << "\t" << cell << "\t" << energy << "\t" << recoPECorr << "\t" << recoMIP << "\t \t" << recoE << std::endl; 

      //std::cout << plane << "\t" << cell << "\t" << recoPECorr << "\t" << recoMIP << "\t \t" << recoE << std::endl; 


      // ***** Get different particle types from cell_hit ..............
      // Returns the sim::Particle object contributing the most light to the specified rb::CellHit
      // Go through the map and find the particle ID that contributed the most light

      const sim::Particle* truth_particle = mc_backtracker->HitToParticle(cellhit, 1);

      ///No photons in this cell are in time with the hit. Returning null pointer
      // if this is a null, add the hit to others category...
      if (truth_particle == NULL){
	slice_others++;
	continue;
      }
      // backtracker seems doesn't have trackid = 0..
      int motherid = -1; 
      motherid = truth_particle->Mother(); // track id of the mother

      //std::cout << j << "\t truth pdg " << truth_particle->PdgCode() << "\t trackid " << truth_particle->TrackId() << "\t mo trackid " << motherid << std::endl;
      
      if (fabs(truth_particle->PdgCode())==11){ // electron
	slice_el++;

	slicer_energy_vec_el.push_back(recoPECorr);
	slicer_plane_vec_el.push_back(plane);
	slicer_cell_vec_el.push_back(cell);
      }
      else if (fabs(truth_particle->PdgCode())==13){ // muon
	slice_mu++;
      }
      else if (fabs(truth_particle->PdgCode())==211){ // pion
	slice_pion++;
      }
      else if (fabs(truth_particle->PdgCode())==2112){ // neutron
	slice_neutron++;
      }
      else if (fabs(truth_particle->PdgCode())==2212){ // proton
	slice_proton++;
      }
      else if (fabs(truth_particle->PdgCode())==22 && motherid>0){
	// backtracker seems doesn't have trackid = 0 .....
	const sim::Particle* truth_mother = mc_backtracker->TrackIDToParticle(motherid);
 	if (fabs(truth_mother->PdgCode())==111){
	  slice_pi0++;

	  slicer_energy_vec_pi0.push_back(recoPECorr);
	  slicer_plane_vec_pi0.push_back(plane);
	  slicer_cell_vec_pi0.push_back(cell);
	}
 	if (fabs(truth_mother->PdgCode())!=111) slice_others++;	
      }
      else{
	slice_others++;
      }

      //const sim::Particle* truth_mother = mc_backtracker->TrackIDToParticle(motherid);	
      //std::cout << "truth pdg " << truth_particle->PdgCode() << "\t trackid " << truth_particle->TrackId() << "\t mo trackid " << truth_mother->TrackId() << "\t mo pdg " << truth_mother->PdgCode() << std::endl;

      slicer_energy_vec.push_back(recoPECorr);
      slicer_plane_vec.push_back(plane);
      slicer_cell_vec.push_back(cell);


      /***********
      double denergy = 0.0;
      const std::vector<sim::FLSHit> fhits = mc_backtracker->HitToFLSHit(cellhit);
      for (unsigned int fls=0; fls<fhits.size(); ++fls){
	const sim::FLSHit& fhit = fhits[fls];
	//denergy += fhits[fls].GetEdep();
	denergy += fhit.fEdep;
      }

      //double efrac = 0.0;
      //if(recoE) efrac = (recoE-denergy)/recoE;
      //std::cout << "reco " << recoE << "\t FLS energy " << denergy << "\t frac " << efrac << std::endl;
      //slicer_energy_vec.push_back(denergy);

      ************/

      //std::cout << "recohit " << j << "\t" << plane << "\t" << cell << "\t" << recoPECorr << std::endl; 
    

    } // loop over all cell_hits in the best slicer
    

    gt_slice_el      = slice_el;
    gt_slice_mu      = slice_mu;
    gt_slice_pion    = slice_pion;
    gt_slice_pi0     = slice_pi0;
    gt_slice_proton  = slice_proton;
    gt_slice_neutron = slice_neutron;
    gt_slice_other   = slice_others;


    //int slice_hits_total = 0;
    //slice_hits_total = slice_el + slice_mu + slice_pion + slice_neutron + slice_proton + slice_pi0 + slice_others;
    //std::cout << "el " << slice_el << "\t mu " << slice_mu << "\t pion " << slice_pion << "\t neutron " << slice_neutron << "\t proton " << slice_proton << "\t pi0 " << slice_pi0 << "\t others " << slice_others << "\t total hits " << slice_hits_total << std::endl;



    /**********************
         Getting Tracks ......
     The module is designed to use any of the tracking algorithms as input :
     kalmantrackmerge, discretetrackmerge, fuzzykvertex, trackmerge
    *********************/
    art::Handle< std::vector<rb::Track> > track_list;
    
    if (fTrackLabel=="fuzzykvertex"){
      evt.getByLabel(fTrackLabel, "Tracks3D", track_list); // fuzzykvertex, gives 3D tracks
      //std::cout << "Analyzer : fuzzykvertex is called : " << fTrackLabel.c_str() << std::endl;
    }else{
      evt.getByLabel(fTrackLabel,track_list);
      //std::cout << "Analyzer : tracking algorithm called : " << fTrackLabel.c_str() << std::endl;
    }
    std::cout << "Analyzer : tracking algorithm used : " << fTrackLabel.c_str() << std::endl;
           
    //std::cout << "track.size() " << track_list->size() << std::endl;
    
    if (!track_list->size()) return;
    // skip events having no tracks, otherwise Module failed due to an exception
    // if there are no object, no ntuple is filled up............

    //the longest track
    art::Ptr<rb::Track> longest_track;
    double tot_energy_event = 0;
    double tot_dep_energy_event = 0;
    int tot_cells_event = 0;
    double length_longest_track = 0;

    for(unsigned int i=0; i<track_list->size(); i++){
      
      art::Ptr<rb::Track> track(track_list,i);
      
      if(track->IsNoise()==true)
	continue;
     
      //identify the longest track
      double length = track->TotalLength();
      if(length>length_longest_track){
	ntracksl++;
	longest_track = track;
	length_longest_track = length;
	if (ntracksl>1) ntracksl--;
      }

      //identify the total number of cells in the event
      int ncells=track->NCell();
      //fnhits_per_track->Fill(ncells);


      //std::cout << "track " << i << "\t" << length << "\t" << ncells << std::endl;

      /***********
      for(int j=0; j<ncells; j++){
	art::Ptr<rb::CellHit> cellhit = track->Cell(j);
      
	double energy = cellhit->PE();
	int plane     = cellhit->Plane();
	int cell      = cellhit->Cell();
	
	//std::cout << "cellhit " << j << "\t" << plane << "\t" << cell << "\t" << energy << std::endl; 

      }
      ***********/

      /******************************
      if(event_type==1)
	fnhits_per_track_nc->Fill(ncells);
      if(event_type==2)
	fnhits_per_track_numu->Fill(ncells);
      if(event_type==3)
	fnhits_per_track_nue->Fill(ncells);
      ******************/

      //if(CalculateMIPPeak(track)<MIP_CUT && isInVertexFiducial(track)==true && length<=max_length && ncells>10)

      ntracks++;

      // identify the total energy
      for(int j=0; j<ncells; j++){
	art::Ptr<rb::CellHit> cellhit=track->Cell(j);
	rb::RecoHit rhit = cal->MakeRecoHit(*cellhit, track->W(cellhit.get()));

	
	if(!rhit.IsCalibrated()){
	  std::cerr << "Not calibrated?!" << std::endl;
	  continue;
	}

	
	// wait : this is for test purpose HMatrix input.... all hits are includes..no longest track
	int plane     = cellhit->Plane();
	int cell      = cellhit->Cell();
	double recoPECorr = rhit.PECorr();

	// ***** Get different particle types from cell_hit ..............
	// Returns the sim::Particle object contributing the most light to the specified rb::CellHit
	// Go through the map and find the particle ID that contributed the most light
	
	const sim::Particle* truth_particle = mc_backtracker->HitToParticle(cellhit, 1);
	
	///No photons in this cell are in time with the hit. Returning null pointer
	// if this is a null, add the hit to others category...
	if (truth_particle == NULL){
	  alg_others++;
	  continue;
	}
	// backtracker seems doesn't have trackid = 0..
	int motherid = -1; 
	motherid = truth_particle->Mother(); // track id of the mother
	
	//double true_particle_energy = truth_particle->E();
	
	//std::cout << j << "\t evt " << event_type << "\t truth pdg " << truth_particle->PdgCode() << "\t trackid " << truth_particle->TrackId() << "\t mo trackid " << motherid << std::endl;
	
	if (fabs(truth_particle->PdgCode())==11){ // electron
	  alg_el++;

	  if (1.0<truth_particle->E() && truth_particle->E()<3.0){
	    //std::cout << " el_E " << truth_particle->E() << std::endl;
	    alg_energy_vec_el.push_back(recoPECorr);
	    alg_plane_vec_el.push_back(plane);
	    alg_cell_vec_el.push_back(cell);
	  }

	}
	else if (fabs(truth_particle->PdgCode())==13){ // muon
	  alg_mu++;
	}
	else if (fabs(truth_particle->PdgCode())==211){ // pion
	  alg_pion++;
	}
	else if (fabs(truth_particle->PdgCode())==2112){ // neutron
	  alg_neutron++;
	}
	else if (fabs(truth_particle->PdgCode())==2212){ // proton
	  alg_proton++;
	}
	else if (fabs(truth_particle->PdgCode())==22 && motherid>0){
	  // backtracker seems doesn't have trackid = 0 .....
	  const sim::Particle* truth_mother = mc_backtracker->TrackIDToParticle(motherid);
	  if (fabs(truth_mother->PdgCode())==111){
	    alg_pi0++;
	    //std::cout << "evt_type " << event_type << "\t truth pdg " << truth_particle->PdgCode() << "\t trackid " << truth_particle->TrackId() << "\t mo trackid " << truth_mother->TrackId() << "\t mo pdg " << truth_mother->PdgCode() << std::endl;	    

	    if (1.0<truth_mother->E() && truth_mother->E()<3.0){
	      //std::cout << "gamma " << truth_particle->E() << "\t pi_zero " << truth_mother->E() << std::endl;
	      alg_energy_vec_pi0.push_back(recoPECorr);
	      alg_plane_vec_pi0.push_back(plane);
	      alg_cell_vec_pi0.push_back(cell);
	    }

	  }
	  if (fabs(truth_mother->PdgCode())!=111) alg_others++;	
	}
	else{
	  alg_others++;
	}
	
	//const sim::Particle* truth_mother = mc_backtracker->TrackIDToParticle(motherid);	
	//std::cout << "truth pdg " << truth_particle->PdgCode() << "\t trackid " << truth_particle->TrackId() << "\t mo trackid " << truth_mother->TrackId() << "\t mo pdg " << truth_mother->PdgCode() << std::endl;
	
	alg_energy_vec.push_back(recoPECorr);
	alg_plane_vec.push_back(plane);
	alg_cell_vec.push_back(cell);
	
	
		
	//energy_vec.push_back(recoPECorr);
	//plane_vec.push_back(plane);
	//cell_vec.push_back(cell);

	double denergy = 0.0;
	const std::vector<sim::FLSHit> fhits = mc_backtracker->HitToFLSHit(cellhit);
	for (unsigned int fls=0; fls<fhits.size(); ++fls){
	  //const sim::FLSHit& fhit = fhits[fls];
	  denergy += fhits[fls].GetEdep();
	  //denergy += fhit.fEdep;
	}


	/**********************************
        // testing purpose
	int view = 0;
	if (cellhit->View() == geo::kX){
	  view = 1; // X view is odd planes
	}
	double x_pos = rhit.X();
	double y_pos = rhit.Y();
	double z_pos = rhit.Z();
	std::cout << plane << "\t" << view << "\t" << cell << "\t" << x_pos << "\t" << y_pos << "\t" << z_pos  << std::endl; 
	******************************/

	double energy = rhit.GeV();

	/************************
	double energy_mev=2*energy*1000; //there is a factor of two fudge factor in the S12.06.17
	fEnergy_hit->Fill(energy_mev);
       
	if(event_type==1)
	  fEnergy_hit_nc->Fill(energy_mev);
	if(event_type==2)
	  fEnergy_hit_numu->Fill(energy_mev);
	if(event_type==3)
	  fEnergy_hit_nue->Fill(energy_mev);
	*********************/

	tot_energy_event=tot_energy_event+energy;
	tot_dep_energy_event = tot_dep_energy_event+denergy;

      } // loop over all cell hits
      
      tot_cells_event=tot_cells_event+ncells;

    } // loop over track_list
    


    gt_alg_el      = alg_el;
    gt_alg_mu      = alg_mu;
    gt_alg_pion    = alg_pion;
    gt_alg_pi0     = alg_pi0;
    gt_alg_proton  = alg_proton;
    gt_alg_neutron = alg_neutron;
    gt_alg_other   = alg_others;
    
    //std::cout << "number of hits from tracks " << tot_cells_event << std::endl;    
    //int alg_hits_total = 0;
    //alg_hits_total = gt_alg_el + gt_alg_mu + gt_alg_pion + gt_alg_neutron + gt_alg_proton + gt_alg_pi0 + gt_alg_other;
    //std::cout << "el " << gt_alg_el << "\t mu " << gt_alg_mu << "\t pion " << gt_alg_pion << "\t neutron " << gt_alg_neutron << "\t proton " << gt_alg_proton << "\t pi0 " << gt_alg_pi0 << "\t others " << gt_alg_other << "\t total hits " << alg_hits_total << std::endl;
    
    //std::cout << "number of tracks " << ntracks << std::endl;
    //std::cout << "Longest tracks " << ntracksl << std::endl;

    int ncells_long_track = longest_track->NCell();


    //std::cout << "longest track " << longest_track->TotalLength() << "\t" << ncells_long_track << std::endl;


    //the MIP Peak is the of cell hits with an energy in the region identified as the MIP Peak 
    // (ie the energy for a minimum ionizing cell hit). It is a number between 0 and 1, tending to be larger 
    // for muons as most hits are minimum ionizing

    double MIP_PEAK = CalculateMIPPeak(longest_track);


    //fill some histograms. Total energy (in GeV), length of the longest track, and MIP Peak
    //ftot_energy_event->Fill(tot_energy_event);
    //flength_longest_track->Fill(length_longest_track);
    //fmip_peak->Fill(MIP_PEAK);
    //fntracks->Fill(ntracks);

    //double efrac_event = -2.0;
    //if(gin_nu_e) efrac_event = (tot_energy_event-gin_nu_e)/gin_nu_e; // denominator should be the true energy

    //double rhit = -2.0;
    //if(gnslice) rhit = (gnhits/gnslice);

    // assign the global tree variables.......
    gtotE    = tot_energy_event;
    gtotE2   = 2.0*tot_energy_event;
    gdepE    = tot_dep_energy_event;
    //////gefrac_event = efrac_event;
    gtrkl    = length_longest_track;
    gmippeak = MIP_PEAK;
    gntrks   = ntracks;
    gntrksl   = ntracksl;
    //gnhits = ncells_long_track; // from longest_track

    gnhits = tot_cells_event; // for whole event
    //////grhit = rhit;

    //std::cout << "slice " << gnslice << " hit " << gnhits << "   hit fraction   " << grhit << std::endl;

    /********************************
    if(event_type==1){
      ftot_energy_event_nc->Fill(tot_energy_event);
      flength_longest_track_nc->Fill(length_longest_track);
      fmip_peak_nc->Fill(MIP_PEAK);
      fntracks_nc->Fill(ntracks);
    }
    if(event_type==2){
      ftot_energy_event_numu->Fill(tot_energy_event);
      flength_longest_track_numu->Fill(length_longest_track);
      fmip_peak_numu->Fill(MIP_PEAK);
      fntracks_numu->Fill(ntracks);
    }
    if(event_type==3){
      ftot_energy_event_nue->Fill(tot_energy_event);
      flength_longest_track_nue->Fill(length_longest_track);
      fmip_peak_nue->Fill(MIP_PEAK);
      fntracks_nue->Fill(ntracks);
    }
    ************************/


    
    //Energy Cut
    pass_cuts[event_type][1]++;
    pass_cuts[event_type][2]++;
    bool PassCuts = true;
    bool PassCuts_initial = true;

    // Aug 15 : PassCuts_inital is commented... events doesnot pass this condition
    // H-matrix chi2 is not filled up...

    // if(PassCuts==true){
//       //vertex fiducial cut on longest track
//       if(HMatrixUser::isInVertexFiducial(longest_track)==false){
// 	PassCuts=false;
// 	PassCuts_initial=false;
      
//       }else{
     
// 	pass_cuts[event_type][2]++;
//       }
//     }
    
    if(PassCuts==true){
      //Energy Cuts
      if(tot_energy_event<EnergyMin || tot_energy_event>EnergyMax){
	PassCuts=false;
	//PassCuts_initial=false; // 
      }else
	pass_cuts[event_type][3]++;
    }
    
    if(PassCuts==true){
      //cut on the length of the longest track
      if(length_longest_track>max_length){
	PassCuts=false;
	//PassCuts_initial=false; // 
      }else
	pass_cuts[event_type][4]++;
    }
    
    
    if(PassCuts==true){
      if(MIP_PEAK>=MIP_CUT)
	PassCuts=false;
      else
	pass_cuts[event_type][5]++;
    }
    if(PassCuts==true){
      if(ncells_long_track<=10)
	PassCuts=false;
      else
	pass_cuts[event_type][6]++;
    }
    if(PassCuts==true){
      if(ntracks>1)
	PassCuts=false;
      else
	pass_cuts[event_type][7]++;
    }
     

    // calculate the mip peak for the longest track
    
    // use only the longest track to make the h-matrix
    if(PassCuts_initial==true){

      // loop over all hits ......
      for(int j=0; j<ncells_long_track; j++){

	// get cell hit
	art::Ptr<rb::CellHit> cellhit = longest_track->Cell(j);
      
	//double energy = cellhit->PE();
	//int plane     = cellhit->Plane();
	//int cell      = cellhit->Cell();

	//std::cout << "cellhit " << j << "\t" << plane << "\t" << cell << "\t" << energy << std::endl; 

	//energy_vec.push_back(energy);
	//plane_vec.push_back(plane);
	//cell_vec.push_back(cell);

	// get reconstructed hit from cellhit and track
	rb::RecoHit recohit = cal->MakeRecoHit(*cellhit, longest_track->W(cellhit.get()));

	// make sure the hit is calibrated
	if(!recohit.IsCalibrated()){
	  std::cerr << "Not calibrated?!" << std::endl;
	  continue;
	}

	//double drift_scale = cal->GetDriftScale(cellhit.get());
	//double MIP_scale = cal->GetMIPScale(cellhit.get()); // PECorr->MIP conversion
	//double GeV_scale = cal->GetGeVScale(cellhit.get()); // MIP->GeV conversion

	//std::cout << "drift " << MIP_scale << "\t" << GeV_scale << std::endl;


	double energy = cellhit->PE();
	int plane     = cellhit->Plane();
	int cell      = cellhit->Cell();

	int view = 0;
	if (cellhit->View() == geo::kX){
	  view = 1; // X view is odd planes
	}
	
	double recoPECorr = recohit.PECorr();
	double recoMIP    = recohit.MIP();
	double recoE      = recohit.GeV();
	
	double x_pos = recohit.X();
	double y_pos = recohit.Y();
	double z_pos = recohit.Z();

	double dist_to_APD = 0;
	if (cellhit->View() == geo::kX){
	  dist_to_APD = recohit.Y();
	}else if (cellhit->View() == geo::kY){
	  dist_to_APD = recohit.X();
	}


	double denergy = 0.0;
	const std::vector<sim::FLSHit> fhits = mc_backtracker->HitToFLSHit(cellhit);
	for (unsigned int fls=0; fls<fhits.size(); ++fls){
	  //const sim::FLSHit& fhit = fhits[fls];
	  denergy += fhits[fls].GetEdep();
	  //denergy += fhit.fEdep;
	}
	double efrac = 0.0;
	if(denergy) efrac = (recoE-denergy)/denergy; // denominator should be the true energy
	//std::cout << "reco " << recoE << "\t FLS energy " << denergy << "\t frac " << efrac << std::endl;


	// wait : PECorr is calculated instead of PE for HMatrix
	//energy_vec.push_back(recoPECorr);
	//plane_vec.push_back(plane);
	//cell_vec.push_back(cell);

	gcellhit_1  = j;
	gview_1     = view;
	gx_pos_1    = x_pos;
	gy_pos_1    = y_pos;
	gz_pos_1    = z_pos;
	gdist_to_APD = dist_to_APD;
	gplane_1    = plane;
	gcell_1     = cell;
	gPE_1       = energy;
	gPECorr_1   = recoPECorr;
	gMIP_1      = recoMIP;
	gGeV_1      = recoE;
	gedep_1     = denergy;
	gefrac_1    = efrac;
	// PE --> PECorr ---> MIP ---> GeV
	gPECorr_PE_1  = gPECorr_1/gPE_1;
	gMIP_PECorr_1 = gMIP_1/gPECorr_1;
	gGeV_MIP_1    = gGeV_1/gMIP_1;
	gGeV_PECorr_1 = gGeV_1/gPECorr_1;
	gPE_PECorr_1  = gPE_1/gPECorr_1;

	tree_1->Fill();

	//std::cout << plane << "\t" << view << "\t" << cell << "\t" << x_pos << "\t" << y_pos << "\t" << z_pos << "\t" << dist_to_APD << std::endl; 

	//std::cout << plane << "\t" << view << "\t" << cell << "\t" << x_pos << "\t" << y_pos << "\t" << z_pos  << std::endl; 

	//std::cout << plane << "\t" << view << "\t" << cell << "\t" << energy << "\t" << recoPECorr << "\t" << recoMIP << "\t \t" << recoE << std::endl; 

	//std::cout << "recohit " << j << "\t" << plane << "\t" << cell << "\t" << recoPECorr << std::endl; 


      }

      // ifirst is filled with the first plane number, it is again changed later...
      //ifirst = longest_track->MinPlane();
      //min_plane_vec.push_back(ifirst);
    
      //std::cout << "first plane " << ifirst << std::endl; 


      //if (event_number!=14 || event_number!=17 || event_number!=72 || event_number!=89) return;  
      //if (event_number!=53) return;


      /**********************************
      HMatrix chi2 calculation :
	This is where the H-Matrix is setup........
        Input to HMatrix : plane_vec, cell_vec and energy_vec of hits from a particle (e- or pi0).
        (make separate vector lists for each particle type) 
        ifirst is the start of the shower, sum_allhits_in a plane > 5 MeV
        energy should not be masked by attenuation corrections, 
        it should be proportional to the true deposited energy in the scintillator.
        chi2 values are initialized to 1.0e-1.. log(..) = -1
        when the input vector is empty, default value is -1 ....
      ******************************/

      // use the H matrix to calculate chi-square for electrons or pi0's.... 
      // compared to a set of pre-made matricies created using single particle MonteCarlo electrons 
      // the chisquare quantifies the transvers and longitudinal shape of the EM showers 


      // This is for slicer ..................
      
      // All particles combined ........
      std::vector<double> slicer_chisq_vec = SetUpEnergyArrays(slicer_plane_vec, slicer_cell_vec, slicer_energy_vec, ifirst);
      //std::vector<double> slicer_chisq_vec = GetHMChisq();
      //std::cout << "size : slicer_chisq_vec " << slicer_chisq_vec.size() << std::endl;

      double slicer_lchisq =  fabs(slicer_chisq_vec[0]); //longitudinal
      double slicer_cchisq =  fabs(slicer_chisq_vec[1]);//combined (a combination of transvers and longitidinal)
      double slicer_tchisq =  fabs(slicer_chisq_vec[2]); //transverse
      //std::cout << "lchisq " << lchisq << "\t tchisq " << tchisq << "\t cchisq " << cchisq << std::endl;
      
      // fill the ntuple with log10(chisq)
      gslc2 = log10(slicer_lchisq);
      gstc2 = log10(slicer_tchisq);
      gscc2 = log10(slicer_cchisq);
      //std::cout << "All : lchisq " << gslc2 << "\t tchisq " << gstc2 << "\t cchisq " << gscc2 << std::endl;

      // only for electrons ............
      std::vector<double> slicer_chisq_vec_el = SetUpEnergyArrays(slicer_plane_vec_el, slicer_cell_vec_el, slicer_energy_vec_el, ifirst);
      //std::cout << "size : slicer_chisq_vec " << slicer_chisq_vec_el.size() << std::endl;

      double slicer_lchi2_el =  log10(fabs(slicer_chisq_vec_el[0])); //longitudinal
      double slicer_cchi2_el =  log10(fabs(slicer_chisq_vec_el[1]));//combined (a combination of transvers and longitidinal)
      double slicer_tchi2_el =  log10(fabs(slicer_chisq_vec_el[2])); //transverse
      gslc2_el = slicer_lchi2_el;
      gstc2_el = slicer_tchi2_el;
      gscc2_el = slicer_cchi2_el;
      //std::cout << "el : lchisq " << gslc2_el << "\t tchisq " << gstc2_el << "\t cchisq " << gscc2_el << std::endl;

      // only for pi0s .............
      std::vector<double> slicer_chisq_vec_pi0 = SetUpEnergyArrays(slicer_plane_vec_pi0, slicer_cell_vec_pi0, slicer_energy_vec_pi0, ifirst);
      //std::cout << "size : slicer_chisq_vec " << slicer_chisq_vec_pi0.size() << std::endl;

      double slicer_lchi2_pi0 =  log10(fabs(slicer_chisq_vec_pi0[0])); //longitudinal
      double slicer_cchi2_pi0 =  log10(fabs(slicer_chisq_vec_pi0[1]));//combined (a combination of transvers and longitidinal)
      double slicer_tchi2_pi0 =  log10(fabs(slicer_chisq_vec_pi0[2])); //transverse
      gslc2_p0 = slicer_lchi2_pi0;
      gstc2_p0 = slicer_tchi2_pi0;
      gscc2_p0 = slicer_cchi2_pi0;
      //std::cout << "pi0 : lchisq " << gslc2_p0 << "\t tchisq " << gstc2_p0 << "\t cchisq " << gscc2_p0 << std::endl;


      // This is for each tracking algorithm ............
      
      // All particles combined ........
      std::vector<double> alg_chisq_vec = SetUpEnergyArrays(alg_plane_vec, alg_cell_vec, alg_energy_vec, ifirst);
      //std::vector<double> alg_chisq_vec = GetHMChisq();
      //std::cout << "size : alg_chisq_vec " << alg_chisq_vec.size() << std::endl;

      double alg_lchisq =  log10(fabs(alg_chisq_vec[0])); //longitudinal
      double alg_cchisq =  log10(fabs(alg_chisq_vec[1]));//combined (a combination of transvers and longitidinal)
      double alg_tchisq =  log10(fabs(alg_chisq_vec[2])); //transverse
      
      galc2 = alg_lchisq;
      gatc2 = alg_tchisq;
      gacc2 = alg_cchisq;
      //std::cout << "All : lchisq " << galc2 << "\t tchisq " << gatc2 << "\t cchisq " << gacc2 << std::endl;

      // only for electrons ............
      std::vector<double> alg_chisq_vec_el = SetUpEnergyArrays(alg_plane_vec_el, alg_cell_vec_el, alg_energy_vec_el, ifirst);
      //std::cout << "size : alg_chisq_vec " << alg_chisq_vec_el.size() << std::endl;

      double alg_lchi2_el =  log10(fabs(alg_chisq_vec_el[0])); //longitudinal
      double alg_cchi2_el =  log10(fabs(alg_chisq_vec_el[1]));//combined (a combination of transvers and longitidinal)
      double alg_tchi2_el =  log10(fabs(alg_chisq_vec_el[2])); //transverse
      galc2_el = alg_lchi2_el;
      gatc2_el = alg_tchi2_el;
      gacc2_el = alg_cchi2_el;
      //std::cout << "el : lchisq " << galc2_el << "\t tchisq " << gatc2_el << "\t cchisq " << gacc2_el << std::endl;

      // only for pi0s .............
      std::vector<double> alg_chisq_vec_pi0 = SetUpEnergyArrays(alg_plane_vec_pi0, alg_cell_vec_pi0, alg_energy_vec_pi0, ifirst);
      //std::cout << "size : alg_chisq_vec " << alg_chisq_vec_pi0.size() << std::endl;

      double alg_lchi2_pi0 =  log10(fabs(alg_chisq_vec_pi0[0])); //longitudinal
      double alg_cchi2_pi0 =  log10(fabs(alg_chisq_vec_pi0[1]));//combined (a combination of transvers and longitidinal)
      double alg_tchi2_pi0 =  log10(fabs(alg_chisq_vec_pi0[2])); //transverse
      galc2_p0 = alg_lchi2_pi0;
      gatc2_p0 = alg_tchi2_pi0;
      gacc2_p0 = alg_cchi2_pi0;
      //std::cout << "pi0 : lchisq " << galc2_p0 << "\t tchisq " << gatc2_p0 << "\t cchisq " << gacc2_p0 << std::endl;




      if(PassCuts==true){ 

	//flchisqL->Fill(log(100*lchisq));
	//flchisqT->Fill(log(100*tchisq));
	//flchisqC->Fill(log(100*cchisq));
	//flchisqLT->Fill(log(100*lchisq),(log(100*tchisq)));
		
	if(PassCuts==true){
	  if(log(100*slicer_lchisq)>=7.8)
	    PassCuts=false;
	  else
	    pass_cuts[event_type][8]++;
	}
	if(PassCuts==true){
	  if(log(100*slicer_tchisq)>=10.6)
	    PassCuts=false;
	  else
	    pass_cuts[event_type][9]++;
	}
	if(PassCuts==true){
	  if(log(100*slicer_cchisq)>=10.8)
	    PassCuts=false;
	  else
	    pass_cuts[event_type][10]++;
	}
      }
      
    
    } // pass_cuts_initial loop

    
    tree_hmatrix->Fill(); 


    return;
  } // end of analyze loop


 void HMatrixUser::endJob(){
    art::ServiceHandle<art::TFileService> tfs;
    TH1* hPOT = tfs->make<TH1F>("hTotalPOT", ";; POT", 1, 0, 1);
    hPOT->Fill(.5, fTotalPOT);
 }

 
  //#########################################################################
  double HMatrixUser::CalculateMIPPeak(art::Ptr<rb::Track> track){

    // the MIP Peak is the of cell hits with an energy in the region identified as the MIP Peak 
    // (ie the energy for a minimum ionizing cell hit). 
    // It is a number between 0 and 1, tending to be larger for muons as most hits are minimum ionizing

    art::ServiceHandle<calib::Calibrator> cal;
    int ncells=track->NCell();
    double cells_in_mip_peak=0;
    double cells_outside_mip_peak=0;

    for(int j=0; j<ncells; j++){
      art::Ptr<rb::CellHit> cellhit=track->Cell(j);
      rb::RecoHit rhit = cal->MakeRecoHit(*cellhit, track->W(cellhit.get()));

      if(!rhit.IsCalibrated()){
	std::cerr << "Not calibrated?!" << std::endl;
	continue;
      }

      double energy = rhit.GeV();
      double energy_mev=2*energy*1000;
      cells_outside_mip_peak++;

      if(4<energy_mev && energy_mev<16)
	cells_in_mip_peak++;      

    }

    cells_outside_mip_peak=cells_in_mip_peak/cells_outside_mip_peak;
    return cells_outside_mip_peak;

  }

  //-------------------------------------------------------------------------

  std::vector<double> HMatrixUser::SetUpEnergyArrays(std::vector<int> plane_vec, std::vector<int>cell_vec, std::vector<double> energy_vec, int ifirst){
  //void HMatrixUser::SetUpEnergyArrays(std::vector<int> plane_vec, std::vector<int>cell_vec, std::vector<double> energy_vec, int ifirst){
    
    //this is where the H-Matrix is setup
    //here, the plane_vec, cell_vec and energy_vec are filled with whatever shower candidate you like. 
    //Ifirst should indicate the first plane you'd like to test in that shower candidate
    //setup up the total energy per layer array


    // Calculate the chi2 ............
    double lchisq  = 1.0e-1;
    double cchisq  = 1.0e-1;
    double tchisq  = 1.0e-1;
    std::vector<double> chisq_vector;
    
    // if there are no candidates, there will be no entries...
    if (energy_vec.empty()){
      chisq_vector.push_back(lchisq);
      chisq_vector.push_back(cchisq);
      chisq_vector.push_back(tchisq);
      
      return chisq_vector;
    }
    
    bool ffound=false;
    
    double lbin[21];
    double transm[42];
    double ltbin[55];

    double etot;
    double ttot; // for tran-only bins
    double esumlay[1000];
    double eplcell[1000][400];
    double Tesum; // total energy sum of hits used in test
    int nhits_tot;

    ifirst = 1;

    //std::cout << "first plane " << ifirst << std::endl; 

    //initialize everything to zero to start
    for(int i=0; i<long_bins; i++){
      lbin[i]=0; 
    }
    for(int i=0; i<trans_bins; i++){
      transm[i]=0;
    }
    for(int i=0; i<comb_bins; i++){
      ltbin[i]=0;
    }

    etot = 0;
    ttot = 0;

    
    for(int i=0; i<max_layers; i++){
      esumlay[i]=0;
      maxHcell[i]=0;
      for(int j=0; j<max_cells; j++){
	eplcell[i][j]=0;
      }      
    }
    

    
    for (unsigned int i=0; i<energy_vec.size(); i++)
      {
	int plane     = 0;
	int cell      = 0;
	double energy = 0;
	
	plane  = plane_vec[i];
	energy = energy_vec[i];
	cell   = cell_vec[i];
	
	//  get energy sum in each layer
	// as well as the total energy sum
	if(plane<max_layers && cell<max_cells){
	  esumlay[plane] +=energy;
	  eplcell[plane][cell] += energy;
	  Tesum += energy;
	}
	
      }

    
    // this finds layer of start of shower if you don't already know it
    // finds the layer where sum of all hit energy is > 5 MeV
    for (unsigned int j=0; j<max_layers; ++j)
      {
	if (!ffound && esumlay[j]>0.005) // used .010 since that's mip value
	  {
	    ifirst = j;
	    ffound = true;
	  }
      }

    // try setting ifirst back one layer
    if (ifirst>0) ifirst = ifirst-1;
    
    //std::cout << "first plane " << ifirst << std::endl;

    // Find the cell in each layer (up to first 50 layers) with maximum energy deposited .....
    for (int i=ifirst; i<ifirst+50; ++i)
      {
	double emax = 0.; // initalize maximum energy variable per layer
	for (int j=0; j<max_cells; ++j)
	  {
	    
	    if (eplcell[i][j] > emax && i<max_layers)
	      {
		emax = eplcell[i][j];
		maxHcell[i] = j;
	      }
	  }
	
      }
    
    
    nhits_tot=0;
    //loop over hits
    for (unsigned int i=0; i<energy_vec.size(); ++i){ 
      
      int plane     = 0;
      int cell      = 0;
      double energy = 0;

      plane  = plane_vec[i];
      energy = energy_vec[i];
      cell   = cell_vec[i];

      //this sets up the bins to be compared against the pre-made hmatricies
      //the longitudinal bins consist of the energy in some number of longitidinal planes
      //the transverse bins then further split up the energy in the central planes by cell
      // it is now 3 layers per longitudinal bin .....
      // transverse inning starts from 3rd layer..... long_layer[2]


      //if (energy>0.0005 && plane<max_layers && cell<max_cells)
      if (plane<max_layers && cell<max_cells) // energy cut is removed ........
	{
	  nhits_tot++;
	  if (plane>ifirst-1 && plane<ifirst+21*nlay)
	    {
	      etot += energy;
	    }

	  if (plane>ifirst-1 && plane<ifirst+nlay)
	    {
	      //  first bin of longitudinal eflow
	      lbin[0] += energy;

	      //std::cout << "max cell " << maxHcell[plane] << std::endl;

	    } else if (plane>ifirst+nlay-1 && plane<ifirst+2*nlay)
	    {
	      lbin[1] += energy;

	    } else if (plane>ifirst+2*nlay-1 && plane<ifirst+3*nlay)
	    {
	      lbin[2] += energy;

	      //  start transverse bins here since 3 layers per bin
	      ttot += energy;
	      //std::cout << "max cell " << maxHcell[plane] << "\t " << ttot << std::endl;
	      if (cell==maxHcell[plane]-3) transm[0] += energy;
	      if (cell==maxHcell[plane]-2) transm[1] += energy;
	      if (cell==maxHcell[plane]-1) transm[2] += energy;
	      if (cell==maxHcell[plane]) transm[3] += energy;
	      if (cell==maxHcell[plane]+1) transm[4] += energy;
	      if (cell==maxHcell[plane]+2) transm[5] += energy;
	      if (cell==maxHcell[plane]+3) transm[6] += energy;

	    } else if (plane>ifirst+3*nlay-1 && plane<ifirst+4*nlay)
	    {
	      lbin[3] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[7] += energy;
	      if (cell==maxHcell[plane]-2) transm[8] += energy;
	      if (cell==maxHcell[plane]-1) transm[9] += energy;
	      if (cell==maxHcell[plane]) transm[10] += energy;
	      if (cell==maxHcell[plane]+1) transm[11] += energy;
	      if (cell==maxHcell[plane]+2) transm[12] += energy;
	      if (cell==maxHcell[plane]+3) transm[13] += energy;

	    } else if (plane>ifirst+4*nlay-1 && plane<ifirst+5*nlay)
	    {
	      lbin[4] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[14] += energy;
	      if (cell==maxHcell[plane]-2) transm[15] += energy;
	      if (cell==maxHcell[plane]-1) transm[16] += energy;
	      if (cell==maxHcell[plane]) transm[17] += energy;
	      if (cell==maxHcell[plane]+1) transm[18] += energy;
	      if (cell==maxHcell[plane]+2) transm[19] += energy;
	      if (cell==maxHcell[plane]+3) transm[20] += energy;

	    } else if (plane>ifirst+5*nlay-1 && plane<ifirst+6*nlay)
	    {
	      lbin[5] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[21] += energy;
	      if (cell==maxHcell[plane]-2) transm[22] += energy;
	      if (cell==maxHcell[plane]-1) transm[23] += energy;
	      if (cell==maxHcell[plane]) transm[24] += energy;
	      if (cell==maxHcell[plane]+1) transm[25] += energy;
	      if (cell==maxHcell[plane]+2) transm[26] += energy;
	      if (cell==maxHcell[plane]+3) transm[27] += energy;

	    } else if (plane>ifirst+6*nlay-1 && plane<ifirst+7*nlay)
	    {
	      lbin[6] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[28] += energy;
	      if (cell==maxHcell[plane]-2) transm[29] += energy;
	      if (cell==maxHcell[plane]-1) transm[30] += energy;
	      if (cell==maxHcell[plane]) transm[31] += energy;
	      if (cell==maxHcell[plane]+1) transm[32] += energy;
	      if (cell==maxHcell[plane]+2) transm[33] += energy;
	      if (cell==maxHcell[plane]+3) transm[34] += energy;

	    } else if (plane>ifirst+7*nlay-1 && plane<ifirst+8*nlay)
	    {
	      lbin[7] += energy;

	      ttot += energy;
	      if (cell==maxHcell[plane]-3) transm[35] += energy;
	      if (cell==maxHcell[plane]-2) transm[36] += energy;
	      if (cell==maxHcell[plane]-1) transm[37] += energy;
	      if (cell==maxHcell[plane]) transm[38] += energy;
	      if (cell==maxHcell[plane]+1) transm[39] += energy;
	      if (cell==maxHcell[plane]+2) transm[40] += energy;
	      if (cell==maxHcell[plane]+3) transm[41] += energy;

	    } else if (plane>ifirst+8*nlay-1 && plane<ifirst+9*nlay)
	    {
	      lbin[8] += energy;
	    } else if (plane>ifirst+9*nlay-1 && plane<ifirst+10*nlay)
	    {
	      lbin[9] += energy;
	    } else if (plane>ifirst+10*nlay-1 && plane<ifirst+11*nlay)
	    {
	      lbin[10] += energy;
	    } else if (plane>ifirst+11*nlay-1 && plane<ifirst+12*nlay)
	    {
	      lbin[11] += energy;
	    } else if (plane>ifirst+12*nlay-1 && plane<ifirst+13*nlay)
	    {
	      lbin[12] += energy;
	    } else if (plane>ifirst+13*nlay-1 && plane<ifirst+14*nlay)
	    {
	      lbin[13] += energy;
	    } else if (plane>ifirst+14*nlay-1 && plane<ifirst+15*nlay)
	    {
	      lbin[14] += energy;
	    } else if (plane>ifirst+15*nlay-1 && plane<ifirst+16*nlay)
	    {
	      lbin[15] += energy;
	    } else if (plane>ifirst+16*nlay-1 && plane<ifirst+17*nlay)
	    {
	      lbin[16] += energy;
	    } else if (plane>ifirst+17*nlay-1 && plane<ifirst+18*nlay)
	    {
	      lbin[17] += energy;
	    } else if (plane>ifirst+18*nlay-1 && plane<ifirst+19*nlay)
	    {
	      lbin[18] += energy;
	    } else if (plane>ifirst+19*nlay-1 && plane<ifirst+20*nlay)
	    {
	      lbin[19] += energy;
	    } else if (plane>ifirst+20*nlay-1 && plane<ifirst+21*nlay)
	    {
	      lbin[20] += energy;
	    }  // end of bin formation

	} // end threshold cut
	
    }  // end loop over Hits
      

    // trans[0] and trans[6] is not assigned in combination.. they are always zero.
    // this is top and lower cell of the shower....
    // get 55 combined long tran bins here-- these bins make the combined matrix
    // first 2 : long bins 0, 1
    // next 39 : trans bin 1 to 41
    // next 12, from lbin 8....

    ltbin[0] = lbin[0];
    ltbin[1] = lbin[1];
    ltbin[2] = transm[1];
    ltbin[3] = transm[2];
    ltbin[4] = transm[3];
    ltbin[5] = transm[4];
    ltbin[6] = transm[5];
    ltbin[7] = transm[7];
    ltbin[8] = transm[8];
    ltbin[9] = transm[9];
    ltbin[10] = transm[10];
    ltbin[11] = transm[11];
    ltbin[12] = transm[12];
    ltbin[13] = transm[13];
    ltbin[14] = transm[14];
    ltbin[15] = transm[15];
    ltbin[16] = transm[16];
    ltbin[17] = transm[17];
    ltbin[18] = transm[18];
    ltbin[19] = transm[19];
    ltbin[20] = transm[20];
    ltbin[21] = transm[21];
    ltbin[22] = transm[22];
    ltbin[23] = transm[23];
    ltbin[24] = transm[24];
    ltbin[25] = transm[25];
    ltbin[26] = transm[26];
    ltbin[27] = transm[27];
    ltbin[28] = transm[28];
    ltbin[29] = transm[29];
    ltbin[30] = transm[30];
    ltbin[31] = transm[31];
    ltbin[32] = transm[32];
    ltbin[33] = transm[33];
    ltbin[34] = transm[34];
    ltbin[35] = transm[35];
    ltbin[36] = transm[36];
    ltbin[37] = transm[37];
    ltbin[38] = transm[38];
    ltbin[39] = transm[39];
    ltbin[40] = transm[40];
    ltbin[41] = transm[41];
    ltbin[42] = lbin[8];
    ltbin[43] = lbin[9];
    ltbin[44] = lbin[10];
    ltbin[45] = lbin[11];
    ltbin[46] = lbin[12];
    ltbin[47] = lbin[13];
    ltbin[48] = lbin[14];
    ltbin[49] = lbin[15];
    ltbin[50] = lbin[16];
    ltbin[51] = lbin[17];
    ltbin[52] = lbin[18];
    ltbin[53] = lbin[19];
    ltbin[54] = lbin[20];
          
    // form E fractions in each layer for long-only bins
      
    fCounter = fCounter+1;
    
    // check if etot and ttot is non-zero, then divide... to avoid nan values..

    for (int i=0; i<long_bins; ++i)
      {
	
	//std::cout << "lbin " << i << "\t " << lbin[i] << "\t " << etot << std::endl;

	if(etot) lbin[i] = lbin[i]/etot;
	  
	fhMeans[i]=fhMeans[i]+lbin[i];
	  
	fenergy_long_bins->SetBinContent(i,fhMeans[i]/fCounter);
	fenergy_long_bins->SetBinError(i,(fhMeans[i]/fCounter)/sqrt(fhMeans[i]));
	 
      }
     
      
    // form E fractions in each layer for combined LT
    //the bins are normalized by the total energy
    for (int i=0; i<comb_bins; ++i)
      {
	if (etot) ltbin[i] = ltbin[i]/etot;
      }
      
      
    // form transverse E fractions in each cell
    //the bins are normalized by the total energy
    //if (ttot<.05) ttot = .05;

    double total_trans_energy = ttot;
    if (ttot<.025) ttot = .025;
    for (int j=0; j<trans_bins; ++j)
      {
	//std::cout << "transbin " << j << "\t " << transm[j] << "\t " << ttot << std::endl;

	transm[j] = transm[j]/ttot;
	fvMeans[j]=fvMeans[j]+transm[j];
	  
	fenergy_trans_bins->SetBinContent(j,fvMeans[j]/fCounter);
	fenergy_trans_bins->SetBinError(j,(fvMeans[j]/fCounter)/sqrt(fvMeans[j]));
	
      }

    /**********************
    std::cout << "etot " << etot << "\t ttot " << ttot << std::endl;
    std::cout << "----------------------------" << std::endl;
    for ( int i=0; i<long_bins; ++i){
      std::cout << lbin[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<trans_bins; ++i){
      std::cout << transm[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<comb_bins; ++i){
      std::cout << ltbin[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    ***********************/

    if (lbin[0]==1.0){ // all energy is deposited in first long bin
      chisq_vector.push_back(lchisq);
      chisq_vector.push_back(cchisq);
      chisq_vector.push_back(tchisq);
      
      return chisq_vector;
    }

    for (int i=0; i<long_bins; ++i){
      for (int j=0; j<long_bins; ++j){
	lchisq += (lbin[i]-lmeans[i])*HM[i][j]*(lbin[j]-lmeans[j]);
      }
    }

    for (int i=0; i<trans_bins; ++i){
      for (int j=0; j<trans_bins; ++j){       
	tchisq += (transm[i]-trmeans[i])*HTM[i][j]*(transm[j]-trmeans[j]);
      }
    }

    for (int i=0; i<comb_bins; ++i){
      for (int j=0; j<comb_bins; ++j){
	cchisq += (ltbin[i]-ltmeans[i])*HC[i][j]*(ltbin[j]-ltmeans[j]);
      }
    }


    // if shower goes upto long[1].., then all transverse bins are zero..., all_trans_energy=0.0
    // if shower stops at long[2], if all transverse energy is in trans[3] bins.. this is middle trans. bin..
    if ((total_trans_energy==0) || (transm[3]==1.0)){
      tchisq = 1.0e-1;
      cchisq = lchisq;
    }

    // old values
    // if all lbins[i] is zero, etot=0, lchi2 = 1.11114
    // if all transm[i] is zero, ttot=0, tchi2 = 3.11088
    // if all ltbin[i] is zero,   cchi2 = 3.9483

    // if all lbins[i] is zero, etot=0, lchi2 = 1.11449
    // if all transm[i] is zero, ttot=0, tchi2 = 3.11091
    // if all ltbin[i] is zero,   cchi2 = 3.9483

    chisq_vector.push_back(lchisq);
    chisq_vector.push_back(cchisq);
    chisq_vector.push_back(tchisq);

    
    return chisq_vector;
    
  } // SetupEnergyArrays loop


  std::vector<double> HMatrixUser::GetHMChisq(){
    //this function calculates a chisqard by comparing the normalized energy per bin as set up 
    //in the SetUpEnergyArrays function against the hmatrix, created using single particle electrons
    // it also calculates the diagonal log chi-squared of the shower

    Double_t lchisq  = 0.;
    Double_t lchisqd = 0.;
    Double_t cchisq  = 0.;
    Double_t cchisqd = 0.;
    Double_t tchisq  = 0.;
    Double_t tchisqd = 0.;
    std::vector<double> chisq_vector;


    /******************************
    std::cout << "----------------------------" << std::endl;
    for ( int i=0; i<long_bins; ++i){
      std::cout << lmeans[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<long_bins; ++i){
      for ( int j=0; j<long_bins; ++j){
	std::cout << HM[i][j] << " "; 
      }
      std::cout << " " << std::endl;
    }


    std::cout << "\n----------------------------" << std::endl;
    for ( int i=0; i<trans_bins; ++i){
      std::cout << trmeans[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;

    for (int i=0; i<trans_bins; ++i){
      for (int j=0; j<trans_bins; ++j){
	std::cout << HTM[i][j] << " ";
      }
      std::cout << "\n" << std::endl;
    }


    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<comb_bins; ++i){
      std::cout << ltmeans[i] << " " ;
    }
    std::cout << "\n ----------------------------" << std::endl;

    for (int i=0; i<comb_bins; ++i){
      for (int j=0; j<comb_bins; ++j){
	std::cout << HC[i][j] << " ";
      }
      std::cout << "\n" << std::endl;
    }



    std::cout << "----------------------------" << std::endl;
    for ( int i=0; i<long_bins; ++i){
      std::cout << lbin[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<trans_bins; ++i){
      std::cout << transm[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;
    for ( int i=0; i<comb_bins; ++i){
      std::cout << ltbin[i] << " "; 
    }
    std::cout << "\n ----------------------------" << std::endl;


    // don't divide by nhits_tot, no energy cut on transverse chi2
    // check the size if number of bins in the HMatrix are changed ....
    for ( int i=0; i<long_bins; ++i)
      {
	//lchisqd += (lbin[i]-lmeans[i])*HM[i][i]*(lbin[i]-lmeans[i])/nhits_tot;
	lchisqd += (lbin[i]-lmeans[i])*HM[i][i]*(lbin[i]-lmeans[i]);
	for ( int j=0; j<long_bins; ++j)
	  {
	    //lchisq += (lbin[i]-lmeans[i])*HM[i][j]*(lbin[j]-lmeans[j])/nhits_tot;
	    lchisq += (lbin[i]-lmeans[i])*HM[i][j]*(lbin[j]-lmeans[j]);
		  
	  }
      }


    // why there is an energy cut for transverse chi2 ............ now it is commented....
    for ( int i=0; i<trans_bins; ++i)
      {
	//if (ttot>0.05) tchisqd += (transm[i]-trmeans[i])*HTM[i][i]*(transm[i]-trmeans[i])/nhits_tot;
	//tchisqd += (transm[i]-trmeans[i])*HTM[i][i]*(transm[i]-trmeans[i])/nhits_tot;
	tchisqd += (transm[i]-trmeans[i])*HTM[i][i]*(transm[i]-trmeans[i]);
	for ( int j=0; j<trans_bins; ++j)
	  {
	    //if (ttot>0.05) tchisq += (transm[i]-trmeans[i])*HTM[i][j]*(transm[j]-trmeans[j])/nhits_tot;
	    //tchisq += (transm[i]-trmeans[i])*HTM[i][j]*(transm[j]-trmeans[j])/nhits_tot;
	    tchisq += (transm[i]-trmeans[i])*HTM[i][j]*(transm[j]-trmeans[j]);
	  }
      }


    for ( int i=0; i<comb_bins; ++i)
      {
	//cchisqd += (ltbin[i]-ltmeans[i])*HC[i][i]*(ltbin[i]-ltmeans[i])/nhits_tot;
	cchisqd += (ltbin[i]-ltmeans[i])*HC[i][i]*(ltbin[i]-ltmeans[i]);
	for ( int j=0; j<comb_bins; ++j)
	  {
	    //cchisq += (ltbin[i]-ltmeans[i])*HC[i][j]*(ltbin[j]-ltmeans[j])/nhits_tot;
	    cchisq += (ltbin[i]-ltmeans[i])*HC[i][j]*(ltbin[j]-ltmeans[j]);
	  }
      }

    ****************************/
    chisq_vector.push_back(lchisq);
    chisq_vector.push_back(cchisq);
    chisq_vector.push_back(tchisq);
    chisq_vector.push_back(lchisqd);
    chisq_vector.push_back(cchisqd);							       
    chisq_vector.push_back(tchisqd);



    return chisq_vector;


  }

  //A function that grabs the MCTruth information
  int HMatrixUser::GetMCTruthM(const art::Event& evt,art::PtrVector<simb::MCTruth>& mctruth){
    mctruth.clear();
    art::PtrVector<simb::MCTruth> temp;
 

    // going to use getManyByType so that we don't have to 
    // worry about using a module label
    std::vector< art::Handle< std::vector<simb::MCTruth> > > tcol;

   
    evt.getManyByType(tcol);
    for(size_t mctc = 0; mctc < tcol.size(); ++mctc){
      for(unsigned int i = 0; i < tcol[mctc]->size(); ++i){
	art::Ptr<simb::MCTruth> t(tcol[mctc], i);
	temp.push_back(t);
      }
    }
    temp.swap(mctruth);
 

    return mctruth.size();
  }

  //------------------------------------------------------------------
  // a function that figures out if events are vertex fiducial. It is probably a outdated.
  bool HMatrixUser::isInVertexFiducial(art::Ptr<rb::Track> trk) const
  {
    // Start and end don't mean much, pick whichever is lowest in y.
    const TVector3 v1 = trk->Start();
    const TVector3 v2 = trk->Stop();
    //const TVector3 start = (v1.Y() < v2.Y()) ? v1 : v2; 
    const TVector3 start = v2;

    const double x = start.X();
    const double y = start.Y();
    const double z = start.Z();

    art::ServiceHandle<geo::Geometry> geom;
    switch(geom->DetId()){

      //case daqchannelmap::DAQChannelMap::NDOS_ID:
      //case daqchannelmap::DAQChannelMap::NEARDET_ID:

      //case daqchannelmap::NDOS_ID:
      //case daqchannelmap::NEARDET_ID:
      
    case novadaq::cnv::kNDOS:
    case novadaq::cnv::kNEARDET:



      if(fabs(x) > 110) return false; // Fiducial X
      if(fabs(y) > 180) return false; // Fiducial Y
      if(z < 20 || z > 1400) return false; // Fiducial Z
      
      if(v1.X()==v2.X()||v1.Y()==v2.Y()) return false; // dx or dy = 0 means bad fit
      break;
      //case daqchannelmap::DAQChannelMap::FARDET_ID:

      //case daqchannelmap::FARDET_ID:
    case novadaq::cnv::kFARDET:

      if(fabs(x) > 730 || fabs(y) > 730) return false;
      if(z < 50 || z > 6300) return false;
      
      if(v1.X()==v2.X()||v1.Y()==v2.Y()) return false;
      break;
    default:
      assert(0 && "Unknown detector");
    }

    return true;
  }

  
  double HMatrixUser::GetOscWeight(const simb::MCNeutrino& nu, art::Ptr<simb::MCFlux>& flux)
  {
    const double L        = 810.;    // km
    const double ldm      = 2.35e-3; // large delta m^2, m23
    const double sin22t13 = 0.1;     // sin^2(2theta_13)
    const double sin22t23 = 1.;      // maximal sin^2(2theta_23)
    const double ssth23 = 0.5;     // sin^2(theta_23) corresponding to above

    // Signal
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 12 && abs(flux->fntype) == 14){
      // Nue appearance
      return sin22t13*ssth23*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }

    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 14 && abs(flux->fntype) == 14){
      // CC mu disappearance
      return 1-sin22t23*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }

    // Background
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 12 && abs(flux->fntype) == 12){
      // Beam nue
      return 1-sin22t13*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }

    if(nu.CCNC() == 1 && nu.Nu().PdgCode() == flux->fntype){
      // NC
      return 1;
    }

    if(nu.CCNC() == 1 && nu.Nu().PdgCode() != flux->fntype){
      // Swapped NC. We play clever reweighting games so this is OK
      return 1;
    }

    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 14 && abs(flux->fntype) == 12){
      // CC mu appearance
      return sin22t13*ssth23*util::sqr(sin(1.267*ldm*L/nu.Nu().E()));
    }

    // Don't know what this is
    return 0;
  }



  int HMatrixUser::GetEventType(const simb::MCNeutrino& nu, art::Ptr<simb::MCFlux>& flux)
  {
    // Signal
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 12 && abs(flux->fntype) == 14){
      // Nue appearance  
      return 1;
    }

    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 14 && abs(flux->fntype) == 14){
      // CC mu disappearance
      return 2;
    }

    // Background
    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 12 && abs(flux->fntype) == 12){
      // Beam nue
      return 4;
    }

    if(nu.CCNC() == 1 && nu.Nu().PdgCode() == flux->fntype){
      // NC
      return 3;
    }

    if(nu.CCNC() == 1 && nu.Nu().PdgCode() != flux->fntype){
      // Swapped NC. We play clever reweighting games so this is OK
      return 3;
    }

    if(nu.CCNC() == 0 && abs(nu.Nu().PdgCode()) == 14 && abs(flux->fntype) == 12){
      // CC mu appearance
      return 2;
    }

    // Don't know what this is
    return 5;
  }







} // end of namespace hme




namespace hme
{
  DEFINE_ART_MODULE(HMatrixUser);
}
