/////////////////////////////////////////////////////////////////////
/// \brief   Run and test Hough transform algorithms
/// \authors  messier@indiana.edu, mibaird@indiana.edu
/// \date
////////////////////////////////////////////////////////////////////

#include "TH2S.h"
#include "art/Framework/Services/Optional/TFileService.h"

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Event.h"

#include "Geometry/Geometry.h"
#include "HoughTransform/Hough2P.h"
#include "HoughTransform/IslandsTH2.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/HitList.h"
#include "RecoBase/HoughResult.h"



/// Hough transform header
namespace hough {
  ///
  /// Run and test Hough transform algorithms
  ///
  class HoughT : public art::EDProducer {
  public:
    explicit HoughT(fhicl::ParameterSet const &p);
    virtual ~HoughT();

    void         beginJob();
    virtual void produce(art::Event &e);
    void         reconfigure(const fhicl::ParameterSet& p);
  private:
    void WriteHoughHistos(const Hough2P& h);
  private:
    int    fWriteHoughHistos; ///< Write the hough maps out?
    int    fWeightByQ;        ///< Weight hits by charge?
    double fRhoBinSz;         ///< Size of rho bins (cm)
    double fThetaBinSz;       ///< Angular size of angle bins (degrees)
    int    fRhoSmoothWin;     ///< Number of bins to smooth over
    int    fThetaSmoothWin;   ///< Number of bins to smooth over
    double fRsqr;             ///< Distance cut on hit pairs
    double fPco;              ///< Number of sigma above average peak height to use as threshold cutoff in Hough space
    unsigned int fHitMin;     ///< Minimum number of hits required per view
    std::string  fSlicerLabel; ///< Label for the process that made slices
  };
}



namespace hough
{
  //.....................................................................
  HoughT::HoughT(fhicl::ParameterSet const &p)
  {
    this->reconfigure(p);
    this->produces< std::vector<rb::HoughResult> >();
  }
  
  //......................................................................
  
  void HoughT::beginJob() 
  {
  }
  
  //......................................................................
  
  void HoughT::reconfigure(const fhicl::ParameterSet& p) 
  {
    fWriteHoughHistos = p.get<int>          ("WriteHoughHistos");
    fWeightByQ        = p.get<int>          ("WeightByQ");
    fRhoBinSz         = p.get<double>       ("RhoBinSz");
    fThetaBinSz       = p.get<double>       ("ThetaBinSz");
    fRhoSmoothWin     = p.get<int>          ("RhoSmoothWin");
    fThetaSmoothWin   = p.get<int>          ("ThetaSmoothWin");
    fRsqr             = p.get<double>       ("Rsqr");
    fPco              = p.get<double>       ("PeakCO");
    fHitMin           = p.get<unsigned int> ("HitMin");
    fSlicerLabel      = p.get<std::string>  ("SlicerLabel");
  }
  
  //......................................................................
  
  HoughT::~HoughT() { }
  
  //......................................................................
  
  void HoughT::produce(art::Event& evt) 
  {
    unsigned register int i;
    art::ServiceHandle<geo::Geometry> geom;
    
    // Container for results to be stored in the event
    std::unique_ptr<std::vector<rb::HoughResult> > 
      htr(new std::vector<rb::HoughResult>);
  
    // Pull out the time slices and shuffle them into a useful container
    art::Handle<std::vector<rb::Cluster> > slice;
    evt.getByLabel(fSlicerLabel,slice);

    for (i=0; i<slice->size(); ++i) {
      const rb::Cluster& s = (*slice)[i];
      //
      // Skip any noise clusters
      //
      if (s.IsNoise()) continue;
      
      //
      // Build a stripped down hit list from the cells in the cluster
      //
      rb::HitList h(s, fWeightByQ);
      
      //
      // Construct the Hough transforms of the X and Y views
      //
      Hough2P hx(geo::kX, fRhoBinSz, fThetaBinSz, fRhoSmoothWin, fThetaSmoothWin, fRsqr, fPco);
      Hough2P hy(geo::kY, fRhoBinSz, fThetaBinSz, fRhoSmoothWin, fThetaSmoothWin, fRsqr, fPco);
      if(s.NXCell() >= fHitMin) hx.Map(h);
      if(s.NYCell() >= fHitMin) hy.Map(h);
      
      /*
      std::cout << "*=== Slice " << i << std::endl;
      std::cout << "Found " << hx.fH.fPeak.size() << " peaks in X" << std::endl;
      std::cout << "Found " << hy.fH.fPeak.size() << " peaks in Y" << std::endl;
      */

      //
      // Put the results into the event
      //
      htr->push_back(hx.fH);
      htr->push_back(hy.fH);
      
    //
    // Optionally write the hough maps to the ROOT histogram file
    //
      if (fWriteHoughHistos) {
	if (s.NXCell() > 0)
	  this->WriteHoughHistos(hx);

	if (s.NYCell() > 0)
	  this->WriteHoughHistos(hy);
      }
    }
    evt.put(std::move(htr));
  }
  
  //......................................................................
  
  void HoughT::WriteHoughHistos(const Hough2P& h) 
  {
    art::ServiceHandle<art::TFileService> f;
    f->make<TH2F>(*h.fHspaceW);
    f->make<TH2S>(*h.fHspaceI);
  }
 

} // end namespace hough  
////////////////////////////////////////////////////////////////////////



#include "art/Framework/Core/ModuleMacros.h"
namespace hough
{
  DEFINE_ART_MODULE(HoughT);
}
