///////////////////////////////////////////////////////////////////////////
/// \brief   Validate anything that creates HoughResults
/// \author  mibaird@indiana.edu, gsdavies@iu.edu
/// \date
//////////////////////////////////////////////////////////////////////////
#include "TTree.h"
#include "TH1F.h"
#include "TH1I.h"

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h" 
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"
#include "HoughTransform/HValidateNt.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/HitMap.h"
#include "RecoBase/HoughResult.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "Utilities/AssociationUtil.h"
#include "Utilities/func/MathUtil.h"

#include "limits.h"
// HoughValidate header
namespace hough {
  ///
  /// \brief Validate Hough transform algorithms
  ///
  class HoughValidate : public art::EDAnalyzer {

  public:
    explicit HoughValidate(fhicl::ParameterSet const &p);
    virtual ~HoughValidate();
    
    void     beginJob();
    void     analyze(const art::Event &e);
    void     reconfigure(const fhicl::ParameterSet& p);

  private:

    void HoughLineToDetector(const rb::HoughResult* hr, 
			     unsigned int hrIdx,
			     double* pz1, 
			     double* px1,
			     double* pz2,
			     double* px2);
      
    TTree*       fNtTree;
    HValidateNt* fNt;

    std::string  fMCLabel;          ///< Label for process that made MCTruth
    std::string  fSliceLabel;       ///< Label for process that made Cluster
    std::string  fHoughResultLabel; ///< Label for process that made HoughResult
    bool         fIsMC;             ///< Flag for is MC or not

    unsigned int fMinHit;  ///< Slices must have at least this many hits
    unsigned int fMinHitX; ///< Minimum hits in x-view/slice
    unsigned int fMinHitY; ///< Minumum hits in y-view/slice

    // Histograms used for production validation
    TH1I* fNumHLX;
    TH1I* fNumHLY;

    TH1F* fThreshX;
    TH1F* fThreshY;

    TH1I* fOnlineX[4];
    TH1I* fOnlineY[4];
    
    TH1F* fRhoX[4];
    TH1F* fRhoY[4];

    TH1F* fThetaX[4];
    TH1F* fThetaY[4];

    TH1F* fPeakX[4];
    TH1F* fPeakY[4];

    TH1F* fPeakThreshX[4];
    TH1F* fPeakThreshY[4];

    TH1F* fPerpDX[4];
    TH1F* fPerpDY[4];

  };
}


namespace hough
{
  //......................................................................
  HoughValidate::HoughValidate(fhicl::ParameterSet const &p)
    : EDAnalyzer(p)
  {
    this->reconfigure(p);
  }
  
  
  //......................................................................
  void HoughValidate::beginJob() 
  {      

      art::ServiceHandle<art::TFileService> tfs;

      // book Ntuple
      fNt = new HValidateNt;
      fNtTree = tfs->make<TTree>("hough", "Hough Analysis Tree");
      fNtTree->Branch("fHValidateNt", "hough::HValidateNt", &fNt, 1600, 99);
      
      // book histos
      const TString val[4] = {"1st","2nd","3rd","4th"};
      const unsigned int maxHists = 4;

      fNumHLX  = tfs->make<TH1I>("fNumHLX",
				"Num HL X-view;# of lines;count",
				11,-0.5,10.5);
      fNumHLY  = tfs->make<TH1I>("fNumHLY",
				"Num HL Y-view;# of lines;count",
				11,-0.5,10.5);

      fThreshX = tfs->make<TH1F>("fThreshX",
				 "Threshold value for Hough space map (X-view);threshold (arbitrary units);count",
				 100,0.0,5.0);
      fThreshY = tfs->make<TH1F>("fThreshY",
				 "Threshold value for Hough space map (Y-view);threshold (arbitrary units);count",
				 100,0.0,5.0);

      
      char title[100];
      for(unsigned int i = 0; i < maxHists; ++i){
	
	sprintf(title,"Num hits on %s line X-view;# of hits online;count",val[i].Data());
	fOnlineX[i]     = tfs->make<TH1I>("fOnlineX"+val[i],
					  title,
					  301,0,300);


	sprintf(title,"Num hits on %s line Y-view;# of hits online;count",val[i].Data());
	fOnlineY[i]     = tfs->make<TH1I>("fOnlineY"+val[i],
					 title,
					 301,0,300);

	sprintf(title,"Rho %s line x-view;rho [cm];count",val[i].Data());
	fRhoX[i]        = tfs->make<TH1F>("fRhoX"+val[i],
					  title,
					  200,-400.0,400.0);

	sprintf(title,"Rho %s line y-view;rho [cm];count",val[i].Data());
	fRhoY[i]        = tfs->make<TH1F>("fRhoY"+val[i],
					  title,
					  200,-400.0,400.0);

	sprintf(title,"Theta %s line x-view;theta [rad];count",val[i].Data());
	fThetaX[i]      = tfs->make<TH1F>("fThetaX"+val[i],
					  title,
					  40,0.0,4.0);

	sprintf(title,"Theta %s line y-view;theta [rad];count",val[i].Data());
	fThetaY[i]      = tfs->make<TH1F>("fTheta"+val[i],
					  title,
					  40,0.0,4.0);

	sprintf(title,"Peak value in Hough space map for %s line (x-view);peak (arbitrary units);count",val[i].Data());
	fPeakX[i]       = tfs->make<TH1F>("fPeakX"+val[i],
					  title,
					  200,0.0,2000.0);

	sprintf(title,"Peak over threshold %s line (x-view);peak/threshold;count",val[i].Data());
	fPeakThreshX[i] = tfs->make<TH1F>("fPeakThreshX"+val[i],
					  title,
					  200,0.0,10000.0);

	sprintf(title,"Peak value in Hough space map for %s line (y-view);peak (arbitrary units);count",val[i].Data());
	fPeakY[i]       = tfs->make<TH1F>("fPeakY"+val[i],
					  title,
					  200,0.0,2000.0);

	sprintf(title,"Peak over threshold %s line (x-view);peak/threshold;count",val[i].Data());
	fPeakThreshY[i] = tfs->make<TH1F>("fPeakThreshY"+val[i],
					  title,
					  200,0.0,10000.0);

	sprintf(title,"Perpendicular dist from true vertex to %s line X-view;dist [cm];count",val[i].Data());
	fPerpDX[i]      = tfs->make<TH1F>("fPerpDX"+val[i],
					  title,
					  100,0.0,100.0);

	sprintf(title,"Perpendicular dist from true vertex to %s line Y-view;dist [cm];count",val[i].Data());
	fPerpDY[i]      = tfs->make<TH1F>("fPerpDY"+val[i],
					  title,
					  100,0.0,100.0);

      } // end loop over histogram arrays

  }


  //......................................................................
  void HoughValidate::reconfigure(const fhicl::ParameterSet& p) 
  {
    fMCLabel          = p.get< std::string > ("MCLabel");
    fSliceLabel       = p.get< std::string > ("SliceLabel");
    fHoughResultLabel = p.get< std::string > ("HoughResultLabel");
    fIsMC             = p.get< bool >        ("IsMC"); 
    fMinHit           = p.get< unsigned int >("MinHit");
    fMinHitX          = p.get< unsigned int >("MinHitX");
    fMinHitY          = p.get< unsigned int >("MinHitY");
  }
  

  //......................................................................
  HoughValidate::~HoughValidate() 
  { 
  }
  
  //......................................................................
  void HoughValidate::HoughLineToDetector(const rb::HoughResult* hr, 
					  unsigned int lineIdx,
					  double* pz1, 
					  double* px1,
					  double* pz2,
					  double* px2)
  {
    art::ServiceHandle<geo::Geometry> geom;

    double slope, icept;
    hr->SlopeIcept(lineIdx, &slope, &icept);

    double x1 = 0, x2 = 0;
    if (hr->fView == geo::kX) {
      x1 =  geom->DetHalfWidth();
      x2 = -geom->DetHalfWidth();
    }
    else {
      x1 =  geom->DetHalfHeight();
      x2 = -geom->DetHalfHeight();
    }
    
    double z1, z2;
    z1 = (x1-icept)/slope;
    z2 = (x2-icept)/slope;
    
    if(z1 < 0.0){ 
      z1 = 0.0; 
      x1 = slope*z1 + icept;
    }
    if(z1 > geom->DetLength()){ 
      z1 = geom->DetLength(); 
      x1 = slope*z1 + icept;
    }
    if(z2 < 0.0){ 
      z2 = 0.0;
      x2 = slope*z2 + icept;
    }
    if(z2 > geom->DetLength()){ 
      z2 = geom->DetLength(); 
      x2 = slope*z2 + icept;
    }

    *pz1 = z1;
    *px1 = x1;
    *pz2 = z2;
    *px2 = x2;
    
  }

  //......................................................................
  void HoughValidate::analyze(const art::Event& evt) 
  {
    
    // get the slices and we'll then get the HoughResults associated with the slice
    art::Handle< std::vector< rb::Cluster > > hSlices;
    evt.getByLabel(fSliceLabel, hSlices);

    // get the MCTruth info
    art::Handle< std::vector< simb::MCTruth > > mclist;
    if(fIsMC == 1) evt.getByLabel(fMCLabel, mclist);

    /*
    // NOTE: I was using this for the MC dot product validation, but I'm
    // temporarily skipping that since sim::ParticleList is no longer being used
    
    // get the particle list for trajectory lengths
    art::Handle< std::vector<sim::ParticleList> > particlelist;
    evt.getByLabel("geantgen", particlelist);
    */
   
    // load event ID info into Ntuple
    fNt->run    = evt.run();
    fNt->subrun = evt.subRun();
    fNt->evt    = evt.id().event();

    // a set of temp vectors to fill the Ntuple
    std::vector<int> numHLXtemp;
    std::vector<int> numHLYtemp;
    
    std::vector<double> nuXtemp;
    std::vector<double> nuYtemp;
    std::vector<double> nuZtemp;
    std::vector<double> nuEtemp;
    
    std::vector<int> nupdgtemp;
    std::vector<int> nuCCNCtemp;

    const unsigned int NUM = 4;
    std::vector<double> perpdistxtemp[NUM];
    std::vector<double> perpdistytemp[NUM];

    std::vector<double> mcdpxtemp[NUM];
    std::vector<double> mcdpytemp[NUM];

    std::vector<double> onxtemp[NUM];
    std::vector<double> onytemp[NUM];

    std::vector<double> rhoxtemp[NUM];
    std::vector<double> rhoytemp[NUM];

    std::vector<double> thetaxtemp[NUM];
    std::vector<double> thetaytemp[NUM];

    std::vector<double> threshxtemp;
    std::vector<double> threshytemp;

    std::vector<double> peakxtemp[NUM]; // 1st, 2nd, 3rd peak in X
    std::vector<double> peakytemp[NUM]; // 1st, 2nd, 3rd peak in Y

    //object to get hough lines associated with the slice
    art::FindManyP<rb::HoughResult> fmhr(hSlices,evt,fHoughResultLabel);
    
    // Loop over slices
    for(unsigned int slIdx = 0; slIdx < hSlices->size(); ++slIdx){
    
      if(rb::IsFiltered(evt,hSlices,slIdx)) continue;

      const rb::Cluster& slice = (*hSlices)[slIdx];

      if(slice.IsNoise()) continue;

      ///Require a minimum number of hits in each view
      if(slice.NXCell() < fMinHitX) continue;
      if(slice.NYCell() < fMinHitY) continue;
      if(slice.NCell()  < fMinHit)  continue;
      
      // Get HitMap containing all slice hits
      const rb::HitMap hitmap(slice.AllCells());
    
      std::vector< art::Ptr<rb::HoughResult> > hrs = fmhr.at(slIdx);
      
      // This validation information is gathered for both data and MC
      for(unsigned int hrIdx = 0; hrIdx < hrs.size(); ++hrIdx) {
	
	// Skip empty HoughResults
	if(hrs.size() == 0) continue;
	
	if(hrs[hrIdx]->fView == geo::kX) {
	  
	const rb::HoughResult* hrxp = hrs[hrIdx].get();
	
	const unsigned int xpMax = hrxp->fPeak.size();
	
	numHLXtemp.push_back(xpMax);
	fNumHLX->Fill(xpMax);

	if(xpMax == 0 || xpMax > NUM) continue;

	fThreshX->Fill(hrxp->fPeak[0].fThr);
	threshxtemp.push_back(hrxp->fPeak[0].fThr); // temp in X
	
	// Loop over Hough Peaks
	for(unsigned int pIdx = 0; pIdx < xpMax; ++pIdx){

	  art::ServiceHandle<geo::Geometry> geom;
	  
	  double z1, z2, x1, x2;
	  std::vector< geo::OfflineChan > hrx_online;
 
	  this->HoughLineToDetector(hrxp, pIdx, &z1, &x1, &z2, &x2);

	  assert(hrxp->fView == geo::kX);
	  geom->CountCellsOnLineFast(hrxp->fView, 
				     x1, z1,
				     x2, z2,
				     hrx_online);

	  int online_x  = -1;
	  int counter_x = 0;
	  for(unsigned int n = 0; n < hrx_online.size(); ++n){
	    const unsigned int plane = hrx_online[n].Plane();
	    const unsigned int cell  = hrx_online[n].Cell();
	    if(hitmap.CellExists(plane,cell)){
	      ++counter_x;
	    }
	  }

	  if(counter_x > 0) online_x = counter_x;

	  fOnlineX[pIdx]->Fill(online_x);
	  fRhoX[pIdx]->Fill(hrxp->fPeak[pIdx].fRho);
	  fThetaX[pIdx]->Fill(hrxp->fPeak[pIdx].fTheta);
	  fPeakX[pIdx]->Fill(hrxp->fPeak[pIdx].fH);
	  fPeakThreshX[pIdx]->Fill((hrxp->fPeak[pIdx].fH)/(hrxp->fPeak[pIdx].fThr));
	  
	  onxtemp[pIdx].push_back(online_x);
	  rhoxtemp[pIdx].push_back(hrxp->fPeak[pIdx].fRho);
	  thetaxtemp[pIdx].push_back(hrxp->fPeak[pIdx].fTheta);
	  peakxtemp[pIdx].push_back(hrxp->fPeak[pIdx].fH);
	} // end loop over peaks
      } // end loop over X-view
      
      if(hrs[hrIdx]->fView == geo::kY) {
	
	const rb::HoughResult* hryp = hrs[hrIdx].get();
	
	const unsigned int ypMax = hryp->fPeak.size();
	
	numHLYtemp.push_back(ypMax);
	fNumHLY->Fill(ypMax);
	if(ypMax == 0 || ypMax > NUM) continue;
	fThreshY->Fill(hryp->fPeak[0].fThr);
	threshytemp.push_back(hryp->fPeak[0].fThr);
	
	for(unsigned int pIdx = 0; pIdx < ypMax; ++pIdx){
	  
	  art::ServiceHandle<geo::Geometry> geom;

	  double z1, z2, x1, x2;
	  std::vector< geo::OfflineChan > hry_online;
 
	  this->HoughLineToDetector(hryp, pIdx, &z1, &x1, &z2, &x2);

	  assert(hryp->fView == geo::kY);
	  geom->CountCellsOnLineFast(hryp->fView, 
				     x1, z1,
				     x2, z2,
				     hry_online);
	  
	  // Initialise number of lines in y to -1 for cases where there are X Hough
	  // lines but no Y and vice-versa
	  int online_y  = -1;
	  int counter_y = 0;	   

	  for(unsigned int n = 0; n < hry_online.size(); ++n){
	    const unsigned int plane = hry_online[n].Plane();
	    const unsigned int cell  = hry_online[n].Cell();
	    if(hitmap.CellExists(plane,cell)){
	      ++counter_y;
	    }
	  }

	  if(counter_y > 0) online_y = counter_y;

	  fOnlineY[pIdx]->Fill(online_y);
	  fRhoY[pIdx]->Fill(hryp->fPeak[pIdx].fRho);
	  fThetaY[pIdx]->Fill(hryp->fPeak[pIdx].fTheta);
	  fPeakY[pIdx]->Fill(hryp->fPeak[pIdx].fH);
	  fPeakThreshY[pIdx]->Fill((hryp->fPeak[pIdx].fH)/(hryp->fPeak[pIdx].fThr));

	  onytemp[pIdx].push_back(online_y);
	  rhoytemp[pIdx].push_back(hryp->fPeak[pIdx].fRho);
	  thetaytemp[pIdx].push_back(hryp->fPeak[pIdx].fTheta);
	  peakytemp[pIdx].push_back(hryp->fPeak[pIdx].fH);
	} // end loop over peaks
      } // end loop over Y-view
    } // end loop over HoughResults
    
    
    // additional validation info gathered for MC files
    if(fIsMC == 1){
      
      // validate perpendicular distance to top four lines
      if(hrs.size() > 0){
	
	for(unsigned int i = 0; i < mclist->size(); i++) {
	
	  simb::MCNeutrino nu = (*mclist)[i].GetNeutrino();
	  simb::MCParticle Nu = nu.Nu();
	
	  // BEWARE OF SEGFAULTS!!!  If this MCTruth did NOT come from a neutrino,
	  // then we must skip the next bit to avoid segfaluts.
	  if(!(Nu.PdgCode() == 12 || Nu.PdgCode() == 14 || Nu.PdgCode() == 16)) continue;
	
	  double xt = Nu.Vx();
	  double yt = Nu.Vy();
	  double zt = Nu.Vz();
	
	  nuXtemp.push_back(xt);
	  nuYtemp.push_back(yt);
	  nuZtemp.push_back(zt);
	  nuEtemp.push_back(Nu.E());
	  nupdgtemp.push_back(Nu.PdgCode());
	  nuCCNCtemp.push_back(nu.CCNC());
	
	
	  for(unsigned int hrIdx = 0; hrIdx < hrs.size(); ++hrIdx) {
	  
	    // variables used in validation
	    double m1 = 0.0;
	    double b1 = 0.0;
	    double distx = 0.0, disty = 0.0;
	  
	    //x-view mc validation
	    if(hrs[hrIdx]->fView == geo::kX) {
	      const rb::HoughResult* hrxp = hrs[hrIdx].get();
	      const unsigned int xpMax = hrxp->fPeak.size();
	    
	      if(xpMax == 0 || xpMax >= NUM) continue;
	      for(unsigned int pIdx = 0;pIdx < xpMax; ++pIdx){
		// perpendicular distance to ith line validation
		hrxp->SlopeIcept(pIdx, &m1, &b1);
		distx = sqrt(geo::DsqrToLine(zt, xt, 0.0, b1, 100.0, 100.0*m1 + b1));
		perpdistxtemp[pIdx].push_back(distx);
	      
		fPerpDX[pIdx]->Fill(distx);
	      } // end loop over x-view Peaks
	    } // end if fView == geo::kX
	  
	    //y-view mc validation
	    if(hrs[hrIdx]->fView == geo::kY) {
	      const rb::HoughResult* hryp = hrs[hrIdx].get();
	      const unsigned int ypMax = hryp->fPeak.size();
	    
	      if(ypMax == 0 || ypMax >= NUM) continue;
	      for(unsigned int pIdx = 0;pIdx < ypMax; ++pIdx){
		// perpendicular distance to ith line validation
		hryp->SlopeIcept(pIdx, &m1, &b1);
		disty = sqrt(geo::DsqrToLine(zt, yt, 0.0, b1, 100.0, 100.0*m1 + b1));
		perpdistytemp[pIdx].push_back(disty);
	      
		fPerpDY[pIdx]->Fill(disty);
	      } // end loop over y-view Peaks
	    } // end if fView == geo::kY
	  
	  } // end for hrIdx
	} // end for i (loop over size of MClist)
      } // if hrs->size() > 0
    } // end if(fIsMC)
    
    
    } // end loop over slices

    // NOTE: I was using this for the MC dot product validation, but I'm
    // temporarily skipping that since sim::ParticleList is no longer being used
    //
    // If I decide to put this back in, it belongs in the if(hr->size() > 0) loop found above
    /*
    if(hr->size() > 0) {

      // MC dot product validation
      //
      // NOTE:  This could be a little more efficient if I had integrated this section into the other loops
     // (instead of looping over mclist two seperate times.)  BUT this already runs very quickly and it
      // was cleaner to code and debug this way...

      for(unsigned int k = 0; k < hr->size(); ++k) {
	if((*hHoughs)[k].fPeak.size() != 0) {

	// variables used in MC-trajectory dot product validation
	double dp = -999.0;
	unsigned int track1 = 999, track2 = 999;

	double xy = 0.0, z = 0.0;
	double m = 0.0, b = 0.0;

	// x-view
	if((*hHoughs)[k].fView == geo::kX) {
	
	  //std::cout << "\n\nevent ID:   " << evt.id() << "\n";

	  // loop over the number of MC reactions
	  for(unsigned int i = 0; i < particlelist->size(); ++i) {

	    art::Ptr<sim::ParticleList> plist(particlelist, i);
	    
	    // loop over all particles and find the one closest to first hough line   
	    (*hHoughs)[k].SlopeIcept(0, &m, &b);
	    for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {
          
	      int pdg = (*plist).fParticles[c].PdgCode();

	      // don't let a hough line match a particle that wouldn't leave a signal in the detector
	      // (nu-e, nu-mu, nu-tau, neutron, pi0)



	      // ADD ANTI-MATTER VERSIONS OF THESE PARTICLES AS WELL!!!!!!!



	      if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		double mcslope = 0.0;
		double xytemp  = 0.0;
		double dptemp  = 0.0;
		double d = 0.0;

		xytemp = (*plist).fParticles[c].Px();
		z      = (*plist).fParticles[c].Pz();
		// make (xy,z) a unit vector
		xy = xytemp/sqrt(xytemp*xytemp + z*z);
		z  = z/sqrt(xytemp*xytemp + z*z);

		if(z != 0.0) mcslope = xy/z;
		else         mcslope = 1.0E9;

		dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		if(dptemp < 0.0) dptemp = -dptemp;
	      
		d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vx(), 0.0, b, 100.0, 100.0*m + b));

		// dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		// AND the track-length of the particle must be > 20.0cm
		if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) {dp = dptemp; track1 = c;}
	      
	      }  // end if(pdg codes)

	    }  // end for c
	    
	    mcdp1xtemp.push_back(dp);
	    
	    //std::cout << "height/threshold = " << (*hHoughs)[k].fPeak[0].fH/(*hHoughs)[k].fPeak[0].fThr 
		      //<< "   peak = " << (*hHoughs)[k].fPeak[0].fH
		      //<< "   threshold = " << (*hHoughs)[k].fPeak[0].fThr
		      //<< "   dp = " << dp << "\n";
	    
	    //if(dp < 0.0) std::cout << "\n\nFirst line X:   (no match)\n";
	    //else std::cout << "\n\nFirst line X:    particle# = " << track1 << "   PDG = " << (*plist).fParticles[track1].PdgCode() << "   dp = " << dp 
	    //		   << "   track length = " << (*plist).fParticles[track1].Trajectory().TotalLength() << "\n";



	    // loop over all particles and find the one closest to second hough line
	    dp = -999.0;
	    if((*hHoughs)[k].fPeak.size() > 1) {
	      
	      (*hHoughs)[k].SlopeIcept(1, &m, &b);
	      for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {


		// skip particle if it is already associated with the first hough line
		if(c != track1) {

		  int pdg = (*plist).fParticles[c].PdgCode();

		  // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
		  if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		    double mcslope = 0.0;
		    double xytemp  = 0.0;
		    double dptemp  = 0.0;
		    double d = 0.0;

		    xytemp = (*plist).fParticles[c].Px();
		    z      = (*plist).fParticles[c].Pz();
		    // make (xy,z) a unit vector
		    xy = xytemp/sqrt(xytemp*xytemp + z*z);
		    z  = z/sqrt(xytemp*xytemp + z*z);

		    if(z != 0.0) mcslope = xy/z;
		    else         mcslope = 1.0E9;

		    dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		    if(dptemp < 0.0) dptemp = -dptemp;
	      
		    d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vx(), 0.0, b, 100.0, 100.0*m + b));

		    // dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		    // AND the track-length of the particle must be > 20.0cm
		    if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) {dp = dptemp; track2 = c;}

		  }  // end if(pdg codes)

		} // end if(c != track1)
	      }  // end for c

	      mcdp2xtemp.push_back(dp);

	      //if(dp < 0.0) std::cout << "\n\nSecond line X:   (no match)\n";
	      //else std::cout << "\n\nSecond line X:    particle# = " << track2 << "   PDG = " << (*plist).fParticles[track2].PdgCode() << "   dp = " << dp 
	      //		     << "   track length = " << (*plist).fParticles[track2].Trajectory().TotalLength() << "\n";
	    }  // end if(peak.size > 1)



	    // loop over all particles and find the one closest to third hough line
	    dp = -999.0;
	    if((*hHoughs)[k].fPeak.size() > 2) {
	      
	      (*hHoughs)[k].SlopeIcept(2, &m, &b);
	      for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {

		// skip particle if it is already associated with the first or second hough line
		if(c != track1 && c != track2) {
		  
		  int pdg = (*plist).fParticles[c].PdgCode();

		  // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
		  if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		    double mcslope = 0.0;
		    double xytemp  = 0.0;
		    double dptemp  = 0.0;
		    double d = 0.0;

		    xytemp = (*plist).fParticles[c].Px();
		    z      = (*plist).fParticles[c].Pz();
		    // make (xy,z) a unit vector
		    xy = xytemp/sqrt(xytemp*xytemp + z*z);
		    z  = z/sqrt(xytemp*xytemp + z*z);

		    if(z != 0.0) mcslope = xy/z;
		    else         mcslope = 1.0E9;

		    dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		    if(dptemp < 0.0) dptemp = -dptemp;
	      
		    d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vx(), 0.0, b, 100.0, 100.0*m + b));

		    // dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		    // AND the track-length of the particle must be > 20.0cm
		    if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) dp = dptemp;

		  }  // end if(pdg codes)

		} // end if(c != track1 && c != track2)
	      }  // end for c

	      mcdp3xtemp.push_back(dp);

	      //if(dp < 0.0) std::cout << "\n\nThird line X:   (no match)\n";
	      //else std::cout << "\n\nThird line X:    particle# = " << track3 << "   PDG = " << (*plist).fParticles[track3].PdgCode() << "   dp = " << dp 
	      //		     << "   track length = " << (*plist).fParticles[track3].Trajectory().TotalLength() << "\n";
	    }  // end if(peak.size > 2)
	    
	  }  // end for i
	}  // end if(x-view) 


	
	// y-view
	if((*hHoughs)[k].fView == geo::kY) {
	
	  // loop over the number of MC reactions
	  for(unsigned int i = 0; i < particlelist->size(); ++i) {

	    art::Ptr<sim::ParticleList> plist(particlelist, i);
	    
	    // loop over all particles and find the one closest to first hough line   
	    (*hHoughs)[k].SlopeIcept(0, &m, &b);
	    for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {
          
	      int pdg = (*plist).fParticles[c].PdgCode();

	      // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
	      if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		double mcslope = 0.0;
		double xytemp  = 0.0;
		double dptemp  = 0.0;
		double d = 0.0;

		xytemp = (*plist).fParticles[c].Py();
		z      = (*plist).fParticles[c].Pz();
		// make (xy,z) a unit vector
		xy = xytemp/sqrt(xytemp*xytemp + z*z);
		z  = z/sqrt(xytemp*xytemp + z*z);

		if(z != 0.0) mcslope = xy/z;
		else         mcslope = 1.0E9;

		dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		if(dptemp < 0.0) dptemp = -dptemp;
	      
		d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vy(), 0.0, b, 100.0, 100.0*m + b));

		// dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		// AND the track-length of the particle must be > 20.0cm
		if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) {dp = dptemp; track1 = c;}
	      
	      }  // end if(pdg codes)

	    }  // end for c

	    mcdp1ytemp.push_back(dp);

	    //if(dp < 0.0) std::cout << "\n\nFirst line Y:   (no match)\n";
	    //else std::cout << "\n\nFirst line Y:    particle# = " << track1 << "   PDG = " << (*plist).fParticles[track1].PdgCode() << "   dp = " << dp 
	    //		   << "   track length = " << (*plist).fParticles[track1].Trajectory().TotalLength() << "\n";


	    // loop over all particles and find the one closest to second hough line
	    dp = -999.0;
	    if((*hHoughs)[k].fPeak.size() > 1) {
	      
	      (*hHoughs)[k].SlopeIcept(1, &m, &b);
	      for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {


		// skip particle if it is already associated with the first hough line
		if(c != track1) {

		  int pdg = (*plist).fParticles[c].PdgCode();

		  // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
		  if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		    double mcslope = 0.0;
		    double xytemp  = 0.0;
		    double dptemp  = 0.0;
		    double d = 0.0;

		    xytemp = (*plist).fParticles[c].Py();
		    z      = (*plist).fParticles[c].Pz();
		    // make (xy,z) a unit vector
		    xy = xytemp/sqrt(xytemp*xytemp + z*z);
		    z  = z/sqrt(xytemp*xytemp + z*z);

		    if(z != 0.0) mcslope = xy/z;
		    else         mcslope = 1.0E9;

		    dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		    if(dptemp < 0.0) dptemp = -dptemp;
	      
		    d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vy(), 0.0, b, 100.0, 100.0*m + b));

		    // dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		    // AND the track-length of the particle must be > 20.0cm
		    if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) {dp = dptemp; track2 = c;}

		  }  // end if(pdg codes)

		} // end if(c != track1)
	      }  // end for c

	      mcdp2ytemp.push_back(dp);

	      //if(dp < 0.0) std::cout << "\n\nSecond line Y:   (no match)\n";
	      //else std::cout << "\n\nSecond line Y:    particle# = " << track2 << "   PDG = " << (*plist).fParticles[track2].PdgCode() << "   dp = " << dp 
	      //		     << "   track length = " << (*plist).fParticles[track2].Trajectory().TotalLength() << "\n";
	    }  // end if(peak.size > 1)



	    // loop over all particles and find the one closest to third hough line
	    dp = -999.0;
	    if((*hHoughs)[k].fPeak.size() > 2) {
	      
	      (*hHoughs)[k].SlopeIcept(2, &m, &b);
	      for (unsigned int c = 0; c != plist->fParticles.size(); ++c) {

		// skip particle if it is already associated with the first or second hough line
		if(c != track1 && c != track2) {
		  
		  int pdg = (*plist).fParticles[c].PdgCode();

		  // don't let a hough line match a particle that wouldn't leave a signal in the detector (nu-e, nu-mu, nu-tau, neutron, pi0)
		  if(pdg != 12 && pdg != 14 && pdg != 16 && pdg != 2112 && pdg != 111) {

		    double mcslope = 0.0;
		    double xytemp  = 0.0;
		    double dptemp  = 0.0;
		    double d = 0.0;

		    xytemp = (*plist).fParticles[c].Py();
		    z      = (*plist).fParticles[c].Pz();
		    // make (xy,z) a unit vector
		    xy = xytemp/sqrt(xytemp*xytemp + z*z);
		    z  = z/sqrt(xytemp*xytemp + z*z);

		    if(z != 0.0) mcslope = xy/z;
		    else         mcslope = 1.0E9;

		    dptemp = (1 + m*mcslope)/sqrt(1 + m*m + mcslope*mcslope + m*m*mcslope*mcslope);
		    if(dptemp < 0.0) dptemp = -dptemp;
	      
		    d = sqrt(geo::DsqrToLine((*plist).fParticles[c].Vz(),(*plist).fParticles[c].Vy(), 0.0, b, 100.0, 100.0*m + b));

		    // dot product must be bigger than the old one AND the hough line must pass within 10.0cm of the true vertex of the particle
		    // AND the track-length of the particle must be > 20.0cm
		    if(dptemp > dp && d < 10.0 && (*plist).fParticles[c].Trajectory().TotalLength() > 10.0) dp = dptemp;

		  }  // end if(pdg codes)

		} // end if(c != track1 && c != track2)
	      }  // end for c

	      mcdp3ytemp.push_back(dp);

	      //if(dp < 0.0) std::cout << "\n\nThird line Y:   (no match)\n";
	      //else std::cout << "\n\nThird line Y:    particle# = " << track3 << "   PDG = " << (*plist).fParticles[track3].PdgCode() << "   dp = " << dp 
	      //		     << "   track length = " << (*plist).fParticles[track3].Trajectory().TotalLength() << "\n";
	    }  // end if(peak.size > 2)
    
	  }  // end for i
	}  // end if(y-view)

	}  // end if((*hHoughs)[k].size() != 0)
      }  // end for k
    }  // end dot product validation
    */




    fNt->numHLX = numHLXtemp;
    fNt->numHLY = numHLYtemp;
    
    fNt->nuX = nuXtemp;
    fNt->nuY = nuYtemp;
    fNt->nuZ = nuZtemp;
    fNt->nuE = nuEtemp;
    
    fNt->nupdg  = nupdgtemp;
    fNt->nuCCNC = nuCCNCtemp;

    fNt->perpdist1x = perpdistxtemp[0];
    fNt->perpdist1y = perpdistytemp[0];
    fNt->perpdist2x = perpdistxtemp[1];
    fNt->perpdist2y = perpdistytemp[1];
    fNt->perpdist3x = perpdistxtemp[2];
    fNt->perpdist3y = perpdistytemp[2];
    fNt->perpdist4x = perpdistxtemp[3];
    fNt->perpdist4y = perpdistytemp[3];

    fNt->mcdp1x = mcdpxtemp[0];
    fNt->mcdp1y = mcdpytemp[0];
    fNt->mcdp2x = mcdpxtemp[1];
    fNt->mcdp2y = mcdpytemp[1];
    fNt->mcdp3x = mcdpxtemp[2];
    fNt->mcdp3y = mcdpytemp[2];
    fNt->mcdp4x = mcdpxtemp[3];
    fNt->mcdp4y = mcdpytemp[3];

    fNt->on1x = onxtemp[0];
    fNt->on2x = onxtemp[1];
    fNt->on3x = onxtemp[2];
    fNt->on4x = onxtemp[3];
    fNt->on1y = onytemp[0];
    fNt->on2y = onytemp[1];
    fNt->on3y = onytemp[2];
    fNt->on4y = onytemp[3];

    fNt->rho1x = rhoxtemp[0];
    fNt->rho2x = rhoxtemp[1];
    fNt->rho3x = rhoxtemp[2];
    fNt->rho4x = rhoxtemp[3];
    fNt->rho1y = rhoytemp[0];
    fNt->rho2y = rhoytemp[1];
    fNt->rho3y = rhoytemp[2];
    fNt->rho4y = rhoytemp[3];

    fNt->theta1x = thetaxtemp[0];
    fNt->theta2x = thetaxtemp[1];
    fNt->theta3x = thetaxtemp[2];
    fNt->theta4x = thetaxtemp[3];
    fNt->theta1y = thetaytemp[0];
    fNt->theta2y = thetaytemp[1];
    fNt->theta3y = thetaytemp[2];
    fNt->theta4y = thetaytemp[3];

    fNt->threshx = threshxtemp;
    fNt->threshy = threshytemp;

    fNt->peak1x  = peakxtemp[0];
    fNt->peak2x  = peakxtemp[1];
    fNt->peak3x  = peakxtemp[2];
    fNt->peak4x  = peakxtemp[3];
    fNt->peak1y  = peakytemp[0];
    fNt->peak2y  = peakytemp[1];
    fNt->peak3y  = peakytemp[2];
    fNt->peak4y  = peakytemp[3];

    fNtTree->Fill();

    return;

  }

  DEFINE_ART_MODULE(HoughValidate);
} // end namespace hough  
///////////////////////////////////////////////////////////////////////
