/////////////////////////////////////////////////////////////////////////
/// \brief   Run and test Hough transform algorithms
/// \authors messier@indiana.edu, mibaird@indiana.edu
/// \date
////////////////////////////////////////////////////////////////////////

#include "TH1F.h"
#include "TH2S.h"

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"

#include "ChannelInfo/BadChanList.h"
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "HoughTransform/MultiHough2P.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/HitList.h"
#include "RecoBase/HoughResult.h"
#include "RecoBase/FilterList.h"
#include "Utilities/func/MathUtil.h"
#include "Utilities/AssociationUtil.h"





// MultiHoughT header
namespace hough {
  ///
  /// Run and test Hough transform algorithms
  ///
  class MultiHoughT : public art::EDProducer {
  public:
    explicit MultiHoughT(fhicl::ParameterSet const &p);
    virtual ~MultiHoughT();
    
    void         beginJob();
    virtual void produce(art::Event &e);
    void         reconfigure(const fhicl::ParameterSet& p);
  private:
    void         WriteHoughHistos(const MultiHough2P& h);
    art::PtrVector<rb::CellHit> Scrub(art::PtrVector<rb::CellHit> c);
 
  private:
    int          fWriteHoughHistos; ///< Write the hough maps out?
    int          fWeightByQ;        ///< Weight hits by charge?
    double       fRhoBinSz;         ///< Size of rho bins (cm)
    double       fThetaBinSz;       ///< Angular size of angle bins (degrees)
    int          fRhoSmoothWin;     ///< Number of bins to smooth over
    int          fThetaSmoothWin;   ///< Number of bins to smooth over
    double       fRsqr;             ///< Distance cut on hit pairs
    double       fPco;              ///< Number of sigma above average peak height to use as threshold cutoff in Hough space
    int          fLoops;            ///< Max number of Multi-Hough loops
    unsigned int fHitMin;           ///< Minimum number of hits required per view
    double       fRemoveDist;       ///< Distance cut for removing points on a Hough line
    std::string  fSlicerLabel;      ///< Label for the process that made the slices
    bool         fObeyPreselection; ///< Check rb::IsFiltered?
    
  };
}



namespace hough
{
  //......................................................................
  MultiHoughT::MultiHoughT(fhicl::ParameterSet const &p)
  {
    this->reconfigure(p);
    this->produces< std::vector<rb::HoughResult> >();
    this->produces<art::Assns<rb::HoughResult, rb::Cluster> >();
  }
  
  //......................................................................
  
  void MultiHoughT::beginJob() 
  { }
  
  //......................................................................
  
  void MultiHoughT::reconfigure(const fhicl::ParameterSet& p) 
  {
    fWriteHoughHistos = p.get<int>            ("WriteHoughHistos");
    fWeightByQ        = p.get<int>            ("WeightByQ");
    fRhoBinSz         = p.get<double>         ("RhoBinSz");
    fThetaBinSz       = p.get<double>         ("ThetaBinSz");
    fRhoSmoothWin     = p.get<int>            ("RhoSmoothWin");
    fThetaSmoothWin   = p.get<int>            ("ThetaSmoothWin");
    fRsqr             = p.get<double>         ("Rsqr");
    fPco              = p.get<double>         ("PeakCO");
    fLoops            = p.get<int>            ("MHLoops");
    fHitMin           = p.get<unsigned int>   ("HitMin");
    fRemoveDist       = p.get<double>         ("RemoveDist");
    fSlicerLabel      = p.get<std::string>    ("SlicerLabel");
    fObeyPreselection = p.get<bool>           ("ObeyPreselection");
  }
  
  
  
  //......................................................................
  MultiHoughT::~MultiHoughT() { }


  
  //.....................................................................
  
  void MultiHoughT::produce(art::Event& evt) 
  {
    unsigned register int i;
    
    // Container for results to be stored in the event
    std::unique_ptr<std::vector<rb::HoughResult> > 
      htr(new std::vector<rb::HoughResult>);
    std::unique_ptr<art::Assns<rb::HoughResult, rb::Cluster> > 
      assns(new art::Assns<rb::HoughResult, rb::Cluster>);
    
    // Pull out the time slices and shuffle them into a useful container
    art::Handle<std::vector<rb::Cluster> > slice;
    evt.getByLabel(fSlicerLabel,slice);
     
    
    for (i=0; i<slice->size(); ++i) {
      if(fObeyPreselection && rb::IsFiltered(evt,slice,i)) continue;
      const rb::Cluster& s = (*slice)[i];
      
      //
      // Skip any noise clusters
      //
      if (s.IsNoise()) continue;


      //
      // Build a stripped down hit list from the cells in the cluster but first,
      // scrub the list of noisy hits.
      //
      rb::Cluster scrubbed(this->Scrub(s.AllCells()));
      rb::HitList h(scrubbed, fWeightByQ);
      
      
      //
      // Construct the MultiHough transforms of the X and Y views
      //
      MultiHough2P hx(geo::kX, fRhoBinSz, fThetaBinSz, fRhoSmoothWin, fThetaSmoothWin, fRsqr, fPco, fLoops, fRemoveDist);
      MultiHough2P hy(geo::kY, fRhoBinSz, fThetaBinSz, fRhoSmoothWin, fThetaSmoothWin, fRsqr, fPco, fLoops, fRemoveDist);

      // require there to be more than fHitMin hits in each view to do the multi-hough
      
      if(scrubbed.NXCell() < fHitMin || scrubbed.NYCell() < fHitMin) continue;
      hx.MultiMap(h);
      hy.MultiMap(h);
      

      // require that there must be at least fHitMin hits in each view SEPERATELY
      /*
      if(scrubbed.NXCell() >= fHitMin) hx.MultiMap(h);
      if(scrubbed.NYCell() >= fHitMin) hy.MultiMap(h);
      */
    
      //
      // Put the results into the event
      //
      art::Ptr<rb::Cluster> sptr(slice, i);
      htr->push_back(hx.fH);
      // Associate most recently added houghresult with the slice
      util::CreateAssn(*this, evt, *htr, sptr, *assns);
      htr->push_back(hy.fH);
      // Associate most recently added houghresult with the slice
      util::CreateAssn(*this, evt, *htr, sptr, *assns);
            
      //
      // Optionally write the hough maps to the ROOT histogram file
      //
      if (fWriteHoughHistos) {
	this->WriteHoughHistos(hx);
	this->WriteHoughHistos(hy);
      }
         
    }

    evt.put(std::move(htr));
    evt.put(std::move(assns));

  }
  
  //......................................................................
  
  void MultiHoughT::WriteHoughHistos(const MultiHough2P& h) 
  {
    // write the Hough space map
    art::ServiceHandle<art::TFileService> f;
    TH2F* hist = h.MapToTH2F();
    if(hist) f->make<TH2F>(*hist);
  }

  //......................................................................
  //function to remove hits with no neighbors from the slice
  art::PtrVector<rb::CellHit> MultiHoughT::Scrub(art::PtrVector<rb::CellHit> c)
  {
    
    std::vector<int> hit_numNeighbors;
    //in an ideal detector a hit must have 1 neighbor
    //inside this distance.
    double maxGap = 60.0; 
    art::PtrVector<rb::CellHit> trimCells;
    int iPlane, fPlane, iCell, fCell;
    double goodCh, totCh;
    
    iPlane = 0;
    iCell  = 0;
    fPlane = 0;
    fCell  = 0;
    
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<chaninfo::BadChanList> badc;
    
    for (unsigned int i=0; i<c.size(); i++){
      hit_numNeighbors.push_back(0);
    }
    
    //loop to find a cells closest neighbor
    for (unsigned int i=0; i<c.size(); i++){
      double minDist = 9999.0;
      int minCell = -1;
      if (hit_numNeighbors[i] > 0) continue;
      for(unsigned int j=0; j<c.size(); j++){
        if (i == j) continue;
        if (c[i]->View() != c[j]->View()) continue;
        double xyz1[3], xyz2[3];
        geom->Plane((c[i])->Plane())->Cell((c[i])->Cell())->GetCenter(xyz1);
        geom->Plane((c[j])->Plane())->Cell((c[j])->Cell())->GetCenter(xyz2);
        double dist = util::pythag(xyz1[0]-xyz2[0],
                                   xyz1[1]-xyz2[1],
                                   xyz1[2]-xyz2[2]);
        if (dist < minDist){
          minDist = dist;
          minCell = j;
        }
      }
      if (minCell == -1) continue;

      //define the rectangle of cells such that a cell and its 
      //nearest neighbor are at diagonal corners
      if (c[i]->Plane() < c[minCell]->Plane()){
        iPlane = c[i]->Plane();
        fPlane = c[minCell]->Plane();
      }
      else {
         iPlane = c[minCell]->Plane();
         fPlane = c[i]->Plane();
      }
      if (c[i]->Cell() < c[minCell]->Cell()) {
        iCell = c[i]->Cell();
        fCell = c[minCell]->Cell();
      }
      else {
        iCell = c[minCell]->Cell();
        fCell = c[i]->Cell();
      }
      
      //within the rectangle find the fraction of cells that are good
      goodCh = 0;
      totCh = 0;
      for (int c=iCell; c<=fCell; c++){
        for (int p=iPlane; p<=fPlane; p+=2){
          totCh++;
          if(!(badc->IsBad(p,c))) ++goodCh;
        }
      }
      if ((minDist*goodCh/totCh) < maxGap){
        hit_numNeighbors[i]++;
        hit_numNeighbors[minCell]++;
      }
    }
    
    //keep the hits that have at least 1 neighbor in the slice
    for (unsigned int i=0; i<c.size(); i++){
      if (hit_numNeighbors[i] > 0){
        trimCells.push_back(c[i]);
      }
    }
   
    return trimCells;
  }


} // end namespace hough
////////////////////////////////////////////////////////////////////////



#include "art/Framework/Core/ModuleMacros.h"
namespace hough
{
  DEFINE_ART_MODULE(MultiHoughT);
}
