////////////////////////////////////////////////////////////////////////
/// \brief  Find event vertex using a Hough transform
/// \author Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "HoughVertex/HoughVertexAlg.h"

#include "HoughVertex/LiteTH2.h"

#include "Utilities/func/MathUtil.h"

#include <cassert>

namespace hv
{
  //......................................................................
  HoughVertexAlg::HoughVertexAlg()
  {
  }

  //......................................................................
  TVector3 HoughVertexAlg::FindVertex(const std::vector<Vector2D>& xhits,
                                      const std::vector<Vector2D>& yhits) const
  {
    // Find the event "bounding box"
    float minx = FLT_MAX, maxx = -FLT_MAX;
    float miny = FLT_MAX, maxy = -FLT_MAX;
    float minz = FLT_MAX, maxz = -FLT_MAX;
    for(const Vector2D& hit: xhits){
      minx = std::min(hit.v, minx);
      maxx = std::max(hit.v, maxx);
      minz = std::min(hit.z, minz);
      maxz = std::max(hit.z, maxz);
    }
    for(const Vector2D& hit: yhits){
      miny = std::min(hit.v, miny);
      maxy = std::max(hit.v, maxy);
      minz = std::min(hit.z, minz);
      maxz = std::max(hit.z, maxz);
    }

    // All these constants were determined by plotting mc.nu.vtx vs slc.boxmin
    // and slc.boxmax from CAFs (for true vertices inside the detector).

    // 50cm padding in x and y
    minx -= 50;
    maxx += 50;
    miny -= 50;
    maxy += 50;

    maxz = std::min(minz+250, maxz);
    minz -= 50;

    // cm-sized bins
    const int Nx = int(maxx-minx)+1;
    const int Ny = int(maxy-miny)+1;
    const int Nz = int(maxz-minz)+1;

    LiteTH2 houghx(Nz, minz, maxz, Nx, minx, maxx);
    LiteTH2 houghy(Nz, minz, maxz, Ny, miny, maxy);

    FillMaps(houghx, houghy, xhits, yhits);
    TVector3 vtx = FindMaximum(houghx, houghy);

    // Zoom into vertex with mm precision
    LiteTH2 houghxzoom(100, vtx.Z()-5, vtx.Z()+5, 100, vtx.X()-5, vtx.X()+5);
    LiteTH2 houghyzoom(100, vtx.Z()-5, vtx.Z()+5, 100, vtx.Y()-5, vtx.Y()+5);

    FillMaps(houghxzoom, houghyzoom, xhits, yhits);
    return FindMaximum(houghxzoom, houghyzoom);
  }

  //......................................................................
  void HoughVertexAlg::FillMaps(LiteTH2& houghx, LiteTH2& houghy,
                                const std::vector<Vector2D>& xhits,
                                const std::vector<Vector2D>& yhits) const
  {
    for(int view: {0, 1}){
      const std::vector<Vector2D>& hits = (view == 0) ? xhits : yhits;
      LiteTH2& hough = (view == 0) ? houghx : houghy;

      const float dy = hough.GetBinCenterY(1)-hough.GetBinCenterY(0);

      for(unsigned int i = 0; i < hits.size()-1; ++i){
        const Vector2D hi = hits[i];

        for(unsigned int j = i+1; j < hits.size(); ++j){
          const Vector2D hj = hits[j];

	  // Due to the slight tilt of planes this basically never happens
          if(hi.z == hj.z) continue;

	  // Solution for y=mx+c
          const float m = (hj.v-hi.v)/(hj.z-hi.z);
          const float c = hi.v - m*hi.z;

          for(int ix = 0; ix < hough.GetNbinsX()+2; ++ix){
            const float x = hough.GetBinCenterX(ix);

	    // Don't fill any weight downstream of the center of the
	    // most-upstream of the two cells.
            if(x > hi.z || x > hj.z) break;

            const float y0 = m*x+c;

            // Error propogation
            const float sigma_y_sq =
              util::sqr( hj.dv * (x-hi.z)/(hj.z-hi.z) ) +
              util::sqr( hi.dv * (hj.z-x)/(hj.z-hi.z) ) +
              util::sqr( hj.dz * (hj.v-hi.v)*(hi.z-x)/util::sqr(hj.z-hi.z) ) +
              util::sqr( hi.dz * (hj.v-hi.v)*(hj.z-x)/util::sqr(hj.z-hi.z) );

            const float sigma_y = sqrt(sigma_y_sq);

            const unsigned int Ny = hough.GetNbinsY();
            const double iy0 = hough.FindBinY(y0-3*sigma_y);
            double y = hough.GetBinCenterY(iy0)-dy;
            const double denom = 1/(9*sigma_y_sq);
            const double inv_sigma = 1/sigma_y;
            for(unsigned int iy = iy0; iy < Ny+2; ++iy){
              y += dy;

              if(y < y0-3*sigma_y) continue;
              if(y > y0+3*sigma_y) break;

              // exp() call in computation of gaussian shows up in profiles
              // const float w = exp(-util::sqr(y-y0)/(2*sigma_y_sq)) / sigma_y;

              // Triweight looks the most like normal, and has the nice
              // property that it goes to zero at 3 when scaled to the same SD
              // as the normal.
              const float w = util::cube(1-util::sqr(y-y0)*denom) * inv_sigma;

              hough.AddBinContent(ix, iy, w);
            }
          } // end for ix
        } // end for j
      } // end for i
    } // end for view
  }

  //......................................................................
  TVector3 HoughVertexAlg::FindMaximum(const LiteTH2& houghx,
                                       const LiteTH2& houghy) const
  {
    float bestw = -1;
    float bestz = 0, bestx = 0, besty = 0;

    // OK, both maps have same binning on x (z) axis.
    for(int iz = 0; iz < houghx.GetNbinsX()+2; ++iz){
      const float z = houghx.GetBinCenterX(iz);

      float local_bestx = 0, local_besty = 0;
      float local_bestwx = -1;
      float local_bestwy = -1;

      for(int ix = 0; ix < houghx.GetNbinsY()+2; ++ix){
        const float x = houghx.GetBinCenterY(ix);

        if(houghx.GetBinContent(iz, ix) > local_bestwx){
          local_bestwx = houghx.GetBinContent(iz, ix);
          local_bestx = x;
        }
      } // end for ix

      for(int iy = 0; iy < houghy.GetNbinsY()+2; ++iy){
        const float y = houghy.GetBinCenterY(iy);
        if(houghy.GetBinContent(iz, iy) > local_bestwy){
          local_bestwy = houghy.GetBinContent(iz, iy);
          local_besty = y;
        }
      } // end for iy

      if(local_bestwx + local_bestwy > bestw){
        bestw = local_bestwx+local_bestwy;
        bestz = z;
        bestx = local_bestx;
        besty = local_besty;
      }
    } // end for iz

    return TVector3(bestx, besty, bestz);
  }
} // namespace
