////////////////////////////////////////////////////////////////////////
///
/// \brief  TODO
/// \author brebel@fnal.gov
///
////////////////////////////////////////////////////////////////////////

#include <string>
#include <vector>

// NOvASoft includes
#include "MCCheater/BackTracker.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "Simulation/ParticleNavigator.h"
#include "Utilities/AssociationUtil.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"

namespace cheat {
  class ClusterCheater : public art::EDProducer {
  public:
    explicit ClusterCheater(fhicl::ParameterSet const& pset);
    virtual ~ClusterCheater();

    void produce(art::Event& evt);

    void reconfigure(fhicl::ParameterSet const& pset);

 private:

    std::string fHitModuleLabel;    ///< label for module creating rb::CellHit objects	   
    std::string fG4ModuleLabel;     ///< label for module running G4 and making particles, etc

  };
}

namespace cheat{

  //--------------------------------------------------------------------
  ClusterCheater::ClusterCheater(fhicl::ParameterSet const& pset)
  {
    this->reconfigure(pset);

    produces< std::vector<rb::Cluster> >();
    produces< art::Assns<rb::Cluster, rb::CellHit> >();
  }

  //--------------------------------------------------------------------
  ClusterCheater::~ClusterCheater()
  {
  }

  //--------------------------------------------------------------------
  void ClusterCheater::reconfigure(fhicl::ParameterSet const& pset)
  {
    fHitModuleLabel    = pset.get< std::string >("HitModuleLabel",    "calhit"  );
    fG4ModuleLabel     = pset.get< std::string >("G4ModuleLabel",     "geantgen");

    return;
  }

  //--------------------------------------------------------------------
  void ClusterCheater::produce(art::Event& evt)
  {
    // get a back tracker
    art::ServiceHandle<cheat::BackTracker> bt;

    sim::ParticleNavigator pnav = bt->ParticleNavigator();

    // grab the hits that have been reconstructed
    art::Handle< std::vector<rb::CellHit> > hitcol;
    evt.getByLabel(fHitModuleLabel, hitcol);

    // make a vector of them
    std::vector< art::Ptr<rb::CellHit> > hits;
    art::fill_ptr_vector(hits, hitcol);
    
    // loop over the hits and figure out which particle contributed to each one
    // make a map of vectors of art::Ptrs keyed by trackID values
    std::map< int, std::vector< art::Ptr<rb::CellHit> > > trackHitMap;

    // loop over all hits and fill in the map
    for(auto itr : hits ){

      std::vector<cheat::TrackIDE> trackides = bt->HitToTrackIDE(itr);

      // loop over all trackides for this hit
      for(auto const& e : trackides){

	// don't worry about eve particles that contribute less than 10% of the
	// energy in the current hit
	if( e.energyFrac < 0.1) continue;

	trackHitMap[e.trackID].push_back(itr);
	
      } // end loop over eve IDs for this hit
    }// end loop over hits

    // loop over the map and make clusters
    std::unique_ptr< std::vector<rb::Cluster> > clustercol(new std::vector<rb::Cluster>);
    std::unique_ptr< art::Assns<rb::Cluster, rb::CellHit> > assn(new art::Assns<rb::Cluster, rb::CellHit>);

    for(auto hitMapItr : trackHitMap){

      // separate out the hits for each particle into the different views
      std::vector< art::Ptr<rb::CellHit> > trackHits( hitMapItr.second );

      art::PtrVector<rb::CellHit> ptrvsX;
      art::PtrVector<rb::CellHit> ptrvsY;
	
      for(auto h : trackHits){

	if(h->View() == geo::kX) ptrvsX.push_back(h);
	if(h->View() == geo::kY) ptrvsY.push_back(h);
	  
      } // end loop over hits for this particle	
      
      // check to see if the particle id for this cluster lives in the
      // ParticleNavigator, if not, don't make a cluster and spew a warning
      if( pnav.find(hitMapItr.first) == pnav.end() ){
	mf::LogWarning("ClusterCheater") << " no particle in ParticleNavigator has track id "
					 << hitMapItr.first
					 << " this could be a problem.  Certainly won't make a"
					 << " Cluster for it.";
	continue;
      }

      if( std::abs(pnav[hitMapItr.first]->PdgCode()) != 11  &&
	  std::abs(pnav[hitMapItr.first]->PdgCode()) != 22  &&
	  std::abs(pnav[hitMapItr.first]->PdgCode()) != 111 )
	if(ptrvsX.size() < 1 || ptrvsY.size() < 1)
	  LOG_WARNING("ClusterCheater") << "no cluster in a view, " << ptrvsX.size()
					<< ", " << ptrvsY.size() << " for particle\n"
					<< *(pnav[hitMapItr.first]);

      if(ptrvsX.size() > 0 ){
	clustercol->push_back(rb::Cluster(geo::kX,
					  ptrvsX,
					  (hitMapItr.first * 1000) + geo::kX*10));
	// associate the hits to this cluster
	util::CreateAssn(*this, evt, *clustercol, ptrvsX, *assn);
      }

      LOG_DEBUG("ClusterCheater") << "adding cluster: \n" 
				  << clustercol->back()
				  << "\nto collection.";

      if(ptrvsY.size() > 0){
	clustercol->push_back(rb::Cluster(geo::kY,
					  ptrvsY,
					  (hitMapItr.first * 1000) + geo::kY*10));
	// associate the hits to this cluster
	util::CreateAssn(*this, evt, *clustercol, ptrvsY, *assn);
      }
                
      LOG_DEBUG("ClusterCheater") << "adding cluster: \n" 
				  << clustercol->back()
				  << "\nto collection.";

    } // end loop over the map

    evt.put(std::move(clustercol));
    evt.put(std::move(assn));

    return;

  } // end produce

} // end namespace cheat

DEFINE_ART_MODULE(cheat::ClusterCheater);
