////////////////////////////////////////////////////////////////////////
/// \brief   Simple module to analyze MC cosmics distributions
/// \author  messier@indiana.edu
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <map>
#include <vector>

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvA includes
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/CellGeo.h"
#include "MCCheater/BackTracker.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"
#include "Simulation/ParticleNavigator.h"
#include "Utilities/func/MathUtil.h"

// ROOT includes
#include "TH1F.h"
#include "TH2F.h"

namespace mcchk {
  /// A module to make some plots of generated cosmic-rays
  class CosmicAna : public art::EDAnalyzer {

  public:
    explicit CosmicAna(fhicl::ParameterSet const& pset);
    virtual ~CosmicAna();

    void analyze(art::Event const& evt);
    void beginRun(art::Run const& run);

  private:
    double fVtxBinSize; ///< Size of bins in cm for neutrino interaction vertex histograms
    double fRockDepthX; ///< Amount of space in cm outside of detector along x to plot interaction points
    double fRockDepthY; ///< Amount of space in cm outside of detector along y to plot interaction points
    double fRockDepthZ; ///< Amount of space in cm outside of detector along z to plot interaction points

    std::map<int, std::string> fParticleMap; ///< Map of particle PDG and name

    TH1F* fDmin;      ///< Closest approach for particles leaving hits
    TH1F* fDminNoHit; ///< Closest approach for particles leaving no hits

    std::map<int, TH2F*> fParticleAngles;
    std::map<int, TH2F*> fParticleAnglesLo;      ///< Particle rate vs angle, low momenta
    std::map<int, TH2F*> fParticleAnglesMi;      ///< Particle rate vs angle, middle momenta
    std::map<int, TH2F*> fParticleAnglesHi;      ///< Particle rate vs angle, high momenta
    std::map<int, TH1F*> fParticleCosTheta;      ///< Particle rate vs cos(Q)
    std::map<int, TH1F*> fParticleEnergy;        ///< Particle energy (GeV)
    std::map<int, TH2F*> fParticleEnergyVsAngle; ///< Particle Energy vs angle
    std::map<int, TH2F*> fParticleVertexXY;      ///< Particle vertex x vs y view
    std::map<int, TH1F*> fParticleVertexZ;       ///< Particle vertex z view

    TH1F* fVerticalMuons; ///< Number of vertical muons

    /// For each particle from ParticleNavigator...
    TH1F* fCRCosX; ///< Particle X Direction cosine
    TH1F* fCRCosY; ///< Particle Y Direction cosine
    TH1F* fCRCosZ; ///< Particle Z Direction cosine
    TH1F* fCRVtxX;
    TH1F* fCRVtxY;
    TH1F* fCRVtxZ;
    TH1F* fCREndX;
    TH1F* fCREndY;
    TH1F* fCREndZ;
    TH1F* fCRAvgCellsX;
    TH1F* fCRAvgCellsY;
    TH1F* fCREnergy;
    TH1F* fCRLowEnergy;
    TH1F* fCRLength;

    TH1F* fSampleTime; ///< Total sample time, number of events times 500 usec
  };
};

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  //......................................................................
  CosmicAna::CosmicAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset),
      fVtxBinSize(pset.get<double>("VtxBinSize")),
      fRockDepthX(pset.get<double>("RockDepthX")),
      fRockDepthY(pset.get<double>("RockDepthY")),
      fRockDepthZ(pset.get<double>("RockDepthZ")),
      fDmin(0) // Flag that histograms are uninitialized
  {
  }

  //......................................................................
  CosmicAna::~CosmicAna()
  {
  }

  //......................................................................
  void CosmicAna::beginRun(art::Run const& )
  {
    // Only make histograms once
    if(fDmin) { return; }

    art::ServiceHandle<art::TFileService> tfs;
    art::ServiceHandle<geo::Geometry> geo;

    // Create bounds and bins for vertex plots
    // Get detector coordinates
    double detX = geo->DetHalfWidth(); // -detX is minimum, +detX is maximum value
    double detY = geo->DetHalfHeight(); // Likewise, use +/- detY
    double detZ = geo->DetLength(); // Here, mininmum value is 0, max value is detZ

    // Add on the space for rock events
    double CRPosX = detX + fRockDepthX;
    double CRPosY = detY + fRockDepthY;
    double CRPosZMin = -fRockDepthZ;
    double CRPosZMax = detZ + fRockDepthZ;

    // This section adds a little extra space to make bins sizes exact
    int nbinsHelper = 0; // Helper to make bin sizes exact

    // First calculate number of complete bins from min to max values
    // Add one more bin for the last partial bin
    // For X/Y, half of those bins are above 0, half below
    // The max value is then the bin size times then number of bins over 2
    // For Z, calculate the adjusted number of bins in the same way
    // Leave the min value unchanged
    // Set the max value as the min value plus the number of bins times the bin size

    // Set the correct bounds
    nbinsHelper = (int)(2.*CRPosX/fVtxBinSize);
    ++nbinsHelper; // Add one extra bin to cover truncated area
    CRPosX = 0.5*fVtxBinSize*(double)nbinsHelper;

    nbinsHelper = (int)(2.*CRPosY/fVtxBinSize);
    ++nbinsHelper; // Add one extra bin to cover truncated area
    CRPosY = 0.5*fVtxBinSize*(double)nbinsHelper;

    nbinsHelper = (int)((CRPosZMax - CRPosZMin)/fVtxBinSize);
    ++nbinsHelper; // Add one extra bin to cover truncated area
    CRPosZMax = CRPosZMin + (fVtxBinSize*(double)nbinsHelper);

    // Set the correct number of bins
    int nBinsX = 2.*CRPosX/fVtxBinSize;
    int nBinsY = 2.*CRPosY/fVtxBinSize;
    int nBinsZ = (CRPosZMax - CRPosZMin)/fVtxBinSize;

    // Helpful for spacing/readability
    // Also helpful in case histograms are made in loops over, say, axis labels
    char HistoName[200];
    char HistoTitle[200];
    char TitleHelper[200];

    sprintf(TitleHelper, "Distance of Closest Approach to Detector Center for Particles That Left");

    sprintf(HistoName,  "fDmin");
    sprintf(HistoTitle, "%s Hits;d (cm);", TitleHelper);
    fDmin      = tfs->make<TH1F>(HistoName, HistoTitle, 100, 0., 500.E2);

    sprintf(HistoName,  "fDminNoHit");
    sprintf(HistoTitle, "%s No Hits;d (cm);", TitleHelper);
    fDminNoHit = tfs->make<TH1F>(HistoName, HistoTitle, 100, 0., 500.E2);

    fParticleMap[11]   = "Electron";
    fParticleMap[13]   = "Muon";
    fParticleMap[22]   = "Photon";
    fParticleMap[2112] = "Neutron";
    fParticleMap[2212] = "Proton";

    for(const auto& particlePair : fParticleMap) {
      int pdg                  = particlePair.first;
      std::string particleName = particlePair.second;

      sprintf(TitleHelper, "%s Angle Distribution", particleName.c_str());

      sprintf(HistoName,  "f%sAngles", particleName.c_str());
      sprintf(HistoTitle, "%s;#phi;cos#theta", TitleHelper);
      fParticleAngles[pdg]   = tfs->make<TH2F>(HistoName, HistoTitle,
                                               36, -180., 180., 50, -1., 1.);

      sprintf(HistoName,  "f%sAnglesLo", particleName.c_str());
      sprintf(HistoTitle, "%s for Low Momentum %ss;#phi;cos#theta", TitleHelper, particleName.c_str());
      fParticleAnglesLo[pdg] = tfs->make<TH2F>(HistoName, HistoTitle,
                                               36, -180., 180., 50, -1., 1.);

      sprintf(HistoName,  "f%sAnglesMi", particleName.c_str());
      sprintf(HistoTitle, "%s for Medium Momentum %ss;#phi;cos#theta", TitleHelper, particleName.c_str());
      fParticleAnglesMi[pdg] = tfs->make<TH2F>(HistoName, HistoTitle,
                                               36, -180., 180., 50, -1., 1.);

      sprintf(HistoName,  "f%sAnglesHi", particleName.c_str());
      sprintf(HistoTitle, "%s for High Momentum %ss;#phi;cos#theta", TitleHelper, particleName.c_str());
      fParticleAnglesHi[pdg] = tfs->make<TH2F>(HistoName, HistoTitle,
                                               36, -180., 180., 50, -1., 1.);

      sprintf(HistoName,  "f%sCosTheta", particleName.c_str());
      sprintf(HistoTitle, "%s cos#theta;cos#theta;", TitleHelper);
      fParticleCosTheta[pdg] = tfs->make<TH1F>(HistoName, HistoTitle, 50, -1., 1.);

      sprintf(HistoName,  "f%sVertexXY", particleName.c_str());
      sprintf(HistoTitle, "%s Vertex, X vs Y View;x (cm);y (cm)", particleName.c_str());
      fParticleVertexXY[pdg] = tfs->make<TH2F>(HistoName, HistoTitle,
                                               nBinsX, -CRPosX, CRPosX, nBinsY, -CRPosY, CRPosY);

      sprintf(HistoName,  "f%sVertexZ", particleName.c_str());
      sprintf(HistoTitle, "%s Vertex, Z View;z (cm);", particleName.c_str());
      fParticleVertexZ[pdg]  = tfs->make<TH1F>(HistoName, HistoTitle, nBinsZ, CRPosZMin, CRPosZMax);
    }

    // Make plots with energy by hand, as particles have vastly different energy scales
    sprintf(TitleHelper, "Energy;E (Gev);");

    sprintf(HistoName,  "fElectronEnergy");
    sprintf(HistoTitle, "Electron %s", TitleHelper);
    fParticleEnergy[11]   = tfs->make<TH1F>(HistoName, HistoTitle, 1000, 0., 1.);

    sprintf(HistoName,  "fMuonEnergy");
    sprintf(HistoTitle, "Muon %s", TitleHelper);
    fParticleEnergy[13]   = tfs->make<TH1F>(HistoName, HistoTitle, 10000, 0., 1000.);

    sprintf(HistoName,  "fPhotonEnergy");
    sprintf(HistoTitle, "Photon %s", TitleHelper);
    fParticleEnergy[22] = tfs->make<TH1F>(HistoName, HistoTitle, 200, 0., 5.);

    sprintf(HistoName,  "fNeutronEnergy");
    sprintf(HistoTitle, "Neutron %s", TitleHelper);
    fParticleEnergy[2112] = tfs->make<TH1F>(HistoName, HistoTitle, 100, 0., 10.);

    sprintf(HistoName,  "fProtonEnergy");
    sprintf(HistoTitle, "Proton %s", TitleHelper);
    fParticleEnergy[2212] = tfs->make<TH1F>(HistoName, HistoTitle, 100, 0., 10.);

    sprintf(TitleHelper, "Energy vs cos#theta for ");

    sprintf(HistoName,  "fElectronEnergyVsAngle");
    sprintf(HistoTitle, "%s Electrons;cos#theta;Energy (GeV)", TitleHelper);
    fParticleEnergyVsAngle[11]    = tfs->make<TH2F>(HistoName, HistoTitle, 200, -1., 1., 100, 0., 0.1);

    sprintf(HistoName,  "fMuonEnergyVsAngle");
    sprintf(HistoTitle, "%s Muons;cos#theta;Energy (GeV)", TitleHelper);
    fParticleEnergyVsAngle[13]    = tfs->make<TH2F>(HistoName, HistoTitle, 200, -1., 1., 10000, 0., 1000.);

    sprintf(HistoName,  "fPhotonEnergyVsAngle");
    sprintf(HistoTitle, "%s Photons;cos#theta;Energy (GeV)", TitleHelper);
    fParticleEnergyVsAngle[22]    = tfs->make<TH2F>(HistoName, HistoTitle, 200, -1., 1., 100, 0., 0.1);

    sprintf(HistoName,  "fNeutronEnergyVsAngle");
    sprintf(HistoTitle, "%s Neutrons;cos#theta;Energy (GeV)", TitleHelper);
    fParticleEnergyVsAngle[2112]  = tfs->make<TH2F>(HistoName, HistoTitle, 200, -1., 1., 100,  0., 10.);

    sprintf(HistoName,  "fProtonEnergyVsAngle");
    sprintf(HistoTitle, "%s Protons;cos#theta;Energy (GeV)", TitleHelper);
    fParticleEnergyVsAngle[2212]   = tfs->make<TH2F>(HistoName, HistoTitle, 200, -1., 1., 100,  0., 10.);

    sprintf(HistoName,  "fVerticalMuons");
    sprintf(HistoTitle, "Number of Vertical Muons;");
    fVerticalMuons = tfs->make<TH1F>(HistoName, HistoTitle, 1, -0.5, 0.5);

    sprintf(HistoName,  "fCRCosX");
    sprintf(HistoTitle, "Particle X Direction Cosine;dx/ds;Tracks");
    fCRCosX      = tfs->make<TH1F>(HistoName, HistoTitle, 200, -1., 1.);

    sprintf(HistoName,  "fCRCosY");
    sprintf(HistoTitle, "Particle Y Direction Cosine;dy/ds;Tracks");
    fCRCosY      = tfs->make<TH1F>(HistoName, HistoTitle, 200, -1., 1.);

    sprintf(HistoName,  "fCRCosZ");
    sprintf(HistoTitle, "Particle Z Direction Cosine;dz/ds;Tracks");
    fCRCosZ      = tfs->make<TH1F>(HistoName, HistoTitle, 200, -1., 1.);

    sprintf(TitleHelper, "Particle Entry Location, ");

    sprintf(HistoName,  "fCRVtxX");
    sprintf(HistoTitle, "%s X;x (cm);", TitleHelper);
    fCRVtxX      = tfs->make<TH1F>(HistoName, HistoTitle, nBinsX, -CRPosX, CRPosX);

    sprintf(HistoName,  "fCRVtxY");
    sprintf(HistoTitle, "%s Y;y (cm);", TitleHelper);
    fCRVtxY      = tfs->make<TH1F>(HistoName, HistoTitle, nBinsY, -CRPosY, CRPosY);

    sprintf(HistoName,  "fCRVtxZ");
    sprintf(HistoTitle, "%s Z;z (cm);", TitleHelper);
    fCRVtxZ      = tfs->make<TH1F>(HistoName, HistoTitle, nBinsZ, CRPosZMin, CRPosZMax);

    sprintf(TitleHelper, "Particle Exit Location, ");

    sprintf(HistoName,  "fCREndX");
    sprintf(HistoTitle, "%s X;x (cm);", TitleHelper);
    fCREndX      = tfs->make<TH1F>(HistoName, HistoTitle, nBinsX, -CRPosX, CRPosX);

    sprintf(HistoName,  "fCREndY");
    sprintf(HistoTitle, "%s Y;y (cm);", TitleHelper);
    fCREndY      = tfs->make<TH1F>(HistoName, HistoTitle, nBinsY, -CRPosY, CRPosY);

    sprintf(HistoName,  "fCREndZ");
    sprintf(HistoTitle, "%s Z;z (cm);", TitleHelper);
    fCREndZ      = tfs->make<TH1F>(HistoName, HistoTitle, nBinsZ, CRPosZMin, CRPosZMax);

    sprintf(TitleHelper, "Average Number of Cells Hit in ");

    sprintf(HistoName,  "fCRAvgCellsX");
    sprintf(HistoTitle, "%s X View;Cells;", TitleHelper);
    fCRAvgCellsX = tfs->make<TH1F>(HistoName, HistoTitle, 100, 0., 100.);

    sprintf(HistoName,  "fCRAvgCellsY");
    sprintf(HistoTitle, "%s Y View;Cells;", TitleHelper);
    fCRAvgCellsY = tfs->make<TH1F>(HistoName, HistoTitle, 100, 0., 100.);

    sprintf(HistoName,  "fCREnergy");
    sprintf(HistoTitle, "Energy of Particle;Energy (GeV);");
    fCREnergy    = tfs->make<TH1F>(HistoName, HistoTitle, 1000, 0., 1000.);

    sprintf(HistoName,  "fCRLowEnergy");
    sprintf(HistoTitle, "Energy of Low Energy Particles;Energy (GeV);");
    fCRLowEnergy = tfs->make<TH1F>(HistoName, HistoTitle, 200, 0., 2.);

    sprintf(HistoName,  "fCRLength");
    sprintf(HistoTitle, "Length of Particle in Detector;Length (cm);");
    fCRLength    = tfs->make<TH1F>(HistoName, HistoTitle, 500, 0., 5000.);

    sprintf(HistoName,  "fSampleTime");
    sprintf(HistoTitle, "Total Sample Time;Time (s);");
    fSampleTime = tfs->make<TH1F>(HistoName, HistoTitle, 1, -0.5, 0.5);
  }

  //......................................................................
  void CosmicAna::analyze(art::Event const& evt)
  {
    fSampleTime->Fill(0., 500.E-6); // Add 500 usec

    art::ServiceHandle<geo::Geometry> geo;

    art::ServiceHandle<cheat::BackTracker> bt;
    const sim::ParticleNavigator &pnav = bt->ParticleNavigator();

    // Flags for "interesting" particles
    double electronE = -1.0;
    double photonE   = -1.0;

    // Loop over particles in the event
    for(sim::ParticleNavigator::const_iterator it = pnav.begin(); it != pnav.end(); ++it) {
      const sim::Particle* particle = (*it).second;

      // Get a list of FLSHits corresponding to the current particle
      const std::vector<sim::FLSHit>& flshits = bt->ParticleToFLSHit(particle->TrackId());
      const int nflshits = flshits.size();

      const TLorentzVector& fourPosition = particle->Position();
      const TLorentzVector& fourMomentum = particle->Momentum();

      // Inputs to ClosestApproach function
      // Closest approach to this point
      double detCenter[3]   = {0., 0., 0.5*geo->DetLength()};
      // Point on the line--calculate closest approach of this line to point above
      double particleVtx[3] = {fourPosition.X(),  fourPosition.Y(),  fourPosition.Z()};
      // Slope of line
      double particleP[3]   = {fourMomentum.Px(), fourMomentum.Py(), fourMomentum.Pz()};
      double closestApp[3]; // Return vector
      double distClosestApp = geo::ClosestApproach(detCenter, particleVtx, particleP, closestApp);

      if(nflshits == 0) {
        fDminNoHit->Fill(distClosestApp);
      }
      else {
        fDmin->Fill(distClosestApp);
        if(distClosestApp > 10000.) {
          mf::LogVerbatim("Dmin > 10000") << "Run " << evt.run() << ", Event " << evt.id().event() << std::endl 
                                          << "Particle PDG " << particle->PdgCode() << std::endl
                                          << "Vertex "  << fourPosition.X()  << " " << fourPosition.Y()  << " " << fourPosition.Z()  << std::endl
                                          << "Momentum" << fourMomentum.Px() << " " << fourMomentum.Py() << " " << fourMomentum.Pz() << std::endl
                                          << "Distance of closest approach to detector center " << distClosestApp << std::endl;
        }
      } // end of (else) conditionals if size of FLSHit list from Particle is 0

      if(nflshits > 0) {
        int pdg = abs(particle->PdgCode());

        double energy = fourMomentum.E();
        double cosTheta = -fourMomentum.Py()/fourMomentum.P();
        double phi = atan2(fourMomentum.Pz(), fourMomentum.Px());
        phi *= 180./M_PI;

        // Add a vertical muon if angle is less than 10 degrees
        if(pdg == 13 && cosTheta > 0.984808) {
          fVerticalMuons->Fill(0.);
        }

        // Proceed only if the current particle is one we care about
        if(fParticleMap.find(pdg) != fParticleMap.end()) {
          fParticleCosTheta[pdg]->Fill(cosTheta);
          fParticleAngles[pdg]  ->Fill(phi, cosTheta);

          // Fill angle plots based on momentum
          if     (energy < 1.)  { fParticleAnglesLo[pdg]->Fill(phi, cosTheta); }
          else if(energy < 10.) { fParticleAnglesMi[pdg]->Fill(phi, cosTheta); }
          else                  { fParticleAnglesHi[pdg]->Fill(phi, cosTheta); }

          // Fill particle energy information
          fParticleEnergy[pdg]       ->Fill(energy);
          fParticleEnergyVsAngle[pdg]->Fill(cosTheta, energy);

          // Fill particle vertex information
          fParticleVertexXY[pdg]->Fill(fourPosition.X(), fourPosition.Y());
          fParticleVertexZ[pdg] ->Fill(fourPosition.Z());

          // Create matrix (vector of vectors) of number of hits in each cell
          std::vector<std::vector<double> > hitsByPlaneCell(geo->GetPlanesByView().size());
          unsigned int numCells = geo->Plane(0)->Ncells();
          if(geo->Plane(1)->Ncells() > numCells) { numCells = geo->Plane(1)->Ncells(); }

          // Resize each vector in the larger plane vector to have the number of cells, initialized to 0
          for(unsigned int i_plane = 0, n_plane = hitsByPlaneCell.size(); i_plane < n_plane; ++i_plane) {
            hitsByPlaneCell[i_plane].resize(numCells, 0.);
          }

          // Vector matching a plane to a view
          std::vector<geo::View_t> planeView(geo->GetPlanesByView().size());

          // Find earliest and latest plane/cell, entry/exit point
          double startT  = flshits[0].GetEntryT();
          double endT    = flshits[0].GetExitT();
          int startPlane = flshits[0].GetPlaneID();
          int endPlane   = startPlane;
          int startCell  = flshits[0].GetCellID();
          int endCell    = startCell;

          double entryPos[3] = {0.};
          double exitPos[3] = {0.};

          double localStartPos[3] = {flshits[0].GetEntryX(),
                                     flshits[0].GetEntryY(),
                                     flshits[0].GetEntryZ()};
          double localEndPos[3]   = {flshits[0].GetExitX(),
                                     flshits[0].GetExitY(),
                                     flshits[0].GetExitZ()};

          // Loop over FLSHits for this particle
          for(const auto& hit : flshits) {
            if     (hit.GetEntryT() < startT) { // Update quantities if this is an earlier hit
              startPlane = hit.GetPlaneID();
              startCell  = hit.GetCellID();
              localStartPos[0] = hit.GetEntryX();
              localStartPos[1] = hit.GetEntryY();
              localStartPos[2] = hit.GetEntryZ();
            }
            else if(hit.GetExitT() > endT) { // Update quantities if this is a later hit
              endPlane = hit.GetPlaneID();
              endCell  = hit.GetCellID();
              localEndPos[0] = hit.GetExitX();
              localEndPos[1] = hit.GetExitY();
              localEndPos[2] = hit.GetExitZ();
            }

            hitsByPlaneCell[hit.GetPlaneID()][hit.GetCellID()] += 1.; // Add a hit to the proper cell
            planeView[hit.GetPlaneID()] = geo->Plane(hit.GetPlaneID())->View();
          } // end of loop over FLSHits

          // Convert entry/exit locations to world coordinates
          geo->Plane(startPlane)->Cell(startCell)->LocalToWorld(localStartPos, entryPos);
          geo->Plane(endPlane)  ->Cell(endCell)  ->LocalToWorld(localEndPos,   exitPos);

          double planesX = 0.;
          double cellsX = 0.;
          double planesY = 0.;
          double cellsY = 0.;

          // Loop over planes, cells
          for(unsigned int i_plane = 0, n_plane = hitsByPlaneCell.size(); i_plane < n_plane; ++i_plane) {
            // Count the number of hit cells in this plane
            double hitCells = 0.;
            for(unsigned int i_cell = 0, n_cell = hitsByPlaneCell[i_plane].size(); i_cell < n_cell; ++i_cell) {
              if(hitsByPlaneCell[i_plane][i_cell] > 0.) { hitCells += 1.; }
            }

            if(planeView[i_plane] == geo::kX && hitCells > 0.) {
              planesX += 1.;
              cellsX += hitCells;
            }
            if(planeView[i_plane] == geo::kY && hitCells > 0.) {
              planesY += 1.;
              cellsY += hitCells;
            }
          } // end loop over hit planes/cells

          // Fill final plots
          fCRCosX->Fill(fourMomentum.Px()/fourMomentum.P());
          fCRCosY->Fill(fourMomentum.Py()/fourMomentum.P());
          fCRCosZ->Fill(fourMomentum.Pz()/fourMomentum.P());

          // Cosmic rays always come from above
          if(exitPos[1] > entryPos[1]) {
            fCREndX->Fill(entryPos[0]);
            fCREndY->Fill(entryPos[1]);
            fCREndZ->Fill(entryPos[2]);
            fCRVtxX->Fill(exitPos[0]);
            fCRVtxY->Fill(exitPos[1]);
            fCRVtxZ->Fill(exitPos[2]);
          }
          else { // Normal condition
            fCRVtxX->Fill(entryPos[0]);
            fCRVtxY->Fill(entryPos[1]);
            fCRVtxZ->Fill(entryPos[2]);
            fCREndX->Fill(exitPos[0]);
            fCREndY->Fill(exitPos[1]);
            fCREndZ->Fill(exitPos[2]);
          }

          if(planesX > 0.) { fCRAvgCellsX->Fill(cellsX/planesX); }
          else             { fCRAvgCellsX->Fill(0.); }

          if(planesY > 0.) { fCRAvgCellsY->Fill(cellsY/planesY); }
          else             { fCRAvgCellsY->Fill(0.); }

          fCREnergy->Fill(energy); // Fill particle energy

          // Fill low energy particles for higher resolution
          if(energy < 2.) {
            fCRLowEnergy->Fill(energy);
          }

          fCRLength->Fill(util::pythag(exitPos[0]-entryPos[0],
                                       exitPos[1]-entryPos[1],
                                       exitPos[2]-entryPos[2]));

          // Check for interesting particles
          if(energy > 0.5 && energy < 5.) {
            if(pdg == 11) { electronE = energy; }
            if(pdg == 22) { photonE   = energy; }
          }
        } // end of condition that current particle is in the particle map
      } // end of condition that size of list of FLSHits is greater than 0
    } // end of loop over particles in ParticleNavigator

    if(electronE > 0. || photonE > 0.) {
      mf::LogVerbatim("CosmicAna") << "Interesting particle:" << std::endl
                                   << "Run" << evt.getRun().run() << ", "
                                   << "Event " << evt.id().event() << std::endl
                                   << "Electron energy " << electronE << std::endl
                                   << "Photon energy " << photonE << std::endl;
    }
  }
} // end of namespace mcchk

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  DEFINE_ART_MODULE(CosmicAna);
}
