///////////////////////////////////////////////////////////////////////////////
/// \brief
/// \author messier@indiana.edu
///////////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <vector>

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvA includes
#include "CMap/CMap.h"
#include "DAQDataFormats/NanoSliceVersionConvention.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RawData/DAQHeader.h"
#include "RawData/RawDigit.h"

// ROOT includes
#include "TH1.h"
#include "TH2.h"
#include "TSystem.h"

namespace mcchk {
  /// A module to check the results from the Monte Carlo generator
  class DetAna : public art::EDAnalyzer {

  public:
    explicit DetAna(fhicl::ParameterSet const& pset);
    virtual ~DetAna();

    void analyze(art::Event const& evt);
    void beginRun(art::Run const& run);

  private:
    std::string fRawDataLabel; ///< module label for the raw data

    double fRawDigit_FDOffset; ///< How far to shift FD Number of digit bins

    TH1F* fNumDigits;     ///< number of digits per event
    TH1F* fADC;           ///< ADC value for each digit
    TH1F* fADCPerSample;  ///< ADC value for each sample
    TH1F* fADCDiff[3];    ///< Difference between ADC(1/2/3) and ADC(0)
    TH2F* fHitCellsXView; ///< cells hit in the x view
    TH2F* fHitCellsYView; ///< cells hit in the y view
  };
};

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  static int msg1cnt = 0;
  
  //......................................................................
  DetAna::DetAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset),
      fRawDataLabel(pset.get<std::string>("RawDataLabel", "daq")),
      fRawDigit_FDOffset(pset.get<double>("RawDigit_FDOffset")),
      fNumDigits(0) // Flag that all histograms are uninitialized
  {
  }

  //......................................................................
  DetAna::~DetAna()
  {
  }

  //......................................................................
  void DetAna::beginRun(art::Run const& )
  {
    if(fNumDigits) { return; } // Only make histograms once

    // Helpful for spacing/readability
    // Also helpful in case histograms are made in loops over, say, axis labels
    char HistoName[200];
    char HistoTitle[200];
    char TitleHelper[200];

    art::ServiceHandle<art::TFileService> tfs;
    art::ServiceHandle<geo::Geometry> geo;

    const std::set<unsigned int> planesX = geo->GetPlanesByView(geo::kX);
    const std::set<unsigned int> planesY = geo->GetPlanesByView(geo::kY);
    int fPlanes = planesX.size() + planesY.size();

    int fCellsX = geo->Plane(*planesX.begin())->Ncells();
    int fCellsY = geo->Plane(*planesY.begin())->Ncells();

    double edgeL =     0.;
    double edgeR = 10000.;
    if(geo->DetId() == novadaq::cnv::DetId::kFARDET) {
      edgeL += fRawDigit_FDOffset;
      edgeR += fRawDigit_FDOffset;
    }
    sprintf(HistoName,  "fNumDigits");
    sprintf(HistoTitle, "Number of RawDigits per Event;Digits;");
    fNumDigits    = tfs->make<TH1F>(HistoName, HistoTitle,
                                    1000, edgeL, edgeR);

    sprintf(HistoName,  "fADC");
    sprintf(HistoTitle, "ADC Value of the RawDigit;ADC;");
    fADC          = tfs->make<TH1F>(HistoName, HistoTitle,
                                    1000, 0., 1000.);

    sprintf(HistoName,  "fADCPerSample");
    sprintf(HistoTitle, "ADC Value for each Sample Read Out in the RawDigit;ADC;");
    fADCPerSample = tfs->make<TH1F>(HistoName, HistoTitle,
                                    1000, 0., 1000.);

    sprintf(TitleHelper, ") and ADC(0);#Delta ADC;");

    for(unsigned int i = 0, n = 3; i < n; ++i) {
      sprintf(HistoName,  "fADCDiff%d", i + 1);
      sprintf(HistoTitle, "Difference Between ADC(%d%s", i + 1, TitleHelper);

      fADCDiff[i] = tfs->make<TH1F>(HistoName, HistoTitle, 1000, 0., 1000.);
    }

    sprintf(TitleHelper, "Cells Hit in the");

    sprintf(HistoName,  "fHitCellsXView");
    sprintf(HistoTitle, "%s XZ View;Plane;Cell", TitleHelper);
    fHitCellsXView = tfs->make<TH2F>(HistoName, HistoTitle,
                                     fPlanes, 0., (float)fPlanes,
                                     fCellsX, 0., (float)fCellsX);

    sprintf(HistoName,  "fHitCellsYView");
    sprintf(HistoTitle, "%s YZ View;Plane;Cell", TitleHelper);
    fHitCellsYView = tfs->make<TH2F>(HistoName, HistoTitle,
                                     fPlanes, 0., (float)fPlanes,
                                     fCellsY, 0., (float)fCellsY);
  }

  //......................................................................
  void DetAna::analyze(art::Event const& evt)
  {
    // Get rawdata::RawDigits from the event
    bool isRawDigitListEmpty = false;
    art::Handle<std::vector<rawdata::RawDigit> > diglist;
    try {
      evt.getByLabel(fRawDataLabel, diglist);
      if(diglist->empty()) { isRawDigitListEmpty = true; }
    }
    catch(...) { isRawDigitListEmpty = true; }

    if(isRawDigitListEmpty) {
      if(msg1cnt < 5) {
        mf::LogWarning("DetAna RawDigit") << "Error retrieving raw digits." << std::endl;
        ++msg1cnt;
      }
      fNumDigits->Fill(0.);
      return;
    }

    std::vector<art::Ptr<rawdata::RawDigit> > rdptrs;
    art::fill_ptr_vector(rdptrs, diglist);

    fNumDigits->Fill(rdptrs.size()); // Fill number of digits in the event

    art::ServiceHandle<geo::Geometry> geo;
    art::ServiceHandle<cmap::CMap> cmap;
    geo::View_t view;

    static daqdataformats::NanoSliceVersionConvention conv;
    const unsigned int nSamples = conv.getNSamples(rdptrs[0]->Version());
    const unsigned int nPretrig = conv.getNPretriggeredSamples(rdptrs[0]->Version());

    // Loop over RawDigits
    // rd should be an art::Ptr<rawdata::RawDigit>
    for(const auto& rd : rdptrs) {
      fADC->Fill(rd->ADC());

      // Fill the ADC value of each sample
      for(unsigned int i_adc = 0, n_adc = rd->NADC(); i_adc < n_adc; ++i_adc) {
        fADCPerSample->Fill(rd->ADC(i_adc));
      }

      if(rd->NADC() == nSamples) {
        if(nSamples >= 4 && nPretrig >= 3) { // Make sure there are enough samples
          for(unsigned int i = 0, n = 3; i < n; ++i) {
            // Fill the difference between the ADC(1/2/3) and ADC(0)
            fADCDiff[i]->Fill(rd->ADC(i + 1) - rd->ADC(0));
          }
        }
        else {
          if(msg1cnt < 5) {
            mf::LogWarning("Multipoint") << "Not enough samples to fill ADC difference plots." << std::endl;
            ++msg1cnt;
          }
        }
      }
      else {
        mf::LogWarning("DetAna") << "The number of ADC samples in the RawDigit did not match"
                                 << " the number of samples in the DAQDataFormat." << std::endl;
      }

      // Ask channel map for plane and cell number
      view = geo->Plane(cmap->GetPlane(rd.get()))->View();
      if(view == geo::kX) {
        fHitCellsXView->Fill(1.*cmap->GetPlane(rd.get())+0.5, 1.*cmap->GetCell(rd.get())+0.5);
      }
      if(view == geo::kY) {
        fHitCellsYView->Fill(1.*cmap->GetPlane(rd.get())+0.5, 1.*cmap->GetCell(rd.get())+0.5);
      }
    } // end of loop over RawDigits
  }
} // end of namespace mcchk

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  DEFINE_ART_MODULE(DetAna);
}
