///////////////////////////////////////////////////////////////////////////////
/// \brief
/// \author messier@indiana.edu
///////////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <vector>

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvA includes
#include "Geometry/Geometry.h"
#include "MCCheater/BackTracker.h"
#include "Simulation/FLSHit.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCTruth.h"

// ROOT includes
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TParticle.h"
#include "TSystem.h"

namespace mcchk {
  /// A module to check the results from the Monte Carlo generator
  class DetSimAna : public art::EDAnalyzer {

  public:
    explicit DetSimAna(fhicl::ParameterSet const& pset);
    virtual ~DetSimAna();

    void analyze(art::Event const& evt);
    void beginJob();

  private:
    bool fDetOnly;

    std::string fMCTruthModuleLabel;

    TH1F* fNParticles_truth;  ///< Number of particles in MCTruth list
    TH1F* fNParticles_detsim; ///< Number of particles after DetSim, in ParticleNavigator list

    TH1F* fPDG_truth;         ///< PDG of all particles in MCTruth list
    TH1F* fPDG_detsim;        ///< PDG of all particles after DetSim, in ParticleNavigator list

    TH1F* fFLS_hits;          ///< FLS hits
    TH1F* fFLS_EDep;          ///< Total FLS energy deposition in event
    TH1F* fFLS_EDep_perhit;   ///< FLS energy deposition per hit

    TH1F* fFLS_Entry_X;       ///< FLS hit cell X entry
    TH1F* fFLS_Entry_Y;       ///<
    TH1F* fFLS_Entry_Z;       ///<

    TH2F* fFLS_Entry_XY;      ///< X vs Y entry
    TH2F* fFLS_Entry_XZ;      ///< X vs Z entry
    TH2F* fFLS_Entry_YZ;      ///< Y vs Z entry

    TH1F* fFLS_Delta_X;       ///< FLS hit passing distance in the cell along X axis
    TH1F* fFLS_Delta_Y;       ///<
    TH1F* fFLS_Delta_Z;       ///<
    TH1F* fFLS_Delta_R;       ///< Total passing distance in the cell
  }; // end of class declaration for DetSimAna
}; // end of namepsace mcchk

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  //......................................................................
  DetSimAna::DetSimAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset),
      fDetOnly(pset.get<bool>("DetOnly")),
      fMCTruthModuleLabel(pset.get<std::string>("MCTruthModuleLabel"))
  {
  }

  //......................................................................
  DetSimAna::~DetSimAna()
  {
  }

  //......................................................................
  void DetSimAna::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    // Helpful for spacing/readability
    // Also helpful in case histograms are made in loops over, say, axis labels
    char HistoName[200];
    char HistoTitle[200];

    sprintf(HistoName,  "fNParticles_truth");
    sprintf(HistoTitle, "Number of Particles in MCTruth List;Number of Particles in Event;");
    fNParticles_truth  = tfs->make<TH1F>(HistoName, HistoTitle, 200, 0., 200.);

    sprintf(HistoName,  "fNParticles_detsim");
    sprintf(HistoTitle, "Number of Particles in ParticleNavigator List;Number of Particles in Event;");
    fNParticles_detsim = tfs->make<TH1F>(HistoName, HistoTitle, 200, 0., 1000.);

    sprintf(HistoName,  "fPDG_truth");
    sprintf(HistoTitle, "In MCTruth;PDG;");
    fPDG_truth         = tfs->make<TH1F>(HistoName, HistoTitle, 10000, -5000., 5000.);

    sprintf(HistoName,  "fPDG_detsim");
    sprintf(HistoTitle, "PDG of Particles in ParticleNavigator List;PDG;");
    fPDG_detsim        = tfs->make<TH1F>(HistoName, HistoTitle, 10000, -5000., 5000.);

    sprintf(HistoName,  "fFLS_hits");
    sprintf(HistoTitle, "Number of FLS Hits in an Event;FLS hits");
    fFLS_hits          = tfs->make<TH1F>(HistoName, HistoTitle, 200, 0., 4000.);

    sprintf(HistoName,  "fFLS_EDep");
    sprintf(HistoTitle, "Total FLS Energy Deposited;Energy (GeV)");
    fFLS_EDep          = tfs->make<TH1F>(HistoName, HistoTitle, 1000, 0., 10.);

    sprintf(HistoName,  "fFLS_EDep_perhit");
    sprintf(HistoTitle, "FLS Energy Deposited per Hit;Energy (GeV)");
    fFLS_EDep_perhit   = tfs->make<TH1F>(HistoName, HistoTitle, 1000, 0., .01);

    sprintf(HistoName,  "fFLS_Entry_X");
    sprintf(HistoTitle, "FLS Hit Cell X Entry;FLS Entry X (cm)");
    fFLS_Entry_X       = tfs->make<TH1F>(HistoName, HistoTitle, 1000, -10., 10.);

    sprintf(HistoName,  "fFLS_Entry_Y");
    sprintf(HistoTitle, "FLS Hit Cell Y Entry;FLS Entry Y (cm)");
    fFLS_Entry_Y       = tfs->make<TH1F>(HistoName, HistoTitle, 1000, -10., 10.);

    sprintf(HistoName,  "fFLS_Entry_Z");
    sprintf(HistoTitle, "FLS Hit Cell Z Entry;FLS Entry Z (cm)");
    fFLS_Entry_Z       = tfs->make<TH1F>(HistoName, HistoTitle, 10000, -210., 210.);

    sprintf(HistoName,  "fFLS_Entry_XY");
    sprintf(HistoTitle, "FLS Hit Cell Entry;FLS Entry X (cm);FLS Entry Y (cm);");
    fFLS_Entry_XY      = tfs->make<TH2F>(HistoName, HistoTitle, 1000, -3.5,  3.5,
                                                                1000, -3.5,  3.5);

    sprintf(HistoName,  "fFLS_Entry_XZ");
    sprintf(HistoTitle, "FLS Hit Cell Entry;FLS Entry X (cm);FLS Entry Z (cm);");
    fFLS_Entry_XZ      = tfs->make<TH2F>(HistoName, HistoTitle, 1000, -3.5,  3.5,
                                                                5000, -210., 210.);

    sprintf(HistoName,  "fFLS_Entry_YZ");
    sprintf(HistoTitle, "FLS Hit Cell Entry;FLS_Entry Y (cm);FLS_Entry Z (cm);");
    fFLS_Entry_YZ      = tfs->make<TH2F>(HistoName, HistoTitle, 1000, -3.5,  3.5,
                                                                5000, -210., 210.);

    sprintf(HistoName,  "fFLS_Delta_X");
    sprintf(HistoTitle, "FLS Hit Passing X Distance in Cell;FLS Delta X (cm)");
    fFLS_Delta_X       = tfs->make<TH1F>(HistoName, HistoTitle, 20000, -10., 10.);

    sprintf(HistoName,  "fFLS_Delta_Y");
    sprintf(HistoTitle, "FLS Hit Passing Y Distance in Cell;FLS Delta Y (cm)");
    fFLS_Delta_Y       = tfs->make<TH1F>(HistoName, HistoTitle, 20000, -10., 10.);

    sprintf(HistoName,  "fFLS_Delta_Z");
    sprintf(HistoTitle, "FLS Hit Passing Z Distance in Cell;FLS Delta Z (cm)");
    fFLS_Delta_Z       = tfs->make<TH1F>(HistoName, HistoTitle, 20000, -10., 10.);

    sprintf(HistoName,  "fFLS_Delta_R");
    sprintf(HistoTitle, "FLS Hit Total Passing Distance in Cell;FLS Delta R (cm)");
    fFLS_Delta_R       = tfs->make<TH1F>(HistoName, HistoTitle, 10000,  0.,   10.);
  }
  
  //......................................................................
  void DetSimAna::analyze(art::Event const& evt)
  {
    // Get the MC generator information out of the event
    bool isMCTruthListEmpty = false;
    art::Handle<std::vector<simb::MCTruth> > mclist;
    try {
      evt.getByLabel(fMCTruthModuleLabel, mclist);
      if(mclist->size() == 0) { isMCTruthListEmpty = true; }
    }
    catch(...) { isMCTruthListEmpty = true; }

    std::vector<art::Ptr<simb::MCTruth> > mcptrs;
    art::fill_ptr_vector(mcptrs, mclist);

    // Create a particle navigator from BackTracker
    art::ServiceHandle<cheat::BackTracker> bt;
    const sim::ParticleNavigator& pnav = bt->ParticleNavigator();

    // Fill histograms based on MC Truth information
    // If there is no truth list, skip this section
    // But don't return, maybe the FLSHit list exists
    if(!isMCTruthListEmpty) {
      art::ServiceHandle<geo::Geometry> geo;
      double detHalfWidth  = geo->DetHalfWidth();
      double detHalfHeight = geo->DetHalfHeight();
      double detLength     = geo->DetLength();

      // Particles are accessed by MCTruth via BackTracker
      // Initialize values that are summed over interactions outside loop
      unsigned int n_particles_bt = 0;

      // Loop over MC Truth interactions
      for(const auto& mc : mcptrs) {
        // Skip non detector events if FCL says to
        if(fDetOnly) {
          const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
          const simb::MCParticle& nu = mc_neutrino.Nu();
          if(std::abs(nu.Vx()) > detHalfWidth ||
             std::abs(nu.Vy()) > detHalfHeight ||
             nu.Vz() < 0. || nu.Vz() > detLength) {
            continue;
          }
        }

        // Number of particles in the interaction
        unsigned int n_particle = mc->NParticles();

        // Fill the number of particles
        fNParticles_truth->Fill((float)n_particle);

        // Loop over particles in the interaction, fill their PDG values
        for(unsigned int i_particle = 0; i_particle < n_particle; ++i_particle) {
          fPDG_truth->Fill((float)mc->GetParticle(i_particle).PdgCode());
        }

        // Get the list of Particles associated with this MCTruth
        std::vector<const sim::Particle*> particles = bt->MCTruthToParticles(mc);
        n_particles_bt += particles.size();

        // Set of particle track IDs (no duplicates)
        std::set<int> trackIds;

        // Loop over particles in this MCTruth
        for(const auto& particle : particles) {
          fPDG_detsim->Fill(particle->PdgCode());

          trackIds.insert(particle->TrackId());
        }

        unsigned int n_FLSHits = 0;
        float EDep_Total = 0.;

        // Loop over all the track IDs
        for(const auto& trackId : trackIds) {
          std::vector<sim::FLSHit> hits = bt->ParticleToFLSHit(trackId);
          
          n_FLSHits += hits.size();
          
          for(const auto& hit : hits) {
            // Sum the total FLS energy deposited in the event
            EDep_Total += hit.GetEdep();

            // Fill FLS energy deposited per hit
            fFLS_EDep_perhit->Fill(hit.GetEdep());

            // Calculate FLS hit passing distance in the cell
            float deltaX = hit.GetEntryX() - hit.GetExitX();
            float deltaY = hit.GetEntryY() - hit.GetExitY();
            float deltaZ = hit.GetEntryZ() - hit.GetExitZ();
            float deltaR = sqrt(deltaX*deltaX + deltaY*deltaY + deltaZ*deltaZ);

            // Fill the FLS hit cell entry points
            fFLS_Entry_X->Fill(hit.GetEntryX());
            fFLS_Entry_Y->Fill(hit.GetEntryY());
            fFLS_Entry_Z->Fill(hit.GetEntryZ());

            fFLS_Entry_XY->Fill(hit.GetEntryX(), hit.GetEntryY());
            fFLS_Entry_XZ->Fill(hit.GetEntryX(), hit.GetEntryZ());
            fFLS_Entry_YZ->Fill(hit.GetEntryY(), hit.GetEntryZ());

            // Fill the FLS hit cell passing distances
            fFLS_Delta_X->Fill(deltaX);
            fFLS_Delta_Y->Fill(deltaY);
            fFLS_Delta_Z->Fill(deltaZ);
            fFLS_Delta_R->Fill(deltaR);
          } // end of loop over individual FLS hits
        } // end of loop over track IDs

        fFLS_hits->Fill(n_FLSHits); // Fill number of FLS Hits

        fFLS_EDep->Fill(EDep_Total); // Fill total energy deposition
      } // end loop over interactions

      fNParticles_detsim->Fill((float)n_particles_bt);
    } // end conditional if there is no mctruth
  }
} // end of namespace mcchk

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  DEFINE_ART_MODULE(DetSimAna);
}
