/// Macro to export histogram content of a ROOT file to a html page
/// Based on $ROOTSYS/bin/hadd.C
/// Raphael Schroeter, May 2011

// C/C++ includes
#include <iostream>
#include <string.h>

// ROOT includes
#include "Riostream.h"
#include "TCanvas.h"
#include "TFile.h"
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TImage.h"
#include "TKey.h"
#include "TString.h"
#include "TSystem.h"

TString htmlpath; ///< Path to html page and local location of saved images
TString subDir;   ///< Name of current subdirectory

ofstream *histhtml; ///< Stream to send html code/text/commands to html file

double NormPOTCurrent; ///< Amount of POT used to generate input file

/// Write the initial portion of the html file and set up top level directory
void InitializeHtml(char* file);

/// Set up subdirectory structure of html file and local area
void WriteDirectoryStruct(TString inPath, TFile* inFile);

/// Get the amount of POT from input file
void GetPOTInfo(TString inPath, TFile* inFile);

/// Recursively look for each histogram in the inPath directory
void ScanRootfile(TString inPath, TFile* inFile);

/// Write the final lines of the html file
void Finalize();

//.............................................................................
void HtmlCosmic(char* file)
{
  TString topPath = "/"; // Path to top directory of histogram locations/subdirectories

  // Open the file
  TFile* inFile = new TFile(file);

  // Get the base file name
  std::string baseFile(file);
  baseFile.erase(0, baseFile.rfind("/"));

  InitializeHtml((char*)baseFile.c_str());
  WriteDirectoryStruct(topPath, inFile);
  GetPOTInfo("cosmicana", inFile);
  ScanRootfile(topPath, inFile);
  Finalize();

  return;
}

//.............................................................................
void InitializeHtml(char* file)
{
  // Set up html files for ease of image viewing
  gSystem->Exec("mkdir html_out/");

  char commandHelper[256];
  sprintf(commandHelper, "mkdir html_out/%s/", file);
  gSystem->Exec(commandHelper);

  sprintf(commandHelper, "html_out/%s/", file);
  htmlpath = commandHelper;
  histhtml = new ofstream(htmlpath + "index.html", ios::out);

  *histhtml << "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
            << "<html>\n <head>\n"
            << "<title>" << file << "</title>\n"
            << "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">\n"
            << "</head>\n<body>\n" << std::endl;

  *histhtml << "<a name=\"pagetop\" ></a>\n"
            << "<h1>" << file << "</h1>\n" << std::endl;

  *histhtml << "Skip to:\n" << std::endl;

  return;
}

//.............................................................................
void WriteDirectoryStruct(TString inPath, TFile* inFile)
{
  inFile->cd(inPath);
  TDirectory* current_sourcedir = gDirectory;

  // Loop over all keys in this directory and write the subdir index on the html file
  TIter nextkey(current_sourcedir->GetListOfKeys());
  TKey* key;
  TKey* oldkey = 0;

  while((key = (TKey*)nextkey())) {
    // Keep only the highest cycle number for each key
    if(oldkey && !strcmp(oldkey->GetName(),key->GetName())) { continue; }
    
    // Read object from the input file
    inFile->cd(inPath);
    TObject* obj = key->ReadObj();

    if(obj->IsA()->InheritsFrom(TDirectory::Class())) { // It's a subdirectory
      subDir = obj->GetName();
      *histhtml << "<li><a href=\"#" << subDir << "\">" << subDir << "</a></li>\n" << std::endl;
    }

    delete obj;
  } // end of while loop over TKeys

  return;
}

//.............................................................................
void GetPOTInfo(TString inPath, TFile* inFile)
{
  // Name of the histogram to get POT information from
  std::string histPOT = "fSampleTime";

  if(inFile->cd(inPath)) {
    TKey* keyPOT = (TKey*)gDirectory->GetListOfKeys()->FindObject(histPOT.c_str());

    if(keyPOT) {
      TH1* hPOT = (TH1*)keyPOT->ReadObj();
      NormPOTCurrent = hPOT->Integral();
      std::cout << "Total sample time in current release " << NormPOTCurrent << std::endl;
    }
    else {
      NormPOTCurrent = 1.;
    }
  }
  else {
    std::cout << "POT info not available!" << std::endl;
    NormPOTCurrent = 1.;
  }

  return;
}

//.............................................................................
void ScanRootfile(TString inPath, TFile* inFile)
{
  TString saveName;

  inFile->cd(inPath);
  TDirectory* current_sourcedir = gDirectory;

  // Loop over all keys in this directory and save all histos found as PNG images
  TIter nextkey(current_sourcedir->GetListOfKeys());
  TKey* key;
  TKey* oldkey = 0;

  while((key = (TKey*)nextkey())) {
    // Keep only the highest cycle number for each key
    if(oldkey && !strcmp(oldkey->GetName(),key->GetName())) { continue; }

    // Read object from the input file
    inFile->cd(inPath);
    TObject* obj = key->ReadObj();

    if(obj->IsA()->InheritsFrom(TH1::Class())) { // Could be TH1, TH2, or TH3
      std::cout << "Exporting histogram " << obj->GetName() << std::endl;
      TH1* hist = (TH1*)obj;

      TCanvas* c = new TCanvas("c", "c", 800, 600);
      saveName = subDir + "/";

      TImage* img = TImage::Create();

      // Add POT scale to energy histogram axis label
      if(strstr(hist->GetName(), "Enu")) {
        char yAxisLabel[256];
        sprintf(yAxisLabel, "Events / %.2E POT", NormPOTCurrent);
        hist->GetYaxis()->SetTitle(yAxisLabel);
      }

      // Use colz option for TH2 objects
      if(obj->IsA()->InheritsFrom(TH2::Class())) {
        hist->Draw("colz");
      }
      else {
        hist->Draw();
      }

      img->FromPad(c);
      img->WriteImage(htmlpath + saveName + hist->GetName() + ".png");

      *histhtml << "<p>" << subDir << "-"<< hist->GetName() << std::endl;
      *histhtml << "<p><p><img src=\"" << saveName +  hist->GetName() << ".png\">" << std::endl;
      *histhtml << "<a href=\"#pagetop\" >Back to top</a>\n" << std::endl;

      delete c;
      delete img;
    } // end of conditonal if object is TH1
    else if(obj->IsA()->InheritsFrom(TDirectory::Class())) { // It's a subdirectory
      subDir = obj->GetName();
      std::cout << "Found subdirectory " << subDir << std::endl;

      // Make local directories for the images
      gSystem->Exec("mkdir " + htmlpath + subDir);

      *histhtml << "<a name=\"" << subDir << "\">\n"
                << "<h2>" << subDir << "</h2>\n" << std::endl;

      // Create a new subdirectory
      TString newDir = obj->GetName();

      ScanRootfile(newDir, inFile);
    }
    else { // Object is of no type that this macro can handle
      std::cout << "Unknown object type, name: " << obj->GetName()
                << " title: " << obj->GetTitle() << std::endl;
    }

    delete obj;
  } // end of while loop over TKeys

  return;
}

//.............................................................................
void Finalize()
{
  *histhtml << "</body></html>" << std::endl;
  histhtml->close();

  return;
}
