///////////////////////////////////////////////////////////////////////////////
/// \brief
/// \author messier@indiana.edu
///////////////////////////////////////////////////////////////////////////////

// C/C++ Includes
#include <cmath>
#include <map>
#include <vector>

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvA includes
#include "Geometry/Geometry.h"
#include "GeometryObjects/CoordinateTransformation.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCTruth.h"

// ROOT includes
#include "TF1.h"
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TSystem.h"

namespace mcchk {
  /// Check the results from the Monte Carlo generator
  /// This module analyzes neutrinos and neutrino interactions
  class NeutrinoAna : public art::EDAnalyzer {

  public:
    explicit NeutrinoAna(fhicl::ParameterSet const& pset);
    virtual ~NeutrinoAna();

    void analyze(art::Event const& evt);
    void beginJob();
    void beginRun(art::Run const& run);

    double DotProduct(double*, double*); ///< Calculate dot product between two vectors

  private:
    /// Calculate true L, E, L/E from neutrino interaction information
    bool AnalyzeNeutrinoInteraction(const art::Ptr<simb::MCTruth>& mc,
                                    const art::Ptr<simb::MCFlux>&  mcflux);

    bool fDetOnly;

    std::string fMCTruthModuleLabel; ///< label for module producing mc truth information

    int fEnu_Nbins;  ///< Number of bins for energy histograms
    double fEnu_min; ///< Minimum energy for energy histograms
    double fEnu_max; ///< Maximum energy for energy histgorams

    double fVtxBinSize; ///< Size of bins in cm for neutrino interaction vertex histograms
    double fRockDepthX; ///< Amount of space in cm outside of detector along x to plot neutrino interaction points
    double fRockDepthY; ///< Amount of space in cm outside of detector along y to plot neutrino interaction points
    double fRockDepthZ; ///< Amount of space in cm outside of detector along z to plot neutrino interaction points

    /// Neutrino information

    TH1F* fEnu_nue_CC;     ///< Neutrino energy for nue CC events
    TH1F* fEnu_nuebar_CC;  ///< Nuetrino energy for anti-nue CC events
    TH1F* fEnu_numu_CC;    ///< Neutrino energy for numu CC events
    TH1F* fEnu_numubar_CC; ///< Neutrino energy for anti-numu CC events
    TH1F* fEnu_nu_NC;      ///< Neutrino energy for neutrino NC events
    TH1F* fEnu_nubar_NC;   ///< Neutrino energy for anti-nu NC events

    std::map<int, TH1F*> fEnu_ByParent;      ///< Neutrino energy by parent species, all neutrinos
    std::map<int, TH1F*> fEnu_nue_ByParent;  ///< Neutrino energy by parent species, only (anti-)nu_e
    std::map<int, TH1F*> fEnu_numu_ByParent; ///< Neutrino energy by parent species, only (anti-)nu_mu
    std::map<int, TH1F*> fEnu_ByMode;        ///< Neutrino energy by interaction mode

    TH1F* fNeutrinoStartingPositionX;  ///< Neutrino Production Point X Coordinate
    TH1F* fNeutrinoStartingPositionY;  ///< Neutrino Production Point Y Coordinate
    TH1F* fNeutrinoStartingPositionZ;  ///< Neutrino Production Point Z Coordinate
    TH2F* fNeutrinoStartingPositionXY; ///< Neutrino Production Point X vs Y Coordinates
    TH2F* fNeutrinoStartingPositionXZ; ///< Neutrino Production Point X vs Z Coordinates
    TH2F* fNeutrinoStartingPositionYZ; ///< Neutrino Production Point Y vs Z Coordinates

    TH1F* fNeutrinoInteractionVertexX;  ///< Neutrino Interaction Vertex X Coordinate
    TH1F* fNeutrinoInteractionVertexY;  ///< Neutrino Interaction Vertex Y Coordinate
    TH1F* fNeutrinoInteractionVertexZ;  ///< Neutrino Interaction Vertex Z Coordinate
    TH2F* fNeutrinoInteractionVertexXY; ///< Neutrino Interaction Vertex X vs Y Coordinates
    TH2F* fNeutrinoInteractionVertexXZ; ///< Neutrino Interaction Vertex X vs Z Coordinates
    TH2F* fNeutrinoInteractionVertexYZ; ///< Neutrino Interaction Vertex Y vs Z Coordinates
    //TH3F* fNeutrinoInteractionVertex; // Neutrino Interaction Vertex

    TH1F* fNuCosX; ///< X direction cosine of incoming neutrino
    TH1F* fNuCosY; ///< Y direction cosine of incoming neutrino
    TH1F* fNuCosZ; ///< Z direction cosine of incoming neutrino

    TH1F* fMode; ///< Interaction mode

    TH1F* fNeutrinoTrueL;      ///< True L for neutrino
    TH2F* fNeutrinoTrueETrueL; ///< True L and E of the neutrino
    TH1F* fNeutrinoTrueLOverE; ///< True L/E of the neutrino

    TH1F* fProtonMultiplicityAllNucl; ///< Proton multiplicity
    std::map<int, TH1F*> fProtonMultiplicity; ///< Proton multiplicity by nucleus PDG

    TH1F* fW;    ///< W (shower mass)
    TH1F* fX;    ///< Bjorken X
    TH1F* fY;    ///< Bjorken Y
    TH1F* fQSqr; ///< Q^2

    TH1F* fYUnscaled; ///< Difference between neutrino and charged lepton energies
  };
};

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  static int msg1cnt = 0;

  //......................................................................
  NeutrinoAna::NeutrinoAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset),
      fDetOnly(pset.get<bool>     ("DetOnly")),
      fMCTruthModuleLabel(pset.get<std::string>("MCTruthModuleLabel")),
      fEnu_Nbins (pset.get<int>   ("Enu_Nbins")),
      fEnu_min   (pset.get<double>("Enu_Emin")),
      fEnu_max   (pset.get<double>("Enu_Emax")),
      fVtxBinSize(pset.get<double>("VtxBinSize")),
      fRockDepthX(pset.get<double>("RockDepthX")),
      fRockDepthY(pset.get<double>("RockDepthY")),
      fRockDepthZ(pset.get<double>("RockDepthZ")),
      fNeutrinoInteractionVertexX(0) // Flag that these histograms are uninitialized
  {
  }

  //......................................................................
  NeutrinoAna::~NeutrinoAna()
  {
    mf::LogInfo("NeutrinoAna") << "~NeutrinoAna" << std::endl;
  }

  //......................................................................
  void NeutrinoAna::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    /// Helper variables for creating histograms
    std::string energyName = "fEnu_";
    std::string energyTitle = "E_{#nu} For ";
    std::string energyLabel = ";E_{#nu} (GeV);Events";

    // Create histograms for neutrino energy by flavor and current
    fEnu_nue_CC     = tfs->make<TH1F>("fEnu_nue_CC",     ("E_{#nu} for #nu_{e} CC events" + energyLabel).c_str(),
                                      fEnu_Nbins, fEnu_min, fEnu_max);
    fEnu_nuebar_CC  = tfs->make<TH1F>("fEnu_nuebar_CC",  ("E_{#nu} for anti-#nu_{e} CC events" + energyLabel).c_str(),
                                      fEnu_Nbins, fEnu_min, fEnu_max);
    fEnu_numu_CC    = tfs->make<TH1F>("fEnu_numu_CC",    ("E_{#nu} for #nu_{#mu} CC events" + energyLabel).c_str(),
                                      fEnu_Nbins, fEnu_min, fEnu_max);
    fEnu_numubar_CC = tfs->make<TH1F>("fEnu_numubar_CC", ("E_{#nu} for anti-#nu_{#mu} CC events" + energyLabel).c_str(),
                                      fEnu_Nbins, fEnu_min, fEnu_max);
    fEnu_nu_NC      = tfs->make<TH1F>("fEnu_nu_NC",      ("E_{#nu} for #nu NC events" + energyLabel).c_str(),
                                      fEnu_Nbins, fEnu_min, fEnu_max);
    fEnu_nubar_NC   = tfs->make<TH1F>("fEnu_nubar_NC",   ("E_{#nu} for anti-#nu NC events" + energyLabel).c_str(),
                                      fEnu_Nbins, fEnu_min, fEnu_max);

    // Helper maps for creating histograms of neutrino energy by parent
    std::map<int, std::string> decayMode;
    decayMode[1]  = "Klong";
    decayMode[5]  = "Kcharged";
    decayMode[11] = "Muon";
    decayMode[13] = "Pion";

    std::map<int, std::string> decayLabel;
    decayLabel[1]  = "K_{L}";
    decayLabel[5]  = "K^{+/-}";
    decayLabel[11] = "#mu^{+/-}";
    decayLabel[13] = "#pi^{+/-}";

    // Create histograms of neutrino energy by parent
    for(std::map<int, std::string>::iterator it = decayMode.begin(); it != decayMode.end(); ++it) {
      std::string histName = energyName + it->second;
      std::string histTitle = energyTitle + "Events Decayed From " + decayLabel[it->first];

      fEnu_ByParent[it->first]      = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                                      fEnu_Nbins, fEnu_min, fEnu_max);

      histName = energyName + "nue_" + it->second;
      histTitle = energyTitle + "#nu_{e} Events Decayed From " + decayLabel[it->first];
      fEnu_nue_ByParent[it->first]  = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                                      fEnu_Nbins, fEnu_min, fEnu_max);

      histName = energyName + "numu_" + it->second;
      histTitle = energyTitle + "#nu_{#mu} Events Decayed From " + decayLabel[it->first];
      fEnu_numu_ByParent[it->first] = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                                      fEnu_Nbins, fEnu_min, fEnu_max);
    }

    // Helper vector for creating histograms of neutrino energy by interaction mode
    std::vector<std::string> modeNames;
    modeNames.push_back("QE");
    modeNames.push_back("Res");
    modeNames.push_back("DIS");
    modeNames.push_back("Coh");

    // Create histograms of neutrino energy by interaction mode
    for(unsigned int i_mode = 0, n_mode = modeNames.size(); i_mode < n_mode; ++i_mode) {
      std::string histName = energyName + modeNames[i_mode];
      std::string histTitle = energyTitle + modeNames[i_mode] + " Events" + energyLabel;
      fEnu_ByMode[i_mode] = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                            fEnu_Nbins, fEnu_min, fEnu_max);
    }

    std::string hNameStart = "fNeutrinoStartingPosition";
    std::string hTitleStart = "Neutrino Production ";
    std::string hLabel = " (cm);Events";

    int nuPosNBins = 200;
    double nuPosXY = 100.;
    double nuPosZMin = -150.;
    double nuPosZMax = 75000.;

    std::string histName = hNameStart + "X";
    std::string histTitle = hTitleStart + "X Coordinate;x" + hLabel;
    fNeutrinoStartingPositionX = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                                 nuPosNBins, -nuPosXY, nuPosXY);

    histName = hNameStart + "Y";
    histTitle = hTitleStart + "Y Coordinate;y" + hLabel;
    fNeutrinoStartingPositionY = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                                 nuPosNBins, -nuPosXY, nuPosXY);

    histName = hNameStart + "Z";
    histTitle = hTitleStart + "Z Coordinate;z" + hLabel;
    fNeutrinoStartingPositionZ = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                                 nuPosNBins, nuPosZMin, nuPosZMax);

    histName = hNameStart + "XY";
    histTitle = hTitleStart + "Coordinates, XY View;x (cm);y (cm)";
    fNeutrinoStartingPositionXY = tfs->make<TH2F>(histName.c_str(), histTitle.c_str(),
                                                  nuPosNBins, -nuPosXY, nuPosXY,
                                                  nuPosNBins, -nuPosXY, nuPosXY);

    histName = hNameStart + "XZ";
    histTitle = hTitleStart + "Coordinates, XZ View;z (cm);x (cm)";
    fNeutrinoStartingPositionXZ = tfs->make<TH2F>(histName.c_str(), histTitle.c_str(),
                                                  nuPosNBins, nuPosZMin, nuPosZMax,
                                                  nuPosNBins, -nuPosXY,  nuPosXY);

    histName = hNameStart + "YZ";
    histTitle = hTitleStart + "Coordinates, YZ View;z (cm);y (cm)";
    fNeutrinoStartingPositionYZ = tfs->make<TH2F>(histName.c_str(), histTitle.c_str(),
                                                  nuPosNBins, nuPosZMin, nuPosZMax,
                                                  nuPosNBins, -nuPosXY,  nuPosXY);

    fNuCosX = tfs->make<TH1F>("fNuCosX", "Incoming #nu X Directron Cosine;dx/ds;Events", 200, -1., 1.);
    fNuCosY = tfs->make<TH1F>("fNuCosY", "Incoming #nu Y Directron Cosine;dy/ds;Events", 200, -1., 1.);
    fNuCosZ = tfs->make<TH1F>("fNuCosZ", "Incoming #nu Z Directron Cosine;dz/ds;Events", 200, -1., 1.);

    fMode = tfs->make<TH1F>("fMode", ";Interaction Mode;", 4, 0., 4.);
    for(unsigned int i_mode = 0, n_mode = modeNames.size(); i_mode < n_mode; ++i_mode) {
      fMode->GetXaxis()->SetBinLabel(i_mode+1, modeNames[i_mode].c_str()); // ROOT bin indices starts at 1, not 0
    }
    fMode->GetXaxis()->CenterLabels();

    fNeutrinoTrueL      = tfs->make<TH1F>("fNeutrinoTrueL", "Neutrino L;L (m);Events",
                                          201, -0.05, 2000.05);
    fNeutrinoTrueETrueL = tfs->make<TH2F>("fNeutrinoTrueETrueL", "Neutrino Energy vs Length;E (GeV);L (m)",
                                          1000, 0., 10., 10000, 100., 1100.);
    fNeutrinoTrueLOverE = tfs->make<TH1F>("fNeutrinoTrueLOverE", "Neutrino L/E;L/E (m/GeV);Events",
                                          2000, 0., 2000.);

    fProtonMultiplicityAllNucl = tfs->make<TH1F>("fProtonMultiplicityAllNucl",   ";Proton Multiplicity All Nuclei;", 15, 0., 15.);

    // Create histograms of proton multiplicity in map, using the PDG of nuclear target as the key
    fProtonMultiplicity[1000010010] = tfs->make<TH1F>("fProtonMultiplicityH",    ";Proton Multiplicity H;",          15, 0., 15.);
    fProtonMultiplicity[1000060120] = tfs->make<TH1F>("fProtonMultiplicityC12",  ";Proton Multiplicity C_{12};",     15, 0., 15.);
    fProtonMultiplicity[1000070140] = tfs->make<TH1F>("fProtonMultiplicityN14",  ";Proton Multiplicity N_{14};",     15, 0., 15.);
    fProtonMultiplicity[1000080160] = tfs->make<TH1F>("fProtonMultiplicityO16",  ";Proton Multiplicity O_{16};",     15, 0., 15.);
    fProtonMultiplicity[1000160320] = tfs->make<TH1F>("fProtonMultiplicityS32",  ";Proton Multiplicity S_{32};",     15, 0., 15.);
    fProtonMultiplicity[1000170350] = tfs->make<TH1F>("fProtonMultiplicityCl35", ";Proton Multiplicity Cl_{35};",    15, 0., 15.);
    fProtonMultiplicity[1000220480] = tfs->make<TH1F>("fProtonMultiplicityTi48", ";Proton Multiplicity Ti_{48};",    15, 0., 15.);
    fProtonMultiplicity[1000260560] = tfs->make<TH1F>("fProtonMultiplicityFe56", ";Proton Multiplicity Fe_{56};",    15, 0., 15.);

    fW    = tfs->make<TH1F>("fW",    ";W (shower mass);Events", 100, 0., 2.);
    fX    = tfs->make<TH1F>("fX",    ";X_{Bj};Events",          100, 0., 1.);
    fY    = tfs->make<TH1F>("fY",    ";Y_{Bj};Events",          100, 0., 1.);
    fQSqr = tfs->make<TH1F>("fQSqr", ";Q^{2};Events",           100, 0., 2.);

    fYUnscaled = tfs->make<TH1F>("fYUnscaled", "E_{#nu} - E_{lepton};#Delta E (GeV);Events", 500, 0., 10.);
  }

  //......................................................................
  void NeutrinoAna::beginRun(art::Run const& )
  {
    if(fNeutrinoInteractionVertexX) { return; }

    art::ServiceHandle<art::TFileService> tfs;
    art::ServiceHandle<geo::Geometry> geo;

    // First get detector coordinates
    double detX = geo->DetHalfWidth(); // -detX is minimum, +detX is maximum value
    double detY = geo->DetHalfHeight(); // Likewise, use +/- detY
    double detZ = geo->DetLength(); // Here, mininmum value is 0, max value is detZ

    // Add on the space for rock events
    double nuPosX = detX + fRockDepthX;
    double nuPosY = detY + fRockDepthY;
    double nuPosZMin = -fRockDepthZ;
    double nuPosZMax = detZ + fRockDepthZ;

    // This section adds a little extra space to make bins sizes exact
    int nbinsHelper = 0; // Helper to make bin sizes exact

    // First calculate number of complete bins from min to max values
    // Add one more bin for the last partial bin
    // For X/Y, half of those bins are above 0, half below
    // The max value is then the bin size times then number of bins over 2
    // For Z, calculate the adjusted number of bins in the same way
    // Leave the min value unchanged
    // Set the max value as the min value plus the number of bins times the bin size

    nbinsHelper = (int)(2.*nuPosX/fVtxBinSize);
    ++nbinsHelper; // Add one extra bin to cover truncated area
    nuPosX = 0.5*fVtxBinSize*(double)nbinsHelper;

    nbinsHelper = (int)(2.*nuPosY/fVtxBinSize);
    ++nbinsHelper; // Add one extra bin to cover truncated area
    nuPosY = 0.5*fVtxBinSize*(double)nbinsHelper;

    nbinsHelper = (int)((nuPosZMax - nuPosZMin)/fVtxBinSize);
    ++nbinsHelper; // Add one extra bin to cover truncated area
    nuPosZMax = nuPosZMin + (fVtxBinSize*(double)nbinsHelper);

    int nBinsX = 2.*nuPosX/fVtxBinSize;
    int nBinsY = 2.*nuPosY/fVtxBinSize;
    int nBinsZ = (nuPosZMax - nuPosZMin)/fVtxBinSize;

    std::string hNameStart = "fNeutrinoInteractionVertex";
    std::string hTitleStart = "Neutrino Vertex ";
    std::string hLabel = " (cm);Events";

    std::string histName = hNameStart + "X";
    std::string histTitle = hTitleStart + "X Coordinate;x" + hLabel;
    fNeutrinoInteractionVertexX = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                                  nBinsX, -nuPosX, nuPosX);

    histName = hNameStart + "Y";
    histTitle = hTitleStart + "Y Coordinate;y" + hLabel;
    fNeutrinoInteractionVertexY = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                                  nBinsY, -nuPosY, nuPosY);

    histName = hNameStart + "Z";
    histTitle = hTitleStart + "Z Coordinate;z" + hLabel;
    fNeutrinoInteractionVertexZ = tfs->make<TH1F>(histName.c_str(), histTitle.c_str(),
                                                  nBinsZ, nuPosZMin, nuPosZMax);

    histName = hNameStart + "XY";
    histTitle = hTitleStart + "Coordinates, XY View;x (cm);y (cm)";
    fNeutrinoInteractionVertexXY = tfs->make<TH2F>(histName.c_str(), histTitle.c_str(),
                                                   nBinsX, -nuPosX, nuPosX,
                                                   nBinsY, -nuPosY, nuPosY);

    histName = hNameStart + "XZ";
    histTitle = hTitleStart + "Coordinates, XZ View;z (cm);x (cm)";
    fNeutrinoInteractionVertexXZ = tfs->make<TH2F>(histName.c_str(), histTitle.c_str(),
                                                   nBinsZ, nuPosZMin, nuPosZMax,
                                                   nBinsX, -nuPosX,   nuPosX);

    histName = hNameStart + "YZ";
    histTitle = hTitleStart + "Coordinates, YZ View;z (cm);y (cm)";
    fNeutrinoInteractionVertexYZ = tfs->make<TH2F>(histName.c_str(), histTitle.c_str(),
                                                   nBinsZ, nuPosZMin, nuPosZMax,
                                                   nBinsY, -nuPosY,   nuPosY);

    /*histName = hNameStart;
    histTitle = hTitleStart + "Coordinates;x (cm);y (cm);z (cm)";
    hNeutrinoInteractionVertex = tfs->make<TH3F>(histName.c_str(), histTitle.c_str(),
                                                 nBinsX, -nuPosX,   nuPosX,
                                                 nBinsY, -nuPosY,   nuPosY,
                                                 nBinsZ, nuPosZMin, nuPosZMax);*/
  }

  //......................................................................
  void NeutrinoAna::analyze(art::Event const& evt)
  {
    // Get the MC generator information out of the event
    bool isMCTruthListEmpty = false;
    art::Handle<std::vector<simb::MCTruth> > mclist;
    try {
      evt.getByLabel(fMCTruthModuleLabel, mclist);
      if(mclist->empty()) { isMCTruthListEmpty = true; }
    }
    catch(...) { isMCTruthListEmpty = true; }
    if(isMCTruthListEmpty) {
      mf::LogError("MCTruth Error") << "Error retrieving MCTruth list." << std::endl;
      return;
    }

    // Get the MCFlux information out of the event
    bool isMCFluxListEmpty = false;
    art::Handle<std::vector<simb::MCFlux> > fllist;
    try {
      evt.getByLabel(fMCTruthModuleLabel, fllist);
      if(fllist->empty()) { isMCFluxListEmpty = true; }
    }
    catch(...) { isMCFluxListEmpty = true; }
    if(isMCFluxListEmpty) {
      if(msg1cnt < 5) {
        mf::LogWarning("MCFlux Error") << "Error retrieving MCFlux list. Let it slide." << std::endl;
        ++msg1cnt;
      }
    }

    if(mclist->size() != fllist->size()) {
      mf::LogWarning("List sizes") << "MCTruth: "  << mclist->size()
                                   << ", MCFlux: " << fllist->size() << std::endl;
    }

    /// Loop over neutrino interactions
    for(unsigned int i_intx = 0, n_intx = mclist->size(); i_intx < n_intx; ++i_intx) {
      /// Pointers to current MCTruth and MCFlux
      art::Ptr<simb::MCTruth> mc    (mclist, i_intx);
      art::Ptr<simb::MCFlux>  mcflux(fllist, i_intx);

      // If there is no MCNeutrino, skip the event
      if(!mc->NeutrinoSet()) {
        if(msg1cnt < 5) {
          mf::LogError("MCNeutrino") << "There is no MCNeutrino in this event." << std::endl;
          ++msg1cnt;
        }

        return;
      }

      // Calculating true L, L/E from neutrino interaction
      // This will return false for non detector events if FCL says to skip them
      if(!AnalyzeNeutrinoInteraction(mc, mcflux)) { continue; }

      // Link to the MCNeutrino class
      // The class contains information not only about the incoming neutrino,
      // but also about the products of the decay
      const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
      const simb::MCParticle& nu  = mc_neutrino.Nu(); // Get the actual neutrino
      const simb::MCParticle& lep = mc_neutrino.Lepton(); // Get the outgoing lepton

      // Neutrino starting positions
      fNeutrinoStartingPositionX ->Fill(mcflux->fvx);
      fNeutrinoStartingPositionY ->Fill(mcflux->fvy);
      fNeutrinoStartingPositionZ ->Fill(mcflux->fvz);
      fNeutrinoStartingPositionXY->Fill(mcflux->fvx, mcflux->fvy);
      fNeutrinoStartingPositionXZ->Fill(mcflux->fvz, mcflux->fvx);
      fNeutrinoStartingPositionYZ->Fill(mcflux->fvz, mcflux->fvy);

      // Neutrino interaction vertex
      fNeutrinoInteractionVertexX ->Fill(nu.Vx());
      fNeutrinoInteractionVertexY ->Fill(nu.Vy());
      fNeutrinoInteractionVertexZ ->Fill(nu.Vz());
      fNeutrinoInteractionVertexXY->Fill(nu.Vx(), nu.Vy());
      fNeutrinoInteractionVertexXZ->Fill(nu.Vz(), nu.Vx());
      fNeutrinoInteractionVertexYZ->Fill(nu.Vz(), nu.Vy());
      //fNeutrinoInteractionVertex->Fill(nu.Vx(), nu.Vy(), nu.Vz());

      // Store these locally to repetitive function calls
      int nuPDG = nu.PdgCode();
      double nuEnergy = nu.E();

      // Fill nu energy histogram by flavor and current
      if(mc_neutrino.CCNC() == simb::kCC) {
        if     (nuPDG == +14) { fEnu_numu_CC   ->Fill(nuEnergy); }
        else if(nuPDG == -14) { fEnu_numubar_CC->Fill(nuEnergy); }
        else if(nuPDG == +12) { fEnu_nue_CC    ->Fill(nuEnergy); }
        else if(nuPDG == -12) { fEnu_nuebar_CC ->Fill(nuEnergy); }
        else { 
          mf::LogWarning("Unknown Flavor") << "Unknown neutrino flavor when trying to fill Enu CC histograms." << std::endl;
        }
      }
      else {
        if(nuPDG > 0) { fEnu_nu_NC   ->Fill(nuEnergy); }
        else          { fEnu_nubar_NC->Fill(nuEnergy); }
      }

      // Find the decay mode (corresponding to parent type)
      // Then set the value to a corresponding map key
      int decayMode = mcflux->fndecay;
      if     (decayMode >= 1  && decayMode <= 4)  { decayMode = 1;  }
      else if(decayMode >= 5  && decayMode <= 10) { decayMode = 5;  }
      else if(decayMode >= 11 && decayMode <= 12) { decayMode = 11; }
      else if(decayMode >= 13 && decayMode <= 14) { decayMode = 13; }
      else { decayMode = 0; }

      // Fill nu energy by parent species histograms
      if(decayMode != 0) {
        fEnu_ByParent[decayMode]->Fill(nuEnergy);
        if(abs(nuPDG) == 12) { fEnu_nue_ByParent[decayMode] ->Fill(nuEnergy); }
        if(abs(nuPDG) == 14) { fEnu_numu_ByParent[decayMode]->Fill(nuEnergy); }
      }

      // Fill nu energy by interaction mode histograms
      int nuMode = mc_neutrino.Mode();
      if(fEnu_ByMode.find(nuMode) != fEnu_ByMode.end()) {
        fEnu_ByMode[nuMode]->Fill(nuEnergy);
      }

      fMode->Fill(nuMode); // Fill the interaction mode

      // Fill neutrino direction cosine histograms
      if(fabs(nu.P()) > 0.) {
        fNuCosX->Fill(nu.Px()/nu.P());
        fNuCosY->Fill(nu.Py()/nu.P());
        fNuCosZ->Fill(nu.Pz()/nu.P());
      } // end of checking whether P is nonzero

      // Proton multiplicity -- Look for protons in the particle stack
      int nprotons = 0; // number of protons

      // Loop over particles produced in the event
      // Get number of protons produced
      for(int i_particle = 0, n_particle = mc->NParticles(); i_particle < n_particle; ++i_particle) {
        const simb::MCParticle& part = mc->GetParticle(i_particle);
        if(part.PdgCode() == 2212) {
          ++nprotons;
        }
      }

      // Fill histograms of proton multiplicity
      fProtonMultiplicityAllNucl->Fill(nprotons);
      int targetPDG = mc_neutrino.Target();
      if(fProtonMultiplicity.find(targetPDG) != fProtonMultiplicity.end()) {
        fProtonMultiplicity[targetPDG]->Fill(nprotons);
      }

      fW->Fill(mc_neutrino.W());
      fX->Fill(mc_neutrino.X());
      fY->Fill(mc_neutrino.Y());
      fQSqr->Fill(mc_neutrino.QSqr());
      fYUnscaled->Fill(nuEnergy - lep.E());
    } // end loop over interactions
  }

  //......................................................................
  bool NeutrinoAna::AnalyzeNeutrinoInteraction(const art::Ptr<simb::MCTruth>& mc,
                                               const art::Ptr<simb::MCFlux>&  mcflux)
  {
    art::ServiceHandle<geo::Geometry> geo;

    // Link to the MCNeutrino class
    // The class contains information not only about the incoming neutrino,
    // but also about the products of the decay
    const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
    //const simb::MCParticle& incoming_nu = mc_neutrino.Nu(); // Incoming neutrino
    const simb::MCParticle& outgoing_lepton = mc_neutrino.Lepton(); // Outgoing lepton

    double detHalfWidth  = geo->DetHalfWidth();
    double detHalfHeight = geo->DetHalfHeight();
    double detLength = geo->DetLength();

    if(fDetOnly) {
      const simb::MCParticle& nu = mc_neutrino.Nu();
      if(std::abs(nu.Vx()) > detHalfWidth ||
         std::abs(nu.Vy()) > detHalfHeight ||
         nu.Vz() < 0. || nu.Vz() > detLength) {
        return false;
      }
    }

    // Neutrino origin vertex (parent decay point) in beam coordinates
    // The CLHEP::cm is required for the coordinate transformation
    double neutrinoTrueStart_BeamCoords[3];
    neutrinoTrueStart_BeamCoords[0] = mcflux->fvx * CLHEP::cm;
    neutrinoTrueStart_BeamCoords[1] = mcflux->fvy * CLHEP::cm;
    neutrinoTrueStart_BeamCoords[2] = mcflux->fvz * CLHEP::cm;

    // Neutrino interaction vertex in detector coordinates
    double neutrinoTrueEnd_DetectorCoords[3];
    neutrinoTrueEnd_DetectorCoords[0] = outgoing_lepton.Vx();
    neutrinoTrueEnd_DetectorCoords[1] = outgoing_lepton.Vy();
    neutrinoTrueEnd_DetectorCoords[2] = outgoing_lepton.Vz();

    // Neutrino origin vertex in detector coordinates
    double neutrinoTrueStart_DetectorCoords[3];

    geo::CoordinateTransformation* coordinateTransformation = geo->getCoordinateTransformation();

    // Try to get start point in detector coordinates
    // The output will have dimensions of CLHEP::cm
    if(!coordinateTransformation->transformBeamToDetectorCoordinates(neutrinoTrueStart_BeamCoords,
                                                                     neutrinoTrueStart_DetectorCoords)) {
      return false;
    }

    // Get neutrino L components
    double vecNeutrinoL[3];
    for(unsigned int i = 0, n = 3; i < n; ++i) {
      // Calculate the distance traveled by the neutrino
      vecNeutrinoL[i] = neutrinoTrueEnd_DetectorCoords[i] - neutrinoTrueStart_DetectorCoords[i]/CLHEP::cm;
      vecNeutrinoL[i] /= 100.; // L is typically given in meters, not cm (especially for L/E)
    }

    // Get neutrino L and E
    double neutrinoL = sqrt(DotProduct(vecNeutrinoL, vecNeutrinoL));
    double neutrinoE = mc_neutrino.Nu().E();

    // Fill histograms related to neutrino L, E
    fNeutrinoTrueL->Fill(neutrinoL);
    fNeutrinoTrueETrueL->Fill(neutrinoE, neutrinoL);
    fNeutrinoTrueLOverE->Fill(neutrinoL/neutrinoE);

    return true;
  }

  //......................................................................
  /// Calculate dot product between two vectors
  double NeutrinoAna::DotProduct(double* v1, double* v2)
  {
    double product = 0.0;

    for(int i = 0; i < 3; ++i) {
      product += v1[i] * v2[i];
    }

    return product;
  }
} // end of namespace mcchk

///////////////////////////////////////////////////////////////////////////////
namespace mcchk {
  DEFINE_ART_MODULE(NeutrinoAna);
}
