/* 
 * File:   DAQSimulator.cpp
 * Author: denis
 * 
 * Created on January 31, 2012, 4:56 PM
 */
#ifdef NOVACMAKE
#include "MockDataDAQ/include/DAQSimulator.h"
#include "MockDataDAQ/include/DCMSimulator.h"
#include "MockDataDAQ/include/BufferNodeSimulator.h"
#include "MockDataDAQ/include/GlobalTriggerSimulator.h"
#include "MockDataDAQ/include/DataLoggerSimulator.h"
#include "MockDataDAQ/include/Configuration.h"
#include "MockDataDAQ/include/Sorter.h"
#include "MockDataDAQ/include/NanoSliceOrder.h"
#include "MockDataDAQ/include/GlobalEventTime.h"
#else
#include "MockDataDAQ/DAQSimulator.h"
#include "MockDataDAQ/DCMSimulator.h"
#include "MockDataDAQ/BufferNodeSimulator.h"
#include "MockDataDAQ/GlobalTriggerSimulator.h"
#include "MockDataDAQ/DataLoggerSimulator.h"
#include "MockDataDAQ/Configuration.h"
#include "MockDataDAQ/Sorter.h"
#include "MockDataDAQ/NanoSliceOrder.h"
#include "MockDataDAQ/GlobalEventTime.h"
#endif

#include "DAQDataFormats/RawMilliSlice.h"
#include "DAQDataFormats/RawTrigger.h"
#include "DAQDataFormats/RawDataBlock.h"

#include "RawData/RawDigit.h"
#include "CMap/CMap.h"


// ART includes
#include "art/Framework/Services/Registry/ServiceHandle.h"


#include <utility>
#include <algorithm>

namespace mcdatadaq{

//------------------------------------------------------------------------------
DAQSimulator::DAQSimulator():
          fGlobalTrigger(0),
          fDataLoggerSimulator(0)
{
    print();
}

//------------------------------------------------------------------------------
DAQSimulator::~DAQSimulator(){

    /// Delete DCM simulators
    for(uint32_t idiblock = 0; idiblock < fDCMs.size()          ; idiblock++)
    for(uint32_t idcm = 0    ; idcm     < fDCMs[idiblock].size(); idcm++)
            delete fDCMs[idiblock][idcm];

    /// Delete BufferNode simulators
    for(uint32_t ibn = 0; ibn < fBufferNodes.size(); ibn++)
            delete fBufferNodes[ibn];
    
    /// Delete Global Trigger simulator
    if(fGlobalTrigger) delete fGlobalTrigger;

    /// Delete DataLogger simulator
    if(fDataLoggerSimulator) delete fDataLoggerSimulator;
}

//------------------------------------------------------------------------------
bool DAQSimulator::init(){

    art::ServiceHandle<cmap::CMap> cmap;
    const Configuration*           config = Configuration::getInstance();
    const int                      detector_id = config->getDetector();
    const uint32_t                 number_ofdiblocks_in_detector = cmap->Map()->getTotalNumberOfDiblocks();

    /// Construct DCM simulators
    for(daqchannelmap::diblock_t idiblock = 1; idiblock <= number_ofdiblocks_in_detector; idiblock++){

        std::vector<DCMSimulator*> dcms_in_diblock_vector;

	// Encode a DChan to for the purpose of getting diblock type (DCM, FEB, and pixel are arbitrary for this.)
	daqchannelmap::dchan              dchan                     = cmap->Map()->encodeDChan(detector_id,idiblock,1,1,1);
        const daqchannelmap::DiBlock_TYPE diblock_type              = cmap->Map()->getDiBlockType(dchan);
        const uint32_t                    number_of_dcms_in_diblock = cmap->Map()->getNumberOfDCM(diblock_type);

        for(daqchannelmap::dcm_id_t idcm = 1; idcm <= number_of_dcms_in_diblock; idcm++){
            DCMSimulator* dcmsim = new DCMSimulator(idiblock, idcm);
            dcms_in_diblock_vector.push_back(dcmsim);
        }// end of loop over diblocks

        fDCMs.push_back(dcms_in_diblock_vector);
    }// end of loop over DCMs

    /// Construct Buffer Nodes
    for(uint32_t ibn = 1; ibn <= config->getNumberOfBufferNodes(); ibn++){
        fBufferNodes.push_back(new BufferNodeSimulator(ibn));
    }// ned of constructing Buffer Nodes

    /// Construct Global Trigger
    fGlobalTrigger       = new GlobalTriggerSimulator();

    /// Construct Data Logger
    fDataLoggerSimulator = new DataLoggerSimulator();




    return true;
}// end of DAQSimulator::init

//------------------------------------------------------------------------------
bool DAQSimulator::processEvent(const std::vector<rawdata::RawDigit>& digits){

    const Configuration* config = Configuration::getInstance();

    std::cout<<"Producing Microslices\n";
    /// Produce MicroSlices in DCMs
    fillMicroSlices(digits);
    std::cout<<"End of Producing Microslices\n";

    /// If we are producing just MicroSlices, don't need to simulate anything else
    if(!config->isProducingJustMicroSlices()) {

        /// Produce MilliSlices in DCMs
        std::cout<<"Producing MilliSlices\n";
        std::vector<daqdataformats::RawMilliSlice*> all_millislices_in_event;
        fillMilliSlices(all_millislices_in_event);

        /// Produce a Trigger in Global Trigger
        std::cout<<"Producing Trigger\n";
        daqdataformats::RawTrigger* trigger = fGlobalTrigger->getNewTrigger();

        
        std::cout<<"Producing DataBlocks!!!!!!!!!!!!!\n";
        /// get DataBlocks from BufferNodes
        std::vector<daqdataformats::RawDataBlock*> all_datablocks;
        fillDataBlocks(all_millislices_in_event, trigger, all_datablocks);

        std::cout<<"Producing Events!!!!!!!!!!!!!!!!!!!!!!!!!\n";
        /// fill Event and Run in DataLogger
        fillEventAndRun(all_datablocks, trigger);
    }// end of checking whether we just need to produce MicroSlices

    /// Now we are ready to advance to the next event
    nextEvent();

    return true;
}


//------------------------------------------------------------------------------
bool DAQSimulator::fillEventAndRun(const std::vector<daqdataformats::RawDataBlock*>& all_datablocks,
                                   daqdataformats::RawTrigger* trigger){
    fDataLoggerSimulator->produceEvent(all_datablocks, trigger);
    return true;
}


//------------------------------------------------------------------------------
bool DAQSimulator::fillDataBlocks(const std::vector<daqdataformats::RawMilliSlice*>& input_millislices,
                                  daqdataformats::RawTrigger*                        input_trigger,
                                  std::vector<daqdataformats::RawDataBlock*>&        output_datablocks){

    /// If no Millislices, nothing to do
    if(input_millislices.size() == 0)return true;

    /// Send all MilliSlices from all DCMs to a destination BufferNode

    /// Empty vector of MilliSlices
    const std::vector<daqdataformats::RawMilliSlice*>  empty_vector_of_millislices;

    const std::vector<daqdataformats::RawMilliSlice*>* pointer_to_the_right_vector;

    /// Figure out destination buffer node from the first MilliSlice
    const uint32_t destination_buffer_node = input_millislices[0]->getDestBuffID();

    /// Loop over Buffer Nodes and construct DataBlocks
    for(uint32_t ibn = 0; ibn < fBufferNodes.size(); ++ibn){

        const uint32_t current_bn_id = fBufferNodes[ibn]->getBufferNodeID();

        if(current_bn_id == destination_buffer_node) pointer_to_the_right_vector = &input_millislices;
        else                                         pointer_to_the_right_vector = &empty_vector_of_millislices;

        daqdataformats::RawDataBlock* datablock = fBufferNodes[ibn]->getDataBlock(*pointer_to_the_right_vector, input_trigger);

        output_datablocks.push_back(datablock);
    }// end of loop over Buffer Nodes

    return true;
}



//------------------------------------------------------------------------------
bool DAQSimulator::fillMilliSlices(std::vector<daqdataformats::RawMilliSlice*>& all_millislices_in_event){

    for(daqchannelmap::diblock_t idiblock = 1; idiblock <= fDCMs.size(); idiblock++)
    for(daqchannelmap::dcm_id_t  idcm     = 1; idcm     <= fDCMs[idiblock-1].size(); idcm++){
        DCMSimulator* dcm_simulator = getDCMSimulator(idiblock, idcm);

        daqdataformats::RawMilliSlice* millislice = dcm_simulator->fillMilliSlice();
        all_millislices_in_event.push_back(millislice);

        /// Write out the MilliSlices for the event
        dcm_simulator->writeMilliSlices();
    }

    return true;
}

//------------------------------------------------------------------------------
bool DAQSimulator::fillMicroSlices(const std::vector<rawdata::RawDigit>& digits){

    const Configuration* config = Configuration::getInstance();
    const daqdataformats::version_t nanoslice_version = config->getNanoSliceVersion();
    
    nanoslice_order_t nanoslice_order_vector;
    getNanoSliceOrder(digits, nanoslice_order_vector);
    
    //print(digits, nanoslice_order_vector);
    
    art::ServiceHandle<cmap::CMap> cmap;
    // Filling NanoSlices
    for(uint32_t idigit = 0  ; idigit < nanoslice_order_vector.size(); ++idigit){

            const uint32_t                 index         = nanoslice_order_vector[idigit];
            const rawdata::RawDigit*       rd            = &(digits[index]);   //! Get Raw digit

            const daqchannelmap::lchan     channel       = rd->Channel();
            const daqchannelmap::dchan     daqchannel    = cmap->Map()->encodeDChan(channel);

            const daqchannelmap::diblock_t diblock       = cmap->Map()->getDiBlock(daqchannel);
            const daqchannelmap::dcm_id_t  dcm           = cmap->Map()->getDCM(daqchannel);

            const feb_timestamp            timestamp     = rd->TDC();
            const uint32_t                 pulse_height  = rd->ADC();

            std::vector<uint32_t> data;
            data.push_back(pulse_height);

            /// Filling other fields of data
            // TODO: If simulation finally sets other ADC fields (greater than 0), then need to change this
            if(nanoslice_version == 1){
                data.push_back(1);
                data.push_back(2);
                data.push_back(1234);
            }// end of filling other fields of data

            DCMSimulator*                  dcm_simulator = getDCMSimulator(diblock, dcm);

            dcm_simulator->generateNanoSlice(daqchannel, timestamp, data);
    }/// end of looping over digits

    /// After we are done, write them out
    for(daqchannelmap::diblock_t idiblock = 1; idiblock <= fDCMs.size(); idiblock++)
    for(daqchannelmap::dcm_id_t  idcm     = 1; idcm     <= fDCMs[idiblock-1].size(); idcm++){
        DCMSimulator* dcm_simulator = getDCMSimulator(idiblock, idcm);

        dcm_simulator->writeMicroSlices();
    }// end of writing out MicroSlices


    return true;
}


//------------------------------------------------------------------------------
bool DAQSimulator::getNanoSliceOrder(const std::vector<rawdata::RawDigit>& digits, nanoslice_order_t& nanoslice_order_vector) const{


    art::ServiceHandle<cmap::CMap> cmap;
    const NanoSliceOrder*          nanoslice_order = NanoSliceOrder::getInstance();

    /// Sorter class. 
    /// It produces the order of nanoslices based on their location and time.
    Sorter sorter;

    // Need to arrange the Raw DigitList in the FEB order and then in the time order
    for(uint32_t i = 0; i < digits.size(); ++i){

            const rawdata::RawDigit* rd               = &(digits[i]);   //! Get Raw digit

            const daqchannelmap::lchan channel        = rd->Channel();
            const daqchannelmap::dchan daqchannel     = cmap->Map()->encodeDChan(channel);

            const daqchannelmap::diblock_t  diblock   = cmap->Map()->getDiBlock(daqchannel);
            const daqchannelmap::dcm_id_t   dcm       = cmap->Map()->getDCM(daqchannel);
            const daqchannelmap::feb_t      feb       = cmap->Map()->getFEB(daqchannel);
            const daqchannelmap::pixel_t    pixel     = cmap->Map()->getPixel(daqchannel);
            const feb_timestamp             timestamp = rd->TDC();

            const unsigned long long weight = nanoslice_order->calculateOrderingWeight(diblock,
										       dcm,
										       feb,
										       pixel,
										       timestamp);
            /// Add the nanoslice index with it's weight
            sorter.addNumber(weight);
    }// end of loop over digits

    /// Sorting out the NanoSlices in the right order
    sorter.sort();

    /// Nanoslice order. First by FEB, then by time
    /// Vector of Digit's index number
    sorter.getIndeces(nanoslice_order_vector);
    
    return true;
}



//------------------------------------------------------------------------------
bool DAQSimulator::print(const std::vector<rawdata::RawDigit>& digits,
                         const nanoslice_order_t&              digits_order) const {

    art::ServiceHandle<cmap::CMap> cmap;

    /// Printing digits in the order assigned by digits_order
    for(uint32_t idigit = 0; idigit < digits.size(); idigit++){

            const rawdata::RawDigit* rd = &(digits[digits_order[idigit]]);   //! Get Raw digit

            const uint32_t daqchannel = rd->DaqChannel();

            const uint32_t diblock    = cmap->Map()->getDiBlock(daqchannel);
            const uint32_t dcm        = cmap->Map()->getDCM(daqchannel);
            const uint32_t feb        = cmap->Map()->getFEB(daqchannel);
            const uint32_t pixel      = cmap->Map()->getPixel(daqchannel);

            const uint32_t TimeStamp  = rd->TDC();

            std::cout<<" DIB "<<diblock;
            std::cout<<" DCM "<<dcm;
            std::cout<<" FEB "<<feb;
            std::cout<<" TIME "<<TimeStamp;
            std::cout<<" PIX "<<pixel<<"\n";
    }// end of loop over digits

    return true;
}

//------------------------------------------------------------------------------
// Print the introductory version of the DAQ Simulator
bool DAQSimulator::print() const {

    std::cout<<"========================================================================================="<<std::endl;
    std::cout<<"DAQ Simulator Version "<<DAQSIMULATION_VERSION<<".00"<<std::endl;
    std::cout<<"========================================================================================="<<std::endl;

    return true;
}

//------------------------------------------------------------------------------
DCMSimulator* DAQSimulator::getDCMSimulator(const daqchannelmap::diblock_t diblock,
                                            const daqchannelmap::dcm_id_t  dcm)const{
    return fDCMs[diblock-1][dcm-1];
}

//------------------------------------------------------------------------------
bool DAQSimulator::nextEvent(){

    const Configuration* config            = Configuration::getInstance();
    GlobalEventTime*     global_event_time = GlobalEventTime::getInstance();

    /// Advance the event global time
    global_event_time->advanceEventTime();

    /// Get DCMs ready for the next event - clear the MicroSlice vector and the MilliSlice
    for(daqchannelmap::diblock_t idiblock = 1; idiblock <= fDCMs.size(); idiblock++)
    for(daqchannelmap::dcm_id_t  idcm = 1    ; idcm     <= fDCMs[idiblock-1].size(); idcm++){
        DCMSimulator* dcm_simulator = getDCMSimulator(idiblock, idcm);

        dcm_simulator->nextEvent();
    }// end of loop over DCMs getting them ready for the next event

    /// If we are producing just MicroSlices, don't need to simulate anything else
    if(!config->isProducingJustMicroSlices()){

        /// Make Buffer Nodes ready for the next event - clear DataBlock
        for(uint32_t ibn = 0; ibn < fBufferNodes.size(); ibn++){
            fBufferNodes[ibn]->nextEvent();
        }// end of making Buffer Nodes ready for the next event

        /// Make Data Logger ready for the next event - clear Event
        fDataLoggerSimulator->nextEvent();
    }// end of advancing event for Buffer nodes and data logger

    return true;
}

} // end of namespace
