// C/C++ & root includes
#include <iostream>
#include "vector"
#include <TTree.h>
#include "TH1D.h"
#include <map>

// ART includes

#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindOneP.h"

//NOvA includes
#include "Utilities/AssociationUtil.h"
#include "MuonRemove/art/MRCCParent.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "ReMId/ReMId.h"
#include "NumuEnergy/NumuE.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Vertex.h"
#include "MCCheater/BackTracker.h"
#include "LEM/PIDDetails.h"
#include "RecoJMShower/EID.h"
#include "RecVarPID/RVP.h"
#include "NumuSandbox/NumuSandFxs.h"
#include "SummaryData/POTSum.h"

namespace murem
{


  class FindParent : public art::EDProducer
  {
  public:
    explicit FindParent(fhicl::ParameterSet const& pset);
    virtual ~FindParent();

    void produce(art::Event& evt);
    void beginRun( art::Run& run);
    void reconfigure(const fhicl::ParameterSet& pset);


  private:

    int  MatchToOrigSlice(art::Ptr<rb::Cluster> mrSlice,
                          art::Handle< std::vector<rb::Cluster> > origSlices,
                          std::vector< art::Ptr< rb::CellHit> > mrHits,
                          float& maxPur,
                          float& maxEff);


    numusand::NumuSandFxs*     fNumuSandFxs;

    std::string fMrccLabel;
    std::string fMRClusterLabel;
    std::string fMRHitLabel;
    std::string fHitLabel;
    std::string fClusterLabel;

  };

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  FindParent::FindParent(fhicl::ParameterSet const& pset)
  {

    produces<std::vector<murem::MRCCParent> >();
    produces< art::Assns<murem::MRCCParent, rb::Cluster> >();
    this->reconfigure(pset);
  }


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  void murem::FindParent::reconfigure(const fhicl::ParameterSet& pset)
  {

    fClusterLabel       = pset.get< std::string      >("ClusterLabel");
    fHitLabel           = pset.get< std::string      >("HitLabel");

    fMrccLabel          = pset.get< std::string      >("MrccLabel");
    fMRClusterLabel     = pset.get< std::string      >("MRClusterLabel");
    fMRHitLabel         = pset.get< std::string      >("MRHitLabel");

  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////


  murem::FindParent::~FindParent()
  {
  }


  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  void murem::FindParent::beginRun( art::Run& run)
  {
    fNumuSandFxs = new numusand::NumuSandFxs();
  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  void FindParent::produce( art::Event& evt)
  {

    std::unique_ptr< std::vector< murem::MRCCParent > >
      mrccParents(new std::vector< murem::MRCCParent >);

    std::unique_ptr< art::Assns< murem::MRCCParent, rb::Cluster > >
      sliceAssassin(new art::Assns< murem::MRCCParent, rb::Cluster >);

    std::vector< murem::MRCCParent> mrccParentsVec;

    art::Handle< std::vector<rb::Cluster> > slicelist;
    evt.getByLabel(fClusterLabel, slicelist);
    if (slicelist->empty()) {
      mf::LogWarning("FindParent") << "Error retrieving slice list" ;
      return;
    }


    art::Handle< std::vector<rb::Cluster> > mrslicelist;
    evt.getByLabel(fMRClusterLabel, mrslicelist);
    if (mrslicelist->empty()) {
      mf::LogWarning("FindParent") << "Error retrieving mrcc slice list" ;
      return;
    }

    art::Handle< std::vector<simb::MCParticle> > removedMuonCol;
    evt.getByLabel( fMrccLabel, removedMuonCol);

    art::PtrVector<simb::MCParticle > removedMuons;
    for(int i = 0; i < (int)removedMuonCol->size(); i++){
      art::Ptr<simb::MCParticle> mu( removedMuonCol, i );
      removedMuons.push_back( mu );
    }


    // Grab all the MRCC cellhits. Need this for parent slice efficiency calculation

    art::Handle< std::vector<rb::CellHit> > mrHitlist;
    evt.getByLabel(fMRHitLabel, mrHitlist);

    std::vector< art::Ptr<rb::CellHit> > mrAllhits;
    for(size_t h = 0; h < mrHitlist->size(); ++h){
      art::Ptr<rb::CellHit> hit(mrHitlist, h);
      mrAllhits.push_back(hit);
    }

    int nSlicesMrcc = mrslicelist->size();

    for(int mrSli = 0; mrSli < nSlicesMrcc; mrSli++ ){

      art::Ptr<rb::Cluster> thisMrSlice(mrslicelist, mrSli);
      if( thisMrSlice->IsNoise() )
	continue;

      float    slicePur     = -5;
      float    sliceEff     = -5;


      int parentSliceIndex = MatchToOrigSlice( thisMrSlice, slicelist, mrAllhits,
					       slicePur, sliceEff);
      if (parentSliceIndex == -5 )
	continue;

      MRCCParent mrccParent;
      mrccParent.SetParentPur( slicePur );
      mrccParent.SetParentEff( sliceEff );
      mrccParent.SetParentIndex( parentSliceIndex );

      mrccParents->push_back( mrccParent );
      util::CreateAssn( *this, evt, *mrccParents, thisMrSlice ,*sliceAssassin);


    }// end loop over mrcc slices


    evt.put( std::move(mrccParents) );
    evt.put( std::move(sliceAssassin) );
  }// End of produce method





  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  int FindParent::MatchToOrigSlice( art::Ptr<rb::Cluster>     mrccSlice,
				    art::Handle<std::vector<rb::Cluster> > origSlices,
				    std::vector< art::Ptr<rb::CellHit> >   mrHits,
				    float& maxPur,
			 	    float& maxEff)
  {

    SortByPlane(mrHits);

    auto minPlane  = [](art::Ptr<rb::CellHit> hit, int plane ){ return (hit->Plane() < plane); };
    auto maxPlane  = [](int plane, art::Ptr<rb::CellHit> hit ){ return (hit->Plane() >= plane); };

    int nSlices           = origSlices->size();
    int nMrCells          = mrccSlice->NCell();
    int mrMin             = mrccSlice->MinPlane();
    int mrMax             = mrccSlice->MaxPlane();
    int matchedSliceIndex = -5;
    maxPur                = -1;

    for( int sli = 0; sli < nSlices; sli++ ){

      art::Ptr<rb::Cluster> thisSlice( origSlices, sli);

      // Don't match mrcc slices to original noise slice, even
      // if it legitimate. Noise slice info is not saved in
      // cafs and therefore, causes trouble at the time of analysis
      if( thisSlice->IsNoise() )
      	continue;

      int origMin = thisSlice->MinPlane();
      int origMax = thisSlice->MaxPlane();


      if( !( ( mrMin >= origMin && mrMin <= origMax ) ||
	     ( mrMax >= origMin && mrMax <= origMax ) ) )
	continue;


      int nCells    = thisSlice->NCell();

      int cellCount = 0;
      for(int h = 0; h < nMrCells; h++){
	art::Ptr<rb::CellHit> mhit  = mrccSlice->Cell(h);
	int plane = mhit->Plane();
	int cell  = mhit->Cell();
	int tdc   = mhit->TDC();

	for(int oh = 0; oh < nCells; oh++){
	  art::Ptr<rb::CellHit> hit  = thisSlice->Cell(oh);
	  if( plane == hit->Plane() &&
	      cell  == hit->Cell() &&
	      tdc   == hit->TDC())
	    cellCount++;
	}// end loop over orig slice hits

      }// end loop over hits in mrccSlice


      float pur = ((float)cellCount)/nMrCells;

      if( pur > maxPur ){
	matchedSliceIndex = sli;
	maxPur            = pur;
      }

    }// end loop over the orig slices

    // Once the original slice for which the mrcc slice has the max
    // purity is found, determine the efficiency of mrcc slice wrt
    // hits in the original matched slice minus the muon hits.

    if(matchedSliceIndex == -5)
      return matchedSliceIndex;
    
      
    art::Ptr<rb::Cluster> origSlice( origSlices, matchedSliceIndex );
    rb::Cluster nonMuonClust = *(origSlice);
    int nCells  = origSlice->NCell();
    int minP    = origSlice->MinPlane();
    int maxP    = origSlice->MaxPlane();


    art::PtrVector<rb::CellHit>::iterator mrhitIter, mrhitBegin, mrhitEnd;
    mrhitEnd    = std::upper_bound( mrHits.begin(), mrHits.end(),
				    maxP, maxPlane );
    mrhitBegin  = std::lower_bound(  mrHits.begin(), mrHits.end(),
				     minP, minPlane );


    for(int oh = 0; oh < nCells; oh++){
      const art::Ptr<rb::CellHit> hit  = origSlice->Cell(oh);
      int plane = hit->Plane();
      int cell  = hit->Cell();
      int tdc   = hit->TDC();
      bool matchedHit = false;

      for( mrhitIter = mrhitBegin; mrhitIter != mrhitEnd; ++mrhitIter){
	const art::Ptr<rb::CellHit> mrHit = *mrhitIter;
	if( mrHit->Plane() == plane &&
	    mrHit->Cell()  == cell  &&
	    mrHit->TDC()   == tdc    )
	  matchedHit = true;

      }// end loop over mrcc hits

      if( !matchedHit )
	nonMuonClust.RemoveHit( hit );

    }// end loop over origSlice hits

    // Now we have the muon removed original cells ,
    // Lets calculate the efficiency we're after.
    int nNonMuonHits = nonMuonClust.NCell();
    int hitCount = 0;

    art::PtrVector<rb::CellHit> nonMuonHits   = nonMuonClust.AllCells();
    art::PtrVector<rb::CellHit> mrccSliceHits = mrccSlice->AllCells();
    rb::SortByPlane(nonMuonHits);
    rb::SortByPlane(mrccSliceHits);

    int start = 0;
    for(int oh = 0; oh < nNonMuonHits; oh++){
      const art::Ptr<rb::CellHit> hit  = nonMuonHits[oh];
      int plane = hit->Plane();
      int cell  = hit->Cell();
      int tdc   = hit->TDC();

      for(int h = start; h < nMrCells; h++){
	const art::Ptr<rb::CellHit> mrHit = mrccSliceHits[h];
	if( mrHit->Plane() > plane ){
	  start = h;
	  break;
	}
	if( mrHit->Plane() == plane &&
	    mrHit->Cell()  == cell  &&
	    mrHit->TDC()   == tdc    ){
	  hitCount++;
	  break;
	}
      }// end loop over mrcc hits
    }// end loop over non-muon hits

    maxEff = (float)hitCount/nNonMuonHits;

    return matchedSliceIndex;

  }// End of MatchToOrigSlice



  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  DEFINE_ART_MODULE(FindParent);

}// End of namespace murem
