////////////////////////////////////////////////////////////////////////                                                         
/// \file  MergeDigits.cxx                                                                                                        
/// \brief A module to copy the original raw digits to a new label. Note that the new label is defined in the job file.
///                                                                                                                              
/// \version $Id: MergeDigits.cxx,v 1.2 2012-09-20 22:01:37 greenc Exp $                                                       
/// \author  ksachdev@physics.umn.edu                                                                                            
////////////////////////////////////////////////////////////////////////             

// C/C++ includes
#include <iostream>
#include <vector>

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"

// NOvASoft includes
#include "RawData/RawDigit.h"
#include "Simulation/FLSHitList.h"



namespace murem 
{
  class MergeDigits : public art::EDProducer 
  {
  public: 
    explicit MergeDigits(fhicl::ParameterSet const& pset);
    ~MergeDigits();
    
    void produce(art::Event& evt);
    
    void reconfigure(const fhicl::ParameterSet& pset);
    
  private: 
    
    std::string      fRawDigitModule;
    std::string      fElectronRawDigitModule;

  };
}


namespace murem
{

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  MergeDigits::MergeDigits(fhicl::ParameterSet const& pset)
  {
    
    this->reconfigure(pset);
    
    produces< std::vector <rawdata::RawDigit> >();
    
    mf::LogInfo("MergeDigits") <<"MergeDigits::MergeDigits()\n";

  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  MergeDigits::~MergeDigits()
  {
  
  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  void MergeDigits::reconfigure(const fhicl::ParameterSet& pset)
  {

    fRawDigitModule         = pset.get< std::string      >("RawDigitModule");
    fElectronRawDigitModule = pset.get< std::string      >("ElectronRawDigitModule");   
  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  void MergeDigits::produce(art::Event& evt)
  {
    bool isMC = !evt.isRealData();

    std::unique_ptr< std::vector<rawdata::RawDigit> > origDigits(new std::vector<rawdata::RawDigit>);
    std::vector< rawdata::RawDigit>                 newDigits;

    // Get the original set of raw digits in the event
  
    art::Handle< std::vector<rawdata::RawDigit> > digitcol;
    evt.getByLabel(fRawDigitModule, digitcol);
    art::PtrVector< rawdata::RawDigit> digits;

    for(unsigned int i = 0; i < digitcol->size(); ++i){
      art::Ptr<rawdata::RawDigit> dig(digitcol, i);
      digits.push_back(dig);
    }

    art::Handle< std::vector<rawdata::RawDigit> > edigitcol;
    evt.getByLabel(fElectronRawDigitModule, edigitcol);
    art::PtrVector< rawdata::RawDigit> edigits;
    
    for(unsigned int i = 0; i < edigitcol->size(); ++i){
      art::Ptr<rawdata::RawDigit> edig(edigitcol, i);
      edigits.push_back(edig);
    }

    int nDigs = digits.size();
    for(int jDig = 0; jDig < nDigs; jDig++){
      rawdata::RawDigit rd = *digits[jDig];
      unsigned int newADC = 0;
      float ratio = 1;

      int nEDigs = edigits.size();
      for(int iDig = 0; iDig < nEDigs; iDig++){
	if(rd.Channel() == edigits[iDig]->Channel()){
	  newADC = rd.ADC(0) + edigits[iDig]->ADC(0);
	  if( rd.ADC(0) != 0 )
	    ratio = newADC/rd.ADC(0); 
	  rd.SetADC(0, newADC);
	  rd.SetADC(1, ratio*rd.ADC(1));
	  rd.SetADC(2, ratio*rd.ADC(2));
	  edigits.erase(edigits.begin()+ iDig);
	}
      }// end loop over electron digits

      rd.SetMC(isMC);
      newDigits.push_back(rd);
    }// end loop over digits
    
    for(int i = 0; i< (int) edigits.size(); i++){
      rawdata::RawDigit edig = *edigits[i]; 
      edig.SetMC(isMC);
      newDigits.push_back(edig);
    }

    std::cout<<" ELEC DIGIT SIZE "<<edigits.size()<<std::endl;    
    std::cout<<" NEW DIG SIZE "<<newDigits.size()<<std::endl;
    
    for(unsigned int digs = 0; digs<newDigits.size(); digs++ ){
      origDigits->push_back( newDigits[digs] );
    }
    evt.put(std::move(origDigits));

    return;

  } // End of producer


  DEFINE_ART_MODULE(MergeDigits);


/*********************************************************************************************/
////////////////////////////////////////////////////////////////////////

}//end of namespace murem
