////////////////////////////////////////////////////////////////////////
/// \file  MRE.cxx
/// \brief Creates an electron in place on the muon removed by MuonRemove job
///
/// \version $Id: MRE.cxx,v 1.3 2012-10-01 23:24:12 bckhouse Exp $
/// \author  ksachdev@physics.umn.edu
////////////////////////////////////////////////////////////////////////



// C/C++ includes
#include <cmath>
#include <iostream>
#include <map>
#include <vector>

// ROOT includes
#include "TMath.h"
#include "TVector3.h"
#include "TLorentzVector.h"
#include "TDatabasePDG.h"

// Framework includes
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib/search_path.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"

// NOvASoft includes
#include "g4nova/G4Alg.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"
#include "MCCheater/BackTracker.h"


namespace art 
{
  class Event;
  class ParameterSet;
}

namespace murem 
{
  class MRE : public art::EDProducer 
  {
    public:
      explicit MRE(fhicl::ParameterSet const& pset); 
      ~MRE();                        
      
      void produce(art::Event& evt); 

      void reconfigure(const fhicl::ParameterSet& pset);

      void beginRun(art::Run& run);

  private:

      g4n::G4Alg*         fG4Alg;     ///< G4Helper object
      fhicl::ParameterSet fG4AlgPSet; ///< parameter set to configure the G4Alg object
      
      std::string         fMCTruthLabel;
      std::string         fMrccInfoLabel;

  };
}


////////////////////////////////////////////////////////////////////////


namespace murem
{

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  
  MRE::MRE(fhicl::ParameterSet const& pset) 
    : fG4Alg(0)
    , fG4AlgPSet(pset.get< fhicl::ParameterSet >("G4AlgPSet") )
  {
    this->reconfigure(pset);
    
    produces< std::vector<simb::MCTruth>     >();
    produces< std::vector<sim::FLSHitList>   >();
    produces< std::vector<sim::Particle>     >();


    mf::LogInfo("MRE") << " MRE::MRE()\n";
  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  MRE::~MRE()
  {
    if(fG4Alg) delete fG4Alg;
  }

  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////

  void MRE::reconfigure(const fhicl::ParameterSet& pset)
  {
    fMrccInfoLabel      = pset.get< std::string      >("MrccInfoLabel");
    fMCTruthLabel       = pset.get< std::string      >("MCTruthLabel");
  }
      
  /*********************************************************************************************/
  ////////////////////////////////////////////////////////////////////////
  void MRE::beginRun(art::Run& run)
  {

    fG4Alg = new g4n::G4Alg(fG4AlgPSet);
    return;
  }

  void MRE::produce(art::Event& evt)
  {
        
    std::unique_ptr<std::vector<simb::MCTruth>     >            mctcol  (new std::vector<simb::MCTruth>    );
    std::unique_ptr<std::vector<sim::FLSHitList>   >            flshlcol(new std::vector<sim::FLSHitList>  );
    std::unique_ptr<std::vector<sim::Particle>     >            pcol    (new std::vector<sim::Particle>    );

    art::Handle< std::vector<simb::MCParticle> > muonDataCol;
    evt.getByLabel(fMrccInfoLabel, muonDataCol);

    static TDatabasePDG*  pdgt = TDatabasePDG::Instance();
    TParticlePDG* pdgpm = pdgt->GetParticle(13);
    assert(pdgpm);

    TParticlePDG* pdgpe = pdgt->GetParticle(11);
    assert(pdgpe);
    const double electronMass = pdgpe->Mass();

    // Set the trackID offset for the electron. 
    // If file is MC, electron trackID's will be offset by
    // the highest pre-existing trackID in the event
    // If data is real, trackID offset is just -1000.

    int trackidOffset = -1000;
    art::ServiceHandle<cheat::BackTracker>  bt;
    if( bt->HaveTruthInfo() ){
      art::Handle< std::vector<sim::Particle> > pListHandle;
      evt.getByLabel(fMCTruthLabel, pListHandle);
      if( pListHandle.failedToGet() ) 
	trackidOffset = -1000;
      else{
	for( int p = 0; p < (int) pListHandle->size(); p++){
	  art::Ptr<sim::Particle> part(pListHandle, p);
	  int partId = part->TrackId();
	  if( partId > trackidOffset)
	    trackidOffset = partId;
	}
      }
    }

    // Loop over all the removed muons
    for( int mData = 0; mData < (int)muonDataCol->size(); mData++){
      
      art::Ptr<simb::MCParticle> muonData(muonDataCol, mData);

      simb::MCTruth truth;
      truth.SetOrigin(simb::kSingleParticle);

      double muonE       = muonData->E(0);
      double electronE   = muonE;
      double electronMom = TMath::Sqrt( (electronE*electronE) - (electronMass*electronMass) ); 
      TVector3 muonDir   = muonData->Momentum(0).Vect().Unit();
      
      // A nearly vertical track usually has unreasonable reconstructed length
      // and energy. Do not introduce a Mr.E for it. 
      if(muonDir.Z() < 1e-6)
	continue;
      TLorentzVector electron4Momentum( electronMom*muonDir.X(), electronMom*muonDir.Y(), 
					electronMom*muonDir.Z(), electronE);
      TLorentzVector origin(muonData->Vx(0), muonData->Vy(0), muonData->Vz(0), muonData->T(0));
 
      simb::MCParticle part(trackidOffset + 1, 11, "primary");
      part.AddTrajectoryPoint(origin, electron4Momentum);
      truth.Add(part);
      mctcol->push_back(truth);

      fG4Alg->RunGeant(&truth, *flshlcol, *pcol, trackidOffset);
	
      for( int p = 0; p < (int) pcol->size(); p++){
	int partId = (*pcol)[p].TrackId();
	if( partId > trackidOffset)
	  trackidOffset = partId;
      }	
    }// end loop over muon data

    
    evt.put(std::move(mctcol));
    evt.put(std::move(flshlcol));
    evt.put(std::move(pcol));
    return;
 
  }// end of producer
  

  DEFINE_ART_MODULE(MRE);


}// end of namespace murem
////////////////////////////////////////////////////////////////////////






