///
/// \brief 
/// \author Xuebing Bu - xbbu@fnal.gov
///
/// select rock muons in ND to study DQ and calibration
///

#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"

// ROOT includes
#include "TVector3.h"
#include "TH1F.h"
#include "TMath.h"
#include "TTree.h"
#include "TProfile2D.h"
#include "time.h"
#include "TString.h"

#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "RecoBase/Prong.h"

#include "Utilities/AssociationUtil.h"
#include "Utilities/func/MathUtil.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Persistency/Common/Assns.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindMany.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cmath>
#include <vector>
#include <string>
#include <vector>
#include <sstream>

class TVector3;
class TH1F;
class TTree;

namespace rb{class Cluster; class Track;}

namespace rockmuon {
  class NDRockMuon : public art::EDAnalyzer {
  public:
    explicit NDRockMuon(fhicl::ParameterSet const& pset);
    virtual ~NDRockMuon();
    
    void beginJob();
    void analyze(const art::Event& evt);
    void reconfigure(const fhicl::ParameterSet& p);
    void endJob(); 

    int Nevts_percell[214][96];

  private:

    //folders to get objects from
    std::string fSliceLabel;     ///label for slices
    std::string fKTrackLabel;    ///label for kalman tracks
    std::string fCellHitLabel;   ///label for cell hits
    std::string fVertexLabel;    ///label for vertex
 
    //tunable parameters
    int fMinHit; //slices must have at least this many hits
    float fMinTrackLength; //tracks must have at least this length in cm
    float fMinTrackCosTheta; //tracks must have at least this cosTheta
    float fMaxTrackStartZ; //tracks must start within this Z in cm
    float fMaxVertexX; //verties must be within this X
    float fMaxVertexY; //verties must be within this Y
    
    //define histograms
    TH1F *vertexX;
    TH1F *vertexY;
    TH1F *vertexZ;

    TH1F *trackStartX;
    TH1F *trackStartY;
    TH1F *trackStartZ;
    TH1F *trackLength;

    TH1F *Ncells_track[6]; //# of track associated cells in all(0), 3DBs(1), DB1(2), DB2(3),DB3(4), and muon-catcher(5);    
    TH1F *Ecells_track[6]; //sum energy of track associated cells

    TH1F *PECorr_track[6];//PECorr distributions
    TH1F *PECorr_track_MCH;//PECorr distributions for MC H modules
    TH1F *PECorr_track_MCV;//PECorr distributions for MC V modules
    
    TH1F *Ecell_track[6];//Ecell distributions
    TH1F *Ecell_track_MCH;//Ecell distributions for MC H modules
    TH1F *Ecell_track_MCV;//Ecell distributions for MC V modules

    TH1F *planeEnergy[214];//energy per plane
    TH1F *moduleADC[214][3];//ADC for each module
    TH1F *modulePECorr[214][3];//PECorr for each module
    TH1F *modulePE[214][3];//PE for each module
    TH1F *moduleEnergy[214][3];//energy for each module
  };
}
 
namespace rockmuon{
  
  //......................................................................
  NDRockMuon::NDRockMuon(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
  {
    mf::LogInfo("NDRockMuon")<<__PRETTY_FUNCTION__<<"\n";
    reconfigure(pset);
  }
  
  //......................................................................
  NDRockMuon::~NDRockMuon()
  {
  }
  
  //......................................................................
  void NDRockMuon::reconfigure(const fhicl::ParameterSet& pset)
  {
    #define SET(T,N) { f##N = pset.get<T>(#N); }
    SET(std::string, SliceLabel);
    SET(std::string, CellHitLabel);
    SET(std::string, KTrackLabel);
    SET(std::string, VertexLabel);

    SET(int, MinHit);
    SET(float, MinTrackLength);
    SET(float, MinTrackCosTheta);
    SET(float, MaxTrackStartZ);
    SET(float, MaxVertexX);
    SET(float, MaxVertexY);
    #undef SET
  }

  //......................................................................
  void NDRockMuon::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;    

    for( int ip=0; ip<214; ++ip ){
       for( int ic=0; ic<96; ++ic ){
         Nevts_percell[ip][ic]=0;
      }
    }

    vertexX=tfs->make<TH1F>("vertexX","",40,-200,200);
    vertexY=tfs->make<TH1F>("vertexY","",40,-200,200);
    vertexZ=tfs->make<TH1F>("vertexZ","",100,0,100);
    vertexX->Sumw2();
    vertexY->Sumw2();
    vertexZ->Sumw2();
    
    trackStartX=tfs->make<TH1F>("trackStartX","",40,-200,200);
    trackStartY=tfs->make<TH1F>("trackStartY","",40,-200,200);
    trackStartZ=tfs->make<TH1F>("trackStartZ","",100,0,100);
    trackStartX->Sumw2();
    trackStartY->Sumw2();
    trackStartZ->Sumw2();

    trackLength=tfs->make<TH1F>("trackLength","",160,0,1600);
    trackLength->Sumw2();
    
    TString detname[6]={"all","DB123","DB1","DB2","DB3","MC"};
    for( int i=0; i<6; ++i ){
      TString name1="Ncells_track_"+detname[i];
      Ncells_track[i]=tfs->make<TH1F>(name1,"",300,0,300);
      Ncells_track[i]->Sumw2();

      TString name2="Ecells_track_"+detname[i];
      Ecells_track[i]=tfs->make<TH1F>(name2,"",300,0,3);
      Ecells_track[i]->Sumw2();

      TString name3="PECorr_track_"+detname[i];
      PECorr_track[i]=tfs->make<TH1F>(name3,"",5000,0,5000);
      PECorr_track[i]->Sumw2();

      TString name4="Ecell_track_"+detname[i];
      Ecell_track[i]=tfs->make<TH1F>(name4,"",1000,0,0.1);
      Ecell_track[i]->Sumw2();
    }//loop det parts

    PECorr_track_MCH=tfs->make<TH1F>("PECorr_track_MCH","",5000,0,5000);
    PECorr_track_MCH->Sumw2();
    PECorr_track_MCV=tfs->make<TH1F>("PECorr_track_MCV","",5000,0,5000);
    PECorr_track_MCV->Sumw2();

    Ecell_track_MCH=tfs->make<TH1F>("Ecell_track_MCH","",1000,0,0.1);
    Ecell_track_MCH->Sumw2();
    Ecell_track_MCV=tfs->make<TH1F>("Ecell_track_MCV","",1000,0,0.1);
    Ecell_track_MCV->Sumw2();

    for( int i=0; i<214; ++i){
     TString perplaneEnergy=Form("Ecells_plane%i",i);
     planeEnergy[i]=tfs->make<TH1F>(perplaneEnergy,"",1000,0,0.1);
     planeEnergy[i]->Sumw2();
    }

    for( int ip=0; ip<214; ++ip ){
     std::stringstream planenumber;
     planenumber << ip;
     std::string str_planenumber=planenumber.str();

     for( int ic=0; ic<3; ++ic ){
       std::stringstream modulenumber;
       modulenumber << ic;
       std::string str_modulenumber=modulenumber.str();

       std::string namePECorr="modulePECorr_plane"+str_planenumber+"_module"+str_modulenumber;
       TString str_namePECorr=namePECorr;
       modulePECorr[ip][ic]=tfs->make<TH1F>(str_namePECorr,"",5000,0,5000);
       modulePECorr[ip][ic]->Sumw2();

       std::string namePE="modulePE_plane"+str_planenumber+"_module"+str_modulenumber;
       TString str_namePE=namePE;
       modulePE[ip][ic]=tfs->make<TH1F>(str_namePE,"",5000,0,5000);
       modulePE[ip][ic]->Sumw2();

       std::string nameADC="moduleADC_plane"+str_planenumber+"_module"+str_modulenumber;
       TString str_nameADC=nameADC;
       moduleADC[ip][ic]=tfs->make<TH1F>(str_nameADC,"",5000,0,5000);
       moduleADC[ip][ic]->Sumw2();

       std::string nameEnergy="moduleEnergy_plane"+str_planenumber+"_module"+str_modulenumber;
       TString str_nameEnergy=nameEnergy;
       moduleEnergy[ip][ic]=tfs->make<TH1F>(str_nameEnergy,"",1000,0,0.1);
       moduleEnergy[ip][ic]->Sumw2();
      }//loop module index
    }//loop plane index

  }

  //......................................................................
  void NDRockMuon::analyze(const art::Event& evt) 
  {
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<geo::Geometry> geom;
    
    //get all hits info
    art::Handle<std::vector<rb::CellHit> > chits;
    evt.getByLabel(fCellHitLabel, chits);

    //get reco slicer info  
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSliceLabel, slices);
    
    //get kalman track
    art::FindManyP<rb::Track> fmKTrack(slices, evt, fKTrackLabel);

    //get reco vertex
    art::FindManyP<rb::Vertex> fmVertex(slices, evt, fVertexLabel);
   
    for(unsigned int sliceIdx = 0; sliceIdx < slices->size(); ++sliceIdx){

      const rb::Cluster& slice = (*slices)[sliceIdx];
      if(slice.IsNoise()) continue; //veto noise slice
      double nCells = slice.NCell();
      if( nCells<fMinHit ) continue; //minimum cells associated to slice

      //select reco vertex
      if (!(fmVertex.isValid())) continue;
      std::vector<art::Ptr<rb::Vertex>> vert = fmVertex.at(sliceIdx);
      if (vert.size() != 1) continue;
      if( fabs(vert[0]->GetX())>fMaxVertexX ) continue;
      if( fabs(vert[0]->GetY())>fMaxVertexY ) continue;

      //select the kalman track
      std::vector<art::Ptr<rb::Track>> Ktracks = fmKTrack.at(sliceIdx);
      if( !(fmKTrack.isValid()) ) continue;
      
      //require there is only one track in the slice
      if( Ktracks.size() != 1 ) continue;

      double Ltrack = Ktracks[0]->TotalLength();
      if( Ltrack<fMinTrackLength ) continue; //track go through the full active detector

      TVector3 dir = Ktracks[0]->Dir();
      double track_theta=dir.Theta();
      if( cos(track_theta)<fMinTrackCosTheta ) continue;//track point forward      
      
      TVector3 start = Ktracks[0]->Start();
      double track_startz = start.Z();
      if( track_startz>fMaxTrackStartZ ) continue;//track starts within first 4 planes
	
      vertexX->Fill(vert[0]->GetX());
      vertexY->Fill(vert[0]->GetY());
      vertexZ->Fill(vert[0]->GetZ());
      trackLength->Fill(Ltrack);
      trackStartX->Fill(start.X());
      trackStartY->Fill(start.Y());
      trackStartZ->Fill(start.Z());

      //save track associated cell information to vectors
      std::vector<float> trkcellE;
      std::vector<int> trkcellPlane;
      std::vector<int> trkcellCell;
      std::vector<int> trkcellADC;
      std::vector<float> trkcellPECorr;
      std::vector<float> trkcellPE;

      for( unsigned int icell=0; icell<Ktracks[0]->NCell(); ++ icell){
	const art::Ptr<rb::CellHit>& chit = Ktracks[0]->Cell(icell);
        const rb::RecoHit rhit(cal->MakeRecoHit(*chit, Ktracks[0]->W(chit.get()))); 

	if( !rhit.IsCalibrated() ) continue;

          trkcellE.push_back(rhit.GeV());
          trkcellADC.push_back(chit->ADC());
          trkcellPlane.push_back(chit->Plane());
          trkcellCell.push_back(chit->Cell());
          trkcellPECorr.push_back(rhit.PECorr());
          trkcellPE.push_back(chit->PE());
      }

      int ncells_track[6]={0};
      double ecells_track[6]={0};
      for( unsigned int ic=0; ic<trkcellE.size(); ++ic ){
         ncells_track[0]++;
         ecells_track[0] += trkcellE[ic];

         planeEnergy[trkcellPlane[ic]]->Fill(trkcellE[ic]);
         Nevts_percell[trkcellPlane[ic]][trkcellCell[ic]]++;

          if(trkcellCell[ic]<32 ){
               modulePE[trkcellPlane[ic]][0]->Fill(trkcellPE[ic]);
               modulePECorr[trkcellPlane[ic]][0]->Fill(trkcellPECorr[ic]);
               moduleADC[trkcellPlane[ic]][0]->Fill(trkcellADC[ic]);
               moduleEnergy[trkcellPlane[ic]][0]->Fill(trkcellE[ic]);
             }//H bottom or V east
             else if( trkcellCell[ic]<64 ){
               modulePE[trkcellPlane[ic]][1]->Fill(trkcellPE[ic]);
               modulePECorr[trkcellPlane[ic]][1]->Fill(trkcellPECorr[ic]);
               moduleADC[trkcellPlane[ic]][1]->Fill(trkcellADC[ic]);
               moduleEnergy[trkcellPlane[ic]][1]->Fill(trkcellE[ic]);
             }//middle
             else{
               modulePE[trkcellPlane[ic]][2]->Fill(trkcellPE[ic]);
               modulePECorr[trkcellPlane[ic]][2]->Fill(trkcellPECorr[ic]);
               moduleADC[trkcellPlane[ic]][2]->Fill(trkcellADC[ic]);
               moduleEnergy[trkcellPlane[ic]][2]->Fill(trkcellE[ic]);
             }//H top or V west

         PECorr_track[0]->Fill(trkcellPECorr[ic]);
         Ecell_track[0]->Fill(trkcellE[ic]);

         if( trkcellPlane[ic]<192 ){
             ncells_track[1]++;
             ecells_track[1] += trkcellE[ic];
             PECorr_track[1]->Fill(trkcellPECorr[ic]);
             Ecell_track[1]->Fill(trkcellE[ic]);

             if( trkcellPlane[ic]<64 ){
                ncells_track[2]++;
                ecells_track[2] += trkcellE[ic];
                PECorr_track[2]->Fill(trkcellPECorr[ic]);
                Ecell_track[2]->Fill(trkcellE[ic]);
             }//DB1
             else if( trkcellPlane[ic]<128 ){
                ncells_track[3]++;
                ecells_track[3] += trkcellE[ic];
                PECorr_track[3]->Fill(trkcellPECorr[ic]);
                Ecell_track[3]->Fill(trkcellE[ic]);
             }//DB2
             else {
                ncells_track[4]++;
                ecells_track[4] += trkcellE[ic];
                PECorr_track[4]->Fill(trkcellPECorr[ic]);
                Ecell_track[4]->Fill(trkcellE[ic]);
             }//DB3
         }//DB1,2,3
         else{
            ncells_track[5]++;
            ecells_track[5] += trkcellE[ic];
            PECorr_track[5]->Fill(trkcellPECorr[ic]);
            Ecell_track[5]->Fill(trkcellE[ic]);
            
            if( trkcellPlane[ic]%2==0 ){
              PECorr_track_MCH->Fill(trkcellPECorr[ic]);
              Ecell_track_MCH->Fill(trkcellE[ic]);
            }
            else{
              PECorr_track_MCV->Fill(trkcellPECorr[ic]);
              Ecell_track_MCV->Fill(trkcellE[ic]);
            }

         }//muon catcher
      }//end loop of track associated cells

     for( int id=0; id<6; ++id ){
       Ncells_track[id]->Fill(ncells_track[id]);
       Ecells_track[id]->Fill(ecells_track[id]);
     }
                 
    }//end loop of slice
    
    return;
  }//end analyze
  
 void NDRockMuon::endJob()
  {
    art::ServiceHandle<art::TFileService> tfs; 
    TH1F *Nevts_incell_perplane[214];//# of events in each cell per each plane
    for( int i=0; i<214; ++i ){ 
     TString perplaneEvts=Form("Nevts_plane%i",i);
     Nevts_incell_perplane[i]=tfs->make<TH1F>(perplaneEvts,"",96,-0.5,95.5);
    }

    for( int ip=0; ip<214; ++ip ){
      for( int ic=0; ic<96; ++ic ){
       Nevts_incell_perplane[ip]->SetBinContent(ic+1,Nevts_percell[ip][ic]);
    }//loop cell number
   }//loop plane number

  }
  
}//end namespace

namespace rockmuon{

  DEFINE_ART_MODULE(NDRockMuon);

}

