#!/bin/env python

import os, sys, stat, pwd
import argparse
import datetime
import samweb_client
import string
import tokenize
import cStringIO

user=os.getenv("USER")
os.environ["GRID_USER"]=user

sam_user=user
sam_station=os.getenv("SAM_STATION")

#this is what we default to if not using CVMFS
# later we change this if --cvmfs option is supplied
cvmfs_distro_base="/grid/fermiapp/nova/novaart"
build_location_arguments=""  # Becomes more specific with --cvmfs 
setup_location=""

jobsub_opts=""

run_nova_sam_cmd="runNovaSAM.py"
run_nova_sam_opts=""
export_to_run_nova_sam=[]

art_sam_wrap_cmd="$NOVAGRIDUTILS_DIR/bin/art_sam_wrap.sh"
art_sam_wrap_opts=""
export_to_art_sam_wrap=[]
export_to_art_sam_wrap.append("SAM_PROJECT_NAME")
export_to_art_sam_wrap.append("SAM_STATION")
export_to_art_sam_wrap.append("IFDH_BASE_URI")
export_to_art_sam_wrap.append("IFDH_DEBUG")
export_to_art_sam_wrap.append("EXPERIMENT")
export_to_art_sam_wrap.append("GRID_USER")

#there must be a better name for this variable
usage_models=["DEDICATED"]

input_files=[]

source_scripts=[]
pre_scripts=[]
post_scripts=[]

def fail(msg):
    sys.stderr.write("\033[91mERROR in %s\033[0m " %prog)
    sys.stderr.write(msg)
    sys.stderr.write("\n")
    sys.exit(-1)
    
def warn(msg):
    sys.stderr.write("\033[93mWARNING in %s\033[0m " %prog)
    sys.stderr.write(msg)
    sys.stderr.write("\n")

    
def check_ticket():
    #is there a nice python bindy way to do this?
    retval=os.system("klist -s")
    if 0==retval: return

    fail("You don't have a kerberos ticket")


def check_dir(dname):
    if not os.path.isdir(dname):
        fail("Invalid directory %s" %dname)


def is_group_writable(fname):
    stat_info = os.stat(fname)
    file_stat_mode  = stat_info.st_mode
    return file_stat_mode & stat.S_IWGRP 


def check_is_group_writable(fname):
    if not is_group_writable(fname):
        fail("%s is not group writable, but should be" % fname)

def get_file_owner(fname):
    stat_info = os.stat(fname)
    file_stat_uid = stat_info.st_uid 
    return pwd.getpwuid(file_stat_uid)[0]


def check_is_production_writable(fname):
    # Production writable means either owned by "novapro" or group writable
    file_owner = get_file_owner(fname)
    if not (file_owner == "novapro" or is_group_writable(fname)):
        fail("%s is neither group writable nor owned by novapro, but should be" % fname)   
                    

def check_env(vname):
    value=os.getenv(vname)
    if None == value or "" == value:
        fail("Environment variable %s not defined" %vname)

#is this check really required?  Or more to the point, I think the check we want to do is really
# to ensure that user specified tag exists.  This is a clunky way to do the check, but I can't
# think of a better way at the moment
def check_tag(tag):
    setup_tag = os.getenv("SRT_BASE_RELEASE")
    if  None == setup_tag or "" == tag:
        fail("You have not set up a novasoft tag")

    elif setup_tag != tag:
        fail("You set up novasoft tag %s, but want job to run as %s"  %(setup_tag,tag))
                
def check_fcl(tag,fcl):
    if "/" in fcl[:1] and os.path.isfile(fcl):
        return # hey I'm ok!
    # Loop over dirs in FHICL_FILE_PATH
    fclPaths = os.environ["FHICL_FILE_PATH"].split(":")
    for path in fclPaths:
        fullPath = os.path.join(path, fcl)
        if os.path.isfile(fullPath):
            return # hey you're ok!

    fail("fcl file %s does not exist" %fcl)


def remove_comments(src):
    """
    This reads tokens using tokenize.generate_tokens and recombines them
    using tokenize.untokenize, and skipping comment/docstring tokens in between
    """
    f = cStringIO.StringIO(src)
    class SkipException(Exception): pass
    processed_tokens = []
    last_token = None
    # go thru all the tokens and try to skip comments and docstrings
    for tok in tokenize.generate_tokens(f.readline):
        t_type, t_string, t_srow_scol, t_erow_ecol, t_line = tok

        try:
            if t_type == tokenize.COMMENT:
                raise SkipException()

            elif t_type == tokenize.STRING:

                if last_token is None or last_token[0] in [tokenize.INDENT]:
                    pass

        except SkipException:
            pass
        else:
            processed_tokens.append(tok)

        last_token = tok

    return tokenize.untokenize(processed_tokens)

        
if __name__=='__main__':

    prog=os.path.basename(sys.argv[0])

    if "-f" in sys.argv or "--file" in sys.argv:
        ### Allow args to be passed in as a plain text file.  
        ### We make a preliminary parser get these arguments out for two reasons: 
        ###    1)  Maintain standard -h, --help functionality 
        ###    2)  Avoid necessity required arguments in initial parsing, 
        ###        allow them to be missing, but find them in the file. 
        preliminary_parser = argparse.ArgumentParser(prog=prog, description='Submit nova art job')

        preliminary_parser.add_argument('-f', '--file',
                        help="""Text file containing any arguments to this utility.  Multiple allowed.  
                        Arguments should look just like they would on the command line, 
                        but the parsing of this file is whitespace insenstive.  
                        Commented lines will be identified with the # character and removed. """, type=str, action='append')
        pre_args, unknown = preliminary_parser.parse_known_args()

        if pre_args.file:
            for filepath in pre_args.file:
                if not os.path.isfile(filepath):
                    fail(filepath + " is not a file.")
                else:
                    text = open(filepath, 'r').read()
                    text = remove_comments(text) # Strip out commented lines 
                    for line in text.splitlines():
                        sys.argv += line.split()


    parser = argparse.ArgumentParser(prog=prog, description='Submit nova art job', add_help=False)
    required = parser.add_argument_group

    ###required options
    required_args = parser.add_argument_group("Required arguments", "These arguments must be supplied.")

    required_args.add_argument('--jobname',
                        required=True,
                        help='Job name',
                        type=str)

    required_args.add_argument('--defname',
                        required=True,
                        help='SAM dataset definition to run over',
                        type=str)
    
    required_args.add_argument('--config', '-c',
                        required=True,
                        help='FHiCL file to use as configuration for nova executable',
                        type=str)

    required_args.add_argument("--tag",
                        required=True,
                        help="Tag of novasoft to use",
                        type=str)
    
    required_args.add_argument("--dest",
                        required=True,
                        help="Destination for output files",
                        type=str)


    ###debugging
    debugging_args = parser.add_argument_group("Debugging options", "These optional arguments can help debug your submission.")
    
    debugging_args.add_argument('--print_jobsub',
                        help='Print jobsub command',
                        action='store_true',default=False)

    debugging_args.add_argument('--test',
                        help='Do not actually do anything, just run tests and print jobsub cmd',
                        action='store_true',default=False)
    
    debugging_args.add_argument('--gdb',
                        help='Run nova executable under gdb, print full stack trace, then quit gdb.',
                        action='store_true',default=False)
    
    job_control_args = parser.add_argument_group("Job control options", "These optional arguments help control where and how your jobs land.") 
    ###number of jobs
    job_control_args.add_argument('--njobs',
                        help='Number of jobs to submit',
                        type=int, default = 0)


    job_control_args.add_argument('--files_per_job',
                        help='Number of files per job - if zero, calculate from number of jobs',
                        type=int, default = 0)

    job_control_args.add_argument('--nevts',
                        help='Number of events per file to process',
                        type=int, default = 0)


    job_control_args.add_argument('--no_multifile',
                        help='Do not use art_sam_wrap.sh multifile mode, which is on by default',
                        action='store_true')

    ###general job control
    job_control_args.add_argument('--opportunistic',
                        help='Run opportunistically on the fermigrid',
                        action='store_true',default=False)
    
    job_control_args.add_argument('--offsite',
                        help='Allow to run on offsite resources as well.  Implies --opportunistic and --cvmfs.',
                        action='store_true',default=False)
    
    job_control_args.add_argument('--offsite_only',
                        help='Allow to run solely on offsite resources.  Implies --cvmfs.',
                        action='store_true',default=False)
    
    job_control_args.add_argument('--site',
                        help='Specify allowed offsite locations.  Omit to allow running at any offsite location',
                        type=str,action='append')

    job_control_args.add_argument('--os',
                        help='Specify OS version of worker node',
                        type=str,action='append')

    job_control_args.add_argument('--disk',
                        help='Local disk space requirement for worker node in MB.',
                        type=int)

    job_control_args.add_argument('--memory',
                        help='Local memory requirement for worker node in MB.',
                        type=int, default=4000)

    job_control_args.add_argument('--group', '-G',
                        help="Specify batch group GROUP -- mainly used to set job priority.  At present, only supportable value is nova",
                        type=str, default="nova")

    job_control_args.add_argument("--role",
                        help="Specify role to run on the grid.  Can be Analysis (default) or Production",
                        type=str,default="Analysis")
    
    ###software control
    novasoft_args = parser.add_argument_group("NOvA software options", "These options control the novasoft setup.")
    novasoft_args.add_argument('--maxopt',
                        help='Run in maxopt mode',
                        action='store_true',default=False)

    novasoft_args.add_argument("--testrel",
                        help="Use testrel at location TESTREL",
                        type=str)
    
    novasoft_args.add_argument('--cvmfs',
                        help='Pull software from CVMFS.',
                        action='store_true')

    novasoft_args.add_argument('--novasoftups',
                        help='Use the ups build of novasoft, must be used with source to setup.',
                        action='store_true')


    ###output control
    output_args = parser.add_argument_group("Output file options", "Note that you must specify either --copyOut or --copyOutScript.")

    
    output_args.add_argument('--copyOutScript',
                        help='Use script COPYOUTSCRIPT to copy back your output',
                        type=str)

    output_args.add_argument('--copyOut',
                        help='Use the built in copy out mechanism.  If used, you must specify --outTier, --cafTier or --histTier', 
                        action='store_true')

    output_args.add_argument('--outTier',
                        help='Data tier of the output file, multiple allowed, formatted as <name_in_fcl_outputs>:<data_tier>',
                        type=str, action='append')
    
    output_args.add_argument('--cafTier',
                        help='Module label for CAF output, multiple allowed.  Format as <cafmaker_module_label>:<data_tier>',
                        type=str, action='append')

    output_args.add_argument('--histTier',
                        help='File identifier string for TFileService output, only one allowed.  Supply as --histTier <id> for output_name.<id>.root, where output_name is assembled based on the input file.',
                        type=str)

    output_args.add_argument('--specialName',
                           help='To specify nova.special name',
                           type=str)

    output_args.add_argument('--hashDirs',
                        help='Use hash directory structure in destination directory.',
                        action='store_true')

    output_args.add_argument('--runDirs',
                        help='Use run directory structure in destination directory, 000XYZ/XYZUW for run number XYZUW.',
                        action='store_true')

    output_args.add_argument('--noCleanup',
                        help='Pass --noCleanup argument to runNovaSAM.py.  Necessary when using a postscript for copyout.',
                        action='store_true')

    output_args.add_argument('--jsonMetadata', help='Create JSON files with metadata corresponding to each output file, and copy them to the same destinations', action='store_true')

    output_args.add_argument('--declareFiles',
                        help='Declare files with metadata on worker node',
                        action='store_true')

    output_args.add_argument('--production',
                        help='Submit production style jobs.  Implies \"--role=Production --hashDirs --jsonMetadata\", and checks that other settings needed for production are specified',
                        action='store_true')

    ###environment  configuration
    environment_args = parser.add_argument_group("Environment options", "These optional arguments allow control of the grid running environment.")

    environment_args.add_argument('--export',
                        help='Export variable EXPORT to art_sam_wrap.sh',
                        type=str, action='append')

    environment_args.add_argument('--source',
                                  help='Source script SOURCE',
                                  type=str, action='append')

    environment_args.add_argument('--prescript',
                                  help='Execute script PRESCRIPT before executing runNovaSAM.py',
                                  type=str, action='append')

    environment_args.add_argument('--postscript',
                                  help='Execute script POSTSCRIPT after executing runNovaSAM.py',
                                  type=str, action='append')
    
    environment_args.add_argument('--inputfile',
                                  help='Copy this extra input file into job area before running executable',
                                  type=str, action='append')
    
    ###support options 
    support_args = parser.add_argument_group("Support options", "These optional arguments using this sybmission utility easier.")
    
    support_args.add_argument("-h", "--help", action="help", help="Show this help message and exit")

    support_args.add_argument('-f', '--file',
                        help="""Text file containing any arguments to this utility.  Multiple allowed.  
                        Arguments should look just like they would on the command line, 
                        but the parsing of this file is whitespace insenstive.  
                        Comments will be identified with the # character and removed. """, type=str, action='append')


    
    args = parser.parse_args()

    jobname=args.jobname
    defname=args.defname
    print_jobsub=args.print_jobsub
    test=args.test
    if test :
        print_jobsub=True
        print ""
        warn("--test was specified, so all we do is run checks and print jobsub cmd.")


    check_ticket()

    check_env("SETUP_IFDH_ART")
    check_env("SETUP_SAM_WEB_CLIENT")
    check_env("SETUP_JOBSUB_CLIENT")
    check_env("SAM_STATION")
    tag=args.tag
    check_tag(tag)    

    if args.testrel:
        check_dir(args.testrel)
        
    fcl=args.config
    check_fcl(tag,fcl)     
    

    role=args.role


    is_production_arg   = args.production 

    # Can specify production through special arg or --role Production 
    # Allow for some harmless whitespace in role, just in case
    is_production       = is_production_arg or "Production" in role 

    dest=args.dest
    check_dir(dest)

    if not is_production:
        # Non-production jobs run as user "novaana", so output dir must 
        # be have group write access 
        check_is_group_writable(dest)
    else:
        # Production writable means either owned by "novapro" or group writable
        check_is_production_writable(dest)

    if "/nova/app" == dest[0:9]:
        fail("Destination directory is in /nova/app.  This is not allowed")
        
    export_to_run_nova_sam.append("DEST=%s"%dest)
    maxopt=args.maxopt
    maxopt_opt=""
    if maxopt:
        maxopt_opt += "-b:maxopt"

    njobs=args.njobs
    files_per_job = args.files_per_job

    print "Definition name: %s" % defname 
    
    if files_per_job > 0 and njobs > 0 :
        ## both njobs and files per job are specified.  Just pass
        ## the settings through to jobsub and art_sam_wrap
        jobsub_opts += ("    -N %d \\\n" %njobs)
        art_sam_wrap_opts += ("      --limit %d \\\n" % files_per_job)

    elif files_per_job > 0:
        ##files/job specified, but not njobs.  Calculate njobs
        ## on the fly

        ##get files in dataset
        samweb = samweb_client.SAMWebClient(experiment='nova')        
        num_project_files=samweb.countFiles(defname=defname)
        print "Definition file count %d" % num_project_files

        njobs=(num_project_files / files_per_job) +1
        jobsub_opts += ("    -N %d \\\n" %njobs)
        art_sam_wrap_opts += ("      --limit %d \\\n" % files_per_job)
            
    elif njobs > 0 :
        ##njobs specified, but not files/job.  Just set njobs
        ## and don't force limits on files per jobs
        jobsub_opts += ("    -N %d \\\n" %njobs)
        

    else :
        warn("Neither --njobs nor --files_per_job specified.  Did you really want to do this?")

    if args.opportunistic or args.offsite:
        usage_models.append("OPPORTUNISTIC")
    if args.offsite :
        usage_models.append("OFFSITE")
    if args.offsite_only:
        if args.offsite:
            fail("Both --offsite and --offsite_only specified, these arguments conflict")

        if args.opportunistic:
            fail("Both --opportunistic and --offsite_only specified, these arguments conflict")

        usage_models = ["OFFSITE"]

        
    if (args.offsite or args.offsite_only) and not args.cvmfs:
        warn("Running offsite, switched to --cvmfs mode")
        args.cvmfs=True

    if (args.offsite or args.offsite_only) and not args.os :
        fail("Running offsite, but OS version not specified!")
        
    resource_opt="    --resource-provides=usage_model=" + string.join(usage_models,",") + " \\\n"
    jobsub_opts += resource_opt

    allowed_sites=["Caltech","Fermigrid", "Fermigridosg1", "FZU", "Harvard", "Michigan",
                   "MIT", "MWT2", "Nebraska", "Omaha", "OSC", "SMU", "SU-OG", "UChicago", "UCSD", "TTU", "Wisconsin"]

    if args.site :
        site_opt="    --site="
        for isite in args.site:
            if isite not in allowed_sites:
                fail("Site %s not allowed" %isite)
                
            site_opt += isite + ","
        site_opt=site_opt[:-1]
        site_opt += " \\\n"
        jobsub_opts += site_opt

    if args.os :
        allowed_os=["SL5","SL6"]
        for ios in args.os:
            if ios not in allowed_os:
                fail("Invalid OS %s" %ios)
                
        os_opt="    --OS=" + string.join(args.os,",") + " \\\n"
        jobsub_opts += os_opt

    if args.disk:
        disk_opt="    --disk=%s \\\n" % (args.disk)
        jobsub_opts += disk_opt        

    if args.memory:
        mem_opt="    --memory=%s \\\n" % (args.memory)
        jobsub_opts += mem_opt        

    if args.cvmfs: #was fnalcvmfs
        cvmfs_distro_base="/cvmfs/nova.opensciencegrid.org"
        build_location_arguments = ":-e:%s/externals:-5:%s/novasoft/slf5/novasoft:-6:%s/novasoft/slf6/novasoft" % (cvmfs_distro_base, cvmfs_distro_base, cvmfs_distro_base)
        setup_location= "%s/novasoft/slf5/novasoft/setup/setup_nova.sh" % (cvmfs_distro_base)
    else:
        setup_location="%s/novasvn/setup/setup_nova.sh" % (cvmfs_distro_base)

    
    group=args.group
    #probably others are ok as well
    allowed_groups=['nova']
    #allowed_groups=['nova_high_prio', 'nova_medium_prio', 'nova_low_prio', 'nova']
    if group not in allowed_groups :

        #here we do not use fail function b/c of extended msg
        sys.stderr.write("\033[91mERROR in %s\033[0m only valid args for --group are: " % prog)
        for allowed_group in allowed_groups:
            sys.stderr.write(allowed_group + " ")
        sys.stderr.write("\n")
        sys.exit(-1)
    jobsub_opts += "    -G %s \\\n" % group

    
    nevts=args.nevts
    if nevts > 0:
        run_nova_sam_opts += "        -n %d \\\n" %nevts

    if args.gdb:
        run_nova_sam_opts += "        --gdb \\\n"
        
    hash_dirs=args.hashDirs
    if hash_dirs or is_production_arg:
        run_nova_sam_opts += "        --hashDirs \\\n"

    run_dirs=args.runDirs
    if run_dirs :
        run_nova_sam_opts += "        --runDirs \\\n"

    no_cleanup=args.noCleanup
    if no_cleanup :
        run_nova_sam_opts += "        --noCleanup \\\n"

    #json_metadata=args.jsonMetadata
    #if json_metadata or is_production_arg:
    #    run_nova_sam_opts += "        --jsonMetadata \\\n"

    copy_out = args.copyOut
    if copy_out :
        run_nova_sam_opts += "        --copyOut \\\n"
        
    
    if args.specialName :
        run_nova_sam_opts += "        --specialName " + args.specialName + " \\\n"

    out_tiers=args.outTier
    if None != out_tiers :
        for tier in out_tiers :
            run_nova_sam_opts += ("        --outTier " + tier + " \\\n")

    caf_tiers=args.cafTier
    if None != caf_tiers :
        for tier in caf_tiers :
            run_nova_sam_opts += ("        --cafTier " + tier + " \\\n")

    hist_tier=args.histTier
    if None != hist_tier :
        run_nova_sam_opts += ("        --histTier " + hist_tier + " \\\n")

    if copy_out and (None==hist_tier and None==caf_tiers and None==out_tiers):
        fail("You specified --copyOut but did not specify --outTier, --cafTier or --histTier")

    if not (copy_out or args.copyOutScript) :
        fail("Did not specify a method to copy back output (--copyOut or --copyOutScript)")
        
    if copy_out and args.copyOutScript :
        fail("The options --copyOut and --copyOutScript conflict")
            
    if is_production:
        if  (None == out_tiers) and (None == caf_tiers):
            fail("Running in production mode, but no tiers (art or caf) specified")
        role="Production"
        sam_user="novapro"
        
    if args.export:
        export_to_art_sam_wrap += args.export

    if args.source:
        source_scripts += args.source

    if args.prescript:
        pre_scripts += args.prescript

    if args.postscript:
        post_scripts += args.postscript

    if args.copyOutScript:
        post_scripts.append(args.copyOutScript)

    for script in source_scripts + pre_scripts + post_scripts :
        if ":" in script:
          script_path = script.split(":")[0]
        else:
            script_path = script
        
        if not os.path.isfile(script_path):
            fail("Script %s does not exist!" % script_path)

    if args.inputfile:
        input_files += args.inputfile

    for input_file in input_files:
        if not os.path.isfile(input_file):
            fail("Input file %s does not exist!" % input_file)
            
    ##start sam project
    timestamp=datetime.datetime.now().strftime("%Y%m%d_%H%M")
    project_name = user + "-" + jobname + "-" + timestamp

    #sam_station=os.getenv("SAM_STATION")

    if args.novasoftups:
        art_sam_wrap_cmd="$NOVASOFT_FQ_DIR/bin/art_sam_wrap.sh"

    if not test :
        start_proj_command ="samweb start-project "
        start_proj_command+=" --defname=%s" %defname
        start_proj_command+=" --group=nova" 
        start_proj_command+=" --station=%s" % sam_station

        start_proj_command+= " %s" %project_name
        start_proj_retval=os.system(start_proj_command)
        print "start proj returned %d" %  start_proj_retval
        if start_proj_retval != 0:
            fail("Couldn't start project")

        print "Station monitor: http://samweb.fnal.gov:8480/station_monitor/nova/stations/nova/projects/"  + project_name        
        os.putenv("SAM_PROJECT_NAME",project_name)
        job_cmd=os.path.expandvars("${CONDOR_EXEC}/%s.sh" % project_name)

        #create symlink so that jobs get a better name
        os.symlink(os.path.expandvars(art_sam_wrap_cmd),job_cmd)
        art_sam_wrap_cmd=job_cmd

    sys.stdout.flush()
    sys.stderr.flush()
    
    for export in export_to_art_sam_wrap:
        jobsub_opts += ("    -e " + export)
    jobsub_opts += " \\\n"

    jobsub_opts += "    --role=%s \\\n " %(role)
    
    if not args.no_multifile:
        art_sam_wrap_opts+="     --multifile \\\n"

    export_to_run_nova_sam.append("CVMFS_DISTRO_BASE='%s'" % cvmfs_distro_base)

    for export in export_to_run_nova_sam :
        art_sam_wrap_opts += ("      --export " + export + " \\\n")

    art_sam_wrap_opts += "      --config " + fcl +" \\\n"

    if not args.novasoftups : 
        art_sam_wrap_opts += "      --source %s:-r:%s:%s%s \\\n" %(setup_location, tag, maxopt_opt, build_location_arguments)
        if args.testrel :
            art_sam_wrap_opts += "      --source setup_testrel:%s \\\n" %(args.testrel)

    for source_script in source_scripts:        
        art_sam_wrap_opts+="      --source " + source_script + " \\\n"
        
    for pre_script in pre_scripts:        
        art_sam_wrap_opts+="      --prescript " + pre_script + " \\\n"
        
    for post_script in post_scripts:        
        art_sam_wrap_opts+="      --postscript " + post_script + " \\\n"
        
    for input_file in input_files:        
        art_sam_wrap_opts+="      --inputfile " + input_file + " \\\n"
        
    jobsub_cmd="jobsub_submit \\\n" + jobsub_opts
    jobsub_cmd += (     "    file://" + art_sam_wrap_cmd + " \\\n " + art_sam_wrap_opts)
    jobsub_cmd += ("      -X " + run_nova_sam_cmd + " \\\n" + run_nova_sam_opts)

    jobsub_cmd = jobsub_cmd[:-2]
    
    
    if print_jobsub:
        print jobsub_cmd
        sys.stdout.flush()
        sys.stderr.flush()
    
            
            
    if not test:
        os.system(jobsub_cmd)
