///////////////////////////////////////////////////////////////////////////
// \file    MergeCollections.cxx
// \brief
// \authors janow@fnal.gov
///////////////////////////////////////////////////////////////////////////  
//#include "NovaSimMixer/MergeCollections.h"

#include <math.h>
#include <string>

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"


#include "DAQChannelMap/DAQChannelMap.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/GTruth.h"


#include "Utilities/AssociationUtil.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"


///group hits according to time and space
namespace MergeCollections {
  
  class MergeCollections : public art::EDProducer {
  public:
    explicit MergeCollections(fhicl::ParameterSet const &pset);
    virtual ~MergeCollections();

    void produce(art::Event& evt);

  protected:
    // internal methods

    // configuration settings
    std::string  fFirstGenCollection;     /// Label on the main collection
    std::string  fFirstG4Collection;     /// Label on the main collection
    std::string  fSecondCollection;     /// Label on the main collection
    int fIsFirstCollectionCosmic; 


  };


  //----------------------------------------------------------------------
  MergeCollections::MergeCollections(fhicl::ParameterSet const &pset) :
    fFirstGenCollection         (pset.get< std::string >("FirstGenCollection")),
    fFirstG4Collection          (pset.get< std::string >("FirstG4Collection")),
    fSecondCollection           (pset.get< std::string >("SecondCollection")),
    fIsFirstCollectionCosmic    (pset.get< int         >("IsFirstCollectionCosmic"))

  {  produces< std::vector<simb::MCTruth>                   >();
     produces< std::vector<simb::MCFlux>                    >();
     produces< std::vector<sim::FLSHitList >                >();
     produces< std::vector<sim::Particle>                   >();
     produces< std::vector<simb::GTruth>                    >();
     produces< art::Assns<simb::MCTruth, simb::MCFlux>      >();
     produces< art::Assns<simb::MCTruth, simb::GTruth>      >();
     produces< art::Assns<sim::Particle, simb::MCTruth>     >();


   }
  
  //----------------------------------------------------------------------
  MergeCollections::~MergeCollections()
  {
  }
  
  //----------------------------------------------------------------------
  void MergeCollections::produce(art::Event& evt)
  {


   std::unique_ptr< std::vector<simb::MCTruth> > truthcol (new std::vector<simb::MCTruth>);
   std::unique_ptr< std::vector<simb::MCFlux>  > fluxcol  (new std::vector<simb::MCFlux >);
   std::unique_ptr< std::vector<simb::GTruth>  > gtruthcol(new std::vector<simb::GTruth>);
   std::unique_ptr< art::Assns<simb::MCTruth, simb::MCFlux> > assns (new art::Assns<simb::MCTruth, simb::MCFlux>);
   std::unique_ptr< art::Assns<simb::MCTruth, simb::GTruth> > gassns(new art::Assns<simb::MCTruth, simb::GTruth>);
   std::unique_ptr< art::Assns<sim::Particle, simb::MCTruth> > passns(new art::Assns<sim::Particle, simb::MCTruth>);


 // Define the FLSHitList and Particle vectors.
    std::unique_ptr<std::vector<sim::FLSHitList>   >    flshlcol(new std::vector<sim::FLSHitList>  );
    std::unique_ptr<std::vector<sim::Particle>     >    pcol    (new std::vector<sim::Particle>    );


    /// Current event number
//    fCurrentEvent = evt.id().event();
    
    // Pull the MC generator information out of the event
    art::Handle< std::vector<simb::MCTruth> > mclist;
    evt.getByLabel(fFirstGenCollection, mclist);
   
    art::Handle< std::vector<simb::MCFlux> > fllist;
    art::Handle< std::vector<simb::GTruth> > glist;

    if(!fIsFirstCollectionCosmic) {
      evt.getByLabel(fFirstGenCollection, fllist);
      evt.getByLabel(fFirstGenCollection, glist);

    }

    art::Handle< std::vector<sim::FLSHitList> > flslist;
    evt.getByLabel(fFirstG4Collection, flslist);

    
    art::Handle< std::vector<sim::Particle> > partlist;
    evt.getByLabel(fFirstG4Collection, partlist);
    

   int IdOffset=0;
        
    for(unsigned int i = 0; i < partlist->size(); ++i){
      
      sim::Particle part ((*partlist)[i]);
      IdOffset=part.TrackId();
      //  std::cout << " Primary Collection ID " << IdOffset << std::endl;
      pcol->push_back(part); 
      
    }
    
    
    /// Loop over neutrino interactions

    //  std::cout << " primary mc list size " << mclist->size() << std::endl;
    
    for(unsigned int i = 0; i < mclist->size(); ++i){
      
      simb::MCTruth truth ((*mclist)[i]);
      truthcol->push_back(truth);
      
      if(! fIsFirstCollectionCosmic){
	simb::MCFlux  flux  ((*fllist)[i]);
	fluxcol ->push_back(flux);
		
	simb::GTruth gtruth ((*glist)[i]);
	gtruthcol->push_back(gtruth);
	
	util::CreateAssn(*this, evt, *truthcol, *fluxcol,   *assns,  fluxcol  ->size()-1, fluxcol  ->size());
	util::CreateAssn(*this, evt, *truthcol, *gtruthcol, *gassns, gtruthcol->size()-1, gtruthcol->size());
	
      }
      
      //      IdOffset=truth.NParticles();
      
      //      std::cout<<"First collection:"<< flslist->size() << std::endl;
      
      if(i<flslist->size()){
	sim::FLSHitList mcfls ((*flslist)[i]);
	flshlcol->push_back(mcfls);
      }
    }
           
    ///Second collection 
    //    std::cout<<"Id Offset = "<<IdOffset<<std::endl;
    
    art::Handle< std::vector<simb::MCTruth> > mc2list;
    evt.getByLabel(fSecondCollection, mc2list);
    
    art::Handle< std::vector<simb::GTruth> > g2list;
    evt.getByLabel(fSecondCollection, g2list);

    art::Handle< std::vector<simb::MCFlux> > fl2list;
    evt.getByLabel(fSecondCollection, fl2list);

    art::Handle< std::vector<sim::FLSHitList> > fls2list;
    evt.getByLabel(fSecondCollection, fls2list);


    art::Handle< std::vector<sim::Particle> > part2list;
    evt.getByLabel(fSecondCollection, part2list);
    
    int LastPartId;
    int *Offset = new int[1000];
    int *OffsetParticles = new int[1000];
    int OffsetIndex;

    LastPartId=0;
    for(int i=0;i<10;i++){
      Offset[i]=0;
      OffsetParticles[i]=0;
    }
    Offset[0]=IdOffset;
    OffsetParticles[0]=partlist->size();
    OffsetIndex=1;
    
    for(unsigned int i = 0; i < part2list->size(); ++i){
      
      OffsetParticles[OffsetIndex]++;
      
      sim::Particle part ((*part2list)[i]);
      //  std::cout<<"partID="<<part.TrackId()<<std::endl;
      if(LastPartId > part.TrackId() ){
	Offset[OffsetIndex]= Offset[OffsetIndex-1] + LastPartId;
	OffsetIndex++ ;
	
	//	std::cout<<"osfet for i="<<i<<" is "<<Offset[OffsetIndex-1]<<std::endl;
	//	std::cin.get();
      }
      
      LastPartId=part.TrackId();
      
      int partmother =0;
      if (part.Mother()!=0)partmother = part.Mother() + Offset[OffsetIndex-1]; 

      sim::Particle gpart(part.TrackId()+ Offset[OffsetIndex-1],
			  part.PdgCode(),
			  part.Process(),
			  partmother,
			  part.Mass(),
			  part.StatusCode());
      //      std::cout <<part.TrackId()<< "\t" << gpart.TrackId()<<"\t"<< Offset[OffsetIndex-1] << " mother " << part.Mother() << " mother offset " << partmother <<  std::endl;//std::cin.get();
      
      gpart.SetGvtx(part.Gvx(), part.Gvy(),part.Gvz(),part.Gvt());
      gpart.SetPolarization(part.Polarization());
      gpart.SetWeight(part.Weight());
      gpart.SetRescatter(part.Rescatter());
      for(int d=0; d<part.NumberDaughters(); d++){
	gpart.AddDaughter(part.Daughter(d)+ Offset[OffsetIndex-1]);
      }
      for(unsigned int t=0; t<part.NumberTrajectoryPoints(); t++){
	//std::cout<<part.Vx(t)<<"\t"<<part.EndX()<<std::endl;
	gpart.AddTrajectoryPoint(part.Position(t),part.Momentum(t));
      }
      pcol->push_back(gpart);
      //  std::cout<< "mc particle " << part.TrackId() << " " << Offset[OffsetIndex-1] << std::endl;
    }
        
    std::cout<<"mc1 size="<<mclist->size()<<std::endl;
    std::cout<<"mc2 size="<<mc2list->size()<<std::endl;
    
    for(unsigned int i = 0; i < mc2list->size(); ++i){
      
      simb::MCTruth truth ((*mc2list)[i]);
      simb::MCFlux  flux  ((*fl2list)[i]);
      simb::GTruth gtruth ((*g2list) [i]);
      
      simb::MCTruth tmc(truth);
      
      
      /// Offset ids for FLSHits
      
      //      std::cout<<"size="<<fls2list->size()<< " mclist " << i << std::endl;
      if(i<fls2list->size()){   
	
	sim::FLSHitList mcfls ((*fls2list)[i]);
	sim::FLSHitList tmcfls;	
	sim::FLSHit fHit;
	
	const std::vector<sim::FLSHit>& hits = mcfls.fHits;
	for (unsigned int j=0; j<hits.size(); ++j)
	  {
	    fHit.Clear();
	    fHit=hits[j];
	    fHit.SetTrackId(hits[j].GetTrackID() + Offset[i]);
	    tmcfls.fHits.push_back(fHit);
	    
	    //    std::cout << " fls ID " << hits[j].GetTrackID() << " " << Offset[i] << " " << hits[j].GetTrackID()+Offset[i] << std::endl;
	  } 
	flshlcol->push_back(tmcfls);
	
      }
      /// End of the FLS offset
      
      truthcol->push_back(tmc);
      gtruthcol->push_back(gtruth);
      fluxcol ->push_back(flux);
      
      util::CreateAssn(*this, evt, *truthcol, *fluxcol, *assns,  fluxcol->size()-1, fluxcol->size());
      util::CreateAssn(*this, evt, *truthcol, *gtruthcol, *gassns,  gtruthcol->size()-1, gtruthcol->size());
      
      
      //      std::cout<<"end of i mcsize"<<std::endl;	
      
      
     }	  
      
      // Define the FLSHitList and Particle vectors.
      
      
      std::unique_ptr< std::vector<simb::MCTruth> >  temptruelist  (new std::vector<simb::MCTruth>);
      std::unique_ptr< std::vector<sim::Particle> >  temppartlist  (new std::vector<sim::Particle>);

      bool NewMCList = false;
      int MCListIndex = -1;
      for(unsigned int j = 0; j < pcol->size(); ++j){
         sim::Particle part ((*pcol)[j]);
	 if(part.Mother()!=0 )NewMCList=false;
         if(part.Mother()==0 && ! NewMCList){
	  MCListIndex++;
	  //	    std::cout << " pushing truthcol index " << MCListIndex << " onto temptrue " << std::endl;

          simb::MCTruth truth ((*truthcol)[MCListIndex]);
          temptruelist->push_back(truth);
	  NewMCList=true;
	 }
	 //	 std::cout << j << " " << temptruelist->size()-1 << std::endl;

	 temppartlist->push_back(part);
	 util::CreateAssn(*this, evt, *temppartlist, *temptruelist, *passns,  temptruelist->size()-1,temptruelist->size());


	}
         
      
      // put the collections in the event
      evt.put(std::move(temptruelist));
      evt.put(std::move(fluxcol));
      evt.put(std::move(gtruthcol));
      evt.put(std::move(assns));
      evt.put(std::move(gassns));
      evt.put(std::move(flshlcol));
      evt.put(std::move(temppartlist));
      evt.put(std::move(passns));
      // int dum;
      //   std::cin >> dum;

       delete [] Offset ;
       delete [] OffsetParticles;

  }
  
  

  DEFINE_ART_MODULE(MergeCollections);
} // end namespace merge
/////////////////////////////////////////////////////////////////////////
