///////////////////////////////////////////////////////////////////////////
// \file    MergeGenCollections.cxx
// \brief
// \authors janow@fnal.gov
///////////////////////////////////////////////////////////////////////////  
//#include "NovaSimMixer/MergeGenCollections.h"

#include <math.h>
#include <string>

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"

#include "DAQChannelMap/DAQChannelMap.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/GTruth.h"

#include "Utilities/AssociationUtil.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"

///group hits according to time and space
namespace MergeGenCollections {
  
  class MergeGenCollections : public art::EDProducer {
  public:
    explicit MergeGenCollections(fhicl::ParameterSet const &pset);
    virtual ~MergeGenCollections();

    void produce(art::Event& evt);

  protected:
    // internal methods

    // configuration settings
    std::string  fFirstGenCollection;     /// Label on the main collection
    std::string  fFirstG4Collection;     /// Label on the main collection
    std::string  fSecondCollection;     /// Label on the main collection
    int fIsFirstCollectionCosmic; 
    int fIsSecondCollectionCosmic; 

  };


  //----------------------------------------------------------------------
  MergeGenCollections::MergeGenCollections(fhicl::ParameterSet const &pset) :
    fFirstGenCollection         (pset.get< std::string >("FirstGenCollection")),
    fFirstG4Collection          (pset.get< std::string >("FirstG4Collection")),
    fSecondCollection           (pset.get< std::string >("SecondCollection")),
    fIsFirstCollectionCosmic    (pset.get< int         >("IsFirstCollectionCosmic")),
    fIsSecondCollectionCosmic    (pset.get< int         >("IsSecondCollectionCosmic"))

  {  produces< std::vector<simb::MCTruth>                   >();
     produces< std::vector<simb::MCFlux>                    >();
     produces< std::vector<simb::GTruth>                    >();
     produces< art::Assns<simb::MCTruth, simb::MCFlux>      >();
     produces< art::Assns<simb::MCTruth, simb::GTruth>      >(); 

   }
  
  //----------------------------------------------------------------------
  MergeGenCollections::~MergeGenCollections()
  {
  }
  
  //----------------------------------------------------------------------
  void MergeGenCollections::produce(art::Event& evt)
  {

   std::unique_ptr< std::vector<simb::MCTruth> > truthcol (new std::vector<simb::MCTruth>);
   std::unique_ptr< std::vector<simb::MCFlux>  > fluxcol  (new std::vector<simb::MCFlux >);
   std::unique_ptr< std::vector<simb::GTruth>  > gtruthcol(new std::vector<simb::GTruth>);
   std::unique_ptr< art::Assns<simb::MCTruth, simb::MCFlux> > assns (new art::Assns<simb::MCTruth, simb::MCFlux>);
   std::unique_ptr< art::Assns<simb::MCTruth, simb::GTruth> > gassns(new art::Assns<simb::MCTruth, simb::GTruth>);
    
    // Pull the MC generator information out of the event
    art::Handle< std::vector<simb::MCTruth> > mclist;
    evt.getByLabel(fFirstGenCollection, mclist);
   
    art::Handle< std::vector<simb::MCFlux> > fllist;
    art::Handle< std::vector<simb::GTruth> > glist;

    if(!fIsFirstCollectionCosmic) {
      evt.getByLabel(fFirstGenCollection, fllist);
      evt.getByLabel(fFirstGenCollection, glist);
    }

    /// Loop over neutrino interactions
    
    for(unsigned int i = 0; i < mclist->size(); ++i){
      
      simb::MCTruth truth ((*mclist)[i]);
      truthcol->push_back(truth);
      
      if(! fIsFirstCollectionCosmic){
	if(i<fllist->size() && i<glist->size()){
	  simb::MCFlux  flux  ((*fllist)[i]);
	  fluxcol ->push_back(flux);	  
	  simb::GTruth gtruth ((*glist)[i]);
	  gtruthcol->push_back(gtruth);
	  util::CreateAssn(*this, evt, *truthcol, *fluxcol,   *assns,  fluxcol  ->size()-1, fluxcol  ->size());
	  util::CreateAssn(*this, evt, *truthcol, *gtruthcol, *gassns, gtruthcol->size()-1, gtruthcol->size());
	}
      }
    }
      
    ///Second collection
    
    art::Handle< std::vector<simb::MCTruth> > mc2list;
    evt.getByLabel(fSecondCollection, mc2list);
    
    art::Handle< std::vector<simb::GTruth> > g2list;
    art::Handle< std::vector<simb::MCFlux> > fl2list;

    if(!fIsSecondCollectionCosmic) {
      evt.getByLabel(fSecondCollection, g2list);
      evt.getByLabel(fSecondCollection, fl2list);
    }
   
    for(unsigned int i = 0; i < mc2list->size(); ++i){
      
      simb::MCTruth truth ((*mc2list)[i]);
      truthcol->push_back(truth);
      
      if(!fIsSecondCollectionCosmic) {
	if(i<fl2list->size() && i<g2list->size()){
	  simb::MCFlux  flux  ((*fl2list)[i]);
	  simb::GTruth gtruth ((*g2list) [i]);
	  gtruthcol->push_back(gtruth);
	  fluxcol ->push_back(flux);
	  util::CreateAssn(*this, evt, *truthcol, *fluxcol, *assns,  fluxcol->size()-1, fluxcol->size());
	  util::CreateAssn(*this, evt, *truthcol, *gtruthcol, *gassns,  gtruthcol->size()-1, gtruthcol->size()); 
	}
      }
    }	  
              
    // put the collections in the event
    evt.put(std::move(truthcol));
    evt.put(std::move(fluxcol));
    evt.put(std::move(gtruthcol));
    evt.put(std::move(assns));
    evt.put(std::move(gassns));
  }
  
  DEFINE_ART_MODULE(MergeGenCollections);
} // end namespace merge
/////////////////////////////////////////////////////////////////////////
