//////////////////////////////////////////////////////////////////////////////
/// \brief  Module to calculate and store sandbox variables in testing but 
///         added to the CAFs.
/// \author rtoner@fnal.gov, bckhouse@fnal.gov
//////////////////////////////////////////////////////////////////////////////
#include <map>
#include <vector>

#include "TVector3.h"

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "fhiclcpp/ParameterSet.h"

#include "Calibrator/Calibrator.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/CellGeo.h"
#include "Geometry/LiveGeometry.h"
#include "MCCheater/BackTracker.h"
#include "NueSandbox/NueSandObj.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "Utilities/AssociationUtil.h"
#include <TParticlePDG.h>
#include <TDatabasePDG.h>

namespace nuesand
{
  class FillNueSandbox: public art::EDProducer
  {
  public:
    explicit FillNueSandbox(const fhicl::ParameterSet& pset);
    virtual ~FillNueSandbox();

    void   produce(art::Event& evt);
    void   beginRun(art::Run& run);
    double GetMincell(const rb::Track& track);
    int    GetMincellWall();
    double GetTrackAngle(const art::Ptr<rb::Track> track);
    void   GetRegression(const float sumz, const float sumz2, const float sumr, const float sumzr, const float totweight, float& slope, float& inter);
    double TrackD(const float pz0,const float pr,const float slope,const float inter);

    /// Fraction of plane pairs where the "track" seems to change angle, ie
    /// have a different number of hits/plane. For a real track this will be
    /// low, for showers it should be higher.
    ///
    /// loose=true allows 1 cell per plane without counting it in the score
    double FracAngChanges(const rb::Cluster* slice, bool loose) const;

    /// Fraction of planes having the modal number of hits/plane. For a
    /// straight track this should be high, for a shower lower.
    double FracModalHits(const rb::Cluster* slice) const;

    /// Fraction of planes where the cells aren't all adjacent. For a track
    /// this should be low, for a shower or two-pronged event, high.
    ///
    /// loose=true allows single cell gaps and still call a plane contiguous

    double FracNonContiguous(const rb::Cluster* slice, bool loose) const;
    /// Maximum number of contiguous unoccupied planes. Should be low for
    /// normal events, high for misslicings (and possibly pizeros).

    int    MaxGap(const rb::Cluster* slice) const;
    
    /// Fraction of energy of the mother neutrino contained inside the detector
    /// In the NC case, the energy of the daughter neutrino is ignored 
    double GetECF(const art::Ptr<simb::MCTruth> mctruth) const;

    /// Calculate the energy in plane
    double GetPlaneEnergy(const art::PtrVector<rb::CellHit>& chits,
			  const art::Ptr<rb::Prong>& prong,
			  const int & iPlane);

    /// Compute dE/dx of a prong, averaged over all planes
    /// Does not count 0's for planes with no hits in them.
    double GetdEdx(const art::Ptr<rb::Prong> prong);
    

  protected:
    void   FillTruthVars(NueSandObj& vars,
			 const art::PtrVector<rb::CellHit> allhits,
			 const art::Ptr<rb::Cluster> slice) const;

    std::string fSlicerLabel;
    std::string fCosmicLabel;
    std::string fKalmanLabel;
    std::string fProngLabel;
    std::string fProng3DInstLabel;
    std::string generatorName;
    float       fMIPlow;
    float       fMIPhigh;

    bool        fSkipPionTruth;

    //Cluster to contain MIP hits
    rb::Cluster slicemip;

    private:
    int mincellwall=0;

  };

  //----------------------------------------------------------------------
  double FillNueSandbox::GetMincell(const rb::Track& track)
  {
    double fwdcell, bakcell;

    art::ServiceHandle<geo::Geometry>      geom;
    art::ServiceHandle<geo::LiveGeometry>  livegeom;
    art::ServiceHandle<cheat::BackTracker> bt;
    
    TVector3 start  = track.Start();
    TVector3 end    = track.Stop();
    TVector3 dirfor = (end-start).Unit();
    TVector3 dirbak = (start-end).Unit();
    double point[3], dir[3];
    int wall;
    
    end.GetXYZ(point);  // starting at track end point projected forwards along track dir = fwd
    dirfor.GetXYZ(dir);
    fwdcell = livegeom->ProjectedLiveCellsToEdge(point,dir);
    wall = livegeom->ProjectedWall(point,dir);
    
    start.GetXYZ(point);  // starting at track start point projected backwards along track dir = bak
    dirbak.GetXYZ(dir);
    bakcell = livegeom->ProjectedLiveCellsToEdge(point,dir);
    
    double mincell = std::min(bakcell,fwdcell);
    
    if(fwdcell < bakcell){ // this wall is assuming the Instrumented Det Ends (downstream); if there are dropped DCMs (or other non-rectangular solid geometry) it might be wrong
      mincellwall = wall;
    }
    else{
      mincellwall = livegeom->ProjectedWall(point,dir);
    }
    
    return mincell;
    
  }
  
  //----------------------------------------------------------------------
  int FillNueSandbox::GetMincellWall()
  {
    return mincellwall;
  }
  
  //----------------------------------------------------------------------
  double FillNueSandbox::GetTrackAngle(const art::Ptr<rb::Track> track)
  {
    art::ServiceHandle<geo::Geometry> geom;
    
    TVector3 trackdir = track->Dir();
    TVector3 beamdir  = geom->NuMIBeamDirection();
    float    ret      = trackdir*beamdir;
    
    return ret;
    
  }
  
  //----------------------------------------------------------------------
  
  void FillNueSandbox::GetRegression(const float sumz, const float sumz2, const float sumr, const float sumzr, const float totweight, float& slope, float& inter)
  {
    
    slope = 0;
    inter = 0;    
    
    if (totweight>0){

      if (sumz2 == (1/totweight)*sumz*sumz){
	slope = 1.0e4;
      } else {
	slope = (sumzr - (1/totweight)*sumz*sumr)/(sumz2 - (1/totweight)*sumz*sumz);
      }
      inter = sumr/totweight - slope*sumz/totweight;
    }
    
  }
  
  //----------------------------------------------------------------------
  
  double FillNueSandbox::TrackD(const float pz0,const float pr,const float slope,const float inter)
  {
    
    double dtrk = 0;
    
    double pz = pz0;
    if (pz0<0) pz=0;
    
    dtrk = (pz + slope*(pr - inter))/TMath::Sqrt(1+slope*slope);

    if (dtrk<0) dtrk=0;
    
    return dtrk;
    
  }
  
  //----------------------------------------------------------------------
  double FillNueSandbox::FracAngChanges(const rb::Cluster* slice,
                                        bool loose) const
  {
    std::map<int, int> hitsPerPlane[2];
    
    const unsigned int cellMax = slice->NCell();
    for(unsigned int cellIdx = 0; cellIdx < cellMax; ++cellIdx){
      const rb::CellHit& chit = *slice->Cell(cellIdx);
      ++hitsPerPlane[chit.View()][chit.Plane()];
    }
    
    int nChanges = 0;
    for(geo::View_t view: {geo::kX, geo::kY}){
      int hitsPrev = -1;
      for(auto it: hitsPerPlane[view]){
        const int hits = it.second;
        if(hitsPrev > 0){
          if(loose){
            if(abs(hits-hitsPrev) > 1) ++nChanges;
          }
          else{
            if(hits != hitsPrev) ++nChanges;
          }
        }
        hitsPrev = hits;
      }
    }
    
    return nChanges/double(hitsPerPlane[0].size()+hitsPerPlane[1].size()-2);
  }
  
  //----------------------------------------------------------------------
  double FillNueSandbox::FracModalHits(const rb::Cluster* slice) const
  {
    std::map<int, int> hitsPerPlane[2];
    
    const unsigned int cellMax = slice->NCell();
    for(unsigned int cellIdx = 0; cellIdx < cellMax; ++cellIdx){
      const rb::CellHit& chit = *slice->Cell(cellIdx);
      ++hitsPerPlane[chit.View()][chit.Plane()];
    }
    
    double ret = 0;
    for(geo::View_t view: {geo::kX, geo::kY}){
      int maxCells = 0;
      for(auto it: hitsPerPlane[view]) maxCells = std::max(it.second, maxCells);
      int numPlanes = 0;
      for(auto it: hitsPerPlane[view]){
        if(it.second == maxCells) ++numPlanes;
      }
      ret += (.5*numPlanes)/hitsPerPlane[view].size();
    }
    
    return ret;
  }
  
  //----------------------------------------------------------------------
  double FillNueSandbox::FracNonContiguous(const rb::Cluster* slice,
                                           bool loose) const
  {
    std::map<int, std::vector<int> > hitsInPlane;
    
    const unsigned int cellMax = slice->NCell();
    for(unsigned int cellIdx = 0; cellIdx < cellMax; ++cellIdx){
      const rb::CellHit& chit = *slice->Cell(cellIdx);
      hitsInPlane[chit.Plane()].push_back(chit.Cell());
    }
    
    int numNonContig = 0;

    const int thresh = loose ? 2 : 1;

    for(auto planeIt: hitsInPlane){
      std::vector<int>& hits = planeIt.second;
      std::sort(hits.begin(), hits.end());

      int lastCell = -1;
      for(int cell: hits){
        if(cell-lastCell > thresh && lastCell >= 0){
          ++numNonContig;
          break;
        }
        lastCell = cell;
      }
    }

    return numNonContig/double(hitsInPlane.size());
  }

  //----------------------------------------------------------------------
  /// Maximum number of contiguous unoccupied planes. Should be low for normal
  /// events, high for misslicings (and possibly pizeros).
  int FillNueSandbox::MaxGap(const rb::Cluster* slice) const
  {
    std::set<int> planes;

    const unsigned int cellMax = slice->NCell();
    for(unsigned int cellIdx = 0; cellIdx < cellMax; ++cellIdx){
      const rb::CellHit& chit = *slice->Cell(cellIdx);
      planes.insert(chit.Plane());
    }

    int prevPlane = -1;
    int maxGap = 0;
    for(int plane: planes){
      if(prevPlane >= 0) maxGap = std::max(maxGap, plane-prevPlane-1);
      prevPlane = plane;
    }

    return maxGap;
  }

  
  //----------------------------------------------------------------------
  FillNueSandbox::FillNueSandbox(const fhicl::ParameterSet& pset):
    fSlicerLabel	(pset.get< std::string >("SlicerLabel")),
    fCosmicLabel	(pset.get< std::string >("CosmicLabel")),
    fKalmanLabel	(pset.get< std::string >("KalmanLabel")),
    fProngLabel		(pset.get< std::string >("ProngLabel")),
    fProng3DInstLabel	(pset.get< std::string >("Prong3DInstLabel")),
    generatorName	(pset.get< std::string >("GeneratorInput")),
    fMIPlow		(pset.get< float       >("MIPlow")),
    fMIPhigh		(pset.get< float       >("MIPhigh")),
    fSkipPionTruth	(pset.get< bool        >("SkipPionTruth"))

  {
    produces<std::vector<nuesand::NueSandObj>>();
    // Associate with the slice
    produces<art::Assns<nuesand::NueSandObj, rb::Cluster>>();
  
    produces< std::vector<rb::Track> >();
    // Associate Tracks with the Slice they came from
    produces< art::Assns<rb::Track, rb::Cluster> >();

  }


  //----------------------------------------------------------------------
  FillNueSandbox::~FillNueSandbox()
  {
  }

  //-------------------------------------------------------------------
  void FillNueSandbox::beginRun(art::Run& run)
  {
  }

  //-------------------------------------------------------------------
  void FillNueSandbox::produce(art::Event& evt)
  {
    
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<calib::Calibrator> cal;
    art::ServiceHandle<cheat::BackTracker> bt;
    
    std::unique_ptr<std::vector<nuesand::NueSandObj>> sandcol(new std::vector<nuesand::NueSandObj>);
    std::unique_ptr<art::Assns<nuesand::NueSandObj, rb::Cluster>> assns(new art::Assns<nuesand::NueSandObj, rb::Cluster>);
    std::unique_ptr< std::vector<rb::Track> > trackcol(new std::vector<rb::Track>);
    std::unique_ptr< art::Assns<rb::Track, rb::Cluster> > assnsslc(new art::Assns<rb::Track, rb::Cluster>);
    
    art::Handle<std::vector<rb::Cluster>> slices;
    evt.getByLabel(fSlicerLabel, slices);
    
    art::FindManyP<rb::Track> cosmicAssnList(slices, evt, fCosmicLabel); 
    art::FindManyP<rb::Track> kalmanAssnList(slices, evt, fKalmanLabel); 
    art::FindManyP<rb::Prong> prongAssnList(slices, evt,
					    art::InputTag(fProngLabel, 
							  fProng3DInstLabel));
    
     const unsigned int sliceMax = slices->size();
    
    // Use property that every hit is in exactly one slice
    art::PtrVector<rb::CellHit> allhits;
    for(unsigned int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      const art::Ptr<rb::Cluster> slice(slices, sliceIdx);
      for(unsigned int i = 0; i < slice->NCell(); ++i){
	allhits.push_back(slice->Cell(i));
      }
    }
    
    std::map<int,std::vector<int> > plane2cell;
  
    for(unsigned int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      
      //Clear out map of planes->cells
      plane2cell.clear();
      
      //Ptr to cluster
      const art::Ptr<rb::Cluster> slice(slices, sliceIdx);
      
      //Continue if noise
      if(slice->IsNoise()) continue;

      //Get slice mean and fid (to define sandbox)
      const TVector3 mean = slice->MeanXYZ();
      const bool inFid = geom->isInsideFiducialVolume(mean.X(),
                                                      mean.Y(),
                                                      mean.Z());
      
      
      //For Calibration:
      const double wx = (slice->NXCell() > 0) ? slice->W(slice->XCell(0).get()) : 0;
      const double wy = (slice->NYCell() > 0) ? slice->W(slice->YCell(0).get()) : 0;
      
      //Count of hits and MIP hits:
      double tothits=0;
      double totmips=0;

      slicemip.Clear();

      //Counters for regression analysis:
      float totweightx=0;
      float totweighty=0;
      int   tothitsx=0;
      int   tothitsy=0;
      float sumz_x  = 0;
      float sumz_y  = 0;
      float sumx  = 0;
      float sumy  = 0;
      float sumz2_x = 0;
      float sumz2_y = 0;
      float sumzx = 0;
      float sumzy = 0;

      float pz, px, py;
      float pecorr = 0;

      // variables used to find extent of planes in 
      // each cluster view
      unsigned int nXplanes = 0, nYplanes = 0;
      unsigned int PXhi  = 0;
      unsigned int PXlo  = UINT_MAX;
      unsigned int PYhi  = 0;
      unsigned int PYlo  = UINT_MAX;

      //loop over hits in each view independently
      for(int view = geo::kX; view <= geo::kY; ++view) {
        const geo::View_t geoview = geo::View_t(view);

	for(unsigned int hitIdx = 0; hitIdx < slice->NCell(geoview); ++hitIdx) {

          const rb::CellHit* h = slice->Cell(geoview,hitIdx).get();
          if(h->View() == geo::kX) {
            if(h->Plane() > PXhi) PXhi = h->Plane();
            if(h->Plane() < PXhi) PXlo = h->Plane();
          }
	  if(h->View() == geo::kY) {
            if(h->Plane() > PYhi) PYhi = h->Plane();
            if(h->Plane() < PYhi) PYlo = h->Plane();
          }
        }
      }

      nXplanes = PXhi - PXlo+1;
      nYplanes = PYhi - PYlo+1;


      //------------->Loop over hits in Slice!
      for(unsigned int hitIdx = 0; hitIdx < slice->NCell(); ++hitIdx)
	{
	  const art::Ptr<rb::CellHit>& chit = slice->Cell(hitIdx);
	  //Get calbrated reco hit:
	  const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));

	  //Make sure things have been calibrated:
	  if(!rhit.IsCalibrated())
	    {
	      LOG_DEBUG("NueSandbox") << "Not calibrated?! "
				      << chit->Plane() << " " << chit->Cell();
	      continue;
	    }

	  //Total Hits:
	  tothits += 1.0;
	  pecorr = rhit.PECorr();
	  pz = rhit.Z();
	  px = rhit.X();
	  py = rhit.Y();

	  //For regression:
	  if (chit->View()==geo::kX){

	    tothitsx++;
	    totweightx += pecorr;
	    sumz_x  += pz*pecorr;
	    sumz2_x += pz*pz*pecorr;
	    sumx += px*pecorr;
	    sumzx += pz*px*pecorr;
	  } else {

	    tothitsy++;
	    totweighty += pecorr;
	    sumz_y  += pz*pecorr;
	    sumz2_y += pz*pz*pecorr;
	    sumy += py*pecorr;
	    sumzy += pz*py*pecorr;
	  }

	  //Total MIP Hits:
	  if (rhit.PECorr()>fMIPlow&&rhit.PECorr()<fMIPhigh) {
	    totmips += 1.0;
	    slicemip.Add(chit); //Add to new slice!
	  }//end mip if
	  
	  //Plane and Cell Number:
	  int fNumPlane = chit->Plane(); 
	  int fNumCell = chit->Cell();

	  //If there is no entry for this plane, add a new key and vector:
	  if ( plane2cell.find(fNumPlane)==plane2cell.end() ) 
	    plane2cell.insert(std::pair<int, std::vector<int> >(fNumPlane, std::vector<int>()));

	  //Push back unique cell number for this Plane (no duplicates!):
	  if (std::find(plane2cell[fNumPlane].begin(),plane2cell[fNumPlane].end(),fNumCell) == plane2cell[fNumPlane].end())
	    plane2cell[fNumPlane].push_back(fNumCell);

	}//<----------END LOOP OVER HITS

      //New fraction of MIP Hits
      //double fracmip=0;
      //if (tothits>0) fracmip = totmips/tothits;

      //For regression:
      float slopex = 0;
      float interx = 0;
      float slopey = 0;
      float intery = 0;

      //Get regression values:
      GetRegression(sumz_x,sumz2_x,sumx,sumzx,totweightx,slopex,interx);
      GetRegression(sumz_y,sumz2_y,sumy,sumzy,totweighty,slopey,intery);

      float dtrk;
      float dtrkx_min=1.0e6;
      float dtrkx_max=0;
      float dtrky_min=1.0e6;
      float dtrky_max=0;
      float dz_start_x=0;
      float dz_stop_x=0;
      float dz_start_y=0;
      float dz_stop_y=0;

      float mincellavg = 1.0e4;

      if (tothitsx > 1 && tothitsy > 1){

	//------------->Loop over hits in Slice AGAIN (regression)
	for(unsigned int hitIdx = 0; hitIdx < slice->NCell(); ++hitIdx)
	  {
	    const art::Ptr<rb::CellHit>& chit = slice->Cell(hitIdx);
	    //Get calbrated reco hit:
	    const rb::RecoHit rhit(cal->MakeRecoHit(*chit, chit->View() == geo::kX ? wx : wy));
	    
	  //Make sure things have been calibrated:
	    if(!rhit.IsCalibrated())
	      {
		LOG_DEBUG("NueSandbox") << "Not calibrated?! "
					<< chit->Plane() << " " << chit->Cell();
		continue;
	      }
	    
	    dtrk = 0;
	    pz = rhit.Z();
	    px = rhit.X();
	    py = rhit.Y();
	   

	    if (chit->View() == geo::kX) { //X VIEW
	 
   
	      //Projected distance from z=0 to point
	      dtrk = TrackD(pz,px,slopex,interx); 
	 
	      //is it a max?
	      if (dtrk>dtrkx_max) {
		dtrkx_max = dtrk;
		dz_stop_x   = dtrk/TMath::Sqrt(1+slopex*slopex);
	      }
	      
	      //is it a min?
	      if (dtrk<dtrkx_min) {
		dtrkx_min = dtrk;
		dz_start_x  = dtrk/TMath::Sqrt(1+slopex*slopex);
	      }
	    } else { //Y VIEW	     

	      //Projected distance from z=0 to point
	      dtrk = TrackD(pz,py,slopey,intery); 

	      //is it a max?
	      if (dtrk>dtrky_max) {
		dtrky_max = dtrk;
		dz_stop_y   = dtrk/TMath::Sqrt(1+slopey*slopey);
	      }
	      
	      //is it a min?
	      if (dtrk<dtrky_min) {
		dtrky_min = dtrk;
		dz_start_y  = dtrk/TMath::Sqrt(1+slopey*slopey);
	      }
	    }
	    
	  }//<------END SECOND HIT LOOP

	float dz_start = TMath::Min(dz_start_x,dz_start_y);
	float dz_stop  = TMath::Max(dz_stop_x,dz_stop_y);
    
	TVector3 dtrkstart;
	dtrkstart[0] = interx + slopex*dz_start;
	dtrkstart[1] = intery + slopey*dz_start;
	dtrkstart[2] = dz_start;
	
	TVector3 dtrkstop;
	dtrkstop[0] = interx + slopex*dz_stop;
	dtrkstop[1] = intery + slopey*dz_stop;
	dtrkstop[2] = dz_stop;

	TVector3 dtrkdir;
	dtrkdir[0] = slopex;
	dtrkdir[1] = slopey;
	dtrkdir[2] = 1;

	//Make very simple track:	
	rb::Track rtrack(*slice,dtrkstart,dtrkdir);
	rtrack.AppendTrajectoryPoint(dtrkstop);
	
	//Mincell for regression line:
        mincellavg = GetMincell(rtrack);

	//Add to vector and make association
	trackcol->push_back(rtrack);
	util::CreateAssn(*this,evt,*trackcol,slice,*assnsslc);

      }


      //Min and Max of box around MIP frac hits:
      TVector3 def(1000, 1000, 1000);
      TVector3 mipmin = slicemip.NCell() == 0 ?  def : slicemip.MinXYZ();
      TVector3 mipmax = slicemip.NCell() == 0 ? -def : slicemip.MaxXYZ();

      unsigned int numplanesfull = 0; //Number of filled planes
      unsigned int maxCellContig = 0; //Max number cells in a row in one plane
      unsigned int maxCellSum = 0; //Max total of cells in one plane

      //------------>Iterate over planes:
      std::map< int, std::vector<int> >::iterator planeIter;
      for (planeIter = plane2cell.begin(); planeIter != plane2cell.end(); ++planeIter) {

	//Count planes:
	numplanesfull++; 

	//Grab vector of hits
	std::vector<int> cellvect = planeIter->second;

	//Sort into correct order
	std::sort(cellvect.begin(),cellvect.end());

	//Look for maximum cells in single plane:
	if (cellvect.size()>maxCellSum)
	  maxCellSum = cellvect.size(); 

	//Counters for max cells in a row:
	unsigned int maxrow=0;
	unsigned int nrow=1;

	//------------>Iterate over cells:
	for (unsigned int icell=0; icell<cellvect.size(); icell++){
          if (icell!=0){
            if (cellvect[icell] == (cellvect[icell-1]+1)){
              nrow++;
              if (nrow>maxrow) maxrow=nrow;
            } else nrow=1;
          }
	}//------------->END LOOP OVER CELLS

	//Most in a row in this plane?
	if (maxrow>maxCellContig)
	  maxCellContig = maxrow; 

      }//<------------END LOOP OVER PLANES

      //Some dummy vars:
      float tempcell = 1.0e4;    
      float trackangle = -5;    
      int maxtrackID = -1;    
      float maxtracklen = 0;    
      float mincellcos = 1.0e4;
      int mincellwallcos = 0;
      
      if(cosmicAssnList.isValid()){
	const std::vector< art::Ptr<rb::Track> > sliceTracksCosmic = cosmicAssnList.at(sliceIdx);
	
	//-----------> LOOP OVER COSMIC TRACKS
	for(size_t trkIdx = 0; trkIdx < sliceTracksCosmic.size(); ++trkIdx){
	  
	  //Track:
	  const art::Ptr<rb::Track> track = sliceTracksCosmic[trkIdx];
	  
	  //Find projected cells to edge:
          tempcell = GetMincell(*track);
	  
	  //Is this the closest to the edge?
	if (tempcell < mincellcos){
          mincellcos = tempcell;
          mincellwallcos = GetMincellWall();
        } 

	  
	  //Find longest cosmic track:
	if (track->TotalLength() > maxtracklen) {
	    maxtracklen = track->TotalLength();
	    maxtrackID = trkIdx;
	  }
	}//<------------END LOOP OVER TRACKS
	
	//CosAngle vs beam for the longest cosmic track (-5 otherwise)
	if (maxtrackID>=0)
	  trackangle = GetTrackAngle(sliceTracksCosmic[maxtrackID]); 
	
      }//<-------End isValid FindManyP

      float mincellkal = 1.0e4;
      
      if(kalmanAssnList.isValid()){
	const std::vector< art::Ptr<rb::Track> > sliceTracksKalman = kalmanAssnList.at(sliceIdx);
	
	//Some dummy vars:
	tempcell = 1.0e4;    
	
	
	//-----------> LOOP OVER KALMAN TRACKS
	for(size_t trkIdx = 0; trkIdx < sliceTracksKalman.size(); ++trkIdx){
	  
	  //Track:
	  const art::Ptr<rb::Track> track = sliceTracksKalman[trkIdx];
	  
	  //Find projected cells to edge:
          tempcell = GetMincell(*track);
	  
	  //Is this the closest to the edge?
	  if (tempcell < mincellkal) mincellkal = tempcell;
	  
	}//<------------END LOOP OVER TRACKS
	
      }//<----End loop over Kalman isValid
      
      double dedx_png1 = 0, dedx_png2 = 0;
      if(prongAssnList.isValid()){
	const std::vector< art::Ptr<rb::Prong> > sliceProngs = 
	  prongAssnList.at(sliceIdx);

	double maxlength = 0, secMaxlength = 0;
	int maxidx = -1, secmaxidx = -1;

	for(unsigned int ipng = 0; ipng < sliceProngs.size(); ++ipng){
	  double len = sliceProngs[ipng]->TotalLength();

	  if(len > maxlength ){
	    secMaxlength = maxlength;
	    maxlength    = len;
	    secmaxidx    = maxidx;
	    maxidx       = ipng;
	  }
	  if(len > secMaxlength && len < maxlength){
	    secMaxlength = len;
	    secmaxidx    = ipng;
	  }
	}// end loop over prongs

	if(maxidx > -1 )
	  dedx_png1 = GetdEdx(sliceProngs[maxidx]);
	if(secmaxidx > -1 )
	  dedx_png2 = GetdEdx(sliceProngs[secmaxidx]);
      }// endif prong assn is valid

      //Make and fill the sandbox!
      nuesand::NueSandObj sandbox(inFid, mean);
      sandbox.fMincellCosmic     = mincellcos;
      sandbox.fMincellWallCosmic = mincellwallcos;
      sandbox.fMincellKalman     = mincellkal;
      sandbox.fAngleCosmic       = trackangle;
      sandbox.fNMIPHits          = totmips;
      sandbox.fMIPMin            = mipmin;
      sandbox.fMIPMax            = mipmax;
      sandbox.fFilledPlanes      = numplanesfull;
      sandbox.fMaxCellsInPlane   = maxCellSum;
      sandbox.fMaxRowOfCells     = maxCellContig;
      sandbox.fMincellAvg        = mincellavg;
      sandbox.fExtentXPlanes     = nXplanes;
      sandbox.fExtentYPlanes     = nYplanes;
      
      sandbox.fFracAngChanges      = FracAngChanges(slice.get(), false);
      sandbox.fFracAngChangesLoose = FracAngChanges(slice.get(), true);
      sandbox.fFracModalHits       = FracModalHits(slice.get());
      sandbox.fFracNonContig       = FracNonContiguous(slice.get(), false);
      sandbox.fFracNonContigLoose  = FracNonContiguous(slice.get(), true);
      sandbox.fMaxGap              = MaxGap(slice.get());

      sandbox.fdEdxProng1          = dedx_png1;
      sandbox.fdEdxProng2          = dedx_png2;

      FillTruthVars(sandbox, allhits, slice);
      
      sandcol->push_back(sandbox);
      util::CreateAssn(*this, evt, *sandcol, slice, *assns);
    } // end for slice idx

    evt.put(std::move(sandcol));
    evt.put(std::move(assns));
    evt.put(std::move(trackcol));
    evt.put(std::move(assnsslc));
  }

  //-------------------------------------------------------------------
  void FillNueSandbox::FillTruthVars(NueSandObj& vars,
                                     const art::PtrVector<rb::CellHit> allhits,
                                     const art::Ptr<rb::Cluster> slice) const
  {
    art::ServiceHandle<cheat::BackTracker> bt;
    if(!bt->HaveTruthInfo()) return;

    art::ServiceHandle<geo::Geometry> geom;

    // Ideally we wouldn't be essentially duplicating CAFMaker's truth matching
    // here. Maybe we can have a truth sandbox that's associated with MCTruths,
    // not slices?

    const std::vector<cheat::NeutrinoEffPur> eps = bt->SliceToNeutrinoInteractions(slice, allhits);
    if(eps.empty()) return;

    const art::Ptr<simb::MCTruth> truth = eps[0].neutrinoInt;
    
    vars.ECFNu = GetECF(truth);

    const std::vector<const sim::Particle*> parts = bt->MCTruthToParticles(truth);

    // Find the biggest pions
    int bestIdx = -1;
    for(unsigned int primIdx = 0; primIdx < parts.size(); ++primIdx){
      const sim::Particle* part = parts[primIdx];
      if(part->Mother()) continue; // Not a primary

      const int pdg = part->PdgCode();
      const double E = part->E();
      if(pdg == 111){
        ++vars.fNPi0;
        if(E > vars.fPi0Energy){
          bestIdx = primIdx;
          vars.fPi0Energy = E;
        }
      }
      if(pdg == +211){
        ++vars.fNPiPlus;
        if(E > vars.fPiPlusEnergy) vars.fPiPlusEnergy = E;
      }
      if(pdg == -211){
        ++vars.fNPiMinus;
        if(E > vars.fPiMinusEnergy) vars.fPiMinusEnergy = E;
      }
    } // end for primIdx

    // No pizeros, leave the rest zeros
    if(bestIdx < 0) return;

    // Otherwise, take the biggest
    const sim::Particle* pizero = parts[bestIdx];

    // TEMPORARILY DISABLING CODE BELOW
    // TO BE REVERTED SOON
    // The code below calls sim::Particle::Daughter() which is sometimes
    // busted.  This will be fixed in newly produced sim files, but for
    // now, just skip the storing of this info.
    if(fSkipPionTruth) return;

    if(pizero->NumberDaughters() != 2) return;

    const sim::ParticleNavigator& nav = bt->ParticleNavigator();
    const sim::Particle* phot0 = nav[pizero->Daughter(0)];
    if(phot0->PdgCode() != 22) return;
    const sim::Particle* phot1 = nav[pizero->Daughter(1)];
    if(phot1->PdgCode() != 22) return;

    vars.fPhot0E = phot0->E();
    vars.fPhot1E = phot1->E();
    vars.fOpenCos = phot0->Momentum(0).Vect().Unit().Dot(phot1->Momentum(0).Vect().Unit());

    // If the photon leaves the detector it can go a very long way
    const TVector3 e0 = phot0->EndPosition().Vect();
    if(geom->isInsideDetectorBigBox(e0.X(), e0.Y(), e0.Z())){
      vars.fConvL0 = (e0-phot0->Position(0).Vect()).Mag();
    }
    else{
      vars.fConvL0 = -1;
    }
    const TVector3 e1 = phot1->EndPosition().Vect();
    if(geom->isInsideDetectorBigBox(e1.X(), e1.Y(), e1.Z())){
      vars.fConvL1 = (e1-phot1->Position(0).Vect()).Mag();
    }
    else{
      vars.fConvL1 = -1;
    }

    // Make sure the biggest photon is first
    if(vars.fPhot1E > vars.fPhot0E){
      std::swap(vars.fPhot0E, vars.fPhot1E);
      std::swap(vars.fConvL0, vars.fConvL1);
    }
  }
  
  //-------------------------------------------------------------------


  double FillNueSandbox::GetECF(const art::Ptr<simb::MCTruth> mc) const
  {
    float       ECFNu(0.0);
    float       DeltaECC(0.0);
    float       DeltaENON(0.0);
    float       TrueEnergyNON(0.0);
    float       ContainmentCC(0.0);
    float       ContainmentNN(0.0);
    float       EscapedEnergyNN(0.0);
    float       EscapedEnergyCC(0.0);
    sim::ParticleNavigator pnav;
    
    const simb::MCNeutrino& mc_neutrino = mc->GetNeutrino();
    const simb::MCParticle& nu = mc_neutrino.Nu();
    art::ServiceHandle<cheat::BackTracker>  bt;
    std::vector<const sim::Particle*> parts=bt->MCTruthToParticles(mc); ///All the primary and other "levels" daughter particles.
    art::ServiceHandle<geo::Geometry>  geom;
      
    ////Yep,"CUT"
    if(abs(nu.Vx()) < geom->DetHalfWidth()  &&        //True Vertex Limitations (Interaction Vertex !)
       abs(nu.Vy())  < geom->DetHalfHeight()  &&
       nu.Vz() < geom->DetLength() &&
       nu.Vz() > 0 ){
      if(mc_neutrino.CCNC()==simb::kCC){
	DeltaECC = 0;
	EscapedEnergyCC = 0.0;
	for (unsigned int j = 0; j < parts.size(); ++j){  ///Loop for All Particles
	  bool inDetStart = fabs(parts[j]->Vx(0)) <= geom->DetHalfWidth() //Check X inside detector
	    && fabs(parts[j]->Vy(0)) <= geom->DetHalfHeight()             // then check Y
	    && parts[j]->Vz(0) >= 0 && parts[j]->Vz(0) <= geom->DetLength();   //then starting and End Z
	  if (!inDetStart) continue;
	  bool inDet = fabs(parts[j]->EndX()) <= geom->DetHalfWidth()           ///Check End X inside detector
	    && fabs(parts[j]->EndY()) <= geom->DetHalfHeight()           ///Check End Y inside detector
	    && parts[j]->EndZ() >= 0 && parts[j]->EndZ() <= geom->DetLength();           ///Check End Z inside detector
	  pnav.Add(new sim::Particle(*parts[j]));                       ///Keep adding each particle to the PaticleNavigator
	  if (!inDet &&                  ///We calculate the escaped energy of all escaped particles
	      abs(parts[j]->PdgCode()) != 12 &&
	      abs(parts[j]->PdgCode()) != 14 &&
	      abs(parts[j]->PdgCode()) != 16 ){
	    for (unsigned int q = 0; q < parts[j]->NumberTrajectoryPoints(); ++q){ ///Loop for All hit Points
	      if (fabs(parts[j]->Vx(q)) > geom -> DetHalfWidth() || ///Check to see if a trahectory point is outside the detector
		  fabs(parts[j]->Vy(q)) > geom -> DetHalfHeight() ||
		  parts[j]->Vz(q) > geom->DetLength() ||
		  parts[j]-> Vz(q) < 0) {
		DeltaECC = (parts[j]->E(q-1) - parts[j] -> Mass());
	      } ///Check to see if trajectory point is in the detector
	    } /// All Hit Point
	    EscapedEnergyCC += DeltaECC;
	  }/// Pick up non-neutrinos particles which are inside the detecter
	}///Loop for All Particles
	ContainmentCC = (nu.E() - EscapedEnergyCC);
	ECFNu = (ContainmentCC/nu.E());
	LOG_DEBUG("NueSandbox") <<"ECF(CC) is "<< ECFNu;
      }///CC CUT
      //////////////&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&    NC    &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&///////
      else if(mc_neutrino.CCNC()==simb::kNC){
	DeltaENON = 0;
	EscapedEnergyNN = 0.0;
	for (unsigned int j = 0; j < parts.size(); ++j){  ///Loop for All Particles
	  bool inDetStart = fabs(parts[j]->Vx(0)) <= geom->DetHalfWidth() //Check X inside detector
	    && fabs(parts[j]->Vy(0)) <= geom->DetHalfHeight()             // then check Y
	    && parts[j]->Vz(0) >= 0 && parts[j]->Vz(0) <= geom->DetLength();   //then starting and End Z
	  if (!inDetStart) continue;
	  bool inDet = fabs(parts[j]->EndX()) <= geom->DetHalfWidth()           ///Check End X inside detector
	    && fabs(parts[j]->EndY()) <= geom->DetHalfHeight()           ///Check End Y inside detector
	    && parts[j]->EndZ() >= 0 && parts[j]->EndZ() <= geom->DetLength();           ///Check End Z inside detector
	  pnav.Add(new sim::Particle(*parts[j]));                       ///Keep adding each particle to the PaticleNavigator
	  if (!inDet &&                  ///We calculate the escaped energy of all escaped particles except neutrinos
	      abs(parts[j]->PdgCode()) != 12 &&
	      abs(parts[j]->PdgCode()) != 14 &&
	      abs(parts[j]->PdgCode()) != 16 ){
	    for (unsigned int q = 0; q < parts[j]->NumberTrajectoryPoints(); ++q){ ///Loop for All hit Points
	      if (fabs(parts[j]->Vx(q)) > geom -> DetHalfWidth() || ///Check to see if a trahectory point is outside the detector
		  fabs(parts[j]->Vy(q)) > geom -> DetHalfHeight() ||
		  parts[j]->Vz(q) > geom->DetLength() ||
		  parts[j]-> Vz(q) < 0) {
		DeltaENON = (parts[j]->E(q-1) - parts[j] -> Mass());
	      } ///Check to see if trajectory point is in the detector
	    } /// All Hit Point
	    EscapedEnergyNN += DeltaENON;
	  }/// Pick up non-neutrnos particles which are inside the detecter
	}///Loop for All Particles
	int numberOfPrimaries = pnav.NumberOfPrimaries();
	///Now, we calculate the ture energy by added all the primaries' energy
	float RealEnergyNN = 0.0;
	TrueEnergyNON = 0.0;
	for (int i = 0; i != numberOfPrimaries; ++i ){     ///Loop for the all primary particles
	  const sim::Particle* myprimary = pnav.Primary(i);
	  if(abs(myprimary->PdgCode()) == 12 ||
	     abs(myprimary->PdgCode()) == 14 ||
	     abs(myprimary->PdgCode()) == 16  ){
	    RealEnergyNN=(myprimary->E() - myprimary->Mass());
	    TrueEnergyNON =nu.E()-RealEnergyNN;
	  }                                           ///Ignoring mother neutrinos from the primaries
	}                                             ///Loop for the all primary particles
	ContainmentNN = (TrueEnergyNON - EscapedEnergyNN);
	ECFNu = (ContainmentNN/nu.E());
	LOG_DEBUG("NueSandbox") <<"ECF(NC) is "<< ECFNu;
	}///NC CUT
	else{
	  LOG_DEBUG("NueSandbox") <<"It is neither CC nor NC";
	}
    }
    return ECFNu;
  }// End of GetECF


  //-------------------------------------------------------------------


  double FillNueSandbox::GetPlaneEnergy(const art::PtrVector<rb::CellHit>&hits,
					const art::Ptr<rb::Prong>& prong,
					const int& iPlane)
  {
    double planeE = 0;
    int nchits = hits.size();

    for(int ihit = 0; ihit < nchits; ++ihit){
      int hitplane = hits[ihit]->Plane();
      
      if( hitplane == iPlane ){
	rb::RecoHit rhit = prong->RecoHit(ihit);
	if(rhit.IsCalibrated()){
	  planeE = planeE + rhit.GeV();
	}
      }

      // hits are ordered by plane. Take advantage
      else if( hitplane > iPlane)
	return planeE;

    }// end loop over chits

    return planeE;
  }// End of GetPlaneEnergy
  

  //-------------------------------------------------------------------

  double FillNueSandbox::GetdEdx(const art::Ptr<rb::Prong> prong)
  {
    art::ServiceHandle<geo::Geometry> geom;
    art::PtrVector<rb::CellHit> pngHits = prong->AllCells();
    rb::SortByPlane(pngHits);
    std::set<int> planes;

    double dedx = 0;    
    double dirz = fabs(prong->Dir().Z());

    if( pngHits.empty() )
      return dedx;

    unsigned int ncell = pngHits.size();
    
    for(unsigned int i = 0; i < ncell; ++i)
      planes.insert( prong->Cell(i)->Plane() );
    
    for(int iplane : planes){
      double celldepth = 2 * geom->Plane(iplane)->Cell(0)->HalfD(); 
      double dx = prong->TotalLength();

      if( dirz > 1e-6)
	dx = celldepth/dirz;

      double de = GetPlaneEnergy(pngHits, prong, iplane);
      dedx = dedx + de/dx;

    }// end loop over planes with hits

    return dedx/planes.size();

  }// End of GetdEdx

  //-------------------------------------------------------------------
  
  DEFINE_ART_MODULE(FillNueSandbox);
} //end namespace
