////////////////////////////////////////////////////////////////////////
// $Id: NumuEAlg.cxx,v 1.10 2012-11-15 20:46:24 lein Exp $
//
// Numu Energy Algorithms
//
// lein@physics.umn.edu
//
////////////////////////////////////////////////////////////////////////

// NOvA includes
#include "NumuEnergy/NumuEAlg.h"
#include "ReMId/ReMId.h"
#include "RecoBase/RecoHit.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "Utilities/func/EnvExpand.h"
#include "Utilities/func/MathUtil.h"
#include "NumuEnergy/ScatterEnergy.h"

#include <cmath>

namespace numue {

  //----------------------------------------------------------------------
  NumuEAlg::NumuEAlg(fhicl::ParameterSet const& pset):
    fPIDModuleLabel     (pset.get< std::string     >("PIDModuleLabel")     ),
    fPIDMethod          (pset.get< bool >("PIDMethod") ),
    fTMVApath           (pset.get< std::string     >("TMVApath") ),
    fFD                 (false),
    fND                 (false),
    fMaxTrkLen          (-1),
    fNumiDir            (-1,-1,-1),
    fTranPlane          (-1),
    fTranZPos           (-1)
  {
    art::ServiceHandle<geo::Geometry>  geom;

    novadaq::cnv::DetId detID = geom->DetId();
    if (detID == novadaq::cnv::kNEARDET) fND = true;
    if (detID == novadaq::cnv::kFARDET)  fFD = true;

    //We want the plane just BEFORE the muon catcher, hence the -1.
    //We want the general center of the z plane - so I picked cell 32.
    //This could be made more advanced / clever in the future if needed.
    if (fND){
      fTranPlane = (geom->FirstPlaneInMuonCatcher())-1; 
      double xyz[3];
      geom->Plane(fTranPlane)->Cell(32)->GetCenter(xyz);
      fTranZPos = xyz[2];
    }

    double detHalfWidth   =  geom->DetHalfWidth();
    double detHalfHeight  =  geom->DetHalfHeight();
    double detLength      =  geom->DetLength();

    fMaxTrkLen = util::pythag(detHalfWidth*2, detHalfHeight*2, detLength);
    fNumiDir   = geom->NuMIBeamDirection();

    fReaderE = new TMVA::Reader;
    fReaderE->AddVariable("slc.calE",&Evars[0]);
    fReaderE->AddVariable("sel.cosrej.anglekal",&Evars[1]);
    fReaderE->AddVariable("sel.cosrej.scatt",&Evars[2]);
    fReaderE->AddVariable("sel.cosrej.eratio",&Evars[3]);
    fReaderE->AddVariable("sel.cosrej.kalfwdcell",&Evars[4]);
    fReaderE->AddVariable("trk.cosmic[0].nhit",&Evars[5]);
    fReaderE->AddVariable("sel.cosrej.ntracks3d",&Evars[6]);
    std::string pidpath = util::EnvExpansion(fTMVApath);
    fReaderE->BookMVA("BDTG",pidpath+"UC_Eres_BDTG.weights.xml");

    return;
  }

  //----------------------------------------------------------------------
  NumuEAlg::~NumuEAlg()  
  {  
  } 

  //----------------------------------------------------------------------
  NumuE NumuEAlg::FDEnergy(std::vector< art::Ptr<rb::Track> > const sliceTracks,
                           art::PtrVector< rb::CellHit >            sliceHits,
                           art::Ptr< rb::Cluster>                   sliceCluster,
                           art::Event                  const&       e,
                           murem::TrackCleanUpAlg*                  trkCleanUpAlg)
  {
    /* This returns multiple neutrino energy estimates, some of which are better than others.
       They are: CalCC, TrkQEE, TrkNonQEE, and TrkCCE. The default E value is set to be indentically
       equal to TrkCCE and comes with the same caveats. At the same time, we are also 
       setting HadCalE and HadTrkE. If no 3D track is found, no energy estimates are made.
       When using the PIDMethod (which is default), the muon track is defined as the 3D track
       with the highest PID. Otherwise, we simply take the longest 3D track as the muon.

       CalCC is a very simple, first pass at energy estimation for the numu cc group.
       ONE SHOULD ALMOST NEVER USE CalCC! THE OTHER ENERGIES ARE USUALLY BETTER CHOICES!
       The weights and methods used were optimized for a contained population that 
       doesn't distinguish between nonqe and qe events. No background existed in the
       optimized sample. It used a linear fit to find the two weights and it preformed
       the fit simultaneously. The optimization was done only for neutrinos with true 
       energy between 1 and 3 GeV and basic activity cuts. The muon energy is calculated 
       by summing the reco GeV for every hit on the track. The vertex energy is simply 
       a summed GeV of every hit NOT on the muon track.
       
       The form of the neutrino energy for CalCC is: 
       Reco Nu Energy = Weight_muon * Summed GeV of Muon Track 
                      + Weight_vertex * Summed GeV of all non-Muon Track hits

       The CalCC method is NOT our best energy estimator and is mostly an accident of history. 

       TrkQE, TrkNonQE and TrkCC are very similar. The weights and methods used were optimized
       for a contained population that distinguishes by truth between QE and NonQE events. 
       The muon energy is found by a spline fit
       to muon track length and muon true energy. This function is MuonEFromTrackLength and more info
       can be found at the top of that function. Then the hadronic energy is defined as the sum of HadCalE
       and HadTrkE. HadCalE is the same vertex energy as used in CalCC - it is just a sum of the GeV
       of all hits not on the muon track. HadTrkE is the hadronic energy determined to be on
       the muon track. This is found by a function in MuonRemove. Summed together, this is the visible hadronic 
       energy. Then this energy is transformed into our reco hadronic energy
       by a spline fit. This fit is different for the NonQE, QE and combined populations. It is a fit between
       the summed hadronic energy and (True Neutrino Energy - Reco Muon Energy). More information can
       be found in the functions HadEQE, HadENonQE, and HadECC. Finally, the neutrino energy is just as 
       sum of the Muon Energy and the adjusted Hadronic Energy.

       Currently, all the energy methods are only really meant for contained events; however, this function 
       and the module that calls it doesn't not attempt to determine containment. Please 
       use with discretion!
    */

    NumuE  sliceEnergy;

    int    bestTrack  = -1;
    double trackLen   = 0.0;

    if (fPIDMethod) bestTrack = remid::HighestPIDTrack(sliceTracks, fPIDModuleLabel, e);
    else bestTrack = LongestTrack(sliceTracks, trackLen);

    if(bestTrack == -1) return sliceEnergy;
    if(bestTrack == 999) return sliceEnergy;

    if (fPIDMethod) trackLen = sliceTracks[bestTrack]->TotalLength();

    //Gets GeV of hits on track
    double recoTrkGeV = sliceTracks[bestTrack]->TotalGeV();

    //Calculates GeV of muon from track length
    double muonGeVFromLen = MuonEFromTrackLength(trackLen);

    art::PtrVector< rb::CellHit > trackHits;
    trackHits = sliceTracks[bestTrack]->AllCells();    

    //Getting all the hits in the slice that are not on our best track
    rb::Cluster vertexHits = MakeVertexCluster(sliceHits,trackHits);
    
    //Gets GeV of hadronic hits NOT on track
    double vertexEnergy = 0.0;
    if (vertexHits.NCell()>0){
      vertexEnergy = vertexHits.TotalGeV(); 
    }

    //Gets GeV of hadronic hits on track
    double extraHadE = trkCleanUpAlg->ExtraEOnTrack(*sliceTracks[bestTrack],*sliceCluster);

    //Get adjusted GeV of total hadronic energy assuming QE/NonQE
    double totHadE = vertexEnergy + extraHadE;
    double hadQEGeV = HadEQE(totHadE);
    double hadNonQEGeV = HadENonQE(totHadE);
    double hadCCGeV = HadECC(totHadE);

    //***************Fill values in our slice energy object

    double calCCMuonWeight = 1.81134; //Using Summed GeV method and Susan's fit from MN talk on 9/7/2012
    double calCCVertWeight = 2.17923;
    double calcc = calCCMuonWeight*recoTrkGeV + calCCVertWeight*vertexEnergy;
    sliceEnergy.SetCalCCE(calcc);
    
    sliceEnergy.SetHadCalE(vertexEnergy);
    sliceEnergy.SetHadTrkE(extraHadE);

    double qeE = muonGeVFromLen + hadQEGeV;
    sliceEnergy.SetTrkQEE(qeE);

    double nonqeE = muonGeVFromLen + hadNonQEGeV;
    sliceEnergy.SetTrkNonQEE(nonqeE);

    double ccE = muonGeVFromLen + hadCCGeV;
    sliceEnergy.SetTrkCCE(ccE);

    sliceEnergy.SetE(ccE); //For now, E is just set to be identical to Trk CC E

    sliceEnergy.SetHadCluster(vertexHits);

    return sliceEnergy;

  }//End of FDEnergy function

  //----------------------------------------------------------------------
  rb::Energy NumuEAlg::QEFormulaEnergy(std::vector< art::Ptr<rb::Track> > const sliceTracks,
				       art::PtrVector< rb::CellHit >            sliceHits,
				       double                                  &error,
                                       art::Event                        const& e)
  {
    /* This estimates the neutrino energy using the quasi-elastic formula.
       If no 3D track is found, it doesn't return an energy.
       It uses the 3D Kalman track with the best ReMId PID as the "muon", or the
       longest track (as set by fcl, best PID recommended).

       Nuclear binding energy is set to 25 MeV. This is consistant with Minerba's 
       cross-section analysis, Susan's work from Feb. 2012, and what GENIE uses
       in the simulation.

       The NuMI beam direction is determined from a geometry function for this purpose.

       Future: one could consider returning a numuE object instead of an Energy as well
       as a separate error value.
    */

    int    bestTrack  = -1;
    double trackLen   = 0.0;

    MuonType trackType;

    if (fPIDMethod) bestTrack = remid::HighestPIDTrack(sliceTracks, fPIDModuleLabel, e);
    else bestTrack = LongestTrack(sliceTracks, trackLen);

    if(bestTrack == -1) return 0;
    if(bestTrack == 999) return 0;

    if (fPIDMethod) trackLen = sliceTracks[bestTrack]->TotalLength();

    art::PtrVector< rb::CellHit > trackHits;

    TVector3 mudir = sliceTracks[bestTrack]->Dir();  // get direction of muon
    if(mudir.Mag()!=1) mudir = mudir.Unit();         // double check is proper unit vector

    double costheta = mudir * fNumiDir;               // get cosine of angle

    double muE      = 0.0;                            // muon energy (GeV)
    if (fFD){
      muE = MuonEFromTrackLength(trackLen);
      trackType = kFDMuon;
    }
    if (fND){muE = NDMuonEFromTrackLength(sliceTracks[bestTrack], &trackType);}

    double Mp       = .938272;                       // proton mass (GeV)
    double Mu       = .105658;                       // muon mass (GeV)
    double Mb       = 0.025;                         // nuclear binding energy (GeV)
    double Mn       = 0.939565;                      // neutron mass
    double Mnp      = Mn - Mb;                       // neutron mass - carbon nuclear binding energy (GeV)
    double pmu_2    = muE*muE - (Mu*Mu);             // Muon momentum squared

    double recoNu = (2*Mnp*muE-(Mnp*Mnp+Mu*Mu-(Mp*Mp))) / (2*(Mnp-muE+sqrt(pmu_2)*costheta));  // QE Formula

    // calculate the error on this energy estimation
    double errMb       = 0.005;// Uncertainty in nuclear binding energy (GeV)

    double resE        = 0.0;//These resolutions are the RMS of (Reco-True)/True Muon Energy plots - should  be updated
    if      (trackType == kFDMuon)       {resE = 0.043;}
    else if (trackType == kNDActiveMuon) {resE = 0.068;}
    else if (trackType == kNDCatcherMuon){resE = 0.102;}
    else if (trackType == kNDActCatMuon) {resE = 0.057;}
    else if (trackType == kNDPiddleMuon) {resE = 0.084;} 
    else                                 {error  = -1;}

    double errE        = resE*muE;

    double rescostheta = 0.05; // resolution of track direction

    // variance of nuclear binding energy, track length, direction
    double varMb       = errMb*errMb;
    double varE        = errE*errE;
    double varcostheta = (rescostheta*costheta)*(rescostheta*costheta);

    double x    = 2*Mnp*muE - (Mnp*Mnp + Mu*Mu - (Mp*Mp));
    double varx = 4*(Mn*Mn+Mb*Mb)*varE + 4*(Mn*Mn+muE*muE +0.5*Mb*Mb)*varMb;
    double vary = 4*varMb + 4*varE + 4*((muE*muE*costheta*costheta*varE)/(2*pmu_2) + pmu_2*varcostheta);

    double sigmaE_2 = recoNu*recoNu*((varx+recoNu*recoNu*vary)/(x*x));
    error = sqrt(sigmaE_2);

    if (std::isnan(error))  error  = -1;
    if (std::isnan(recoNu)) recoNu = -1;

    rb::Energy sliceEnergy(recoNu);
    return sliceEnergy;

  }//End of QEFormulaEnergy function


  //----------------------------------------------------------------------
  NumuE NumuEAlg::NDEnergy(std::vector< art::Ptr<rb::Track> > const sliceTracks,
                           art::PtrVector< rb::CellHit >            sliceHits,
			   art::Ptr< rb::Cluster>                   sliceCluster,
			   art::Event                        const& e,
			   murem::TrackCleanUpAlg*                  trkCleanUpAlg)
  {
    /* This returns multiple neutrino energy estimates, some of which are better than others.
       They are: CalCC, TrkQEE, TrkNonQEE and TrkCCE. The default E value is set to be indentically
       equal to TrkCCE and comes with the same caveats. At the same time, we are also 
       setting HadCalE and ND specific information about the distribution of energy relative
       to the muon catcher. If no 3D track is found, no energy estimates are made.
       When using the PIDMethod (which is default), the muon track is defined as the 3D track
       with the highest PID. Otherwise, we simply take the longest 3D track as the muon.

       CalCC is a very simple, first pass at energy estimation for the numu cc group.
       ONE SHOULD ALMOST NEVER USE CalCC! THE OTHER ENERGIES ARE USUALLY BETTER CHOICES!
       The weights and methods used were optimized for a contained population that 
       doesn't distinguish between nonqe and qe events. No background existed in the
       optimized sample. It used a linear fit to find the weights and it performed
       the fit simultaneously. It was optimized using Kalaman tracks. 
       The optimization was done only for neutrinos with true energy between 1 and 3
       GeV and basic activity cuts.
       
       The form of the neutrino energy for CalCC is: 
       Reco Nu Energy = Weight_muon_active * Summed GeV of Muon Track in active region 
                      + Weight_muon_tran * Summed GeV of Muon Track in transition plane
                      + Weight_muon_steel * Summed GeV of Muon Track in steel section
                      + Weight_vertex * Summed GeV of all non-Muon Track hits
                      + Weight_offset

       The weights are listed directly following this comment block.

       The CalCC method is NOT our best energy estimator and exists as an accident of history.

       TrkQE, TrkNonQE and TrkCCE are very similar. The weights and methods used were optimized
       for a contained population that distinguishes by truth between QE and NonQE events.  
       The muon energy is found by a spline fit
       to muon track length and muon true energy. This relies heavily on MuonEFromTrackLength but first
       takes into account the muon track length in the active region vs. the catcher region.
       Then the hadronic energy is defined as the sum of HadCalE and HadTrkE. HadCalE is the same vertex 
       energy as used in CalCC - it is just a sum of the GeV of all hits not on the muon track. HadTrkE 
       is the hadronic energy determined to be on the muon track. This is found by a function in 
       MuonRemove. Summed together, this is the visible hadronic energy. This energy
       is broken down into three regions - the active region, the transition plane, and the muon
       catcher. For now, all fits have been done on a population that has primarily hadronic energy in 
       the active region. However, in order to return an energy for every case, we actually use
       the total hadronic region without regards to where it was located. The slices which have
       hadronic energy in regions other than fully active will have a worse and systematically biased energy 
       resolution and should be used with caution. The system can be updated in the future to deal with these
       cases more carefully. The summed GeV of the hadronic energy is transformed into our reco hadronic energy
       by a spline fit. This fit is different for the NonQE, QE and combined populations. It is a fit between
       the summed hadronic energy and (True Neutrino Energy - Reco Muon Energy). More information can
       be found in the functions NDHadEQE, NDHadENonQE and NDHadECC. Finally, the neutrino energy is just as 
       sum of the Muon Energy and the reco Hadronic Energy.

       The transition plane is the plane that divides the active region from the beginning of 
       the muon catcher. It is the very last plane in the active region - after that comes 
       steel and planes which are fully encased in the muon catcher.

       Currently, all the described energy methods are only really meant for contained events; however, this 
       function and the module that calls it doesn't not attempt to determine containment. Please 
       use with discretion!
    */

    double muonWeight_act   = 0.900903; //For CalCC Energy - Susan's fit from Collaboration Meeting (docdb 7738) on 7/30/2012
    double muonWeight_tran  = 3.84225;
    double muonWeight_steel = 6.58836;
    double vertWeight       = 1.61335;
    double offsetWeight     = 0.985003;

    NumuE  sliceEnergy;

    int    bestTrack  = -1;
    double trackLen   = 0.0;
    
    //Need to have valid locations of transition plane!
    if ((fTranPlane < 0)||(fTranZPos < 0)) return sliceEnergy;

    if (fPIDMethod) bestTrack = remid::HighestPIDTrack(sliceTracks, fPIDModuleLabel, e);
    else bestTrack = LongestTrack(sliceTracks, trackLen);

    if(bestTrack == -1) return sliceEnergy;
    if(bestTrack == 999) return sliceEnergy;

    if (fPIDMethod) trackLen = sliceTracks[bestTrack]->TotalLength();

    //Obtaining values of track length / energy in different regions - written to CAF and can be
    //used to reoptimize the energy later

    double recoTrkGeV_act   = 0.0;
    double recoTrkGeV_tran  = 0.0;
    double recoTrkGeV_steel = 0.0;
    double actTrkLen        = 0.0;
    double catTrkLen        = 0.0;
    double tranX            = 999.0;
    double tranY            = 999.0;
    double hadCalGeV_act    = 0.0;
    double hadCalGeV_tran   = 0.0;
    double hadCalGeV_steel  = 0.0;
    double hadTrkGeV_act    = 0.0;
    double hadTrkGeV_tran   = 0.0;
    double hadTrkGeV_steel  = 0.0;
    double vertexEnergy     = 0.0;
    double vertexTrkEnergy  = 0.0;
    
    //Need to sum the muon track energy in three different regions
    NDTrackEnergySplitting(sliceTracks[bestTrack],&recoTrkGeV_act,&recoTrkGeV_tran,&recoTrkGeV_steel);
  
    //Need to determine how much track length is in each region
    NDTrackLength(sliceTracks[bestTrack],&actTrkLen,&catTrkLen,&tranX,&tranY);
  
    art::PtrVector< rb::CellHit > trackHits;
    trackHits = sliceTracks[bestTrack]->AllCells();
    
    //Getting all the hits in the slice that are not on our best track
    rb::Cluster vertexHits = MakeVertexCluster(sliceHits,trackHits);
  
    //Gets GeV of hadronic hits NOT on track  
    if (vertexHits.NCell()>0){
      vertexEnergy = vertexHits.TotalGeV();
    }
  
    //Need to sum the hadron energy in three different regions
    for (unsigned int hitHad = 0; hitHad <vertexHits.NCell(); ++hitHad){
      if (vertexHits.RecoHit(hitHad).IsCalibrated()){
        if (vertexHits.Cell(hitHad)->Plane() < fTranPlane){
          hadCalGeV_act += vertexHits.RecoHit(hitHad).GeV(); 
        }
        else if (vertexHits.Cell(hitHad)->Plane() == fTranPlane){
          hadCalGeV_tran += vertexHits.RecoHit(hitHad).GeV(); 
        }
        else {
          hadCalGeV_steel += vertexHits.RecoHit(hitHad).GeV(); 
        }
      }//End of loop over calibrated nd hadron hits
    }//End of loop over nd hadron hits

    //Need hadronic contamination of muon track
    std::map<int,float> extraHadE = trkCleanUpAlg->ExtraEOnTrackPlanes(*sliceTracks[bestTrack],*sliceCluster);
 
    for (std::map<int,float>::iterator it = extraHadE.begin(); it!=extraHadE.end(); ++it){
      if (it->first < fTranPlane){
	hadTrkGeV_act += it->second; 
      }
      else if (it->first == fTranPlane){
	hadTrkGeV_tran += it->second; 
      }
      else {
	hadTrkGeV_steel += it->second; 
      }
    }//End of loop over extra hadronic energy

    vertexTrkEnergy = hadTrkGeV_act + hadTrkGeV_tran + hadTrkGeV_steel;
    
    //Calculates GeV of muon from track length in regions of ND
    double muonGeVFromLen = NDMuonEFromTrackLength(actTrkLen,catTrkLen,recoTrkGeV_act,recoTrkGeV_tran,recoTrkGeV_steel);

    //Gets adjusted GeV of total hadronic energy assuming QE/NonQE
    double totHadE     = vertexEnergy + vertexTrkEnergy;
    double hadQEGeV    = NDHadEQE(totHadE);
    double hadNonQEGeV = NDHadENonQE(totHadE);
    double hadCCGeV = NDHadECC(totHadE);

    //***************Fill values in our slice energy object

    double calcc = muonWeight_act*recoTrkGeV_act + muonWeight_tran*recoTrkGeV_tran + muonWeight_steel*recoTrkGeV_steel
                    + vertWeight*vertexEnergy + offsetWeight; 
    sliceEnergy.SetCalCCE(calcc);

    double qeE = muonGeVFromLen + hadQEGeV;
    sliceEnergy.SetTrkQEE(qeE);

    double nonqeE = muonGeVFromLen + hadNonQEGeV;
    sliceEnergy.SetTrkNonQEE(nonqeE);

    double ccE = muonGeVFromLen + hadCCGeV;
    sliceEnergy.SetTrkCCE(ccE);

    sliceEnergy.SetE(ccE); //For now, E is just set to be identical to Trk CC E

    sliceEnergy.SetNDTrkLenAct(actTrkLen);
    sliceEnergy.SetNDTrkLenCat(catTrkLen);

    sliceEnergy.SetNDTrkCalAct(recoTrkGeV_act);
    sliceEnergy.SetNDTrkCalTran(recoTrkGeV_tran);
    sliceEnergy.SetNDTrkCalCat(recoTrkGeV_steel);

    sliceEnergy.SetHadCalE(vertexEnergy);
    sliceEnergy.SetHadTrkE(vertexTrkEnergy);

    sliceEnergy.SetNDHadCalAct(hadCalGeV_act);
    sliceEnergy.SetNDHadCalTran(hadCalGeV_tran);
    sliceEnergy.SetNDHadCalCat(hadCalGeV_steel);
    sliceEnergy.SetNDHadTrkAct(hadTrkGeV_act);
    sliceEnergy.SetNDHadTrkTran(hadTrkGeV_tran);
    sliceEnergy.SetNDHadTrkCat(hadTrkGeV_steel);

    sliceEnergy.SetNDTrkTranX(tranX);
    sliceEnergy.SetNDTrkTranY(tranY);
    
    sliceEnergy.SetHadCluster(vertexHits);

    return sliceEnergy;

  }//End of SimpleNDEnergy function

  //-------------------------------------------------------------------

  rb::Cluster NumuEAlg::MakeVertexCluster(art::PtrVector< rb::CellHit > const& sliceHits,
                                          art::PtrVector< rb::CellHit > const& trackHits) const
    
  {
    rb::Cluster vertexCluster;
    
    for (unsigned int sliceHit = 0; sliceHit < sliceHits.size(); ++sliceHit){
      
      bool match = 0;
      
      for (unsigned int trackHit = 0; trackHit < trackHits.size(); ++trackHit){
        match = (sliceHits[sliceHit] == trackHits[trackHit]);
        if (match) break;
      }//End of loop over hits in track   
      
      if (!match) vertexCluster.Add(sliceHits[sliceHit]);
      
    }//End of loop over hits in the slice
    
    return vertexCluster;
    
  }//End of MakeVertexCluster

  //-------------------------------------------------------------------
  int NumuEAlg::LongestTrack(std::vector< art::Ptr<rb::Track> > const& sliceTracks,
                             double&                                   longestTrackLength)  const
  {
    int bestTrack = -1;

    for(size_t c = 0; c < sliceTracks.size(); ++c){
      
      geo::View_t view = sliceTracks[c]->View();
      
      //Only use 3D tracks
      if (view != geo::kXorY){  
        continue;
      } //End of loop over 2d tracks
      
      if ((sliceTracks[c]->TotalLength() > longestTrackLength)&&(sliceTracks[c]->TotalLength() < fMaxTrkLen)) {
        longestTrackLength  = sliceTracks[c]->TotalLength();
        bestTrack = c;
      }
      
    }//End of for loop over tracks

    return bestTrack;
    
  }//End of longest track

  //-------------------------------------------------------------------
  void NumuEAlg::NDTrackLength(art::Ptr<rb::Track>       const& track,
                               double*                          actLength,
                               double*                          catLength,
                               double*                          transitionX,
                               double*                          transitionY)  const
  {
    if(track->NTrajectoryPoints() == 1) {
      std::cout<<"You have a problem! Your track only has one trajectory point!"<<std::endl;
      return;
    }

    bool foundActRegion = false;
    bool foundCatRegion = false;

    double activeLen  = 0.0;
    double catcherLen = 0.0;

    TVector3 actTraj;
    TVector3 catTraj;

    double tranX = 999.0;
    double tranY = 999.0;

    const int N = track->NTrajectoryPoints();

    for(int n = 0; n < N-1; ++n){
      
      if (track->TrajectoryPoint(n+1).Z() <= fTranZPos){
        foundActRegion = true;
        activeLen += (track->TrajectoryPoint(n+1)-track->TrajectoryPoint(n)).Mag();
      }
      if ((track->TrajectoryPoint(n+1).Z() > fTranZPos)&&!foundCatRegion){
        foundCatRegion = true;
        if (track->TrajectoryPoint(n).Z() <= fTranZPos){
          foundActRegion = true;
          actTraj = track->TrajectoryPoint(n);
          catTraj = track->TrajectoryPoint(n+1);
        }
        else {
          catcherLen += (track->TrajectoryPoint(n+1)-track->TrajectoryPoint(n)).Mag();
        }
        continue;
      }
      if ((track->TrajectoryPoint(n+1).Z() > fTranZPos)&&foundCatRegion){
        catcherLen += (track->TrajectoryPoint(n+1)-track->TrajectoryPoint(n)).Mag();
      }
    }

    if (foundActRegion && foundCatRegion){
      TVector3 diff = catTraj-actTraj;
      double addActLen = (fTranZPos-actTraj.Z())*diff.Mag()/diff.Z();
      double addCatLen = (catTraj.Z()-fTranZPos)*diff.Mag()/diff.Z();
      activeLen  += addActLen;
      catcherLen += addCatLen;
      
      //Finding x, y location when crosses transition plane
      tranX = actTraj.X() + (fTranZPos - actTraj.Z())*diff.X()/diff.Z();
      tranY = actTraj.Y() + (fTranZPos - actTraj.Z())*diff.Y()/diff.Z();
    }

    if (actLength) *actLength = activeLen;
    if (catLength) *catLength = catcherLen;
    if (transitionX) *transitionX = tranX;
    if (transitionY) *transitionY = tranY;

  }//End of NDTrackLength

  //-------------------------------------------------------------------
  void NumuEAlg::NDTrackEnergySplitting(art::Ptr<rb::Track>       const& track,
                                        double*                          trkCalAct,
                                        double*                          trkCalTran,
			                double*                          trkCalCat) const
  {
    double trkEnergyInActive  = 0.0;
    double trkEnergyInTran    = 0.0;
    double trkEnergyInCatcher = 0.0;

    //Need to sum the muon track energy in three different regions
    for (unsigned int hitIdx = 0; hitIdx <track->NCell(); ++hitIdx){
      if (track->RecoHit(hitIdx).IsCalibrated()){
        if (track->Cell(hitIdx)->Plane() < fTranPlane){
          trkEnergyInActive += track->RecoHit(hitIdx).GeV(); 
        }
        else if (track->Cell(hitIdx)->Plane() == fTranPlane){
          trkEnergyInTran += track->RecoHit(hitIdx).GeV(); 
        }
        else {
          trkEnergyInCatcher += track->RecoHit(hitIdx).GeV(); 
        }
      }//End of loop over calibrated nd track hits
    }//End of loop over nd track hits

    if (trkCalAct)  *trkCalAct  = trkEnergyInActive;
    if (trkCalTran) *trkCalTran = trkEnergyInTran;
    if (trkCalCat)  *trkCalCat  = trkEnergyInCatcher;

  }//End of NDTrackEnergySplitting

  //-------------------------------------------------------------------
  double NumuEAlg::MuonEFromTrackLength(double trkLen)  const
  {

    //This function returns muon energy, in GeV, given the muon track length in cm.
    //It uses a spline fit from May 2015.
    
    double muonE = 0.0;

    if (trkLen <= 0.0) return muonE;

    if (trkLen < FDMstitch1){
      muonE = FDMslope1*trkLen + FDMoffset;
    }
    else if (trkLen < FDMstitch2){
      muonE = FDMslope2*trkLen + ((FDMslope1-FDMslope2)*FDMstitch1 + FDMoffset);
    }
    else if (trkLen < FDMstitch3){
      muonE = FDMslope3*trkLen + ((FDMslope2-FDMslope3)*FDMstitch2 + (FDMslope1-FDMslope2)*FDMstitch1 + FDMoffset);
    }
    else {
      muonE = FDMslope4*trkLen + ((FDMslope3-FDMslope4)*FDMstitch3 + (FDMslope2-FDMslope3)*FDMstitch2 
				   +(FDMslope1-FDMslope2)*FDMstitch1 + FDMoffset);
    }
    
    return muonE;

  }//End of MuonEFromTrackLength
  //-------------------------------------------------------------------
  double NumuEAlg::ActiveTrackLengthFromMuonE(double muonE) const
  {
  
    //This function returns effective active-region track length, in cm, given the muon energy in GeV.
    //It reverses a spline fit from May 2015.  
  
    double trkLen = 0.0;

    if (muonE <= 0.0) return trkLen;
    
    if (muonE <= FDMoffset ) return trkLen;
    
    else if (muonE < (FDMslope1*FDMstitch1 + FDMoffset)){
      trkLen = (muonE - FDMoffset)/FDMslope1;
    }
    else if (muonE < (FDMslope2*FDMstitch2 + ((FDMslope1-FDMslope2)*FDMstitch1 + FDMoffset))){
      trkLen = (muonE - ((FDMslope1-FDMslope2)*FDMstitch1 + FDMoffset))/FDMslope2;
    }
    else if (muonE < (FDMslope3*FDMstitch3 + ((FDMslope2-FDMslope3)*FDMstitch2 + (FDMslope1-FDMslope2)*FDMstitch1 + FDMoffset))){
      trkLen = (muonE - ((FDMslope2-FDMslope3)*FDMstitch2 +(FDMslope1-FDMslope2)*FDMstitch1 + FDMoffset))/FDMslope3; 
    }
    else {
      trkLen = (muonE - ((FDMslope3-FDMslope4)*FDMstitch3 + (FDMslope2-FDMslope3)*FDMstitch2 + (FDMslope1-FDMslope2)*FDMstitch1 + FDMoffset))/FDMslope4; 
    }
    
    return trkLen;
    
  }//End of ActiveTrackLengthFromMuonE
  //-------------------------------------------------------------------
  double NumuEAlg:: NDMuonEFromTrackLength(double actTrkLen, double catTrkLen, double trkCalAct, double trkCalTran, double trkCalCat, 
					   MuonType* muonType)   const 
  {

    //This function returns muon energy, in GeV, given the muon track length in cm.
    //It distinguishes between the four types of muon tracks and calls the 
    //correct energy estimator for each. This fitting work was done in May 2015.

    double muonE = 0.0;

    if ((catTrkLen <= 0.0) && (actTrkLen <= 0.0)) return muonE;

    if (((trkCalAct > 0.0)||(trkCalTran > 0.0))&&(trkCalCat == 0.0)){
      muonE = NDMuonEInActiveOnly(actTrkLen);
      if (muonType){*muonType = kNDActiveMuon;}
    }//Entirely contained in active region and transition plane

    if ((trkCalAct == 0.0)&&(trkCalTran == 0.0)&&(trkCalCat > 0.0)){
      muonE = NDMuonEInCatcherOnly(catTrkLen);
      if (muonType){*muonType = kNDCatcherMuon;}
    }//Entirely contained in catcher region

    if (((trkCalAct > 0.0)||(trkCalTran > 0.0))&&(trkCalCat > 0.0)){
      muonE = NDMuonEInActiveAndCatcher(actTrkLen,catTrkLen);
      if (muonType){*muonType = kNDActCatMuon;}
    }//Length in active and catcher region
 
    return muonE;
    
  }//End of NDMuonEFromTrackLength
  //-------------------------------------------------------------------
  double NumuEAlg:: NDMuonEFromTrackLength(art::Ptr<rb::Track>       const& track,
					   MuonType*                        muonType)   const 
  {

    //This function returns muon energy, in GeV, given the muon track length in cm.
    //It distinguishes between the four types of muon tracks and calls the 
    //correct energy estimator for each. This fitting work was done in May 2015.

    double muonE      = 0.0;

    double actTrkLen  = 0.0;
    double catTrkLen  = 0.0;
    double trkCalAct  = 0.0;
    double trkCalTran = 0.0;
    double trkCalCat  = 0.0;

    NDTrackLength(track,&actTrkLen,&catTrkLen);
    NDTrackEnergySplitting(track,&trkCalAct,&trkCalTran,&trkCalCat);
    muonE = NDMuonEFromTrackLength(actTrkLen, catTrkLen, trkCalAct, trkCalTran, trkCalCat, muonType);

    return muonE;
    
  }//End of NDMuonEFromTrackLength
  //-------------------------------------------------------------------
  double NumuEAlg:: NDMuonEInActiveOnly(double actTrkLen)   const 
  {

    //This function returns muon energy, in GeV, given the muon track length in cm.
    //It is for ND muons that are entirely contained in the active region of the 
    //detector. It uses the same fit used for FD muons, since these are also
    //entirely contained in the active material.

    double muonE = 0.0;

    if (actTrkLen <= 0.0) return muonE;

    muonE = MuonEFromTrackLength(actTrkLen);
    
    return muonE;
    
  }//End of NDMuonEInActiveOnly
  //-------------------------------------------------------------------
  double NumuEAlg:: NDMuonEInCatcherOnly(double catTrkLen)   const 
  {

    //This function returns muon energy, in GeV, given the muon track length in cm.
    //It is for ND muons that are entirely contained in the muon catcher region of the 
    //detector. It is a linear fit since the catcher removes our ability to see
    //fine structure like we can in the active region. This fit is from July 2013.
    //It could use a new tune, but for FA, none of these muons were used so the tuning 
    //wasn't done.

    double muonE = 0.0;

    if (catTrkLen <= 0.0) return muonE;

    double slope  = 0.00513341; // GeV/cm
    double offset = 0.230294;   // GeV

    muonE = slope*catTrkLen + offset;
    
    return muonE;
    
  }//End of NDMuonEInCatcherOnly
  //-------------------------------------------------------------------
  double NumuEAlg:: NDMuonEInActiveAndCatcher(double actTrkLen, double catTrkLen)   const 
  {

    //This function returns muon energy, in GeV, given the muon track length in cm.
    //It is for ND muons that have energy in both the active and catcher region.
    //It uses a linear fit to determine the muon energy when it enters the muon catcher.
    //This fit is from May 2015..
    //This is then converted into an effective active track length, which is added
    //to the existing active track length for a total active track length. 
    //This length is then fed into the FD muon fit.

    double muonE = 0.0;

    if ((catTrkLen <= 0.0) &&  (actTrkLen <= 0.0)) return muonE;

    double offset  = 0.152;    // GeV
    double slope   = 0.00536;  // GeV/cm

    double muonECat = slope*catTrkLen + offset;
    double catcherActLen = ActiveTrackLengthFromMuonE(muonECat);
    double totalActLen = catcherActLen + actTrkLen;

    muonE = MuonEFromTrackLength(totalActLen);
    
    return muonE;
    
  }//End of NDMuonEInActiveAndCatcher
  //-------------------------------------------------------------------
  std::vector<rb::Energy> NumuEAlg::MuonEnergies(std::vector< art::Ptr<rb::Track> > const sliceTracks) const
  {
  
    //This function returns energy for each track, given the assumption
    //that the track is a muon.

    std::vector<rb::Energy> energiesOfTracks;
   
    for(size_t c = 0; c < sliceTracks.size(); ++c){

      rb::Energy trackEnergy(-1.0);
     
      geo::View_t view = sliceTracks[c]->View();
      
      //Only give 3D tracks valid energies
      if (view != geo::kXorY){  
	energiesOfTracks.push_back(trackEnergy);
	continue;
      } //End of loop over 2d tracks
  
      double energy = -1.0;

      if (fFD){
	double trkLen = sliceTracks[c]->TotalLength();
	energy = MuonEFromTrackLength(trkLen);
	trackEnergy.SetE(energy);
	energiesOfTracks.push_back(trackEnergy);
      }//End of FD track

      if (fND){
	energy = NDMuonEFromTrackLength(sliceTracks[c]);
	trackEnergy.SetE(energy);
	energiesOfTracks.push_back(trackEnergy);
      }//End of ND track      
   
    }//End of loop over tracks

    return energiesOfTracks;

  }//End of MuonEnergies
  //-------------------------------------------------------------------
  int NumuEAlg:: NDTransitionPlane()   const 
  {

    //Function that returns the plane number of the transition plane - 
    //the last active plane before the steel of the muon catcher.
    
    return fTranPlane;
    
  }//End of NDTransitionPlane
  //-------------------------------------------------------------------
  double NumuEAlg:: NDTransitionZPos()   const 
  {

    //Function that returns the z position of the middle of the transition plane - 
    //the last active plane before the steel of the muon catcher.
    
    return fTranZPos;
    
  }//End of NDTransitionZPos
  //-------------------------------------------------------------------
  double NumuEAlg:: HadEQE(double calE)   const 
  {

    //This function returns weighted hadron energy, in GeV, given the sum of energy not on 
    //the muon track and energy from hadron on muon track.
    //It uses a spline fit from May 2015. It was fit for a contained QE-only reco population.

    double hadE = 0.0;

    if (calE < 0.0) return hadE;
  
    double slope1  = 0.621;     // Unitless
    double slope2  = 1.47;      // Unitless
    double slope3  = 1.81;      // Unitless
    double slope4  = 2.06;      // Unitless
    double offset  = 0.0515;    // GeV
    double stitch1 = 0.0597;    // GeV
    double stitch2 = 0.139;     // GeV
    double stitch3 = 0.800;     // GeV

    if (calE < stitch1){
      hadE = slope1*calE + offset;
    }
    else if (calE < stitch2){
      hadE = slope2*calE + ((slope1-slope2)*stitch1 + offset);
    }
    else if (calE < stitch3){
      hadE = slope3*calE + ((slope2-slope3)*stitch2 + (slope1-slope2)*stitch1 + offset);
    }
    else {
      hadE = slope4*calE + ((slope3-slope4)*stitch3 + (slope2-slope3)*stitch2 
			    +(slope1-slope2)*stitch1 + offset);
    }
    
    return hadE;
    
  }//End of HadEQE
  //-------------------------------------------------------------------
  double NumuEAlg:: HadENonQE(double calE)   const
  {

    //This function returns weighted hadron energy, in GeV, given the sum of energy not on 
    //the muon track and energy from hadron on muon track.
    //It uses a spline fit from May 2015. It was fit for a contained NonQE-only reco population.

    double hadE = 0.0;
    
    if (calE < 0.0) return hadE;

    double slope1  = 0.94;      // Unitless
    double slope2  = 1.21;      // Unitless
    double slope3  = 1.78;      // Unitless
    double slope4  = 2.20;      // Unitless
    double offset  = 0.241;     // GeV
    double stitch1 = 0.132;     // GeV
    double stitch2 = 0.223;     // GeV
    double stitch3 = 0.964;     // GeV
    
    if (calE < stitch1){
      hadE = slope1*calE + offset;
    }
    else if (calE < stitch2){
      hadE = slope2*calE + ((slope1-slope2)*stitch1 + offset);
    }
    else if (calE < stitch3){
      hadE = slope3*calE + ((slope2-slope3)*stitch2 + (slope1-slope2)*stitch1 + offset);
    }
    else {
      hadE = slope4*calE + ((slope3-slope4)*stitch3 + (slope2-slope3)*stitch2 
			    +(slope1-slope2)*stitch1 + offset);
    }
    
    return hadE;
    
  }//End of HadENonQE
  //-------------------------------------------------------------------
  double NumuEAlg:: HadECC(double calE)   const 
  {

    //This function returns weighted hadron energy, in GeV, given the sum of energy not on 
    //the muon track and energy from hadron on muon track.
    //It uses a spline fit from May 2015. It was fit for a contained reco population that includes 
    //both QE and nonQE events.

    double hadE = 0.0;

    if (calE < 0.0) return hadE;

    double slope1  = 0.525;     // Unitless
    double slope2  = 1.36;      // Unitless
    double slope3  = 2.88;      // Unitless
    double slope4  = 1.90;      // Unitless 
    double offset  = 0.0535;    // GeV
    double stitch1 = 0.0500;    // GeV
    double stitch2 = 0.100;     // GeV
    double stitch3 = 0.161;     // GeV
    
    if (calE < stitch1){
      hadE = slope1*calE + offset;
    }
    else if (calE < stitch2){
      hadE = slope2*calE + ((slope1-slope2)*stitch1 + offset);
    }
    else if (calE < stitch3){
      hadE = slope3*calE + ((slope2-slope3)*stitch2 + (slope1-slope2)*stitch1 + offset);
    }
    else {
      hadE = slope4*calE + ((slope3-slope4)*stitch3 + (slope2-slope3)*stitch2 
			    +(slope1-slope2)*stitch1 + offset);
    }
    
    return hadE;
    
  }//End of HadECC
  //-------------------------------------------------------------------
  double NumuEAlg:: NDHadEQE(double calE)   const 
  {

    //This function returns weighted hadron energy, in GeV, given the sum of energy not on 
    //the muon track and energy from hadron on muon track.
    //It uses a spline fit from May 2015. It was fit for a contained QE-only reco population.
    //The population used for the fit had ALMOST NO hadronic activity in the transition plane or muon catcher!

    double hadE = 0.0;

    if (calE < 0.0) return hadE;
    
    double slope1  = 1.567;    // Unitless
    double slope2  = 1.822;    // Unitless
    double offset  = 0.0450;   // GeV
    double stitch1 = 0.0820;   // GeV
    
    if (calE < stitch1){
      hadE = slope1*calE + offset;
    }
    else {
      hadE = slope2*calE + ((slope1-slope2)*stitch1 + offset);
    }
    
    return hadE;
    
  }//End of NDHadEQE
  //-------------------------------------------------------------------
  double NumuEAlg:: NDHadENonQE(double calE)   const 
  {

    //This function returns weighted hadron energy, in GeV, given the sum of energy not on 
    //the muon track and energy from hadron on muon track.
    //It uses a spline fit from May 2015. It was fit for a contained nonQE-only reco population.
    //The population used for the fit had ALMOST NO hadronic activity in the transition plane or muon catcher!

    double hadE = 0.0;

    if (calE < 0.0) return hadE;
    
    double slope1  = 1.080;    // Unitless
    double slope2  = 1.889;    // Unitless
    double offset  = 0.254;    // GeV
    double stitch1 = 0.169;    // GeV
    
    if (calE < stitch1){
      hadE = slope1*calE + offset;
    }
    else {
      hadE = slope2*calE + ((slope1-slope2)*stitch1 + offset);
    }

    
    return hadE;
    
  }//End of NDHadENonQE
  //-------------------------------------------------------------------
  double NumuEAlg:: NDHadECC(double calE)   const 
  {

    //This function returns weighted hadron energy, in GeV, given the sum of energy not on 
    //the muon track and energy from hadron on muon track.
    //It uses a spline fit from May 2015. It was fit for a reco population that includes 
    //QE and nonQE events.
    //The fit does not try to account for large amounts of hadronic activity in the muon catcher.

    double hadE = 0.0;

    if (calE < 0.0) return hadE;

    double slope1  = 0.64;    // Unitless
    double slope2  = 2.072;    // Unitless
    double offset  = 0.061;    // GeV
    double stitch1 = 0.015;    // GeV
    
    
    if (calE < stitch1){
      hadE = slope1*calE + offset;
    }
    else {
      hadE = slope2*calE + ((slope1-slope2)*stitch1 + offset);
    }
    
    return hadE;
    
  }//End of NDHadECC
  //-------------------------------------------------------------------
  double NumuEAlg::GetUCE(art::FindManyP<cosrej::CosRejObj> CRO, size_t slicenum, float totalgev, int nhitslice)
  {

    std::vector< art::Ptr<cosrej::CosRejObj> > cosmicrej = CRO.at(slicenum);
    Evars[0] = totalgev;
    Evars[1] = cosmicrej[0]->AngleKal();
    Evars[2] = cosmicrej[0]->Scatt();
    Evars[3] = cosmicrej[0]->ERatio();
    Evars[4] = cosmicrej[0]->KalFwdCell();
    Evars[5] = cosmicrej[0]->CosHitRatio()*1.0*nhitslice;
    Evars[6] = cosmicrej[0]->NTracks3D();
    return fReaderE->EvaluateRegression("BDTG")[0];

  }//End of GetUCE
  //-------------------------------------------------------------------
  double NumuEAlg::GetFSE(art::FindManyP<cosrej::CosRejObj> CRO, size_t slicenum)
  {
    std::vector< art::Ptr<cosrej::CosRejObj> > cosmicrej = CRO.at(slicenum);
    double pars[4];
    pars[0] = cosmicrej[0]->FScattExt();
    pars[1] = cosmicrej[0]->FScattSig();
    pars[2] = cosmicrej[0]->FScattMax();
    pars[3] = cosmicrej[0]->FScattSum();
    ScatterEnergy *nnEnergy = new ScatterEnergy();
    double scatterMuEnergy = nnEnergy->Value(0, pars);
    delete nnEnergy;
    return scatterMuEnergy;
  }//End of GetFSE
} // End of namespace
