////////////////////////////////////////////////////////////////////////
// Class:       NumuEnergy
// Module Type: producer
// File:        NumuEnergy_module.cc
//
// Generated at Fri Sep 21 08:53:53 2012 by Susan Lein using artmod
// from art v1_00_11.
////////////////////////////////////////////////////////////////////////

#include <string>
#include "TMath.h"

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Persistency/Common/Assns.h"

// NOvA includes 
#include "NumuEnergy/NumuEAlg.h"
#include "NumuEnergy/NumuE.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "RecoBase/Energy.h"
#include "Utilities/AssociationUtil.h"
#include "Utilities/func/MathUtil.h"
#include "NovaDAQConventions/DAQConventions.h"

namespace numue {

  class NumuEnergy : public art::EDProducer {
  public:
    explicit NumuEnergy(fhicl::ParameterSet const & pset);
    virtual ~NumuEnergy();
    
    void produce(art::Event & e);
    void beginRun(art::Run& run);
    
  private:
    
    NumuEAlg*                fNumuEAlg;                ///< Numu Energy Algorithm Object
    murem::TrackCleanUpAlg*  fTrackCleanUpAlg;         ///< Track Clean Up Algorithm Object
    fhicl::ParameterSet      fNumuEAlgPSet;            ///< Parameter Set to configure the NumuEAlg object
    fhicl::ParameterSet      fTrackCleanUpAlgPSet;     ///< Parameter Set to configure the TrackCleanUpAlg object
    std::string              fSlicerModuleLabel;       ///< Label for module doing slicing
    std::string              fTrackModuleLabel;        ///< Label for module doing tracking
    
    bool                     fND;                      ///< Is detector ND?
    bool                     fFD;                      ///< Is detector FD?
    
  };
  
  //----------------------------------------------------------------------
  NumuEnergy::NumuEnergy(fhicl::ParameterSet const & pset):
    fNumuEAlg(0),
    fTrackCleanUpAlg(0),
    fNumuEAlgPSet(pset.get< fhicl::ParameterSet >("NumuEAlgPSet")   ),
    fTrackCleanUpAlgPSet(pset.get< fhicl::ParameterSet >("TrackCleanUpAlgPSet")   ),
    fSlicerModuleLabel(pset.get< std::string >("SlicerModuleLabel")),
    fTrackModuleLabel(pset.get< std::string >("TrackModuleLabel")),
    fND(false),
    fFD(false)
  { 
    produces< std::vector<NumuE> >();
    produces< art::Assns<NumuE, rb::Cluster> >();

    produces< std::vector<rb::Energy> >();
    produces< art::Assns<rb::Energy, rb::Track> >();
  }
  
  
  //----------------------------------------------------------------------
  NumuEnergy::~NumuEnergy()
  {
    if (fNumuEAlg)        delete fNumuEAlg;
    if (fTrackCleanUpAlg) delete fTrackCleanUpAlg;
  }
  
  //----------------------------------------------------------------------
  void NumuEnergy::beginRun(art::Run& run)
  {
    
    fNumuEAlg        = new NumuEAlg(fNumuEAlgPSet);
    fTrackCleanUpAlg = new murem::TrackCleanUpAlg(fTrackCleanUpAlgPSet);
    
    art::ServiceHandle<geo::Geometry>  geom; 
    
    novadaq::cnv::DetId detID = geom->DetId();
    if (detID == novadaq::cnv::kNEARDET) fND = true;
    if (detID == novadaq::cnv::kFARDET)  fFD = true;

    return;

  }//End of Begin Run

  //----------------------------------------------------------------------
  void NumuEnergy::produce(art::Event& e)
  {
    if (!fND && !fFD) {
      std::cerr << "Nope - this file isn't near or far detector, algorithm doesn't work on any other" << std::endl;
      return;
    }

    //Load the slicer list from the event
    art::Handle<std::vector<rb::Cluster> > slicecol;
    e.getByLabel(fSlicerModuleLabel,slicecol);
    
    if(slicecol->empty()){
      mf::LogWarning ("No Slices")<<"No Slices in the input file";
      return;
    }

    art::PtrVector<rb::Cluster>  slicelist;
    for(unsigned int i = 0; i<slicecol->size();++i){
      art::Ptr<rb::Cluster> clust(slicecol,i);
      slicelist.push_back(clust);
    }
    
    //Getting tracks associated with slices
    art::FindManyP<rb::Track> sliceToTracks(slicecol, e, fTrackModuleLabel);
   
    //Declaring containers for things to be stored in event
    std::unique_ptr< std::vector<NumuE> > energyCol(new std::vector<NumuE>);
    std::unique_ptr< art::Assns<NumuE, rb::Cluster> > assoc(new art::Assns<NumuE, rb::Cluster>);

    std::unique_ptr< std::vector<rb::Energy> > trackECol(new std::vector<rb::Energy>);
    std::unique_ptr< art::Assns<rb::Energy, rb::Track> > trackAssoc(new art::Assns<rb::Energy, rb::Track>);

    art::FindManyP<cosrej::CosRejObj> nCosRej(slicecol, e, "cosrej");

    for(size_t s = 0; s < slicelist.size(); ++s){
  
      bool noiseSlice = slicelist[s]->IsNoise();
      if (noiseSlice) continue;
     
      art::PtrVector< rb::CellHit > sliceHits;
      sliceHits = slicelist[s]->AllCells();

      // Get all of the tracks in the slice
      const std::vector< art::Ptr<rb::Track> > sliceTracks = sliceToTracks.at(s);
      if(sliceTracks.empty()) continue; 
    
      NumuE                     sliceEnergy;
      rb::Energy                angleqeE;
      double                    errAngleQEE = -1.0;
      std::vector<rb::Energy>   trackEnergies;

      if (fFD){
	trackEnergies = fNumuEAlg->MuonEnergies(sliceTracks);
        sliceEnergy   = fNumuEAlg->FDEnergy(sliceTracks,sliceHits,slicelist[s],e,fTrackCleanUpAlg);
        angleqeE      = fNumuEAlg->QEFormulaEnergy(sliceTracks,sliceHits,errAngleQEE,e);
        sliceEnergy.SetAngleQEE(angleqeE.E());
        sliceEnergy.SetAngleQEError(errAngleQEE);
        if(nCosRej.at(s).size() > 0) {
          sliceEnergy.SetUCCCE(fNumuEAlg->GetUCE(nCosRej,s,slicelist[s]->TotalGeV(),slicelist[s]->NCell()));
        }
      }//End of fd energy

      if (fND){
	trackEnergies = fNumuEAlg->MuonEnergies(sliceTracks);
        sliceEnergy   = fNumuEAlg->NDEnergy(sliceTracks,sliceHits,slicelist[s],e,fTrackCleanUpAlg);
	angleqeE      = fNumuEAlg->QEFormulaEnergy(sliceTracks,sliceHits,errAngleQEE,e);
        sliceEnergy.SetAngleQEE(angleqeE.E());
        sliceEnergy.SetAngleQEError(errAngleQEE);
      }//End of nd energy

      if ((sliceEnergy.E() > 0)||(sliceEnergy.CalCCE() > 0)){ 
        energyCol->push_back(sliceEnergy);
        util::CreateAssn(*this,e,*(energyCol.get()),slicelist[s],*(assoc.get()));
      }//End of loop over non-zero energy

      for(size_t c = 0; c < trackEnergies.size(); ++c){
	trackECol->push_back(trackEnergies[c]);
	util::CreateAssn(*this,e,*(trackECol.get()),sliceTracks[c],*(trackAssoc.get()));
      }//End of loop over track energies for the slice
      
    }//End of loop over slices

    //Put objects and associations in file
    e.put(std::move(energyCol));
    e.put(std::move(assoc));
    e.put(std::move(trackECol));
    e.put(std::move(trackAssoc));
    return;
  }//End of Produce

  //----------------------------------------------------------------------
  DEFINE_ART_MODULE(NumuEnergy)

}// End of namespace
