////////////////////////////////////////////////////////////////////////
/// \brief   This slightly less simple photon transport uses a template for photon collection in 
///          time and in distance along the fiber.
///
/// \author  aurisano@fnal.gov
/// \date
////////////////////////////////////////////////////////////////////////
//Framework Includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Persistency/Common/Ptr.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib/search_path.h"
#include "CLHEP/Random/RandExponential.h"
#include "CLHEP/Random/RandPoissonQ.h"
#include "CLHEP/Random/RandGaussQ.h"

//NOvA includes
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "Simulation/FLSHit.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/PhotonSignal.h"
#include "Simulation/Simulation.h"
#include "Utilities/func/MathUtil.h"
#include "Utilities/func/RandHisto.h"

//stdlib Includes
#include <string>
#include <list>
#include <vector>
#include <cmath>

//ROOT Includes
#include "TH1D.h"
#include "TH2D.h"
#include "TProfile.h"
#include "TFile.h"
#include "TMath.h"

//System Includes
#include <sys/stat.h>

///photon transport simulation
namespace photrans
{
  /// a struct to hold collection rate info
  struct RateInfo
  {
    double rate;
    double z;
    double zWidth;
    TH1D   timePDF;
  };

  /// a class for transporting photons in a roughly realistic way
  class ImprovedTransport : public art::EDProducer
  {
  public:
    explicit ImprovedTransport(fhicl::ParameterSet const &pset);
    virtual ~ImprovedTransport();
    
    void beginRun(art::Run& run);
    void produce(art::Event& evt);
    
  protected:
    double FiberTransmission(double x);
    bool   LoadHistos();
    void   StepAlongHit(sim::FLSHit& hit);
    /// \param meanPE   The photons entering the fiber
    /// \param stepTime When they entered
    /// \param dist     How far to the readout
    /// \param lambda   Accumulates total mean expected at readout into here
    void   CreatePhotoElectrons(double meanPE, double stepTime, double dist, RateInfo& info, double& lambda);
    void   ClusterPhotoElectrons(int planeId, int cellId, int trackId);
    
  private:
    //fcl parameters
    std::string         fFilterModuleLabel;      ///< module that filtered the fls hits
    fhicl::ParameterSet fPSetNDOS, fPSetND, fPSetFD;

    double              fPhotonsPerMeV;          ///< blue photon production scale
    double              fXViewFactor;
    double              fYViewFactor;
    double              fQuantumEff;             ///< APD quantum efficiency
    double              fFiberIndexOfRefraction; ///< n for fiber core
    std::vector<double> fAttenPars;              ///< from Leon's fit
    double              fTimeClustering;         ///< groups photons closer than this many ns apart
    bool                fWriteEmptySignals;      ///< Write an empty PhotonSignal if there's 0 PE
    double              fStep;                   ///< step size for walking along the track (cm)
    bool                fApplyBirks;             ///< Use EdepBirks from FLSHits instead of calculating here
    std::string         fCollectionRateFile;     ///< Relative path to the collection rate histo file
    bool                fApplyFiberSmearing;     ///< Turn on fiber smearing
    std::string         fSmearingHistoFile;      ///< Relative path ot the fiber smearing histo file
    int                 msglevel;
    
    //non-fcl members
    double                                fFiberSpeedOfLight;    ///< Speed of light in the fiber
    TH2D                                  fCollectionRateHisto;  ///< Collection rate for given dT vs. dZ
    TH1D                                  fFiberSmearingHisto;   ///< Smearing histo for photons bouncing around in fiber
    art::ServiceHandle<geo::Geometry>     fGeo;                  ///< Handle to the geometry service
    std::list< double > fPETimes;                                ///< List of photo-electron times (each representing one observed PE)
    std::list< RateInfo >                 fRateInfo;             ///< List of RateInfo object
   
    std::unique_ptr<std::vector<sim::PhotonSignal> >   fSignals; ///< Produced vector of PhotonSignals
    CLHEP::RandPoissonQ*    fPoisson;
    CLHEP::RandFlat*        fFlat;
    RandHisto*              fRandHisto;
    CLHEP::RandGaussQ*      fGauss;
    bool                    fFirstTime;
  };

  //............................................................
  ImprovedTransport::ImprovedTransport(fhicl::ParameterSet const &pset)
    : fFilterModuleLabel     (pset.get< std::string>("FilterModuleLabel"))
    , fPSetNDOS              (pset.get<fhicl::ParameterSet>("ndos"))
    , fPSetND                (pset.get<fhicl::ParameterSet>("nd"))
    , fPSetFD                 (pset.get<fhicl::ParameterSet>("fd"))
    , fSignals               (new std::vector<sim::PhotonSignal>)
    , fFirstTime(true)
  {
    // get the random number seed, use a random default if not specified    
    // in the configuration file.  
    unsigned int seed = pset.get< unsigned int >("Seed", sim::GetRandomNumberSeed());
    
    createEngine( seed );

    // get the RandomNumberGenerator service
    art::ServiceHandle<art::RandomNumberGenerator> rng;
    CLHEP::HepRandomEngine &engine = rng->getEngine();
    fPoisson = new CLHEP::RandPoissonQ(engine);
    fFlat = new CLHEP::RandFlat(engine);
    fRandHisto = new RandHisto(engine);
    fGauss = new CLHEP::RandGaussQ(engine);
    
    produces< std::vector<sim::PhotonSignal> >();
  }
  
  //............................................................
  ImprovedTransport::~ImprovedTransport()
  {
    delete fPoisson;
    delete fFlat;
    delete fRandHisto;
    delete fGauss;
  }

  void ImprovedTransport::beginRun(art::Run& /*run*/)
  {
    if (fFirstTime){
      // Figure out which parameter set we're using
      fhicl::ParameterSet pset;
      art::ServiceHandle<geo::Geometry> geom;
      switch(geom->DetId()){
      case novadaq::cnv::kNDOS:
	pset = fPSetNDOS;
	break;
      case novadaq::cnv::kNEARDET:
	pset = fPSetND;
	break;
      case novadaq::cnv::kFARDET:
	pset = fPSetFD;
	break;
      default:
	assert(0 && "Unknown detector");
      }

      fPhotonsPerMeV          = pset.get< double              >("PhotonsPerMeV");
      fXViewFactor            = pset.get< double              >("XViewFactor");
      fYViewFactor            = pset.get< double              >("YViewFactor");
      fQuantumEff             = pset.get< double              >("QuantumEff");
      fFiberIndexOfRefraction = pset.get< double              >("FiberIndexOfRefraction");
      fAttenPars              = pset.get< std::vector<double> >("AttenPars");
      fTimeClustering         = pset.get< double              >("TimeClustering"); 
      fWriteEmptySignals      = pset.get< bool                >("WriteEmptySignals");
      fStep                   = pset.get< double              >("Step");
      fApplyBirks             = pset.get< bool                >("ApplyBirks"); 
      fCollectionRateFile     = pset.get< std::string         >("CollectionRateFile");
      fSmearingHistoFile      = pset.get< std::string         >("SmearingHistoFile");
      fApplyFiberSmearing     = pset.get< bool                >("ApplyFiberSmearing");    
      msglevel                = pset.get< int                 >("MessageLevel");

      // constructor decides if initialized value is a path or an environment variable
      cet::search_path sp("FW_SEARCH_PATH");
      
      const double c = TMath::Ccgs()*1e-9; // c in cm/ns
      fFiberSpeedOfLight = c/fFiberIndexOfRefraction;
      
      sp.find_file(pset.get< std::string >("CollectionRateFile"), fCollectionRateFile);
      // Check the collection rate file
      struct stat sb;
      if ( fCollectionRateFile.empty() || stat(fCollectionRateFile.c_str(), &sb)!=0 ){
	// failed to resolve the file name
	throw cet::exception("NoCollectionRateFile") << "collection rate file " << fCollectionRateFile << " not found!\n" << __FILE__ << ":" << __LINE__ << "\n";
      }
      
      sp.find_file(pset.get< std::string >("SmearingHistoFile"), fSmearingHistoFile);
      if ( fSmearingHistoFile.empty() || stat(fSmearingHistoFile.c_str(), &sb)!=0 ){
	// failed to resolve the file name
	throw cet::exception("NoSmearingHistoFile") << "smearing histo file " << fSmearingHistoFile << " not found!\n" << __FILE__ << ":" << __LINE__ << "\n";
      }

      if ( !LoadHistos() ) 
	{
	  throw cet::exception("NoCollectionRateHisto") 
	  << "can't load histos\n"
          << __FILE__ << ":" << __LINE__ << "\n";
	return;
	}

      //don't reinitialize after the first time through
      fFirstTime = false;
    }
  }


  //............................................................
  void ImprovedTransport::produce(art::Event& evt) 
  {
    // get list of hits
    art::Handle< std::vector<sim::FLSHitList> > flslistHandle;
    evt.getByLabel(fFilterModuleLabel, flslistHandle);
    
    //clear output vector
    fSignals->clear();
    
    for (unsigned int ilist = 0; ilist < flslistHandle->size(); ++ilist)
      {
	const std::vector<sim::FLSHit> hits = flslistHandle->at(ilist).fHits;
	for (unsigned int ihit = 0; ihit < hits.size(); ++ihit)
	  {
	    sim::FLSHit hit = hits[ihit];
	    //step along the hit and fill fPETimes (clustering done at end of StepAlongHit)
	    StepAlongHit( hit );
	  }
      }

    // make a new collection into which we will put the sim::PhotonSignals.  That new collection
    // will then be put into the event.  If we tried to put fSignals in the event, the std::move
    // call changes ownership of the vector and the next time we try to access fSignals would
    // cause a seg fault
    std::unique_ptr<std::vector<sim::PhotonSignal>> pscol(new std::vector<sim::PhotonSignal>);
    pscol->swap(*fSignals);

    //put vector<sim::PhotonSignal> into the event
    evt.put(std::move( pscol ));
    
    return;
  }
  
  //............................................................
  void ImprovedTransport::StepAlongHit( sim::FLSHit& hit )
  {
    //clear photo-electron times for this hit
    fPETimes.clear();
    
    const double MeVPerGeV = 1000.;
    const double green_scale = MeVPerGeV * fPhotonsPerMeV * fQuantumEff;
    
    //for cleaner-looking code below...
    double Xa       = hit.GetEntryX();
    double Ya       = hit.GetEntryY();
    double Za       = hit.GetEntryZ();
    double Ta       = hit.GetEntryT();

    double Xb       = hit.GetExitX();
    double Yb       = hit.GetExitY();
    double Zb       = hit.GetExitZ();
    double Tb       = hit.GetExitT();
    
    //if ( (std::isinf(Ta) || std::isinf(Tb) ) ) 
    //{
    //mf::LogInfo("InfiniteTime") << "Found infinite time";
    //throw cet::exception("InfiniteTime") << "Found an infinite time" << "\n";
    //}
    
    int   planeId   = hit.GetPlaneID();
    int   cellId    = hit.GetCellID();
    int   trackId   = hit.GetTrackID();
    
    double seglength = util::pythag(Xb - Xa, Yb - Ya, Zb - Za);

    //cell half-width
    double DZ = fGeo->Plane((unsigned int)(planeId))->Cell ((unsigned int)(cellId ))->HalfL();
    //pigtail for cell
    double pigtail = fGeo->GetPigtail(fGeo->Plane(planeId)->Cell(cellId)->Id());
    
    double factor;
    if (fGeo->Plane(planeId)->View() == geo::kX) factor = fXViewFactor;
    else factor = fYViewFactor;

    //find out how many steps to take
    int nsteps = ceil( seglength/fStep );

    // The total number of PE that would have been produced are accumulated in
    // here, whether or not fluctuations allowed a PhotonSignal to be created
    // for them to live in.
    double lambda = 0;
    
    for (int istep = 0; istep < nsteps; ++istep)
      {
	//at midpoint of the step
	double stepZ = Za + ((Zb-Za)/nsteps)*(istep+0.5);
	double stepT = Ta + ((Tb-Ta)/nsteps)*(istep+0.5);
	double stepE = (fApplyBirks) ? hit.GetEdepBirks()/nsteps : hit.GetEdep()/nsteps;

	//green_scale*stepE*collectionRate = N Photo-electrons (before attenuation)
	std::list<RateInfo>::iterator rate_it;
	for (rate_it = fRateInfo.begin(); rate_it != fRateInfo.end(); ++rate_it)
	  {
	    double collectionRate = rate_it->rate;
	    double currZ = stepZ + rate_it->z;
	    double fraction(1.0);

	    //if photons collected is small, stop working on this step
	    //if ( green_scale*stepE*collectionRate < 0.1 ) break;

	    //is this currZ inside the cell?
	    if ( std::fabs(currZ) - 0.5*rate_it->zWidth > DZ ) continue; 
	    else if ( std::fabs(currZ) + 0.5*rate_it->zWidth < DZ ) fraction = 1.0;
	    else fraction = (DZ - ( std::fabs(currZ) - 0.5*rate_it->zWidth ))/rate_it->zWidth;
	    
	    double muGreen = green_scale*stepE*collectionRate*fraction*factor;
	    //short and long length to the electronics
	    double dist0 = DZ  - currZ + pigtail;
	    double dist1 = 3*DZ + currZ + pigtail;			    

	    CreatePhotoElectrons( 0.5*muGreen, stepT, dist0, *rate_it, lambda );
	    CreatePhotoElectrons( 0.5*muGreen, stepT, dist1, *rate_it, lambda );
	  }
      }

    // Keep track of how many we have so far so we know which ones get newly
    // added by ClusterPhotoElectrons().
    unsigned int start = fSignals->size();

    ClusterPhotoElectrons(planeId, cellId, trackId);

    // Divide the poission lambda evenly among all PhotonSignals. When someone
    // adds them up downstream they'll get the right overall answer for this
    // FLSHit.
    const int N = fSignals->size()-start;
    if(N > 0){
      for(unsigned int i = start; i < fSignals->size(); ++i){
	(*fSignals)[i].SetPoissonLambda(lambda/N);
      }
    }
    else{
      if(fWriteEmptySignals){
	// Create a PhotonSignal with zero PE, just to have somewhere to attach
	// the PoissonLambda to.
	sim::PhotonSignal photon(hit.GetEntryT(), 0);
	photon.SetPlaneCell(planeId, cellId);
	photon.SetTrackId(trackId);
	photon.SetPoissonLambda(lambda);
	fSignals->push_back(photon);
      }
    }
    
    return;
  }
    
  //............................................................ 
  void ImprovedTransport::CreatePhotoElectrons(double meanPE, double stepTime, double dist, RateInfo& info, double& lambda)
  {
    double expect = meanPE*FiberTransmission(dist);
    lambda += expect;

    int nPE = fPoisson->fire( expect );
    double rand0(0), rand1(0), rand2(0);
    for (int iPE=0; iPE < nPE; ++iPE)
      {
	double time = stepTime;
	time += dist/fFiberSpeedOfLight;
	
	fRandHisto->GetRandom( &info.timePDF, rand0, rand1, rand2 );	
	time += rand0;

	double fiberspread = 0;
	if (fApplyFiberSmearing) 
	  {
	    double weight(-1.0), uweight(0.0);
	    while (weight < uweight) 
	      {
		// Draw a Dt/z value then multiply by "dist" to get time (in ns)
		fRandHisto->GetRandom( &fFiberSmearingHisto, rand0, rand1, rand2 );
		fiberspread = dist*rand0;
		
		// Use fiber attenuation to deweight longer times (<==>longer distances)
		// This is an approximation, but should be pretty close.
		double Delta_dist = fiberspread*fFiberSpeedOfLight;
		weight = FiberTransmission(dist+Delta_dist)/FiberTransmission(dist);
		uweight = fFlat->fire(0,1);
	      }
	  }
	time += fiberspread;
	
	fPETimes.push_back( time );
      }
    
    return;
  }
  
  
  //............................................................ 
  void ImprovedTransport::ClusterPhotoElectrons(int planeId, int cellId, int trackId)
  {
    if (fPETimes.empty()) return;
    //cluster photons into PhotonSignals fTimeClustering long

    sim::PhotonSignal photon;
    photon.SetPlaneCell(planeId, cellId);
    photon.SetTrackId(trackId);

    double meanTime(0);
    int nPE(0);

    fPETimes.sort();
    std::list< double >::iterator it;
    for (it = fPETimes.begin(); it != fPETimes.end(); ++it)
      {
	double currTime = *it;
	if (nPE == 0)
	  {
	    meanTime = currTime;
	    nPE = 1;
	  }
	else if ( fabs(meanTime - currTime) < fTimeClustering )
	  {
	    //add to cluster
	    meanTime = meanTime*nPE + currTime;
	    ++nPE;
	    meanTime /= nPE;
	  }
	else 
	  {
	    photon.SetTimeMean(meanTime);
	    photon.SetNPhoton(nPE);
	    fSignals->push_back(photon);

	    //Set variables with current iteration
	    nPE = 1;
	    meanTime = currTime;
	  }
      }
    //there is one PhotonSignal that hasn't been saved yet
    photon.SetTimeMean(meanTime);
    photon.SetNPhoton(nPE);
    fSignals->push_back(photon);
    
    return;
  }
  
  //............................................................
  bool ImprovedTransport::LoadHistos() 
  {
    // just read a file from disk for now.  should eventually be in the DB.
    // use FileInPath to get the histogram file
    struct stat sb;
    if ( stat(fCollectionRateFile.c_str(), &sb) !=0 ) 
      {
	throw cet::exception("NoCollectionRateFile")
	  << "ImprovedTransport: Unable to open " 
	  << fCollectionRateFile << "\n"
	  << __FILE__ << ":" << __LINE__ << "\n";
	return false;
      }

    if ( stat(fSmearingHistoFile.c_str(), &sb) !=0 )
      {
        throw cet::exception("NoSmearingHistoFile") << "ImprovedTransport: Unable to open " << fSmearingHistoFile << "\n" << __FILE__ << ":" << __LINE__ << "\n";
        return false;
      }

    TFile f1(fCollectionRateFile.c_str());
    TH2D *htemp = (TH2D*)f1.Get("dT_dZ_CollectionRate")->Clone();
    if (!htemp)
      {
	std::cerr << "ImprovedTransport: Histogram dT_dZ_CollectionRate not found in " << fCollectionRateFile << std::endl;
        return false;
      }

    fCollectionRateHisto = *htemp;
    fCollectionRateHisto.SetDirectory(0);
    f1.Close();

    TFile f2(fSmearingHistoFile.c_str());
    TH1D *htemp2 = (TH1D*)f2.Get("Dt_per_z")->Clone();
    if (!htemp2)
      {
	std::cerr << "ImprovedTransport: Histogram Dt_per_z_distribution not found in " << fSmearingHistoFile << std::endl;
        return false;
      }

    fFiberSmearingHisto = *htemp2;
    fFiberSmearingHisto.SetDirectory(0);
    f2.Close();

    fRateInfo.clear();
    std::stringstream hName;
    int nbinsY = fCollectionRateHisto.GetNbinsY();
    double ylo = fCollectionRateHisto.GetYaxis()->GetXmin();
    double yhi = fCollectionRateHisto.GetYaxis()->GetXmax();
    for (int iZ = 1; iZ <= fCollectionRateHisto.GetNbinsX(); ++iZ)
      {
	RateInfo info_pos;
	info_pos.rate = 0.0;
	info_pos.z = fCollectionRateHisto.GetXaxis()->GetBinCenter(iZ);
	info_pos.zWidth = fCollectionRateHisto.GetXaxis()->GetBinWidth(iZ);
	hName.str("timePDF_Pos_");
	hName << iZ;
	info_pos.timePDF.SetName(hName.str().c_str());
	info_pos.timePDF.SetBins(nbinsY, ylo, yhi);
	info_pos.timePDF.SetDirectory(0);

	RateInfo info_neg;
	info_neg.rate = 0.0;
	info_neg.z = -fCollectionRateHisto.GetXaxis()->GetBinCenter(iZ);
	info_neg.zWidth = fCollectionRateHisto.GetXaxis()->GetBinWidth(iZ);
	hName.str("timePDF_Neg_");
	hName << iZ;
	info_neg.timePDF.SetName(hName.str().c_str());
	info_neg.timePDF.SetBins(nbinsY, ylo, yhi);
	info_neg.timePDF.SetDirectory(0);

	for (int iT = 1; iT <= fCollectionRateHisto.GetNbinsY(); ++iT)
	  {
	    double collectionRate = 0.5*fCollectionRateHisto.GetBinContent(iZ,iT);
	    info_pos.rate += collectionRate;
	    info_pos.timePDF.SetBinContent(iT, collectionRate);

	    info_neg.rate += collectionRate;
	    info_neg.timePDF.SetBinContent(iT, collectionRate);
	  }
	fRateInfo.push_back(info_pos);
	fRateInfo.push_back(info_neg);
      }
    return true;
  }
  
  //............................................................
  double ImprovedTransport::FiberTransmission(double x) 
  {
    // currently, all fibers are the same
    
    // fAttenPars holds a series of (c,L) pairs that form the
    // function:
    //    transmission = c0*exp(-x/L0) + c1*exp(-x/L1) + ...
    //
    // Thus, we should have an even number of parameters

    double answer = 0;
    
    int nPar = fAttenPars.size();
    if (!nPar) 
      {
	mf::LogWarning("AttenPars") << "No AttenPars entries.  Fiber transmission set to 1!";
	answer = 1;
      }
    else if (nPar%2) 
      {
	mf::LogWarning("AttenPars") << "Odd number of AttenPars entries.  Fiber transmission set to 1!";
	answer = 1;
      }
    else 
      {
	answer = 0;    
	for (Int_t iP=0;iP<nPar;iP+=2) answer += fAttenPars[iP]*exp(-x/fAttenPars[iP+1]);
      }
    return answer;
  }
  
  DEFINE_ART_MODULE(ImprovedTransport);
} //namespace
