////////////////////////////////////////////////////////////////////////////
// \file    NumiFilteringAna_module.cc
// \brief   Module to filter events for neutrino search.
//          Output files used for neutrino event scanning in Far Detector
//          Select Contained/Exiting/Entering tracks after cuts.
// \author  Gavin Davies - gsdavies@iastate.edu
///////////////////////////////////////////////////////////////////////////
#include <algorithm>
#include <cassert>
#include <string>

// ROOT includes
#include "TH1D.h"
#include "TH1I.h"
#include "TMath.h"
#include "TVector3.h"

//Novasoft includes
#include "CosRej/CosRejFxs.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Track.h"
#include "SummaryData/CosmicExposure.h"
#include "Utilities/AssociationUtil.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"


namespace comi
{
  class NumiFilteringAna: public art::EDAnalyzer
  {
  public:
    explicit NumiFilteringAna(fhicl::ParameterSet const &pset);
    virtual ~NumiFilteringAna();
    
    void beginRun(const art::Run& run);
    void analyze(art::Event const& evt);
    virtual void beginSubRun(const art::SubRun& sr);

    void beginJob();
    void endJob();

  protected:

    std::string  fSlicerLabel;
    std::string  fTrackType;
    std::string  fProngType;

    float        fMinXFid;
    float        fMinYFid;
    float        fMinZFid;
    float        fMaxXFid;
    float        fMaxYFid;
    float        fMaxZFid;


    fhicl::ParameterSet fPSetNDOS, fPSetND, fPSetFD;
  private:

    cosrej::CosRejFxs*         fCosRejFxs;

    //event summary info
    int run;
    int subrun;
    int event;
    
    double fTotalLivetime;
    
    TH1D *fLivetime;
    // Histograms
    TH1D *fSMaxAnyHitUn;

    TH1D *fSMinXHitUn;
    TH1D *fSMinYHitUn;
    TH1D *fSMinZHitUn;

    TH1D *fSMaxXHitUn;
    TH1D *fSMaxYHitUn;
    TH1D *fSMaxZHitUn;

    TH1D *fSMinX;
    TH1D *fSMinY;
    TH1D *fSMinZ;

    TH1D *fSMaxX;
    TH1D *fSMaxY;
    TH1D *fSMaxZ;

    TH1D *fFMinX;
    TH1D *fFMinY;
    TH1D *fFMinZ;

    TH1D *fFMaxX;
    TH1D *fFMaxY;
    TH1D *fFMaxZ;

    TH1D *fSNXplanes;
    TH1D *fSNYplanes;
    TH1D *fFNXplanes;
    TH1D *fFNYplanes;

    
    
  };

}//End namespace comi

namespace comi
{
  //----------------------------------------------------------------------
  NumiFilteringAna::NumiFilteringAna(fhicl::ParameterSet const &pset) : 
    EDAnalyzer(pset),
    fSlicerLabel       (pset.get< std::string         >("SlicerLabel")  ),
    fTrackType         (pset.get< std::string         >("TrackType")    ),
    fProngType         (pset.get< std::string         >("ProngType")    ),
    fPSetNDOS          (pset.get< fhicl::ParameterSet >("ndos")         ),
    fPSetND            (pset.get< fhicl::ParameterSet >("nd")           ),
    fPSetFD            (pset.get< fhicl::ParameterSet >("fd")           )
  {
  }

  //----------------------------------------------------------------------
  NumiFilteringAna::~NumiFilteringAna()
  {
  }
  
  //----------------------------------------------------------------------
  void NumiFilteringAna::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    
    double nhitmax    = 100.0;
    int    nhitbins   = 100;
    double nplanemax  = 500.0;
    int    nplanebins = 500;
    double xmax       = 1200.0;
    double ymax       = 1200.0;
    double zmax       = 6500.0;
    int    xbins      = 480;
    int    ybins      = 480;
    int    zbins      = 1300;

    fLivetime     = tfs->make<TH1D>("livetime", "TotalLiveTime;;Seconds", 1, 0, 1);
    // # Uncontained hits in x/y/z min/max for Slicer Clusters
    fSMaxAnyHitUn = tfs->make<TH1D>("smaxanyhitun",";# Slicer Cluster Uncontained Hits Max; # of Slices",nhitbins, 0., nhitmax);

    fSMinXHitUn = tfs->make<TH1D>("sminxhitun",";# Slicer Cluster Uncontained Hits minX; # of Slices",nhitbins, 0., nhitmax);
    fSMinYHitUn = tfs->make<TH1D>("sminyhitun",";# Slicer Cluster Uncontained Hits minY; # of Slices",nhitbins, 0., nhitmax);
    fSMinZHitUn = tfs->make<TH1D>("sminzhitun",";# Slicer Cluster Uncontained Hits minZ; # of Slices",nhitbins, 0., nhitmax);

    fSMaxXHitUn = tfs->make<TH1D>("smaxxhitun",";# Slicer Cluster Uncontained Hits maxX; # of Slices",nhitbins, 0., nhitmax);
    fSMaxYHitUn = tfs->make<TH1D>("smaxyhitun",";# Slicer Cluster Uncontained Hits maxY; # of Slices",nhitbins, 0., nhitmax);
    fSMaxZHitUn = tfs->make<TH1D>("smaxzhitun",";# Slicer Cluster Uncontained Hits maxZ; # of Slices",nhitbins, 0., nhitmax);

    // Min/Max X/Y/Z position of Slicer Cluster
    fSMinX = tfs->make<TH1D>("sminx",";Min. X Slicer Cluster; # of Slices",xbins, -xmax, xmax);
    fSMinY = tfs->make<TH1D>("sminy",";Min. Y Slicer Cluster; # of Slices",ybins, -ymax, ymax);
    fSMinZ = tfs->make<TH1D>("sminz",";Min. Z Slicer Cluster; # of Slices",zbins, 0., zmax);

    // Min/Max X/Y/Z position of FuzzyK Cluster
    fFMinX = tfs->make<TH1D>("fminx",";Min. X FuzzyK Cluster; # of Slices",xbins, -xmax, xmax);
    fFMinY = tfs->make<TH1D>("fminy",";Min. Y FuzzyK Cluster; # of Slices",ybins, -ymax, ymax);
    fFMinZ = tfs->make<TH1D>("fminz",";Min. Z FuzzyK Cluster; # of Slices",zbins, 0., zmax);

    // Min/Max X/Y/Z position of Slicer Cluster
    fSMaxX = tfs->make<TH1D>("smaxx",";Max. X Slicer Cluster; # of Slices",xbins, -xmax, xmax);
    fSMaxY = tfs->make<TH1D>("smaxy",";Max. Y Slicer Cluster; # of Slices",ybins, -ymax, ymax);
    fSMaxZ = tfs->make<TH1D>("smaxz",";Max. Z Slicer Cluster; # of Slices",zbins, 0., zmax);

    // Min/Max X/Y/Z position of FuzzyK Cluster
    fFMaxX = tfs->make<TH1D>("fmaxx",";Max. X FuzzyK Cluster; # of Slices",xbins, -xmax, xmax);
    fFMaxY = tfs->make<TH1D>("fmaxy",";Max. Y FuzzyK Cluster; # of Slices",ybins, -ymax, ymax);
    fFMaxZ = tfs->make<TH1D>("fmaxz",";Max. Z FuzzyK Cluster; # of Slices",zbins, 0., zmax);

    // # X/Y planes for Slicer Clusters
    fSNXplanes = tfs->make<TH1D>("snxplanes",";# X Planes Slicer Cluster Extends; # of Slices",nplanebins, 0., nplanemax);
    fSNYplanes = tfs->make<TH1D>("snyplanes",";# Y Planes Slicer Cluster Extends; # of Slices",nplanebins, 0., nplanemax);

    fFNXplanes = tfs->make<TH1D>("fnxplanes",";# X Planes FuzzyK Cluster Extends; # of Slices",nplanebins, 0., nplanemax);
    fFNYplanes = tfs->make<TH1D>("fnyplanes",";# Y Planes FuzzyK Cluster Extends; # of Slices",nplanebins, 0., nplanemax);
  }

  //----------------------------------------------------------------------
  void NumiFilteringAna::beginRun(const art::Run& run)
  {
    art::ServiceHandle<geo::Geometry>  geom;

    fhicl::ParameterSet pset;
    
    switch(geom->DetId()){
    case novadaq::cnv::kNDOS:
      pset = fPSetNDOS;
      break;
    case novadaq::cnv::kNEARDET:
      pset = fPSetND;
      break;
    case novadaq::cnv::kFARDET:
      pset = fPSetFD;
      break;
    default:
      assert(0 && "Unknown detector");
    }

    fMinXFid = pset.get< float >("MinXFid");
    fMaxXFid = pset.get< float >("MaxXFid");
    fMinYFid = pset.get< float >("MinYFid");
    fMaxYFid = pset.get< float >("MaxYFid");
    fMinZFid = pset.get< float >("MinZFid");
    fMaxZFid = pset.get< float >("MaxZFid");
  }
  
  //----------------------------------------------------------------------
  void NumiFilteringAna::analyze(art::Event const& evt)
  {
    
    run = evt.run();
    subrun = evt.subRun();
    event = evt.id().event();

    //Grab event's Slicer slices
    art::Handle< std::vector<rb::Cluster> > slices;
    evt.getByLabel(fSlicerLabel, slices);
    
    // Object holding prongs associated with cluster
    art::FindManyP<rb::Prong> fmpProng(slices, evt, fProngType);
      
    // Object holding tracks associated with cluster
    art::FindManyP<rb::Track> fmpTrack(slices, evt, fTrackType); 
    
    const int sliceMax = slices->size();

    for(int sliceIdx = 0; sliceIdx < sliceMax; ++sliceIdx){
      if(rb::IsFiltered(evt, slices, sliceIdx)) continue;
      
      const rb::Cluster& slice = (*slices)[sliceIdx];
      
      // Skip noise slices
      if(slice.IsNoise()) continue;

      std::vector< art::Ptr<rb::Prong> > prongs;
      
      double       longestObj = 0;
      unsigned int longObjIdx = 0;

      // If there are associations between Prong and Cluster use them
      if(fmpProng.isValid()){
	prongs = fmpProng.at(sliceIdx);
      }
      else continue; // otherwise skip this slice

      // 3D prongs container
      std::vector< art::Ptr<rb::Prong> > prongs3D;
      for(unsigned int itr = 0; itr < prongs.size(); ++itr){
	if(prongs[itr]->Is3D()) prongs3D.push_back(prongs[itr]);
      }
            
      //Find the longest 3D prong in the slice:
      for(unsigned int n = 0; n < prongs3D.size(); ++n){
	const double L = prongs3D[n]->TotalLength();
	if(L > longestObj) {
	  longestObj = L;
	  longObjIdx = n;
	}
      }
      
      const double nCells = slice.NCell(); //Number of cells in slice

      unsigned int PXhi = 0, PXhiFK = 0;
      unsigned int PXlo = UINT_MAX, PXloFK = UINT_MAX;
      unsigned int PYhi = 0, PYhiFK = 0;
      unsigned int PYlo = UINT_MAX, PYloFK = UINT_MAX;
      
      bool is3d = false;


      if(fmpTrack.isValid()){
        std::vector< art::Ptr<rb::Track> > tracks = fmpTrack.at(sliceIdx);
        if(tracks.size()>0) {
          const art::Ptr<rb::Track> track = tracks[0];
	  is3d = track->Is3D();
	}
      }
      else {
	mf::LogWarning("NumiFilteringAna") << "No Tracks. Skip this slice!";
	continue;
      }
      
      //----------------Make any cut selections here----------------
      // Do the NCell cut
      if(nCells <= 20 || !is3d) continue;
      if(prongs3D.size() < 1) continue;
      
      double sminx = -9999., sminy = -9999., sminz = -9999.;
      double smaxx = 9999., smaxy = 9999., smaxz = 9999.;

      double fminx = -9999., fminy = -9999., fminz = -9999.;
      double fmaxx = 9999., fmaxy = 9999., fmaxz = 9999.;

      sminx    = slice.MinXYZ().X();
      sminy    = slice.MinXYZ().Y();
      sminz    = slice.MinXYZ().Z();
      
      smaxx    = slice.MaxXYZ().X();
      smaxy    = slice.MaxXYZ().Y();
      smaxz    = slice.MaxXYZ().Z();

      // Skip 4 diblock slices
      if(sminz > 1702 || smaxz > 1702) continue;
      
      
      fminx    = prongs3D[longObjIdx]->MinXYZ().X();
      fminy    = prongs3D[longObjIdx]->MinXYZ().Y();
      fminz    = prongs3D[longObjIdx]->MinXYZ().Z();
      
      fmaxx    = prongs3D[longObjIdx]->MaxXYZ().X();
      fmaxy    = prongs3D[longObjIdx]->MaxXYZ().Y();
      fmaxz    = prongs3D[longObjIdx]->MaxXYZ().Z();
      
      unsigned int nXplanes = 0, nYplanes = 0;

      unsigned int stot[6] = {0,0,0,0,0,0};

      //loop over all hits
      for(int view = geo::kX; view <= geo::kY; ++view) {
	const geo::View_t geoview = geo::View_t(view);
	
	for(unsigned int j = 0; j < slice.NCell(geoview); ++j) {
	  
	  const rb::CellHit* h = slice.Cell(geoview,j).get();

	  art::ServiceHandle<geo::Geometry>  geom;

	  double xyz[3];
	  geom->Plane(h->Plane())->Cell(h->Cell())->GetCenter(xyz);
	  const double x = xyz[0];
	  const double y = xyz[1];
	  const double z = xyz[2];
	  
	  if(h->View() == geo::kX) {
	    if(h->Plane() > PXhi) PXhi = h->Plane();
	    if(h->Plane() < PXlo) PXlo = h->Plane();
	  }
	  if(h->View() == geo::kY) {
	    if(h->Plane() > PYhi) PYhi = h->Plane();
	    if(h->Plane() < PYlo) PYlo = h->Plane();
	  }
	  // Total up number of cells > Max (X/Y/Z) and < Min (X/Y/Z)
	  //TOP
	  if(y > fMaxYFid) {
	    stot[0]++;
	    continue;
	  }
	  //BOTTOM
	  if(y < fMinYFid) {
	    stot[1]++;
	    continue;
	  }
	  //LEFT
	  if(x > fMaxXFid) {
	    stot[2]++;
	    continue;
	  }
	  //RIGHT
	  if(x < fMinXFid) {
	    stot[3]++;
	    continue;
	  }
	  //FRONT
	  if(z < fMinZFid) {
	    stot[4]++;
	    continue;
	  }
	  //BACK
	  if(z > fMaxZFid) {
	    stot[5]++;
	    continue;
	  }

	}
      }
      
      
      nXplanes = PXhi - PXlo+1;
      nYplanes = PYhi - PYlo+1;
      
      unsigned int nXplanesFK = 0, nYplanesFK = 0;

      //loop over all hits
      for(int view = geo::kX; view <= geo::kY; ++view) {
	const geo::View_t geoview = geo::View_t(view);
	const unsigned int ncellview = prongs3D[longObjIdx]->NCell(geoview);

	for(unsigned int j = 0; j < ncellview; ++j) {
	  
	  art::ServiceHandle<geo::Geometry>  geom;
	  const rb::CellHit* h = prongs3D[longObjIdx]->Cell(geoview,j).get();
	  
	  // Find max/min plane number
	  if(h->View() == geo::kX) {
	    if(h->Plane() > PXhiFK) PXhiFK = h->Plane();
	    if(h->Plane() < PXloFK) PXloFK = h->Plane();
	  }
	  if(h->View() == geo::kY) {
	    if(h->Plane() > PYhiFK) PYhiFK = h->Plane();
	    if(h->Plane() < PYloFK) PYloFK = h->Plane();
	  }
	}
      }
      
      nXplanesFK = PXhiFK - PXloFK+1;
      nYplanesFK = PYhiFK - PYloFK+1;
      
      
      // Fill Histograms
      fSNXplanes->Fill(nXplanes);
      fSNYplanes->Fill(nYplanes);
      
      fFNXplanes->Fill(nXplanesFK);
      fFNYplanes->Fill(nYplanesFK);

      // Min/Max postions
      fSMinX->Fill(sminx);
      fSMinY->Fill(sminy);
      fSMinZ->Fill(sminz);

      fFMinX->Fill(fminx);
      fFMinY->Fill(fminy);
      fFMinZ->Fill(fminz);

      fSMaxX->Fill(smaxx);
      fSMaxY->Fill(smaxy);
      fSMaxZ->Fill(smaxz);

      fFMaxX->Fill(fmaxx);
      fFMaxY->Fill(fmaxy);
      fFMaxZ->Fill(fmaxz);

      // Total # of hits above Max, below Min X/Y/Z
      fSMaxYHitUn->Fill(stot[0]);
      fSMinYHitUn->Fill(stot[1]);
      fSMaxXHitUn->Fill(stot[2]);
      fSMinXHitUn->Fill(stot[3]);
      fSMinZHitUn->Fill(stot[4]);
      fSMaxZHitUn->Fill(stot[5]);

      //std::cout << "The largest number is: " << *std::max_element(stot,stot+6) << "\n";

      const unsigned int stotmax = *std::max_element(stot,stot+6);
      fSMaxAnyHitUn->Fill(stotmax);

    } // end for sliceIdx    
  }
  

  void NumiFilteringAna::beginSubRun(const art::SubRun& sr)
  {
    art::Handle<sumdata::CosmicExposure> expo;
    
    sr.getByLabel("exposure", expo);
    if(!expo.failedToGet())
      {
	fTotalLivetime += expo->totlivetime;
	mf::LogInfo("CAFMaker") << "Exposure in subrun " << sr.subRun()
				<< " = " << expo->totlivetime;
      }
  }

  void NumiFilteringAna::endJob()
  {
    fLivetime->Fill(.5, fTotalLivetime);
  }
  
  DEFINE_ART_MODULE(NumiFilteringAna)  
}//end namespace comi
