////////////////////////////////////////////////////////////////////////
// \file    ReMId.cxx
// \brief   A pid object for muon identification
// \version $Id: ReMId.cxx,v 1.1.1.1 2012-09-26 18:38:44 raddatz Exp $
// \author  Nicholas Raddatz - raddatz@physics.umn.edu
////////////////////////////////////////////////////////////////////////
#include <limits>
#include <iostream>

#include "ReMId.h"
#include "TVector3.h"

#include "art/Framework/Core/FindOneP.h"
#include "cetlib/exception.h"

namespace remid
{

  //......................................
  ReMId::ReMId()
    : rb::PID()
  {
    SetMuonDedxLL(0);
    SetPionDedxLL(0);
    SetMuonScatLL(0);
    SetMuonScatLL(0);
  }

  //......................................
  ReMId::ReMId(int pdg, double val)
    : rb::PID(pdg,val)
  {
    SetMuonDedxLL(0);
    SetPionDedxLL(0);
    SetMuonScatLL(0);
    SetMuonScatLL(0);
  }

  //......................................
  void    ReMId::AddScatter(TVector3 pos,double cosScat)
  {
    fscatLocation.push_back(pos);
    fcosScatAngle.push_back(cosScat);
  }

  //.....................................
  void    ReMId::AddDedx(double dedx, double distFromStart, bool used, bool vert)
  {
    fdedx.push_back(dedx);
    ffracDistFromStart.push_back(distFromStart);
    fdedxused.push_back(used);
    fdedxvert.push_back(vert);
  }

  //.....................................
  void    ReMId::SetMuonDedxLL(double ll)
  {
    fMuonDedxLL = ll;
  }

  //.....................................
  void    ReMId::SetPionDedxLL(double ll)
  {
    fPionDedxLL = ll;
  }

  //.....................................
  double  ReMId::AvgDedx() const
  {
    if(fdedx.size() == 0){ return 0;}
    else{
      double totDedx = 0;
      for(unsigned int i = 0; i<fdedx.size();++i){
        totDedx+=fdedx[i];
      } 
      return totDedx/((double)fdedx.size());
    }
  }

  //.....................................
  double  ReMId::AvgDedxUsed() const
  {
    int ndedxused = 0;
    double totDedx = 0;
    for(unsigned int i = 0; i < fdedx.size(); ++i){
      if(fdedxused[i]){
	++ndedxused;
	totDedx+=fdedx[i];
      }
    }
    if(ndedxused == 0){ return 0;}
    else{return totDedx/((double)ndedxused); }
  }

  //.....................................
  unsigned int  ReMId::NScatters() const
  {
    return fscatLocation.size();
  }

  //.....................................
  unsigned int ReMId::NDedx() const
  {
    return fdedx.size();
  }

  //.....................................
  TVector3  ReMId::ScatLocation(unsigned int i) const
  {
    return fscatLocation[i];
  }

  //.....................................
  double  ReMId::CosScat(unsigned int i) const
  {
    return fcosScatAngle[i];
  }

  //.....................................
  double  ReMId::DedxLocation(unsigned int i) const
  {
    return ffracDistFromStart[i];
  }

  //.....................................
  double  ReMId::Dedx(unsigned int i) const
  {
    return fdedx[i];
  }

  //.....................................
  bool  ReMId::DedxUsed(unsigned int i) const
  {
    return fdedxused[i];
  }

  //.....................................
  bool  ReMId::DedxVertex(unsigned int i) const
  {
    return fdedxvert[i];
  }

  //.....................................
  double  ReMId::MuonDedxLL() const
  {
    return fMuonDedxLL;
  }

  //.....................................
  double  ReMId::PionDedxLL() const
  {
    return fPionDedxLL;
  }

  //.....................................
  void    ReMId::SetMuonScatLL(double ll)
  {
    fMuonScatLL = ll;
  }

  //.....................................
  void    ReMId::SetPionScatLL(double ll)
  {
    fPionScatLL = ll;
  }

  //.....................................
  void    ReMId::SetDedxUsed(unsigned int idx, bool used)
  {
    fdedxused[idx] = used;
  }

  //.....................................
  void    ReMId::SetDedxVertex(unsigned int idx, bool vert)
  {
    fdedxvert[idx] = vert;
  }

  //.....................................
  void    ReMId::SetContained(bool cont)
  {
    fCont = cont;
  }

  //.....................................
  double  ReMId::MuonScatLL() const
  {
    return fMuonScatLL;
  }

  //.....................................
  double  ReMId::PionScatLL() const
  {
    return fPionScatLL;
  }

  //.....................................
  /// Return the dE/dx separation variable used as an input to the kNN that determines the pid value
  double  ReMId::DedxSeparation() const
  {
    return fMuonDedxLL-fPionDedxLL;
  }

  //.....................................
  /// Return the scattering separation variable used as an input to the kNN that determines the pid value
  double  ReMId::ScatSeparation() const
  {
    return fMuonScatLL-fPionScatLL;
  }

  //.....................................
  /// Return the measurement fraction variable used as an input to the kNN that determines the pid value
  double ReMId::MeasurementFraction() const
  {
    double frac = -1;
    if(this->NDedx() > 0){
      frac = ((double)this->NMeasurementPlanes())/((double)this->NDedx());
    }
    return frac;
  }

  //.....................................
  int  ReMId::NMeasurementPlanes() const
  {
    int npl = 0;
    for(unsigned int i = 0; i < fdedxused.size(); ++i){
      if(fdedxused[i]){ ++npl; }
    }
    return npl;
  }

  //.....................................
  int  ReMId::NVertexPlanes() const
  {
    int npl = 0;
    for(unsigned int i = 0; i< fdedxvert.size(); ++i){
      if(fdedxvert[i]){ ++npl; }
    }
    return npl;
  }

  //.....................................
  bool    ReMId::Contained() const
  {
    return fCont;
  }

  //.....................................
  unsigned int HighestPIDTrack(
                const std::vector< art::Ptr<rb::Track> >& sliceTracks,
                const std::string& remidModuleLabel, const art::Event& e)
  {
    int    bestTrack        = -1;
    double highestPID       = -2.0;
    double bestTrackLength  = 0.0;
    double bestStartZ       = std::numeric_limits<double>::max();
    double bestADC          = std::numeric_limits<double>::max();

    art::FindOneP<remid::ReMId> tracksToReMId(sliceTracks, e, remidModuleLabel);

    for(size_t c = 0; c < sliceTracks.size(); ++c){
      
      geo::View_t view = sliceTracks[c]->View();
      
      //Only use 3D tracks
      if (view != geo::kXorY){  
        continue;
      } //End of loop over 2d tracks

      if(tracksToReMId.isValid()){
        art::Ptr<remid::ReMId> trackReMId = tracksToReMId.at(c);   

        // If the PID is lower, we can bail, nothing to update 
        if(trackReMId->Value() < highestPID) continue; 

        if(trackReMId->Value() == highestPID)
        { // If the new guy is shorter, we can bail 
          if (sliceTracks[c]->TotalLength() < bestTrackLength) continue;

          // If equal, need to fall to the next level, low start Z
          if (sliceTracks[c]->TotalLength() == bestTrackLength){

            // If start Z is higher, we can bail 
            if (sliceTracks[c]->Start().Z() > bestStartZ) continue;

            // If equal, need to fall to the next level, total ADC
            // You might think this is stupid, but it's equivalent to flipping 
            // a coin which always lands on the same side.  This is good, this 
            // will happen so rarely that you should just pee up a rope.  
            // Use lower ADC 
            if (sliceTracks[c]->Start().Z() == bestStartZ){

              if(sliceTracks[c]->TotalADC() > bestADC) continue;

              if(sliceTracks[c]->TotalADC() == bestADC)
                throw cet::exception("ReMId Matched Track Exception")<<std::endl 
                      << "Can't break a tie between two ReMId/tracks.  "
                      << "This should have never happend, but it did.  " 
                      << "The numu group is going to have to reflect on this "
                      << "and find a way to fix it." << std::endl;                
              
            }


          }
        }


        
        highestPID = trackReMId->Value();
        bestTrackLength = sliceTracks[c]->TotalLength();
        bestStartZ = sliceTracks[c]->Start().Z(); 
        bestADC    = sliceTracks[c]->TotalADC(); 
        bestTrack = c;

        
      }//End of valid ReMId object
      
    }//End of for loop over tracks


    if (bestTrack < 0)
      return 999;

    return (unsigned int)bestTrack;
    

  }



}
