////////////////////////////////////////////////////////////////////////
// \file    RecoMuon_module.cc
// \brief   A module for creating reconstructed muon id objects
// \version $Id: RecoMuon_module.cxx,v 1.2 2012-10-06 01:56:18 gsdavies Exp $
// \author  Nicholas Raddatz - raddatz@physics.umn.edu
////////////////////////////////////////////////////////////////////////
#include <vector>
#include <string>

// ROOT includes
#include "TH2D.h"
#include "TVector3.h"
#include "TFile.h"
#include "TMath.h"
#include "TMVA/MethodKNN.h"
#include "TMVA/Factory.h"
#include "TMVA/Tools.h"
#include "TMVA/Reader.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Persistency/Common/Assns.h"
#include "Utilities/AssociationUtil.h"
#include "art/Framework/Core/FindManyP.h"

// NOvA includes
#include "ReMId/ReMId.h"
#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/PlaneGeo.h"
#include "RecoBase/Track.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "Utilities/func/EnvExpand.h"
#include "Utilities/func/MathUtil.h"
#include "MuonRemove/art/TrackCleanUpAlg.h"
#include "NumuEnergy/NumuEAlg.h"


/// A PID for muons
namespace remid {
  class RecoMuon : public art::EDProducer {
  public:
    
    explicit RecoMuon(fhicl::ParameterSet const& pset);
    virtual ~RecoMuon();
    void beginJob();
    void beginRun(art::Run& run);
    void produce(art::Event& evt);

  private:

    std::string fLibPath;  ///<Path to library samples
    std::string fXMLPath;  ///<Path to KNN XML file
    std::string fSliceLabel;  ///<Where to find reconstructed slices
    std::string fTrackLabel;  ///<Where to find reconstructed tracks
    fhicl::ParameterSet fTrackCleanUpAlgPSet;  ///< Parameter set to configure the TrackCleanUpAlg object
    fhicl::ParameterSet fNumuEAlgPSet;  ///< Parameter set to configure the NumuEAlg object

    std::string fTrainingSampleFile; 

    TH2D*  dEdxVxMuonSample;
    TH2D*  dEdxVxPionSample;

    TH2D*  ScatVxMuonSample;
    TH2D*  ScatVxPionSample;

    std::vector<double> fmuondedxnorm;
    std::vector<double> fpiondedxnorm;
    std::vector<double> fmuonscatnorm;
    std::vector<double> fpionscatnorm;

    bool trainfileloaded;
    TMVA::Reader* reader;
    Float_t  LLE, LLScat, length, measfrac;

    double fDetWidth, fDetHeight, fDetLength;
    double  maxpionlength;

    murem::TrackCleanUpAlg*   fTrackCleanUpAlg;
    numue::NumuEAlg* fNumuEAlg;

    bool IsTrackContained(art::Ptr<rb::Track> track);

    void Get2DHistoBins(TH2D* histogram, double xval, double yval, int &xbin, int &ybin);

  };

  //....................................................................................

  RecoMuon::RecoMuon(fhicl::ParameterSet const& pset):
    fTrackCleanUpAlg(0),
    fNumuEAlg(0)
  {
    fLibPath    = pset.get< std::string >("LibraryPath");
    fXMLPath    = pset.get< std::string >("XMLPath");
    fSliceLabel = pset.get< std::string >("SliceLabel");
    fTrackLabel = pset.get< std::string >("TrackLabel");
    fTrackCleanUpAlgPSet = pset.get< fhicl::ParameterSet >("TrackCleanUpAlgPSet");
    fNumuEAlgPSet = pset.get< fhicl::ParameterSet >("NumuEAlgPSet");

    produces< std::vector<remid::ReMId> >();
    produces< art::Assns<remid::ReMId,rb::Track> >();

  }

  //....................................................................................

  RecoMuon::~RecoMuon()
  {
    if(fTrackCleanUpAlg){ delete fTrackCleanUpAlg; }
    if(fNumuEAlg){ delete fNumuEAlg; }
  }

  //....................................................................................

  void RecoMuon::beginJob()
  {

    fTrackCleanUpAlg = new murem::TrackCleanUpAlg(fTrackCleanUpAlgPSet);

    trainfileloaded = false;    

    // Exapnd fcl input library path env. variable
    std::string libpath = util::EnvExpansion(fLibPath);

    // Grab the Dedx histograms
    TFile muondedxfile((libpath+"DedxVxSignal.root").c_str(),"READ");
    dEdxVxMuonSample = (TH2D*)muondedxfile.Get("DedxVxS");
    TFile piondedxfile((libpath+"DedxVxBackground.root").c_str(),"READ");
    dEdxVxPionSample = (TH2D*)piondedxfile.Get("DedxVxB");

    // normalize these histograms to give probabilities in each bin
    for(int i = 1;i<=dEdxVxMuonSample->GetNbinsX();++i){
      double xbintot = 0;
      for(int j = 0; j<=dEdxVxMuonSample->GetNbinsY()+1;++j){
        xbintot+=dEdxVxMuonSample->GetBinContent(i,j);
      }
      //scale the y bins in this x bin column to total number in this x bin, i.e. normalize to total probability of 1
      for(int j = 0; j<=dEdxVxMuonSample->GetNbinsY()+1;++j){
        if(xbintot != 0){dEdxVxMuonSample->SetBinContent(i,j,dEdxVxMuonSample->GetBinContent(i,j)/xbintot);}
        else{dEdxVxMuonSample->SetBinContent(i,j,-1);}
      }
      fmuondedxnorm.push_back(xbintot);
    }

    // normalize these histograms to give probabilities in each bin
    for(int i = 1;i<=dEdxVxPionSample->GetNbinsX();++i){
      double xbintot = 0;
      for(int j = 0; j<=dEdxVxPionSample->GetNbinsY()+1;++j){
        xbintot+=dEdxVxPionSample->GetBinContent(i,j);
      }
      //scale the y bins in this x bin column to total number in this x bin, i.e. normalize to total probability of 1
      for(int j = 0; j<=dEdxVxPionSample->GetNbinsY()+1;++j){
        if(xbintot != 0){dEdxVxPionSample->SetBinContent(i,j,dEdxVxPionSample->GetBinContent(i,j)/xbintot);}
        else{dEdxVxPionSample->SetBinContent(i,j,-1);}
      }
      fpiondedxnorm.push_back(xbintot);
    }
    maxpionlength = 550.0; // Maximum length to sample the pion ll dedx histogram

    // Now grab the scattering histograms
    TFile muonscatfile((libpath+"ScatVxSignal.root").c_str(),"READ");
    ScatVxMuonSample = (TH2D*)muonscatfile.Get("ScatVxS");
    TFile pionscatfile((libpath+"ScatVxBackground.root").c_str(),"READ");
    ScatVxPionSample = (TH2D*)pionscatfile.Get("ScatVxB");

    // normalize these histograms to give probabilities in each bin
    for(int i = 1;i<=ScatVxMuonSample->GetNbinsX();++i){
      double xbintot = 0;
      for(int j = 0; j<=ScatVxMuonSample->GetNbinsY()+1;++j){
        xbintot+=ScatVxMuonSample->GetBinContent(i,j);
      }
      //scale the y bins in this x bin column to total number in this x bin, i.e. normalize to total probability of 1
      for(int j = 0; j<=ScatVxMuonSample->GetNbinsY()+1;++j){
        if(xbintot != 0){ScatVxMuonSample->SetBinContent(i,j,ScatVxMuonSample->GetBinContent(i,j)/xbintot);}
        else{ScatVxMuonSample->SetBinContent(i,j,-1);}
      }
      fmuonscatnorm.push_back(xbintot);
    }

    // normalize these histograms to give probabilities in each bin
    for(int i = 1; i<=ScatVxPionSample->GetNbinsX(); ++i){
      double xbintot = 0;
      for(int j = 0; j<=ScatVxPionSample->GetNbinsY()+1; ++j){
        xbintot+=ScatVxPionSample->GetBinContent(i,j);
      }
      //scale the y bins in this x bin column to total number in this x bin, i.e. normalize to total probability of 1
      for(int j = 0; j<=ScatVxPionSample->GetNbinsY()+1;++j){
        if(xbintot != 0){ScatVxPionSample->SetBinContent(i,j,ScatVxPionSample->GetBinContent(i,j)/xbintot);}
        else{ScatVxPionSample->SetBinContent(i,j,-1);}
      }
      fpionscatnorm.push_back(xbintot);
    }


  }

  //....................................................................................
  void RecoMuon::beginRun(art::Run& run)
  {
    fNumuEAlg = new numue::NumuEAlg(fNumuEAlgPSet);

    std::string xmlpath = util::EnvExpansion(fXMLPath);

    art::ServiceHandle<geo::Geometry> geom;

    // set up FD training file
    if(geom->DetId() == novadaq::cnv::kFARDET){ fTrainingSampleFile = xmlpath+"TMVAClassification_KNN.weights.xml"; }
    // set up ND training file
    else if(geom->DetId() == novadaq::cnv::kNEARDET){ fTrainingSampleFile = xmlpath+"TMVAClassification_KNN.nd.weights.xml"; }
    else{ std::cerr << "This file isn't near or far detector, algorithm doesn't work on any other" << std::endl;}

    // Book Method
    if(!trainfileloaded){
      // Set up the TMVA::Reader to read in kNN weight file
      reader = new TMVA::Reader();
      // Add the pid variables
      reader->AddVariable("trackLength",&length);
      reader->AddVariable("dedxSep",&LLE);
      reader->AddVariable("scatSep",&LLScat);
      reader->AddVariable("measFrac",&measfrac);
      reader->BookMVA("KNN method",fTrainingSampleFile);
      trainfileloaded = true;
    }
  }
  //....................................................................................


  void RecoMuon::produce(art::Event&  evt)
  {
    std::unique_ptr< std::vector<remid::ReMId> > pidcol(new std::vector<remid::ReMId>);
    std::unique_ptr< art::Assns<remid::ReMId,rb::Track> > assncol(new art::Assns<remid::ReMId,rb::Track>);

    // get a geometry handle
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<calib::Calibrator> cal;

    fDetWidth  = geom->DetHalfWidth();
    fDetHeight = geom->DetHalfHeight();
    fDetLength = geom->DetLength();

    double tranZPos = -1.0;
    if(geom->DetId() == novadaq::cnv::kNEARDET){ tranZPos = fNumuEAlg->NDTransitionZPos(); }

    art::Handle<std::vector<rb::Cluster> > slicecol;
    evt.getByLabel(fSliceLabel,slicecol);
    art::PtrVector<rb::Cluster> slicelist;
    for(unsigned int i = 0; i<slicecol->size();++i){
      art::Ptr<rb::Cluster>slice(slicecol,i);
      slicelist.push_back(slice);
    }
  

    art::FindManyP<rb::Track> fmt(slicecol, evt, fTrackLabel);

    // loop over all of the slices
    for(unsigned int iSlice = 0; iSlice<slicelist.size(); ++iSlice){
      // get all of the tracks in this slice
      const std::vector< art::Ptr<rb::Track> > slicetrks = fmt.at(iSlice);
      for(unsigned int iTrack = 0;iTrack<slicetrks.size(); ++iTrack){
	// make a pid object for this track
	ReMId pid;
	pid.SetVal(-1); // Set default pid value 
	pid.SetPdg(13); // Set that this is a pid for muons

	art::Ptr<rb::Track> track = slicetrks[iTrack];

	// store some useful track info
	double trklen = track->TotalLength();

	unsigned int ntraj = track->NTrajectoryPoints();
	// vector holds effective active track length from the end of the track
	std::vector<double> lengths(ntraj); 
	// vector holds catcher track length from the end of the track 
	std::vector<double> catlengths(ntraj); 
	
	// loop over trajectory points starting from the end
	for(int itrj = ntraj-1; itrj >= 0; --itrj){
	  TVector3 trajpoint = track->TrajectoryPoint(itrj);
	  catlengths[itrj] = 0.0;
	  if(itrj == (int)ntraj-1){
	    double lend = 0.0;
	    if(geom->DetId() == novadaq::cnv::kNEARDET && trajpoint.Z() > tranZPos){
	      double catlength = 0.0;
	      // convert the catcher track length to an energy
	      double catenergy = fNumuEAlg->NDMuonEInCatcherOnly(catlength);
	      // get the effective active length of this catcher energy
	      lend = fNumuEAlg->ActiveTrackLengthFromMuonE(catenergy);
	      catlengths[itrj] = catlength;
	    }
	    lengths[itrj] = lend;
	  }
	  else{
	    TVector3 lasttrajpoint = track->TrajectoryPoint(itrj+1);
	    lengths[itrj] = (trajpoint-lasttrajpoint).Mag() + lengths[itrj+1];
	    catlengths[itrj] = catlengths[itrj+1];
	    if(geom->DetId() == novadaq::cnv::kNEARDET){
	      double catlength = 0.0;
	      double actlength = 0.0;
	      // check if this is all or partially in the muon catcher
	      if(lasttrajpoint.Z() > tranZPos){
		// all in the muon catcher
		if(trajpoint.Z() > tranZPos){ catlength = (trajpoint-lasttrajpoint).Mag(); }
		else{
		  // only part of this length is in the muon catcher, 
		  // split it up appropriately between active and catcher length
		  TVector3 diff = lasttrajpoint-trajpoint;
		  catlength = fabs(tranZPos-lasttrajpoint.Z())*diff.Mag()/fabs(diff.Z());
		  actlength = fabs(tranZPos-trajpoint.Z())*diff.Mag()/fabs(diff.Z());
		}
	      }
	      else{ actlength = (trajpoint-lasttrajpoint).Mag(); }
	      // convert the catlength to an active det length
	      double effactlen = 0.0;
	      if(catlength > 0.0){
		double totcatlength = catlength + catlengths[itrj+1];
		catlengths[itrj] = totcatlength;
		// convert the total catcher track length to an energy
		double catenergy = fNumuEAlg->NDMuonEInCatcherOnly(totcatlength);
		// get the effective active length of this catcher energy
		effactlen = fNumuEAlg->ActiveTrackLengthFromMuonE(catenergy);
	      }
	      else{ effactlen = lengths[itrj+1]; }
	      lengths[itrj] = (effactlen + actlength);
	    } // end of if ND
	  }
	} // end over loop over trajectory points
	
	trklen = lengths[0];

	unsigned int minPlane = track->MinPlane();
        unsigned int maxPlane = track->MaxPlane();

	int maxXPlane = -1;
	int maxYPlane = -1;

        double hitCosx[track->NXCell()];
        double hitCosy[track->NYCell()];
        for(unsigned int ihit = 0; ihit<track->NXCell(); ++ihit){
	  if(track->XCell(ihit)->Plane() > maxXPlane){ maxXPlane = track->XCell(ihit)->Plane(); }
	  // get direction with this x hit
          double xyzplane[3];
          geom->Plane(track->XCell(ihit)->Plane())->Cell(0)->GetCenter(xyzplane);
          double dz = geom->Plane(track->XCell(ihit)->Plane())->Cell(0)->HalfD();
          double minz = xyzplane[2]-dz;
          double maxz = xyzplane[2]+dz;
          unsigned int currtrj = 0;
          // find the matching trajectory point for this plane;
          for(unsigned int itraj = 0;itraj<track->NTrajectoryPoints();++itraj){
            if(track->TrajectoryPoint(itraj).Z()>minz && track->TrajectoryPoint(itraj).Z()<maxz){currtrj = itraj;}
          }
	  TVector3 direction = track->Dir();
	  if(currtrj>0 && currtrj<track->NTrajectoryPoints()-2){
	    direction = track->TrajectoryPoint(currtrj) - track->TrajectoryPoint(currtrj-1);
	  }
	  direction = direction.Unit();
	  hitCosx[ihit] = sqrt(1.0/(direction.X()*direction.X()+direction.Z()*direction.Z()))*direction.Z();
        }

	for(unsigned int ihit = 0; ihit<track->NYCell(); ++ihit){
	  if(track->YCell(ihit)->Plane() > maxYPlane){ maxYPlane = track->YCell(ihit)->Plane(); }
	  // get direction with this y hit
          double xyzplane[3];
          geom->Plane(track->YCell(ihit)->Plane())->Cell(0)->GetCenter(xyzplane);
          double dz = geom->Plane(track->YCell(ihit)->Plane())->Cell(0)->HalfD();
          double minz = xyzplane[2]-dz;
          double maxz = xyzplane[2]+dz;
          unsigned int currtrj = 0;
          // find the matching trajectory point for this plane;
          for(unsigned int itraj = 0;itraj<track->NTrajectoryPoints();++itraj){
            if(track->TrajectoryPoint(itraj).Z()>minz && track->TrajectoryPoint(itraj).Z()<maxz){currtrj = itraj;}
          }
	  TVector3 direction = track->Dir();
	  if(currtrj>0 && currtrj<track->NTrajectoryPoints()-2){
	    direction = track->TrajectoryPoint(currtrj) - track->TrajectoryPoint(currtrj-1);
	  }
	  direction = direction.Unit();
	  hitCosy[ihit] = sqrt(1.0/(direction.Y()*direction.Y()+direction.Z()*direction.Z()))*direction.Z();
        }

        // calculate dedx info for this track
	std::vector<double> dedxs;
	std::vector<double> xdedxs;
	std::vector<unsigned int> dedxpls;

        for(unsigned int iPlane = minPlane;iPlane<=maxPlane;++iPlane){
          double planeGeV = 0;
          double totcosinview = 0;
          double avgCosOtherView = 1.0;
          bool useOtherView = true;
          double nhits = 0;
          double xyzplane[3];
          geom->Plane(iPlane)->Cell(0)->GetCenter(xyzplane);
          double dz = geom->Plane(iPlane)->Cell(0)->HalfD();
          double minz = xyzplane[2]-dz;
          double maxz = xyzplane[2]+dz;
          int matchpt = 0;
          int minCell = 10000;
          int maxCell = 0;
          double nhitsother = 0;
          int view = 0;
          // find the matching trajectory point for this plane;
          for(unsigned int itraj = 0;itraj<track->NTrajectoryPoints();++itraj){
            if(track->TrajectoryPoint(itraj).Z()>minz && track->TrajectoryPoint(itraj).Z()<maxz){matchpt = itraj;}
          }

          // find the track length up to this trajectory point
          double xlength = trklen - lengths[matchpt];

	  if(minPlane == maxPlane){xlength = trklen/2;}
          if(track->Stop().Z()<track->Start().Z()){xlength = (trklen-xlength);}
          if(geom->Plane(iPlane)->View() == geo::kX){
            for(unsigned int iHit = 0; iHit<track->NXCell();++iHit){
              if(track->XCell(iHit)->Plane() == iPlane && track->XCell(iHit)->View() != geo::kY){
                rb::RecoHit rhit = track->RecoHit(track->XCell(iHit));
		if(minPlane == maxPlane){
		  rhit = cal->MakeRecoHit(*(track->XCell(iHit)), 0.25);
		}
                if(rhit.IsCalibrated()){
                  planeGeV+=rhit.GeV();
                  totcosinview+=hitCosx[iHit];
                  nhits+=1.0;
                  if(track->XCell(iHit)->Cell()<minCell){minCell = track->XCell(iHit)->Cell();}
                  if(track->XCell(iHit)->Cell()>maxCell){maxCell = track->XCell(iHit)->Cell();}
                }
              }
            }
            // try to get the avgcos from the other view
            double nhitsbefore = 0;
            double nhitsafter = 0;
            double totCosBefore = 0;
            double totCosAfter = 0;
            for(unsigned int iHit = 0; iHit<track->NYCell();++iHit){
              if(track->YCell(iHit)->Plane() == iPlane-1){
                ++nhitsbefore;
                totCosBefore+=hitCosy[iHit];
              }
              else if(track->YCell(iHit)->Plane() == iPlane+1){
                ++nhitsafter;
                totCosAfter+=hitCosy[iHit];
              }
            }
            if(nhitsbefore > 0 && nhitsafter > 0){
              avgCosOtherView = ((totCosAfter+totCosBefore)/(nhitsbefore+nhitsafter));
              nhitsother = (nhitsbefore+nhitsafter)/2;
            }
            else if(nhitsbefore > 0){ 
              avgCosOtherView = totCosBefore/nhitsbefore;
              nhitsother = nhitsbefore;
            }
            else if(nhitsafter > 0){ 
              avgCosOtherView = totCosAfter/nhitsafter;
              nhitsother = nhitsafter;
            }
            else{ useOtherView = false; }
          }
          else if(geom->Plane(iPlane)->View() == geo::kY){
            view = 1;
            for(unsigned int iHit = 0; iHit<track->NYCell();++iHit){
              if(track->YCell(iHit)->Plane() == iPlane && track->YCell(iHit)->View() != geo::kX 
		 && track->YCell(iHit)->View() != geo::kXorY){
                rb::RecoHit rhit = track->RecoHit(track->YCell(iHit));
		if(minPlane == maxPlane){
		  rhit = cal->MakeRecoHit(*(track->YCell(iHit)), 0.25);
		}
                if(rhit.IsCalibrated()){
                  planeGeV+=rhit.GeV();
                  totcosinview+=hitCosy[iHit];
                  nhits+=1.0;
                  if(track->YCell(iHit)->Cell()<minCell){minCell = track->YCell(iHit)->Cell();}
                  if(track->YCell(iHit)->Cell()>maxCell){maxCell = track->YCell(iHit)->Cell();}
                }
              }
            }
            // try to get the avgcos from the other view
            double nhitsbefore = 0;
            double nhitsafter = 0;
            double totCosBefore = 0;
            double totCosAfter = 0;
            for(unsigned int iHit = 0; iHit<track->NXCell();++iHit){
              if(track->XCell(iHit)->Plane() == iPlane-1){
                ++nhitsbefore;
                totCosBefore+=hitCosx[iHit];
              }
              else if(track->XCell(iHit)->Plane() == iPlane+1){
                ++nhitsafter;
                totCosAfter+=hitCosx[iHit];
              }
            }
            if(nhitsbefore > 0 && nhitsafter > 0){
              avgCosOtherView = TMath::Abs(((totCosAfter+totCosBefore)/(nhitsbefore+nhitsafter)));
              nhitsother = (nhitsbefore+nhitsafter)/2;
            }
            else if(nhitsbefore > 0){
              avgCosOtherView = TMath::Abs(totCosBefore/nhitsbefore);
              nhitsother = nhitsbefore;
            }
            else if(nhitsafter > 0){
              avgCosOtherView = TMath::Abs(totCosAfter/nhitsafter);
              nhitsother = nhitsbefore;
            }
            else{ useOtherView = false; }
          }

          if(nhits >0){
            double avgCos = TMath::Abs(totcosinview/nhits);
            double xyzp[3];
            geom->Plane(iPlane)->Cell(minCell)->GetCenter(xyzp);
            double xyzp1[3];
            geom->Plane(iPlane)->Cell(maxCell)->GetCenter(xyzp1);
            double planewidth = 2*geom->Plane(iPlane)->Cell(1)->HalfD();

	    // calculate the amount of dead material to subtract off
	    // first add up live material 
	    double liveLength = 0;
	    for(int i = minCell; i<maxCell+1;++i){
	      // get the cell half width
	      const geo::PlaneGeo* p = geom->Plane(iPlane);
	      const geo::CellGeo* c = p->Cell(i);
	      if(i == minCell || i == maxCell){ liveLength+=c->HalfW(); }
	      else{ liveLength+=(2.0*c->HalfW()); }
	    }
	    // dead material is total length - live length only if min and max cell are not the same
	    double deadLength = 0;
	    if(minCell != maxCell){ deadLength = (TMath::Abs(xyzp1[view]-xyzp[view])-liveLength); }

	    double planeheight = planewidth*TMath::Tan(TMath::ACos(avgCos))-deadLength;
            double planePathLength = TMath::Sqrt(planewidth*planewidth + planeheight*planeheight); 
	    // in the case of completely vertical track  be a bit careful
            if(avgCos == 0 || avgCos!=avgCos){ planePathLength = liveLength; }
            if(useOtherView){
              double deltaz = planewidth;
              double deltav = planewidth*TMath::Tan(TMath::ACos(avgCos))-deadLength;
              if(avgCos == 0 || avgCos!=avgCos){
                deltav = liveLength;
		// if the track looks completely vertical in both views,
		// than no part of the track length should be in z direction
                if(avgCosOtherView == 0 || avgCosOtherView != avgCosOtherView){ deltaz = 0; }
              }
	      // in opposite view dead length 
              double deltaov = (planewidth-deadLength)*TMath::Tan(TMath::ACos(avgCosOtherView));
	      // in the case that the other view looks completely vertical scale up
	      // the length between cells in this view by the number of hits in the other view
	      // this case should really never happen
              if(avgCosOtherView == 0 || avgCosOtherView != avgCosOtherView){
		deltaov = nhitsother*(TMath::Abs(xyzp1[view]-xyzp[view]))/nhits;
	      }
              planePathLength = TMath::Sqrt(deltaz*deltaz+deltav*deltav+deltaov*deltaov);
	    }
	    // also the case of completely vertical track
	    if(minPlane == maxPlane){ planePathLength = liveLength; }
	    assert(planeGeV/(planePathLength)>=0);

	    dedxs.push_back(planeGeV/(planePathLength));
	    xdedxs.push_back(xlength);
	    dedxpls.push_back(iPlane);
            
          } // nhits > 0

        } // end of loop over planes of the track

	// check if this track is contained
	bool trackcont = IsTrackContained(track);
	pid.SetContained(trackcont);

	// dedx log-likelihoods for muon and pion
        double llmuon = 0;
        double nllmuon = 0;
        double llpion = 0;
        double nllpion = 0;

	// get the vertex region planes of this track
	int vertexPlanes[2];
	fTrackCleanUpAlg->LastVertexPlanes(*(track),vertexPlanes);

	// is there any vertex free planes?
	int minVert[2] = {0, 0};
	if(vertexPlanes[0] < maxXPlane || vertexPlanes[1] < maxYPlane){
	  minVert[0] = vertexPlanes[0];
	  minVert[1] = vertexPlanes[1];
	} 

	// calculate log-likelihood for every dedx measurement made not in vertex planes
	for(unsigned int idedx = 0; idedx<dedxs.size(); ++idedx){
	  
	  // check if in vertex region
	  if(geom->Plane(dedxpls[idedx])->View() == geo::kX){
	    if((int)dedxpls[idedx] < minVert[0]){ continue; }
	  }
	  else{
	    if((int)dedxpls[idedx] < minVert[1]){ continue; }
	  }
	  if(dedxpls[idedx] == maxPlane && maxPlane != minPlane){ continue; }
	  
	  // get the distance from track end of this dedx measurement
	  double distFromEnd = trklen - xdedxs[idedx];
	  
	  int xbin = 0;
	  int dedxbin = 0;
	  Get2DHistoBins(dEdxVxMuonSample,distFromEnd,dedxs[idedx],xbin,dedxbin);
	  
	  // compute the probability from the muon sample histogram
	  double prob = dEdxVxMuonSample->GetBinContent(xbin,dedxbin);
	  if(prob <= 0){ 
	    // find the normalization of this distance bin
	    double norm = fmuondedxnorm[xbin-1];
	    prob = 1/norm;
	  }
	  llmuon+=TMath::Log(prob);
	  ++nllmuon;	    
	  
	} // end of loop over dedxs
	
	// get pion dedx log-likelihood
	int maxpionbin = dEdxVxPionSample->GetXaxis()->FindFixBin(maxpionlength); 
	for(unsigned int idedx = 0; idedx<dedxs.size(); ++idedx){

	  pid.AddDedx(dedxs[idedx], xdedxs[idedx]/trklen);

	  // check if in vertex region
	  if(geom->Plane(dedxpls[idedx])->View() == geo::kX){
	    if((int)dedxpls[idedx] < minVert[0]){
	      pid.SetDedxVertex(idedx,true);
	      pid.SetDedxUsed(idedx,false);
	      continue;
	    }
	  }
	  else{
	    if((int)dedxpls[idedx] < minVert[1]){
	      pid.SetDedxVertex(idedx,true);
	      pid.SetDedxUsed(idedx,false);
	      continue;
	    }
	  }

	  if(dedxpls[idedx] == maxPlane && maxPlane != minPlane){
	    pid.SetDedxUsed(idedx,false);
	    continue;
	  }

	  // get the distance from track end of this dedx measurement
	  double distFromEnd = trklen - xdedxs[idedx];
	  int xbin = 0;
	  int dedxbin = 0;
	  Get2DHistoBins(dEdxVxPionSample,distFromEnd,dedxs[idedx],xbin,dedxbin);
	  
	  if(xbin > maxpionbin){ xbin = maxpionbin; }
	  // compute the probability from the pion sample histogram
	  double prob = dEdxVxPionSample->GetBinContent(xbin,dedxbin);
	  if(prob <= 0){
	    // find the normalization of this distance bin
	    double norm = fmuondedxnorm[xbin-1];
	    prob = 1/norm;
	  }
	  llpion+=TMath::Log(prob);
	  ++nllpion;
	}


        double llscatmuon = 0;
        double nllscatmuon = 0;
        double llscatpion = 0;
        double nllscatpion = 0;

	for(unsigned int idir = 1; idir < track->NTrajectoryPoints()-1; ++idir){
	  // get the scattering angle
	  TVector3 initDir = track->TrajectoryPoint(idir) - track->TrajectoryPoint(idir-1);
	  TVector3 currDir = track->TrajectoryPoint(idir+1) - track->TrajectoryPoint(idir);
          double cosScat = initDir.Dot(currDir)/(initDir.Mag()*currDir.Mag());

	  // get position where this scatter occurs
	  double lastdist = fabs(lengths[idir] - lengths[idir-1]);

          pid.AddScatter(track->TrajectoryPoint(idir),cosScat);

          double scatVar = TMath::ACos(cosScat)*TMath::ACos(cosScat)/lastdist;

	  double distFromEnd = lengths[idir];

	  // get the muon likelihood of this scat at this position
	  bool useMu = true;
	  int xbin = 0;
	  int scatbin = 0;
	  Get2DHistoBins(ScatVxMuonSample,distFromEnd,scatVar,xbin,scatbin);
	  double prob = ScatVxMuonSample->GetBinContent(xbin,scatbin);
	  if(useMu){
	    if(prob <= 0){
	      // find the normalization of this distance bin
	      double norm = fmuonscatnorm[xbin-1];
	      prob = 1/norm;
	    }
	    llscatmuon+=TMath::Log(prob);
	    ++nllscatmuon;
	  }
	  
	  // get the pion likelihood of this scat at this position
	  bool usePi = true;
	  xbin = 0;
	  scatbin = 0;
	  Get2DHistoBins(ScatVxPionSample,distFromEnd,scatVar,xbin,scatbin);
	  prob = ScatVxPionSample->GetBinContent(xbin,scatbin);
	  if(usePi){
	    if(prob <= 0){
	      // find the normalization of this distance bin
	      double norm = fmuonscatnorm[xbin-1];
	      prob = 1/norm;
	    }
	    llscatpion+=TMath::Log(prob);
	    ++nllscatpion;
	  }
        }

	// Set the Muon and the Pion Log-Likelihoods
	// First default values
	pid.SetMuonDedxLL(-100);
        pid.SetPionDedxLL(-100);
	pid.SetMuonScatLL(0);
	pid.SetPionScatLL(0);

	if(nllmuon > 0 && nllpion > 0 && nllscatmuon > 0 && nllscatpion > 0){
	  // If non-zero measurement planes, then get real values of Log-Likelihoods
	  pid.SetMuonDedxLL(llmuon/nllmuon);
	  pid.SetPionDedxLL(llpion/nllpion);
	  pid.SetMuonScatLL(llscatmuon/nllscatmuon);
	  pid.SetPionScatLL(llscatpion/nllscatpion);
	  	  
	  LLScat = pid.ScatSeparation();
	  LLE = pid.DedxSeparation();
	  measfrac = ((float)pid.NMeasurementPlanes())/((float)pid.NMeasurementPlanes()+(float)pid.NVertexPlanes());

	  // get the kNN output value for this track
	  std::vector<float> knnInput(4);
	  knnInput[0] = trklen;
	  knnInput[1] = LLE;
	  knnInput[2] = LLScat;
	  knnInput[3] = measfrac;
	  
	  // get pid value
	  double kNN = reader->EvaluateMVA(knnInput,"KNN method");
	  pid.SetVal(kNN);
	}

        pidcol->push_back(pid);
        util::CreateAssn(*this,evt,*(pidcol.get()),track,*(assncol.get()));

      } // end of loop over tracks on slice
    } // end of loop over slice
      
    evt.put(std::move(pidcol));
    evt.put(std::move(assncol));
  }

  //..........................................................................
  bool  RecoMuon::IsTrackContained(art::Ptr<rb::Track> track)
  {

    // only check end of track containment, asssume uncontained until proven otherwise
    bool stopcontained = false;

    art::ServiceHandle<geo::Geometry> geom;

    // FD track end containment file
    if(geom->DetId() == novadaq::cnv::kFARDET){
      double xmin,xmax,ymin,ymax,zmin,zmax;
      double voffset = 25.0;
      xmin = -fDetWidth+voffset;
      xmax = fDetWidth-voffset;
      ymin = -fDetHeight+voffset;
      ymax = fDetHeight-voffset;
      zmin = 0+voffset;
      zmax = fDetLength-voffset;
      TVector3 mins(xmin,ymin,zmin);
      TVector3 maxs(xmax,ymax,zmax);

      // track stop containment
      TVector3 stop = track->Stop();
      // if outside of detector bounds, not contained
      if(!(stop.X() > maxs.X() || stop.X()< mins.X() ||
	   stop.Y() > maxs.Y() || stop.Y()< mins.Y() ||
	   stop.Z() > maxs.Z() || stop.Z()< mins.Z())){
	if(track->NTrajectoryPoints() > 2){
	  double end[3] = {stop.X(), stop.Y(), stop.Z()};
	  TVector3 traj = track->TrajectoryPoint(track->NTrajectoryPoints()-2);
	  TVector3 enddir = (stop-traj).Unit();
	  double dir[3] = {enddir.X(),enddir.Y(),enddir.Z()};
	  double edge[3];
	  geo::ProjectToBoxEdge(end,dir,-fDetWidth,fDetWidth,-fDetHeight,fDetHeight,0.0,fDetLength,edge);
	  // check distance from track end to detector edge
	  TVector3 projedge(edge[0],edge[1],edge[2]);
	  double distToEdge = (projedge-stop).Mag();
	  if(distToEdge > 50){stopcontained = true; }
	}
      }
    }
    else if(geom->DetId() == novadaq::cnv::kNEARDET){
      // ND track end containment
      TVector3 stop = track->Stop();
      double xmin,xmax,ymin,ymax,zmin,zmax;
      double voffset = 25.0;
      // check if z position places us in the muon catcher
      double muoncxyz[3];
      geom->Plane(geom->FirstPlaneInMuonCatcher()-1)->Cell(0)->GetCenter(muoncxyz);
      if(stop.Z() > muoncxyz[2]){
	// stop is in the muon catcher
	voffset = 50.0;
	xmin = -fDetWidth+voffset;
	xmax = fDetWidth-voffset;
	ymin = -fDetHeight+voffset;
	ymax = (fDetHeight/3.0)-voffset;
	zmin = 0+voffset;
	zmax = fDetLength-voffset;
      }
      else{
	// end is not in the muon catcher
	xmin = -fDetWidth+voffset;
	xmax = fDetWidth-voffset;
	ymin = -fDetHeight+voffset;
	ymax = fDetHeight-voffset;
	zmin = 0+voffset;
	zmax = fDetLength-voffset;
      }
      TVector3 mins(xmin,ymin,zmin);
      TVector3 maxs(xmax,ymax,zmax);
      // if outside of detector bounds, not contained
      if(!(stop.X() > maxs.X() || stop.X()< mins.X() ||
	   stop.Y() > maxs.Y() || stop.Y()< mins.Y() ||
	   stop.Z() > maxs.Z() || stop.Z()< mins.Z())){
	if(track->NTrajectoryPoints() > 2){
	  double end[3] = {stop.X(), stop.Y(), stop.Z()};
	  TVector3 traj = track->TrajectoryPoint(track->NTrajectoryPoints()-2);
	  TVector3 enddir = (stop-traj).Unit();
	  double dir[3] = {enddir.X(),enddir.Y(),enddir.Z()};
	  double edge[3];
	  if(stop.Z() < muoncxyz[2]){ geo::ProjectToBoxEdge(end,dir,-fDetWidth,fDetWidth,-fDetHeight,fDetHeight,0.0,muoncxyz[2],edge); }
	  else{ geo::ProjectToBoxEdge(end,dir,-fDetWidth,fDetWidth,-fDetHeight,fDetHeight,0.0,fDetLength,edge); }
	  // check distance from track end to detector edge
	  TVector3 projedge(edge[0],edge[1],edge[2]);
	  double distToEdge = (projedge-stop).Mag();
	  if(distToEdge > 50){stopcontained = true; }
	}
      }
    }

    return stopcontained;
  } // End of IsTrackContained function

  //..........................................................................

  void  RecoMuon::Get2DHistoBins(TH2D* histogram, double xval, double yval, int &xbin, int &ybin)
  {

    // get the x bin for the histogram
    TAxis* xaxis = histogram->GetXaxis();
    if(xval < xaxis->GetXmin()){ xbin = 1; }
    else if(xval > xaxis->GetXmax()){ xbin = histogram->GetNbinsX(); }
    else{ xbin = xaxis->FindFixBin(xval); }

    // get the y bin for the histogram
    TAxis* yaxis = histogram->GetYaxis();
    if(yval < yaxis->GetXmin()){ ybin = 1; }
    else if(yval > yaxis->GetXmax()){ ybin = histogram->GetNbinsY(); }
    else{ ybin  =  yaxis->FindFixBin(yval); }

  } // End of Get2DHistoBins

  //..........................................................................
  
  DEFINE_ART_MODULE(RecoMuon);

}
