/////////////////////////////////////////////////////////////////////////////////////
/// \brief  TODO
/// \author ??
/// \date
/////////////////////////////////////////////////////////////////////////////////////

#include "ReadoutSim/FPGAAlgorithms.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib/exception.h"
#include "cetlib/search_path.h"
#include "CLHEP/Random/RandomEngine.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "art/Framework/Services/Optional/TFileService.h"

#include "DAQDataFormats/NanoSliceVersionConvention.h"
#include "Geometry/Geometry.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RunHistory/service/RunHistoryService.h"
#include "Utilities/func/RandHisto.h"

#include "sys/stat.h"

#include <algorithm>
#include <limits>

#include "TSystem.h"
#include "TFile.h"

namespace rsim
{
  IFPGAAlgorithm::IFPGAAlgorithm(const fhicl::ParameterSet& pset) :
    CommonParameters(pset)
    ,fThresholdADC(0)
    ,fThresholdPDF(0)
    ,fThresholds(0)
    ,fPlane(0)
    ,fCell(0)
  {
    //cannot get thresholds from the database and randomly
    if (fUseRandomThresholds && fUseDBThresholds){
      throw cet::exception("Configuration") << "FPGAAlgorithm: UseRandomThreshold and UseDBThreshold cannot both be true \n";
    }

    //fetch the threshold PDF if we are setup to use random thresholds
    if (fUseRandomThresholds){
      // constructor decides if initialized value is a path or an environment variable
      cet::search_path sp("FW_SEARCH_PATH");

      sp.find_file(pset.get<std::string>("ThresholdFile"), fThresholdFile);

      struct stat sb;
      if (fThresholdFile.empty() || stat(fThresholdFile.c_str(), &sb) != 0){
	// failed to resolve the file name
	throw cet::exception("FileNotOpen") << "FPGAAlgorithms: Unable to find " << fThresholdFile << " with the threshold PDF \n "
					    <<__FILE__ << ":" << __LINE__ << "\n";
      }
	
      if (gSystem->AccessPathName(fThresholdFile.c_str())){
	throw cet::exception("FileNotOpen") << "FPGAAlgorithms: Unable to open " << fThresholdFile << " to get the threshold PDF \n";
      }

      TFile f(fThresholdFile.c_str());
      TH1D *htemp = (TH1D*)f.Get("thresholds");
      if (!htemp){
	throw cet::exception("FatalRootError") << "FPGAAlgorithms: Unable to get thresholds histogram from " << fThresholdFile << " \n";
      }
      fThresholdPDF = (TH1D*)htemp->Clone("ThresholdPDF");
      fThresholdPDF->SetDirectory(0);
      f.Close();
    }
  }

  IFPGAAlgorithm::~IFPGAAlgorithm()
  {
    if (fThresholdPDF) delete fThresholdPDF;
  }
  
  FPGA_DualCorrelatedSampling::
  FPGA_DualCorrelatedSampling(const fhicl::ParameterSet& pset) :
    IFPGAAlgorithm(pset),
    fLookbackSamples   (pset.get<int   >("FPGA_DCS_LookbackSamples")),
    fRetriggerThreshold(pset.get<double>("FPGA_DCS_RetriggerThreshold")),
    fVersion           (pset.get<int   >("FPGA_DCS_NanosliceVersion"))
  {
    fThresholdADC = pset.get<double>("FPGA_DCS_ThresholdADC");
  }

  FPGA_MatchedFiltering::
  FPGA_MatchedFiltering(const fhicl::ParameterSet& pset) :
    IFPGAAlgorithm(pset),
    fBaselineTime(pset.get<double>("FPGA_MatchedFiltering_BaselineTime")),
    fEndTime     (pset.get<double>("FPGA_MatchedFiltering_EndTime"))
  {
    fThresholdADC = pset.get<double>("FPGA_MatchedFiltering_ThresholdMatchValue");
  }

  FPGA_DumpASICOutput::FPGA_DumpASICOutput(const fhicl::ParameterSet& pset) :
    IFPGAAlgorithm(pset)
  {
    fUseDBThresholds = false;
    fThresholdADC    = 0;
  }
	   

  double IFPGAAlgorithm::ThresholdADC() const
  {
    //should correctly not  use the database for DumpASICOutput
    if (!(fUseDBThresholds || fUseRandomThresholds)) return fThresholdADC;

    return fThresholdMap[fPlane][fCell];
  }

  void IFPGAAlgorithm::FetchThresholds()
  {
    fThresholdMap.clear();

    art::ServiceHandle<geo::Geometry> geom;

    if (fUseRandomThresholds){
      // get the RandomNumberGenerator service
      art::ServiceHandle<art::RandomNumberGenerator> rng;
      CLHEP::HepRandomEngine &engine = rng->getEngine();
      RandHisto randHisto(engine);
      double thresh(0.0), dummy1(0.0), dummy2(0.0);

      for (unsigned int plane = 0; plane < geom->NPlanes(); plane++){
	std::vector< int > plane_thresholds;
	for (unsigned int cell = 0; cell < geom->Plane(plane)->Ncells(); cell++){
	  randHisto.GetRandom( fThresholdPDF, thresh, dummy1, dummy2 );
	  plane_thresholds.push_back(thresh);
	}
	fThresholdMap.push_back(plane_thresholds);
      }
    }
    else if (fUseDBThresholds){
      art::ServiceHandle<nova::dbi::RunHistoryService> RH;

      mf::LogInfo("ThresholdDB") << "Fetching thresholds for run " << RH->RunNumber();
	
      if (!RH->LoadDAQRunHistory()){
	throw cet::exception("DatabaseFailure")  << "RunHistory::LoadDAQRunHistory failed";
      }
      if (!RH->LoadPixelInfo()){
	throw cet::exception("DatabaseFailure")  << "RunHistory::LoadPixelInfo failed";
      }
      if (!RH->LoadAPDSettings()){
	throw cet::exception("DatabaseFailure")  << "RunHistory::LoadAPDSettings failed";
      }
      if (!RH->LoadHardwareInfo()){
	throw cet::exception("DatabaseFailure")  << "RunHistory::LoadHardwareInfo failed";
      }
	
      for (unsigned int plane = 0; plane < geom->NPlanes(); plane++) {
	std::vector< int > plane_thresholds;
	for (unsigned int cell = 0; cell < geom->Plane(plane)->Ncells(); cell++){
	  bool ok(false);
	  geo::OfflineChan oChan(plane, cell);
	  int threshold = RH->GetThreshold(oChan, ok);
	  //only look at channels with threshold >= 30 because of problems with the noise model
	  if (ok && threshold >= 30) plane_thresholds.push_back( threshold );
	  else plane_thresholds.push_back(4095);
	}
	fThresholdMap.push_back(plane_thresholds);
      }
    }

    if (!fThresholds){
      art::ServiceHandle<art::TFileService> tfs;
      fThresholds = tfs->make<TH1D>("fThresholds", ";Thresholds", 501, 0, 500);
      for (unsigned int i = 0; i < fThresholdMap.size(); ++i){
	for (unsigned int j = 0; j < fThresholdMap[i].size(); ++j){
	  fThresholds->Fill( fThresholdMap[i][j] );
	}
      }
    }

    return;
  }


  IFPGAAlgorithm* CreateFPGAAlgorithm(const fhicl::ParameterSet& pset)
  {
    const std::string algo = pset.get<std::string>("FPGAAlgorithm");


    if(algo == "DualCorrelatedSampling")
      return new FPGA_DualCorrelatedSampling(pset);
    if(algo == "FPGA_MatchedFiltering")
      return new FPGA_MatchedFiltering(pset);
    if(algo == "DumpASICOutput")
      return new FPGA_DumpASICOutput(pset);


    assert(0 && "Unknown FPGA algorithm");
  }


  //-------------- DualCorrelatedSampling --------------
  std::vector<rawdata::RawDigit> FPGA_DualCorrelatedSampling::
  ASICToDigits(double *ASIC_Output,
	       int firstDigitizationClocktick)
  {
    std::vector<rawdata::RawDigit> digitlist(0);

    const double threshold = ThresholdADC();

    static daqdataformats::NanoSliceVersionConvention conv;

    const int nSamples = conv.getNSamples(fVersion);
    const int nPretrig = conv.getNPretriggeredSamples(fVersion);

    // Fill DCS space with the differences, skipping over one each time:
    //   Digitizations:  Z0      Z1      Z2      ...
    //   DCS space:      Z2-Z0   Z3-Z1   Z4-Z2   ...
    //
    // Look through DCS space searching for above-threshold values.  When you
    // find one, find the biggest DCS value that occurs and return that as a
    // hit (with the time from the first >thresh point and the amplitude from
    // the peak.)

    bool   triggered = false;
    double peakDCSValue = 0;
    int    peakDCSTime  = 0;

    double DCS_Space[fNumClockticksInSpill];

    // Make sure neither the lookback (which actually looks forwards from the
    // baseline to the peak the way we have it set up) nor the pretrig or any
    // postrig samples can overflow the array bounds. This does cause us to
    // miss hits right at the edge of the trigger window, which doesn't happen
    // in real data. Should be a sub-percent effect on the overall rate, at the
    // edges where hits can be legitimately cut off too.
    const int iMin = firstDigitizationClocktick + std::max(nPretrig-fLookbackSamples, 0)*fNumClockticksPerDigitization;
    const int iMax = fNumClockticksInSpill - (nSamples-nPretrig+fLookbackSamples-1)*fNumClockticksPerDigitization;

    for (int iClocktick = iMin; iClocktick < iMax; iClocktick += fNumClockticksPerDigitization) {

      DCS_Space[iClocktick] = ASIC_Output[iClocktick + fLookbackSamples*fNumClockticksPerDigitization] - ASIC_Output[iClocktick];

      // are we triggered?
      double value = DCS_Space[iClocktick];

      if (value > threshold) {
	triggered = true;
      }
      else if (triggered && value < fRetriggerThreshold*threshold) {
	// End of triggered period declared when back below retrigger
	// threshold.  Report what we found and reset.

	rawdata::RawDigit digit;
	digit.SetVersion(fVersion);

	digit.SetTDC(peakDCSTime<<2); // shift by two bits to match real DAQ

	if(fVersion == 0){
	  // Version zero is a single DCS value
	  digit.SetADC(0, peakDCSValue);
	}
	else{
	  // Other version are multiple raw values
	  const int dt = fNumClockticksPerDigitization;
	  for(int i = fLookbackSamples-nPretrig; i < nSamples+fLookbackSamples-nPretrig; ++i){
	    const int idx = peakDCSTime+i*dt;
	    // Check we got iMin and iMax correct above
	    assert(idx >= 0 && idx < fNumClockticksInSpill);
	    digit.SetADC(digit.NADC(), ASIC_Output[idx]);
	  }
	}

	digitlist.push_back(digit);

	triggered = false;
	peakDCSValue = 0;
	peakDCSTime  = 0;
      }

      // looking for peak?
      if (triggered) {
	if (value > peakDCSValue) {
	  peakDCSTime  = iClocktick;
	  peakDCSValue = value;
	}
      }

    }

    return digitlist;
  }


  //-------------- MatchedFiltering --------------
  std::vector<rawdata::RawDigit> FPGA_MatchedFiltering::
  ASICToDigits(double *ASIC_Output,
	       int firstDigitizationClocktick)
  {
    std::vector<rawdata::RawDigit> digitlist(0);
    rawdata::RawDigit digit;

    const double threshold = ThresholdADC();

    // Convert the algo parameters into clocktick units (rounded to the nearest
    // numClockticksPerDigitization) or into ADC counts per DCS gap (from PE/ns)
    //
    // The clocktick numbers here;
    //   BaselineWindow: Number of clockticks for baseline averaging
    //   MatchingWindow: Number of clockticks for matching calculation

    double MatchStartTime          = 0; // Since f(t)=0 for t<0, no reason to include t<0 in sum
    int    MatchStartTimeFineticks = int(MatchStartTime / fClocktick);

    int BaselineWindow =
      fNumClockticksPerDigitization *
      std::max( 1, int( (MatchStartTime - fBaselineTime)
			    / fClocktick / fNumClockticksPerDigitization) );
    int MatchingWindow =
      fNumClockticksPerDigitization *
      std::max( 2, int( (fEndTime - MatchStartTime)
			    / fClocktick / fNumClockticksPerDigitization) );


    // Matched filtering provides precise time determination, but it doesn't
    // help with the magnitude of the hit.  We'll need to use the ASIC output
    // directly to get that.
    //
    // Loop over the relevant digitizations for this channel

    bool   triggered = false;
    double peakMatchValue = 0;
    int    peakMatchTime  = 0;

    // RBP debug bool dumpit = 0;
    // RBP debug double tmpvalue[fNumClockticksInSpill];
    // RBP debug for (int ii=0;ii<fNumClockticksInSpill;ii++) tmpvalue[ii] = 0;

    for (int iClocktick = firstDigitizationClocktick;
	 iClocktick < fNumClockticksInSpill - BaselineWindow - MatchingWindow; // ensure there's room left in the window (maybe change this later)
	 iClocktick += fNumClockticksPerDigitization) {

      // The first BaselineWindow ticks are devoted to calculating the baseline
      // average
      double baseline = 0;
      int baseline_count = 0;
      for (int iBaselinetick = 0;
	   iBaselinetick < BaselineWindow;
	   iBaselinetick += fNumClockticksPerDigitization) {
	baseline += ASIC_Output[iClocktick+iBaselinetick];
	baseline_count++;
      }
      baseline /= baseline_count; // With what precision can the FPGA do this?
      // Also, not checking for div-by-0 (shouldn't be possible).

      // Loop through possible relative alignments of the test signal.
      // There is no real reason this has to have anything to do with
      // the clock.  This finer looping can be on any scale.  But, this
      // is how it was described to me.
      for (int iFinetick = 0;
	   iFinetick < fNumClockticksPerDigitization;
	   iFinetick++) {

	// Calculate the match value (inner product).  Use the
	// PerfectASICCurve function, although the ASIC output
	// generation and this FPGA algorithm would, in general,
	// use different functions (since the *real* ASIC could
	// behave non-ideally.)
	double value = 0;
	double norm1 = 0;
	double norm2 = 0;
	int    count = 0;
	for (int iM=0; iM<MatchingWindow; iM+=fNumClockticksPerDigitization) {
	  double y1 = (ASIC_Output[iClocktick + BaselineWindow + iM] - baseline);
	  double y2 = PerfectASICCurve(fClocktick * (iM - iFinetick) + MatchStartTime);
	  value += y1 * y2;
	  norm1 += y1 * y1;
	  norm2 += y2 * y2;
	  count++;
	}

	// Remove the effect of different numbers of digitizations used
	// and simultaneously make the value represent a simple scaling of
	// the input:
	value /= norm2;

	// RBP debug int tmptime  = iClocktick+BaselineWindow+iFinetick-MatchStartTimeFineticks;
	// RBP debug if (tmptime<fNumClockticksInSpill&&tmptime>=0) {
	// RBP debug   tmpvalue[tmptime] = value;
	// RBP debug }
	// RBP debug else {
	// RBP debug   cout << "RBP - tmptime = " << tmptime << endl;
	// RBP debug }

	// are we triggered?
	if (value > threshold) {
	  triggered = true;
	}
	else if (triggered && value < 0.67*threshold) {
	  // End of triggered period declared when back below two-thirds
	  // threshold.  Report what we found and reset.  The "0" says to put
	  // the ADC/TDC info into the first element of the RawDigit vector.
	  // The vector nature of ADC and TDC in RawDigit isn't used at the
	  // moment.  It could be used to store waveforms.
	  // RBP debug if (1) {
	  // RBP debug std::cout << "==== " << peakMatchTime << endl; RBP debug
	  // dumpit = 1; RBP debug }
	  digit.SetADC(0, uint16_t(peakMatchValue));
	  digit.SetTDC((unsigned int)peakMatchTime<<2); // shift by two bits to match real DAQ
	  digitlist.push_back(digit);
	  triggered = false;
	  peakMatchValue = 0;
	  peakMatchTime  = 0;
	}

	// looking for peak?
	if (triggered) {
	  if (value > peakMatchValue) {
	    // For time, subtract MatchStartTimeFineticks so that
	    // MatchStartTime doesn't appear as a global shift.  (Wouldn't
	    // really matter, but might be less confusing this way.)  Also add
	    // in the baseline window to get the peak more or less in line with
	    // the ASIC output's ramp up.
	    peakMatchTime  = iClocktick+BaselineWindow+iFinetick-MatchStartTimeFineticks;
	    peakMatchValue = value;
	  }
	}

      }
    }

    // RBP debug // RBP
    // RBP debug if (dumpit&&0) {
    // RBP debug   FPGA_DumpASICOutput(ASIC_Output);
    // RBP debug   cout << "++" << endl;
    // RBP debug   FPGA_DumpASICOutput(&tmpvalue[0]);
    // RBP debug }


    return digitlist;
  }

  double FPGA_MatchedFiltering::PerfectASICCurve(double t) const
  {
    // TODO: This is copy-and-pasted from SimpleReadout, which sucks

    double R = fASICRiseTime;
    double F = fASICFallTime;

    static double peak = 0;
    if(peak == 0) {
      peak = pow(R/(R+F),R/F)-pow(R/(R+F),1+R/F);
    }

    if (t<0) {
      return 0;
    }
    else {
      return exp(-t/F)*(1-exp(-t/R))/peak;
    }
  }


  //-------------- DumpASICOutput (for debug) --------------
  std::vector<rawdata::RawDigit> FPGA_DumpASICOutput::
  ASICToDigits(double *ASIC_Output,
	       int firstDigitizationClocktick)
  {
    std::vector<rawdata::RawDigit> digitlist(0);
    for (int iClocktick=0; iClocktick<fNumClockticksInSpill; iClocktick++) {
      double tickTime = iClocktick*fClocktick;
      mf::LogDebug("SimpleReadout") << tickTime << " " << ASIC_Output[iClocktick];
    }

    return digitlist;
  }

} // end namespace
