////////////////////////////////////////////////////////////////////////
/// \brief   Code to make electronics channel noise
///
/// \author  rbpatter@caltech.edu
/// \date
////////////////////////////////////////////////////////////////////////
#include "ReadoutSim/NoiseMaker.h"

#include "NovaDAQConventions/DAQConventions.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "CLHEP/Random/RandGaussQ.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "TMath.h"

#include <iostream>

namespace rsim{

  NoiseMaker::NoiseMaker(const fhicl::ParameterSet& pset) :
    CommonParameters(pset),
    fNumNormSamples     (pset.get< int    >("NumNormSamples")),
    fVoltageNoiseDensity(pset.get< double >("VoltageNoiseDensity")),
    fDetectorCapacitance(pset.get< double >("DetectorCapacitance")),
    fLeakageCurrent     (pset.get< double >("LeakageCurrent")),
    fGain               (pset.get< double >("Gain")),
    fThresholdSigma     (pset.get< double >("ThresholdSigma")),
    fFPGA_DCS_LookbackSamples(pset.get< int >("FPGA_DCS_LookbackSamples"))
  {
    fNeedInit = true;
  }


  double NoiseMaker::GetCurrentNoise(CLHEP::RandGaussQ* gauss) {
    if (fNeedInit) Init(gauss);

    fLastCDraw = fTread/(fTfall+fTread)*gauss->fire() + fTfall/(fTfall+fTread)*fLastCDraw;

    return fLastCDraw * fCurrentNoiseScale;
  }


  double NoiseMaker::GetVoltageNoise(CLHEP::RandGaussQ* gauss) {
    if (fNeedInit) Init(gauss);

    fLastVDraw = fTread/(fTrise+fTread)*gauss->fire() + fTrise/(fTrise+fTread)*fLastVDraw;

    return fLastVDraw * fVoltageNoiseScale;
  }


  void NoiseMaker::Init(CLHEP::RandGaussQ* gauss) {

    fNeedInit = false;

    // Draw a bunch of noise, apply DCS to it, get RMS, and use
    // that to set the scale factors.  See doc-4401.

    mf::LogInfo("NoiseMaker") << "rsim::NoiseMaker Initializing..." << std::endl;

    fTrise = fASICRiseTime;
    fTfall = fASICFallTime;
    fTread = fClocktick*fNumClockticksPerDigitization*fFPGA_DCS_LookbackSamples;

    // Start with arbitrary scales of 1
    fVoltageNoiseScale = 1;
    fCurrentNoiseScale = 1;

    fLastVDraw = 0;
    fLastCDraw = 0;

    std::vector<double> rawtrace(0);
    std::vector<double> dcstrace(0);

    const double Qe = 1.6022E-10; // e in nC

    // a "leakage rate", of sorts, in GHz:
    double I_L =
      fLeakageCurrent / Qe
      * fGain * fAPDExcessNoiseFactor * 1E-9; // last factor for 1/s-->1/ns

    // --- first, current noise:
    rawtrace.clear();
    for (int iS=0;iS<fNumNormSamples;iS++) {
      rawtrace.push_back(GetCurrentNoise(gauss));
    }
    ApplyDCS(rawtrace,dcstrace);
    double rms = GetRMS(dcstrace);
    double enc = TMath::Sqrt(I_L*(2*fTread - fTrise));
    mf::LogInfo("NoiseMaker") << "rsim::NoiseMaker: equivalent noise charge (current) = " << enc << " electrons" << std::endl;
    fCurrentNoiseScale = enc / rms / fGain;

    // --- then, voltage noise:
    rawtrace.clear();
    for (int iS=0;iS<fNumNormSamples;iS++) {
      rawtrace.push_back(GetVoltageNoise(gauss));
    }
    ApplyDCS(rawtrace,dcstrace);
    rms = GetRMS(dcstrace);
    enc =
      fVoltageNoiseDensity * fDetectorCapacitance
      / Qe / TMath::Sqrt(2*fTrise*1E-9) * 1E-12; // last factor for pF-->F
    mf::LogInfo("NoiseMaker") << "rsim::NoiseMaker: equivalent noise charge (voltage) = " << enc << " electrons" << std::endl;
    fVoltageNoiseScale = enc / rms / fGain;

    // --- then, extract nominal threshold
    rawtrace.clear();
    for (int iS=0; iS<fNumNormSamples;iS++) {
      rawtrace.push_back(GetNoise(gauss)/fADCMaxPE*( (1<<12) - 1 ));
    }
    ApplyDCS(rawtrace,dcstrace);
    rms = GetRMS(dcstrace);
    fNominalThreshold = fThresholdSigma*rms;
    mf::LogInfo("NoiseMaker") << "rsim::NoiseMaker: nominal threshold (" << fThresholdSigma << "*sigma_dcs) = " << fNominalThreshold << std::endl;
  }


  void NoiseMaker::ApplyDCS(std::vector<double> &rawtrace, std::vector<double> &dcstrace) {
    dcstrace.clear();
    int N = rawtrace.size();
    for (int iS=fFPGA_DCS_LookbackSamples;iS<N;iS++) {
      dcstrace.push_back(rawtrace[iS] - rawtrace[iS-fFPGA_DCS_LookbackSamples]);
    }
    return;
  }


  double NoiseMaker::GetRMS(std::vector<double> &dcstrace) {
    int N = dcstrace.size();
    double sum = 0;
    double sumsq = 0;
    for (int iS=0;iS<N;iS++) {
      sum   += dcstrace[iS];
      sumsq += dcstrace[iS]*dcstrace[iS];
    }
    return ( (N>0) ? TMath::Sqrt(sumsq/N - (sum/N)*(sum/N)) : 0 );
  }

}//namespace
