////////////////////////////////////////////////////////////////////////
// \file    HitMap.cxx
// \brief   Efficient lookup of hits by plane and cell number
// \version $Id: HitMap.cxx,v 1.7 2012-05-11 00:39:30 bckhouse Exp $
// \author $Author: bckhouse $
////////////////////////////////////////////////////////////////////////
#include "RecoBase/HitMap.h"

#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"

namespace rb
{
  //------------------------------------------------------------
  HitMap::HitMap()
  {
  }

  //------------------------------------------------------------
  HitMap::HitMap(const art::PtrVector<rb::CellHit>& chits)
  {
    Add(chits);
  }

  //------------------------------------------------------------
  HitMap::HitMap(const rb::Cluster* clust)
  {
    Add(clust);
  }
  
  //------------------------------------------------------------
  HitMap::~HitMap()
  {
  }
  
  //------------------------------------------------------------
  void HitMap::Add(const art::Ptr<rb::CellHit>& chit)
  {
    const geo::OfflineChan chan(chit->Plane(), chit->Cell());

    typedef std::map<geo::OfflineChan, art::Ptr<rb::CellHit> >::iterator it_t;
    it_t it = fMap.find(chan);

    if(it == fMap.end()){
      // If there's no hit on this channel, store this one there
      fMap.insert(std::make_pair(chan, chit));
    }
    else{
      // If there already was a hit, highest energy wins
      if(chit->PE() > it->second->PE()){
        it->second = chit;
      }
      else{
        // Discard this hit
      }
    }
  }

  //------------------------------------------------------------
  void HitMap::Add(const art::PtrVector<rb::CellHit>& chits)
  {
    const unsigned int N = chits.size();
    for(unsigned int n = 0; n < N; ++n) Add(chits[n]);
  }
  
  //------------------------------------------------------------
  void HitMap::Add(const rb::Cluster* clust)
  {
    const int maxIdx = clust->NCell();
    for(int idx = 0; idx < maxIdx; ++idx)
      Add(clust->Cell(idx));
  }

  //------------------------------------------------------------
  void HitMap::Remove(unsigned int plane, unsigned int cell)
  {
    fMap.erase(geo::OfflineChan(plane, cell));
  }

  //------------------------------------------------------------
  bool HitMap::CellExists(unsigned int plane,
			  unsigned int cell) const
  {
    return fMap.find(geo::OfflineChan(plane, cell)) != fMap.end();
  }
  
  //------------------------------------------------------------
  art::Ptr<rb::CellHit> HitMap::Cell(unsigned int plane,
				     unsigned int cell) const
  {
    std::map<geo::OfflineChan, art::Ptr<rb::CellHit> >::const_iterator it =
      fMap.find(geo::OfflineChan(plane, cell));
    
    if(it == fMap.end()) return art::Ptr<rb::CellHit>();
    
    return it->second;
  }
  
  //------------------------------------------------------------
  art::Ptr<rb::CellHit> HitMap::operator()(unsigned int planeIdx,
					   unsigned int cellIdx) const
  {
    return Cell(planeIdx, cellIdx);
  }

} // end namespace rb
//////////////////////////////////////////////////////////////////////
