////////////////////////////////////////////////////////////////////////
// \file    RecoHit.h
// \brief   A RecoHit contains information about the hit that requires some level
//          of reconstruction.  In particular, the location of the hit in
//          the orthogonal view is used to set this stuff.
// \version $Id: RecoHit.h,v 1.6 2012-03-07 22:13:19 bckhouse Exp $
// \author  $Author: bckhouse $
////////////////////////////////////////////////////////////////////////
#ifndef RBRECOHIT_H
#define RBRECOHIT_H

#include <ostream>

namespace rb
{  
  /// \brief Calibrated quantities relying on position in the orthogonal view.
  ///        To generate a \ref rb::CellHit from a \ref rb::RecoHit you need a 
  ///        reconstructed object to provide the assumed W position.
  class RecoHit
  {
  public:
    RecoHit();  // Default constructor
    ~RecoHit(); 

#ifndef __GCCXML__

    /// You MUST check here before accessing \ref PECorr, \ref MIP or \ref GeV
    bool IsCalibrated() const;

    void Print() const;
    
    // Getters
    float PECorr() const;
    float MIP   () const;
    float GeV   () const;
    float X     () const { return fX;      } 
    float Y     () const { return fY;      }
    float Z     () const { return fZ;      }
    float T     () const { return fT;      }
    
    // Setters
    void SetPECorr(float PECorr) { fPECorr = PECorr; }
    void SetMIP   (float MIP   ) { fMIP   = MIP;     }
    void SetGeV   (float GeV   ) { fGeV   = GeV;     }
    void SetX     (float X     ) { fX     = X;       }
    void SetY     (float Y     ) { fY     = Y;       }
    void SetZ     (float Z     ) { fZ     = Z;       }
    void SetT     (float T     ) { fT     = T;       }
    
    friend std::ostream& operator << (std::ostream& o, const RecoHit& r);
#endif // __GCCXML__
 
  protected:
    float fPECorr; ///< PE with attenuation corrections added
    float fMIP;    ///< pulseheight normalized to a MIP
    float fGeV;    ///< energy deposition represented by this hit
    float fX;	   ///< spacetime location of deposition, x (cm)
    float fY;	   ///< spacetime location of deposition, y (cm)
    float fZ;	   ///< spacetime location of deposition, z (cm)
    float fT;	   ///< spacetime location of deposition, t (ns)
  };
  
} // end namespace rb

#endif // RBRECOHIT_H

////////////////////////////////////////////////////////////////////////
