////////////////////////////////////////////////////////////////////////
///
/// \brief  Modified from ClusterMerge 
///
/// \author $Author: bianjm $
////////////////////////////////////////////////////////////////////////
#include <cmath>
#include <iostream>
#include <list>
#include <string>
#include <vector>

// ROOT includes
#include "TVector3.h"
#include "TH1D.h"
#include "TTree.h"

// ART includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Persistency/Common/PtrVector.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"  
#include "messagefacility/MessageLogger/MessageLogger.h" 

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"	
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/PlaneGeo.h"	
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Track.h"
#include "Utilities/func/MathUtil.h"



namespace jmshower { 
  class JMClusterMerge : public art::EDProducer
  {
  public:
    explicit JMClusterMerge(fhicl::ParameterSet const& pset); 
    ~JMClusterMerge(); 
    void reconfigure(const fhicl::ParameterSet& pset);
    void produce(art::Event& evt); 
    void beginJob();
    
    
  private:
    std::string pSliceDir;
    std::string fClusterLabel;
    int         fDebug;               ///< Controls debug output 0=none, 99=scream
    double      fDHitGood;            ///< Maximum distance from the line for a hit to be considered part of it  
    int         fMinHit;
    double      fGoodness;
    int         fMaxPlane;
    bool        fMCCheater;          ///< True/false MCCheater SimHit/SimTrack input
    TH1F*       fHistoGoodness;
    TH1F*       fHistoGoodnessLog; ///< Easier to read in the reco validation.
    TH1F*       fHistoEnergy;
    TH1F*       fHistoPurity;
    TH1F*       fHistoRemainingCells;
    
    TTree*      fTrack;
    int    m_trackNHit1;
    int    m_trackNHit2;
    int    m_trackNPlane1;
    int    m_trackNPlane2;
    double    m_trackChisq1;
    double    m_trackChisq2;
    double    m_trackChisq;
    double    m_trackDiffChisq;
    double    m_trackGap;
    double    m_trackDHit;
    
    int    m_trackTrueNuCCNC;
    int    m_trackTrueNuMode;
    int    m_trackTrueNuPdg;
    double m_trackTrueNuEnergy;
    int    m_trackTrueNuIsFid;
    double m_trackTrueNuP4[4];
    double m_trackTrueNuVtx[3];
    int    m_trackEvtTruePdg[20];
    double m_trackEvtTrueEnergy[20];
    double m_trackEvtTrueElecPhotons;
    
    
    double m_trackTrueDang;
    double m_trackTrueEDang;
    int m_trackTruePdg;
    int m_trackTrueId;
    double m_trackTrueVtx[3];
    double m_trackTrueP4[4];
    double m_trackTruePartHit[10];
    double m_trackTruePartEnergy[10];
    double m_trackTrueDepEnergy;
    double m_trackTrueElecPhotons;
    double m_trackTrueTotPhotons;
    double m_trackTrueTrkElecPhotons;
    double m_trackTrueTrkTotPhotons;
    double m_trackTrueRWElecPhotons;
    double m_trackTrueRWTotPhotons;
    double m_trackTrueTrkRWElecPhotons;
    double m_trackTrueTrkRWTotPhotons;
        
  };
}


namespace jmshower{

  static bool hitSort(art::Ptr<rb::Cluster> c1, art::Ptr<rb::Cluster> c2)
  {
    return c1->NCell() > c2->NCell();
  }

  JMClusterMerge::JMClusterMerge(fhicl::ParameterSet const& pset)
  {
    reconfigure(pset);
    produces< std::vector<rb::Track>   >();
    
  }
  
  
  JMClusterMerge::~JMClusterMerge()
  {
  }
  
  void JMClusterMerge::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    fHistoGoodness = tfs->make<TH1F>("fHistoGoodness","Goodness of candidates for merging;Goodness;events", 1000,0.,1000.);
    fHistoGoodnessLog = tfs->make<TH1F>("fHistoGoodnessLog","Goodness of candidates for merging;Log10[Goodness+1];events", 100,0.,3.);
    fHistoEnergy = tfs->make<TH1F>("fHistoEnergy","Total Energy of Cells;PE;events", 100,10000.,20000.);
    fHistoPurity = tfs->make<TH1F>("fHistoPurity","Purity of Clusters;(Hits of true particle/total hits in cluster);events",101,0.,1.01);
    fHistoRemainingCells = tfs->make<TH1F>("fHistoRemainingCells","True hits not put into a cluster;Missing Cells;events",100,0.,100.);

    fTrack = tfs->make<TTree>("fTrack","fTrack");

    fTrack->Branch("trackNHit1", &m_trackNHit1);
    fTrack->Branch("trackNHit2",&m_trackNHit2);;
    fTrack->Branch("trackNPlane1",&m_trackNPlane1);
    fTrack->Branch("trackNPlane2",&m_trackNPlane2);
    fTrack->Branch("trackChisq1",&m_trackChisq1);
    fTrack->Branch("trackChisq2",&m_trackChisq2);
    fTrack->Branch("trackChisq",&m_trackChisq);
    fTrack->Branch("trackDiffChisq",&m_trackDiffChisq);
    fTrack->Branch("trackGap",&m_trackGap);
    fTrack->Branch("trackDHit",&m_trackDHit);
    
  }
  
  
  void JMClusterMerge::reconfigure(const fhicl::ParameterSet& pset)
  {
    pSliceDir      = pset.get< std::string >("SliceDir" );
    fClusterLabel  = pset.get< std::string >("ClusterLabel" );
    fDebug         = pset.get< int    >("Debug"    );
    fDHitGood      = pset.get< double >("DHitGood" );	
    fMinHit        = pset.get< int    >("MinHit"   );
    fGoodness      = pset.get< double >("Goodness" );
    fMaxPlane      = pset.get< int    >("MaxPlane" );
    fMCCheater     = pset.get< bool   >("MCCheater");
  }
  
  
  //......................................................................
  
  void JMClusterMerge::produce(art::Event& evt)
  { 
    art::Handle< std::vector<rb::Cluster> > clusthandle;
    evt.getByLabel(fClusterLabel, clusthandle);
    
    std::unique_ptr< std::vector<rb::Track> > trackcol(new std::vector<rb::Track> );

    std::vector< std::vector<double> > starts;
    starts.resize(clusthandle->size());
    std::vector< std::vector<double> > ends;
    ends.resize(clusthandle->size());
    std::vector< std::vector<double> > dirs;
    dirs.resize(clusthandle->size());
    std::vector<double> sig;
    sig.resize(clusthandle->size());

    std::vector<bool> isnoise;
    isnoise.resize(clusthandle->size());

    std::vector<rb::Cluster> vCluster;
    std::vector< int > trId;
    
    unsigned int cc = 0;
    double energy = 0;
    
    //get slices  for timing information
    art::Handle<std::vector<rb::Cluster> > slicecol;
    evt.getByLabel(pSliceDir,slicecol);
    std::vector< std::pair< double, double> > tMap; //contains min and max time for each slice in event, first number is min, secont max
    for (unsigned int i=0; i<slicecol->size(); ++i) {
      const rb::Cluster& s = (*slicecol)[i];
      if (s.IsNoise()) continue;
      std::pair< double, double> p(s.MinTNS(), s.MaxTNS()); //add min and max time as a pair for a given slice
      tMap.push_back(p);  
    }
    
    for (unsigned int cv=0;cv<2;cv++) {
      geo::View_t view;
      if (cv == 0) {
	view = geo::kX;
      }
      else {
	view = geo::kY;
      }
   
      std::vector< art::Ptr<rb::Cluster> > cluster;



      for(unsigned int i = 0; i < clusthandle->size(); i++){
	art::Ptr<rb::Cluster> clust(clusthandle, i);
	if ((cv == 0) && (clust->NXCell() > 0))cluster.push_back(clust);
	if ((cv == 1) && (clust->NYCell() > 0)) cluster.push_back(clust);
      }
      
      
      std::sort(cluster.begin(), cluster.end(), hitSort);
      art::ServiceHandle<geo::Geometry> geom;   
      std::vector< std::vector< art::Ptr< rb::CellHit> > > cells(cluster.size());

      std::vector< std::vector<int> > motherids(cluster.size());
      std::vector< std::vector<int> > sonids(cluster.size());
      std::vector< unsigned int > samemotherson(cluster.size());

      for (unsigned int i=0;i<cluster.size();i++) {
        samemotherson[i]=0;
	for ( unsigned int nh=0;nh<cluster[i]->NCell();nh++) {
	  cells[i].push_back(cluster[i]->Cell(view,nh));
          energy += cluster[i]->Cell(view,nh)->PE();
	}
        if (cells[i].size() > 1) rb::SortByPlane(cells[i]);
      }

      for ( unsigned int nc=0;nc<cluster.size();nc++) {
        if(cluster[nc]->NCell()==0)continue;
	std::vector<art::Ptr< rb::CellHit> >::iterator itr = cells[nc].begin();
	art::PtrVector<rb::CellHit> mergecells;
        for(unsigned int i = 0; i < cluster[nc]->NCell(); i++){
          mergecells.push_back(cluster[nc]->Cell(view,i));
        }
	std::vector<double> xzViewX;
	std::vector<double> xzViewZ;
	std::vector<double> yzViewY;
	std::vector<double> yzViewZ;
	std::vector<double> xzViewW; // W = weight
	std::vector<double> yzViewW; // W = weight
	double xyz[3]; // xyz[0]=x, xyz[1]=y, xyz[2]=z
	int planeMax = -9999;
	int planeMin = 9999;

        //initialize min and max time for a slice window
        double minTNS = 0.;
        double maxTNS = 0.;
        //determine which slice the cluster candidate is a part of
        for( unsigned int i = 0; i < tMap.size(); ++i){
          if ((cluster[nc]->MinTNS() >= tMap[i].first) && (cluster[nc]->MaxTNS() <= tMap[i].second)){
            //setting min and max time to the size of the slice the cluster belongs to
            minTNS = tMap[i].first;
            maxTNS = tMap[i].second;
            break;
          }
        }
	
	while( itr != cells[nc].end() ){
	  double sig = 1.;
	  geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->GetCenter(xyz);
	  if((*itr)->View() == geo::kX){
	    if ((*itr)->Plane() > planeMax) planeMax = (*itr)->Plane();
	    if ((*itr)->Plane() < planeMin) planeMin = (*itr)->Plane();
	    xzViewX.push_back(xyz[0]);
	    xzViewZ.push_back(xyz[2]);
	    xzViewW.push_back(sig);
	  }
	  if((*itr)->View() == geo::kY){
	    if ((*itr)->Plane() > planeMax) planeMax = (*itr)->Plane();
	    if ((*itr)->Plane() < planeMin) planeMin = (*itr)->Plane();
	    yzViewY.push_back(xyz[1]);
	    yzViewZ.push_back(xyz[2]);
	    yzViewW.push_back(sig);
	  }
	  itr++;
	}// end loop over hits in this track
	
	double Chi2 = 0.;
	
	double xzStart[2]    = {0.,0.}; 
	double xzEnd[2]      = {0.,0.}; 
	double slopeXZ     = 0.;
	double interceptXZ = 0.;
	if (xzViewX.size() >1){
	  Chi2 = geo::LinFit(xzViewZ, xzViewX, xzViewW, &xzStart[0], 
			     &xzStart[1], &xzEnd[0], &xzEnd[1]);
	  slopeXZ     = 999999.;
	  interceptXZ = xzEnd[1] - slopeXZ*xzEnd[0];
	  if(fabs(xzStart[0] - xzEnd[0]) > 1.e-2){
	    slopeXZ     = (xzEnd[1] - xzStart[1])/(xzEnd[0] - xzStart[0]);
	    interceptXZ = xzEnd[1] - slopeXZ*xzEnd[0];
	  }
	}
	
	double yzStart[2]    = {0.,0.}; 
	double yzEnd[2]      = {0.,0.}; 
	double slopeYZ     = 0.;
	double interceptYZ = 0.;
	if (yzViewY.size() >1){
	  Chi2 = geo::LinFit(yzViewZ, yzViewY, yzViewW, &yzStart[0], 
			     &yzStart[1], &yzEnd[0], &yzEnd[1]);
	  slopeYZ     = 999999.;
	  interceptYZ = yzEnd[1] - slopeYZ*yzEnd[0];
	  if(fabs(yzStart[0] - yzEnd[0]) > 1.e-2){
	    slopeYZ     = (yzEnd[1] - yzStart[1])/(yzEnd[0] - yzStart[0]);
	    interceptYZ = yzEnd[1] - slopeYZ*yzEnd[0];
	  }
	}
	
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////
	unsigned int clustcount = cluster.size();

	for ( unsigned int numclust=0;numclust<cluster.size();numclust++) {
          if(cluster[numclust]->NCell()==0)continue;
	  if (numclust <= nc) continue;

          //check to make sure merge candidate is part of the same time slice as cluster we are trying to add
          if ((cluster[numclust]->MinTNS() < minTNS) || (cluster[numclust]->MaxTNS() > maxTNS)) continue;
	  
	  std::vector<double> xzViewXtest;
	  std::vector<double> xzViewZtest;
	  std::vector<double> yzViewYtest;
	  std::vector<double> yzViewZtest;
	  std::vector<double> xzViewWtest; // W = weight
	  std::vector<double> yzViewWtest; // W = weight
	  double xzStarttest[2]    = {0.,0.}; 
	  double xzEndtest[2]      = {0.,0.}; 
	  double yzStarttest[2]    = {0.,0.}; 
	  double yzEndtest[2]      = {0.,0.}; 
	  int tplaneMax = -9999;
	  int tplaneMin = 9999;
	  bool add = true;
	  
	  if (((int)cells[numclust].size() >= fMinHit) && ((int)cells[nc].size() >= fMinHit)){
	    itr = cells[numclust].begin();
	    while( itr != cells[numclust].end() ){
	      double sig = 1.;
	      geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->GetCenter(xyz);
	      if((*itr)->View() == geo::kX){
		if ((*itr)->Plane() > tplaneMax) tplaneMax = (*itr)->Plane();
		if ((*itr)->Plane() < tplaneMin) tplaneMin = (*itr)->Plane();
		xzViewXtest.push_back(xyz[0]);
		xzViewZtest.push_back(xyz[2]);
		xzViewWtest.push_back(sig);
	      }
	      if((*itr)->View() == geo::kY){
		if ((*itr)->Plane() > tplaneMax) tplaneMax = (*itr)->Plane();
		if ((*itr)->Plane() < tplaneMin) tplaneMin = (*itr)->Plane();
		yzViewYtest.push_back(xyz[1]);
		yzViewZtest.push_back(xyz[2]);
		yzViewWtest.push_back(sig);
	      }
	      itr++;
	    }// end loop over hits in this track
	    
	    double Chi2test = 0.;
	    
	    if (xzViewXtest.size() >1){ 
	      Chi2test = geo::LinFit(xzViewZtest, xzViewXtest, xzViewWtest, &xzStarttest[0], &xzStarttest[1], &xzEndtest[0], &xzEndtest[1]);
	    }
	    
	    if (yzViewYtest.size() >1){
	      Chi2test = geo::LinFit(yzViewZtest, yzViewYtest, yzViewWtest, &yzStarttest[0], &yzStarttest[1], &yzEndtest[0], &yzEndtest[1]);
	    }
	    
	    itr = cells[nc].begin();
	    while( itr != cells[nc].end() ){
	      double sig = 1.;
	      geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->GetCenter(xyz);
	      if((*itr)->View() == geo::kX){
		xzViewXtest.push_back(xyz[0]);
		xzViewZtest.push_back(xyz[2]);
		xzViewWtest.push_back(sig);
	      }
	      if((*itr)->View() == geo::kY){
		yzViewYtest.push_back(xyz[1]);
		yzViewZtest.push_back(xyz[2]);
		yzViewWtest.push_back(sig);
	      }
	      itr++;
	    }// end loop over hits in this track
	    
	    double Chi2merge = 0.;
	    
	    if (xzViewXtest.size() >1){
	      Chi2merge = geo::LinFit(xzViewZtest, xzViewXtest, xzViewWtest, &xzStarttest[0],
				      &xzStarttest[1], &xzEndtest[0], &xzEndtest[1]);
	    }
	    
	    if (yzViewYtest.size() >1){
	      Chi2merge = geo::LinFit(yzViewZtest, yzViewYtest, yzViewWtest, &yzStarttest[0], 
				      &yzStarttest[1], &yzEndtest[0], &yzEndtest[1]);
	    }
	    
	    double diffChi2 = (Chi2merge - Chi2 - Chi2test)/(cells[numclust].size() + 
							     cells[nc].size());

            if(cluster[nc]->NCell()+cluster[nc]->NCell()>=20){
              if(diffChi2 > 7.) add = false;
            }

            if(cluster[nc]->NCell()+cluster[nc]->NCell()<20){
              if(diffChi2 > 30.) add = false;
            }
	    
	    if (tplaneMax < planeMin) {
	      if (planeMin - tplaneMax > fMaxPlane) {add = false;}
	    }
	    if (tplaneMin > planeMax) {
	      if (tplaneMin - planeMax > fMaxPlane) {add = false;}
	    }

            for(unsigned int m1=0;m1<motherids[nc].size();m1++){
              for(unsigned int m2=0;m2<motherids[numclust].size();m2++){
                if(motherids[nc][m1]==motherids[numclust][m2]) add = false;
              }
            }


	    if (add) {

              motherids[numclust].push_back(nc);
              sonids[nc].push_back(numclust);

	      xzStart[0] = xzStarttest[0];
	      xzStart[1] = xzStarttest[1];
	      yzStart[0] = yzStarttest[0];
	      yzStart[1] = yzStarttest[1];
	      xzEnd[0]   = xzEndtest[0];
	      xzEnd[1]   = xzEndtest[1];
	      yzEnd[0]   = yzEndtest[0];
	      yzEnd[1]   = yzEndtest[1];
	      
	      for ( unsigned int nh=0;nh<cluster[numclust]->NCell();nh++) {
		cells[nc].push_back(cluster[numclust]->Cell(view,nh));
                mergecells.push_back(cluster[numclust]->Cell(view,nh));
	      }

	      rb::SortByPlane(cells[nc]);
	      rb::SortByPlane(mergecells);
	      
	      Chi2 = Chi2merge;



	      planeMax = -9999;
	      planeMin = 9999;
	      itr = cells[nc].begin();
	      while( itr != cells[nc].end() ){
		if((*itr)->View() == geo::kX){
		  if ((*itr)->Plane() > planeMax) planeMax = (*itr)->Plane();
		  if ((*itr)->Plane() < planeMin) planeMin = (*itr)->Plane();
		}
		if((*itr)->View() == geo::kY){
		  if ((*itr)->Plane() > planeMax) planeMax = (*itr)->Plane();
		  if ((*itr)->Plane() < planeMin) planeMin = (*itr)->Plane();
		}
		itr++;
	      }// end loop over hits in this track
	      
	      slopeXZ     = 999999.;
	      interceptXZ = xzEnd[1] - slopeXZ*xzEnd[0];
	      if(fabs(xzStart[0] - xzEnd[0]) > 1.e-2){
		slopeXZ     = (xzEnd[1] - xzStart[1])/(xzEnd[0] - xzStart[0]);
		interceptXZ = xzEnd[1] - slopeXZ*xzEnd[0];
	      }      
	      
	      slopeYZ     = 999999.;
	      interceptYZ = yzEnd[1] - slopeYZ*yzEnd[0];
	      if(fabs(yzStart[0] - yzEnd[0]) > 1.e-2){
		slopeYZ     = (yzEnd[1] - yzStart[1])/(yzEnd[0] - yzStart[0]);
		interceptYZ = yzEnd[1] - slopeYZ*yzEnd[0];
	      } 
	    }
	  }
	  

	  if (((int)cells[numclust].size() < fMinHit) && ((int)cells[nc].size() >= fMinHit)){

	    double dHit       = 0.;  // distance of current hit from line
	    double slope      = 0.;  // slope for the hit view
	    double intercept  = 0.;  // intercept in the hit view
	    double point      = 0.;
	    tplaneMax = -9999;
	    tplaneMin = 9999;
	    itr = cells[numclust].begin();
	    while( itr != cells[numclust].end() ){
	      geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->GetCenter(xyz);
	      if((*itr)->View() == geo::kX){
		if ((*itr)->Plane() > tplaneMax) tplaneMax = (*itr)->Plane();
		if ((*itr)->Plane() < tplaneMin) tplaneMin = (*itr)->Plane();
		slope      = slopeXZ;
		intercept  = interceptXZ;
		point      = xyz[0];

	      }
	      else if((*itr)->View() == geo::kY){
		if ((*itr)->Plane() > tplaneMax) tplaneMax = (*itr)->Plane();
		if ((*itr)->Plane() < tplaneMin) tplaneMin = (*itr)->Plane();
		slope      = slopeYZ;
		intercept  = interceptYZ;
		point      = xyz[1];
	      }
	      dHit += fabs(-slope*xyz[2] + point - intercept)/sqrt(slope*slope + 1.);
	      itr++;
	    }
	    dHit = dHit / cells[numclust].size();

            if(fabs(cluster[numclust]->MinPlane()-cluster[nc]->MinPlane())<8){
              if(dHit > 30.) add = false;
            }
            if(fabs(cluster[numclust]->MinPlane()-cluster[nc]->MinPlane())>=8){
              if(dHit > 30.) add = false;
            }
	    
	    if (tplaneMax < planeMin) {
	      if (planeMin - tplaneMax > fMaxPlane) {add = false;}
	    }
	    if (tplaneMin > planeMax) {
	      if (tplaneMin - planeMax > fMaxPlane) {add = false;}
	    }

            for(unsigned int m1=0;m1<motherids[nc].size();m1++){
              for(unsigned int m2=0;m2<motherids[numclust].size();m2++){
                if(motherids[nc][m1]==motherids[numclust][m2]) add = false; 
              }
            }


	    if (add){

              motherids[numclust].push_back(nc);
              sonids[nc].push_back(numclust);

	      for ( unsigned int nh=0;nh<cluster[numclust]->NCell();nh++) {
		cells[nc].push_back(cluster[numclust]->Cell(view,nh));
                mergecells.push_back(cluster[numclust]->Cell(view,nh));
	      }

	      rb::SortByPlane(cells[nc]);
	      rb::SortByPlane(mergecells);

	      planeMax = -9999;
	      planeMin = 9999;
	      itr = cells[nc].begin();
	      while( itr != cells[nc].end() ){
		double sig = 1.;
		geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->GetCenter(xyz);
		
		if((*itr)->View() == geo::kX){
		  if ((*itr)->Plane() > planeMax) planeMax = (*itr)->Plane();
		  if ((*itr)->Plane() < planeMin) planeMin = (*itr)->Plane();
		  xzViewXtest.push_back(xyz[0]);
		  xzViewZtest.push_back(xyz[2]);
		  xzViewWtest.push_back(sig);
		}
		if((*itr)->View() == geo::kY){
		  if ((*itr)->Plane() > planeMax) planeMax = (*itr)->Plane();
		  if ((*itr)->Plane() < planeMin) planeMin = (*itr)->Plane();
		  yzViewYtest.push_back(xyz[1]);
		  yzViewZtest.push_back(xyz[2]);
		  yzViewWtest.push_back(sig);
		}
		itr++;
	      }// end loop over hits in this track
	      
	      if (xzViewXtest.size() >1){
		Chi2 = geo::LinFit(xzViewZtest, xzViewXtest, xzViewWtest, &xzStarttest[0], &xzStarttest[1], &xzEndtest[0], &xzEndtest[1]);
	      }
	      
	      if (yzViewYtest.size() >1){
		Chi2 = geo::LinFit(yzViewZtest, yzViewYtest, yzViewWtest, &yzStarttest[0], &yzStarttest[1], &yzEndtest[0], &yzEndtest[1]);
	      }
	      
	      xzStart[0] = xzStarttest[0];
	      xzStart[1] = xzStarttest[1];
	      yzStart[0] = yzStarttest[0];
	      yzStart[1] = yzStarttest[1];
	      xzEnd[0]   = xzEndtest[0];
	      xzEnd[1]   = xzEndtest[1];
	      yzEnd[0]   = yzEndtest[0];
	      yzEnd[1]   = yzEndtest[1];
	      
	      // reset the end points based on startz and endz
	      slopeXZ     = 999999.;
	      interceptXZ = xzEnd[1] - slopeXZ*xzEnd[0];
	      if(fabs(xzStart[0] - xzEnd[0]) > 1.e-2){
		slopeXZ     = (xzEnd[1] - xzStart[1])/(xzEnd[0] - xzStart[0]);
		interceptXZ = xzEnd[1] - slopeXZ*xzEnd[0];
	      }      
	      
	      slopeYZ     = 999999.;
	      interceptYZ = yzEnd[1] - slopeYZ*yzEnd[0];
	      if(fabs(yzStart[0] - yzEnd[0]) > 1.e-2){
		slopeYZ     = (yzEnd[1] - yzStart[1])/(yzEnd[0] - yzStart[0]);
		interceptYZ = yzEnd[1] - slopeYZ*yzEnd[0];
	      }   
	    }

	    if ((numclust == (cluster.size()-1)) && (clustcount != cluster.size())){
	      numclust=0;
	      clustcount = cluster.size();
	    }

	  }


	  if ((int)cells[numclust].size() < 3 && (int)cells[nc].size() < 3 && (int)cells[numclust].size()>0 && (int)cells[nc].size()>0 ) {
	    double dHit       = 0.;  // distance of current hit from line
	    double point      = 0.;
	    tplaneMax = -9999;
	    tplaneMin = 9999;
            dHit = 9999;
	    itr = cells[numclust].begin();
	    while( itr != cells[numclust].end() ){
	      geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->GetCenter(xyz);
	      if((*itr)->View() == geo::kX){
		if ((*itr)->Plane() > tplaneMax) tplaneMax = (*itr)->Plane();
		if ((*itr)->Plane() < tplaneMin) tplaneMin = (*itr)->Plane();
		point      = xyz[0];
	      }
	      else if((*itr)->View() == geo::kY){
		if ((*itr)->Plane() > tplaneMax) tplaneMax = (*itr)->Plane();
		if ((*itr)->Plane() < tplaneMin) tplaneMin = (*itr)->Plane();
		point      = xyz[1];
	      }
              for(unsigned int ii = 0; ii<cells[nc].size(); ii++){
                double point0 = 0;
                double z0 = 0;
                if((*itr)->View() == geo::kX){
                  point0     = xzViewX[ii];
                  z0     = xzViewZ[ii];
                }else{
                  point0     = yzViewY[ii];
                  z0 = yzViewZ[ii];
                }

	        double Hitd = sqrt((point-point0)*(point-point0)+(xyz[2]-z0)*(xyz[2]-z0)); 
                if(Hitd<dHit)dHit = Hitd;
              }
	      itr++;
	    }

            int trackGap = 0; 
            if (tplaneMax <= planeMin) {
              trackGap = planeMin - tplaneMax;
            }
            if (tplaneMin > planeMax) {
              trackGap = tplaneMin - planeMax;
            }

	    if (trackGap ==0  && dHit > 15){add = false;}
	    if (trackGap !=0  && dHit > 20){add = false;} 

	    if (add){

              motherids[numclust].push_back(nc);
              sonids[nc].push_back(numclust);

              for ( unsigned int nh=0;nh<cluster[numclust]->NCell();nh++) {
                cells[nc].push_back(cluster[numclust]->Cell(view,nh));
                mergecells.push_back(cluster[numclust]->Cell(view,nh));
              }

	      rb::SortByPlane(cells[nc]);
	      rb::SortByPlane(mergecells);

	      planeMax = -9999;
	      planeMin = 9999;


	      itr = cells[nc].begin();
	      while( itr != cells[nc].end() ){
		double sig = 1.;
		geom->Plane((*itr)->Plane())->Cell((*itr)->Cell())->GetCenter(xyz);
		
		if((*itr)->View() == geo::kX){
		  if ((*itr)->Plane() > planeMax) planeMax = (*itr)->Plane();
		  if ((*itr)->Plane() < planeMin) planeMin = (*itr)->Plane();
		  xzViewXtest.push_back(xyz[0]);
		  xzViewZtest.push_back(xyz[2]);
		  xzViewWtest.push_back(sig);
		}
		if((*itr)->View() == geo::kY){
		  if ((*itr)->Plane() > planeMax) planeMax = (*itr)->Plane();
		  if ((*itr)->Plane() < planeMin) planeMin = (*itr)->Plane();
		  yzViewYtest.push_back(xyz[1]);
		  yzViewZtest.push_back(xyz[2]);
		  yzViewWtest.push_back(sig);
		}
		itr++;
	      }// end loop over hits in this track
	      
	      if (xzViewXtest.size() >1){
		Chi2 = geo::LinFit(xzViewZtest, xzViewXtest, xzViewWtest, &xzStarttest[0], &xzStarttest[1], &xzEndtest[0], &xzEndtest[1]);
	      }
	      
	      if (yzViewYtest.size() >1){
		Chi2 = geo::LinFit(yzViewZtest, yzViewYtest, yzViewWtest, &yzStarttest[0], &yzStarttest[1], &yzEndtest[0], &yzEndtest[1]);
	      }
	      
	      xzStart[0] = xzStarttest[0];
	      xzStart[1] = xzStarttest[1];
	      yzStart[0] = yzStarttest[0];
	      yzStart[1] = yzStarttest[1];
	      xzEnd[0]   = xzEndtest[0];
	      xzEnd[1]   = xzEndtest[1];
	      yzEnd[0]   = yzEndtest[0];
	      yzEnd[1]   = yzEndtest[1];
	      
	      // reset the end points based on startz and endz
	      slopeXZ     = 999999.;
	      interceptXZ = xzEnd[1] - slopeXZ*xzEnd[0];
	      if(fabs(xzStart[0] - xzEnd[0]) > 1.e-2){
		slopeXZ     = (xzEnd[1] - xzStart[1])/(xzEnd[0] - xzStart[0]);
		interceptXZ = xzEnd[1] - slopeXZ*xzEnd[0];
	      }      
	      
	      slopeYZ     = 999999.;
	      interceptYZ = yzEnd[1] - slopeYZ*yzEnd[0];
	      if(fabs(yzStart[0] - yzEnd[0]) > 1.e-2){
		slopeYZ     = (yzEnd[1] - yzStart[1])/(yzEnd[0] - yzStart[0]);
		interceptYZ = yzEnd[1] - slopeYZ*yzEnd[0];
	      }   
	    }

	    if ((numclust == (cluster.size()-1)) && (clustcount != cluster.size())){
	      numclust=0;
	      clustcount = cluster.size();
	    }
	  }

        }//end numclust
	////////////////////////////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////////////////////////////
	
	
	// are there hits in each view? if not, reset the start and end z positions to
	// be based on the view with hits

        if(mergecells.size()<3&&motherids[nc].size()>0)continue;
        if(mergecells.size()<2)continue;

        bool sontag = false;

	if(motherids[nc].size()>0)continue;	
	double startz = 0.;
	double endz = 0.;
	if(xzViewZ.size() < 1){
	  startz = yzStart[0];
	  endz   = yzEnd[0];
	  yzStart[1] = slopeYZ*startz + interceptYZ;
	  yzEnd[1]   = slopeYZ*endz   + interceptYZ;
	  xzStart[1] = geom->DetHalfWidth();
	  xzEnd[1] = geom->DetHalfWidth();
	}
	if(yzViewZ.size() < 1){
	  startz = xzStart[0];
	  endz   = xzEnd[0];
	  xzStart[1] = slopeXZ*startz + interceptXZ;
	  xzEnd[1]   = slopeXZ*endz   + interceptXZ;
	  yzStart[1] = geom->DetHalfHeight();
	  yzEnd[1] = geom->DetHalfHeight();
	}
	
	
	
	double xextremes[2] = {9999., -9999.}; // first is min, second max
	double yextremes[2] = {9999., -9999.};
	for(unsigned int i = 0; i < xzViewX.size(); ++i){
	  if(xzViewX[i] < xextremes[0]) xextremes[0] = xzViewX[i];
	  if(xzViewX[i] > xextremes[1]) xextremes[1] = xzViewX[i];
	}
	
	for(unsigned int i = 0; i < yzViewY.size(); ++i){
	  if(yzViewY[i] < yextremes[0]) yextremes[0] = yzViewY[i];
	  if(yzViewY[i] > yextremes[1]) yextremes[1] = yzViewY[i];
	}// end loop over hits in this track
	// check that the end points are consistent with the extremes of hits used in each view
	if(yzStart[1] > yzEnd[1]   && yextremes[1] > yzStart[1]) yzStart[1] = yextremes[1];
	if(yzStart[1] < yzEnd[1]   && yextremes[0] < yzStart[1]) yzStart[1] = yextremes[0];
	if(yzEnd[1]   > yzStart[1] && yextremes[1] > yzEnd[1]  ) yzEnd[1]   = yextremes[1];
	if(yzEnd[1]   < yzStart[1] && yextremes[0] < yzEnd[1]  ) yzEnd[1]   = yextremes[0];
	if(xzStart[1] > xzEnd[1]   && xextremes[1] > xzStart[1]) xzStart[1] = xextremes[1];
	if(xzStart[1] < xzEnd[1]   && xextremes[0] < xzStart[1]) xzStart[1] = xextremes[0];
	if(xzEnd[1]   > xzStart[1] && xextremes[1] > xzEnd[1]  ) xzEnd[1]   = xextremes[1];
	if(xzEnd[1]   < xzStart[1] && xextremes[0] < xzEnd[1]  ) xzEnd[1]   = xextremes[0];
	
	//check that the end points are within the detector
	if(xzStart[1] < -geom->DetHalfWidth())  xzStart[1] = -geom->DetHalfWidth();
	if(xzStart[1] >  geom->DetHalfWidth())  xzStart[1] =  geom->DetHalfWidth();
	if(xzEnd[1]   < -geom->DetHalfWidth())  xzEnd[1]   = -geom->DetHalfWidth();
	if(xzEnd[1]   >  geom->DetHalfWidth())  xzEnd[1]   =  geom->DetHalfWidth();
	
	if(yzStart[1] < -geom->DetHalfHeight()) yzStart[1] = -geom->DetHalfHeight();
	if(yzStart[1] >  geom->DetHalfHeight()) yzStart[1] =  geom->DetHalfHeight();
	if(yzEnd[1]   < -geom->DetHalfHeight()) yzEnd[1]   = -geom->DetHalfHeight();
	if(yzEnd[1]   >  geom->DetHalfHeight()) yzEnd[1]   =  geom->DetHalfHeight();
	
	double length = util::pythag(endz-startz, xzEnd[1]-xzStart[1], yzEnd[1]-yzStart[1]);
	double dcosx  = (xzEnd[1] - xzStart[1])/length;
	double dcosy  = (yzEnd[1] - yzStart[1])/length;
	double dcosz  = (endz - startz)/length;
	
	double evis = 0;
        for(unsigned int i=0; i<mergecells.size(); i++){
          evis += mergecells[i]->PE();
        }
        rb::Cluster Clust(mergecells);
        vCluster.push_back(Clust);
        starts[cc].push_back(xzStart[1]); 
        starts[cc].push_back(yzStart[1]); 
        starts[cc].push_back(startz);
        ends[cc].push_back(xzEnd[1]); 
        ends[cc].push_back(yzEnd[1]); 
        ends[cc].push_back(endz);
        dirs[cc].push_back(dcosx);
        dirs[cc].push_back(dcosy);
        dirs[cc].push_back(dcosz);
        sig[cc] = evis;
        isnoise[cc] = sontag;
        cc++;
        std::vector< int > trackcount;
        
      }
    }
    
    std::unique_ptr< std::vector<rb::Cluster> > clustcol(new std::vector<rb::Cluster> );
    for(unsigned int i = 0; i < vCluster.size(); i++)
    {
	TVector3 start(starts[i][0], starts[i][1], starts[i][2]);
	TVector3 end(ends[i][0], ends[i][1], ends[i][2]);
        TVector3 dir(dirs[i][0], dirs[i][1], dirs[i][2]);
	
	if(fDebug > 0){
	  std::cout << "3D end points are " << std::endl;
	  start.Print();
	  end.Print();
	}
	
      rb::Track track(vCluster[i], start, dir);
      track.AppendTrajectoryPoint(end);
      track.SetNoise(isnoise[i]);
	if(fDebug){
	std::cout << "track start at " << track.Start().X() << " " 
		  << track.Start().Y() << " " << track.Start().Z() << std::endl;
	std::cout << "track end at "   << track.Stop().X()   << " " 
		  << track.Stop().Y()  << " " << track.Stop().Z()   << std::endl;
	}
	
      trackcol->push_back(track);
    }
    evt.put(std::move(trackcol));
    return;
  }//End of produce section
}//end namespace

namespace jmshower{
  DEFINE_ART_MODULE(JMClusterMerge)
    }
