///////////////////////////////////////////////////////////////////////
/// \brief   Example to use RecoJMShower 
///          for particle identification
/// \author  bianjm@.physics.umn.edu
////////////////////////////////////////////////////////////////////////

#include <cmath>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

// ROOT includes
#include "TLorentzVector.h"
#include "TVector3.h"

// ART includes
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"

// NOvA includes
#include "Geometry/Geometry.h"
#include "RecoJMShower/JMShower.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCTruth.h"


namespace jmshower {

  class RecoJMShowerFilter : public art::EDFilter {
    public:
      explicit RecoJMShowerFilter(fhicl::ParameterSet const& pset); 
      virtual ~RecoJMShowerFilter();                        
      virtual bool filter(art::Event & evt);

      void beginJob();
      bool beginSubRun(art::SubRun& sr);
      void reconfigure(const fhicl::ParameterSet& pset);

      std::string   fPotLabel;  ///< Module that produced the POTSum object

      int fNJMShower;      
      int fEvtNuPdg;       
      int fEvtCCNC;        

  };
}


namespace jmshower{
  //_________________________________________________________________________

  RecoJMShowerFilter::RecoJMShowerFilter(fhicl::ParameterSet const& pset)
  {
    reconfigure(pset);
  }
 
  //......................................................................
  
  void RecoJMShowerFilter::reconfigure(const fhicl::ParameterSet& pset)
  {
    fPotLabel  = pset.get< std::string>("PotLabel");
    fNJMShower = pset.get<int>("NJMShower");
    fEvtNuPdg  = pset.get<int>("EvtNuPdg");
    fEvtCCNC   = pset.get<int>("EvtCCNC");
  }
  
  //......................................................................
  
  RecoJMShowerFilter::~RecoJMShowerFilter()
  {
  }
  
  //......................................................................
  
  void RecoJMShowerFilter::beginJob()
  {
  }
  bool RecoJMShowerFilter::beginSubRun(art::SubRun& sr) {
    //    art::Handle< sumdata::POTSum > p;
    //    sr.getByLabel(fPotLabel,p);
    //    std::cout<<"pot = "<<p->totgoodpot<<std::endl;
    return true;
  } 
  //......................................................................
  bool RecoJMShowerFilter::filter(art::Event& evt)
  {
    //----------------------------------------------------
    //  Run, subrun and event information
    //----------------------------------------------------
    //    int run   = evt.run();
    //    int srun  = evt.subRun();
    //    int event = evt.id().event();
    art::ServiceHandle<geo::Geometry> geom;
    //------------------------------------------------------------------------
    //  MC true information
    //----------------------------------------------------------------------
    
    
    //Neutrino MC truth

    TLorentzVector evtTrueNuP4(0,0,0,0);
    TVector3 evtTrueNuVtx(0,0,-100);
    int evtTrueNuCCNC = -1;
    //    int evtTrueNuMode = -1;
    int evtTrueNuPdg = 0;


    if(!evt.isRealData()){
      art::Handle< std::vector<simb::MCTruth> > mc;
      try {
        evt.getByLabel("generator", mc);
        if (mc->size()!=1) return false;
      }
      catch (...) {
        return false;
      }
      if ((*mc)[0].NeutrinoSet()) {
        const simb::MCNeutrino& gen = (*mc)[0].GetNeutrino();
        const simb::MCParticle& nu  = gen.Nu();
        evtTrueNuP4 = nu.Momentum();
        evtTrueNuVtx.SetXYZ(nu.EndX(),nu.EndY(),nu.EndZ());
        evtTrueNuCCNC = gen.CCNC() ;
        //        evtTrueNuMode = gen.Mode() ;
        evtTrueNuPdg = nu.PdgCode() ;
        std::cout<<"Nuetrino, Energy "<<nu.PdgCode()<<" "<<evtTrueNuP4.E()<<std::endl;
        std::cout<<"CCNC, Mode "<<gen.CCNC()<<" "<<gen.Mode()<<std::endl;
        enum simb::int_type_ outMode = (enum simb::int_type_) gen.Mode();
//      enum simb::int_type_ outMode1 = (enum simb::int_type_) InteractionType();
        std::cout<<"Mode enum  "<<outMode<<std::endl;
//        std::cout<<"Type enum  "<<outMode1<<std::endl;
      }
      if(evtTrueNuPdg!=fEvtNuPdg) return false;
      if(evtTrueNuCCNC!=fEvtCCNC) return false;
    }

    std::cout<<"before veto evtTrueNuPdg,evtTrueNuCCNC,recoshowercol.size  "<<evtTrueNuPdg<<" "<<evtTrueNuCCNC<<" "<<std::endl;


//      if(evtTrueNuMode!=3) return false;

    art::Handle< std::vector<jmshower::JMShower> > showerhandle;
    evt.getByLabel("recojmshower", showerhandle);
    std::vector<art::Ptr<jmshower::JMShower> >  svcshowercol;
    for(unsigned int i = 0; i < showerhandle->size(); i++){
      art::Ptr<jmshower::JMShower> p(showerhandle, i);
      svcshowercol.push_back(p);
    }

    std::cout<<"Pdg, CCNC, size "<<evtTrueNuPdg<<" "<<evtTrueNuCCNC<<" "<<svcshowercol.size()<<std::endl;
    bool /*seltag0=false,seltag1=false,seltag2=false, seltag3=false,*/seltag4=false; 
    if(svcshowercol.size()>0){


      double llemu = svcshowercol[0]->DedxLongLL(int(jmshower::kElectron)) - svcshowercol[0]->DedxLongLL(int(jmshower::kMuon));
      double ltemu = svcshowercol[0]->DedxTransLL(int(jmshower::kElectron)) - svcshowercol[0]->DedxTransLL(int(jmshower::kMuon));

      double lleproton = svcshowercol[0]->DedxLongLL(int(jmshower::kElectron)) - svcshowercol[0]->DedxLongLL(int(jmshower::kProton));
      //      double lteproton = svcshowercol[0]->DedxTransLL(int(jmshower::kElectron)) - svcshowercol[0]->DedxTransLL(int(jmshower::kProton));

      double llepion = svcshowercol[0]->DedxLongLL(int(jmshower::kElectron)) - svcshowercol[0]->DedxLongLL(int(jmshower::kPion));
      //      double ltepion = svcshowercol[0]->DedxTransLL(int(jmshower::kElectron)) - svcshowercol[0]->DedxTransLL(int(jmshower::kPion));

      //      double llepi0 = svcshowercol[0]->DedxLongLL(int(jmshower::kElectron)) - svcshowercol[0]->DedxLongLL(int(jmshower::kPi0));
      //      double ltepi0 = svcshowercol[0]->DedxTransLL(int(jmshower::kElectron)) - svcshowercol[0]->DedxTransLL(int(jmshower::kPi0));



      //      if(svcshowercol.size()==1 && lleproton<0 && lteproton<0) seltag1 = true;
      //      if(svcshowercol.size()==1 && llepion<0 && ltepion<0) seltag2 = true;
      //      if(svcshowercol.size()==1 && llepi0<0 && ltepi0<0) seltag3 = true;
      if(svcshowercol.size()==1 && llemu<0 && ltemu<0) seltag4 = true;
      std::cout<<"llemu, lleproton, llepion "<<llemu<<" "<<lleproton<<" "<<llepion<<std::endl;
    }
    //    if(svcshowercol.size()==0) seltag0 = true;
    return seltag4;
  }
}// end namespace jmshower


//......................................................................
namespace jmshower
{
  DEFINE_ART_MODULE(RecoJMShowerFilter);
}
