///////////////////////////////////////////////////////////////////////////
/// \brief  Simple service to provide a RunHistory configured to the right run
/// \author Christopher Backhouse - bckhouse@caltech.edu
//////////////////////////////////////////////////////////////////////////

// NOvA includes
#include "RunHistory/service/RunHistoryService.h"
#ifdef NOVACMAKE
#include "Database/cxx/include/Table.h"
#else
#include "Database/Table.h"
#endif

#include "RunHistory/service/DetectorService.h"

// Framework includes
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// ROOT includes
#include "TTimeStamp.h"

namespace nova
{
namespace dbi
{

  //------------------------------------------------------------
  RunHistoryService::RunHistoryService(const fhicl::ParameterSet& pset,
                                       art::ActivityRegistry &reg)
  {
    reconfigure(pset);
    reg.sPreBeginRun.watch(this, &RunHistoryService::preBeginRun);
  }

  //----------------------------------------------------------

  RunHistoryService::~RunHistoryService()
  {
  }

  //-----------------------------------------------------------
  void RunHistoryService::reconfigure(const fhicl::ParameterSet& pset)
  {
    fCalHitLabel            = pset.get< std::string >("CalHitLabel");
    fFakeRunNumber          = pset.get< int >("FakeRunNumber", 0);
    fTimeQueries            = pset.get< bool >("TimeQueries", false);
    fUseMasking             = pset.get< bool >("UseMasking", false);
    fAnalysisDiBlockMasking = pset.get< bool >("AnalysisDiBlockMasking", false);
    fCustomMasking          = pset.get< bool >("CustomMasking", false);
    fCustomMask             = pset.get< std::vector<int> >("CustomMask");
    fAbort                  = pset.get< bool >("AbortIfNoMask", false);
    fAbortIfMissingRunInDB     = pset.get< bool >("AbortIfMissingRunInDB", true);
  }

  //----------------------------------------------------------
  void RunHistoryService::preBeginRun(const art::Run& run)
  {
    art::ServiceHandle<rh::DetectorService> ds;
    RunHistory::SetDetector(ds->DetId(run));

    int runNumber = (fFakeRunNumber == 0) ? run.run() : fFakeRunNumber;

    RunHistory::SetRunNumber(runNumber);

    this->SetTimeQueries(fTimeQueries);
    this->SetTimeParsing(fTimeQueries);
    this->SetIfAnyMasking(fUseMasking);
    this->SetAnalysisDiBlockMasking(fAnalysisDiBlockMasking);
    this->SetIfCustomMasking(fCustomMasking);
    this->SetAbort(fAbort);
    this->SetAbortIfMissingRunInDB(fAbortIfMissingRunInDB);

    if (novadaq::cnv::DetInfo::GetId(RunHistory::Detector()) == novadaq::cnv::kFARDET) { // only worry about masking for FD
      std::vector<bool> cmask;  // let people put 1s and 0s into fcl instead of trues and falses
      for(unsigned int i = 0; i < 14; i++) {
        if(fCustomMask[i] == 1) cmask.push_back(true);
        else cmask.push_back(false);
      }
      this->SetCustomMask(cmask);

      art::Handle<int> mask;
      run.getByLabel(fCalHitLabel,mask);
      if(mask.failedToGet()) {
        if(fCustomMasking) {
          std::cout << "RunHistory: No mask in file.  Custom mask requested." << std::endl;
          RunHistory::GoodDiBlockMask();
        }
        else if(runNumber >= 1000000 || runNumber < 10000) {  // clear MC that can't access DB
          std::cout << "RunHistory: this looks like ideal conditions MC.  No mask applied: " << RunHistory::DiBlockMaskToString((1<<15)-1) << std::endl;
          RunHistory::GoodDiBlockMask();
        }
        else {
          if(fAbort) assert(RunHistory::GoodDiBlockMask()!=(1<<16)-1); // running in mode that absolutely requires a mask, and there was not one
          if(RunHistory::GoodDiBlockMask()==0) std::cout << "RunHistory: did not find mask in file, using current fcl settings of: detector fully masked off! " << RunHistory::DiBlockMaskToString(0) << std::endl;
          else std::cout << "RunHistory: did not find mask in file, using current fcl settings of: " << RunHistory::DiBlockMaskToString(RunHistory::GoodDiBlockMask()) << std::endl;
        }
      }
      else {
        if(fAbort && runNumber < 1000000 && runNumber > 10000) assert(*mask!=(1<<16)-1); // running in mode that absolutely requires a mask, and there was not one
        RunHistory::SetGoodDiBlockMask(*mask);
        if(*mask==0) {
          std::cout << "RunHistory: found mask in file: detector fully masked off! " << RunHistory::DiBlockMaskToString(RunHistory::GoodDiBlockMask()) << std::endl;
        }
        else if(*mask==(1<<16)-1) {
          std::cout << "RunHistory: no run was found in good runs DB when this file was processed, so we assume perfect.  LiveGeometry should still catch any empty regions, but check badboxes in the event display to be sure."  << std::endl;
        }
        else if(*mask==(1<<17)-1) {
          std::cout << "RunHistory: masking was turned off when file was processed, so apply no mask. " << RunHistory::DiBlockMaskToString(RunHistory::GoodDiBlockMask()) << std::endl;
        }
        else {
          std::cout << "RunHistory: found mask in file: " << RunHistory::DiBlockMaskToString(*mask) << std::endl;
        }
      }
    }

    // determine whether we are dealing with MC, data, or a mixture of the two
    // for now just test on the date of the file.  MC has a timestamp before 2010
    art::Timestamp ts = run.beginTime();

    // make it into a TTimeStamp
    TTimeStamp tts(ts.timeHigh(), ts.timeLow());

    // Get the year of the timestamp from YYYYMMDD format
    uint32_t year = tts.GetDate()/10000;

    this->SetDataType(nova::dbi::kNDataType);
    if(year < 2010) this->SetDataType(nova::dbi::kMCOnly);
    else            this->SetDataType(nova::dbi::kDataOnly);

    if (runNumber > 1000000 || runNumber < 10000) 
      this->SetDataType(nova::dbi::kMCOnly);

    return;
  }

  DEFINE_ART_SERVICE(RunHistoryService)

}
}
////////////////////////////////////////////////////////////////////////
