////////////////////////////////////////////////////////////////////////
/// \brief   Build and write out the ShowerLID objects and make nue PID
///           
/// \TODO    Jianming's oscillation weighting:  Need to find standard
//           infrastructure to do this and replace his
///           code, which I stripped out of this class
///
/// \author  Alex Smith smith@physics.umn.edu
////////////////////////////////////////////////////////////////////////

// ROOT includes
#include "TFile.h"
#include "TH1D.h"
#include "TLorentzVector.h"
#include "TMath.h"
#include "TMultiLayerPerceptron.h"
#include "TRandom.h"
#include "TTree.h"

// NOvA includes
#include "DAQChannelMap/DAQChannelMapConstants.h"
#include "Geometry/Geometry.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/PID.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/Vertex.h"
#include "Utilities/AssociationUtil.h"
#include "Calibrator/Calibrator.h"

// ART includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindMany.h"
#include "art/Persistency/Common/Assns.h"
#include "art/Persistency/Common/Ptr.h"
#include "messagefacility/MessageLogger/MessageLogger.h"


#include "fhiclcpp/ParameterSet.h"
#include "ShowerLID/DedxDistribution.h"
#include "ShowerLID/ParticleIDAlg.h"
#include "ShowerLID/NuEEnergyAlg.h"
#include "ShowerLID/LIDAlg.h"
#include "ShowerLID/EventLID.h"
#include "ShowerLID/ShowerLID.h"
#include "Utilities/func/EnvExpand.h"

#include <cmath>

namespace slid {

  class LIDBuilder : public art::EDProducer {
  public:
    explicit LIDBuilder(const fhicl::ParameterSet& pset);
    ~LIDBuilder();

    virtual void produce(art::Event& evt);

    virtual void reconfigure(const fhicl::ParameterSet& pset);

    //virtual void beginJob();
    virtual void beginRun(art::Run& run);
    //virtual void endJob();

  protected:
    std::string fShowerLabel;
    std::string fSlicerLabel;
    std::string fVertexLabel;
    bool fPIDOnlyForHighestEShower;
    bool fObeyPreselection;
    bool fSkipNoiseSlices;
    bool fTrainingMode;

    // future alg       const std::string kKNNWithENu = "knnwithenu";
    // future alg       const std::string kKNNNoENu = "knnnoenu";

    std::string fLibPath;
    int kIDNuE = 12; // TEMP, should find where stadard defs are.
    /// Particle ID alorithm (loglikelihoods and dE/dx)
    ParticleIDAlg* fParticleIDAlg;
    NuEEnergyAlg* fNuEEnergyAlg;

    /// MVAAlgorithm (Neural Net, KNN, etc. May include any algorithm that is
    /// a subclass of slid::LIDAlg)
    slid::LIDAlg* fMVAAlg;
    slid::LIDAlg* fMVAAlgWithE;
    /// FCL parameters for NuE Energy alorithm
    fhicl::ParameterSet fNuEEnergyAlgPSet;

    /// FCL parameters for particle ID alorithm (loglikelihoods and dE/dx)
    fhicl::ParameterSet fParticleIDAlgPSet;

    /// FCL parameters for LID alorithm
    fhicl::ParameterSet fLIDAlgPSet;

    /// determine if library load is necessary or not when moving to a new run
    bool libLoad = false;

  };
  bool CompareByE( const slid::ShowerLID& a, const slid::ShowerLID& b);
} // namespace




namespace slid {
  int Ncount0 = 0, Ncount1 = 0, Ncount2 = 0;
  int Nfill = 0, Nput = 0;
  //......................................................................

  LIDBuilder::LIDBuilder(fhicl::ParameterSet const& pset) 
    : fParticleIDAlg(0), fNuEEnergyAlg(0),
      fMVAAlg(0),
      fMVAAlgWithE(0),
      fNuEEnergyAlgPSet(pset.get< fhicl::ParameterSet >("NuEEnergyAlgPSet")),
      fParticleIDAlgPSet(pset.get< fhicl::ParameterSet >("ParticleIDAlgPSet")),
      fLIDAlgPSet(pset.get< fhicl::ParameterSet >("LIDAlgPSet")) 
  {
    reconfigure(pset);
    // Define output structures produced.
    produces<std::vector<slid::ShowerLID> >();
    produces<std::vector<slid::EventLID> >();
    produces< art::Assns<slid::EventLID,  rb::Cluster> >();
    produces< art::Assns<slid::ShowerLID, rb::Shower> >();
    produces< art::Assns<slid::ShowerLID, rb::Cluster> >();
  }

  //......................................................................

  LIDBuilder::~LIDBuilder() 
  {
    if (fParticleIDAlg) delete fParticleIDAlg;
    if (fNuEEnergyAlg)  delete fNuEEnergyAlg;
    if (fMVAAlg)        delete fMVAAlg;
    if (fMVAAlgWithE)   delete fMVAAlgWithE;
  }

  //......................................................................

  void LIDBuilder::reconfigure(const fhicl::ParameterSet& pset) 
  {
    //
    //  Read FCL parameters
    //
    fNuEEnergyAlgPSet = pset.get< fhicl::ParameterSet >("NuEEnergyAlgPSet");
    fParticleIDAlgPSet = pset.get< fhicl::ParameterSet >("ParticleIDAlgPSet");
    fLIDAlgPSet = pset.get< fhicl::ParameterSet >("LIDAlgPSet");
    fLibPath = util::EnvExpansion(pset.get< std::string >("LibraryPath"));
    fSlicerLabel = pset.get< std::string >("SlicerLabel");
    fShowerLabel = pset.get< std::string >("ShowerLabel");
    fVertexLabel = pset.get< std::string >("VertexLabel");
    fPIDOnlyForHighestEShower = pset.get<bool>("PIDOnlyForHighestEShower");
    fObeyPreselection = pset.get<bool>("ObeyPreselection");
    fSkipNoiseSlices = pset.get<bool>("SkipNoiseSlices");
    fTrainingMode = pset.get<bool>("TrainingMode");

  }

  void LIDBuilder::beginRun(art::Run& run) 
  {
    art::ServiceHandle<geo::Geometry> fGeom;
    novadaq::cnv::DetId detId = fGeom->DetId();

    //NOTE: This code is here because the geometry service only becomes available to determine the
    //detector id at the begin run step.  However at the present moment the libraries being loaded
    //by the particle id and ann algorithms will not very from run to run and only depend on the detector id.
    //If there are more then one run in a file this loading process takes a lot of time, only do it once per job.
    if (!libLoad){
      if (fParticleIDAlg) delete fParticleIDAlg;
      if (fNuEEnergyAlg)  delete fNuEEnergyAlg;
      if (fMVAAlg)        delete fMVAAlg;
      if (fMVAAlgWithE)   delete fMVAAlgWithE;
      fNuEEnergyAlg     = new NuEEnergyAlg(fNuEEnergyAlgPSet);
      fParticleIDAlg    = new ParticleIDAlg(fParticleIDAlgPSet, detId, fNuEEnergyAlg);
      bool withE = true;
      if (!fTrainingMode){
        fMVAAlg           = new LIDAlg(fLIDAlgPSet, fNuEEnergyAlg, !withE);
        fMVAAlgWithE      = new LIDAlg(fLIDAlgPSet, fNuEEnergyAlg, withE);
      }
      libLoad = true;
    }
  }


  //......................................................................

  void LIDBuilder::produce(art::Event & evt) 
  {

    art::ServiceHandle<geo::Geometry> fGeom;

    LOG_DEBUG("LIDBuilderDEBUG") << "In LIDBuilder::produce" << '\n';

    std::unique_ptr< std::vector<slid::ShowerLID> >             showerlidcol(new std::vector<slid::ShowerLID>);
    std::unique_ptr< std::vector<slid::EventLID> >              elidcol(new std::vector<slid::EventLID>);
    std::unique_ptr< art::Assns<slid::ShowerLID, rb::Shower> >  assnSlidShower(new art::Assns<slid::ShowerLID, rb::Shower>);
    std::unique_ptr< art::Assns<slid::ShowerLID, rb::Cluster> > assnSlidSlice(new art::Assns<slid::ShowerLID, rb::Cluster>);
    std::unique_ptr< art::Assns<slid::EventLID, rb::Cluster> >  assnElidSlice(new art::Assns<slid::EventLID, rb::Cluster>);


    art::Handle<std::vector<rb::Cluster> > slicecol;
    evt.getByLabel(fSlicerLabel, slicecol);
    art::PtrVector<rb::Cluster> slicelist;
    for (unsigned int i = 0; i < slicecol->size(); ++i) {
      art::Ptr<rb::Cluster> slice(slicecol, i);
      slicelist.push_back(slice);
    }
    //
    //  Grab list of rb::Shower* for calculations and reading and list of art::Ptr<rb::Shower>
    //  for writing the output data products.  Since these **should** be indexed the same, this
    //  should be OK.
    //
    art::FindManyP<rb::Shower> fmtP(slicecol, evt, fShowerLabel);
    art::FindMany<rb::Shower> fmt(slicecol, evt, fShowerLabel);
    art::FindManyP<rb::Vertex> fmv(slicecol, evt, fVertexLabel);

    // loop over all of the slices
    for (unsigned int iSlice = 0; iSlice < slicelist.size(); ++iSlice) {
      const size_t nextShwIdx = showerlidcol->size();
      //
      //  Skip noise slices if this option is selected in FCL
      //
      if (fSkipNoiseSlices && slicelist[iSlice]->IsNoise()) continue;
      //
      //  If there was nue preselection, skip slice if it failed if this option is selected in FCL
      //
      if (fObeyPreselection && rb::IsFiltered(evt, slicecol, iSlice)) continue;

      // get all of the showers in this slice
      if ((!fmt.isValid())||(!fmtP.isValid())) continue;
      const std::vector< const rb::Shower* > showerCollection = fmt.at(iSlice);
      const std::vector < art::Ptr < rb::Shower >> showerCollectionPtr = fmtP.at(iSlice);
      if ((showerCollection.empty())||(showerCollectionPtr.empty())) continue;

      //hold slid objects temporarily before sorting
      std::vector<slid::ShowerLID> tempslidcol;
      //keep track of shower index associationed with the slid object after sorting
      std::vector<std::pair<double,int> > associationlist;


      //get the vertex for the slice
      const std::vector<art::Ptr<rb::Vertex>> vertCol = fmv.at(iSlice);
      //currently only one vertex per slice
      TVector3 vert = vertCol[0]->GetXYZ();

      //
      // Get indices of highest and second-highest energy showers
      //
      double eShHighest = 0;
      double sliceShwE = 0;
      unsigned int iHighestEnergySh = -1;
      for (unsigned int jShower = 0; jShower < showerCollection.size(); ++jShower) {
	double thisESh = fNuEEnergyAlg->ShowerEnergy(showerCollection[jShower], showerCollection, evt);
        sliceShwE += thisESh;
	if (thisESh > eShHighest) {
	  eShHighest = thisESh;
	  iHighestEnergySh = jShower;
	}
      }

      TLorentzVector evtP4(0.0,0.0,0.0,0.0);
      //
      //  Loop over showers to fill PID info
      //
      std::map<int, int> showerLidToShower;
      unsigned int nShowers = showerCollection.size();
      for (unsigned int iShower = 0; iShower < nShowers; ++iShower) {
	
	
	slid::ShowerLID showerlid;

        //we only have to calculate the electron longitudinal and transverse likelihoods once, and can save the values
        double ellmax = -9999;
        double electronLongLL = -9999;
        double electronTransLL = -9999;
        for(int itype=0; itype != int(slid::DedxParticleType::kDEDXPARTICLETYPESIZE); ++itype){
          if(itype == int(slid::DedxParticleType::kELECTRON)||itype == int(slid::DedxParticleType::kELECTRONCCQE)||itype == int(slid::DedxParticleType::kELECTRONCCRES)||
             itype == int(slid::DedxParticleType::kELECTRONCCDIS)||itype == int(slid::DedxParticleType::kELECTRONCCCOH)){
            double eLongLLTest = fParticleIDAlg->DedxLongLL(slid::DedxParticleType(itype), showerCollection[iShower], showerCollection, evt);
            double eTransLLTest = fParticleIDAlg->DedxTransLL(slid::DedxParticleType(itype), showerCollection[iShower], showerCollection, evt);
            if(ellmax< eLongLLTest+eTransLLTest){
              ellmax = eLongLLTest+eTransLLTest;
              electronLongLL = eLongLLTest;
              electronTransLL = eTransLLTest;
            }
          }
        }
	showerlid.SetEGLLL(electronLongLL
			   - fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kPHOTON, showerCollection[iShower],
							showerCollection, evt));
	showerlid.SetEGLLT(electronTransLL
			   - fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kPHOTON, showerCollection[iShower],
							 showerCollection, evt));
	showerlid.SetEMuLLL(electronLongLL
			    - fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kMUON, showerCollection[iShower], 
							 showerCollection, evt));
	showerlid.SetEMuLLT(electronTransLL
			    - fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kMUON, showerCollection[iShower], 
							  showerCollection, evt));
	showerlid.SetEPi0LLL(electronLongLL
			     - fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kPI0, showerCollection[iShower], 
							  showerCollection, evt));
	showerlid.SetEPi0LLT(electronTransLL
			     - fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kPI0, showerCollection[iShower], 
							   showerCollection, evt));
	showerlid.SetEPLLL(electronLongLL
			   - fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kPROTON, showerCollection[iShower], 
							showerCollection, evt));
	showerlid.SetEPLLT(electronTransLL
			   - fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kPROTON, showerCollection[iShower], 
							 showerCollection, evt));
	showerlid.SetENLLL(electronLongLL
			   - fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kNEUTRON, showerCollection[iShower], 
							showerCollection, evt));
	showerlid.SetENLLT(electronTransLL
			   - fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kNEUTRON, showerCollection[iShower], 
							 showerCollection, evt));
	showerlid.SetEPiLLL(electronLongLL
			    - fParticleIDAlg->DedxLongLL(slid::DedxParticleType::kPION, showerCollection[iShower], 
							 showerCollection, evt));
	showerlid.SetEPiLLT(electronTransLL
			    - fParticleIDAlg->DedxTransLL(slid::DedxParticleType::kPION, showerCollection[iShower], 
							  showerCollection, evt));
	showerlid.SetInverseGLLL( fParticleIDAlg->DedxInverseLongLL(slid::DedxParticleType::kPHOTON, showerCollection[iShower],
								    showerCollection, evt));
	showerlid.SetPi0PhotonID(fParticleIDAlg->Pi0SecondPhoton(showerCollection, iShower, evt) );
	
	showerlid.SetGap(fParticleIDAlg->GetGapVertexToShowerStart(showerCollection[iShower], vert, evt));

	showerlid.SetNMIPPlanes( fParticleIDAlg->GetMIPPlanes( showerCollection[iShower], showerCollection, evt));

	showerlid.SetPi0mass(std::max(fParticleIDAlg->Pi0Mass(showerCollection, iShower, evt), 0.0));
        showerlid.SetRadius(fParticleIDAlg->Radius(showerCollection, iShower, evt));

        double shwE = fNuEEnergyAlg->ShowerEnergy(showerCollection.at(iShower), showerCollection, evt);
	if (sliceShwE > 1e-10) showerlid.SetShowerEFrac(shwE/sliceShwE);
        else showerlid.SetShowerEFrac(0.0);

        if (shwE < 1e-10) shwE = 0.0;
        showerlid.SetShowerEnergy(shwE);
        TLorentzVector showerptrk(showerCollection[iShower]->Dir()[0]*shwE,
				  showerCollection[iShower]->Dir()[1]*shwE,
				  showerCollection[iShower]->Dir()[2]*shwE,
				  shwE);

        TVector3 nuDir = fGeom->NuMIBeamDirection();
        double theta = showerptrk.Angle(nuDir);

        showerlid.SetCosTheta(cos(theta));
        showerlid.SetVertexEnergy(fNuEEnergyAlg->VertexEnergy(showerCollection.at(iShower),
							      showerCollection, vertCol.at(0).get(), slicelist.at(iSlice).get(), evt));
	evtP4 += showerptrk;
	
	if (std::isnan(sliceShwE) || std::isnan(shwE) || std::isinf(sliceShwE) || std::isinf(shwE)) {
	  mf::LogError("LIDBuilder")<< "sliceShwE/shwE is nan or inf";
	  continue;
	}
	
	showerlid.SetIsMuon(!(showerlid.EMuLLL()>-0.2&&showerlid.EMuLLT()>-99999));
	showerlid.SetPdg(kIDNuE);
	showerlid.SetAlgDescription("dummy");
	showerlid.SetShowerDepEnergy(fNuEEnergyAlg->ShowerDepEnergy(showerCollection.at(iShower), 
								    showerCollection, evt));

	showerlid.SetNueEnergy(fNuEEnergyAlg->NuEEnergy(showerCollection,
							slicelist.at(iSlice).get(), 
							iShower, evt));

	showerlid.SetHadronicEnergy(fNuEEnergyAlg->HadronicEnergy(showerCollection,
								  slicelist.at(iSlice).get(), 
								  iShower, evt));

	if (!fPIDOnlyForHighestEShower || iShower == iHighestEnergySh){ 
          //in training mode just write -5, since the ann weights have not been determined
          if (fTrainingMode){
	    showerlid.SetAlgDescription(""); // not needed?
            showerlid.SetVal(-5);
            showerlid.SetValWithE(-5);
          }
          //standard running, calculate the ann
          else{
	    showerlid.SetAlgDescription(fMVAAlg->Description()); // not needed?
	    showerlid.SetVal(fMVAAlg->CalcMVAResult(showerlid));
	    showerlid.SetValWithE(fMVAAlgWithE->CalcMVAResult(showerlid));
          }
	}

        showerlid.SetVtxDoca(fParticleIDAlg->PointDoca(vertCol.at(0)->GetXYZ(),showerCollection[iShower]->Start(),showerCollection[iShower]->Stop()));

        showerlid.fPartLongLL = fParticleIDAlg->fPartLongLL;
        showerlid.fPartTransLL = fParticleIDAlg->fPartTransLL;
       
        tempslidcol.push_back(showerlid);
        associationlist.push_back(std::make_pair(showerlid.ShowerEnergy(),iShower));
      }// end loop over showers
      
      std::sort( tempslidcol.begin(), tempslidcol.end(), CompareByE);
      std::sort(associationlist.rbegin(), associationlist.rend());
      
      for (unsigned int islid=0; islid<tempslidcol.size(); ++islid){
	showerlidcol->push_back(tempslidcol[islid]);
	util::CreateAssn(*this, evt, *(showerlidcol.get()), showerCollectionPtr[associationlist[islid].second], *(assnSlidShower.get()));
      }

      //if we have added new showers, make event pid and associations to slice
      if (showerlidcol->size() > nextShwIdx){
        //for the event lid nextShwIdx points to the most energetic shower, which is what we want
	slid::EventLID elid;
	elid.SetPdg( showerlidcol->at(nextShwIdx).Pdg());
	elid.SetVal( showerlidcol->at(nextShwIdx).Value());
	elid.SetValWithE( showerlidcol->at(nextShwIdx).ValueWithE() );
	elidcol->push_back(elid);
	util::CreateAssn(*this, evt, *(elidcol.get()), slicelist[iSlice], *(assnElidSlice.get()));
        //create new associations for the most recent set of showers
	util::CreateAssn(*this, evt, *showerlidcol, slicelist[iSlice], *(assnSlidSlice.get()), nextShwIdx, showerlidcol->size());
      }
      
    } //end loop over slices
    
    evt.put(std::move(showerlidcol));
    evt.put(std::move(elidcol));
    evt.put(std::move(assnSlidShower));
    evt.put(std::move(assnElidSlice));
    evt.put(std::move(assnSlidSlice));

  }// end of producer

  ////////////////////////////////////////////////////////////////////////

  bool CompareByE( const slid::ShowerLID& a, const slid::ShowerLID& b)

  {
    return a.ShowerEnergy() > b.ShowerEnergy();
  }
  
  DEFINE_ART_MODULE(LIDBuilder);
}
