///////////////////////////////////////////////////////////////////////
// LIDTraining.cxx
//
// A module to reconstruct shower energy and topological variables  
// for particle identification
//
// \author $Author: Jianming Bian $
////////////////////////////////////////////////////////////////////////
#include <cmath>
#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>

// ROOT includes
#include "TFile.h"
#include "TMath.h"
#include "TH1D.h"
#include "TNtuple.h"
#include "TFitter.h"
#include "TMinuit.h"
#include "TVector3.h"
#include "TGraph2DErrors.h"

// NOvA includes
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "GeometryObjects/Geo.h"
#include "RecoBase/FilterList.h"
#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "SimulationBase/MCFlux.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCTrajectory.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/GTruth.h"
#include "Simulation/FLSHit.h"

#include "SimulationBase/MCTruth.h"
#include "Calibrator/Calibrator.h"
#include "CalibrationUtils/CalibUtil.h"
#include "MCCheater/BackTracker.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Shower.h"
#include "ShowerLID/ShowerLID.h"
#include "ShowerLID/ParticleIDAlg.h"
#include "ShowerLID/NuEEnergyAlg.h"

// ART includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Core/FindOneP.h"
#include "art/Framework/Core/FindMany.h"


#include "fhiclcpp/ParameterSet.h"  
#include "messagefacility/MessageLogger/MessageLogger.h" 

#include "OscLib/services/OscCalcService.h"
#include "OscLib/services/OscCalcGenService.h"

#include "SummaryData/POTSum.h"
#include "TRandom.h"
#include "TMatrixD.h"

#include "Utilities/func/MathUtil.h"
//#include "CosmicCalib/CalibUtil.h"


namespace slid{
  class LIDTraining : public art::EDAnalyzer
    {
    public:
      explicit LIDTraining(fhicl::ParameterSet const& pset); 
      ~LIDTraining();                        
      virtual void analyze(const art::Event& evt);
      virtual void beginJob();
      virtual void beginSubRun(const art::SubRun& sr);
      virtual void beginRun(const art::Run& run);
      virtual void reconfigure(const fhicl::ParameterSet& pset);

      virtual void endJob();

      typedef std::vector<TLorentzVector> Vp4;
      typedef std::vector<int> Vint;
      typedef std::vector<double> Vdouble;

      TH1D *ht_nue_rw;
      TH1D *ht_bkg_nue_rw;
      TH1D *ht_bkg_numu_rw;

//McTruth
    std::vector<int> getFaveIds(std::vector<cheat::NeutrinoWithIndex>& allNus,  
                                std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceEffPurs);

      static const int GAMMAID;
      static const int PIZEROID;
      static const int PIPLUSID;
      static const int PIMINUSID;
      static const int NUEID;
      static const int NUMUID;
      static const int NUTAUID;

    private:
      TTree*      fEvent;
      int    m_evtTrueNuCCNC;
      int    m_evtTrueNuMode;
      int    m_evtTrueNuPdg;
      double m_evtTrueNuEnergy;
      double m_evtTrueNuP4[4];
      double m_evtTrueNuVtx[3];
      double m_evtRwProb[20];
      double m_evtRwRand[20];
      double m_evtTotalRecoGeV;
      double m_evtPi0Mgg;
      double m_evtPi0StartDist;
      double m_evtPi0LineDist;
      double m_evtEtot;
      double m_evtSh1Energy; 
      double m_evtSh1Energy0; 
      double m_evtSh1Length; 
      double m_evtSh1NueEnergy; 
      double m_evtSh1Gap; 
      double m_evtSh1IsFid; 
      double m_evtSh1IsFidAna;
      double m_evtSh1P4[4];
      double m_evtSh1Start[3];
      double m_evtSh1DedxLLL[20];
      double m_evtSh1DedxLLT[20];
      double m_evtSh1NPlane;
      double m_evtSh1NMIPPlane;
      double m_evtSh1VtxGeV;
      double m_evtSh1Pi0Mgg;
      double m_evtSh1TrueDang;
      double m_evtSh1TrueEDang;
      int  m_evtSh1TruePdg;
      double m_evtSh1CosTheta;
      double m_evtSh1SumPt;
      double m_evtSh2Energy; 
      double m_evtSheEnergy; 

      TTree*        fShower;
      int    m_showerTrueNuCCNC;
      int    m_showerTrueNuMode;
      int    m_showerTrueNuPdg;
      double m_showerTrueDang;
      int m_showerTruePdg;
      int m_showerNPlane;
      double m_showerStart[3];
      double m_showerStop[3];
      double m_showerDir[3];
      double m_showerPlaneDedx[200];
      double m_showerTransCellDedx[20];
      double m_showerEnergy;
      double m_showerTrueP4[4];
      double m_showerIsFid;
      double m_showerIsFidAna;
      double m_showerRun;
      double m_showerSubRun;
      double m_showerEvent;
      int m_showerSubEvent;
      double m_showerId;
      double m_showerEff;
      double m_showerPur;

      int           fCutPlane;
      unsigned int  fMinTrack;
      unsigned int  fMaxTrack;
      double        fXThresh;
      double        fYThresh;
      double        fZMin;
      double        fZMax;
      std::string   fPotLabel;           ///< Module that produced the POTSum object
      bool          fRespectFilter;
      bool          fTruthMatchByEnergy;
      bool          fRecordMC;
      std::string   fSliceLabel;
      std::string   fProngLabel;
      std::string   fInstLabel;
      std::string   fVertexLabel;
      std::string   fHitLabel;
      std::string   fShowerLabel;
      std::string   fShowerLIDLabel;
      double        fCellHitPHThresh;

      ParticleIDAlg* fParticleIDAlg;
      NuEEnergyAlg* fNuEEnergyAlg;
      /// FCL parameters for NuE Energy alorithm
      fhicl::ParameterSet fNuEEnergyAlgPSet;
      /// FCL parameters for particle ID alorithm (loglikelihoods and dE/dx)
      fhicl::ParameterSet fParticleIDAlgPSet;


    };
}
  //______________________________________________________________________________
namespace slid{
  const int LIDTraining::GAMMAID   = 22;
  const int LIDTraining::PIZEROID  = 111;
  const int LIDTraining::PIPLUSID  = 211;
  const int LIDTraining::PIMINUSID = -211;
  const int LIDTraining::NUEID     = 12;
  const int LIDTraining::NUMUID    = 14;
  const int LIDTraining::NUTAUID   = 16;
  const double xmass[7] = {0.000511, 0, 0.105658, 0.1349766, 0.938272, 0.939565379, 0.139570};
  int Ncount0 = 0, Ncount1 = 0, Ncount2 = 0;

  LIDTraining::LIDTraining(fhicl::ParameterSet const& pset):
    EDAnalyzer(pset),
    fParticleIDAlg(0), 
    fNuEEnergyAlg(0),
    fNuEEnergyAlgPSet(pset.get< fhicl::ParameterSet >("NuEEnergyAlgPSet")),
    fParticleIDAlgPSet(pset.get< fhicl::ParameterSet >("ParticleIDAlgPSet"))
  {
    std::cout<<"LIDTraining::LIDTraining(fhicl::ParameterSet const& pset)"<<std::endl;
    reconfigure(pset);
  }

 
  //......................................................................
  
  void LIDTraining::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSliceLabel      = pset.get< std::string >("SliceLabel"     );
    fProngLabel      = pset.get< std::string >("ProngLabel"     );
    fInstLabel       = pset.get< std::string >("InstLabel"      );
    fVertexLabel     = pset.get< std::string >("VertexLabel"    );
    fHitLabel        = pset.get< std::string >("HitLabel"       );
    fShowerLabel     = pset.get< std::string >("ShowerLabel"    );
    fShowerLIDLabel  = pset.get< std::string >("ShowerLIDLabel" );
    
    fCellHitPHThresh = pset.get< double      >("CellHitPHThresh");
    fMinTrack        = pset.get< int         >("MinTrack"       );
    fMaxTrack        = pset.get< int         >("MaxTrack"       );
    fXThresh         = pset.get< double      >("XThresh"        );
    fYThresh         = pset.get< double      >("YThresh"        );
    fZMin            = pset.get< double      >("ZMin"           );
    fZMax            = pset.get< double      >("ZMax"           );
    fPotLabel        = pset.get< std::string >("PotLabel"       );
    fCutPlane        = pset.get< int         >("CutPlane"       );
    fRespectFilter   = pset.get< bool        >("RespectFilter"  );
    fTruthMatchByEnergy  = pset.get< bool        >("TruthMatchByEnergy");
    fRecordMC            = pset.get< bool        >("RecordMC"       );
    fNuEEnergyAlgPSet    = pset.get< fhicl::ParameterSet >("NuEEnergyAlgPSet");
    fParticleIDAlgPSet   = pset.get< fhicl::ParameterSet >("ParticleIDAlgPSet");
  }
  
  //......................................................................
  
  LIDTraining::~LIDTraining()
  {
    if (fParticleIDAlg) delete fParticleIDAlg;
    if (fNuEEnergyAlg)  delete fNuEEnergyAlg;
  }
  
  //......................................................................
  
  void LIDTraining::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;
    fEvent = tfs->make<TTree>("fEvent","fEvent");
    //True neutrino information
    fEvent->Branch("evtTrueNuCCNC", &m_evtTrueNuCCNC);
    fEvent->Branch("evtTrueNuMode", &m_evtTrueNuMode);
    fEvent->Branch("evtTrueNuPdg", &m_evtTrueNuPdg);
    fEvent->Branch("evtTrueNuEnergy", &m_evtTrueNuEnergy);
    fEvent->Branch("evtTrueNuP4[4]", m_evtTrueNuP4);
    fEvent->Branch("evtTrueNuVtx[3]", m_evtTrueNuVtx);
    fEvent->Branch("evtRwRand[20]", m_evtRwRand);
    fEvent->Branch("evtRwProb[20]", m_evtRwProb);
    fEvent->Branch("evtTotalRecoGeV", &m_evtTotalRecoGeV);
    fEvent->Branch("evtPi0Mgg",  &m_evtPi0Mgg);
    fEvent->Branch("evtPi0StartDist", &m_evtPi0StartDist);
    fEvent->Branch("evtPi0LineDist", &m_evtPi0LineDist); 
    fEvent->Branch("evtEtot", &m_evtEtot);
    fEvent->Branch("evtSh1Energy", &m_evtSh1Energy);
    fEvent->Branch("evtSh1Length", &m_evtSh1Length);
    fEvent->Branch("evtSh1NueEnergy", &m_evtSh1NueEnergy);
    fEvent->Branch("evtSh1SumPt", &m_evtSh1SumPt);
    fEvent->Branch("evtSh1Gap", &m_evtSh1Gap);
    fEvent->Branch("evtSh1IsFid", &m_evtSh1IsFid);
    fEvent->Branch("evtSh1IsFidAna", &m_evtSh1IsFidAna);
    fEvent->Branch("evtSh1P4[4]",m_evtSh1P4);
    fEvent->Branch("evtSh1Start[3]",m_evtSh1Start);
    fEvent->Branch("evtSh1DedxLLL[20]", m_evtSh1DedxLLL);
    fEvent->Branch("evtSh1DedxLLT[20]", m_evtSh1DedxLLT);
    fEvent->Branch("evtSh1NPlane", &m_evtSh1NPlane);
    fEvent->Branch("evtSh1NMIPPlane", &m_evtSh1NMIPPlane);
    fEvent->Branch("evtSh1VtxGeV", &m_evtSh1VtxGeV);
    fEvent->Branch("evtSh1Pi0Mgg", &m_evtSh1Pi0Mgg);
    fEvent->Branch("evtSh1TrueDang", &m_evtSh1TrueDang);
    fEvent->Branch("evtSh1TrueEDang", &m_evtSh1TrueEDang);
    fEvent->Branch("evtSh1TruePdg", &m_evtSh1TruePdg);
    fEvent->Branch("evtSh1CosTheta", &m_evtSh1CosTheta);
    fEvent->Branch("evtSh2Energy", &m_evtSh2Energy);

    fShower = tfs->make<TTree>("fShower","fShower");
    fShower->Branch("showerId",&m_showerId);
    fShower->Branch("showerRun",&m_showerRun);
    fShower->Branch("showerSubRun",&m_showerSubRun);
    fShower->Branch("showerEvent",&m_showerEvent);
    fShower->Branch("showerSubEvent",&m_showerSubEvent);
    fShower->Branch("showerTrueNuCCNC", &m_showerTrueNuCCNC);
    fShower->Branch("showerTrueNuMode", &m_showerTrueNuMode);
    fShower->Branch("showerTrueNuPdg", &m_showerTrueNuPdg);
    fShower->Branch("showerNPlane", &m_showerNPlane); 
    fShower->Branch("showerStart[3]", m_showerStart);
    fShower->Branch("showerStop[3]", m_showerStop);
    fShower->Branch("showerDir[3]", m_showerDir);
    fShower->Branch("showerEnergy", &m_showerEnergy);
    fShower->Branch("showerIsFid", &m_showerIsFid);
    fShower->Branch("showerIsFidAna", &m_showerIsFidAna);
    fShower->Branch("showerPlaneDedx[200]", m_showerPlaneDedx);
    fShower->Branch("showerTransCellDedx[20]", m_showerTransCellDedx);
    fShower->Branch("showerTrueDang", &m_showerTrueDang);
    fShower->Branch("showerTruePdg", &m_showerTruePdg);
    fShower->Branch("showerTrueP4[4]", m_showerTrueP4);
    fShower->Branch("showerEff", &m_showerEff);
    fShower->Branch("showerPur", &m_showerPur);




    TFile hfileExtra("/nova/ana/users/bianjm/share/development/RecoJMShower/histograms/extrahist_fd.root");
    ht_nue_rw = (TH1D*)hfileExtra.Get("ht_nue_rw");
    ht_bkg_nue_rw = (TH1D*)hfileExtra.Get("ht_bkg_nue_rw");
    ht_bkg_numu_rw = (TH1D*)hfileExtra.Get("ht_bkg_numu_rw");

  }

  void LIDTraining::beginSubRun(art::SubRun const& sr) {
/*
    art::Handle< sumdata::POTSum > p;
    sr.getByLabel(fPotLabel,p);
    std::cout<<"pot = "<<p->totgoodpot<<std::endl;
*/
  } 

  void LIDTraining::beginRun(const art::Run& run) 
  {
    art::ServiceHandle<geo::Geometry> fGeom;
    novadaq::cnv::DetId detId = fGeom->DetId();
    if (fParticleIDAlg) delete fParticleIDAlg;
    if (fNuEEnergyAlg)  delete fNuEEnergyAlg;
    fNuEEnergyAlg     = new NuEEnergyAlg(fNuEEnergyAlgPSet);
    fParticleIDAlg    = new ParticleIDAlg(fParticleIDAlgPSet, detId, fNuEEnergyAlg, false);

  }

  //......................................................................
  void LIDTraining::analyze(const art::Event& evt)
  {

    Ncount0++;

    //----------------------------------------------------
    //  Run, subrun and event information
    //----------------------------------------------------
    int run   = evt.run();
    int srun  = evt.subRun();
    int event = evt.id().event();
    std::cout<<"Event "<<event<<" ==============================================="<<std::endl;
    art::Handle<std::vector<rb::Cluster> > slicecol;
    evt.getByLabel(fSliceLabel, slicecol);
    art::PtrVector<rb::Cluster> slicelist;
    for(unsigned int i = 0; i < slicecol->size(); ++i){
      slicelist.push_back(art::Ptr<rb::Cluster>(slicecol, i));
    }

    art::ServiceHandle<cheat::BackTracker> bt;

    //Get all of the hits from calhits
    art::Handle< std::vector<rb::CellHit> > hitsHandle;
    evt.getByLabel(fHitLabel, hitsHandle);
    art::PtrVector<rb::CellHit> allHits;
    for(unsigned int i = 0; i < hitsHandle->size(); ++i){
      art::Ptr<rb::CellHit> hit(hitsHandle, i);
      allHits.push_back(hit);
    }
    // We're gonna get a vector of NeutrinoEffPurs for allHits
    std::vector<cheat::NeutrinoWithIndex> allNus = bt->allNeutrinoInteractions();
    
    std::vector< std::vector<cheat::NeutrinoEffPur> > sliceEffPurs;
    sliceEffPurs.resize(slicelist.size());

    for(unsigned int iSubvt = 0; iSubvt < slicelist.size(); ++iSubvt)
    {
      sliceEffPurs[iSubvt] = bt->SliceToMCTruth(slicelist[iSubvt]->AllCells(), allHits);//slicer4d
      
    }
    std::vector<int> faveIds = getFaveIds(allNus, sliceEffPurs);

    for(unsigned int sliceIdx = 0; sliceIdx < slicecol->size(); ++sliceIdx){//Reconstruct EID slice by slice
      if(rb::IsFiltered(evt,slicecol,sliceIdx) && fRespectFilter) continue;
      const rb::Cluster& slice = (*slicecol)[sliceIdx];
      if(slice.IsNoise()) continue;

      if (fRecordMC && !(evt.isRealData()) &&(faveIds[sliceIdx] == -1)) continue;

      m_evtTrueNuEnergy = 0;
      m_evtTrueNuP4[0]  = 0;
      m_evtTrueNuP4[1]  = 0;
      m_evtTrueNuP4[2]  = 0;
      m_evtTrueNuP4[3]  = 0;
      m_evtTrueNuVtx[0] = 0;
      m_evtTrueNuVtx[1] = 0;
      m_evtTrueNuVtx[2] = 0;

      for(int i=0;i<20;i++){
        m_evtRwRand[i] = -1;
        m_evtRwProb[i] = -1;
      }
      
      
      art::ServiceHandle<geo::Geometry> geom;
      art::ServiceHandle<calib::Calibrator> cal;
      
      
      //--  ---------------------------------------------------------------------------------------------
      //  MC true information
      //-----------------------------------------------------------------------------------------------
      //Neutrino MC truth
      TLorentzVector evtTrueNuP4(0,0,0,0);
      TVector3 evtTrueNuVtx(0,0,-100);
      int evtTrueNuCCNC = -1;
      int evtTrueNuMode = -1;
      int evtTrueNuPdg = 0;

      art::Ptr<simb::MCTruth> mc;
      if(!evt.isRealData()&&fRecordMC) mc = sliceEffPurs[sliceIdx][0].neutrinoInt;
      std::vector<const sim::Particle*> plist;
      plist.clear();
      if(!evt.isRealData()&&fRecordMC)plist = bt->MCTruthToParticles(mc);

      if(!evt.isRealData()&&fRecordMC){
        if (mc->NeutrinoSet()) {
          const simb::MCNeutrino& gen = mc->GetNeutrino();
          const simb::MCParticle& nu  = gen.Nu();
          evtTrueNuP4 = nu.Momentum();
          evtTrueNuVtx.SetXYZ(nu.EndX(),nu.EndY(),nu.EndZ());
          evtTrueNuCCNC = gen.CCNC() ;
          evtTrueNuMode = gen.Mode() ;
          evtTrueNuPdg = nu.PdgCode() ;
        }
      }
      
      
      
      m_evtTrueNuCCNC = evtTrueNuCCNC;
      m_evtTrueNuMode = evtTrueNuMode;
      m_evtTrueNuPdg = evtTrueNuPdg;
      
      
      m_evtTrueNuEnergy = evtTrueNuP4.E();
      m_evtTrueNuP4[0] = evtTrueNuP4.Px();
      m_evtTrueNuP4[1] = evtTrueNuP4.Py();
      m_evtTrueNuP4[2] = evtTrueNuP4.Pz();
      m_evtTrueNuP4[3] = evtTrueNuP4.E();
      m_evtTrueNuVtx[0] = evtTrueNuVtx[0];
      m_evtTrueNuVtx[1] = evtTrueNuVtx[1];
      m_evtTrueNuVtx[2] = evtTrueNuVtx[2];
    

      TRandom rand0(run*1100000000+srun*1000000+event);
      double randnum0 = rand0.Uniform(0, 1);
      TRandom rand1(run*1200000000+srun*100000+event);
      double randnum1 = rand1.Uniform(0, 1);
      TRandom rand2(run*1300000000+srun*1000000+event);
      double randnum2 = rand2.Uniform(0, 1);
 
      double p_nue_rw       = ht_nue_rw->Interpolate(evtTrueNuP4.E());
      double p_bkg_nue_rw   = 1;//ht_bkg_nue_rw->Interpolate(evtTrueNuP4.E());
      double p_bkg_numu_rw  = 1;//ht_bkg_numu_rw->Interpolate(evtTrueNuP4.E());
 
 
      m_evtRwRand[0] = randnum0;
      m_evtRwRand[1] = randnum1;
      m_evtRwRand[2] = randnum2;
      m_evtRwProb[0] = p_nue_rw;
      m_evtRwProb[1] = p_bkg_nue_rw;
      m_evtRwProb[2] = p_bkg_numu_rw;

      std::map<geo::OfflineChan, Vint> truePartPdg;
      truePartPdg.clear();

      TLorentzVector electruep4(0,0,0,0);
      std::vector<TLorentzVector> evtntpcol;
      std::vector<TVector3> evtntvtxcol;
      std::vector<TVector3> evtntpxzcol;
      std::vector<TVector3> evtntpyzcol;
      std::vector<int> evtntpdgcol;
      std::vector<int> evtntmothercol;

      evtntpcol.clear(); 
      evtntvtxcol.clear();
      evtntpxzcol.clear();
      evtntpyzcol.clear();
      evtntpdgcol.clear();
      evtntmothercol.clear();

      if(!evt.isRealData()&&fRecordMC){
        for (unsigned int kk=0;kk<plist.size();kk++){    
          TLorentzVector ntp(0,0,0,0);                   
          TVector3 ntvtx(0,0,0);                         
          TVector3 ntpxz(0,0,0);                         
          TVector3 ntpyz(0,0,0);                         
          int ntpdg =0;                                  
          int ntmother = -1;                             
          if(abs(plist[kk]->PdgCode()) == 11 && plist[kk]->Mother()==0){
            electruep4 = plist[kk]->Momentum();
          }
          ntpdg = plist[kk]->PdgCode();                  
          ntp = plist[kk]->Momentum();                   
          ntpxz.SetXYZ(ntp.Px(),0,ntp.Pz());             
          ntpyz.SetXYZ(0,ntp.Py(),ntp.Pz());             
          ntmother = plist[kk]->Mother();                
          ntvtx.SetXYZ(plist[kk]->Vx(),plist[kk]->Vy(),  plist[kk]->Vz());
          evtntpcol.push_back(ntp);                      
          evtntpxzcol.push_back(ntpxz);                  
          evtntpyzcol.push_back(ntpyz);                  
          evtntvtxcol.push_back(ntvtx);
          evtntpdgcol.push_back(ntpdg); 
          evtntmothercol.push_back(ntmother);
        }
      }

      m_evtTotalRecoGeV = 0; 
      m_evtEtot = 0;
      m_evtPi0Mgg  = 0;
      m_evtPi0StartDist = 9999;
      m_evtPi0LineDist = 9999;
      m_evtSh1Energy = 0;
      m_evtSh1Energy0 = 0;
      m_evtSh1Length = 0;
      m_evtSh1NueEnergy = 0;
      m_evtSh1Gap = 0;
      m_evtSh1P4[0] = 0;
      m_evtSh1P4[1] = 0;
      m_evtSh1P4[2] = 0;
      m_evtSh1P4[3] = 0;
      m_evtSh1Start[0] = 0;
      m_evtSh1Start[1] = 0;
      m_evtSh1Start[2] = 0;
      m_evtSh1TrueDang = 0;
      m_evtSh1TrueEDang = 0;
      m_evtSh1TruePdg = 0;
      m_evtSh1IsFid = 0;
      m_evtSh1IsFidAna = 0;
      for(int itype=0;itype<20;itype++){
        m_evtSh1DedxLLL[itype] = -9999;
        m_evtSh1DedxLLT[itype] = -9999;
      }
      m_evtSh1NPlane = 0;
      m_evtSh1NMIPPlane = 0;
      m_evtSh1VtxGeV = 0;
      m_evtSh1Pi0Mgg = 0;
      m_evtSh1CosTheta = 0;
      m_evtSh1SumPt=0;
      m_evtSh2Energy = 0;
      m_evtSheEnergy = 0;

      //-----------------------------------------------------
      //  Read  slice, hit and track;
      //-----------------------------------------------------

      std::map<geo::OfflineChan, int> hitslicemap;

      std::set<art::Ptr<rb::CellHit> > cellhitlist;
      std::vector<const rb::Cluster* >   cluster;


      double totalRecoGeV = 0;
      art::Ptr<rb::Cluster>clust(slicecol,sliceIdx);

      //get verticies associated with each slice
      art::FindManyP<rb::Vertex> fmv(slicecol, evt, fVertexLabel);

      const double wx = (clust->NXCell() > 0) ? clust->W(clust->XCell(0).get()) : 0;
      const double wy = (clust->NYCell() > 0) ? clust->W(clust->YCell(0).get()) : 0;
      for(unsigned int i=0;i<clust->NXCell();i++){
        const art::Ptr<rb::CellHit>& chit = clust->XCell(i);

        const rb::RecoHit rhit(cal->MakeRecoHit(*chit, wx ));
        if(rhit.IsCalibrated()){
          if((clust->XCell(i)->PE())>fCellHitPHThresh ){
            totalRecoGeV += rhit.GeV();
          }
        }
      }
      for(unsigned int i=0;i<clust->NYCell();i++){
        const art::Ptr<rb::CellHit>& chit = clust->YCell(i);
        const rb::RecoHit rhit(cal->MakeRecoHit(*chit, wy));

        if(rhit.IsCalibrated()){
           if((clust->YCell(i)->PE())>fCellHitPHThresh ){
             totalRecoGeV += rhit.GeV();
           }
        }
      }
      m_evtTotalRecoGeV = totalRecoGeV;

      //After read slice, hit and tracks.
      // Determination of start point, end point and direction
      //  Fill true information of showers
      std::vector<double> ntdangcol;
      std::vector<TLorentzVector> ntpcol;
      std::vector<TVector3> ntvtxcol;
      std::vector<TVector3> ntstopcol;
      std::vector<int> ntpdgcol;
      ntdangcol.clear();
      ntpcol.clear();
      ntvtxcol.clear();
      ntstopcol.clear();
      ntpdgcol.clear();
   
      std::vector<art::Ptr<slid::ShowerLID> > slidcol;
      slidcol.clear();
      art::FindManyP<slid::ShowerLID> fmslid(slicecol, evt, fShowerLIDLabel);
      if(fmslid.isValid()){
        slidcol = fmslid.at(sliceIdx);
      }
      std::vector<const rb::Shower* > showercol;
      showercol.clear();
      art::FindMany<rb::Shower> fmshw(slicecol, evt, fShowerLabel);
      if(fmshw.isValid()){
        showercol = fmshw.at(sliceIdx);
      }

      art::FindOneP<rb::Shower> fos(slidcol, evt, fShowerLIDLabel);

      std::vector<double>  showersumptcol;
      std::vector<TLorentzVector>  showersump4col;
      showersumptcol.clear();
      showersump4col.clear();
      
      for(unsigned int i = 0; i < slidcol.size(); i++){

	std::cout<<"Shower "<<i<<" ==============================================="<<std::endl;

        //For cell energy check and mc truth 
        cet::maybe_ref<art::Ptr<rb::Shower> const> rshw(fos.at(i));
        art::Ptr<rb::Shower> shw = rshw.ref();
	art::PtrVector< rb::CellHit > showerHits = shw->AllCells();
        TVector3 trkdir=shw->Dir();
        TVector3 trkstart=shw->Start();
        TVector3 trkstop=shw->Stop();
        double ntdang= 200.0;
        TLorentzVector ntp(0,0,0,0);
        TVector3 ntvtx(0,0,0);
        TVector3 ntstop(0,0,0);
        int ntpdg =0;
	int bestang =-1;
        if(!evt.isRealData()&&fRecordMC){
	
	  if( fTruthMatchByEnergy ){

	    const std::vector< cheat::TrackIDE > trackides = bt->HitsToTrackIDE( showerHits );
	    if( !trackides.empty() ){
	      const sim::Particle * part = bt->TrackIDToParticle( trackides[0].trackID );
	      ntpdg  = part->PdgCode();
	      double angle     = part->Momentum().Vect().Angle( trkdir ) ;
	      double backangle = part->Momentum().Vect().Angle( -trkdir ) ;
	      ntdang = std::min( angle, backangle );
	      std::set< int > setid;
	      setid.insert( trackides[0].trackID );
	      m_showerPur = bt->HitCollectionPurity( setid, showerHits);
	      m_showerEff = bt->HitCollectionEfficiency( setid, showerHits, slicelist[sliceIdx]->AllCells(), geo::kXorY);
	      std::cout<<"Pdg code from energy "<<ntpdg<<std::endl;
	    }
	    else ntpdg = -1;
	  }// end if truthMatchByEnergy
	  else{
	    int partid;
	    for (unsigned int kk=0;kk<plist.size();kk++){
	      int abspdg = abs(plist[kk]->PdgCode());
	      if(abspdg==12||abspdg==14||abspdg==16)continue;

	      TLorentzVector true_4mom = plist[kk]->Momentum();
	      if(true_4mom.Vect().Mag()<0.000001)continue;
	      double ntangd = trkdir.Angle(true_4mom.Vect());
	      ntangd=fmod(ntangd, TMath::Pi());
	      // if(ntangd>TMath::Pi()/2)
	      // 	ntangd = TMath::Pi()-ntangd;
	      if(ntangd<ntdang){
		ntdang = ntangd;
		bestang = plist[kk]->PdgCode();
		ntpdg = plist[kk]->PdgCode();
		ntp = true_4mom;
		ntvtx.SetXYZ(plist[kk]->Vx(),plist[kk]->Vy(),plist[kk]->Vz());
		ntstop.SetXYZ(plist[kk]->EndX(),plist[kk]->EndY(),plist[kk]->EndZ());
		partid = plist[kk]->TrackId();
	      }
	    }
	    std::cout<<"Pdg code from angle "<<bestang<<std::endl;
	    std::set< int > setid;
	    setid.insert( partid );
	    m_showerPur = bt->HitCollectionPurity( setid, showerHits);
	    m_showerEff = bt->HitCollectionEfficiency( setid, showerHits, slicelist[sliceIdx]->AllCells(), geo::kXorY );
	      
	  }
	}
        ntdang = ntdang*TMath::RadToDeg();
	std::cout<<"dangle "<<ntdang<<std::endl;
        ntdangcol.push_back(ntdang);
        ntpcol.push_back(ntp);
        ntvtxcol.push_back(ntvtx);
        ntstopcol.push_back(ntstop);
        ntpdgcol.push_back(ntpdg);
      
      }


      //   return;//cell only
      //**********************************************************
      //         Fill showers information
      //**********************************************************

      Vp4 showerpGam;
      showerpGam.clear();
     
      double showeretot = 0;

      int ish1 = -1;
      int ish2 = -1;

      double sh1gev = 0;
      double sh2gev = 0;
      TLorentzVector sh1p4(0,0,0,0);
      TLorentzVector trackpTot(0,0,0,0);
      TLorentzVector showerSumP4(0,0,0,0);

      for(int i = 0; i < (int)slidcol.size(); i++){
	cet::maybe_ref<art::Ptr<rb::Shower> const> rshw(fos.at(i));
        art::Ptr<rb::Shower> shw = rshw.ref();
        double testE = fNuEEnergyAlg->ShowerEnergy(shw.get(), showercol, evt);
	if(testE > sh1gev){
          ish1 = i;
          sh1gev = testE; 
        }
      }

      for(int i = 0; i < (int)slidcol.size(); i++){
        if(i==ish1)continue;
        cet::maybe_ref<art::Ptr<rb::Shower> const> rshw(fos.at(i));
        art::Ptr<rb::Shower> shw = rshw.ref();
        double testE = fNuEEnergyAlg->ShowerEnergy(shw.get(), showercol, evt);
        if(testE > sh2gev){
          ish2 = i;
          sh2gev = testE; 
        }
      }

      TVector3 nuDir = geom->NuMIBeamDirection(); //(-0.845e-3, 52.552e-3, 0.998618);
      for(int i = 0; i < (int)slidcol.size(); i++){
        cet::maybe_ref<art::Ptr<rb::Shower> const> rshw(fos.at(i));
        art::Ptr<rb::Shower> shw = rshw.ref();
        // Set 4-Vector momentum
        double showerEnergy = fNuEEnergyAlg->ShowerEnergy(shw.get(), showercol, evt);
        double showereraw = showerEnergy;
        TVector3 p3(shw->Dir()[0]*showereraw,shw->Dir()[1]*showereraw,shw->Dir()[2]*showereraw);
        double maxll = -9999;
        int lltype = -1;
        std::map<int, float> tempLLL;
        std::map<int, float> tempLLT;
        tempLLL.clear();
        tempLLT.clear();
        tempLLL = slidcol[i]->fPartLongLL;
        tempLLT = slidcol[i]->fPartTransLL;
        for( int itype = 0; itype <11; ++itype){
          if(tempLLL[itype] + tempLLT[itype]  > maxll){
            lltype = itype;
            maxll = tempLLL[itype] + tempLLT[itype];
          }
        }
        double m = xmass[0];
        if(lltype>=0&&lltype<7) m=xmass[lltype];
        double showerEtot0 = sqrt(p3.X()*p3.X()+p3.Y()*p3.Y()+p3.Z()*p3.Z());
        double P3 = sqrt(showerEtot0*showerEtot0+2*showerEtot0*m);
        double E = showerEtot0+m;
        TLorentzVector showerptrk(shw->Dir()[0]*P3,shw->Dir()[1]*P3,shw->Dir()[2]*P3,E);
        showerpGam.push_back(showerptrk);
        showeretot+= showerEnergy;
        showerSumP4+=showerptrk;
      }

      for(unsigned int i = 0; i < slidcol.size(); i++){
        cet::maybe_ref<art::Ptr<rb::Shower> const> rshw(fos.at(i));
        art::Ptr<rb::Shower> shw = rshw.ref();

        fParticleIDAlg->SetShower(shw.get(), showercol, evt);

        // General information 
        double showerEnergy    = fNuEEnergyAlg->ShowerEnergy(shw.get(), showercol, evt);

        int showerIsFidAna = 0;
        if ((fabs(shw->Start()[0]) < fXThresh) &&
            (fabs(shw->Stop()[0])  < fXThresh) &&
            (fabs(shw->Start()[1]) < fYThresh) &&
            (fabs(shw->Stop()[1])  < fYThresh) &&
            (std::min(shw->Start()[2],shw->Stop()[2]) > fZMin) &&
            (std::max(shw->Start()[2],shw->Stop()[2]) < fZMax)) showerIsFidAna = 1;

        int showerIsFid = 1;
        for (unsigned int pln=0; pln<shw->ExtentPlane(); ++pln){
          double plnrad = fParticleIDAlg->PlaneRadius(pln);
          TVector3 plncoord = fParticleIDAlg->PlaneHitXYZ(pln);
          double xdistedge = std::min(fabs(geom->DetHalfWidth()-plncoord[0]),fabs(-geom->DetHalfWidth()-plncoord[0]));
          double ydistedge = std::min(fabs(geom->DetHalfHeight()-plncoord[1]),fabs(-geom->DetHalfHeight()-plncoord[1]));
          double zdistedge = std::min(plncoord[2],fabs(geom->DetLength()-plncoord[2]));
          if ((xdistedge > 2.0*plnrad) && (ydistedge > 2.0*plnrad) && (zdistedge > 2.0*plnrad)) continue;
          showerIsFid = 0;
          break; //can stop checking once we have found one plane outside volume
        }

        m_showerTrueNuCCNC = evtTrueNuCCNC;
        m_showerTrueNuMode = evtTrueNuMode;
        m_showerTrueNuPdg = evtTrueNuPdg;
        m_showerTrueDang = ntdangcol[i];
        m_showerTruePdg = ntpdgcol[i];
        m_showerTrueP4[0] = ntpcol[i].Px();
        m_showerTrueP4[1] = ntpcol[i].Py();
        m_showerTrueP4[2] = ntpcol[i].Pz();
        m_showerTrueP4[3] = ntpcol[i].E();
        m_showerNPlane = shw->ExtentPlane();
        m_showerEnergy = showerEnergy;
        m_showerStart[0]  = (shw->Start())[0];
        m_showerStart[1]  = (shw->Start())[1];
        m_showerStart[2]  = (shw->Start())[2];
        m_showerStop[0]  = (shw->Stop())[0];
        m_showerStop[1]  = (shw->Stop())[1];
        m_showerStop[2]  = (shw->Stop())[2];
        m_showerDir[0]   = (shw->Dir())[0];
        m_showerDir[1]   = (shw->Dir())[1];
        m_showerDir[2]   = (shw->Dir())[2];
        m_showerIsFidAna = showerIsFidAna;
        m_showerIsFid = showerIsFid;

        // Fill longitudinal and transverse information
        for(int itd=0;itd<20;itd++){
          m_showerTransCellDedx[itd] = 0;
        }
        for(unsigned int plane=0;plane<200;plane++){
          m_showerPlaneDedx[plane] = 0;
        }

        for(unsigned int plane=0;plane<shw->ExtentPlane();plane++){
          if(plane>=200)break;
          double tdedx = fParticleIDAlg->PlaneLongDedx(plane);
          if (tdedx< 0) tdedx=0;
          m_showerPlaneDedx[plane] = tdedx;
        }
        for(unsigned itp = 0;itp<20;itp++){
          double tdedx = fParticleIDAlg->PlaneTransDedx(itp);
          if (tdedx< 0) tdedx=0;
          m_showerTransCellDedx[itp] = tdedx;
        }

	m_showerRun= run;
	m_showerSubRun= srun;
	m_showerEvent= event;
	m_showerSubEvent= (int)sliceIdx;
	m_showerId= i;
	std::cout<<"GOING TO FILL SHOWER TREE\n";
        fShower->Fill();
      }

      //****************************************
      // Fill event level information
      //****************************************

      if(ish1!=-1){
        cet::maybe_ref<art::Ptr<rb::Shower> const> rshw(fos.at(ish1));
        art::Ptr<rb::Shower> shw = rshw.ref();

        fParticleIDAlg->SetShower(shw.get(), showercol, evt);

        m_evtSh1Energy = fNuEEnergyAlg->ShowerEnergy(shw.get(), showercol, evt);
        if ((fabs(shw->Start()[0]) < fXThresh) &&
            (fabs(shw->Stop()[0])  < fXThresh) &&
            (fabs(shw->Start()[1]) < fYThresh) &&
            (fabs(shw->Stop()[1])  < fYThresh) &&
            (std::min(shw->Start()[2],shw->Stop()[2]) > fZMin) &&
            (std::max(shw->Start()[2],shw->Stop()[2]) < fZMax)) m_evtSh1IsFidAna = 1;

        m_evtSh1IsFid = 1;
        for (unsigned int pln=0; pln<shw->ExtentPlane(); ++pln){
          double plnrad = fParticleIDAlg->PlaneRadius(pln);
          TVector3 plncoord = fParticleIDAlg->PlaneHitXYZ(pln);
          double xdistedge = std::min(fabs(geom->DetHalfWidth()-plncoord[0]),fabs(-geom->DetHalfWidth()-plncoord[0]));
          double ydistedge = std::min(fabs(geom->DetHalfHeight()-plncoord[1]),fabs(-geom->DetHalfHeight()-plncoord[1]));
          double zdistedge = std::min(plncoord[2],fabs(geom->DetLength()-plncoord[2]));
          if ((xdistedge > 2.0*plnrad) && (ydistedge > 2.0*plnrad) && (zdistedge > 2.0*plnrad)) continue;
          m_evtSh1IsFid = 0;
          break;  //can stop checking once we have found one plane outside the volume
        }

        m_evtSh1P4[0] = showerpGam[ish1].Px();
        m_evtSh1P4[1] = showerpGam[ish1].Py();
        m_evtSh1P4[2] = showerpGam[ish1].Pz();
        m_evtSh1P4[3] = showerpGam[ish1].E();
        m_evtSh1Start[0] = shw->Start()[0];
        m_evtSh1Start[1] = shw->Start()[1];
        m_evtSh1Start[2] = shw->Start()[2];
        m_evtSh1TrueDang = TMath::RadToDeg()*showerpGam[ish1].Angle(ntpcol[ish1].Vect());//ntdangcol[ish1];
	m_evtSh1TrueEDang = TMath::RadToDeg()*showerpGam[ish1].Angle(electruep4.Vect());
        m_evtSh1TruePdg = ntpdgcol[ish1];

        double ellmax = -9999;
        double elllmax = -9999;
        double elltmax = -9999;
        std::map<int, float> tempLLL;
        std::map<int, float> tempLLT;
        tempLLL.clear();
        tempLLT.clear();
        tempLLL = slidcol[ish1]->fPartLongLL;
        tempLLT = slidcol[ish1]->fPartTransLL;
        for(int itype=0; itype != int(slid::DedxParticleType::kDEDXPARTICLETYPESIZE); ++itype){
          m_evtSh1DedxLLL[itype] = tempLLL[itype];
          m_evtSh1DedxLLT[itype] = tempLLT[itype];
          if(itype == int(slid::DedxParticleType::kELECTRON)||itype == int(slid::DedxParticleType::kELECTRONCCQE)||
             itype == int(slid::DedxParticleType::kELECTRONCCRES)||itype == int(slid::DedxParticleType::kELECTRONCCDIS)||
             itype == int(slid::DedxParticleType::kELECTRONCCCOH))
          {
            if(ellmax<tempLLL[itype]+tempLLT[itype]){
              ellmax = tempLLL[itype]+tempLLT[itype];
              elllmax = tempLLL[itype];
              elltmax = tempLLT[itype];
            }
          }
        }
        m_evtSh1DedxLLL[0] = elllmax;
        m_evtSh1DedxLLT[0] = elltmax;
        m_evtSh1VtxGeV = slidcol[ish1]->VertexEnergy();
        m_evtSh1Pi0Mgg = slidcol[ish1]->Pi0mass();
        m_evtSh1Gap = slidcol[ish1]->Gap();
        m_evtSh1NPlane = shw->ExtentPlane();
        m_evtSh1NMIPPlane = slidcol[ish1]->NMIPPlanes();
        m_evtSh1NueEnergy = slidcol[ish1]->NueEnergy();
        double showerSumTheta = showerSumP4.Angle(nuDir); 
        double showerSumPt = showerSumP4.Vect().Mag()*sin(showerSumTheta);
        m_evtSh1SumPt = showerSumPt; 
        m_evtSh1CosTheta = slidcol[ish1]->CosTheta();
        m_evtSh1Length = shw->TotalLength();

        if(ish2!=-1){
          cet::maybe_ref<art::Ptr<rb::Shower> const> rshw2(fos.at(ish2));
          art::Ptr<rb::Shower> shw2 = rshw2.ref();
          m_evtSh2Energy = fNuEEnergyAlg->ShowerEnergy(shw.get(), showercol, evt);

          cet::maybe_ref<art::Ptr<rb::Shower> const> rshw1(fos.at(ish1));
          art::Ptr<rb::Shower> shw1 = rshw1.ref();

          m_evtPi0StartDist = (shw1->Start()-shw2->Start()).Mag();

          double x1 = shw1->Start()[0];
          double y1 = shw1->Start()[1];
          double z1 = shw1->Start()[2];
          double p1 = (shw1->Stop()[0] - shw1->Start()[0])/(shw1->Stop() - shw1->Start()).Mag();
          double q1 = (shw1->Stop()[1] - shw1->Start()[1])/(shw1->Stop() - shw1->Start()).Mag();
          double r1 = (shw1->Stop()[2] - shw1->Start()[2])/(shw1->Stop() - shw1->Start()).Mag();
          double x2 = shw2->Start()[0];
          double y2 = shw2->Start()[1];
          double z2 = shw2->Start()[2];
          double p2 = (shw2->Stop()[0] - shw2->Start()[0])/(shw2->Stop() - shw2->Start()).Mag();
          double q2 = (shw2->Stop()[1] - shw2->Start()[1])/(shw2->Stop() - shw2->Start()).Mag();
          double r2 = (shw2->Stop()[2] - shw2->Start()[2])/(shw2->Stop() - shw2->Start()).Mag();
          
          TMatrixD dNorm(3,3);
          dNorm[0][0] = x2-x1;
          dNorm[0][1] = y2-y1;
          dNorm[0][2] = z2-z1;
          dNorm[1][0] = p1;
          dNorm[1][1] = q1;
          dNorm[1][2] = r1;
          dNorm[2][0] = p2;
          dNorm[2][1] = q2;
          dNorm[2][2] = r2;

          double ma = x2-x1;
          double mb = y2-y1;
          double mc = z2-z1;
          double md = p1;
          double me = q1;
          double mf = r1;
          double mg = p2;
          double mh = q2;
          double mi = r2;

          
          TMatrixD dDeNorm1(2,2);
          TMatrixD dDeNorm2(2,2);
          TMatrixD dDeNorm3(2,2);
          dDeNorm1[0][0] = p1;
          dDeNorm1[0][1] = q1;
          dDeNorm1[1][0] = p2;
          dDeNorm1[1][1] = q2;
          double det1 = dDeNorm1[0][0]*dDeNorm1[1][1] - dDeNorm1[0][1]*dDeNorm1[1][0];
          dDeNorm2[0][0] = q1;
          dDeNorm2[0][1] = r1;
          dDeNorm2[1][0] = q2;
          dDeNorm2[1][1] = r2;
          double det2 = dDeNorm2[0][0]*dDeNorm2[1][1] - dDeNorm2[0][1]*dDeNorm2[1][0];
          dDeNorm3[0][0] = r1;
          dDeNorm3[0][1] = p1;
          dDeNorm3[1][0] = r2;
          dDeNorm3[1][1] = p2;
          double det3 = dDeNorm3[0][0]*dDeNorm3[1][1] - dDeNorm3[0][1]*dDeNorm3[1][0];
          double det0 = ma*me*mi + mb*mf*mg + mc*md*mh - mc*me*mg - mb*md*mi - ma*mf*mh;
          double pi0LineDist = fabs(det0)/sqrt(det1*det1+det2*det2+det3*det3); 
          m_evtPi0LineDist = pi0LineDist; 
        }
      }
      m_evtEtot = showeretot; 
      fEvent->Fill();
    }
    return; 
  }

  void LIDTraining::endJob()
  {
    std::cout<<"Total number of events   "<<Ncount0<<std::endl;
  }
  
  std::vector<int> LIDTraining::getFaveIds(std::vector<cheat::NeutrinoWithIndex>& allNus, 
                                        std::vector< std::vector<cheat::NeutrinoEffPur> >& subevtNus)
  {
    // This method associates each subevt with its favorite neutrino. 
    // The first argument is all the neutrinos, the second has an vector for each subevt (in the same order)
    // The inner vectors contain all the neutrinos that made hits in the subevt
    // with efficiency and purity calculated by backtracker.  
    
    
    // We'll make an array with an entry for every subevt to store the unique identifier of 
    // the favourite neutrino within the subevt, with the correct size and all entries initialized to -1
    std::vector<int> faveIds(subevtNus.size(), -1);
    
    //loop over subevts
    // loop over nus within the subevts to find the one with the highest "metric"
    // if that slice has the highest "metric" value (efficiency for now)
    //store its uniqueId in the array.
    for(unsigned int iSubvt = 0; iSubvt < subevtNus.size(); ++iSubvt)
    {
      //for the moment we can just take the first neutrino in each subevt's vectors, 
      //since these will be sorted by efficiency.  
      // Later there might be another metric.
      if(subevtNus[iSubvt].size() > 0) {
        faveIds[iSubvt] = subevtNus[iSubvt][0].truthColIndex;
      }
    }
    
    
    // then loop over nus
    //  if a nuUniqueId exists more than once, 
    //determine which subevt has the highest efficiency
    // restore the other subevt entries in faveIds to -1.
    for(unsigned int iNu = 0; iNu < allNus.size(); ++iNu){
      std::vector<int> associatedSubevts;
      // loop over faveIds to find out how many times each nu was stored.
      for(unsigned int iSubvt = 0; iSubvt < faveIds.size(); ++iSubvt)
      {
        if((int)allNus[iNu].truthColIndex == faveIds[iSubvt])
        {
          associatedSubevts.push_back(iSubvt);
        }
        
      }
      
      
      if(associatedSubevts.size() > 1){// then there are subevts with the same faveNu.
        float maxEff = 0;
        int bestSubevt = -2; // -2 for debugging.
        // loop over all associatedSubevts to determine which has hightest efficiency        
        for(unsigned int iAssc = 0; iAssc < associatedSubevts.size(); ++iAssc){
          if(subevtNus[associatedSubevts[iAssc]][0].efficiency > maxEff){
            maxEff = subevtNus[associatedSubevts[iAssc]][0].efficiency;
            bestSubevt = associatedSubevts[iAssc];
          }//end if
        }// end for(iAssc ... )
        assert(bestSubevt >= 0 && "Best subevt not found"); 
        for(unsigned int iAssc = 0; iAssc < associatedSubevts.size(); ++iAssc){
          if(associatedSubevts[iAssc] != bestSubevt){ 
            // then this has been incorrectly identified.
            faveIds[associatedSubevts[iAssc]] = -1; 
          }//end if
        }// end for(iAssc ... )
        
      }// end if(#asscSubevts > 1)
      
    }
    
    return faveIds;
  }

  DEFINE_ART_MODULE(LIDTraining);

}
//......................................................................
