////////////////////////////////////////////////////////////////////////
// \file    ShowerLID.h
// \brief   A Shower class that records the log likelihood information 
//          for every shower.
// \version 
// \author  Kanika Sachdev (ksachdev@physics.umn.edu)
////////////////////////////////////////////////////////////////////////
#ifndef SHOWERLID_H
#define SHOWERLID_H

#include "RecoBase/PID.h"

#include <string>
#include <map>

namespace slid 
{

  class ShowerLID : public rb::PID
  {
  public:
    
    inline ShowerLID();
    ~ShowerLID(){}
#ifndef __GCCXML__
    ShowerLID( const rb::PID & pid);


    /// Setters
    void SetValWithE     (float in) { fPIDWithE = in; }
    void SetEGLLL        (float in) { fEGLLL    = in; }
    void SetEGLLT        (float in) { fEGLLT    = in; }
    void SetEMuLLL       (float in) { fEMuLLL   = in; }
    void SetEMuLLT       (float in) { fEMuLLT   = in; }
    void SetEPi0LLL      (float in) { fEPi0LLL  = in; }
    void SetEPi0LLT      (float in) { fEPi0LLT  = in; }
    void SetEPLLL        (float in) { fEPLLL    = in; }
    void SetEPLLT        (float in) { fEPLLT    = in; }
    void SetENLLL        (float in) { fENLLL    = in; }
    void SetENLLT        (float in) { fENLLT    = in; }
    void SetEPiLLL       (float in) { fEPiLLL   = in; }
    void SetEPiLLT       (float in) { fEPiLLT   = in; }
    void SetGap          (float in) { fGap      = in; }
    void SetPi0mass      (float in) { fPi0mass  = in; }
    void SetPi0PhotonID  (float in) { fPi0Id    = in; }
    void SetShowerEFrac  (float in) { fShwEFrac = in; }
    void SetCosTheta     (float in) { fCosTheta = in; }
    void SetIsMuon       (bool  in) { fIsMuon   = in; }

    void SetRadius       (float in) { fRadius = in; }

    void SetInverseGLLL  (float in) { fInvGLLL= in; }
    void SetNMIPPlanes   (float in) { fNMIP   = in; }

    void SetVtxDoca      (float in) { fVtxDoca = in; }

    /// \brief set algorithm name
    void SetAlgName(const std::string algName) { fAlgName = algName; }

    /// \brief set algorithm description
    void SetAlgDescription(const std::string algDescription) { fAlgDescription = algDescription; }

    void SetShowerDepEnergy (float in) { fShwDepE    = in; }
    void SetVertexEnergy    (float in) { fVtxE       = in; }
    void SetShowerEnergy    (float in) { fShwE       = in; }
    void SetNueEnergy       (float in) { fNueEnergy  = in; }
    void SetHadronicEnergy  (float in) { fHadE       = in; }

   
    // Getters                                                                                                                                  
    std::string AlgName        () const  { return fAlgName;        }
    std::string AlgDescription () const  { return fAlgDescription; }

    float ValueWithE   () const { return fPIDWithE; }

    float EGLLL      () const { return fEGLLL;    }
    float EGLLT        () const { return fEGLLT;    }
    float EMuLLL       () const { return fEMuLLL;   }
    float EMuLLT       () const { return fEMuLLT;   }
    float EPi0LLL      () const { return fEPi0LLL;  }
    float EPi0LLT      () const { return fEPi0LLT;  }
    float EPLLL        () const { return fEPLLL;    }
    float EPLLT        () const { return fEPLLT;    }
    float ENLLL        () const { return fENLLL;    }
    float ENLLT        () const { return fENLLT;    }
    float EPiLLL       () const { return fEPiLLL;   }
    float EPiLLT       () const { return fEPiLLT;   }
    float Gap          () const { return fGap;      }
    float Pi0mass      () const { return fPi0mass;  }
    int   Pi0PhotonID  () const { return fPi0Id;  }
    float ShowerEFrac  () const { return fShwEFrac; }
    float CosTheta     () const { return fCosTheta; }
    bool  IsMuon       () const { return fIsMuon; }

    float Radius       () const { return fRadius; }

    float InverseGLLL  () const { return fInvGLLL;}
    int   NMIPPlanes    () const { return fNMIP;   }

    float VtxDoca      () const { return fVtxDoca; }

    float VertexEnergy      () const  { return fVtxE;          }
    float ShowerDepEnergy   () const  { return fShwDepE;       }
    float ShowerEnergy      () const  { return fShwE;          }
    float NueEnergy         () const  { return fNueEnergy;     }
    float HadronicEnergy    () const  { return fHadE;          }

#endif // __GCCXML__

    bool IsInversePhoton() const;
  protected:
    /// Algorithm name
    std::string fAlgName;
    
    /// Algorithm description
    std::string fAlgDescription;


  public:
    /// \brief training variables

    float fPIDWithE;

    float fEGLLL;
    float fEGLLT;
    float fEMuLLL;
    float fEMuLLT;
    float fEPi0LLL;
    float fEPi0LLT;
    float fEPLLL;
    float fEPLLT;
    float fENLLL;
    float fENLLT;
    float fEPiLLL;
    float fEPiLLT;

    float fInvGLLL;
    int   fNMIP;

    float fGap;
    float fPi0mass;
    int   fPi0Id;
    float fRadius;
    bool  fIsMuon;
    float fCosTheta;
    float fShwEFrac;

    float fVtxDoca;


    float fVtxE;
    float fShwDepE;
    float fShwE;
    float fNueEnergy;
    float fHadE;

    /// Map of the longitudinal ll by paricle type
    std::map<int, float> fPartLongLL;

    /// Map of the transverse ll by particle type
    std::map<int, float> fPartTransLL;
  };

  inline ShowerLID::ShowerLID()
    :
    fPIDWithE(-5.0),
    fEGLLL(-5.0), fEGLLT(-5.0), 
    fEMuLLL(-5.0), fEMuLLT(-5.0), 
    fEPi0LLL(-5.0), fEPi0LLT(-5.0), 
    fEPLLL(-5.0), fEPLLT(-5.0), 
    fENLLL(-5.0), fENLLT(-5.0), 
    fEPiLLL(-5.0), fEPiLLT(-5.0), 
    fInvGLLL(-5.0), fNMIP(-5), fGap(-5.0), 
    fPi0mass(-5.0), fPi0Id(-5),
    fRadius(-5.0), fIsMuon(false),
    fCosTheta(-5.0), fShwEFrac(-5.0),
    fVtxDoca(-5.0), 
    fVtxE(-5.0),fShwDepE(-5.0), 
    fShwE(-5.0), fNueEnergy(-5.0),
    fHadE(-5.0)
    {
  }

  bool CompareByEnergy( const slid::ShowerLID* a, const slid::ShowerLID* b);
}
#endif
