
/**  TMVA-Based Electron ID classification training and diagnostics.
 *
 *    This class is used to train a specified classifier to recognize charged current nue events and their
 * backgrounds.  This code is based on JMShower EID and is intended to take advantage of the diagnostics and
 * ability to use different classifiers in TMVA.
 */
#include <TString.h>
#include <vector>

#include "TMVAEIDTraining.h"


#if not defined(__CINT__) || defined(__MAKECINT__)
// needs to be included when makecint runs (ACLIC)
#endif

TMVAEIDTraining::TMVAEIDTraining(TString jobNameIn, TString outPathIn,
        TString myMethodListIn, TString useENuIn) {

    //---------------------------------------------------------------
    // This loads the library
    TMVA::Tools::Instance();
    //    systemPtr = new TUnixSystem();
    fOutPath = outPathIn;
    methodsBooked = false;
    TMVAEIDTraining::fJobName = jobNameIn;
    TMVAEIDTraining::myMethodList = myMethodListIn;
    TMVAEIDTraining::useENu = false;
    if (useENuIn == "useENu") {
        TMVAEIDTraining::useENu = true;
        std::cout << "###Using ENu in training" << std::endl;
    }
    if (myMethodList == "") {
        myMethodList = "TMlpANN";
    }
    TMVAEIDTraining::weightDir = fOutPath + TString("/output/") + fJobName + TString("/weights/");
    TMVAEIDTraining::plotDir = fOutPath + TString("/output/") + fJobName + TString("/plots/");
    TMVAEIDTraining::histDir = fOutPath + TString("/output/") + fJobName + TString("/hist/");

    gSystem->mkdir(weightDir.Data(), true);
    gSystem->mkdir(plotDir.Data(), true);
    gSystem->mkdir(histDir.Data(), true);

    TMVAEIDTraining::outfileName = histDir + TString("TMVA") + TString(".root");

    //
    //  Set TMVA options
    //
    (TMVA::gConfig().GetIONames()).fWeightFileDir = weightDir;

    // to get access to the GUI and all tmva macros
    TString tmva_dir(TString(gRootDir) + "/tmva");
    if (gSystem->Getenv("TMVASYS"))
        tmva_dir = TString(gSystem->Getenv("TMVASYS"));
    std::cout << tmva_dir << std::endl;
    gROOT->SetMacroPath("$ROOTSYS/root/tmva/test/:.:" + tmva_dir + "/test/:" + gROOT->GetMacroPath());
    gROOT->ProcessLine(".L TMVAGui.C");

}

TMVAEIDTraining::TMVAEIDTraining(const TMVAEIDTraining& orig) {
}

TMVAEIDTraining::~TMVAEIDTraining() {

    delete factory;
    delete signalIn;
    delete backgroundIn;
    delete trainingSample;
}

void TMVAEIDTraining::runTraining(TString cSVFileNameIn) {

    cSVFileName = cSVFileNameIn;
    TMVAEIDTraining::setTMVAAlgorithms();
    TMVAEIDTraining::internalTMVASetup();
    TMVAEIDTraining::preTrainingSetup();
    if (!isTransformedInput) {
        TMVAEIDTraining::transformInputTrees(!isTransformedInput);
        if (cSVFileName != NULL && cSVFileName != "") {
            return;
        }
        TMVAEIDTraining::preTrainHistograms();
    }
    TMVAEIDTraining::setupTrainingAndTestTrees();
    TMVAEIDTraining::bookTMVAMethods();
    TMVAEIDTraining::trainTestAndEvaluateTMVAMethods();
    TMVAEIDTraining::saveOutput();
}

void TMVAEIDTraining::setupForReader() {

    TMVAEIDTraining::setTMVAAlgorithms();
}

void TMVAEIDTraining::setSignalFilesUntransformed(std::vector<TString> inputFileNameIn) {
    std::cout << "IN setSignalFilesUntransformed" << std::endl;
    isTransformedInput = false;
    signalIn = new TChain("showerrec/fEvent");
    for (unsigned int ii = 0; ii < inputFileNameIn.size(); ii++) {
        signalIn->Add(inputFileNameIn[ii]);
    }

    //    if (inputFileNameIn == "FULL") {
    //        // temp hack
    //        signalIn->Add("/export/data/local/novadata/nuevt/mc/S12-10-07/FD/posfkcaf_swap_fd_nhc_[1-9].root");
    //        signalIn->Add("/export/data/local/novadata/nuevt/mc/S12-10-07/FD/posfkcaf_swap_fd_nhc_[1-9][0-9].root");
    //        signalIn->Add("/export/data/local/novadata/nuevt/mc/S12-10-07/FD/posfkcaf_swap_fd_nhc_[1-1][0-4][0-9].root");
    //        signalIn->Add("/export/data/local/novadata/nuevt/mc/S12-10-07/FD/posfkcaf_swap_fd_nhc_150.root");
    //        std::cout << "Added FULL set of 150 signal files" << std::endl;
    //    } else {
    //        signalIn->Add(inputFileNameIn);
    //        std::cout << "Added signal file " << inputFileNameIn << std::endl;
    //    }
}

void TMVAEIDTraining::setBackgroundFilesUntransformed(std::vector<TString> inputFileNameIn) {
    std::cout << "IN setBackgroundFilesUntransformed" << std::endl;
    isTransformedInput = false;
    backgroundIn = new TChain("showerrec/fEvent");
    for (unsigned int ii = 0; ii < inputFileNameIn.size(); ii++) {
        backgroundIn->Add(inputFileNameIn[ii]);
    }
    //    if (inputFileNameIn == "FULL") {
    //        //temp hack
    //        backgroundIn->Add("/export/data/local/novadata/nuevt/mc/S12-10-07/FD/posfkcaf_fd_nhc_[1-9].root");
    //        backgroundIn->Add("/export/data/local/novadata/nuevt/mc/S12-10-07/FD/posfkcaf_fd_nhc_[1-9][0-9].root");
    //        backgroundIn->Add("/export/data/local/novadata/nuevt/mc/S12-10-07/FD/posfkcaf_fd_nhc_[1-2][0-9][0-9].root");
    //        backgroundIn->Add("/export/data/local/novadata/nuevt/mc/S12-10-07/FD/posfkcaf_fd_nhc_300.root");
    //        std::cout << "Added FULL set of 300 background files" << std::endl;
    //    } else {
    //        backgroundIn->Add(inputFileNameIn);
    //        std::cout << "Added background file " << inputFileNameIn << std::endl;
    //    }
}

void TMVAEIDTraining::setTMVAAlgorithms() {
    std::cout << "IN setTMVAAlgorithms" << std::endl;

    // Default MVA methods to be trained + tested

    // --- Cut optimisation
    Use["Cuts"] = 0;
    Use["CutsD"] = 0;
    Use["CutsPCA"] = 0;
    Use["CutsGA"] = 0;
    Use["CutsSA"] = 0;
    //
    // --- 1-dimensional likelihood ("naive Bayes estimator")
    Use["Likelihood"] = 0;
    Use["LikelihoodD"] = 0; // the "D" extension indicates decorrelated input variables (see option strings)
    Use["LikelihoodPCA"] = 0; // the "PCA" extension indicates PCA-transformed input variables (see option strings)
    Use["LikelihoodKDE"] = 0;
    Use["LikelihoodMIX"] = 0;
    //
    // --- Mutidimensional likelihood and Nearest-Neighbour methods
    Use["PDERS"] = 0;
    Use["PDERSD"] = 0;
    Use["PDERSPCA"] = 0;
    Use["PDEFoam"] = 0;
    Use["PDEFoamBoost"] = 0; // uses generalised MVA method boosting
    //    if (myMethodList == "KNN") {
    //        Use["KNN"] = 1; // k-nearest neighbour method
    //    } else {
    Use["KNN"] = 0; // k-nearest neighbour method
    //    }
    //
    // --- Linear Discriminant Analysis
    Use["LD"] = 0; // Linear Discriminant identical to Fisher
    Use["Fisher"] = 0;
    Use["FisherG"] = 0;
    Use["BoostedFisher"] = 0; // uses generalised MVA method boosting
    Use["HMatrix"] = 0;
    //
    // --- Function Discriminant analysis
    Use["FDA_GA"] = 0; // minimisation of user-defined function using Genetics Algorithm
    Use["FDA_SA"] = 0;
    Use["FDA_MC"] = 0;
    Use["FDA_MT"] = 0;
    Use["FDA_GAMT"] = 0;
    Use["FDA_MCMT"] = 0;
    //
    // --- Neural Networks (all are feed-forward Multilayer Perceptrons)
    //    if (myMethodList == "MLP") {
    Use["MLP"] = 0; // Recommended ANN
    //    } else {
    //        Use["MLP"] = 0; // Recommended ANN
    //    }
    Use["MLPBFGS"] = 0; // Recommended ANN with optional training method
    Use["MLPBNN"] = 0; // Recommended ANN with BFGS training method and bayesian regulator
    Use["CFMlpANN"] = 0; // Depreciated ANN from ALEPH
    //    if (myMethodList == "TMlpANN") {
    //        Use["TMlpANN"] = 1; // ROOT's own ANN
    //    } else {
    Use["TMlpANN"] = 1; // ROOT's own ANN
    //    }
    //
    // --- Support Vector Machine
    Use["SVM"] = 0;
    //
    // --- Boosted Decision Trees
    //    if (myMethodList == "BDT") {
    //        Use["BDT"] = 1; // uses Adaptive Boost
    //    } else {
    Use["BDT"] = 0; // uses Adaptive Boost
    //    }
    Use["BDTG"] = 0; // uses Gradient Boost
    Use["BDTB"] = 0; // uses Bagging
    Use["BDTD"] = 0; // decorrelation + Adaptive Boost
    Use["BDTF"] = 0; // allow usage of fisher discriminant for node splitting
    //
    // --- Friedman's RuleFit method, ie, an optimised series of cuts ("rules")
    Use["RuleFit"] = 0;

}

void TMVAEIDTraining::internalTMVASetup() {
    std::cout << "IN internalTMVASetup" << std::endl;
    std::cout << std::endl;
    std::cout << "==> Start TMVAClassification" << std::endl;

    // Select methods (don't look at this code - not of interest)
    if (myMethodList != "") {
        for (std::map<std::string, int>::iterator it = Use.begin(); it != Use.end(); it++) it->second = 0;

        std::vector<TString> mlist = TMVA::gTools().SplitString(myMethodList, ',');
        if (mlist.size() > 1) {
            std::cout << "Only one MVA method may be specified at a time";
            //            exit;
        }

        for (UInt_t i = 0; i < mlist.size(); i++) {
            std::string regMethod(mlist[i]);

            if (Use.find(regMethod) == Use.end()) {
                std::cout << "Method \"" << regMethod << "\" not known in TMVA under this name. Choose among the following:" << std::endl;
                for (std::map<std::string, int>::iterator it = Use.begin(); it != Use.end(); it++) std::cout << it->first << " ";
                std::cout << std::endl;
                return;
            }
            Use[regMethod] = 1;
        }
    }

}

void TMVAEIDTraining::preTrainingSetup() {
    std::cout << "IN preTrainingSetup" << std::endl;
    // --- Here the preparation phase begins

    // Create a ROOT output file where TMVA will store ntuples, histograms, etc.
    //  TString outfileName( "TMVA.root" );
    //    TFile*
    outputFile = TFile::Open(TMVAEIDTraining::outfileName, "RECREATE");

    // Create the factory object. Later you can choose the methods
    // whose performance you'd like to investigate. The factory is
    // the only TMVA object you have to interact with
    //
    // The first argument is the base of the name of all the
    // weightfiles in the directory weight/
    //
    // The second argument is the output file for the training results
    // All TMVA output can be suppressed by removing the "!" (not) in
    // front of the "Silent" argument in the option string
    //    TMVA::Factory *
    factory = new TMVA::Factory("TMVAClassification", outputFile,
            "!V:!Silent:Color:DrawProgressBar:Transformations=I;D;P;G,D:AnalysisType=Classification");

    // If you wish to modify default settings
    // (please check "src/Config.h" to see all available global options)
    //    (TMVA::gConfig().GetVariablePlotting()).fTimesRMS = 8.0;
    //    (TMVA::gConfig().GetIONames()).fWeightFileDir = "myWeightDirectory";

    // Define the input variables that shall be used for the MVA training
    // note that you may also use variable expressions, such as: "3*var1/var2*abs(var3)"
    // [all types of expressions that can also be parsed by TTree::Draw( "expression" )]

    factory->AddVariable("egLLL", 'F'); //"spec1 := var1*2",  "Spectator 1", "units", 'F'
    factory->AddVariable("egLLT", 'F');
    factory->AddVariable("emuLLL", 'F');
    factory->AddVariable("emuLLT", 'F');
    factory->AddVariable("epi0LLL", 'F');
    factory->AddVariable("epi0LLT", 'F');
    factory->AddVariable("epLLL", 'F');
    factory->AddVariable("epLLT", 'F');
    factory->AddVariable("enLLL", 'F');
    factory->AddVariable("enLLT", 'F');
    factory->AddVariable("epiLLL", 'F');
    factory->AddVariable("epiLLT", 'F');
    factory->AddVariable("gap", 'F');
    factory->AddVariable("pi0mass", 'F');
    factory->AddVariable("vtxgev", 'F');
    factory->AddVariable("shE", 'F');
    std::cout << "### useENu = " << useENu << std::endl;
    if (useENu) {
        factory->AddVariable("nueRecEnergy", 'F');
    }
    //    factory->AddVariable("costheta", 'F');
    //    factory->AddVariable("length", 'F');



    // You can add so-called "Spectator variables", which are not used in the MVA training,
    // but will appear in the final "TestTree" produced by TMVA. This TestTree will contain the
    // input variables, the response values of all trained MVAs, and the spectator variables
    factory->AddSpectator("sbtype", 'I');
    factory->AddSpectator("specType", 'I');
    //    factory->AddSpectator("spec2 := var1*3", "Spectator 2", "units", 'F');

}

void TMVAEIDTraining::setupTrainingAndTestTrees() {
    std::cout << "IN setupTrainingAndTestTrees" << std::endl;
    // global event weights per tree (see below for setting event-wise weights)
//    Double_t signalWeight = 1.0;
//    Double_t backgroundWeight = 1.0;

    // You can add an arbitrary number of signal or background trees
    TCut signalCut = "sbtype==1"; // how to identify signal events
    TCut backgrCut = "sbtype==0"; // how to identify background events
    factory->SetInputTrees(trainingSample, signalCut, backgrCut);


    // To give different trees for training and testing, do as follows:
    //    factory->AddSignalTree( signalTrainingTree, signalTrainWeight, "Training" );
    //    factory->AddSignalTree( signalTestTree,     signalTestWeight,  "Test" );

    // Use the following code instead of the above two or four lines to add signal and background
    // training and test events "by hand"
    // NOTE that in this case one should not give expressions (such as "var1+var2") in the input
    //      variable definition, but simply compute the expression before adding the event
    //
    //     // --- begin ----------------------------------------------------------
    //     std::vector<Double_t> vars( 4 ); // vector has size of number of input variables
    //     Float_t  treevars[4], weight;
    //
    //     // Signal
    //     for (UInt_t ivar=0; ivar<4; ivar++) signal->SetBranchAddress( Form( "var%i", ivar+1 ), &(treevars[ivar]) );
    //     for (UInt_t i=0; i<signal->GetEntries(); i++) {
    //        signal->GetEntry(i);
    //        for (UInt_t ivar=0; ivar<4; ivar++) vars[ivar] = treevars[ivar];
    //        // add training and test events; here: first half is training, second is testing
    //        // note that the weight can also be event-wise
    //        if (i < signal->GetEntries()/2.0) factory->AddSignalTrainingEvent( vars, signalWeight );
    //        else                              factory->AddSignalTestEvent    ( vars, signalWeight );
    //     }
    //
    //     // Background (has event weights)
    //     background->SetBranchAddress( "weight", &weight );
    //     for (UInt_t ivar=0; ivar<4; ivar++) background->SetBranchAddress( Form( "var%i", ivar+1 ), &(treevars[ivar]) );
    //     for (UInt_t i=0; i<background->GetEntries(); i++) {
    //        background->GetEntry(i);
    //        for (UInt_t ivar=0; ivar<4; ivar++) vars[ivar] = treevars[ivar];
    //        // add training and test events; here: first half is training, second is testing
    //        // note that the weight can also be event-wise
    //        if (i < background->GetEntries()/2) factory->AddBackgroundTrainingEvent( vars, backgroundWeight*weight );
    //        else                                factory->AddBackgroundTestEvent    ( vars, backgroundWeight*weight );
    //     }
    // --- end ------------------------------------------------------------
    //
    // --- end of tree registration

    // Set individual event weights (the variables must exist in the original TTree)
    //    for signal    : factory->SetSignalWeightExpression    ("weight1*weight2");
    //    for background: factory->SetBackgroundWeightExpression("weight1*weight2");
    //   factory->SetBackgroundWeightExpression("weight");

    // Apply additional cuts on the signal and background samples (can be different)
    TCut mycuts = ""; // for example: TCut mycuts = "abs(var1)<0.5 && abs(var2-0.5)<1";
    TCut mycutb = ""; // for example: TCut mycutb = "abs(var1)<0.5";




    // Tell the factory how to use the training and testing events
    //
    // If no numbers of events are given, half of the events in the tree are used
    // for training, and the other half for testing:
    //    factory->PrepareTrainingAndTestTree( mycut, "SplitMode=random:!V" );
    // To also specify the number of testing events, use:
    //    factory->PrepareTrainingAndTestTree( mycut,
    //                                         "NSigTrain=3000:NBkgTrain=3000:NSigTest=3000:NBkgTest=3000:SplitMode=Random:!V" );
    //    factory->PrepareTrainingAndTestTree(mycuts, mycutb,
    //            "nTrain_Signal=0:nTrain_Background=0:SplitMode=Random:NormMode=NumEvents:!V");
    TString optionsString = TString::Format(
            "nTrain_Signal=%d:nTrain_Background=%d:SplitMode=Alternate:NormMode=NumEvents:!V", 0, 0);

    factory->PrepareTrainingAndTestTree(mycuts, mycutb,
            //            "nTrain_Signal=0:nTrain_Background=0:nTest_Signal=0:nTest_Background=0:SplitMode=Random:NormMode=NumEvents:!V");
            //            "nTrain_Signal=0:nTrain_Background=0:SplitMode=Alternate:NormMode=NumEvents:!V");
            optionsString.Data());
    //            "nTrain_Signal=0:nTrain_Background=0:SplitMode=Alternate:NormMode=NumEvents:!V");
}

void TMVAEIDTraining::bookTMVAMethods() {
    std::cout << "IN bookTMVAMethods" << std::endl;
    // ---- Book MVA methods
    //
    // Please lookup the various method configuration options in the corresponding cxx files, eg:
    // src/MethoCuts.cxx, etc, or here: http://tmva.sourceforge.net/optionRef.html
    // it is possible to preset ranges in the option string in which the cut optimisation should be done:
    // "...:CutRangeMin[2]=-1:CutRangeMax[2]=1"...", where [2] is the third input variable

    // Cut optimisation
    if (Use["Cuts"])
        factory->BookMethod(TMVA::Types::kCuts, "Cuts",
            "!H:!V:FitMethod=MC:EffSel:SampleSize=200000:VarProp=FSmart");

    if (Use["CutsD"])
        factory->BookMethod(TMVA::Types::kCuts, "CutsD",
            "!H:!V:FitMethod=MC:EffSel:SampleSize=200000:VarProp=FSmart:VarTransform=Decorrelate");

    if (Use["CutsPCA"])
        factory->BookMethod(TMVA::Types::kCuts, "CutsPCA",
            "!H:!V:FitMethod=MC:EffSel:SampleSize=200000:VarProp=FSmart:VarTransform=PCA");

    if (Use["CutsGA"])
        factory->BookMethod(TMVA::Types::kCuts, "CutsGA",
            "H:!V:FitMethod=GA:CutRangeMin[0]=-10:CutRangeMax[0]=10:VarProp[1]=FMax:EffSel:Steps=30:Cycles=3:PopSize=400:SC_steps=10:SC_rate=5:SC_factor=0.95");

    if (Use["CutsSA"])
        factory->BookMethod(TMVA::Types::kCuts, "CutsSA",
            "!H:!V:FitMethod=SA:EffSel:MaxCalls=150000:KernelTemp=IncAdaptive:InitialTemp=1e+6:MinTemp=1e-6:Eps=1e-10:UseDefaultScale");

    // Likelihood ("naive Bayes estimator")
    if (Use["Likelihood"])
        factory->BookMethod(TMVA::Types::kLikelihood, "Likelihood",
            "H:!V:TransformOutput:PDFInterpol=Spline2:NSmoothSig[0]=20:NSmoothBkg[0]=20:NSmoothBkg[1]=10:NSmooth=1:NAvEvtPerBin=50");

    // Decorrelated likelihood
    if (Use["LikelihoodD"])
        factory->BookMethod(TMVA::Types::kLikelihood, "LikelihoodD",
            "!H:!V:TransformOutput:PDFInterpol=Spline2:NSmoothSig[0]=20:NSmoothBkg[0]=20:NSmooth=5:NAvEvtPerBin=50:VarTransform=Decorrelate");

    // PCA-transformed likelihood
    if (Use["LikelihoodPCA"])
        factory->BookMethod(TMVA::Types::kLikelihood, "LikelihoodPCA",
            "!H:!V:!TransformOutput:PDFInterpol=Spline2:NSmoothSig[0]=20:NSmoothBkg[0]=20:NSmooth=5:NAvEvtPerBin=50:VarTransform=PCA");

    // Use a kernel density estimator to approximate the PDFs
    if (Use["LikelihoodKDE"])
        factory->BookMethod(TMVA::Types::kLikelihood, "LikelihoodKDE",
            "!H:!V:!TransformOutput:PDFInterpol=KDE:KDEtype=Gauss:KDEiter=Adaptive:KDEFineFactor=0.3:KDEborder=None:NAvEvtPerBin=50");

    // Use a variable-dependent mix of splines and kernel density estimator
    if (Use["LikelihoodMIX"])
        factory->BookMethod(TMVA::Types::kLikelihood, "LikelihoodMIX",
            "!H:!V:!TransformOutput:PDFInterpolSig[0]=KDE:PDFInterpolBkg[0]=KDE:PDFInterpolSig[1]=KDE:PDFInterpolBkg[1]=KDE:PDFInterpolSig[2]=Spline2:PDFInterpolBkg[2]=Spline2:PDFInterpolSig[3]=Spline2:PDFInterpolBkg[3]=Spline2:KDEtype=Gauss:KDEiter=Nonadaptive:KDEborder=None:NAvEvtPerBin=50");

    // Test the multi-dimensional probability density estimator
    // here are the options strings for the MinMax and RMS methods, respectively:
    //      "!H:!V:VolumeRangeMode=MinMax:DeltaFrac=0.2:KernelEstimator=Gauss:GaussSigma=0.3" );
    //      "!H:!V:VolumeRangeMode=RMS:DeltaFrac=3:KernelEstimator=Gauss:GaussSigma=0.3" );
    if (Use["PDERS"])
        factory->BookMethod(TMVA::Types::kPDERS, "PDERS",
            "!H:!V:NormTree=T:VolumeRangeMode=Adaptive:KernelEstimator=Gauss:GaussSigma=0.3:NEventsMin=400:NEventsMax=600");

    if (Use["PDERSD"])
        factory->BookMethod(TMVA::Types::kPDERS, "PDERSD",
            "!H:!V:VolumeRangeMode=Adaptive:KernelEstimator=Gauss:GaussSigma=0.3:NEventsMin=400:NEventsMax=600:VarTransform=Decorrelate");

    if (Use["PDERSPCA"])
        factory->BookMethod(TMVA::Types::kPDERS, "PDERSPCA",
            "!H:!V:VolumeRangeMode=Adaptive:KernelEstimator=Gauss:GaussSigma=0.3:NEventsMin=400:NEventsMax=600:VarTransform=PCA");

    // Multi-dimensional likelihood estimator using self-adapting phase-space binning
    if (Use["PDEFoam"])
        factory->BookMethod(TMVA::Types::kPDEFoam, "PDEFoam",
            "!H:!V:SigBgSeparate=F:TailCut=0.001:VolFrac=0.0666:nActiveCells=500:nSampl=2000:nBin=5:Nmin=100:Kernel=None:Compress=T");

    if (Use["PDEFoamBoost"])
        factory->BookMethod(TMVA::Types::kPDEFoam, "PDEFoamBoost",
            "!H:!V:Boost_Num=30:Boost_Transform=linear:SigBgSeparate=F:MaxDepth=4:UseYesNoCell=T:DTLogic=MisClassificationError:FillFoamWithOrigWeights=F:TailCut=0:nActiveCells=500:nBin=20:Nmin=400:Kernel=None:Compress=T");

    // K-Nearest Neighbour classifier (KNN)
    if (Use["KNN"])
        factory->BookMethod(TMVA::Types::kKNN, "KNN",
            "H:CreateMVAPdfs:nkNN=20:ScaleFrac=0.8:SigmaFact=1.0:Kernel=Gaus:UseKernel=F:UseWeight=T:!Trim");

    // H-Matrix (chi2-squared) method
    if (Use["HMatrix"])
        factory->BookMethod(TMVA::Types::kHMatrix, "HMatrix", "!H:!V:VarTransform=None");

    // Linear discriminant (same as Fisher discriminant)
    if (Use["LD"])
        factory->BookMethod(TMVA::Types::kLD, "LD", "H:!V:VarTransform=None:CreateMVAPdfs:PDFInterpolMVAPdf=Spline2:NbinsMVAPdf=50:NsmoothMVAPdf=10");

    // Fisher discriminant (same as LD)
    if (Use["Fisher"])
        factory->BookMethod(TMVA::Types::kFisher, "Fisher", "H:!V:Fisher:VarTransform=None:CreateMVAPdfs:PDFInterpolMVAPdf=Spline2:NbinsMVAPdf=50:NsmoothMVAPdf=10");

    // Fisher with Gauss-transformed input variables
    if (Use["FisherG"])
        factory->BookMethod(TMVA::Types::kFisher, "FisherG", "H:!V:VarTransform=Gauss");

    // Composite classifier: ensemble (tree) of boosted Fisher classifiers
    if (Use["BoostedFisher"])
        factory->BookMethod(TMVA::Types::kFisher, "BoostedFisher",
            "H:!V:Boost_Num=20:Boost_Transform=log:Boost_Type=AdaBoost:Boost_AdaBoostBeta=0.2:!Boost_DetailedMonitoring");

    // Function discrimination analysis (FDA) -- test of various fitters - the recommended one is Minuit (or GA or SA)
    if (Use["FDA_MC"])
        factory->BookMethod(TMVA::Types::kFDA, "FDA_MC",
            "H:!V:Formula=(0)+(1)*x0+(2)*x1+(3)*x2+(4)*x3:ParRanges=(-1,1);(-10,10);(-10,10);(-10,10);(-10,10):FitMethod=MC:SampleSize=100000:Sigma=0.1");

    if (Use["FDA_GA"]) // can also use Simulated Annealing (SA) algorithm (see Cuts_SA options])
        factory->BookMethod(TMVA::Types::kFDA, "FDA_GA",
            "H:!V:Formula=(0)+(1)*x0+(2)*x1+(3)*x2+(4)*x3:ParRanges=(-1,1);(-10,10);(-10,10);(-10,10);(-10,10):FitMethod=GA:PopSize=300:Cycles=3:Steps=20:Trim=True:SaveBestGen=1");

    if (Use["FDA_SA"]) // can also use Simulated Annealing (SA) algorithm (see Cuts_SA options])
        factory->BookMethod(TMVA::Types::kFDA, "FDA_SA",
            "H:!V:Formula=(0)+(1)*x0+(2)*x1+(3)*x2+(4)*x3:ParRanges=(-1,1);(-10,10);(-10,10);(-10,10);(-10,10):FitMethod=SA:MaxCalls=15000:KernelTemp=IncAdaptive:InitialTemp=1e+6:MinTemp=1e-6:Eps=1e-10:UseDefaultScale");

    if (Use["FDA_MT"])
        factory->BookMethod(TMVA::Types::kFDA, "FDA_MT",
            "H:!V:Formula=(0)+(1)*x0+(2)*x1+(3)*x2+(4)*x3:ParRanges=(-1,1);(-10,10);(-10,10);(-10,10);(-10,10):FitMethod=MINUIT:ErrorLevel=1:PrintLevel=-1:FitStrategy=2:UseImprove:UseMinos:SetBatch");

    if (Use["FDA_GAMT"])
        factory->BookMethod(TMVA::Types::kFDA, "FDA_GAMT",
            "H:!V:Formula=(0)+(1)*x0+(2)*x1+(3)*x2+(4)*x3:ParRanges=(-1,1);(-10,10);(-10,10);(-10,10);(-10,10):FitMethod=GA:Converger=MINUIT:ErrorLevel=1:PrintLevel=-1:FitStrategy=0:!UseImprove:!UseMinos:SetBatch:Cycles=1:PopSize=5:Steps=5:Trim");

    if (Use["FDA_MCMT"])
        factory->BookMethod(TMVA::Types::kFDA, "FDA_MCMT",
            "H:!V:Formula=(0)+(1)*x0+(2)*x1+(3)*x2+(4)*x3:ParRanges=(-1,1);(-10,10);(-10,10);(-10,10);(-10,10):FitMethod=MC:Converger=MINUIT:ErrorLevel=1:PrintLevel=-1:FitStrategy=0:!UseImprove:!UseMinos:SetBatch:SampleSize=20");

    // TMVA ANN: MLP (recommended ANN) -- all ANNs in TMVA are Multilayer Perceptrons
    if (Use["MLP"])
        //        factory->BookMethod(TMVA::Types::kMLP, "MLP",
        //                "H:!V:CreateMVAPdfs:NeuronType=tanh:VarTransform=N:NCycles=600:HiddenLayers=N+5:TestRate=5:!UseRegulator");
        factory->BookMethod(TMVA::Types::kMLP, "MLP",
            "H:!V:CreateMVAPdfs:NeuronType=tanh:TrainingMethod=BFGS:VarTransform=N_AllClasses:NCycles=600:HiddenLayers=22,12,6:TestRate=10:!UseRegulator");

    if (Use["MLPBFGS"])
        factory->BookMethod(TMVA::Types::kMLP, "MLPBFGS", "H:!V:NeuronType=tanh:VarTransform=N:NCycles=600:HiddenLayers=N+5:TestRate=5:TrainingMethod=BFGS:!UseRegulator");

    if (Use["MLPBNN"])
        factory->BookMethod(TMVA::Types::kMLP, "MLPBNN", "H:!V:NeuronType=tanh:VarTransform=N:NCycles=600:HiddenLayers=N+5:TestRate=5:TrainingMethod=BFGS:UseRegulator"); // BFGS training with bayesian regulators

    // CF(Clermont-Ferrand)ANN
    if (Use["CFMlpANN"])
        factory->BookMethod(TMVA::Types::kCFMlpANN, "CFMlpANN", "!H:!V:NCycles=2000:HiddenLayers=N+1,N"); // n_cycles:#nodes:#nodes:...

    // Tmlp(Root)ANN
    if (Use["TMlpANN"])
        //        factory->BookMethod(TMVA::Types::kTMlpANN, "TMlpANN", "H:!V:NCycles=200:HiddenLayers=N+1,N:LearningMethod=BFGS:ValidationFraction=0.3"); // n_cycles:#nodes:#nodes:...
        //        factory->BookMethod(TMVA::Types::kTMlpANN, "TMlpANN",
        //                "H:!V:NCycles=600:HiddenLayers=N+1,N:LearningMethod=BFGS:ValidationFraction=0.5"); // n_cycles:#nodes:#nodes:...
        //        factory->BookMethod(TMVA::Types::kTMlpANN, "TMlpANN",
        //                "H:!V:NCycles=600:HiddenLayers=22,12,6:LearningMethod=BFGS:ValidationFraction=0.5:VarTransform=N_AllClasses"); // n_cycles:#nodes:#nodes:...
        factory->BookMethod(TMVA::Types::kTMlpANN, "TMlpANN",
            "H:V:NCycles=600:HiddenLayers=22,12,6:LearningMethod=BFGS:ValidationFraction=0.0:VarTransform=N_AllClasses"); // n_cycles:#nodes:#nodes:...
    //        factory->BookMethod(TMVA::Types::kTMlpANN, "TMlpANN",
    //                "H:V:NCycles=600:HiddenLayers=22,12,6:LearningMethod=BFGS:ValidationFraction=0.0:VarTransform=N_AllClasses"); // n_cycles:#nodes:#nodes:...

    // Support Vector Machine
    if (Use["SVM"])
        factory->BookMethod(TMVA::Types::kSVM, "SVM", "Gamma=0.25:Tol=0.001:VarTransform=Norm");

    // Boosted Decision Trees
    if (Use["BDTG"]) // Gradient Boost
        factory->BookMethod(TMVA::Types::kBDT, "BDTG",
            "!H:!V:NTrees=1000:BoostType=Grad:Shrinkage=0.10:UseBaggedGrad:GradBaggingFraction=0.5:nCuts=20:NNodesMax=5");

    if (Use["BDT"]) // Adaptive Boost
        factory->BookMethod(TMVA::Types::kBDT, "BDT",
            "!H:!V:NTrees=850:nEventsMin=150:MaxDepth=3:BoostType=AdaBoost:AdaBoostBeta=0.5:SeparationType=GiniIndex:nCuts=20:PruneMethod=NoPruning");


    if (Use["BDTB"]) // Bagging
        factory->BookMethod(TMVA::Types::kBDT, "BDTB",
            "!H:!V:NTrees=400:BoostType=Bagging:SeparationType=GiniIndex:nCuts=20:PruneMethod=NoPruning");

    if (Use["BDTD"]) // Decorrelation + Adaptive Boost
        factory->BookMethod(TMVA::Types::kBDT, "BDTD",
            "!H:!V:NTrees=400:nEventsMin=400:MaxDepth=3:BoostType=AdaBoost:SeparationType=GiniIndex:nCuts=20:PruneMethod=NoPruning:VarTransform=Decorrelate");

    if (Use["BDTF"]) // Allow Using Fisher discriminant in node splitting for (strong) linearly correlated variables
        factory->BookMethod(TMVA::Types::kBDT, "BDTMitFisher",
            "!H:!V:NTrees=50:nEventsMin=150:UseFisherCuts:MaxDepth=3:BoostType=AdaBoost:AdaBoostBeta=0.5:SeparationType=GiniIndex:nCuts=20:PruneMethod=NoPruning");

    // RuleFit -- TMVA implementation of Friedman's method
    if (Use["RuleFit"])
        factory->BookMethod(TMVA::Types::kRuleFit, "RuleFit",
            "H:!V:RuleFitModule=RFTMVA:Model=ModRuleLinear:MinImp=0.001:RuleMinDist=0.001:NTrees=20:fEventsMin=0.01:fEventsMax=0.5:GDTau=-1.0:GDTauPrec=0.01:GDStep=0.01:GDNSteps=10000:GDErrScale=1.02");
    methodsBooked = true;
}

void TMVAEIDTraining::trainTestAndEvaluateTMVAMethods() {
    std::cout << "IN trainTestAndEvaluateTMVAMethods" << std::endl;
    // For an example of the category classifier usage, see: TMVAClassificationCategory

    // --------------------------------------------------------------------------------------------------

    // ---- Now you can optimize the setting (configuration) of the MVAs using the set of training events

    // factory->OptimizeAllMethods("SigEffAt001","Scan");
    // factory->OptimizeAllMethods("ROCIntegral","GA");

    // --------------------------------------------------------------------------------------------------

    // ---- Now you can tell the factory to train, test, and evaluate the MVAs

    std::cout << "PRE-TRAIN: trainingSample->GetEntries() = " << trainingSample->GetEntries() << std::endl;

    // Train MVAs using the set of training events
    factory->TrainAllMethods();

    // ---- Evaluate all MVAs using the set of test events
    factory->TestAllMethods();

    // ----- Evaluate and compare performance of all configured MVAs
    factory->EvaluateAllMethods();
}

void TMVAEIDTraining::saveOutput() {

    std::cout << "IN saveOutput" << std::endl;

    // Save the output
    outputFile->Close();

    std::cout << "==> Wrote root file: " << outputFile->GetName() << std::endl;
    std::cout << "==> TMVAClassification is done!" << std::endl;

}

void TMVAEIDTraining::preTrainHistograms() {
    std::cout << "IN preTrainHistograms" << std::endl;
    TCut MCNUECCQE = "evtTrueNuPdg==12&&(evtRwRand[0]>0&&evtRwRand[0]<evtRwProb[0])";
    TCut MCBKG = "(evtRwRand[2]>0&&evtRwRand[2]<evtRwProb[2])";
    htruenue = new TH1F("htruenue", "htruenue", 100, 0, 5);
    htruenupi0 = new TH1F("htruenupi0", "htruenupi0", 100, 0, 5);
    signalIn->Draw("evtTrueNuEnergy>>htruenue", MCNUECCQE);
    backgroundIn->Draw("evtTrueNuEnergy>>htruenupi0", MCBKG);

//    double nsig = htruenue->GetEntries();
//    double nbkg = htruenupi0->GetEntries();

    TCanvas *c0 = new TCanvas("c0", "c0", 200, 10, 1200, 600);
    c0->Divide(2, 1);
    c0->cd(1);
    htruenue->Draw();
    c0->cd(2);
    htruenupi0->Draw();
    c0->Update();
    c0->SaveAs(plotDir + "JMCanvas0.png");
    c0->SaveAs(plotDir + "JMCanvas0.eps");
}

void TMVAEIDTraining::printTrainingVars() {
    std::cout << "--------------------------------------------" << std::endl;
    std::cout << "egLLL/T" << egLLL << "," << egLLT << std::endl;
    std::cout << "emuLLL/T" << emuLLL << "," << emuLLT << std::endl;
    std::cout << "epi0LLL/T" << epi0LLL << "," << epi0LLT << std::endl;
    std::cout << "epLLL/T" << epLLL << "," << epLLT << std::endl;
    std::cout << "enLLL/T" << enLLL << "," << enLLT << std::endl;
    std::cout << "epiLLL/T" << epiLLL << "," << epiLLT << std::endl;
    std::cout << "gap" << gap << std::endl;
    std::cout << "pi0mass" << pi0mass << std::endl;
    std::cout << "vtxgev" << vtxgev << std::endl;
    std::cout << "shE" << shE << std::endl;
    if (useENu) {
        std::cout << "nueRecEnergy" << nueRecEnergy << std::endl;
    }
    std::cout << "costheta" << costheta << std::endl;
    std::cout << "length" << length << std::endl;
    std::cout << "type" << type << std::endl;
}

void TMVAEIDTraining::dumpTrainingSampleToCSV(CSVOperation step, TString cSVFileName) {

    if (step == CSVOperation::INIT) {
        std::cout << "Dumping training sample to CSV File:" << cSVFileName << std::endl;
        lineBufferSize = 0;
        outfile.open((cSVFileName.Data()));
        // Header
        outfile << "egLLL\t" << "egLLT\t" << "emuLLL\t" << "emuLLT\t"
                << "epi0LLL\t" << "epi0LLT\t" << "epLLL\t" << "epLLT\t"
                << "enLLL\t" << "enLLT\t" << "epiLLL\t" << "epiLLT\t"
                << "gap\t" << "pi0mass\t" << "vtxgev\t" << "shE\t"
                << "nueRecEnergy\t" << "costheta\t" << "length\t" << "type\t"
                << "specType\t" << "reWeight" << "\n";
    } else if (step == CSVOperation::FILL) {

        std::string sigback;
        if (type == 0) {
            sigback = "Background";
        } else {
            sigback = "Signal";
        }
        outfile << egLLL << "\t" << egLLT << "\t" << emuLLL << "\t" << emuLLT << "\t"
                << epi0LLL << "\t" << epi0LLT << "\t" << epLLL << "\t" << epLLT << "\t"
                << enLLL << "\t" << enLLT << "\t" << epiLLL << "\t" << epiLLT << "\t"
                << gap << "\t" << pi0mass << "\t" << vtxgev << "\t" << shE << "\t"
                << nueRecEnergy << "\t" << costheta << "\t" << length << "\t" << sigback << "\t"
                << specType << "\t" << reWeight << "\n";


    } else if (step == CSVOperation::CLOSE) {
        outfile.close();
    }
}

/**
 *
 * Transform input trees to internal trees used for training.
 *
 * @param fillTrainingSample
 */
void TMVAEIDTraining::transformInputTrees(bool fillTrainingSample) {
    std::cout << "IN transformInputTrees" << std::endl;

//    double Elow = 1.0;
//    double Ehigh = 3.0;
    double rsig = 1.0;
    double rbg = 0.9;

    TMVAEIDTraining::setTreeBranchAddresses(signalIn);
    TMVAEIDTraining::setTreeBranchAddresses(backgroundIn);

    TMVAEIDTraining::dumpTrainingSampleToCSV(CSVOperation::INIT, cSVFileName);

    std::cout << "fillTrainingSample = " << fillTrainingSample << std::endl;
    if (fillTrainingSample) {
        trainingSample = new TTree("TrainingSample", "Filtered Monte Carlo Events");
        trainingSample->Branch("egLLL", &egLLL, "egLLL/F");
        trainingSample->Branch("egLLT", &egLLT, "egLLT/F");
        trainingSample->Branch("emuLLL", &emuLLL, "emuLLL/F");
        trainingSample->Branch("emuLLT", &emuLLT, "emuLLT/F");
        trainingSample->Branch("epi0LLL", &epi0LLL, "epi0LLL/F");
        trainingSample->Branch("epi0LLT", &epi0LLT, "epi0LLT/F");
        trainingSample->Branch("epLLL", &epLLL, "epLLL/F");
        trainingSample->Branch("epLLT", &epLLT, "epLLT/F");
        trainingSample->Branch("enLLL", &enLLL, "enLLL/F");
        trainingSample->Branch("enLLT", &enLLT, "enLLT/F");
        trainingSample->Branch("epiLLL", &epiLLL, "epiLLL/F");
        trainingSample->Branch("epiLLT", &epiLLT, "epiLLT/F");
        trainingSample->Branch("gap", &gap, "gap/F");
        trainingSample->Branch("pi0mass", &pi0mass, "pi0mass/F");
        trainingSample->Branch("vtxgev", &vtxgev, "vtxgev/F");
        trainingSample->Branch("shE", &shE, "shE/F");
        trainingSample->Branch("nueRecEnergy", &nueRecEnergy, "nueRecEnergy/F");
        trainingSample->Branch("costheta", &costheta, "costheta/F");
        trainingSample->Branch("length", &length, "length/F");
        trainingSample->Branch("sbtype", &type, "sbtype/I");
        trainingSample->Branch("specType", &specType, "specType/C");

    }



    numBackgroundNuMuOrNuE0 = 0, numBackgroundPassingFirstShCuts = 0;
    numBackgroundPassingNotObviousMuonsShCuts2 = 0, nb3 = 0;
    numBackgroundNueCC0 = 0, numBackgroundPassingFirstShCutsNueCC1 = 0;
    numBackgroundPassingNotObviousMuonsShCutsNueCC2 = 0, nbNueCC2_a = 0, nbNueCC3 = 0;
    numBackgroundNumuCC0 = 0, numBackgroundPassingFirstShCutsNumuCC1 = 0;
    numBackgroundPassingNotObviousMuonsShCutsNumuCC2 = 0, nbNumuCC3 = 0;
    numBackgroundNC0 = 0, numBackgroundPassingFirstShCutsNC1 = 0;
    numBackgroundPassingNotObviousMuonsShCutsNC2 = 0, nbNC3 = 0;

    numSignalCorrectProcess = 0, numSignalCorrectShower = 0, numSignalCorrectShowerNotObviousMuon = 0, ns3 = 0;
    normSig = 0, normBkg = 0, totBkg = 0;
    //
    //  Signal
    //
    type = 1;
    std::cout << "signalIn->GetEntries() = " << signalIn->GetEntries() << std::endl;
    for (int i = 0; i < signalIn->GetEntries(); i++) {

        signalIn->GetEntry(i);
        //
        //  If not QE interaction AND NuE AND not selected by reweight for oscillation then skip event
        //
        if (!(evtTrueNuCCNC == kQE && evtTrueNuPdg == kIDNuE &&
                (evtRwRand[0] > 0 && evtRwRand[0] < evtRwProb[0]))) continue;
        if (i % 2 == 1) numSignalCorrectProcess++;
        //
        //  If leading shower is not and electron or positron then skip event
        //
        if (fabs(evtSh1TruePdg) != kIDElectron) continue;
        //
        //  If the angle between the leading shower and the true Nue is greater than 30 degrees,
        //  (ie., shower not matched to true MC NuE in MC) then skip
        //
        if (evtSh1TrueDang > 30) continue;
        //
        //  If leading shower energy not positive, then skip
        //
        if (!(evtSh1Energy > 0.0))continue;

        if (i % 2 == 1) numSignalCorrectShower++;
        //
        //  Cut out shower candidates that are obvious muons
        //
        if (evtSh1DedxLLL[0] - evtSh1DedxLLL[2]<-0.2)continue;
        if (i % 2 == 1) numSignalCorrectShowerNotObviousMuon++;

        //
        //  Cut events to get desired proportion of signal
        //
        if ((double) i > (double) signalIn->GetEntries() * rsig)continue;

        if (fabs(evtSh1TruePdg) != 11) {
            specType = "BeamNuE\0";
            if (evtRwRand[0] < evtRwProb[0]) {
                reWeight = 1.0;
            } else {
                reWeight = 0.0;
            }
        } else {
            specType = "Signal\0";
            if (evtRwRand[0] < evtRwProb[0]) {
                reWeight = 1.0;
            } else {
                reWeight = 0.0;
            }
        }
        //       std::cout << "SIG specType = " << specType << std::endl;

        TMVAEIDTraining::dumpTrainingSampleToCSV(CSVOperation::FILL, cSVFileName);
        TMVAEIDTraining::fillTrainingVariables();
        if (fillTrainingSample) trainingSample->Fill();
        //        printTrainingVars();

        normSig++;
    }
    std::cout << "normSig = " << normSig << std::endl;
    //
    //   Background
    //
    type = 0;
    std::cout << "backgroundIn->GetEntries() = " << backgroundIn->GetEntries() << std::endl;
    for (int i = 0; i < backgroundIn->GetEntries(); i++) {

        
        
//        if (iEvt % 2 == 0)continue;
//        backgroundIn->GetEntry(iEvt);
//        if (!((fabs(evtTrueNuPdg) == 14 && evtRwRand[2] > 0 && evtRwRand[2] < evtRwProb[2])
//                || (fabs(evtTrueNuPdg) == 12 && evtRwRand[1] > 0 && evtRwRand[1] < evtRwProb[1])))continue;
//        if (!(evtSh1NPlane>-15 && evtSh1Energy > 0.0))continue;
//        if (evtSh1DedxLLL[0] - evtSh1DedxLLL[2]<-0.2)continue;
//        if (!((evtSh1Energy + 0.282525 + 1.0766 * (evtEtot - evtSh1Energy)) / 1.015 > Elow
//                && (evtSh1Energy + 0.282525 + 1.0766 * (evtEtot - evtSh1Energy)) / 1.015 < Ehigh))continue;        
        
        
        
        backgroundIn->GetEntry(i);
        //
        //  If not selected for NuMu OR NuE oscillation reweight then skip
        //
        if (!((fabs(evtTrueNuPdg) == kIDNuMu && evtRwRand[2] > 0 && evtRwRand[2] < evtRwProb[2])
                || (fabs(evtTrueNuPdg) == kIDNuE && evtRwRand[1] > 0 && evtRwRand[1] < evtRwProb[1])))continue;
        if (i % 2 == 1) numBackgroundNuMuOrNuE0++;
        //
        //  If not NuMu or NuE event
        //
        if (!((fabs(evtTrueNuPdg) == kIDNuMu) || (fabs(evtTrueNuPdg) == kIDNuE)))continue;
        if (i % 2 == 1 && evtTrueNuCCNC == 0 && fabs(evtTrueNuPdg) == kIDNuMu
                && evtRwRand[2] > 0 && evtRwRand[2] < evtRwProb[2]) numBackgroundNumuCC0++;
        if (i % 2 == 1 && evtTrueNuCCNC == 1 && ((fabs(evtTrueNuPdg) == kIDNuMu)
                || (fabs(evtTrueNuPdg) == kIDNuE && evtRwRand[1] > 0 && evtRwRand[1] < evtRwProb[1]))) numBackgroundNC0++;
        if (i % 2 == 1 && evtTrueNuCCNC == 0 && fabs(evtTrueNuPdg) == kIDNuE
                && evtRwRand[1] > 0 && evtRwRand[1] < evtRwProb[1]) numBackgroundNueCC0++;
        //
        //  Leading shower energy > 0, number of planes > -15
        //
        if (!(evtSh1NPlane>-15 && evtSh1Energy > 0.0))continue;
        if (i % 2 == 1) numBackgroundPassingFirstShCuts++;

        if (i % 2 == 1 && evtTrueNuCCNC == 0 && fabs(evtTrueNuPdg) == kIDNuMu
                && evtRwRand[2] > 0 && evtRwRand[2] < evtRwProb[2]) numBackgroundPassingFirstShCutsNumuCC1++;
        if (i % 2 == 1 && evtTrueNuCCNC == 1 && ((fabs(evtTrueNuPdg) == kIDNuMu
                && evtRwRand[2] > 0 && evtRwRand[2] < evtRwProb[2])
                || (fabs(evtTrueNuPdg) == kIDNuE && evtRwRand[1] > 0 && evtRwRand[1] < evtRwProb[1])))
            numBackgroundPassingFirstShCutsNC1++;
        if (i % 2 == 1 && evtTrueNuCCNC == 0 && fabs(evtTrueNuPdg) == kIDNuE && evtRwRand[1] > 0
                && evtRwRand[1] < evtRwProb[1]) numBackgroundPassingFirstShCutsNueCC1++;
        //
        //  Cut out shower candidates that are obvious muons
        //
        if (evtSh1DedxLLL[0] - evtSh1DedxLLL[2]<-0.2) continue;

        if (i % 2 == 1) numBackgroundPassingNotObviousMuonsShCuts2++;
        if (i % 2 == 1 && evtTrueNuCCNC == 0 && fabs(evtTrueNuPdg) == kIDNuMu && evtRwRand[2] > 0
                && evtRwRand[2] < evtRwProb[2]) numBackgroundPassingNotObviousMuonsShCutsNumuCC2++;
        if (i % 2 == 1 && evtTrueNuCCNC == 1 && ((fabs(evtTrueNuPdg) == kIDNuMu && evtRwRand[2] > 0
                && evtRwRand[2] < evtRwProb[2]) || (fabs(evtTrueNuPdg) == kIDNuE && evtRwRand[1] > 0
                && evtRwRand[1] < evtRwProb[1]))) numBackgroundPassingNotObviousMuonsShCutsNC2++;
        if (i % 2 == 1 && evtTrueNuCCNC == 0 && fabs(evtTrueNuPdg) == kIDNuE && evtRwRand[1] > 0
                && evtRwRand[1] < evtRwProb[1]) numBackgroundPassingNotObviousMuonsShCutsNueCC2++;
        //        //
        //        //  Skip if quasi-elastic NuE
        //        //
        //        if (evtTrueNuCCNC == kQE && fabs(evtTrueNuPdg) == kIDNuE) continue;
        //        //      nb3++;

        //
        //  Cut events to get desired proportion of background
        //
        if ((double) i > (double) backgroundIn->GetEntries() * rbg)continue;


        //  For CSV file, write out CCQE NuE  (beam Nue) before skipping over
        if (evtTrueNuCCNC == kQE && fabs(evtTrueNuPdg) == kIDNuE) {
            specType = "CCQEE\0";
            if (evtRwRand[1] < evtRwProb[1]) {
                reWeight = 1.0;
            } else {
                reWeight = 0.0;
            }
        }

        //
        //  Skip if quasi-elastic NuE
        //
        if (evtTrueNuCCNC == kQE && fabs(evtTrueNuPdg) == kIDNuE) continue;
        //      nb3++;

        if (evtTrueNuCCNC == 1 && ((fabs(evtTrueNuPdg) == kIDNuMu)
                || (fabs(evtTrueNuPdg) == kIDNuE))) {
            specType = "NCMuOrE\0";
            if (fabs(evtTrueNuPdg) == kIDNuE) {
                if (evtRwRand[1] < evtRwProb[1] ) {
                    reWeight = 1.0;
                } else {
                    reWeight = 0.0;
                }
            } else if (fabs(evtTrueNuPdg) == kIDNuMu) {
                if (evtRwRand[2] < evtRwProb[2]) {
                    reWeight = 1.0;
                } else {
                    reWeight = 0.0;
                }
            }

        } else if (evtTrueNuCCNC == kQE && fabs(evtTrueNuPdg) == kIDNuMu) {
            specType = "CCQEMu\0";
            if (evtRwRand[2] < evtRwProb[2]) {
                reWeight = 1.0;
            } else {
                reWeight = 0.0;
            }

        }
        //        std::cout << "BG specType = " << specType << std::endl;
        TMVAEIDTraining::dumpTrainingSampleToCSV(CSVOperation::FILL, cSVFileName);
        TMVAEIDTraining::fillTrainingVariables();

        if (fillTrainingSample) trainingSample->Fill();
        normBkg++;
    }
    std::cout << "normBkg = " << normBkg << std::endl;
    TMVAEIDTraining::dumpTrainingSampleToCSV(CSVOperation::CLOSE, cSVFileName);

    std::cout << "Nsig, Nbig " << normSig << " " << normBkg << " " << std::endl;
    scalepotsig = 18.0e+20 / ((numSignalCorrectProcess / (2.87128765282905034e-02 * 6938.035))*2.99649003626026520e+21); //2.38399959425089536e+19);//NDOS
    scalepotNumuCC = 18.0e+20 / ((numBackgroundNumuCC0 / (6938.035 - 199.21 - 5511.0658))*3.02657807308939830e+21); //2.38399959425089536e+19);//NDOS
    scalepotNueCC = 18.0e+20 / ((numBackgroundNueCC0 / 82.02)*3.02657807308939830e+21); //2.38399959425089536e+19);//NDOS
    scalepotNC = 18.0e+20 / ((numBackgroundNC0 / 2664.42968)*3.02657807308939830e+21); //2.38399959425089536e+19);//NDOS
    //
    //  Write scale factors out to a CSV file to be used in R
    //
    ofstream potScaleFile;
    TString potScaleFileName = "POTScales.csv";
    potScaleFile.open((potScaleFileName.Data()));
    // Header
    potScaleFile << "scalepotsig\t" << "scalepotNumuCC\t" << "scalepotNueCC\t" << "scalepotNC" << std::endl;
    potScaleFile << scalepotsig << "\t" << scalepotNumuCC << "\t" << scalepotNueCC << "\t" << scalepotNC << std::endl;
    potScaleFile.close();



    trainingSample->Write("trainingSample");
}

void TMVAEIDTraining::setTreeBranchAddresses(TTree* treeIncoming) {

    treeIncoming->SetBranchAddress("evtTrueNuCCNC", &evtTrueNuCCNC);
    treeIncoming->SetBranchAddress("evtTrueNuMode", &evtTrueNuMode);
    treeIncoming->SetBranchAddress("evtTrueNuPdg", &evtTrueNuPdg);
    treeIncoming->SetBranchAddress("evtTrueNuEnergy", &evtTrueNuEnergy);
    treeIncoming->SetBranchAddress("evtTrueNuIsFid", &evtTrueNuIsFid);
    treeIncoming->SetBranchAddress("evtTrueNuP4[4]", evtTrueNuP4);
    treeIncoming->SetBranchAddress("evtTrueNuVtx[3]", evtTrueNuVtx);
    treeIncoming->SetBranchAddress("evtOsTag", &evtOsTag);
    treeIncoming->SetBranchAddress("evtOsProb[20]", evtOsProb);
    treeIncoming->SetBranchAddress("evtOsRand", &evtOsRand);
    treeIncoming->SetBranchAddress("evtRwProb[20]", evtRwProb);
    treeIncoming->SetBranchAddress("evtRwRand[20]", evtRwRand);
    treeIncoming->SetBranchAddress("evtTruePdg[20]", evtTruePdg);
    treeIncoming->SetBranchAddress("evtTrueMother[20]", evtTrueMother);
    treeIncoming->SetBranchAddress("evtTrueEnergy[20]", evtTrueEnergy);
    treeIncoming->SetBranchAddress("evtSh1TruePdg", &evtSh1TruePdg);
    treeIncoming->SetBranchAddress("evtSh1TrueDang", &evtSh1TrueDang);
    treeIncoming->SetBranchAddress("evtSh1TrueEDang", &evtSh1TrueEDang);
    treeIncoming->SetBranchAddress("evtSh1Energy", &evtSh1Energy);
    treeIncoming->SetBranchAddress("evtSh1Energy0", &evtSh1Energy0);
    treeIncoming->SetBranchAddress("evtTotalRecoGeV", &evtTotalRecoGeV);
    treeIncoming->SetBranchAddress("evtSh1SliceGeV", &evtSh1SliceGeV);
    treeIncoming->SetBranchAddress("evtSheEnergy", &evtSheEnergy);
    treeIncoming->SetBranchAddress("evtSh1Gap", &evtSh1Gap);
    treeIncoming->SetBranchAddress("evtSh1IsFid", &evtSh1IsFid);
    treeIncoming->SetBranchAddress("evtSh1P4[4]", evtSh1P4);
    treeIncoming->SetBranchAddress("evtSh1Start[3]", evtSh1Start);
    treeIncoming->SetBranchAddress("evtSh1DedxLLL[20]", evtSh1DedxLLL);
    treeIncoming->SetBranchAddress("evtSh1DedxLLT[20]", evtSh1DedxLLT);
    treeIncoming->SetBranchAddress("evtSh1NPlane", &evtSh1NPlane);
    treeIncoming->SetBranchAddress("evtSh1NMIPPlane", &evtSh1NMIPPlane);
    treeIncoming->SetBranchAddress("evtSh1ContStartPlane", &evtSh1ContStartPlane);
    treeIncoming->SetBranchAddress("evtEtot", &evtEtot);
    treeIncoming->SetBranchAddress("evtSh1SumPt", &evtSh1SumPt);
    treeIncoming->SetBranchAddress("evtNTrk", &evtNTrk);
    treeIncoming->SetBranchAddress("evtSh1Enue", &evtSh1Enue);
    treeIncoming->SetBranchAddress("evtSh2Energy", &evtSh2Energy);
    treeIncoming->SetBranchAddress("evtPi0Mgg", &evtPi0Mgg);
    treeIncoming->SetBranchAddress("evtSh1Pi0Mgg", &evtSh1Pi0Mgg);
    treeIncoming->SetBranchAddress("evtSh1VtxGeV", &evtSh1VtxGeV);
    treeIncoming->SetBranchAddress("evtSh1CosTheta", &evtSh1CosTheta);
    treeIncoming->SetBranchAddress("evtSh1Length", &evtSh1Length);
    treeIncoming->SetBranchAddress("evtPi0StartDist", &evtPi0StartDist);
    treeIncoming->SetBranchAddress("evtPi0LineDist", &evtPi0LineDist);
}

void TMVAEIDTraining::fillTrainingVariables() {

    egLLL = (float) evtSh1DedxLLL[0] - (float) evtSh1DedxLLL[1];
    egLLT = (float) evtSh1DedxLLT[0] - (float) evtSh1DedxLLT[1];
    emuLLL = (float) evtSh1DedxLLL[0] - (float) evtSh1DedxLLL[2];
    emuLLT = (float) evtSh1DedxLLT[0] - (float) evtSh1DedxLLT[2];
    epi0LLL = (float) evtSh1DedxLLL[0] - (float) evtSh1DedxLLL[3];
    epi0LLT = (float) evtSh1DedxLLT[0] - (float) evtSh1DedxLLT[3];
    epLLL = (float) evtSh1DedxLLL[0] - (float) evtSh1DedxLLL[5];
    epLLT = (float) evtSh1DedxLLT[0] - (float) evtSh1DedxLLT[5];
    enLLL = (float) evtSh1DedxLLL[0] - (float) evtSh1DedxLLL[6];
    enLLT = (float) evtSh1DedxLLT[0] - (float) evtSh1DedxLLT[6];
    epiLLL = (float) evtSh1DedxLLL[0] - (float) evtSh1DedxLLL[7];
    epiLLT = (float) evtSh1DedxLLT[0] - (float) evtSh1DedxLLT[7];
    gap = (float) evtSh1Gap;
    pi0mass = TMath::Max((float) evtSh1Pi0Mgg, (float) 0.0);
    vtxgev = (float) evtSh1VtxGeV;
    //      costheta = evtSh1SumPt/evtSh1SliceGeV;
    costheta = (float) evtSh1CosTheta;
    length = (float) evtSh1Length / (float) evtSh1Energy;
    shE = (float) evtSh1Energy / (float) evtSh1SliceGeV;
    nueRecEnergy = ((float) evtSh1Energy + 0.282525 + 1.0766 * ((float) evtEtot - (float) evtSh1Energy));

    ntrk = (float) evtNTrk;
    annPi0LineDist = (float) evtPi0LineDist;
    annPi0StartDist = (float) evtPi0StartDist;
}

void TMVAEIDTraining::launchGUI() {
    // Launch the GUI for the root macros
    //    if (!gROOT->IsBatch()) TMVAGui(outfileName);
}

void TMVAEIDTraining::runDiagnostics() {

    //    currentAlgorithm = "MLP method";
    //    currentAlgorithm = "TMlpANN method";
    //    currentAlgorithm = "KNN method";
    //    currentAlgorithm = "BDT method";

    TMVAEIDTraining::transformInputTrees(!isTransformedInput);

    //    TMVAEIDTraining::defineEvaluationTree();

    double Elow = 1.0;
    double Ehigh = 3.0;
    double mVAPlotLow = -1.5;
    double mVAPlotHigh = 1.5;
    int nBinsANN = 50;
    std::cout << std::endl;
    std::cout << "==> Start TMVAClassificationApplication" << std::endl;

    if (!methodsBooked) {
        TMVAEIDTraining::bookTMVAMethods();
    }


    // Select methods (don't look at this code - not of interest)
    if (myMethodList != "") {
        for (std::map<std::string, int>::iterator it = Use.begin(); it != Use.end(); it++) it->second = 0;

        std::vector<TString> mlist = TMVA::gTools().SplitString(myMethodList, ',');
        for (UInt_t i = 0; i < mlist.size(); i++) {
            std::string regMethod(mlist[i]);

            if (Use.find(regMethod) == Use.end()) {
                std::cout << "Method \"" << regMethod
                        << "\" not known in TMVA under this name. Choose among the following:" << std::endl;
                for (std::map<std::string, int>::iterator it = Use.begin(); it != Use.end(); it++) {
                    std::cout << it->first << " ";
                }
                std::cout << std::endl;
                return;
            }
            Use[regMethod] = 1;
            currentAlgorithm = regMethod + " method";
            if (regMethod == "TMlpANN") {
                nBinsANN = 50;
                mVAPlotLow = -0.5;
                mVAPlotHigh = 1.5;
            } else if (regMethod == "MLP") {
                nBinsANN = 50;
                mVAPlotLow = -0.5;
                mVAPlotHigh = 1.5;
            } else if (regMethod == "BDT") {
                nBinsANN = 100;
                mVAPlotLow = -1.0;
                mVAPlotHigh = 1.0;
            } else if (regMethod == "KNN") {
                nBinsANN = 100;
                mVAPlotLow = -1.5;
                mVAPlotHigh = 1.5;
            }
        }
    }


    //    double rsig = 1.0;
    //    double rbg = 0.9;
    //    double trueNuE, etot, enue, sh1Energy, ANN, wSig, wBkg;

    // Use the NN to plot the results for each sample
    // This will give approx. the same result as DrawNetwork.
    // All entries are used, while DrawNetwork focuses on
    // the test sample. Also the xaxis range is manually set.
    bg = new TH1F("bgh", "Background", nBinsANN, mVAPlotLow, mVAPlotHigh);
    bgNumuCC = new TH1F("bgNumuCC", "NuMu CC Background", nBinsANN, mVAPlotLow, mVAPlotHigh);
    bgNC = new TH1F("bgNC", "Neutral Current Background", nBinsANN, mVAPlotLow, mVAPlotHigh);
    bgNueCC = new TH1F("bgNueCC", "Beam NuE Charged Current Background", nBinsANN, mVAPlotLow, mVAPlotHigh);

    sig = new TH1F("sigh", "Signal Nue CC QE", nBinsANN, mVAPlotLow, mVAPlotHigh);
    sigbg = new TH1F("sigbgh", "Mis-reconstructed Signal \"Background\"", nBinsANN, mVAPlotLow, mVAPlotHigh);
    TCanvas* mlpa_canvas2 = new TCanvas("mlpa_canvas2", "Network analysis 2", 10, 10, 1200, 600);
    TLine *line1, *line2;

    //Int_t type;


    // --------------------------------------------------------------------------------------------------

    // --- Create the Reader object
    std::cout << "Defining MTVA::Reader object..." << std::endl;
    reader = new TMVA::Reader("!Color:!Silent");
    std::cout << "Adding variables to Reader..." << std::endl;



    reader->AddVariable("egLLL", &egLLL);
    reader->AddVariable("egLLT", &egLLT);
    reader->AddVariable("emuLLL", &emuLLL);
    reader->AddVariable("emuLLT", &emuLLT);
    reader->AddVariable("epi0LLL", &epi0LLL);
    reader->AddVariable("epi0LLT", &epi0LLT);
    reader->AddVariable("epLLL", &epLLL);
    reader->AddVariable("epLLT", &epLLT);
    reader->AddVariable("enLLL", &enLLL);
    reader->AddVariable("enLLT", &enLLT);
    reader->AddVariable("epiLLL", &epiLLL);
    reader->AddVariable("epiLLT", &epiLLT);
    reader->AddVariable("gap", &gap);
    reader->AddVariable("pi0mass", &pi0mass);
    reader->AddVariable("vtxgev", &vtxgev);
    reader->AddVariable("shE", &shE);
    if (useENu) {
        reader->AddVariable("nueRecEnergy", &nueRecEnergy);
    }
    reader->AddSpectator("sbtype", &type);
    //    reader->AddSpectator("specType", &specType);
    //    reader->AddSpectator("reWeight", &reWeight);

    // --- Book the MVA methods

    //    TString dir = "weights/";
    TString prefix = "TMVAClassification";

    // Book method(s)
    std::cout << "Booking TMVA methods..." << std::endl;
    for (std::map<std::string, int>::iterator it = Use.begin(); it != Use.end(); it++) {
        if (it->second) {
            TString methodName = TString(it->first) + TString(" method");
            TString weightfile = weightDir + prefix + TString("_") + TString(it->first) + TString(".weights.xml");
            reader->BookMVA(methodName, weightfile);
        }
    }




    TMVAEIDTraining::setTreeBranchAddresses(signalIn);
    TMVAEIDTraining::setTreeBranchAddresses(backgroundIn);


    std::cout << "Looping over background input tree to fill hists...Algorithm = " << currentAlgorithm << std::endl;
    std::cout << "backgroundIn->GetEntries() = " << backgroundIn->GetEntries() << std::endl;
    int bgcnt = 0;
    for (int iEvt = 0; iEvt < backgroundIn->GetEntries(); iEvt++) {
        //      if((double)i>(double)backgroundIn->GetEntries()*rbg)continue;
        if (iEvt % 2 == 0)continue;
        //      if(i>36383)continue;
        backgroundIn->GetEntry(iEvt);
        if (!((fabs(evtTrueNuPdg) == 14 && evtRwRand[2] > 0 && evtRwRand[2] < evtRwProb[2])
                || (fabs(evtTrueNuPdg) == 12 && evtRwRand[1] > 0 && evtRwRand[1] < evtRwProb[1])))continue;
        if (!(evtSh1NPlane>-15 && evtSh1Energy > 0.0))continue;
        if (evtSh1DedxLLL[0] - evtSh1DedxLLL[2]<-0.2)continue;
        if (!((evtSh1Energy + 0.282525 + 1.0766 * (evtEtot - evtSh1Energy)) / 1.015 > Elow
                && (evtSh1Energy + 0.282525 + 1.0766 * (evtEtot - evtSh1Energy)) / 1.015 < Ehigh))continue;

        //
        //  Fill training variables
        //
        TMVAEIDTraining::fillTrainingVariables();

        if (evtTrueNuCCNC == 0 && fabs(evtTrueNuPdg) == 14 && evtRwRand[2] > 0 && evtRwRand[2] < evtRwProb[2]) {
            bgNumuCC->Fill(reader->EvaluateMVA(currentAlgorithm), scalepotNumuCC);
            bg->Fill(reader->EvaluateMVA(currentAlgorithm), scalepotNumuCC);
        }
        if (evtTrueNuCCNC == 1 && ((fabs(evtTrueNuPdg) == 14 && evtRwRand[2] > 0 && evtRwRand[2] < evtRwProb[2]) 
                || (fabs(evtTrueNuPdg) == 12 && evtRwRand[1] > 0 && evtRwRand[1] < evtRwProb[1]))) {
            bgNC->Fill(reader->EvaluateMVA(currentAlgorithm), scalepotNC);
            bg->Fill(reader->EvaluateMVA(currentAlgorithm), scalepotNC);
        }
        if (evtTrueNuCCNC == 0 && fabs(evtTrueNuPdg) == 12 && evtRwRand[1] > 0 && evtRwRand[1] < evtRwProb[1]) {
            bgNueCC->Fill(reader->EvaluateMVA(currentAlgorithm), scalepotNueCC);
            bg->Fill(reader->EvaluateMVA(currentAlgorithm), scalepotNueCC);
        }
        bgcnt++;
    }
    std::cout << "bgcnt = " << bgcnt << std::endl;

    std::cout << "Looping over signal input tree to fill hists..." << std::endl;
    std::cout << "signalIn->GetEntries() = " << signalIn->GetEntries() << std::endl;
    int sgcnt = 0;
    for (int iSigEvt = 0; iSigEvt < signalIn->GetEntries(); iSigEvt++) {
        if (iSigEvt % 2 == 0)continue;
        signalIn->GetEntry(iSigEvt);
        if (!(evtTrueNuCCNC == 0 && evtTrueNuPdg == 12 && (evtRwRand[0] > 0 && evtRwRand[0] < evtRwProb[0])))continue;
        //      if(evtSh1TrueDang>10||fabs(evtSh1TruePdg)!=11)continue;
        if (!(evtSh1NPlane>-15 && evtSh1Energy > 0.0))continue;
        if (evtSh1DedxLLL[0] - evtSh1DedxLLL[2]<-0.2)continue;
        if (!((evtSh1Energy + 0.282525 + 1.0766 * (evtEtot - evtSh1Energy)) / 1.015 > Elow
                && (evtSh1Energy + 0.282525 + 1.0766 * (evtEtot - evtSh1Energy)) / 1.015 < Ehigh))continue;

        // fill training variables from Tree entry
        TMVAEIDTraining::fillTrainingVariables();
        sig->Fill(reader->EvaluateMVA(currentAlgorithm), scalepotsig);
        if (fabs(evtSh1TruePdg) != 11)sigbg->Fill(reader->EvaluateMVA(currentAlgorithm), scalepotsig);
        sgcnt++;
    }
    std::cout << "sgcnt = " << sgcnt << std::endl;


    hfom = new TH1F("hfom", "Figure of Merit", nBinsANN, mVAPlotLow, mVAPlotHigh);
    hfomBeamNueBG = new TH1F("hfomBeamNueBG", "Figure of Merit, Beam Nue as BG", nBinsANN, mVAPlotLow, mVAPlotHigh);
    heff = new TH1F("heff", "Signal Efficiency", nBinsANN, mVAPlotLow, mVAPlotHigh);
    hbgeff = new TH1F("hbgeff", "Mis-reconstructed Signal \"Background\" Efficiency", nBinsANN, mVAPlotLow, mVAPlotHigh);
    hbgNumuCCeff = new TH1F("hbgNumuCCeff", "NuMu CC Background Efficiency", nBinsANN, mVAPlotLow, mVAPlotHigh);
    hbgNCeff = new TH1F("hbgNCeff", "Neutral Current Background Efficiency", nBinsANN, mVAPlotLow, mVAPlotHigh);
    hbgNueCCeff = new TH1F("hbgNueCCeff", "Beam NuE Charged Current Background Efficiency", nBinsANN, mVAPlotLow, mVAPlotHigh);
    for (int iBin = 0; iBin < nBinsANN; iBin++) {
        double sumnbg = bg->Integral(iBin + 1, nBinsANN);
        double sumnbgNueCC = bgNueCC->Integral(iBin + 1, nBinsANN);
        double sumnsig = sig->Integral(iBin + 1, nBinsANN);
        double eff = double(sig->Integral(iBin + 1, nBinsANN)) / double(sig->Integral(1, nBinsANN));
        double bgeff = double(bg->Integral(iBin + 1, nBinsANN)) / double(bg->Integral(1, nBinsANN));
        double bgNumuCCeff = double(bgNumuCC->Integral(iBin + 1, nBinsANN)) / double(bgNumuCC->Integral(1, nBinsANN));
        double bgNCeff = double(bgNC->Integral(iBin + 1, nBinsANN)) / double(bgNC->Integral(1, nBinsANN));
        double bgNueCCeff = double(bgNueCC->Integral(iBin + 1, nBinsANN)) / double(bgNueCC->Integral(1, nBinsANN));
        heff->SetBinContent(iBin + 1, eff);
        hbgeff->SetBinContent(iBin + 1, bgeff);
        hbgNumuCCeff->SetBinContent(iBin + 1, bgNumuCCeff);
        hbgNCeff->SetBinContent(iBin + 1, bgNCeff);
        hbgNueCCeff->SetBinContent(iBin + 1, bgNueCCeff);
        //
        //  Fill hfom
        //
        if (sumnsig + sumnbg - sumnbgNueCC < 0.000001) {
            hfom->SetBinContent(iBin + 1, 0);
            continue;
        }
        double significance = 0;
        if (sumnsig + sumnbg - sumnbgNueCC > 0) significance = double(sumnsig) / sqrt(sumnsig + sumnbg - sumnbgNueCC);
        std::cout << "FOM, sig, bkg  " << significance << " " << sumnsig << " " << sumnbg << " " << std::endl;
        hfom->SetBinContent(iBin + 1, significance);
        //
        //  Fill hfomBeamNueBG
        //
        if (sumnsig + sumnbg < 0.000001) {
            hfomBeamNueBG->SetBinContent(iBin + 1, 0);
            continue;
        }
        double significanceV2 = 0;

        if (sumnsig + sumnbg > 0) significanceV2 = double(sumnsig) / sqrt(sumnsig + sumnbg);
        std::cout << "FOM(Beam Nue as BG), sig, bkg  " << significanceV2 << " " << sumnsig << " " << sumnbg << " " << std::endl;
        hfomBeamNueBG->SetBinContent(iBin + 1, significanceV2);
    }

    bg->SetLineColor(kBlue);
    sig->SetLineColor(kRed);
    sig->SetFillStyle(3003);
    sig->SetFillColor(kRed);

    bg->Draw();
    sig->Draw("same");
    sigbg->Draw("same");
    //   sig->DrawNormalized("same", sig->GetEntries()*scalesig);
    TLegend *legend = new TLegend(.75, .80, .95, .95);
    legend->AddEntry(bg, "Background (nhc)");
    legend->AddEntry(sig, "Signal (Nue_CCQE)");
    legend->Draw();
    mlpa_canvas2->cd(0);
    mlpa_canvas2->Update();
    mlpa_canvas2->SaveAs(plotDir + "JMCanvas2.png");
    mlpa_canvas2->SaveAs(plotDir + "JMCanvas2.eps");

    std::cout << "scalepotsig " << scalepotsig << std::endl;
    std::cout << "scalepotNumuCC " << scalepotNumuCC << std::endl;
    std::cout << "scalepotNC " << scalepotNC << std::endl;
    std::cout << "scalepotNueCC " << scalepotNueCC << std::endl;

    TCanvas* mlpa_canvas3 = new TCanvas("mlpa_canvas3", "Network analysis 3", 10, 10, 800, 400);
    mlpa_canvas3->Divide(2, 1);
    mlpa_canvas3->cd(1);
    Double_t maxFOM = hfom->GetBinContent(hfom->GetMaximumBin());
    Double_t maxFOMx = hfom->GetBinCenter(hfom->GetMaximumBin());
    hfom->GetYaxis()->SetTitle("FOM");
    legend = new TLegend(.75, .80, .95, .95);
    legend->AddEntry(hfom, TString::Format("Max FOM = %4.4f at %4.2f", maxFOM, maxFOMx));
    hfom->GetXaxis()->SetTitle("ANN");
    //    hfom->GetXaxis()->SetRangeUser(0.4, 1.2);
    hfom->Draw("l");
    hfomBeamNueBG->SetLineColor(kRed);
    hfomBeamNueBG->Draw("lsame");
    line1 = new TLine(hfom->GetBinCenter(hfom->GetMaximumBin()), 0, hfom->GetBinCenter(hfom->GetMaximumBin()), hfom->GetMaximum());
    line1->SetLineWidth(4);
    line1->SetLineColor(kRed);
    line1->Draw("same");


    mlpa_canvas3->Update();

    mlpa_canvas3->cd(2);
    heff->GetYaxis()->SetTitle("PID Eff.");
    heff->GetXaxis()->SetTitle("ANN");
    heff->GetYaxis()->SetRangeUser(0, 1);
    //    heff->GetXaxis()->SetRangeUser(0.4, 1.2);
    heff->Draw("l");
    line2 = new TLine(hfom->GetBinCenter(hfom->GetMaximumBin()), 0, hfom->GetBinCenter(hfom->GetMaximumBin()), 1.0);
    line2->SetLineWidth(4);
    line2->SetLineColor(kRed);
    line2->Draw("same");

    mlpa_canvas3->Update();
    mlpa_canvas3->SaveAs(plotDir + "JMCanvas3.png");
    mlpa_canvas3->SaveAs(plotDir + "JMCanvas3.eps");

    TCanvas* mlpa_canvas4 = new TCanvas("mlpa_canvas4", "Network analysis 4", 10, 10, 600, 400);
    mlpa_canvas4->Divide(1, 1);
    mlpa_canvas4->cd(1);
    bg->GetYaxis()->SetTitle("Events");
    bg->GetXaxis()->SetTitle("ANN");
    bg->SetLineColor(kBlue);
    bgNumuCC->SetLineColor(9);
    bgNC->SetLineColor(7);
    bgNueCC->SetLineColor(6);
    sig->SetLineColor(kRed);
    sig->SetFillStyle(3003);
    sig->SetFillColor(kRed);

    bg->Draw();
    bgNumuCC->Draw("same");
    bgNC->Draw("same");
    bgNueCC->Draw("same");
    sig->Draw("same");
    //   sig->DrawNormalized("same", sig->GetEntries()*scalesig);
    legend = new TLegend(.75, .80, .95, .95);
    legend->AddEntry(bg, "Background (nhc)");
    legend->AddEntry(bgNumuCC, "Beam #nu_{#mu} CC");
    std::cout << "bgNC = " << bgNC << std::endl;
    legend->AddEntry(bgNC, "Beam NC");
    legend->AddEntry(bgNueCC, "Beam #nu_{e} CC");
    legend->AddEntry(sig, "Signal #nu_{e} CC");
    legend->Draw();
    mlpa_canvas4->Update();
    mlpa_canvas4->SaveAs(plotDir + "JMCanvas4.png");
    mlpa_canvas4->SaveAs(plotDir + "JMCanvas4.eps");

    // Get elapsed time
    sw.Stop();
    std::cout << "--- End of event loop: ";
    sw.Print();

    TMVAEIDTraining::saveDiagnosticHists();

    std::cout << "ns0,ns1,ns2,ns3,scalepotsig " << numSignalCorrectProcess << " " << numSignalCorrectShower << " " << numSignalCorrectShowerNotObviousMuon << " " << ns3 << " " << scalepotsig << std::endl;
    double Dns0 = double(numSignalCorrectProcess * scalepotsig);
//    double Dns1 = double(numSignalCorrectShower * scalepotsig);
//    double Dns2 = double(numSignalCorrectShowerNotObviousMuon * scalepotsig);
    //    double Dns3 = double(ns3 * scalepotsig);

//    double DnbNumuCC0 = double(numBackgroundNumuCC0 * scalepotNumuCC);
//    double DnbNumuCC1 = double(numBackgroundPassingFirstShCutsNumuCC1 * scalepotNumuCC);
//    double DnbNumuCC2 = double(numBackgroundPassingNotObviousMuonsShCutsNumuCC2 * scalepotNumuCC);
    //    double DnbNumuCC3 = double(nbNumuCC3 * scalepotNumuCC);

//    double DnbNC0 = double(numBackgroundNC0 * scalepotNC);
//    double DnbNC1 = double(numBackgroundPassingFirstShCutsNC1 * scalepotNC);
//    double DnbNC2 = double(numBackgroundPassingNotObviousMuonsShCutsNC2 * scalepotNC);
    //    double DnbNC3 = double(nbNC3 * scalepotNC);

//    double DnbNueCC0 = double(numBackgroundNueCC0 * scalepotNueCC);
//    double DnbNueCC1 = double(numBackgroundPassingFirstShCutsNueCC1 * scalepotNueCC);
//    double DnbNueCC2 = double(numBackgroundPassingNotObviousMuonsShCutsNueCC2 * scalepotNueCC);
    //    double DnbNueCC3 = double(nbNueCC3 * scalepotNueCC);

//    double Dnb0 = double(DnbNumuCC0 + DnbNC0 + DnbNueCC0);
//    double Dnb1 = double(DnbNumuCC1 + DnbNC1 + DnbNueCC1);
//    double Dnb2 = double(DnbNumuCC2 + DnbNC2 + DnbNueCC2);
    //    double Dnb3 = double(DnbNumuCC3 + DnbNC3 + DnbNueCC3);

    //   double scalepot = 18.0e+20/((nbNumuCC0/7175.2)*2.8123e+21);//2.38399959425089536e+19);//NDOS
    std::cout << "ANN  =      " << hfom->GetBinCenter(hfom->GetMaximumBin()) << std::endl;
    std::cout << "FOM  hist =      " << hfom->GetBinContent(hfom->GetMaximumBin()) << std::endl;
    //    std::cout << "Dns3 = "<< Dns3  << std::endl;
    //    std::cout << "Dnb3 = "<< Dnb3  << std::endl;
    std::cout << "Dns0 = " << Dns0 << std::endl;
    //   std::cout << "FOM       =      " << double(Dns3) / sqrt(Dns3 + Dnb3) << std::endl;

    //    std::cout << "Eff. =      " << double(Dns3) / double(Dns0) << std::endl;
    std::cout << "ns2 " << numSignalCorrectShowerNotObviousMuon << std::endl;
    std::cout << "nb2, nb3 " << numBackgroundPassingNotObviousMuonsShCuts2 << " " << nb3 << std::endl;

    //    std::cout << "BG Eff       . =   " << double(Dnb3) / double(Dnb0) << std::endl;
    //    std::cout << "BG NumuCC Eff. =   " << double(DnbNumuCC3) / double(DnbNumuCC0) << std::endl;
    //   std::cout << "BG NC     Eff. =   " << double(DnbNC3) / double(DnbNC0) << std::endl;
    //   std::cout << "BG NueCC  Eff. =   " << double(DnbNueCC3) / double(DnbNueCC0) << std::endl;

}

void TMVAEIDTraining::saveDiagnosticHists() {


    TFile f(histDir + "annDiagnosticHists.root", "recreate");
    hbgeff->Write("hbgeff");
    hbgNumuCCeff->Write("hbgNumuCCeff");
    hbgNCeff->Write("hbgNCeff");
    hbgNueCCeff->Write("hbgNueCCeff");
    heff->Write("heff");
    hfom->Write("hfom");
    //   hfomBeamNueBG->Write("hfomBeamNueBG");
    sigbg->Write("sigbg");
    bg->Write("bg");
    bgNumuCC->Write("bgNumuCC");
    bgNC->Write("bgNC");
    bgNueCC->Write("bgNueCC");
    sig->Write("sig");
    //   annMC->Write("annMC");
    f.Close();

}
