/////////////////////////////////////////////////////////////////////////
/// \brief   Class to help Slicer4D manage the collection of hits
/// \authors mbaird42@indiana.edu
/// \date    2013/05/01
/////////////////////////////////////////////////////////////////////////

#include "Slicer/PointManager.h"

#include <cmath>
#include <cstdlib>
#include <iostream>

namespace slicer {

  //......................................................................
  PointManager::PointManager()
  { }

  //......................................................................
  PointManager::~PointManager()
  { }

  //......................................................................
  void PointManager::Initalize(double eps, double pepen,
			       double distpen, double opvplpen,
			       bool usepepen)
  {
    fEps      = eps;
    fPEPen    = pepen;
    fDistPenP = distpen/6.2;
    fDistPenC = distpen/4.1;
    fOpVPlPen = opvplpen;
    fUsePEPen = usepepen;
  }

  //......................................................................
  void PointManager::Reset(unsigned int n, bool skip)
  {
    // initialize cluster IDs to -1 (-1 = "unassigned")
    fClustIDs.clear();
    for(unsigned int i = 0; i < n; ++i) fClustIDs.push_back(-1);

    fNhits = n;

    fTime .clear();
    fTres .clear();
    fPlane.clear();
    fCell .clear();
    fPE   .clear();
    fView .clear();

    fSkip = skip;

    NScount = 0;
    RQcount = 0;
  }

  //......................................................................
  void PointManager::FindNeighbors()
  {

    fNeighborList.clear();
    fNeighborList.resize(fNhits);

    // First add each point to its own list (the DBSCAN algorithm needs
    // this to work correctly.)
    for(unsigned int i = 0; i < fNhits; ++i) fNeighborList[i].push_back(i);    

    // use i+1 < fNhits to protect against the case where fNhits = 0
    for(unsigned int i = 0; i+1 < fNhits; ++i) {
      for(unsigned int j = i+1; j < fNhits; ++j) {

	// If hits are too far apart in time, then we don't need to determine
	// the neighbor score.  Since the hits are time sorted, this will
	// also be true of all hits later in the list.
	if( fSkip && ((fTime[j]-fTime[i]) > 1000.0) ) break;

	double score = this->NeighborScore(i,j);
	if(score <= fEps) {
	  // i and j are neighbors, so they go into each others lists
	  fNeighborList[i].push_back(j);
	  fNeighborList[j].push_back(i);
	} // end if

      } // end for j
    } // end for i

  }

  //......................................................................
  void PointManager::CleanUpNeighborhood()
  {
    // An optional function to clean up the neighbor assignments (to
    // be run AFTER the FindNeighbors function.)
    //
    // Currently, this function will remove points from neighborhoods
    // iff all of their neighbors are in the opposite view.

    for(unsigned int i = 0; i < fNhits; ++i) {

      if(fNeighborList[i].size() <= 1) continue;

      bool goodneighbors = false;

      // The first point in the neighbor list is the original point,
      // so start with the second point in the list.
      for(unsigned int j = 1; j < fNeighborList[i].size(); ++j) {
	if(fView[i] == fView[fNeighborList[i][j]]) {
	  goodneighbors = true;
	  break;
	}
      } // end for j (loop over all neighbors)

      if(!goodneighbors) {
	
	// erase point i from the neighbor list of all of point i's neighbors
	for(unsigned int j = 1; j < fNeighborList[i].size(); ++j) {

	  unsigned int n = fNeighborList[i][j];
	  std::vector<unsigned int> nList = fNeighborList[n];

	  fNeighborList[n].clear();

	  for(unsigned int k = 0; k < nList.size(); ++k) {

	    if(nList[k] != i) fNeighborList[n].push_back(nList[k]);

	  } // end for k (loop through the neighbor list of the j-th neighbor of i)

	} // end for j (loop over all neighbors of i)

	// clear point i's neighbor list and add point i back to that list
	fNeighborList[i].clear();
	fNeighborList[i].push_back(i);

      } // end if no good neighbors

    } // end for i (loop over all hits)

  }

  //......................................................................
  std::vector<unsigned int> PointManager::regionQuery(unsigned int point)
  {    
    RQcount++;
    return fNeighborList[point]; 
  }

  //......................................................................
  void PointManager::changeClustID(unsigned int point, unsigned int id)
  {
    fClustIDs[point] = id;
  }

  //......................................................................
  void PointManager::AddHit(double t, double tres,
			    unsigned int p, unsigned int c, double pe,
			    geo::View_t v)
  {
    fTime .push_back(t);
    fTres .push_back(tres);
    fPlane.push_back(p);
    fCell .push_back(c);
    fPE   .push_back(pe);
    fView .push_back(v);
  }

  //......................................................................
  double PointManager::NeighborScore(unsigned int i, unsigned int pt)
  {
    NScount++;
    
    double score = 0.0;
    double T = 0.0, R = 0.0, Z = 0.0, XY = 0.0, PE = 0.0;

    double dT = fabs(fTime[i] - fTime[pt]);
    double dC =  abs(fCell[i] - fCell[pt]);
    double dP =  abs(fPlane[i]- fPlane[pt]);

    ///\ TODO: perhaps some of the below calculations can be made more
    //         efficient...
    if(fView[i] == fView[pt]) {
      XY = dC/fDistPenC;
      Z  = dP/fDistPenP;
      R  = sqrt(dC*dC*16.81 + dP*dP*38.44);
      if(fUsePEPen) {
	double denom = sqrt(fPE[i]*fPE[i]+fPE[pt]*fPE[pt]);
	if(denom == 0.0) denom = 1.0e-6;
	PE = pow(fPEPen/denom, 5);
      }
    }
    else {
      XY = 0.0;
      Z  = dP/fOpVPlPen;
      R  = dP*6.2;
      if(fUsePEPen) {
	double denom = sqrt(fPE[i]*fPE[i]+fPE[pt]*fPE[pt]);
	if(denom == 0.0) denom = 1.0e-6;
	PE = pow(fPEPen/denom, 5);
      }
    }

    T = (dT - R/(30.0))/sqrt(fTres[i]*fTres[i]+fTres[pt]*fTres[pt]);

    score = T*T + Z*Z + XY*XY + PE;
    
    return score;
  }

  //......................................................................
  void PointManager::DebugPrint()
  {
    std::cout << "\n\n==========\nDebug Print:\n\n";
    std::cout << "Number of hits = " << fNhits << "\n";
    std::cout << "Number of Neighbor Score calculations = " << NScount << "\n";
    std::cout << "Number of Region Queries = " << RQcount << "\n";
    std::cout << "\n==========\n\n";
  }

  //......................................................................
  void PointManager::PreparekDistInfo(unsigned int kmax)
  {
    // This function calculates the nearest neighbor distances for all
    // hits.  This function is NOT used for clustering, it is only used to
    // help determine the correct input parameters for Slicer4D.

    // BTW, this is NOT maximally efficient but it doesn't have to be since
    // it is only intended to be used to help tune the Slicer4D input
    // parameters.

    fkDistList.clear();
    fkDistList.resize(fNhits);

    for(unsigned int i = 0; i < fNhits; ++i) {

      std::vector<double> list;
      for(unsigned int k = 0; k < kmax; ++k) list.push_back(1.0e9);

      for(unsigned int j = 0; j < fNhits; ++j) {

	double dist = this->NeighborScore(i,j);

	if(dist < list[kmax-1]) {
	  // add this dist to the list, resort, and resize
	  list.push_back(dist);
	  
	  bool done = false;
	  while(!done) {
	    done = true;
	    for(unsigned int k = 1; k <= kmax; ++k) {  
	      if(list[k-1] > list[k]) {  
		double temp = list[k];
		list[k]   = list[k-1];
		list[k-1] = temp;	    
		done = false;
	      } // end if
	    } // end for k
	  } // end while !done

	  list.resize(kmax);

	} // end if
      } // end for j

      fkDistList[i] = list;

    } // end for i

  }

  //......................................................................
  double PointManager::kDist(unsigned int k, unsigned int pt)
  {
    // This function returns the k-dist value (distance to the k-th nearest
    // neighbor) for a point.  This function is NOT used for clustering, it
    // is only used to help determine the correct input parameters for Slicer4D.

    std::vector<double> list = fkDistList[pt];

    return list[k];
  }

} // end namespace slicer
/////////////////////////////////////////////////////////////////////////
