////////////////////////////////////////////////////////////////////////
// \file    S4DParamCalc.cxx
// \brief   Module calculate the best input parameters for Slicer4D.
// \version $Id: S4DParamCalc_module.cc,v 1.4 2013-05-06 10:30:00 mbaird42 Exp $
// \authors mbaird42@fnal.gov 
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <vector>
#include <algorithm>

// ROOT includes
#include "TH1F.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "Calibrator/Calibrator.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/CellGeo.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "Slicer/PointManager.h"



// S4DParamCalc header
namespace slicer {
  class S4DParamCalc : public art::EDAnalyzer {
  public:
    explicit S4DParamCalc(fhicl::ParameterSet const& pset);    
    ~S4DParamCalc();                        
    
    void analyze(const art::Event& evt); 
    void reconfigure(const fhicl::ParameterSet& pset);
    void beginJob();
    
  private:
    std::string fCellHitLabel;
    std::string fSlicerLabel;
    bool        fMakeKDistPlots;

    // input parameters to Point Manager
    double       fEpsilon;            ///< Minimum causal separation value
    unsigned int fMinPts;             ///< minimum number of casual neighbors for a hit to be in a cluster

    // input parameters for calculating the causal separation value
    bool         fUseLowPEPen;        ///< Add a causal distance penalty for PE (low PE is INCREASES the distance)
    double       fPEPen;              ///< PE value for which the low PE penalty will be 1.
    double       fDistPen;            ///< Distance (in cm) for which the distance penalty will be 1 (roughly one strong interaction length.)
    double       fOpVPlPen;           ///< Allowed plane gap for hits in opposite views for which the distance penalty will be 1.

    // helper class for calculating neighbor distances
    PointManager fPtMan;
    
    // Histograms
    TH1F *fPEnoise;
    TH1F *fPEphys;

    TH1F *fDeltaT0;
    TH1F *fDeltaT1;
    TH1F *fDeltaX0;
    TH1F *fDeltaX1;
    TH1F *fDeltaY0;
    TH1F *fDeltaY1;
    TH1F *fDeltaZ0;
    TH1F *fDeltaZ1;
    TH1F *fDeltaP0;
    TH1F *fDeltaP1;

    TH1F *fAveTClose;
    TH1F *fAveDClose;
    TH1F *fAvePClose;

    TH1F *fTClose3;
    TH1F *fDClose3;
    TH1F *fPClose3;

    TH1F *fTClose4;
    TH1F *fDClose4;
    TH1F *fPClose4;

    TH1F *fTClose5;
    TH1F *fDClose5;
    TH1F *fPClose5;

    TH1F *fTres;
    TH1F *fTresNoise;
  };
}



namespace slicer
{
  //.......................................................................
  S4DParamCalc::S4DParamCalc(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset)
  {
    this->reconfigure(pset);
    fPtMan.Initalize(fEpsilon, fPEPen, fDistPen, fOpVPlPen, fUseLowPEPen);
  }

  //......................................................................
  S4DParamCalc::~S4DParamCalc()
  { }

  //......................................................................
  void S4DParamCalc::reconfigure(const fhicl::ParameterSet& pset)
  {
    fCellHitLabel   = pset.get< std::string  >("CellHitLabel");
    fSlicerLabel    = pset.get< std::string  >("SlicerLabel");
    fMakeKDistPlots = pset.get< bool         >("MakeKDistPlots");
    fEpsilon        = pset.get< double       >("Epsilon");
    fMinPts         = pset.get< unsigned int >("MinPts");
    fUseLowPEPen    = pset.get< bool         >("UseLowPEPen");
    fPEPen          = pset.get< double       >("PEPen");
    fDistPen        = pset.get< double       >("DistPen");
    fOpVPlPen       = pset.get< double       >("OppViewPlanePen");
  }
      
  //......................................................................
  void S4DParamCalc::beginJob()
  {
    art::ServiceHandle<art::TFileService> tfs;

    // book the histograms
    fPEnoise = tfs->make<TH1F>("fPEnoise",
			       "PE for hits in the noise cluster",
			       200,0.0,2000.0);
    fPEphys  = tfs->make<TH1F>("fPEphys",
			       "PE for all hits in the non-noise clusters",
			       200,0.0,2000.0);
    fDeltaT0 = tfs->make<TH1F>("fDeltaT0",
			       "Delta T for all hits;delta T [ns];count",
			       1100000,-550000.0,550000.0);
    fDeltaT1 = tfs->make<TH1F>("fDeltaT1",
			       "Delta T for all hits within 10 planes;delta T [ns];count",
			       1100000,-550000.0,550000.0);
    fDeltaX0 = tfs->make<TH1F>("fDeltaX0",
			       "Delta X for all hits in the X-view;delta X [cm];count",
			       400,-1600,1600);  // 8 cm bins
    fDeltaX1 = tfs->make<TH1F>("fDeltaX1",
			       "Delta X for all hits in the X-view within 1 usec;delta X [cm];count",
			       400,-1600,1600);  // 8 cm bins
    fDeltaY0 = tfs->make<TH1F>("fDeltaY0",
			       "Delta Y for all hits in the Y-view;delta Y [cm];count",
			       400,-1600,1600);  // 8 cm bins
    fDeltaY1 = tfs->make<TH1F>("fDeltaY1",
			       "Delta Y for all hits in the Y-view within 1 usec;delta Y [cm];count",
			       400,-1600,1600);  // 8 cm bins
    fDeltaZ0 = tfs->make<TH1F>("fDeltaZ0",
			       "Delta Z for all hits in the same view;delta Z [cm];count",
			       1000,-6000,6000);  // 12 cm bins
    fDeltaZ1 = tfs->make<TH1F>("fDeltaZ1",
			       "Delta Z for all hits in the same view within 1 usec;delta Z [cm];count",
			       1000,-6000,6000);  // 12 cm bins
    fDeltaP0 = tfs->make<TH1F>("fDeltaP0",
			       "Delta Plane for all hits in opposite views;delta Plane;count",
			       1000,-1000,1000);  // 2 plane bins
    fDeltaP1 = tfs->make<TH1F>("fDeltaP1",
			       "Delta Plane for all hits in opposite views within 1 usec;delta Plane;count",
			       1000,-1000,1000);  // 2 plane bins

    fAveTClose = tfs->make<TH1F>("fAveTClose",
				 "Average delta-time for 4 closest neighbors in the slice;time [ns];count",
				 1000,0.0,1000.0);
    fAveDClose = tfs->make<TH1F>("fAveDClose",
				 "Average distance for 4 closest neighbors in the same view and in the same slice;dist [cm];count",
				 160,0.0,1600.0);
    fAvePClose = tfs->make<TH1F>("fAvePClose",
				 "Average delta plane for 4 closest neighbors in opposite views and in the same slice;# of planes;count",
				 100,0.0,200.0);
    fTClose3 = tfs->make<TH1F>("fTClose3",
			       "Delta-time for 3rd closest neighbor in the slice;time [ns];count",
			       1000,0.0,1000.0);
    fDClose3 = tfs->make<TH1F>("fDClose3",
			       "Distance to 3rd closest neighbor in the same view and in the same slice;dist [cm];count",
			       160,0.0,1600.0);
    fPClose3 = tfs->make<TH1F>("fPClose3",
			       "Delta plane for 3rd closest neighbor in opposite view and in the same slice;# of planes;count",
			       100,0.0,200.0);
    fTClose4 = tfs->make<TH1F>("fTClose4",
			       "Delta-time for 4th closest neighbor in the slice;time [ns];count",
			       1000,0.0,1000.0);
    fDClose4 = tfs->make<TH1F>("fDClose4",
			       "Distance to 4th closest neighbor in the same view and in the same slice;dist [cm];count",
			       160,0.0,1600.0);
    fPClose4 = tfs->make<TH1F>("fPClose4",
			       "Delta plane for 4th closest neighbor in opposite view and in the same slice;# of planes;count",
			       100,0.0,200.0);
    fTClose5 = tfs->make<TH1F>("fTClose5",
			       "Delta-time for 5th closest neighbor in the slice;time [ns];count",
			       1000,0.0,1000.0);
    fDClose5 = tfs->make<TH1F>("fDClose5",
			       "Distance to 5th closest neighbor in the same view and in the same slice;dist [cm];count",
			       160,0.0,1600.0);
    fPClose5 = tfs->make<TH1F>("fPClose5",
			       "Delta plane for 5th closest neighbor in opposite view and in the same slice;# of planes;count",
			       100,0.0,200.0);

    fTres      = tfs->make<TH1F>("fTres",
				 "Computed timing resolution for hits in non-noise slices;time [ns];count",
				 500,0.0,500.0);
    fTresNoise = tfs->make<TH1F>("fTresNoise",
				 "Computed timing resolution for hits in the noise slice;time [ns];count",
				 500,0.0,500.0);

  }

  //......................................................................
  void S4DParamCalc::analyze(const art::Event& evt)
  {
    // std::cout << "\n\n====================\nBegin processing for event: "
    // << evt.id() << "\n";

    // declare geometry service handle
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<calib::Calibrator> cal;

    // get the CellHits
    art::Handle< std::vector<rb::CellHit> > hitcol;
    evt.getByLabel(fCellHitLabel, hitcol);

    std::vector<art::Ptr<rb::CellHit > > hitlist;
    for(unsigned int i = 0; i < hitcol->size(); ++i){
      art::Ptr<rb::CellHit> hit(hitcol, i);
      hitlist.push_back(hit);
    }

    // get the Slices
    art::Handle< std::vector< rb::Cluster > > slices;
    evt.getByLabel(fSlicerLabel, slices);

    // Set the PointManager for this event.  (Use false for skipping the
    // delta T calculation since we want the correct distances for all of
    // the noise hits.)
    fPtMan.Reset(hitcol->size(), false);

    // General Plan:
    //
    // loop over all slices, get all hits in the slice
    //   loop over all hits
    //      - add hit to PtMan
    //      - add PE to noise or phys histo

    // std::cout << "\nMaking PE plots...";

    for(unsigned int sli = 0; sli < slices->size(); ++sli) {
      if((*slices)[sli].IsNoise()) {
	for(unsigned int hiti = 0; hiti < (*slices)[sli].NCell(); ++hiti) {
	  const rb::CellHit* hit = (*slices)[sli].Cell(hiti).get();
	  fPtMan.AddHit(hit->TNS(), cal->GetTimeRes(*hit),
			hit->Plane(), hit->Cell(),
			hit->PE(), hit->View());
	  fPEnoise->Fill(hit->PE());
	  fTresNoise->Fill(cal->GetTimeRes(*hit));
	}
      }
      else {
	for(unsigned int hiti = 0; hiti < (*slices)[sli].NCell(); ++hiti) {
	  const rb::CellHit* hit = (*slices)[sli].Cell(hiti).get();
	  fPtMan.AddHit(hit->TNS(), cal->GetTimeRes(*hit),
			hit->Plane(), hit->Cell(),
			hit->PE(), hit->View());
	  fPEphys->Fill(hit->PE());
	  fTres->Fill(cal->GetTimeRes(*hit));
	}
      }
    }



    // std::cout << "\t(done)\nMaking timing/spatial resolution plots...";

    //
    // Make plots form pairs of hits in the same non-noise slice
    //

    // define number of neighbors to keep track of
    unsigned int Nn = 5;

    for(unsigned int sli = 0; sli < slices->size(); ++sli) {
      if((*slices)[sli].IsNoise()) continue;

      std::vector< std::vector< double > > D;
      std::vector< std::vector< double > > P;
      std::vector< std::vector< double > > T;
      for(unsigned int i = 0; i < (*slices)[sli].NCell(); ++i) {
	std::vector< double > temp;
	for(unsigned int j = 0; j < Nn; ++j) temp.push_back(1.0e6);
	D.push_back(temp);
	P.push_back(temp);
	T.push_back(temp);
      }

      for(unsigned int hi = 0; hi < (*slices)[sli].NCell(); ++hi) {
	const rb::CellHit* hiti = (*slices)[sli].Cell(hi).get();
	for(unsigned int hj = 0; hj < (*slices)[sli].NCell(); ++hj) {
	  if(hi == hj) continue;
	  const rb::CellHit* hitj = (*slices)[sli].Cell(hj).get();
	  
	  // compute delta Time
	  T[hi].push_back(fabs(hiti->TNS()-hitj->TNS()));
	  std::sort(T[hi].begin(),T[hi].end());
	  T[hi].resize(Nn);

	  // if same view, compute 2D Distance
	  if(hiti->View() == hitj->View()) {
	    
	    // get xyz info about i-th hit
	    unsigned int iplane = hiti->Plane();
	    unsigned int icell  = hiti->Cell();
	    const geo::PlaneGeo* ip = geom->Plane(iplane);
	    const geo::CellGeo*  ic = ip->  Cell (icell);
	    double ixyz[3];
	    ic->GetCenter(ixyz);
	    double iz  = ixyz[2];
	    double ixy = 0.0;
	    if (ip->View()==geo::kX) ixy = ixyz[0];
	    if (ip->View()==geo::kY) ixy = ixyz[1];
	    
	    // get xyz info about j-th hit
	    unsigned int jplane = hitj->Plane();
	    unsigned int jcell  = hitj->Cell();
	    const geo::PlaneGeo* jp = geom->Plane(jplane);
	    const geo::CellGeo*  jc = jp->  Cell (jcell);
	    double jxyz[3];
	    jc->GetCenter(jxyz);
	    double jz  = jxyz[2];
	    double jxy = 0.0;
	    if (jp->View()==geo::kX) jxy = jxyz[0];
	    if (jp->View()==geo::kY) jxy = jxyz[1];

	    double dist = sqrt((iz-jz)*(iz-jz) + (ixy-jxy)*(ixy-jxy));
	    D[hi].push_back(dist);
	    std::sort(D[hi].begin(),D[hi].end());
	    D[hi].resize(Nn);

	  }

	  // if opposite view, compute delta Plane
	  if(hiti->View() != hitj->View()) {
	    P[hi].push_back(abs((int)hiti->Plane()-(int)hitj->Plane()));
	    std::sort(P[hi].begin(),P[hi].end());
	    P[hi].resize(Nn);
	  }	  

	}  // end loop over hits (hj)

	double Tave = 0.0;
	int    Tn   = 0;
	std::vector<double> Ttemp = T[hi];
	double Dave = 0.0;
	int    Dn   = 0;
	std::vector<double> Dtemp = D[hi];
	double Pave = 0.0;
	int    Pn   = 0;
	std::vector<double> Ptemp = P[hi];
	for(unsigned int a = 0; a < Nn; ++a) {
	  if(Ttemp[a] < 1.0e6) {
	    Tave += Ttemp[a];
	    Tn++;
	  }
	  if(Dtemp[a] < 1.0e6) {
	    Dave += Dtemp[a];
	    Dn++;
	  }
	  if(Ptemp[a] < 1.0e6) {
	    Pave += Ptemp[a];
	    Pn++;
	  }
	} // end loop over a
	if(Tn > 0) Tave = Tave/(double)Tn;
	else       Tave = -1.0;
	if(Dn > 0) Dave = Dave/(double)Dn;
	else       Dave = -1.0;
	if(Pn > 0) Pave = Pave/(double)Pn;
	else       Pave = -1.0;

	fAveTClose->Fill(Tave);
	fAveDClose->Fill(Dave);
	fAvePClose->Fill(Pave);
	
	fTClose3->Fill(Ttemp[2]);
	fDClose3->Fill(Dtemp[2]);
	fPClose3->Fill(Ptemp[2]);

	fTClose4->Fill(Ttemp[3]);
	fDClose4->Fill(Dtemp[3]);
	fPClose4->Fill(Ptemp[3]);

	fTClose5->Fill(Ttemp[4]);
	fDClose5->Fill(Dtemp[4]);
	fPClose5->Fill(Ptemp[4]);

      }  // end loop over hits (hi)
    }  // end loop over (slices sli)
    

    //
    // Make plots from ALL pairs of hits
    //

    // use i+1 < fNhits to protect against the case where fNhits = 0
    for(unsigned int i = 0; i+1 < hitlist.size(); ++i) {

      // get xyz info about i-th hit
      unsigned int iplane = hitlist[i]->Plane();
      unsigned int icell  = hitlist[i]->Cell();
      const geo::PlaneGeo* ip = geom->Plane(iplane);
      const geo::CellGeo*  ic = ip->  Cell (icell);
      double ixyz[3];
      ic->GetCenter(ixyz);
      double iz  = ixyz[2];
      double ixy = 0.0;
      if (ip->View()==geo::kX) ixy = ixyz[0];
      if (ip->View()==geo::kY) ixy = ixyz[1];

      for(unsigned int j = i+1; j < hitlist.size(); ++j) {

	// get xyz info about j-th hit
	unsigned int jplane = hitlist[j]->Plane();
	unsigned int jcell  = hitlist[j]->Cell();
	const geo::PlaneGeo* jp = geom->Plane(jplane);
	const geo::CellGeo*  jc = jp->  Cell (jcell);
	double jxyz[3];
	jc->GetCenter(jxyz);
	double jz  = jxyz[2];
	double jxy = 0.0;
	if (jp->View()==geo::kX) jxy = jxyz[0];
	if (jp->View()==geo::kY) jxy = jxyz[1];

	//
	// Make delta-TXYZ plots...
	//
	double dt = hitlist[i]->TNS()-hitlist[j]->TNS();
	int    dp = (int)hitlist[i]->Plane()-(int)hitlist[j]->Plane();

	fDeltaT0->Fill(dt);
	if(abs(dp) <= 10) fDeltaT1->Fill(dt);

	// same view X
	if(ip->View() == geo::kX && jp->View() == geo::kX) {
	  fDeltaX0->Fill(ixy-jxy);
	  fDeltaZ0->Fill(iz-jz);
	  if(fabs(dt) <= 1000.0) {
	    fDeltaX1->Fill(ixy-jxy);
	    fDeltaZ1->Fill(iz-jz);
	  }
	}

	// same view Y
	if(ip->View() == geo::kY && jp->View() == geo::kY) {
	  fDeltaY0->Fill(ixy-jxy);
	  fDeltaZ0->Fill(iz-jz);
	  if(fabs(dt) <= 1000.0) {
	    fDeltaY1->Fill(ixy-jxy);
	    fDeltaZ1->Fill(iz-jz);
	  }
	}

	// opposite views
	if(ip->View() != jp->View()) {
	  fDeltaP0->Fill(dp);
	  if(fabs(dt) <= 1000.0) {
	    fDeltaP1->Fill(dp);
	  }
	}

      }
    }


    // std::cout << "\t(done)\n";

    if(fMakeKDistPlots) {

      // std::cout << "Preparing k-Dist information...";
      
      // have PtMan initialize what it needs to do the k-dist calculations
      // loop over all hits
      //   - get the k-dist values for k = 2 to 7 and fill std::vectors of doubles
      //   - sort each vector and fill the histo
      //   - MAKE SURE THESE HISTOS ARE WRITTEN AND RESET PER EVENT!!!!!
      
      TH1F *f2dist = new TH1F("f2dist","2-dist for all hits",20001,-0.5,20000.5);
      TH1F *f3dist = new TH1F("f3dist","3-dist for all hits",20001,-0.5,20000.5);
      TH1F *f4dist = new TH1F("f4dist","4-dist for all hits",20001,-0.5,20000.5);
      TH1F *f5dist = new TH1F("f5dist","5-dist for all hits",20001,-0.5,20000.5);
      
      fPtMan.PreparekDistInfo(10);
      
      // std::cout << "\t(done)\n";
      
      for(unsigned int k = 2; k <= 5; ++k) {
	// std::cout << "Filling kDist plots for k = " << k << "...";
	double kdist[hitcol->size()];
	for(unsigned int i = 0; i < hitcol->size(); ++i) {
	  kdist[i] = fPtMan.kDist(k,i);
	} // end for i
	
	// sort kdist
	bool done = false;
	while(!done) {
	  done = true;
	  for(unsigned int j = 1; j < hitcol->size(); ++j) {  
	    if(kdist[j-1] < kdist[j]) {  
	      double temp = kdist[j];
	      kdist[j]    = kdist[j-1];
	      kdist[j-1]  = temp;	    
	      done = false;
	    } // end if
	  } // end for j
	} // end while !done
	
	if(k == 2) {
	  for(unsigned int i = 0; i < hitcol->size(); ++i) {
	    f2dist->Fill(i,kdist[i]);
	  } // end for i
	}
	if(k == 3) {
	  for(unsigned int i = 0; i < hitcol->size(); ++i) {
	    f3dist->Fill(i,kdist[i]);
	  } // end for i
	}     
	if(k == 4) {
	  for(unsigned int i = 0; i < hitcol->size(); ++i) {
	    f4dist->Fill(i,kdist[i]);
	  } // end for i
	} 
	if(k == 5) {
	  for(unsigned int i = 0; i < hitcol->size(); ++i) {
	    f5dist->Fill(i,kdist[i]);
	  } // end for i
	}
	// std::cout << "\t(done)\n";
      } // end for k
      
      art::ServiceHandle<art::TFileService> tfs;
      tfs->make<TH1F>(*f2dist);
      tfs->make<TH1F>(*f3dist);
      tfs->make<TH1F>(*f4dist);
      tfs->make<TH1F>(*f5dist);
      
      // std::cout << "\n\n====================\nFinished processing.\n\n";
      
      if(f2dist) { delete f2dist; f2dist = 0; }
      if(f3dist) { delete f3dist; f3dist = 0; }
      if(f4dist) { delete f4dist; f4dist = 0; }
      if(f5dist) { delete f5dist; f5dist = 0; }
      
    } // end if{fMakeKDistPlots)
    
    // std::cout << "\n\n";

    return;
  }

} // end namespace slicer
////////////////////////////////////////////////////////////////////////



#include "art/Framework/Core/ModuleMacros.h"

namespace slicer
{
  DEFINE_ART_MODULE(S4DParamCalc);
}
