////////////////////////////////////////////////////////////////////////
// \file    SlicerAna_module.cc
// \brief   Analysis module to validate slicer performance
// \version $Id: SlicerAna.cxx,v 1.4 2012-12-19 14:51:38 mbaird42 Exp $
// \author  $Author: mbaird42 $
////////////////////////////////////////////////////////////////////////

// C/C++ includes
#include <cmath>
#include <iostream>
#include <map>
#include <vector>

// ROOT includes
#include "TTree.h"
#include "TH1F.h"
#include "TH2F.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/FindManyP.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// NOvASoft includes
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "Utilities/AssociationUtil.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"

#include "Calibrator/Calibrator.h"
#include "MCCheater/BackTracker.h"





// SlicerAna header
namespace slicer {
  class SlicerAna : public art::EDAnalyzer {
  public:
    explicit SlicerAna(fhicl::ParameterSet const& pset);    
    ~SlicerAna();                        
    
    void analyze(const art::Event& evt); 
    void reconfigure(const fhicl::ParameterSet& pset);
    void beginJob();
    void endJob();

    std::vector<cheat::NeutrinoEffPur> sortEffPur(const std::vector<cheat::NeutrinoEffPur>& nuEP);
    
  private:
    void ClearVectors();

    std::string   fSlicerLabel;
    bool          fIsMC;
    
    TTree*        fOutTree;


    
    // Histograms created for production validation
    TH1F *fNumNu;
    TH1F *fNumMCT;
    TH1F *fNumSlices;
    TH2F *fNumMCTSlices;
    TH1F *fNumNoise;
    TH1F *fNumHits;
    TH1F *fNumHitsNoise;
    
    TH1F *fComp;
    TH1F *fPur;
    TH2F *fCompPur;
    TH1F *fCompNoise;
    TH1F *fPurNoise;

    TH1F *fNSvsVE;
    TH1F *fNGvsVE;
    TH1F *fEffVsVisE;

    
    TH2F *fPCXallhits;
    TH2F *fPCYallhits;
    TH1F *fTallhits;
    TH2F *fPCXave;
    TH2F *fPCYave;
    TH1F *fTave;
    TH2F *fPCXdiff;
    TH2F *fPCYdiff;
    TH1F *fTsd;
    TH1F *fSlDur;
    
    TH2F *fPCXallhitsNoise;
    TH2F *fPCYallhitsNoise;
    TH1F *fTallhitsNoise;
    TH2F *fPCXaveNoise;
    TH2F *fPCYaveNoise;
    TH1F *fTaveNoise;

    TH1F *fEvtNoiseFraction;

    TH1F *fTres;
    TH1F *fTresNoise;



    //
    // info to go into TTree
    //

    // event level info
    int run;               ///< run number
    int subrun;            ///< subrun number
    int event;             ///< event number

    int NumNoise;          ///< number of noise slices
    int NumSlice;          ///< number of non-noise slices

    int NumNuFLSHitALL;    ///< number of nu counted by FLSHits
    int NumNuCellHitALL;   ///< number of nu counted by CellHits
    int NumNuCellHit;      ///< number of nu in non-noise slices counted by CellHits

    int NumMCTFLSHitALL;   ///< total number of MCTruths counted by FLSHits
    int NumMCTCellHitALL;  ///< total number of MCTruths counted by CellHits
    int NumMCTCellHit;     ///< total number of MCTruths in non-noise slices counted by CellHits

    // slice level info
    std::vector<unsigned int> IsNoiseVec;  ///< 0 = non-noise slice, 1 = noise slice
    std::vector<int>          NumHitsXVec; ///< number of X hits per slice
    std::vector<int>          NumHitsYVec; ///< number of Y hits per slice
    std::vector<double>       slEVec;      ///< slice.TotalGeV()
    std::vector<float>        slPEVec;     ///< sum of PE for all hits in the slice

    // geometric info about all slices
    std::vector<double> PaveVec;           ///< average plane number for hits in a non-noise slice
    std::vector<double> TaveVec;           ///< average time for hits in a non-noise slice
    std::vector<double> CXaveVec;          ///< average cell number for X hits in a non-noise slice
    std::vector<double> CYaveVec;          ///< average cell number for Y hits in a non-noise slice

    std::vector<double> PsdVec;            ///< st. dev. of plane number for hits in a non-noise slice
    std::vector<double> TsdVec;            ///< st. dev. time for hits in a non-noise slice
    std::vector<double> CXsdVec;           ///< st. dev. cell number for X hits in a non-noise slice
    std::vector<double> CYsdVec;           ///< st. dev. cell number for Y hits in a non-noise slice

    std::vector<double> PhiVec;            ///< highest plane in the slice
    std::vector<double> PloVec;            ///< lowest plane in the slice
    std::vector<double> ThiVec;            ///< highest time in the slice
    std::vector<double> TloVec;            ///< lowest plane in the slice
    std::vector<double> CXhiVec;           ///< highest X cell in the slice
    std::vector<double> CXloVec;           ///< lowest X cell in the slice
    std::vector<double> CYhiVec;           ///< highest Y cell in the slice
    std::vector<double> CYloVec;           ///< lowest Y cell in the slice

    // info about slicer performance (matching to MCTruths...)
    std::vector<float>  nu1XVec;            ///< x location for primary nu associated with slice
    std::vector<float>  nu1YVec;            ///< y location for primary nu associated with slice
    std::vector<float>  nu1ZVec;            ///< z location for primary nu associated with slice
    std::vector<float>  nu1EVec;            ///< Energy for primary nu associated with slice

    std::vector<float>  nu2XVec;            ///< x location for second nu associated with slice
    std::vector<float>  nu2YVec;            ///< y location for second nu associated with slice
    std::vector<float>  nu2ZVec;            ///< z location for second nu associated with slice
    std::vector<float>  nu2EVec;            ///< Energy for second nu associated with slice

    std::vector<float>  nu3XVec;            ///< x location for third nu associated with slice
    std::vector<float>  nu3YVec;            ///< y location for third nu associated with slice
    std::vector<float>  nu3ZVec;            ///< z location for third nu associated with slice
    std::vector<float>  nu3EVec;            ///< Energy for third nu associated with slice

    std::vector<float>  nu4XVec;            ///< x location for fourth nu associated with slice
    std::vector<float>  nu4YVec;            ///< y location for fourth nu associated with slice
    std::vector<float>  nu4ZVec;            ///< z location for fourth nu associated with slice
    std::vector<float>  nu4EVec;            ///< Energy for fourth nu associated with slice

    std::vector<float>  nu5XVec;            ///< x location for fifth nu associated with slice
    std::vector<float>  nu5YVec;            ///< y location for fifth nu associated with slice
    std::vector<float>  nu5ZVec;            ///< z location for fifth nu associated with slice
    std::vector<float>  nu5EVec;            ///< Energy for fifth nu associated with slice

    std::vector<int>    nu1pdgVec;          ///< PDG code for primary nu associated with slice
    std::vector<int>    nu1CCNCVec;         ///< CCNC for primary nu associated with slice
    std::vector<int>    nu1IntTypeVec;      ///< Interaction type for primary neutrino associated with slice
    std::vector<double> nu1VisEslVec;       ///< visible energy for the best matched nu in the slice
    std::vector<double> nu1VisEtotVec;      ///< total visible energy for the best matched nu in the entire event

    std::vector<int>    nu2pdgVec;          ///< PDG code for second nu associated with slice
    std::vector<int>    nu2CCNCVec;         ///< CCNC for second nu associated with slice
    std::vector<int>    nu2IntTypeVec;      ///< Interaction type for second neutrino associated with slice
    std::vector<double> nu2VisEslVec;       ///< visible energy for the second best matched nu in the slice
    std::vector<double> nu2VisEtotVec;      ///< total visible energy for the second best matched nu in the entire event

    std::vector<int>    nu3pdgVec;          ///< PDG code for third nu associated with slice
    std::vector<int>    nu3CCNCVec;         ///< CCNC for third nu associated with slice
    std::vector<int>    nu3IntTypeVec;      ///< Interaction type for third neutrino associated with slice
    std::vector<double> nu3VisEslVec;       ///< visible energy for the third best matched nu in the slice
    std::vector<double> nu3VisEtotVec;      ///< total visible energy for the third best matched nu in the entire event

    std::vector<int>    nu4pdgVec;          ///< PDG code for fourth nu associated with slice
    std::vector<int>    nu4CCNCVec;         ///< CCNC for fourth nu associated with slice
    std::vector<int>    nu4IntTypeVec;      ///< Interaction type for fourth neutrino associated with slice
    std::vector<double> nu4VisEslVec;       ///< visible energy for the fourth best matched nu in the slice
    std::vector<double> nu4VisEtotVec;      ///< total visible energy for the fourth best matched nu in the entire event

    std::vector<int>    nu5pdgVec;          ///< PDG code for fifth nu associated with slice
    std::vector<int>    nu5CCNCVec;         ///< CCNC for fifth nu associated with slice
    std::vector<int>    nu5IntTypeVec;      ///< Interaction type for fifth neutrino associated with slice
    std::vector<double> nu5VisEslVec;       ///< visible energy for the fifth best matched nu in the slice
    std::vector<double> nu5VisEtotVec;      ///< total visible energy for the fifth best matched nu in the entire event

    std::vector<int>    pdgPartMostVec;    ///< the PDG code for the particle that contributed the most energy to this slice
    std::vector<int>    NumMatchesVec;     ///< number of MCTruth matches per slice

    std::vector<double> nu1EffVec;         ///< efficiency for primary nu associated with slice
    std::vector<double> nu1PurVec;         ///< purity for primary nu associated with slice 
    std::vector<double> nu2EffVec;         ///< efficiency for second nu associated with slice
    std::vector<double> nu2PurVec;         ///< purity for second nu associated with slice
    std::vector<double> nu3EffVec;         ///< efficiency for third nu associated with slice
    std::vector<double> nu3PurVec;         ///< purity for third nu associated with slice
    std::vector<double> nu4EffVec;         ///< efficiency for fourth nu associated with slice
    std::vector<double> nu4PurVec;         ///< purity for fourth nu associated with slice
    std::vector<double> nu5EffVec;         ///< efficiency for fifth nu associated with slice
    std::vector<double> nu5PurVec;         ///< purity for fifth nu associated with slice

    std::vector<unsigned int> IsMichelPrimaryVec; ///< is this slice a Michel e? (the primary contribution by energy was from a michel e)
    std::vector<unsigned int> IsMichelPresentVec; ///< did a Michel e contribute to this slice?
    std::vector<unsigned int> IsLateSliceVec;     ///< is the average time for this slice > 200 ns after the time of the neutrino interaction?

    std::vector<double> PercentNoiseVec;          ///< Percentage of slice that is true noise (by hit)
    std::vector<double> PercentNoiseByPEVec;      ///< Percentage of slice that is true noise (weighted by PE)

    // MCTruth level info

    // the main F.O.M. number
    unsigned int NgoodSL;

  };
}



namespace slicer
{
  //.......................................................................
  SlicerAna::SlicerAna(fhicl::ParameterSet const& pset) 
    : EDAnalyzer(pset)
  {
    this->reconfigure(pset);
  }

  //......................................................................
  SlicerAna::~SlicerAna()
  {
    //======================================================================
    // Clean up any memory allocated by your module
    //======================================================================
  }

  //......................................................................
  void SlicerAna::reconfigure(const fhicl::ParameterSet& pset)
  {
    fSlicerLabel = pset.get<std::string> ("SlicerLabel");
    fIsMC        = pset.get<bool>        ("IsMC");
  }
      
  //......................................................................
  void SlicerAna::beginJob()
  {

    NgoodSL = 0;

    art::ServiceHandle<art::TFileService> tfs;

    // define the TTree and branch variables
    fOutTree = tfs->make<TTree>("SlicerAna", "Slicer Analysis Tree");
    fOutTree->Branch("run", &run);
    fOutTree->Branch("subrun", &subrun);
    fOutTree->Branch("event", &event);
    
    fOutTree->Branch("NumNoise", &NumNoise);
    fOutTree->Branch("NumSlice", &NumSlice);
    fOutTree->Branch("NumNuFLSHitALL", &NumNuFLSHitALL);
    fOutTree->Branch("NumNuCellHitALL", &NumNuCellHitALL);
    fOutTree->Branch("NumNuCellHit", &NumNuCellHit);
    fOutTree->Branch("NumMCTFLSHitALL", &NumMCTFLSHitALL);
    fOutTree->Branch("NumMCTCellHitALL", &NumMCTCellHitALL);
    fOutTree->Branch("NumMCTCellHit", &NumMCTCellHit);

    fOutTree->Branch("IsNoise", &IsNoiseVec);
    fOutTree->Branch("NumHitsX", &NumHitsXVec);
    fOutTree->Branch("NumHitsY", &NumHitsYVec);
    fOutTree->Branch("slE", &slEVec);
    fOutTree->Branch("slPE", &slPEVec);

    fOutTree->Branch("Pave", &PaveVec);
    fOutTree->Branch("Tave", &TaveVec);
    fOutTree->Branch("CXave", &CXaveVec);
    fOutTree->Branch("CYave", &CYaveVec);    

    fOutTree->Branch("Psd", &PsdVec);
    fOutTree->Branch("Tsd", &TsdVec);
    fOutTree->Branch("CXsd", &CXsdVec);
    fOutTree->Branch("CYsd", &CYsdVec); 

    fOutTree->Branch("Phi", &PhiVec);
    fOutTree->Branch("Thi", &ThiVec);
    fOutTree->Branch("CXhi", &CXhiVec);
    fOutTree->Branch("CYhi", &CYhiVec); 

    fOutTree->Branch("Plo", &PloVec);
    fOutTree->Branch("Tlo", &TloVec);
    fOutTree->Branch("CXlo", &CXloVec);
    fOutTree->Branch("CYlo", &CYloVec);

    fOutTree->Branch("nu1X", &nu1XVec);
    fOutTree->Branch("nu1Y", &nu1YVec);
    fOutTree->Branch("nu1Z", &nu1ZVec);
    fOutTree->Branch("nu1E", &nu1EVec);
    fOutTree->Branch("nu2X", &nu2XVec);
    fOutTree->Branch("nu2Y", &nu2YVec);
    fOutTree->Branch("nu2Z", &nu2ZVec);
    fOutTree->Branch("nu2E", &nu2EVec);
    fOutTree->Branch("nu3X", &nu3XVec);
    fOutTree->Branch("nu3Y", &nu3YVec);
    fOutTree->Branch("nu3Z", &nu3ZVec);
    fOutTree->Branch("nu3E", &nu3EVec);
    fOutTree->Branch("nu4X", &nu4XVec);
    fOutTree->Branch("nu4Y", &nu4YVec);
    fOutTree->Branch("nu4Z", &nu4ZVec);
    fOutTree->Branch("nu4E", &nu4EVec);
    fOutTree->Branch("nu5X", &nu5XVec);
    fOutTree->Branch("nu5Y", &nu5YVec);
    fOutTree->Branch("nu5Z", &nu5ZVec);
    fOutTree->Branch("nu5E", &nu5EVec);

    fOutTree->Branch("nu1pdg", &nu1pdgVec);
    fOutTree->Branch("nu1CCNC", &nu1CCNCVec);
    fOutTree->Branch("nu1IntType", &nu1IntTypeVec);
    fOutTree->Branch("nu1VisEsl", &nu1VisEslVec);
    fOutTree->Branch("nu1VisEtot", &nu1VisEtotVec);
    fOutTree->Branch("nu2pdg", &nu2pdgVec);
    fOutTree->Branch("nu2CCNC", &nu2CCNCVec);
    fOutTree->Branch("nu2IntType", &nu2IntTypeVec);
    fOutTree->Branch("nu2VisEsl", &nu2VisEslVec);
    fOutTree->Branch("nu2VisEtot", &nu2VisEtotVec);
    fOutTree->Branch("nu3pdg", &nu3pdgVec);
    fOutTree->Branch("nu3CCNC", &nu3CCNCVec);
    fOutTree->Branch("nu3IntType", &nu3IntTypeVec);
    fOutTree->Branch("nu3VisEsl", &nu3VisEslVec);
    fOutTree->Branch("nu3VisEtot", &nu3VisEtotVec);
    fOutTree->Branch("nu4pdg", &nu4pdgVec);
    fOutTree->Branch("nu4CCNC", &nu4CCNCVec);
    fOutTree->Branch("nu4IntType", &nu4IntTypeVec);
    fOutTree->Branch("nu4VisEsl", &nu4VisEslVec);
    fOutTree->Branch("nu4VisEtot", &nu4VisEtotVec);
    fOutTree->Branch("nu5pdg", &nu5pdgVec);
    fOutTree->Branch("nu5CCNC", &nu5CCNCVec);
    fOutTree->Branch("nu5IntType", &nu5IntTypeVec);
    fOutTree->Branch("nu5VisEsl", &nu5VisEslVec);
    fOutTree->Branch("nu5VisEtot", &nu5VisEtotVec);

    fOutTree->Branch("pdgPartMost", &pdgPartMostVec);
    fOutTree->Branch("NumMatches", &NumMatchesVec);

    fOutTree->Branch("nu1Eff", &nu1EffVec);
    fOutTree->Branch("nu1Pur", &nu1PurVec);
    fOutTree->Branch("nu2Eff", &nu2EffVec);
    fOutTree->Branch("nu2Pur", &nu2PurVec);
    fOutTree->Branch("nu3Eff", &nu3EffVec);
    fOutTree->Branch("nu3Pur", &nu3PurVec);
    fOutTree->Branch("nu4Eff", &nu4EffVec);
    fOutTree->Branch("nu4Pur", &nu4PurVec);
    fOutTree->Branch("nu5Eff", &nu5EffVec);
    fOutTree->Branch("nu5Pur", &nu5PurVec);

    fOutTree->Branch("IsMichelPrimary", &IsMichelPrimaryVec);
    fOutTree->Branch("IsMichelPresent", &IsMichelPresentVec);
    fOutTree->Branch("IsLateSlice", &IsLateSliceVec);

    fOutTree->Branch("PercentNoise", &PercentNoiseVec);
    fOutTree->Branch("PercentNoiseByPE", &PercentNoiseByPEVec);



    // book the histograms
    fNumNu        = tfs->make<TH1F>("fNumNu",
				    "# of Nu;# of nu;count",
				    101,-0.5,100.5);
    fNumMCT       = tfs->make<TH1F>("fNumMCT",
				    "# of MCTruths;# of MCT;count",
				    201,-0.5,200.5);
    fNumSlices    = tfs->make<TH1F>("fNumSlices",
				    "# of non-noise slices;# of non-noise slices;count",
				    101,-0.5,100.5);
    fNumMCTSlices = tfs->make<TH2F>("fNumMCTSlices",
				    "# of MCTruths that contributed to non-noise slices vs. # of non-noise slices;# of MCT;# of slices",
				    101,-0.5,100.5,101,-0.5,100.5);
    fNumNoise     = tfs->make<TH1F>("fNumNoise",
				    "# of noise slices;# of noise slices;count",
				    21,-0.5,20.5);
    fNumHits      = tfs->make<TH1F>("fNumHits",
				    "# of hits per non-noise slice;# of hits;count",
				    500,0.0,5000.0);
    fNumHitsNoise = tfs->make<TH1F>("fNumHitsNoise",
				    "# of hits in the noise slice;# of hits;count",
				    500,0.0,5000.0);

    fComp = tfs->make<TH1F>("fComp",
			    "Completeness for most pure MCTruth in slice; completeness; count",
			    101,-0.005,1.005);
    fPur  = tfs->make<TH1F>("fPur",
			    "Purity for most pure MCTruth in slice; purity; count",
			    101,-0.005,1.005);
    fCompPur = tfs->make<TH2F>("fCompPur",
			       "Completeness/Purity for 'most pure' MCTruth by slice;Pur;Comp",
			       101,-0.005,1.005,101,-0.005,1.005);
    fCompNoise = tfs->make<TH1F>("fCompNoise",
				 "Completeness for all MCTruths in NOISE slice; completeness; count",
				 101,-0.005,1.005);
    fPurNoise  = tfs->make<TH1F>("fPurNoise",
				 "Purity for all MCTruths in NOISE slice; purity; count",
				 101,-0.005,1.005);

    fNSvsVE    = new TH1F("fNSvsVE","number of slices vs. VisE",80,0.0,20.0);
    fNGvsVE    = new TH1F("fNGvsVE","number of good slices vs. VisE",80,0.0,20.0);
    fNSvsVE->Reset();
    fNGvsVE->Reset();
    fEffVsVisE = tfs->make<TH1F>("fEffVsVisE",
				 "Efficiency vs. visible energy of best matched MCTruth for all non-noise slices;VisE [GeV];eff",
				 80,0.0,20.0);
    
    fPCXallhits  = tfs->make<TH2F>("fPCXallhits",
				   "Plane vs. X cell for all hits;plane;X cell",
				   1320,-20.0,1300.0,510,-10.0,500.0);
    fPCYallhits  = tfs->make<TH2F>("fPCYallhits",
				   "Plane vs. Y cell for all hits;plane;Y cell",
				   1320,-20.0,1300.0,510,-10.0,500.0);
    fTallhits    = tfs->make<TH1F>("fTallhits",
				   "Time for all hits; time [ns]; count",
				   12000,-50000.0,550000.0);
    fPCXave  = tfs->make<TH2F>("fPCXave",
			       "ave hit plane vs. ave hit X cell;plane;X cell",
			       660,-20.0,1300.0,255,-10.0,500.0);
    fPCYave  = tfs->make<TH2F>("fPCYave",
			       "ave hit plane vs. ave hit Y cell;plane;Y cell",
			       660,-20.0,1300.0,255,-10.0,500.0);
    fTave    = tfs->make<TH1F>("fTave",
			       "ave hit time per slice; time [ns]; count",
			       10000,0.0,500000.0);
    fPCXdiff = tfs->make<TH2F>("fPCXdiff",
			       "diff btwn hi and lo for plane vs X cell per slice;delta plane;delta X cell",
			       500,0.0,1000.0,250,0.0,500.0);
    fPCYdiff = tfs->make<TH2F>("fPCYdiff",
			       "diff btwn hi and lo for plane vs Y cell per slice;delta plane;delta Y cell",
			       500,0.0,1000.0,250,0.0,500.0);
    fTsd     = tfs->make<TH1F>("fTsd",
			       "st. dev. of  hit time per slice; sigma time [ns]; count",
			       500000,0.0,500000.0);
    fSlDur   = tfs->make<TH1F>("fSlDur",
			       "Slice Duration (last hit time - first hit time);deltaT [ns];count",
			       200,0.0,2000.0);

    fPCXallhitsNoise  = tfs->make<TH2F>("fPCXallhitsNoise",
					"Plane vs. X cell for all hits (noise slice);plane;X cell",
					1320,-20.0,1300.0,510,-10.0,500.0);
    fPCYallhitsNoise  = tfs->make<TH2F>("fPCYallhitsNoise",
					"Plane vs. Y cell for all hits (noise slice);plane;Y cell",
					1320,-20.0,1300.0,510,-10.0,500.0);
    fTallhitsNoise    = tfs->make<TH1F>("fTallhitsNoise",
					"Time for all hits (noise slice); time [ns]; count",
					12000,-50000.0,550000.0);
    fPCXaveNoise = tfs->make<TH2F>("fPCXaveNoise",
				   "ave hit plane vs. ave hit X cell (noise slice);plane;X cell",
				   660,-20.0,1300.0,255,-10.0,500.0);
    fPCYaveNoise = tfs->make<TH2F>("fPCYaveNoise",
				   "ave hit plane vs. ave hit Y cell (noise slice);plane;Y cell",
				   660,-20.0,1300.0,255,-10.0,500.0);
    fTaveNoise   = tfs->make<TH1F>("fTaveNoise",
				   "ave hit time per slice (noise slice); time [ns]; count",
				   10000,0.0,500000.0);

    fEvtNoiseFraction  = tfs->make<TH1F>("fEvtNoiseFraction",
					 "Fraction of all hits in the noise slice; noise fraction; count",
					 10000,0.0,1.0001);

    fTres      = tfs->make<TH1F>("fTres",
				 "Computed timing resolution for hits in non-noise slices;time [ns];count",
				 500,0.0,500.0);
    fTresNoise = tfs->make<TH1F>("fTresNoise",
				 "Computed timing resolution for hits in the noise slice;time [ns];count",
				 500,0.0,500.0);
    
  }

  //......................................................................
  void SlicerAna::endJob()
  {
    if(fIsMC) {
      std::cout << "FOM: " << NgoodSL
		<< "\n========== End Job ==========\n\n";

      for(int i = 1; i <= fNSvsVE->GetNbinsX(); ++i) {
	double ratio;
	if(fNSvsVE->GetBinContent(i) != 0.0) {
	  ratio = (fNGvsVE->GetBinContent(i))/(fNSvsVE->GetBinContent(i));
	  fEffVsVisE->Fill(fNSvsVE->GetBinCenter(i),ratio);
	}
      }

    }
  }

  //......................................................................
  void SlicerAna::analyze(const art::Event& evt)
  {

    this->ClearVectors();
    int NSlice    = 0;
    int NhitNoise = 0;
    int NhitTotal = 0;

    // get the MCTruth info
    art::Handle< std::vector< simb::MCTruth > > mclist;
    if(fIsMC) evt.getByLabel("generator", mclist);

    // get the slices
    art::Handle< std::vector< rb::Cluster > > slices;
    evt.getByLabel(fSlicerLabel, slices);

    // make the needed services
    art::ServiceHandle<cheat::BackTracker> BT;
    art::ServiceHandle<calib::Calibrator>  cal;

    art::PtrVector<rb::Cluster> ARTslices;
    for(unsigned int i = 0; i < slices->size(); ++i)
      ARTslices.push_back(art::Ptr<rb::Cluster>(slices,i));

    std::vector< std::vector< cheat::NeutrinoEffPur > > ALLnuEffPur;
    if(fIsMC)    
      ALLnuEffPur = BT->SlicesToMCTruthsTable(ARTslices);

    // sets to record info about neutrinos/MCTruths that interact in the detector
    std::set< art::Ptr< simb::MCTruth > > NuListFLSHitALL;
    std::set< art::Ptr< simb::MCTruth > > NuListCellHitALL;
    std::set< art::Ptr< simb::MCTruth > > NuListCellHit;

    std::set< art::Ptr< simb::MCTruth > > MCTListFLSHitALL;
    std::set< art::Ptr< simb::MCTruth > > MCTListCellHitALL;
    std::set< art::Ptr< simb::MCTruth > > MCTListCellHit;



    run    = evt.run();
    subrun = evt.subRun();
    event  = evt.id().event();



    //
    // Loop over all slices and gather info for each slice...
    //

    // std::cout << "\n\n" << evt.id() << "\nsize of MClist = " << mclist->size() << "\n\n";

    for(unsigned int i = 0; i < slices->size(); ++i) {
     
      // define default values for the slice info
      unsigned int IsNoise  = 0;
      int          NumHitsX = -1;
      int          NumHitsY = -1;
      double       slE      = 0.0;
      float        slPE     = 0.0;

      double Pave  = 0.0;
      double CXave = 0.0;
      double CYave = 0.0;
      double Tave  = 0.0;

      double Psd   = 0.0;
      double CXsd  = 0.0;
      double CYsd  = 0.0;
      double Tsd   = 0.0;

      double Phi  = -1.0e6;
      double Plo  = 1.0e6;
      double Thi  = -1.0e6;
      double Tlo  = 1.0e6;
      double CXhi = -1.0e6;
      double CXlo = 1.0e6;
      double CYhi = -1.0e6;
      double CYlo = 1.0e6;

      float  nu1X = 0.0, nu1Y = 0.0, nu1Z = 0.0, nu1E = -1.0;
      int    nu1pdg = 0, nu1CCNC = -1, nu1IntType = 0;
      double nu1VisEsl = -1.0, nu1VisEtot = -1.0;

      float  nu2X = 0.0, nu2Y = 0.0, nu2Z = 0.0, nu2E = -1.0;
      int    nu2pdg = 0, nu2CCNC = -1, nu2IntType = 0;
      double nu2VisEsl = -1.0, nu2VisEtot = -1.0;

      float  nu3X = 0.0, nu3Y = 0.0, nu3Z = 0.0, nu3E = -1.0;
      int    nu3pdg = 0, nu3CCNC = -1, nu3IntType = 0;
      double nu3VisEsl = -1.0, nu3VisEtot = -1.0;

      float  nu4X = 0.0, nu4Y = 0.0, nu4Z = 0.0, nu4E = -1.0;
      int    nu4pdg = 0, nu4CCNC = -1, nu4IntType = 0;
      double nu4VisEsl = -1.0, nu4VisEtot = -1.0;

      float  nu5X = 0.0, nu5Y = 0.0, nu5Z = 0.0, nu5E = -1.0;
      int    nu5pdg = 0, nu5CCNC = -1, nu5IntType = 0;
      double nu5VisEsl = -1.0, nu5VisEtot = -1.0;

      int    pdgPartMost = 0, NumMatches = 0;
      
      double nu1Eff = 0.0, nu1Pur = 0.0;
      double nu2Eff = 0.0, nu2Pur = 0.0;
      double nu3Eff = 0.0, nu3Pur = 0.0;
      double nu4Eff = 0.0, nu4Pur = 0.0;
      double nu5Eff = 0.0, nu5Pur = 0.0;

      unsigned int IsMichelPrimary = 0;
      unsigned int IsMichelPresent = 0;
      unsigned int IsLateSlice     = 0;

      double PercentNoise     = -1.0;
      double PercentNoiseByPE = -1.0;
      

      
      if((*slices)[i].IsNoise()) IsNoise = 1;
      NumHitsX = (*slices)[i].NXCell();
      NumHitsY = (*slices)[i].NYCell();
      // slE      = (*slices)[i].TotalGeV();

      // count the non-noise slices (to verify that there is only 1 noise slice)
      if(!(*slices)[i].IsNoise()) NSlice++;



      // Calculate average and st. dev. for plane #, cell #, and time for
      // hits in the slice

      double nx = 0.0, ny = 0.0;

      // loop over all hits
      for(unsigned int j = 0; j < (*slices)[i].NCell(); ++j) {

	const rb::CellHit* h = (*slices)[i].Cell(j).get();

	// Fill hit level histos
	if((*slices)[i].IsNoise()) {
	  fTresNoise->Fill(cal->GetTimeRes(*h));
	  fTallhitsNoise->Fill(h->TNS());
	  if(h->View() == geo::kX) fPCXallhitsNoise->Fill(h->Plane(),h->Cell());
	  if(h->View() == geo::kY) fPCYallhitsNoise->Fill(h->Plane(),h->Cell());
	}
	else {
	  fTres->Fill(cal->GetTimeRes(*h));
	  fTallhits->Fill(h->TNS());
	  if(h->View() == geo::kX) fPCXallhits->Fill(h->Plane(),h->Cell());
	  if(h->View() == geo::kY) fPCYallhits->Fill(h->Plane(),h->Cell());
	}

	// Problems occur if you do the time calculations in ns since the numbers
	// are huge when you square them (even if you use doubles.)  So I will
	// convert to microseconds to do my sums for Tave and Tsd.
	double Tus = (h->TNS())/1000.0;

	slPE += h->PE();

	Pave += h->Plane();
	Tave += Tus;
	Psd  += (h->Plane())*(h->Plane());
	Tsd  += Tus*Tus;

	if(h->Plane() > Phi) Phi = h->Plane();
	if(h->Plane() < Plo) Plo = h->Plane();
	if(h->TNS() > Thi) Thi = h->TNS();
	if(h->TNS() < Tlo) Tlo = h->TNS();

	if(h->View() == geo::kX) {
	  CXave += h->Cell();
	  CXsd  += (h->Cell())*(h->Cell());
	  nx += 1.0;
	  if(h->Cell() > CXhi) CXhi = h->Cell();
	  if(h->Cell() < CXlo) CXlo = h->Cell();
	}
	else if(h->View() == geo::kY) {
	  CYave += h->Cell();
	  CYsd  += (h->Cell())*(h->Cell());
	  ny += 1.0;
	  if(h->Cell() > CYhi) CYhi = h->Cell();
	  if(h->Cell() < CYlo) CYlo = h->Cell();
	} 
	else {
	  // Trigger some kind of warning that an invalid view occured.
	  CXave = 1.0e9;
	  CYave = 1.0e9;
	}
      } // end loop over j

      if(nx == 0.0 && ny == 0.0) {
	// assign a negative number to attract attention
	Pave = -100.0;
	Tave = -100.0;
	Psd  = -100.0;
	Tsd  = -100.0;
      }
      else {
	Pave = Pave/(nx+ny);
	Tave = Tave/(nx+ny);
	Psd  = sqrt(Psd/(nx+ny) - Pave*Pave);
	Tsd  = sqrt(Tsd/(nx+ny) - Tave*Tave);

	// convert back to ns
	Tave = Tave*1000.0;
	Tsd  = Tsd*1000.0;
      }

      if(nx != 0.0) {
	CXave = CXave/nx;
	CXsd  = sqrt(CXsd/nx - CXave*CXave);
      }
      else {
	CXave = -100.0;
	CXsd  = -100.0;
      }

      if(ny != 0.0) {
	CYave = CYave/ny;
	CYsd  = sqrt(CYsd/ny - CYave*CYave);
      }
      else {
	CYave = -100.0;    
	CYsd  = -100.0;
      }

      
      
      // additional analysis for comparison to MCTruth
      if(fIsMC) {

	// Get the list for this slice, keep only the neutrinos that contributed,
	// and sort the remaining list by purity.
	std::vector<cheat::NeutrinoEffPur> nuEffPur;
	nuEffPur = this->sortEffPur(ALLnuEffPur[i]);

	if(nuEffPur.size() > 0) {

	  // if this is a noise slice, count all Eff and Pur
	  if((*slices)[i].IsNoise()) {
	    for(unsigned int a = 0; a < nuEffPur.size(); ++a) {
	      fCompNoise->Fill(nuEffPur[a].efficiency);
	      fPurNoise ->Fill(nuEffPur[a].purity);
	    }
	  }
	  
	  simb::MCNeutrino nu = nuEffPur[0].neutrinoInt->GetNeutrino();
	  simb::MCParticle Nu = nu.Nu();
	
	  /*
	  std::cout << i << ":\tsize of nuEffPur = " << nuEffPur.size() << "\t"
		    << "Nu pdg = " << Nu.PdgCode() << "\n";
	  */

	  // count neutrinos/MCTruths that contributed cell hits
	  // keep track of the nu that contribute to ALL slices
	  for(unsigned int a = 0; a < nuEffPur.size(); ++a) {
	    MCTListCellHitALL.insert(nuEffPur[a].neutrinoInt);
	    if(nuEffPur[a].neutrinoInt->NeutrinoSet())
	      NuListCellHitALL.insert(nuEffPur[a].neutrinoInt);
	  }
	  
	  // keep track of only the nu that contribute to the non-noise slices
	  if( !((*slices)[i].IsNoise()) ) {
	    for(unsigned int a = 0; a < nuEffPur.size(); ++a) {
	      MCTListCellHit.insert(nuEffPur[a].neutrinoInt);
	      if(nuEffPur[a].neutrinoInt->NeutrinoSet())
		  NuListCellHit.insert(nuEffPur[a].neutrinoInt);
	    }
	  }
	  


	  // BEWARE OF SEGFAULTS!!!  If this MCTruth did NOT come from a neutrino,
	  // then we must skip the next section to avoid segfaluts.
	  if(nuEffPur[0].neutrinoInt->NeutrinoSet()) {

	    nu1X = Nu.Vx();
	    nu1Y = Nu.Vy();
	    nu1Z = Nu.Vz();
	    nu1E = Nu.E();
	    nu1pdg = Nu.PdgCode();
	    
	    nu1CCNC = nu.CCNC();
	    nu1IntType = nu.InteractionType();
	    
	  } // end if this MCTruth = neutrino

	  nu1Eff = nuEffPur[0].efficiency;
	  nu1Pur = nuEffPur[0].purity;
	  
	  nu1VisEsl  = nuEffPur[0].energySlice;
	  nu1VisEtot = nuEffPur[0].energyTotal;
	  
	  if(nuEffPur.size() > 1) {
	    if(nuEffPur[1].neutrinoInt->NeutrinoSet()) {
	      
	      nu = nuEffPur[1].neutrinoInt->GetNeutrino();
	      Nu = nu.Nu();

	      nu2X = Nu.Vx();
	      nu2Y = Nu.Vy();
	      nu2Z = Nu.Vz();
	      nu2E = Nu.E();
	      nu2pdg = Nu.PdgCode();
	      
	      nu2CCNC = nu.CCNC();
	      nu2IntType = nu.InteractionType();
	      
	    } // end if this MCTruth = neutrino

	    nu2Eff = nuEffPur[1].efficiency;
	    nu2Pur = nuEffPur[1].purity;
	    nu2VisEsl  = nuEffPur[1].energySlice;
	    nu2VisEtot = nuEffPur[1].energyTotal;
	  }
	  if(nuEffPur.size() > 2) {
	    if(nuEffPur[2].neutrinoInt->NeutrinoSet()) {
	      
	      nu = nuEffPur[2].neutrinoInt->GetNeutrino();
	      Nu = nu.Nu();

	      nu3X = Nu.Vx();
	      nu3Y = Nu.Vy();
	      nu3Z = Nu.Vz();
	      nu3E = Nu.E();
	      nu3pdg = Nu.PdgCode();
	      
	      nu3CCNC = nu.CCNC();
	      nu3IntType = nu.InteractionType();
	      
	    } // end if this MCTruth = neutrino

	    nu3Eff = nuEffPur[2].efficiency;
	    nu3Pur = nuEffPur[2].purity;
	    nu3VisEsl  = nuEffPur[2].energySlice;
	    nu3VisEtot = nuEffPur[2].energyTotal;
	  }
	  if(nuEffPur.size() > 3) {
	    if(nuEffPur[3].neutrinoInt->NeutrinoSet()) {
	      
	      nu = nuEffPur[3].neutrinoInt->GetNeutrino();
	      Nu = nu.Nu();

	      nu4X = Nu.Vx();
	      nu4Y = Nu.Vy();
	      nu4Z = Nu.Vz();
	      nu4E = Nu.E();
	      nu4pdg = Nu.PdgCode();
	      
	      nu4CCNC = nu.CCNC();
	      nu4IntType = nu.InteractionType();
	      
	    } // end if this MCTruth = neutrino
	    nu4Eff = nuEffPur[3].efficiency;
	    nu4Pur = nuEffPur[3].purity;
	    nu4VisEsl  = nuEffPur[3].energySlice;
	    nu4VisEtot = nuEffPur[3].energyTotal;
	  }
	  if(nuEffPur.size() > 4) {
	    if(nuEffPur[4].neutrinoInt->NeutrinoSet()) {
	      
	      nu = nuEffPur[4].neutrinoInt->GetNeutrino();
	      Nu = nu.Nu();

	      nu5X = Nu.Vx();
	      nu5Y = Nu.Vy();
	      nu5Z = Nu.Vz();
	      nu5E = Nu.E();
	      nu5pdg = Nu.PdgCode();
	      
	      nu5CCNC = nu.CCNC();
	      nu5IntType = nu.InteractionType();
	      
	    } // end if this MCTruth = neutrino
	    nu5Eff = nuEffPur[4].efficiency;
	    nu5Pur = nuEffPur[4].purity;
	    nu5VisEsl  = nuEffPur[4].energySlice;
	    nu5VisEtot = nuEffPur[4].energyTotal;
	  }
	} // end if nuEffPur.size() > 0

	NumMatches = nuEffPur.size();

	// get info about primary particle contributing to the slice
	const std::vector< const sim::Particle * > particles = BT->HitsToParticle((*slices)[i].AllCells());
	
	if(particles.size() > 0) {
	  pdgPartMost = particles[0]->PdgCode();

	  // decide if michel e is present or primary in this slice
	  if(pdgPartMost == 11 && particles[0]->Process() == "Decay") IsMichelPrimary = 1;
	  
	  for(unsigned int a = 0; a < particles.size(); ++a) {
	    if(particles[a]->PdgCode() == 11 && particles[a]->Process() == "Decay") IsMichelPresent = 1;
	  }

	} // end if(particles.size() > 0)

	// decide if this is a "late" slice
	if(nuEffPur.size() > 0) {
	  if(nuEffPur[0].neutrinoInt->NParticles() > 0) {
	    const simb::MCParticle p = nuEffPur[0].neutrinoInt->GetParticle(0);
	    double T = p.T();
	    if( (T-Tave) > 200.0) IsLateSlice = 1;
	  }
	}

	if(NumHitsX >= 3 && NumHitsY >= 3 &&
	   nu1Eff >= 0.9 && nu1Pur >= 0.9) NgoodSL++;

	// count the percentage of this slice that is true noise
	double Nhit    = 0.0;
	double NhitPE  = 0.0;
	double NNhit   = 0.0;
	double NNhitPE = 0.0;

	for(unsigned int hi = 0; hi < (*slices)[i].NCell(); ++hi) {

	  art::Ptr<rb::CellHit> hit = (*slices)[i].Cell(hi);
	
	  Nhit   += 1.0;
	  NhitPE += hit->PE();

	  if(BT->IsNoise(hit)) {
	    NNhit   += 1.0;
	    NNhitPE += hit->PE();
	  }

	} // end loop over hits in this slice

	PercentNoise     = NNhit/Nhit;
	PercentNoiseByPE = NNhitPE/NhitPE;

      } // end if(IsMC)
      

      
      
      
      
      // fill slice vectors with values
      
      IsNoiseVec .push_back(IsNoise);
      NumHitsXVec.push_back(NumHitsX);
      NumHitsYVec.push_back(NumHitsY);
      slEVec     .push_back(slE);
      slPEVec    .push_back(slPE);
      
      PaveVec .push_back(Pave);
      TaveVec .push_back(Tave);
      CXaveVec.push_back(CXave);
      CYaveVec.push_back(CYave);
      
      PsdVec .push_back(Psd);
      TsdVec .push_back(Tsd);
      CXsdVec.push_back(CXsd);
      CYsdVec.push_back(CYsd);
      
      PhiVec .push_back(Phi);
      ThiVec .push_back(Thi);
      CXhiVec.push_back(CXhi);
      CYhiVec.push_back(CYhi);
      PloVec .push_back(Plo);
      TloVec .push_back(Tlo);
      CXloVec.push_back(CXlo);
      CYloVec.push_back(CYlo);
      
      nu1XVec.push_back(nu1X);
      nu1YVec.push_back(nu1Y);
      nu1ZVec.push_back(nu1Z);
      nu1EVec.push_back(nu1E);
      nu2XVec.push_back(nu2X);
      nu2YVec.push_back(nu2Y);
      nu2ZVec.push_back(nu2Z);
      nu2EVec.push_back(nu2E);
      nu3XVec.push_back(nu3X);
      nu3YVec.push_back(nu3Y);
      nu3ZVec.push_back(nu3Z);
      nu3EVec.push_back(nu3E);
      nu4XVec.push_back(nu4X);
      nu4YVec.push_back(nu4Y);
      nu4ZVec.push_back(nu4Z);
      nu4EVec.push_back(nu4E);
      nu5XVec.push_back(nu5X);
      nu5YVec.push_back(nu5Y);
      nu5ZVec.push_back(nu5Z);
      nu5EVec.push_back(nu5E);

      nu1pdgVec      .push_back(nu1pdg);
      nu1CCNCVec     .push_back(nu1CCNC);
      nu1IntTypeVec  .push_back(nu1IntType);
      nu1VisEslVec   .push_back(nu1VisEsl);
      nu1VisEtotVec  .push_back(nu1VisEtot);
      nu2pdgVec      .push_back(nu2pdg);
      nu2CCNCVec     .push_back(nu2CCNC);
      nu2IntTypeVec  .push_back(nu2IntType);
      nu2VisEslVec   .push_back(nu2VisEsl);
      nu2VisEtotVec  .push_back(nu2VisEtot);
      nu3pdgVec      .push_back(nu3pdg);
      nu3CCNCVec     .push_back(nu3CCNC);
      nu3IntTypeVec  .push_back(nu3IntType);
      nu3VisEslVec   .push_back(nu3VisEsl);
      nu3VisEtotVec  .push_back(nu3VisEtot);
      nu4pdgVec      .push_back(nu4pdg);
      nu4CCNCVec     .push_back(nu4CCNC);
      nu4IntTypeVec  .push_back(nu4IntType);
      nu4VisEslVec   .push_back(nu4VisEsl);
      nu4VisEtotVec  .push_back(nu4VisEtot);
      nu5pdgVec      .push_back(nu5pdg);
      nu5CCNCVec     .push_back(nu5CCNC);
      nu5IntTypeVec  .push_back(nu5IntType);
      nu5VisEslVec   .push_back(nu5VisEsl);
      nu5VisEtotVec  .push_back(nu5VisEtot);

      pdgPartMostVec.push_back(pdgPartMost);
      NumMatchesVec .push_back(NumMatches);
      
      nu1EffVec.push_back(nu1Eff);
      nu1PurVec.push_back(nu1Pur);
      nu2EffVec.push_back(nu2Eff);
      nu2PurVec.push_back(nu2Pur);
      nu3EffVec.push_back(nu3Eff);
      nu3PurVec.push_back(nu3Pur);
      nu4EffVec.push_back(nu4Eff);
      nu4PurVec.push_back(nu4Pur);
      nu5EffVec.push_back(nu5Eff);
      nu5PurVec.push_back(nu5Pur);

      IsMichelPrimaryVec.push_back(IsMichelPrimary);
      IsMichelPresentVec.push_back(IsMichelPresent);
      IsLateSliceVec    .push_back(IsLateSlice);

      PercentNoiseVec    .push_back(PercentNoise);
      PercentNoiseByPEVec.push_back(PercentNoiseByPE);

      // fill histos
      if((*slices)[i].IsNoise()) {

	if(NumHitsX+NumHitsY == 0.0)
	  std::cout << "\n\nEmpty Noise Slice!   " << evt.id() << "\n\n\n";
	fNumHitsNoise->Fill(NumHitsX+NumHitsY);
	fPCXaveNoise ->Fill(Pave,CXave);
	fPCYaveNoise ->Fill(Pave,CYave);
	fTaveNoise   ->Fill(Tave);

	NhitNoise += NumHitsX+NumHitsY;
	NhitTotal += NumHitsX+NumHitsY;

      }
      else {

	fNumHits->Fill(NumHitsX+NumHitsY);
	fComp   ->Fill(nu1Eff);
	fPur    ->Fill(nu1Pur);
	fCompPur->Fill(nu1Pur,nu1Eff);
	fPCXave ->Fill(Pave,CXave);
	fPCYave ->Fill(Pave,CYave);
	fPCXdiff->Fill(Phi-Plo,CXhi-CXlo);
	fPCYdiff->Fill(Phi-Plo,CYhi-CYlo);
	fTave   ->Fill(Tave);       
	fTsd    ->Fill(Tsd);
	fSlDur  ->Fill(Thi-Tlo);

	NhitTotal += NumHitsX+NumHitsY;
	
	/*
	if(Thi-Tlo >= 1000.0 && Thi-Tlo <= 1070.0)
	  std::cout << "\nevent = " << evt.id() << "   slice = " << i << "   time = " << Thi-Tlo;
	*/

	if(fIsMC) {
	  fNSvsVE->Fill(nu1VisEtot);
	  if(nu1Eff >= 0.9 && nu1Pur >= 0.9) fNGvsVE->Fill(nu1VisEtot);
	}

      }

    } // end loop over slices (i)



    // accumulate the set of MCTruths that contributed FLSHits
    if(fIsMC) {

      // Get the particle list vector
      art::Handle< std::vector<sim::Particle> > parts;
      evt.getByLabel("geantgen",parts);

      // Retrive the associated MCTruth(s)
      art::FindManyP<simb::MCTruth> truthAssn(parts,evt,"geantgen");

      // Loop through MCTruth object to accumulate list of keepers
      for(size_t p = 0; p < parts->size(); ++p) {
	for(size_t t = 0; t < truthAssn.at(p).size(); ++t) {
	  art::Ptr<simb::MCTruth> mcTru = truthAssn.at(p)[t];
	  MCTListFLSHitALL.insert(mcTru);
	  if(mcTru->NeutrinoSet())
	    NuListFLSHitALL.insert(mcTru);
	}
      }

    }



    /// TODO: Add vectors of info about each neutrino?



    fNumNu->Fill(NuListFLSHitALL.size());
    fNumMCT->Fill(MCTListFLSHitALL.size());

    NumSlice = NSlice;
    fNumSlices->Fill(NSlice);

    fNumMCTSlices->Fill(MCTListCellHit.size(),NSlice);

    NumNoise = slices->size() - NSlice;
    fNumNoise->Fill(slices->size() - NSlice);

    NumNuFLSHitALL  = NuListFLSHitALL.size();
    NumNuCellHitALL = NuListCellHitALL.size();
    NumNuCellHit    = NuListCellHit.size();

    NumMCTFLSHitALL  = MCTListFLSHitALL.size();
    NumMCTCellHitALL = MCTListCellHitALL.size();
    NumMCTCellHit    = MCTListCellHit.size();

    fEvtNoiseFraction->Fill((double)NhitNoise/(double)NhitTotal);

    fOutTree->Fill();

    /*
    std::cout << "\n\nFSL = " << NuListFLSHitALL.size() << "\t"
	      << "cellALL = " << NuListCellHitALL.size() << "\t"
	      << "cell = " << NuListCellHit.size() << "\n";
    std::cout << "FSL = " << MCTListFLSHitALL.size() << "\t"
	      << "cellALL = " << MCTListCellHitALL.size() << "\t"
	      << "cell = " << MCTListCellHit.size() << "\n\n";
    */

    return;
  }

  //----------------------------------------------------------------------
  std::vector<cheat::NeutrinoEffPur> SlicerAna::sortEffPur(const std::vector<cheat::NeutrinoEffPur>& nuEP)
  {
    std::vector<cheat::NeutrinoEffPur> results;

    // only push in stuff with non-zero purity
    for(unsigned int i = 0; i < nuEP.size(); ++i) {
      if(nuEP[i].purity > 0.0) results.push_back(nuEP[i]);
    }

    // sort by purity (bubble sort)

    if(results.size() > 0) {
      bool done = false;
      while(!done) {
	done = true;
	for(unsigned int i = 1; i < results.size(); ++i) {
	  if(results[i-1].purity < results[i].purity) {
	    cheat::NeutrinoEffPur temp = results[i];
	    results[i]   = results[i-1];
	    results[i-1] = temp;
	    done = false;
	  } // end if
	} // end for i
      } // end while !done
    } // end if(results.size() > 0)

    return results;

  }

  //----------------------------------------------------------------------
  void SlicerAna::ClearVectors()
  {
    IsNoiseVec.clear();
    NumHitsXVec.clear();
    NumHitsYVec.clear();
    slEVec.clear();
    slPEVec.clear();

    PaveVec.clear();
    TaveVec.clear();
    CXaveVec.clear();
    CYaveVec.clear();

    PsdVec.clear();
    TsdVec.clear();
    CXsdVec.clear();
    CYsdVec.clear();

    PhiVec.clear();
    PloVec.clear();
    ThiVec.clear();
    TloVec.clear();
    CXhiVec.clear();
    CXloVec.clear();
    CYhiVec.clear();
    CYloVec.clear();

    nu1XVec.clear();
    nu1YVec.clear();
    nu1ZVec.clear();
    nu1EVec.clear();
    nu2XVec.clear();
    nu2YVec.clear();
    nu2ZVec.clear();
    nu2EVec.clear();
    nu3XVec.clear();
    nu3YVec.clear();
    nu3ZVec.clear();
    nu3EVec.clear();
    nu4XVec.clear();
    nu4YVec.clear();
    nu4ZVec.clear();
    nu4EVec.clear();
    nu5XVec.clear();
    nu5YVec.clear();
    nu5ZVec.clear();
    nu5EVec.clear();

    nu1pdgVec.clear();
    nu1CCNCVec.clear();
    nu1IntTypeVec.clear();
    nu1VisEslVec.clear();
    nu1VisEtotVec.clear();
    nu2pdgVec.clear();
    nu2CCNCVec.clear();
    nu2IntTypeVec.clear();
    nu2VisEslVec.clear();
    nu2VisEtotVec.clear();
    nu3pdgVec.clear();
    nu3CCNCVec.clear();
    nu3IntTypeVec.clear();
    nu3VisEslVec.clear();
    nu3VisEtotVec.clear();
    nu4pdgVec.clear();
    nu4CCNCVec.clear();
    nu4IntTypeVec.clear();
    nu4VisEslVec.clear();
    nu4VisEtotVec.clear();
    nu5pdgVec.clear();
    nu5CCNCVec.clear();
    nu5IntTypeVec.clear();
    nu5VisEslVec.clear();
    nu5VisEtotVec.clear();

    pdgPartMostVec.clear();
    NumMatchesVec.clear();

    nu1EffVec.clear();
    nu1PurVec.clear(); 
    nu2EffVec.clear();
    nu2PurVec.clear();
    nu3EffVec.clear();
    nu3PurVec.clear();
    nu4EffVec.clear();
    nu4PurVec.clear();
    nu5EffVec.clear();
    nu5PurVec.clear();

    IsMichelPrimaryVec.clear();
    IsMichelPresentVec.clear();
    IsLateSliceVec.clear();

    PercentNoiseVec.clear();
    PercentNoiseByPEVec.clear();

  }

} // end namespace slicer
////////////////////////////////////////////////////////////////////////



#include "art/Framework/Core/ModuleMacros.h"

namespace slicer
{
  DEFINE_ART_MODULE(SlicerAna);
}
