///////////////////////////////////////////////////////////////////////////
// \file    TruthSlicer.cxx
// \brief   This is a slicer that will separate things into slices
//          according to MCTruth.  It is NOT part of the other slicers so
//          that they do not have to depend on backtracker.
// \version $Id: TruthSlicer.cxx,v 1.0 2013-05-05 17:00:00 mbaird42 Exp $
// \authors mbaird42@fnal.gov
///////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
// #include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"

#include "CMap/CMap.h"
#include "Geometry/Geometry.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"

#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "SimulationBase/MCTruth.h"
#include "MCCheater/BackTracker.h"

#include <cmath>
#include <string>
#include <vector>


// TruthSlicer header
namespace slicer {
  
  class TruthSlicer : public art::EDProducer {
  public:
    explicit TruthSlicer(fhicl::ParameterSet const &pset);
    virtual ~TruthSlicer();

    void produce(art::Event& evt);
    
    // configuration settings
    std::string  fCellHitInput;   ///< Read CellHits from Cal/[fCellHitFolderIn]
    std::string  fMCTruthInput;   ///< Read MCTruth from this label
    bool         fMerge;          ///< Try to merge small ( < fMinHit) slices into bigger ones
    unsigned int fMinHits;        ///< Minimum number of hits for a "non-small" slice
  };

}





namespace slicer
{
  //----------------------------------------------------------------------
  TruthSlicer::TruthSlicer(fhicl::ParameterSet const &pset) :
    fCellHitInput (pset.get< std::string  >("CellHitInput")    ),
    fMCTruthInput (pset.get< std::string  >("MCTruthInput")    ),
    fMerge        (pset.get< bool         >("MergeSmallSlices")),
    fMinHits      (pset.get< unsigned int >("MinHits")         )
  {
    produces< std::vector<rb::Cluster> >();
  }

  //----------------------------------------------------------------------
  TruthSlicer::~TruthSlicer()
  { }

  //----------------------------------------------------------------------
  void TruthSlicer::produce(art::Event& evt)
  {
    // Load hit list (CellHits)
    art::Handle< std::vector<rb::CellHit> > hitcol;
    evt.getByLabel(fCellHitInput, hitcol);

    std::vector<art::Ptr<rb::CellHit > > hitlist;
    for(unsigned int i = 0; i < hitcol->size(); ++i){
      art::Ptr<rb::CellHit> hit(hitcol, i);
      hitlist.push_back(hit);
    }

    // Load the MCTruth list
    art::Handle< std::vector< simb::MCTruth > > mclist;
    evt.getByLabel(fMCTruthInput, mclist);

    // Make the backtracker.
    art::ServiceHandle<cheat::BackTracker> BT;

    // make the final slice container that will be put into the event
    std::unique_ptr< std::vector<rb::Cluster> >
      clustercol(new std::vector<rb::Cluster>);


    
    //
    // Do the truth slicing...
    //
    std::vector<rb::Cluster> clusters = BT->ClusterByTruth(hitlist);



    ///\ TODO: fill this is...
    if(fMerge) {
      // TruthSlicer will make a separate slice for each neutrino
      // that leaves hits in the detector.  This will leave many small
      // slices that in "reality" would have been either swallowed up
      // by some larger cluster or labeled as noise.
      //
      // This section attempts to merge the smaller clusters into
      // the larger ones trying to make a more realistic picture of what
      // the "ideal" slicing case might be.


    } // end if fMerge


    
    // add the clusters to clustercol
    for(unsigned int i = 0; i < clusters.size(); ++i) {
      if(clusters[i].NCell() > 0) clustercol->push_back(clusters[i]);
    }

    // put the slices into the event
    evt.put(std::move(clustercol));

  } // end Produce function

} // end namespace slicer
/////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/ModuleMacros.h"
namespace slicer
{
  DEFINE_ART_MODULE(TruthSlicer);
}
