////////////////////////////////////////////////////////////////////////
// \file    TimingFitAlg.cxx
// \brief   Algorithm to determine the direction of tracks using timing
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "TimingFit/TimingFitAlg.h"

#include "Calibrator/Calibrator.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "Utilities/func/MathUtil.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"

#include <cmath>

#include "TGraphErrors.h"
#include "TH2.h"
#include "TPad.h"

#include "Minuit2/FCNBase.h"
#include "Minuit2/FunctionMinimum.h"
#include "Minuit2/MnMigrad.h"
#include "Minuit2/MnUserParameters.h"

namespace tf
{
  //......................................................................
  TimingFitAlg::TimingFitAlg()
  {
  }

  class HoughScore: public ROOT::Minuit2::FCNBase
  {
  public:
    HoughScore(const std::vector<Pt>& pts) : fPts(pts) {}

    double Eval(double t0, double cinv) const
    {
      double ret = 0;
      const unsigned int I = fPts.size();
      for(unsigned int i = 0; i < I; ++i){
        const Pt& pt = fPts[i];

        const double dt = pt.t-(t0+pt.L*cinv);
        const double w = exp(-util::sqr(dt/pt.err)/2)/pt.err;
        ret += w;
      }
      return ret;
    }

    virtual double operator()(const std::vector<double>& xs) const
    {
    // Minuit is trying to minimize, so use a minus sign to maximize score
      return -Eval(xs[0], xs[1]);
    }

    virtual double Up() const {return 1;}
  protected:
    const std::vector<Pt>& fPts;
  };

  //......................................................................
  TimingFitResult TimingFitAlg::HoughFitPts(const std::vector<Pt>& pts) const
  {
    TimingFitResult ret;
    if(pts.size() < 2){
      ret.valid = false;
      return ret;
    }

    HoughScore hs(pts);

    ROOT::Minuit2::MnUserParameters mnPars;
    mnPars.Add("t0", 0, 100);
    mnPars.Add("cinv", 0, 1./30);

    // Best fit
    ROOT::Minuit2::MnMigrad mnMigrad(hs, mnPars);
    gErrorIgnoreLevel = 1001;
    ROOT::Minuit2::FunctionMinimum funcMin = mnMigrad();
    ret.bestInvSpeed = mnMigrad.Parameter(1).Value();
    ret.bestScore = -funcMin.Fval()/pts.size();
    ret.valid = funcMin.IsValid();

    // Degugging code to look at fit failures
    /*
    if(!ret.valid){
      TH2F h("", ";t_{0} (ns);1/c (ns/cm)", 100, -300, 300, 100, -.2, +.2);

      for(int x = 0; x < h.GetNbinsX()+2; ++x){
        const double t0 = h.GetXaxis()->GetBinCenter(x);
        for(int y = 0; y < h.GetNbinsY()+2; ++y){
          const double cinv = h.GetYaxis()->GetBinCenter(y);

          h.SetBinContent(x, y, hs.Eval(t0, cinv));
        } // end for y (cinv)
      } // end for x (t0)

      h.Draw("colz");
      gPad->Print("map.eps");

      TGraphErrors* g = new TGraphErrors;
      for(unsigned int i = 0; i < pts.size(); ++i){
        g->SetPoint(g->GetN(), pts[i].L, pts[i].t);
        g->SetPointError(g->GetN()-1, 0, pts[i].err);
      }

      g->SetMarkerStyle(kFullCircle);
      g->Draw("ap");
      gPad->Print("t_vs_z.eps");

      exit(0);
    }
    */

    // Best fit constrained to +c
    mnMigrad.SetValue(0u, 0.);
    mnMigrad.SetValue(1, +1./30);
    mnMigrad.Fix(1);
    funcMin = mnMigrad();
    ret.fwdScore = -funcMin.Fval()/pts.size();

    // Best fit constrained to -c
    mnMigrad.SetValue(0u, 0.);
    mnMigrad.SetValue(1, -1./30);
    mnMigrad.Fix(1);
    funcMin = mnMigrad();
    ret.revScore = -funcMin.Fval()/pts.size();

    return ret;
  }

  //......................................................................
  TimingFitResult TimingFitAlg::HoughFit(const rb::Track* track) const
  {
    std::vector<Pt> pts;

    art::ServiceHandle<calib::Calibrator> cal;

    const unsigned int cellMax = track->NCell();
    const TVector3 dir = track->Dir();
    // Center Pts around (0,0). Makes natural choice of MINUIT seeding
    const TVector3 origin = track->MeanXYZ();
    const double t0 = track->MeanTNS();
    for(unsigned int cellIdx = 0; cellIdx < cellMax; ++cellIdx){
      const art::Ptr<rb::CellHit> chit = track->Cell(cellIdx);

      const rb::RecoHit rhit = track->RecoHit(chit);
      const double t = rhit.T();

      const TVector3 pos(rhit.X(), rhit.Y(), rhit.Z());

      Pt pt((pos-origin).Dot(dir), t-t0, cal->GetTimeRes(*chit));
      pts.push_back(pt);
    } // end for cellIdx

    return HoughFitPts(pts);
  }

} // namespace
