////////////////////////////////////////////////////////////////////////
// \file    TimingFit_module.cc
// \brief   Determine the direction of tracks using timing
// \author  Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

// NOvASoft includes
#include "RecoBase/Track.h"
#include "Utilities/AssociationUtil.h"

// Framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "fhiclcpp/ParameterSet.h"

#include "TimingFit/TimingFitAlg.h"

/// Timing fit
namespace tf
{
  class Pt;

  /// Determine the direction of tracks using timing
  class TimingFit: public art::EDProducer
  {
  public:
    explicit TimingFit(const fhicl::ParameterSet& pset);
    ~TimingFit();

    virtual void reconfigure(const fhicl::ParameterSet& pset);
    virtual void produce(art::Event& evt);
  protected:
    TimingFitAlg fTimingFitAlg;

    std::string fTrackLabel;
  };

  //......................................................................
  TimingFit::TimingFit(const fhicl::ParameterSet& pset)
  {
    reconfigure(pset);

    produces<std::vector<tf::TimingFitResult>>();
    produces<art::Assns<tf::TimingFitResult, rb::Track>>();
  }

  //......................................................................
  TimingFit::~TimingFit()
  {
  }

  //......................................................................
  void TimingFit::reconfigure(const fhicl::ParameterSet& pset)
  {
    fTrackLabel = pset.get<std::string>("TrackLabel");
  }

  //......................................................................
  void TimingFit::produce(art::Event& evt)
  {
    std::unique_ptr<std::vector<tf::TimingFitResult>> rescol(new std::vector<tf::TimingFitResult>);
    std::unique_ptr<art::Assns<tf::TimingFitResult, rb::Track>> assns(new art::Assns<tf::TimingFitResult, rb::Track>);

    art::Handle<std::vector<rb::Track>> tracks;
    evt.getByLabel(fTrackLabel, tracks);

    const unsigned int trackMax = tracks->size();
    for(unsigned int trackIdx = 0; trackIdx < trackMax; ++trackIdx){
      art::Ptr<rb::Track> track(tracks, trackIdx);

      rescol->push_back(fTimingFitAlg.HoughFit(track.get()));
      util::CreateAssn(*this, evt, *rescol, track, *assns);
    } // end for trackIdx

    evt.put(std::move(rescol));
    evt.put(std::move(assns));
  }

  DEFINE_ART_MODULE(TimingFit);

} // namespace tf
////////////////////////////////////////////////////////////////////////

