//////////////////////////////////////////////////////////////////////////
/// \brief  Cosmic track analysis module CosmicTrackAna.
///         Makes some validation plots.
/// \author Gavin S. Davies - gsdavies@iastate.edu
/// \date 
//////////////////////////////////////////////////////////////////////////
#include <string>
#include <vector>
#include <algorithm>
#include <initializer_list>

// ROOT includes
#include "TH1D.h"
#include "TH2D.h"
#include "TTree.h"
#include "TVector3.h"
#include "TSpline.h"
#include "TStopwatch.h"

// NOvA includes
#include "Calibrator/Calibrator.h"
#include "GeometryObjects/Geo.h"
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "Geometry/LiveGeometry.h"
#include "MCCheater/BackTracker.h"
#include "NovaDAQConventions/DAQConventions.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/RecoHit.h"
#include "RecoBase/Track.h"
#include "Simulation/Particle.h"
#include "Simulation/FLSHit.h"
#include "Simulation/ParticleNavigator.h"
#include "SummaryData/CosmicExposure.h"
#include "Utilities/func/MathUtil.h"
#include "TrackFit/CosmicTrackStruct.h"
#include "TrackFit/CosmicTrackSelection.h"
#include "TrackFit/CosmicTrackUtilities.h"
#include "BreakPointFitter/func/Path.h"

// Framework includes
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Persistency/Common/Ptr.h"
#include "art/Persistency/Common/PtrVector.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art/Framework/Core/FindOne.h" 

enum _oneDRecoHists{
  kDCosX = 0,                 ///< direction cosine in x
  kDCosY,                     ///< direction cosine in y
  kDCosZ,                     ///< direction cosine in z
  kEndX,                      ///< track end position in x
  kEndY,                      ///< track end position in y
  kEndZ,                      ///< track end position in z
  kStartX,                    ///< track start position in x
  kStartY,                    ///< track start position in y
  kStartZ,                    ///< track start position in z
  kDiffX,                     ///< difference in track start and end position in x
  kDiffY,                     ///< difference in track start and end position in y
  kDiffZ,                     ///< difference in track start and end position in z
  kDirX,                      ///< direction in x
  kDirY,                      ///< direction in y
  kDirZ,                      ///< direction in z
  kFractionHitsUsed,          ///< Fraction of hits in event used in track
  kFractionSigUsed,           ///< Fraction of hits in event used in track
  kTrkCells,                  ///< number of cells in a track
  kTracksFound,               ///< number of tracks found
  kResidualsX,                ///< residuals for the tracks
  kResidualsY,                ///< residuals for the tracks
  kCellHitSig,                ///< signal size of calibrated cell hits
  kTrkSig,                    ///< track signal (Q)
  kVtxDistToEdge,             ///< distance in cm to closest edge of detector at track vertex
  kEndDistToEdge,             ///< distance in cm to closest edge of detector at track end
  kHitXPosition,              ///< position of reco hits in x
  kHitYPosition,              ///< position of reco hits in y
  kHitZPosition,              ///< position of reco hits in z
  kTrackLength,               ///< length of track
  kCellHitX,	   	      ///<
  kCellHitY,	   	      ///<
  kPlaneHitX,      	      ///<
  kPlaneHitY,      	      ///<
  kCosCosmic,                 ///< cosine theta of cosmic track with respect to zenith
  kAzimuth,                   ///< azimuthal angle of the cosmic ray
  kTNSX,                      ///< distribution of x/y cellhit times in nano seconds
  kTNSY,                      ///< distribution of x/y cellhit times in nano seconds
  kTrackRangeMomentum,        ///< momentum of track from range
  kCellsPerPlaneX,            ///< average number of cells hit per plane in x
  kCellsPerPlaneY,            ///< average number of cells hit per plane in y
  kStepRelativeToMedian,      ///< size of steps along trajectory normalized by the median step
  kMax1DRecoHist
};

enum _oneDTruthHists{
  kDCosXDiff = 0,             ///< Fraction difference in direction cosine in x
  kDCosYDiff,                 ///< Fraction difference in direction cosine in y
  kDCosZDiff,                 ///< Fraction difference in direction cosine in z
  kRecoEff,
  kPrimaryTrueEnergy,         ///< True energy of particle zero (GeV)
  kDeltaTrackLength,          ///< difference between true and reco length in cm
  kDeltaRangeMomentum,        ///< difference in reco momentum from range and true momentum
  kDeltaRangeMomentumUncor,   ///< difference in reco momentum from range uncorrected for track length
  kDeltaZEnd,                 ///< difference in z end point of reconstructed track vs simulated
  kDeltaEnd,                  ///< difference in end point of reconstructed track vs simulated
  kDeltaBeg,                  ///< difference in start point of reconstructed track vs simulated
  kTrackEfficiency,           ///< efficiency in selecting hits for the track 
  kTrackPurity,               ///< purity in selecting hits for the track    
  kEndTrueX,                  ///< truth end position in x
  kEndTrueY,                  ///< truth end position in y
  kEndTrueZ,                  ///< truth end position in z
  kStartTrueX,                ///< truth start position in x
  kStartTrueY,                ///< truth start position in y
  kStartTrueZ,                ///< truth start position in z
  kCosCosmicTrue,             ///< cosine theta of cosmic track with respect to zenith
  kAzimuthTrue,               ///< azimuthal angle of the cosmic ray
  kMax1DTruthHist
};

enum _twoDRecoHists{
  kFracHitsUsedVsNum = 0, ///< Fraction hits used vs the number of hits that could be used
  kEndXY,		  ///< track end position in xy
  kEndZX,                 ///< track end position in zx
  kEndZY,		  ///< track end position in zy
  kStartXY,		  ///< track start position in xy
  kStartZX,               ///< track start position in zx
  kStartZY,		  ///< track start position in zy
  kDirXY,                 ///< direction in x versus y
  kResidVsCosX,           ///< residual for each hit vs cos(theta)
  kResidVsCosY,		  
  kResidVsAzX,            ///< residual for each hit vs azimuth   
  kResidVsAzY,
  kHitXYPosition,         ///< reco hit positions in xy
  kHitXZPosition,	  ///< reco hit positions in xz
  kHitYZPosition,	  ///< reco hit positions in yz
  kCellPlaneX,            ///< 
  kCellPlaneY,            ///< 
  kCellsPerPlaneXY,       ///< cells hit per plane in Y vs X
  kPEPerCMVsWX,           ///< PE/cm deposited as a function of W in X view
  kPEPerCMVsWY,           ///< PE/cm deposited as a function of W in Y view
  kStepRelativeToMedianVsAz, ///< largest trajectory step relative to the median vs azimuth
  kMax2DRecoHist
};

enum _twoDTruthHists{
  kDCosXRecoVsTrue = 0,   ///< Direction cosine in x				      
  kDCosYRecoVsTrue,  	  ///< Direction cosine in y				      
  kDCosZRecoVsTrue,	  ///< Direction cosine in z				      
  kDCosXDiffVsZLen,	  ///< Fraction difference in direction cosine in x vs z length   
  kDCosYDiffVsZLen,	  ///< Fraction difference in direction cosine in y vs z length   
  kDCosZDiffVsZLen,	  ///< Fraction difference in direction cosine in z vs z length 
  kTrackFindEff,          
  kTracksFoundVsLeft,     ///< number of tracks found in event vs particles leaving tracks
  kDeltaZEndVsDCosZ,      ///< difference in z endpoint of recoed track & simulated vs dcosZ    
  kDeltaZBegVsDCosZ,  	  ///< difference in z start point of recoed track & simulated vs dcosZ 
  kDeltaZEndVsDistToEdge, ///< difference in z endpoint of recoed track & simulated vs dcosZ    
  kDiffEndVsDCosZ,        ///< difference in end point of recoed track & simulated vs dcosZ    
  kDiffBegVsDCosZ,  	  ///< difference in start point of recoed track & simulated vs dcosZ 
  kDiffEndVsAzimuth,      ///< difference in end point of recoed track & simulated vs azimuth
  kDiffBegVsAzimuth,  	  ///< difference in start point of recoed track & simulated vs azimuth 
  kDiffEndVsCosTheta,     ///< difference in end point of recoed track & simulated vs cosTheta    
  kDiffBegVsCosTheta,  	  ///< difference in start point of recoed track & simulated vs cosTheta 
  kDeltaRangeMomVsLength, ///< difference in reco and true range momentum as a function of reco length
  kDeltaRangeMomUncorVsLength,  ///< difference in reco and true range momentum as a function of reco length
  kDeltaRangeMomUncorVsCosTheta,///< difference in reco and true range momentum as a function of cos theta
  kDeltaRangeMomUncorVsAzimuth, ///< difference in reco and true range momentum as a function of azimuth
  kDeltaLengthVsLength,         ///< difference in recon and true length as a function of reco length
  kdEdxVsRange,                 ///< dE/dx vs remaining length in a track 
  kDeltaLengthVsPlaneAsymmetry, ///< fractional length vs asymmetry in number of planes in each view
  kDeltaLengthVsDeltaStartPlane,///< fractional length vs difference in start plane in each view
  kDeltaLengthVsDeltaEndPlane,  ///< fractional length vs difference in end   plane in each view
  kDeltaLengthVsCellsPerPlane,  ///< fractional length vs maximum cells per plane
  kMax2DTruthHist
};


namespace trk {
  // A module to analyse cosmic tracks produced by CosmicTrack module
  class CosmicTrackAna : public art::EDAnalyzer {
  public:
    explicit CosmicTrackAna(fhicl::ParameterSet const& pset);
    virtual ~CosmicTrackAna();

    void analyze(art::Event const& evt);
    void reconfigure(const fhicl::ParameterSet& pset);

    virtual void beginSubRun(const art::SubRun& sr);

    void beginJob();
    void endJob();
        
  private:

    void Make1DRecoHistograms(std::vector<TH1D *> &hists, 
			      TString const        sample);
    void Make1DTruthHistograms(std::vector<TH1D *> &hists, 
			       TString const        sample);
    void Make2DRecoHistograms(std::vector<TH2D *> &hists,
			      TString const        sample);
    void Make2DTruthHistograms(std::vector<TH2D *> &hists,
			       TString const        sample);

    std::pair<CosmicTrackInfo, 
	      CosmicSignalHitFraction> FillRecoInfo(rb::Track             const& track,
						    rb::Cluster 	  const& slice);
    CosmicTrackInfo                    FillTrueInfo(rb::Track                          const& track,
						    rb::Cluster 	  	       const& slice,
						    std::map<geo::CellUniqueId, float>      & trueCellPathLength);
    
    void FillEventMCTruthHistograms(size_t trkMax);   

    void FillTrackMCTruthHistograms(CosmicTrackInfo  const& ti,
				    CosmicTrackInfo  const& ri);
    void FillTrackHistograms(CosmicTrackInfo                    const& ri,	  
			     CosmicSignalHitFraction 		const& shf,  
			     rb::Track               		const& track,
			     std::map<geo::CellUniqueId, float> const& trueCellPathLength);   

    void FilldEdxHistogram(simb::MCParticle const& part);

    CosmicTrackSelection                   fSelection;               ///< object encapsulating selection criteria
    CosmicTrackUtilities                   fUtilities;               ///< object encapsulating utilitie methods
    std::string                            fTrackLabel;        	     ///< Where to find Tracks to reconstruct
    std::string   			   fExposureLabel;     	     ///< Where to find CosmicExposure info   
    double        			   fMinSig;                  ///< Minimum signal of cell hit to use in a 
                                                                     ///< track - using PE values	    
    bool          			   fFindPurityAndEfficiency; ///< Set to true in fcl file to calculate them, 
                                                                     ///< false by default	    
                  			                             ///< because the calculation is so slow
    int           			   fTicks;                   ///< # of clock ticks to read out one event
    int           			   fTickWindow;        	     ///< Size of window for ticks of TDC to slice event in time 
    double        			   fLiveTime;          	     ///< Livetime in sec according to file
    unsigned int  			   fNevents;	      	     ///< total number of events counter	
    int           			   fNTracks;           	     ///< total number of reconstructed tracks counter
    int           			   fNTracksMC;         	     ///< total number of MC particles	
    int           			   fRun;		     ///< Run number
    int           			   fSubRun;            	     ///< Subrun number             
    double        			   fCellHalfDepthZ;          ///< extent of a cell in the z direction	
    double        			   fDetHalfWidth;            ///< detector half width, cached	
    double        			   fDetHalfHeight;           ///< detector half height, cached	
    double        			   fDetLength;               ///< detector length, cached	       
					                                                                      
    TH1D*                                  fCosmicRate;              ///< histogram holding single value for cosmic rate in Hz at end of job
    TH1D*                                  fExposure;                ///< histogram holding single value for exposure in s at end of job

    std::vector<TH1D *> 		   fOneDHistsRecoAll;        ///< one dimensional histograms for all tracks	   
    std::vector<TH1D *> 		   fOneDHistsRecoStop;       ///< one dimensional histograms for stopping tracks
    std::vector<TH1D *> 		   fOneDHistsTruthAll;       ///< one dimensional histograms for all tracks	   
    std::vector<TH1D *> 		   fOneDHistsTruthStop;      ///< one dimensional histograms for stopping tracks
    std::vector<TH2D *> 		   fTwoDHistsRecoAll;        ///< two dimensional histograms for all tracks	   
    std::vector<TH2D *> 		   fTwoDHistsRecoStop;       ///< two dimensional histograms for stopping tracks
    std::vector<TH2D *> 		   fTwoDHistsTruthAll;       ///< two dimensional histograms for all tracks	   
    std::vector<TH2D *> 		   fTwoDHistsTruthStop;      ///< two dimensional histograms for stopping tracks

    CosmicTrackTree                        fTrigger;                 ///< event information			       
    CosmicTrackInfo                        fTrackTruth;              ///< single track mc truth information	       
    CosmicTrackInfo                        fTrackReco;               ///< single track reco information	       
    CosmicSignalHitFraction                fSigHitFrac;              ///< fraction of signal and hits used	       
    CosmicHitInfo                          fCosmicHit;               ///< information about hits
    TTree*                                 fTree;                    ///< initializer of TTree for this module        
    TTree*                                 fHitTree;                 ///< initializer of Hit TTree for this module        
    art::ServiceHandle<geo::Geometry>      fGeom;
    art::ServiceHandle<geo::LiveGeometry>  fLiveGeom;
    art::ServiceHandle<cheat::BackTracker> fBT;
    art::ServiceHandle<art::TFileService>  fTFS;

  };
};

static const double ns            = 1.0E-9;

namespace trk{

  bool sortTVector3ByZ(TVector3 const& a, TVector3 const& b)
  {
    return a.Z() < b.Z();
  }

  bool sortTVector3ByY(TVector3 const& a, TVector3 const& b)
  {
    return a.Y() > b.Y();
  }

  bool sortFLSHits(sim::FLSHit const& a, sim::FLSHit const& b)
  {
    if(a.GetPlaneID() != b.GetPlaneID())
      return a.GetPlaneID() < b.GetPlaneID();
    
    return a.GetCellID() < b.GetCellID();    
  }
  
  //......................................................................
  CosmicTrackAna::CosmicTrackAna(fhicl::ParameterSet const& pset)
    : EDAnalyzer(pset)
    , fMinSig          (0)
    , fLiveTime        (0) 
    , fNevents         (0) 
    , fNTracks         (0) 
    , fNTracksMC       (0) 
    , fRun             (-999) 
    , fSubRun          (-999)
    , fCellHalfDepthZ  (-999.)
    , fDetHalfWidth    (-999.)
    , fDetHalfHeight   (-999.)
    , fDetLength       (-999.)
  {
    reconfigure(pset);    
  }
  
  //......................................................................  
  CosmicTrackAna::~CosmicTrackAna()
  {
  }
  
  //......................................................................  
  void CosmicTrackAna::beginJob()
  {
      
    LOG_DEBUG("CosmicTrackAna") << "cell half depth: " << fGeom->Plane(0)->Cell(0)->HalfD() << "\n"
				<< "cell half width: " << fGeom->Plane(0)->Cell(0)->HalfW() << "\n"
				<< "cell half leng:  " << fGeom->Plane(0)->Cell(0)->HalfL();

    // will always want the reco histograms, not sure we will want the truth 
    // histograms until we determine we have MC
    this->Make1DRecoHistograms(fOneDHistsRecoAll,  "AllReco");
    this->Make1DRecoHistograms(fOneDHistsRecoStop, "StopReco");
    this->Make2DRecoHistograms(fTwoDHistsRecoAll,  "AllReco");
    this->Make2DRecoHistograms(fTwoDHistsRecoStop, "StopReco");

    fCosmicRate = fTFS->make<TH1D>("cosmicrate", ";Cosmic Rate (kHz)", 500, 0., 100);
    fExposure   = fTFS->make<TH1D>("exposure",   ";Exposure (s)",      1000, 0., 10);


    std::string trigst("run/I:subrun/I:event/I:nevents/I:ntracks/I");
    std::string shfst ("signalFraction/F:hitFraction/F:signal/F:totcell/F:hitsX/F:hitsY/F:signalX/F:signalY/F");
    std::string ctist ("efficiency/F:purity/F:pdg/I");
    ctist.append(":length/F:momentum/F:momentumCor/F:momentumBPF/F:calorE/F");
    ctist.append(":dcosX/F:dcosY/F:dcosZ/F");
    ctist.append(":vtxX/F:vtxY/F:vtxZ/F");
    ctist.append(":endX/F:endY/F:endZ/F");
    ctist.append(":deltaVtxPos/F:deltaEndPos/F");
    ctist.append(":cosTheta/F:azimuth/F");
    ctist.append(":vtxToEdge/F:endToEdge/F");
    ctist.append(":cellsPerPlaneX/F:cellsPerPlaneY/F");
    ctist.append(":planesX/F:planesY/F:xStartPlane/F:yStartPlane/F:xEndPlane/F:yEndPlane/F");
    ctist.append(":avgResidual/F:maxResidual/F:triCellFrac/F");
    ctist.append(":stop/O:goodVtx/O:isCalibTrk/O");
    fTree = fTFS->make<TTree>("Cosmics","CosmicTrackAna Tree");
    fTree->Branch("trigger",           &fTrigger,    trigst.c_str());
    fTree->Branch("truth",             &fTrackTruth, ctist.c_str());
    fTree->Branch("reco",              &fTrackReco,  ctist.c_str());
    fTree->Branch("signalHitFraction", &fSigHitFrac, shfst.c_str());

    std::string chist ("pathLength/F:truePathLength/F:triCellPathLength/F:PE/F:PECorr/F:w/F:distFromEnd/F:cell/I:plane/I:module/I:view/I");
    fHitTree = fTFS->make<TTree>("Hits","CosmicHit Tree");
    fHitTree->Branch("hits", &fCosmicHit, chist.c_str());
    
  }

  //......................................................................
  void CosmicTrackAna::Make1DRecoHistograms(std::vector<TH1D *> &hists,
					    TString const        sample)
  {

    hists.resize(kMax1DRecoHist);

    double maxlength = 6600.0;
    double xmax      = 1200.0;
    double ymax      = 1200.0;
    double zmax      = 6500.0;
    int    xbins     = 240;
    int    ybins     = 240;
    int    zbins     = 650;
    double ncell     = 384;
    double nplane    = 960;

    hists[kDCosX]             = fTFS->make<TH1D>("dcosX"+sample,            ";dcos_{X};",                  101, -1., 1.);  
    hists[kDCosY]             = fTFS->make<TH1D>("dcosY"+sample,  	    ";dcos_{Y};",                  101, -1., 1.); 
    hists[kDCosZ]             = fTFS->make<TH1D>("dcosZ"+sample,  	    ";dcos_{Z};",                  101, -1., 1.);
    hists[kEndX] 	      = fTFS->make<TH1D>("endX"+sample,   	    ";End X (cm)",                 xbins, -xmax, xmax);
    hists[kEndY] 	      = fTFS->make<TH1D>("endY"+sample,   	    ";End Y (cm)",                 ybins, -ymax, ymax);
    hists[kEndZ] 	      = fTFS->make<TH1D>("endZ"+sample,   	    ";End Z (cm)",                 zbins, 0., zmax);	 
    hists[kDirX] 	      = fTFS->make<TH1D>("dirX"+sample,   	    ";Direction X (cm)",           100, -1.1, 1.1); 
    hists[kDirY] 	      = fTFS->make<TH1D>("dirY"+sample,   	    ";Direction Y (cm)",           100, -1.1, 1.1); 
    hists[kDirZ] 	      = fTFS->make<TH1D>("dirZ"+sample,   	    ";Direction Z (cm)",           100, -1.1, 1.1); 
    hists[kStartX]            = fTFS->make<TH1D>("startX"+sample, 	    ";Start X (cm)",               xbins, -xmax, xmax); 
    hists[kStartY]            = fTFS->make<TH1D>("startY"+sample, 	    ";Start Y (cm)",               ybins, -ymax, ymax); 
    hists[kStartZ]            = fTFS->make<TH1D>("startZ"+sample, 	    ";Start Z (cm)",               zbins, 0., zmax);	 
    hists[kDiffX]             = fTFS->make<TH1D>("diffX"+sample,  	    ";End-Start X (cm)",           100, 0., 1000.);
    hists[kDiffY]             = fTFS->make<TH1D>("diffY"+sample,  	    ";End-Start Y (cm)",           100, 0., 1000.);
    hists[kDiffZ]             = fTFS->make<TH1D>("diffZ"+sample,  	    ";End-Start Z (cm)",           150, 0., 1500.);
    hists[kFractionHitsUsed]  = fTFS->make<TH1D>("fractionHitsUsed"+sample, ";Used/Total (Hits);Spills",   100, 0., 1.);
    hists[kFractionSigUsed]   = fTFS->make<TH1D>("fractionSigUsed"+sample,  ";Used/Total (Signal);Spills", 100, 0., 1.);
    hists[kTracksFound]       = fTFS->make<TH1D>("tracksFound"+sample,      ";Reco tracks;",               25, 0, 125);
    hists[kCellHitSig]        = fTFS->make<TH1D>("cellHitSig"+sample,       ";PE;Cell Hits",               100, 0., 100.);
    hists[kTrkSig]            = fTFS->make<TH1D>("trksig"+sample,           ";Track Signal (Q);counts",    100,0.,100000.);	   
    hists[kTrkCells]          = fTFS->make<TH1D>("fTrkCells"+sample,        ";Number of Track Cell Hits",  500,0,500);	   
    hists[kEndDistToEdge]     = fTFS->make<TH1D>("endDistToEdge"+sample,    ";End Distance (cm);",         300,  -1500., 1500.);
    hists[kVtxDistToEdge]     = fTFS->make<TH1D>("vtxDistToEdge"+sample,    ";Vertex Distance (cm);",      300,  -1500., 1500.);
    hists[kTrackLength]       = fTFS->make<TH1D>("trackLength"+sample,      ";Track Length (cm);",         264, 0., 2.*maxlength);
    hists[kCosCosmic]         = fTFS->make<TH1D>("coscosmic"+sample,        ";cos#theta",                  100, 0., 1.);	
    hists[kAzimuth]           = fTFS->make<TH1D>("azimuth"+sample,          ";#phi",                       36,  0., 360.);
    hists[kHitXPosition]      = fTFS->make<TH1D>("hitXPosition"+sample,     ";Hit X (cm);Reco Hits",       xbins, -xmax, xmax);
    hists[kHitYPosition]      = fTFS->make<TH1D>("hitYPosition"+sample,     ";Hit Y (cm);Reco Hits",       ybins, -ymax, ymax);
    hists[kHitZPosition]      = fTFS->make<TH1D>("hitZPosition"+sample,     ";Hit Z (cm);Reco Hits",       zbins, 0., zmax);	   
    hists[kCellHitX]          = fTFS->make<TH1D>("CellX"+sample,            ";cell #",                     ncell, -0.5, ncell-0.5);
    hists[kCellHitY]          = fTFS->make<TH1D>("CellY"+sample,            ";cell #",  		   ncell, -0.5, ncell-0.5);
    hists[kPlaneHitX]         = fTFS->make<TH1D>("PlaneX"+sample,           ";plane #",		           nplane, 0, nplane-0.5);
    hists[kPlaneHitY]         = fTFS->make<TH1D>("PlaneY"+sample,           ";plane #",		           nplane, 0, nplane-0.5);
    hists[kResidualsX]        = fTFS->make<TH1D>("residualsX"+sample,       ";Residual (cm);Hits",         400, 0., 1000.);
    hists[kResidualsY]        = fTFS->make<TH1D>("residualsY"+sample,       ";Residual (cm);Hits",         400, 0., 1000.);
    hists[kTNSX]              = fTFS->make<TH1D>("HitTNSX"+sample,          ";Time (ns);hits",             70000, -1.e5, 6.e5);  
    hists[kTNSY]              = fTFS->make<TH1D>("HitTNSY"+sample,          ";Time (ns);hits",             70000, -1.e5, 6.e5);  
    hists[kTrackRangeMomentum]= fTFS->make<TH1D>("rangeMomentum"+sample,    ";p_{range} (GeV);Tracks",     100, 0., 100.);
    hists[kCellsPerPlaneX]    = fTFS->make<TH1D>("cellsPerPlaneX"+sample,   ";<Cells/Plane>_{X view}",     200, 0., 50. );
    hists[kCellsPerPlaneY]    = fTFS->make<TH1D>("cellsPerPlaneY"+sample,   ";<Cells/Plane>_{X view}",     200, 0., 50. );
    hists[kStepRelativeToMedian] = fTFS->make<TH1D>("stepsRelativeToMedian"+sample, ";Step Size/Median Step", 200, 0., 50.);
    for( auto hist : hists) hist->Sumw2();

    return;
  }

  //......................................................................
  void CosmicTrackAna::Make1DTruthHistograms(std::vector<TH1D *> &hists,
					     TString const        sample)
  {

    hists.resize(kMax1DTruthHist);

    double maxlength = 6600.0;

    hists[kDCosXDiff]         = fTFS->make<TH1D>("dcosXDiff"+sample,         ";dcos_{X} Fractional Difference;", 200, -1., 3.);
    hists[kDCosYDiff]         = fTFS->make<TH1D>("dcosYDiff"+sample,         ";dcos_{Y} Fractional Difference;", 200, -1., 3.);
    hists[kDCosZDiff]         = fTFS->make<TH1D>("dcosZDiff"+sample,         ";dcos_{Z} Fractional Difference;", 200, -1., 3.);
    hists[kPrimaryTrueEnergy] = fTFS->make<TH1D>("primaryTrueEnergy"+sample, ";True Energy (GeV);",              200, 0., 50.);
    hists[kRecoEff]           = fTFS->make<TH1D>("recoeff"+sample,           ";Reco Efficiency",                 100, 0., 1.);
    hists[kTrackPurity]       = fTFS->make<TH1D>("trackPurity"+sample,       ";Tracking Purity; Tracks",         100, 0., 2.);
    hists[kTrackEfficiency]   = fTFS->make<TH1D>("trackEfficiency"+sample,   ";Tracking Efficiency; Tracks",     100, 0., 2.);
    hists[kDeltaTrackLength]  = fTFS->make<TH1D>("deltaTrackLength"+sample,  ";#Delta_{Track Length} (cm);",     264, -maxlength, maxlength);
    hists[kDeltaRangeMomentum]= fTFS->make<TH1D>("deltaRangeMom"+sample,     ";#Delta p_{range} (GeV/c);Tracks", 400, -2, 2.);
    hists[kDeltaRangeMomentumUncor]= fTFS->make<TH1D>("deltaRangeMomUncor"+sample, ";#Delta p_{range} (GeV/c);Tracks", 100, -2, 2.);
    hists[kDeltaZEnd]         = fTFS->make<TH1D>("deltaZEnd"+sample,         ";#Delta Z_{end} (cm);Tracks",      200, -100., 100.);
    hists[kDeltaEnd]          = fTFS->make<TH1D>("deltaEnd"+sample,          ";#Delta_{end} (cm);Tracks",        200, -100., 100.);
    hists[kDeltaBeg]          = fTFS->make<TH1D>("deltaBeg"+sample,          ";#Delta_{beg} (cm);Tracks",        200, -100., 100.);
    hists[kEndTrueX] 	      = fTFS->make<TH1D>("endX"+sample,   	     ";End X (cm)",                      80, -800., 800.);
    hists[kEndTrueY] 	      = fTFS->make<TH1D>("endY"+sample,   	     ";End Y (cm)",                      80, -800., 800.); 
    hists[kEndTrueZ] 	      = fTFS->make<TH1D>("endZ"+sample,   	     ";End Z (cm)",                 	 300, 0., 6000.);	 
    hists[kStartTrueX]        = fTFS->make<TH1D>("startX"+sample, 	     ";Start X (cm)",               	 80, -800., 800.);
    hists[kStartTrueY]        = fTFS->make<TH1D>("startY"+sample, 	     ";Start Y (cm)",               	 80, -800., 800.);
    hists[kStartTrueZ]        = fTFS->make<TH1D>("startZ"+sample, 	     ";Start Z (cm)",               	 300, 0., 6000.);
    hists[kCosCosmicTrue]     = fTFS->make<TH1D>("coscosmic"+sample,        ";cos#theta",                  100, 0., 1.);	
    hists[kAzimuthTrue]       = fTFS->make<TH1D>("azimuth"+sample,          ";#phi",                       36,  0., 360.);

    for( auto hist : hists) hist->Sumw2();

    return;
  }
  
  //......................................................................
  void CosmicTrackAna::Make2DRecoHistograms(std::vector<TH2D *> &hists,
					    TString const        sample)
  {    

    hists.resize(kMax2DRecoHist);

    double xmax      = 1200.0;
    double ymax      = 1200.0;
    double zmax      = 6500.0;
    int    xbins     = 24;
    int    ybins     = 24;
    int    zbins     = 130;
    double ncell     = 384;
    double nplane    = 960;

    hists[kEndXY]             = fTFS->make<TH2D>("endXY"+sample,                 ";End X (cm); End Y (cm)",              xbins, -xmax, xmax, ybins, -ymax, ymax);
    hists[kEndZY]             = fTFS->make<TH2D>("endZY"+sample,                 ";End Z (cm); End Y (cm)",		 zbins, 0., zmax, ybins, -ymax, ymax);   
    hists[kEndZX]             = fTFS->make<TH2D>("endZX"+sample,                 ";End Z (cm); End X (cm)",		 zbins, 0., zmax, xbins, -xmax, xmax);   
    hists[kStartXY]           = fTFS->make<TH2D>("startXY"+sample,               ";Start X (cm); Start Y (cm)",          xbins, -xmax, xmax, ybins, -ymax, ymax);
    hists[kStartZY]           = fTFS->make<TH2D>("startZY"+sample,               ";Start Z (cm); Start Y (cm)",		 zbins, 0., zmax, ybins, -ymax, ymax);   
    hists[kStartZX]           = fTFS->make<TH2D>("startZX"+sample,               ";Start Z (cm); Start X (cm)",		 zbins, 0., zmax, xbins, -xmax, xmax);   
    hists[kDirXY]             = fTFS->make<TH2D>("dirXY"+sample,                 ";Prong Direction X;Prong Direction Y", 101, -1., 1., 101, -1., 1.);
    hists[kFracHitsUsedVsNum] = fTFS->make<TH2D>("fractionHitsUsedVsNum"+sample, ";Total Hits; Fraction Used",           250, 0., 500., 100, 0., 1.);    
    hists[kHitXYPosition]     = fTFS->make<TH2D>("hitXYPosition"+sample,         ";Hit X (cm);Hit Y (cm)",               20, -200., 200., 50, -250., 250.);	
    hists[kHitXZPosition]     = fTFS->make<TH2D>("hitXZPosition"+sample,         ";Hit Z (cm);Hit X (cm)",               150, 0.,  1500., 40, -200., 200.);
    hists[kHitYZPosition]     = fTFS->make<TH2D>("hitYZPosition"+sample,         ";Hit Z (cm);Hit Y (cm)",               150, 0.,  1500., 50, -250., 250.);
    hists[kCellPlaneX]        = fTFS->make<TH2D>("ViewX"+sample,                 ";Plane no. ;Cell no.",                 nplane, 0., nplane-0.5, ncell, 0., ncell-0.5);
    hists[kCellPlaneY]        = fTFS->make<TH2D>("ViewY"+sample,                 ";Plane no. ;Cell no.",                 nplane, 0., nplane-0.5, ncell, 0., ncell-0.5);      
    hists[kResidVsCosX]       = fTFS->make<TH2D>("residVsCosX"+sample,           ";Residual (cm);cos(#theta)",           200, 0., 1000., 100, -1., 1.); 
    hists[kResidVsCosY]       = fTFS->make<TH2D>("residVsCosY"+sample,           ";Residual (cm);cos(#theta)", 		 200, 0., 1000., 100, -1., 1.); 
    hists[kResidVsAzX]        = fTFS->make<TH2D>("residVsAzX"+sample,            ";Residual (cm);#phi",        		 200, 0., 1000., 180, 0., 360.);
    hists[kResidVsAzY]        = fTFS->make<TH2D>("residVsAzY"+sample,            ";Residual (cm);#phi",        		 200, 0., 1000., 180, 0., 360.);
    hists[kCellsPerPlaneXY]   = fTFS->make<TH2D>("cellsPerPlaneXY"+sample,       ";<Cells/Plane>_{X view};<Cells/Plane>_{Y view}",    200, 0., 50.,   200,  0., 50.);
    hists[kPEPerCMVsWX]        = fTFS->make<TH2D>("pePerCMVsWX"+sample,          ";W (cm);PE/cm",                        xbins, -xmax, xmax, 60, 0., 300.);
    hists[kPEPerCMVsWY]        = fTFS->make<TH2D>("pePerCMVsWY"+sample,          ";W (cm);PE/cm",                        ybins, -ymax, ymax, 60, 0., 300.);
    hists[kStepRelativeToMedianVsAz] = fTFS->make<TH2D>("stepRelativeToMedianVsAz"+sample, ";#phi;Largest Step/Median Step", 120, 0., 360., 100, 0., 10.);

    for( auto hist : hists) hist->Sumw2();

    return;
  }

  //......................................................................
  void CosmicTrackAna::Make2DTruthHistograms(std::vector<TH2D *> &hists,
					     TString const        sample)
  {    

    hists.resize(kMax2DTruthHist);

    double maxlength = 6600.0;

    hists[kDCosXRecoVsTrue]       = fTFS->make<TH2D>("dcosXRecoVsTrue"+sample,       ";dcos X_{true};dcos X_{reco}",                  250, -1., 1., 250, -1., 1.);
    hists[kDCosYRecoVsTrue]       = fTFS->make<TH2D>("dcosYRecoVsTrue"+sample,       ";dcos Y_{true};dcos Y_{reco}", 		      250, -1., 1., 250, -1., 1.);
    hists[kDCosZRecoVsTrue]       = fTFS->make<TH2D>("dcosZRecoVsTrue"+sample,       ";dcos Z_{true};dcos Z_{reco}", 		      250, -1., 1., 250, -1., 1.);
    hists[kDCosXDiffVsZLen]       = fTFS->make<TH2D>("dcosXDiffVsZLen"+sample,       ";Z length (cm);dcos_{X} Fractional Difference", 1320, 0., maxlength, 200, -1., 5.);
    hists[kDCosYDiffVsZLen]       = fTFS->make<TH2D>("dcosYDiffVsZLen"+sample,       ";Z length (cm);dcos_{Y} Fractional Difference", 1320, 0., maxlength, 200, -1., 3.);
    hists[kDCosZDiffVsZLen]       = fTFS->make<TH2D>("dcosZDiffVsZLen"+sample,       ";Z length (cm);dcos_{Z} Fractional Difference", 1320, 0., maxlength, 200, -1., 3.);
    hists[kTracksFoundVsLeft]     = fTFS->make<TH2D>("tracksFoundVsLeft"+sample,     ";True tracks;Reco tracks",                      25, 0, 125, 25, 0, 125);
    hists[kTrackFindEff]          = fTFS->make<TH2D>("trackfindeff"+sample,          ";True tracks;Reco Efficiency",                  20, 0, 10, 100, 0, 1);
    hists[kDeltaZEndVsDCosZ]      = fTFS->make<TH2D>("deltaZEndVsDCosZ"+sample,      ";#Delta Z_{end} (cm);dZ/dS;",                   200, -100., 100., 20, -1., 1.);
    hists[kDeltaZBegVsDCosZ]      = fTFS->make<TH2D>("deltaZBegVsDCosZ"+sample,      ";#Delta Z_{start} (cm);dZ/dS;",                 200, -100., 100., 20, -1., 1.);
    hists[kDiffEndVsDCosZ]        = fTFS->make<TH2D>("diffEndVsDCosZ"+sample,        ";dZ/dS;#Delta_{end} (cm)",                      20,  -1., 1., 200, -100., 100.);
    hists[kDiffBegVsDCosZ]        = fTFS->make<TH2D>("diffBegVsDCosZ"+sample,        ";dZ/dS;#Delta_{start} (cm)",                    20,  -1., 1., 200, -100., 100.);
    hists[kDiffEndVsAzimuth]      = fTFS->make<TH2D>("diffEndVsAzimuth"+sample,      ";#phi;#Delta_{end} (cm);",                      120,  0., 360., 200, -100., 100.);
    hists[kDiffBegVsAzimuth]      = fTFS->make<TH2D>("diffBegVsAzimuth"+sample,      ";#phi;#Delta_{start} (cm);",                    120,  0., 360., 200, -100., 100.);
    hists[kDiffEndVsCosTheta]     = fTFS->make<TH2D>("diffEndVsCosTheta"+sample,     ";cos(#theta);#Delta_{end} (cm);",               20,  -1., 1., 200, -100., 100.);
    hists[kDiffBegVsCosTheta]     = fTFS->make<TH2D>("diffBegVsCosTheta"+sample,     ";cos(#theta);#Delta_{start} (cm);",             20,  -1., 1., 200, -100., 100.);
    hists[kDeltaZEndVsDistToEdge] = fTFS->make<TH2D>("deltaZEndVsDistToEdge"+sample, ";#Delta Z_{end} (cm);Dist To Edge (cm);",       200, -100., 100., 200, -100., 100.);
    hists[kDeltaRangeMomVsLength] = fTFS->make<TH2D>("deltaMomVsLength"+sample,      ";l_{reco} (cm);#Deltap/p",                      100, 0., 2500., 2000, -1., 1.);
    hists[kDeltaRangeMomUncorVsLength]   = fTFS->make<TH2D>("deltaMomUncorVsLength"+sample,        ";l_{reco} (cm);#Deltap/p",        100, 0., 2500., 2000, -1., 1.);
    hists[kDeltaRangeMomUncorVsCosTheta] = fTFS->make<TH2D>("deltaMomUncorVsCosTheta"+sample,      ";cos#theta_{reco};#Deltap/p",     100, 0., 1.,    2000, -1., 1.);
    hists[kDeltaRangeMomUncorVsAzimuth]  = fTFS->make<TH2D>("deltaMomUncorVsAzimuth"+sample,       ";#phi_{reco} (deg);#Deltap/p",    100, 0., 360.,  2000, -1., 1.);
    hists[kDeltaLengthVsLength]          = fTFS->make<TH2D>("deltaLengthVsLength"+sample,          ";l_{reco} (cm);#Deltal/l",        100, 0., 2500., 2000, -1., 1.);
    hists[kdEdxVsRange]                  = fTFS->make<TH2D>("dEdxVsRange"+sample,                  ";Range (cm);dE/dx (MeV)",         5000, 0., 5000., 2000, 0., 20.);
    hists[kDeltaLengthVsPlaneAsymmetry]  = fTFS->make<TH2D>("deltaLengthVsPlaneAsymmetry"+sample,  ";Plane Asymmetry;#Deltal/l",      200, -1., 1.,  2000, -1., 1.);
    hists[kDeltaLengthVsDeltaStartPlane] = fTFS->make<TH2D>("deltaLengthVsDeltaStartPlane"+sample, ";#Delta Start Plane;#Deltal/l",   50,   0., 50., 2000, -1., 1.);
    hists[kDeltaLengthVsDeltaEndPlane]   = fTFS->make<TH2D>("deltaLengthVsDeltaEndPlane"+sample,   ";#Delta Start Plane;#Deltal/l",   50,   0., 50., 2000, -1., 1.);
    hists[kDeltaLengthVsCellsPerPlane]   = fTFS->make<TH2D>("deltaLengthVsCellsPerPlane"+sample,   ";Maximum Cells/Plane;#Deltal/l",  50,   0., 50., 2000, -1., 1.);

    for( auto hist : hists) hist->Sumw2();

    return;
  }

  //......................................................................
  void CosmicTrackAna::beginSubRun(const art::SubRun& sr)
  {
    art::Handle<sumdata::CosmicExposure> ce;
    sr.getByLabel(fExposureLabel, ce);
    if(ce.failedToGet()){
      mf::LogError("CosmicTrackAna") << "No Cosmic Exposure Info";
      return;
    }
    fLiveTime += ce->totlivetime;

    LOG_INFO("CosmicTrackAna") << "The live time is now " << fLiveTime;

    return;
  }
  
  //......................................................................
  ///
  /// The analysis method is intended to perform checks of the
  /// performance of the reconstruction for the purposes of
  /// optimization, testing, and validation.
  ///
  /// \param evt - Read-only access to the event data
  ///
  // checks to make include - 
  // 1. are the direction cosines of the track correct
  // 2. fraction of hits used in tracks
  // 3. number of particles leaving tracks vs number tracks found
  // 4. fraction of hits in a track that belong to a given particle
  // 5. cosmic flux rate and angular distribtion stability
  // 6. cosmic track efficiency
  void CosmicTrackAna::analyze(art::Event const& evt)
  {  
    if(fDetLength < 0){
      fDetHalfWidth   = fGeom->DetHalfWidth();
      fDetHalfHeight  = fGeom->DetHalfHeight();
      fDetLength      = fGeom->DetLength(); 
      fCellHalfDepthZ = fGeom->Plane(0)->Cell(0)->HalfD();    
    }

    // make the histograms to record Truth related information 
    // if they havent already been made and if there is truth 
    // information in the event
    if( fBT->HaveTruthInfo() && fOneDHistsTruthAll.size() < 1){
      this->Make1DTruthHistograms(fOneDHistsTruthAll,  "AllTruth");
      this->Make1DTruthHistograms(fOneDHistsTruthStop, "StopTruth");
      this->Make2DTruthHistograms(fTwoDHistsTruthAll,  "AllTruth");
      this->Make2DTruthHistograms(fTwoDHistsTruthStop, "StopTruth");
    }


    ++fNevents; //Updating counter of number of events in sub-run
    if(fRun < 0)    fRun = evt.run();
    if(fSubRun < 0) fSubRun = evt.subRun();
    
    fTrigger.run      = evt.run();	
    fTrigger.subrun   = evt.subRun();	
    fTrigger.event    = evt.id().event(); 
    fTrigger.nevents  = fNevents;
    
    // now get the tracks that were made by CosmicTrack
    // and test that there are tracks in the event
    art::Handle< std::vector<rb::Track> > trks;
    evt.getByLabel(fTrackLabel,trks);
    if (trks->size() < 1) {
      mf::LogWarning("CTAWarn") << "There are " << trks->size() 
 	                        << " Tracks in this event, skip";
      return;
    }
    LOG_DEBUG("CosmicTrackAna") << "There are " << trks->size() << " Tracks in this event";

    const unsigned int trkMax = trks->size();

    if( fBT->HaveTruthInfo() )
      this->FillEventMCTruthHistograms(trkMax);

    // get a FindMany to grab the slice associated with each track
    art::FindOne<rb::Cluster> foc(trks, evt, fTrackLabel);
        
    unsigned int numtracks = 0;
    unsigned int numstop   = 0;

    // TStopwatch ts;
    // double mcTime = 0.;
    // double recoTime = 0.;

    // Loop over tracks
    for(unsigned int trkIdx = 0; trkIdx < trkMax; ++trkIdx){

      // find the slice associated to the track
      cet::maybe_ref<rb::Cluster const> sliceRef(foc.at(trkIdx));
      if( !sliceRef.isValid() ) continue;
      rb::Cluster const& slice(sliceRef.ref());
                  
      rb::Track const& track = trks->at(trkIdx);
      if(!track.Is3D()) continue;

      LOG_DEBUG("CosmicTrackAna") << "-----------------------------------------Track starts at " 
				     << track.Start().X() << " " 
				     << track.Start().Y() << " " 
				     << track.Start().Z() << "---------------------------------";

      // fill the histogram for the step sizes compared to median for a track
      if(std::abs(track.Stop().Z() - track.Start().Z()) >= 50.   &&
	 -track.Dir().Y() < 0.95*track.Dir().Mag()                 ){
	    
	std::vector<float> steps;
	float median = -999.;
	float largestStep = -999.;
	fSelection.GoodSteps(track, steps, median);
	
	if(median > 0.){
	  for(auto const& step : steps){
	    fOneDHistsRecoAll[kStepRelativeToMedian]->Fill(step/median);	 
	    largestStep = std::max(largestStep, step/median);
	  }
	  fTwoDHistsRecoAll[kStepRelativeToMedianVsAz]->Fill(fUtilities.Azimuth(track.Dir().X(), track.Dir().Z(), track.Dir().Mag()), 
							     largestStep);
	}
      }// end filling of step size histogram

      // check the quality of the reconstruction
      if(fSelection.GoodReco(track) ){

	//count number of tracks that cross fTrkLengthCut
	++numtracks;

	std::pair<CosmicTrackInfo, 
		  CosmicSignalHitFraction> cticshf = this->FillRecoInfo(track, slice);

	fTrackReco  = cticshf.first;
	fSigHitFrac = cticshf.second;

	std::map<geo::CellUniqueId, float> trueCellPathLength;
	fTrackTruth = this->FillTrueInfo(track, slice, trueCellPathLength);

	if(fTrackReco.stop) ++numstop;

	fTree->Fill();

	// ts.Start();
	if( fBT->HaveTruthInfo() && std::abs(fTrackTruth.pdg) == 13 )	  
	  this->FillTrackMCTruthHistograms(fTrackTruth, fTrackReco);

	// ts.Stop(); 
	// mcTime += ts.RealTime();
	// ts.Reset();

	LOG_DEBUG("CosmicTrackAna") << "----------------Start find track path length in cell--------------";
	this->FillTrackHistograms(fTrackReco, fSigHitFrac, track, trueCellPathLength);
	LOG_DEBUG("CosmicTrackAna") << "----------------End find track path length in cell--------------";

      }// end if passes track length in z cut

    }//end loop over tracks

    // LOG_VERBATIM("CosmicTrackAna") << "mcTime: " << mcTime
    // 				   << " recoTime: " << recoTime;
    

    fOneDHistsRecoAll[kTracksFound] ->Fill(numtracks);
    fOneDHistsRecoStop[kTracksFound]->Fill(numstop);
    
    fNTracks +=numtracks;
    
    fTrigger.ntracks = numtracks;

    return;
  }

  //......................................................................  
  void CosmicTrackAna::FillEventMCTruthHistograms(size_t trkMax)
  {

    unsigned int numtracksMC = 0;
    
    std::vector<sim::Particle> detParticles;
    
    const sim::ParticleNavigator& pnav = fBT->ParticleNavigator();
    
    // count how many tracks were left by primary particles in the event
    for(int p = 0; p < pnav.NumberOfPrimaries(); ++p){
      if( fBT->ParticleToFLSHit(pnav.Primary(p)->TrackId()).size() > 0){
	if(std::abs(pnav.Primary(p)->PdgCode()) == 13){
	  ++fNTracksMC;
	  ++numtracksMC;
	  detParticles.push_back(*(pnav.Primary(p)));
	}
      }
    }
    
    // no point filling histograms for MC if particles don't leave hits
    if (detParticles.size() < 1 ) {
      mf::LogWarning("CTAWarn") << "Failed to find any Mother particles leaving"
				<< " hits in the detector, bail";
      return;
    }
    else {
      LOG_DEBUG("CosmicTrackAna") << "There are " << detParticles.size() 
				  << " Mother particles leaving hits in the detector";

      for(auto detPart : detParticles){
	auto lengthEndPoints = fUtilities.TrueLengthEndPoints(detPart);

	fOneDHistsTruthAll[kPrimaryTrueEnergy]->Fill(detPart.E());
	fOneDHistsTruthAll[kAzimuthTrue]      ->Fill(fUtilities.Azimuth(detPart.Px(), detPart.Pz(), detPart.P())); 
	fOneDHistsTruthAll[kCosCosmicTrue]    ->Fill(fUtilities.CosTheta(detPart.Py(), detPart.P()));
	fOneDHistsTruthAll[kStartTrueX]       ->Fill(lengthEndPoints.second.first.X());
	fOneDHistsTruthAll[kStartTrueY]       ->Fill(lengthEndPoints.second.first.Y());
	fOneDHistsTruthAll[kStartTrueZ]       ->Fill(lengthEndPoints.second.first.Z());
	fOneDHistsTruthAll[kEndTrueX]         ->Fill(lengthEndPoints.second.second.X());
	fOneDHistsTruthAll[kEndTrueY]         ->Fill(lengthEndPoints.second.second.Y());
	fOneDHistsTruthAll[kEndTrueZ]         ->Fill(lengthEndPoints.second.second.Z());
	if( fSelection.Stopper( detPart.EndPosition().Vect() ) ){ 
	  fOneDHistsTruthStop[kPrimaryTrueEnergy]->Fill(detPart.E());
	  fOneDHistsTruthStop[kAzimuthTrue]      ->Fill(fUtilities.Azimuth(detPart.Px(), detPart.Pz(), detPart.P())); 
	  fOneDHistsTruthStop[kCosCosmicTrue]    ->Fill(fUtilities.CosTheta(detPart.Py(), detPart.P()));
	  fOneDHistsTruthStop[kStartTrueX]       ->Fill(lengthEndPoints.second.first.X());
	  fOneDHistsTruthStop[kStartTrueY]       ->Fill(lengthEndPoints.second.first.Y());
	  fOneDHistsTruthStop[kStartTrueZ]       ->Fill(lengthEndPoints.second.first.Z());
	  fOneDHistsTruthStop[kEndTrueX]         ->Fill(lengthEndPoints.second.second.X());
	  fOneDHistsTruthStop[kEndTrueY]         ->Fill(lengthEndPoints.second.second.Y());
	  fOneDHistsTruthStop[kEndTrueZ]         ->Fill(lengthEndPoints.second.second.Z());
	}
      }// end loop over particles (muons)
      
      double ratio = 0.;
      fTwoDHistsTruthAll[kTracksFoundVsLeft]->Fill(numtracksMC*1., trkMax*1.); // fill only for MC
      
      if(trkMax > numtracksMC) 
	LOG_VERBATIM("CosmicTrackAna") << "reconstructed more tracks than generated: "
				       << trkMax << " vs " << numtracksMC;
      
      ratio = (trkMax*1.)/(numtracksMC*1.);
      fTwoDHistsTruthAll[kTrackFindEff]->Fill(numtracksMC*1.,ratio);
      fOneDHistsTruthAll[kRecoEff]->Fill(ratio);
      
    }// end if particles were found
    
    return;
  }

  //......................................................................  
  CosmicTrackInfo CosmicTrackAna::FillTrueInfo(rb::Track   const& track,
					       rb::Cluster const& slice,
					       std::map<geo::CellUniqueId, float>& trueCellPathLength)
  {

    CosmicTrackInfo ti;

    ti.purity           = -999.;
    ti.efficiency       = -999.;
    ti.length           = -999.;
    ti.momentum         = -999.;
    ti.momentumCor      = -999.;
    ti.dcosX            = -999.;
    ti.dcosY            = -999.;
    ti.dcosZ            = -999.;
    ti.vtxX             = -999.;
    ti.vtxY             = -999.;
    ti.vtxZ             = -999.;
    ti.endX             = -999.;
    ti.endY             = -999.;
    ti.endZ             = -999.;
    ti.deltaVtxPos      = -999.;
    ti.deltaEndPos      = -999.;
    ti.cosTheta         = -999.;
    ti.azimuth          = -999.;
    ti.stop             = false;
    ti.goodVtx          = false;
    ti.vtxToEdge        = -999.;
    ti.endToEdge        = -999.;
    ti.cellsPerPlaneX   = -999.;
    ti.cellsPerPlaneY   = -999.;

    if( !fBT->HaveTruthInfo() ) return ti;
      
    // get the truth information for this track
    std::vector< const sim::Particle* > parts = fBT->HitsToParticle(track.AllCells());
    
    if(parts.size() > 0){
      
      // get the particle for the first IDE as it represents the track ID with the most 
      // energy in the collection of hits
      auto part = parts.front();
      
      // 	if(track.Dir().Z() > 0)
      // 	LOG_VERBATIM("CosmicTrackAna") << "simulation z end point: " << simEnd.Z()
      // 				       << "\ntrack z end point: " << track.Stop().Z();
      
      this->FilldEdxHistogram(*part);
      auto lengthEndPoints = fUtilities.TrueLengthEndPoints(*part);

      // get the FLSHits for this track, then make a map
      // of the cell unique id to the path length of the particle in the cell
      std::vector<sim::FLSHit> flsHits = fBT->ParticleToFLSHit(part->TrackId());
      // for( auto flsHit : flsHits)
      // 	trueCellPathLength[flsHit.GetCellUniqueId()] = flsHit.GetTotalPathLength();
      LOG_DEBUG("CosmicTrackAna") << "------------Fill true cell path lengths-------------";
      trueCellPathLength = fUtilities.TruePathLengthInCell(*part, flsHits);
      LOG_DEBUG("CosmicTrackAna") << "------------End Fill true cell path lengths-------------";
      
      // calculating purity and efficiency take forever.  
      // control whether it is done with a flag
      if(fFindPurityAndEfficiency){
	// check the efficiency and purity of the track hits for
	// this MC particle
	std::set<int> trackID;
	trackID.insert(part->TrackId());
	ti.purity     = fBT->HitCollectionPurity(trackID, track.AllCells());
	ti.efficiency = fBT->HitCollectionEfficiency(trackID, track.AllCells(), slice.AllCells(), geo::kXorY);
      }
      
      ti.pdg              = part->PdgCode();
      ti.length           = lengthEndPoints.first.first;				  
      ti.momentum   	  = lengthEndPoints.first.second;				  
      ti.momentumCor   	  = 0.;
      ti.dcosX       	  = part->Px()/part->P();
      ti.dcosY       	  = part->Py()/part->P();
      ti.dcosZ       	  = part->Pz()/part->P();
      ti.vtxX       	  = lengthEndPoints.second.first.X();			  
      ti.vtxY       	  = lengthEndPoints.second.first.Y();			  
      ti.vtxZ       	  = lengthEndPoints.second.first.Z();			  
      ti.endX       	  = lengthEndPoints.second.second.X();			  
      ti.endY       	  = lengthEndPoints.second.second.Y();			  
      ti.endZ       	  = lengthEndPoints.second.second.Z();			  
      ti.cosTheta   	  = fUtilities.CosTheta(part->Py(), part->P());						  
      ti.azimuth    	  = fUtilities.Azimuth(part->Px(), part->Pz(), part->P()); 
      ti.stop             = fSelection.Stopper(lengthEndPoints.second.second);
      ti.goodVtx          = (std::abs(fSelection.DistToEdge(track.Start())) < 10.) ? true : false;
      ti.vtxToEdge        = fSelection.DistToEdge(lengthEndPoints.second.first); 
      ti.endToEdge        = fSelection.DistToEdge(lengthEndPoints.second.second); 
      ti.cellsPerPlaneX   = 0.;
      ti.cellsPerPlaneY   = 0.;

      // fill information about the difference between the true and reconstructed end points
      TVector3 simEnd(ti.endX, ti.endY, ti.endZ);
      TVector3 simBeg(ti.vtxX, ti.vtxY, ti.vtxZ);
      TVector3 recEnd(track.Stop().X(),  track.Stop().Y(),  track.Stop().Z());
      TVector3 recBeg(track.Start().X(), track.Start().Y(), track.Start().Z());
      TVector3 diffEnd = recEnd - simEnd;
      TVector3 diffBeg = recBeg - simBeg;
      
      float diffEndSign = 1.;
      float recoDCosZ = track.Dir().Z()/track.Dir().Mag();
      if( (recoDCosZ > 0. && recEnd.Z() > simEnd.Z()) ||
	  (recoDCosZ < 0. && recEnd.Z() < simEnd.Z()) ) diffEndSign = -1.;
      
      float diffBegSign = 1.;
      if( (recoDCosZ > 0. && recBeg.Z() < simBeg.Z()) ||
	  (recoDCosZ < 0. && recBeg.Z() > simBeg.Z()) ) diffBegSign = -1.;
      
      ti.deltaVtxPos = diffBegSign*diffBeg.Mag();
      ti.deltaEndPos = diffEndSign*diffEnd.Mag();

    }

    return ti;

  }

  //......................................................................  
  void CosmicTrackAna::FillTrackMCTruthHistograms(CosmicTrackInfo const& ti,
						  CosmicTrackInfo const& ri)
  {
    

    float rangeMomUncor = ri.momentum/fUtilities.LengthCorrection(ri.length);
    float endDiffSign   = ti.deltaEndPos/std::abs(ti.deltaEndPos);
    float begDiffSign   = ti.deltaVtxPos/std::abs(ti.deltaVtxPos);
    
    fOneDHistsTruthAll[kDeltaZEnd]                 ->Fill(endDiffSign*std::abs(ri.endZ - ti.endZ));
    fOneDHistsTruthAll[kDeltaEnd]                  ->Fill(ti.deltaEndPos);						         
    fOneDHistsTruthAll[kDeltaBeg]                  ->Fill(ti.deltaVtxPos);						         
    fOneDHistsTruthAll[kDCosXDiff]                 ->Fill((ti.dcosX - ri.dcosX)/ti.dcosX);				   
    fOneDHistsTruthAll[kDCosYDiff]	           ->Fill((ti.dcosY - ri.dcosY)/ti.dcosX);				   
    fOneDHistsTruthAll[kDCosZDiff]	           ->Fill((ti.dcosZ - ri.dcosZ)/ti.dcosX);				   
    fOneDHistsTruthAll[kTrackPurity]               ->Fill(ti.purity);								   
    fOneDHistsTruthAll[kTrackEfficiency]	   ->Fill(ti.efficiency);		    					   
    fOneDHistsTruthAll[kDeltaTrackLength]          ->Fill(ti.length - ri.length);
    fOneDHistsTruthAll[kDeltaRangeMomentum]        ->Fill((std::abs(ti.momentum) - ri.momentum)/std::abs(ti.momentum));		   
    fOneDHistsTruthAll[kDeltaRangeMomentumUncor]   ->Fill((std::abs(ti.momentum) - rangeMomUncor)/std::abs(ti.momentum));	   
    fTwoDHistsTruthAll[kDCosXRecoVsTrue]	   ->Fill(ti.dcosX, ri.dcosX);						   
    fTwoDHistsTruthAll[kDCosYRecoVsTrue]	   ->Fill(ti.dcosY, ri.dcosY);						   
    fTwoDHistsTruthAll[kDCosZRecoVsTrue]	   ->Fill(ti.dcosZ, ri.dcosZ);						   
    fTwoDHistsTruthAll[kDCosXDiffVsZLen]	   ->Fill(std::abs(ri.vtxZ - ri.endZ), (ti.dcosX - ri.dcosZ)/ti.dcosX);		   
    fTwoDHistsTruthAll[kDCosYDiffVsZLen]	   ->Fill(std::abs(ri.vtxZ - ri.endZ), (ti.dcosY - ri.dcosY)/ti.dcosY);		   
    fTwoDHistsTruthAll[kDCosZDiffVsZLen]	   ->Fill(std::abs(ri.vtxZ - ri.endZ), (ti.dcosZ - ri.dcosZ)/ti.dcosZ);		   
    fTwoDHistsTruthAll[kDeltaZEndVsDCosZ]          ->Fill(endDiffSign * std::abs(ri.endZ - ti.endZ), ri.dcosZ);
    fTwoDHistsTruthAll[kDeltaZEndVsDistToEdge]     ->Fill(endDiffSign * std::abs(ri.endZ - ti.endZ), ti.endToEdge);
    fTwoDHistsTruthAll[kDeltaZBegVsDCosZ]          ->Fill(begDiffSign * std::abs(ri.vtxZ - ti.vtxZ), ri.dcosZ);
    fTwoDHistsTruthAll[kDiffBegVsDCosZ]            ->Fill(ri.dcosZ,    ti.deltaVtxPos);
    fTwoDHistsTruthAll[kDiffEndVsDCosZ]            ->Fill(ri.dcosZ,    ti.deltaEndPos);
    fTwoDHistsTruthAll[kDiffBegVsCosTheta]         ->Fill(ri.cosTheta, ti.deltaVtxPos);
    fTwoDHistsTruthAll[kDiffEndVsCosTheta]         ->Fill(ri.cosTheta, ti.deltaEndPos);
    fTwoDHistsTruthAll[kDiffBegVsAzimuth]          ->Fill(ri.azimuth,  ti.deltaVtxPos);
    fTwoDHistsTruthAll[kDiffEndVsAzimuth]          ->Fill(ri.azimuth,  ti.deltaEndPos);
    fTwoDHistsTruthAll[kDeltaRangeMomVsLength]     ->Fill(ri.length, (std::abs(ti.momentum) - ri.momentum)/std::abs(ti.momentum));
    fTwoDHistsTruthAll[kDeltaRangeMomUncorVsLength]->Fill(ri.length, (std::abs(ti.momentum) - rangeMomUncor)/std::abs(ti.momentum));
    fTwoDHistsTruthAll[kDeltaLengthVsLength]       ->Fill(ri.length, (std::abs(ti.length)   - ri.length)/std::abs(ti.length));
    fTwoDHistsTruthAll[kDeltaLengthVsPlaneAsymmetry]->Fill((ri.planesX - ri.planesY)/(ri.planesX + ri.planesY), (std::abs(ti.length) - ri.length)/std::abs(ti.length));
    fTwoDHistsTruthAll[kDeltaLengthVsDeltaStartPlane]->Fill(std::abs(ri.xStartPlane - ri.yStartPlane),     (std::abs(ti.length) - ri.length)/std::abs(ti.length));
    fTwoDHistsTruthAll[kDeltaLengthVsDeltaEndPlane]  ->Fill(std::abs(ri.xEndPlane   - ri.yEndPlane),       (std::abs(ti.length) - ri.length)/std::abs(ti.length));
    fTwoDHistsTruthAll[kDeltaLengthVsCellsPerPlane]  ->Fill(std::max(ri.cellsPerPlaneX,ri.cellsPerPlaneY), (std::abs(ti.length) - ri.length)/std::abs(ti.length));
    
    // check and see how well the reconstruction works for stoppers selected based on the reconstruction
    if( fSelection.CleanSample(ri) ){
      fOneDHistsTruthStop[kDCosXDiff]                   ->Fill((ti.dcosX - ri.dcosX)/ti.dcosX);			      
      fOneDHistsTruthStop[kDCosYDiff]		        ->Fill((ti.dcosY - ri.dcosY)/ti.dcosY);			      
      fOneDHistsTruthStop[kDCosZDiff]		        ->Fill((ti.dcosZ - ri.dcosZ)/ti.dcosZ);			      
      fOneDHistsTruthStop[kDeltaZEnd]                   ->Fill(endDiffSign*std::abs(ri.endZ - ti.endZ));
      fOneDHistsTruthStop[kDeltaEnd]                    ->Fill(ti.deltaEndPos);
      fOneDHistsTruthStop[kDeltaBeg]                    ->Fill(ti.deltaVtxPos);
      fOneDHistsTruthStop[kTrackPurity]                 ->Fill(ti.purity);							      
      fOneDHistsTruthStop[kTrackEfficiency]             ->Fill(ti.efficiency);
      fOneDHistsTruthStop[kDeltaTrackLength]            ->Fill(ti.length - ri.length);
      fOneDHistsTruthStop[kDeltaRangeMomentum]          ->Fill((std::abs(ti.momentum) - ri.momentum)/std::abs(ti.momentum));	      
      fOneDHistsTruthStop[kDeltaRangeMomentumUncor]     ->Fill((std::abs(ti.momentum) - rangeMomUncor)/std::abs(ti.momentum));
      fTwoDHistsTruthStop[kDCosXRecoVsTrue]             ->Fill(ti.dcosX, ri.dcosX);					      
      fTwoDHistsTruthStop[kDCosYRecoVsTrue]             ->Fill(ti.dcosY, ri.dcosY);					      
      fTwoDHistsTruthStop[kDCosZRecoVsTrue]             ->Fill(ti.dcosZ, ri.dcosZ);					      
      fTwoDHistsTruthStop[kDCosXDiffVsZLen]             ->Fill(std::abs(ri.vtxZ - ri.endZ), (ti.dcosX - ri.dcosX)/ti.dcosX);	      
      fTwoDHistsTruthStop[kDCosYDiffVsZLen]             ->Fill(std::abs(ri.vtxZ - ri.endZ), (ti.dcosY - ri.dcosY)/ti.dcosY);	      
      fTwoDHistsTruthStop[kDCosZDiffVsZLen]             ->Fill(std::abs(ri.vtxZ - ri.endZ), (ti.dcosZ - ri.dcosZ)/ti.dcosZ);	      
      fTwoDHistsTruthStop[kDeltaZEndVsDCosZ]            ->Fill(endDiffSign*std::abs(ri.endZ - ti.endZ), ri.dcosZ);			
      fTwoDHistsTruthStop[kDeltaZEndVsDistToEdge]       ->Fill(endDiffSign*std::abs(ri.endZ - ti.endZ), ri.endToEdge);
      fTwoDHistsTruthStop[kDeltaZBegVsDCosZ]            ->Fill(begDiffSign*std::abs(ri.vtxZ - ti.vtxZ), ri.dcosZ);	
      fTwoDHistsTruthStop[kDiffBegVsDCosZ]              ->Fill(ri.dcosZ,    ti.deltaVtxPos);
      fTwoDHistsTruthStop[kDiffEndVsDCosZ]              ->Fill(ri.dcosZ,    ti.deltaEndPos);
      fTwoDHistsTruthStop[kDiffBegVsCosTheta]           ->Fill(ri.cosTheta, ti.deltaVtxPos);
      fTwoDHistsTruthStop[kDiffEndVsCosTheta]           ->Fill(ri.cosTheta, ti.deltaEndPos);
      fTwoDHistsTruthStop[kDiffBegVsAzimuth]            ->Fill(ri.azimuth,  ti.deltaVtxPos);
      fTwoDHistsTruthStop[kDiffEndVsAzimuth]            ->Fill(ri.azimuth,  ti.deltaEndPos);
      fTwoDHistsTruthStop[kDeltaRangeMomVsLength]       ->Fill(ri.length,   (std::abs(ti.momentum) - ri.momentum)/std::abs(ti.momentum));
      fTwoDHistsTruthStop[kDeltaRangeMomUncorVsLength]  ->Fill(ri.length,   (std::abs(ti.momentum) - rangeMomUncor)/std::abs(ti.momentum));
      fTwoDHistsTruthStop[kDeltaRangeMomUncorVsCosTheta]->Fill(ri.cosTheta, (std::abs(ti.momentum) - rangeMomUncor)/std::abs(ti.momentum));
      fTwoDHistsTruthStop[kDeltaRangeMomUncorVsAzimuth] ->Fill(ri.azimuth,  (std::abs(ti.momentum) - rangeMomUncor)/std::abs(ti.momentum));
      fTwoDHistsTruthStop[kDeltaLengthVsLength]         ->Fill(ri.length,   (std::abs(ti.length) - ri.length)/std::abs(ti.length));
      fTwoDHistsTruthStop[kDeltaLengthVsPlaneAsymmetry]->Fill((ri.planesX - ri.planesY)/(ri.planesX + ri.planesY), (std::abs(ti.length) - ri.length)/std::abs(ti.length));
      fTwoDHistsTruthStop[kDeltaLengthVsDeltaStartPlane]->Fill(std::abs(ri.xStartPlane - ri.yStartPlane),     (std::abs(ti.length) - ri.length)/std::abs(ti.length));
      fTwoDHistsTruthStop[kDeltaLengthVsDeltaEndPlane]  ->Fill(std::abs(ri.xEndPlane   - ri.yEndPlane),       (std::abs(ti.length) - ri.length)/std::abs(ti.length));
      fTwoDHistsTruthStop[kDeltaLengthVsCellsPerPlane]  ->Fill(std::max(ri.cellsPerPlaneX,ri.cellsPerPlaneY), (std::abs(ti.length) - ri.length)/std::abs(ti.length));
    }

    return;
  }

  //......................................................................  
  void CosmicTrackAna::FilldEdxHistogram(simb::MCParticle const& part)
  {

    // loop over the particle trajectory points and determine the 
    // energy loss for each set
    TVector3 curPos, prevPos;
    float    totalLength = 0.;

    std::vector< std::pair<float, float> > dEdXVsTotalL;

    for(size_t t = 1; t < part.NumberTrajectoryPoints(); ++t){

      if( std::abs(part.Position(t).X()) < fDetHalfWidth  &&
	  std::abs(part.Position(t).Y()) < fDetHalfHeight &&
	  part.Position(t).Z() > 0. && part.Position(t).Z() < fDetLength &&
	  std::abs(part.Position(t-1).X()) < fDetHalfWidth  &&
	  std::abs(part.Position(t-1).Y()) < fDetHalfHeight &&
	  part.Position(t-1).Z() > 0. && part.Position(t-1).Z() < fDetLength ){

	curPos.SetXYZ(part.Position(t).X(), part.Position(t).Y(), part.Position(t).Z());
	prevPos.SetXYZ(part.Position(t-1).X(), part.Position(t-1).Y(), part.Position(t-1).Z());

	float dx = (curPos - prevPos).Mag();
	float dE = 1.e3*(part.E(t-1) - part.E(t)); // in MeV

	totalLength += dx;

	if(dE < 0. && std::abs(dE) > 1.e-3) 
	  LOG_WARNING("CosmicTrackAna") << "MC dE is negative!! " << dE 
					<< " " << dx 
					<< " " << part.E(t) << " " << part.E(t-1);
	if(dx <= 0.){
	  LOG_WARNING("CosmicTrackAna") << "MC indicates a bad dx: " << dx;
	  continue;
	}

	dEdXVsTotalL.push_back(std::pair<float, float>(dE/dx, totalLength));
	// LOG_VERBATIM("CosmicTrackAna") << dE/dx << " " << totalLength;

      }      
    }

    // loop over the vector of pairs to plot dE/dx vs range
    for(auto const& dEdxTotalL : dEdXVsTotalL){
      fTwoDHistsTruthAll[kdEdxVsRange]->Fill(totalLength - dEdxTotalL.second, dEdxTotalL.first);
      if( fSelection.Stopper(curPos) )
	fTwoDHistsTruthStop[kdEdxVsRange]->Fill(totalLength - dEdxTotalL.second, dEdxTotalL.first);
    }

    return;

  }

  //......................................................................  
  std::pair<CosmicTrackInfo, CosmicSignalHitFraction> CosmicTrackAna::FillRecoInfo(rb::Track             const& track,
										   rb::Cluster 		 const& slice)
  {
    // determine the number of planes with hits in each view
    std::set<unsigned short> xPlanes;
    std::set<unsigned short> yPlanes;
    for(size_t c = 0; c < track.NCell(); ++c){
      if     (track.Cell(c)->View() == geo::kX) xPlanes.insert(track.Cell(c)->Plane());
      else if(track.Cell(c)->View() == geo::kY) yPlanes.insert(track.Cell(c)->Plane());
    }// end loop over cells in track

    // get the residuals along the track
    std::vector<float> residuals = fUtilities.TrackResiduals(track);
    std::sort(residuals.begin(), residuals.end());
    float avResid  = 0.;
    for(auto resid : residuals) avResid += resid;

    // find the number of tricells in this track
    std::map<std::pair<uint32_t, uint32_t>, bool> planeCellMap;
    fUtilities.FindTriCells(track, planeCellMap);
    float numTriCell = 0.;
    for(auto itr : planeCellMap) 
      if(itr.second) numTriCell += 1.;

    // find the momentum from the BreakPointFitter
    bpfit::Path path(*fGeom, track.Trajectory());
    
    double bpfMom = 0.;
    path.MuonParams(0, 0, &bpfMom, 0, 0);

    CosmicTrackInfo ri;
    ri.planesX        = xPlanes.size();
    ri.planesY        = yPlanes.size();
    ri.length         = track.TotalLength();    
    ri.momentum       = fUtilities.RangeMomentum(track);
    ri.momentumCor    = fUtilities.LengthCorrection(ri.length);
    ri.momentumBPF    = bpfMom;
    ri.calorE         = track.TotalGeV();
    ri.vtxX           = track.Start().X();		    
    ri.vtxY           = track.Start().Y();		    
    ri.vtxZ           = track.Start().Z();		    
    ri.endX           = track.Stop().X();		    
    ri.endY           = track.Stop().Y();		    
    ri.endZ           = track.Stop().Z();
    ri.dcosX          = track.Dir().X()/track.Dir().Mag();
    ri.dcosY          = track.Dir().Y()/track.Dir().Mag();
    ri.dcosZ          = track.Dir().Z()/track.Dir().Mag();
    ri.cosTheta       = fUtilities.CosTheta(track.Dir().Y(), track.Dir().Mag());
    ri.azimuth        = fUtilities.Azimuth(track.Dir().X(), track.Dir().Z(), track.Dir().Mag());
    ri.vtxToEdge      = fSelection.DistToEdge(track.Start());
    ri.endToEdge      = fSelection.DistToEdge(track.Stop());  
    ri.xStartPlane    = 1. * track.MaxPlane(geo::kX);
    ri.yStartPlane    = 1. * track.MaxPlane(geo::kY);
    ri.xEndPlane      = 1. * track.MinPlane(geo::kX);
    ri.yEndPlane      = 1. * track.MinPlane(geo::kY);
    ri.cellsPerPlaneX = (ri.planesX > 0.) ? (1. * track.NXCell())/ri.planesX : -999.;
    ri.cellsPerPlaneY = (ri.planesY > 0.) ? (1. * track.NYCell())/ri.planesY : -999.;
    ri.avgResidual    = avResid/track.NCell();						   
    ri.maxResidual    = residuals.back();							   
    ri.triCellFrac    = 1.*numTriCell/(1.*track.NCell());					   
    ri.stop           = fSelection.Stopper(track.Stop());					   
    ri.goodVtx        = (std::abs(fSelection.DistToEdge(track.Start())) < 10.) ? true : false;
    ri.isCalibTrk     = fSelection.IsCalibTrack(track, slice);                                

    float trkcells  = 0.;
    float trksig    = 0.;
    float totcell   = 0.;
    float totsig    = 0.;
    float xsig      = 0.;
    float ysig      = 0.;

    // loop over the cell hits from the cluster
    for(size_t i = 0; i < slice.NCell(); ++i){
      art::Ptr<rb::CellHit> chit = slice.Cell(i);
      if(chit->PE() > fMinSig){
	fOneDHistsRecoAll[kCellHitSig]->Fill(chit->PE()); 
	if( fSelection.CleanSample(ri) ) fOneDHistsRecoStop[kCellHitSig]->Fill(chit->PE()); 
	totcell += 1.;
	totsig  += chit->PE();
      }
    }

    // get the total signal in each view of the track
    for(size_t c = 0; c < track.NCell(); ++c){
      if(track.Cell(c)->PE() > fMinSig){
	trkcells += 1.;
	if     (track.Cell(c)->View() == geo::kX) xsig += track.Cell(c)->PE()*track.Weight(c);
	else if(track.Cell(c)->View() == geo::kY) ysig += track.Cell(c)->PE()*track.Weight(c);
	//LOG_VERBATIM("CosmicTrackAna") << xsig << " " << ysig << " " << " " << track.Cell(c)->PE() << " " << track.Weight(c);
      }
    }

    trksig   = xsig + ysig;  

    CosmicSignalHitFraction shf;
    shf.signalFraction = trksig/totsig;
    shf.hitFraction    = trkcells/totcell;
    shf.signal         = trksig;
    shf.hitsX          = track.NXCell();
    shf.hitsY          = track.NYCell();
    shf.totcell        = totcell;
    shf.signalX        = xsig;
    shf.signalY        = ysig;

    return std::pair<CosmicTrackInfo, CosmicSignalHitFraction>(ri, shf);
  }

  //......................................................................  
  void CosmicTrackAna::FillTrackHistograms(CosmicTrackInfo                    const& ri,	
					   CosmicSignalHitFraction 	      const& shf,	
					   rb::Track               	      const& track,
					   std::map<geo::CellUniqueId, float> const& trueCellPathLength)
  {

    fOneDHistsRecoAll[kTrkSig]             ->Fill(shf.signal);	   
    fOneDHistsRecoAll[kTrkCells]	   ->Fill(shf.hitsX+shf.hitsY);	   
    fOneDHistsRecoAll[kDCosX]   	   ->Fill(ri.dcosX);   
    fOneDHistsRecoAll[kDCosY]   	   ->Fill(ri.dcosY);   
    fOneDHistsRecoAll[kDCosZ]      	   ->Fill(ri.dcosZ);   
    fOneDHistsRecoAll[kAzimuth]            ->Fill(ri.azimuth);		     
    fOneDHistsRecoAll[kCosCosmic]  	   ->Fill(ri.cosTheta);	     
    fOneDHistsRecoAll[kDirX]               ->Fill(ri.dcosX);
    fOneDHistsRecoAll[kDirY]               ->Fill(ri.dcosY);
    fOneDHistsRecoAll[kDirZ]               ->Fill(ri.dcosZ);      
    fOneDHistsRecoAll[kStartX]     	   ->Fill(ri.vtxX); 	     
    fOneDHistsRecoAll[kStartY]     	   ->Fill(ri.vtxY);	     
    fOneDHistsRecoAll[kStartZ]     	   ->Fill(ri.vtxZ);	     
    fOneDHistsRecoAll[kEndX]       	   ->Fill(ri.endX);		     
    fOneDHistsRecoAll[kEndY]       	   ->Fill(ri.endY);		     
    fOneDHistsRecoAll[kEndZ]       	   ->Fill(ri.endZ);		     
    fOneDHistsRecoAll[kTrackLength]	   ->Fill(ri.length);               
    fOneDHistsRecoAll[kTrackRangeMomentum]->Fill(ri.momentum);
    fOneDHistsRecoAll[kDiffX]             ->Fill(std::abs(ri.vtxX - ri.endX));
    fOneDHistsRecoAll[kDiffY]             ->Fill(std::abs(ri.vtxY - ri.endY));
    fOneDHistsRecoAll[kDiffZ]             ->Fill(std::abs(ri.vtxZ - ri.endZ));
    fOneDHistsRecoAll[kFractionHitsUsed]  ->Fill(shf.hitFraction);
    fOneDHistsRecoAll[kFractionSigUsed]   ->Fill(shf.signalFraction);  
    fOneDHistsRecoAll[kVtxDistToEdge]     ->Fill(ri.vtxToEdge);
    fOneDHistsRecoAll[kEndDistToEdge]     ->Fill(ri.endToEdge);
    fOneDHistsRecoAll[kCellsPerPlaneX]    ->Fill(ri.cellsPerPlaneX);
    fOneDHistsRecoAll[kCellsPerPlaneY]    ->Fill(ri.cellsPerPlaneY);
    fTwoDHistsRecoAll[kDirXY]             ->Fill(ri.dcosX,ri.dcosY);
    fTwoDHistsRecoAll[kStartXY]    	  ->Fill(ri.vtxX, ri.vtxY); 
    fTwoDHistsRecoAll[kStartZX]    	  ->Fill(ri.vtxZ, ri.vtxX); 
    fTwoDHistsRecoAll[kStartZY]    	  ->Fill(ri.vtxZ, ri.vtxY); 
    fTwoDHistsRecoAll[kEndXY]      	  ->Fill(ri.endX, ri.endY);     
    fTwoDHistsRecoAll[kEndZX]      	  ->Fill(ri.endZ, ri.endX);     
    fTwoDHistsRecoAll[kEndZY]      	  ->Fill(ri.endZ, ri.endY);     
    fTwoDHistsRecoAll[kFracHitsUsedVsNum] ->Fill(shf.totcell, shf.hitFraction);
    fTwoDHistsRecoAll[kCellsPerPlaneXY]   ->Fill(ri.cellsPerPlaneX, ri.cellsPerPlaneY);

    if( fSelection.CleanSample(ri) ){
      fOneDHistsRecoStop[kTrkSig]            ->Fill(shf.signal);	   
      fOneDHistsRecoStop[kTrkCells]	     ->Fill(shf.hitsX+shf.hitsY);	   
      fOneDHistsRecoStop[kDCosX]   	     ->Fill(ri.dcosX);   
      fOneDHistsRecoStop[kDCosY]   	     ->Fill(ri.dcosY);   
      fOneDHistsRecoStop[kDCosZ]      	     ->Fill(ri.dcosZ);   
      fOneDHistsRecoStop[kAzimuth]           ->Fill(ri.azimuth);		     
      fOneDHistsRecoStop[kCosCosmic]  	     ->Fill(ri.cosTheta);	     
      fOneDHistsRecoStop[kDirX]              ->Fill(ri.dcosX);
      fOneDHistsRecoStop[kDirY]              ->Fill(ri.dcosY);
      fOneDHistsRecoStop[kDirZ]              ->Fill(ri.dcosZ);      
      fOneDHistsRecoStop[kStartX]     	     ->Fill(ri.vtxX);      	     
      fOneDHistsRecoStop[kStartY]     	     ->Fill(ri.vtxY);      	     
      fOneDHistsRecoStop[kStartZ]     	     ->Fill(ri.vtxZ);      	     
      fOneDHistsRecoStop[kEndX]       	     ->Fill(ri.endX);      		     
      fOneDHistsRecoStop[kEndY]       	     ->Fill(ri.endY);      		     
      fOneDHistsRecoStop[kEndZ]       	     ->Fill(ri.endZ);      		     
      fOneDHistsRecoStop[kTrackLength]	     ->Fill(ri.length);           
      fOneDHistsRecoStop[kTrackRangeMomentum]->Fill(ri.momentum);
      fOneDHistsRecoStop[kDiffX]             ->Fill(std::abs(ri.vtxX - ri.endX));
      fOneDHistsRecoStop[kDiffY]             ->Fill(std::abs(ri.vtxY - ri.endY));
      fOneDHistsRecoStop[kDiffZ]             ->Fill(std::abs(ri.vtxZ - ri.endZ));
      fOneDHistsRecoStop[kFractionHitsUsed]  ->Fill(shf.hitFraction);
      fOneDHistsRecoStop[kFractionSigUsed]   ->Fill(shf.signalFraction);  
      fOneDHistsRecoStop[kVtxDistToEdge]     ->Fill(ri.vtxToEdge);
      fOneDHistsRecoStop[kEndDistToEdge]     ->Fill(ri.endToEdge);
      fOneDHistsRecoStop[kCellsPerPlaneX]    ->Fill(ri.cellsPerPlaneX);
      fOneDHistsRecoStop[kCellsPerPlaneY]    ->Fill(ri.cellsPerPlaneY);
      fTwoDHistsRecoStop[kDirXY]             ->Fill(ri.dcosX,ri.dcosY);
      fTwoDHistsRecoStop[kStartXY]    	     ->Fill(ri.vtxX, ri.vtxY); 
      fTwoDHistsRecoStop[kStartZX]    	     ->Fill(ri.vtxZ, ri.vtxX); 
      fTwoDHistsRecoStop[kStartZY]    	     ->Fill(ri.vtxZ, ri.vtxY); 
      fTwoDHistsRecoStop[kEndXY]      	     ->Fill(ri.endX, ri.endY);     
      fTwoDHistsRecoStop[kEndZX]      	     ->Fill(ri.endZ, ri.endX);     
      fTwoDHistsRecoStop[kEndZY]      	     ->Fill(ri.endZ, ri.endY);     
      fTwoDHistsRecoStop[kFracHitsUsedVsNum] ->Fill(shf.totcell, shf.hitFraction);
      fTwoDHistsRecoStop[kCellsPerPlaneXY]   ->Fill(ri.cellsPerPlaneX, ri.cellsPerPlaneY);
    }

    // make a map of the z position of each trajectory point of the
    // track to its corresponding trajectory point.  the map will
    // be used to find the residual of the reconstruction to the 
    // cell hits.
    // TVector3 cellCenter;
    // TVector3 cellExtent;
    // TVector3 lowerBound;
    // TVector3 upperBound;
    TVector3 point;
    float    distFromEnd      = 0.;
    float    lengthInCell     = 0.;
    float    trueLengthInCell = 0.;
    // double   xyz[3]           = {0.};

    auto trueCellItr = trueCellPathLength.begin();
    zBoundMap zToPlaneBounds = fUtilities.MakeZBoundaryMap(track.Trajectory());

    std::map< std::pair<uint32_t, uint32_t>, bool > triCellMap;
    std::pair<uint32_t, uint32_t> pc;
    double cellWidth = 0.;
    fUtilities.FindTriCells(track, triCellMap);

    auto residuals = fUtilities.TrackResiduals(track);

    // Loop through both views and grab every CellHit in this track
    for(size_t c = 0; c < track.NCell(); ++c){	
      const art::Ptr<rb::CellHit> chit = track.Cell(c);
      const geo::View_t view = chit->View();
      
      pc.first  = chit->Plane();
      pc.second = chit->Cell();
      cellWidth = 2. * fGeom->Plane(pc.first)->Cell(pc.second)->HalfW();

      trueCellItr = trueCellPathLength.find(fGeom->Plane(pc.first)->Cell(pc.second)->Id());
      if( trueCellItr != trueCellPathLength.end() )
	trueLengthInCell = trueCellItr->second;
      else
	trueLengthInCell = -9999.;

      // LOG_VERBATIM("CosmicTrackAna") << chit->Plane() << " " << chit->Cell() <<  "cell hit view is " << view;

      if(view == geo::kX){
	fOneDHistsRecoAll[kCellHitX]  ->Fill(pc.second);
	fOneDHistsRecoAll[kPlaneHitX] ->Fill(pc.first);
	fTwoDHistsRecoAll[kCellPlaneX]->Fill(pc.first, pc.second);
	if( fSelection.CleanSample(ri) ){
	  fOneDHistsRecoStop[kCellHitX]  ->Fill(pc.second);
	  fOneDHistsRecoStop[kPlaneHitX] ->Fill(pc.first);
	  fTwoDHistsRecoStop[kCellPlaneX]->Fill(pc.first, pc.second);
	}
      }
      else if(view == geo::kY){
	fOneDHistsRecoAll[kCellHitY]  ->Fill(pc.second);
	fOneDHistsRecoAll[kPlaneHitY] ->Fill(pc.first);
	fTwoDHistsRecoAll[kCellPlaneY]->Fill(pc.first, pc.second);
	if( fSelection.CleanSample(ri) ){
	  fOneDHistsRecoStop[kCellHitY]  ->Fill(pc.second);
	  fOneDHistsRecoStop[kPlaneHitY] ->Fill(pc.first);
	  fTwoDHistsRecoStop[kCellPlaneY]->Fill(pc.first, pc.second);
	}
      }

      const rb::RecoHit rhit = track.RecoHit(chit);
      if(!rhit.IsCalibrated()) continue;

      point = TVector3(rhit.X(), rhit.Y(), rhit.Z());
      distFromEnd = track.DistanceFromEnd(rhit.Z());

      if(view == geo::kX){
	fOneDHistsRecoAll[kResidualsX] ->Fill(residuals[c]);
	fOneDHistsRecoAll[kTNSX]       ->Fill(chit->TNS());
	fTwoDHistsRecoAll[kResidVsCosX]->Fill(residuals[c], ri.cosTheta);
	fTwoDHistsRecoAll[kResidVsAzX] ->Fill(residuals[c], ri.azimuth);
      }
      else if(view == geo::kY){
	fOneDHistsRecoAll[kResidualsY] ->Fill(residuals[c]);
	fOneDHistsRecoAll[kTNSY]       ->Fill(chit->TNS());
	fTwoDHistsRecoAll[kResidVsCosY]->Fill(residuals[c], ri.cosTheta);
	fTwoDHistsRecoAll[kResidVsAzY] ->Fill(residuals[c], ri.azimuth);
      }
      fOneDHistsRecoAll[kHitXPosition] ->Fill(rhit.X());
      fOneDHistsRecoAll[kHitYPosition] ->Fill(rhit.Y());
      fOneDHistsRecoAll[kHitZPosition] ->Fill(rhit.Z());
      fTwoDHistsRecoAll[kHitXYPosition]->Fill(rhit.X(), rhit.Y());
      fTwoDHistsRecoAll[kHitXZPosition]->Fill(rhit.Z(), rhit.X());
      fTwoDHistsRecoAll[kHitYZPosition]->Fill(rhit.Z(), rhit.Y());	
      if( fSelection.CleanSample(ri) ){
	if(distFromEnd >= 100. && distFromEnd <= 200.){
	  // find the pathlength through the cell
	  try{
	    lengthInCell = fUtilities.PathLengthInCell(zToPlaneBounds, point, pc);
	    fCosmicHit.pathLength        = lengthInCell;
	    fCosmicHit.truePathLength    = trueLengthInCell;
	    fCosmicHit.triCellPathLength = (triCellMap[pc]) ? fUtilities.TriCellPathLength(track, pc.first, cellWidth, view) : -9999.;
	    fCosmicHit.PE                = chit->PE();			
	    fCosmicHit.PECorr         	 = rhit.PECorr();			
	    fCosmicHit.distFromEnd    	 = distFromEnd;			
	    fCosmicHit.view           	 = view;				
	    fCosmicHit.cell           	 = pc.second;			
	    fCosmicHit.plane          	 = pc.first;			
	    fCosmicHit.module         	 = std::floor(fCosmicHit.cell/32.);
	    if(view == geo::kX){
	      fCosmicHit.w = rhit.Y();
	      fTwoDHistsRecoStop[kPEPerCMVsWX]->Fill(rhit.Y(), rhit.PECorr()/lengthInCell);
	    }
	    else if(view == geo::kY){
	      fCosmicHit.w = rhit.X();
	      fTwoDHistsRecoStop[kPEPerCMVsWY]->Fill(rhit.X(), rhit.PECorr()/lengthInCell);
	    }
	    fHitTree->Fill();
	  }
	  catch(...){}
	} // end if in the correct window from end point of track
	if(view == geo::kX){
	  fOneDHistsRecoStop[kResidualsX] ->Fill(residuals[c]);
	  fOneDHistsRecoStop[kTNSX]       ->Fill(chit->TNS());
	  fTwoDHistsRecoStop[kResidVsCosX]->Fill(residuals[c], ri.cosTheta);
	  fTwoDHistsRecoStop[kResidVsAzX] ->Fill(residuals[c], ri.azimuth);
	}
	else if(view == geo::kY){
	  fOneDHistsRecoStop[kResidualsY] ->Fill(residuals[c]);
	  fOneDHistsRecoStop[kTNSY]       ->Fill(chit->TNS());
	  fTwoDHistsRecoStop[kResidVsCosY]->Fill(residuals[c], ri.cosTheta);
	  fTwoDHistsRecoStop[kResidVsAzY] ->Fill(residuals[c], ri.azimuth);
	}
	fOneDHistsRecoStop[kHitXPosition] ->Fill(rhit.X());
	fOneDHistsRecoStop[kHitYPosition] ->Fill(rhit.Y());
	fOneDHistsRecoStop[kHitZPosition] ->Fill(rhit.Z());
	fTwoDHistsRecoStop[kHitXYPosition]->Fill(rhit.X(), rhit.Y());
	fTwoDHistsRecoStop[kHitXZPosition]->Fill(rhit.Z(), rhit.X());
	fTwoDHistsRecoStop[kHitYZPosition]->Fill(rhit.Z(), rhit.Y());	
      }

    } // end for cells
      
//     if(cosCosmic < 0.1){
//       LOG_VERBATIM("CosmicTrackAna") << "track params " << track.TotalLength() << " " << cosCosmic << " " << azimuth
// 				     << "\ndirection " << dir.X() << " " << dir.Y() << " " << dir.Z()
// 				     << "\nminPlane: " << track.MinPlane() << " maxPlane: " << track.MaxPlane();
//       for(size_t p = 0; p < track.NTrajectoryPoints(); ++p){
// 	TVector3 point = track.TrajectoryPoint(p);
// 	LOG_VERBATIM("CosmicTrackAna") << point.X() << " " << point.Y() << " " << point.Z();
//       }
//     }
                  
    return;
  }

  //......................................................................  
  void CosmicTrackAna::reconfigure(const fhicl::ParameterSet& pset)
  {


    fTrackLabel               = pset.get< std::string         >("TrackLabel"                     );
    fExposureLabel            = pset.get< std::string 	      >("ExposureLabel"                  );
    fMinSig                   = pset.get< double      	      >("MinSig"                         );
    fFindPurityAndEfficiency  = pset.get< bool        	      >("FindPurityAndEfficiency",  false);

    fSelection.reconfigure(pset.get<fhicl::ParameterSet>("SelectionCriteria"));
    fUtilities.reconfigure(pset.get<fhicl::ParameterSet>("Utilities"        ));

    return;
  } 
  
  //......................................................................  
  void CosmicTrackAna::endJob()
  { 
    fExposure->Fill(fLiveTime);

    if(fNevents != 0 && fLiveTime != 0){
      float cosmicrate = 1.e-3*fNTracks/fLiveTime;

      LOG_VERBATIM("CosmicTrackAna") << "There are: " << fNTracks 
				     << " cosmic tracks selected in this job.\n"
				     << "Livetime is " << fLiveTime 
				     << " according to CosmicExposure.\n"
				     << "Note, livetime is for FULL SUBRUN, "
				     << "so rate is wrong if haven't run full subrun.\n"
				     << "The cosmic rate is therefore " 
				     << cosmicrate << " Hz \n";
      
      fCosmicRate->Fill(cosmicrate);
    }

    return;
  } 


  
} // end namespace trk

namespace trk{
  DEFINE_ART_MODULE(CosmicTrackAna);
}

