///////////////////////////////////////////////////////////////////////
/// \file WindowTrackingAlg.h
/// \brief   Track finder for cosmic rays
/// \author  brebel@fnal.gov
/// \date
///////////////////////////////////////////////////////////////////////
#ifndef TRACKFIT_WINDOWTRACKINGALG_H
#define TRACKFIT_WINDOWTRACKINGALG_H

// C++ includes
#include <vector>

// ART includes
#include "art/Persistency/Common/Ptr.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "Geometry/Geometry.h"
#include "TrackFit/TrackAlg.h"

#include "TH1.h"
#include "TVector3.h"

namespace fhicl { class ParameterSet; }
namespace rb {
  class CellHit;
  class Track;
  class Cluster;
}

namespace trk {

  /// An algorithm to perform cosmic ray track fitting
  class WindowTrackingAlg : public TrackAlg{
  public:
    explicit WindowTrackingAlg(fhicl::ParameterSet const &pset);
    virtual ~WindowTrackingAlg();

    // Take a slice and return a track
    std::vector<rb::Track> MakeTrack( const rb::Cluster* slice );

  private:

    rb::Track    Make3DTrack(std::vector<art::Ptr<rb::CellHit> > & allHits,
			     std::vector<art::Ptr<rb::CellHit> > & removedHits,
			     double                              & chiSqr,
			     bool(*comp)(unsigned short int, unsigned short int));
			 
    rb::Track    FitView(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >, 
			          bool(*)(unsigned short int, unsigned short int) >         & hits,
			 std::map<unsigned short int, std::pair<float, float>, 
			          bool(*)(unsigned short int, unsigned short int) >    const& planeCoG,
			 std::vector<art::Ptr<rb::CellHit> >                                & removedHits,
			 geo::View_t                                                          view,
			 double                                                             & chiSqr);

    rb::Track    MakeViewTrack(std::vector< art::Ptr<rb::CellHit> > const& trkHits,
			       geo::View_t                          const& view,
			       double                                    & chiSqr,
			       bool                                 const& dir,
			       std::vector< art::Ptr<rb::CellHit> >      & removedHits);

    void         LookForBremsstrahlungHits(rb::Track                           & track,
					   std::vector<art::Ptr<rb::CellHit> > & removedHits);

    unsigned int FitWindow(std::vector<art::Ptr<rb::CellHit> >                        const& prevWindow,
			   std::vector<art::Ptr<rb::CellHit> >                             & currWindow,
			   std::vector<art::Ptr<rb::CellHit> >                             & removedHits,
			   std::map<unsigned short int, std::pair<float, float>, 
			            bool(*)(unsigned short int, unsigned short int) > const& planeCoG,
			   geo::View_t                                                const& view);

    rb::Track    ShortTrack(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
			             bool(*)(unsigned short int, unsigned short int) >       & xViewHits,
			    std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
			             bool(*)(unsigned short int, unsigned short int) >       & yViewHits,
			    std::map<unsigned short int, std::pair<float, float>,
                                     bool(*)(unsigned short int, unsigned short int) >  const& planeCoG,
                            double                                                           & chiSqr);

    rb::Track    ShortViewTrack(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
			                 bool(*)(unsigned short int, unsigned short int) >       & viewHits,
				std::map<unsigned short int, std::pair<float, float>,
				         bool(*)(unsigned short int, unsigned short int) >  const& planeCoG,
				bool                                                        const& twoPlanes,
				int                                                         const& firstPlane,
				int                                                         const& secondPlane,
				geo::View_t                                                 const& view);
    void         ShortTrackExtraPlane(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
				         bool(*)(unsigned short int, unsigned short int) >   const& viewHits,
				      std::map<unsigned short int, std::pair<float, float>,
				         bool(*)(unsigned short int, unsigned short int) >  const& planeCoG,
				      std::vector<std::pair<geo::View_t, TVector3> >             & trajPoints,
				      rb::Track                                                  & viewTrack,
				      unsigned short int                                    const& removedPlane);
    void         ShortTrackRemoveExtraPlane(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
				               bool(*)(unsigned short int, unsigned short int) >       & viewHits,
					    std::map<unsigned short int, std::pair<float, float>,
					       bool(*)(unsigned short int, unsigned short int) >       & planeCoG,
					    unsigned short int                                         & removedPlane);
				      

    void         SetTrackEndPoints(rb::Track & track);
    void         FindEndPoint(geo::View_t const& view, 
			      double           * cellCenter,
			      TVector3    const& endPoint,
			      TVector3    const& endDir,
			      bool        const& start,
			      rb::Track        & track);

    void         CheckTrackDirectionInY(rb::Track & track,
					TVector3  & start,
					TVector3  & end);

    void         DetermineInitialDirection(rb::Track & track);

    bool         ValidTrack(std::map<unsigned short int, std::vector<art::Ptr<rb::CellHit> >,
			             bool(*)(unsigned short int, unsigned short int) >        const& planeToHit);

    double                            fCellHalfDepth;      ///< half of cell size in z
    double                            fDHitGood;           ///< Maximum distance from the line for a 
                                                           ///< hit to be considered part of it
    unsigned int                      fWindowSize;         ///< Number of planes in the sliding window                  
    unsigned int                      fMinViewPlanes;      ///< Minimum number of planes in a view to fit
    unsigned int                      fMinHitsTryAgain;    ///< Minimum number of hits left to look for another track
    double                            fMinHitFraction;     ///< Minimum value of hits/plane to look for a track
    unsigned int                      fNTrajectory;        ///< number of trajectory points to use in determining incoming direction
    int                               fMaxPlaneSeparation; ///< Maximum number of planes allowed to separate planes with hits in slice
    art::ServiceHandle<geo::Geometry> fGeo;                ///< handle to the geometry

  };

}

#endif // TRACKFIT_WINDOWTRACKINGALG_H
///////////////////////////////////////////////////////////////////////
