////////////////////////////////////////////////////////////////////////
/// \brief   This is the version 2, works for multi-track event
/// \author  ricken@fnal.gov
/// \date    Sept. 2012
////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <vector>

#include "TVector3.h"
#include "TNtuple.h"

#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"

#include "RecoBase/Track.h"
#include "RecoBase/Vertex.h"
#include "VertexFinder/DOCAInfo.h"


namespace vf {
  class VertexFinder : public art::EDProducer
  {
  public:
    explicit VertexFinder(fhicl::ParameterSet const &pset);
    virtual ~VertexFinder();
    
    void produce(art::Event &evt);
    void beginJob();
    void reconfigure(fhicl::ParameterSet const &pset);
    
    static const int GAMMAID;
    static const int PIZEROID;
    static const int PIPLUSID;
    static const int PIMINUSID;
    static const int NUEID;
    static const int NUMUID;
    static const int NUTAUID;
    
    
  private:
    
    TNtuple*    ntuple;
    
    int         fMinTrack;
    bool        fPi0Reco;
    double      fXMax;
    double      fYMax;
    double      fZMin;
    double      fZMax;  
    double      fDOCA_cut;
    double      fAngle_cut;
  };
}


namespace vf
{
  //................................................................
  const int VertexFinder::GAMMAID = 22;
  const int VertexFinder::PIZEROID = 111;
  const int VertexFinder::PIPLUSID = 211;
  const int VertexFinder::PIMINUSID = -211;
  const int VertexFinder::NUEID = 12;
  const int VertexFinder::NUMUID = 14;
  const int VertexFinder::NUTAUID = 16;

  //................................................................
  VertexFinder::VertexFinder(fhicl::ParameterSet const& pset) 
  {
    reconfigure(pset);
    produces< std::vector<rb::Vertex> >();
  }
  

  //......................................................................
  void VertexFinder::reconfigure(fhicl::ParameterSet const &pset)
  {
    fPi0Reco        =pset.get< bool        >("Pi0Reco"    );   
    fMinTrack       =pset.get< int         >("MinTrack"   );
    fXMax           =pset.get< double      >("XMax"       );
    fYMax           =pset.get< double      >("YMax"       );
    fZMin           =pset.get< double      >("ZMin"       );
    fZMax           =pset.get< double      >("ZMax"       );
    fDOCA_cut       =pset.get< double      >("DOCA_cut"   );
    fAngle_cut      =pset.get< double      >("Angle_cut"  );
  }

  //......................................................................
  VertexFinder::~VertexFinder()
  {
  }
 
  //......................................................................
  void VertexFinder::beginJob()
  {
  }
  
  //.........................................................................
  void VertexFinder::produce(art::Event& evt)
  {

    std::unique_ptr< std::vector<rb::Vertex> > vert(new std::vector<rb::Vertex> );
    
    art::Handle< std::vector<rb::Track> > trackhandle;
    try{ evt.getByLabel("trackmerge",trackhandle); }
    catch(...){ std::cout<<"couldn't get the trackhandle "<<std::endl; return;}
    
    TVector3 mc_vertex,rc_vertex, delta;
    double doca, angle_reco;      
    std::vector<int> trackID;
    
    int NbTrack = trackhandle->size();
    if (NbTrack<fMinTrack) return;
    
    //Get vertex time: finding the earliest cell hit in the track   
    

    double vertime;
    art::Ptr<rb::Track> track(trackhandle,0);
    vertime = track->Cell(0)->TNS();
    int cell_Nb = track->NCell();
    for (int idx =1 ; idx!=cell_Nb; idx++){
      if (vertime > track->Cell(idx)->TNS()) 
	vertime = track->Cell(idx)->TNS();
    }

    if (NbTrack==1) rc_vertex = track->Start();
 
    // Let's do this with the qualified Tracks:
    int combNB=0;
    for (int k=0; k<NbTrack; k++) trackID.push_back(k);

    //    rc_vertex.SetXYZ(0.,0.,0.);     
    
    for (int i=0; i<NbTrack-1;i++){
      for(int j=i+1;j<NbTrack;j++ )
	{
	  
	  art::Ptr<rb::Track> track1(trackhandle, trackID[i]);
	  art::Ptr<rb::Track> track2(trackhandle, trackID[j]);
	  DOCAInfo doca_info(*track1, *track2);
	  
	  doca       = doca_info.Doca();
	  angle_reco = doca_info.AngleReco();

          //Choosing right combinations
	  if (doca > fDOCA_cut                        ||
              fabs(angle_reco) > fAngle_cut           || 
              fabs(doca_info.VertexReco().x())>fXMax || 
              fabs(doca_info.VertexReco().y())>fYMax ||
              doca_info.VertexReco().z()<fZMin       ||
	      doca_info.VertexReco().z()>fZMax         )   continue;

	  else { 
	    rc_vertex = rc_vertex+doca_info.VertexReco();
	    combNB++; 
	  }

	}
    }


    if (combNB<1) return;
    rc_vertex  = rc_vertex*(1.0/(combNB+0.0)); 
    
    vert->push_back(rb::Vertex(rc_vertex,vertime));
    
    evt.put(std::move(vert));
    return;
  }
  
  DEFINE_ART_MODULE(VertexFinder);
  
  
} // end namespace vf
//..........................................................................
