////////////////////////////////////////////////////////////////////////
// $Id: G4Alg.cxx,v 1.6 2012-12-03 23:52:01 rhatcher Exp $
//
// Geant4 Driver Module
//
// brebel@fnal.gov
//
////////////////////////////////////////////////////////////////////////
#include <cassert>
#include <cstdlib>
#include <string>
#include <sstream>
#include <sys/stat.h>
#include <vector>
#include <map>
#include <unistd.h>

// NOvA includes
#include "Geometry/Geometry.h"
#include "g4nova/G4Alg.h"
#include "g4nova/FLSHitListAction.h"
#include "g4nova/ParticleListAction.h"
#include "G4Base/UserActionManager.h"
#include "G4Base/UserActionFactory.h"
#include "G4Base/G4Helper.h"
#include "SimulationBase/MCTruth.h"
#include "Simulation/FLSHitList.h"
#include "Simulation/Particle.h"

// Framework includes
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib/search_path.h"

// Geant4 includes
#include "Geant4/G4ParticleTable.hh"

namespace g4n {

  //____________________________________________________________________________
  // Constructor
  G4Alg::G4Alg(fhicl::ParameterSet const& pset)
    : fG4Help(0)
    , fG4PhysListName     (pset.get< std::string     >("G4PhysListName", "QGSP_BERT"  ) )
    , fEnergyThresh       (pset.get< double          >("G4EnergyThreshold")  )
    , fManyParticles      (pset.get< bool            >("ManyParticles")  )
    , fSparseTrajectories (pset.get< bool            >("SparseTrajectories") )
    , fIsRockVetoUsed     (pset.get< bool            >("G4CheckRockVeto")    )
    , fSlimSimParticles   (pset.get< bool            >("SlimSimParticles")   )
    , fGenModuleLabel     (pset.get< std::string     >("GenModuleLabel")     )
    , fPlaIndex(0)
    , fFlsIndex(0)
  {

    /// In principle the next two should just be always the start of the 
    /// UserAction list, but for now handle these separately until they 
    /// learn to config from Config()
    //    fUserActions.push_back("ParticleListAction");
    //  fUserActions.push_back("FLSHitListAction");

    /// dummy vector in case user didn't set "AddedUserActions" in fcl file
    std::vector<std::string>  nullstrvec;  // empty vector of strings
    const std::vector<std::string>& psetstrvec = 
      pset.get< std::vector<std::string> >("AddedUserActions",nullstrvec);

    for (size_t j=0; j < psetstrvec.size(); ++j) {
      fUserActions.push_back(psetstrvec[j]); // append user entries to list of actions
    }

    /// dummy vectors in case user didn't set "OverrideLifetimePDG"
    /// or "OverrideLifetimeNS"
    std::vector<int>    nullintvec;
    std::vector<double> nulldblvec;
    fOverrideLifetimePDG = pset.get< std::vector<int>    >("OverrideLifetimePDG",nullintvec);
    fOverrideLifetimeNS  = pset.get< std::vector<double> >("OverrideLifetimeNS",nulldblvec);
    size_t npdg  = fOverrideLifetimePDG.size();
    size_t ntime = fOverrideLifetimeNS.size();

    if ( npdg != ntime ) {
      mf::LogError("G4Alg")
        << "Attempt to override PDG lifetime in Geant4 failed\n"
        << "due to size mismatch: "
        << npdg << " PDG vs. " << ntime << " times";

      fOverrideLifetimePDG.clear();
      fOverrideLifetimeNS.clear();

      throw cet::exception("OverrideLifetimePairMismatch")
        << "Can't override G4 particle lifetimes without matching set: "
        << npdg << " PDG vs. " << ntime << " times" << "\n"
        << __FILE__ << ":" << __LINE__ << "\n";
    }

    // Constructor decides if initialized value is a path or 
    // an environment variable
    cet::search_path sp("FW_SEARCH_PATH");

    sp.find_file(pset.get< std::string >("G4MacroPath"), fG4MacroPath);
    struct stat sb;
    if ( fG4MacroPath.empty() || stat(fG4MacroPath.c_str(), &sb)!=0 ) {
      // failed to resolve the file name
      throw cet::exception("NoG4Macro")
        << "G4 macro file " << fG4MacroPath << " not found!\n"
        << __FILE__ << ":" << __LINE__ << "\n";
    }

    // need to instantiate the G4Helper in order to make the 
    // G4RunManager and pass it the physics list - G4 really dislikes
    // it if you try to make a G4UserRunAction (ie UserActionManager)
    // before passing a physics list to the run manager
    art::ServiceHandle<geo::Geometry> geo;
    fG4Help = new g4b::G4Helper(fG4MacroPath,
				fG4PhysListName,
				geo->GDMLFile());

    // more control over GDML processing (before InitPhysics)
    const bool dfltOverlap = false;
    const bool dfltSchema  = true;
    bool overlapCheck       = pset.get< bool >("G4OverlapCheck",dfltOverlap);
    bool validateGDMLSchema = pset.get< bool >("G4ValidateGDMLSchema",dfltSchema);
    if ( dfltOverlap != overlapCheck || dfltSchema != validateGDMLSchema ) {
      mf::LogInfo("G4Alg") << "non-standard GDML processing: "
                           << std::boolalpha
                           << "overlapCheck " << overlapCheck
                           << " [" << dfltOverlap << "] "
                           << "validateGDMLSchema " << validateGDMLSchema
                           << " ]" << dfltSchema << "] "
                           << std::noboolalpha;
    }
    fG4Help->SetOverlapCheck(overlapCheck);
    fG4Help->SetValidateGDMLSchema(validateGDMLSchema);

    // Initialize the helper (physics stage)
    fG4Help->InitPhysics();

    // Geant4 comes with "user hooks" that allows users to perform
    // special tasks at the beginning and end of runs, events, tracks,
    // steps.  By using the UserActionManager, we've separated each
    // set of user tasks into their own class; e.g., there can be one
    // class for processing voxels, one class for processing
    // particles, one class for histograms, etc.

    // All UserActions must be adopted after the InitMC call

    // Setup the user actions that we want to use.
    g4b::UserActionManager*  uam = g4b::UserActionManager::Instance();
    
    g4n::FLSHitListAction*   fls = 
      new g4n::FLSHitListAction(fEnergyThresh, fIsRockVetoUsed);
    fls->SetName("g4n::FLSHitListAction");
    fls->Config( pset );

    g4n::ParticleListAction* pl  = 
      new g4n::ParticleListAction(fEnergyThresh,fManyParticles);
    pl->SetName("g4n::ParticleListAction");
    pl->Config( pset );

    // the ParticleListAction must be added to the UserActionManager 
    // first as it has to define the track ID in the case that the 
    // current particle to track is from an EM process that would cause
    // it not to be added to the list and the track ID has to be 
    // reassigned to be the same as its mother.
    uam->AddAndAdoptAction(pl);
    fPlaIndex = uam->GetSize() - 1;
    //if (false) std::cerr << "RWH: fPlaIndex " << fPlaIndex
    //          << " vs " << uam->GetIndex("g4n::ParticleListAction")
    //          << std::endl;


    uam->AddAndAdoptAction(fls);
    fFlsIndex = uam->GetSize() - 1;
    //if (false) std::cerr << "RWH: fFlsIndex " << fFlsIndex
    //          << " vs " << uam->GetIndex("g4n::FLSHitListAction")
    //          << std::endl;

    ConfigUserActionManager(fUserActions,pset);

    uam->PrintActionList("config");

    // Complete the initialize the helper (actions+finalize stage)
    fG4Help->SetUserAction();

    // Allow user to adjust particle lifetimes
    OverridePDGLifetimes();

    return;
  }// end of constructor

  //____________________________________________________________________________
  // Destructor
  G4Alg::~G4Alg()  
  {  
    if (fG4Help) delete fG4Help;
  } // end of destructor

  //______________________________________________________________________________
  void G4Alg::ConfigUserActionManager(std::vector<std::string> const& actionList,
                                      fhicl::ParameterSet const& pset)
  {

    /// additional user supplied UserAction classes

    // UserActionManager is a singleton, so get a handle
    g4b::UserActionManager*  uam = g4b::UserActionManager::Instance();

    // The factory is also a singleton
    g4b::UserActionFactory& uafactory = g4b::UserActionFactory::Instance();


    // given a list of classes, ask the factory for a copy and configure
    // it before handing it to the manager

    for (size_t j = 0; j < actionList.size(); ++j ) {

      std::string uaname = actionList[j];  // will use name for finding pset too
      g4b::UserAction* ua = uafactory.GetUserAction(uaname);
    
      if ( ua ) {
	// factory constructed a UserAction, now configure it
      
	// parameter set name can't have colons 
	// (ie not namespace qualified class used to create our object)
	// take only the class name portion
	size_t icolon = uaname.find_last_of(":");
	if ( icolon == std::string::npos ) icolon = -1;
	std::string psetname = uaname.substr(icolon+1); // don't include colon
      
	// check if there is a sub-pset with the same name as the
	// (non-namespaced) class name we're trying to configure
	fhicl::ParameterSet uapset;  // for holding results of fetching sub-pset
	bool found = pset.get_if_present< fhicl::ParameterSet >(psetname,uapset);
	mf::LogInfo("G4Alg")
	  << "UserAction \"" << uaname << "\" to be Config() with "
	  << ( ( found ) ? "\"" : "G4AlgPSet, couldn't find \"" ) 
	  << psetname << "\" pset";
	// if no named sub-pset available, pass the one G4Alg got 
	fhicl::ParameterSet const& pset2pass = ( found ) ? uapset : pset;
      
	// configure using _some_ pset
	ua->Config(pset2pass);
	// add to the manager
	uam->AddAndAdoptAction(ua);
      
      } 
      else {
	// ooops, factory failed ... let someone know
      
	const std::vector<std::string>& availActions = uafactory.AvailableUserActions();
	std::cout << "g4n::G4Alg requested UserAction \"" << uaname << "\" from "
		  << "g4b::UserActionFactory" << std::endl
		  << "but it only knows about: " << std::endl;
	for (size_t k=0; k<availActions.size(); ++k) 
	  std::cout << "   " << availActions[k] << std::endl;
	std::cout << "proceeding without this action" << std::endl;
      }
    }
  }

  //______________________________________________________________________________
  void G4Alg::RunGeant(std::vector< art::Handle< std::vector<simb::MCTruth> > >& mclists,
		       std::vector< sim::FLSHitList >&                           flshitlist,
		       std::vector< sim::Particle >&                             particlelist,
		       std::vector< std::vector< std::pair<size_t, size_t> > >&  pListLimits)  
  {  
    g4b::UserActionManager*  uam = g4b::UserActionManager::Instance();
    dynamic_cast<ParticleListAction *>(uam->GetAction(fPlaIndex))->ResetTrackIDOffset();

    particlelist.clear();
    flshitlist  .clear();

    // pListLimits keeps track of the index of first particle for a given MCTruth 
    // in the particlelist vector and the index beyond the last particle
    pListLimits .clear();
    pListLimits .resize(mclists.size());

    // Need to process Geant4 simulation for each interaction separately.
    for(size_t mcl = 0; mcl < mclists.size(); ++mcl){
      
      art::Handle< std::vector<simb::MCTruth> > mclistHandle = mclists[mcl];

      for(size_t i = 0; i < mclistHandle->size(); ++i){
	const simb::MCTruth* mct(&(*mclistHandle)[i]);

	size_t start = particlelist.size();
	this->RunGeant(mct, flshitlist, particlelist);
	size_t end = particlelist.size();

	pListLimits[mcl].push_back(std::pair<size_t, size_t>(start, end));
	
      }// end of loop over interactions in the current handle
    }// end loop over handles
 
    return;
  }// end of RunGeant

  //______________________________________________________________________________
  void G4Alg::RunGeant(std::vector< const simb::MCTruth* >& mctruths,
		       std::vector< sim::FLSHitList >&      flshitlist,
		       std::vector< sim::Particle >&        particlelist,
		       std::map<int, size_t>&               trackIDToMCTruthIndex)  
  {  
    g4b::UserActionManager*  uam = g4b::UserActionManager::Instance();
    ParticleListAction* pla = dynamic_cast<ParticleListAction *>(uam->GetAction(fPlaIndex));
    pla->ResetTrackIDOffset();

    particlelist.clear();
    flshitlist  .clear();

    // trackIDToMCTruthIndex keeps track of which trackIDs correspond to which MCTruth
    // in mctruths
    trackIDToMCTruthIndex.clear();

    fG4Help->G4Run(mctruths);
    
    this->CollectFLSHitsAndParticles(flshitlist, particlelist);
    trackIDToMCTruthIndex = pla->TrackIDToMCTruthIndexMap();
 
    return;
  }// end of RunGeant

  //______________________________________________________________________________
  void G4Alg::RunGeant(art::Ptr<simb::MCTruth>         mctruth,
		       std::vector< sim::FLSHitList >& flshitlist,
		       std::vector< sim::Particle >&   particlelist,
		       int                             trackIDOffset)  
  {  
    this->RunGeant(mctruth.get(), flshitlist, particlelist, trackIDOffset);
  }

  //______________________________________________________________________________
  void G4Alg::RunGeant(const simb::MCTruth*            mctruth,
		       std::vector< sim::FLSHitList >& flshitlist,
		       std::vector< sim::Particle >&   particlelist,
		       int                             trackIDOffset)  
  {  
    LOG_DEBUG("G4Alg") << *mctruth;
    
    if(trackIDOffset > 0){
      g4b::UserActionManager*  uam = g4b::UserActionManager::Instance();
      dynamic_cast<ParticleListAction *>(uam->GetAction(fPlaIndex))->ResetTrackIDOffset(trackIDOffset);
    }

    // The following tells Geant4 to track the particles in this interaction.
    fG4Help->G4Run(mctruth);
    
    this->CollectFLSHitsAndParticles(flshitlist, particlelist);
    
    return;
  }// end RunGeant for a single simb::MCTruth

  //______________________________________________________________________________
  // this method does the work of sorting out the various fls hits and particles that
  // we want returned from the simulation, ie removing those that leave no trace in 
  // a detector, etc
  void G4Alg::CollectFLSHitsAndParticles(std::vector< sim::FLSHitList >& flshitlist,
					 std::vector< sim::Particle >&   particlelist)
  {
    // Make auto_pts of the lists.
    g4b::UserActionManager*  uam = g4b::UserActionManager::Instance();
    sim::FLSHitList& flshl = *(dynamic_cast<FLSHitListAction *>(uam->GetAction(fFlsIndex))->GetList());
    std::vector<sim::Particle> pl   = dynamic_cast<ParticleListAction *>(uam->GetAction(fPlaIndex))->GetList();

    if(fSparseTrajectories){
      for(unsigned int n = 0; n < pl.size(); ++n){
	pl[n].SparsifyTrajectory();
      }
    }

    /* If the list of FLS hits for this MCTruth is empty, then we
       should not store any of the associated sim::Particle objects.
       (Just be done at this point and return.) */
    if(fSlimSimParticles){
      if ( flshl.fHits.empty() ) return;
    }

    // Push them to the total vector of FLSHitList
    flshitlist.push_back(flshl);

    if ( fSlimSimParticles ) {
      // Keep only particles associated with FLS hits, either directly
      // or through an ancestor (e.g. if it entered or exited the detector)
      std::set<int> IDsToStore;
      // First start with IDs directly associated with FLS hits
      for ( size_t h = 0; h < flshl.fHits.size(); ++h ) {
	IDsToStore.insert(flshl.fHits[h].GetTrackID());
      }
      // Loop through particles to find descendants of FLS-hit-makers
      std::set<sim::Particle> partsToStore;
      for ( size_t p = 0; p < pl.size(); ++p ) {
	sim::Particle *thisParticle = &pl[p];
	bool assocWithFLS(false);
	if ( IDsToStore.find( thisParticle->TrackId() ) != IDsToStore.end() ) {
	  // Directly associated with an FLS hit! Keep it.
	  assocWithFLS = true;
	}
	int motherTrkID = thisParticle->Mother();
	while ( !assocWithFLS && motherTrkID != 0 ) {
	  // Is this ancestor associated with an FLS hit?
	  if ( IDsToStore.find( motherTrkID ) != IDsToStore.end() ) assocWithFLS = true;
	  for ( size_t i = 0; i < pl.size(); ++i ) {
	    if ( pl[i].TrackId() == motherTrkID ) motherTrkID = pl[i].Mother();
	  }
	}
	if ( assocWithFLS ) {
	  partsToStore.insert( *thisParticle );
	  IDsToStore.insert( thisParticle->TrackId() );
	}
      }
      // Loop through selected particles to find and include all of their ancestors
      for ( size_t p = 0; p < pl.size(); ++p ) {
	if ( IDsToStore.find(pl[p].TrackId()) != IDsToStore.end() ) {
	  // Add this particle and its ancestors to the list of particles to keep
	  sim::Particle *thisParticle = &pl[p];
	  while ( thisParticle->Mother() != 0 ) {
	    // Get the mother
	    for ( size_t i = 0; i < pl.size(); ++i ) {
	      if ( pl[i].TrackId() == thisParticle->Mother() ) {
		thisParticle = &pl[i];
	      }
	    }
	    partsToStore.insert( *thisParticle );
	    IDsToStore.insert( thisParticle->TrackId() );
	  }
	}
      }
      // One more loop to keep all primaries (particles with Mother ID = 0)
      for ( size_t p = 0; p < pl.size(); ++p ) {
	sim::Particle *thisParticle = &pl[p];
	if ( thisParticle->Mother() == 0 ) {
	  partsToStore.insert( *thisParticle );
	  IDsToStore.insert( thisParticle->TrackId() );
	}
      }
      // Transfer the selected particles to the particle list
      for ( std::set<sim::Particle>::iterator p = partsToStore.begin();
	    p != partsToStore.end(); ++p ) {
	// Recreate this particle in order to slim down the daughter list, too
	sim::Particle newParticle( p->TrackId(),
				   p->PdgCode(),
				   p->Process(),
				   p->Mother(),
				   p->Mass(),
				   p->StatusCode() );
	newParticle.SetPolarization( p->Polarization() );
	sim::Particle part = *p; newParticle.SetGvtx( part.GetGvtx() );
	newParticle.SetRescatter( p->Rescatter() );
	newParticle.SetWeight( p->Weight() );
	simb::MCTrajectory traj = p->Trajectory();
	for ( int t = 0; t < (int)traj.size(); ++t ) {
	  newParticle.AddTrajectoryPoint( traj.Position(t), traj.Momentum(t) );
	}
	for ( int d = 0; d < p->NumberDaughters(); ++d ) {
	  int trkID = p->Daughter(d);
	  if ( IDsToStore.find( trkID ) != IDsToStore.end() ) {
	    newParticle.AddDaughter( trkID );
	  }
	}
	// Save it!
	particlelist.push_back( newParticle );
      }
    } else {
      // Keep all sim::Particles from this event
      for(size_t p = 0; p < pl.size(); ++p){
	particlelist.push_back(pl[p]);
      }
    }

    return;
  }
					 

  //______________________________________________________________________________
  void G4Alg::OverridePDGLifetimes()
  {
    int n = fOverrideLifetimePDG.size();
    for (int k=0; k<n; ++k) {
      int    pdg      = fOverrideLifetimePDG[k];
      double lifetime = fOverrideLifetimeNS[k];
      G4ParticleDefinition* g4pdef = 
        G4ParticleTable::GetParticleTable()->FindParticle(pdg);
      if ( ! g4pdef ) {
        mf::LogError("G4Alg")
          << "Attempt to override PDG lifetime in Geant4 failed "
          << "to get a G4ParticleDefintions for PDG=" << pdg;
        continue;
      }
      mf::LogInfo("G4Alg")
        << "Modifying the G4ParticleDefinition for PDG=" << pdg 
        << ", lifetime was " << g4pdef->GetPDGLifeTime()/ns << " (ns) to "
        << lifetime;
      g4pdef->SetPDGLifeTime(lifetime*ns);
    }

  }

 
} // end namespace
