//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//
// $Id: NovaG4MonopoleFieldSetup.cc,v 1.2 2012-09-20 21:45:42 greenc Exp $
// GEANT4 tag $Name: not supported by cvs2svn $
//
//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......
//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo......
//
//
// NovaG4MonopoleFieldSetup is responsible for setting up a magnetic field
// and the ability to use it with two different equation of motions, 
// one for monopoles and another for the rest of the particles. 
// 
// 

// =======================================================================
// Created:  13 May 2010, B. Bozsogi
//  2012-02-04:  R. Hatcher - incorporate into NOvA framework from base G4 v4.9.4.p02
// =======================================================================

#include "g4nova/NovaG4MonopoleFieldSetup.hh"
#include "g4nova/NovaG4MonopoleFieldMessenger.hh"

#include "Geant4/G4MagneticField.hh"
#include "Geant4/G4UniformMagField.hh"
#include "Geant4/G4FieldManager.hh"
#include "Geant4/G4TransportationManager.hh"
#include "Geant4/G4Mag_UsualEqRhs.hh"

#include "g4nova/NovaG4MonopoleEquation.hh"

#include "Geant4/G4MagIntegratorStepper.hh"
#include "Geant4/G4ChordFinder.hh"

// #include "G4ExplicitEuler.hh"
// #include "G4ImplicitEuler.hh"
// #include "G4SimpleRunge.hh"
// #include "G4SimpleHeum.hh"
#include "Geant4/G4ClassicalRK4.hh"
// #include "G4HelixExplicitEuler.hh"
// #include "G4HelixImplicitEuler.hh"
// #include "G4HelixSimpleRunge.hh"
// #include "G4CashKarpRKF45.hh"
// #include "G4RKG3_Stepper.hh"

// #include "G4SIunits.hh"

g4n::NovaG4MonopoleFieldSetup* g4n::NovaG4MonopoleFieldSetup::fMonopoleFieldSetup=0;

namespace g4n {

//////////////////////////////////////////////////////////////////////////
//
//  Constructor
NovaG4MonopoleFieldSetup::NovaG4MonopoleFieldSetup()
{
  if (!fMonopoleFieldSetup)
  {
    fChordFinder = 0;
    fStepper = 0;
    fMagneticField = 0;
    
    fMonopoleFieldMessenger = new NovaG4MonopoleFieldMessenger(this);
  }
  else
  {
    G4cerr << "Only ONE instance of NovaG4MonopoleFieldSetup is allowed!"
           << G4endl;
    G4Exception("NovaG4MonopoleFieldSetup::NovaG4MonopoleFieldSetup()",
                "InvalidSetup", FatalException,
                "Only ONE instance of MagneticFieldSetup is allowed!");
  }
}

//////////////////////////////////////////////////////////////////////////
// Retrieve the static instance of the singleton
//
NovaG4MonopoleFieldSetup* NovaG4MonopoleFieldSetup::GetMonopoleFieldSetup()
{
   static NovaG4MonopoleFieldSetup theInstance;
   if (!fMonopoleFieldSetup)
   {
     fMonopoleFieldSetup = &theInstance;
   }
   
   return fMonopoleFieldSetup;
}

////////////////////////////////////////////////////////////////////////////////

NovaG4MonopoleFieldSetup::~NovaG4MonopoleFieldSetup()
{
  delete fMonopoleFieldMessenger;
  if(fMagneticField) delete fMagneticField;
  if(fChordFinder)   delete fChordFinder;
  if(fStepper)       delete fStepper;
  if(fMonopoleStepper)  delete fMonopoleStepper;
}

////////////////////////////////////////////////////////////////////////////////

void NovaG4MonopoleFieldSetup::SetMagField(G4double fieldValue)
{
  //apply a global uniform magnetic field along Z axis  
  if (fMagneticField) { delete fMagneticField; }  //delete the existing magn field

  if (fieldValue != 0.)     // create a new one if non nul
    {
      fMagneticField = new G4UniformMagField(G4ThreeVector(0., 0., fieldValue));        
      InitialiseAll();
    }
   else
    {
      fMagneticField = 0;
      fFieldManager->SetDetectorField(fMagneticField);
    }
}

////////////////////////////////////////////////////////////////////////////////

void
NovaG4MonopoleFieldSetup::InitialiseAll()
{
  fFieldManager = G4TransportationManager::GetTransportationManager()
                                          ->GetFieldManager();
                                          
  fEquation = new G4Mag_UsualEqRhs(fMagneticField); 
  fMonopoleEquation = new NovaG4MonopoleEquation(fMagneticField);
 
  fMinStep     = 0.01*mm ; // minimal step of 1 mm is default
                                        
  fMonopoleStepper = new G4ClassicalRK4( fMonopoleEquation, 8 ); // for time information..
  fStepper = new G4ClassicalRK4( fEquation );                                         
                                         
  SetStepperAndChordFinder(0);
}

/////////////////////////////////////////////////////////////////////////////
//
// Update field
//
void NovaG4MonopoleFieldSetup::SetStepperAndChordFinder(G4int val)
{
  if (fMagneticField)
  {
    fFieldManager->SetDetectorField(fMagneticField );

    if(fChordFinder) delete fChordFinder;

    switch (val)
    {
      case 0:
        fChordFinder = new G4ChordFinder( fMagneticField, fMinStep, fStepper);      
        break;
      case 1: 
	fChordFinder = new G4ChordFinder( fMagneticField, fMinStep, fMonopoleStepper);
        break;
    }   
  
    fFieldManager->SetChordFinder( fChordFinder );
  }
}

////////////////////////////////////////////////////////////////////////////////
//
//  Utility method
G4FieldManager*  NovaG4MonopoleFieldSetup::GetGlobalFieldManager()
{
  return G4TransportationManager::GetTransportationManager()
                          ->GetFieldManager();
}

} // end-of-namespace g4n
