///
/// \file    ViewMatchAlg.h
/// \brief   Algorithm to handle 3d matching of prongs from FuzzyKVertex
/// \version $Id: ViewMatchAlg.h,v 1.3 2012-11-05 22:36:04 edniner Exp $
/// \author  eniner@indiana.edu
///
#ifndef FUZZ_VIEWMATCHALG_H
#define FUZZ_VIEWMATCHALG_H

#include "fhiclcpp/ParameterSet.h"
#include "GeometryObjects/PlaneGeo.h"
#include "TVector3.h"

namespace rb 
{
  class Cluster; 
  class Prong; 
}

namespace fuzz {
  class ViewMatchAlg
  {
  public:
    explicit ViewMatchAlg(fhicl::ParameterSet const& pset);
    ~ViewMatchAlg();
    void reconfigure(const fhicl::ParameterSet& pset);

    /// \brief Function to load the vector of 2d prongs to be matched
    /// \param prongs Collection of prongs to match
    void LoadProngs(std::vector<rb::Prong> prongs);

    /// \brief Function to perform the view matching
    void   Matching();

    /// \brief Function to calculate the cumulative energy profile of a test prong
    /// \param testprong The 3D candidate prong being considered for a match
    /// \param dir The direction vector for the test prong
    /// \param start The startpoint of the energy distribution (usually the vertex)
    /// \param totpecorr The total PECorr for the prong in a given view
    /// \param v The view, either kX or kY of the prong
    /// Function returns a sorted list of point pairs distance,pecorr showing the deposition of energy as a function of distance from the vertex
    std::vector<std::pair<double, double> > CalcEnergyProfile(rb::Prong testprong, TVector3 dir, TVector3 start, double *totpecorr, geo::View_t v);

    /// \brief Function to combine a 2d prong from each view into a 3d test prong to compute a matching score for
    /// \param xprong Input prong from the XZ view
    /// \param yprong Input prong from the YZ view
    /// \param newdir Pointer to the new 3vector direction of the candidate
    /// \param vert Pointer to the 3vector vertex of the new prong
    rb::Prong   MakeTestProng(rb::Prong& xprong, rb::Prong& yprong, TVector3* newdir, TVector3* vert);

    /// \brief Function to perform the Kuiper test to compare the energy profiles of the track in each view
    /// \param xlist Sorted list of (distance_to_vertex,pecorr) for the track in the xz view
    /// \param ylist Sorted list of (distance_to_vertex,pecorr) for the track in the yz view
    /// \param shift Base distance to shift the track in one view to better align the profiles
    /// \param xpecorr The total pecorr in XZ view for the track
    /// \param ypecorr The total pecorr in YZ view for the track
    double KuiperTest(std::vector<std::pair<double,double>> xlist, std::vector<std::pair<double,double>> ylist, double shift, double xpecorr, double ypecorr);

  public:

    /// Container for all the candidate XZ prongs
    std::vector<rb::Prong> fXProngs;

    /// Container for all the candidate YZ prongs
    std::vector<rb::Prong> fYProngs;

    /// Container for all the 3D prongs produced after matching
    std::vector<rb::Prong> fMatchedProngs;

    /// Container for all the unmatched 2D prongs left over
    std::vector<rb::Prong> fUnmatchedProngs;
    

    /// store information related to the view matching for a summary ntuple
    std::vector<int> xzid;
    std::vector<int> yzid;
    std::vector<double> ksscore;
    std::vector<double> ksshift;
    std::vector<std::vector<double>> xzs;
    std::vector<std::vector<double>> yzs;
    std::vector<std::vector<double>> xzpe;
    std::vector<std::vector<double>> yzpe;

  };
}


#endif
////////////////////////////////////////////////////////////////////////
