///
/// \file    FuzzyKVertex.cxx
/// \brief   Fuzzy k-means to optimize elastic arms vertex
/// \author  eniner@indiana.edu
/// \version $Id: FuzzyKVertex.cxx,v 1.15 2012-11-07 18:20:04 edniner Exp $
///
#include "FuzzyKVertex/FuzzyKMeanAlg.h"
#include "FuzzyKVertex/ViewMatchAlg.h"

#include "ChannelInfo/BadChanList.h"
#include "GeometryObjects/CellGeo.h"
#include "Geometry/Geometry.h"
#include "GeometryObjects/PlaneGeo.h"
#include "GeometryObjects/CellUniqueId.h"
#include "RecoBase/CellHit.h"
#include "RecoBase/Cluster.h"
#include "RecoBase/Prong.h"
#include "RecoBase/Vertex.h"
#include "RecoBase/FilterList.h"
#include "Utilities/func/MathUtil.h"
#include "Utilities/AssociationUtil.h"
#include "Simulation/Simulation.h"
#include "SimulationBase/MCTruth.h"
#include "SimulationBase/MCNeutrino.h"
#include "SimulationBase/MCParticle.h"
#include "Simulation/Particle.h"
#include "MCCheater/BackTracker.h"

#include "art/Framework/Core/FindManyP.h"
#include "CLHEP/Random/RandGaussQ.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/TFileService.h"

#include "TVector3.h"
#include "TTree.h"

#include <cmath>

/// Fuzzy k-Means prong-finding algorithm
namespace fuzz {
  class FuzzyKVertex : public art::EDProducer 
  {
  public:
    explicit FuzzyKVertex(fhicl::ParameterSet const &p);
    virtual ~FuzzyKVertex();
    
    void  beginJob();
    void  reconfigure(const fhicl::ParameterSet& p);
    void  produce(art::Event &e);

  private: 

    /// \brief Function that runs the FuzzyKMean algorithm and produces 2d clusters for each view
    /// \param prongs A pointer vector collecting the resulting 2d prongs after clustering
    /// \param slice The input rb::Cluster object to use in clustering
    /// \param vert The candidate vertex to cluster around
    void  MakeProngs(std::vector<rb::Prong>* prongs, 
                     const rb::Cluster slice,
                     TVector3 vert);

    /// \brief Function for perform splitting after the 2d clustering stage and return prong candidates
    /// \param s rb::Cluster object containing hits that are added to the prong candidates
    /// \param fuzzy The FuzzyKMeanAlg object with the clustering results
    /// \param cindx The index of the cluster center to turn into a prong
    /// \param view The view, either kX or kY in which the prong takes place
    /// \param v The vertex seed used in the clustering
    /// \param prongs Pointer vector to the collections of prongs the new addition will be added to
    void AddProng(const rb::Cluster s, 
		  FuzzyKMeanAlg* fuzzy,
		  unsigned int cindx,
		  geo::View_t view, 
                  TVector3 v,
                  std::vector<rb::Prong>* prongs);

    /// \brief Function to clean up the slice and remove hits that have no nearby neighbors to reduce noise
    /// \param c Inupt is a PtrVector to the collection of cellhits in the slice.
    /// The output is also a PtrVector of cellhits after the cleaning process
    art::PtrVector<rb::CellHit> Scrub(art::PtrVector<rb::CellHit> c);

    /// \brief Function to find the best matching neutrino for a slice, used to make artificial verticies
    /// \param allTruth Input collection of all truth objects associated with the event
    /// \param sliceTruthTable Input table relating the efficiency/purtity of each slice for each truth object
    /// Function returns a vector of ints with the best mc truth object by efficiency for each slice
    std::vector<int> getFaveIdsByEfficiency(std::vector<cheat::NeutrinoWithIndex>& allTruth, 
                                            std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable);

    std::set<int> FindVisibleProngs3D(const art::PtrVector<rb::CellHit>& allhits,
                                      std::set<int> ids,
                                      std::vector<int> *cellsX,
                                      std::vector<int> *cellsY);
  private:

    /// An object of FuzzyKMeanAlg to do the main clustering
    FuzzyKMeanAlg* fFuzzyClustAlg;

    /// ParameterSet of FuzzyKMeanAlg
    fhicl::ParameterSet fFuzzyKMeanPSet;

    /// An object for recalculating cluster centers during the splitting stage
    FuzzyKMeanAlg* fFuzzyTestAlg;

    ///An object of ViewMatchAlg to do the 3D view matching
    ViewMatchAlg* fViewMatchAlg;

    /// ParameterSet of ViewMatchAlg
    fhicl::ParameterSet fViewMatchPSet;

    /// Slices must have at least this many hits to cluster (FHiCL parameter)
    unsigned int fMinHit;      

    /// Miniumm hits in x view / slice to cluster (FHiCL parameter)
    unsigned int fMinHitX;     

    /// Minimum hit in y view / slice to cluster (FHiCL parameter)
    unsigned int fMinHitY;     
    
    /// Maximum distance allowed between a hit and its closest neighbor in the same view (FHiCL parameter)
    double fScrubDist;

    /// 1=make fake vertex from truth, 0= use ElasticArms vertex (FHiCL parameter)
    int fArtificialVert;

    /// 1 sigma distance, in cm, to smear fake vertex x coordinate from truth (FHiCL parameter)
    double fSmearX;

    /// 1 sigma distance, in cm, to smear fake vertex y coordinate from truth (FHiCL parameter)
    double fSmearY;

    /// 1 sigma distance, in cm, to smear fake vertex z coordinate from truth (FHiCL parameter)
    double fSmearZ;
    
    /// Maximum plane away from vertex a hit can be to use aggresive splitting (FHiCL parameter)
    int fVertPlaneGap;

    /// Minimum gap size, in cm, to split for hits near vertex (FHiCL parameter)
    double fVertGapSize;

    /// Minimum number of planes between two hits in a view at the split point in vertex region, 
    /// there can be no hit activity on that track in that view between those hits (FHiCL parameter)
    int fVertIntPGap;

    /// Minimum gap size, in cm, to split for hits away from vertex region (FHiCL parameter)
    double fNonVertGapSize;

    /// Minimum number of planes between two hits in a view at the split point away from the vertex region,
    /// there can be no hit activity on that track in that view between those hits (FHiCL parameter)
    int fNonVertIntPGap;

    /// Minimum number of clusters to attempt in a view
    unsigned int fMinClust;

    /// Maximum number of clusters to attempt in a view
    unsigned int fMaxClust;

    /// Minimum hit membership in a cluster
    double fMemb;
    
    /// Input label to grab slices (FHiCL parameter)
    std::string fSliceLabel;

    /// Input label to grab verticies (FHiCL parameter)
    std::string fVertexLabel;

    /// Output label for 3d prongs (FHiCL parameter)
    std::string fPngInst3D;

    /// Output label for 2d prongs (FHiCL parameter)
    std::string fPngInst2D;

    /// label for 3d prong associations
    std::string fPngAssn3D;
 
    /// lable for 2d prong associations
    std::string fPngAssn2D;

    /// If true, will skip slice if preselection did not pass (FHiCL parameter)
    bool fObeyPreselection;

    /// Tree to output debugging info for view merging process
    TTree* fOutTree;
 
    /// ntuple variables to debug view matching
    int run;
    int subrun;
    int event;
    int xzid;
    int yzid;
    int slc;
    double score;
    double shiftd;
    std::vector<double> xzs;
    std::vector<double> yzs;
    std::vector<double> xzpe;
    std::vector<double> yzpe;

  };
}
////////////////////////////////////////////////////////////////////////////////////////////////

namespace fuzz{

  FuzzyKVertex::FuzzyKVertex(fhicl::ParameterSet const &p) :
    fFuzzyClustAlg(0),
    fFuzzyKMeanPSet(p.get<fhicl::ParameterSet>("FuzzyKMeanPSet")),
    fFuzzyTestAlg(0),
    fViewMatchAlg(0),
    fViewMatchPSet(p.get<fhicl::ParameterSet>("ViewMatchPSet"))
  {
    this->reconfigure(p);

    this->produces< std::vector<rb::Prong>  >(fPngInst3D);
    this->produces< std::vector<rb::Prong>  >(fPngInst2D);
    if (fArtificialVert == 0){
      this->produces<art::Assns<rb::Prong, rb::Vertex> >(fPngAssn3D);
      this->produces<art::Assns<rb::Prong, rb::Vertex> >(fPngAssn2D);
    }
    if (fArtificialVert == 1){
      this->produces<art::Assns<rb::Vertex, rb::Prong> >(fPngAssn3D);
      this->produces<art::Assns<rb::Vertex, rb::Prong> >(fPngAssn2D);
    }
    this->produces<art::Assns<rb::Prong, rb::Cluster> >(fPngAssn3D);
    this->produces<art::Assns<rb::Prong, rb::Cluster> >(fPngAssn2D);

    //if we want to make use of a fake vertex instead of the real one
    if (fArtificialVert == 1){
      this->produces<std::vector<rb::Vertex> >("FakeVert");
      this->produces<art::Assns<rb::Vertex,rb::Cluster> >();

      // get the random number seed, use a random default if not specified    
      // in the configuration file.  
      unsigned int seed = sim::GetRandomNumberSeed();
      createEngine( seed );
    }


    
  }

  //......................................................................

  void FuzzyKVertex::beginJob() 
  {
    //art::ServiceHandle<art::TFileService> tfs;

    // define ta tree containing the results of the view matching for diagnostic purposes
    //fOutTree = tfs->make<TTree>("ViewMatch","View Matching");

    //store view matching summary information
    //fOutTree->Branch("xzid", &xzid, "xzid/I");
    //fOutTree->Branch("yzid", &yzid, "yzid/I");
    //fOutTree->Branch("slc", &slc, "slc/I");
    //fOutTree->Branch("score", &score);
    //fOutTree->Branch("shiftd", &shiftd);
    //fOutTree->Branch("xzs", &xzs);
    //fOutTree->Branch("yzs", &yzs);
    //fOutTree->Branch("xzpe", &xzpe);
    //fOutTree->Branch("yzpe", &yzpe);

  }

  //......................................................................

  void FuzzyKVertex::reconfigure(const fhicl::ParameterSet& p) 
  {
    fMinHit  = p.get<unsigned int>("MinHit");
    fMinHitX = p.get<unsigned int>("MinHitX");
    fMinHitY = p.get<unsigned int>("MinHitY");

    fMinClust = p.get<unsigned int>("MinClust");
    fMaxClust = p.get<unsigned int>("MaxClust");

    fMemb = p.get<double>("Memb");
    
    fScrubDist = p.get<double>("ScrubDist");

    fVertPlaneGap   = p.get<int   >("VertPlaneGap");
    fVertGapSize    = p.get<double>("VertGapSize");
    fVertIntPGap    = p.get<int   >("VertIntPGap");
    fNonVertGapSize = p.get<double>("NonVertGapSize");
    fNonVertIntPGap    = p.get<int   >("NonVertIntPGap");

    fArtificialVert = p.get<int   >("ArtificialVert");
    fSmearX         = p.get<double>("SmearX");
    fSmearY         = p.get<double>("SmearY");
    fSmearZ         = p.get<double>("SmearZ");
    
    fSliceLabel  = p.get<std::string>("SliceLabel");
    fVertexLabel = p.get<std::string>("VertexLabel");
    fPngInst3D    = p.get<std::string>("PngInst3D");
    fPngInst2D    = p.get<std::string>("PngInst2D");
    fPngAssn3D    = p.get<std::string>("PngAssn3D");
    fPngAssn2D    = p.get<std::string>("PngAssn2D");

    fObeyPreselection = p.get<bool>("ObeyPreselection");
  }

  //......................................................................

  FuzzyKVertex::~FuzzyKVertex() 
  { 
    if(fFuzzyClustAlg) delete fFuzzyClustAlg;
    if(fFuzzyTestAlg)  delete fFuzzyTestAlg;
    if(fViewMatchAlg)  delete fViewMatchAlg;
  }

  //......................................................................

  void FuzzyKVertex::produce(art::Event& evt) 
  {
    fFuzzyClustAlg = new FuzzyKMeanAlg(fFuzzyKMeanPSet);
    fFuzzyTestAlg  = new FuzzyKMeanAlg(fFuzzyKMeanPSet);
    fViewMatchAlg  = new ViewMatchAlg(fViewMatchPSet);

    unsigned register int i;

    std::unique_ptr< std::vector<rb::Prong>  > png3D(new std::vector<rb::Prong>);
    std::unique_ptr<art::Assns<rb::Prong, rb::Vertex> > png3DAssns(new art::Assns<rb::Prong, rb::Vertex>);
    std::unique_ptr<art::Assns<rb::Prong, rb::Vertex> > png2DAssns(new art::Assns<rb::Prong, rb::Vertex>);
    std::unique_ptr< std::vector<rb::Prong>  > png2D(new std::vector<rb::Prong>);
    std::unique_ptr<art::Assns<rb::Prong, rb::Cluster> > png3DSlcAssns(new art::Assns<rb::Prong, rb::Cluster>);
    std::unique_ptr<art::Assns<rb::Prong, rb::Cluster> > png2DSlcAssns(new art::Assns<rb::Prong, rb::Cluster>);

    //in case we want to use an artificial vertex instead of the real one, useful for testing algorithm and decoupling from ElasticArms
    std::unique_ptr< std::vector<rb::Vertex> > aVert(new std::vector<rb::Vertex>);
    std::unique_ptr<art::Assns<rb::Vertex, rb::Cluster> > avAssns(new art::Assns<rb::Vertex, rb::Cluster>);
    std::unique_ptr<art::Assns<rb::Vertex, rb::Prong> > fpng3DAssns(new art::Assns<rb::Vertex, rb::Prong>);
    std::unique_ptr<art::Assns<rb::Vertex, rb::Prong> > fpng2DAssns(new art::Assns<rb::Vertex, rb::Prong>);
    
    art::ServiceHandle<geo::Geometry> fGeo;

    // Pull out the time slices and hough results for the event
    art::Handle<std::vector<rb::Cluster> > sHandle;
    evt.getByLabel(fSliceLabel,sHandle);
    art::PtrVector<rb::Cluster> slices;
    for(unsigned int i=0; i<sHandle->size(); ++i){
      slices.push_back(art::Ptr<rb::Cluster>(sHandle, i));
    }


    std::vector<int> bestMCId;
    std::vector<cheat::NeutrinoWithIndex> truths;
    std::vector<std::vector<cheat::NeutrinoEffPur>> sEffPur;
    //if we want to use an artificial vertex based off of the true interaction vertex
    //build a matrix to match the proper mctruth object to a slice
    if(fArtificialVert == 1){
      art::ServiceHandle<cheat::BackTracker> bt;
      //get vector of neutrinos in event
      truths = bt->allMCTruth();
      sEffPur = bt->SlicesToMCTruthsTable(slices);
      //get best neutrino match for each slice
      bestMCId = getFaveIdsByEfficiency(truths, sEffPur);
    }
    
    //object to get vertex associated with a slice
    art::FindManyP<rb::Vertex> fmv(sHandle, evt, fVertexLabel);
    
    //container to hold prongs
    std::vector<rb::Prong> prongs;

    // containers needed for making associations between fake verticies and tracks.  
    // Need to keep track of the slice and the start and stop indicies of the 
    //tracks in that slice out of the entire collection
    std::vector<size_t> slIndx3D;
    std::vector<size_t> slIndx2D;
    std::vector<size_t> sIndx3D;
    std::vector<size_t> eIndx3D;
    std::vector<size_t> sIndx2D;
    std::vector<size_t> eIndx2D;
    
    //loop over slices to do the work
    for (i=0; i<sHandle->size(); ++i) {


       //obey any preselection if desired
       if(fObeyPreselection && rb::IsFiltered(evt,sHandle,i)) continue;

       art::Ptr<rb::Cluster> sptr(sHandle, i);
      
      // Skip the noise cluster and require some minimum number of hits
      if (sptr->IsNoise()) continue;
      
      // perform scrubbing
      rb::Cluster neighbors(this->Scrub(sptr->AllCells()));
      
      //make sure slice is reconstructable before proceeding
      if (neighbors.NXCell()<fMinHitX) continue;
      if (neighbors.NYCell()<fMinHitY) continue;
      if (neighbors.NCell() <fMinHit)  continue;

      // Find any verticies that go with this time slice
      TVector3 vert;
      std::vector<art::Ptr<rb::Vertex> >  v;

      //if a fake vertex is needed, make it now
      if (fArtificialVert == 1){
        if (bestMCId[i] == -1) continue;

        // get the RandomNumberGenerator service
        art::ServiceHandle<art::RandomNumberGenerator> rng;
        CLHEP::HepRandomEngine &engine = rng->getEngine();
        CLHEP::RandGaussQ       gauss(engine);

        art::Ptr<simb::MCTruth> nuTru = sEffPur[i][bestMCId[i]].neutrinoInt;
        // Go ahead and shoot gaussians
        double Vx = gauss.fire(0.0, fSmearX);
        double Vy = gauss.fire(0.0, fSmearY);
        double Vz = gauss.fire(0.0, fSmearZ);
        double T  = 0.0;
        // Now shift to the original vertex
        // If neutrino is set, get to the neutrino vertex
        if(nuTru->NeutrinoSet()){
          Vx += nuTru->GetNeutrino().Nu().Vx();	
          Vy += nuTru->GetNeutrino().Nu().Vy();	
          Vz += nuTru->GetNeutrino().Nu().Vz();	
          T  += nuTru->GetNeutrino().Nu().Position().T();	
        }	
        else{	
          // Otherwise, get to the first particle vertex	
          Vx += nuTru->GetParticle(0).Vx();
          Vy += nuTru->GetParticle(0).Vy();	
          Vz += nuTru->GetParticle(0).Vz();
          T  += nuTru->GetParticle(0).Position().T();
        }

        //make fake vertex and put it in the event
        rb::Vertex vtx(Vx, Vy, Vz, T);
        aVert->push_back(vtx);
        util::CreateAssn(*this, evt, *aVert, slices[i], *avAssns, UINT_MAX, "FakeVert");
        vert = vtx.GetXYZ();
      }
      //else get the vertex associated with the slice
      else{
        v = fmv.at(i);
        //NOTE: Currently only one vertex is produced per slice, if this ever changes 
        //then this logic must also change.
        if (v.size() != 1) continue;
        vert = v[0]->GetXYZ();
      }

      //make sure all vertex values are real and not nan or inf.  Make sure vertex is not too far outside detector(this cut is far too loose right now
      //but for rock events a vertex outside the detector is possible, don't want to risk being too restrictive without study.
      if(std::isnan(vert.X()) || std::isinf(vert.X()) || std::abs(vert.X())>10000) continue;
      if(std::isnan(vert.Y()) || std::isinf(vert.Y()) || std::abs(vert.Y())>10000) continue;
      if(std::isnan(vert.Z()) || std::isinf(vert.Z()) || std::abs(vert.Z())>10000) continue;

      //store ntuple information for view match debugging
      run = evt.run();
      subrun = evt.subRun();
      event = evt.id().event();

 
      prongs.clear();
      //make the 2D prong candidates for this slice
      this->MakeProngs(&prongs, neighbors, vert);

      //perform the view matching with this set of 2D prongs
      fViewMatchAlg->LoadProngs(prongs);
      fViewMatchAlg->Matching();

      //record view matching information into the ntuple
      //for (unsigned int k=0; k<fViewMatchAlg->xzid.size(); ++k){
      //  xzid = -1;
      //  yzid = -1;
      //  score = -1;
      //  shiftd = -11.0;
      //  xzpe.clear();
      //  yzpe.clear();
      //  xzs.clear();
      //  yzs.clear();

      //  xzid = fViewMatchAlg->xzid[k];
      //  yzid = fViewMatchAlg->yzid[k];
      //  score = fViewMatchAlg->ksscore[k];
      //  shiftd = fViewMatchAlg->ksshift[k];
      //  xzs = fViewMatchAlg->xzs[k];
      //  yzs = fViewMatchAlg->yzs[k];
      //  xzpe = fViewMatchAlg->xzpe[k];
      //  yzpe = fViewMatchAlg->yzpe[k];

      //  fOutTree->Fill();
      //}

      //loop over 3D matched prongs to push into the event
      for(unsigned int k=0; k<fViewMatchAlg->fMatchedProngs.size(); ++k){

        rb::Prong png(fViewMatchAlg->fMatchedProngs[k]);

        //move the prong start point from the elastic arms vertex to the closest hit
        TVector3 start;
        double minZ = png.MinZ();
        double maxZ = png.MaxZ();
        if ((png.Dir().Z() > 0.0) &&(minZ < png.Start().Z())) minZ = png.Start().Z();
        if ((png.Dir().Z() < 0.0) &&(maxZ > png.Start().Z())) maxZ = png.Start().Z();

        if (png.Dir().Z() < 0.0){
          double temp = maxZ;
          maxZ = minZ;
          minZ = temp;
        }

        start[0] = (fViewMatchAlg->fMatchedProngs[k].Dir().X()/fViewMatchAlg->fMatchedProngs[k].Dir().Z())*(minZ-fViewMatchAlg->fMatchedProngs[k].Start().Z())
                    + fViewMatchAlg->fMatchedProngs[k].Start().X();
        start[1] = (fViewMatchAlg->fMatchedProngs[k].Dir().Y()/fViewMatchAlg->fMatchedProngs[k].Dir().Z())*(minZ-fViewMatchAlg->fMatchedProngs[k].Start().Z()) 
                    + fViewMatchAlg->fMatchedProngs[k].Start().Y();
        start[2] = minZ;


        png.SetStart(start);

        png3D->push_back(png);
        if (fArtificialVert == 0) util::CreateAssn(*this, evt, *png3D, v[0], *png3DAssns, UINT_MAX, fPngInst3D);
        util::CreateAssn(*this, evt, *png3D, sptr, *png3DSlcAssns, UINT_MAX, fPngInst3D);
      } // end loop over matched prongs

      //if we are making a fake vertex, calculate the start and ending index for this slice's collection of prongs/tracks
      //so that associations can be made
      if ((fArtificialVert == 1) && (fViewMatchAlg->fMatchedProngs.size() > 0)){
        slIndx3D.push_back(i);
        sIndx3D.push_back((*png3D).size()-fViewMatchAlg->fMatchedProngs.size());
        eIndx3D.push_back((*png3D).size());
      }

      //Loop over the unmatched 2d prongs to also put them into the event
      for(unsigned int k=0; k<fViewMatchAlg->fUnmatchedProngs.size(); ++k){
        rb::Prong png(fViewMatchAlg->fUnmatchedProngs[k]);
        double xyStart = 0.0;
        double zStart = 0.0;
        double minZ = png.MinZ();
        double maxZ = png.MaxZ();
        if ((png.Dir().Z() > 0.0) &&(minZ < png.Start().Z())) minZ = png.Start().Z();
        if ((png.Dir().Z() < 0.0) &&(maxZ > png.Start().Z())) maxZ = png.Start().Z();

        if (png.Dir().Z() < 0.0){
          double temp = maxZ;
          maxZ = minZ;
          minZ = temp;
        }

        if (fViewMatchAlg->fUnmatchedProngs[k].View() == geo::kX){
          xyStart = (fViewMatchAlg->fUnmatchedProngs[k].Dir().X()/fViewMatchAlg->fUnmatchedProngs[k].Dir().Z())*(minZ-fViewMatchAlg->fUnmatchedProngs[k].Start().Z())
                     + fViewMatchAlg->fUnmatchedProngs[k].Start().X();
          zStart = minZ;
          if ((png.MinPlane() == png.MaxPlane())){ 
            if (png.Dir().X() > 0.0) {xyStart = png.MinV(geo::kX);}
            else if (png.Dir().X() < 0.0) {xyStart = png.MaxV(geo::kX);}
          }
        }

        if (fViewMatchAlg->fUnmatchedProngs[k].View() == geo::kY){
          xyStart = (fViewMatchAlg->fUnmatchedProngs[k].Dir().Y()/fViewMatchAlg->fUnmatchedProngs[k].Dir().Z())*(minZ-fViewMatchAlg->fUnmatchedProngs[k].Start().Z())
                     + fViewMatchAlg->fUnmatchedProngs[k].Start().Y();
          zStart = minZ;
          if ((png.MinPlane() == png.MaxPlane())) { 
            if (png.Dir().Y() > 0.0) {xyStart = png.MinV(geo::kY); }
            else if (png.Dir().Y() < 0.0) {xyStart = png.MaxV(geo::kY);}
          }
        }

        png.SetStart(xyStart,zStart);
        
        png2D->push_back(png);
        if (fArtificialVert == 0) util::CreateAssn(*this, evt, *png2D, v[0], *png2DAssns, UINT_MAX, fPngInst2D);
        util::CreateAssn(*this, evt, *png2D, sptr, *png2DSlcAssns, UINT_MAX, fPngInst2D);
      } // end loop over unmatched 2d tracks

      //if we are making a fake vertex, prepare accordingly for the associations
      if ((fArtificialVert == 1) && (fViewMatchAlg->fUnmatchedProngs.size() > 0)){
        slIndx2D.push_back(i);
        sIndx2D.push_back((*png2D).size()-fViewMatchAlg->fUnmatchedProngs.size());
        eIndx2D.push_back((*png2D).size());
      }

    } // Overall loop on slices

    //if a fake vertex is produced, now make the associations between the vertex and tracks
    if (fArtificialVert == 1){
      for(i=0; i<slIndx3D.size(); ++i){
        util::CreateAssn(*this, evt, *aVert, *png3D, *fpng3DAssns, sIndx3D[i],eIndx3D[i],slIndx3D[i],"FakeVert",fPngInst3D);
      }
      for(i=0; i<slIndx2D.size(); ++i){
        util::CreateAssn(*this, evt, *aVert, *png2D, *fpng2DAssns, sIndx2D[i],eIndx2D[i],slIndx2D[i],"FakeVert",fPngInst2D);
      }
    }
    //put products into the event
    evt.put(std::move(png3D),fPngInst3D);
    if (fArtificialVert == 0){
      evt.put(std::move(png3DAssns),fPngAssn3D);
      evt.put(std::move(png2DAssns),fPngAssn2D);
    }
    if (fArtificialVert == 1){
      evt.put(std::move(fpng3DAssns),fPngAssn3D);
      evt.put(std::move(fpng2DAssns),fPngAssn2D);
    }
    evt.put(std::move(png3DSlcAssns),fPngAssn3D);
    evt.put(std::move(png2DSlcAssns),fPngAssn2D);
    evt.put(std::move(png2D),fPngInst2D);
    if (fArtificialVert == 1){
      evt.put(std::move(aVert),"FakeVert");
      evt.put(std::move(avAssns));
    }

  }

  //......................................................................

  /// Function to produce the 2D prongs in both views before the matching
  void FuzzyKVertex::MakeProngs(std::vector<rb::Prong>* prongs,
                                const rb::Cluster slice,
                                TVector3 vert)
  {
    
    //first cluster in the xz view
    fFuzzyClustAlg->LoadCluster(slice);
    fFuzzyClustAlg->DoClustering(fMinClust,fMaxClust, vert.X(),vert.Z(),geo::kX);
    
    //loop from back to front.  FuzzyKMean makes clusters with the weakest cluster seed
    //having the last index.  When producing prongs start with the weak cluster so hits can be pushed
    //into the more dominant prongs before being dropped.
    for (int i=fFuzzyClustAlg->fA.size()-1; i>-1; i--){
      this->AddProng(slice, fFuzzyClustAlg, i, geo::kX, vert, prongs);
    }
    
    //now cluster in the yz view
    fFuzzyClustAlg->DoClustering(fMinClust,fMaxClust, vert.Y(),vert.Z(),geo::kY); 
    
    //loop from back to front.  FuzzyKMean makes clusters with the weakest cluster seed
    //having the last index.  When producing prongs start with the weak cluster so hits can be pushed
    //into the more dominant prongs before being dropped.
    for ( int i=fFuzzyClustAlg->fA.size()-1; i>-1; i--){
      this->AddProng(slice, fFuzzyClustAlg, i, geo::kY, vert, prongs);
    }

  }
  
  //......................................................................
  
    void FuzzyKVertex::AddProng(const rb::Cluster s, 
      			  	FuzzyKMeanAlg* fuzzy,
				unsigned int cindx,
				geo::View_t view, 
                                TVector3 v,
                                std::vector<rb::Prong>* prongs)
  {
    art::ServiceHandle<geo::Geometry> geom;

    //initialize variables
    double theta = fuzzy->fA[cindx];
    double dCosXY = sin(theta);
    double dCosZ = cos(theta);
    double zStart = v.Z();
    double xyStart = v[view];
    unsigned int vertplane = 999;

    //snap vertex into detector boundaries so that it can be determined which view the vertex plane lies in
    double vx = v.X();
    double vy = v.Y();
    double vz = v.Z();
    if (v.X() < -geom->DetHalfWidth()) vx = -geom->DetHalfWidth()+15.0;
    if (v.X() > geom->DetHalfWidth()) vx = geom->DetHalfWidth()-15.0;
    if (v.Y() < -geom->DetHalfHeight()) vy = -geom->DetHalfHeight()+15.0;
    if (v.Y() > geom->DetHalfHeight()) vy = geom->DetHalfHeight()-15.0;
    if (v.Z() < 0.0) vz = 2.0;
    if (v.Z() > geom->DetLength()) vz = geom->DetLength()-2.0;

    if ((std::abs(v.X()) < geom->DetHalfWidth()) && (std::abs(v.Y()) < geom->DetHalfHeight()) && (v.Z() > 0.0) && (v.Z() < geom->DetLength())){
      geo::CellUniqueId cid = geom->CellId(vx, vy, vz, 1.5, 1.5, 6., 1.0);
      if (cid != 0) vertplane = geom->getPlaneID(cid);
    }

    //calculate distance of each hit to the vertex and sort
    unsigned int cellcount = s.NCell(view);
    std::vector<std::pair<float, unsigned int> > distvec;
    for (unsigned int i=0; i<cellcount; ++i){
      if (fuzzy->fU[cindx][i] >= fMemb){
        distvec.push_back(std::make_pair(fuzzy->fDist[i], i));
      }
    }
    sort(distvec.begin(), distvec.end());

    //loop over collection of cells in the cluster, break into smaller groups if needed
    std::vector<std::vector<std::pair<float, unsigned int> > > cluslist;
    std::vector<std::pair<float, unsigned int> > cluscand;
    cluscand.push_back(distvec[0]); //first cell always put in candidate cluster
    for (unsigned int i=1; i<distvec.size(); ++i){
      
      //check if hit is near the vertex. in this special case gap size for splitting is different
      if (abs(s.Cell(view,distvec[i-1].second)->Plane() - vertplane) <= fVertPlaneGap) {
        if ((distvec[i].first - distvec[i-1].first > fVertGapSize) && (abs(s.Cell(view,distvec[i].second)->Plane() - s.Cell(view,distvec[i-1].second)->Plane()) >= fVertIntPGap)){
          cluslist.push_back(cluscand);
          cluscand.clear();
          cluscand.push_back(distvec[i]); 
          continue;
        }
        else {
          cluscand.push_back(distvec[i]);
          continue;
        }
      }
      //in non vertex region use different splitting criteria
      else if ((distvec[i].first - distvec[i-1].first >= fNonVertGapSize) && 
               (abs(s.Cell(view,distvec[i].second)->Plane() - s.Cell(view,distvec[i-1].second)->Plane()) >= fNonVertIntPGap)){
        cluslist.push_back(cluscand);
        cluscand.clear();
        cluscand.push_back(distvec[i]);
        continue;
      }
      //otherwise add cluster unchanged
      else {
        cluscand.push_back(distvec[i]);
        continue;
      }
    }
    //add final cluster to list
    cluslist.push_back(cluscand);

    //if the list is of size 1, there is no splitting and all cells can be added to one prong
    if (cluslist.size() == 1){
      rb::Cluster clus(view);
      for (unsigned int i=0; i<cluslist[0].size(); ++i){
        unsigned int hit = cluslist[0][i].second;
        clus.Add(s.Cell(view,hit));
      }
      rb::Prong p(clus, xyStart, zStart, dCosXY, dCosZ);
      (*prongs).push_back(p);
    }
    //if splits occurred we must re cluster accordingly
    else{
      for (unsigned int i=0; i<cluslist.size(); ++i){
        //TODO: There should be some minimum hit/energy requirement
        //for keeping a cluster so that small things get thrown away, but want
        //to preserve neutrons.  For now no cuts are made on energy or hits.

        //Check fraction of cells with minimum membership to a different cluster.
        //This bit is tricky since cells in the same candidate grouping may not
        //belong to the same different cluster.  In current model, if a hit has the
        //minimum membership in another cluster let it go there.
        for (unsigned int j=0; j<cluslist[i].size(); ++j){
          unsigned int hit = cluslist[i][j].second;
          for(unsigned int k=0; k<fuzzy->fA.size(); ++k){
            if((k != cindx) && (fuzzy->fU[k][hit] >= fMemb)) {fuzzy->fU[cindx][hit] = 0.0; break;}
          }
        }
 
        //now put this into a prong.  We lost hits somewhere, so make a special FuzzyKMean object to calculate new angle for
        //for this cluster.
        rb::Cluster clus(view);
        for (unsigned int j=0; j<cluslist[i].size(); ++j){
          unsigned int hit = cluslist[i][j].second;
          if (fuzzy->fU[cindx][hit] >= fMemb) clus.Add(s.Cell(view,hit));
        }
        if (clus.NCell() > 0){
          fFuzzyTestAlg->LoadCluster(clus);
          //we know all the hits belong together, just need to recalculate angle, 
          //so set min and max cluster size to 1
          fFuzzyTestAlg->DoClustering(1,1, xyStart,zStart,view);
          theta = fFuzzyTestAlg->fA[0];
          dCosXY = sin(theta);
          dCosZ = cos(theta);
          //make sure the prong is physical before adding it to cluster
          if (!std::isnan(xyStart) && !std::isnan(zStart) && !std::isnan(dCosXY) && !std::isnan(dCosZ) &&
              !std::isinf(xyStart) && !std::isinf(zStart) && !std::isinf(dCosXY) && !std::isinf(dCosZ)){
            rb::Prong p(clus, xyStart, zStart, dCosXY, dCosZ);
            (*prongs).push_back(p);
          }
        }//end if

      }// end for loop
    }//end else statement

  }
  
  //......................................................................

  /// Function to remove hits with no neighbors from the slice
  art::PtrVector<rb::CellHit> FuzzyKVertex::Scrub(art::PtrVector<rb::CellHit> c)
  {
    
    std::vector<int> hit_numNeighbors;
    //in an ideal detector a hit must have 1 neighbor
    //inside this distance.
    double maxGap = fScrubDist; 
    art::PtrVector<rb::CellHit> trimCells;
    int iPlane, fPlane, iCell, fCell;
    double goodCh, totCh;
    
    iPlane = 0;
    iCell  = 0;
    fPlane = 0;
    fCell  = 0;
    
    art::ServiceHandle<geo::Geometry> geom;
    art::ServiceHandle<chaninfo::BadChanList> badc;
    
    for (unsigned int i=0; i<c.size(); i++){
      hit_numNeighbors.push_back(0);
    }
    
    //loop to find a cells closest neighbor
    for (unsigned int i=0; i<c.size(); i++){
      double minDist = 9999.0;
      int minCell = -1;
      if (hit_numNeighbors[i] > 0) continue;
      for(unsigned int j=0; j<c.size(); ++j){
        if (i == j) continue;
        if (c[i]->View() != c[j]->View()) continue;
        double xyz1[3], xyz2[3];
        geom->Plane((c[i])->Plane())->Cell((c[i])->Cell())->GetCenter(xyz1);
        geom->Plane((c[j])->Plane())->Cell((c[j])->Cell())->GetCenter(xyz2);
        double dist = util::pythag(xyz1[0]-xyz2[0],
                                   xyz1[1]-xyz2[1],
                                   xyz1[2]-xyz2[2]);
        if (dist < minDist){
          minDist = dist;
          minCell = j;
        }
      }
      if (minCell == -1) continue;
      
      //define the rectangle of cells such that a cell and its 
      //nearest neighbor are at diagonal corners
      if (c[i]->Plane() < c[minCell]->Plane()){
        iPlane = c[i]->Plane();
        fPlane = c[minCell]->Plane();
      }
      else {
         iPlane = c[minCell]->Plane();
         fPlane = c[i]->Plane();
      }
      if (c[i]->Cell() < c[minCell]->Cell()) {
        iCell = c[i]->Cell();
        fCell = c[minCell]->Cell();
      }
      else {
        iCell = c[minCell]->Cell();
        fCell = c[i]->Cell();
      }
      
      //within the rectangle find the fraction of cells that are good
      goodCh = 0;
      totCh = 0;
      for (int c=iCell; c<=fCell; c++){
        for (int p=iPlane; p<=fPlane; p+=2){
          totCh++;
          if(!(badc->IsBad(p,c))) ++goodCh;
        }
      }
      if ((minDist*goodCh/totCh) < maxGap){
        hit_numNeighbors[i]++;
        hit_numNeighbors[minCell]++;
      }
    }
    
    //keep the hits that have at least 1 neighbor in the slice
    for (unsigned int i=0; i<c.size(); i++){
      if (hit_numNeighbors[i] > 0){
        trimCells.push_back(c[i]);
      }
    }
   
    return trimCells;
  }

  //.........................................................................
 std::vector<int> FuzzyKVertex::getFaveIdsByEfficiency(std::vector<cheat::NeutrinoWithIndex>& allTruth, 
                                                       std::vector< std::vector<cheat::NeutrinoEffPur> >& sliceTruthTable)
  {
    // double check that slice truth table is safe
    for(auto slc = sliceTruthTable.begin(); slc != sliceTruthTable.end(); ++slc)
    {
      assert(slc->size() == allTruth.size() && "Slice/truth table must be properly shaped");
    }
    
    // We'll make an array with an entry for every subevt to store the unique identifier of 
    // the favourite neutrino within the subevt, with the correct size and all entries initialized to -1
    std::vector<int> faveIds(sliceTruthTable.size(), -1);
    
    // if no neutrinos, there are no favorites 
    if(allTruth.size()  == 0) 
      return faveIds;
    
    
    // List of best slice for each nu by slicer efficiency 
    std::vector<int> bestSlices(allTruth.size());
    //First we find the best slice for each neutrino
    for(size_t iNu = 0; iNu < allTruth.size(); ++iNu)
    {
      
      float maxValue =  0;  // we will use total slicer efficiency  
      int bestSlice   = -1;
      for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].efficiency > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].efficiency;
            bestSlice = iSlc;
          }
        }
      }
      
      bestSlices[iNu] = bestSlice;
    }
    
    // List of best nu for each slice
    
    // now we find the best neutrino for each slice.  
    for(size_t iSlc = 0; iSlc < sliceTruthTable.size(); ++iSlc)
    {
      
      float maxValue =  0;  // we will use slicer efficiency 
      int bestNu   = -1;
      for(size_t iNu = 0; iNu < sliceTruthTable[iSlc].size(); ++iNu)
      {
        
        if(!sliceTruthTable[iSlc][iNu].neutrinoInt.isNull())
        {
          if(sliceTruthTable[iSlc][iNu].efficiency > maxValue)
          {
            maxValue = sliceTruthTable[iSlc][iNu].efficiency;
            bestNu = iNu;
          }
        }
      }
      if(bestSlices[bestNu] == (int)iSlc)
      {
        faveIds[iSlc] = bestNu;
      }
    }
    
    
    
    return faveIds;
    
  }   

  struct Counts
  {
    Counts()
    {
      nHits[0] = nHits[1] = 0;
      nOnlyHits[0] = nOnlyHits[1] = 0;
    }
    int nHits[2]; // Indexed by view
    int nOnlyHits[2]; // Hit by only one particle
  };

//......................................................................
  std::set<int> FuzzyKVertex::FindVisibleProngs3D(const art::PtrVector<rb::CellHit>& allhits,
                                                  std::set<int> ids,
                                                  std::vector<int> *cellsX,
                                                  std::vector<int> *cellsY) {

    std::map<int, Counts> counts;
    art::ServiceHandle<cheat::BackTracker> bt;

    // loop over all hits and check if this track id contributed to them
    for(unsigned int i = 0; i < allhits.size(); ++i){

      //get vector of track id's that contribute to this hit
      const art::Ptr<rb::CellHit> hit = allhits[i];
      const std::vector<cheat::TrackIDE> trackIDs = bt->HitToTrackIDE(hit);

      // loop over all the returned track ids
      int nTracks = 0; // number of tracks that contributed energy to the hit
      int loneID = -1; // Which ID it was if there was only one

      for(unsigned int j = 0; j < trackIDs.size(); ++j){

        const int trackID = trackIDs[j].trackID;
        const float eFrac = trackIDs[j].energyFrac;
        //detector energy resolution is around 10%
        //to be a visible track we are looking for at least 1 hit in each view 
        //where a true particle deposited 20% (2 sigma) of the energy in that cell
        //this is a very minimum definition but leaves the most room for optimization
        if(eFrac >= 0.2){
          ++nTracks;
          loneID = trackID;

	  // check if the current id matches the input track id
          if(ids.find(trackID) != ids.end()){
            ++counts[trackID].nHits[hit->View()];
          }
        }
      } // end for j

      if(nTracks == 1){
        ++counts[loneID].nOnlyHits[hit->View()];
      }

    } // end for i

    const int kMinHits = 1;    // minimum number of hits to be considered visible
    const int kMinOnlyHit = 0; // minimum number of hits in which it is only particle contributing energy to be visible
    std::set<int> ret;

    for(std::set<int>::iterator it = ids.begin(); it != ids.end(); ++it){

      const Counts c = counts[*it];
      if((c.nOnlyHits[geo::kX] >= kMinOnlyHit) && (c.nOnlyHits[geo::kY] >= kMinOnlyHit)){
        if((c.nHits[geo::kX] >= kMinHits) && (c.nHits[geo::kY] >= kMinHits)){
          //std::cout<<"For track: "<<(*it)<<" xcells: "<<c.nHits[geo::kX]<<" ycells "<<c.nHits[geo::kY]<<std::endl;
          cellsX->push_back(c.nHits[geo::kX]);
          cellsY->push_back(c.nHits[geo::kY]);
          ret.insert(*it);
        }
      }
      
    }
    return ret;
  }  // end FindVisibleProngs

}
////////////////////////////////////////////////////////////////////////

namespace fuzz
{
  DEFINE_ART_MODULE(FuzzyKVertex);
}
////////////////////////////////////////////////////////////////////////
